% Prolog lexicon for SHARDS, from OALD machine-readable dictionary
% Produced by asc2lex, Matthew Purver 19/04/2001
%
% Manually edited for irregulars, CMTs, determiners etc.
%
% As this material is obtained from the OALD, it is freely available
% for RESEARCH PURPOSES ONLY. See the OTA's TEI header in
% ascii_0710-2.txt for more details.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Morphological interface predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% noun/6
% noun( ?Stem, ?Word, ?Type, ?Number, ?Gender, ?Case )
%
% Morphological interface to noun/4 - relates word stem
% Stem to surface word Word.
% Number will be 'sing' or 'plur'
% Gender currently undefined
% Case currently always 'case' (i.e. any)
% Type will be 'mass' or 'count' (both may succeed)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

noun( Word, Word, Type, sing, _Gender, case ) :-
      noun( Word, _Plural, RawType, _SemClass ),
      noun_type( RawType, Type ).

noun( Stem, Word, Type, plur, _Gender, case ) :-
      noun( Stem, Word, RawType, _SemClass ),
      noun_type( RawType, Type ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% noun_type( +RawType, ?Type )
%
% Returns type as defined in lexicon, except for 'both'
% which gets converted to 'mass' or 'count'
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

noun_type( Type, Type ) :-
	\+ Type = both.
noun_type( both, mass ).
noun_type( both, count ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% verb/5
% verb( ?Stem, ?Word, ?VForm, ?Number, ?Type )
%
% Morphological interface to verb/7 - relates word stem
% Stem to surface word Word.
% VForm currently 'inf', 'pres' or 'past'
% Number currently 's3', 'nons3' or 'person' (i.e. undefined)
% Type will be 'intran', 'tran', 'ditran', etc.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

verb( Word, Word, inf, person, Type ) :-
	verb( Word, _, _, _, _, Cat, CatList ),
	verb_type( Cat, CatList, Type ).

verb( Word, Word, pres, nons3, Type ) :-
	verb( Word, _, _, _, _, Cat, CatList ),
	verb_type( Cat, CatList, Type ).

verb( Stem, Word, pres, s3, Type ) :-
	verb( Stem, Word, _, _, _, Cat, CatList ),
	verb_type( Cat, CatList, Type ).

verb( Stem, Word, past, person, Type ) :-
	verb( Stem, _, _, Word, _, Cat, CatList ),
	verb_type( Cat, CatList, Type ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% verb_type( +RawType, +CatList, ?Type )
%
% Type is a verb subcategory type determined from the
% RawType atom and CatList list of numbers defined in
% the lexicon
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

verb_type( intran, _, intran ).        % VP -> V
verb_type( tran, _, tran ).            % VP -> V NP
verb_type( _, CatList, ditran ) :-     % VP -> V NP NP
	memberchk( '12A', CatList );
	memberchk( '12B', CatList );
	memberchk( '12C', CatList ).
verb_type( _, CatList, stran ) :-      % VP -> V [question]
	memberchk( '10', CatList ).
verb_type( _, CatList, pp_to ) :-      % VP -> V PP
	memberchk( '3A', CatList ).
verb_type( _, CatList, pp_for ) :-     % VP -> V PP
	memberchk( '3A', CatList ).
verb_type( _, CatList, subjraise ) :-  % VP -> V [to+inf] (subject raising)
	memberchk( '4E', CatList ).
verb_type( _, CatList, subjcon ) :-    % VP -> V [to+inf] (subject control)
	memberchk( '7A', CatList ).
verb_type( _, CatList, aux ) :-        % VP -> V [inf] (auxiliary)
	memberchk( '5', CatList ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% adj/3
% adj( ?Stem, ?Word, ?Type )
%
% Morphological interface to adj/4 - relates word stem
% Stem to surface word Word.
% Type will be 'simple', 'comparative' or 'superlative'
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

adj( Word, Word, simple ) :-
	adj( Word, _, _, _ ).

adj( Stem, Word, comparative ) :-
	adj( Stem, Word, _, _ ).

adj( Stem, Word, superlative ) :-
	adj( Stem, _, Word, _ ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% General predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% word( ?Word )
%
% Succeeds if Word is a word defined in the lexicon
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
word( Word ) :-
	noun( _, Word, _, _, _, _ );
	pron( Word, _, _, _ );
	verb( _, Word, _, _, _ );
	adj( _, Word, _ );
	adv( Word, _ );
	prep( Word, _ );
	conj( Word, _ );
	det( Word, _, _ );
	misc( Word, _, _ ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% open_class_word( ?Word )
%
% Succeeds if Word is a noun/verb/adj/adv in the lexicon
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
open_class_word( Word ) :-
	noun( _, Word, _, _, _, _ );
	verb( _, Word, _, _, _ );
	adj( _, Word, _ );
	adv( Word, _ ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% unknown_word( +Word )
%
% Succeeds if Word is NOT defined in the lexicon
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
unknown_word( Word ) :-
	\+ word( Word ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Lexicon
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

verb( 'hoover', 'hoovers', 'hoovering', 'hoovered', 'hoovered', tran, ['6A'] ).
verb( 'latinize', 'latinizes', 'latinizing', 'latinized', 'latinized', tran, [] ).
verb( 'roneo', 'roneos', 'roneoing', 'roneoed', 'roneoed', tran, [] ).
verb( 'x-ray', 'x-rays', 'x-raying', 'x-rayed', 'x-rayed', tran, ['6A'] ).
verb( 'xerox', 'xeroxes', 'xeroxing', 'xeroxed', 'xeroxed', tran, [] ).
verb( 'abandon', 'abandons', 'abandoning', 'abandoned', 'abandoned', tran, ['6A','14'] ).
verb( 'abase', 'abases', 'abasing', 'abased', 'abased', tran, ['6B'] ).
verb( 'abash', 'abashes', 'abashing', 'abashed', 'abashed', tran, ['6A'] ).
verb( 'abate', 'abates', 'abating', 'abated', 'abated', _, ['2A','6A'] ).
verb( 'abbreviate', 'abbreviates', 'abbreviating', 'abbreviated', 'abbreviated', tran, ['6A','14'] ).
verb( 'abdicate', 'abdicates', 'abdicating', 'abdicated', 'abdicated', _, ['2A','6A'] ).
verb( 'abduct', 'abducts', 'abducting', 'abducted', 'abducted', tran, ['6A'] ).
verb( 'abet', 'abets', 'abetting', 'abetted', 'abetted', tran, ['6A','14'] ).
verb( 'abhor', 'abhors', 'abhorring', 'abhorred', 'abhorred', tran, ['6A'] ).
verb( 'abide', 'abides', 'abiding', 'abided', 'abided', _, ['2C','3A','6A'] ).
verb( 'abjure', 'abjures', 'abjuring', 'abjured', 'abjured', tran, ['6A'] ).
verb( 'abolish', 'abolishes', 'abolishing', 'abolished', 'abolished', tran, ['6A'] ).
verb( 'abominate', 'abominates', 'abominating', 'abominated', 'abominated', tran, ['6A','6C'] ).
verb( 'abort', 'aborts', 'aborting', 'aborted', 'aborted', _, ['2A','6A'] ).
verb( 'abound', 'abounds', 'abounding', 'abounded', 'abounded', intran, ['3A'] ).
verb( 'about-face', 'about-faces', 'about-facing', 'about-faced', 'about-faced', intran, [] ).
verb( 'abrade', 'abrades', 'abrading', 'abraded', 'abraded', tran, ['6A'] ).
verb( 'abridge', 'abridges', 'abridging', 'abridged', 'abridged', tran, ['6A'] ).
verb( 'abrogate', 'abrogates', 'abrogating', 'abrogated', 'abrogated', tran, ['6A'] ).
verb( 'abscond', 'absconds', 'absconding', 'absconded', 'absconded', intran, ['2A','3A'] ).
verb( 'absent', 'absents', 'absenting', 'absented', 'absented', tran, ['6B','14'] ).
verb( 'absolve', 'absolves', 'absolving', 'absolved', 'absolved', tran, ['6A','14'] ).
verb( 'absorb', 'absorbs', 'absorbing', 'absorbed', 'absorbed', tran, ['6A'] ).
verb( 'abstain', 'abstains', 'abstaining', 'abstained', 'abstained', intran, ['2A','3A'] ).
verb( 'abstract', 'abstracts', 'abstracting', 'abstracted', 'abstracted', tran, ['6A','14'] ).
verb( 'abuse', 'abuses', 'abusing', 'abused', 'abused', tran, ['6A'] ).
verb( 'abut', 'abuts', 'abutting', 'abutted', 'abutted', intran, ['3A'] ).
verb( 'accede', 'accedes', 'acceding', 'acceded', 'acceded', intran, ['2A','3A'] ).
verb( 'accelerate', 'accelerates', 'accelerating', 'accelerated', 'accelerated', _, ['2A','6A'] ).
verb( 'accent', 'accents', 'accenting', 'accented', 'accented', tran, ['6A'] ).
verb( 'accentuate', 'accentuates', 'accentuating', 'accentuated', 'accentuated', tran, ['6A'] ).
verb( 'accept', 'accepts', 'accepting', 'accepted', 'accepted', _, ['2A','6A','9','16B'] ).
verb( 'acclaim', 'acclaims', 'acclaiming', 'acclaimed', 'acclaimed', tran, ['6A','16B','23'] ).
verb( 'acclimate', 'acclimates', 'acclimating', 'acclimated', 'acclimated', _, [] ).
verb( 'acclimatize', 'acclimatizes', 'acclimatizing', 'acclimatized', 'acclimatized', _, ['2A','14'] ).
verb( 'accommodate', 'accommodates', 'accommodating', 'accommodated', 'accommodated', tran, ['6A','14'] ).
verb( 'accompany', 'accompanies', 'accompanying', 'accompanied', 'accompanied', tran, ['6A','14'] ).
verb( 'accomplish', 'accomplishes', 'accomplishing', 'accomplished', 'accomplished', tran, ['6A'] ).
verb( 'accord', 'accords', 'according', 'accorded', 'accorded', _, ['2A','2C','3A','12A','13A'] ).
verb( 'accost', 'accosts', 'accosting', 'accosted', 'accosted', tran, ['6A'] ).
verb( 'account', 'accounts', 'accounting', 'accounted', 'accounted', _, ['3A','25'] ).
verb( 'accredit', 'accredits', 'accrediting', 'accredited', 'accredited', tran, ['14'] ).
verb( 'accrue', 'accrues', 'accruing', 'accrued', 'accrued', intran, ['2A','3A'] ).
verb( 'accumulate', 'accumulates', 'accumulating', 'accumulated', 'accumulated', _, ['2A','6A'] ).
verb( 'accuse', 'accuses', 'accusing', 'accused', 'accused', tran, ['6A','14'] ).
verb( 'accustom', 'accustoms', 'accustoming', 'accustomed', 'accustomed', tran, ['14'] ).
verb( 'ache', 'aches', 'aching', 'ached', 'ached', intran, ['2A','3A','4A'] ).
verb( 'achieve', 'achieves', 'achieving', 'achieved', 'achieved', tran, ['6A'] ).
verb( 'acidify', 'acidifies', 'acidifying', 'acidified', 'acidified', _, ['2A','6A'] ).
verb( 'acknowledge', 'acknowledges', 'acknowledging', 'acknowledged', 'acknowledged', tran, ['6A','6C','9','16B','24A','25'] ).
verb( 'acquaint', 'acquaints', 'acquainting', 'acquainted', 'acquainted', tran, ['14'] ).
verb( 'acquiesce', 'acquiesces', 'acquiescing', 'acquiesced', 'acquiesced', intran, ['2A','3A'] ).
verb( 'acquire', 'acquires', 'acquiring', 'acquired', 'acquired', tran, ['6A'] ).
verb( 'acquit', 'acquits', 'acquitting', 'acquitted', 'acquitted', tran, ['6A','14','16B'] ).
verb( 'act', 'acts', 'acting', 'acted', 'acted', _, ['2A','2C','3A','6A','15B'] ).
verb( 'activate', 'activates', 'activating', 'activated', 'activated', tran, ['6A'] ).
verb( 'actuate', 'actuates', 'actuating', 'actuated', 'actuated', tran, ['6A'] ).
verb( 'ad-lib', 'ad-libs', 'ad-libbing', 'ad-libbed', 'ad-libbed', intran, ['2A'] ).
verb( 'adapt', 'adapts', 'adapting', 'adapted', 'adapted', tran, ['6A','14'] ).
verb( 'add', 'adds', 'adding', 'added', 'added', _, ['2C','3A','6A','9','14','15B'] ).
verb( 'addict', 'addicts', 'addicting', 'addicted', 'addicted', tran, [] ).
verb( 'addle', 'addles', 'addling', 'addled', 'addled', _, ['2A','6A'] ).
verb( 'address', 'addresses', 'addressing', 'addressed', 'addressed', tran, ['6A','14','16B'] ).
verb( 'adduce', 'adduces', 'adducing', 'adduced', 'adduced', tran, ['6A'] ).
verb( 'adhere', 'adheres', 'adhering', 'adhered', 'adhered', intran, ['2A','3A'] ).
verb( 'adjoin', 'adjoins', 'adjoining', 'adjoined', 'adjoined', _, ['2A','6A'] ).
verb( 'adjourn', 'adjourns', 'adjourning', 'adjourned', 'adjourned', _, ['2A','2C','6A'] ).
verb( 'adjudge', 'adjudges', 'adjudging', 'adjudged', 'adjudged', tran, ['9','14','25'] ).
verb( 'adjudicate', 'adjudicates', 'adjudicating', 'adjudicated', 'adjudicated', _, ['2A','3A','6A','14','25'] ).
verb( 'adjure', 'adjures', 'adjuring', 'adjured', 'adjured', tran, ['17'] ).
verb( 'adjust', 'adjusts', 'adjusting', 'adjusted', 'adjusted', tran, ['6A','14'] ).
verb( 'administer', 'administers', 'administering', 'administered', 'administered', _, ['6A','14'] ).
verb( 'admire', 'admires', 'admiring', 'admired', 'admired', tran, ['6A'] ).
verb( 'admit', 'admits', 'admitting', 'admitted', 'admitted', _, ['3A','6A','6C','9','14','25'] ).
verb( 'admix', 'admixes', 'admixing', 'admixed', 'admixed', _, ['2A','6A'] ).
verb( 'admonish', 'admonishes', 'admonishing', 'admonished', 'admonished', tran, ['6A','14'] ).
verb( 'adopt', 'adopts', 'adopting', 'adopted', 'adopted', tran, ['6A'] ).
verb( 'adore', 'adores', 'adoring', 'adored', 'adored', tran, ['6A','6C'] ).
verb( 'adorn', 'adorns', 'adorning', 'adorned', 'adorned', tran, ['6A','14'] ).
verb( 'adulterate', 'adulterates', 'adulterating', 'adulterated', 'adulterated', tran, ['6A','14'] ).
verb( 'adumbrate', 'adumbrates', 'adumbrating', 'adumbrated', 'adumbrated', tran, ['6A'] ).
verb( 'advance', 'advances', 'advancing', 'advanced', 'advanced', _, ['2A','2B','3A','6A','12A','13A','14'] ).
verb( 'advantage', 'advantages', 'advantaging', 'advantaged', 'advantaged', tran, ['6A'] ).
verb( 'adventure', 'adventures', 'adventuring', 'adventured', 'adventured', tran, [] ).
verb( 'advert', 'adverts', 'adverting', 'adverted', 'adverted', intran, ['3A'] ).
verb( 'advertise', 'advertises', 'advertising', 'advertised', 'advertised', _, ['2A','3A','6A'] ).
verb( 'advise', 'advises', 'advising', 'advised', 'advised', _, ['3A','6A','6C','14','17','20','21'] ).
verb( 'advocate', 'advocates', 'advocating', 'advocated', 'advocated', tran, ['6A','6C'] ).
verb( 'aerate', 'aerates', 'aerating', 'aerated', 'aerated', tran, ['6A'] ).
verb( 'affect', 'affects', 'affecting', 'affected', 'affected', tran, ['6A','7A'] ).
verb( 'affiance', 'affiances', 'affiancing', 'affianced', 'affianced', tran, ['6A'] ).
verb( 'affiliate', 'affiliates', 'affiliating', 'affiliated', 'affiliated', _, ['2A','6A','14'] ).
verb( 'affirm', 'affirms', 'affirming', 'affirmed', 'affirmed', _, ['2A','6A','9','14'] ).
verb( 'affix', 'affixes', 'affixing', 'affixed', 'affixed', tran, ['6A','14'] ).
verb( 'afflict', 'afflicts', 'afflicting', 'afflicted', 'afflicted', tran, ['6A','14'] ).
verb( 'afford', 'affords', 'affording', 'afforded', 'afforded', tran, ['6A','7A','12A','13A'] ).
verb( 'afforest', 'afforests', 'afforesting', 'afforested', 'afforested', tran, ['6A'] ).
verb( 'affranchise', 'affranchises', 'affranchising', 'affranchised', 'affranchised', tran, ['6A'] ).
verb( 'affront', 'affronts', 'affronting', 'affronted', 'affronted', tran, ['6A'] ).
verb( 'age', 'ages', 'aging', 'aged', 'aged', _, ['2A','6A'] ).
verb( 'agglomerate', 'agglomerates', 'agglomerating', 'agglomerated', 'agglomerated', _, ['2A','6A'] ).
verb( 'agglutinate', 'agglutinates', 'agglutinating', 'agglutinated', 'agglutinated', tran, ['2A','6A'] ).
verb( 'aggrandize', 'aggrandizes', 'aggrandizing', 'aggrandized', 'aggrandized', tran, ['6A'] ).
verb( 'aggravate', 'aggravates', 'aggravating', 'aggravated', 'aggravated', tran, ['6A'] ).
verb( 'aggregate', 'aggregates', 'aggregating', 'aggregated', 'aggregated', _, ['2A','2E','6A'] ).
verb( 'aggrieve', 'aggrieves', 'aggrieving', 'aggrieved', 'aggrieved', tran, [] ).
verb( 'agitate', 'agitates', 'agitating', 'agitated', 'agitated', _, ['3A','6A'] ).
verb( 'agree', 'agrees', 'agreeing', 'agreed', 'agreed', _, ['2A','2C','3A','3B','4C','6A','7A'] ).
verb( 'aid', 'aids', 'aiding', 'aided', 'aided', tran, ['6A','14','17'] ).
verb( 'ail', 'ails', 'ailing', 'ailed', 'ailed', _, ['2A','2B','6A'] ).
verb( 'aim', 'aims', 'aiming', 'aimed', 'aimed', _, ['2A','3A','4A','6A','14'] ).
verb( 'air', 'airs', 'airing', 'aired', 'aired', tran, ['6A'] ).
verb( 'alarm', 'alarms', 'alarming', 'alarmed', 'alarmed', tran, ['6A'] ).
verb( 'alert', 'alerts', 'alerting', 'alerted', 'alerted', tran, ['6A'] ).
verb( 'alienate', 'alienates', 'alienating', 'alienated', 'alienated', tran, ['6A','14'] ).
verb( 'alight', 'alights', 'alighting', 'alighted', 'alighted', intran, ['2A','3A'] ).
verb( 'align', 'aligns', 'aligning', 'aligned', 'aligned', _, ['2A','3A','6A','14'] ).
verb( 'allay', 'allays', 'allaying', 'allayed', 'allayed', tran, ['6A'] ).
verb( 'allege', 'alleges', 'alleging', 'alleged', 'alleged', tran, ['6A','9'] ).
verb( 'alleviate', 'alleviates', 'alleviating', 'alleviated', 'alleviated', tran, ['6A'] ).
verb( 'allocate', 'allocates', 'allocating', 'allocated', 'allocated', tran, ['6A','14'] ).
verb( 'allot', 'allots', 'allotting', 'allotted', 'allotted', tran, ['6A','12A','13A','14'] ).
verb( 'allow', 'allows', 'allowing', 'allowed', 'allowed', _, ['3A','6A','6C','9','12A','13A','14','15B','17','25'] ).
verb( 'alloy', 'alloys', 'alloying', 'alloyed', 'alloyed', tran, ['6A'] ).
verb( 'allude', 'alludes', 'alluding', 'alluded', 'alluded', intran, ['3A'] ).
verb( 'allure', 'allures', 'alluring', 'allured', 'allured', tran, ['6A','14','17'] ).
verb( 'ally', 'allies', 'allying', 'allied', 'allied', tran, ['14'] ).
verb( 'alter', 'alters', 'altering', 'altered', 'altered', _, ['2A','6A'] ).
verb( 'alternate', 'alternates', 'alternating', 'alternated', 'alternated', _, ['3A','6A','14'] ).
verb( 'amalgamate', 'amalgamates', 'amalgamating', 'amalgamated', 'amalgamated', _, ['2A','6A'] ).
verb( 'amass', 'amasses', 'amassing', 'amassed', 'amassed', tran, ['6A'] ).
verb( 'amaze', 'amazes', 'amazing', 'amazed', 'amazed', tran, ['6A'] ).
verb( 'amble', 'ambles', 'ambling', 'ambled', 'ambled', intran, ['2A','2C'] ).
verb( 'ambuscade', 'ambuscades', 'ambuscading', 'ambuscaded', 'ambuscaded', tran, [] ).
verb( 'ambush', 'ambushes', 'ambushing', 'ambushed', 'ambushed', tran, ['6A'] ).
verb( 'ameliorate', 'ameliorates', 'ameliorating', 'ameliorated', 'ameliorated', _, ['2A','6A'] ).
verb( 'amend', 'amends', 'amending', 'amended', 'amended', _, ['2A','6A'] ).
verb( 'amortize', 'amortizes', 'amortizing', 'amortized', 'amortized', tran, ['6A'] ).
verb( 'amount', 'amounts', 'amounting', 'amounted', 'amounted', intran, ['3A'] ).
verb( 'amplify', 'amplifies', 'amplifying', 'amplified', 'amplified', tran, ['6A'] ).
verb( 'amputate', 'amputates', 'amputating', 'amputated', 'amputated', tran, ['6A'] ).
verb( 'amuse', 'amuses', 'amusing', 'amused', 'amused', tran, ['6A'] ).
verb( 'anaesthetize', 'anaesthetizes', 'anaesthetizing', 'anaesthetized', 'anaesthetized', tran, ['6A'] ).
verb( 'analyse', 'analyses', 'analysing', 'analysed', 'analysed', tran, ['6A'] ).
verb( 'analyze', 'analyzes', 'analyzing', 'analyzed', 'analyzed', tran, ['6A'] ).
verb( 'anathematize', 'anathematizes', 'anathematizing', 'anathematized', 'anathematized', _, [] ).
verb( 'anchor', 'anchors', 'anchoring', 'anchored', 'anchored', _, ['2A','6A'] ).
verb( 'anesthetize', 'anesthetizes', 'anesthetizing', 'anesthetized', 'anesthetized', tran, ['6A'] ).
verb( 'anger', 'angers', 'angering', 'angered', 'angered', tran, ['6A'] ).
verb( 'angle', 'angles', 'angling', 'angled', 'angled', _, ['2A','3A','6A'] ).
verb( 'angle-park', 'angle-parks', 'angle-parking', 'angle-parked', 'angle-parked', _, [] ).
verb( 'anglicize', 'anglicizes', 'anglicizing', 'anglicized', 'anglicized', tran, ['6A'] ).
verb( 'animadvert', 'animadverts', 'animadverting', 'animadverted', 'animadverted', intran, ['3A'] ).
verb( 'animate', 'animates', 'animating', 'animated', 'animated', tran, ['6A','14'] ).
verb( 'anneal', 'anneals', 'annealing', 'annealed', 'annealed', tran, ['6A'] ).
verb( 'annex', 'annexes', 'annexing', 'annexed', 'annexed', tran, ['6A','14'] ).
verb( 'annihilate', 'annihilates', 'annihilating', 'annihilated', 'annihilated', tran, ['6A'] ).
verb( 'annotate', 'annotates', 'annotating', 'annotated', 'annotated', tran, ['6A'] ).
verb( 'announce', 'announces', 'announcing', 'announced', 'announced', tran, ['6A','9','14'] ).
verb( 'annoy', 'annoys', 'annoying', 'annoyed', 'annoyed', tran, ['6A'] ).
verb( 'annul', 'annuls', 'annulling', 'annulled', 'annulled', tran, ['6A'] ).
verb( 'annunciate', 'annunciates', 'annunciating', 'annunciated', 'annunciated', tran, ['6A'] ).
verb( 'anoint', 'anoints', 'anointing', 'anointed', 'anointed', tran, ['6A','14','23'] ).
verb( 'answer', 'answers', 'answering', 'answered', 'answered', _, ['2A','2C','3A','6A','9','12A','15B'] ).
verb( 'antagonize', 'antagonizes', 'antagonizing', 'antagonized', 'antagonized', tran, ['6A'] ).
verb( 'antedate', 'antedates', 'antedating', 'antedated', 'antedated', tran, ['6A'] ).
verb( 'anticipate', 'anticipates', 'anticipating', 'anticipated', 'anticipated', tran, ['6A','6C','9'] ).
verb( 'ape', 'apes', 'aping', 'aped', 'aped', tran, [] ).
verb( 'apologize', 'apologizes', 'apologizing', 'apologized', 'apologized', intran, ['2A','3A'] ).
verb( 'apostrophize', 'apostrophizes', 'apostrophizing', 'apostrophized', 'apostrophized', tran, [] ).
verb( 'appal', 'appals', 'appalling', 'appalled', 'appalled', tran, ['6A'] ).
verb( 'apparel', 'apparels', 'apparelling', 'apparelled', 'apparelled', tran, [] ).
verb( 'appeal', 'appeals', 'appealing', 'appealed', 'appealed', intran, ['2A','3A'] ).
verb( 'appear', 'appears', 'appearing', 'appeared', 'appeared', intran, ['2A','2C','4D','4E'] ).
verb( 'appease', 'appeases', 'appeasing', 'appeased', 'appeased', tran, ['6A'] ).
verb( 'append', 'appends', 'appending', 'appended', 'appended', tran, ['6A','14'] ).
verb( 'appertain', 'appertains', 'appertaining', 'appertained', 'appertained', intran, ['3A'] ).
verb( 'applaud', 'applauds', 'applauding', 'applauded', 'applauded', _, ['2A','2B','6A'] ).
verb( 'appliqu_e', 'appliqu_es', 'appliqu_eing', 'appliqu_eed', 'appliqu_eed', tran, [] ).
verb( 'apply', 'applies', 'applying', 'applied', 'applied', _, ['2C','3A','6A','14'] ).
verb( 'appoint', 'appoints', 'appointing', 'appointed', 'appointed', tran, ['6A','9','14','16A','23','25'] ).
verb( 'apportion', 'apportions', 'apportioning', 'apportioned', 'apportioned', tran, ['6A','12B','13B','14'] ).
verb( 'appraise', 'appraises', 'appraising', 'appraised', 'appraised', tran, ['6A'] ).
verb( 'appreciate', 'appreciates', 'appreciating', 'appreciated', 'appreciated', _, ['2B','6A','9','10'] ).
verb( 'apprehend', 'apprehends', 'apprehending', 'apprehended', 'apprehended', tran, ['6A','9'] ).
verb( 'apprentice', 'apprentices', 'apprenticing', 'apprenticed', 'apprenticed', tran, ['6A','14'] ).
verb( 'apprise', 'apprises', 'apprising', 'apprised', 'apprised', tran, ['14'] ).
verb( 'approach', 'approaches', 'approaching', 'approached', 'approached', _, ['2A','6A'] ).
verb( 'appropriate', 'appropriates', 'appropriating', 'appropriated', 'appropriated', tran, ['6A','14'] ).
verb( 'approve', 'approves', 'approving', 'approved', 'approved', _, ['3A','6A'] ).
verb( 'approximate', 'approximates', 'approximating', 'approximated', 'approximated', _, ['3A','6A'] ).
verb( 'aquaplane', 'aquaplanes', 'aquaplaning', 'aquaplaned', 'aquaplaned', intran, [] ).
verb( 'arbitrate', 'arbitrates', 'arbitrating', 'arbitrated', 'arbitrated', _, ['2A','6A'] ).
verb( 'arch', 'arches', 'arching', 'arched', 'arched', _, ['2C','6A'] ).
verb( 'argue', 'argues', 'arguing', 'argued', 'argued', _, ['2A','2C','3A','6A','9','14'] ).
verb( 'arise', 'arises', 'arising', 'arose', 'arisen', intran, ['2A','3A'] ).
verb( 'arm', 'arms', 'arming', 'armed', 'armed', _, ['2A','6A','14'] ).
verb( 'arouse', 'arouses', 'arousing', 'aroused', 'aroused', tran, ['6A','14'] ).
verb( 'arraign', 'arraigns', 'arraigning', 'arraigned', 'arraigned', tran, ['6A','14'] ).
verb( 'arrange', 'arranges', 'arranging', 'arranged', 'arranged', _, ['3A','4C','6A','14','15A'] ).
verb( 'array', 'arrays', 'arraying', 'arrayed', 'arrayed', tran, ['6A','15A'] ).
verb( 'arrest', 'arrests', 'arresting', 'arrested', 'arrested', tran, ['6A'] ).
verb( 'arrive', 'arrives', 'arriving', 'arrived', 'arrived', intran, ['2A','2C','3A'] ).
verb( 'arrogate', 'arrogates', 'arrogating', 'arrogated', 'arrogated', tran, ['14'] ).
verb( 'article', 'articles', 'articling', 'articled', 'articled', tran, [] ).
verb( 'articulate', 'articulates', 'articulating', 'articulated', 'articulated', _, ['2C','6A','15A'] ).
verb( 'ascend', 'ascends', 'ascending', 'ascended', 'ascended', _, ['2A','2C','6A'] ).
verb( 'ascertain', 'ascertains', 'ascertaining', 'ascertained', 'ascertained', tran, ['6A','8','9','10','17'] ).
verb( 'ascribe', 'ascribes', 'ascribing', 'ascribed', 'ascribed', tran, ['14'] ).
verb( 'ask', 'asks', 'asking', 'asked', 'asked', _, ['3A','6A','7A','8','9','10','12C','14','15B','17','20','21'] ).
verb( 'asperse', 'asperses', 'aspersing', 'aspersed', 'aspersed', tran, ['6A'] ).
verb( 'asphalt', 'asphalts', 'asphalting', 'asphalted', 'asphalted', tran, ['6A'] ).
verb( 'asphyxiate', 'asphyxiates', 'asphyxiating', 'asphyxiated', 'asphyxiated', tran, ['6A'] ).
verb( 'aspirate', 'aspirates', 'aspirating', 'aspirated', 'aspirated', tran, [] ).
verb( 'aspire', 'aspires', 'aspiring', 'aspired', 'aspired', intran, ['3A','4A'] ).
verb( 'assail', 'assails', 'assailing', 'assailed', 'assailed', tran, ['6A','14'] ).
verb( 'assassinate', 'assassinates', 'assassinating', 'assassinated', 'assassinated', tran, ['6A'] ).
verb( 'assault', 'assaults', 'assaulting', 'assaulted', 'assaulted', tran, ['6A'] ).
verb( 'assay', 'assays', 'assaying', 'assayed', 'assayed', tran, ['6A','7A'] ).
verb( 'assemble', 'assembles', 'assembling', 'assembled', 'assembled', _, ['2A','6A'] ).
verb( 'assent', 'assents', 'assenting', 'assented', 'assented', intran, ['2A','3A'] ).
verb( 'assert', 'asserts', 'asserting', 'asserted', 'asserted', tran, ['6A','9','25'] ).
verb( 'assess', 'assesses', 'assessing', 'assessed', 'assessed', tran, ['6A','14'] ).
verb( 'asseverate', 'asseverates', 'asseverating', 'asseverated', 'asseverated', tran, ['6A','9'] ).
verb( 'assign', 'assigns', 'assigning', 'assigned', 'assigned', tran, ['12A','13A','13B','14','17'] ).
verb( 'assimilate', 'assimilates', 'assimilating', 'assimilated', 'assimilated', _, ['2A','3A','6A'] ).
verb( 'assist', 'assists', 'assisting', 'assisted', 'assisted', _, ['2A','3A','6A','14','17'] ).
verb( 'associate', 'associates', 'associating', 'associated', 'associated', _, ['3A','14'] ).
verb( 'assuage', 'assuages', 'assuaging', 'assuaged', 'assuaged', tran, ['6A'] ).
verb( 'assume', 'assumes', 'assuming', 'assumed', 'assumed', tran, ['6A','9','25'] ).
verb( 'assure', 'assures', 'assuring', 'assured', 'assured', tran, ['6A','11','14'] ).
verb( 'astonish', 'astonishes', 'astonishing', 'astonished', 'astonished', tran, ['6A'] ).
verb( 'astound', 'astounds', 'astounding', 'astounded', 'astounded', tran, ['6A'] ).
verb( 'atomize', 'atomizes', 'atomizing', 'atomized', 'atomized', tran, [] ).
verb( 'atone', 'atones', 'atoning', 'atoned', 'atoned', intran, ['2A','3A'] ).
verb( 'atrophy', 'atrophies', 'atrophying', 'atrophied', 'atrophied', _, ['2A','2B','6A'] ).
verb( 'attach', 'attaches', 'attaching', 'attached', 'attached', _, ['3A','6A','14'] ).
verb( 'attack', 'attacks', 'attacking', 'attacked', 'attacked', tran, ['6A'] ).
verb( 'attain', 'attains', 'attaining', 'attained', 'attained', _, ['3A','6A'] ).
verb( 'attempt', 'attempts', 'attempting', 'attempted', 'attempted', tran, ['6A','7A'] ).
verb( 'attend', 'attends', 'attending', 'attended', 'attended', _, ['2A','3A','6A'] ).
verb( 'attenuate', 'attenuates', 'attenuating', 'attenuated', 'attenuated', tran, ['6A'] ).
verb( 'attest', 'attests', 'attesting', 'attested', 'attested', _, ['2A','3A','6A'] ).
verb( 'attire', 'attires', 'attiring', 'attired', 'attired', tran, ['6A'] ).
verb( 'attitudinize', 'attitudinizes', 'attitudinizing', 'attitudinized', 'attitudinized', intran, ['2A'] ).
verb( 'attract', 'attracts', 'attracting', 'attracted', 'attracted', tran, ['6A'] ).
verb( 'attribute', 'attributes', 'attributing', 'attributed', 'attributed', tran, ['14'] ).
verb( 'attune', 'attunes', 'attuning', 'attuned', 'attuned', tran, ['14'] ).
verb( 'auction', 'auctions', 'auctioning', 'auctioned', 'auctioned', tran, ['6A','15B'] ).
verb( 'audit', 'audits', 'auditing', 'audited', 'audited', tran, ['6A'] ).
verb( 'audition', 'auditions', 'auditioning', 'auditioned', 'auditioned', tran, ['6A'] ).
verb( 'augment', 'augments', 'augmenting', 'augmented', 'augmented', _, ['2A','6A'] ).
verb( 'augur', 'augurs', 'auguring', 'augured', 'augured', _, ['2A','6A'] ).
verb( 'authenticate', 'authenticates', 'authenticating', 'authenticated', 'authenticated', tran, ['6A'] ).
verb( 'authorize', 'authorizes', 'authorizing', 'authorized', 'authorized', tran, ['6A','17'] ).
verb( 'autograph', 'autographs', 'autographing', 'autographed', 'autographed', tran, ['6A'] ).
verb( 'automate', 'automates', 'automating', 'automated', 'automated', tran, [] ).
verb( 'avail', 'avails', 'availing', 'availed', 'availed', _, ['2A','3A','14'] ).
verb( 'avenge', 'avenges', 'avenging', 'avenged', 'avenged', tran, ['6A','14'] ).
verb( 'aver', 'avers', 'averring', 'averred', 'averred', tran, ['6A','9'] ).
verb( 'average', 'averages', 'averaging', 'averaged', 'averaged', _, ['2B','6A'] ).
verb( 'avert', 'averts', 'averting', 'averted', 'averted', tran, ['6A','14'] ).
verb( 'avoid', 'avoids', 'avoiding', 'avoided', 'avoided', tran, ['6A','6C'] ).
verb( 'avouch', 'avouches', 'avouching', 'avouched', 'avouched', _, ['3A','6A','9'] ).
verb( 'avow', 'avows', 'avowing', 'avowed', 'avowed', tran, ['6A','25'] ).
verb( 'await', 'awaits', 'awaiting', 'awaited', 'awaited', tran, ['6A'] ).
verb( 'awake', 'awakes', 'awaking', 'awoke', 'awoken', intran, ['2A','3A','4B'] ).
verb( 'awaken', 'awakens', 'awakening', 'awakened', 'awakened', tran, ['6A','14'] ).
verb( 'award', 'awards', 'awarding', 'awarded', 'awarded', tran, ['6A','12A','13A'] ).
verb( 'awe', 'awes', 'awing', 'awed', 'awed', tran, ['6A','14'] ).
verb( 'ax', 'axes', 'axing', 'axed', 'axed', tran, ['6A'] ).
verb( 'axe', 'axes', 'axing', 'axed', 'axed', tran, ['6A'] ).
verb( 'baa', 'baas', 'baaing', 'baaed', 'baaed', intran, [] ).
verb( 'babble', 'babbles', 'babbling', 'babbled', 'babbled', _, ['2A','2B','2C','6A','15B'] ).
verb( 'baby', 'babies', 'babying', 'babied', 'babied', tran, ['6A'] ).
verb( 'babysit', 'babysits', 'babysitting', 'babysat', 'babysat', intran, [] ).
verb( 'back', 'backs', 'backing', 'backed', 'backed', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'backbite', 'backbites', 'backbiting', 'backbit', 'backbitten', _, [] ).
verb( 'backdate', 'backdates', 'backdating', 'backdated', 'backdated', tran, [] ).
verb( 'backfire', 'backfires', 'backfiring', 'backfired', 'backfired', intran, [] ).
verb( 'backpedal', 'backpedals', 'backpedalling', 'backpedalled', 'backpedalled', intran, [] ).
verb( 'backslide', 'backslides', 'backsliding', 'backslid', 'backslid', intran, ['2A'] ).
verb( 'backspace', 'backspaces', 'backspacing', 'backspaced', 'backspaced', intran, [] ).
verb( 'badger', 'badgers', 'badgering', 'badgered', 'badgered', tran, ['6A','14','16A'] ).
verb( 'baffle', 'baffles', 'baffling', 'baffled', 'baffled', tran, ['6A'] ).
verb( 'bag', 'bags', 'bagging', 'bagged', 'bagged', _, ['2A','2C','6A','15B'] ).
verb( 'bail', 'bails', 'bailing', 'bailed', 'bailed', _, ['2A','2C','6A','15B'] ).
verb( 'bait', 'baits', 'baiting', 'baited', 'baited', _, ['2A','6A'] ).
verb( 'bake', 'bakes', 'baking', 'baked', 'baked', _, ['2A','2C','6A','22'] ).
verb( 'balance', 'balances', 'balancing', 'balanced', 'balanced', _, ['2A','6A','14','15A'] ).
verb( 'bale', 'bales', 'baling', 'baled', 'baled', tran, ['6A'] ).
verb( 'balk', 'balks', 'balking', 'balked', 'balked', _, ['2A','3A','6A','14'] ).
verb( 'ball', 'balls', 'balling', 'balled', 'balled', _, [] ).
verb( 'ballast', 'ballasts', 'ballasting', 'ballasted', 'ballasted', tran, ['6A'] ).
verb( 'balloon', 'balloons', 'ballooning', 'ballooned', 'ballooned', intran, ['2A','2C'] ).
verb( 'ballot', 'ballots', 'balloting', 'balloted', 'balloted', intran, ['2A','3A'] ).
verb( 'bamboozle', 'bamboozles', 'bamboozling', 'bamboozled', 'bamboozled', tran, ['6A','14'] ).
verb( 'ban', 'bans', 'banning', 'banned', 'banned', tran, ['6A','14'] ).
verb( 'band', 'bands', 'banding', 'banded', 'banded', _, ['2C','6A','14','15B'] ).
verb( 'bandage', 'bandages', 'bandaging', 'bandaged', 'bandaged', tran, ['6A','15B'] ).
verb( 'bandy', 'bandies', 'bandying', 'bandied', 'bandied', tran, ['6A','14','15B'] ).
verb( 'bang', 'bangs', 'banging', 'banged', 'banged', _, ['2A','2C','6A','15B'] ).
verb( 'banish', 'banishes', 'banishing', 'banished', 'banished', tran, ['6A','14'] ).
verb( 'bank', 'banks', 'banking', 'banked', 'banked', _, ['2A','2C','3A','6A','15B'] ).
verb( 'bankrupt', 'bankrupts', 'bankrupting', 'bankrupted', 'bankrupted', tran, ['6A'] ).
verb( 'banquet', 'banquets', 'banqueting', 'banqueted', 'banqueted', _, ['2A','6A'] ).
verb( 'bant', 'bants', 'banting', 'banted', 'banted', intran, ['2A'] ).
verb( 'banter', 'banters', 'bantering', 'bantered', 'bantered', _, ['2A','6A'] ).
verb( 'baptize', 'baptizes', 'baptizing', 'baptized', 'baptized', tran, ['6A','23'] ).
verb( 'bar', 'bars', 'barring', 'barred', 'barred', tran, ['6A','6C','12C','14','15B'] ).
verb( 'barbarize', 'barbarizes', 'barbarizing', 'barbarized', 'barbarized', tran, ['6A'] ).
verb( 'barbecue', 'barbecues', 'barbecuing', 'barbecued', 'barbecued', tran, [] ).
verb( 'bare', 'bares', 'baring', 'bared', 'bared', tran, ['6A'] ).
verb( 'bargain', 'bargains', 'bargaining', 'bargained', 'bargained', _, ['2A','3A','9','15B'] ).
verb( 'barge', 'barges', 'barging', 'barged', 'barged', intran, ['2C','3A'] ).
verb( 'bark', 'barks', 'barking', 'barked', 'barked', _, ['2A','2C','3A','6A','15B'] ).
verb( 'barnstorm', 'barnstorms', 'barnstorming', 'barnstormed', 'barnstormed', intran, [] ).
verb( 'barrack', 'barracks', 'barracking', 'barracked', 'barracked', _, ['2A','6A'] ).
verb( 'barrel', 'barrels', 'barrelling', 'barrelled', 'barrelled', tran, [] ).
verb( 'barricade', 'barricades', 'barricading', 'barricaded', 'barricaded', tran, ['6A','15B'] ).
verb( 'barter', 'barters', 'bartering', 'bartered', 'bartered', _, ['2A','14','15B'] ).
verb( 'base', 'bases', 'basing', 'based', 'based', tran, ['14'] ).
verb( 'bash', 'bashes', 'bashing', 'bashed', 'bashed', tran, ['6A','15A','15B'] ).
verb( 'bask', 'basks', 'basking', 'basked', 'basked', intran, ['2C'] ).
verb( 'bastardize', 'bastardizes', 'bastardizing', 'bastardized', 'bastardized', tran, [] ).
verb( 'baste', 'bastes', 'basting', 'basted', 'basted', tran, ['6A','15B'] ).
verb( 'bastinado', 'bastinados', 'bastinadoing', 'bastinadoed', 'bastinadoed', tran, ['6A'] ).
verb( 'bat', 'bats', 'batting', 'batted', 'batted', _, ['2A','2B','2C','6A'] ).
verb( 'bate', 'bates', 'bating', 'bated', 'bated', tran, [] ).
verb( 'bath', 'baths', 'bathing', 'bathed', 'bathed', _, ['2A','6A'] ).
verb( 'bathe', 'bathes', 'bathing', 'bathed', 'bathed', _, ['2A','6A'] ).
verb( 'batten', 'battens', 'battening', 'battened', 'battened', _, ['3A','6A','15B'] ).
verb( 'batter', 'batters', 'battering', 'battered', 'battered', _, ['2C','6A','15A','15B'] ).
verb( 'battle', 'battles', 'battling', 'battled', 'battled', intran, ['3A'] ).
verb( 'baulk', 'baulks', 'baulking', 'baulked', 'baulked', _, ['2A','3A','6A','14'] ).
verb( 'bawl', 'bawls', 'bawling', 'bawled', 'bawled', _, ['2C','3A','6A','15A'] ).
verb( 'bay', 'bays', 'baying', 'bayed', 'bayed', intran, ['2A'] ).
verb( 'bayonet', 'bayonets', 'bayoneting', 'bayoneted', 'bayoneted', tran, ['6A'] ).
verb( 'be', 'is', 'being', 'was', 'been', unknown, ['1','4F'] ).
verb( 'be', 'is', 'being', 'was', 'been', intran, ['1','4F'] ).
verb( 'beach', 'beaches', 'beaching', 'beached', 'beached', tran, ['6A'] ).
verb( 'beam', 'beams', 'beaming', 'beamed', 'beamed', _, ['2C','6A','15A'] ).
verb( 'bear', 'bears', 'bearing', 'bore', 'borne', _, ['2A','2C','3A','6A','6D','6E','7A','11','12C','14','15B','16B','17'] ).
verb( 'beard', 'beards', 'bearding', 'bearded', 'bearded', tran, ['6A'] ).
verb( 'beat', 'beats', 'beating', 'beat', 'beaten', _, ['2A','2C','6A','14','15A','15B','22'] ).
verb( 'beatify', 'beatifies', 'beatifying', 'beatified', 'beatified', tran, ['6A'] ).
verb( 'beautify', 'beautifies', 'beautifying', 'beautified', 'beautified', tran, ['6A'] ).
verb( 'beaver', 'beavers', 'beavering', 'beavered', 'beavered', intran, ['2A','2C'] ).
verb( 'beckon', 'beckons', 'beckoning', 'beckoned', 'beckoned', _, ['2A','6A','15B','16A'] ).
verb( 'become', 'becomes', 'becoming', 'became', 'become', _, ['2D','3A','6A'] ).
verb( 'bed', 'beds', 'bedding', 'bedded', 'bedded', tran, ['6A','15A','15B'] ).
verb( 'bedevil', 'bedevils', 'bedevilling', 'bedevilled', 'bedevilled', tran, [] ).
verb( 'beef', 'beefs', 'beefing', 'beefed', 'beefed', intran, [] ).
verb( 'beeswax', 'beeswaxes', 'beeswaxing', 'beeswaxed', 'beeswaxed', tran, [] ).
verb( 'beetle', 'beetles', 'beetling', 'beetled', 'beetled', intran, ['2A'] ).
verb( 'befall', 'befalls', 'befalling', 'befell', 'befallen', _, ['6A'] ).
verb( 'befit', 'befits', 'befitting', 'befitted', 'befitted', tran, ['6A'] ).
verb( 'befoul', 'befouls', 'befouling', 'befouled', 'befouled', tran, [] ).
verb( 'befriend', 'befriends', 'befriending', 'befriended', 'befriended', tran, ['6A'] ).
verb( 'beg', 'begs', 'begging', 'begged', 'begged', _, ['2A','2C','3A','6A','7A','9','14','17'] ).
verb( 'beget', 'begets', 'begetting', 'begat', 'begotten', tran, ['6A'] ).
verb( 'beggar', 'beggars', 'beggaring', 'beggared', 'beggared', tran, ['6A'] ).
verb( 'begin', 'begins', 'beginning', 'began', 'begun', _, ['2A','3A','6A','6D','7A'] ).
verb( 'begrudge', 'begrudges', 'begrudging', 'begrudged', 'begrudged', tran, ['6C','12A','13A'] ).
verb( 'beguile', 'beguiles', 'beguiling', 'beguiled', 'beguiled', tran, ['6A','14'] ).
verb( 'behave', 'behaves', 'behaving', 'behaved', 'behaved', intran, ['2A','2C','6B'] ).
verb( 'behead', 'beheads', 'beheading', 'beheaded', 'beheaded', tran, ['6A'] ).
verb( 'behold', 'beholds', 'beholding', 'beheld', 'beheld', tran, ['6A'] ).
verb( 'behove', 'behoves', 'behoving', 'behoved', 'behoved', tran, [] ).
verb( 'belabour', 'belabours', 'belabouring', 'belaboured', 'belaboured', tran, ['6A'] ).
verb( 'belay', 'belays', 'belaying', 'belayed', 'belayed', tran, ['6A'] ).
verb( 'belch', 'belches', 'belching', 'belched', 'belched', _, ['2A','6A','15B'] ).
verb( 'beleaguer', 'beleaguers', 'beleaguering', 'beleaguered', 'beleaguered', tran, ['6A'] ).
verb( 'belie', 'belies', 'belying', 'belied', 'belied', tran, ['6A'] ).
verb( 'believe', 'believes', 'believing', 'believed', 'believed', _, ['3A','6A','9','10','25'] ).
verb( 'belittle', 'belittles', 'belittling', 'belittled', 'belittled', tran, ['6A'] ).
verb( 'bell', 'bells', 'belling', 'belled', 'belled', tran, [] ).
verb( 'bellow', 'bellows', 'bellowing', 'bellowed', 'bellowed', _, ['2A','6A','15B'] ).
verb( 'belly', 'bellies', 'bellying', 'bellied', 'bellied', _, ['2A','2C','6A','15B'] ).
verb( 'bellyache', 'bellyaches', 'bellyaching', 'bellyached', 'bellyached', intran, [] ).
verb( 'bellyland', 'bellylands', 'bellylanding', 'bellylanded', 'bellylanded', intran, [] ).
verb( 'bellylaugh', 'bellylaughs', 'bellylaughing', 'bellylaughed', 'bellylaughed', intran, [] ).
verb( 'belong', 'belongs', 'belonging', 'belonged', 'belonged', intran, ['2C','3A'] ).
verb( 'belt', 'belts', 'belting', 'belted', 'belted', tran, ['2C','3A','6A','15B'] ).
verb( 'bemoan', 'bemoans', 'bemoaning', 'bemoaned', 'bemoaned', tran, ['6A'] ).
verb( 'bend', 'bends', 'bending', 'bended', 'bended', _, ['2A','2C','6A','15A','15B'] ).
verb( 'benefit', 'benefits', 'benefiting', 'benefited', 'benefited', _, ['3A','6A'] ).
verb( 'bequeath', 'bequeaths', 'bequeathing', 'bequeathed', 'bequeathed', tran, ['6A','12A','13A'] ).
verb( 'berate', 'berates', 'berating', 'berated', 'berated', tran, ['6A'] ).
verb( 'bereave', 'bereaves', 'bereaving', 'bereaved', 'bereaved', tran, ['14'] ).
verb( 'berth', 'berths', 'berthing', 'berthed', 'berthed', _, ['2C','6A','15A'] ).
verb( 'beseech', 'beseeches', 'beseeching', 'beseeched', 'beseeched', tran, ['6A','11','13B','17'] ).
verb( 'beseem', 'beseems', 'beseeming', 'beseemed', 'beseemed', tran, [] ).
verb( 'beset', 'besets', 'besetting', 'beset', 'beset', tran, ['6A'] ).
verb( 'beshrew', 'beshrews', 'beshrewing', 'beshrewed', 'beshrewed', tran, [] ).
verb( 'besiege', 'besieges', 'besieging', 'besieged', 'besieged', tran, ['6A','14'] ).
verb( 'besmear', 'besmears', 'besmearing', 'besmeared', 'besmeared', tran, [] ).
verb( 'besmirch', 'besmirches', 'besmirching', 'besmirched', 'besmirched', tran, [] ).
verb( 'bespeak', 'bespeaks', 'bespeaking', 'bespoke', 'bespoken', tran, ['6A','25'] ).
verb( 'best', 'bests', 'besting', 'bested', 'bested', tran, ['6A'] ).
verb( 'bestir', 'bestirs', 'bestirring', 'bestirred', 'bestirred', tran, ['6A','17'] ).
verb( 'bestow', 'bestows', 'bestowing', 'bestowed', 'bestowed', tran, ['6A','14'] ).
verb( 'bestrew', 'bestrews', 'bestrewing', 'bestrewed', 'bestrewed', tran, ['6A','14'] ).
verb( 'bestride', 'bestrides', 'bestriding', 'bestrode', 'bestridden', tran, ['6A'] ).
verb( 'bet', 'bets', 'betting', 'betted', 'betted', _, ['2A','3A','9','11','12C'] ).
verb( 'betake', 'betakes', 'betaking', 'betook', 'betaken', tran, ['14'] ).
verb( 'bethink', 'bethinks', 'bethinking', 'bethought', 'bethought', tran, ['11','14','17','20','21'] ).
verb( 'betide', 'betides', 'betiding', 'betided', 'betided', tran, [] ).
verb( 'betoken', 'betokens', 'betokening', 'betokened', 'betokened', tran, ['6A'] ).
verb( 'betray', 'betrays', 'betraying', 'betrayed', 'betrayed', tran, ['6A','14','25'] ).
verb( 'betroth', 'betroths', 'betrothing', 'betrothed', 'betrothed', tran, ['6A','14'] ).
verb( 'better', 'betters', 'bettering', 'bettered', 'bettered', tran, ['6A'] ).
verb( 'bevel', 'bevels', 'bevelling', 'bevelled', 'bevelled', tran, [] ).
verb( 'bewail', 'bewails', 'bewailing', 'bewailed', 'bewailed', tran, ['6A'] ).
verb( 'beware', '-', '-', '-', '-', _, ['2A','3A','10'] ).
verb( 'bewilder', 'bewilders', 'bewildering', 'bewildered', 'bewildered', tran, ['6A'] ).
verb( 'bewitch', 'bewitches', 'bewitching', 'bewitched', 'bewitched', tran, ['6A'] ).
verb( 'bias', 'biases', 'biasing', 'biased', 'biased', tran, ['6A','14'] ).
verb( 'bib', 'bibs', 'bibbing', 'bibbed', 'bibbed', intran, [] ).
verb( 'bicker', 'bickers', 'bickering', 'bickered', 'bickered', intran, ['2A','2C','3A'] ).
verb( 'bicycle', 'bicycles', 'bicycling', 'bicycled', 'bicycled', intran, ['2A','2C'] ).
verb( 'bid', 'bids', 'bidding', 'bid', 'bid', _, ['2A','3A','6A','12A','13A','14','15B','17','18B'] ).
verb( 'bide', 'bides', 'biding', 'bided', 'bided', tran, [] ).
verb( 'biff', 'biffs', 'biffing', 'biffed', 'biffed', tran, [] ).
verb( 'bifurcate', 'bifurcates', 'bifurcating', 'bifurcated', 'bifurcated', _, ['2A','6A'] ).
verb( 'bike', 'bikes', 'biking', 'biked', 'biked', intran, [] ).
verb( 'bilk', 'bilks', 'bilking', 'bilked', 'bilked', tran, ['6A','14'] ).
verb( 'bill', 'bills', 'billing', 'billed', 'billed', _, ['6A','14'] ).
verb( 'billet', 'billets', 'billeting', 'billeted', 'billeted', tran, ['6A','14'] ).
verb( 'billow', 'billows', 'billowing', 'billowed', 'billowed', intran, ['2C'] ).
verb( 'bind', 'binds', 'binding', 'bound', 'bound', _, ['2A','6A','14','15A','15B','16B','17'] ).
verb( 'birch', 'birches', 'birching', 'birched', 'birched', tran, ['6A'] ).
verb( 'bisect', 'bisects', 'bisecting', 'bisected', 'bisected', tran, ['6A'] ).
verb( 'bitch', 'bitches', 'bitching', 'bitched', 'bitched', intran, ['2A'] ).
verb( 'bite', 'bites', 'biting', 'bit', 'bitten', _, ['2A','3A','6A','15B'] ).
verb( 'bivouac', 'bivouacs', 'bivouacking', 'bivouacked', 'bivouacked', intran, ['2A'] ).
verb( 'blab', 'blabs', 'blabbing', 'blabbed', 'blabbed', _, ['2A','6A','15B'] ).
verb( 'blabber', 'blabbers', 'blabbering', 'blabbered', 'blabbered', _, [] ).
verb( 'black', 'blacks', 'blacking', 'blacked', 'blacked', tran, ['6A'] ).
verb( 'black-lead', 'black-leads', 'black-leading', 'black-leaded', 'black-leaded', tran, [] ).
verb( 'blackball', 'blackballs', 'blackballing', 'blackballed', 'blackballed', tran, [] ).
verb( 'blacken', 'blackens', 'blackening', 'blackened', 'blackened', _, ['2A','6A'] ).
verb( 'blackguard', 'blackguards', 'blackguarding', 'blackguarded', 'blackguarded', tran, [] ).
verb( 'blackleg', 'blacklegs', 'blacklegging', 'blacklegged', 'blacklegged', _, [] ).
verb( 'blacklist', 'blacklists', 'blacklisting', 'blacklisted', 'blacklisted', tran, [] ).
verb( 'blackmail', 'blackmails', 'blackmailing', 'blackmailed', 'blackmailed', tran, [] ).
verb( 'blame', 'blames', 'blaming', 'blamed', 'blamed', tran, ['6A','14'] ).
verb( 'blanch', 'blanches', 'blanching', 'blanched', 'blanched', _, ['2A','6A'] ).
verb( 'blanket', 'blankets', 'blanketing', 'blanketed', 'blanketed', tran, ['6A','14'] ).
verb( 'blare', 'blares', 'blaring', 'blared', 'blared', _, ['2A','2C','15B'] ).
verb( 'blaspheme', 'blasphemes', 'blaspheming', 'blasphemed', 'blasphemed', _, ['2A','6A'] ).
verb( 'blast', 'blasts', 'blasting', 'blasted', 'blasted', tran, ['2A','2C','6A','15B'] ).
verb( 'blather', 'blathers', 'blathering', 'blathered', 'blathered', intran, [] ).
verb( 'blaze', 'blazes', 'blazing', 'blazed', 'blazed', _, ['2A','2C','6A','15B'] ).
verb( 'blazon', 'blazons', 'blazoning', 'blazoned', 'blazoned', tran, [] ).
verb( 'bleach', 'bleaches', 'bleaching', 'bleached', 'bleached', _, ['2A','6A'] ).
verb( 'bleat', 'bleats', 'bleating', 'bleated', 'bleated', _, ['2A','6A','15B'] ).
verb( 'bleed', 'bleeds', 'bleeding', 'bled', 'bled', _, ['2A','2C','3A','6A','14'] ).
verb( 'bleep', 'bleeps', 'bleeping', 'bleeped', 'bleeped', intran, [] ).
verb( 'blemish', 'blemishes', 'blemishing', 'blemished', 'blemished', tran, ['6A'] ).
verb( 'blench', 'blenches', 'blenching', 'blenched', 'blenched', intran, ['2A'] ).
verb( 'blend', 'blends', 'blending', 'blended', 'blended', _, ['2A','3A','6A'] ).
verb( 'bless', 'blesses', 'blessing', 'blessed', 'blessed', tran, ['6A'] ).
verb( 'blether', 'blethers', 'blethering', 'blethered', 'blethered', intran, ['2A','2C'] ).
verb( 'blight', 'blights', 'blighting', 'blighted', 'blighted', tran, ['6A'] ).
verb( 'blind', 'blinds', 'blinding', 'blinded', 'blinded', tran, ['6A','14'] ).
verb( 'blindfold', 'blindfolds', 'blindfolding', 'blindfolded', 'blindfolded', tran, ['6A'] ).
verb( 'blink', 'blinks', 'blinking', 'blinked', 'blinked', _, ['2A','2C','6A','15B'] ).
verb( 'blister', 'blisters', 'blistering', 'blistered', 'blistered', _, ['2A','6A'] ).
verb( 'blitz', 'blitzes', 'blitzing', 'blitzed', 'blitzed', tran, ['6A'] ).
verb( 'block', 'blocks', 'blocking', 'blocked', 'blocked', tran, ['6A','15B'] ).
verb( 'blockade', 'blockades', 'blockading', 'blockaded', 'blockaded', tran, ['6A'] ).
verb( 'blood', 'bloods', 'blooding', 'blooded', 'blooded', tran, ['6A'] ).
verb( 'bloom', 'blooms', 'blooming', 'bloomed', 'bloomed', intran, ['2A','2C'] ).
verb( 'blossom', 'blossoms', 'blossoming', 'blossomed', 'blossomed', intran, ['2A','2C'] ).
verb( 'blot', 'blots', 'blotting', 'blotted', 'blotted', tran, ['6A','15B'] ).
verb( 'blow', 'blows', 'blowing', 'blew', 'blown', _, ['2A','2B','2C','2E','3A','6A','12A','15A','15B'] ).
verb( 'blow-dry', 'blow-dries', 'blow-drying', 'blow-dried', 'blow-dried', tran, ['6A'] ).
verb( 'blubber', 'blubbers', 'blubbering', 'blubbered', 'blubbered', _, ['2A','15B'] ).
verb( 'bludgeon', 'bludgeons', 'bludgeoning', 'bludgeoned', 'bludgeoned', tran, ['6A','14'] ).
verb( 'blue', 'blues', 'bluing', 'blued', 'blued', tran, [] ).
verb( 'blue-pencil', 'blue-pencils', 'blue-pencilling', 'blue-pencilled', 'blue-pencilled', tran, [] ).
verb( 'bluff', 'bluffs', 'bluffing', 'bluffed', 'bluffed', _, ['2A','6A','14','15B'] ).
verb( 'blunder', 'blunders', 'blundering', 'blundered', 'blundered', _, ['2A','2C','3A'] ).
verb( 'blunt', 'blunts', 'blunting', 'blunted', 'blunted', tran, ['6A'] ).
verb( 'blur', 'blurs', 'blurring', 'blurred', 'blurred', _, ['2A','6A'] ).
verb( 'blurt', 'blurts', 'blurting', 'blurted', 'blurted', tran, ['15B'] ).
verb( 'blush', 'blushes', 'blushing', 'blushed', 'blushed', intran, ['2A','2C','3A','4B'] ).
verb( 'bluster', 'blusters', 'blustering', 'blustered', 'blustered', _, ['2A','2C','15B'] ).
verb( 'board', 'boards', 'boarding', 'boarded', 'boarded', _, ['3A','6A','15B'] ).
verb( 'boast', 'boasts', 'boasting', 'boasted', 'boasted', _, ['2A','3A','3B','6A'] ).
verb( 'boat', 'boats', 'boating', 'boated', 'boated', intran, ['2A','2C'] ).
verb( 'bob', 'bobs', 'bobbing', 'bobbed', 'bobbed', _, ['2C','6A'] ).
verb( 'bode', 'bodes', 'boding', 'boded', 'boded', _, ['12B','13B'] ).
verb( 'bog', 'bogs', 'bogging', 'bogged', 'bogged', _, ['2E','15B'] ).
verb( 'boggle', 'boggles', 'boggling', 'boggled', 'boggled', intran, ['2A','3A'] ).
verb( 'boil', 'boils', 'boiling', 'boiled', 'boiled', _, ['2A','2B','2C','2D','6A','15B','22'] ).
verb( 'bolster', 'bolsters', 'bolstering', 'bolstered', 'bolstered', tran, ['6A','15B'] ).
verb( 'bolt', 'bolts', 'bolting', 'bolted', 'bolted', _, ['2A','2C','6A','15B'] ).
verb( 'bomb', 'bombs', 'bombing', 'bombed', 'bombed', _, ['2C','6A','15B'] ).
verb( 'bombard', 'bombards', 'bombarding', 'bombarded', 'bombarded', tran, ['6A','14'] ).
verb( 'bond', 'bonds', 'bonding', 'bonded', 'bonded', tran, ['6A'] ).
verb( 'bone', 'bones', 'boning', 'boned', 'boned', tran, ['6A'] ).
verb( 'boo', 'boos', 'booing', 'booed', 'booed', _, ['2A','6A','15B'] ).
verb( 'boob', 'boobs', 'boobing', 'boobed', 'boobed', intran, [] ).
verb( 'book', 'books', 'booking', 'booked', 'booked', tran, ['6A'] ).
verb( 'boom', 'booms', 'booming', 'boomed', 'boomed', _, ['2A','2C','15B'] ).
verb( 'boost', 'boosts', 'boosting', 'boosted', 'boosted', tran, ['6A'] ).
verb( 'boot', 'boots', 'booting', 'booted', 'booted', tran, ['6A','15A','15B'] ).
verb( 'bootleg', 'bootlegs', 'bootlegging', 'bootlegged', 'bootlegged', tran, [] ).
verb( 'booze', 'boozes', 'boozing', 'boozed', 'boozed', intran, ['2A','2C'] ).
verb( 'border', 'borders', 'bordering', 'bordered', 'bordered', _, ['3A','6A'] ).
verb( 'bore', 'bores', 'boring', 'bored', 'bored', _, ['2A','2C','3A','6A','6D','6E','7A','11','12C','14','15A','15B','16B','17'] ).
verb( 'borrow', 'borrows', 'borrowing', 'borrowed', 'borrowed', tran, ['6A','14'] ).
verb( 'boss', 'bosses', 'bossing', 'bossed', 'bossed', tran, ['6A','15B'] ).
verb( 'botanize', 'botanizes', 'botanizing', 'botanized', 'botanized', intran, [] ).
verb( 'botch', 'botches', 'botching', 'botched', 'botched', tran, ['6A','15B'] ).
verb( 'bother', 'bothers', 'bothering', 'bothered', 'bothered', _, ['2A','3A','4C','6A','14','16A'] ).
verb( 'bottle', 'bottles', 'bottling', 'bottled', 'bottled', tran, ['6A','15B'] ).
verb( 'bottom', 'bottoms', 'bottoming', 'bottomed', 'bottomed', intran, [] ).
verb( 'bounce', 'bounces', 'bouncing', 'bounced', 'bounced', _, ['2A','2C','6A'] ).
verb( 'bound', 'bounds', 'bounding', 'bounded', 'bounded', _, ['2A','2C','4A','6A','14','15A','15B','16B','17'] ).
verb( 'bow', 'bows', 'bowing', 'bowed', 'bowed', tran, [] ).
verb( 'bow', 'bows', 'bowing', 'bowed', 'bowed', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'bowdlerize', 'bowdlerizes', 'bowdlerizing', 'bowdlerized', 'bowdlerized', tran, [] ).
verb( 'bowl', 'bowls', 'bowling', 'bowled', 'bowled', _, ['2A','2C','6A','15B'] ).
verb( 'box', 'boxes', 'boxing', 'boxed', 'boxed', _, ['2A','3A','6A','15B'] ).
verb( 'boycott', 'boycotts', 'boycotting', 'boycotted', 'boycotted', tran, ['6A'] ).
verb( 'brace', 'braces', 'bracing', 'braced', 'braced', _, ['6A','15B','16A'] ).
verb( 'bracket', 'brackets', 'bracketing', 'bracketed', 'bracketed', tran, ['6A','15B'] ).
verb( 'brag', 'brags', 'bragging', 'bragged', 'bragged', intran, ['2A','3A'] ).
verb( 'braid', 'braids', 'braiding', 'braided', 'braided', tran, ['6A'] ).
verb( 'brain', 'brains', 'braining', 'brained', 'brained', tran, ['6A'] ).
verb( 'brainwash', 'brainwashes', 'brainwashing', 'brainwashed', 'brainwashed', tran, ['6A'] ).
verb( 'braise', 'braises', 'braising', 'braised', 'braised', tran, ['6A'] ).
verb( 'brake', 'brakes', 'braking', 'braked', 'braked', _, ['2A','2C','2D','3A','6A','15A','15B','22'] ).
verb( 'branch', 'branches', 'branching', 'branched', 'branched', intran, ['2A','2C'] ).
verb( 'brand', 'brands', 'branding', 'branded', 'branded', tran, ['6A','16B'] ).
verb( 'brandish', 'brandishes', 'brandishing', 'brandished', 'brandished', tran, ['6A'] ).
verb( 'brave', 'braves', 'braving', 'braved', 'braved', tran, ['6A','15B'] ).
verb( 'brawl', 'brawls', 'brawling', 'brawled', 'brawled', intran, ['2A'] ).
verb( 'bray', 'brays', 'braying', 'brayed', 'brayed', tran, ['2A'] ).
verb( 'braze', 'brazes', 'brazing', 'brazed', 'brazed', tran, ['6A'] ).
verb( 'brazen', 'brazens', 'brazening', 'brazened', 'brazened', tran, [] ).
verb( 'breach', 'breaches', 'breaching', 'breached', 'breached', tran, ['6A'] ).
verb( 'break', 'breaks', 'breaking', 'broke', 'broken', _, ['2A','2C','2D','3A','6A','15A','15B','22'] ).
verb( 'breakfast', 'breakfasts', 'breakfasting', 'breakfasted', 'breakfasted', intran, [] ).
verb( 'breast', 'breasts', 'breasting', 'breasted', 'breasted', tran, ['6A'] ).
verb( 'breastfeed', 'breastfeeds', 'breastfeeding', 'breastfed', 'breastfed', _, [] ).
verb( 'breathe', 'breathes', 'breathing', 'breathed', 'breathed', _, ['2A','2C','6A','15B'] ).
verb( 'breed', 'breeds', 'breeding', 'bred', 'bred', _, ['2A','6A'] ).
verb( 'breeze', 'breezes', 'breezing', 'breezed', 'breezed', intran, ['2C'] ).
verb( 'brew', 'brews', 'brewing', 'brewed', 'brewed', _, ['2A','2C','6A'] ).
verb( 'bribe', 'bribes', 'bribing', 'bribed', 'bribed', tran, ['6A','15A','17'] ).
verb( 'brick', 'bricks', 'bricking', 'bricked', 'bricked', tran, ['15B'] ).
verb( 'bridge', 'bridges', 'bridging', 'bridged', 'bridged', tran, ['6A','15B'] ).
verb( 'bridle', 'bridles', 'bridling', 'bridled', 'bridled', _, ['2A','2C','6A'] ).
verb( 'brief', 'briefs', 'briefing', 'briefed', 'briefed', tran, ['6A'] ).
verb( 'brighten', 'brightens', 'brightening', 'brightened', 'brightened', _, ['2A','2C','6A','15B'] ).
verb( 'brim', 'brims', 'brimming', 'brimmed', 'brimmed', intran, ['2A','2C'] ).
verb( 'bring', 'brings', 'bringing', 'brought', 'brought', tran, ['6A','12A','12C','13A','14','15B','17','19B'] ).
verb( 'bristle', 'bristles', 'bristling', 'bristled', 'bristled', intran, ['2A','2C'] ).
verb( 'broach', 'broaches', 'broaching', 'broached', 'broached', _, ['2C','6A','15B'] ).
verb( 'broadcast', 'broadcasts', 'broadcasting', 'broadcasted', 'broadcasted', _, ['2A','6A'] ).
verb( 'broaden', 'broadens', 'broadening', 'broadened', 'broadened', _, ['2A','2C','6A','15B'] ).
verb( 'brocade', 'brocades', 'brocading', 'brocaded', 'brocaded', tran, ['6A'] ).
verb( 'broil', 'broils', 'broiling', 'broiled', 'broiled', _, ['2A','6A'] ).
verb( 'bronze', 'bronzes', 'bronzing', 'bronzed', 'bronzed', _, ['2A','6A'] ).
verb( 'brood', 'broods', 'brooding', 'brooded', 'brooded', intran, ['2A','2C','3A'] ).
verb( 'brook', 'brooks', 'brooking', 'brooked', 'brooked', tran, ['6A','6B'] ).
verb( 'browbeat', 'browbeats', 'browbeating', 'browbeat', 'browbeaten', tran, ['6A','14'] ).
verb( 'brown', 'browns', 'browning', 'browned', 'browned', _, ['2A','6A'] ).
verb( 'browse', 'browses', 'browsing', 'browsed', 'browsed', intran, ['2A','2C'] ).
verb( 'bruise', 'bruises', 'bruising', 'bruised', 'bruised', _, ['2A','6A'] ).
verb( 'bruit', 'bruits', 'bruiting', 'bruited', 'bruited', tran, ['15B'] ).
verb( 'brush', 'brushes', 'brushing', 'brushed', 'brushed', _, ['2A','2C','6A','15B','22'] ).
verb( 'brutalize', 'brutalizes', 'brutalizing', 'brutalized', 'brutalized', tran, ['6A'] ).
verb( 'bubble', 'bubbles', 'bubbling', 'bubbled', 'bubbled', intran, ['2A','2C'] ).
verb( 'buck', 'bucks', 'bucking', 'bucked', 'bucked', _, ['2A','2C','6A','15B'] ).
verb( 'bucket', 'buckets', 'bucketing', 'bucketed', 'bucketed', intran, [] ).
verb( 'buckle', 'buckles', 'buckling', 'buckled', 'buckled', _, ['2A','2C','6A','15B'] ).
verb( 'bud', 'buds', 'budding', 'budded', 'budded', intran, [] ).
verb( 'budge', 'budges', 'budging', 'budged', 'budged', _, ['2A','2C','6A'] ).
verb( 'budget', 'budgets', 'budgeting', 'budgeted', 'budgeted', intran, ['3A'] ).
verb( 'buff', 'buffs', 'buffing', 'buffed', 'buffed', tran, [] ).
verb( 'buffet', 'buffets', 'buffeting', 'buffeted', 'buffeted', _, ['3A','6A'] ).
verb( 'bug', 'bugs', 'bugging', 'bugged', 'bugged', tran, ['6A'] ).
verb( 'bugger', 'buggers', 'buggering', 'buggered', 'buggered', _, ['6A'] ).
verb( 'build', 'builds', 'building', 'built', 'built', _, ['2C','3A','6A','12B','13B','14','15B'] ).
verb( 'bulge', 'bulges', 'bulging', 'bulged', 'bulged', _, ['2A','2C','6A'] ).
verb( 'bulk', 'bulks', 'bulking', 'bulked', 'bulked', intran, [] ).
verb( 'bulldoze', 'bulldozes', 'bulldozing', 'bulldozed', 'bulldozed', tran, ['6A','14'] ).
verb( 'bullshit', 'bullshits', 'bullshitting', 'bullshitted', 'bullshitted', intran, [] ).
verb( 'bully', 'bullies', 'bullying', 'bullied', 'bullied', _, ['2C','6A','14'] ).
verb( 'bum', 'bums', 'bumming', 'bummed', 'bummed', _, ['2C','6A','14'] ).
verb( 'bump', 'bumps', 'bumping', 'bumped', 'bumped', _, ['2C','3A','6A','14','15B'] ).
verb( 'bunch', 'bunches', 'bunching', 'bunched', 'bunched', _, ['2A','2C','15B'] ).
verb( 'bundle', 'bundles', 'bundling', 'bundled', 'bundled', _, ['2C','15A','15B'] ).
verb( 'bung', 'bungs', 'bunging', 'bunged', 'bunged', tran, ['6A','15B'] ).
verb( 'bungle', 'bungles', 'bungling', 'bungled', 'bungled', _, ['2A','6A'] ).
verb( 'bunk', 'bunks', 'bunking', 'bunked', 'bunked', intran, [] ).
verb( 'bunker', 'bunkers', 'bunkering', 'bunkered', 'bunkered', _, ['2A','6A'] ).
verb( 'buoy', 'buoys', 'buoying', 'buoyed', 'buoyed', intran, ['6A','15B'] ).
verb( 'burble', 'burbles', 'burbling', 'burbled', 'burbled', intran, [] ).
verb( 'burden', 'burdens', 'burdening', 'burdened', 'burdened', tran, ['6A','14'] ).
verb( 'burgeon', 'burgeons', 'burgeoning', 'burgeoned', 'burgeoned', intran, [] ).
verb( 'burgle', 'burgles', 'burgling', 'burgled', 'burgled', _, ['2A','6A'] ).
verb( 'burke', 'burkes', 'burking', 'burked', 'burked', tran, ['6A'] ).
verb( 'burlesque', 'burlesques', 'burlesquing', 'burlesqued', 'burlesqued', tran, ['6A'] ).
verb( 'burn', 'burns', 'burning', 'burned', 'burned', _, ['2A','2B','2C','4A','6A','14','15B'] ).
verb( 'burnish', 'burnishes', 'burnishing', 'burnished', 'burnished', _, ['2A','6A'] ).
verb( 'burp', 'burps', 'burping', 'burped', 'burped', _, ['2A','6A'] ).
verb( 'burrow', 'burrows', 'burrowing', 'burrowed', 'burrowed', _, ['2A','2C','6A'] ).
verb( 'burst', 'bursts', 'bursting', 'burst', 'burst', _, ['2A','2C','3A','6A','22'] ).
verb( 'burthen', 'burthens', 'burthening', 'burthened', 'burthened', tran, [] ).
verb( 'bury', 'buries', 'burying', 'buried', 'buried', tran, ['6A','15A','22'] ).
verb( 'bus', 'buses', 'busing', 'bused', 'bused', _, ['2A','6A'] ).
verb( 'bust', 'busts', 'busting', 'busted', 'busted', _, [] ).
verb( 'bustle', 'bustles', 'bustling', 'bustled', 'bustled', _, ['2A','2C','15B'] ).
verb( 'busy', 'busies', 'busying', 'busied', 'busied', tran, ['6A','14'] ).
verb( 'butcher', 'butchers', 'butchering', 'butchered', 'butchered', tran, ['6A'] ).
verb( 'butt', 'butts', 'butting', 'butted', 'butted', _, ['2C','3A','6A','15A'] ).
verb( 'butter', 'butters', 'buttering', 'buttered', 'buttered', tran, ['6A','15B'] ).
verb( 'button', 'buttons', 'buttoning', 'buttoned', 'buttoned', _, ['2A','6A','15B'] ).
verb( 'buttonhole', 'buttonholes', 'buttonholing', 'buttonholed', 'buttonholed', tran, ['6A'] ).
verb( 'buttress', 'buttresses', 'buttressing', 'buttressed', 'buttressed', tran, ['6A','15B'] ).
verb( 'buy', 'buys', 'buying', 'bought', 'bought', _, ['2A','2C','6A','12B','13B','15B'] ).
verb( 'buzz', 'buzzes', 'buzzing', 'buzzed', 'buzzed', _, ['2A','2C','6A'] ).
verb( 'bypass', 'bypasses', 'bypassing', 'bypassed', 'bypassed', tran, ['6A'] ).
verb( 'cable', 'cables', 'cabling', 'cabled', 'cabled', _, ['2A','6A'] ).
verb( 'cache', 'caches', 'caching', 'cached', 'cached', tran, [] ).
verb( 'cackle', 'cackles', 'cackling', 'cackled', 'cackled', intran, [] ).
verb( 'cadge', 'cadges', 'cadging', 'cadged', 'cadged', _, ['2A','6A','14'] ).
verb( 'cage', 'cages', 'caging', 'caged', 'caged', tran, ['6A'] ).
verb( 'cajole', 'cajoles', 'cajoling', 'cajoled', 'cajoled', tran, ['6A','14'] ).
verb( 'cake', 'cakes', 'caking', 'caked', 'caked', _, ['2A','6A'] ).
verb( 'calcify', 'calcifies', 'calcifying', 'calcified', 'calcified', _, ['2A','6A'] ).
verb( 'calcine', 'calcines', 'calcining', 'calcined', 'calcined', _, ['2A','6A'] ).
verb( 'calculate', 'calculates', 'calculating', 'calculated', 'calculated', _, ['2A','3A','6A','8','9','10'] ).
verb( 'calender', 'calenders', 'calendering', 'calendered', 'calendered', tran, ['6A'] ).
verb( 'calibrate', 'calibrates', 'calibrating', 'calibrated', 'calibrated', tran, ['6A'] ).
verb( 'calk', 'calks', 'calking', 'calked', 'calked', tran, [] ).
verb( 'call', 'calls', 'calling', 'called', 'called', _, ['2A','2B','2C','3A','4A','6A','12B','13B','15B','22','23'] ).
verb( 'calm', 'calms', 'calming', 'calmed', 'calmed', _, ['2C','6A','15B'] ).
verb( 'calumniate', 'calumniates', 'calumniating', 'calumniated', 'calumniated', tran, ['6A'] ).
verb( 'calve', 'calves', 'calving', 'calved', 'calved', intran, [] ).
verb( 'camber', 'cambers', 'cambering', 'cambered', 'cambered', _, [] ).
verb( 'camouflage', 'camouflages', 'camouflaging', 'camouflaged', 'camouflaged', tran, ['6A'] ).
verb( 'camp', 'camps', 'camping', 'camped', 'camped', _, ['2A','2C'] ).
verb( 'campaign', 'campaigns', 'campaigning', 'campaigned', 'campaigned', intran, ['2A','3A'] ).
verb( 'can', 'can', '-', 'could', '-', unknown, ['5','6A'] ).
verb( 'can', 'cans', 'canning', 'canned', 'canned', tran, ['5','6A'] ).
verb( 'canalize', 'canalizes', 'canalizing', 'canalized', 'canalized', tran, ['6A','14'] ).
verb( 'cancel', 'cancels', 'cancelling', 'cancelled', 'cancelled', _, ['2C','6A','15B'] ).
verb( 'candy', 'candies', 'candying', 'candied', 'candied', _, ['2A','6A'] ).
verb( 'cane', 'canes', 'caning', 'caned', 'caned', tran, ['6A'] ).
verb( 'canker', 'cankers', 'cankering', 'cankered', 'cankered', tran, [] ).
verb( 'cannibalize', 'cannibalizes', 'cannibalizing', 'cannibalized', 'cannibalized', tran, [] ).
verb( 'canoe', 'canoes', 'canoeing', 'canoed', 'canoed', tran, ['2A','2C'] ).
verb( 'canonize', 'canonizes', 'canonizing', 'canonized', 'canonized', tran, ['6A'] ).
verb( 'cant', 'cants', 'canting', 'canted', 'canted', _, ['2A','2C','6A','15B'] ).
verb( 'canter', 'canters', 'cantering', 'cantered', 'cantered', _, [] ).
verb( 'canvass', 'canvasses', 'canvassing', 'canvassed', 'canvassed', _, ['2A','3A','6A'] ).
verb( 'cap', 'caps', 'capping', 'capped', 'capped', tran, ['6A'] ).
verb( 'caparison', 'caparisons', 'caparisoning', 'caparisoned', 'caparisoned', tran, [] ).
verb( 'caper', 'capers', 'capering', 'capered', 'capered', intran, [] ).
verb( 'capitalize', 'capitalizes', 'capitalizing', 'capitalized', 'capitalized', _, ['3A','6A'] ).
verb( 'capitulate', 'capitulates', 'capitulating', 'capitulated', 'capitulated', tran, ['2A'] ).
verb( 'capsize', 'capsizes', 'capsizing', 'capsized', 'capsized', _, ['2A','6A'] ).
verb( 'captain', 'captains', 'captaining', 'captained', 'captained', tran, ['6A'] ).
verb( 'captivate', 'captivates', 'captivating', 'captivated', 'captivated', tran, ['6A'] ).
verb( 'capture', 'captures', 'capturing', 'captured', 'captured', tran, ['6A'] ).
verb( 'carbonize', 'carbonizes', 'carbonizing', 'carbonized', 'carbonized', tran, ['6A'] ).
verb( 'card', 'cards', 'carding', 'carded', 'carded', tran, [] ).
verb( 'care', 'cares', 'caring', 'cared', 'cared', intran, ['2A','3A','3B','4C'] ).
verb( 'careen', 'careens', 'careening', 'careened', 'careened', _, ['2A','6A'] ).
verb( 'career', 'careers', 'careering', 'careered', 'careered', intran, ['2C','3A'] ).
verb( 'caress', 'caresses', 'caressing', 'caressed', 'caressed', tran, ['6A'] ).
verb( 'caricature', 'caricatures', 'caricaturing', 'caricatured', 'caricatured', tran, ['6A'] ).
verb( 'carol', 'carols', 'carolling', 'carolled', 'carolled', tran, [] ).
verb( 'carouse', 'carouses', 'carousing', 'caroused', 'caroused', tran, ['2A'] ).
verb( 'carp', 'carps', 'carping', 'carped', 'carped', tran, ['2A','3A'] ).
verb( 'carpet', 'carpets', 'carpeting', 'carpeted', 'carpeted', tran, ['6A'] ).
verb( 'carry', 'carries', 'carrying', 'carried', 'carried', _, ['2B','2C','6A','15A','15B','16B'] ).
verb( 'cart', 'carts', 'carting', 'carted', 'carted', tran, ['6A','15B'] ).
verb( 'cartoon', 'cartoons', 'cartooning', 'cartooned', 'cartooned', tran, [] ).
verb( 'carve', 'carves', 'carving', 'carved', 'carved', _, ['6A','14','15A','15B'] ).
verb( 'cascade', 'cascades', 'cascading', 'cascaded', 'cascaded', intran, [] ).
verb( 'case', 'cases', 'casing', 'cased', 'cased', tran, ['6A'] ).
verb( 'cash', 'cashes', 'cashing', 'cashed', 'cashed', _, ['2C','6A','12B','13B'] ).
verb( 'cashier', 'cashiers', 'cashiering', 'cashiered', 'cashiered', tran, ['6A'] ).
verb( 'cast', 'casts', 'casting', 'cast', 'cast', _, ['2C','6A','15A','15B'] ).
verb( 'castigate', 'castigates', 'castigating', 'castigated', 'castigated', tran, ['6A'] ).
verb( 'castle', 'castles', 'castling', 'castled', 'castled', intran, [] ).
verb( 'castrate', 'castrates', 'castrating', 'castrated', 'castrated', tran, ['6A'] ).
verb( 'catalogue', 'catalogues', 'cataloguing', 'catalogued', 'catalogued', tran, ['6A'] ).
verb( 'catapult', 'catapults', 'catapulting', 'catapulted', 'catapulted', tran, [] ).
verb( 'catcall', 'catcalls', 'catcalling', 'catcalled', 'catcalled', intran, [] ).
verb( 'catch', 'catches', 'catching', 'caught', 'caught', _, ['2C','3A','6A','12C','14','15A','15B','19B','22'] ).
verb( 'catechize', 'catechizes', 'catechizing', 'catechized', 'catechized', tran, ['6A'] ).
verb( 'categorize', 'categorizes', 'categorizing', 'categorized', 'categorized', tran, ['6A'] ).
verb( 'cater', 'caters', 'catering', 'catered', 'catered', intran, ['3A'] ).
verb( 'caterwaul', 'caterwauls', 'caterwauling', 'caterwauled', 'caterwauled', intran, [] ).
verb( 'caulk', 'caulks', 'caulking', 'caulked', 'caulked', tran, ['6A'] ).
verb( 'cause', 'causes', 'causing', 'caused', 'caused', tran, ['6A','12A','13A','17'] ).
verb( 'cauterize', 'cauterizes', 'cauterizing', 'cauterized', 'cauterized', tran, ['6A'] ).
verb( 'caution', 'cautions', 'cautioning', 'cautioned', 'cautioned', tran, ['6A','14','17'] ).
verb( 'cave', 'caves', 'caving', 'caved', 'caved', _, ['2C','15B'] ).
verb( 'cavil', 'cavils', 'cavilling', 'cavilled', 'cavilled', intran, ['2A','3A'] ).
verb( 'cavort', 'cavorts', 'cavorting', 'cavorted', 'cavorted', intran, [] ).
verb( 'caw', 'caws', 'cawing', 'cawed', 'cawed', _, ['2A','15B'] ).
verb( 'cease', 'ceases', 'ceasing', 'ceased', 'ceased', _, ['2A','3A','6A','6D','7A'] ).
verb( 'cede', 'cedes', 'ceding', 'ceded', 'ceded', tran, ['6A','14'] ).
verb( 'celebrate', 'celebrates', 'celebrating', 'celebrated', 'celebrated', tran, ['6A'] ).
verb( 'cement', 'cements', 'cementing', 'cemented', 'cemented', tran, ['6A','15B'] ).
verb( 'censor', 'censors', 'censoring', 'censored', 'censored', tran, ['6A'] ).
verb( 'censure', 'censures', 'censuring', 'censured', 'censured', tran, ['6A','14'] ).
verb( 'centralize', 'centralizes', 'centralizing', 'centralized', 'centralized', _, ['2A','6A'] ).
verb( 'centre', 'centres', 'centring', 'centred', 'centred', _, ['3A','6A','14','15A'] ).
verb( 'certificate', 'certificates', 'certificating', 'certificated', 'certificated', tran, [] ).
verb( 'certify', 'certifies', 'certifying', 'certified', 'certified', _, ['3A','6A','9','16B','25'] ).
verb( 'chafe', 'chafes', 'chafing', 'chafed', 'chafed', _, ['2A','3A','6A'] ).
verb( 'chaff', 'chaffs', 'chaffing', 'chaffed', 'chaffed', tran, ['2A','2C','6A','15A'] ).
verb( 'chagrin', 'chagrins', 'chagrining', 'chagrined', 'chagrined', tran, ['6A'] ).
verb( 'chain', 'chains', 'chaining', 'chained', 'chained', tran, ['6A','15A','15B'] ).
verb( 'chair', 'chairs', 'chairing', 'chaired', 'chaired', tran, ['6A'] ).
verb( 'chalk', 'chalks', 'chalking', 'chalked', 'chalked', tran, ['15B'] ).
verb( 'challenge', 'challenges', 'challenging', 'challenged', 'challenged', tran, ['6A','14','17'] ).
verb( 'champ', 'champs', 'champing', 'champed', 'champed', _, ['2A','2C','4A','6A'] ).
verb( 'champion', 'champions', 'championing', 'championed', 'championed', tran, ['6A'] ).
verb( 'chance', 'chances', 'chancing', 'chanced', 'chanced', _, ['2A','3A','4E','6A','6C'] ).
verb( 'change', 'changes', 'changing', 'changed', 'changed', _, ['2A','2C','6A','14'] ).
verb( 'channel', 'channels', 'channelling', 'channelled', 'channelled', tran, ['6A','14'] ).
verb( 'chant', 'chants', 'chanting', 'chanted', 'chanted', _, ['2A','6A'] ).
verb( 'chap', 'chaps', 'chapping', 'chapped', 'chapped', _, ['2A','6A'] ).
verb( 'chaperon', 'chaperons', 'chaperoning', 'chaperoned', 'chaperoned', tran, ['6A'] ).
verb( 'char', 'chars', 'charring', 'charred', 'charred', _, ['2A','6A'] ).
verb( 'characterize', 'characterizes', 'characterizing', 'characterized', 'characterized', tran, ['6A'] ).
verb( 'charge', 'charges', 'charging', 'charged', 'charged', _, ['2A','2B','2C','6A','14','15A','15B','17'] ).
verb( 'charm', 'charms', 'charming', 'charmed', 'charmed', _, ['2A','6A','15A'] ).
verb( 'chart', 'charts', 'charting', 'charted', 'charted', tran, ['6A'] ).
verb( 'charter', 'charters', 'chartering', 'chartered', 'chartered', tran, ['6A'] ).
verb( 'chase', 'chases', 'chasing', 'chased', 'chased', _, ['2C','3A','6A','15A','15B'] ).
verb( 'chasten', 'chastens', 'chastening', 'chastened', 'chastened', tran, ['6A'] ).
verb( 'chastise', 'chastises', 'chastising', 'chastised', 'chastised', tran, ['6A'] ).
verb( 'chat', 'chats', 'chatting', 'chatted', 'chatted', _, ['2A','2C','15B'] ).
verb( 'chatter', 'chatters', 'chattering', 'chattered', 'chattered', intran, ['2A','2C'] ).
verb( 'chaw', 'chaws', 'chawing', 'chawed', 'chawed', tran, [] ).
verb( 'cheapen', 'cheapens', 'cheapening', 'cheapened', 'cheapened', _, ['2A','6A'] ).
verb( 'cheat', 'cheats', 'cheating', 'cheated', 'cheated', _, ['2A','2C','6A','14'] ).
verb( 'check', 'checks', 'checking', 'checked', 'checked', _, ['2C','6A','15B'] ).
verb( 'checker', 'checkers', 'checkering', 'checkered', 'checkered', tran, [] ).
verb( 'checkmate', 'checkmates', 'checkmating', 'checkmated', 'checkmated', tran, ['6A'] ).
verb( 'cheek', 'cheeks', 'cheeking', 'cheeked', 'cheeked', tran, ['6A'] ).
verb( 'cheep', 'cheeps', 'cheeping', 'cheeped', 'cheeped', intran, [] ).
verb( 'cheer', 'cheers', 'cheering', 'cheered', 'cheered', _, ['2A','2C','6A','15B'] ).
verb( 'chequer', 'chequers', 'chequering', 'chequered', 'chequered', tran, [] ).
verb( 'cherish', 'cherishes', 'cherishing', 'cherished', 'cherished', tran, ['6A'] ).
verb( 'chew', 'chews', 'chewing', 'chewed', 'chewed', _, ['2A','2C','3A','6A','15B'] ).
verb( 'chide', 'chides', 'chiding', 'chided', 'chided', _, ['6A','14'] ).
verb( 'chill', 'chills', 'chilling', 'chilled', 'chilled', _, ['2A','2C','6A'] ).
verb( 'chime', 'chimes', 'chiming', 'chimed', 'chimed', _, ['2A','2C','6A','15A'] ).
verb( 'chink', 'chinks', 'chinking', 'chinked', 'chinked', _, ['2A','2C','6A','15B'] ).
verb( 'chip', 'chips', 'chipping', 'chipped', 'chipped', _, ['2A','2C','6A','15A','15B'] ).
verb( 'chirp', 'chirps', 'chirping', 'chirped', 'chirped', _, [] ).
verb( 'chirrup', 'chirrups', 'chirruping', 'chirruped', 'chirruped', _, [] ).
verb( 'chisel', 'chisels', 'chiselling', 'chiselled', 'chiselled', tran, ['6A','14','15B'] ).
verb( 'chivvy', 'chivvies', 'chivvying', 'chivvied', 'chivvied', tran, ['15B'] ).
verb( 'chivy', 'chivies', 'chivying', 'chivied', 'chivied', tran, ['15B'] ).
verb( 'chlorinate', 'chlorinates', 'chlorinating', 'chlorinated', 'chlorinated', tran, [] ).
verb( 'chock', 'chocks', 'chocking', 'chocked', 'chocked', tran, ['6A','15B'] ).
verb( 'choke', 'chokes', 'choking', 'choked', 'choked', _, ['2A','3A','6A','14','15B'] ).
verb( 'choose', 'chooses', 'choosing', 'chose', 'chosen', _, ['2A','2C','3A','6A','7A','16A','16B','23'] ).
verb( 'chop', 'chops', 'chopping', 'chopped', 'chopped', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'chortle', 'chortles', 'chortling', 'chortled', 'chortled', intran, [] ).
verb( 'chorus', 'choruses', 'chorusing', 'chorused', 'chorused', tran, [] ).
verb( 'christen', 'christens', 'christening', 'christened', 'christened', tran, ['6A','23'] ).
verb( 'chronicle', 'chronicles', 'chronicling', 'chronicled', 'chronicled', tran, ['6A'] ).
verb( 'chuck', 'chucks', 'chucking', 'chucked', 'chucked', tran, ['6A','12A','13A','15A','15B'] ).
verb( 'chuckle', 'chuckles', 'chuckling', 'chuckled', 'chuckled', intran, ['2A','2C'] ).
verb( 'chug', 'chugs', 'chugging', 'chugged', 'chugged', intran, ['2C'] ).
verb( 'chum', 'chums', 'chumming', 'chummed', 'chummed', intran, ['2C'] ).
verb( 'churn', 'churns', 'churning', 'churned', 'churned', _, ['2A','2C','6A','15B'] ).
verb( 'cipher', 'ciphers', 'ciphering', 'ciphered', 'ciphered', _, ['2A','6A'] ).
verb( 'circle', 'circles', 'circling', 'circled', 'circled', _, ['2A','2C','6A'] ).
verb( 'circularize', 'circularizes', 'circularizing', 'circularized', 'circularized', tran, ['6A'] ).
verb( 'circulate', 'circulates', 'circulating', 'circulated', 'circulated', _, ['2A','2C','6A'] ).
verb( 'circumcise', 'circumcises', 'circumcising', 'circumcised', 'circumcised', tran, [] ).
verb( 'circumnavigate', 'circumnavigates', 'circumnavigating', 'circumnavigated', 'circumnavigated', tran, ['6A'] ).
verb( 'circumscribe', 'circumscribes', 'circumscribing', 'circumscribed', 'circumscribed', tran, ['6A'] ).
verb( 'circumvent', 'circumvents', 'circumventing', 'circumvented', 'circumvented', tran, ['6A'] ).
verb( 'cite', 'cites', 'citing', 'cited', 'cited', tran, ['6A'] ).
verb( 'civilize', 'civilizes', 'civilizing', 'civilized', 'civilized', tran, ['6A'] ).
verb( 'clack', 'clacks', 'clacking', 'clacked', 'clacked', intran, [] ).
verb( 'claim', 'claims', 'claiming', 'claimed', 'claimed', _, ['2A','4A','7A','9'] ).
verb( 'clam', 'clams', 'clamming', 'clammed', 'clammed', intran, ['2A','2C'] ).
verb( 'clamber', 'clambers', 'clambering', 'clambered', 'clambered', intran, ['2C'] ).
verb( 'clamour', 'clamours', 'clamouring', 'clamoured', 'clamoured', _, ['2A','2C','4A'] ).
verb( 'clamp', 'clamps', 'clamping', 'clamped', 'clamped', _, ['2C','6A','15B'] ).
verb( 'clang', 'clangs', 'clanging', 'clanged', 'clanged', _, [] ).
verb( 'clank', 'clanks', 'clanking', 'clanked', 'clanked', _, [] ).
verb( 'clap', 'claps', 'clapping', 'clapped', 'clapped', _, ['2A','2B','6A','14','15A','15B'] ).
verb( 'clarify', 'clarifies', 'clarifying', 'clarified', 'clarified', _, ['2A','6A'] ).
verb( 'clash', 'clashes', 'clashing', 'clashed', 'clashed', _, ['2A','2C','3A','6A','15B'] ).
verb( 'clasp', 'clasps', 'clasping', 'clasped', 'clasped', _, ['2A','6A','15A'] ).
verb( 'class', 'classes', 'classing', 'classed', 'classed', tran, ['6A','14'] ).
verb( 'classify', 'classifies', 'classifying', 'classified', 'classified', tran, ['6A'] ).
verb( 'clatter', 'clatters', 'clattering', 'clattered', 'clattered', _, ['2A','2C','6A'] ).
verb( 'claw', 'claws', 'clawing', 'clawed', 'clawed', tran, ['3A','6A','15B'] ).
verb( 'clean', 'cleans', 'cleaning', 'cleaned', 'cleaned', _, ['2A','2C','6A','15B'] ).
verb( 'cleanse', 'cleanses', 'cleansing', 'cleansed', 'cleansed', tran, ['6A'] ).
verb( 'clear', 'clears', 'clearing', 'cleared', 'cleared', _, ['2C','6A','14','15B'] ).
verb( 'cleave', 'cleaves', 'cleaving', 'cleaved', 'cleaved', _, ['2A','3A','6A','14','15A','15B','22'] ).
verb( 'clench', 'clenches', 'clenching', 'clenched', 'clenched', tran, ['6A','14'] ).
verb( 'clerk', 'clerks', 'clerking', 'clerked', 'clerked', intran, [] ).
verb( 'clew', 'clews', 'clewing', 'clewed', 'clewed', tran, [] ).
verb( 'click', 'clicks', 'clicking', 'clicked', 'clicked', intran, ['2A'] ).
verb( 'climax', 'climaxes', 'climaxing', 'climaxed', 'climaxed', _, [] ).
verb( 'climb', 'climbs', 'climbing', 'climbed', 'climbed', _, ['2A','2C','6A'] ).
verb( 'clinch', 'clinches', 'clinching', 'clinched', 'clinched', _, ['2A','6A'] ).
verb( 'cling', 'clings', 'clinging', 'clung', 'clung', intran, ['2C','3A'] ).
verb( 'clink', 'clinks', 'clinking', 'clinked', 'clinked', _, [] ).
verb( 'clip', 'clips', 'clipping', 'clipped', 'clipped', tran, ['2C','6A','15A','15B','22'] ).
verb( 'cloak', 'cloaks', 'cloaking', 'cloaked', 'cloaked', tran, ['6A'] ).
verb( 'clobber', 'clobbers', 'clobbering', 'clobbered', 'clobbered', tran, ['6A'] ).
verb( 'clock', 'clocks', 'clocking', 'clocked', 'clocked', _, ['2C','6A','15B'] ).
verb( 'clog', 'clogs', 'clogging', 'clogged', 'clogged', _, ['2A','2C','6A','15B'] ).
verb( 'cloister', 'cloisters', 'cloistering', 'cloistered', 'cloistered', tran, ['6A'] ).
verb( 'close', 'closes', 'closing', 'closed', 'closed', _, ['2A','2C','3A','6A','15B'] ).
verb( 'closet', 'closets', 'closeting', 'closeted', 'closeted', tran, [] ).
verb( 'clot', 'clots', 'clotting', 'clotted', 'clotted', _, ['2A','6A'] ).
verb( 'clothe', 'clothes', 'clothing', 'clothed', 'clothed', tran, ['6A'] ).
verb( 'cloud', 'clouds', 'clouding', 'clouded', 'clouded', _, ['2A','2C','3A','6A'] ).
verb( 'clout', 'clouts', 'clouting', 'clouted', 'clouted', tran, [] ).
verb( 'clown', 'clowns', 'clowning', 'clowned', 'clowned', intran, ['2A'] ).
verb( 'cloy', 'cloys', 'cloying', 'cloyed', 'cloyed', _, ['2A','6A'] ).
verb( 'club', 'clubs', 'clubbing', 'clubbed', 'clubbed', _, ['2C','6A','16A'] ).
verb( 'cluck', 'clucks', 'clucking', 'clucked', 'clucked', intran, [] ).
verb( 'clump', 'clumps', 'clumping', 'clumped', 'clumped', _, ['2A','2C'] ).
verb( 'clunk', 'clunks', 'clunking', 'clunked', 'clunked', intran, [] ).
verb( 'cluster', 'clusters', 'clustering', 'clustered', 'clustered', intran, ['2A','2C','3A'] ).
verb( 'clutch', 'clutches', 'clutching', 'clutched', 'clutched', _, ['2A','3A','15A'] ).
verb( 'clutter', 'clutters', 'cluttering', 'cluttered', 'cluttered', tran, ['6A','15B'] ).
verb( 'co-opt', 'co-opts', 'co-opting', 'co-opted', 'co-opted', tran, ['6A','14'] ).
verb( 'co-star', 'co-stars', 'co-starring', 'co-starred', 'co-starred', _, ['6A','14'] ).
verb( 'coach', 'coaches', 'coaching', 'coached', 'coached', _, ['2A','6A','14'] ).
verb( 'coagulate', 'coagulates', 'coagulating', 'coagulated', 'coagulated', _, ['2A','6A'] ).
verb( 'coal', 'coals', 'coaling', 'coaled', 'coaled', _, ['2A','6A'] ).
verb( 'coalesce', 'coalesces', 'coalescing', 'coalesced', 'coalesced', intran, [] ).
verb( 'coarsen', 'coarsens', 'coarsening', 'coarsened', 'coarsened', _, ['2A','6A'] ).
verb( 'coast', 'coasts', 'coasting', 'coasted', 'coasted', _, ['2A','2C','3A','6A'] ).
verb( 'coat', 'coats', 'coating', 'coated', 'coated', tran, ['6A','14'] ).
verb( 'coax', 'coaxes', 'coaxing', 'coaxed', 'coaxed', _, ['2A','6A','14','15B','17'] ).
verb( 'cobble', 'cobbles', 'cobbling', 'cobbled', 'cobbled', tran, ['6A'] ).
verb( 'cock', 'cocks', 'cocking', 'cocked', 'cocked', tran, ['6A','15B'] ).
verb( 'cocoon', 'cocoons', 'cocooning', 'cocooned', 'cocooned', tran, ['6A'] ).
verb( 'cod', 'cods', 'codding', 'codded', 'codded', _, ['2A','6A'] ).
verb( 'coddle', 'coddles', 'coddling', 'coddled', 'coddled', tran, ['6A'] ).
verb( 'code', 'codes', 'coding', 'coded', 'coded', tran, [] ).
verb( 'codify', 'codifies', 'codifying', 'codified', 'codified', tran, ['6A'] ).
verb( 'coerce', 'coerces', 'coercing', 'coerced', 'coerced', tran, ['6A','14'] ).
verb( 'coexist', 'coexists', 'coexisting', 'coexisted', 'coexisted', intran, ['2A','3A'] ).
verb( 'cogitate', 'cogitates', 'cogitating', 'cogitated', 'cogitated', _, ['2A','3A','6A','14'] ).
verb( 'cohabit', 'cohabits', 'cohabiting', 'cohabited', 'cohabited', intran, [] ).
verb( 'cohere', 'coheres', 'cohering', 'cohered', 'cohered', intran, ['2A'] ).
verb( 'coil', 'coils', 'coiling', 'coiled', 'coiled', _, ['2A','2C','6A','15A','15B'] ).
verb( 'coin', 'coins', 'coining', 'coined', 'coined', tran, ['6A'] ).
verb( 'coincide', 'coincides', 'coinciding', 'coincided', 'coincided', intran, ['2A','3A'] ).
verb( 'coke', 'cokes', 'coking', 'coked', 'coked', tran, [] ).
verb( 'cold-shoulder', 'cold-shoulders', 'cold-shouldering', 'cold-shouldered', 'cold-shouldered', tran, ['6A'] ).
verb( 'collaborate', 'collaborates', 'collaborating', 'collaborated', 'collaborated', intran, ['2A','3A'] ).
verb( 'collapse', 'collapses', 'collapsing', 'collapsed', 'collapsed', intran, ['2A','6A'] ).
verb( 'collar', 'collars', 'collaring', 'collared', 'collared', tran, ['6A'] ).
verb( 'collate', 'collates', 'collating', 'collated', 'collated', tran, ['6A'] ).
verb( 'collect', 'collects', 'collecting', 'collected', 'collected', _, ['2A','2C','6A','15B'] ).
verb( 'collectivize', 'collectivizes', 'collectivizing', 'collectivized', 'collectivized', tran, ['6A'] ).
verb( 'collide', 'collides', 'colliding', 'collided', 'collided', intran, ['2A','2C','3A'] ).
verb( 'collocate', 'collocates', 'collocating', 'collocated', 'collocated', intran, ['2A','3A'] ).
verb( 'colonize', 'colonizes', 'colonizing', 'colonized', 'colonized', tran, ['6A'] ).
verb( 'colour', 'colours', 'colouring', 'coloured', 'coloured', _, ['2A','2C','6A','22'] ).
verb( 'comb', 'combs', 'combing', 'combed', 'combed', _, ['2C','3A','6A','15B'] ).
verb( 'combat', 'combats', 'combating', 'combated', 'combated', _, ['3A','6A'] ).
verb( 'combine', 'combines', 'combining', 'combined', 'combined', _, ['2A','3A','4A','6A','14'] ).
verb( 'come', 'comes', 'coming', 'came', 'come', intran, ['2A','2B','2C','2D','2E','3A','4A'] ).
verb( 'comfort', 'comforts', 'comforting', 'comforted', 'comforted', tran, ['6A'] ).
verb( 'command', 'commands', 'commanding', 'commanded', 'commanded', _, ['2A','6A','9','17'] ).
verb( 'commandeer', 'commandeers', 'commandeering', 'commandeered', 'commandeered', tran, ['6A'] ).
verb( 'commemorate', 'commemorates', 'commemorating', 'commemorated', 'commemorated', tran, ['6A'] ).
verb( 'commence', 'commences', 'commencing', 'commenced', 'commenced', _, ['2A','3A','6A','6C'] ).
verb( 'commend', 'commends', 'commending', 'commended', 'commended', tran, ['6A','14'] ).
verb( 'comment', 'comments', 'commenting', 'commented', 'commented', intran, ['2A','3A'] ).
verb( 'commentate', 'commentates', 'commentating', 'commentated', 'commentated', intran, ['3A'] ).
verb( 'commercialize', 'commercializes', 'commercializing', 'commercialized', 'commercialized', tran, ['6A'] ).
verb( 'commingle', 'commingles', 'commingling', 'commingled', 'commingled', _, ['2A','6A'] ).
verb( 'commiserate', 'commiserates', 'commiserating', 'commiserated', 'commiserated', intran, ['3A'] ).
verb( 'commission', 'commissions', 'commissioning', 'commissioned', 'commissioned', tran, ['6A','17'] ).
verb( 'commit', 'commits', 'committing', 'committed', 'committed', tran, ['6A','14','16A'] ).
verb( 'commune', 'communes', 'communing', 'communed', 'communed', intran, ['2C','3A'] ).
verb( 'communicate', 'communicates', 'communicating', 'communicated', 'communicated', _, ['2A','3A','6A','14'] ).
verb( 'commute', 'commutes', 'commuting', 'commuted', 'commuted', _, ['2A','6A','14'] ).
verb( 'comp`ere', 'comp`eres', 'comp`ering', 'comp`ered', 'comp`ered', tran, [] ).
verb( 'compact', 'compacts', 'compacting', 'compacted', 'compacted', tran, [] ).
verb( 'compare', 'compares', 'comparing', 'compared', 'compared', _, ['3A','6A','14'] ).
verb( 'compartmentalize', 'compartmentalizes', 'compartmentalizing', 'compartmentalized', 'compartmentalized', tran, [] ).
verb( 'compass', 'compasses', 'compassing', 'compassed', 'compassed', tran, [] ).
verb( 'compel', 'compels', 'compelling', 'compelled', 'compelled', tran, ['6A','14','17'] ).
verb( 'compensate', 'compensates', 'compensating', 'compensated', 'compensated', _, ['3A','6A','14'] ).
verb( 'compete', 'competes', 'competing', 'competed', 'competed', intran, ['2A','3A'] ).
verb( 'compile', 'compiles', 'compiling', 'compiled', 'compiled', tran, ['6A'] ).
verb( 'complain', 'complains', 'complaining', 'complained', 'complained', intran, ['2A','3A','3B'] ).
verb( 'complement', 'complements', 'complementing', 'complemented', 'complemented', tran, ['6A'] ).
verb( 'complete', 'completes', 'completing', 'completed', 'completed', tran, ['6A'] ).
verb( 'complicate', 'complicates', 'complicating', 'complicated', 'complicated', tran, ['6A'] ).
verb( 'compliment', 'compliments', 'complimenting', 'complimented', 'complimented', tran, ['6A','14'] ).
verb( 'comply', 'complies', 'complying', 'complied', 'complied', tran, ['2A','3A'] ).
verb( 'comport', 'comports', 'comporting', 'comported', 'comported', _, ['3A','15A'] ).
verb( 'compose', 'composes', 'composing', 'composed', 'composed', _, ['2A','6A','16A'] ).
verb( 'compost', 'composts', 'composting', 'composted', 'composted', tran, ['6A'] ).
verb( 'compound', 'compounds', 'compounding', 'compounded', 'compounded', _, ['2A','3A','6A'] ).
verb( 'comprehend', 'comprehends', 'comprehending', 'comprehended', 'comprehended', tran, ['6A'] ).
verb( 'compress', 'compresses', 'compressing', 'compressed', 'compressed', tran, ['6A','14'] ).
verb( 'comprise', 'comprises', 'comprising', 'comprised', 'comprised', tran, ['6A'] ).
verb( 'compromise', 'compromises', 'compromising', 'compromised', 'compromised', _, ['2A','6A'] ).
verb( 'compute', 'computes', 'computing', 'computed', 'computed', _, ['2A','6A','14'] ).
verb( 'computerize', 'computerizes', 'computerizing', 'computerized', 'computerized', tran, ['6A'] ).
verb( 'con', 'cons', 'conning', 'conned', 'conned', tran, ['6A','14'] ).
verb( 'conceal', 'conceals', 'concealing', 'concealed', 'concealed', tran, ['6A','14'] ).
verb( 'concede', 'concedes', 'conceding', 'conceded', 'conceded', tran, ['6A','9','12A','13A'] ).
verb( 'conceive', 'conceives', 'conceiving', 'conceived', 'conceived', _, ['2A','3A','6A','9','10','14'] ).
verb( 'concentrate', 'concentrates', 'concentrating', 'concentrated', 'concentrated', _, ['2A','3A','6A','14'] ).
verb( 'concern', 'concerns', 'concerning', 'concerned', 'concerned', tran, ['6A','14'] ).
verb( 'concert', 'concerts', 'concerting', 'concerted', 'concerted', tran, [] ).
verb( 'conciliate', 'conciliates', 'conciliating', 'conciliated', 'conciliated', tran, ['6A'] ).
verb( 'conclude', 'concludes', 'concluding', 'concluded', 'concluded', _, ['2A','3A','6A','7A','9','14'] ).
verb( 'concoct', 'concocts', 'concocting', 'concocted', 'concocted', tran, ['6A'] ).
verb( 'concrete', 'concretes', 'concreting', 'concreted', 'concreted', _, ['2A','6A'] ).
verb( 'concur', 'concurs', 'concurring', 'concurred', 'concurred', intran, ['2A','3A','4A'] ).
verb( 'concuss', 'concusses', 'concussing', 'concussed', 'concussed', tran, ['6A'] ).
verb( 'condemn', 'condemns', 'condemning', 'condemned', 'condemned', tran, ['6A','14','16B','17'] ).
verb( 'condense', 'condenses', 'condensing', 'condensed', 'condensed', _, ['2A','3A','6A','14'] ).
verb( 'condescend', 'condescends', 'condescending', 'condescended', 'condescended', intran, ['2A','3A','4A'] ).
verb( 'condition', 'conditions', 'conditioning', 'conditioned', 'conditioned', tran, ['6A'] ).
verb( 'condole', 'condoles', 'condoling', 'condoled', 'condoled', intran, ['3A'] ).
verb( 'condone', 'condones', 'condoning', 'condoned', 'condoned', tran, ['6A','6C'] ).
verb( 'conduce', 'conduces', 'conducing', 'conduced', 'conduced', intran, ['3A'] ).
verb( 'conduct', 'conducts', 'conducting', 'conducted', 'conducted', _, ['2A','6A','14','15A','15B','16A'] ).
verb( 'cone', 'cones', 'coning', 'coned', 'coned', tran, ['15B'] ).
verb( 'confab', 'confabs', 'confabbing', 'confabbed', 'confabbed', intran, [] ).
verb( 'confabulate', 'confabulates', 'confabulating', 'confabulated', 'confabulated', intran, ['2A','3A'] ).
verb( 'confederate', 'confederates', 'confederating', 'confederated', 'confederated', _, ['2A','3A','6A','14'] ).
verb( 'confer', 'confers', 'conferring', 'conferred', 'conferred', _, ['2A','3A','14'] ).
verb( 'confess', 'confesses', 'confessing', 'confessed', 'confessed', _, ['2A','3A','3B','6A','9','14','25'] ).
verb( 'confide', 'confides', 'confiding', 'confided', 'confided', _, ['3A','14'] ).
verb( 'configure', 'configures', 'configuring', 'configured', 'configured', tran, ['6A'] ).
verb( 'confine', 'confines', 'confining', 'confined', 'confined', tran, ['6A','14'] ).
verb( 'confirm', 'confirms', 'confirming', 'confirmed', 'confirmed', tran, ['6A','9'] ).
verb( 'confiscate', 'confiscates', 'confiscating', 'confiscated', 'confiscated', tran, ['6A'] ).
verb( 'conflict', 'conflicts', 'conflicting', 'conflicted', 'conflicted', intran, ['2A','3A'] ).
verb( 'conform', 'conforms', 'conforming', 'conformed', 'conformed', _, ['2A','3A','14'] ).
verb( 'confound', 'confounds', 'confounding', 'confounded', 'confounded', tran, ['6A','14'] ).
verb( 'confront', 'confronts', 'confronting', 'confronted', 'confronted', tran, ['6A','14'] ).
verb( 'confuse', 'confuses', 'confusing', 'confused', 'confused', tran, ['6A','14'] ).
verb( 'confute', 'confutes', 'confuting', 'confuted', 'confuted', tran, ['6A'] ).
verb( 'congeal', 'congeals', 'congealing', 'congealed', 'congealed', _, ['2A','6A'] ).
verb( 'conglomerate', 'conglomerates', 'conglomerating', 'conglomerated', 'conglomerated', _, ['2A','6A'] ).
verb( 'congratulate', 'congratulates', 'congratulating', 'congratulated', 'congratulated', tran, ['6A','14'] ).
verb( 'congregate', 'congregates', 'congregating', 'congregated', 'congregated', _, ['2A','2C','6A'] ).
verb( 'conjecture', 'conjectures', 'conjecturing', 'conjectured', 'conjectured', _, ['2A','6A','9','25'] ).
verb( 'conjoin', 'conjoins', 'conjoining', 'conjoined', 'conjoined', _, ['2A','6A'] ).
verb( 'conjugate', 'conjugates', 'conjugating', 'conjugated', 'conjugated', _, ['2A','6A'] ).
verb( 'conjure', 'conjures', 'conjuring', 'conjured', 'conjured', _, ['2A','15A','15B','17'] ).
verb( 'conk', 'conks', 'conking', 'conked', 'conked', intran, [] ).
verb( 'connect', 'connects', 'connecting', 'connected', 'connected', _, ['2A','2C','3A','6A','14','15A','15B'] ).
verb( 'connive', 'connives', 'conniving', 'connived', 'connived', intran, ['3A'] ).
verb( 'connote', 'connotes', 'connoting', 'connoted', 'connoted', tran, ['6A'] ).
verb( 'conquer', 'conquers', 'conquering', 'conquered', 'conquered', tran, ['6A'] ).
verb( 'conscript', 'conscripts', 'conscripting', 'conscripted', 'conscripted', tran, ['6A','14'] ).
verb( 'consecrate', 'consecrates', 'consecrating', 'consecrated', 'consecrated', tran, ['6A','14','23'] ).
verb( 'consent', 'consents', 'consenting', 'consented', 'consented', intran, ['2A','3A','7A'] ).
verb( 'conserve', 'conserves', 'conserving', 'conserved', 'conserved', tran, ['6A'] ).
verb( 'consider', 'considers', 'considering', 'considered', 'considered', tran, ['6A','6C','8','9','10','25'] ).
verb( 'consign', 'consigns', 'consigning', 'consigned', 'consigned', tran, ['6A','14'] ).
verb( 'consist', 'consists', 'consisting', 'consisted', 'consisted', intran, ['3A'] ).
verb( 'console', 'consoles', 'consoling', 'consoled', 'consoled', tran, ['6A','14'] ).
verb( 'consolidate', 'consolidates', 'consolidating', 'consolidated', 'consolidated', _, ['2A','6A'] ).
verb( 'consort', 'consorts', 'consorting', 'consorted', 'consorted', intran, ['2C','3A'] ).
verb( 'conspire', 'conspires', 'conspiring', 'conspired', 'conspired', _, ['2A','3A','4A','6A'] ).
verb( 'constipate', 'constipates', 'constipating', 'constipated', 'constipated', tran, ['6A'] ).
verb( 'constitute', 'constitutes', 'constituting', 'constituted', 'constituted', tran, ['6A','23'] ).
verb( 'constitutionalize', 'constitutionalizes', 'constitutionalizing', 'constitutionalized', 'constitutionalized', tran, [] ).
verb( 'constrain', 'constrains', 'constraining', 'constrained', 'constrained', tran, ['6A','17'] ).
verb( 'constrict', 'constricts', 'constricting', 'constricted', 'constricted', tran, ['6A'] ).
verb( 'construct', 'constructs', 'constructing', 'constructed', 'constructed', tran, ['6A'] ).
verb( 'construe', 'construes', 'construing', 'construed', 'construed', _, ['2A','6A'] ).
verb( 'consult', 'consults', 'consulting', 'consulted', 'consulted', _, ['3A','6A','14'] ).
verb( 'consume', 'consumes', 'consuming', 'consumed', 'consumed', _, ['6A'] ).
verb( 'consummate', 'consummates', 'consummating', 'consummated', 'consummated', tran, ['6A'] ).
verb( 'contact', 'contacts', 'contacting', 'contacted', 'contacted', tran, ['6A'] ).
verb( 'contain', 'contains', 'containing', 'contained', 'contained', tran, ['6A'] ).
verb( 'contaminate', 'contaminates', 'contaminating', 'contaminated', 'contaminated', tran, ['6A'] ).
verb( 'contemn', 'contemns', 'contemning', 'contemned', 'contemned', tran, ['6A'] ).
verb( 'contemplate', 'contemplates', 'contemplating', 'contemplated', 'contemplated', _, ['2A','6A','6C','19C'] ).
verb( 'contend', 'contends', 'contending', 'contended', 'contended', _, ['3A','9'] ).
verb( 'content', 'contents', 'contenting', 'contented', 'contented', tran, ['6A','14'] ).
verb( 'contest', 'contests', 'contesting', 'contested', 'contested', _, ['3A','6A','9'] ).
verb( 'continue', 'continues', 'continuing', 'continued', 'continued', _, ['2A','2B','2D','2E','6A','6D','7A','14'] ).
verb( 'contort', 'contorts', 'contorting', 'contorted', 'contorted', tran, ['6A','14'] ).
verb( 'contour', 'contours', 'contouring', 'contoured', 'contoured', tran, ['6A'] ).
verb( 'contract', 'contracts', 'contracting', 'contracted', 'contracted', _, ['2A','2C','4A','6A','14'] ).
verb( 'contradict', 'contradicts', 'contradicting', 'contradicted', 'contradicted', tran, ['6A'] ).
verb( 'contradistinguish', 'contradistinguishes', 'contradistinguishing', 'contradistinguished', 'contradistinguished', tran, ['14'] ).
verb( 'contrast', 'contrasts', 'contrasting', 'contrasted', 'contrasted', _, ['2C','3A','6A','14'] ).
verb( 'contravene', 'contravenes', 'contravening', 'contravened', 'contravened', tran, ['6A'] ).
verb( 'contribute', 'contributes', 'contributing', 'contributed', 'contributed', _, ['2A','3A','6A','14'] ).
verb( 'contrive', 'contrives', 'contriving', 'contrived', 'contrived', _, ['2A','6A','7A'] ).
verb( 'control', 'controls', 'controlling', 'controlled', 'controlled', tran, ['6A'] ).
verb( 'controvert', 'controverts', 'controverting', 'controverted', 'controverted', tran, ['6A'] ).
verb( 'contuse', 'contuses', 'contusing', 'contused', 'contused', tran, ['6A'] ).
verb( 'convalesce', 'convalesces', 'convalescing', 'convalesced', 'convalesced', intran, ['2A'] ).
verb( 'convene', 'convenes', 'convening', 'convened', 'convened', _, ['2A','6A'] ).
verb( 'converge', 'converges', 'converging', 'converged', 'converged', intran, ['2A','2C','3A'] ).
verb( 'converse', 'converses', 'conversing', 'conversed', 'conversed', intran, ['2A','2C','3A'] ).
verb( 'convert', 'converts', 'converting', 'converted', 'converted', tran, ['6A','14'] ).
verb( 'convey', 'conveys', 'conveying', 'conveyed', 'conveyed', tran, ['6A','14'] ).
verb( 'convict', 'convicts', 'convicting', 'convicted', 'convicted', tran, ['6A','14'] ).
verb( 'convince', 'convinces', 'convincing', 'convinced', 'convinced', tran, ['6A','11','14'] ).
verb( 'convoke', 'convokes', 'convoking', 'convoked', 'convoked', tran, ['6A'] ).
verb( 'convoy', 'convoys', 'convoying', 'convoyed', 'convoyed', tran, ['6A'] ).
verb( 'convulse', 'convulses', 'convulsing', 'convulsed', 'convulsed', tran, ['6A'] ).
verb( 'coo', 'coos', 'cooing', 'cooed', 'cooed', _, [] ).
verb( 'cook', 'cooks', 'cooking', 'cooked', 'cooked', _, ['2A','6A','12B','13B','15B'] ).
verb( 'cool', 'cools', 'cooling', 'cooled', 'cooled', _, ['2A','2C','6A'] ).
verb( 'coop', 'coops', 'cooping', 'cooped', 'cooped', tran, ['6A','15B'] ).
verb( 'cooperate', 'cooperates', 'cooperating', 'cooperated', 'cooperated', intran, ['2A','3A','4A'] ).
verb( 'coordinate', 'coordinates', 'coordinating', 'coordinated', 'coordinated', tran, ['6A'] ).
verb( 'cop', 'cops', 'copping', 'copped', 'copped', _, ['2C'] ).
verb( 'cope', 'copes', 'coping', 'coped', 'coped', intran, ['2A','3A'] ).
verb( 'copper', 'coppers', 'coppering', 'coppered', 'coppered', tran, [] ).
verb( 'copper-bottom', 'copper-bottoms', 'copper-bottoming', 'copper-bottomed', 'copper-bottomed', tran, [] ).
verb( 'copulate', 'copulates', 'copulating', 'copulated', 'copulated', intran, ['2A','3A'] ).
verb( 'copy', 'copies', 'copying', 'copied', 'copied', _, ['2A','6A','15A','15B'] ).
verb( 'copyright', 'copyrights', 'copyrighting', 'copyrighted', 'copyrighted', tran, [] ).
verb( 'cord', 'cords', 'cording', 'corded', 'corded', tran, [] ).
verb( 'cordon', 'cordons', 'cordoning', 'cordoned', 'cordoned', tran, ['15B'] ).
verb( 'core', 'cores', 'coring', 'cored', 'cored', tran, ['6A'] ).
verb( 'cork', 'corks', 'corking', 'corked', 'corked', tran, ['6A','15A'] ).
verb( 'corn', 'corns', 'corning', 'corned', 'corned', tran, [] ).
verb( 'corner', 'corners', 'cornering', 'cornered', 'cornered', _, ['2A','6A'] ).
verb( 'corral', 'corrals', 'corralling', 'corralled', 'corralled', tran, ['6A'] ).
verb( 'correct', 'corrects', 'correcting', 'corrected', 'corrected', tran, ['6A','14'] ).
verb( 'correlate', 'correlates', 'correlating', 'correlated', 'correlated', _, ['2A','3A','6A','14'] ).
verb( 'correspond', 'corresponds', 'corresponding', 'corresponded', 'corresponded', intran, ['2A','3A'] ).
verb( 'corroborate', 'corroborates', 'corroborating', 'corroborated', 'corroborated', tran, ['6A'] ).
verb( 'corrode', 'corrodes', 'corroding', 'corroded', 'corroded', _, ['2A','6A'] ).
verb( 'corrugate', 'corrugates', 'corrugating', 'corrugated', 'corrugated', _, ['2A','6A'] ).
verb( 'corrupt', 'corrupts', 'corrupting', 'corrupted', 'corrupted', _, ['2A','6A'] ).
verb( 'coruscate', 'coruscates', 'coruscating', 'coruscated', 'coruscated', intran, ['2A'] ).
verb( 'cosh', 'coshes', 'coshing', 'coshed', 'coshed', tran, [] ).
verb( 'cosset', 'cossets', 'cosseting', 'cosseted', 'cosseted', tran, [] ).
verb( 'cost', 'costs', 'costing', 'costed', 'costed', _, ['2B','6A'] ).
verb( 'cotton', 'cottons', 'cottoning', 'cottoned', 'cottoned', intran, [] ).
verb( 'couch', 'couches', 'couching', 'couched', 'couched', _, ['2A','6A','14'] ).
verb( 'cough', 'coughs', 'coughing', 'coughed', 'coughed', _, ['2A','15B'] ).
verb( 'counsel', 'counsels', 'counselling', 'counselled', 'counselled', tran, ['6A','6B','17'] ).
verb( 'count', 'counts', 'counting', 'counted', 'counted', _, ['2A','2C','3A','6A','14','16B','25'] ).
verb( 'countenance', 'countenances', 'countenancing', 'countenanced', 'countenanced', tran, ['6A'] ).
verb( 'counter', 'counters', 'countering', 'countered', 'countered', _, ['2A','3A','6A','14'] ).
verb( 'counteract', 'counteracts', 'counteracting', 'counteracted', 'counteracted', tran, ['6A'] ).
verb( 'counterattack', 'counterattacks', 'counterattacking', 'counterattacked', 'counterattacked', _, [] ).
verb( 'counterbalance', 'counterbalances', 'counterbalancing', 'counterbalanced', 'counterbalanced', tran, ['6A'] ).
verb( 'counterfeit', 'counterfeits', 'counterfeiting', 'counterfeited', 'counterfeited', tran, ['6A'] ).
verb( 'countermand', 'countermands', 'countermanding', 'countermanded', 'countermanded', tran, ['6A'] ).
verb( 'countermine', 'countermines', 'countermining', 'countermined', 'countermined', _, [] ).
verb( 'counterplot', 'counterplots', 'counterplotting', 'counterplotted', 'counterplotted', _, [] ).
verb( 'counterpoise', 'counterpoises', 'counterpoising', 'counterpoised', 'counterpoised', tran, [] ).
verb( 'countersign', 'countersigns', 'countersigning', 'countersigned', 'countersigned', tran, ['6A'] ).
verb( 'countersink', 'countersinks', 'countersinking', 'countersunk', 'countersunk', tran, [] ).
verb( 'countervail', 'countervails', 'countervailing', 'countervailed', 'countervailed', _, ['2A','6A'] ).
verb( 'couple', 'couples', 'coupling', 'coupled', 'coupled', _, ['2A','6A','14'] ).
verb( 'course', 'courses', 'coursing', 'coursed', 'coursed', _, ['2A','2C','6A'] ).
verb( 'court', 'courts', 'courting', 'courted', 'courted', _, ['2A','6A'] ).
verb( 'court-martial', 'court-martials', 'court-martialing', 'court-martialed', 'court-martialed', tran, ['6A'] ).
verb( 'covenant', 'covenants', 'covenanting', 'covenanted', 'covenanted', _, ['3A','6A','7A','9','14'] ).
verb( 'cover', 'covers', 'covering', 'covered', 'covered', tran, ['6A','15A','15B'] ).
verb( 'covet', 'covets', 'coveting', 'coveted', 'coveted', tran, ['6A'] ).
verb( 'cow', 'cows', 'cowing', 'cowed', 'cowed', tran, ['6A'] ).
verb( 'cower', 'cowers', 'cowering', 'cowered', 'cowered', intran, ['2A','2C'] ).
verb( 'cox', 'coxes', 'coxing', 'coxed', 'coxed', _, ['2A','6A'] ).
verb( 'cozen', 'cozens', 'cozening', 'cozened', 'cozened', tran, ['6A','14'] ).
verb( 'crab', 'crabs', 'crabbing', 'crabbed', 'crabbed', _, [] ).
verb( 'crack', 'cracks', 'cracking', 'cracked', 'cracked', _, ['2A','2C','6A','14'] ).
verb( 'crackle', 'crackles', 'crackling', 'crackled', 'crackled', intran, ['2A','2C'] ).
verb( 'cradle', 'cradles', 'cradling', 'cradled', 'cradled', tran, ['6A','14'] ).
verb( 'cram', 'crams', 'cramming', 'crammed', 'crammed', _, ['2A','6A','14','15B'] ).
verb( 'cramp', 'cramps', 'cramping', 'cramped', 'cramped', tran, ['6A'] ).
verb( 'crane', 'cranes', 'craning', 'craned', 'craned', _, ['2A','2C','6A','16A'] ).
verb( 'crank', 'cranks', 'cranking', 'cranked', 'cranked', tran, ['6A','15B'] ).
verb( 'crap', 'craps', 'crapping', 'crapped', 'crapped', intran, [] ).
verb( 'crash', 'crashes', 'crashing', 'crashed', 'crashed', _, ['2A','2C','6A'] ).
verb( 'crash-dive', 'crash-dives', 'crash-diving', 'crash-dived', 'crash-dived', intran, [] ).
verb( 'crash-land', 'crash-lands', 'crash-landing', 'crash-landed', 'crash-landed', _, [] ).
verb( 'crate', 'crates', 'crating', 'crated', 'crated', tran, [] ).
verb( 'crave', 'craves', 'craving', 'craved', 'craved', _, ['2A','3A'] ).
verb( 'crawl', 'crawls', 'crawling', 'crawled', 'crawled', intran, ['2A','2C'] ).
verb( 'crayon', 'crayons', 'crayoning', 'crayoned', 'crayoned', tran, [] ).
verb( 'creak', 'creaks', 'creaking', 'creaked', 'creaked', intran, [] ).
verb( 'cream', 'creams', 'creaming', 'creamed', 'creamed', tran, [] ).
verb( 'crease', 'creases', 'creasing', 'creased', 'creased', _, ['2A','6A'] ).
verb( 'create', 'creates', 'creating', 'created', 'created', tran, ['6A','23'] ).
verb( 'credit', 'credits', 'crediting', 'credited', 'credited', tran, ['6A','14'] ).
verb( 'creep', 'creeps', 'creeping', 'crept', 'crept', intran, ['2A','2B','2C'] ).
verb( 'cremate', 'cremates', 'cremating', 'cremated', 'cremated', tran, ['6A'] ).
verb( 'crepitate', 'crepitates', 'crepitating', 'crepitated', 'crepitated', intran, ['2A'] ).
verb( 'crest', 'crests', 'cresting', 'crested', 'crested', _, [] ).
verb( 'crew', 'crews', 'crewing', 'crewed', 'crewed', intran, ['3A','6A'] ).
verb( 'crib', 'cribs', 'cribbing', 'cribbed', 'cribbed', _, [] ).
verb( 'crick', 'cricks', 'cricking', 'cricked', 'cricked', tran, [] ).
verb( 'crime', 'crimes', 'criming', 'crimed', 'crimed', tran, [] ).
verb( 'crimp', 'crimps', 'crimping', 'crimped', 'crimped', tran, [] ).
verb( 'crimson', 'crimsons', 'crimsoning', 'crimsoned', 'crimsoned', _, [] ).
verb( 'cringe', 'cringes', 'cringing', 'cringed', 'cringed', intran, ['2A','2C'] ).
verb( 'crinkle', 'crinkles', 'crinkling', 'crinkled', 'crinkled', _, ['2A','2C','6A','15B'] ).
verb( 'cripple', 'cripples', 'crippling', 'crippled', 'crippled', tran, ['6A'] ).
verb( 'crisp', 'crisps', 'crisping', 'crisped', 'crisped', _, [] ).
verb( 'crisscross', 'crisscrosses', 'crisscrossing', 'crisscrossed', 'crisscrossed', _, [] ).
verb( 'criticize', 'criticizes', 'criticizing', 'criticized', 'criticized', _, ['2A','6A','14'] ).
verb( 'croak', 'croaks', 'croaking', 'croaked', 'croaked', _, ['2A','6A','15B'] ).
verb( 'crochet', 'crochets', 'crocheting', 'crocheted', 'crocheted', _, ['2A','6A'] ).
verb( 'crock', 'crocks', 'crocking', 'crocked', 'crocked', _, ['2C','15B'] ).
verb( 'crook', 'crooks', 'crooking', 'crooked', 'crooked', _, ['2A','6A'] ).
verb( 'croon', 'croons', 'crooning', 'crooned', 'crooned', _, ['2C','6A','13A','15A'] ).
verb( 'crop', 'crops', 'cropping', 'cropped', 'cropped', _, ['2A','2C','6A','14','22'] ).
verb( 'cross', 'crosses', 'crossing', 'crossed', 'crossed', _, ['2A','2C','6A','14','15A','15B'] ).
verb( 'cross-examine', 'cross-examines', 'cross-examining', 'cross-examined', 'cross-examined', tran, ['6A'] ).
verb( 'cross-fertilize', 'cross-fertilizes', 'cross-fertilizing', 'cross-fertilized', 'cross-fertilized', tran, [] ).
verb( 'cross-index', 'cross-indexes', 'cross-indexing', 'cross-indexed', 'cross-indexed', tran, [] ).
verb( 'cross-question', 'cross-questions', 'cross-questioning', 'cross-questioned', 'cross-questioned', tran, ['6A'] ).
verb( 'crossbreed', 'crossbreeds', 'crossbreeding', 'crossbred', 'crossbred', tran, ['2A','6A'] ).
verb( 'crosscheck', 'crosschecks', 'crosschecking', 'crosschecked', 'crosschecked', _, ['2A','6A'] ).
verb( 'crouch', 'crouches', 'crouching', 'crouched', 'crouched', intran, ['2A','2C','4A'] ).
verb( 'crow', 'crows', 'crowing', 'crowed', 'crowed', intran, ['3A','6A'] ).
verb( 'crowd', 'crowds', 'crowding', 'crowded', 'crowded', _, ['2B','2C','6A','14','15A','15B'] ).
verb( 'crown', 'crowns', 'crowning', 'crowned', 'crowned', tran, ['6A','14','23'] ).
verb( 'crucify', 'crucifies', 'crucifying', 'crucified', 'crucified', tran, [] ).
verb( 'cruise', 'cruises', 'cruising', 'cruised', 'cruised', intran, ['2A','2C'] ).
verb( 'crumble', 'crumbles', 'crumbling', 'crumbled', 'crumbled', _, ['2A','2C','6A'] ).
verb( 'crumple', 'crumples', 'crumpling', 'crumpled', 'crumpled', _, ['2A','2C','6A','15B'] ).
verb( 'crunch', 'crunches', 'crunching', 'crunched', 'crunched', _, ['2A','2C','6A'] ).
verb( 'crusade', 'crusades', 'crusading', 'crusaded', 'crusaded', intran, ['2A','3A'] ).
verb( 'crush', 'crushes', 'crushing', 'crushed', 'crushed', _, ['2A','2C','6A','15A','15B'] ).
verb( 'crust', 'crusts', 'crusting', 'crusted', 'crusted', _, ['2A','2C','6A'] ).
verb( 'cry', 'cries', 'crying', 'cried', 'cried', _, ['2A','2B','2C','3A','4A','6A','9','14','15A','15B'] ).
verb( 'crystallize', 'crystallizes', 'crystallizing', 'crystallized', 'crystallized', _, ['2A','6A'] ).
verb( 'cube', 'cubes', 'cubing', 'cubed', 'cubed', tran, ['6A'] ).
verb( 'cuckold', 'cuckolds', 'cuckolding', 'cuckolded', 'cuckolded', tran, ['6A'] ).
verb( 'cuddle', 'cuddles', 'cuddling', 'cuddled', 'cuddled', _, ['2C','6A','15B'] ).
verb( 'cudgel', 'cudgels', 'cudgeling', 'cudgeled', 'cudgeled', tran, [] ).
verb( 'cuff', 'cuffs', 'cuffing', 'cuffed', 'cuffed', tran, [] ).
verb( 'cull', 'culls', 'culling', 'culled', 'culled', tran, ['6A'] ).
verb( 'culminate', 'culminates', 'culminating', 'culminated', 'culminated', tran, ['3A'] ).
verb( 'cultivate', 'cultivates', 'cultivating', 'cultivated', 'cultivated', tran, ['6A'] ).
verb( 'cumber', 'cumbers', 'cumbering', 'cumbered', 'cumbered', tran, ['6A','14'] ).
verb( 'cup', 'cups', 'cupping', 'cupped', 'cupped', tran, ['6A'] ).
verb( 'curb', 'curbs', 'curbing', 'curbed', 'curbed', tran, ['6A'] ).
verb( 'curdle', 'curdles', 'curdling', 'curdled', 'curdled', _, ['2A','6A'] ).
verb( 'cure', 'cures', 'curing', 'cured', 'cured', _, ['6A','14'] ).
verb( 'curl', 'curls', 'curling', 'curled', 'curled', _, ['2A','2C','6A','15A','15B'] ).
verb( 'curry', 'curries', 'currying', 'curried', 'curried', tran, ['6A'] ).
verb( 'curse', 'curses', 'cursing', 'cursed', 'cursed', _, ['2A','3A','6A'] ).
verb( 'curtail', 'curtails', 'curtailing', 'curtailed', 'curtailed', tran, ['6A'] ).
verb( 'curtain', 'curtains', 'curtaining', 'curtained', 'curtained', tran, ['6A','15B'] ).
verb( 'curtsey', 'curtseys', 'curtseying', 'curtseyed', 'curtseyed', intran, [] ).
verb( 'curtsy', 'curtsies', 'curtsying', 'curtsied', 'curtsied', intran, [] ).
verb( 'curve', 'curves', 'curving', 'curved', 'curved', _, ['2A','6A'] ).
verb( 'cushion', 'cushions', 'cushioning', 'cushioned', 'cushioned', tran, ['6A','14'] ).
verb( 'cut', 'cuts', 'cutting', 'cut', 'cut', _, ['2A','2C','3A','6A','12B','13B','15A','15B','22'] ).
verb( 'cycle', 'cycles', 'cycling', 'cycled', 'cycled', intran, ['2A','2B','2C'] ).
verb( 'cyclostyle', 'cyclostyles', 'cyclostyling', 'cyclostyled', 'cyclostyled', tran, [] ).
verb( 'cypher', 'cyphers', 'cyphering', 'cyphered', 'cyphered', _, ['2A','6A'] ).
verb( 'dab', 'dabs', 'dabbing', 'dabbed', 'dabbed', _, ['2C','3A','6A','14','15A','15B'] ).
verb( 'dabble', 'dabbles', 'dabbling', 'dabbled', 'dabbled', _, ['2C','3A','6A','15A','15B'] ).
verb( 'dally', 'dallies', 'dallying', 'dallied', 'dallied', intran, ['2A','3A'] ).
verb( 'dam', 'dams', 'damming', 'dammed', 'dammed', tran, ['6A','15B'] ).
verb( 'damage', 'damages', 'damaging', 'damaged', 'damaged', tran, ['6A'] ).
verb( 'damascene', 'damascenes', 'damascening', 'damascened', 'damascened', tran, [] ).
verb( 'damn', 'damns', 'damning', 'damned', 'damned', tran, ['6A'] ).
verb( 'damp', 'damps', 'damping', 'damped', 'damped', _, ['2C','6A','15B'] ).
verb( 'dampen', 'dampens', 'dampening', 'dampened', 'dampened', _, [] ).
verb( 'dance', 'dances', 'dancing', 'danced', 'danced', _, ['2A','2C','6A','15A'] ).
verb( 'dandle', 'dandles', 'dandling', 'dandled', 'dandled', tran, ['6A','15A','15B'] ).
verb( 'dangle', 'dangles', 'dangling', 'dangled', 'dangled', _, ['2A','2C','6A','15A'] ).
verb( 'dapple', 'dapples', 'dappling', 'dappled', 'dappled', tran, ['6A'] ).
verb( 'dare', 'dares', 'daring', 'dared', 'dared', unknown, ['4A','5','6A','17'] ).
verb( 'dare', 'dares', 'daring', 'dared', 'dared', _, ['4A','5','6A','17'] ).
verb( 'darken', 'darkens', 'darkening', 'darkened', 'darkened', _, ['2A','6A'] ).
verb( 'darn', 'darns', 'darning', 'darned', 'darned', _, ['2A','6A'] ).
verb( 'dart', 'darts', 'darting', 'darted', 'darted', _, ['2A','2C','6A','15A','15B'] ).
verb( 'dash', 'dashes', 'dashing', 'dashed', 'dashed', _, ['2C','6A','15A','15B'] ).
verb( 'date', 'dates', 'dating', 'dated', 'dated', _, ['2A','2C','3A','6A'] ).
verb( 'daub', 'daubs', 'daubing', 'daubed', 'daubed', _, ['2A','6A','14','15A','15B'] ).
verb( 'daunt', 'daunts', 'daunting', 'daunted', 'daunted', tran, ['6A'] ).
verb( 'dawdle', 'dawdles', 'dawdling', 'dawdled', 'dawdled', _, ['2A','2C','15B'] ).
verb( 'dawn', 'dawns', 'dawning', 'dawned', 'dawned', intran, ['2A','3A'] ).
verb( 'daydream', 'daydreams', 'daydreaming', 'daydreamed', 'daydreamed', intran, [] ).
verb( 'daze', 'dazes', 'dazing', 'dazed', 'dazed', tran, ['6A'] ).
verb( 'dazzle', 'dazzles', 'dazzling', 'dazzled', 'dazzled', tran, ['6A'] ).
verb( 'de-escalate', 'de-escalates', 'de-escalating', 'de-escalated', 'de-escalated', tran, ['6A'] ).
verb( 'de-ice', 'de-ices', 'de-icing', 'de-iced', 'de-iced', tran, ['6A'] ).
verb( 'deaden', 'deadens', 'deadening', 'deadened', 'deadened', tran, ['6A'] ).
verb( 'deafen', 'deafens', 'deafening', 'deafened', 'deafened', tran, ['6A'] ).
verb( 'deal', 'deals', 'dealing', 'dealt', 'dealt', _, ['2A','3A','6A','12A','13A','15B'] ).
verb( 'debag', 'debags', 'debagging', 'debagged', 'debagged', tran, ['6A'] ).
verb( 'debar', 'debars', 'debarring', 'debarred', 'debarred', tran, ['14'] ).
verb( 'debark', 'debarks', 'debarking', 'debarked', 'debarked', _, [] ).
verb( 'debase', 'debases', 'debasing', 'debased', 'debased', tran, ['6A'] ).
verb( 'debate', 'debates', 'debating', 'debated', 'debated', _, ['2A','2C','6A','8','10'] ).
verb( 'debauch', 'debauches', 'debauching', 'debauched', 'debauched', tran, ['6A'] ).
verb( 'debilitate', 'debilitates', 'debilitating', 'debilitated', 'debilitated', tran, ['6A'] ).
verb( 'debit', 'debits', 'debiting', 'debited', 'debited', tran, ['6A','14'] ).
verb( 'debouch', 'debouches', 'debouching', 'debouched', 'debouched', _, ['2A','6A'] ).
verb( 'debrief', 'debriefs', 'debriefing', 'debriefed', 'debriefed', tran, ['6A'] ).
verb( 'debug', 'debugs', 'debugging', 'debugged', 'debugged', tran, ['6A'] ).
verb( 'debunk', 'debunks', 'debunking', 'debunked', 'debunked', tran, ['6A'] ).
verb( 'decamp', 'decamps', 'decamping', 'decamped', 'decamped', intran, ['2A','3A'] ).
verb( 'decant', 'decants', 'decanting', 'decanted', 'decanted', tran, ['6A'] ).
verb( 'decapitate', 'decapitates', 'decapitating', 'decapitated', 'decapitated', tran, ['6A'] ).
verb( 'decarbonize', 'decarbonizes', 'decarbonizing', 'decarbonized', 'decarbonized', tran, [] ).
verb( 'decay', 'decays', 'decaying', 'decayed', 'decayed', intran, ['2A'] ).
verb( 'decease', 'deceases', 'deceasing', 'deceased', 'deceased', intran, [] ).
verb( 'deceive', 'deceives', 'deceiving', 'deceived', 'deceived', tran, ['6A','14'] ).
verb( 'decelerate', 'decelerates', 'decelerating', 'decelerated', 'decelerated', _, [] ).
verb( 'decentralize', 'decentralizes', 'decentralizing', 'decentralized', 'decentralized', tran, ['6A'] ).
verb( 'decide', 'decides', 'deciding', 'decided', 'decided', _, ['2A','3A','6A','7A','8','9','10','14','17'] ).
verb( 'decimalize', 'decimalizes', 'decimalizing', 'decimalized', 'decimalized', tran, ['6A'] ).
verb( 'decimate', 'decimates', 'decimating', 'decimated', 'decimated', tran, ['6A'] ).
verb( 'decipher', 'deciphers', 'deciphering', 'deciphered', 'deciphered', tran, ['6A'] ).
verb( 'deck', 'decks', 'decking', 'decked', 'decked', tran, ['6A','14','15A'] ).
verb( 'declaim', 'declaims', 'declaiming', 'declaimed', 'declaimed', _, ['2A','3A','6A'] ).
verb( 'declare', 'declares', 'declaring', 'declared', 'declared', _, ['3A','6A','9','14','25'] ).
verb( 'declassify', 'declassifies', 'declassifying', 'declassified', 'declassified', tran, ['6A'] ).
verb( 'decline', 'declines', 'declining', 'declined', 'declined', _, ['2A','2C','6A','7A'] ).
verb( 'declutch', 'declutches', 'declutching', 'declutched', 'declutched', intran, ['2A'] ).
verb( 'decode', 'decodes', 'decoding', 'decoded', 'decoded', tran, ['6A'] ).
verb( 'decoke', 'decokes', 'decoking', 'decoked', 'decoked', tran, [] ).
verb( 'decolonize', 'decolonizes', 'decolonizing', 'decolonized', 'decolonized', tran, ['6A'] ).
verb( 'decompose', 'decomposes', 'decomposing', 'decomposed', 'decomposed', _, ['2A','6A'] ).
verb( 'decompress', 'decompresses', 'decompressing', 'decompressed', 'decompressed', tran, ['6A'] ).
verb( 'decontaminate', 'decontaminates', 'decontaminating', 'decontaminated', 'decontaminated', tran, ['6A'] ).
verb( 'decontrol', 'decontrols', 'decontrolling', 'decontrolled', 'decontrolled', tran, ['6A'] ).
verb( 'decorate', 'decorates', 'decorating', 'decorated', 'decorated', tran, ['6A','14'] ).
verb( 'decoy', 'decoys', 'decoying', 'decoyed', 'decoyed', tran, ['6A','14'] ).
verb( 'decrease', 'decreases', 'decreasing', 'decreased', 'decreased', _, ['2A','6A'] ).
verb( 'decree', 'decrees', 'decreeing', 'decreed', 'decreed', tran, ['6A','9'] ).
verb( 'decry', 'decries', 'decrying', 'decried', 'decried', tran, ['6A'] ).
verb( 'dedicate', 'dedicates', 'dedicating', 'dedicated', 'dedicated', tran, ['6A','14'] ).
verb( 'deduce', 'deduces', 'deducing', 'deduced', 'deduced', tran, ['6A','9','14'] ).
verb( 'deduct', 'deducts', 'deducting', 'deducted', 'deducted', tran, ['6A','14'] ).
verb( 'deem', 'deems', 'deeming', 'deemed', 'deemed', tran, ['9','25'] ).
verb( 'deep-freeze', 'deep-freezes', 'deep-freezing', 'deep-froze', 'deep-frozen', tran, [] ).
verb( 'deepen', 'deepens', 'deepening', 'deepened', 'deepened', _, [] ).
verb( 'deface', 'defaces', 'defacing', 'defaced', 'defaced', tran, ['6A'] ).
verb( 'defame', 'defames', 'defaming', 'defamed', 'defamed', tran, ['6A'] ).
verb( 'default', 'defaults', 'defaulting', 'defaulted', 'defaulted', intran, ['2A'] ).
verb( 'defeat', 'defeats', 'defeating', 'defeated', 'defeated', tran, ['6A'] ).
verb( 'defecate', 'defecates', 'defecating', 'defecated', 'defecated', intran, ['2A'] ).
verb( 'defect', 'defects', 'defecting', 'defected', 'defected', intran, ['2A','2C','3A'] ).
verb( 'defend', 'defends', 'defending', 'defended', 'defended', tran, ['6A','14'] ).
verb( 'defer', 'defers', 'deferring', 'deferred', 'deferred', _, ['3A','6A','6C'] ).
verb( 'defile', 'defiles', 'defiling', 'defiled', 'defiled', _, ['6A'] ).
verb( 'define', 'defines', 'defining', 'defined', 'defined', tran, ['6A'] ).
verb( 'deflate', 'deflates', 'deflating', 'deflated', 'deflated', tran, ['6A'] ).
verb( 'deflect', 'deflects', 'deflecting', 'deflected', 'deflected', _, ['2A','3A','6A','14'] ).
verb( 'deflower', 'deflowers', 'deflowering', 'deflowered', 'deflowered', tran, [] ).
verb( 'defoliate', 'defoliates', 'defoliating', 'defoliated', 'defoliated', tran, ['6A'] ).
verb( 'deforest', 'deforests', 'deforesting', 'deforested', 'deforested', tran, [] ).
verb( 'deform', 'deforms', 'deforming', 'deformed', 'deformed', tran, ['6A'] ).
verb( 'defraud', 'defrauds', 'defrauding', 'defrauded', 'defrauded', tran, ['6A','14'] ).
verb( 'defray', 'defrays', 'defraying', 'defrayed', 'defrayed', tran, ['6A'] ).
verb( 'defrock', 'defrocks', 'defrocking', 'defrocked', 'defrocked', tran, [] ).
verb( 'defrost', 'defrosts', 'defrosting', 'defrosted', 'defrosted', tran, ['6A'] ).
verb( 'defuse', 'defuses', 'defusing', 'defused', 'defused', tran, ['6A'] ).
verb( 'defy', 'defies', 'defying', 'defied', 'defied', tran, ['6A','17'] ).
verb( 'degauss', 'degausses', 'degaussing', 'degaussed', 'degaussed', tran, ['6A'] ).
verb( 'degenerate', 'degenerates', 'degenerating', 'degenerated', 'degenerated', intran, ['2A','3A'] ).
verb( 'degrade', 'degrades', 'degrading', 'degraded', 'degraded', tran, ['6A'] ).
verb( 'dehorn', 'dehorns', 'dehorning', 'dehorned', 'dehorned', tran, ['6A'] ).
verb( 'dehumanize', 'dehumanizes', 'dehumanizing', 'dehumanized', 'dehumanized', tran, ['6A'] ).
verb( 'dehydrate', 'dehydrates', 'dehydrating', 'dehydrated', 'dehydrated', tran, ['6A'] ).
verb( 'deify', 'deifies', 'deifying', 'deified', 'deified', tran, ['6A'] ).
verb( 'deign', 'deigns', 'deigning', 'deigned', 'deigned', intran, ['4A'] ).
verb( 'deject', 'dejects', 'dejecting', 'dejected', 'dejected', tran, [] ).
verb( 'delay', 'delays', 'delaying', 'delayed', 'delayed', _, ['2A','2B','6A','6C'] ).
verb( 'delegate', 'delegates', 'delegating', 'delegated', 'delegated', tran, ['14','17'] ).
verb( 'delete', 'deletes', 'deleting', 'deleted', 'deleted', tran, ['6A','14'] ).
verb( 'deliberate', 'deliberates', 'deliberating', 'deliberated', 'deliberated', _, ['2A','3A','6A','8','10'] ).
verb( 'delight', 'delights', 'delighting', 'delighted', 'delighted', _, ['3A','4C','6A'] ).
verb( 'delimit', 'delimits', 'delimiting', 'delimited', 'delimited', tran, ['6A'] ).
verb( 'delimitate', 'delimitates', 'delimitating', 'delimitated', 'delimitated', tran, ['6A'] ).
verb( 'delineate', 'delineates', 'delineating', 'delineated', 'delineated', tran, ['6A'] ).
verb( 'deliver', 'delivers', 'delivering', 'delivered', 'delivered', tran, ['6A','14','15B'] ).
verb( 'delouse', 'delouses', 'delousing', 'deloused', 'deloused', tran, [] ).
verb( 'delude', 'deludes', 'deluding', 'deluded', 'deluded', tran, ['6A','14'] ).
verb( 'deluge', 'deluges', 'deluging', 'deluged', 'deluged', tran, ['6A','14'] ).
verb( 'delve', 'delves', 'delving', 'delved', 'delved', _, ['2A','3A','6A'] ).
verb( 'demagnetize', 'demagnetizes', 'demagnetizing', 'demagnetized', 'demagnetized', tran, ['6A'] ).
verb( 'demand', 'demands', 'demanding', 'demanded', 'demanded', tran, ['6A','7A','9'] ).
verb( 'demarcate', 'demarcates', 'demarcating', 'demarcated', 'demarcated', tran, ['6A'] ).
verb( 'demean', 'demeans', 'demeaning', 'demeaned', 'demeaned', tran, ['6A'] ).
verb( 'demilitarize', 'demilitarizes', 'demilitarizing', 'demilitarized', 'demilitarized', tran, ['6A'] ).
verb( 'demist', 'demists', 'demisting', 'demisted', 'demisted', tran, [] ).
verb( 'demob', 'demobs', 'demobbing', 'demobbed', 'demobbed', tran, [] ).
verb( 'demobilize', 'demobilizes', 'demobilizing', 'demobilized', 'demobilized', tran, ['6A'] ).
verb( 'democratize', 'democratizes', 'democratizing', 'democratized', 'democratized', tran, ['6A'] ).
verb( 'demolish', 'demolishes', 'demolishing', 'demolished', 'demolished', tran, ['6A'] ).
verb( 'demonetize', 'demonetizes', 'demonetizing', 'demonetized', 'demonetized', tran, [] ).
verb( 'demonstrate', 'demonstrates', 'demonstrating', 'demonstrated', 'demonstrated', _, ['2A','3A','6A','9'] ).
verb( 'demoralize', 'demoralizes', 'demoralizing', 'demoralized', 'demoralized', tran, ['6A'] ).
verb( 'demote', 'demotes', 'demoting', 'demoted', 'demoted', tran, ['6A'] ).
verb( 'demur', 'demurs', 'demurring', 'demurred', 'demurred', intran, ['2A','3A'] ).
verb( 'denationalize', 'denationalizes', 'denationalizing', 'denationalized', 'denationalized', tran, ['6A'] ).
verb( 'denigrate', 'denigrates', 'denigrating', 'denigrated', 'denigrated', tran, ['6A'] ).
verb( 'denominate', 'denominates', 'denominating', 'denominated', 'denominated', tran, ['23'] ).
verb( 'denote', 'denotes', 'denoting', 'denoted', 'denoted', tran, ['6A','9'] ).
verb( 'denounce', 'denounces', 'denouncing', 'denounced', 'denounced', tran, ['6A','16B'] ).
verb( 'dent', 'dents', 'denting', 'dented', 'dented', _, ['2A','6A'] ).
verb( 'denude', 'denudes', 'denuding', 'denuded', 'denuded', tran, ['6A','14'] ).
verb( 'deny', 'denies', 'denying', 'denied', 'denied', tran, ['6A','6C','9','12A','13A','25'] ).
verb( 'deodorize', 'deodorizes', 'deodorizing', 'deodorized', 'deodorized', tran, ['6A'] ).
verb( 'depart', 'departs', 'departing', 'departed', 'departed', intran, ['2A','3A'] ).
verb( 'depend', 'depends', 'depending', 'depended', 'depended', intran, ['3A'] ).
verb( 'depict', 'depicts', 'depicting', 'depicted', 'depicted', tran, ['6A'] ).
verb( 'deplane', 'deplanes', 'deplaning', 'deplaned', 'deplaned', intran, [] ).
verb( 'deplete', 'depletes', 'depleting', 'depleted', 'depleted', tran, ['6A','14'] ).
verb( 'deplore', 'deplores', 'deploring', 'deplored', 'deplored', tran, ['6A'] ).
verb( 'deploy', 'deploys', 'deploying', 'deployed', 'deployed', _, ['2A','6A'] ).
verb( 'depopulate', 'depopulates', 'depopulating', 'depopulated', 'depopulated', tran, ['6A'] ).
verb( 'deport', 'deports', 'deporting', 'deported', 'deported', tran, ['6A','16B'] ).
verb( 'depose', 'deposes', 'deposing', 'deposed', 'deposed', _, ['3A','6A','9'] ).
verb( 'deposit', 'deposits', 'depositing', 'deposited', 'deposited', tran, ['6A'] ).
verb( 'deprave', 'depraves', 'depraving', 'depraved', 'depraved', tran, ['6A'] ).
verb( 'deprecate', 'deprecates', 'deprecating', 'deprecated', 'deprecated', tran, ['6A','6C'] ).
verb( 'depreciate', 'depreciates', 'depreciating', 'depreciated', 'depreciated', _, ['2A','6A'] ).
verb( 'depress', 'depresses', 'depressing', 'depressed', 'depressed', tran, ['6A'] ).
verb( 'deprive', 'deprives', 'depriving', 'deprived', 'deprived', tran, ['14'] ).
verb( 'depute', 'deputes', 'deputing', 'deputed', 'deputed', tran, ['14','17'] ).
verb( 'deputize', 'deputizes', 'deputizing', 'deputized', 'deputized', intran, ['2A','3A'] ).
verb( 'derail', 'derails', 'derailing', 'derailed', 'derailed', tran, ['6A'] ).
verb( 'derange', 'deranges', 'deranging', 'deranged', 'deranged', tran, ['6A'] ).
verb( 'derate', 'derates', 'derating', 'derated', 'derated', tran, ['6A'] ).
verb( 'derequisition', 'derequisitions', 'derequisitioning', 'derequisitioned', 'derequisitioned', tran, ['6A'] ).
verb( 'derestrict', 'derestricts', 'derestricting', 'derestricted', 'derestricted', tran, ['6A'] ).
verb( 'deride', 'derides', 'deriding', 'derided', 'derided', tran, ['6A','16B'] ).
verb( 'derive', 'derives', 'deriving', 'derived', 'derived', _, ['3A','14'] ).
verb( 'derogate', 'derogates', 'derogating', 'derogated', 'derogated', intran, ['3A'] ).
verb( 'desalinate', 'desalinates', 'desalinating', 'desalinated', 'desalinated', tran, ['6A'] ).
verb( 'desalinize', 'desalinizes', 'desalinizing', 'desalinized', 'desalinized', tran, ['6A'] ).
verb( 'desalt', 'desalts', 'desalting', 'desalted', 'desalted', tran, [] ).
verb( 'descale', 'descales', 'descaling', 'descaled', 'descaled', tran, ['6A'] ).
verb( 'descant', 'descants', 'descanting', 'descanted', 'descanted', intran, ['3A'] ).
verb( 'descend', 'descends', 'descending', 'descended', 'descended', _, ['2A','2C','3A','6A'] ).
verb( 'describe', 'describes', 'describing', 'described', 'described', tran, ['6A','10','14','16B'] ).
verb( 'descry', 'descries', 'descrying', 'descried', 'descried', tran, ['6A'] ).
verb( 'desecrate', 'desecrates', 'desecrating', 'desecrated', 'desecrated', tran, ['6A'] ).
verb( 'desegregate', 'desegregates', 'desegregating', 'desegregated', 'desegregated', tran, ['6A'] ).
verb( 'desensitize', 'desensitizes', 'desensitizing', 'desensitized', 'desensitized', tran, ['6A'] ).
verb( 'desert', 'deserts', 'deserting', 'deserted', 'deserted', _, ['2A','6A'] ).
verb( 'deserve', 'deserves', 'deserving', 'deserved', 'deserved', _, ['6A','7A'] ).
verb( 'desiccate', 'desiccates', 'desiccating', 'desiccated', 'desiccated', tran, ['6A'] ).
verb( 'design', 'designs', 'designing', 'designed', 'designed', _, ['2A','2C','6A','14','16A','16B'] ).
verb( 'designate', 'designates', 'designating', 'designated', 'designated', tran, ['6A','16B','17'] ).
verb( 'desire', 'desires', 'desiring', 'desired', 'desired', tran, ['6A','7A','9','17'] ).
verb( 'desist', 'desists', 'desisting', 'desisted', 'desisted', intran, ['2A','3A'] ).
verb( 'desolate', 'desolates', 'desolating', 'desolated', 'desolated', tran, ['6A'] ).
verb( 'despair', 'despairs', 'despairing', 'despaired', 'despaired', intran, ['2A','3A'] ).
verb( 'despatch', 'despatches', 'despatching', 'despatched', 'despatched', tran, ['6A','14'] ).
verb( 'despise', 'despises', 'despising', 'despised', 'despised', tran, ['6A'] ).
verb( 'despoil', 'despoils', 'despoiling', 'despoiled', 'despoiled', tran, ['6A','14'] ).
verb( 'destine', 'destines', 'destining', 'destined', 'destined', tran, ['14','17'] ).
verb( 'destroy', 'destroys', 'destroying', 'destroyed', 'destroyed', tran, ['6A'] ).
verb( 'detach', 'detaches', 'detaching', 'detached', 'detached', tran, ['6A','14','16A'] ).
verb( 'detail', 'details', 'detailing', 'detailed', 'detailed', tran, ['6A','14','16A'] ).
verb( 'detain', 'detains', 'detaining', 'detained', 'detained', tran, ['6A','16A'] ).
verb( 'detect', 'detects', 'detecting', 'detected', 'detected', tran, ['6A'] ).
verb( 'deter', 'deters', 'deterring', 'deterred', 'deterred', tran, ['6A','14'] ).
verb( 'deteriorate', 'deteriorates', 'deteriorating', 'deteriorated', 'deteriorated', _, ['2A','6A'] ).
verb( 'determine', 'determines', 'determining', 'determined', 'determined', _, ['3A','6A','7A','8','9','10','14','17'] ).
verb( 'detest', 'detests', 'detesting', 'detested', 'detested', tran, ['6A','6C'] ).
verb( 'dethrone', 'dethrones', 'dethroning', 'dethroned', 'dethroned', tran, ['6A'] ).
verb( 'detonate', 'detonates', 'detonating', 'detonated', 'detonated', _, ['2A','6A'] ).
verb( 'detour', 'detours', 'detouring', 'detoured', 'detoured', tran, ['6A'] ).
verb( 'detract', 'detracts', 'detracting', 'detracted', 'detracted', intran, ['3A'] ).
verb( 'detrain', 'detrains', 'detraining', 'detrained', 'detrained', _, ['2A','6A'] ).
verb( 'detribalize', 'detribalizes', 'detribalizing', 'detribalized', 'detribalized', tran, ['6A'] ).
verb( 'devaluate', 'devaluates', 'devaluating', 'devaluated', 'devaluated', tran, ['6A'] ).
verb( 'devalue', 'devalues', 'devaluing', 'devalued', 'devalued', tran, ['6A'] ).
verb( 'devastate', 'devastates', 'devastating', 'devastated', 'devastated', tran, ['6A'] ).
verb( 'develop', 'develops', 'developing', 'developed', 'developed', _, ['2A','2C','3A','6A'] ).
verb( 'deviate', 'deviates', 'deviating', 'deviated', 'deviated', intran, ['3A'] ).
verb( 'devil', 'devils', 'devilling', 'devilled', 'devilled', _, ['2A','3A','6A'] ).
verb( 'devise', 'devises', 'devising', 'devised', 'devised', tran, ['6A','8','14'] ).
verb( 'devitalize', 'devitalizes', 'devitalizing', 'devitalized', 'devitalized', tran, ['6A'] ).
verb( 'devolve', 'devolves', 'devolving', 'devolved', 'devolved', _, ['3A','6A','14'] ).
verb( 'devote', 'devotes', 'devoting', 'devoted', 'devoted', tran, ['14'] ).
verb( 'devour', 'devours', 'devouring', 'devoured', 'devoured', tran, ['6A'] ).
verb( 'diagnose', 'diagnoses', 'diagnosing', 'diagnosed', 'diagnosed', tran, ['6A','16B'] ).
verb( 'dial', 'dials', 'dialling', 'dialled', 'dialled', tran, ['6A'] ).
verb( 'dibble', 'dibbles', 'dibbling', 'dibbled', 'dibbled', tran, ['15B'] ).
verb( 'dice', 'dices', 'dicing', 'diced', 'diced', _, ['2A','6A'] ).
verb( 'dicker', 'dickers', 'dickering', 'dickered', 'dickered', intran, ['2A','3A'] ).
verb( 'dictate', 'dictates', 'dictating', 'dictated', 'dictated', _, ['2A','3A','6A','14'] ).
verb( 'diddle', 'diddles', 'diddling', 'diddled', 'diddled', tran, ['6A','14'] ).
verb( 'die', 'dies', 'dying', 'died', 'died', intran, ['2A','2C','2D','3A','4C'] ).
verb( 'diet', 'diets', 'dieting', 'dieted', 'dieted', _, ['2A','6A'] ).
verb( 'differ', 'differs', 'differing', 'differed', 'differed', intran, ['2A','2C','3A'] ).
verb( 'differentiate', 'differentiates', 'differentiating', 'differentiated', 'differentiated', tran, ['3A','6A','14'] ).
verb( 'diffract', 'diffracts', 'diffracting', 'diffracted', 'diffracted', tran, ['6A'] ).
verb( 'diffuse', 'diffuses', 'diffusing', 'diffused', 'diffused', _, ['2A','6A'] ).
verb( 'dig', 'digs', 'digging', 'dug', 'dug', _, ['2C','3A','6A','15B'] ).
verb( 'digest', 'digests', 'digesting', 'digested', 'digested', _, ['2A','6A'] ).
verb( 'dignify', 'dignifies', 'dignifying', 'dignified', 'dignified', tran, ['6A','14'] ).
verb( 'digress', 'digresses', 'digressing', 'digressed', 'digressed', intran, ['2A','3A'] ).
verb( 'dike', 'dikes', 'diking', 'diked', 'diked', _, [] ).
verb( 'dilate', 'dilates', 'dilating', 'dilated', 'dilated', _, ['2A','3A','6A'] ).
verb( 'dilly-dally', 'dilly-dallies', 'dilly-dallying', 'dilly-dallied', 'dilly-dallied', intran, ['2A'] ).
verb( 'dilute', 'dilutes', 'diluting', 'diluted', 'diluted', tran, ['6A','14'] ).
verb( 'dim', 'dims', 'dimming', 'dimmed', 'dimmed', _, ['2A','6A'] ).
verb( 'diminish', 'diminishes', 'diminishing', 'diminished', 'diminished', _, ['2A','6A'] ).
verb( 'dimple', 'dimples', 'dimpling', 'dimpled', 'dimpled', _, ['2A','6A'] ).
verb( 'din', 'dins', 'dinning', 'dinned', 'dinned', _, ['2C'] ).
verb( 'dine', 'dines', 'dining', 'dined', 'dined', _, ['2A','6A'] ).
verb( 'dip', 'dips', 'dipping', 'dipped', 'dipped', _, ['2A','2C','3A','6A','14'] ).
verb( 'direct', 'directs', 'directing', 'directed', 'directed', _, ['2A','6A','9','14','17'] ).
verb( 'dirty', 'dirties', 'dirtying', 'dirtied', 'dirtied', _, ['2A','6A'] ).
verb( 'disable', 'disables', 'disabling', 'disabled', 'disabled', tran, ['6A'] ).
verb( 'disabuse', 'disabuses', 'disabusing', 'disabused', 'disabused', tran, ['6A','14'] ).
verb( 'disafforest', 'disafforests', 'disafforesting', 'disafforested', 'disafforested', tran, [] ).
verb( 'disagree', 'disagrees', 'disagreeing', 'disagreed', 'disagreed', tran, ['2A','3A'] ).
verb( 'disallow', 'disallows', 'disallowing', 'disallowed', 'disallowed', tran, ['6A'] ).
verb( 'disappear', 'disappears', 'disappearing', 'disappeared', 'disappeared', intran, ['2A'] ).
verb( 'disappoint', 'disappoints', 'disappointing', 'disappointed', 'disappointed', tran, ['6A'] ).
verb( 'disapprove', 'disapproves', 'disapproving', 'disapproved', 'disapproved', _, ['2A','3A','6A'] ).
verb( 'disarm', 'disarms', 'disarming', 'disarmed', 'disarmed', _, ['2A','6A'] ).
verb( 'disarrange', 'disarranges', 'disarranging', 'disarranged', 'disarranged', tran, ['6A'] ).
verb( 'disarray', 'disarrays', 'disarraying', 'disarrayed', 'disarrayed', tran, [] ).
verb( 'disassociate', 'disassociates', 'disassociating', 'disassociated', 'disassociated', tran, ['14'] ).
verb( 'disavow', 'disavows', 'disavowing', 'disavowed', 'disavowed', tran, ['6A'] ).
verb( 'disband', 'disbands', 'disbanding', 'disbanded', 'disbanded', _, ['2A','6A'] ).
verb( 'disbelieve', 'disbelieves', 'disbelieving', 'disbelieved', 'disbelieved', _, ['2A','3A','6A'] ).
verb( 'disbud', 'disbuds', 'disbudding', 'disbudded', 'disbudded', tran, ['6A'] ).
verb( 'disburden', 'disburdens', 'disburdening', 'disburdened', 'disburdened', tran, ['6A','14'] ).
verb( 'disburse', 'disburses', 'disbursing', 'disbursed', 'disbursed', _, ['2A','6A'] ).
verb( 'discard', 'discards', 'discarding', 'discarded', 'discarded', tran, ['6A'] ).
verb( 'discern', 'discerns', 'discerning', 'discerned', 'discerned', tran, ['6A'] ).
verb( 'discharge', 'discharges', 'discharging', 'discharged', 'discharged', _, ['6A','14'] ).
verb( 'discipline', 'disciplines', 'disciplining', 'disciplined', 'disciplined', tran, ['6A'] ).
verb( 'disclaim', 'disclaims', 'disclaiming', 'disclaimed', 'disclaimed', tran, ['6A','6C'] ).
verb( 'disclose', 'discloses', 'disclosing', 'disclosed', 'disclosed', tran, ['6A','14'] ).
verb( 'discolour', 'discolours', 'discolouring', 'discoloured', 'discoloured', _, ['2A','6A'] ).
verb( 'discomfit', 'discomfits', 'discomfiting', 'discomfited', 'discomfited', tran, ['6A'] ).
verb( 'discommode', 'discommodes', 'discommoding', 'discommoded', 'discommoded', tran, ['6A'] ).
verb( 'discompose', 'discomposes', 'discomposing', 'discomposed', 'discomposed', tran, ['6A'] ).
verb( 'disconcert', 'disconcerts', 'disconcerting', 'disconcerted', 'disconcerted', tran, ['6A'] ).
verb( 'disconnect', 'disconnects', 'disconnecting', 'disconnected', 'disconnected', tran, ['6A','14'] ).
verb( 'discontent', 'discontents', 'discontenting', 'discontented', 'discontented', tran, ['6A'] ).
verb( 'discontinue', 'discontinues', 'discontinuing', 'discontinued', 'discontinued', _, ['2A','6A','6C'] ).
verb( 'discount', 'discounts', 'discounting', 'discounted', 'discounted', tran, ['6A'] ).
verb( 'discountenance', 'discountenances', 'discountenancing', 'discountenanced', 'discountenanced', tran, ['6A'] ).
verb( 'discourage', 'discourages', 'discouraging', 'discouraged', 'discouraged', tran, ['6A','14'] ).
verb( 'discourse', 'discourses', 'discoursing', 'discoursed', 'discoursed', intran, [] ).
verb( 'discover', 'discovers', 'discovering', 'discovered', 'discovered', tran, ['6A','8','9','10','25'] ).
verb( 'discredit', 'discredits', 'discrediting', 'discredited', 'discredited', tran, ['6A'] ).
verb( 'discriminate', 'discriminates', 'discriminating', 'discriminated', 'discriminated', _, ['2A','3A','14'] ).
verb( 'discuss', 'discusses', 'discussing', 'discussed', 'discussed', tran, ['6A','8','10','14'] ).
verb( 'disdain', 'disdains', 'disdaining', 'disdained', 'disdained', tran, ['6A','6C','7A'] ).
verb( 'disembark', 'disembarks', 'disembarking', 'disembarked', 'disembarked', _, ['2A','2C','6A','14'] ).
verb( 'disembarrass', 'disembarrasses', 'disembarrassing', 'disembarrassed', 'disembarrassed', tran, ['14'] ).
verb( 'disembody', 'disembodies', 'disembodying', 'disembodied', 'disembodied', tran, ['6A'] ).
verb( 'disembowel', 'disembowels', 'disembowelling', 'disembowelled', 'disembowelled', tran, ['6A'] ).
verb( 'disenchant', 'disenchants', 'disenchanting', 'disenchanted', 'disenchanted', tran, ['6A'] ).
verb( 'disencumber', 'disencumbers', 'disencumbering', 'disencumbered', 'disencumbered', tran, ['6A','14'] ).
verb( 'disenfranchise', 'disenfranchises', 'disenfranchising', 'disenfranchised', 'disenfranchised', tran, [] ).
verb( 'disengage', 'disengages', 'disengaging', 'disengaged', 'disengaged', _, ['2A','2C','6A','14'] ).
verb( 'disentangle', 'disentangles', 'disentangling', 'disentangled', 'disentangled', _, ['2A','6A','14'] ).
verb( 'disestablish', 'disestablishes', 'disestablishing', 'disestablished', 'disestablished', tran, ['6A'] ).
verb( 'disfavour', 'disfavours', 'disfavouring', 'disfavoured', 'disfavoured', tran, ['6A'] ).
verb( 'disfigure', 'disfigures', 'disfiguring', 'disfigured', 'disfigured', tran, ['6A'] ).
verb( 'disforest', 'disforests', 'disforesting', 'disforested', 'disforested', tran, ['6A'] ).
verb( 'disfranchise', 'disfranchises', 'disfranchising', 'disfranchised', 'disfranchised', tran, ['6A'] ).
verb( 'disgorge', 'disgorges', 'disgorging', 'disgorged', 'disgorged', tran, ['6A'] ).
verb( 'disgrace', 'disgraces', 'disgracing', 'disgraced', 'disgraced', tran, ['6A'] ).
verb( 'disguise', 'disguises', 'disguising', 'disguised', 'disguised', tran, ['6A','16B'] ).
verb( 'disgust', 'disgusts', 'disgusting', 'disgusted', 'disgusted', tran, ['6A'] ).
verb( 'dish', 'dishes', 'dishing', 'dished', 'dished', tran, ['6A','15B'] ).
verb( 'dishearten', 'disheartens', 'disheartening', 'disheartened', 'disheartened', tran, ['6A'] ).
verb( 'dishonour', 'dishonours', 'dishonouring', 'dishonoured', 'dishonoured', tran, ['6A'] ).
verb( 'disillusion', 'disillusions', 'disillusioning', 'disillusioned', 'disillusioned', tran, ['6A'] ).
verb( 'disincline', 'disinclines', 'disinclining', 'disinclined', 'disinclined', tran, ['14','17'] ).
verb( 'disinfect', 'disinfects', 'disinfecting', 'disinfected', 'disinfected', tran, ['6A'] ).
verb( 'disinfest', 'disinfests', 'disinfesting', 'disinfested', 'disinfested', tran, ['6A'] ).
verb( 'disinherit', 'disinherits', 'disinheriting', 'disinherited', 'disinherited', tran, ['6A'] ).
verb( 'disintegrate', 'disintegrates', 'disintegrating', 'disintegrated', 'disintegrated', _, ['2A','6A'] ).
verb( 'disinter', 'disinters', 'disinterring', 'disinterred', 'disinterred', tran, ['6A'] ).
verb( 'disjoint', 'disjoints', 'disjointing', 'disjointed', 'disjointed', tran, ['6A'] ).
verb( 'dislike', 'dislikes', 'disliking', 'disliked', 'disliked', tran, ['6A','6C'] ).
verb( 'dislocate', 'dislocates', 'dislocating', 'dislocated', 'dislocated', tran, ['6A'] ).
verb( 'dislodge', 'dislodges', 'dislodging', 'dislodged', 'dislodged', tran, ['6A','14'] ).
verb( 'dismantle', 'dismantles', 'dismantling', 'dismantled', 'dismantled', tran, ['6A'] ).
verb( 'dismay', 'dismays', 'dismaying', 'dismayed', 'dismayed', tran, ['6A'] ).
verb( 'dismember', 'dismembers', 'dismembering', 'dismembered', 'dismembered', tran, ['6A'] ).
verb( 'dismiss', 'dismisses', 'dismissing', 'dismissed', 'dismissed', tran, ['6A','14'] ).
verb( 'dismount', 'dismounts', 'dismounting', 'dismounted', 'dismounted', _, ['2A','3A','6A'] ).
verb( 'disobey', 'disobeys', 'disobeying', 'disobeyed', 'disobeyed', tran, ['2A','6A'] ).
verb( 'disoblige', 'disobliges', 'disobliging', 'disobliged', 'disobliged', tran, ['6A'] ).
verb( 'disorder', 'disorders', 'disordering', 'disordered', 'disordered', tran, ['6A'] ).
verb( 'disorganize', 'disorganizes', 'disorganizing', 'disorganized', 'disorganized', tran, ['6A'] ).
verb( 'disorient', 'disorients', 'disorienting', 'disoriented', 'disoriented', tran, ['6A'] ).
verb( 'disorientate', 'disorientates', 'disorientating', 'disorientated', 'disorientated', tran, ['6A'] ).
verb( 'disown', 'disowns', 'disowning', 'disowned', 'disowned', tran, ['6A'] ).
verb( 'disparage', 'disparages', 'disparaging', 'disparaged', 'disparaged', tran, ['6A'] ).
verb( 'dispatch', 'dispatches', 'dispatching', 'dispatched', 'dispatched', tran, ['6A','14'] ).
verb( 'dispel', 'dispels', 'dispelling', 'dispelled', 'dispelled', tran, ['6A'] ).
verb( 'dispense', 'dispenses', 'dispensing', 'dispensed', 'dispensed', _, ['3A','6A','14'] ).
verb( 'disperse', 'disperses', 'dispersing', 'dispersed', 'dispersed', _, ['2A','6A'] ).
verb( 'dispirit', 'dispirits', 'dispiriting', 'dispirited', 'dispirited', tran, ['6A'] ).
verb( 'displace', 'displaces', 'displacing', 'displaced', 'displaced', tran, ['6A'] ).
verb( 'display', 'displays', 'displaying', 'displayed', 'displayed', tran, ['6A'] ).
verb( 'displease', 'displeases', 'displeasing', 'displeased', 'displeased', tran, ['6A'] ).
verb( 'disport', 'disports', 'disporting', 'disported', 'disported', tran, ['6A'] ).
verb( 'dispose', 'disposes', 'disposing', 'disposed', 'disposed', _, ['2A','3A','6A','17'] ).
verb( 'dispossess', 'dispossesses', 'dispossessing', 'dispossessed', 'dispossessed', tran, ['14'] ).
verb( 'disprove', 'disproves', 'disproving', 'disproved', 'disproved', tran, ['6A'] ).
verb( 'dispute', 'disputes', 'disputing', 'disputed', 'disputed', _, ['2A','3A','6A','8','10'] ).
verb( 'disqualify', 'disqualifies', 'disqualifying', 'disqualified', 'disqualified', tran, ['6A','14'] ).
verb( 'disquiet', 'disquiets', 'disquieting', 'disquieted', 'disquieted', tran, ['6A'] ).
verb( 'disregard', 'disregards', 'disregarding', 'disregarded', 'disregarded', tran, ['6A'] ).
verb( 'disrobe', 'disrobes', 'disrobing', 'disrobed', 'disrobed', _, ['2A','6A'] ).
verb( 'disrupt', 'disrupts', 'disrupting', 'disrupted', 'disrupted', tran, ['6A'] ).
verb( 'dissatisfy', 'dissatisfies', 'dissatisfying', 'dissatisfied', 'dissatisfied', tran, ['6A'] ).
verb( 'dissect', 'dissects', 'dissecting', 'dissected', 'dissected', tran, ['6A'] ).
verb( 'dissemble', 'dissembles', 'dissembling', 'dissembled', 'dissembled', _, ['2A','6A'] ).
verb( 'disseminate', 'disseminates', 'disseminating', 'disseminated', 'disseminated', tran, ['6A'] ).
verb( 'dissent', 'dissents', 'dissenting', 'dissented', 'dissented', intran, ['2A','3A'] ).
verb( 'dissever', 'dissevers', 'dissevering', 'dissevered', 'dissevered', tran, ['6A'] ).
verb( 'dissimulate', 'dissimulates', 'dissimulating', 'dissimulated', 'dissimulated', _, ['2A','6A'] ).
verb( 'dissipate', 'dissipates', 'dissipating', 'dissipated', 'dissipated', _, ['2A','6A'] ).
verb( 'dissociate', 'dissociates', 'dissociating', 'dissociated', 'dissociated', tran, ['6A','14'] ).
verb( 'dissolve', 'dissolves', 'dissolving', 'dissolved', 'dissolved', _, ['2A','2C','3A','6A','14'] ).
verb( 'dissuade', 'dissuades', 'dissuading', 'dissuaded', 'dissuaded', tran, ['6A','14'] ).
verb( 'distance', 'distances', 'distancing', 'distanced', 'distanced', tran, ['6A','14'] ).
verb( 'distemper', 'distempers', 'distempering', 'distempered', 'distempered', tran, ['6A','22'] ).
verb( 'distend', 'distends', 'distending', 'distended', 'distended', _, ['2A','6A'] ).
verb( 'distil', 'distils', 'distilling', 'distilled', 'distilled', _, ['2A','6A','14','15B'] ).
verb( 'distinguish', 'distinguishes', 'distinguishing', 'distinguished', 'distinguished', _, ['3A','6A','14'] ).
verb( 'distort', 'distorts', 'distorting', 'distorted', 'distorted', tran, ['6A'] ).
verb( 'distract', 'distracts', 'distracting', 'distracted', 'distracted', tran, ['6A','14'] ).
verb( 'distrain', 'distrains', 'distraining', 'distrained', 'distrained', intran, ['2A','3A'] ).
verb( 'distress', 'distresses', 'distressing', 'distressed', 'distressed', tran, ['6A'] ).
verb( 'distribute', 'distributes', 'distributing', 'distributed', 'distributed', tran, ['6A','14'] ).
verb( 'distrust', 'distrusts', 'distrusting', 'distrusted', 'distrusted', tran, ['6A'] ).
verb( 'disturb', 'disturbs', 'disturbing', 'disturbed', 'disturbed', tran, ['6A'] ).
verb( 'disunite', 'disunites', 'disuniting', 'disunited', 'disunited', _, ['2A','6A'] ).
verb( 'ditch', 'ditches', 'ditching', 'ditched', 'ditched', _, ['2A','6A'] ).
verb( 'dither', 'dithers', 'dithering', 'dithered', 'dithered', intran, ['2A','2C'] ).
verb( 'divagate', 'divagates', 'divagating', 'divagated', 'divagated', intran, ['2A','3A'] ).
verb( 'dive', 'dives', 'diving', 'dived', 'dived', intran, ['2A','2C'] ).
verb( 'dive-bomb', 'dive-bombs', 'dive-bombing', 'dive-bombed', 'dive-bombed', _, [] ).
verb( 'diverge', 'diverges', 'diverging', 'diverged', 'diverged', intran, ['2A','3A'] ).
verb( 'diversify', 'diversifies', 'diversifying', 'diversified', 'diversified', tran, ['6A'] ).
verb( 'divert', 'diverts', 'diverting', 'diverted', 'diverted', tran, ['6A','14'] ).
verb( 'divest', 'divests', 'divesting', 'divested', 'divested', tran, ['14'] ).
verb( 'divide', 'divides', 'dividing', 'divided', 'divided', _, ['2A','2C','6A','14','15B'] ).
verb( 'divine', 'divines', 'divining', 'divined', 'divined', _, ['2A','6A','10'] ).
verb( 'divorce', 'divorces', 'divorcing', 'divorced', 'divorced', tran, ['6A','14'] ).
verb( 'divulge', 'divulges', 'divulging', 'divulged', 'divulged', tran, ['6A','14'] ).
verb( 'dizzy', 'dizzies', 'dizzying', 'dizzied', 'dizzied', tran, [] ).
verb( 'do', 'does', 'doing', 'did', 'done', unknown, ['2A','2B','2C','3A','6A','6C','7B','12B','13B','15A','15B'] ).
verb( 'do', 'does', 'doing', 'did', 'done', _, ['2A','2B','2C','3A','6A','6C','7B','12B','13B','15A','15B'] ).
verb( 'dock', 'docks', 'docking', 'docked', 'docked', _, ['2A','6A','14'] ).
verb( 'docket', 'dockets', 'docketing', 'docketed', 'docketed', tran, ['6A'] ).
verb( 'doctor', 'doctors', 'doctoring', 'doctored', 'doctored', tran, ['6A'] ).
verb( 'document', 'documents', 'documenting', 'documented', 'documented', tran, ['6A'] ).
verb( 'dodder', 'dodders', 'doddering', 'doddered', 'doddered', intran, ['2A','2C'] ).
verb( 'dodge', 'dodges', 'dodging', 'dodged', 'dodged', _, ['2A','3A','6A'] ).
verb( 'doff', 'doffs', 'doffing', 'doffed', 'doffed', tran, ['6A'] ).
verb( 'dog', 'dogs', 'dogging', 'dogged', 'dogged', tran, ['6A'] ).
verb( 'dogmatize', 'dogmatizes', 'dogmatizing', 'dogmatized', 'dogmatized', _, ['2A','6A'] ).
verb( 'dole', 'doles', 'doling', 'doled', 'doled', tran, ['15B'] ).
verb( 'doll', 'dolls', 'dolling', 'dolled', 'dolled', _, ['2C','15B'] ).
verb( 'domesticate', 'domesticates', 'domesticating', 'domesticated', 'domesticated', tran, ['6A'] ).
verb( 'dominate', 'dominates', 'dominating', 'dominated', 'dominated', _, ['2A','3A','6A'] ).
verb( 'domineer', 'domineers', 'domineering', 'domineered', 'domineered', intran, ['2A','3A'] ).
verb( 'don', 'dons', 'donning', 'donned', 'donned', tran, ['6A'] ).
verb( 'donate', 'donates', 'donating', 'donated', 'donated', tran, ['6A','14'] ).
verb( 'doodle', 'doodles', 'doodling', 'doodled', 'doodled', intran, ['2A'] ).
verb( 'doom', 'dooms', 'dooming', 'doomed', 'doomed', tran, ['6A','14','17'] ).
verb( 'dope', 'dopes', 'doping', 'doped', 'doped', tran, ['6A'] ).
verb( 'dose', 'doses', 'dosing', 'dosed', 'dosed', tran, ['6A','14','15B'] ).
verb( 'doss', 'dosses', 'dossing', 'dossed', 'dossed', intran, ['2C'] ).
verb( 'dot', 'dots', 'dotting', 'dotted', 'dotted', tran, [] ).
verb( 'dote', 'dotes', 'doting', 'doted', 'doted', intran, ['3A'] ).
verb( 'double', 'doubles', 'doubling', 'doubled', 'doubled', _, ['2A','2C','6A','15B'] ).
verb( 'double-check', 'double-checks', 'double-checking', 'double-checked', 'double-checked', tran, [] ).
verb( 'double-cross', 'double-crosses', 'double-crossing', 'double-crossed', 'double-crossed', tran, ['6A'] ).
verb( 'double-park', 'double-parks', 'double-parking', 'double-parked', 'double-parked', _, [] ).
verb( 'doubt', 'doubts', 'doubting', 'doubted', 'doubted', tran, ['6A','9','10'] ).
verb( 'douse', 'douses', 'dousing', 'doused', 'doused', tran, ['6A'] ).
verb( 'dovetail', 'dovetails', 'dovetailing', 'dovetailed', 'dovetailed', _, ['2A','3A','6A'] ).
verb( 'dower', 'dowers', 'dowering', 'dowered', 'dowered', tran, [] ).
verb( 'down', 'downs', 'downing', 'downed', 'downed', tran, ['6A'] ).
verb( 'downgrade', 'downgrades', 'downgrading', 'downgraded', 'downgraded', tran, ['6A'] ).
verb( 'dowse', 'dowses', 'dowsing', 'dowsed', 'dowsed', tran, ['6A'] ).
verb( 'doze', 'dozes', 'dozing', 'dozed', 'dozed', intran, ['2A','2C'] ).
verb( 'draft', 'drafts', 'drafting', 'drafted', 'drafted', tran, ['6A'] ).
verb( 'drag', 'drags', 'dragging', 'dragged', 'dragged', _, ['2A','2C','6A','14','15B'] ).
verb( 'dragoon', 'dragoons', 'dragooning', 'dragooned', 'dragooned', tran, ['6A','14'] ).
verb( 'drain', 'drains', 'draining', 'drained', 'drained', _, ['2A','2C','6A','14','15B','22'] ).
verb( 'dramatize', 'dramatizes', 'dramatizing', 'dramatized', 'dramatized', tran, ['6A'] ).
verb( 'drape', 'drapes', 'draping', 'draped', 'draped', tran, ['6A','14'] ).
verb( 'drat', 'drats', 'dratting', 'dratted', 'dratted', tran, [] ).
verb( 'draught', 'draughts', 'draughting', 'draughted', 'draughted', tran, [] ).
verb( 'draw', 'draws', 'drawing', 'drew', 'drawn', _, ['2A','2B','2C','2D','3A','6A','14','15B'] ).
verb( 'drawl', 'drawls', 'drawling', 'drawled', 'drawled', _, ['2A','2C','6A','15B'] ).
verb( 'dread', 'dreads', 'dreading', 'dreaded', 'dreaded', _, ['6A','6C','7A'] ).
verb( 'dream', 'dreams', 'dreaming', 'dreamed', 'dreamed', _, ['2A','3A','6A','8','9','10','15B'] ).
verb( 'dredge', 'dredges', 'dredging', 'dredged', 'dredged', _, ['2A','3A','6A','14','15B'] ).
verb( 'drench', 'drenches', 'drenching', 'drenched', 'drenched', tran, ['6A'] ).
verb( 'dress', 'dresses', 'dressing', 'dressed', 'dressed', _, ['2A','2C','6A','15B'] ).
verb( 'dribble', 'dribbles', 'dribbling', 'dribbled', 'dribbled', _, ['2A','6A'] ).
verb( 'drift', 'drifts', 'drifting', 'drifted', 'drifted', _, ['2A','2C','6A','14','15B'] ).
verb( 'drill', 'drills', 'drilling', 'drilled', 'drilled', _, ['2A','6A','14'] ).
verb( 'drink', 'drinks', 'drinking', 'drank', 'drunk', _, ['2A','3A','6A','15B'] ).
verb( 'drip', 'drips', 'dripping', 'dripped', 'dripped', _, ['2A','2C','6A'] ).
verb( 'drip-dry', 'drip-dries', 'drip-drying', 'drip-dried', 'drip-dried', tran, [] ).
verb( 'drive', 'drives', 'driving', 'drove', 'driven', _, ['2A','2C','3A','6A','14','15B','17','22'] ).
verb( 'drivel', 'drivels', 'drivelling', 'drivelled', 'drivelled', intran, ['2A','2C'] ).
verb( 'drizzle', 'drizzles', 'drizzling', 'drizzled', 'drizzled', intran, ['2A'] ).
verb( 'drone', 'drones', 'droning', 'droned', 'droned', _, ['2C','15B'] ).
verb( 'drool', 'drools', 'drooling', 'drooled', 'drooled', intran, [] ).
verb( 'droop', 'droops', 'drooping', 'drooped', 'drooped', _, ['2A','2C','6A'] ).
verb( 'drop', 'drops', 'dropping', 'dropped', 'dropped', _, ['2A','2C','3B','6A','12A','13A','14','15A','15B'] ).
verb( 'drown', 'drowns', 'drowning', 'drowned', 'drowned', _, ['2A','6A','15B'] ).
verb( 'drowse', 'drowses', 'drowsing', 'drowsed', 'drowsed', _, ['2A','2C','15B'] ).
verb( 'drub', 'drubs', 'drubbing', 'drubbed', 'drubbed', tran, ['6A','14'] ).
verb( 'drudge', 'drudges', 'drudging', 'drudged', 'drudged', intran, ['2A','2C','3A'] ).
verb( 'drug', 'drugs', 'drugging', 'drugged', 'drugged', tran, ['6A'] ).
verb( 'drum', 'drums', 'drumming', 'drummed', 'drummed', _, ['2A','2C','3A','6A','14','15B'] ).
verb( 'dry', 'dries', 'drying', 'dried', 'dried', _, ['2A','2C','6A','15B'] ).
verb( 'dry-clean', 'dry-cleans', 'dry-cleaning', 'dry-cleaned', 'dry-cleaned', tran, [] ).
verb( 'dub', 'dubs', 'dubbing', 'dubbed', 'dubbed', tran, ['6A','22','23'] ).
verb( 'duck', 'ducks', 'ducking', 'ducked', 'ducked', _, ['2A','6A'] ).
verb( 'duel', 'duels', 'duelling', 'duelled', 'duelled', intran, [] ).
verb( 'dull', 'dulls', 'dulling', 'dulled', 'dulled', _, ['2A','6A'] ).
verb( 'dumbfound', 'dumbfounds', 'dumbfounding', 'dumbfounded', 'dumbfounded', tran, ['6A'] ).
verb( 'dump', 'dumps', 'dumping', 'dumped', 'dumped', tran, ['6A','15A'] ).
verb( 'dun', 'duns', 'dunning', 'dunned', 'dunned', tran, [] ).
verb( 'dunk', 'dunks', 'dunking', 'dunked', 'dunked', tran, ['6A','14'] ).
verb( 'dupe', 'dupes', 'duping', 'duped', 'duped', tran, ['6A'] ).
verb( 'duplicate', 'duplicates', 'duplicating', 'duplicated', 'duplicated', tran, ['6A'] ).
verb( 'dust', 'dusts', 'dusting', 'dusted', 'dusted', tran, ['6A'] ).
verb( 'dwarf', 'dwarfs', 'dwarfing', 'dwarfed', 'dwarfed', tran, ['6A'] ).
verb( 'dwell', 'dwells', 'dwelling', 'dwelt', 'dwelt', intran, ['3A'] ).
verb( 'dwindle', 'dwindles', 'dwindling', 'dwindled', 'dwindled', intran, ['2A'] ).
verb( 'dye', 'dyes', 'dying', 'dyed', 'dyed', _, ['2A','6A','22'] ).
verb( 'dyke', 'dykes', 'dyking', 'dyked', 'dyked', _, [] ).
verb( 'dynamite', 'dynamites', 'dynamiting', 'dynamited', 'dynamited', tran, ['6A'] ).
verb( 'earmark', 'earmarks', 'earmarking', 'earmarked', 'earmarked', tran, ['6A','14'] ).
verb( 'earn', 'earns', 'earning', 'earned', 'earned', tran, ['6A','12B','13B'] ).
verb( 'earth', 'earths', 'earthing', 'earthed', 'earthed', tran, ['6A','15B'] ).
verb( 'ease', 'eases', 'easing', 'eased', 'eased', _, ['2A','2C','6A','14','15A','15B'] ).
verb( 'eat', 'eats', 'eating', 'ate', 'eaten', _, ['2A','2C','3A','4A','6A','15B'] ).
verb( 'eavesdrop', 'eavesdrops', 'eavesdropping', 'eavesdropped', 'eavesdropped', intran, ['2A','3A'] ).
verb( 'ebb', 'ebbs', 'ebbing', 'ebbed', 'ebbed', intran, ['2A','2C'] ).
verb( 'echo', 'echos', 'echoing', 'echoed', 'echoed', _, ['2A','2C','6A','15B'] ).
verb( 'eclipse', 'eclipses', 'eclipsing', 'eclipsed', 'eclipsed', tran, ['6A'] ).
verb( 'economize', 'economizes', 'economizing', 'economized', 'economized', _, ['2A','3A','6A'] ).
verb( 'eddy', 'eddies', 'eddying', 'eddied', 'eddied', intran, ['2A','2C'] ).
verb( 'edge', 'edges', 'edging', 'edged', 'edged', _, ['2C','6A','14','15A','15B'] ).
verb( 'edify', 'edifies', 'edifying', 'edified', 'edified', tran, ['6A'] ).
verb( 'edit', 'edits', 'editing', 'edited', 'edited', tran, ['6A'] ).
verb( 'educate', 'educates', 'educating', 'educated', 'educated', tran, ['6A','15A','16A'] ).
verb( 'educe', 'educes', 'educing', 'educed', 'educed', tran, ['6A'] ).
verb( 'eff', 'effs', 'effing', 'effed', 'effed', intran, [] ).
verb( 'efface', 'effaces', 'effacing', 'effaced', 'effaced', tran, ['6A'] ).
verb( 'effect', 'effects', 'effecting', 'effected', 'effected', tran, ['6A'] ).
verb( 'effervesce', 'effervesces', 'effervescing', 'effervesced', 'effervesced', intran, ['2A'] ).
verb( 'egg', 'eggs', 'egging', 'egged', 'egged', tran, ['15B'] ).
verb( 'egotrip', 'egotrips', 'egotripping', 'egotripped', 'egotripped', intran, [] ).
verb( 'ejaculate', 'ejaculates', 'ejaculating', 'ejaculated', 'ejaculated', tran, ['6A'] ).
verb( 'eject', 'ejects', 'ejecting', 'ejected', 'ejected', _, ['2A','6A','14'] ).
verb( 'eke', 'ekes', 'eking', 'eked', 'eked', tran, ['15B'] ).
verb( 'elaborate', 'elaborates', 'elaborating', 'elaborated', 'elaborated', tran, ['6A'] ).
verb( 'elapse', 'elapses', 'elapsing', 'elapsed', 'elapsed', intran, ['2A'] ).
verb( 'elate', 'elates', 'elating', 'elated', 'elated', tran, ['6A'] ).
verb( 'elbow', 'elbows', 'elbowing', 'elbowed', 'elbowed', tran, ['6A','14','15B'] ).
verb( 'elect', 'elects', 'electing', 'elected', 'elected', tran, ['6A','7A','14','23','25'] ).
verb( 'electrify', 'electrifies', 'electrifying', 'electrified', 'electrified', tran, ['6A'] ).
verb( 'electrocute', 'electrocutes', 'electrocuting', 'electrocuted', 'electrocuted', tran, ['6A'] ).
verb( 'electroplate', 'electroplates', 'electroplating', 'electroplated', 'electroplated', tran, [] ).
verb( 'elevate', 'elevates', 'elevating', 'elevated', 'elevated', tran, ['6A','14'] ).
verb( 'elicit', 'elicits', 'eliciting', 'elicited', 'elicited', tran, ['6A','14'] ).
verb( 'elide', 'elides', 'eliding', 'elided', 'elided', tran, ['6A'] ).
verb( 'eliminate', 'eliminates', 'eliminating', 'eliminated', 'eliminated', tran, ['6A','14'] ).
verb( 'elongate', 'elongates', 'elongating', 'elongated', 'elongated', _, ['2A','6A'] ).
verb( 'elope', 'elopes', 'eloping', 'eloped', 'eloped', intran, ['2A','2C','3A'] ).
verb( 'elucidate', 'elucidates', 'elucidating', 'elucidated', 'elucidated', tran, ['6A'] ).
verb( 'elude', 'eludes', 'eluding', 'eluded', 'eluded', tran, ['6A'] ).
verb( 'emaciate', 'emaciates', 'emaciating', 'emaciated', 'emaciated', tran, ['6A'] ).
verb( 'emanate', 'emanates', 'emanating', 'emanated', 'emanated', intran, ['3A'] ).
verb( 'emancipate', 'emancipates', 'emancipating', 'emancipated', 'emancipated', tran, ['6A','14'] ).
verb( 'emasculate', 'emasculates', 'emasculating', 'emasculated', 'emasculated', tran, ['6A'] ).
verb( 'embalm', 'embalms', 'embalming', 'embalmed', 'embalmed', tran, ['6A'] ).
verb( 'embargo', 'embargos', 'embargoing', 'embargoed', 'embargoed', tran, ['6A'] ).
verb( 'embark', 'embarks', 'embarking', 'embarked', 'embarked', _, ['2A','2C','3A','6A'] ).
verb( 'embarrass', 'embarrasses', 'embarrassing', 'embarrassed', 'embarrassed', tran, ['6A'] ).
verb( 'embed', 'embeds', 'embedding', 'embedded', 'embedded', tran, ['6A','14'] ).
verb( 'embellish', 'embellishes', 'embellishing', 'embellished', 'embellished', tran, ['6A','14'] ).
verb( 'embezzle', 'embezzles', 'embezzling', 'embezzled', 'embezzled', tran, ['6A'] ).
verb( 'embitter', 'embitters', 'embittering', 'embittered', 'embittered', tran, ['6A'] ).
verb( 'emblazon', 'emblazons', 'emblazoning', 'emblazoned', 'emblazoned', tran, ['6A','14'] ).
verb( 'embody', 'embodies', 'embodying', 'embodied', 'embodied', tran, ['6A','14'] ).
verb( 'embolden', 'emboldens', 'emboldening', 'emboldened', 'emboldened', tran, ['6A','17'] ).
verb( 'emboss', 'embosses', 'embossing', 'embossed', 'embossed', tran, ['6A','14'] ).
verb( 'embrace', 'embraces', 'embracing', 'embraced', 'embraced', _, ['2A','6A'] ).
verb( 'embroider', 'embroiders', 'embroidering', 'embroidered', 'embroidered', _, ['2A','6A'] ).
verb( 'embroil', 'embroils', 'embroiling', 'embroiled', 'embroiled', tran, ['6A','14'] ).
verb( 'emend', 'emends', 'emending', 'emended', 'emended', tran, ['6A'] ).
verb( 'emerge', 'emerges', 'emerging', 'emerged', 'emerged', intran, ['2A','3A'] ).
verb( 'emigrate', 'emigrates', 'emigrating', 'emigrated', 'emigrated', intran, ['2A','3A'] ).
verb( 'emit', 'emits', 'emitting', 'emitted', 'emitted', tran, ['6A'] ).
verb( 'empale', 'empales', 'empaling', 'empaled', 'empaled', tran, [] ).
verb( 'empanel', 'empanels', 'empanelling', 'empanelled', 'empanelled', tran, ['6A'] ).
verb( 'emphasize', 'emphasizes', 'emphasizing', 'emphasized', 'emphasized', tran, ['6A'] ).
verb( 'emplane', 'emplanes', 'emplaning', 'emplaned', 'emplaned', _, ['2A','6A'] ).
verb( 'employ', 'employs', 'employing', 'employed', 'employed', tran, ['6A','14','16B'] ).
verb( 'empower', 'empowers', 'empowering', 'empowered', 'empowered', tran, ['17'] ).
verb( 'empty', 'empties', 'emptying', 'emptied', 'emptied', _, ['2A','2C','3A','6A','15B'] ).
verb( 'emulate', 'emulates', 'emulating', 'emulated', 'emulated', tran, ['6A'] ).
verb( 'emulsify', 'emulsifies', 'emulsifying', 'emulsified', 'emulsified', tran, ['6A'] ).
verb( 'enable', 'enables', 'enabling', 'enabled', 'enabled', tran, ['17'] ).
verb( 'enact', 'enacts', 'enacting', 'enacted', 'enacted', tran, ['6A','9'] ).
verb( 'enamel', 'enamels', 'enamelling', 'enamelled', 'enamelled', tran, [] ).
verb( 'enamour', 'enamours', 'enamouring', 'enamoured', 'enamoured', tran, ['6A'] ).
verb( 'encamp', 'encamps', 'encamping', 'encamped', 'encamped', _, ['2A','6A'] ).
verb( 'encase', 'encases', 'encasing', 'encased', 'encased', tran, ['6A','14'] ).
verb( 'enchain', 'enchains', 'enchaining', 'enchained', 'enchained', tran, ['6A'] ).
verb( 'enchant', 'enchants', 'enchanting', 'enchanted', 'enchanted', tran, ['6A'] ).
verb( 'encircle', 'encircles', 'encircling', 'encircled', 'encircled', tran, ['6A'] ).
verb( 'enclose', 'encloses', 'enclosing', 'enclosed', 'enclosed', tran, ['6A','14'] ).
verb( 'encode', 'encodes', 'encoding', 'encoded', 'encoded', tran, ['6A'] ).
verb( 'encompass', 'encompasses', 'encompassing', 'encompassed', 'encompassed', tran, ['6A'] ).
verb( 'encore', 'encores', 'encoring', 'encored', 'encored', tran, [] ).
verb( 'encounter', 'encounters', 'encountering', 'encountered', 'encountered', tran, ['6A'] ).
verb( 'encourage', 'encourages', 'encouraging', 'encouraged', 'encouraged', tran, ['6A','14','17'] ).
verb( 'encroach', 'encroaches', 'encroaching', 'encroached', 'encroached', intran, ['3A'] ).
verb( 'encrust', 'encrusts', 'encrusting', 'encrusted', 'encrusted', _, ['2A','6A','14'] ).
verb( 'encumber', 'encumbers', 'encumbering', 'encumbered', 'encumbered', tran, ['6A','14'] ).
verb( 'end', 'ends', 'ending', 'ended', 'ended', _, ['2A','2C','3A','6A','15B'] ).
verb( 'endanger', 'endangers', 'endangering', 'endangered', 'endangered', tran, ['6A'] ).
verb( 'endear', 'endears', 'endearing', 'endeared', 'endeared', tran, ['14'] ).
verb( 'endeavour', 'endeavours', 'endeavouring', 'endeavoured', 'endeavoured', intran, ['4A'] ).
verb( 'endorse', 'endorses', 'endorsing', 'endorsed', 'endorsed', tran, ['6A'] ).
verb( 'endow', 'endows', 'endowing', 'endowed', 'endowed', tran, ['6A','14'] ).
verb( 'endue', 'endues', 'enduing', 'endued', 'endued', tran, ['14'] ).
verb( 'endure', 'endures', 'enduring', 'endured', 'endured', _, ['2A','2C','6A','6D','17'] ).
verb( 'enervate', 'enervates', 'enervating', 'enervated', 'enervated', tran, ['6A'] ).
verb( 'enfeeble', 'enfeebles', 'enfeebling', 'enfeebled', 'enfeebled', tran, ['6A'] ).
verb( 'enfold', 'enfolds', 'enfolding', 'enfolded', 'enfolded', tran, ['6A','14'] ).
verb( 'enforce', 'enforces', 'enforcing', 'enforced', 'enforced', tran, ['6A','14'] ).
verb( 'enfranchise', 'enfranchises', 'enfranchising', 'enfranchised', 'enfranchised', tran, ['6A'] ).
verb( 'engage', 'engages', 'engaging', 'engaged', 'engaged', _, ['2A','3A','6A','7A','9','14','16B','17'] ).
verb( 'engender', 'engenders', 'engendering', 'engendered', 'engendered', tran, ['6A'] ).
verb( 'engineer', 'engineers', 'engineering', 'engineered', 'engineered', _, ['2A','6A'] ).
verb( 'engraft', 'engrafts', 'engrafting', 'engrafted', 'engrafted', tran, ['6A','14'] ).
verb( 'engrave', 'engraves', 'engraving', 'engraved', 'engraved', tran, ['6A','14'] ).
verb( 'engross', 'engrosses', 'engrossing', 'engrossed', 'engrossed', tran, ['6A'] ).
verb( 'engulf', 'engulfs', 'engulfing', 'engulfed', 'engulfed', tran, ['6A'] ).
verb( 'enhance', 'enhances', 'enhancing', 'enhanced', 'enhanced', tran, ['6A'] ).
verb( 'enjoin', 'enjoins', 'enjoining', 'enjoined', 'enjoined', tran, ['6A','9','14','17'] ).
verb( 'enjoy', 'enjoys', 'enjoying', 'enjoyed', 'enjoyed', tran, ['6A','6C'] ).
verb( 'enkindle', 'enkindles', 'enkindling', 'enkindled', 'enkindled', tran, ['6A'] ).
verb( 'enlarge', 'enlarges', 'enlarging', 'enlarged', 'enlarged', _, ['2A','3A','6A'] ).
verb( 'enlighten', 'enlightens', 'enlightening', 'enlightened', 'enlightened', tran, ['6A','14'] ).
verb( 'enlist', 'enlists', 'enlisting', 'enlisted', 'enlisted', _, ['2A','2C','6A','14','16B'] ).
verb( 'enliven', 'enlivens', 'enlivening', 'enlivened', 'enlivened', tran, ['6A'] ).
verb( 'enmesh', 'enmeshes', 'enmeshing', 'enmeshed', 'enmeshed', tran, ['6A','14'] ).
verb( 'ennoble', 'ennobles', 'ennobling', 'ennobled', 'ennobled', tran, ['6A'] ).
verb( 'enplane', 'enplanes', 'enplaning', 'enplaned', 'enplaned', _, [] ).
verb( 'enquire', 'enquires', 'enquiring', 'enquired', 'enquired', _, ['2A','3A','6A','8','10','14'] ).
verb( 'enrage', 'enrages', 'enraging', 'enraged', 'enraged', tran, ['6A'] ).
verb( 'enrapture', 'enraptures', 'enrapturing', 'enraptured', 'enraptured', tran, ['6A'] ).
verb( 'enrich', 'enriches', 'enriching', 'enriched', 'enriched', tran, ['6A','14'] ).
verb( 'enrol', 'enrols', 'enroling', 'enroled', 'enroled', _, ['2A','2C','6A','14','16B'] ).
verb( 'enroll', 'enrolls', 'enrolling', 'enrolled', 'enrolled', _, ['2A','2C','6A','14','16B'] ).
verb( 'ensconce', 'ensconces', 'ensconcing', 'ensconced', 'ensconced', tran, ['14'] ).
verb( 'enshrine', 'enshrines', 'enshrining', 'enshrined', 'enshrined', tran, ['6A','14'] ).
verb( 'enshroud', 'enshrouds', 'enshrouding', 'enshrouded', 'enshrouded', tran, ['6A'] ).
verb( 'enslave', 'enslaves', 'enslaving', 'enslaved', 'enslaved', tran, ['6A'] ).
verb( 'ensnare', 'ensnares', 'ensnaring', 'ensnared', 'ensnared', tran, ['6A','14'] ).
verb( 'ensue', 'ensues', 'ensuing', 'ensued', 'ensued', intran, ['2A','3A'] ).
verb( 'ensure', 'ensures', 'ensuring', 'ensured', 'ensured', _, ['3A','9','12A','13A','14'] ).
verb( 'entail', 'entails', 'entailing', 'entailed', 'entailed', tran, ['6A','14'] ).
verb( 'entangle', 'entangles', 'entangling', 'entangled', 'entangled', tran, ['6A','14','15A'] ).
verb( 'enter', 'enters', 'entering', 'entered', 'entered', _, ['2A','3A','6A','14','15B'] ).
verb( 'entertain', 'entertains', 'entertaining', 'entertained', 'entertained', tran, ['2A','6A','14'] ).
verb( 'enthral', 'enthrals', 'enthralling', 'enthralled', 'enthralled', tran, ['6A'] ).
verb( 'enthrall', 'enthralls', 'enthralling', 'enthralled', 'enthralled', tran, ['6A'] ).
verb( 'enthrone', 'enthrones', 'enthroning', 'enthroned', 'enthroned', tran, ['6A'] ).
verb( 'enthuse', 'enthuses', 'enthusing', 'enthused', 'enthused', intran, ['3A'] ).
verb( 'entice', 'entices', 'enticing', 'enticed', 'enticed', tran, ['6A','15A','17'] ).
verb( 'entitle', 'entitles', 'entitling', 'entitled', 'entitled', tran, ['14','17'] ).
verb( 'entomb', 'entombs', 'entombing', 'entombed', 'entombed', tran, ['6A'] ).
verb( 'entrain', 'entrains', 'entraining', 'entrained', 'entrained', _, ['2A','6A'] ).
verb( 'entrance', 'entrances', 'entrancing', 'entranced', 'entranced', tran, ['6A'] ).
verb( 'entrap', 'entraps', 'entrapping', 'entrapped', 'entrapped', tran, ['6A'] ).
verb( 'entreat', 'entreats', 'entreating', 'entreated', 'entreated', tran, ['6A','14','17'] ).
verb( 'entrench', 'entrenches', 'entrenching', 'entrenched', 'entrenched', tran, ['6A'] ).
verb( 'entrust', 'entrusts', 'entrusting', 'entrusted', 'entrusted', tran, ['14'] ).
verb( 'entwine', 'entwines', 'entwining', 'entwined', 'entwined', tran, ['6A','14'] ).
verb( 'enumerate', 'enumerates', 'enumerating', 'enumerated', 'enumerated', tran, ['6A'] ).
verb( 'enunciate', 'enunciates', 'enunciating', 'enunciated', 'enunciated', _, ['2A','6A'] ).
verb( 'envelop', 'envelops', 'enveloping', 'enveloped', 'enveloped', tran, ['6A','14'] ).
verb( 'envenom', 'envenoms', 'envenoming', 'envenomed', 'envenomed', tran, ['6A'] ).
verb( 'environ', 'environs', 'environing', 'environed', 'environed', tran, ['6A'] ).
verb( 'envisage', 'envisages', 'envisaging', 'envisaged', 'envisaged', tran, ['6A'] ).
verb( 'envy', 'envies', 'envying', 'envied', 'envied', tran, ['6A','12C'] ).
verb( 'enwrap', 'enwraps', 'enwrapping', 'enwrapped', 'enwrapped', tran, [] ).
verb( 'epitomize', 'epitomizes', 'epitomizing', 'epitomized', 'epitomized', tran, ['6A'] ).
verb( 'equal', 'equals', 'equalling', 'equalled', 'equalled', tran, ['6A','15A'] ).
verb( 'equalize', 'equalizes', 'equalizing', 'equalized', 'equalized', tran, ['6A'] ).
verb( 'equate', 'equates', 'equating', 'equated', 'equated', tran, ['6A','14'] ).
verb( 'equip', 'equips', 'equipping', 'equipped', 'equipped', tran, ['6A','14'] ).
verb( 'eradicate', 'eradicates', 'eradicating', 'eradicated', 'eradicated', tran, ['6A'] ).
verb( 'erase', 'erases', 'erasing', 'erased', 'erased', tran, ['6A'] ).
verb( 'erect', 'erects', 'erecting', 'erected', 'erected', tran, ['6A'] ).
verb( 'erode', 'erodes', 'eroding', 'eroded', 'eroded', tran, ['6A'] ).
verb( 'err', 'errs', 'erring', 'erred', 'erred', intran, ['2A','2C'] ).
verb( 'erupt', 'erupts', 'erupting', 'erupted', 'erupted', intran, ['2A'] ).
verb( 'escalate', 'escalates', 'escalating', 'escalated', 'escalated', _, ['2A','6A'] ).
verb( 'escape', 'escapes', 'escaping', 'escaped', 'escaped', _, ['2A','3A','6A','6C'] ).
verb( 'eschew', 'eschews', 'eschewing', 'eschewed', 'eschewed', tran, ['6A'] ).
verb( 'escort', 'escorts', 'escorting', 'escorted', 'escorted', tran, ['6A','15B'] ).
verb( 'espouse', 'espouses', 'espousing', 'espoused', 'espoused', tran, ['6A'] ).
verb( 'espy', 'espies', 'espying', 'espied', 'espied', tran, ['6A'] ).
verb( 'essay', 'essays', 'essaying', 'essayed', 'essayed', _, ['4A','6A'] ).
verb( 'establish', 'establishes', 'establishing', 'established', 'established', tran, ['6A','14','16B'] ).
verb( 'esteem', 'esteems', 'esteeming', 'esteemed', 'esteemed', tran, ['6A','25'] ).
verb( 'estimate', 'estimates', 'estimating', 'estimated', 'estimated', _, ['3A','9','14'] ).
verb( 'estrange', 'estranges', 'estranging', 'estranged', 'estranged', tran, ['6A','14'] ).
verb( 'etch', 'etches', 'etching', 'etched', 'etched', _, ['2A','6A'] ).
verb( 'eulogize', 'eulogizes', 'eulogizing', 'eulogized', 'eulogized', tran, ['6A'] ).
verb( 'evacuate', 'evacuates', 'evacuating', 'evacuated', 'evacuated', tran, ['6A','14'] ).
verb( 'evade', 'evades', 'evading', 'evaded', 'evaded', tran, ['6A','6C'] ).
verb( 'evaluate', 'evaluates', 'evaluating', 'evaluated', 'evaluated', tran, ['6A'] ).
verb( 'evaporate', 'evaporates', 'evaporating', 'evaporated', 'evaporated', _, ['2A','6A'] ).
verb( 'even', 'evens', 'evening', 'evened', 'evened', tran, ['6A','15B'] ).
verb( 'evict', 'evicts', 'evicting', 'evicted', 'evicted', tran, ['14'] ).
verb( 'evidence', 'evidences', 'evidencing', 'evidenced', 'evidenced', tran, ['6A'] ).
verb( 'evince', 'evinces', 'evincing', 'evinced', 'evinced', tran, ['6A','9'] ).
verb( 'eviscerate', 'eviscerates', 'eviscerating', 'eviscerated', 'eviscerated', tran, ['6A'] ).
verb( 'evoke', 'evokes', 'evoking', 'evoked', 'evoked', tran, ['6A'] ).
verb( 'evolve', 'evolves', 'evolving', 'evolved', 'evolved', _, ['2A','6A'] ).
verb( 'exacerbate', 'exacerbates', 'exacerbating', 'exacerbated', 'exacerbated', tran, ['6A'] ).
verb( 'exact', 'exacts', 'exacting', 'exacted', 'exacted', tran, ['6A','14'] ).
verb( 'exaggerate', 'exaggerates', 'exaggerating', 'exaggerated', 'exaggerated', _, ['2A','6A'] ).
verb( 'exalt', 'exalts', 'exalting', 'exalted', 'exalted', tran, ['6A'] ).
verb( 'examine', 'examines', 'examining', 'examined', 'examined', tran, ['6A','14'] ).
verb( 'exasperate', 'exasperates', 'exasperating', 'exasperated', 'exasperated', tran, ['6A'] ).
verb( 'excavate', 'excavates', 'excavating', 'excavated', 'excavated', tran, ['6A'] ).
verb( 'exceed', 'exceeds', 'exceeding', 'exceeded', 'exceeded', tran, ['6A'] ).
verb( 'excel', 'excels', 'excelling', 'excelled', 'excelled', _, ['2C','3A','6A','15A'] ).
verb( 'except', 'excepts', 'excepting', 'excepted', 'excepted', tran, ['6A','14'] ).
verb( 'exchange', 'exchanges', 'exchanging', 'exchanged', 'exchanged', tran, ['6A','14'] ).
verb( 'excise', 'excises', 'excising', 'excised', 'excised', tran, ['6A'] ).
verb( 'excite', 'excites', 'exciting', 'excited', 'excited', tran, ['6A','14','17'] ).
verb( 'exclaim', 'exclaims', 'exclaiming', 'exclaimed', 'exclaimed', _, ['2A','9'] ).
verb( 'exclude', 'excludes', 'excluding', 'excluded', 'excluded', tran, ['6A','14'] ).
verb( 'excogitate', 'excogitates', 'excogitating', 'excogitated', 'excogitated', tran, ['6A'] ).
verb( 'excommunicate', 'excommunicates', 'excommunicating', 'excommunicated', 'excommunicated', tran, ['6A'] ).
verb( 'excoriate', 'excoriates', 'excoriating', 'excoriated', 'excoriated', tran, ['6A'] ).
verb( 'excrete', 'excretes', 'excreting', 'excreted', 'excreted', tran, ['6A'] ).
verb( 'exculpate', 'exculpates', 'exculpating', 'exculpated', 'exculpated', tran, ['6A','14'] ).
verb( 'excuse', 'excuses', 'excusing', 'excused', 'excused', tran, ['6A','6C','12B','13B','14','19C'] ).
verb( 'execrate', 'execrates', 'execrating', 'execrated', 'execrated', tran, ['6A'] ).
verb( 'execute', 'executes', 'executing', 'executed', 'executed', tran, ['6A'] ).
verb( 'exemplify', 'exemplifies', 'exemplifying', 'exemplified', 'exemplified', tran, ['6A'] ).
verb( 'exempt', 'exempts', 'exempting', 'exempted', 'exempted', tran, ['6A','14'] ).
verb( 'exercise', 'exercises', 'exercising', 'exercised', 'exercised', _, ['2A','6A','15A'] ).
verb( 'exert', 'exerts', 'exerting', 'exerted', 'exerted', tran, ['6A','14','16A'] ).
verb( 'exhale', 'exhales', 'exhaling', 'exhaled', 'exhaled', _, ['2A','6A'] ).
verb( 'exhaust', 'exhausts', 'exhausting', 'exhausted', 'exhausted', tran, ['6A'] ).
verb( 'exhibit', 'exhibits', 'exhibiting', 'exhibited', 'exhibited', tran, ['2A','6A'] ).
verb( 'exhilarate', 'exhilarates', 'exhilarating', 'exhilarated', 'exhilarated', tran, ['6A'] ).
verb( 'exhort', 'exhorts', 'exhorting', 'exhorted', 'exhorted', tran, ['6A','14','17'] ).
verb( 'exhume', 'exhumes', 'exhuming', 'exhumed', 'exhumed', tran, ['6A'] ).
verb( 'exile', 'exiles', 'exiling', 'exiled', 'exiled', tran, ['6A','15A'] ).
verb( 'exist', 'exists', 'existing', 'existed', 'existed', intran, ['2A','2C','3A'] ).
verb( 'exit', 'exits', 'exiting', 'exited', 'exited', intran, [] ).
verb( 'exonerate', 'exonerates', 'exonerating', 'exonerated', 'exonerated', tran, ['6A','14'] ).
verb( 'exorcize', 'exorcizes', 'exorcizing', 'exorcized', 'exorcized', tran, ['6A','14'] ).
verb( 'expand', 'expands', 'expanding', 'expanded', 'expanded', _, ['2A','2C','6A','14'] ).
verb( 'expatiate', 'expatiates', 'expatiating', 'expatiated', 'expatiated', intran, ['3A'] ).
verb( 'expatriate', 'expatriates', 'expatriating', 'expatriated', 'expatriated', tran, ['6A'] ).
verb( 'expect', 'expects', 'expecting', 'expected', 'expected', tran, ['6A','7A','9','14','17'] ).
verb( 'expectorate', 'expectorates', 'expectorating', 'expectorated', 'expectorated', _, ['2A','6A'] ).
verb( 'expedite', 'expedites', 'expediting', 'expedited', 'expedited', tran, ['6A'] ).
verb( 'expel', 'expels', 'expelling', 'expelled', 'expelled', tran, ['6A','14'] ).
verb( 'expend', 'expends', 'expending', 'expended', 'expended', tran, ['6A','14'] ).
verb( 'experience', 'experiences', 'experiencing', 'experienced', 'experienced', tran, ['6A'] ).
verb( 'experiment', 'experiments', 'experimenting', 'experimented', 'experimented', intran, ['2A','2C','3A'] ).
verb( 'expiate', 'expiates', 'expiating', 'expiated', 'expiated', tran, ['6A'] ).
verb( 'expire', 'expires', 'expiring', 'expired', 'expired', intran, ['2A'] ).
verb( 'explain', 'explains', 'explaining', 'explained', 'explained', tran, ['6A','8','9','10','14','15B'] ).
verb( 'explicate', 'explicates', 'explicating', 'explicated', 'explicated', tran, ['6A'] ).
verb( 'explode', 'explodes', 'exploding', 'exploded', 'exploded', _, ['2A','2C','6A'] ).
verb( 'exploit', 'exploits', 'exploiting', 'exploited', 'exploited', tran, ['6A'] ).
verb( 'explore', 'explores', 'exploring', 'explored', 'explored', tran, ['6A'] ).
verb( 'export', 'exports', 'exporting', 'exported', 'exported', tran, ['6A'] ).
verb( 'expose', 'exposes', 'exposing', 'exposed', 'exposed', tran, ['6A','14','15A'] ).
verb( 'expostulate', 'expostulates', 'expostulating', 'expostulated', 'expostulated', intran, ['2A','3A'] ).
verb( 'expound', 'expounds', 'expounding', 'expounded', 'expounded', tran, ['6A','14'] ).
verb( 'express', 'expresses', 'expressing', 'expressed', 'expressed', tran, ['6A','10','14','15A'] ).
verb( 'expropriate', 'expropriates', 'expropriating', 'expropriated', 'expropriated', tran, ['6A','14'] ).
verb( 'expunge', 'expunges', 'expunging', 'expunged', 'expunged', tran, ['6A','14'] ).
verb( 'expurgate', 'expurgates', 'expurgating', 'expurgated', 'expurgated', tran, ['6A'] ).
verb( 'extemporize', 'extemporize', 'extemporize', 'extemporize', 'extemporize', _, ['2A','6A'] ).
verb( 'extend', 'extends', 'extending', 'extended', 'extended', _, ['2B','2C','6A','14','15A'] ).
verb( 'extenuate', 'extenuates', 'extenuating', 'extenuated', 'extenuated', tran, ['6A'] ).
verb( 'exteriorize', 'exteriorizes', 'exteriorizing', 'exteriorized', 'exteriorized', tran, [] ).
verb( 'exterminate', 'exterminates', 'exterminating', 'exterminated', 'exterminated', tran, ['6A'] ).
verb( 'externalize', 'externalizes', 'externalizing', 'externalized', 'externalized', tran, ['6A'] ).
verb( 'extinguish', 'extinguishes', 'extinguishing', 'extinguished', 'extinguished', tran, ['6A'] ).
verb( 'extirpate', 'extirpates', 'extirpating', 'extirpated', 'extirpated', tran, ['6A'] ).
verb( 'extol', 'extols', 'extolling', 'extolled', 'extolled', tran, ['6A','15A'] ).
verb( 'extort', 'extorts', 'extorting', 'extorted', 'extorted', tran, ['6A','14'] ).
verb( 'extract', 'extracts', 'extracting', 'extracted', 'extracted', tran, ['6A','14'] ).
verb( 'extradite', 'extradites', 'extraditing', 'extradited', 'extradited', tran, ['6A'] ).
verb( 'extrapolate', 'extrapolate', 'extrapolate', 'extrapolate', 'extrapolate', _, ['6A'] ).
verb( 'extricate', 'extricates', 'extricating', 'extricated', 'extricated', tran, ['6A','14'] ).
verb( 'extrude', 'extrudes', 'extruding', 'extruded', 'extruded', tran, ['6A','14'] ).
verb( 'exude', 'exudes', 'exuding', 'exuded', 'exuded', _, ['2A','2C','6A'] ).
verb( 'exult', 'exults', 'exulting', 'exulted', 'exulted', intran, ['2A','3A','4C'] ).
verb( 'eye', 'eyes', 'eying', 'eyed', 'eyed', tran, ['6A','15A'] ).
verb( 'f^ete', 'f^etes', 'f^eting', 'f^eted', 'f^eted', tran, ['6A'] ).
verb( 'fabricate', 'fabricates', 'fabricating', 'fabricated', 'fabricated', tran, ['6A'] ).
verb( 'face', 'faces', 'facing', 'faced', 'faced', _, ['2C','6A','6C','14','15B'] ).
verb( 'facilitate', 'facilitates', 'facilitating', 'facilitated', 'facilitated', tran, ['6A'] ).
verb( 'factorize', 'factorizes', 'factorizing', 'factorized', 'factorized', tran, ['6A'] ).
verb( 'fade', 'fades', 'fading', 'faded', 'faded', _, ['2A','2C','3A','6A','15B'] ).
verb( 'fag', 'fags', 'fagging', 'fagged', 'fagged', _, ['2A','2C','3A','6A','15B'] ).
verb( 'fail', 'fails', 'failing', 'failed', 'failed', _, ['2A','2C','3A','4A','6A'] ).
verb( 'faint', 'faints', 'fainting', 'fainted', 'fainted', intran, ['2A','2C'] ).
verb( 'fake', 'fakes', 'faking', 'faked', 'faked', tran, ['6A','15B'] ).
verb( 'fall', 'falls', 'falling', 'fell', 'fallen', intran, ['2A','2B','2C','2D','3A'] ).
verb( 'falsify', 'falsifies', 'falsifying', 'falsified', 'falsified', tran, ['6A'] ).
verb( 'falter', 'falters', 'faltering', 'faltered', 'faltered', _, ['2A','2C','15B'] ).
verb( 'familiarize', 'familiarizes', 'familiarizing', 'familiarized', 'familiarized', tran, ['6A','14'] ).
verb( 'famish', 'famishes', 'famishing', 'famished', 'famished', _, ['2A','3A','6A'] ).
verb( 'fan', 'fans', 'fanning', 'fanned', 'fanned', _, ['2C','6A'] ).
verb( 'fancy', 'fancies', 'fancying', 'fancied', 'fancied', tran, ['6A','6C','9','16B','19C','25'] ).
verb( 'fare', 'fares', 'faring', 'fared', 'fared', intran, ['2C'] ).
verb( 'farm', 'farms', 'farming', 'farmed', 'farmed', _, ['2A','6A','15B'] ).
verb( 'farrow', 'farrows', 'farrowing', 'farrowed', 'farrowed', intran, [] ).
verb( 'fart', 'farts', 'farting', 'farted', 'farted', intran, [] ).
verb( 'fascinate', 'fascinates', 'fascinating', 'fascinated', 'fascinated', tran, ['6A'] ).
verb( 'fashion', 'fashions', 'fashioning', 'fashioned', 'fashioned', tran, ['6A','15A'] ).
verb( 'fast', 'fasts', 'fasting', 'fasted', 'fasted', intran, ['2A','2B'] ).
verb( 'fasten', 'fastens', 'fastening', 'fastened', 'fastened', _, ['2A','2C','3A','6A','14','15A','15B'] ).
verb( 'fat', 'fats', 'fatting', 'fatted', 'fatted', tran, [] ).
verb( 'fate', 'fates', 'fating', 'fated', 'fated', tran, ['17'] ).
verb( 'father', 'fathers', 'fathering', 'fathered', 'fathered', tran, ['6A','14'] ).
verb( 'fathom', 'fathoms', 'fathoming', 'fathomed', 'fathomed', tran, ['6A'] ).
verb( 'fatigue', 'fatigues', 'fatiguing', 'fatigued', 'fatigued', tran, ['6A'] ).
verb( 'fatten', 'fattens', 'fattening', 'fattened', 'fattened', _, ['2A','2C','6A','15B'] ).
verb( 'fault', 'faults', 'faulting', 'faulted', 'faulted', tran, ['6A'] ).
verb( 'favour', 'favours', 'favouring', 'favoured', 'favoured', tran, ['6A','14'] ).
verb( 'fawn', 'fawns', 'fawning', 'fawned', 'fawned', intran, ['2A','3A'] ).
verb( 'fear', 'fears', 'fearing', 'feared', 'feared', _, ['2A','3A','4A','6A','6C','9'] ).
verb( 'feast', 'feasts', 'feasting', 'feasted', 'feasted', _, ['2A','2B','6A','14'] ).
verb( 'feather', 'feathers', 'feathering', 'feathered', 'feathered', tran, ['6A'] ).
verb( 'featherbed', 'featherbeds', 'featherbedding', 'featherbedded', 'featherbedded', tran, [] ).
verb( 'feature', 'features', 'featuring', 'featured', 'featured', tran, ['6A'] ).
verb( 'federate', 'federates', 'federating', 'federated', 'federated', _, ['2A','6A'] ).
verb( 'fee', 'fees', 'feeing', 'feed', 'feed', tran, ['6A'] ).
verb( 'feed', 'feeds', 'feeding', 'fed', 'fed', _, ['2A','2C','3A','6A','14','15A','15B'] ).
verb( 'feel', 'feels', 'feeling', 'felt', 'felt', _, ['2A','2C','2D','3A','6A','6C','9','10','15B','18A','19A','25'] ).
verb( 'feign', 'feigns', 'feigning', 'feigned', 'feigned', tran, ['6A','9'] ).
verb( 'feint', 'feints', 'feinting', 'feinted', 'feinted', intran, ['2A','3A'] ).
verb( 'felicitate', 'felicitates', 'felicitating', 'felicitated', 'felicitated', tran, ['6A','14'] ).
verb( 'fell', 'fells', 'felling', 'felled', 'felled', tran, ['2A','2B','2C','2D','3A','6A'] ).
verb( 'fence', 'fences', 'fencing', 'fenced', 'fenced', tran, ['2A','2C','3A','6A','15B'] ).
verb( 'fend', 'fends', 'fending', 'fended', 'fended', _, ['15B'] ).
verb( 'ferment', 'ferments', 'fermenting', 'fermented', 'fermented', _, ['2A','6A'] ).
verb( 'ferret', 'ferrets', 'ferreting', 'ferreted', 'ferreted', _, ['2A','2C','15B'] ).
verb( 'ferry', 'ferries', 'ferrying', 'ferried', 'ferried', _, ['2A','2C','6A','15A','15B'] ).
verb( 'fertilize', 'fertilizes', 'fertilizing', 'fertilized', 'fertilized', tran, ['6A'] ).
verb( 'fester', 'festers', 'festering', 'festered', 'festered', intran, ['2A'] ).
verb( 'festoon', 'festoons', 'festooning', 'festooned', 'festooned', tran, ['6A'] ).
verb( 'fetch', 'fetches', 'fetching', 'fetched', 'fetched', _, ['6A','12B','12C','13B','15A','15B'] ).
verb( 'fetter', 'fetters', 'fettering', 'fettered', 'fettered', tran, ['6A'] ).
verb( 'fib', 'fibs', 'fibbing', 'fibbed', 'fibbed', intran, ['2A'] ).
verb( 'fiddle', 'fiddles', 'fiddling', 'fiddled', 'fiddled', tran, ['2A','2C','6A'] ).
verb( 'fidget', 'fidgets', 'fidgeting', 'fidgeted', 'fidgeted', _, ['2A','2C','6A'] ).
verb( 'field', 'fields', 'fielding', 'fielded', 'fielded', _, ['2A','6A'] ).
verb( 'fight', 'fights', 'fighting', 'fought', 'fought', _, ['2A','2B','2C','3A','4A','6A','15A','15B'] ).
verb( 'figure', 'figures', 'figuring', 'figured', 'figured', _, ['2C','3A','9','15A','15B','25'] ).
verb( 'filch', 'filches', 'filching', 'filched', 'filched', tran, ['6A'] ).
verb( 'file', 'files', 'filing', 'filed', 'filed', _, ['2C','6A','15A','15B','22'] ).
verb( 'filibuster', 'filibusters', 'filibustering', 'filibustered', 'filibustered', intran, [] ).
verb( 'fill', 'fills', 'filling', 'filled', 'filled', _, ['2A','2C','6A','12B','13B','14','15B'] ).
verb( 'fillet', 'fillets', 'filleting', 'filleted', 'filleted', tran, ['6A'] ).
verb( 'film', 'films', 'filming', 'filmed', 'filmed', _, ['2A','2C','6A'] ).
verb( 'filter', 'filters', 'filtering', 'filtered', 'filtered', _, ['2A','2C','6A','14','15B'] ).
verb( 'filtrate', 'filtrates', 'filtrating', 'filtrated', 'filtrated', _, [] ).
verb( 'finalize', 'finalizes', 'finalizing', 'finalized', 'finalized', tran, ['6A'] ).
verb( 'finance', 'finances', 'financing', 'financed', 'financed', tran, ['6A'] ).
verb( 'find', 'finds', 'finding', 'found', 'found', tran, ['6A','8','9','10','12A','12B','13A','13B','15A','15B','19B','22','25'] ).
verb( 'fine', 'fines', 'fining', 'fined', 'fined', tran, ['6A','14'] ).
verb( 'finger', 'fingers', 'fingering', 'fingered', 'fingered', tran, ['6A'] ).
verb( 'finish', 'finishes', 'finishing', 'finished', 'finished', _, ['2A','2C','6A','6C','15B','24B'] ).
verb( 'fire', 'fires', 'firing', 'fired', 'fired', _, ['2A','2C','3A','6A','15B'] ).
verb( 'firm', 'firms', 'firming', 'firmed', 'firmed', _, [] ).
verb( 'fish', 'fishes', 'fishing', 'fished', 'fished', _, ['2A','2C','6A','15A','15B'] ).
verb( 'fit', 'fits', 'fitting', 'fitted', 'fitted', _, ['2A','2C','6A','14','15A','15B','16A'] ).
verb( 'fix', 'fixes', 'fixing', 'fixed', 'fixed', _, ['3A','6A','14','15A','15B'] ).
verb( 'fixate', 'fixates', 'fixating', 'fixated', 'fixated', tran, ['6A'] ).
verb( 'fizz', 'fizzes', 'fizzing', 'fizzed', 'fizzed', intran, ['2A','2C'] ).
verb( 'fizzle', 'fizzles', 'fizzling', 'fizzled', 'fizzled', intran, ['2A','2C'] ).
verb( 'flabbergast', 'flabbergasts', 'flabbergasting', 'flabbergasted', 'flabbergasted', tran, ['6A'] ).
verb( 'flag', 'flags', 'flagging', 'flagged', 'flagged', _, ['2A','6A','15B'] ).
verb( 'flagellate', 'flagellates', 'flagellating', 'flagellated', 'flagellated', tran, ['6A'] ).
verb( 'flail', 'flails', 'flailing', 'flailed', 'flailed', tran, ['6A'] ).
verb( 'flake', 'flakes', 'flaking', 'flaked', 'flaked', intran, ['2A','2C'] ).
verb( 'flame', 'flames', 'flaming', 'flamed', 'flamed', intran, ['2A','2C'] ).
verb( 'flank', 'flanks', 'flanking', 'flanked', 'flanked', tran, ['6A'] ).
verb( 'flap', 'flaps', 'flapping', 'flapped', 'flapped', _, ['2A','2C','6A','15B'] ).
verb( 'flare', 'flares', 'flaring', 'flared', 'flared', _, ['2A','2C','6A'] ).
verb( 'flash', 'flashes', 'flashing', 'flashed', 'flashed', _, ['2A','2C','6A','12C','15A'] ).
verb( 'flatten', 'flattens', 'flattening', 'flattened', 'flattened', _, ['2A','2C','6A','15A','15B'] ).
verb( 'flatter', 'flatters', 'flattering', 'flattered', 'flattered', tran, ['6A'] ).
verb( 'flaunt', 'flaunts', 'flaunting', 'flaunted', 'flaunted', _, ['2A','2C','6A'] ).
verb( 'flavour', 'flavours', 'flavouring', 'flavoured', 'flavoured', tran, ['6A'] ).
verb( 'flay', 'flays', 'flaying', 'flayed', 'flayed', tran, ['6A'] ).
verb( 'fleck', 'flecks', 'flecking', 'flecked', 'flecked', tran, ['6A'] ).
verb( 'flee', 'flees', 'fleeing', 'fled', 'flown', _, ['2A','2C','6A'] ).
verb( 'fleece', 'fleeces', 'fleecing', 'fleeced', 'fleeced', tran, ['6A','14'] ).
verb( 'flex', 'flexes', 'flexing', 'flexed', 'flexed', tran, ['6A'] ).
verb( 'flick', 'flicks', 'flicking', 'flicked', 'flicked', tran, ['6A','15A','15B','22'] ).
verb( 'flicker', 'flickers', 'flickering', 'flickered', 'flickered', intran, ['2A','2C'] ).
verb( 'flight', 'flights', 'flighting', 'flighted', 'flighted', tran, ['6A'] ).
verb( 'flinch', 'flinches', 'flinching', 'flinched', 'flinched', intran, ['2A','3A'] ).
verb( 'fling', 'flings', 'flinging', 'flung', 'flung', _, ['2C','6A','12A','13A','15A','15B','22'] ).
verb( 'flip', 'flips', 'flipping', 'flipped', 'flipped', _, ['6A','15A','15B'] ).
verb( 'flirt', 'flirts', 'flirting', 'flirted', 'flirted', intran, ['2A','3A'] ).
verb( 'flit', 'flits', 'flitting', 'flitted', 'flitted', intran, ['2C'] ).
verb( 'float', 'floats', 'floating', 'floated', 'floated', _, ['2A','2C','6A','15A','15B'] ).
verb( 'flock', 'flocks', 'flocking', 'flocked', 'flocked', intran, ['2C','4A'] ).
verb( 'flog', 'flogs', 'flogging', 'flogged', 'flogged', tran, ['6A'] ).
verb( 'flood', 'floods', 'flooding', 'flooded', 'flooded', _, ['3A','6A','14','15B','16A'] ).
verb( 'floodlight', 'floodlights', 'floodlighting', 'floodlighted', 'floodlighted', tran, [] ).
verb( 'floor', 'floors', 'flooring', 'floored', 'floored', tran, ['6A'] ).
verb( 'flop', 'flops', 'flopping', 'flopped', 'flopped', _, ['2A','2C','15A','15B'] ).
verb( 'flounce', 'flounces', 'flouncing', 'flounced', 'flounced', _, ['2C','6A'] ).
verb( 'flounder', 'flounders', 'floundering', 'floundered', 'floundered', intran, ['2A','2C'] ).
verb( 'flour', 'flours', 'flouring', 'floured', 'floured', tran, ['6A'] ).
verb( 'flourish', 'flourishes', 'flourishing', 'flourished', 'flourished', _, ['2A','6A'] ).
verb( 'flout', 'flouts', 'flouting', 'flouted', 'flouted', tran, ['6A'] ).
verb( 'flow', 'flows', 'flowing', 'flowed', 'flowed', intran, ['2A','2C'] ).
verb( 'flower', 'flowers', 'flowering', 'flowered', 'flowered', intran, ['2A','2C'] ).
verb( 'fluctuate', 'fluctuates', 'fluctuating', 'fluctuated', 'fluctuated', intran, ['2A','2C'] ).
verb( 'fluff', 'fluffs', 'fluffing', 'fluffed', 'fluffed', tran, ['6A','15B'] ).
verb( 'flummox', 'flummoxes', 'flummoxing', 'flummoxed', 'flummoxed', tran, ['6A'] ).
verb( 'flunk', 'flunks', 'flunking', 'flunked', 'flunked', _, ['2A','2C','6A','15B'] ).
verb( 'fluoridate', 'fluoridates', 'fluoridating', 'fluoridated', 'fluoridated', tran, ['6A'] ).
verb( 'fluoridize', 'fluoridizes', 'fluoridizing', 'fluoridized', 'fluoridized', tran, [] ).
verb( 'flurry', 'flurries', 'flurrying', 'flurried', 'flurried', tran, ['6A'] ).
verb( 'flush', 'flushes', 'flushing', 'flushed', 'flushed', _, ['2A','2C','2D','6A','14'] ).
verb( 'fluster', 'flusters', 'flustering', 'flustered', 'flustered', tran, ['6A'] ).
verb( 'flute', 'flutes', 'fluting', 'fluted', 'fluted', _, ['6A'] ).
verb( 'flutter', 'flutters', 'fluttering', 'fluttered', 'fluttered', _, ['2A','2C','6A','15A','15B'] ).
verb( 'fly', 'flies', 'flying', 'flew', 'flew', _, ['2A','2B','2C','2D','4A','6A','15A','15B'] ).
verb( 'fly-fish', 'fly-fishes', 'fly-fishing', 'fly-fished', 'fly-fished', intran, ['2A'] ).
verb( 'foal', 'foals', 'foaling', 'foaled', 'foaled', intran, ['2A'] ).
verb( 'foam', 'foams', 'foaming', 'foamed', 'foamed', intran, ['2A','2C'] ).
verb( 'fob', 'fobs', 'fobbing', 'fobbed', 'fobbed', tran, ['15B'] ).
verb( 'focus', 'focuses', 'focusing', 'focused', 'focused', _, ['2A','2C','6A','14'] ).
verb( 'fog', 'fogs', 'fogging', 'fogged', 'fogged', tran, [] ).
verb( 'foil', 'foils', 'foiling', 'foiled', 'foiled', tran, ['6A'] ).
verb( 'foist', 'foists', 'foisting', 'foisted', 'foisted', tran, ['14','15A'] ).
verb( 'fold', 'folds', 'folding', 'folded', 'folded', _, ['2A','2C','6A','15A','15B'] ).
verb( 'follow', 'follows', 'following', 'followed', 'followed', _, ['2A','2B','2C','6A','15B'] ).
verb( 'foment', 'foments', 'fomenting', 'fomented', 'fomented', tran, ['6A'] ).
verb( 'fondle', 'fondles', 'fondling', 'fondled', 'fondled', tran, ['6A'] ).
verb( 'fool', 'fools', 'fooling', 'fooled', 'fooled', _, ['2A','2C','6A','14','15B'] ).
verb( 'foot', 'foots', 'footing', 'footed', 'footed', _, ['6A'] ).
verb( 'footle', 'footles', 'footling', 'footled', 'footled', _, [] ).
verb( 'footslog', 'footslogs', 'footslogging', 'footslogged', 'footslogged', intran, [] ).
verb( 'forage', 'forages', 'foraging', 'foraged', 'foraged', intran, ['2A','3A'] ).
verb( 'foray', 'forays', 'foraying', 'forayed', 'forayed', intran, ['2A'] ).
verb( 'forbear', 'forbears', 'forbearing', 'forbore', 'forborne', _, ['2A','3A','6C','7A'] ).
verb( 'forbid', 'forbids', 'forbidding', 'forbad', 'forbidden', tran, ['6A','12C','17'] ).
verb( 'force', 'forces', 'forcing', 'forced', 'forced', tran, ['6A','15A','15B','17','22'] ).
verb( 'force-feed', 'force-feeds', 'force-feeding', 'force-fed', 'force-fed', tran, ['6A'] ).
verb( 'force-land', 'force-lands', 'force-landing', 'force-landed', 'force-landed', _, [] ).
verb( 'ford', 'fords', 'fording', 'forded', 'forded', tran, ['6A'] ).
verb( 'forearm', 'forearms', 'forearming', 'forearmed', 'forearmed', tran, ['6A'] ).
verb( 'forebode', 'forebodes', 'foreboding', 'foreboded', 'foreboded', tran, ['6A','9'] ).
verb( 'forecast', 'forecasts', 'forecasting', 'forecasted', 'forecasted', tran, ['6A'] ).
verb( 'foreclose', 'forecloses', 'foreclosing', 'foreclosed', 'foreclosed', _, ['2A','3A','6A'] ).
verb( 'foredoom', 'foredooms', 'foredooming', 'foredoomed', 'foredoomed', tran, ['6A','14'] ).
verb( 'foregather', 'foregathers', 'foregathering', 'foregathered', 'foregathered', intran, ['2A','2C'] ).
verb( 'forego', 'foregoes', 'foregoing', 'forewent', 'foregone', _, [] ).
verb( 'foreknow', 'foreknows', 'foreknowing', 'foreknew', 'foreknown', _, ['2A','3A','6A','8','9','10','17','18B','25'] ).
verb( 'foreordain', 'foreordains', 'foreordaining', 'foreordained', 'foreordained', tran, ['6A','14','17'] ).
verb( 'foresee', 'foresees', 'foreseeing', 'foresaw', 'foreseen', tran, ['6A','9','10'] ).
verb( 'foreshadow', 'foreshadows', 'foreshadowing', 'foreshadowed', 'foreshadowed', tran, ['6A'] ).
verb( 'foreshorten', 'foreshortens', 'foreshortening', 'foreshortened', 'foreshortened', tran, ['6A'] ).
verb( 'forestall', 'forestalls', 'forestalling', 'forestalled', 'forestalled', tran, ['6A'] ).
verb( 'foreswear', 'foreswears', 'foreswearing', 'foreswore', 'foresworn', tran, ['6A'] ).
verb( 'foretell', 'foretells', 'foretelling', 'foretold', 'foretold', tran, ['6A','9','10','12A','13A'] ).
verb( 'forewarn', 'forewarns', 'forewarning', 'forewarned', 'forewarned', tran, ['6A'] ).
verb( 'forfeit', 'forfeits', 'forfeiting', 'forfeited', 'forfeited', tran, ['6A'] ).
verb( 'forgather', 'forgathers', 'forgathering', 'forgathered', 'forgathered', intran, ['2A','2C'] ).
verb( 'forge', 'forges', 'forging', 'forged', 'forged', _, ['2C','6A'] ).
verb( 'forget', 'forgets', 'forgetting', 'forgot', 'forgotten', _, ['2A','3A','6A','6C','6D','7A','8','9','10'] ).
verb( 'forgive', 'forgives', 'forgiving', 'forgave', 'forgiven', _, ['2A','6A','12C','14'] ).
verb( 'forgo', 'forgoes', 'forgoing', 'forwent', 'forgone', tran, [] ).
verb( 'fork', 'forks', 'forking', 'forked', 'forked', _, ['2A','2C','6A','15A','15B'] ).
verb( 'form', 'forms', 'forming', 'formed', 'formed', _, ['2A','2C','6A','14','15A'] ).
verb( 'formalize', 'formalizes', 'formalizing', 'formalized', 'formalized', tran, ['6A'] ).
verb( 'formulate', 'formulates', 'formulating', 'formulated', 'formulated', tran, ['6A'] ).
verb( 'fornicate', 'fornicates', 'fornicating', 'fornicated', 'fornicated', intran, ['21'] ).
verb( 'forsake', 'forsakes', 'forsaking', 'forsook', 'forsaken', tran, ['6A'] ).
verb( 'forswear', 'forswears', 'forswearing', 'forswore', 'forsworn', tran, ['6A'] ).
verb( 'fortify', 'fortifies', 'fortifying', 'fortified', 'fortified', tran, ['6A','14'] ).
verb( 'forward', 'forwards', 'forwarding', 'forwarded', 'forwarded', tran, ['6A','12A','13A','15A'] ).
verb( 'fossilize', 'fossilizes', 'fossilizing', 'fossilized', 'fossilized', _, ['2A','6A'] ).
verb( 'foster', 'fosters', 'fostering', 'fostered', 'fostered', tran, ['6A'] ).
verb( 'foul', 'fouls', 'fouling', 'fouled', 'fouled', _, ['2A','2C','6A'] ).
verb( 'found', 'founds', 'founding', 'founded', 'founded', tran, ['6A','8','9','10','12A','12B','13A','13B','14','15A','15B','19B','22','25'] ).
verb( 'founder', 'founders', 'foundering', 'foundered', 'foundered', _, ['2A','6A'] ).
verb( 'fowl', 'fowls', 'fowling', 'fowled', 'fowled', intran, [] ).
verb( 'fox', 'foxes', 'foxing', 'foxed', 'foxed', tran, ['6A'] ).
verb( 'foxhunt', 'foxhunts', 'foxhunting', 'foxhunted', 'foxhunted', intran, [] ).
verb( 'fracture', 'fractures', 'fracturing', 'fractured', 'fractured', _, ['2A','6A'] ).
verb( 'fragment', 'fragments', 'fragmenting', 'fragmented', 'fragmented', intran, ['2A'] ).
verb( 'frame', 'frames', 'framing', 'framed', 'framed', _, ['2A','2C','6A'] ).
verb( 'frank', 'franks', 'franking', 'franked', 'franked', tran, ['6A'] ).
verb( 'fraternize', 'fraternizes', 'fraternizing', 'fraternized', 'fraternized', intran, ['2A','2C','3A'] ).
verb( 'fray', 'frays', 'fraying', 'frayed', 'frayed', _, ['2A','2C','6A'] ).
verb( 'freak', 'freaks', 'freaking', 'freaked', 'freaked', _, ['2C','15B'] ).
verb( 'freckle', 'freckles', 'freckling', 'freckled', 'freckled', _, ['2A','6A'] ).
verb( 'free', 'frees', 'freeing', 'freed', 'freed', tran, ['6A','14'] ).
verb( 'freelance', 'freelances', 'freelancing', 'freelanced', 'freelanced', intran, [] ).
verb( 'freewheel', 'freewheels', 'freewheeling', 'freewheeled', 'freewheeled', intran, ['2A','2C'] ).
verb( 'freeze', 'freezes', 'freezing', 'froze', 'frozen', _, ['2A','2C','3A','6A','15B'] ).
verb( 'freight', 'freights', 'freighting', 'freighted', 'freighted', tran, ['6A','14'] ).
verb( 'frequent', 'frequents', 'frequenting', 'frequented', 'frequented', tran, ['6A'] ).
verb( 'fresco', 'frescos', 'frescoing', 'frescoed', 'frescoed', tran, [] ).
verb( 'freshen', 'freshens', 'freshening', 'freshened', 'freshened', _, ['2A','2C','6A','15B'] ).
verb( 'fret', 'frets', 'fretting', 'fretted', 'fretted', _, ['2A','2C','3A','6A','15A'] ).
verb( 'fricassee', 'fricassees', 'fricasseeing', 'fricasseed', 'fricasseed', tran, [] ).
verb( 'fright', 'frights', 'frighting', 'frighted', 'frighted', tran, [] ).
verb( 'frighten', 'frightens', 'frightening', 'frightened', 'frightened', tran, ['6A','14','15B'] ).
verb( 'fringe', 'fringes', 'fringing', 'fringed', 'fringed', tran, ['6A'] ).
verb( 'frisk', 'frisks', 'frisking', 'frisked', 'frisked', _, ['2A','2C','6A'] ).
verb( 'fritter', 'fritters', 'frittering', 'frittered', 'frittered', tran, ['15B'] ).
verb( 'frivol', 'frivols', 'frivolling', 'frivolled', 'frivolled', _, ['2A','15B'] ).
verb( 'frizz', 'frizzes', 'frizzing', 'frizzed', 'frizzed', tran, ['6A'] ).
verb( 'frizzle', 'frizzles', 'frizzling', 'frizzled', 'frizzled', _, ['2A','2C','6A','15B'] ).
verb( 'frogmarch', 'frogmarches', 'frogmarching', 'frogmarched', 'frogmarched', tran, ['6A'] ).
verb( 'frolic', 'frolics', 'frolicking', 'frolicked', 'frolicked', intran, ['2A','2C'] ).
verb( 'front', 'fronts', 'fronting', 'fronted', 'fronted', _, ['2A','2C','6A'] ).
verb( 'frost', 'frosts', 'frosting', 'frosted', 'frosted', _, ['2A','2C','6A'] ).
verb( 'froth', 'froths', 'frothing', 'frothed', 'frothed', intran, ['2A','2C'] ).
verb( 'frown', 'frowns', 'frowning', 'frowned', 'frowned', intran, ['2A','3A'] ).
verb( 'fructify', 'fructifies', 'fructifying', 'fructified', 'fructified', _, ['2A','6A'] ).
verb( 'fruit', 'fruits', 'fruiting', 'fruited', 'fruited', intran, [] ).
verb( 'frustrate', 'frustrates', 'frustrating', 'frustrated', 'frustrated', tran, ['6A','15A'] ).
verb( 'fry', 'fries', 'frying', 'fried', 'fried', _, ['2A','6A'] ).
verb( 'fuck', 'fucks', 'fucking', 'fucked', 'fucked', _, ['2A','6A'] ).
verb( 'fuddle', 'fuddles', 'fuddling', 'fuddled', 'fuddled', tran, ['6A','15A'] ).
verb( 'fuel', 'fuels', 'fuelling', 'fuelled', 'fuelled', _, ['2A','6A'] ).
verb( 'fulfil', 'fulfils', 'fulfilling', 'fulfilled', 'fulfilled', tran, ['6A'] ).
verb( 'fulminate', 'fulminates', 'fulminating', 'fulminated', 'fulminated', intran, ['2A','3A'] ).
verb( 'fumble', 'fumbles', 'fumbling', 'fumbled', 'fumbled', _, ['2A','2C','6A'] ).
verb( 'fume', 'fumes', 'fuming', 'fumed', 'fumed', _, ['2A','2C','3A','6A'] ).
verb( 'fumigate', 'fumigates', 'fumigating', 'fumigated', 'fumigated', tran, ['6A'] ).
verb( 'function', 'functions', 'functioning', 'functioned', 'functioned', intran, ['2A','2C'] ).
verb( 'fund', 'funds', 'funding', 'funded', 'funded', tran, ['6A'] ).
verb( 'funk', 'funks', 'funking', 'funked', 'funked', _, ['2A','6A'] ).
verb( 'funnel', 'funnels', 'funnelling', 'funnelled', 'funnelled', _, ['2A','6A'] ).
verb( 'furbish', 'furbishes', 'furbishing', 'furbished', 'furbished', tran, ['6A'] ).
verb( 'furl', 'furls', 'furling', 'furled', 'furled', _, ['2A','6A'] ).
verb( 'furnish', 'furnishes', 'furnishing', 'furnished', 'furnished', tran, ['6A','14'] ).
verb( 'furrow', 'furrows', 'furrowing', 'furrowed', 'furrowed', tran, ['6A'] ).
verb( 'further', 'furthers', 'furthering', 'furthered', 'furthered', tran, ['6A'] ).
verb( 'fuse', 'fuses', 'fusing', 'fused', 'fused', _, ['2A','6A','15A','15B'] ).
verb( 'fuss', 'fusses', 'fussing', 'fussed', 'fussed', _, ['2A','2C','6A'] ).
verb( 'gabble', 'gabbles', 'gabbling', 'gabbled', 'gabbled', _, ['2A','2C','6A','15B'] ).
verb( 'gad', 'gads', 'gadding', 'gadded', 'gadded', intran, ['2C'] ).
verb( 'gag', 'gags', 'gagging', 'gagged', 'gagged', _, ['2A','6A'] ).
verb( 'gage', 'gages', 'gaging', 'gaged', 'gaged', tran, ['6A'] ).
verb( 'gain', 'gains', 'gaining', 'gained', 'gained', _, ['2A','2B','2C','3A','6A','12B','13B','14'] ).
verb( 'gainsay', 'gainsays', 'gainsaying', 'gainsaid', 'gainsaid', tran, ['6A'] ).
verb( 'gall', 'galls', 'galling', 'galled', 'galled', tran, ['6A'] ).
verb( 'gallivant', 'gallivants', 'gallivanting', 'gallivanted', 'gallivanted', intran, ['2C'] ).
verb( 'gallop', 'gallops', 'galloping', 'galloped', 'galloped', _, ['2A','2B','2C','6A'] ).
verb( 'galumph', 'galumphs', 'galumphing', 'galumphed', 'galumphed', intran, [] ).
verb( 'galvanize', 'galvanizes', 'galvanizing', 'galvanized', 'galvanized', tran, ['6A','14'] ).
verb( 'gamble', 'gambles', 'gambling', 'gambled', 'gambled', _, ['2A','2B','2C','15B'] ).
verb( 'gambol', 'gambols', 'gambolling', 'gambolled', 'gambolled', intran, ['2A','2C'] ).
verb( 'game', 'games', 'gaming', 'gamed', 'gamed', _, ['2A','2C','15B'] ).
verb( 'gang', 'gangs', 'ganging', 'ganged', 'ganged', intran, ['2C'] ).
verb( 'gangrene', 'gangrenes', 'gangrening', 'gangrened', 'gangrened', _, ['2A','6A'] ).
verb( 'gaol', 'gaols', 'gaoling', 'gaoled', 'gaoled', tran, ['6A'] ).
verb( 'gape', 'gapes', 'gaping', 'gaped', 'gaped', intran, ['2A','2C'] ).
verb( 'garage', 'garages', 'garaging', 'garaged', 'garaged', tran, ['6A'] ).
verb( 'garb', 'garbs', 'garbing', 'garbed', 'garbed', tran, ['6A'] ).
verb( 'garble', 'garbles', 'garbling', 'garbled', 'garbled', tran, ['6A'] ).
verb( 'garden', 'gardens', 'gardening', 'gardened', 'gardened', intran, ['2A'] ).
verb( 'gargle', 'gargles', 'gargling', 'gargled', 'gargled', _, ['2A','6A'] ).
verb( 'garland', 'garlands', 'garlanding', 'garlanded', 'garlanded', tran, ['6A'] ).
verb( 'garner', 'garners', 'garnering', 'garnered', 'garnered', tran, ['6A','15B'] ).
verb( 'garnish', 'garnishes', 'garnishing', 'garnished', 'garnished', tran, ['6A','14'] ).
verb( 'garotte', 'garottes', 'garotting', 'garotted', 'garotted', tran, ['6A'] ).
verb( 'garrison', 'garrisons', 'garrisoning', 'garrisoned', 'garrisoned', tran, ['6A'] ).
verb( 'garrotte', 'garrottes', 'garrotting', 'garrotted', 'garrotted', tran, ['6A'] ).
verb( 'gas', 'gasses', 'gassing', 'gassed', 'gassed', _, ['2A','2C','6A'] ).
verb( 'gash', 'gashes', 'gashing', 'gashed', 'gashed', tran, ['6A'] ).
verb( 'gasify', 'gasifies', 'gasifying', 'gasified', 'gasified', _, ['2A','6A'] ).
verb( 'gasp', 'gasps', 'gasping', 'gasped', 'gasped', _, ['2A','2C','6A','15B'] ).
verb( 'gate', 'gates', 'gating', 'gated', 'gated', tran, ['6A'] ).
verb( 'gatecrash', 'gatecrashes', 'gatecrashing', 'gatecrashed', 'gatecrashed', tran, ['6A'] ).
verb( 'gather', 'gathers', 'gathering', 'gathered', 'gathered', _, ['2A','2C','6A','9','12B','13B','15A','15B'] ).
verb( 'gauge', 'gauges', 'gauging', 'gauged', 'gauged', tran, ['6A'] ).
verb( 'gawp', 'gawps', 'gawping', 'gawped', 'gawped', intran, ['2A','3A'] ).
verb( 'gaze', 'gazes', 'gazing', 'gazed', 'gazed', intran, ['2A','2C','3A'] ).
verb( 'gazette', 'gazettes', 'gazetting', 'gazetted', 'gazetted', tran, [] ).
verb( 'gazump', 'gazumps', 'gazumping', 'gazumped', 'gazumped', _, ['2A','6A'] ).
verb( 'gear', 'gears', 'gearing', 'geared', 'geared', _, ['2A','2C','3A','14','15A'] ).
verb( 'gel', 'gels', 'gelling', 'gelled', 'gelled', intran, [] ).
verb( 'geld', 'gelds', 'gelding', 'gelded', 'gelded', tran, ['6A'] ).
verb( 'gen', 'gens', 'genning', 'genned', 'genned', tran, ['15B'] ).
verb( 'generalize', 'generalizes', 'generalizing', 'generalized', 'generalized', _, ['2A','3A','6A','14'] ).
verb( 'generate', 'generates', 'generating', 'generated', 'generated', tran, ['6A'] ).
verb( 'genuflect', 'genuflects', 'genuflecting', 'genuflected', 'genuflected', intran, ['2A'] ).
verb( 'germinate', 'germinates', 'germinating', 'germinated', 'germinated', _, ['2A','6A'] ).
verb( 'gerrymander', 'gerrymanders', 'gerrymandering', 'gerrymandered', 'gerrymandered', tran, ['6A'] ).
verb( 'gesticulate', 'gesticulates', 'gesticulating', 'gesticulated', 'gesticulated', intran, ['2A'] ).
verb( 'gesture', 'gestures', 'gesturing', 'gestured', 'gestured', intran, ['2A'] ).
verb( 'get', 'gets', 'getting', 'got', 'got', _, ['2C','2D','2E','3A','4A','6A','7A','7B','12B','13B','14','15A','15B','17','19B','22','24C'] ).
verb( 'ghost', 'ghosts', 'ghosting', 'ghosted', 'ghosted', _, [] ).
verb( 'gibber', 'gibbers', 'gibbering', 'gibbered', 'gibbered', intran, ['2A','2C'] ).
verb( 'gibbet', 'gibbets', 'gibbeting', 'gibbeted', 'gibbeted', tran, [] ).
verb( 'gibe', 'gibes', 'gibing', 'gibed', 'gibed', intran, ['2A','3A'] ).
verb( 'gift', 'gifts', 'gifting', 'gifted', 'gifted', tran, ['6A'] ).
verb( 'giggle', 'giggles', 'giggling', 'giggled', 'giggled', intran, ['2A','6A'] ).
verb( 'gild', 'gilds', 'gilding', 'gilded', 'gilded', tran, ['6A'] ).
verb( 'gin', 'gins', 'ginning', 'ginned', 'ginned', tran, ['6A'] ).
verb( 'ginger', 'gingers', 'gingering', 'gingered', 'gingered', tran, ['6A','15B'] ).
verb( 'gird', 'girds', 'girding', 'girded', 'girded', tran, ['15B'] ).
verb( 'girdle', 'girdles', 'girdling', 'girdled', 'girdled', tran, ['15A','15B'] ).
verb( 'give', 'gives', 'giving', 'gave', 'given', _, ['2A','2C','3A','6A','12A','12B','13A','15B','16A'] ).
verb( 'gladden', 'gladdens', 'gladdening', 'gladdened', 'gladdened', tran, ['6A'] ).
verb( 'glamorize', 'glamorizes', 'glamorizing', 'glamorized', 'glamorized', tran, ['6A'] ).
verb( 'glance', 'glances', 'glancing', 'glanced', 'glanced', _, ['2C','3A','15A'] ).
verb( 'glare', 'glares', 'glaring', 'glared', 'glared', _, ['2A','2C','3A','6A','14'] ).
verb( 'glass', 'glasses', 'glassing', 'glassed', 'glassed', tran, ['6A','15B'] ).
verb( 'glaze', 'glazes', 'glazing', 'glazed', 'glazed', _, ['2A','2C','6A','15B'] ).
verb( 'gleam', 'gleams', 'gleaming', 'gleamed', 'gleamed', intran, ['2A','2C'] ).
verb( 'glean', 'gleans', 'gleaning', 'gleaned', 'gleaned', _, ['2A','6A'] ).
verb( 'glide', 'glides', 'gliding', 'glided', 'glided', intran, ['2A','2C'] ).
verb( 'glimmer', 'glimmers', 'glimmering', 'glimmered', 'glimmered', intran, ['2A','2C'] ).
verb( 'glimpse', 'glimpses', 'glimpsing', 'glimpsed', 'glimpsed', tran, ['6A','19A'] ).
verb( 'glint', 'glints', 'glinting', 'glinted', 'glinted', intran, [] ).
verb( 'glissade', 'glissades', 'glissading', 'glissaded', 'glissaded', intran, [] ).
verb( 'glisten', 'glistens', 'glistening', 'glistened', 'glistened', intran, ['2A','2C'] ).
verb( 'glister', 'glisters', 'glistering', 'glistered', 'glistered', intran, [] ).
verb( 'glitter', 'glitters', 'glittering', 'glittered', 'glittered', intran, ['2A','2C'] ).
verb( 'gloat', 'gloats', 'gloating', 'gloated', 'gloated', intran, ['2A','3A'] ).
verb( 'globetrot', 'globetrots', 'globetrotting', 'globetrotted', 'globetrotted', intran, ['2A'] ).
verb( 'glorify', 'glorifies', 'glorifying', 'glorified', 'glorified', tran, ['6A'] ).
verb( 'glory', 'glories', 'glorying', 'gloried', 'gloried', intran, ['3A'] ).
verb( 'gloss', 'glosses', 'glossing', 'glossed', 'glossed', tran, ['6A','15A'] ).
verb( 'glow', 'glows', 'glowing', 'glowed', 'glowed', intran, ['2A','2C'] ).
verb( 'glower', 'glowers', 'glowering', 'glowered', 'glowered', intran, ['2A','3A'] ).
verb( 'glue', 'glues', 'gluing', 'glued', 'glued', tran, ['6A','15A','15B'] ).
verb( 'glut', 'gluts', 'glutting', 'glutted', 'glutted', tran, ['6A','14'] ).
verb( 'gnash', 'gnashes', 'gnashing', 'gnashed', 'gnashed', _, ['2A','6A'] ).
verb( 'gnaw', 'gnaws', 'gnawing', 'gnawed', 'gnawed', _, ['3A','6A','15B'] ).
verb( 'go', 'goes', 'going', 'went', 'gone', intran, ['2A','2B','2C','2D','2E','3A','4A','6A','15B'] ).
verb( 'goad', 'goads', 'goading', 'goaded', 'goaded', tran, ['6A','14','15B','17'] ).
verb( 'gobble', 'gobbles', 'gobbling', 'gobbled', 'gobbled', _, ['2A','2C','6A','15B'] ).
verb( 'goggle', 'goggles', 'goggling', 'goggled', 'goggled', intran, ['2A','3A'] ).
verb( 'golf', 'golfs', 'golfing', 'golfed', 'golfed', intran, [] ).
verb( 'gong', 'gongs', 'gonging', 'gonged', 'gonged', tran, [] ).
verb( 'goof', 'goofs', 'goofing', 'goofed', 'goofed', _, ['2A','6A'] ).
verb( 'gore', 'gores', 'goring', 'gored', 'gored', tran, ['6A'] ).
verb( 'gorge', 'gorges', 'gorging', 'gorged', 'gorged', _, ['2A','2C','6A','14'] ).
verb( 'gormandize', 'gormandizes', 'gormandizing', 'gormandized', 'gormandized', intran, [] ).
verb( 'gossip', 'gossips', 'gossiping', 'gossiped', 'gossiped', intran, ['2A','2C'] ).
verb( 'gouge', 'gouges', 'gouging', 'gouged', 'gouged', tran, ['6A','15B'] ).
verb( 'govern', 'governs', 'governing', 'governed', 'governed', _, ['2A','6A'] ).
verb( 'gown', 'gowns', 'gowning', 'gowned', 'gowned', tran, [] ).
verb( 'grab', 'grabs', 'grabbing', 'grabbed', 'grabbed', _, ['3A','6A'] ).
verb( 'grace', 'graces', 'gracing', 'graced', 'graced', tran, ['6A'] ).
verb( 'grade', 'grades', 'grading', 'graded', 'graded', tran, ['6A','15B'] ).
verb( 'graduate', 'graduates', 'graduating', 'graduated', 'graduated', _, ['2A','2C','6A'] ).
verb( 'graft', 'grafts', 'grafting', 'grafted', 'grafted', _, ['2A','6A','15A','15B'] ).
verb( 'grant', 'grants', 'granting', 'granted', 'granted', tran, ['6A','9','12A','13A','25'] ).
verb( 'granulate', 'granulates', 'granulating', 'granulated', 'granulated', _, ['2A','6A'] ).
verb( 'grapple', 'grapples', 'grappling', 'grappled', 'grappled', intran, ['2A','2C','3A'] ).
verb( 'grasp', 'grasps', 'grasping', 'grasped', 'grasped', _, ['3A','6A'] ).
verb( 'grass', 'grasses', 'grassing', 'grassed', 'grassed', _, ['2A','3A','6A','15B'] ).
verb( 'grate', 'grates', 'grating', 'grated', 'grated', _, ['2A','3A','6A','15A'] ).
verb( 'gratify', 'gratifies', 'gratifying', 'gratified', 'gratified', tran, ['6A'] ).
verb( 'grave', 'graves', 'graving', 'graved', 'graven', tran, [] ).
verb( 'gravel', 'gravels', 'gravelling', 'gravelled', 'gravelled', tran, ['6A'] ).
verb( 'gravitate', 'gravitates', 'gravitating', 'gravitated', 'gravitated', intran, ['3A'] ).
verb( 'gray', 'grays', 'graying', 'grayed', 'grayed', _, ['2A','6A'] ).
verb( 'graze', 'grazes', 'grazing', 'grazed', 'grazed', _, ['2A','2C','6A'] ).
verb( 'grease', 'greases', 'greasing', 'greased', 'greased', tran, [] ).
verb( 'greet', 'greets', 'greeting', 'greeted', 'greeted', tran, ['6A','14'] ).
verb( 'grey', 'greys', 'greying', 'greyed', 'greyed', _, ['2A','6A'] ).
verb( 'grieve', 'grieves', 'grieving', 'grieved', 'grieved', _, ['2A','2C','6A'] ).
verb( 'grill', 'grills', 'grilling', 'grilled', 'grilled', _, ['2A','2C','6A'] ).
verb( 'grimace', 'grimaces', 'grimacing', 'grimaced', 'grimaced', intran, ['2A'] ).
verb( 'grime', 'grimes', 'griming', 'grimed', 'grimed', tran, ['6A'] ).
verb( 'grin', 'grins', 'grinning', 'grinned', 'grinned', _, ['2A','2C','6A'] ).
verb( 'grind', 'grinds', 'grinding', 'ground', 'ground', _, ['2A','2C','6A','15A','15B'] ).
verb( 'grip', 'grips', 'gripping', 'gripped', 'gripped', _, ['2A','6A'] ).
verb( 'grit', 'grits', 'gritting', 'gritted', 'gritted', tran, [] ).
verb( 'grizzle', 'grizzles', 'grizzling', 'grizzled', 'grizzled', intran, [] ).
verb( 'groan', 'groans', 'groaning', 'groaned', 'groaned', _, ['2A','2C','6A','15B'] ).
verb( 'groin', 'groins', 'groining', 'groined', 'groined', tran, [] ).
verb( 'groom', 'grooms', 'grooming', 'groomed', 'groomed', tran, ['6A'] ).
verb( 'groove', 'grooves', 'grooving', 'grooved', 'grooved', tran, [] ).
verb( 'grope', 'gropes', 'groping', 'groped', 'groped', _, ['2A','2C','3A','15A'] ).
verb( 'gross', 'grosses', 'grossing', 'grossed', 'grossed', tran, ['6A'] ).
verb( 'grouch', 'grouches', 'grouching', 'grouched', 'grouched', intran, [] ).
verb( 'ground', 'grounds', 'grounding', 'grounded', 'grounded', _, ['2A','2C','6A','14','15A','15B'] ).
verb( 'group', 'groups', 'grouping', 'grouped', 'grouped', _, ['2C','6A','15A','15B'] ).
verb( 'grouse', 'grouses', 'grousing', 'groused', 'groused', intran, ['2A','2C'] ).
verb( 'grovel', 'grovels', 'grovelling', 'grovelled', 'grovelled', intran, ['2A','2C'] ).
verb( 'grow', 'grows', 'growing', 'grew', 'grown', _, ['2A','2C','2D','3A','4A','6A','12B','13B'] ).
verb( 'growl', 'growls', 'growling', 'growled', 'growled', _, ['2A','2C','6A','15B'] ).
verb( 'grub', 'grubs', 'grubbing', 'grubbed', 'grubbed', _, ['2C','6A','15B'] ).
verb( 'grudge', 'grudges', 'grudging', 'grudged', 'grudged', tran, ['6C','12A','13A'] ).
verb( 'grumble', 'grumbles', 'grumbling', 'grumbled', 'grumbled', _, ['2A','2C','3A','6A','15B'] ).
verb( 'grunt', 'grunts', 'grunting', 'grunted', 'grunted', _, ['2A','6A','15B'] ).
verb( 'guarantee', 'guarantees', 'guaranteeing', 'guaranteed', 'guaranteed', tran, ['6A','7A','9','12A','13A','25'] ).
verb( 'guard', 'guards', 'guarding', 'guarded', 'guarded', _, ['3A','6A','15A'] ).
verb( 'guess', 'guesses', 'guessing', 'guessed', 'guessed', _, ['2A','2C','3A','6A','8','9','10','25'] ).
verb( 'guffaw', 'guffaws', 'guffawing', 'guffawed', 'guffawed', intran, [] ).
verb( 'guide', 'guides', 'guiding', 'guided', 'guided', tran, ['6A','15A','15B'] ).
verb( 'guillotine', 'guillotines', 'guillotining', 'guillotined', 'guillotined', tran, ['6A'] ).
verb( 'gull', 'gulls', 'gulling', 'gulled', 'gulled', tran, ['6A','15A'] ).
verb( 'gulp', 'gulps', 'gulping', 'gulped', 'gulped', _, ['2A','6A','15B'] ).
verb( 'gum', 'gums', 'gumming', 'gummed', 'gummed', tran, ['6A','15A','15B'] ).
verb( 'gun', 'guns', 'gunning', 'gunned', 'gunned', tran, ['6A','15B'] ).
verb( 'gurgle', 'gurgles', 'gurgling', 'gurgled', 'gurgled', intran, [] ).
verb( 'gush', 'gushes', 'gushing', 'gushed', 'gushed', intran, ['2A','2C','3A'] ).
verb( 'gut', 'guts', 'gutting', 'gutted', 'gutted', tran, ['6A'] ).
verb( 'gutter', 'gutters', 'guttering', 'guttered', 'guttered', intran, ['2A'] ).
verb( 'guy', 'guys', 'guying', 'guyed', 'guyed', tran, ['6A'] ).
verb( 'guzzle', 'guzzles', 'guzzling', 'guzzled', 'guzzled', _, ['2A','6A','15B'] ).
verb( 'gybe', 'gybes', 'gybing', 'gybed', 'gybed', _, ['2A','6A'] ).
verb( 'gyp', 'gyps', 'gypping', 'gypped', 'gypped', tran, ['6A'] ).
verb( 'gyrate', 'gyrates', 'gyrating', 'gyrated', 'gyrated', intran, [] ).
verb( 'h\'m', 'h\'ms', 'h\'mming', 'h\'mmed', 'h\'mmed', intran, [] ).
verb( 'habituate', 'habituates', 'habituating', 'habituated', 'habituated', tran, ['14'] ).
verb( 'hack', 'hacks', 'hacking', 'hacked', 'hacked', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'haggle', 'haggles', 'haggling', 'haggled', 'haggled', intran, ['2A','3A'] ).
verb( 'hail', 'hails', 'hailing', 'hailed', 'hailed', _, ['2C','3A','6A','15B','16B','23'] ).
verb( 'halloo', 'halloos', 'hallooing', 'hallooed', 'hallooed', intran, [] ).
verb( 'hallow', 'hallows', 'hallowing', 'hallowed', 'hallowed', tran, ['6A'] ).
verb( 'halt', 'halts', 'halting', 'halted', 'halted', _, ['2A','2C','6A'] ).
verb( 'halve', 'halves', 'halving', 'halved', 'halved', tran, ['6A'] ).
verb( 'ham', 'hams', 'hamming', 'hammed', 'hammed', _, ['2A','6A','15B'] ).
verb( 'hammer', 'hammers', 'hammering', 'hammered', 'hammered', _, ['2A','2C','3A','6A','15B','22'] ).
verb( 'hamper', 'hampers', 'hampering', 'hampered', 'hampered', tran, ['6A'] ).
verb( 'hamstring', 'hamstrings', 'hamstringing', 'hamstrung', 'hamstrung', tran, ['6A'] ).
verb( 'hand', 'hands', 'handing', 'handed', 'handed', tran, ['12A','13A','15A','15B'] ).
verb( 'handcuff', 'handcuffs', 'handcuffing', 'handcuffed', 'handcuffed', tran, [] ).
verb( 'handicap', 'handicaps', 'handicapping', 'handicapped', 'handicapped', tran, ['6A'] ).
verb( 'handle', 'handles', 'handling', 'handled', 'handled', tran, ['6A'] ).
verb( 'hang', 'hangs', 'hanging', 'hanged', 'hanged', _, ['2A','2B','2C','3A','6A','15A','15B'] ).
verb( 'hanker', 'hankers', 'hankering', 'hankered', 'hankered', intran, ['3A'] ).
verb( 'hap', 'haps', 'happing', 'happed', 'happed', intran, [] ).
verb( 'happen', 'happens', 'happening', 'happened', 'happened', intran, ['2A','3A','4E'] ).
verb( 'harangue', 'harangues', 'haranguing', 'harangued', 'harangued', _, ['2A','6A'] ).
verb( 'harass', 'harasses', 'harassing', 'harassed', 'harassed', tran, ['6A'] ).
verb( 'harbour', 'harbours', 'harbouring', 'harboured', 'harboured', _, ['2A','6A'] ).
verb( 'harden', 'hardens', 'hardening', 'hardened', 'hardened', _, ['2A','2C','6A','15B'] ).
verb( 'hare', 'hares', 'haring', 'hared', 'hared', intran, ['2C'] ).
verb( 'hark', 'harks', 'harking', 'harked', 'harked', intran, [] ).
verb( 'harm', 'harms', 'harming', 'harmed', 'harmed', tran, ['6A'] ).
verb( 'harmonize', 'harmonizes', 'harmonizing', 'harmonized', 'harmonized', _, ['2A','2C','6A','14'] ).
verb( 'harness', 'harnesses', 'harnessing', 'harnessed', 'harnessed', tran, ['6A'] ).
verb( 'harp', 'harps', 'harping', 'harped', 'harped', intran, [] ).
verb( 'harpoon', 'harpoons', 'harpooning', 'harpooned', 'harpooned', tran, [] ).
verb( 'harrow', 'harrows', 'harrowing', 'harrowed', 'harrowed', tran, ['6A'] ).
verb( 'harry', 'harries', 'harrying', 'harried', 'harried', tran, ['6A'] ).
verb( 'harvest', 'harvests', 'harvesting', 'harvested', 'harvested', tran, ['6A'] ).
verb( 'hash', 'hashes', 'hashing', 'hashed', 'hashed', tran, ['6A','15B'] ).
verb( 'hassle', 'hassles', 'hassling', 'hassled', 'hassled', _, ['2A','3A','6A'] ).
verb( 'hasten', 'hastens', 'hastening', 'hastened', 'hastened', _, ['2A','2C','4A','6A'] ).
verb( 'hatch', 'hatches', 'hatching', 'hatched', 'hatched', _, ['2A','6A'] ).
verb( 'hate', 'hates', 'hating', 'hated', 'hated', tran, ['6A','6D','7A','17','19C'] ).
verb( 'haul', 'hauls', 'hauling', 'hauled', 'hauled', _, ['2C','3A','6A','15A','15B'] ).
verb( 'haunt', 'haunts', 'haunting', 'haunted', 'haunted', tran, ['6A'] ).
verb( 'have', 'has', 'having', 'had', 'had', unknown, ['6A','6B','7B','15B','18C','19B','24B','24C'] ).
verb( 'have', 'has', 'having', 'had', 'had', _, ['6A','6B','7B','15B','18C','19B','24B','24C'] ).
verb( 'haw', 'haws', 'hawing', 'hawed', 'hawed', intran, [] ).
verb( 'hawk', 'hawks', 'hawking', 'hawked', 'hawked', tran, ['6A','15B'] ).
verb( 'hazard', 'hazards', 'hazarding', 'hazarded', 'hazarded', tran, ['6A'] ).
verb( 'haze', 'hazes', 'hazing', 'hazed', 'hazed', tran, [] ).
verb( 'head', 'heads', 'heading', 'headed', 'headed', _, ['2C','6A','15B'] ).
verb( 'heal', 'heals', 'healing', 'healed', 'healed', _, ['2A','2C','6A'] ).
verb( 'heap', 'heaps', 'heaping', 'heaped', 'heaped', tran, ['6A','14','15A','15B'] ).
verb( 'hear', 'hears', 'hearing', 'heard', 'heard', _, ['2A','3A','6A','9','10','15A','18A','19A','24A'] ).
verb( 'hearken', 'hearkens', 'hearkening', 'hearkened', 'hearkened', intran, [] ).
verb( 'hearten', 'heartens', 'heartening', 'heartened', 'heartened', tran, ['6A'] ).
verb( 'heat', 'heats', 'heating', 'heated', 'heated', _, ['2C','6A','15B'] ).
verb( 'heave', 'heaves', 'heaving', 'heaved', 'heaved', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'heckle', 'heckles', 'heckling', 'heckled', 'heckled', tran, ['6A'] ).
verb( 'hector', 'hectors', 'hectoring', 'hectored', 'hectored', _, [] ).
verb( 'hedge', 'hedges', 'hedging', 'hedged', 'hedged', _, ['2A','6A','15A','15B'] ).
verb( 'hedgehop', 'hedgehops', 'hedgehopping', 'hedgehopped', 'hedgehopped', intran, [] ).
verb( 'heed', 'heeds', 'heeding', 'heeded', 'heeded', tran, ['6A'] ).
verb( 'heel', 'heels', 'heeling', 'heeled', 'heeled', _, ['2A','2C','6A','15B'] ).
verb( 'heighten', 'heightens', 'heightening', 'heightened', 'heightened', _, ['2A','6A'] ).
verb( 'heliograph', 'heliographs', 'heliographing', 'heliographed', 'heliographed', tran, [] ).
verb( 'help', 'helps', 'helping', 'helped', 'helped', _, ['2A','2C','6A','6C','14','15A','15B','17','18B'] ).
verb( 'hem', 'hems', 'hemming', 'hemmed', 'hemmed', _, ['6A','15B'] ).
verb( 'hemstitch', 'hemstitches', 'hemstitching', 'hemstitched', 'hemstitched', tran, [] ).
verb( 'herald', 'heralds', 'heralding', 'heralded', 'heralded', tran, ['6A'] ).
verb( 'herd', 'herds', 'herding', 'herded', 'herded', _, ['2C','15A','15B'] ).
verb( 'hesitate', 'hesitates', 'hesitating', 'hesitated', 'hesitated', intran, ['2A','3A','3B','4C'] ).
verb( 'hew', 'hews', 'hewing', 'hewed', 'hewed', _, ['2A','2C','6A','15A','15B'] ).
verb( 'hibernate', 'hibernates', 'hibernating', 'hibernated', 'hibernated', intran, ['2A'] ).
verb( 'hiccough', 'hiccoughs', 'hiccoughing', 'hiccoughed', 'hiccoughed', intran, [] ).
verb( 'hiccup', 'hiccups', 'hiccupping', 'hiccupped', 'hiccupped', intran, [] ).
verb( 'hide', 'hides', 'hiding', 'hid', 'hidden', _, ['2A','6A','14'] ).
verb( 'hie', 'hies', 'hieing', 'hied', 'hied', intran, [] ).
verb( 'highjack', 'highjacks', 'highjacking', 'highjacked', 'highjacked', tran, ['6A'] ).
verb( 'highlight', 'highlights', 'highlighting', 'highlighted', 'highlighted', tran, [] ).
verb( 'hijack', 'hijacks', 'hijacking', 'hijacked', 'hijacked', tran, ['6A'] ).
verb( 'hike', 'hikes', 'hiking', 'hiked', 'hiked', intran, [] ).
verb( 'hinder', 'hinders', 'hindering', 'hindered', 'hindered', tran, ['6A','6C','15A'] ).
verb( 'hinge', 'hinges', 'hinging', 'hinged', 'hinged', _, ['3A','6A'] ).
verb( 'hint', 'hints', 'hinting', 'hinted', 'hinted', _, ['3A','6A','9'] ).
verb( 'hire', 'hires', 'hiring', 'hired', 'hired', tran, ['6A','15B'] ).
verb( 'hiss', 'hisses', 'hissing', 'hissed', 'hissed', _, ['2A','3A','6A','15A'] ).
verb( 'hit', 'hits', 'hitting', 'hit', 'hit', _, ['2C','3A','6A','12C','15A','15B'] ).
verb( 'hitch', 'hitches', 'hitching', 'hitched', 'hitched', _, ['2A','2C','6A','15A','15B'] ).
verb( 'hitchhike', 'hitchhikes', 'hitchhiking', 'hitchhiked', 'hitchhiked', intran, ['2A'] ).
verb( 'hive', 'hives', 'hiving', 'hived', 'hived', _, ['2C','6A'] ).
verb( 'hoard', 'hoards', 'hoarding', 'hoarded', 'hoarded', _, ['6A','15B'] ).
verb( 'hoax', 'hoaxes', 'hoaxing', 'hoaxed', 'hoaxed', tran, ['6A','14'] ).
verb( 'hobble', 'hobbles', 'hobbling', 'hobbled', 'hobbled', _, ['2A','2C','6A'] ).
verb( 'hobnob', 'hobnobs', 'hobnobbing', 'hobnobbed', 'hobnobbed', intran, ['2A','2C','3A'] ).
verb( 'hock', 'hocks', 'hocking', 'hocked', 'hocked', tran, ['6A'] ).
verb( 'hoe', 'hoes', 'hoeing', 'hoed', 'hoed', _, ['2A','6A','15B'] ).
verb( 'hog', 'hogs', 'hogging', 'hogged', 'hogged', tran, ['6A'] ).
verb( 'hoist', 'hoists', 'hoisting', 'hoisted', 'hoisted', tran, ['6A','15B'] ).
verb( 'hold', 'holds', 'holding', 'held', 'held', _, ['2A','2C','2D','3A','6A','9','14','15A','15B','22','25'] ).
verb( 'hole', 'holes', 'holing', 'holed', 'holed', _, ['2C','6A','15B'] ).
verb( 'holiday', 'holidays', 'holidaying', 'holidayed', 'holidayed', intran, [] ).
verb( 'holler', 'hollers', 'hollering', 'hollered', 'hollered', _, [] ).
verb( 'hollow', 'hollows', 'hollowing', 'hollowed', 'hollowed', tran, ['6A','15A','15B'] ).
verb( 'holystone', 'holystones', 'holystoning', 'holystoned', 'holystoned', tran, [] ).
verb( 'homogenize', 'homogenizes', 'homogenizing', 'homogenized', 'homogenized', tran, ['6A'] ).
verb( 'hone', 'hones', 'honing', 'honed', 'honed', tran, ['6A'] ).
verb( 'honeycomb', 'honeycombs', 'honeycombing', 'honeycombed', 'honeycombed', tran, ['6A'] ).
verb( 'honeymoon', 'honeymoons', 'honeymooning', 'honeymooned', 'honeymooned', intran, [] ).
verb( 'honk', 'honks', 'honking', 'honked', 'honked', intran, [] ).
verb( 'honour', 'honours', 'honouring', 'honoured', 'honoured', tran, ['6A'] ).
verb( 'hood', 'hoods', 'hooding', 'hooded', 'hooded', tran, [] ).
verb( 'hoodoo', 'hoodoos', 'hoodooing', 'hoodooed', 'hoodooed', tran, [] ).
verb( 'hoodwink', 'hoodwinks', 'hoodwinking', 'hoodwinked', 'hoodwinked', tran, ['6A','14'] ).
verb( 'hook', 'hooks', 'hooking', 'hooked', 'hooked', _, ['2C','6A','15A','15B'] ).
verb( 'hoop', 'hoops', 'hooping', 'hooped', 'hooped', tran, [] ).
verb( 'hoot', 'hoots', 'hooting', 'hooted', 'hooted', _, ['2A','2C','6A','15A','15B'] ).
verb( 'hop', 'hops', 'hopping', 'hopped', 'hopped', _, ['2A','2C','6A'] ).
verb( 'hope', 'hopes', 'hoping', 'hoped', 'hoped', _, ['2A','3A','7A','9'] ).
verb( 'horn', 'horns', 'horning', 'horned', 'horned', intran, [] ).
verb( 'horrify', 'horrifies', 'horrifying', 'horrified', 'horrified', tran, ['6A'] ).
verb( 'horsewhip', 'horsewhips', 'horsewhipping', 'horsewhipped', 'horsewhipped', tran, [] ).
verb( 'hose', 'hoses', 'hosing', 'hosed', 'hosed', tran, ['6A','15B'] ).
verb( 'hospitalize', 'hospitalizes', 'hospitalizing', 'hospitalized', 'hospitalized', tran, [] ).
verb( 'host', 'hosts', 'hosting', 'hosted', 'hosted', tran, ['6A'] ).
verb( 'hot', 'hots', 'hotting', 'hotted', 'hotted', _, ['2C','15B'] ).
verb( 'hotfoot', 'hotfoots', 'hotfooting', 'hotfooted', 'hotfooted', intran, [] ).
verb( 'hound', 'hounds', 'hounding', 'hounded', 'hounded', tran, ['6A'] ).
verb( 'house', 'houses', 'housing', 'housed', 'housed', tran, ['6A'] ).
verb( 'hover', 'hovers', 'hovering', 'hovered', 'hovered', intran, ['2A','2C'] ).
verb( 'howl', 'howls', 'howling', 'howled', 'howled', _, ['2A','2C','6A','15A','15B'] ).
verb( 'huddle', 'huddles', 'huddling', 'huddled', 'huddled', _, ['2C','15A','15B'] ).
verb( 'huff', 'huffs', 'huffing', 'huffed', 'huffed', intran, [] ).
verb( 'hug', 'hugs', 'hugging', 'hugged', 'hugged', tran, ['6A'] ).
verb( 'hull', 'hulls', 'hulling', 'hulled', 'hulled', tran, [] ).
verb( 'hum', 'hums', 'humming', 'hummed', 'hummed', _, ['2A','2C','6A'] ).
verb( 'humanize', 'humanizes', 'humanizing', 'humanized', 'humanized', _, ['2A','6A'] ).
verb( 'humble', 'humbles', 'humbling', 'humbled', 'humbled', tran, ['6A'] ).
verb( 'humbug', 'humbugs', 'humbugging', 'humbugged', 'humbugged', tran, ['6A','14'] ).
verb( 'humidify', 'humidifies', 'humidifying', 'humidified', 'humidified', tran, [] ).
verb( 'humiliate', 'humiliates', 'humiliating', 'humiliated', 'humiliated', tran, ['6A'] ).
verb( 'humour', 'humours', 'humouring', 'humoured', 'humoured', tran, ['6A'] ).
verb( 'hump', 'humps', 'humping', 'humped', 'humped', tran, ['6A','15B'] ).
verb( 'hunch', 'hunches', 'hunching', 'hunched', 'hunched', tran, ['6A','15B'] ).
verb( 'hunger', 'hungers', 'hungering', 'hungered', 'hungered', intran, ['2A','3A','4C'] ).
verb( 'hunt', 'hunts', 'hunting', 'hunted', 'hunted', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'hurdle', 'hurdles', 'hurdling', 'hurdled', 'hurdled', _, ['2A','15B'] ).
verb( 'hurl', 'hurls', 'hurling', 'hurled', 'hurled', tran, ['6A','15A','15B'] ).
verb( 'hurrah', 'hurrahs', 'hurrahing', 'hurrahed', 'hurrahed', intran, [] ).
verb( 'hurry', 'hurries', 'hurrying', 'hurried', 'hurried', _, ['2A','2C','6A','15A','15B'] ).
verb( 'hurt', 'hurts', 'hurting', 'hurt', 'hurt', _, ['2A','6A','6B'] ).
verb( 'hurtle', 'hurtles', 'hurtling', 'hurtled', 'hurtled', intran, ['2C'] ).
verb( 'husband', 'husbands', 'husbanding', 'husbanded', 'husbanded', tran, ['1','6A'] ).
verb( 'hush', 'hushes', 'hushing', 'hushed', 'hushed', _, ['2A','15A','15B'] ).
verb( 'husk', 'husks', 'husking', 'husked', 'husked', tran, [] ).
verb( 'hustle', 'hustles', 'hustling', 'hustled', 'hustled', _, ['2A','2C','6A','15A','15B'] ).
verb( 'hybridize', 'hybridizes', 'hybridizing', 'hybridized', 'hybridized', _, ['2A','6A'] ).
verb( 'hydrate', 'hydrates', 'hydrating', 'hydrated', 'hydrated', _, [] ).
verb( 'hymn', 'hymns', 'hymning', 'hymned', 'hymned', tran, [] ).
verb( 'hyphen', 'hyphens', 'hyphening', 'hyphened', 'hyphened', tran, [] ).
verb( 'hyphenate', 'hyphenates', 'hyphenating', 'hyphenated', 'hyphenated', tran, [] ).
verb( 'hypnotize', 'hypnotizes', 'hypnotizing', 'hypnotized', 'hypnotized', tran, ['6A'] ).
verb( 'hypothecate', 'hypothecates', 'hypothecating', 'hypothecated', 'hypothecated', tran, [] ).
verb( 'ice', 'ices', 'icing', 'iced', 'iced', _, ['2C','6A','15B'] ).
verb( 'ice-skate', 'ice-skates', 'ice-skating', 'ice-skated', 'ice-skated', intran, [] ).
verb( 'idealize', 'idealizes', 'idealizing', 'idealized', 'idealized', tran, ['6A'] ).
verb( 'identify', 'identifies', 'identifying', 'identified', 'identified', tran, ['3A','6A','14'] ).
verb( 'idle', 'idles', 'idling', 'idled', 'idled', _, ['2A','2C','15B'] ).
verb( 'idolize', 'idolizes', 'idolizing', 'idolized', 'idolized', tran, ['6A'] ).
verb( 'ignite', 'ignites', 'igniting', 'ignited', 'ignited', _, ['2A','6A'] ).
verb( 'ignore', 'ignores', 'ignoring', 'ignored', 'ignored', tran, ['6A'] ).
verb( 'ill-treat', 'ill-treats', 'ill-treating', 'ill-treated', 'ill-treated', tran, [] ).
verb( 'ill-use', 'ill-uses', 'ill-using', 'ill-used', 'ill-used', tran, [] ).
verb( 'illume', 'illumes', 'illuming', 'illumed', 'illumed', tran, [] ).
verb( 'illuminate', 'illuminates', 'illuminating', 'illuminated', 'illuminated', tran, ['6A'] ).
verb( 'illumine', 'illumines', 'illumining', 'illumined', 'illumined', tran, ['6A'] ).
verb( 'illustrate', 'illustrates', 'illustrating', 'illustrated', 'illustrated', tran, ['6A'] ).
verb( 'image', 'images', 'imaging', 'imaged', 'imaged', tran, ['6A'] ).
verb( 'imagine', 'imagines', 'imagining', 'imagined', 'imagined', tran, ['6A','6C','9','10','16B','19A','19C','25'] ).
verb( 'imbed', 'imbeds', 'imbedding', 'imbedded', 'imbedded', tran, [] ).
verb( 'imbibe', 'imbibes', 'imbibing', 'imbibed', 'imbibed', tran, [] ).
verb( 'imbue', 'imbues', 'imbuing', 'imbued', 'imbued', tran, ['14'] ).
verb( 'imitate', 'imitates', 'imitating', 'imitated', 'imitated', tran, ['6A'] ).
verb( 'immerse', 'immerses', 'immersing', 'immersed', 'immersed', tran, ['6A','14'] ).
verb( 'immigrate', 'immigrates', 'immigrating', 'immigrated', 'immigrated', intran, ['2A','3A'] ).
verb( 'immobilize', 'immobilizes', 'immobilizing', 'immobilized', 'immobilized', tran, ['6A'] ).
verb( 'immolate', 'immolates', 'immolating', 'immolated', 'immolated', tran, ['6A','14'] ).
verb( 'immortalize', 'immortalizes', 'immortalizing', 'immortalized', 'immortalized', tran, ['6A'] ).
verb( 'immunize', 'immunizes', 'immunizing', 'immunized', 'immunized', tran, ['6A','14'] ).
verb( 'immure', 'immures', 'immuring', 'immured', 'immured', tran, ['6A'] ).
verb( 'impact', 'impacts', 'impacting', 'impacted', 'impacted', tran, [] ).
verb( 'impair', 'impairs', 'impairing', 'impaired', 'impaired', tran, ['6A'] ).
verb( 'impale', 'impales', 'impaling', 'impaled', 'impaled', tran, ['6A','15A'] ).
verb( 'impanel', 'impanels', 'impanelling', 'impanelled', 'impanelled', tran, ['6A'] ).
verb( 'impart', 'imparts', 'imparting', 'imparted', 'imparted', tran, ['6A','14'] ).
verb( 'impeach', 'impeaches', 'impeaching', 'impeached', 'impeached', tran, ['6A','14'] ).
verb( 'impede', 'impedes', 'impeding', 'impeded', 'impeded', tran, ['6A'] ).
verb( 'impel', 'impels', 'impelling', 'impelled', 'impelled', tran, ['14','17'] ).
verb( 'impend', 'impends', 'impending', 'impended', 'impended', intran, [] ).
verb( 'imperil', 'imperils', 'imperilling', 'imperilled', 'imperilled', tran, ['6A'] ).
verb( 'impersonate', 'impersonates', 'impersonating', 'impersonated', 'impersonated', tran, ['6A'] ).
verb( 'impinge', 'impinges', 'impinging', 'impinged', 'impinged', intran, ['3A'] ).
verb( 'implant', 'implants', 'implanting', 'implanted', 'implanted', tran, ['6A','14'] ).
verb( 'implement', 'implements', 'implementing', 'implemented', 'implemented', tran, ['6A'] ).
verb( 'implicate', 'implicates', 'implicating', 'implicated', 'implicated', tran, ['6A','14'] ).
verb( 'implore', 'implores', 'imploring', 'implored', 'implored', tran, ['6A','14','17'] ).
verb( 'imply', 'implies', 'implying', 'implied', 'implied', tran, ['6A','9'] ).
verb( 'import', 'imports', 'importing', 'imported', 'imported', tran, ['6A','9','14'] ).
verb( 'importune', 'importunes', 'importuning', 'importuned', 'importuned', tran, ['6A','9','14','17'] ).
verb( 'impose', 'imposes', 'imposing', 'imposed', 'imposed', _, ['3A','14'] ).
verb( 'impound', 'impounds', 'impounding', 'impounded', 'impounded', tran, ['6A'] ).
verb( 'impoverish', 'impoverishes', 'impoverishing', 'impoverished', 'impoverished', tran, ['6A'] ).
verb( 'imprecate', 'imprecates', 'imprecating', 'imprecated', 'imprecated', tran, ['14'] ).
verb( 'impregnate', 'impregnates', 'impregnating', 'impregnated', 'impregnated', tran, ['6A','14'] ).
verb( 'impress', 'impresses', 'impressing', 'impressed', 'impressed', tran, ['6A','14'] ).
verb( 'imprint', 'imprints', 'imprinting', 'imprinted', 'imprinted', tran, ['14'] ).
verb( 'imprison', 'imprisons', 'imprisoning', 'imprisoned', 'imprisoned', tran, ['6A'] ).
verb( 'improve', 'improves', 'improving', 'improved', 'improved', _, ['2A','3A','6A'] ).
verb( 'improvise', 'improvises', 'improvising', 'improvised', 'improvised', _, ['2A','6A'] ).
verb( 'impugn', 'impugns', 'impugning', 'impugned', 'impugned', tran, ['6A'] ).
verb( 'impulse-buy', 'impulse-buys', 'impulse-buying', 'impulse-bought', 'impulse-bought', _, [] ).
verb( 'impute', 'imputes', 'imputing', 'imputed', 'imputed', tran, ['14'] ).
verb( 'inactivate', 'inactivates', 'inactivating', 'inactivated', 'inactivated', tran, [] ).
verb( 'inaugurate', 'inaugurates', 'inaugurating', 'inaugurated', 'inaugurated', tran, ['6A'] ).
verb( 'incapacitate', 'incapacitates', 'incapacitating', 'incapacitated', 'incapacitated', tran, ['6A','14'] ).
verb( 'incarcerate', 'incarcerates', 'incarcerating', 'incarcerated', 'incarcerated', tran, ['6A'] ).
verb( 'incarnate', 'incarnates', 'incarnating', 'incarnated', 'incarnated', tran, ['6A'] ).
verb( 'incense', 'incenses', 'incensing', 'incensed', 'incensed', tran, ['6A'] ).
verb( 'inch', 'inches', 'inching', 'inched', 'inched', _, ['2C','15A','15B'] ).
verb( 'incinerate', 'incinerates', 'incinerating', 'incinerated', 'incinerated', tran, ['6A'] ).
verb( 'incise', 'incises', 'incising', 'incised', 'incised', tran, ['6A'] ).
verb( 'incite', 'incites', 'inciting', 'incited', 'incited', tran, ['6A','14','17'] ).
verb( 'incline', 'inclines', 'inclining', 'inclined', 'inclined', _, ['2A','3A','4C','6A','15A','17'] ).
verb( 'inclose', 'incloses', 'inclosing', 'inclosed', 'inclosed', tran, ['6A','14'] ).
verb( 'include', 'includes', 'including', 'included', 'included', tran, ['6A','6C'] ).
verb( 'incommode', 'incommodes', 'incommoding', 'incommoded', 'incommoded', tran, ['6A'] ).
verb( 'inconvenience', 'inconveniences', 'inconveniencing', 'inconvenienced', 'inconvenienced', tran, ['6A'] ).
verb( 'incorporate', 'incorporates', 'incorporating', 'incorporated', 'incorporated', _, ['2A','3A','6A','14','23'] ).
verb( 'increase', 'increases', 'increasing', 'increased', 'increased', _, ['2A','6A'] ).
verb( 'incriminate', 'incriminates', 'incriminating', 'incriminated', 'incriminated', tran, ['6A'] ).
verb( 'incubate', 'incubates', 'incubating', 'incubated', 'incubated', _, ['2A','6A'] ).
verb( 'inculcate', 'inculcates', 'inculcating', 'inculcated', 'inculcated', tran, ['6A','14'] ).
verb( 'inculpate', 'inculpates', 'inculpating', 'inculpated', 'inculpated', tran, ['6A'] ).
verb( 'incur', 'incurs', 'incurring', 'incurred', 'incurred', tran, ['6A'] ).
verb( 'indemnify', 'indemnifies', 'indemnifying', 'indemnified', 'indemnified', tran, ['6A','14'] ).
verb( 'indent', 'indents', 'indenting', 'indented', 'indented', _, ['3A','6A'] ).
verb( 'indenture', 'indentures', 'indenturing', 'indentured', 'indentured', tran, ['6A'] ).
verb( 'index', 'indexes', 'indexing', 'indexed', 'indexed', tran, ['6A'] ).
verb( 'indicate', 'indicates', 'indicating', 'indicated', 'indicated', tran, ['6A','9','14'] ).
verb( 'indict', 'indicts', 'indicting', 'indicted', 'indicted', tran, ['6A','14','16B'] ).
verb( 'indite', 'indites', 'inditing', 'indited', 'indited', tran, ['6A'] ).
verb( 'individualize', 'individualizes', 'individualizing', 'individualized', 'individualized', tran, ['6A'] ).
verb( 'indoctrinate', 'indoctrinates', 'indoctrinating', 'indoctrinated', 'indoctrinated', tran, ['6A','14'] ).
verb( 'indorse', 'indorses', 'indorsing', 'indorsed', 'indorsed', tran, ['6A'] ).
verb( 'induce', 'induces', 'inducing', 'induced', 'induced', tran, ['6A','14','17'] ).
verb( 'induct', 'inducts', 'inducting', 'inducted', 'inducted', tran, ['6A','14','16B'] ).
verb( 'indue', 'indues', 'induing', 'indued', 'indued', tran, ['14'] ).
verb( 'indulge', 'indulges', 'indulging', 'indulged', 'indulged', _, ['3A','6A'] ).
verb( 'inebriate', 'inebriates', 'inebriating', 'inebriated', 'inebriated', tran, ['6A'] ).
verb( 'infatuate', 'infatuates', 'infatuating', 'infatuated', 'infatuated', tran, [] ).
verb( 'infect', 'infects', 'infecting', 'infected', 'infected', tran, ['6A','14'] ).
verb( 'infer', 'infers', 'inferring', 'inferred', 'inferred', tran, ['6A','9','14'] ).
verb( 'infest', 'infests', 'infesting', 'infested', 'infested', tran, ['6A'] ).
verb( 'infiltrate', 'infiltrates', 'infiltrating', 'infiltrated', 'infiltrated', _, ['2A','3A','6A','14'] ).
verb( 'inflame', 'inflames', 'inflaming', 'inflamed', 'inflamed', _, ['2A','6A'] ).
verb( 'inflate', 'inflates', 'inflating', 'inflated', 'inflated', tran, ['6A','14'] ).
verb( 'inflect', 'inflects', 'inflecting', 'inflected', 'inflected', tran, ['6A'] ).
verb( 'inflict', 'inflicts', 'inflicting', 'inflicted', 'inflicted', tran, ['6A','14'] ).
verb( 'influence', 'influences', 'influencing', 'influenced', 'influenced', tran, ['6A'] ).
verb( 'inform', 'informs', 'informing', 'informed', 'informed', _, ['3A','6A','11','14','21'] ).
verb( 'infringe', 'infringes', 'infringing', 'infringed', 'infringed', _, ['3A','6A'] ).
verb( 'infuriate', 'infuriates', 'infuriating', 'infuriated', 'infuriated', tran, ['6A'] ).
verb( 'infuse', 'infuses', 'infusing', 'infused', 'infused', _, ['2A','6A','14'] ).
verb( 'ingest', 'ingests', 'ingesting', 'ingested', 'ingested', tran, ['6A'] ).
verb( 'ingraft', 'ingrafts', 'ingrafting', 'ingrafted', 'ingrafted', tran, ['6A','14'] ).
verb( 'ingratiate', 'ingratiates', 'ingratiating', 'ingratiated', 'ingratiated', tran, ['14'] ).
verb( 'inhabit', 'inhabits', 'inhabiting', 'inhabited', 'inhabited', tran, ['6A'] ).
verb( 'inhale', 'inhales', 'inhaling', 'inhaled', 'inhaled', _, ['2A','6A'] ).
verb( 'inherit', 'inherits', 'inheriting', 'inherited', 'inherited', _, ['2A','6A'] ).
verb( 'inhibit', 'inhibits', 'inhibiting', 'inhibited', 'inhibited', tran, ['6A','14'] ).
verb( 'initial', 'initials', 'initialling', 'initialled', 'initialled', tran, ['6A'] ).
verb( 'initiate', 'initiates', 'initiating', 'initiated', 'initiated', tran, ['6A','14'] ).
verb( 'inject', 'injects', 'injecting', 'injected', 'injected', tran, ['6A','14'] ).
verb( 'injure', 'injures', 'injuring', 'injured', 'injured', tran, ['6A'] ).
verb( 'ink', 'inks', 'inking', 'inked', 'inked', tran, ['6A','15B'] ).
verb( 'inlay', 'inlays', 'inlaying', 'inlaid', 'inlaid', tran, ['6A','14'] ).
verb( 'innovate', 'innovates', 'innovating', 'innovated', 'innovated', intran, ['2A'] ).
verb( 'inoculate', 'inoculates', 'inoculating', 'inoculated', 'inoculated', tran, ['6A','14'] ).
verb( 'inquire', 'inquires', 'inquiring', 'inquired', 'inquired', _, ['2A','3A','6A','8','10','14'] ).
verb( 'inscribe', 'inscribes', 'inscribing', 'inscribed', 'inscribed', tran, ['6A','15A'] ).
verb( 'inseminate', 'inseminates', 'inseminating', 'inseminated', 'inseminated', tran, ['6A'] ).
verb( 'insert', 'inserts', 'inserting', 'inserted', 'inserted', tran, ['6A','15A'] ).
verb( 'inset', 'insets', 'insetting', 'inset', 'inset', tran, [] ).
verb( 'insinuate', 'insinuates', 'insinuating', 'insinuated', 'insinuated', tran, ['6A','9','14'] ).
verb( 'insist', 'insists', 'insisting', 'insisted', 'insisted', _, ['3A','3B'] ).
verb( 'inspan', 'inspans', 'inspanning', 'inspanned', 'inspanned', tran, [] ).
verb( 'inspect', 'inspects', 'inspecting', 'inspected', 'inspected', tran, ['6A'] ).
verb( 'inspire', 'inspires', 'inspiring', 'inspired', 'inspired', tran, ['6A','14','17'] ).
verb( 'install', 'installs', 'installing', 'installed', 'installed', tran, ['6A','14'] ).
verb( 'instance', 'instances', 'instancing', 'instanced', 'instanced', tran, ['6A'] ).
verb( 'instigate', 'instigates', 'instigating', 'instigated', 'instigated', tran, ['6A','17'] ).
verb( 'instil', 'instils', 'instilling', 'instilled', 'instilled', tran, ['6A','14'] ).
verb( 'institute', 'institutes', 'instituting', 'instituted', 'instituted', tran, ['6A','14'] ).
verb( 'institutionalize', 'institutionalizes', 'institutionalizing', 'institutionalized', 'institutionalized', tran, ['6A'] ).
verb( 'instruct', 'instructs', 'instructing', 'instructed', 'instructed', tran, ['6A','11','15A','17','20','21'] ).
verb( 'insulate', 'insulates', 'insulating', 'insulated', 'insulated', tran, ['6A','14'] ).
verb( 'insult', 'insults', 'insulting', 'insulted', 'insulted', tran, ['6A'] ).
verb( 'insure', 'insures', 'insuring', 'insured', 'insured', tran, ['6A','14'] ).
verb( 'integrate', 'integrates', 'integrating', 'integrated', 'integrated', tran, ['2A','6A'] ).
verb( 'intend', 'intends', 'intending', 'intended', 'intended', tran, ['6A','6D','7A','9','14','17'] ).
verb( 'intensify', 'intensifies', 'intensifying', 'intensified', 'intensified', _, ['2A','6A'] ).
verb( 'inter', 'inters', 'interring', 'interred', 'interred', tran, [] ).
verb( 'interact', 'interacts', 'interacting', 'interacted', 'interacted', intran, [] ).
verb( 'interbreed', 'interbreeds', 'interbreeding', 'interbred', 'interbred', _, ['2A','6A'] ).
verb( 'intercede', 'intercedes', 'interceding', 'interceded', 'interceded', intran, ['3A'] ).
verb( 'intercept', 'intercepts', 'intercepting', 'intercepted', 'intercepted', tran, ['6A'] ).
verb( 'interchange', 'interchanges', 'interchanging', 'interchanged', 'interchanged', tran, ['2A','6A'] ).
verb( 'intercommunicate', 'intercommunicates', 'intercommunicating', 'intercommunicated', 'intercommunicated', intran, [] ).
verb( 'interconnect', 'interconnects', 'interconnecting', 'interconnected', 'interconnected', _, ['6A'] ).
verb( 'interdict', 'interdicts', 'interdicting', 'interdicted', 'interdicted', tran, ['6A'] ).
verb( 'interest', 'interests', 'interesting', 'interested', 'interested', tran, ['6A','14'] ).
verb( 'interfere', 'interferes', 'interfering', 'interfered', 'interfered', intran, ['2A','3A'] ).
verb( 'interject', 'interjects', 'interjecting', 'interjected', 'interjected', tran, ['6A'] ).
verb( 'interlace', 'interlaces', 'interlacing', 'interlaced', 'interlaced', _, ['2A','6A','14'] ).
verb( 'interlard', 'interlards', 'interlarding', 'interlarded', 'interlarded', tran, ['14'] ).
verb( 'interleave', 'interleaves', 'interleaving', 'interleaved', 'interleaved', tran, ['6A','14'] ).
verb( 'interlink', 'interlinks', 'interlinking', 'interlinked', 'interlinked', _, ['2A','6A'] ).
verb( 'interlock', 'interlocks', 'interlocking', 'interlocked', 'interlocked', _, ['2A','6A'] ).
verb( 'intermarry', 'intermarries', 'intermarrying', 'intermarried', 'intermarried', intran, ['2A','3A'] ).
verb( 'intermingle', 'intermingles', 'intermingling', 'intermingled', 'intermingled', _, ['6A','14'] ).
verb( 'intermix', 'intermixes', 'intermixing', 'intermixed', 'intermixed', _, [] ).
verb( 'intern', 'interns', 'interning', 'interned', 'interned', tran, ['6A'] ).
verb( 'internalize', 'internalizes', 'internalizing', 'internalized', 'internalized', tran, ['6A'] ).
verb( 'internationalize', 'internationalizes', 'internationalizing', 'internationalized', 'internationalized', tran, ['6A'] ).
verb( 'interpellate', 'interpellates', 'interpellating', 'interpellated', 'interpellated', tran, ['6A'] ).
verb( 'interpolate', 'interpolates', 'interpolating', 'interpolated', 'interpolated', tran, ['6A'] ).
verb( 'interpose', 'interposes', 'interposing', 'interposed', 'interposed', _, ['2A','3A','6A','14'] ).
verb( 'interpret', 'interprets', 'interpreting', 'interpreted', 'interpreted', _, ['2A','6A','16B'] ).
verb( 'interrelate', 'interrelates', 'interrelating', 'interrelated', 'interrelated', _, ['2A','6A'] ).
verb( 'interrogate', 'interrogates', 'interrogating', 'interrogated', 'interrogated', tran, ['6A'] ).
verb( 'interrupt', 'interrupts', 'interrupting', 'interrupted', 'interrupted', _, ['2A','6A'] ).
verb( 'intersect', 'intersects', 'intersecting', 'intersected', 'intersected', _, ['2A','6A'] ).
verb( 'intersperse', 'intersperses', 'interspersing', 'interspersed', 'interspersed', tran, ['14'] ).
verb( 'intertwine', 'intertwines', 'intertwining', 'intertwined', 'intertwined', _, ['2A','6A'] ).
verb( 'intervene', 'intervenes', 'intervening', 'intervened', 'intervened', intran, ['2A','3A'] ).
verb( 'interview', 'interviews', 'interviewing', 'interviewed', 'interviewed', tran, ['6A'] ).
verb( 'interweave', 'interweaves', 'interweaving', 'interwove', 'interwoven', tran, ['6A','14'] ).
verb( 'intimate', 'intimates', 'intimating', 'intimated', 'intimated', tran, ['6A','9','14'] ).
verb( 'intimidate', 'intimidates', 'intimidating', 'intimidated', 'intimidated', tran, ['6A','14'] ).
verb( 'intone', 'intones', 'intoning', 'intoned', 'intoned', _, ['2A','6A'] ).
verb( 'intoxicate', 'intoxicates', 'intoxicating', 'intoxicated', 'intoxicated', tran, ['6A'] ).
verb( 'intrench', 'intrenches', 'intrenching', 'intrenched', 'intrenched', tran, ['6A'] ).
verb( 'intrigue', 'intrigues', 'intriguing', 'intrigued', 'intrigued', _, ['2A','3A','6A'] ).
verb( 'introduce', 'introduces', 'introducing', 'introduced', 'introduced', tran, ['6A','14','15A'] ).
verb( 'introspect', 'introspects', 'introspecting', 'introspected', 'introspected', intran, ['2A'] ).
verb( 'introvert', 'introverts', 'introverting', 'introverted', 'introverted', tran, ['6A'] ).
verb( 'intrude', 'intrudes', 'intruding', 'intruded', 'intruded', _, ['2A','3A','14'] ).
verb( 'intrust', 'intrusts', 'intrusting', 'intrusted', 'intrusted', tran, ['14'] ).
verb( 'intuit', 'intuits', 'intuiting', 'intuited', 'intuited', _, ['2A','6A'] ).
verb( 'inundate', 'inundates', 'inundating', 'inundated', 'inundated', tran, ['6A','14'] ).
verb( 'inure', 'inures', 'inuring', 'inured', 'inured', tran, ['14'] ).
verb( 'invade', 'invades', 'invading', 'invaded', 'invaded', tran, ['6A'] ).
verb( 'invalid', 'invalids', 'invaliding', 'invalided', 'invalided', tran, ['15A'] ).
verb( 'invalidate', 'invalidates', 'invalidating', 'invalidated', 'invalidated', tran, ['6A'] ).
verb( 'inveigh', 'inveighs', 'inveighing', 'inveighed', 'inveighed', intran, ['3A'] ).
verb( 'inveigle', 'inveigles', 'inveigling', 'inveigled', 'inveigled', tran, ['14'] ).
verb( 'invent', 'invents', 'inventing', 'invented', 'invented', tran, ['6A'] ).
verb( 'invert', 'inverts', 'inverting', 'inverted', 'inverted', tran, ['6A'] ).
verb( 'invest', 'invests', 'investing', 'invested', 'invested', _, ['3A','6A','14'] ).
verb( 'investigate', 'investigates', 'investigating', 'investigated', 'investigated', tran, ['6A'] ).
verb( 'invigilate', 'invigilates', 'invigilating', 'invigilated', 'invigilated', intran, ['2A','6A'] ).
verb( 'invigorate', 'invigorates', 'invigorating', 'invigorated', 'invigorated', tran, ['6A'] ).
verb( 'invite', 'invites', 'inviting', 'invited', 'invited', tran, ['6A','15A','15B','17'] ).
verb( 'invoice', 'invoices', 'invoicing', 'invoiced', 'invoiced', tran, [] ).
verb( 'invoke', 'invokes', 'invoking', 'invoked', 'invoked', tran, ['6A','14'] ).
verb( 'involve', 'involves', 'involving', 'involved', 'involved', tran, ['6A','6B','14','19C'] ).
verb( 'ionize', 'ionizes', 'ionizing', 'ionized', 'ionized', _, ['2A','6A'] ).
verb( 'irk', 'irks', 'irking', 'irked', 'irked', tran, [] ).
verb( 'iron', 'irons', 'ironing', 'ironed', 'ironed', _, ['2A','2C','6A','15A','15B'] ).
verb( 'irradiate', 'irradiates', 'irradiating', 'irradiated', 'irradiated', tran, ['6A'] ).
verb( 'irrigate', 'irrigates', 'irrigating', 'irrigated', 'irrigated', tran, ['6A'] ).
verb( 'irritate', 'irritates', 'irritating', 'irritated', 'irritated', tran, ['6A'] ).
verb( 'isolate', 'isolates', 'isolating', 'isolated', 'isolated', tran, ['6A','14'] ).
verb( 'issue', 'issues', 'issuing', 'issued', 'issued', _, ['2A','3A','6A','14'] ).
verb( 'italicize', 'italicizes', 'italicizing', 'italicized', 'italicized', tran, [] ).
verb( 'itch', 'itches', 'itching', 'itched', 'itched', intran, ['2A','3A','4A'] ).
verb( 'itemize', 'itemizes', 'itemizing', 'itemized', 'itemized', tran, ['6A'] ).
verb( 'iterate', 'iterates', 'iterating', 'iterated', 'iterated', tran, ['6A'] ).
verb( 'jab', 'jabs', 'jabbing', 'jabbed', 'jabbed', _, ['3A','14','15B'] ).
verb( 'jabber', 'jabbers', 'jabbering', 'jabbered', 'jabbered', _, ['2A','2C','6A','15B'] ).
verb( 'jack', 'jacks', 'jacking', 'jacked', 'jacked', tran, ['15B'] ).
verb( 'jack-knife', 'jack-knifes', 'jack-knifing', 'jack-knifed', 'jack-knifed', intran, ['2A'] ).
verb( 'jag', 'jags', 'jagging', 'jagged', 'jagged', tran, ['6A'] ).
verb( 'jail', 'jails', 'jailing', 'jailed', 'jailed', tran, ['6A'] ).
verb( 'jam', 'jams', 'jamming', 'jammed', 'jammed', _, ['2A','2C','6A','14','15A','15B'] ).
verb( 'jampack', 'jampacks', 'jampacking', 'jampacked', 'jampacked', tran, ['6A'] ).
verb( 'jangle', 'jangles', 'jangling', 'jangled', 'jangled', _, ['2A','6A'] ).
verb( 'japan', 'japans', 'japanning', 'japanned', 'japanned', tran, [] ).
verb( 'jar', 'jars', 'jarring', 'jarred', 'jarred', _, ['2A','3A','6A'] ).
verb( 'jaundice', 'jaundices', 'jaundicing', 'jaundiced', 'jaundiced', tran, [] ).
verb( 'jaunt', 'jaunts', 'jaunting', 'jaunted', 'jaunted', intran, ['2A','2C'] ).
verb( 'jaw', 'jaws', 'jawing', 'jawed', 'jawed', intran, ['2A','2C','3A'] ).
verb( 'jaywalk', 'jaywalks', 'jaywalking', 'jaywalked', 'jaywalked', intran, [] ).
verb( 'jazz', 'jazzes', 'jazzing', 'jazzed', 'jazzed', tran, ['6A','15B'] ).
verb( 'jeer', 'jeers', 'jeering', 'jeered', 'jeered', _, ['2A','3A','6A'] ).
verb( 'jell', 'jells', 'jelling', 'jelled', 'jelled', _, ['2A','6A'] ).
verb( 'jelly', 'jellies', 'jellying', 'jellied', 'jellied', _, ['2A','6A'] ).
verb( 'jeopardize', 'jeopardizes', 'jeopardizing', 'jeopardized', 'jeopardized', tran, ['6A'] ).
verb( 'jerk', 'jerks', 'jerking', 'jerked', 'jerked', _, ['2C','6A','15A','15B'] ).
verb( 'jest', 'jests', 'jesting', 'jested', 'jested', intran, ['2A','3A'] ).
verb( 'jet', 'jets', 'jetting', 'jetted', 'jetted', _, [] ).
verb( 'jettison', 'jettisons', 'jettisoning', 'jettisoned', 'jettisoned', tran, ['6A'] ).
verb( 'jewel', 'jewels', 'jewelling', 'jewelled', 'jewelled', tran, [] ).
verb( 'jib', 'jibs', 'jibbing', 'jibbed', 'jibbed', intran, ['2A','3A'] ).
verb( 'jibe', 'jibes', 'jibing', 'jibed', 'jibed', intran, ['2A','3A'] ).
verb( 'jig', 'jigs', 'jigging', 'jigged', 'jigged', _, ['2A','2C','15B'] ).
verb( 'jiggle', 'jiggles', 'jiggling', 'jiggled', 'jiggled', _, [] ).
verb( 'jilt', 'jilts', 'jilting', 'jilted', 'jilted', tran, ['6A'] ).
verb( 'jingle', 'jingles', 'jingling', 'jingled', 'jingled', _, ['2A','2C','6A','15B'] ).
verb( 'jive', 'jives', 'jiving', 'jived', 'jived', intran, [] ).
verb( 'job', 'jobs', 'jobbing', 'jobbed', 'jobbed', _, ['2A','6A','14'] ).
verb( 'jockey', 'jockeys', 'jockeying', 'jockeyed', 'jockeyed', _, ['3A','15A'] ).
verb( 'jog', 'jogs', 'jogging', 'jogged', 'jogged', _, ['2A','2C','6A','15B'] ).
verb( 'joggle', 'joggles', 'joggling', 'joggled', 'joggled', _, ['2A','6A'] ).
verb( 'join', 'joins', 'joining', 'joined', 'joined', _, ['2A','2C','3A','6A','14','15A','15B'] ).
verb( 'joint', 'joints', 'jointing', 'jointed', 'jointed', tran, ['6A'] ).
verb( 'joke', 'jokes', 'joking', 'joked', 'joked', intran, ['2A','2C'] ).
verb( 'jolly', 'jollies', 'jollying', 'jollied', 'jollied', tran, ['6A','15A','15B'] ).
verb( 'jolt', 'jolts', 'jolting', 'jolted', 'jolted', _, ['2A','2C','6A','15A','15B'] ).
verb( 'jostle', 'jostles', 'jostling', 'jostled', 'jostled', _, ['2C','6A'] ).
verb( 'jot', 'jots', 'jotting', 'jotted', 'jotted', tran, ['15B'] ).
verb( 'journey', 'journeys', 'journeying', 'journeyed', 'journeyed', intran, ['2A','2C'] ).
verb( 'joust', 'jousts', 'jousting', 'jousted', 'jousted', intran, [] ).
verb( 'joy', 'joys', 'joying', 'joyed', 'joyed', intran, [] ).
verb( 'judder', 'judders', 'juddering', 'juddered', 'juddered', intran, [] ).
verb( 'judge', 'judges', 'judging', 'judged', 'judged', _, ['2A','3A','6A','9','10','22','25'] ).
verb( 'jug', 'jugs', 'jugging', 'jugged', 'jugged', tran, ['6A'] ).
verb( 'juggle', 'juggles', 'juggling', 'juggled', 'juggled', _, ['2A','3A','6A','16A'] ).
verb( 'jumble', 'jumbles', 'jumbling', 'jumbled', 'jumbled', _, ['2C','15B'] ).
verb( 'jump', 'jumps', 'jumping', 'jumped', 'jumped', _, ['2A','2C','6A'] ).
verb( 'junket', 'junkets', 'junketing', 'junketed', 'junketed', intran, [] ).
verb( 'justify', 'justifies', 'justifying', 'justified', 'justified', tran, ['6A','19C'] ).
verb( 'jut', 'juts', 'jutting', 'jutted', 'jutted', intran, ['2C'] ).
verb( 'juxtapose', 'juxtaposes', 'juxtaposing', 'juxtaposed', 'juxtaposed', tran, ['6A'] ).
verb( 'keel', 'keels', 'keeling', 'keeled', 'keeled', _, ['2C','6A','15B'] ).
verb( 'keen', 'keens', 'keening', 'keened', 'keened', _, [] ).
verb( 'keep', 'keeps', 'keeping', 'kept', 'kept', _, ['2A','2C','2E','3A','6A','14','15A','15B','19B','22'] ).
verb( 'ken', 'kens', 'kenning', 'kenned', 'kenned', tran, ['6A','9'] ).
verb( 'kennel', 'kennels', 'kennelling', 'kennelled', 'kennelled', _, [] ).
verb( 'key', 'keys', 'keying', 'keyed', 'keyed', tran, ['6A'] ).
verb( 'kick', 'kicks', 'kicking', 'kicked', 'kicked', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'kid', 'kids', 'kidding', 'kidded', 'kidded', _, [] ).
verb( 'kidnap', 'kidnaps', 'kidnapping', 'kidnapped', 'kidnapped', tran, [] ).
verb( 'kill', 'kills', 'killing', 'killed', 'killed', _, ['2A','6A','15B'] ).
verb( 'kindle', 'kindles', 'kindling', 'kindled', 'kindled', _, ['2A','6A'] ).
verb( 'kink', 'kinks', 'kinking', 'kinked', 'kinked', _, [] ).
verb( 'kip', 'kips', 'kipping', 'kipped', 'kipped', intran, ['2A','2C'] ).
verb( 'kiss', 'kisses', 'kissing', 'kissed', 'kissed', _, ['2A','6A','15A','15B'] ).
verb( 'kit', 'kits', 'kitting', 'kitted', 'kitted', tran, ['15B'] ).
verb( 'knap', 'knaps', 'knapping', 'knapped', 'knapped', tran, [] ).
verb( 'knead', 'kneads', 'kneading', 'kneaded', 'kneaded', tran, ['6A'] ).
verb( 'kneel', 'kneels', 'kneeling', 'kneeled', 'kneeled', intran, ['2A','2C'] ).
verb( 'knife', 'knifes', 'knifing', 'knifed', 'knifed', tran, ['6A'] ).
verb( 'knight', 'knights', 'knighting', 'knighted', 'knighted', tran, ['6A'] ).
verb( 'knit', 'knits', 'knitting', 'knitted', 'knitted', _, ['2A','2C','6A','15A','15B'] ).
verb( 'knock', 'knocks', 'knocking', 'knocked', 'knocked', _, ['2A','2C','6A','14','15A','15B'] ).
verb( 'knot', 'knots', 'knotting', 'knotted', 'knotted', _, ['2A','6A','15B'] ).
verb( 'know', 'knows', 'knowing', 'knew', 'known', _, ['2A','3A','6A','8','9','10','17','18B','25'] ).
verb( 'knuckle', 'knuckles', 'knuckling', 'knuckled', 'knuckled', intran, [] ).
verb( 'kotow', 'kotows', 'kotowing', 'kotowed', 'kotowed', intran, [] ).
verb( 'kowtow', 'kowtows', 'kowtowing', 'kowtowed', 'kowtowed', intran, [] ).
verb( 'label', 'labels', 'labelling', 'labelled', 'labelled', tran, ['6A'] ).
verb( 'labour', 'labours', 'labouring', 'laboured', 'laboured', _, ['2A','2C','3A','4A','6A'] ).
verb( 'lace', 'laces', 'lacing', 'laced', 'laced', _, ['2A','2C','3A','6A','14','15A','15B'] ).
verb( 'lacerate', 'lacerates', 'lacerating', 'lacerated', 'lacerated', tran, ['6A'] ).
verb( 'lack', 'lacks', 'lacking', 'lacked', 'lacked', _, ['3A','6B'] ).
verb( 'lacquer', 'lacquers', 'lacquering', 'lacquered', 'lacquered', tran, ['6A'] ).
verb( 'ladder', 'ladders', 'laddering', 'laddered', 'laddered', intran, [] ).
verb( 'lade', 'lades', 'lading', 'laded', 'laden', tran, ['6A'] ).
verb( 'ladle', 'ladles', 'ladling', 'ladled', 'ladled', tran, ['6A','15B'] ).
verb( 'lag', 'lags', 'lagging', 'lagged', 'lagged', _, ['2A','2C','6A','14'] ).
verb( 'laicize', 'laicizes', 'laicizing', 'laicized', 'laicized', tran, ['6A'] ).
verb( 'lam', 'lams', 'lamming', 'lammed', 'lammed', _, ['3A','6A'] ).
verb( 'lamb', 'lambs', 'lambing', 'lambed', 'lambed', intran, [] ).
verb( 'lambaste', 'lambastes', 'lambasting', 'lambasted', 'lambasted', tran, [] ).
verb( 'lame', 'lames', 'laming', 'lamed', 'lamed', tran, [] ).
verb( 'lament', 'laments', 'lamenting', 'lamented', 'lamented', _, ['2A','3A','6A'] ).
verb( 'laminate', 'laminates', 'laminating', 'laminated', 'laminated', _, ['2A','6A'] ).
verb( 'lampoon', 'lampoons', 'lampooning', 'lampooned', 'lampooned', tran, ['6A'] ).
verb( 'lance', 'lances', 'lancing', 'lanced', 'lanced', tran, ['6A'] ).
verb( 'land', 'lands', 'landing', 'landed', 'landed', _, ['2A','2C','6A','12C'] ).
verb( 'landscape', 'landscapes', 'landscaping', 'landscaped', 'landscaped', tran, ['6A'] ).
verb( 'languish', 'languishes', 'languishing', 'languished', 'languished', intran, ['2A','2C'] ).
verb( 'lap', 'laps', 'lapping', 'lapped', 'lapped', _, ['2A','2C','6A','15B'] ).
verb( 'lapse', 'lapses', 'lapsing', 'lapsed', 'lapsed', intran, ['2A','3A'] ).
verb( 'lard', 'lards', 'larding', 'larded', 'larded', tran, ['6A'] ).
verb( 'lark', 'larks', 'larking', 'larked', 'larked', intran, ['2A','2C'] ).
verb( 'larn', 'larns', 'larning', 'larned', 'larned', _, [] ).
verb( 'lash', 'lashes', 'lashing', 'lashed', 'lashed', _, ['2C','6A','14','15A','15B'] ).
verb( 'lasso', 'lassos', 'lassoing', 'lassoed', 'lassoed', tran, [] ).
verb( 'last', 'lasts', 'lasting', 'lasted', 'lasted', intran, ['2A','2B','2C'] ).
verb( 'latch', 'latches', 'latching', 'latched', 'latched', _, ['2A','2C','3A','6A'] ).
verb( 'lather', 'lathers', 'lathering', 'lathered', 'lathered', _, ['2A','6A'] ).
verb( 'laud', 'lauds', 'lauding', 'lauded', 'lauded', tran, ['6A'] ).
verb( 'laugh', 'laughs', 'laughing', 'laughed', 'laughed', _, ['2A','2B','2C','3A','6B','15A','15B','22'] ).
verb( 'launch', 'launches', 'launching', 'launched', 'launched', _, ['2C','3A','6A','15A'] ).
verb( 'launder', 'launders', 'laundering', 'laundered', 'laundered', _, ['2A','6A'] ).
verb( 'lave', 'laves', 'laving', 'laved', 'laved', tran, [] ).
verb( 'lavish', 'lavishes', 'lavishing', 'lavished', 'lavished', tran, ['14'] ).
verb( 'lay', 'lays', 'laying', 'laid', 'laid', _, ['2A','2B','2C','2D','3A','6A','12C','14','15A','15B','22'] ).
verb( 'layer', 'layers', 'layering', 'layered', 'layered', tran, ['6A'] ).
verb( 'laze', 'lazes', 'lazing', 'lazed', 'lazed', _, ['2A','2C','15B'] ).
verb( 'leach', 'leaches', 'leaching', 'leached', 'leached', tran, ['6A','15B'] ).
verb( 'lead', 'leads', 'leading', 'leant', 'leant', _, ['2A','2C','3A','6A','12C','14','15A','15B','17'] ).
verb( 'leaf', 'leafs', 'leafing', 'leafed', 'leafed', intran, [] ).
verb( 'league', 'leagues', 'leaguing', 'leagued', 'leagued', _, ['2C','6A','15A','15B'] ).
verb( 'leak', 'leaks', 'leaking', 'leaked', 'leaked', _, ['2A','2C','6A','14'] ).
verb( 'lean', 'leans', 'leaning', 'leaned', 'leaned', _, ['2A','2C','3A','15A'] ).
verb( 'leap', 'leaps', 'leaping', 'leaped', 'leaped', _, ['2A','2C','3A','6A','15A'] ).
verb( 'leapfrog', 'leapfrogs', 'leapfrogging', 'leapfrogged', 'leapfrogged', tran, [] ).
verb( 'learn', 'learns', 'learning', 'learned', 'learned', _, ['2A','3A','6A','7A','8','9','10','15A','15B'] ).
verb( 'lease', 'leases', 'leasing', 'leased', 'leased', tran, ['6A'] ).
verb( 'leave', 'leaves', 'leavening', 'leavened', 'leavened', _, ['2A','2C','3A','6A','12B','13B','14','15A','15B','16A','19B','22','24B','25'] ).
verb( 'leaven', 'leavens', 'leavening', 'leavened', 'leavened', tran, ['6A'] ).
verb( 'lecture', 'lectures', 'lecturing', 'lectured', 'lectured', _, ['2A','3A','6A','14'] ).
verb( 'leer', 'leers', 'leering', 'leered', 'leered', intran, ['2A','3A'] ).
verb( 'legalize', 'legalizes', 'legalizing', 'legalized', 'legalized', tran, ['6A'] ).
verb( 'legislate', 'legislates', 'legislating', 'legislated', 'legislated', intran, ['2A','3A'] ).
verb( 'legitimatize', 'legitimatizes', 'legitimatizing', 'legitimatized', 'legitimatized', tran, [] ).
verb( 'lend', 'lends', 'lending', 'lent', 'lent', tran, ['6A','12A','13A','14'] ).
verb( 'lengthen', 'lengthens', 'lengthening', 'lengthened', 'lengthened', _, ['2A','6A'] ).
verb( 'lessen', 'lessens', 'lessening', 'lessened', 'lessened', _, ['2A','6A'] ).
verb( 'let', 'lets', 'letting', 'let', 'let', _, ['2C','6A','14','15A','15B','18B','22','24A'] ).
verb( 'levant', 'levants', 'levanting', 'levanted', 'levanted', intran, [] ).
verb( 'level', 'levels', 'levelling', 'levelled', 'levelled', _, ['2C','3A','6A','14','15A','15B'] ).
verb( 'lever', 'levers', 'levering', 'levered', 'levered', tran, ['6A','15B'] ).
verb( 'levitate', 'levitates', 'levitating', 'levitated', 'levitated', _, ['2A','6A'] ).
verb( 'levy', 'levies', 'levying', 'levied', 'levied', _, ['3A','6A','14'] ).
verb( 'liaise', 'liaises', 'liaising', 'liaised', 'liaised', intran, ['2A','3A'] ).
verb( 'libel', 'libels', 'libelling', 'libelled', 'libelled', tran, ['6A'] ).
verb( 'liberalize', 'liberalizes', 'liberalizing', 'liberalized', 'liberalized', tran, [] ).
verb( 'liberate', 'liberates', 'liberating', 'liberated', 'liberated', tran, ['6A','14'] ).
verb( 'licence', 'licences', 'licencing', 'licenced', 'licenced', tran, ['6A','17'] ).
verb( 'license', 'licenses', 'licensing', 'licensed', 'licensed', tran, ['6A','17'] ).
verb( 'lick', 'licks', 'licking', 'licked', 'licked', _, ['2A','6A','15A','15B','22'] ).
verb( 'lie', 'lies', 'lying', 'lay', 'lain', intran, ['2A','2C','2D','3A'] ).
verb( 'lie', 'lies', 'lying', 'lied', 'lied', intran, ['2A','2C','2D','3A'] ).
verb( 'lift', 'lifts', 'lifting', 'lifted', 'lifted', _, ['2A','2C','6A','15A','15B'] ).
verb( 'light', 'lights', 'lighting', 'lighted', 'lighted', _, ['2C','3A','6A','15A','15B'] ).
verb( 'lighten', 'lightens', 'lightening', 'lightened', 'lightened', _, ['2A','6A'] ).
verb( 'lighter', 'lighters', 'lightering', 'lightered', 'lightered', tran, [] ).
verb( 'like', 'likes', 'liking', 'liked', 'liked', tran, ['6A','6D','7A','17','19B','19C','22'] ).
verb( 'liken', 'likens', 'likening', 'likened', 'likened', tran, ['14'] ).
verb( 'lilt', 'lilts', 'lilting', 'lilted', 'lilted', _, [] ).
verb( 'limber', 'limbers', 'limbering', 'limbered', 'limbered', _, ['2C','15B'] ).
verb( 'lime', 'limes', 'liming', 'limed', 'limed', tran, ['6A'] ).
verb( 'limit', 'limits', 'limiting', 'limited', 'limited', tran, ['6A','14'] ).
verb( 'limn', 'limns', 'limning', 'limned', 'limned', tran, [] ).
verb( 'limp', 'limps', 'limping', 'limped', 'limped', intran, ['2A','2C'] ).
verb( 'line', 'lines', 'lining', 'lined', 'lined', _, ['2C','6A','14','15B'] ).
verb( 'linger', 'lingers', 'lingering', 'lingered', 'lingered', intran, ['2A','2C'] ).
verb( 'link', 'links', 'linking', 'linked', 'linked', _, ['2A','2C','6A','15A','15B'] ).
verb( 'lionize', 'lionizes', 'lionizing', 'lionized', 'lionized', tran, ['6A'] ).
verb( 'lip-read', 'lip-reads', 'lip-reading', 'lip-read', 'lip-read', tran, [] ).
verb( 'liquefy', 'liquefies', 'liquefying', 'liquefied', 'liquefied', _, ['2A','6A'] ).
verb( 'liquidate', 'liquidates', 'liquidating', 'liquidated', 'liquidated', _, ['2A','6A'] ).
verb( 'liquidize', 'liquidizes', 'liquidizing', 'liquidized', 'liquidized', tran, ['6A'] ).
verb( 'lisp', 'lisps', 'lisping', 'lisped', 'lisped', _, ['2A','6A','15B'] ).
verb( 'list', 'lists', 'listing', 'listed', 'listed', _, ['2A','2C','6A'] ).
verb( 'listen', 'listens', 'listening', 'listened', 'listened', intran, ['2A','2C','3A'] ).
verb( 'lithograph', 'lithographs', 'lithographing', 'lithographed', 'lithographed', _, [] ).
verb( 'litigate', 'litigates', 'litigating', 'litigated', 'litigated', _, ['2A','6A'] ).
verb( 'litter', 'litters', 'littering', 'littered', 'littered', _, ['2A','6A','14','15A','15B'] ).
verb( 'live', 'lives', 'living', 'lived', 'lived', _, ['2A','2B','2C','2D','3A','4A','6B','15B'] ).
verb( 'liven', 'livens', 'livening', 'livened', 'livened', _, ['2C','15B'] ).
verb( 'load', 'loads', 'loading', 'loaded', 'loaded', _, ['2A','2C','6A','14','15A','15B'] ).
verb( 'loaf', 'loafs', 'loafing', 'loafed', 'loafed', _, ['2A','2C','15A','15B'] ).
verb( 'loan', 'loans', 'loaning', 'loaned', 'loaned', tran, ['6A','14'] ).
verb( 'loathe', 'loathes', 'loathing', 'loathed', 'loathed', tran, ['6A','6C'] ).
verb( 'lob', 'lobs', 'lobbing', 'lobbed', 'lobbed', _, ['2C','6A','15A'] ).
verb( 'lobby', 'lobbies', 'lobbying', 'lobbied', 'lobbied', _, ['2A','2C','6A','15A'] ).
verb( 'localize', 'localizes', 'localizing', 'localized', 'localized', tran, ['6A'] ).
verb( 'locate', 'locates', 'locating', 'located', 'located', tran, ['6A','15A'] ).
verb( 'lock', 'locks', 'locking', 'locked', 'locked', _, ['2A','2C','6A','15A','15B'] ).
verb( 'lodge', 'lodges', 'lodging', 'lodged', 'lodged', _, ['3A','6A','14','15A'] ).
verb( 'loft', 'lofts', 'lofting', 'lofted', 'lofted', tran, ['6A'] ).
verb( 'log', 'logs', 'logging', 'logged', 'logged', tran, ['6A'] ).
verb( 'loiter', 'loiters', 'loitering', 'loitered', 'loitered', _, ['2A','2C','15B'] ).
verb( 'loll', 'lolls', 'lolling', 'lolled', 'lolled', _, ['2A','2C','15B'] ).
verb( 'long', 'longs', 'longing', 'longed', 'longed', intran, ['3A','4C'] ).
verb( 'look', 'looks', 'looking', 'looked', 'looked', _, ['2A','2C','2D','3A','4A','4D','6A','8','15B'] ).
verb( 'loom', 'looms', 'looming', 'loomed', 'loomed', intran, ['2C','2D'] ).
verb( 'loop', 'loops', 'looping', 'looped', 'looped', _, ['2A','6A','15B'] ).
verb( 'loose', 'looses', 'loosing', 'loosed', 'loosed', tran, ['6A'] ).
verb( 'loosen', 'loosens', 'loosening', 'loosened', 'loosened', _, ['2A','6A','15B'] ).
verb( 'loot', 'loots', 'looting', 'looted', 'looted', _, ['2A','6A'] ).
verb( 'lop', 'lops', 'lopping', 'lopped', 'lopped', _, ['6A','15B'] ).
verb( 'lope', 'lopes', 'loping', 'loped', 'loped', intran, ['2A','2C'] ).
verb( 'lord', 'lords', 'lording', 'lorded', 'lorded', tran, [] ).
verb( 'lose', 'loses', 'losing', 'lost', 'lost', _, ['2A','2B','3A','6A','12C','15A'] ).
verb( 'lounge', 'lounges', 'lounging', 'lounged', 'lounged', intran, ['2A','2C'] ).
verb( 'lour', 'lours', 'louring', 'loured', 'loured', intran, ['2A','3A'] ).
verb( 'love', 'loves', 'loving', 'loved', 'loved', tran, ['6A','6D','7A','17'] ).
verb( 'low', 'lows', 'lowing', 'lowed', 'lowed', intran, [] ).
verb( 'lower', 'lowers', 'lowering', 'lowered', 'lowered', _, ['2A','6A','15A'] ).
verb( 'lower', 'lowers', 'lowering', 'lowered', 'lowered', intran, ['2A','3A'] ).
verb( 'lubricate', 'lubricates', 'lubricating', 'lubricated', 'lubricated', tran, ['6A'] ).
verb( 'luff', 'luffs', 'luffing', 'luffed', 'luffed', _, [] ).
verb( 'lug', 'lugs', 'lugging', 'lugged', 'lugged', tran, ['6A','15A','15B'] ).
verb( 'lull', 'lulls', 'lulling', 'lulled', 'lulled', _, ['2A','6A','15A'] ).
verb( 'lumber', 'lumbers', 'lumbering', 'lumbered', 'lumbered', _, ['2C','6A','14','15A','15B'] ).
verb( 'lump', 'lumps', 'lumping', 'lumped', 'lumped', tran, ['2A','6A','15B'] ).
verb( 'lunch', 'lunches', 'lunching', 'lunched', 'lunched', _, [] ).
verb( 'lunge', 'lunges', 'lunging', 'lunged', 'lunged', intran, ['2A','2C'] ).
verb( 'lurch', 'lurches', 'lurching', 'lurched', 'lurched', intran, ['2C'] ).
verb( 'lure', 'lures', 'luring', 'lured', 'lured', tran, ['6A','15B'] ).
verb( 'lurk', 'lurks', 'lurking', 'lurked', 'lurked', intran, ['2C'] ).
verb( 'lust', 'lusts', 'lusting', 'lusted', 'lusted', intran, ['3A'] ).
verb( 'luxuriate', 'luxuriates', 'luxuriating', 'luxuriated', 'luxuriated', intran, ['3A'] ).
verb( 'lynch', 'lynches', 'lynching', 'lynched', 'lynched', tran, ['6A'] ).
verb( 'macadamize', 'macadamizes', 'macadamizing', 'macadamized', 'macadamized', tran, [] ).
verb( 'macerate', 'macerates', 'macerating', 'macerated', 'macerated', _, ['2A','6A'] ).
verb( 'machine', 'machines', 'machining', 'machined', 'machined', tran, ['6A'] ).
verb( 'madden', 'maddens', 'maddening', 'maddened', 'maddened', tran, ['6A'] ).
verb( 'maffick', 'mafficks', 'mafficking', 'mafficked', 'mafficked', intran, [] ).
verb( 'magnetize', 'magnetizes', 'magnetizing', 'magnetized', 'magnetized', tran, ['6A'] ).
verb( 'magnify', 'magnifies', 'magnifying', 'magnified', 'magnified', tran, ['6A'] ).
verb( 'mail', 'mails', 'mailing', 'mailed', 'mailed', tran, ['6A'] ).
verb( 'maim', 'maims', 'maiming', 'maimed', 'maimed', tran, ['6A'] ).
verb( 'maintain', 'maintains', 'maintaining', 'maintained', 'maintained', tran, ['6A','9','25'] ).
verb( 'major', 'majors', 'majoring', 'majored', 'majored', intran, ['3A'] ).
verb( 'make', 'makes', 'making', 'made', 'made', _, ['2A','2C','2D','3A','6A','12A','12B','13A','13B','14','15B','16A','18B','22','23','24A','25'] ).
verb( 'malfunction', 'malfunctions', 'malfunctioning', 'malfunctioned', 'malfunctioned', intran, ['2A'] ).
verb( 'malign', 'maligns', 'maligning', 'maligned', 'maligned', tran, ['6A'] ).
verb( 'malinger', 'malingers', 'malingering', 'malingered', 'malingered', intran, ['2A'] ).
verb( 'malt', 'malts', 'malting', 'malted', 'malted', _, ['2A','6A'] ).
verb( 'maltreat', 'maltreats', 'maltreating', 'maltreated', 'maltreated', tran, ['6A'] ).
verb( 'man', 'mans', 'manning', 'manned', 'manned', tran, ['6A'] ).
verb( 'manacle', 'manacles', 'manacling', 'manacled', 'manacled', tran, ['6A'] ).
verb( 'manage', 'manages', 'managing', 'managed', 'managed', _, ['2A','2C','3A','4A','6A'] ).
verb( 'mandate', 'mandates', 'mandating', 'mandated', 'mandated', tran, ['6A'] ).
verb( 'mangle', 'mangles', 'mangling', 'mangled', 'mangled', tran, ['6A'] ).
verb( 'manhandle', 'manhandles', 'manhandling', 'manhandled', 'manhandled', tran, [] ).
verb( 'manicure', 'manicures', 'manicuring', 'manicured', 'manicured', tran, ['6A'] ).
verb( 'manifest', 'manifests', 'manifesting', 'manifested', 'manifested', tran, ['6A'] ).
verb( 'manifold', 'manifolds', 'manifolding', 'manifolded', 'manifolded', tran, ['6A'] ).
verb( 'manipulate', 'manipulates', 'manipulating', 'manipulated', 'manipulated', tran, ['6A'] ).
verb( 'manoeuvre', 'manoeuvres', 'manoeuvring', 'manoeuvred', 'manoeuvred', _, ['2A','2C','6A','15A'] ).
verb( 'mantle', 'mantles', 'mantling', 'mantled', 'mantled', _, ['2C','6A'] ).
verb( 'manufacture', 'manufactures', 'manufacturing', 'manufactured', 'manufactured', tran, ['6A'] ).
verb( 'manumit', 'manumits', 'manumitting', 'manumitted', 'manumitted', tran, [] ).
verb( 'manure', 'manures', 'manuring', 'manured', 'manured', tran, ['6A'] ).
verb( 'map', 'maps', 'mapping', 'mapped', 'mapped', tran, ['6A','15B'] ).
verb( 'mar', 'mars', 'marring', 'marred', 'marred', tran, ['6A'] ).
verb( 'maraud', 'marauds', 'marauding', 'marauded', 'marauded', intran, ['2A'] ).
verb( 'march', 'marches', 'marching', 'marched', 'marched', _, ['2A','2B','2C','3A','15A','15B'] ).
verb( 'marinade', 'marinades', 'marinading', 'marinaded', 'marinaded', tran, [] ).
verb( 'marinate', 'marinates', 'marinating', 'marinated', 'marinated', tran, [] ).
verb( 'mark', 'marks', 'marking', 'marked', 'marked', tran, ['2A','6A','8','10','15A','15B','22'] ).
verb( 'market', 'markets', 'marketing', 'marketed', 'marketed', _, ['2A','6A'] ).
verb( 'maroon', 'maroons', 'marooning', 'marooned', 'marooned', tran, ['6A'] ).
verb( 'marry', 'marries', 'marrying', 'married', 'married', _, ['2A','2D','4A','6A','15B'] ).
verb( 'marshal', 'marshals', 'marshalling', 'marshalled', 'marshalled', tran, ['6A','15A','15B'] ).
verb( 'martyr', 'martyrs', 'martyring', 'martyred', 'martyred', tran, ['6A'] ).
verb( 'marvel', 'marvels', 'marvelling', 'marvelled', 'marvelled', intran, ['3A','3B'] ).
verb( 'mash', 'mashes', 'mashing', 'mashed', 'mashed', tran, ['6A'] ).
verb( 'mask', 'masks', 'masking', 'masked', 'masked', tran, ['6A'] ).
verb( 'masquerade', 'masquerades', 'masquerading', 'masqueraded', 'masqueraded', intran, ['2A','2C'] ).
verb( 'mass', 'masses', 'massing', 'massed', 'massed', _, ['2A','6A'] ).
verb( 'mass-produce', 'mass-produces', 'mass-producing', 'mass-produced', 'mass-produced', tran, [] ).
verb( 'massacre', 'massacres', 'massacring', 'massacred', 'massacred', tran, ['6A'] ).
verb( 'massage', 'massages', 'massaging', 'massaged', 'massaged', tran, ['6A'] ).
verb( 'master', 'masters', 'mastering', 'mastered', 'mastered', tran, ['6A'] ).
verb( 'mastermind', 'masterminds', 'masterminding', 'masterminded', 'masterminded', tran, [] ).
verb( 'masticate', 'masticates', 'masticating', 'masticated', 'masticated', tran, ['6A'] ).
verb( 'masturbate', 'masturbates', 'masturbating', 'masturbated', 'masturbated', _, ['2A','6A'] ).
verb( 'mat', 'mats', 'matting', 'matted', 'matted', _, ['2A','2C','6A','15A'] ).
verb( 'match', 'matches', 'matching', 'matched', 'matched', _, ['2A','6A','12B','13B','14'] ).
verb( 'mate', 'mates', 'mating', 'mated', 'mated', _, ['2A','3A','6A','14'] ).
verb( 'materialize', 'materializes', 'materializing', 'materialized', 'materialized', intran, ['2A','6A'] ).
verb( 'matriculate', 'matriculates', 'matriculating', 'matriculated', 'matriculated', _, ['2A','2C','6A'] ).
verb( 'matter', 'matters', 'mattering', 'mattered', 'mattered', intran, ['2A','2C'] ).
verb( 'maturate', 'maturates', 'maturating', 'maturated', 'maturated', intran, ['2A'] ).
verb( 'mature', 'matures', 'maturing', 'matured', 'matured', _, ['2A','6A'] ).
verb( 'maul', 'mauls', 'mauling', 'mauled', 'mauled', tran, ['6A','15B'] ).
verb( 'maunder', 'maunders', 'maundering', 'maundered', 'maundered', intran, ['2A','2C'] ).
verb( 'maximize', 'maximizes', 'maximizing', 'maximized', 'maximized', tran, ['6A'] ).
verb( 'may', 'may', '-', '-', '-', unknown, ['5'] ).
verb( 'mean', 'means', 'meaning', 'meant', 'meant', tran, ['6A','6A','6C','7A','9','12A','13A','14','16B','17'] ).
verb( 'meander', 'meanders', 'meandering', 'meandered', 'meandered', intran, ['2A','2C'] ).
verb( 'measure', 'measures', 'measuring', 'measured', 'measured', _, ['2A','2B','6A','15A','15B'] ).
verb( 'mechanize', 'mechanizes', 'mechanizing', 'mechanized', 'mechanized', tran, ['6A'] ).
verb( 'meddle', 'meddles', 'meddling', 'meddled', 'meddled', intran, ['2A','3A'] ).
verb( 'mediate', 'mediates', 'mediating', 'mediated', 'mediated', _, ['2A','3A','6A'] ).
verb( 'medicate', 'medicates', 'medicating', 'medicated', 'medicated', tran, ['6A'] ).
verb( 'meditate', 'meditates', 'meditating', 'meditated', 'meditated', _, ['2A','3A','6A'] ).
verb( 'meet', 'meets', 'meeting', 'met', 'met', _, ['2A','2C','6A'] ).
verb( 'meliorate', 'meliorates', 'meliorating', 'meliorated', 'meliorated', _, ['2A','6A'] ).
verb( 'mellow', 'mellows', 'mellowing', 'mellowed', 'mellowed', _, ['2A','6A'] ).
verb( 'melt', 'melts', 'melting', 'melted', 'melted', _, ['2A','2C','6A','15B'] ).
verb( 'memorialize', 'memorializes', 'memorializing', 'memorialized', 'memorialized', tran, ['6A'] ).
verb( 'memorize', 'memorizes', 'memorizing', 'memorized', 'memorized', tran, ['6A'] ).
verb( 'menace', 'menaces', 'menacing', 'menaced', 'menaced', tran, ['6A'] ).
verb( 'mend', 'mends', 'mending', 'mended', 'mended', _, ['2A','6A'] ).
verb( 'menstruate', 'menstruates', 'menstruating', 'menstruated', 'menstruated', intran, ['2A'] ).
verb( 'mention', 'mentions', 'mentioning', 'mentioned', 'mentioned', tran, ['6A','6C','9','13A'] ).
verb( 'mercerize', 'mercerizes', 'mercerizing', 'mercerized', 'mercerized', tran, ['6A'] ).
verb( 'merge', 'merges', 'merging', 'merged', 'merged', _, ['2A','3A','6A','14'] ).
verb( 'merit', 'merits', 'meriting', 'merited', 'merited', tran, ['6A'] ).
verb( 'mesh', 'meshes', 'meshing', 'meshed', 'meshed', _, ['2A','3A','6A'] ).
verb( 'mesmerize', 'mesmerizes', 'mesmerizing', 'mesmerized', 'mesmerized', tran, ['6A'] ).
verb( 'mess', 'messes', 'messing', 'messed', 'messed', _, ['2C','6A','15B'] ).
verb( 'metal', 'metals', 'metalling', 'metalled', 'metalled', tran, ['6A'] ).
verb( 'metamorphose', 'metamorphoses', 'metamorphosing', 'metamorphosed', 'metamorphosed', _, ['3A','6A','14'] ).
verb( 'mete', 'metes', 'meting', 'meted', 'meted', tran, ['15B'] ).
verb( 'metricize', 'metricizes', 'metricizing', 'metricized', 'metricized', tran, [] ).
verb( 'mew', 'mews', 'mewing', 'mewed', 'mewed', intran, ['2A'] ).
verb( 'miaou', 'miaous', 'miaouing', 'miaoued', 'miaoued', intran, [] ).
verb( 'miaow', 'miaows', 'miaowing', 'miaowed', 'miaowed', intran, ['2A'] ).
verb( 'microfilm', 'microfilms', 'microfilming', 'microfilmed', 'microfilmed', tran, ['6A'] ).
verb( 'migrate', 'migrates', 'migrating', 'migrated', 'migrated', intran, ['2A','3A'] ).
verb( 'mildew', 'mildews', 'mildewing', 'mildewed', 'mildewed', _, ['2A','6A'] ).
verb( 'militate', 'militates', 'militating', 'militated', 'militated', intran, ['3A'] ).
verb( 'milk', 'milks', 'milking', 'milked', 'milked', _, ['2A','6A'] ).
verb( 'mill', 'mills', 'milling', 'milled', 'milled', _, ['2C','6A'] ).
verb( 'mime', 'mimes', 'miming', 'mimed', 'mimed', _, ['2A','6A'] ).
verb( 'mimeograph', 'mimeographs', 'mimeographing', 'mimeographed', 'mimeographed', tran, ['6A'] ).
verb( 'mimic', 'mimics', 'mimicking', 'mimicked', 'mimicked', tran, ['6A'] ).
verb( 'mince', 'minces', 'mincing', 'minced', 'minced', _, ['2A','6A'] ).
verb( 'mind', 'minds', 'minding', 'minded', 'minded', _, ['2A','6A','6C','9','19C'] ).
verb( 'mine', 'mines', 'mining', 'mined', 'mined', _, ['2A','3A','6A'] ).
verb( 'mingle', 'mingles', 'mingling', 'mingled', 'mingled', _, ['2A','2C','6A','14'] ).
verb( 'miniaturize', 'miniaturizes', 'miniaturizing', 'miniaturized', 'miniaturized', tran, ['6A'] ).
verb( 'minimize', 'minimizes', 'minimizing', 'minimized', 'minimized', tran, ['6A'] ).
verb( 'minister', 'ministers', 'ministering', 'ministered', 'ministered', intran, ['3A'] ).
verb( 'mint', 'mints', 'minting', 'minted', 'minted', tran, ['6A'] ).
verb( 'minute', 'minutes', 'minuting', 'minuted', 'minuted', tran, ['6A'] ).
verb( 'mire', 'mires', 'miring', 'mired', 'mired', _, ['2A','6A'] ).
verb( 'mirror', 'mirrors', 'mirroring', 'mirrored', 'mirrored', tran, ['6A'] ).
verb( 'misadvise', 'misadvises', 'misadvising', 'misadvised', 'misadvised', tran, ['6A'] ).
verb( 'misapply', 'misapplies', 'misapplying', 'misapplied', 'misapplied', tran, ['6A'] ).
verb( 'misapprehend', 'misapprehends', 'misapprehending', 'misapprehended', 'misapprehended', tran, ['6A'] ).
verb( 'misappropriate', 'misappropriates', 'misappropriating', 'misappropriated', 'misappropriated', tran, ['6A'] ).
verb( 'misbehave', 'misbehaves', 'misbehaving', 'misbehaved', 'misbehaved', _, ['2A','6B'] ).
verb( 'miscalculate', 'miscalculates', 'miscalculating', 'miscalculated', 'miscalculated', _, [] ).
verb( 'miscall', 'miscalls', 'miscalling', 'miscalled', 'miscalled', tran, [] ).
verb( 'miscarry', 'miscarries', 'miscarrying', 'miscarried', 'miscarried', tran, ['2A'] ).
verb( 'miscast', 'miscasts', 'miscasting', 'miscast', 'miscast', tran, ['6A'] ).
verb( 'misconceive', 'misconceives', 'misconceiving', 'misconceived', 'misconceived', _, ['3A','6A'] ).
verb( 'misconduct', 'misconducts', 'misconducting', 'misconducted', 'misconducted', tran, ['6A','6B','14'] ).
verb( 'misconstrue', 'misconstrues', 'misconstruing', 'misconstrued', 'misconstrued', tran, ['6A'] ).
verb( 'miscount', 'miscounts', 'miscounting', 'miscounted', 'miscounted', _, ['2A','6A'] ).
verb( 'misdate', 'misdates', 'misdating', 'misdated', 'misdated', tran, ['6A'] ).
verb( 'misdeal', 'misdeals', 'misdealing', 'misdealt', 'misdealt', _, ['2A','6A'] ).
verb( 'misdirect', 'misdirects', 'misdirecting', 'misdirected', 'misdirected', tran, ['6A'] ).
verb( 'misfire', 'misfires', 'misfiring', 'misfired', 'misfired', intran, ['2A'] ).
verb( 'misgive', 'misgives', 'misgiving', 'misgave', 'misgiven', tran, [] ).
verb( 'misgovern', 'misgoverns', 'misgoverning', 'misgoverned', 'misgoverned', tran, ['6A'] ).
verb( 'misguide', 'misguides', 'misguiding', 'misguided', 'misguided', tran, ['6A','14'] ).
verb( 'mishandle', 'mishandles', 'mishandling', 'mishandled', 'mishandled', tran, [] ).
verb( 'misinform', 'misinforms', 'misinforming', 'misinformed', 'misinformed', tran, ['6A'] ).
verb( 'misinterpret', 'misinterprets', 'misinterpreting', 'misinterpreted', 'misinterpreted', tran, ['6A'] ).
verb( 'misjudge', 'misjudges', 'misjudging', 'misjudged', 'misjudged', _, ['2A','6A'] ).
verb( 'mislay', 'mislays', 'mislaying', 'mislaid', 'mislaid', tran, ['6A'] ).
verb( 'mislead', 'misleads', 'misleading', 'misled', 'misled', tran, ['6A'] ).
verb( 'mismanage', 'mismanages', 'mismanaging', 'mismanaged', 'mismanaged', tran, ['6A'] ).
verb( 'misname', 'misnames', 'misnaming', 'misnamed', 'misnamed', tran, ['6A'] ).
verb( 'misplace', 'misplaces', 'misplacing', 'misplaced', 'misplaced', tran, ['6A'] ).
verb( 'misprint', 'misprints', 'misprinting', 'misprinted', 'misprinted', tran, ['6A'] ).
verb( 'mispronounce', 'mispronounces', 'mispronouncing', 'mispronounced', 'mispronounced', tran, ['6A'] ).
verb( 'misquote', 'misquotes', 'misquoting', 'misquoted', 'misquoted', tran, ['6A'] ).
verb( 'misread', 'misreads', 'misreading', 'misread', 'misread', tran, ['6A'] ).
verb( 'misrepresent', 'misrepresents', 'misrepresenting', 'misrepresented', 'misrepresented', tran, [] ).
verb( 'miss', 'misses', 'missing', 'missed', 'missed', _, ['2A','2C','6A','6B','6C','15B'] ).
verb( 'misspell', 'misspells', 'misspelling', 'misspelled', 'misspelled', tran, ['6A'] ).
verb( 'misspend', 'misspends', 'misspending', 'misspent', 'misspent', tran, ['6A'] ).
verb( 'misstate', 'misstates', 'misstating', 'misstated', 'misstated', tran, ['6A'] ).
verb( 'mist', 'mists', 'misting', 'misted', 'misted', _, ['2C','6A'] ).
verb( 'mistake', 'mistakes', 'mistaking', 'mistook', 'mistaken', _, ['6A','10','14'] ).
verb( 'mistime', 'mistimes', 'mistiming', 'mistimed', 'mistimed', tran, [] ).
verb( 'mistranslate', 'mistranslates', 'mistranslating', 'mistranslated', 'mistranslated', tran, ['6A'] ).
verb( 'mistrust', 'mistrusts', 'mistrusting', 'mistrusted', 'mistrusted', tran, ['6A'] ).
verb( 'misunderstand', 'misunderstands', 'misunderstanding', 'misunderstood', 'misunderstood', tran, ['6A'] ).
verb( 'misuse', 'misuses', 'misusing', 'misused', 'misused', tran, ['6A'] ).
verb( 'mitigate', 'mitigates', 'mitigating', 'mitigated', 'mitigated', tran, ['6A'] ).
verb( 'mix', 'mixes', 'mixing', 'mixed', 'mixed', _, ['2A','2C','3A','6A','12B','13B','14','15B'] ).
verb( 'mizzle', 'mizzles', 'mizzling', 'mizzled', 'mizzled', intran, ['2A'] ).
verb( 'moan', 'moans', 'moaning', 'moaned', 'moaned', _, ['2A','2C','15B'] ).
verb( 'mob', 'mobs', 'mobbing', 'mobbed', 'mobbed', tran, ['6A'] ).
verb( 'mobilize', 'mobilizes', 'mobilizing', 'mobilized', 'mobilized', _, ['2A','6A'] ).
verb( 'mock', 'mocks', 'mocking', 'mocked', 'mocked', _, ['3A','6A'] ).
verb( 'model', 'models', 'modelling', 'modelled', 'modelled', _, ['2A','6A','14','15A'] ).
verb( 'moderate', 'moderates', 'moderating', 'moderated', 'moderated', _, ['2A','6A'] ).
verb( 'modernize', 'modernizes', 'modernizing', 'modernized', 'modernized', tran, ['6A'] ).
verb( 'modify', 'modifies', 'modifying', 'modified', 'modified', tran, ['6A'] ).
verb( 'modulate', 'modulates', 'modulating', 'modulated', 'modulated', _, ['2C','6A'] ).
verb( 'moil', 'moils', 'moiling', 'moiled', 'moiled', intran, [] ).
verb( 'moisten', 'moistens', 'moistening', 'moistened', 'moistened', _, ['2A','6A'] ).
verb( 'molest', 'molests', 'molesting', 'molested', 'molested', tran, ['6A'] ).
verb( 'mollify', 'mollifies', 'mollifying', 'mollified', 'mollified', tran, ['6A'] ).
verb( 'mollycoddle', 'mollycoddles', 'mollycoddling', 'mollycoddled', 'mollycoddled', tran, ['6A'] ).
verb( 'monetize', 'monetizes', 'monetizing', 'monetized', 'monetized', tran, ['6A'] ).
verb( 'monitor', 'monitors', 'monitoring', 'monitored', 'monitored', _, [] ).
verb( 'monkey', 'monkeys', 'monkeying', 'monkeyed', 'monkeyed', intran, ['2C'] ).
verb( 'monopolize', 'monopolizes', 'monopolizing', 'monopolized', 'monopolized', tran, ['6A'] ).
verb( 'moo', 'moos', 'mooing', 'mooed', 'mooed', intran, [] ).
verb( 'mooch', 'mooches', 'mooching', 'mooched', 'mooched', intran, ['2C'] ).
verb( 'moon', 'moons', 'mooning', 'mooned', 'mooned', _, ['2C','15B'] ).
verb( 'moor', 'moors', 'mooring', 'moored', 'moored', tran, ['6A','15A'] ).
verb( 'moot', 'moots', 'mooting', 'mooted', 'mooted', tran, ['6A'] ).
verb( 'mop', 'mops', 'mopping', 'mopped', 'mopped', _, ['6A','15B'] ).
verb( 'mope', 'mopes', 'moping', 'moped', 'moped', intran, ['2A','2C'] ).
verb( 'moralize', 'moralizes', 'moralizing', 'moralized', 'moralized', _, ['2A','3A','6A'] ).
verb( 'mortar', 'mortars', 'mortaring', 'mortared', 'mortared', tran, ['6A'] ).
verb( 'mortgage', 'mortgages', 'mortgaging', 'mortgaged', 'mortgaged', tran, ['6A','14'] ).
verb( 'mortice', 'mortices', 'morticing', 'morticed', 'morticed', tran, ['15A','15B'] ).
verb( 'mortify', 'mortifies', 'mortifying', 'mortified', 'mortified', _, ['2A','6A'] ).
verb( 'mortise', 'mortises', 'mortising', 'mortised', 'mortised', tran, ['15A','15B'] ).
verb( 'mosey', 'moseys', 'moseying', 'moseyed', 'moseyed', intran, ['2A','2C'] ).
verb( 'mother', 'mothers', 'mothering', 'mothered', 'mothered', tran, ['6A'] ).
verb( 'mothproof', 'mothproofs', 'mothproofing', 'mothproofed', 'mothproofed', tran, [] ).
verb( 'motion', 'motions', 'motioning', 'motioned', 'motioned', _, ['3A','15A','15B','17'] ).
verb( 'motivate', 'motivates', 'motivating', 'motivated', 'motivated', tran, ['6A'] ).
verb( 'motor', 'motors', 'motoring', 'motored', 'motored', intran, ['2A','2C'] ).
verb( 'motorize', 'motorizes', 'motorizing', 'motorized', 'motorized', tran, ['6A'] ).
verb( 'mottle', 'mottles', 'mottling', 'mottled', 'mottled', tran, ['6A'] ).
verb( 'mould', 'moulds', 'moulding', 'moulded', 'moulded', _, ['2A','6A','14'] ).
verb( 'moulder', 'moulders', 'mouldering', 'mouldered', 'mouldered', intran, ['2A','2C'] ).
verb( 'moult', 'moults', 'moulting', 'moulted', 'moulted', _, ['2A','6A'] ).
verb( 'mount', 'mounts', 'mounting', 'mounted', 'mounted', _, ['2A','2C','6A'] ).
verb( 'mourn', 'mourns', 'mourning', 'mourned', 'mourned', _, ['3A','6A'] ).
verb( 'mouse', 'mouses', 'mousing', 'moused', 'moused', intran, ['2A'] ).
verb( 'mouth', 'mouths', 'mouthing', 'mouthed', 'mouthed', _, ['2A','6A'] ).
verb( 'move', 'moves', 'moving', 'moved', 'moved', _, ['2A','2C','3A','6A','9','15A','15B','17'] ).
verb( 'mow', 'mows', 'mowing', 'mowed', 'mowed', tran, ['2A','6A','15B'] ).
verb( 'mow', 'mows', 'mowing', 'mowed', 'mowed', intran, [] ).
verb( 'muck', 'mucks', 'mucking', 'mucked', 'mucked', _, ['2C','6A','15B'] ).
verb( 'mud', 'muds', 'mudding', 'mudded', 'mudded', tran, ['6A'] ).
verb( 'muddle', 'muddles', 'muddling', 'muddled', 'muddled', _, ['2C','6A','15B'] ).
verb( 'muddy', 'muddies', 'muddying', 'muddied', 'muddied', tran, ['6A'] ).
verb( 'muff', 'muffs', 'muffing', 'muffed', 'muffed', tran, ['6A'] ).
verb( 'muffle', 'muffles', 'muffling', 'muffled', 'muffled', tran, ['6A','15B'] ).
verb( 'mug', 'mugs', 'mugging', 'mugged', 'mugged', tran, ['6A','15B'] ).
verb( 'mulch', 'mulches', 'mulching', 'mulched', 'mulched', tran, [] ).
verb( 'mulct', 'mulcts', 'mulcting', 'mulcted', 'mulcted', tran, ['12C','14'] ).
verb( 'mull', 'mulls', 'mulling', 'mulled', 'mulled', tran, ['6A','15A'] ).
verb( 'multiply', 'multiplies', 'multiplying', 'multiplied', 'multiplied', _, ['2A','6A','14'] ).
verb( 'mumble', 'mumbles', 'mumbling', 'mumbled', 'mumbled', _, ['2A','2C','6A'] ).
verb( 'mummify', 'mummifies', 'mummifying', 'mummified', 'mummified', tran, ['6A'] ).
verb( 'munch', 'munches', 'munching', 'munched', 'munched', _, ['2A','2C','6A'] ).
verb( 'munition', 'munitions', 'munitioning', 'munitioned', 'munitioned', tran, ['6A'] ).
verb( 'murder', 'murders', 'murdering', 'murdered', 'murdered', tran, ['6A'] ).
verb( 'murmur', 'murmurs', 'murmuring', 'murmured', 'murmured', _, ['2A','2C','3A','6A'] ).
verb( 'muscle', 'muscles', 'muscling', 'muscled', 'muscled', intran, ['2C'] ).
verb( 'muse', 'muses', 'musing', 'mused', 'mused', intran, ['2A','3A'] ).
verb( 'mushroom', 'mushrooms', 'mushrooming', 'mushroomed', 'mushroomed', intran, ['2C'] ).
verb( 'muss', 'musses', 'mussing', 'mussed', 'mussed', tran, ['6A','15B'] ).
verb( 'must', 'must', '-', '-', '-', unknown, ['5'] ).
verb( 'muster', 'musters', 'mustering', 'mustered', 'mustered', _, ['2A','6A','15B'] ).
verb( 'mute', 'mutes', 'muting', 'muted', 'muted', tran, ['6A'] ).
verb( 'mutilate', 'mutilates', 'mutilating', 'mutilated', 'mutilated', tran, ['6A'] ).
verb( 'mutiny', 'mutinies', 'mutinying', 'mutinied', 'mutinied', intran, ['2A','3A'] ).
verb( 'mutter', 'mutters', 'muttering', 'muttered', 'muttered', _, ['2A','2C','6A','14'] ).
verb( 'muzzle', 'muzzles', 'muzzling', 'muzzled', 'muzzled', tran, ['6A'] ).
verb( 'mystify', 'mystifies', 'mystifying', 'mystified', 'mystified', tran, ['6A'] ).
verb( 'nab', 'nabs', 'nabbing', 'nabbed', 'nabbed', tran, [] ).
verb( 'nag', 'nags', 'nagging', 'nagged', 'nagged', _, ['2A','2C','3A','6A'] ).
verb( 'nail', 'nails', 'nailing', 'nailed', 'nailed', tran, ['15A','15B'] ).
verb( 'name', 'names', 'naming', 'named', 'named', tran, ['6A','14','23'] ).
verb( 'name-drop', 'name-drops', 'name-dropping', 'name-dropped', 'name-dropped', intran, ['2A'] ).
verb( 'nap', 'naps', 'napping', 'napped', 'napped', intran, [] ).
verb( 'nark', 'narks', 'narking', 'narked', 'narked', tran, [] ).
verb( 'narrate', 'narrates', 'narrating', 'narrated', 'narrated', tran, ['6A'] ).
verb( 'narrow', 'narrows', 'narrowing', 'narrowed', 'narrowed', _, ['2A','6A'] ).
verb( 'nasalize', 'nasalizes', 'nasalizing', 'nasalized', 'nasalized', tran, ['6A'] ).
verb( 'nationalize', 'nationalizes', 'nationalizing', 'nationalized', 'nationalized', tran, ['6A'] ).
verb( 'natter', 'natters', 'nattering', 'nattered', 'nattered', intran, ['2A','2C'] ).
verb( 'naturalize', 'naturalizes', 'naturalizing', 'naturalized', 'naturalized', _, ['2A','6A'] ).
verb( 'nauseate', 'nauseates', 'nauseating', 'nauseated', 'nauseated', tran, ['6A'] ).
verb( 'navigate', 'navigates', 'navigating', 'navigated', 'navigated', _, ['2A','6A'] ).
verb( 'near', 'nears', 'nearing', 'neared', 'neared', _, ['2A','6A'] ).
verb( 'necessitate', 'necessitates', 'necessitating', 'necessitated', 'necessitated', tran, ['6A','6C'] ).
verb( 'neck', 'necks', 'necking', 'necked', 'necked', intran, [] ).
verb( 'need', 'needs', 'needing', 'needed', 'needed', unknown, ['5','6A','6E','7A'] ).
verb( 'need', 'needs', 'needing', 'needed', 'needed', tran, ['5','6A','6E','7A'] ).
verb( 'needle', 'needles', 'needling', 'needled', 'needled', tran, ['6A','15A'] ).
verb( 'negate', 'negates', 'negating', 'negated', 'negated', tran, ['6A'] ).
verb( 'negative', 'negatives', 'negativing', 'negatived', 'negatived', tran, ['6A'] ).
verb( 'neglect', 'neglects', 'neglecting', 'neglected', 'neglected', tran, ['6A','6C','7A'] ).
verb( 'negotiate', 'negotiates', 'negotiating', 'negotiated', 'negotiated', _, ['2A','3A','6A','14'] ).
verb( 'neigh', 'neighs', 'neighing', 'neighed', 'neighed', intran, [] ).
verb( 'neighbour', 'neighbours', 'neighbouring', 'neighboured', 'neighboured', _, ['3A','6A'] ).
verb( 'nerve', 'nerves', 'nerving', 'nerved', 'nerved', tran, ['6A','14','16A'] ).
verb( 'nest', 'nests', 'nesting', 'nested', 'nested', intran, ['2A','2C'] ).
verb( 'nestle', 'nestles', 'nestling', 'nestled', 'nestled', _, ['2C','15A'] ).
verb( 'net', 'nets', 'netting', 'netted', 'netted', tran, ['6A'] ).
verb( 'nett', 'netts', 'netting', 'netted', 'netted', tran, ['6A'] ).
verb( 'nettle', 'nettles', 'nettling', 'nettled', 'nettled', tran, ['6A'] ).
verb( 'neuter', 'neuters', 'neutering', 'neutered', 'neutered', tran, [] ).
verb( 'neutralize', 'neutralizes', 'neutralizing', 'neutralized', 'neutralized', tran, ['6A'] ).
verb( 'nibble', 'nibbles', 'nibbling', 'nibbled', 'nibbled', _, ['2A','3A','6A'] ).
verb( 'nick', 'nicks', 'nicking', 'nicked', 'nicked', tran, [] ).
verb( 'nickel', 'nickels', 'nickelling', 'nickelled', 'nickelled', tran, [] ).
verb( 'nickname', 'nicknames', 'nicknaming', 'nicknamed', 'nicknamed', tran, ['6A','23'] ).
verb( 'niggle', 'niggles', 'niggling', 'niggled', 'niggled', intran, [] ).
verb( 'nip', 'nips', 'nipping', 'nipped', 'nipped', _, ['2A','2C','6A','15A','15B'] ).
verb( 'nobble', 'nobbles', 'nobbling', 'nobbled', 'nobbled', tran, ['6A'] ).
verb( 'nod', 'nods', 'nodding', 'nodded', 'nodded', _, ['2A','2C','3A','4A','6A','12A','13A'] ).
verb( 'noise', 'noises', 'noising', 'noised', 'noised', tran, [] ).
verb( 'nominate', 'nominates', 'nominating', 'nominated', 'nominated', tran, ['6A','14','23'] ).
verb( 'nonplus', 'nonplusses', 'nonplussing', 'nonplussed', 'nonplussed', tran, ['6A'] ).
verb( 'noose', 'nooses', 'noosing', 'noosed', 'noosed', tran, [] ).
verb( 'normalize', 'normalizes', 'normalizing', 'normalized', 'normalized', tran, [] ).
verb( 'nose', 'noses', 'nosing', 'nosed', 'nosed', _, ['2C','3A','15A','15B'] ).
verb( 'nosedive', 'nosedives', 'nosediving', 'nosedived', 'nosedived', intran, [] ).
verb( 'nosh', 'noshes', 'noshing', 'noshed', 'noshed', intran, [] ).
verb( 'notch', 'notches', 'notching', 'notched', 'notched', tran, ['6A','15B'] ).
verb( 'note', 'notes', 'noting', 'noted', 'noted', tran, ['6A','8','9','10','15B'] ).
verb( 'notice', 'notices', 'noticing', 'noticed', 'noticed', _, ['2A','6A','8','9','10','18A','19A'] ).
verb( 'notify', 'notifies', 'notifying', 'notified', 'notified', tran, ['6A','11','14'] ).
verb( 'nourish', 'nourishes', 'nourishing', 'nourished', 'nourished', tran, ['6A'] ).
verb( 'nudge', 'nudges', 'nudging', 'nudged', 'nudged', tran, ['6A'] ).
verb( 'nullify', 'nullifies', 'nullifying', 'nullified', 'nullified', tran, ['6A'] ).
verb( 'numb', 'numbs', 'numbing', 'numbed', 'numbed', tran, ['6A'] ).
verb( 'number', 'numbers', 'numbering', 'numbered', 'numbered', tran, ['2C','6A','14'] ).
verb( 'nurse', 'nurses', 'nursing', 'nursed', 'nursed', tran, ['6A'] ).
verb( 'nurture', 'nurtures', 'nurturing', 'nurtured', 'nurtured', tran, [] ).
verb( 'nut', 'nuts', 'nutting', 'nutted', 'nutted', intran, [] ).
verb( 'nuzzle', 'nuzzles', 'nuzzling', 'nuzzled', 'nuzzled', _, ['2C','6A'] ).
verb( 'obey', 'obeys', 'obeying', 'obeyed', 'obeyed', _, ['2A','6A'] ).
verb( 'obfuscate', 'obfuscates', 'obfuscating', 'obfuscated', 'obfuscated', tran, ['6A'] ).
verb( 'object', 'objects', 'objecting', 'objected', 'objected', _, ['2A','3A','9'] ).
verb( 'objurgate', 'objurgates', 'objurgating', 'objurgated', 'objurgated', tran, ['6A'] ).
verb( 'obligate', 'obligates', 'obligating', 'obligated', 'obligated', tran, ['17'] ).
verb( 'oblige', 'obliges', 'obliging', 'obliged', 'obliged', tran, ['6A','14','17'] ).
verb( 'obliterate', 'obliterates', 'obliterating', 'obliterated', 'obliterated', tran, ['6A'] ).
verb( 'obscure', 'obscures', 'obscuring', 'obscured', 'obscured', tran, ['6A'] ).
verb( 'observe', 'observes', 'observing', 'observed', 'observed', _, ['2A','6A','8','9','10','18A','19A','25'] ).
verb( 'obsess', 'obsesses', 'obsessing', 'obsessed', 'obsessed', tran, ['6A'] ).
verb( 'obstruct', 'obstructs', 'obstructing', 'obstructed', 'obstructed', tran, ['6A'] ).
verb( 'obtain', 'obtains', 'obtaining', 'obtained', 'obtained', _, ['2A','6A'] ).
verb( 'obtrude', 'obtrudes', 'obtruding', 'obtruded', 'obtruded', _, ['2A','14'] ).
verb( 'obviate', 'obviates', 'obviating', 'obviated', 'obviated', tran, ['6A'] ).
verb( 'occasion', 'occasions', 'occasioning', 'occasioned', 'occasioned', tran, ['6A','12A','13A'] ).
verb( 'occupy', 'occupies', 'occupying', 'occupied', 'occupied', tran, ['6A'] ).
verb( 'occur', 'occurs', 'occurring', 'occurred', 'occurred', intran, ['2A','2C','3A'] ).
verb( 'offend', 'offends', 'offending', 'offended', 'offended', _, ['3A','6A'] ).
verb( 'offer', 'offers', 'offering', 'offered', 'offered', _, ['2A','6A','7A','12A','13A','14','15B'] ).
verb( 'officiate', 'officiates', 'officiating', 'officiated', 'officiated', intran, ['2A','2C','3A'] ).
verb( 'offset', 'offsets', 'offsetting', 'offset', 'offset', tran, ['6A','14'] ).
verb( 'ogle', 'ogles', 'ogling', 'ogled', 'ogled', _, ['3A','6A'] ).
verb( 'oil', 'oils', 'oiling', 'oiled', 'oiled', tran, ['6A'] ).
verb( 'okay', 'okays', 'okaying', 'okayed', 'okayed', tran, ['6A'] ).
verb( 'omen', 'omens', 'omening', 'omened', 'omened', tran, ['6A'] ).
verb( 'omit', 'omits', 'omitting', 'omitted', 'omitted', tran, ['6A','6C','7A'] ).
verb( 'ooze', 'oozes', 'oozing', 'oozed', 'oozed', _, ['2C','6A'] ).
verb( 'open', 'opens', 'opening', 'opened', 'opened', _, ['2A','2C','3A','6A','12C','14','15A','15B','16A'] ).
verb( 'operate', 'operates', 'operating', 'operated', 'operated', _, ['2A','2C','3A','4A','6A'] ).
verb( 'opine', 'opines', 'opining', 'opined', 'opined', tran, ['9'] ).
verb( 'oppose', 'opposes', 'opposing', 'opposed', 'opposed', tran, ['6A','14'] ).
verb( 'oppress', 'oppresses', 'oppressing', 'oppressed', 'oppressed', tran, ['6A'] ).
verb( 'oppugn', 'oppugns', 'oppugning', 'oppugned', 'oppugned', tran, ['6A'] ).
verb( 'opt', 'opts', 'opting', 'opted', 'opted', intran, ['3A'] ).
verb( 'orate', 'orates', 'orating', 'orated', 'orated', intran, ['2A'] ).
verb( 'orbit', 'orbits', 'orbiting', 'orbited', 'orbited', _, ['2A','2C','6A'] ).
verb( 'orchestrate', 'orchestrates', 'orchestrating', 'orchestrated', 'orchestrated', tran, ['6A'] ).
verb( 'ordain', 'ordains', 'ordaining', 'ordained', 'ordained', tran, ['6A','9','23'] ).
verb( 'order', 'orders', 'ordering', 'ordered', 'ordered', tran, ['6A','9','12B','13B','15A','15B','17'] ).
verb( 'organize', 'organizes', 'organizing', 'organized', 'organized', tran, ['6A'] ).
verb( 'orient', 'orients', 'orienting', 'oriented', 'oriented', tran, [] ).
verb( 'orientate', 'orientates', 'orientating', 'orientated', 'orientated', tran, ['6A'] ).
verb( 'originate', 'originates', 'originating', 'originated', 'originated', _, ['2C','3A','6A'] ).
verb( 'ornament', 'ornaments', 'ornamenting', 'ornamented', 'ornamented', tran, ['6A','14'] ).
verb( 'orphan', 'orphans', 'orphaning', 'orphaned', 'orphaned', tran, ['6A'] ).
verb( 'oscillate', 'oscillates', 'oscillating', 'oscillated', 'oscillated', _, ['2A','6A'] ).
verb( 'ossify', 'ossifies', 'ossifying', 'ossified', 'ossified', _, ['2A','6A'] ).
verb( 'ostracize', 'ostracizes', 'ostracizing', 'ostracized', 'ostracized', tran, ['6A'] ).
verb( 'ought', 'ought', '-', '-', '-', unknown, ['7B'] ).
verb( 'oust', 'ousts', 'ousting', 'ousted', 'ousted', tran, ['6A','14'] ).
verb( 'out', 'outs', 'outing', 'outed', 'outed', tran, [] ).
verb( 'out-herod', 'out-herods', 'out-heroding', 'out-heroded', 'out-heroded', tran, ['6A'] ).
verb( 'outbalance', 'outbalances', 'outbalancing', 'outbalanced', 'outbalanced', tran, ['6A'] ).
verb( 'outbid', 'outbids', 'outbidding', 'outbid', 'outbid', tran, ['6A'] ).
verb( 'outbrave', 'outbraves', 'outbraving', 'outbraved', 'outbraved', tran, ['6A'] ).
verb( 'outclass', 'outclasses', 'outclassing', 'outclassed', 'outclassed', tran, ['6A'] ).
verb( 'outdistance', 'outdistances', 'outdistancing', 'outdistanced', 'outdistanced', tran, ['6A'] ).
verb( 'outdo', 'outdoes', 'outdoing', 'outdid', 'outdone', tran, ['6A'] ).
verb( 'outface', 'outfaces', 'outfacing', 'outfaced', 'outfaced', tran, ['6A'] ).
verb( 'outfight', 'outfights', 'outfighting', 'outfought', 'outfought', tran, ['6A'] ).
verb( 'outfit', 'outfits', 'outfitting', 'outfitted', 'outfitted', tran, [] ).
verb( 'outflank', 'outflanks', 'outflanking', 'outflanked', 'outflanked', tran, ['6A'] ).
verb( 'outfox', 'outfoxes', 'outfoxing', 'outfoxed', 'outfoxed', tran, ['6A'] ).
verb( 'outgo', 'outgoes', 'outgoing', 'outwent', 'outgone', intran, [] ).
verb( 'outgrow', 'outgrows', 'outgrowing', 'outgrew', 'outgrown', tran, ['6A'] ).
verb( 'outlast', 'outlasts', 'outlasting', 'outlasted', 'outlasted', tran, ['6A'] ).
verb( 'outlaw', 'outlaws', 'outlawing', 'outlawed', 'outlawed', tran, ['6A'] ).
verb( 'outline', 'outlines', 'outlining', 'outlined', 'outlined', tran, ['6A'] ).
verb( 'outlive', 'outlives', 'outliving', 'outlived', 'outlived', tran, ['6A'] ).
verb( 'outmanoeuvre', 'outmanoeuvres', 'outmanoeuvring', 'outmanoeuvred', 'outmanoeuvred', tran, ['6A'] ).
verb( 'outmarch', 'outmarches', 'outmarching', 'outmarched', 'outmarched', tran, ['6A'] ).
verb( 'outmatch', 'outmatches', 'outmatching', 'outmatched', 'outmatched', tran, ['6A'] ).
verb( 'outnumber', 'outnumbers', 'outnumbering', 'outnumbered', 'outnumbered', tran, ['6A'] ).
verb( 'outplay', 'outplays', 'outplaying', 'outplayed', 'outplayed', tran, ['6A'] ).
verb( 'outpoint', 'outpoints', 'outpointing', 'outpointed', 'outpointed', tran, ['6A'] ).
verb( 'outrage', 'outrages', 'outraging', 'outraged', 'outraged', tran, ['6A'] ).
verb( 'outrange', 'outranges', 'outranging', 'outranged', 'outranged', tran, ['6A'] ).
verb( 'outrank', 'outranks', 'outranking', 'outranked', 'outranked', tran, ['6A'] ).
verb( 'outride', 'outrides', 'outriding', 'outrode', 'outridden', tran, ['6A'] ).
verb( 'outrival', 'outrivals', 'outrivalling', 'outrivalled', 'outrivalled', tran, ['6A'] ).
verb( 'outrun', 'outruns', 'outrunning', 'outran', 'outrun', tran, ['6A'] ).
verb( 'outsail', 'outsails', 'outsailing', 'outsailed', 'outsailed', tran, ['6A'] ).
verb( 'outshine', 'outshines', 'outshining', 'outshone', 'outshone', tran, ['6A'] ).
verb( 'outsmart', 'outsmarts', 'outsmarting', 'outsmarted', 'outsmarted', tran, ['6A'] ).
verb( 'outspan', 'outspans', 'outspanning', 'outspanned', 'outspanned', _, ['2A','6A'] ).
verb( 'outstay', 'outstays', 'outstaying', 'outstayed', 'outstayed', tran, ['6A'] ).
verb( 'outstrip', 'outstrips', 'outstripping', 'outstripped', 'outstripped', tran, ['6A'] ).
verb( 'outvie', 'outvies', 'outvying', 'outvied', 'outvied', tran, ['6A'] ).
verb( 'outvote', 'outvotes', 'outvoting', 'outvoted', 'outvoted', tran, ['6A'] ).
verb( 'outwear', 'outwears', 'outwearing', 'outwore', 'outworn', tran, ['6A'] ).
verb( 'outweigh', 'outweighs', 'outweighing', 'outweighed', 'outweighed', tran, ['6A'] ).
verb( 'outwit', 'outwits', 'outwitting', 'outwitted', 'outwitted', tran, ['6A'] ).
verb( 'overact', 'overacts', 'overacting', 'overacted', 'overacted', _, ['2A','6A'] ).
verb( 'overarch', 'overarches', 'overarching', 'overarched', 'overarched', _, ['2A','6A'] ).
verb( 'overawe', 'overawes', 'overawing', 'overawed', 'overawed', tran, ['6A'] ).
verb( 'overbalance', 'overbalances', 'overbalancing', 'overbalanced', 'overbalanced', _, ['2A','6A'] ).
verb( 'overbear', 'overbears', 'overbearing', 'overbore', 'overborne', tran, ['6A'] ).
verb( 'overbid', 'overbids', 'overbidding', 'overbid', 'overbid', _, ['2A','6A'] ).
verb( 'overburden', 'overburdens', 'overburdening', 'overburdened', 'overburdened', tran, ['6A'] ).
verb( 'overcall', 'overcalls', 'overcalling', 'overcalled', 'overcalled', _, [] ).
verb( 'overcapitalize', 'overcapitalizes', 'overcapitalizing', 'overcapitalized', 'overcapitalized', tran, [] ).
verb( 'overcharge', 'overcharges', 'overcharging', 'overcharged', 'overcharged', _, ['2A','6A'] ).
verb( 'overclothe', 'overclothes', 'overclothing', 'overclothed', 'overclothed', tran, ['6A'] ).
verb( 'overcloud', 'overclouds', 'overclouding', 'overclouded', 'overclouded', _, ['2A','6A'] ).
verb( 'overcome', 'overcomes', 'overcoming', 'overcame', 'overcome', tran, ['6A'] ).
verb( 'overcook', 'overcooks', 'overcooking', 'overcooked', 'overcooked', tran, [] ).
verb( 'overcrop', 'overcrops', 'overcropping', 'overcropped', 'overcropped', tran, ['6A'] ).
verb( 'overcrowd', 'overcrowds', 'overcrowding', 'overcrowded', 'overcrowded', tran, ['6A'] ).
verb( 'overdo', 'overdoes', 'overdoing', 'overdid', 'overdone', tran, ['6A'] ).
verb( 'overdraw', 'overdraws', 'overdrawing', 'overdrew', 'overdrawn', _, ['2A','6A'] ).
verb( 'overdress', 'overdresses', 'overdressing', 'overdressed', 'overdressed', _, ['2A','6A'] ).
verb( 'overeat', 'overeats', 'overeating', 'overate', 'overeaten', intran, [] ).
verb( 'overemphasize', 'overemphasizes', 'overemphasizing', 'overemphasized', 'overemphasized', tran, [] ).
verb( 'overestimate', 'overestimates', 'overestimating', 'overestimated', 'overestimated', tran, [] ).
verb( 'overexert', 'overexerts', 'overexerting', 'overexerted', 'overexerted', tran, [] ).
verb( 'overexpose', 'overexposes', 'overexposing', 'overexposed', 'overexposed', tran, [] ).
verb( 'overfeed', 'overfeeds', 'overfeeding', 'overfeeded', 'overfeeded', _, ['2A','2C','3A','6A','14','15A','15B'] ).
verb( 'overflow', 'overflows', 'overflowing', 'overflowed', 'overflowed', _, ['2A','3A','6A'] ).
verb( 'overfly', 'overflies', 'overflying', 'overflew', 'overflown', tran, ['2A','2B','2C','2D','4A','6A','15A','15B'] ).
verb( 'overhang', 'overhangs', 'overhanging', 'overhung', 'overhung', _, ['2A','6A'] ).
verb( 'overhaul', 'overhauls', 'overhauling', 'overhauled', 'overhauled', tran, ['6A'] ).
verb( 'overhear', 'overhears', 'overhearing', 'overheard', 'overheard', tran, ['6A','18A','19A'] ).
verb( 'overheat', 'overheats', 'overheating', 'overheated', 'overheated', tran, [] ).
verb( 'overindulge', 'overindulges', 'overindulging', 'overindulged', 'overindulged', _, [] ).
verb( 'overlap', 'overlaps', 'overlapping', 'overlapped', 'overlapped', _, ['2A','6A'] ).
verb( 'overlay', 'overlays', 'overlaying', 'overlaid', 'overlain', tran, [] ).
verb( 'overleap', 'overleaps', 'overleaping', 'overleaped', 'overleaped', tran, ['6A'] ).
verb( 'overlie', 'overlies', 'overlying', 'overlaid', 'overlaid', intran, ['2A','2C','2D','3A'] ).
verb( 'overload', 'overloads', 'overloading', 'overloaded', 'overloaded', tran, ['6A'] ).
verb( 'overlook', 'overlooks', 'overlooking', 'overlooked', 'overlooked', tran, ['6A'] ).
verb( 'overmaster', 'overmasters', 'overmastering', 'overmastered', 'overmastered', tran, ['6A'] ).
verb( 'overpay', 'overpays', 'overpaying', 'overpaid', 'overpaid', tran, ['6A','14'] ).
verb( 'overplay', 'overplays', 'overplaying', 'overplayed', 'overplayed', tran, [] ).
verb( 'overpower', 'overpowers', 'overpowering', 'overpowered', 'overpowered', tran, ['6A'] ).
verb( 'overpraise', 'overpraises', 'overpraising', 'overpraised', 'overpraised', tran, [] ).
verb( 'overprint', 'overprints', 'overprinting', 'overprinted', 'overprinted', tran, ['6A'] ).
verb( 'overproduce', 'overproduces', 'overproducing', 'overproduced', 'overproduced', _, [] ).
verb( 'overrate', 'overrates', 'overrating', 'overrated', 'overrated', tran, ['6A'] ).
verb( 'overreach', 'overreaches', 'overreaching', 'overreached', 'overreached', tran, ['6A'] ).
verb( 'override', 'overrides', 'overriding', 'overrode', 'overridden', tran, ['6A'] ).
verb( 'overrule', 'overrules', 'overruling', 'overruled', 'overruled', tran, ['6A'] ).
verb( 'overrun', 'overruns', 'overrunning', 'overran', 'overrun', tran, ['6A'] ).
verb( 'oversee', 'oversees', 'overseeing', 'oversaw', 'overseen', tran, ['6A'] ).
verb( 'oversew', 'oversews', 'oversewing', 'oversewed', 'oversewn', tran, ['2A','2B','2C','6A','15B'] ).
verb( 'overshadow', 'overshadows', 'overshadowing', 'overshadowed', 'overshadowed', tran, ['6A'] ).
verb( 'overshoot', 'overshoots', 'overshooting', 'overshot', 'overshot', tran, ['6A'] ).
verb( 'oversimplify', 'oversimplifies', 'oversimplifying', 'oversimplified', 'oversimplified', tran, [] ).
verb( 'oversleep', 'oversleeps', 'oversleeping', 'overslept', 'overslept', intran, ['2A'] ).
verb( 'overspend', 'overspends', 'overspending', 'overspent', 'overspent', _, ['2A','6A','14','19B'] ).
verb( 'overstate', 'overstates', 'overstating', 'overstated', 'overstated', tran, ['6A'] ).
verb( 'overstay', 'overstays', 'overstaying', 'overstayed', 'overstayed', tran, ['6A'] ).
verb( 'overstep', 'oversteps', 'overstepping', 'overstepped', 'overstepped', tran, ['6A'] ).
verb( 'overstock', 'overstocks', 'overstocking', 'overstocked', 'overstocked', tran, ['6A'] ).
verb( 'overstrain', 'overstrains', 'overstraining', 'overstrained', 'overstrained', tran, [] ).
verb( 'overtake', 'overtakes', 'overtaking', 'overtook', 'overtaken', tran, ['6A'] ).
verb( 'overtax', 'overtaxes', 'overtaxing', 'overtaxed', 'overtaxed', tran, ['6A'] ).
verb( 'overthrow', 'overthrows', 'overthrowing', 'overthrew', 'overthrown', tran, ['6A'] ).
verb( 'overtop', 'overtops', 'overtopping', 'overtopped', 'overtopped', tran, ['6A'] ).
verb( 'overtrump', 'overtrumps', 'overtrumping', 'overtrumped', 'overtrumped', tran, ['6A'] ).
verb( 'overturn', 'overturns', 'overturning', 'overturned', 'overturned', _, ['2A','6A'] ).
verb( 'overvalue', 'overvalues', 'overvaluing', 'overvalued', 'overvalued', tran, [] ).
verb( 'overwhelm', 'overwhelms', 'overwhelming', 'overwhelmed', 'overwhelmed', tran, ['6A'] ).
verb( 'overwork', 'overworks', 'overworking', 'overworked', 'overworked', _, ['2A','6A'] ).
verb( 'owe', 'owes', 'owing', 'owed', 'owed', _, ['2A','3A','6A','12A','13A','14'] ).
verb( 'own', 'owns', 'owning', 'owned', 'owned', _, ['2C','3B','6A','9'] ).
verb( 'oxidize', 'oxidizes', 'oxidizing', 'oxidized', 'oxidized', _, ['2A','6A'] ).
verb( 'oxygenate', 'oxygenates', 'oxygenating', 'oxygenated', 'oxygenated', tran, [] ).
verb( 'oxygenize', 'oxygenizes', 'oxygenizing', 'oxygenized', 'oxygenized', tran, [] ).
verb( 'pace', 'paces', 'pacing', 'paced', 'paced', _, ['2A','2C','6A','15B'] ).
verb( 'pacify', 'pacifies', 'pacifying', 'pacified', 'pacified', tran, ['6A'] ).
verb( 'pack', 'packs', 'packing', 'packed', 'packed', _, ['2A','2C','3A','6A','14','15A','15B'] ).
verb( 'package', 'packages', 'packaging', 'packaged', 'packaged', tran, [] ).
verb( 'pad', 'pads', 'padding', 'padded', 'padded', _, ['2A','2C','6A','15B'] ).
verb( 'paddle', 'paddles', 'paddling', 'paddled', 'paddled', _, ['2A','6A'] ).
verb( 'padlock', 'padlocks', 'padlocking', 'padlocked', 'padlocked', tran, ['6A'] ).
verb( 'page', 'pages', 'paging', 'paged', 'paged', tran, ['6A'] ).
verb( 'pain', 'pains', 'paining', 'pained', 'pained', tran, ['6A'] ).
verb( 'paint', 'paints', 'painting', 'painted', 'painted', _, ['2A','2C','6A','15A','15B','22'] ).
verb( 'pair', 'pairs', 'pairing', 'paired', 'paired', _, ['2A','2C','6A','15B'] ).
verb( 'pal', 'pals', 'palling', 'palled', 'palled', intran, [] ).
verb( 'palaver', 'palavers', 'palavering', 'palavered', 'palavered', intran, ['2A'] ).
verb( 'pale', 'pales', 'paling', 'paled', 'paled', intran, ['2A','2C'] ).
verb( 'palisade', 'palisades', 'palisading', 'palisaded', 'palisaded', tran, ['6A'] ).
verb( 'pall', 'palls', 'palling', 'palled', 'palled', intran, ['2A','3A'] ).
verb( 'palliate', 'palliates', 'palliating', 'palliated', 'palliated', tran, ['6A'] ).
verb( 'palm', 'palms', 'palming', 'palmed', 'palmed', tran, ['6A','15B'] ).
verb( 'palpitate', 'palpitates', 'palpitating', 'palpitated', 'palpitated', intran, ['2A','2C'] ).
verb( 'palsy', 'palsies', 'palsying', 'palsied', 'palsied', tran, [] ).
verb( 'palter', 'palters', 'paltering', 'paltered', 'paltered', intran, ['3A'] ).
verb( 'pamper', 'pampers', 'pampering', 'pampered', 'pampered', tran, ['6A'] ).
verb( 'pan', 'pans', 'panning', 'panned', 'panned', _, ['2C','3A','6A','15B'] ).
verb( 'pander', 'panders', 'pandering', 'pandered', 'pandered', intran, ['3A'] ).
verb( 'panel', 'panels', 'panelling', 'panelled', 'panelled', tran, [] ).
verb( 'panhandle', 'panhandles', 'panhandling', 'panhandled', 'panhandled', intran, [] ).
verb( 'panic', 'panics', 'panicking', 'panicked', 'panicked', intran, [] ).
verb( 'pant', 'pants', 'panting', 'panted', 'panted', _, ['2A','2C','3A','6A','15B'] ).
verb( 'paper', 'papers', 'papering', 'papered', 'papered', tran, ['6A','15B'] ).
verb( 'parachute', 'parachutes', 'parachuting', 'parachuted', 'parachuted', _, ['2A','2C','6A','15A','15B'] ).
verb( 'parade', 'parades', 'parading', 'paraded', 'paraded', _, ['2A','6A'] ).
verb( 'paragraph', 'paragraphs', 'paragraphing', 'paragraphed', 'paragraphed', tran, [] ).
verb( 'parallel', 'parallels', 'parallelling', 'parallelled', 'parallelled', tran, ['6A'] ).
verb( 'paralyse', 'paralyses', 'paralysing', 'paralysed', 'paralysed', tran, ['6A'] ).
verb( 'paralyze', 'paralyzes', 'paralyzing', 'paralyzed', 'paralyzed', tran, ['6A'] ).
verb( 'paraphrase', 'paraphrases', 'paraphrasing', 'paraphrased', 'paraphrased', tran, ['6A'] ).
verb( 'parboil', 'parboils', 'parboiling', 'parboiled', 'parboiled', tran, ['6A'] ).
verb( 'parcel', 'parcels', 'parcelling', 'parcelled', 'parcelled', tran, ['6A','15B'] ).
verb( 'parch', 'parches', 'parching', 'parched', 'parched', tran, ['6A'] ).
verb( 'pardon', 'pardons', 'pardoning', 'pardoned', 'pardoned', tran, ['6A','12B','13B'] ).
verb( 'pare', 'pares', 'paring', 'pared', 'pared', tran, ['6A','15B'] ).
verb( 'park', 'parks', 'parking', 'parked', 'parked', _, ['2A','6A','15A'] ).
verb( 'parley', 'parleys', 'parleying', 'parleyed', 'parleyed', intran, ['2A','3A'] ).
verb( 'parody', 'parodies', 'parodying', 'parodied', 'parodied', tran, ['6A'] ).
verb( 'parole', 'paroles', 'paroling', 'paroled', 'paroled', tran, ['6A'] ).
verb( 'parry', 'parries', 'parrying', 'parried', 'parried', tran, ['6A'] ).
verb( 'parse', 'parses', 'parsing', 'parsed', 'parsed', tran, ['6A'] ).
verb( 'part', 'parts', 'parting', 'parted', 'parted', _, ['2A','2D','3A','6A'] ).
verb( 'partake', 'partakes', 'partaking', 'partook', 'partaken', _, ['3A'] ).
verb( 'participate', 'participates', 'participating', 'participated', 'participated', intran, ['2A','3A'] ).
verb( 'particularize', 'particularizes', 'particularizing', 'particularized', 'particularized', _, ['2A','6A'] ).
verb( 'partition', 'partitions', 'partitioning', 'partitioned', 'partitioned', tran, ['6A','15B'] ).
verb( 'partner', 'partners', 'partnering', 'partnered', 'partnered', tran, ['6A','15A'] ).
verb( 'pass', 'passes', 'passing', 'passed', 'passed', _, ['2A','2C','3A','6A','12A','13A','14','15A','15B'] ).
verb( 'paste', 'pastes', 'pasting', 'pasted', 'pasted', tran, ['6A','15A','15B'] ).
verb( 'pasteurize', 'pasteurizes', 'pasteurizing', 'pasteurized', 'pasteurized', tran, ['6A'] ).
verb( 'pasture', 'pastures', 'pasturing', 'pastured', 'pastured', _, ['2A','6A'] ).
verb( 'pat', 'pats', 'patting', 'patted', 'patted', _, ['2A','6A','15A'] ).
verb( 'patch', 'patches', 'patching', 'patched', 'patched', tran, ['6A','15B'] ).
verb( 'patent', 'patents', 'patenting', 'patented', 'patented', tran, ['6A'] ).
verb( 'patrol', 'patrols', 'patrolling', 'patrolled', 'patrolled', _, ['2A','6A'] ).
verb( 'patronize', 'patronizes', 'patronizing', 'patronized', 'patronized', tran, ['6A'] ).
verb( 'patter', 'patters', 'pattering', 'pattered', 'pattered', _, ['2A','2C','6A'] ).
verb( 'pattern', 'patterns', 'patterning', 'patterned', 'patterned', tran, ['14'] ).
verb( 'pauperize', 'pauperizes', 'pauperizing', 'pauperized', 'pauperized', tran, ['6A'] ).
verb( 'pause', 'pauses', 'pausing', 'paused', 'paused', intran, ['2A','4A'] ).
verb( 'pave', 'paves', 'paving', 'paved', 'paved', tran, ['6A'] ).
verb( 'paw', 'paws', 'pawing', 'pawed', 'pawed', tran, ['6A','15B'] ).
verb( 'pawn', 'pawns', 'pawning', 'pawned', 'pawned', tran, ['6A'] ).
verb( 'pay', 'pays', 'paying', 'paid', 'paid', _, ['2A','3A','6A','12A','12B','13A','13B','14','15B'] ).
verb( 'peach', 'peaches', 'peaching', 'peached', 'peached', _, ['2A','3A','6A'] ).
verb( 'peak', 'peaks', 'peaking', 'peaked', 'peaked', intran, ['2A'] ).
verb( 'peal', 'peals', 'pealing', 'pealed', 'pealed', _, ['2A','2C','6A'] ).
verb( 'pearl', 'pearls', 'pearling', 'pearled', 'pearled', intran, [] ).
verb( 'peck', 'pecks', 'pecking', 'pecked', 'pecked', _, ['2A','2C','3A','6A'] ).
verb( 'peculate', 'peculates', 'peculating', 'peculated', 'peculated', _, [] ).
verb( 'pedal', 'pedals', 'pedalling', 'pedalled', 'pedalled', _, ['2A','2C','6A'] ).
verb( 'peddle', 'peddles', 'peddling', 'peddled', 'peddled', _, ['2A','6A'] ).
verb( 'pee', 'pees', 'peeing', 'peed', 'peed', intran, ['2A'] ).
verb( 'peek', 'peeks', 'peeking', 'peeked', 'peeked', intran, [] ).
verb( 'peel', 'peels', 'peeling', 'peeled', 'peeled', _, ['2A','2C','6A','15B'] ).
verb( 'peep', 'peeps', 'peeping', 'peeped', 'peeped', intran, ['2A','2C'] ).
verb( 'peer', 'peers', 'peering', 'peered', 'peered', intran, ['2A','3A'] ).
verb( 'peeve', 'peeves', 'peeving', 'peeved', 'peeved', tran, [] ).
verb( 'peg', 'pegs', 'pegging', 'pegged', 'pegged', _, ['2C','6A','15B'] ).
verb( 'pelt', 'pelts', 'pelting', 'pelted', 'pelted', _, ['2C','6A','14'] ).
verb( 'pen', 'pens', 'penning', 'penned', 'penned', tran, ['15A','15B'] ).
verb( 'penalize', 'penalizes', 'penalizing', 'penalized', 'penalized', tran, ['6A','14'] ).
verb( 'pencil', 'pencils', 'pencilling', 'pencilled', 'pencilled', tran, ['6A'] ).
verb( 'penetrate', 'penetrates', 'penetrating', 'penetrated', 'penetrated', _, ['3A','6A'] ).
verb( 'pension', 'pensions', 'pensioning', 'pensioned', 'pensioned', tran, ['15B'] ).
verb( 'people', 'peoples', 'peopling', 'peopled', 'peopled', tran, ['6A'] ).
verb( 'pep', 'peps', 'pepping', 'pepped', 'pepped', tran, ['15B'] ).
verb( 'pepper', 'peppers', 'peppering', 'peppered', 'peppered', tran, ['6A'] ).
verb( 'perambulate', 'perambulates', 'perambulating', 'perambulated', 'perambulated', _, ['2A','6A'] ).
verb( 'perceive', 'perceives', 'perceiving', 'perceived', 'perceived', tran, ['6A','8','9','10','18A','19A','25'] ).
verb( 'perch', 'perches', 'perching', 'perched', 'perched', _, ['2C'] ).
verb( 'percolate', 'percolates', 'percolating', 'percolated', 'percolated', _, ['2A','3A','6A'] ).
verb( 'perfect', 'perfects', 'perfecting', 'perfected', 'perfected', tran, ['6A'] ).
verb( 'perforate', 'perforates', 'perforating', 'perforated', 'perforated', tran, ['6A'] ).
verb( 'perform', 'performs', 'performing', 'performed', 'performed', _, ['2A','6A'] ).
verb( 'perfume', 'perfumes', 'perfuming', 'perfumed', 'perfumed', tran, ['6A'] ).
verb( 'peril', 'perils', 'perilling', 'perilled', 'perilled', tran, [] ).
verb( 'perish', 'perishes', 'perishing', 'perished', 'perished', _, ['2A','2C','6A'] ).
verb( 'perjure', 'perjures', 'perjuring', 'perjured', 'perjured', tran, ['6A'] ).
verb( 'perk', 'perks', 'perking', 'perked', 'perked', _, ['2C','15B'] ).
verb( 'perm', 'perms', 'perming', 'permed', 'permed', tran, [] ).
verb( 'permeate', 'permeates', 'permeating', 'permeated', 'permeated', _, ['3A','6A'] ).
verb( 'permit', 'permits', 'permitting', 'permitted', 'permitted', _, ['3A','6A','6C','17','19C'] ).
verb( 'permute', 'permutes', 'permuting', 'permuted', 'permuted', tran, ['6A'] ).
verb( 'perpetrate', 'perpetrates', 'perpetrating', 'perpetrated', 'perpetrated', tran, ['6A'] ).
verb( 'perpetuate', 'perpetuates', 'perpetuating', 'perpetuated', 'perpetuated', tran, ['6A'] ).
verb( 'perplex', 'perplexes', 'perplexing', 'perplexed', 'perplexed', tran, ['6A','14'] ).
verb( 'persecute', 'persecutes', 'persecuting', 'persecuted', 'persecuted', tran, ['6A'] ).
verb( 'persevere', 'perseveres', 'persevering', 'persevered', 'persevered', intran, ['2A','3A'] ).
verb( 'persist', 'persists', 'persisting', 'persisted', 'persisted', intran, ['2A','3A'] ).
verb( 'personalize', 'personalizes', 'personalizing', 'personalized', 'personalized', tran, ['6A'] ).
verb( 'personate', 'personates', 'personating', 'personated', 'personated', tran, ['6A'] ).
verb( 'personify', 'personifies', 'personifying', 'personified', 'personified', tran, ['6A'] ).
verb( 'perspire', 'perspires', 'perspiring', 'perspired', 'perspired', intran, ['2A'] ).
verb( 'persuade', 'persuades', 'persuading', 'persuaded', 'persuaded', tran, ['11','14','17'] ).
verb( 'pertain', 'pertains', 'pertaining', 'pertained', 'pertained', intran, ['3A'] ).
verb( 'perturb', 'perturbs', 'perturbing', 'perturbed', 'perturbed', tran, ['6A'] ).
verb( 'peruse', 'peruses', 'perusing', 'perused', 'perused', tran, ['6A'] ).
verb( 'pervade', 'pervades', 'pervading', 'pervaded', 'pervaded', tran, ['6A'] ).
verb( 'pervert', 'perverts', 'perverting', 'perverted', 'perverted', tran, ['6A'] ).
verb( 'pester', 'pesters', 'pestering', 'pestered', 'pestered', tran, ['6A','14','17'] ).
verb( 'pestle', 'pestles', 'pestling', 'pestled', 'pestled', tran, [] ).
verb( 'pet', 'pets', 'petting', 'petted', 'petted', tran, [] ).
verb( 'peter', 'peters', 'petering', 'petered', 'petered', intran, ['2C'] ).
verb( 'petition', 'petitions', 'petitioning', 'petitioned', 'petitioned', _, ['3A','6A','11','14','17'] ).
verb( 'petrify', 'petrifies', 'petrifying', 'petrified', 'petrified', _, ['2A','6A'] ).
verb( 'phase', 'phases', 'phasing', 'phased', 'phased', tran, ['6A','15B'] ).
verb( 'philander', 'philanders', 'philandering', 'philandered', 'philandered', intran, ['2A'] ).
verb( 'philosophize', 'philosophizes', 'philosophizing', 'philosophized', 'philosophized', intran, ['2A'] ).
verb( 'phone', 'phones', 'phoning', 'phoned', 'phoned', _, ['2A','4A','6A','11','12A','13A'] ).
verb( 'photocopy', 'photocopies', 'photocopying', 'photocopied', 'photocopied', tran, ['6A'] ).
verb( 'photograph', 'photographs', 'photographing', 'photographed', 'photographed', tran, ['6A'] ).
verb( 'photosensitize', 'photosensitizes', 'photosensitizing', 'photosensitized', 'photosensitized', tran, [] ).
verb( 'photostat', 'photostats', 'photostatting', 'photostatted', 'photostatted', tran, [] ).
verb( 'phrase', 'phrases', 'phrasing', 'phrased', 'phrased', tran, ['6A'] ).
verb( 'pick', 'picks', 'picking', 'picked', 'picked', _, ['3A','6A','15B'] ).
verb( 'picket', 'pickets', 'picketing', 'picketed', 'picketed', _, ['2A','6A'] ).
verb( 'pickle', 'pickles', 'pickling', 'pickled', 'pickled', tran, ['6A'] ).
verb( 'picnic', 'picnics', 'picnicking', 'picnicked', 'picnicked', intran, [] ).
verb( 'picture', 'pictures', 'picturing', 'pictured', 'pictured', tran, ['6A','14'] ).
verb( 'piddle', 'piddles', 'piddling', 'piddled', 'piddled', intran, [] ).
verb( 'piece', 'pieces', 'piecing', 'pieced', 'pieced', tran, ['6A','15A','15B'] ).
verb( 'pierce', 'pierces', 'piercing', 'pierced', 'pierced', _, ['2C','6A'] ).
verb( 'piffle', 'piffles', 'piffling', 'piffled', 'piffled', intran, [] ).
verb( 'pig', 'pigs', 'pigging', 'pigged', 'pigged', intran, [] ).
verb( 'pigeonhole', 'pigeonholes', 'pigeonholing', 'pigeonholed', 'pigeonholed', tran, [] ).
verb( 'pile', 'piles', 'piling', 'piled', 'piled', _, ['2C','6A','15A','15B'] ).
verb( 'pilfer', 'pilfers', 'pilfering', 'pilfered', 'pilfered', _, ['2A','6A'] ).
verb( 'pillage', 'pillages', 'pillaging', 'pillaged', 'pillaged', tran, [] ).
verb( 'pillow', 'pillows', 'pillowing', 'pillowed', 'pillowed', tran, ['6A'] ).
verb( 'pilot', 'pilots', 'piloting', 'piloted', 'piloted', tran, ['6A','15A'] ).
verb( 'pimp', 'pimps', 'pimping', 'pimped', 'pimped', intran, ['2A','3A'] ).
verb( 'pin', 'pins', 'pinning', 'pinned', 'pinned', tran, ['15A','15B'] ).
verb( 'pinch', 'pinches', 'pinching', 'pinched', 'pinched', _, ['2A','6A','15A','15B'] ).
verb( 'pine', 'pines', 'pining', 'pined', 'pined', intran, ['2A','2C','3A','4C'] ).
verb( 'ping', 'pings', 'pinging', 'pinged', 'pinged', intran, [] ).
verb( 'pinion', 'pinions', 'pinioning', 'pinioned', 'pinioned', tran, ['6A','15A','15B'] ).
verb( 'pink', 'pinks', 'pinking', 'pinked', 'pinked', _, ['6A','15B'] ).
verb( 'pinnacle', 'pinnacles', 'pinnacling', 'pinnacled', 'pinnacled', tran, [] ).
verb( 'pinpoint', 'pinpoints', 'pinpointing', 'pinpointed', 'pinpointed', tran, [] ).
verb( 'pioneer', 'pioneers', 'pioneering', 'pioneered', 'pioneered', _, ['2A','6A'] ).
verb( 'pip', 'pips', 'pipping', 'pipped', 'pipped', tran, ['6A'] ).
verb( 'pipe', 'pipes', 'piping', 'piped', 'piped', _, ['2A','2C','6A','15A'] ).
verb( 'pique', 'piques', 'piquing', 'piqued', 'piqued', tran, ['6A'] ).
verb( 'pirate', 'pirates', 'pirating', 'pirated', 'pirated', tran, ['6A'] ).
verb( 'pirouette', 'pirouettes', 'pirouetting', 'pirouetted', 'pirouetted', intran, [] ).
verb( 'piss', 'pisses', 'pissing', 'pissed', 'pissed', _, [] ).
verb( 'pit', 'pits', 'pitting', 'pitted', 'pitted', tran, ['6A','14'] ).
verb( 'pitch', 'pitches', 'pitching', 'pitched', 'pitched', _, ['2A','2C','6A','15A','15B','22'] ).
verb( 'pitchfork', 'pitchforks', 'pitchforking', 'pitchforked', 'pitchforked', tran, [] ).
verb( 'pity', 'pities', 'pitying', 'pitied', 'pitied', tran, ['6A'] ).
verb( 'pivot', 'pivots', 'pivoting', 'pivoted', 'pivoted', _, ['3A','6A'] ).
verb( 'placard', 'placards', 'placarding', 'placarded', 'placarded', tran, ['6A'] ).
verb( 'placate', 'placates', 'placating', 'placated', 'placated', tran, ['6A'] ).
verb( 'place', 'places', 'placing', 'placed', 'placed', tran, ['6A','15A','15B'] ).
verb( 'plagiarize', 'plagiarizes', 'plagiarizing', 'plagiarized', 'plagiarized', tran, ['6A'] ).
verb( 'plague', 'plagues', 'plaguing', 'plagued', 'plagued', tran, ['6A','14'] ).
verb( 'plain', 'plains', 'plaining', 'plained', 'plained', _, [] ).
verb( 'plait', 'plaits', 'plaiting', 'plaited', 'plaited', tran, ['6A'] ).
verb( 'plan', 'plans', 'planning', 'planned', 'planned', tran, ['6A','7A','15B'] ).
verb( 'plane', 'planes', 'planing', 'planed', 'planed', _, ['2A','2D','15B','22'] ).
verb( 'plank', 'planks', 'planking', 'planked', 'planked', tran, ['6A','15B'] ).
verb( 'plant', 'plants', 'planting', 'planted', 'planted', tran, ['6A','15A','15B'] ).
verb( 'plash', 'plashes', 'plashing', 'plashed', 'plashed', _, ['6A'] ).
verb( 'plaster', 'plasters', 'plastering', 'plastered', 'plastered', tran, ['6A','14'] ).
verb( 'plate', 'plates', 'plating', 'plated', 'plated', tran, ['6A','14'] ).
verb( 'play', 'plays', 'playing', 'played', 'played', _, ['2A','2B','2C','3A','6A','12B','12C','13B','14','15A','15B','16B'] ).
verb( 'pleach', 'pleaches', 'pleaching', 'pleached', 'pleached', tran, ['6A'] ).
verb( 'plead', 'pleads', 'pleading', 'pleaded', 'pleaded', _, ['3A','6A'] ).
verb( 'please', 'pleases', 'pleasing', 'pleased', 'pleased', _, ['2A','6A'] ).
verb( 'pleat', 'pleats', 'pleating', 'pleated', 'pleated', tran, ['6A'] ).
verb( 'pledge', 'pledges', 'pledging', 'pledged', 'pledged', tran, ['6A'] ).
verb( 'plight', 'plights', 'plighting', 'plighted', 'plighted', tran, ['6A'] ).
verb( 'plod', 'plods', 'plodding', 'plodded', 'plodded', _, ['2C'] ).
verb( 'plonk', 'plonks', 'plonking', 'plonked', 'plonked', intran, [] ).
verb( 'plop', 'plops', 'plopping', 'plopped', 'plopped', intran, [] ).
verb( 'plot', 'plots', 'plotting', 'plotted', 'plotted', _, ['2A','3A','4A','6A','8','10','15B'] ).
verb( 'plough', 'ploughs', 'ploughing', 'ploughed', 'ploughed', _, ['3A','6A','15B'] ).
verb( 'pluck', 'plucks', 'plucking', 'plucked', 'plucked', _, ['3A','6A','15B'] ).
verb( 'plug', 'plugs', 'plugging', 'plugged', 'plugged', _, ['2C','6A','15B'] ).
verb( 'plumb', 'plumbs', 'plumbing', 'plumbed', 'plumbed', tran, ['6A'] ).
verb( 'plume', 'plumes', 'pluming', 'plumed', 'plumed', tran, ['6A'] ).
verb( 'plummet', 'plummets', 'plummetting', 'plummetted', 'plummetted', intran, ['2A'] ).
verb( 'plump', 'plumps', 'plumping', 'plumped', 'plumped', _, ['2A','2C','3A','6A','15B'] ).
verb( 'plunder', 'plunders', 'plundering', 'plundered', 'plundered', _, ['2A','6A','14'] ).
verb( 'plunge', 'plunges', 'plunging', 'plunged', 'plunged', _, ['2A','2C','6A','14'] ).
verb( 'ply', 'plies', 'plying', 'plied', 'plied', _, ['2C','6A'] ).
verb( 'poach', 'poaches', 'poaching', 'poached', 'poached', _, ['2A','3A','6A'] ).
verb( 'pocket', 'pockets', 'pocketing', 'pocketed', 'pocketed', tran, ['6A'] ).
verb( 'pod', 'pods', 'podding', 'podded', 'podded', _, ['2A','2C','6A'] ).
verb( 'point', 'points', 'pointing', 'pointed', 'pointed', _, ['2A','3A','6A','14','15B'] ).
verb( 'poise', 'poises', 'poising', 'poised', 'poised', _, ['2C','6A','15A'] ).
verb( 'poison', 'poisons', 'poisoning', 'poisoned', 'poisoned', tran, ['6A'] ).
verb( 'poke', 'pokes', 'poking', 'poked', 'poked', _, ['2C','6A','15A','15B'] ).
verb( 'polarize', 'polarizes', 'polarizing', 'polarized', 'polarized', tran, ['6A'] ).
verb( 'poleax', 'poleaxes', 'poleaxing', 'poleaxed', 'poleaxed', tran, ['6A'] ).
verb( 'poleaxe', 'poleaxes', 'poleaxing', 'poleaxed', 'poleaxed', tran, ['6A'] ).
verb( 'police', 'polices', 'policing', 'policed', 'policed', tran, ['6A'] ).
verb( 'polish', 'polishes', 'polishing', 'polished', 'polished', _, ['2A','6A','15B'] ).
verb( 'politicize', 'politicizes', 'politicizing', 'politicized', 'politicized', _, ['2A','6A'] ).
verb( 'politick', 'politicks', 'politicking', 'politicked', 'politicked', intran, ['2A'] ).
verb( 'poll', 'polls', 'polling', 'polled', 'polled', _, ['2A','2C','6A'] ).
verb( 'pollard', 'pollards', 'pollarding', 'pollarded', 'pollarded', tran, ['6A'] ).
verb( 'pollinate', 'pollinates', 'pollinating', 'pollinated', 'pollinated', tran, ['6A'] ).
verb( 'pollute', 'pollutes', 'polluting', 'polluted', 'polluted', tran, ['6A'] ).
verb( 'pomade', 'pomades', 'pomading', 'pomaded', 'pomaded', tran, [] ).
verb( 'pommel', 'pommels', 'pommelling', 'pommelled', 'pommelled', tran, [] ).
verb( 'ponder', 'ponders', 'pondering', 'pondered', 'pondered', _, ['2A','3A','6A','8','10'] ).
verb( 'poniard', 'poniards', 'poniarding', 'poniarded', 'poniarded', tran, ['6A'] ).
verb( 'pontificate', 'pontificates', 'pontificating', 'pontificated', 'pontificated', intran, ['2A'] ).
verb( 'pooh-pooh', 'pooh-poohs', 'pooh-poohing', 'pooh-poohed', 'pooh-poohed', tran, ['6A'] ).
verb( 'pool', 'pools', 'pooling', 'pooled', 'pooled', tran, ['6A'] ).
verb( 'pop', 'pops', 'popping', 'popped', 'popped', _, ['2A','2C','6A','15A','15B'] ).
verb( 'popularize', 'popularizes', 'popularizing', 'popularized', 'popularized', tran, ['6A'] ).
verb( 'populate', 'populates', 'populating', 'populated', 'populated', tran, ['6A'] ).
verb( 'pore', 'pores', 'poring', 'pored', 'pored', intran, ['3A'] ).
verb( 'port', 'ports', 'porting', 'ported', 'ported', tran, ['6A'] ).
verb( 'portend', 'portends', 'portending', 'portended', 'portended', tran, ['6A'] ).
verb( 'portion', 'portions', 'portioning', 'portioned', 'portioned', tran, ['14','15B'] ).
verb( 'portray', 'portrays', 'portraying', 'portrayed', 'portrayed', tran, ['6A'] ).
verb( 'pose', 'poses', 'posing', 'posed', 'posed', _, ['2A','2C','3A','6A'] ).
verb( 'posh', 'poshes', 'poshing', 'poshed', 'poshed', tran, ['15B'] ).
verb( 'posit', 'posits', 'positing', 'posited', 'posited', tran, ['6A'] ).
verb( 'position', 'positions', 'positioning', 'positioned', 'positioned', tran, ['6A'] ).
verb( 'possess', 'possesses', 'possessing', 'possessed', 'possessed', tran, ['6A'] ).
verb( 'post', 'posts', 'posting', 'posted', 'posted', _, ['3A','6A','15A','15B'] ).
verb( 'postdate', 'postdates', 'postdating', 'postdated', 'postdated', tran, ['6A'] ).
verb( 'postmark', 'postmarks', 'postmarking', 'postmarked', 'postmarked', tran, [] ).
verb( 'postpone', 'postpones', 'postponing', 'postponed', 'postponed', tran, ['6A','6C'] ).
verb( 'postulate', 'postulates', 'postulating', 'postulated', 'postulated', tran, ['6A'] ).
verb( 'posture', 'postures', 'posturing', 'postured', 'postured', _, ['2A','6A'] ).
verb( 'pot', 'pots', 'potting', 'potted', 'potted', _, ['3A','6A','15B'] ).
verb( 'potter', 'potters', 'pottering', 'pottered', 'pottered', intran, ['2A','2C'] ).
verb( 'pouch', 'pouches', 'pouching', 'pouched', 'pouched', tran, ['2A','6A'] ).
verb( 'poultice', 'poultices', 'poulticing', 'poulticed', 'poulticed', tran, ['6A'] ).
verb( 'pounce', 'pounces', 'pouncing', 'pounced', 'pounced', intran, ['3A'] ).
verb( 'pound', 'pounds', 'pounding', 'pounded', 'pounded', _, ['2C','3A','6A','15A'] ).
verb( 'pour', 'pours', 'pouring', 'poured', 'poured', _, ['2C','6A','12B','13B','14','15A','15B'] ).
verb( 'pout', 'pouts', 'pouting', 'pouted', 'pouted', _, ['2A','6A'] ).
verb( 'powder', 'powders', 'powdering', 'powdered', 'powdered', _, ['2A','6A'] ).
verb( 'power', 'powers', 'powering', 'powered', 'powered', tran, ['6A'] ).
verb( 'power-dive', 'power-dives', 'power-diving', 'power-dived', 'power-dived', tran, [] ).
verb( 'powwow', 'powwows', 'powwowing', 'powwowed', 'powwowed', intran, [] ).
verb( 'pr_ecis', 'pr_ecises', 'pr_ecising', 'pr_ecised', 'pr_ecised', tran, [] ).
verb( 'practise', 'practises', 'practising', 'practised', 'practised', _, ['2A','2B','3A','4A','6A','6C'] ).
verb( 'praise', 'praises', 'praising', 'praised', 'praised', tran, ['6A'] ).
verb( 'prance', 'prances', 'prancing', 'pranced', 'pranced', intran, ['2A','2C'] ).
verb( 'prate', 'prates', 'prating', 'prated', 'prated', intran, ['2A','2C'] ).
verb( 'prattle', 'prattles', 'prattling', 'prattled', 'prattled', intran, ['2A','2C'] ).
verb( 'prawn', 'prawns', 'prawning', 'prawned', 'prawned', intran, [] ).
verb( 'pray', 'prays', 'praying', 'prayed', 'prayed', _, ['2A','3A','11','14','17'] ).
verb( 'pre-empt', 'pre-empts', 'pre-empting', 'pre-empted', 'pre-empted', tran, ['6A'] ).
verb( 'pre-exist', 'pre-exists', 'pre-existing', 'pre-existed', 'pre-existed', intran, ['2A'] ).
verb( 'preach', 'preaches', 'preaching', 'preached', 'preached', _, ['2A','2B','2C','3A','6A','12A','13A'] ).
verb( 'preachify', 'preachifies', 'preachifying', 'preachified', 'preachified', intran, [] ).
verb( 'prearrange', 'prearranges', 'prearranging', 'prearranged', 'prearranged', tran, [] ).
verb( 'precede', 'precedes', 'preceding', 'preceded', 'preceded', _, ['2A','6A'] ).
verb( 'precipitate', 'precipitates', 'precipitating', 'precipitated', 'precipitated', tran, ['6A','14'] ).
verb( 'preclude', 'precludes', 'precluding', 'precluded', 'precluded', tran, ['6A','6C','14'] ).
verb( 'preconceive', 'preconceives', 'preconceiving', 'preconceived', 'preconceived', tran, ['6A'] ).
verb( 'predecease', 'predeceases', 'predeceasing', 'predeceased', 'predeceased', tran, [] ).
verb( 'predestinate', 'predestinates', 'predestinating', 'predestinated', 'predestinated', tran, [] ).
verb( 'predestine', 'predestines', 'predestining', 'predestined', 'predestined', tran, ['14','17'] ).
verb( 'predetermine', 'predetermines', 'predetermining', 'predetermined', 'predetermined', tran, ['6A','17'] ).
verb( 'predicate', 'predicates', 'predicating', 'predicated', 'predicated', tran, ['6A','9','17'] ).
verb( 'predict', 'predicts', 'predicting', 'predicted', 'predicted', tran, ['6A','9','10'] ).
verb( 'predigest', 'predigests', 'predigesting', 'predigested', 'predigested', tran, [] ).
verb( 'predispose', 'predisposes', 'predisposing', 'predisposed', 'predisposed', tran, ['14','17'] ).
verb( 'predominate', 'predominates', 'predominating', 'predominated', 'predominated', intran, ['2A','3A'] ).
verb( 'preen', 'preens', 'preening', 'preened', 'preened', tran, ['6A','14'] ).
verb( 'prefabricate', 'prefabricates', 'prefabricating', 'prefabricated', 'prefabricated', tran, ['6A'] ).
verb( 'preface', 'prefaces', 'prefacing', 'prefaced', 'prefaced', tran, ['14'] ).
verb( 'prefer', 'prefers', 'preferring', 'preferred', 'preferred', tran, ['6A','6D','7A','9','14','17'] ).
verb( 'prefigure', 'prefigures', 'prefiguring', 'prefigured', 'prefigured', tran, ['6A','9','10'] ).
verb( 'prefix', 'prefixes', 'prefixing', 'prefixed', 'prefixed', tran, ['6A','14'] ).
verb( 'preheat', 'preheats', 'preheating', 'preheated', 'preheated', tran, ['6A'] ).
verb( 'prejudge', 'prejudges', 'prejudging', 'prejudged', 'prejudged', tran, ['6A'] ).
verb( 'prejudice', 'prejudices', 'prejudicing', 'prejudiced', 'prejudiced', tran, ['6A','15A'] ).
verb( 'prelude', 'preludes', 'preluding', 'preluded', 'preluded', tran, ['6A'] ).
verb( 'premeditate', 'premeditates', 'premeditating', 'premeditated', 'premeditated', tran, ['6A'] ).
verb( 'premise', 'premises', 'premising', 'premised', 'premised', tran, ['6A','9'] ).
verb( 'premiss', 'premisses', 'premissing', 'premissed', 'premissed', tran, ['6A','9'] ).
verb( 'preoccupy', 'preoccupies', 'preoccupying', 'preoccupied', 'preoccupied', tran, ['6A'] ).
verb( 'preordain', 'preordains', 'preordaining', 'preordained', 'preordained', tran, ['6A','9'] ).
verb( 'prepare', 'prepares', 'preparing', 'prepared', 'prepared', _, ['3A','6A','7A','14'] ).
verb( 'prepay', 'prepays', 'prepaying', 'prepaid', 'prepaid', tran, ['6A'] ).
verb( 'preponderate', 'preponderates', 'preponderating', 'preponderated', 'preponderated', intran, ['2A','2C'] ).
verb( 'prepossess', 'prepossesses', 'prepossessing', 'prepossessed', 'prepossessed', tran, ['6A','15A'] ).
verb( 'prerecord', 'prerecords', 'prerecording', 'prerecorded', 'prerecorded', tran, ['6A'] ).
verb( 'presage', 'presages', 'presaging', 'presaged', 'presaged', tran, ['6A'] ).
verb( 'prescribe', 'prescribes', 'prescribing', 'prescribed', 'prescribed', _, ['2A','3A','6A','8','10','14','21'] ).
verb( 'present', 'presents', 'presenting', 'presented', 'presented', tran, ['6A','14','15A'] ).
verb( 'preserve', 'preserves', 'preserving', 'preserved', 'preserved', tran, ['6A','14'] ).
verb( 'preside', 'presides', 'presiding', 'presided', 'presided', intran, ['2A','2C','3A'] ).
verb( 'press', 'presses', 'pressing', 'pressed', 'pressed', _, ['2A','2C','3A','4A','6A','14','15A','15B','17','22'] ).
verb( 'presume', 'presumes', 'presuming', 'presumed', 'presumed', _, ['3A','6A','7A','9','25'] ).
verb( 'presuppose', 'presupposes', 'presupposing', 'presupposed', 'presupposed', tran, ['6A','9'] ).
verb( 'pretend', 'pretends', 'pretending', 'pretended', 'pretended', _, ['3A','6A','7A','9'] ).
verb( 'prettify', 'prettifies', 'prettifying', 'prettified', 'prettified', tran, ['6A'] ).
verb( 'prevail', 'prevails', 'prevailing', 'prevailed', 'prevailed', intran, ['2A','3A'] ).
verb( 'prevaricate', 'prevaricates', 'prevaricating', 'prevaricated', 'prevaricated', intran, ['2A'] ).
verb( 'prevent', 'prevents', 'preventing', 'prevented', 'prevented', tran, ['6A','14','19C'] ).
verb( 'preview', 'previews', 'previewing', 'previewed', 'previewed', tran, [] ).
verb( 'prey', 'preys', 'preying', 'preyed', 'preyed', intran, ['3A'] ).
verb( 'price', 'prices', 'pricing', 'priced', 'priced', tran, ['6A'] ).
verb( 'prick', 'pricks', 'pricking', 'pricked', 'pricked', _, ['2A','6A','15B'] ).
verb( 'prickle', 'prickles', 'prickling', 'prickled', 'prickled', _, [] ).
verb( 'pride', 'prides', 'priding', 'prided', 'prided', tran, [] ).
verb( 'prim', 'prims', 'primming', 'primmed', 'primmed', tran, [] ).
verb( 'prime', 'primes', 'priming', 'primed', 'primed', tran, ['6A'] ).
verb( 'primp', 'primps', 'primping', 'primped', 'primped', tran, [] ).
verb( 'prink', 'prinks', 'prinking', 'prinked', 'prinked', tran, [] ).
verb( 'print', 'prints', 'printing', 'printed', 'printed', _, ['2A','6A','15B'] ).
verb( 'prise', 'prises', 'prising', 'prised', 'prised', tran, ['15A','15B'] ).
verb( 'prize', 'prizes', 'prizing', 'prized', 'prized', tran, ['15A','15B'] ).
verb( 'probate', 'probates', 'probating', 'probated', 'probated', tran, [] ).
verb( 'probe', 'probes', 'probing', 'probed', 'probed', tran, ['6A'] ).
verb( 'proceed', 'proceeds', 'proceeding', 'proceeded', 'proceeded', intran, ['2A','3A','4C'] ).
verb( 'process', 'processes', 'processing', 'processed', 'processed', tran, ['6A'] ).
verb( 'process', 'processes', 'processing', 'processed', 'processed', intran, [] ).
verb( 'proclaim', 'proclaims', 'proclaiming', 'proclaimed', 'proclaimed', tran, ['6A','9','23','25'] ).
verb( 'procrastinate', 'procrastinates', 'procrastinating', 'procrastinated', 'procrastinated', intran, ['2A'] ).
verb( 'procreate', 'procreates', 'procreating', 'procreated', 'procreated', tran, ['6A'] ).
verb( 'procure', 'procures', 'procuring', 'procured', 'procured', tran, ['6A','12B','13B'] ).
verb( 'prod', 'prods', 'prodding', 'prodded', 'prodded', _, ['3A','6A'] ).
verb( 'produce', 'produces', 'producing', 'produced', 'produced', _, ['2A','6A'] ).
verb( 'profane', 'profanes', 'profaning', 'profaned', 'profaned', tran, ['6A'] ).
verb( 'profess', 'professes', 'professing', 'professed', 'professed', _, ['6A','7A','9','25'] ).
verb( 'proffer', 'proffers', 'proffering', 'proffered', 'proffered', tran, ['6A','7A'] ).
verb( 'profile', 'profiles', 'profiling', 'profiled', 'profiled', tran, [] ).
verb( 'profit', 'profits', 'profiting', 'profited', 'profited', _, ['3A','6A','13A'] ).
verb( 'profiteer', 'profiteers', 'profiteering', 'profiteered', 'profiteered', intran, ['2A'] ).
verb( 'prognosticate', 'prognosticates', 'prognosticating', 'prognosticated', 'prognosticated', tran, ['6A','9'] ).
verb( 'program', 'programs', 'programming', 'programmed', 'programmed', tran, ['6A'] ).
verb( 'programme', 'programmes', 'programming', 'programmed', 'programmed', tran, ['6A'] ).
verb( 'progress', 'progresses', 'progressing', 'progressed', 'progressed', intran, ['2A','2C'] ).
verb( 'prohibit', 'prohibits', 'prohibiting', 'prohibited', 'prohibited', tran, ['6A','14'] ).
verb( 'project', 'projects', 'projecting', 'projected', 'projected', _, ['2A','2C','6A','14','15A'] ).
verb( 'prolapse', 'prolapses', 'prolapsing', 'prolapsed', 'prolapsed', intran, [] ).
verb( 'proliferate', 'proliferates', 'proliferating', 'proliferated', 'proliferated', _, ['2A','6A'] ).
verb( 'prolong', 'prolongs', 'prolonging', 'prolonged', 'prolonged', tran, ['6A'] ).
verb( 'promenade', 'promenades', 'promenading', 'promenaded', 'promenaded', _, ['2A','2C','6A','15A'] ).
verb( 'promise', 'promises', 'promising', 'promised', 'promised', _, ['2A','6A','7A','9','11','12A','13A','17'] ).
verb( 'promote', 'promotes', 'promoting', 'promoted', 'promoted', tran, ['6A','14'] ).
verb( 'prompt', 'prompts', 'prompting', 'prompted', 'prompted', tran, ['6A','17'] ).
verb( 'promulgate', 'promulgates', 'promulgating', 'promulgated', 'promulgated', tran, ['6A'] ).
verb( 'pronounce', 'pronounces', 'pronouncing', 'pronounced', 'pronounced', _, ['2A','3A','6A','9','22','25'] ).
verb( 'proof', 'proofs', 'proofing', 'proofed', 'proofed', tran, ['6A'] ).
verb( 'proofread', 'proofreads', 'proofreading', 'proofread', 'proofread', _, ['2A','6A'] ).
verb( 'prop', 'props', 'propping', 'propped', 'propped', tran, ['6A','15A','15B','22'] ).
verb( 'propagandize', 'propagandizes', 'propagandizing', 'propagandized', 'propagandized', intran, [] ).
verb( 'propagate', 'propagates', 'propagating', 'propagated', 'propagated', _, ['2A','6A'] ).
verb( 'propel', 'propels', 'propelling', 'propelled', 'propelled', tran, ['6A','15A'] ).
verb( 'prophesy', 'prophesies', 'prophesying', 'prophesied', 'prophesied', _, ['2A','2C','6A','9','10'] ).
verb( 'propitiate', 'propitiates', 'propitiating', 'propitiated', 'propitiated', tran, ['6A'] ).
verb( 'proportion', 'proportions', 'proportioning', 'proportioned', 'proportioned', tran, ['6A','14'] ).
verb( 'propose', 'proposes', 'proposing', 'proposed', 'proposed', _, ['2A','6A','6D','7A','9','14'] ).
verb( 'proposition', 'propositions', 'propositioning', 'propositioned', 'propositioned', tran, ['6A'] ).
verb( 'propound', 'propounds', 'propounding', 'propounded', 'propounded', tran, ['6A'] ).
verb( 'prorogue', 'prorogues', 'proroguing', 'prorogued', 'prorogued', tran, ['6A'] ).
verb( 'proscribe', 'proscribes', 'proscribing', 'proscribed', 'proscribed', tran, ['6A'] ).
verb( 'prosecute', 'prosecutes', 'prosecuting', 'prosecuted', 'prosecuted', tran, ['6A','14'] ).
verb( 'proselytize', 'proselytizes', 'proselytizing', 'proselytized', 'proselytized', _, ['2A','6A'] ).
verb( 'prospect', 'prospects', 'prospecting', 'prospected', 'prospected', intran, ['2A','3A'] ).
verb( 'prosper', 'prospers', 'prospering', 'prospered', 'prospered', _, ['2A','6A'] ).
verb( 'prostitute', 'prostitutes', 'prostituting', 'prostituted', 'prostituted', tran, ['6A'] ).
verb( 'prostrate', 'prostrates', 'prostrating', 'prostrated', 'prostrated', tran, ['6A'] ).
verb( 'protect', 'protects', 'protecting', 'protected', 'protected', tran, ['6A','14'] ).
verb( 'protest', 'protests', 'protesting', 'protested', 'protested', _, ['2A','3A','6A','9'] ).
verb( 'protract', 'protracts', 'protracting', 'protracted', 'protracted', tran, ['6A'] ).
verb( 'protrude', 'protrudes', 'protruding', 'protruded', 'protruded', _, ['2A','6A'] ).
verb( 'prove', 'proves', 'proving', 'proved', 'proved', _, ['4D','6A','9','14','25'] ).
verb( 'provide', 'provides', 'providing', 'provided', 'provided', _, ['3A','6A','9','14'] ).
verb( 'provision', 'provisions', 'provisioning', 'provisioned', 'provisioned', tran, ['6A'] ).
verb( 'provoke', 'provokes', 'provoking', 'provoked', 'provoked', tran, ['6A','14','17'] ).
verb( 'prowl', 'prowls', 'prowling', 'prowled', 'prowled', _, ['2A','2C','6A'] ).
verb( 'prune', 'prunes', 'pruning', 'pruned', 'pruned', tran, ['6A','14','15B'] ).
verb( 'pry', 'pries', 'prying', 'pried', 'pried', _, ['2A','2C','3A','15A','15B','22'] ).
verb( 'psychoanalyse', 'psychoanalyses', 'psychoanalysing', 'psychoanalysed', 'psychoanalysed', tran, [] ).
verb( 'psychoanalyze', 'psychoanalyzes', 'psychoanalyzing', 'psychoanalyzed', 'psychoanalyzed', tran, [] ).
verb( 'pub-crawl', 'pub-crawls', 'pub-crawling', 'pub-crawled', 'pub-crawled', intran, [] ).
verb( 'publicize', 'publicizes', 'publicizing', 'publicized', 'publicized', tran, ['6A'] ).
verb( 'publish', 'publishes', 'publishing', 'published', 'published', tran, ['6A'] ).
verb( 'pucker', 'puckers', 'puckering', 'puckered', 'puckered', _, ['2A','2C','6A','15B'] ).
verb( 'puddle', 'puddles', 'puddling', 'puddled', 'puddled', tran, [] ).
verb( 'puff', 'puffs', 'puffing', 'puffed', 'puffed', _, ['2A','2C','6A','15A','15B'] ).
verb( 'puke', 'pukes', 'puking', 'puked', 'puked', _, [] ).
verb( 'pule', 'pules', 'puling', 'puled', 'puled', intran, ['2A'] ).
verb( 'pull', 'pulls', 'pulling', 'pulled', 'pulled', _, ['2A','2C','3A','6A','15A','15B','22'] ).
verb( 'pullulate', 'pullulates', 'pullulating', 'pullulated', 'pullulated', intran, [] ).
verb( 'pulp', 'pulps', 'pulping', 'pulped', 'pulped', _, ['2A','6A'] ).
verb( 'pulsate', 'pulsates', 'pulsating', 'pulsated', 'pulsated', _, ['2A','6A'] ).
verb( 'pulse', 'pulses', 'pulsing', 'pulsed', 'pulsed', intran, ['2C'] ).
verb( 'pulverize', 'pulverizes', 'pulverizing', 'pulverized', 'pulverized', _, ['2A','6A'] ).
verb( 'pummel', 'pummels', 'pummelling', 'pummelled', 'pummelled', tran, ['6A','15B'] ).
verb( 'pump', 'pumps', 'pumping', 'pumped', 'pumped', _, ['2A','2C','6A','15A','15B','22'] ).
verb( 'pun', 'puns', 'punning', 'punned', 'punned', intran, ['2A','3A'] ).
verb( 'punch', 'punches', 'punching', 'punched', 'punched', tran, ['6A','15A','15B'] ).
verb( 'punctuate', 'punctuates', 'punctuating', 'punctuated', 'punctuated', tran, ['6A','15A'] ).
verb( 'puncture', 'punctures', 'puncturing', 'punctured', 'punctured', _, ['2A','6A'] ).
verb( 'punish', 'punishes', 'punishing', 'punished', 'punished', tran, ['6A','14'] ).
verb( 'punt', 'punts', 'punting', 'punted', 'punted', _, ['2A','6A'] ).
verb( 'purchase', 'purchases', 'purchasing', 'purchased', 'purchased', tran, ['6A'] ).
verb( 'purge', 'purges', 'purging', 'purged', 'purged', tran, ['6A','14','15A','15B'] ).
verb( 'purify', 'purifies', 'purifying', 'purified', 'purified', tran, ['6A','14'] ).
verb( 'purl', 'purls', 'purling', 'purled', 'purled', _, [] ).
verb( 'purloin', 'purloins', 'purloining', 'purloined', 'purloined', tran, ['6A'] ).
verb( 'purport', 'purports', 'purporting', 'purported', 'purported', tran, ['6A','7A','9'] ).
verb( 'purpose', 'purposes', 'purposing', 'purposed', 'purposed', tran, ['6A','6D','7A','9'] ).
verb( 'purr', 'purrs', 'purring', 'purred', 'purred', _, ['2A','2C','6A'] ).
verb( 'purse', 'purses', 'pursing', 'pursed', 'pursed', tran, ['6A','15B'] ).
verb( 'pursue', 'pursues', 'pursuing', 'pursued', 'pursued', tran, ['6A'] ).
verb( 'purvey', 'purveys', 'purveying', 'purveyed', 'purveyed', _, ['3A','6A','14'] ).
verb( 'push', 'pushes', 'pushing', 'pushed', 'pushed', _, ['2A','2C','3A','6A','14','15A','15B','17','22'] ).
verb( 'pussyfoot', 'pussyfoots', 'pussyfooting', 'pussyfooted', 'pussyfooted', intran, ['2A','2C'] ).
verb( 'put', 'puts', 'putting', 'put', 'put', _, ['2C','6A','14','15A','15B','22'] ).
verb( 'put', 'puts', 'putting', 'putted', 'putted', _, [] ).
verb( 'putrefy', 'putrefies', 'putrefying', 'putrefied', 'putrefied', _, ['2A','6A'] ).
verb( 'putt', 'putts', 'putting', 'putted', 'putted', _, ['2A','6A'] ).
verb( 'putter', 'putters', 'puttering', 'puttered', 'puttered', _, [] ).
verb( 'putty', 'putties', 'puttying', 'puttied', 'puttied', tran, ['6A','15B'] ).
verb( 'puzzle', 'puzzles', 'puzzling', 'puzzled', 'puzzled', _, ['3A','6A','15B'] ).
verb( 'quack', 'quacks', 'quacking', 'quacked', 'quacked', intran, [] ).
verb( 'quadruple', 'quadruples', 'quadrupling', 'quadrupled', 'quadrupled', _, ['2A','6A'] ).
verb( 'quadruplicate', 'quadruplicates', 'quadruplicating', 'quadruplicated', 'quadruplicated', tran, ['6A'] ).
verb( 'quaff', 'quaffs', 'quaffing', 'quaffed', 'quaffed', _, ['2A','6A','15B'] ).
verb( 'quail', 'quails', 'quailing', 'quailed', 'quailed', intran, ['2A','3A'] ).
verb( 'quake', 'quakes', 'quaking', 'quaked', 'quaked', intran, ['2A','2C'] ).
verb( 'qualify', 'qualifies', 'qualifying', 'qualified', 'qualified', _, ['2C','3A','4A','6A','14','16B','17'] ).
verb( 'quantify', 'quantifies', 'quantifying', 'quantified', 'quantified', tran, ['6A'] ).
verb( 'quarantine', 'quarantines', 'quarantining', 'quarantined', 'quarantined', tran, ['6A'] ).
verb( 'quarrel', 'quarrels', 'quarrelling', 'quarrelled', 'quarrelled', intran, ['2A','2C','3A'] ).
verb( 'quarry', 'quarries', 'quarrying', 'quarried', 'quarried', _, ['2A','2C','6A','15A','15B'] ).
verb( 'quarter', 'quarters', 'quartering', 'quartered', 'quartered', tran, ['6A','15A'] ).
verb( 'quash', 'quashes', 'quashing', 'quashed', 'quashed', tran, ['6A'] ).
verb( 'quaver', 'quavers', 'quavering', 'quavered', 'quavered', _, ['2A','6A','15B'] ).
verb( 'queen', 'queens', 'queening', 'queened', 'queened', tran, [] ).
verb( 'queer', 'queers', 'queering', 'queered', 'queered', tran, ['6A'] ).
verb( 'quell', 'quells', 'quelling', 'quelled', 'quelled', tran, ['6A'] ).
verb( 'quench', 'quenches', 'quenching', 'quenched', 'quenched', tran, ['6A'] ).
verb( 'query', 'queries', 'querying', 'queried', 'queried', tran, ['6A','10'] ).
verb( 'quest', 'quests', 'questing', 'quested', 'quested', intran, ['3A'] ).
verb( 'question', 'questions', 'questioning', 'questioned', 'questioned', tran, ['6A','10'] ).
verb( 'queue', 'queues', 'queueing', 'queued', 'queued', intran, ['2A','2C','3A'] ).
verb( 'quibble', 'quibbles', 'quibbling', 'quibbled', 'quibbled', intran, [] ).
verb( 'quick-freeze', 'quick-freezes', 'quick-freezing', 'quick-froze', 'quick-frozen', tran, [] ).
verb( 'quicken', 'quickens', 'quickening', 'quickened', 'quickened', _, ['2A','6A'] ).
verb( 'quiet', 'quiets', 'quieting', 'quieted', 'quieted', _, [] ).
verb( 'quieten', 'quietens', 'quietening', 'quietened', 'quietened', _, ['2C','6A','15B'] ).
verb( 'quilt', 'quilts', 'quilting', 'quilted', 'quilted', tran, [] ).
verb( 'quip', 'quips', 'quipping', 'quipped', 'quipped', intran, [] ).
verb( 'quit', 'quits', 'quitting', 'quitted', 'quitted', tran, ['2A','6A','6D'] ).
verb( 'quiver', 'quivers', 'quivering', 'quivered', 'quivered', _, ['2A','6A'] ).
verb( 'quiz', 'quizzes', 'quizzing', 'quizzed', 'quizzed', tran, ['6A'] ).
verb( 'quote', 'quotes', 'quoting', 'quoted', 'quoted', tran, ['6A','13A','14'] ).
verb( 'rabbit', 'rabbits', 'rabbiting', 'rabbited', 'rabbited', intran, [] ).
verb( 'race', 'races', 'racing', 'raced', 'raced', _, ['2A','2C','3A','4A','6A','15A'] ).
verb( 'rack', 'racks', 'racking', 'racked', 'racked', tran, ['6A','15A'] ).
verb( 'racket', 'rackets', 'racketing', 'racketed', 'racketed', intran, ['2A','2C'] ).
verb( 'radiate', 'radiates', 'radiating', 'radiated', 'radiated', _, ['2A','3A','6A'] ).
verb( 'raffle', 'raffles', 'raffling', 'raffled', 'raffled', tran, ['6A','15B'] ).
verb( 'raft', 'rafts', 'rafting', 'rafted', 'rafted', _, ['2C','6A','15A','15B'] ).
verb( 'rag', 'rags', 'ragging', 'ragged', 'ragged', tran, ['6A'] ).
verb( 'rage', 'rages', 'raging', 'raged', 'raged', intran, ['2A','2C'] ).
verb( 'raid', 'raids', 'raiding', 'raided', 'raided', _, ['2A','6A'] ).
verb( 'rail', 'rails', 'railing', 'railed', 'railed', _, ['2A','3A','6A','15B'] ).
verb( 'railroad', 'railroads', 'railroading', 'railroaded', 'railroaded', tran, ['15A','15B'] ).
verb( 'rain', 'rains', 'raining', 'rained', 'rained', _, ['2C','14'] ).
verb( 'raise', 'raises', 'raising', 'raised', 'raised', tran, ['6A','15A','15B'] ).
verb( 'rake', 'rakes', 'raking', 'raked', 'raked', _, ['2A','2C','3A','6A','14','15A','15B','22'] ).
verb( 'rally', 'rallies', 'rallying', 'rallied', 'rallied', _, ['2A','2C','6A','15A'] ).
verb( 'ram', 'rams', 'ramming', 'rammed', 'rammed', tran, ['6A','15A','15B'] ).
verb( 'ramble', 'rambles', 'rambling', 'rambled', 'rambled', intran, ['2A','2C'] ).
verb( 'ramify', 'ramifies', 'ramifying', 'ramified', 'ramified', _, ['2A','6A'] ).
verb( 'ramp', 'ramps', 'ramping', 'ramped', 'ramped', tran, ['2C'] ).
verb( 'rampage', 'rampages', 'rampaging', 'rampaged', 'rampaged', intran, ['2A'] ).
verb( 'range', 'ranges', 'ranging', 'ranged', 'ranged', _, ['2C','3A','6A','15A'] ).
verb( 'rank', 'ranks', 'ranking', 'ranked', 'ranked', _, ['3A','6A','15A','16B'] ).
verb( 'rankle', 'rankles', 'rankling', 'rankled', 'rankled', intran, ['2A'] ).
verb( 'ransack', 'ransacks', 'ransacking', 'ransacked', 'ransacked', tran, ['6A','14','16A'] ).
verb( 'ransom', 'ransoms', 'ransoming', 'ransomed', 'ransomed', tran, ['6A'] ).
verb( 'rant', 'rants', 'ranting', 'ranted', 'ranted', _, ['2A','6A'] ).
verb( 'rap', 'raps', 'rapping', 'rapped', 'rapped', _, ['2A','2C','6A','15B'] ).
verb( 'rape', 'rapes', 'raping', 'raped', 'raped', tran, ['6A'] ).
verb( 'rarefy', 'rarefies', 'rarefying', 'rarefied', 'rarefied', _, ['2A','6A'] ).
verb( 'rase', 'rases', 'rasing', 'rased', 'rased', tran, ['6A'] ).
verb( 'rasp', 'rasps', 'rasping', 'rasped', 'rasped', _, ['2A','2C','6A','15A','15B','22'] ).
verb( 'rat', 'rats', 'ratting', 'ratted', 'ratted', tran, ['2A','3A'] ).
verb( 'rate', 'rates', 'rating', 'rated', 'rated', _, ['2C','2D','6A','14','15A','16B'] ).
verb( 'ratify', 'ratifies', 'ratifying', 'ratified', 'ratified', tran, ['6A'] ).
verb( 'ration', 'rations', 'rationing', 'rationed', 'rationed', tran, ['6A','15B'] ).
verb( 'rationalize', 'rationalizes', 'rationalizing', 'rationalized', 'rationalized', tran, ['6A'] ).
verb( 'rattle', 'rattles', 'rattling', 'rattled', 'rattled', _, ['2A','2C','6A','15A','15B'] ).
verb( 'ravage', 'ravages', 'ravaging', 'ravaged', 'ravaged', _, ['6A'] ).
verb( 'rave', 'raves', 'raving', 'raved', 'raved', intran, ['2A','2C','3A','15B'] ).
verb( 'ravel', 'ravels', 'ravelling', 'ravelled', 'ravelled', _, ['2A','2C','6A','15B'] ).
verb( 'ravish', 'ravishes', 'ravishing', 'ravished', 'ravished', tran, ['6A'] ).
verb( 'ray', 'rays', 'raying', 'rayed', 'rayed', _, [] ).
verb( 'raze', 'razes', 'razing', 'razed', 'razed', tran, ['6A'] ).
verb( 'razor', 'razors', 'razoring', 'razored', 'razored', tran, [] ).
verb( 're-address', 're-addresses', 're-addressing', 're-addressed', 're-addressed', tran, ['6A'] ).
verb( 're-afforest', 're-afforests', 're-afforesting', 're-afforested', 're-afforested', tran, [] ).
verb( 're-count', 're-counts', 're-counting', 're-counted', 're-counted', tran, ['6A'] ).
verb( 're-cover', 're-covers', 're-covering', 're-covered', 're-covered', tran, ['6A'] ).
verb( 're-echo', 're-echos', 're-echoing', 're-echoed', 're-echoed', intran, ['2A'] ).
verb( 're-form', 're-forms', 're-forming', 're-formed', 're-formed', _, [] ).
verb( 're-join', 're-joins', 're-joining', 're-joined', 're-joined', tran, ['6A'] ).
verb( 'reach', 'reaches', 'reaching', 'reached', 'reached', _, ['2C','3A','6A','12B','13B','15B'] ).
verb( 'react', 'reacts', 'reacting', 'reacted', 'reacted', intran, ['2A','3A'] ).
verb( 'read', 'reads', 'reading', 'read', 'read', _, ['2A','2B','2C','6A','12A','13A','15A','15B','16B'] ).
verb( 'readjust', 'readjusts', 'readjusting', 'readjusted', 'readjusted', tran, ['3A','6A','15A'] ).
verb( 'readmit', 'readmits', 'readmitting', 'readmitted', 'readmitted', tran, ['3A','6A','6C','9','14','25'] ).
verb( 'reaffirm', 'reaffirms', 'reaffirming', 'reaffirmed', 'reaffirmed', tran, ['6A','9'] ).
verb( 'realize', 'realizes', 'realizing', 'realized', 'realized', tran, ['6A','9','10','14'] ).
verb( 'reanimate', 'reanimates', 'reanimating', 'reanimated', 'reanimated', tran, ['6A'] ).
verb( 'reap', 'reaps', 'reaping', 'reaped', 'reaped', _, ['2A','6A'] ).
verb( 'reappear', 'reappears', 'reappearing', 'reappeared', 'reappeared', intran, ['2A'] ).
verb( 'rear', 'rears', 'rearing', 'reared', 'reared', _, ['2A','2C','6A','15B'] ).
verb( 'rearm', 'rearms', 'rearming', 'rearmed', 'rearmed', _, ['2A','6A'] ).
verb( 'rearrange', 'rearranges', 'rearranging', 'rearranged', 'rearranged', tran, ['3A','4C','6A','14','15A'] ).
verb( 'reason', 'reasons', 'reasoning', 'reasoned', 'reasoned', _, ['2A','3A','6A','9','14'] ).
verb( 'reassemble', 'reassembles', 'reassembling', 'reassembled', 'reassembled', _, ['2A','6A'] ).
verb( 'reassess', 'reassesses', 'reassessing', 'reassessed', 'reassessed', tran, ['6A','14'] ).
verb( 'reassure', 'reassures', 'reassuring', 'reassured', 'reassured', tran, ['6A'] ).
verb( 'reattribute', 'reattributes', 'reattributing', 'reattributed', 'reattributed', tran, ['14'] ).
verb( 'rebel', 'rebels', 'rebelling', 'rebelled', 'rebelled', intran, ['2A','3A'] ).
verb( 'rebind', 'rebinds', 'rebinding', 'rebound', 'rebound', tran, ['6A'] ).
verb( 'rebound', 'rebounds', 'rebounding', 'rebounded', 'rebounded', intran, ['2A','3A'] ).
verb( 'rebuff', 'rebuffs', 'rebuffing', 'rebuffed', 'rebuffed', tran, ['6A'] ).
verb( 'rebuild', 'rebuilds', 'rebuilding', 'rebuilt', 'rebuilt', tran, ['6A'] ).
verb( 'rebuke', 'rebukes', 'rebuking', 'rebuked', 'rebuked', tran, ['6A','14'] ).
verb( 'rebut', 'rebuts', 'rebutting', 'rebutted', 'rebutted', tran, ['6A'] ).
verb( 'recall', 'recalls', 'recalling', 'recalled', 'recalled', tran, ['6A','6C','8','9','10','14','19C'] ).
verb( 'recant', 'recants', 'recanting', 'recanted', 'recanted', _, ['2A','6A'] ).
verb( 'recap', 'recaps', 'recapping', 'recapped', 'recapped', _, [] ).
verb( 'recap', 'recaps', 'recapping', 'recapped', 'recapped', tran, [] ).
verb( 'recapitulate', 'recapitulates', 'recapitulating', 'recapitulated', 'recapitulated', _, ['2A','6A'] ).
verb( 'recapture', 'recaptures', 'recapturing', 'recaptured', 'recaptured', tran, ['6A'] ).
verb( 'recast', 'recasts', 'recasting', 'recast', 'recast', tran, ['6A'] ).
verb( 'recede', 'recedes', 'receding', 'receded', 'receded', intran, ['2A','3A'] ).
verb( 'receipt', 'receipts', 'receipting', 'receipted', 'receipted', tran, ['6A'] ).
verb( 'receive', 'receives', 'receiving', 'received', 'received', _, ['2A','6A'] ).
verb( 'recess', 'recesses', 'recessing', 'recessed', 'recessed', tran, ['6A'] ).
verb( 'reciprocate', 'reciprocates', 'reciprocating', 'reciprocated', 'reciprocated', _, ['2A','6A'] ).
verb( 'recite', 'recites', 'reciting', 'recited', 'recited', _, ['2A','6A','15A'] ).
verb( 'reckon', 'reckons', 'reckoning', 'reckoned', 'reckoned', _, ['2A','6A','7A','9','15B','16B','25'] ).
verb( 'reclaim', 'reclaims', 'reclaiming', 'reclaimed', 'reclaimed', tran, ['6A'] ).
verb( 'recline', 'reclines', 'reclining', 'reclined', 'reclined', _, ['2A','2C','15A'] ).
verb( 'recognize', 'recognizes', 'recognizing', 'recognized', 'recognized', tran, ['6A','9','16A','25'] ).
verb( 'recoil', 'recoils', 'recoiling', 'recoiled', 'recoiled', intran, ['2A','3A'] ).
verb( 'recollect', 'recollects', 'recollecting', 'recollected', 'recollected', _, ['2A','6A','6C','8','9','10'] ).
verb( 'recommend', 'recommends', 'recommending', 'recommended', 'recommended', tran, ['6A','6C','9','12A','13A','14','16A','17'] ).
verb( 'recommit', 'recommits', 'recommitting', 'recommitted', 'recommitted', tran, ['6A','14','16A'] ).
verb( 'recompense', 'recompenses', 'recompensing', 'recompensed', 'recompensed', tran, ['6A','14'] ).
verb( 'reconcile', 'reconciles', 'reconciling', 'reconciled', 'reconciled', tran, ['6A','14'] ).
verb( 'recondition', 'reconditions', 'reconditioning', 'reconditioned', 'reconditioned', tran, ['6A'] ).
verb( 'reconnoitre', 'reconnoitres', 'reconnoitring', 'reconnoitred', 'reconnoitred', _, ['2A','6A'] ).
verb( 'reconsecrate', 'reconsecrates', 'reconsecrating', 'reconsecrated', 'reconsecrated', tran, ['6A','14','23'] ).
verb( 'reconsider', 'reconsiders', 'reconsidering', 'reconsidered', 'reconsidered', tran, ['6A','6C','8','9','10','25'] ).
verb( 'reconstruct', 'reconstructs', 'reconstructing', 'reconstructed', 'reconstructed', tran, ['6A'] ).
verb( 'reconvict', 'reconvicts', 'reconvicting', 'reconvicted', 'reconvicted', tran, ['6A','14'] ).
verb( 'record', 'records', 'recording', 'recorded', 'recorded', tran, ['6A'] ).
verb( 'recount', 'recounts', 'recounting', 'recounted', 'recounted', tran, ['6A'] ).
verb( 'recoup', 'recoups', 'recouping', 'recouped', 'recouped', tran, ['6A','14'] ).
verb( 'recover', 'recovers', 'recovering', 'recovered', 'recovered', _, ['2A','3A','6A'] ).
verb( 'recreate', 'recreates', 'recreating', 'recreated', 'recreated', tran, ['6A','23'] ).
verb( 'recriminate', 'recriminates', 'recriminating', 'recriminated', 'recriminated', intran, ['2A','3A'] ).
verb( 'recruit', 'recruits', 'recruiting', 'recruited', 'recruited', _, ['6A'] ).
verb( 'rectify', 'rectifies', 'rectifying', 'rectified', 'rectified', tran, ['6A'] ).
verb( 'recuperate', 'recuperates', 'recuperating', 'recuperated', 'recuperated', _, ['2A','6A'] ).
verb( 'recur', 'recurs', 'recurring', 'recurred', 'recurred', intran, ['2A','3A'] ).
verb( 'recurve', 'recurves', 'recurving', 'recurved', 'recurved', _, [] ).
verb( 'recycle', 'recycles', 'recycling', 'recycled', 'recycled', tran, ['6A'] ).
verb( 'redact', 'redacts', 'redacting', 'redacted', 'redacted', tran, ['6A'] ).
verb( 'redden', 'reddens', 'reddening', 'reddened', 'reddened', _, ['2A','6A'] ).
verb( 'redecorate', 'redecorates', 'redecorating', 'redecorated', 'redecorated', tran, ['6A','14'] ).
verb( 'redeem', 'redeems', 'redeeming', 'redeemed', 'redeemed', tran, ['6A','14'] ).
verb( 'redefine', 'redefines', 'redefining', 'redefined', 'redefined', tran, ['6A'] ).
verb( 'redeploy', 'redeploys', 'redeploying', 'redeployed', 'redeployed', tran, ['6A'] ).
verb( 'redesign', 'redesigns', 'redesigning', 'redesigned', 'redesigned', tran, ['2A','2C','6A','14','16A','16B'] ).
verb( 'redevelop', 'redevelops', 'redeveloping', 'redeveloped', 'redeveloped', tran, ['2A','2C','3A','6A'] ).
verb( 'rediscover', 'rediscovers', 'rediscovering', 'rediscovered', 'rediscovered', tran, ['6A','8','9','10','25'] ).
verb( 'redistribute', 'redistributes', 'redistributing', 'redistributed', 'redistributed', tran, ['6A','14'] ).
verb( 'redo', 'redoes', 'redoing', 'redid', 'redone', tran, ['6A'] ).
verb( 'redouble', 'redoubles', 'redoubling', 'redoubled', 'redoubled', _, ['2A','6A'] ).
verb( 'redound', 'redounds', 'redounding', 'redounded', 'redounded', intran, ['3A'] ).
verb( 'redress', 'redresses', 'redressing', 'redressed', 'redressed', tran, ['6A'] ).
verb( 'reduce', 'reduces', 'reducing', 'reduced', 'reduced', _, ['2A','2B','6A','14'] ).
verb( 'reduplicate', 'reduplicates', 'reduplicating', 'reduplicated', 'reduplicated', tran, ['6A'] ).
verb( 'reef', 'reefs', 'reefing', 'reefed', 'reefed', tran, ['6A'] ).
verb( 'reek', 'reeks', 'reeking', 'reeked', 'reeked', intran, ['3A'] ).
verb( 'reel', 'reels', 'reeling', 'reeled', 'reeled', _, ['2A','2C','6A','15A','15B'] ).
verb( 'reface', 'refaces', 'refacing', 'refaced', 'refaced', tran, [] ).
verb( 'refashion', 'refashions', 'refashioning', 'refashioned', 'refashioned', tran, ['6A','15A'] ).
verb( 'refer', 'refers', 'referring', 'referred', 'referred', _, ['3A','14','15B'] ).
verb( 'referee', 'referees', 'refereeing', 'refereed', 'refereed', _, ['2A','6A'] ).
verb( 'refill', 'refills', 'refilling', 'refilled', 'refilled', tran, [] ).
verb( 'refine', 'refines', 'refining', 'refined', 'refined', _, ['2A','6A'] ).
verb( 'refit', 'refits', 'refitting', 'refitted', 'refitted', _, ['2A','6A'] ).
verb( 'reflate', 'reflates', 'reflating', 'reflated', 'reflated', tran, ['6A'] ).
verb( 'reflect', 'reflects', 'reflecting', 'reflected', 'reflected', _, ['2A','3A','6A','8','9','10','14'] ).
verb( 'refloat', 'refloats', 'refloating', 'refloated', 'refloated', _, ['2A','6A'] ).
verb( 'reforest', 'reforests', 'reforesting', 'reforested', 'reforested', tran, [] ).
verb( 'reform', 'reforms', 'reforming', 'reformed', 'reformed', _, ['2A','6A'] ).
verb( 'refract', 'refracts', 'refracting', 'refracted', 'refracted', tran, ['6A'] ).
verb( 'refrain', 'refrains', 'refraining', 'refrained', 'refrained', intran, ['2A','3A'] ).
verb( 'refresh', 'refreshes', 'refreshing', 'refreshed', 'refreshed', tran, ['6A'] ).
verb( 'refrigerate', 'refrigerates', 'refrigerating', 'refrigerated', 'refrigerated', tran, ['6A'] ).
verb( 'refuel', 'refuels', 'refuelling', 'refuelled', 'refuelled', _, ['2A','6A'] ).
verb( 'refund', 'refunds', 'refunding', 'refunded', 'refunded', tran, ['6A'] ).
verb( 'refurbish', 'refurbishes', 'refurbishing', 'refurbished', 'refurbished', tran, ['6A'] ).
verb( 'refurnish', 'refurnishes', 'refurnishing', 'refurnished', 'refurnished', tran, ['6A','14'] ).
verb( 'refuse', 'refuses', 'refusing', 'refused', 'refused', _, ['2A','6A','7A','12C'] ).
verb( 'refute', 'refutes', 'refuting', 'refuted', 'refuted', tran, ['6A'] ).
verb( 'regain', 'regains', 'regaining', 'regained', 'regained', tran, ['6A'] ).
verb( 'regale', 'regales', 'regaling', 'regaled', 'regaled', tran, ['6A','14'] ).
verb( 'regard', 'regards', 'regarding', 'regarded', 'regarded', tran, ['6A','14','16B'] ).
verb( 'regenerate', 'regenerates', 'regenerating', 'regenerated', 'regenerated', _, ['2A','6A'] ).
verb( 'regiment', 'regiments', 'regimenting', 'regimented', 'regimented', tran, ['6A'] ).
verb( 'register', 'registers', 'registering', 'registered', 'registered', _, ['2A','3A','6A','14'] ).
verb( 'regress', 'regresses', 'regressing', 'regressed', 'regressed', intran, ['2A'] ).
verb( 'regret', 'regrets', 'regretting', 'regretted', 'regretted', tran, ['6A','6D','7A','9'] ).
verb( 'regroup', 'regroups', 'regrouping', 'regrouped', 'regrouped', _, ['2A','6A'] ).
verb( 'regularize', 'regularizes', 'regularizing', 'regularized', 'regularized', tran, ['6A'] ).
verb( 'regulate', 'regulates', 'regulating', 'regulated', 'regulated', tran, ['6A'] ).
verb( 'regurgitate', 'regurgitates', 'regurgitating', 'regurgitated', 'regurgitated', _, ['2A','6A'] ).
verb( 'rehabilitate', 'rehabilitates', 'rehabilitating', 'rehabilitated', 'rehabilitated', tran, ['6A'] ).
verb( 'rehash', 'rehashes', 'rehashing', 'rehashed', 'rehashed', tran, [] ).
verb( 'rehear', 'rehears', 'rehearing', 'reheard', 'reheard', tran, ['6A'] ).
verb( 'rehearse', 'rehearses', 'rehearsing', 'rehearsed', 'rehearsed', _, ['2A','6A'] ).
verb( 'rehouse', 'rehouses', 'rehousing', 'rehoused', 'rehoused', tran, ['6A'] ).
verb( 'reign', 'reigns', 'reigning', 'reigned', 'reigned', intran, ['2A','3A'] ).
verb( 'reimburse', 'reimburses', 'reimbursing', 'reimbursed', 'reimbursed', tran, ['6A','12A','13A','14'] ).
verb( 'rein', 'reins', 'reining', 'reined', 'reined', tran, ['6A','15B'] ).
verb( 'reincarnate', 'reincarnates', 'reincarnating', 'reincarnated', 'reincarnated', tran, [] ).
verb( 'reinforce', 'reinforces', 'reinforcing', 'reinforced', 'reinforced', tran, ['6A'] ).
verb( 'reinstate', 'reinstates', 'reinstating', 'reinstated', 'reinstated', tran, ['6A','14'] ).
verb( 'reinsure', 'reinsures', 'reinsuring', 'reinsured', 'reinsured', tran, ['6A'] ).
verb( 'reintegrate', 'reintegrates', 'reintegrating', 'reintegrated', 'reintegrated', tran, ['2A','6A'] ).
verb( 'reinterpret', 'reinterprets', 'reinterpreting', 'reinterpreted', 'reinterpreted', tran, ['2A','6A','16B'] ).
verb( 'reissue', 'reissues', 'reissuing', 'reissued', 'reissued', tran, ['6A'] ).
verb( 'reiterate', 'reiterates', 'reiterating', 'reiterated', 'reiterated', tran, ['6A'] ).
verb( 'reject', 'rejects', 'rejecting', 'rejected', 'rejected', tran, ['6A'] ).
verb( 'rejig', 'rejigs', 'rejigging', 'rejigged', 'rejigged', tran, ['6A'] ).
verb( 'rejoice', 'rejoices', 'rejoicing', 'rejoiced', 'rejoiced', _, ['2A','2C','3B','4C','6A'] ).
verb( 'rejoin', 'rejoins', 'rejoining', 'rejoined', 'rejoined', tran, ['6A'] ).
verb( 'rejoin', 'rejoins', 'rejoining', 'rejoined', 'rejoined', _, ['2A','6A'] ).
verb( 'rejuvenate', 'rejuvenates', 'rejuvenating', 'rejuvenated', 'rejuvenated', _, ['2A','6A'] ).
verb( 'rekindle', 'rekindles', 'rekindling', 'rekindled', 'rekindled', _, ['2A','6A'] ).
verb( 'relapse', 'relapses', 'relapsing', 'relapsed', 'relapsed', intran, ['2A','3A'] ).
verb( 'relate', 'relates', 'relating', 'related', 'related', _, ['3A','6A','14'] ).
verb( 'relax', 'relaxes', 'relaxing', 'relaxed', 'relaxed', _, ['2A','2C','6A'] ).
verb( 'relay', 'relays', 'relaying', 'relayed', 'relayed', tran, [] ).
verb( 'relay', 'relays', 'relaying', 'relayed', 'relayed', tran, [] ).
verb( 'release', 'releases', 'releasing', 'released', 'released', tran, ['6A','14'] ).
verb( 'relegate', 'relegates', 'relegating', 'relegated', 'relegated', tran, ['14'] ).
verb( 'relent', 'relents', 'relenting', 'relented', 'relented', intran, ['2A'] ).
verb( 'relieve', 'relieves', 'relieving', 'relieved', 'relieved', tran, ['6A'] ).
verb( 'reline', 'relines', 'relining', 'relined', 'relined', tran, [] ).
verb( 'relinquish', 'relinquishes', 'relinquishing', 'relinquished', 'relinquished', tran, ['6A','14'] ).
verb( 'relish', 'relishes', 'relishing', 'relished', 'relished', tran, ['6A','6D'] ).
verb( 'relive', 'relives', 'reliving', 'relived', 'relived', tran, [] ).
verb( 'relocate', 'relocates', 'relocating', 'relocated', 'relocated', _, [] ).
verb( 'rely', 'relies', 'relying', 'relied', 'relied', intran, ['3A'] ).
verb( 'remain', 'remains', 'remaining', 'remained', 'remained', intran, ['2A','2B','2C','4A'] ).
verb( 'remake', 'remakes', 'remaking', 'remade', 'remade', tran, [] ).
verb( 'remand', 'remands', 'remanding', 'remanded', 'remanded', tran, ['6A'] ).
verb( 'remark', 'remarks', 'remarking', 'remarked', 'remarked', _, ['3A','6A','9','10'] ).
verb( 'remarry', 'remarries', 'remarrying', 'remarried', 'remarried', _, [] ).
verb( 'remedy', 'remedies', 'remedying', 'remedied', 'remedied', tran, ['6A'] ).
verb( 'remember', 'remembers', 'remembering', 'remembered', 'remembered', _, ['6A','6C','7A','8','9','10','14','16B','19C'] ).
verb( 'remilitarize', 'remilitarizes', 'remilitarizing', 'remilitarized', 'remilitarized', tran, ['6A'] ).
verb( 'remind', 'reminds', 'reminding', 'reminded', 'reminded', tran, ['6A','11','14','17','20','21'] ).
verb( 'reminisce', 'reminisces', 'reminiscing', 'reminisced', 'reminisced', intran, ['2A','3A'] ).
verb( 'remit', 'remits', 'remitting', 'remitted', 'remitted', _, ['2C','6A','12A','13A','14'] ).
verb( 'remodel', 'remodels', 'remodelling', 'remodelled', 'remodelled', tran, ['2A','6A','14','15A'] ).
verb( 'remonstrate', 'remonstrates', 'remonstrating', 'remonstrated', 'remonstrated', intran, ['2A','3A'] ).
verb( 'remould', 'remoulds', 'remoulding', 'remoulded', 'remoulded', tran, ['2A','6A','14'] ).
verb( 'remount', 'remounts', 'remounting', 'remounted', 'remounted', _, ['2A','6A'] ).
verb( 'remove', 'removes', 'removing', 'removed', 'removed', _, ['2A','2C','6A','14'] ).
verb( 'remunerate', 'remunerates', 'remunerating', 'remunerated', 'remunerated', tran, ['6A','14'] ).
verb( 'rename', 'renames', 'renaming', 'renamed', 'renamed', tran, ['6A'] ).
verb( 'rend', 'rends', 'rending', 'rent', 'rent', tran, ['6A','14','15A'] ).
verb( 'render', 'renders', 'rendering', 'rendered', 'rendered', tran, ['6A','12A','13A','14','15A','15B','22'] ).
verb( 'rendezvous', 'rendezvous', 'rendezvouing', 'rendezvoued', 'rendezvoued', intran, ['2A','2C'] ).
verb( 'renegade', 'renegades', 'renegading', 'renegaded', 'renegaded', intran, [] ).
verb( 'renege', 'reneges', 'reneging', 'reneged', 'reneged', intran, ['3A'] ).
verb( 'renegue', 'renegues', 'reneguing', 'renegued', 'renegued', intran, ['3A'] ).
verb( 'renew', 'renews', 'renewing', 'renewed', 'renewed', tran, ['6A'] ).
verb( 'renounce', 'renounces', 'renouncing', 'renounced', 'renounced', tran, ['6A'] ).
verb( 'renovate', 'renovates', 'renovating', 'renovated', 'renovated', tran, ['6A'] ).
verb( 'rent', 'rents', 'renting', 'rented', 'rented', _, ['2A','2C','6A','14','15A'] ).
verb( 'reopen', 'reopens', 'reopening', 'reopened', 'reopened', _, ['2A','6A'] ).
verb( 'reorganize', 'reorganizes', 'reorganizing', 'reorganized', 'reorganized', _, [] ).
verb( 'reorient', 'reorients', 'reorienting', 'reoriented', 'reoriented', _, [] ).
verb( 'reorientate', 'reorientates', 'reorientating', 'reorientated', 'reorientated', _, [] ).
verb( 'repaint', 'repaints', 'repainting', 'repainted', 'repainted', tran, ['2A','2C','6A','15A','15B','22'] ).
verb( 'repair', 'repairs', 'repairing', 'repaired', 'repaired', _, ['3A','6A'] ).
verb( 'repatriate', 'repatriates', 'repatriating', 'repatriated', 'repatriated', tran, ['6A'] ).
verb( 'repay', 'repays', 'repaying', 'repaid', 'repaid', _, ['2A','6A','14'] ).
verb( 'repeal', 'repeals', 'repealing', 'repealed', 'repealed', tran, ['6A'] ).
verb( 'repeat', 'repeats', 'repeating', 'repeated', 'repeated', _, ['2A','6A','9'] ).
verb( 'repel', 'repels', 'repelling', 'repelled', 'repelled', tran, ['6A'] ).
verb( 'repent', 'repents', 'repenting', 'repented', 'repented', _, ['2A','3A','6A','6D'] ).
verb( 'repine', 'repines', 'repining', 'repined', 'repined', intran, ['2A','3A'] ).
verb( 'replace', 'replaces', 'replacing', 'replaced', 'replaced', tran, ['6A','14','15A'] ).
verb( 'replant', 'replants', 'replanting', 'replanted', 'replanted', tran, ['6A','15A','15B'] ).
verb( 'replay', 'replays', 'replaying', 'replayed', 'replayed', tran, ['6A'] ).
verb( 'replenish', 'replenishes', 'replenishing', 'replenished', 'replenished', tran, ['6A','14'] ).
verb( 'reply', 'replies', 'replying', 'replied', 'replied', _, ['2A','3A','3B'] ).
verb( 'repoint', 'repoints', 'repointing', 'repointed', 'repointed', tran, ['6A'] ).
verb( 'report', 'reports', 'reporting', 'reported', 'reported', _, ['2A','3A','6A','6D','9','14','15A','15B','25'] ).
verb( 'repose', 'reposes', 'reposing', 'reposed', 'reposed', _, ['2A','2C','3A','6A','14','15A'] ).
verb( 'repot', 'repots', 'repotting', 'repotted', 'repotted', tran, [] ).
verb( 'reprehend', 'reprehends', 'reprehending', 'reprehended', 'reprehended', tran, ['6A'] ).
verb( 'represent', 'represents', 'representing', 'represented', 'represented', tran, ['6A','9','14','16A','25'] ).
verb( 'represent', 'represents', 'representing', 'represented', 'represented', tran, [] ).
verb( 'repress', 'represses', 'repressing', 'repressed', 'repressed', tran, ['6A'] ).
verb( 'reprieve', 'reprieves', 'reprieving', 'reprieved', 'reprieved', tran, ['6A'] ).
verb( 'reprimand', 'reprimands', 'reprimanding', 'reprimanded', 'reprimanded', tran, ['6A'] ).
verb( 'reprint', 'reprints', 'reprinting', 'reprinted', 'reprinted', tran, ['6A'] ).
verb( 'reproach', 'reproaches', 'reproaching', 'reproached', 'reproached', tran, ['6A','14'] ).
verb( 'reprobate', 'reprobates', 'reprobating', 'reprobated', 'reprobated', tran, ['6A'] ).
verb( 'reproduce', 'reproduces', 'reproducing', 'reproduced', 'reproduced', _, ['2A','6A'] ).
verb( 'reproof', 'reproofs', 'reproofing', 'reproofed', 'reproofed', tran, ['6A'] ).
verb( 'reprove', 'reproves', 'reproving', 'reproved', 'reproved', tran, ['6A','14'] ).
verb( 'repudiate', 'repudiates', 'repudiating', 'repudiated', 'repudiated', tran, ['6A'] ).
verb( 'repulse', 'repulses', 'repulsing', 'repulsed', 'repulsed', tran, ['6A'] ).
verb( 'repute', 'reputes', 'reputing', 'reputed', 'reputed', tran, ['25'] ).
verb( 'request', 'requests', 'requesting', 'requested', 'requested', tran, ['6A','9','17'] ).
verb( 'require', 'requires', 'requiring', 'required', 'required', tran, ['6A','6D','9','14','17'] ).
verb( 'requisition', 'requisitions', 'requisitioning', 'requisitioned', 'requisitioned', tran, ['6A','14'] ).
verb( 'requite', 'requites', 'requiting', 'requited', 'requited', tran, ['6A','14'] ).
verb( 'reread', 'rereads', 'rereading', 'reread', 'reread', tran, ['2A','2B','2C','6A','12A','13A','15A','15B','16B'] ).
verb( 'rerun', 'reruns', 'rerunning', 'reran', 'rerun', tran, [] ).
verb( 'rescind', 'rescinds', 'rescinding', 'rescinded', 'rescinded', tran, ['6A'] ).
verb( 'rescue', 'rescues', 'rescuing', 'rescued', 'rescued', tran, ['6A','14'] ).
verb( 'research', 'researches', 'researching', 'researched', 'researched', intran, ['2A','3A'] ).
verb( 'reseat', 'reseats', 'reseating', 'reseated', 'reseated', tran, ['6A'] ).
verb( 'reseed', 'reseeds', 'reseeding', 'reseeded', 'reseeded', tran, ['2A','6A'] ).
verb( 'resell', 'resells', 'reselling', 'resold', 'resold', tran, ['2A','2C','6A','12A','13A','15B'] ).
verb( 'resemble', 'resembles', 'resembling', 'resembled', 'resembled', tran, ['6B'] ).
verb( 'resent', 'resents', 'resenting', 'resented', 'resented', tran, ['6A','6C','19C'] ).
verb( 'reserve', 'reserves', 'reserving', 'reserved', 'reserved', tran, ['6A','14'] ).
verb( 'reset', 'resets', 'resetting', 'reset', 'reset', tran, ['6A'] ).
verb( 'resettle', 'resettles', 'resettling', 'resettled', 'resettled', _, ['2A','6A'] ).
verb( 'reshape', 'reshapes', 'reshaping', 'reshaped', 'reshaped', tran, ['2A','2C','6A','15A'] ).
verb( 'reshuffle', 'reshuffles', 'reshuffling', 'reshuffled', 'reshuffled', tran, ['6A'] ).
verb( 'reside', 'resides', 'residing', 'resided', 'resided', intran, ['2C','3A'] ).
verb( 'resign', 'resigns', 'resigning', 'resigned', 'resigned', _, ['2A','3A','6A','14'] ).
verb( 'resist', 'resists', 'resisting', 'resisted', 'resisted', _, ['2A','6A','6C'] ).
verb( 'resole', 'resoles', 'resoling', 'resoled', 'resoled', tran, ['6A'] ).
verb( 'resolve', 'resolves', 'resolving', 'resolved', 'resolved', _, ['3A','6A','7A','9','14'] ).
verb( 'resonate', 'resonates', 'resonating', 'resonated', 'resonated', _, [] ).
verb( 'resort', 'resorts', 'resorting', 'resorted', 'resorted', intran, ['3A'] ).
verb( 'resound', 'resounds', 'resounding', 'resounded', 'resounded', _, ['2A','2C','3A'] ).
verb( 'respect', 'respects', 'respecting', 'respected', 'respected', tran, ['6A'] ).
verb( 'respire', 'respires', 'respiring', 'respired', 'respired', intran, ['2A'] ).
verb( 'respite', 'respites', 'respiting', 'respited', 'respited', tran, ['6A'] ).
verb( 'respond', 'responds', 'responding', 'responded', 'responded', intran, ['2A','2C','3A','3B'] ).
verb( 'rest', 'rests', 'resting', 'rested', 'rested', _, ['2A','2B','2C','2D','3A','6A','14'] ).
verb( 'restart', 'restarts', 'restarting', 'restarted', 'restarted', _, ['2A','2C','3A','6A','6D','7A','15A','19B'] ).
verb( 'restate', 'restates', 'restating', 'restated', 'restated', tran, ['6A'] ).
verb( 'restock', 'restocks', 'restocking', 'restocked', 'restocked', tran, ['6A'] ).
verb( 'restore', 'restores', 'restoring', 'restored', 'restored', tran, ['6A','14'] ).
verb( 'restrain', 'restrains', 'restraining', 'restrained', 'restrained', tran, ['6A','14'] ).
verb( 'restrict', 'restricts', 'restricting', 'restricted', 'restricted', tran, ['6A','14'] ).
verb( 'restructure', 'restructures', 'restructuring', 'restructured', 'restructured', tran, ['6A'] ).
verb( 'result', 'results', 'resulting', 'resulted', 'resulted', intran, ['2A','3A'] ).
verb( 'resume', 'resumes', 'resuming', 'resumed', 'resumed', tran, ['6A','6D'] ).
verb( 'resurface', 'resurfaces', 'resurfacing', 'resurfaced', 'resurfaced', _, ['2A','6A'] ).
verb( 'resurrect', 'resurrects', 'resurrecting', 'resurrected', 'resurrected', _, ['2A','6A'] ).
verb( 'resuscitate', 'resuscitates', 'resuscitating', 'resuscitated', 'resuscitated', _, ['2A','6A'] ).
verb( 'ret', 'rets', 'retting', 'retted', 'retted', tran, ['6A'] ).
verb( 'retail', 'retails', 'retailing', 'retailed', 'retailed', _, ['3A','6A'] ).
verb( 'retain', 'retains', 'retaining', 'retained', 'retained', tran, ['6A'] ).
verb( 'retake', 'retakes', 'retaking', 'retook', 'retaken', tran, ['6A'] ).
verb( 'retaliate', 'retaliates', 'retaliating', 'retaliated', 'retaliated', intran, ['2A','3A'] ).
verb( 'retard', 'retards', 'retarding', 'retarded', 'retarded', tran, ['6A'] ).
verb( 'retch', 'retches', 'retching', 'retched', 'retched', intran, ['2A'] ).
verb( 'retell', 'retells', 'retelling', 'retold', 'retold', tran, ['6A'] ).
verb( 'rethink', 'rethinks', 'rethinking', 'rethought', 'rethought', _, ['2A','6A'] ).
verb( 'reticulate', 'reticulates', 'reticulating', 'reticulated', 'reticulated', _, ['2A','6A'] ).
verb( 'retire', 'retires', 'retiring', 'retired', 'retired', _, ['2A','2C','3A','6A'] ).
verb( 'retool', 'retools', 'retooling', 'retooled', 'retooled', tran, ['6A'] ).
verb( 'retort', 'retorts', 'retorting', 'retorted', 'retorted', _, ['2A','6A','9','14'] ).
verb( 'retouch', 'retouches', 'retouching', 'retouched', 'retouched', tran, ['6A'] ).
verb( 'retrace', 'retraces', 'retracing', 'retraced', 'retraced', tran, ['6A'] ).
verb( 'retract', 'retracts', 'retracting', 'retracted', 'retracted', _, ['2A','6A'] ).
verb( 'retransmit', 'retransmits', 'retransmitting', 'retransmitted', 'retransmitted', tran, ['6A','14'] ).
verb( 'retread', 'retreads', 'retreading', 'retreaded', 'retreaded', tran, [] ).
verb( 'retreat', 'retreats', 'retreating', 'retreated', 'retreated', intran, ['2A','2C','3A'] ).
verb( 'retrench', 'retrenches', 'retrenching', 'retrenched', 'retrenched', _, ['2A','6A'] ).
verb( 'retrieve', 'retrieves', 'retrieving', 'retrieved', 'retrieved', _, ['2A','6A','14'] ).
verb( 'retrograde', 'retrogrades', 'retrograding', 'retrograded', 'retrograded', intran, ['2A'] ).
verb( 'retrogress', 'retrogresses', 'retrogressing', 'retrogressed', 'retrogressed', intran, ['2A'] ).
verb( 'return', 'returns', 'returning', 'returned', 'returned', _, ['2A','2C','3A','4A','6A','12A','13A','15A','16A'] ).
verb( 'reunite', 'reunites', 'reuniting', 'reunited', 'reunited', _, ['2A','6A'] ).
verb( 'rev', 'revs', 'revving', 'revved', 'revved', _, ['2A','2C','6A','15B'] ).
verb( 'revalue', 'revalues', 'revaluing', 'revalued', 'revalued', tran, ['6A'] ).
verb( 'revamp', 'revamps', 'revamping', 'revamped', 'revamped', tran, ['6A'] ).
verb( 'reveal', 'reveals', 'revealing', 'revealed', 'revealed', tran, ['6A','9','14','25'] ).
verb( 'revel', 'revels', 'revelling', 'revelled', 'revelled', intran, ['2A','2B','2C','3A','15B'] ).
verb( 'revenge', 'revenges', 'revenging', 'revenged', 'revenged', tran, ['6A'] ).
verb( 'reverberate', 'reverberates', 'reverberating', 'reverberated', 'reverberated', _, ['2A','6A'] ).
verb( 'revere', 'reveres', 'revering', 'revered', 'revered', tran, ['6A'] ).
verb( 'reverence', 'reverences', 'reverencing', 'reverenced', 'reverenced', tran, ['6A'] ).
verb( 'reverse', 'reverses', 'reversing', 'reversed', 'reversed', _, ['2A','6A'] ).
verb( 'revert', 'reverts', 'reverting', 'reverted', 'reverted', intran, ['2A','3A'] ).
verb( 'review', 'reviews', 'reviewing', 'reviewed', 'reviewed', _, ['2A','2C','6A'] ).
verb( 'revile', 'reviles', 'reviling', 'reviled', 'reviled', _, ['3A','6A'] ).
verb( 'revise', 'revises', 'revising', 'revised', 'revised', tran, ['6A'] ).
verb( 'revisit', 'revisits', 'revisiting', 'revisited', 'revisited', tran, ['2C','3A','6A','14'] ).
verb( 'revitalize', 'revitalizes', 'revitalizing', 'revitalized', 'revitalized', tran, [] ).
verb( 'revive', 'revives', 'reviving', 'revived', 'revived', _, ['2A','6A'] ).
verb( 'revivify', 'revivifies', 'revivifying', 'revivified', 'revivified', tran, [] ).
verb( 'revoke', 'revokes', 'revoking', 'revoked', 'revoked', _, ['2A','6A'] ).
verb( 'revolt', 'revolts', 'revolting', 'revolted', 'revolted', _, ['2A','3A','6A'] ).
verb( 'revolutionize', 'revolutionizes', 'revolutionizing', 'revolutionized', 'revolutionized', tran, ['6A'] ).
verb( 'revolve', 'revolves', 'revolving', 'revolved', 'revolved', _, ['2A','3A','6A'] ).
verb( 'reward', 'rewards', 'rewarding', 'rewarded', 'rewarded', tran, ['6A','14'] ).
verb( 'rewire', 'rewires', 'rewiring', 'rewired', 'rewired', tran, [] ).
verb( 'reword', 'rewords', 'rewording', 'reworded', 'reworded', tran, ['6A'] ).
verb( 'rewrite', 'rewrites', 'rewriting', 'rewrote', 'rewritten', tran, ['6A'] ).
verb( 'rhapsodize', 'rhapsodizes', 'rhapsodizing', 'rhapsodized', 'rhapsodized', intran, ['2A','3A'] ).
verb( 'rhyme', 'rhymes', 'rhyming', 'rhymed', 'rhymed', _, ['2A','3A','6A'] ).
verb( 'rib', 'ribs', 'ribbing', 'ribbed', 'ribbed', tran, ['6A'] ).
verb( 'rick', 'ricks', 'ricking', 'ricked', 'ricked', tran, ['6A'] ).
verb( 'ricochet', 'ricochets', 'ricocheting', 'ricocheted', 'ricocheted', _, ['2A','6A'] ).
verb( 'rid', 'rids', 'riding', 'rid', 'rid', tran, ['14'] ).
verb( 'riddle', 'riddles', 'riddling', 'riddled', 'riddled', tran, ['6A','14'] ).
verb( 'ride', 'rides', 'riding', 'rode', 'ridden', _, ['2A','2B','2C','2D','3A','4A','6A','15A'] ).
verb( 'ridge', 'ridges', 'ridging', 'ridged', 'ridged', tran, ['6A'] ).
verb( 'ridicule', 'ridicules', 'ridiculing', 'ridiculed', 'ridiculed', tran, [] ).
verb( 'riffle', 'riffles', 'riffling', 'riffled', 'riffled', _, ['3A','6A'] ).
verb( 'rifle', 'rifles', 'rifling', 'rifled', 'rifled', tran, ['6A'] ).
verb( 'rig', 'rigs', 'rigging', 'rigged', 'rigged', _, ['2A','2C','6A','14','15B'] ).
verb( 'right', 'rights', 'righting', 'righted', 'righted', tran, ['6A'] ).
verb( 'rile', 'riles', 'riling', 'riled', 'riled', tran, ['6A'] ).
verb( 'rim', 'rims', 'rimming', 'rimmed', 'rimmed', tran, ['6A'] ).
verb( 'rime', 'rimes', 'riming', 'rimed', 'rimed', tran, ['6A'] ).
verb( 'ring', 'rings', 'ringing', 'ringed', 'ringed', _, ['2A','2B','2C','2D','3A','6A','15A','15B'] ).
verb( 'ring', 'rings', 'ringing', 'rang', 'rung', _, ['2A','2B','2C','2D','3A','6A','15A','15B'] ).
verb( 'rinse', 'rinses', 'rinsing', 'rinsed', 'rinsed', tran, ['6A','15A','15B'] ).
verb( 'riot', 'riots', 'rioting', 'rioted', 'rioted', intran, ['2A','2B','2C','3A'] ).
verb( 'rip', 'rips', 'ripping', 'ripped', 'ripped', _, ['2A','2C','6A','15A','15B','22'] ).
verb( 'ripen', 'ripens', 'ripening', 'ripened', 'ripened', _, ['2A','6A'] ).
verb( 'riposte', 'ripostes', 'riposting', 'riposted', 'riposted', intran, [] ).
verb( 'ripple', 'ripples', 'rippling', 'rippled', 'rippled', _, ['2A','6A'] ).
verb( 'rise', 'rises', 'rising', 'rose', 'risen', intran, ['2A','2B','2C','3A','4A'] ).
verb( 'risk', 'risks', 'risking', 'risked', 'risked', tran, ['6A','6D'] ).
verb( 'rival', 'rivals', 'rivalling', 'rivalled', 'rivalled', tran, ['6A','14'] ).
verb( 'rive', 'rives', 'riving', 'rived', 'rived', _, ['6A','15A','15B'] ).
verb( 'rivet', 'rivets', 'riveting', 'riveted', 'riveted', tran, ['6A','15A','15B'] ).
verb( 'roam', 'roams', 'roaming', 'roamed', 'roamed', _, ['2A','2C','6A'] ).
verb( 'roar', 'roars', 'roaring', 'roared', 'roared', _, ['2A','2C','3A','6A','15B','22'] ).
verb( 'roast', 'roasts', 'roasting', 'roasted', 'roasted', _, ['2A','6A'] ).
verb( 'rob', 'robs', 'robbing', 'robbed', 'robbed', tran, ['6A','14'] ).
verb( 'robe', 'robes', 'robing', 'robed', 'robed', _, ['2A','6A','14'] ).
verb( 'rock', 'rocks', 'rocking', 'rocked', 'rocked', _, ['2A','6A','15A'] ).
verb( 'rocket', 'rockets', 'rocketing', 'rocketed', 'rocketed', intran, ['2A'] ).
verb( 'roll', 'rolls', 'rolling', 'rolled', 'rolled', _, ['2A','2B','2C','6A','12B','13B','14','15A','15B','22'] ).
verb( 'romance', 'romances', 'romancing', 'romanced', 'romanced', intran, ['2A'] ).
verb( 'romanticize', 'romanticizes', 'romanticizing', 'romanticized', 'romanticized', _, ['2A','6A'] ).
verb( 'romp', 'romps', 'romping', 'romped', 'romped', intran, ['2A','2C'] ).
verb( 'roof', 'roofs', 'roofing', 'roofed', 'roofed', tran, ['6A','15A','15B'] ).
verb( 'rook', 'rooks', 'rooking', 'rooked', 'rooked', tran, ['6A'] ).
verb( 'room', 'rooms', 'rooming', 'roomed', 'roomed', intran, ['2C'] ).
verb( 'roost', 'roosts', 'roosting', 'roosted', 'roosted', intran, ['2A'] ).
verb( 'root', 'roots', 'rooting', 'rooted', 'rooted', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'rootle', 'rootles', 'rootling', 'rootled', 'rootled', intran, ['2C'] ).
verb( 'rope', 'ropes', 'roping', 'roped', 'roped', tran, ['15A','15B'] ).
verb( 'rosin', 'rosins', 'rosinning', 'rosinned', 'rosinned', tran, ['6A'] ).
verb( 'rot', 'rots', 'rotting', 'rotted', 'rotted', _, ['2A','2C','6A'] ).
verb( 'rotate', 'rotates', 'rotating', 'rotated', 'rotated', _, ['2A','6A'] ).
verb( 'rouge', 'rouges', 'rouging', 'rouged', 'rouged', _, ['2A','6A'] ).
verb( 'rough', 'roughs', 'roughing', 'roughed', 'roughed', tran, ['6A','15B'] ).
verb( 'rough-dry', 'rough-dries', 'rough-drying', 'rough-dried', 'rough-dried', tran, [] ).
verb( 'rough-house', 'rough-houses', 'rough-housing', 'rough-housed', 'rough-housed', _, ['2A','6A'] ).
verb( 'roughcast', 'roughcasts', 'roughcasting', 'roughcast', 'roughcast', tran, [] ).
verb( 'roughen', 'roughens', 'roughening', 'roughened', 'roughened', _, [] ).
verb( 'round', 'rounds', 'rounding', 'rounded', 'rounded', _, ['2C','3A','6A','15B'] ).
verb( 'rouse', 'rouses', 'rousing', 'roused', 'roused', _, ['2A','6A','15A'] ).
verb( 'rout', 'routs', 'routing', 'routed', 'routed', tran, ['6A','15B'] ).
verb( 'route', 'routes', 'routing', 'routed', 'routed', tran, [] ).
verb( 'rove', 'roves', 'roving', 'roved', 'roved', _, ['2A','2C','6A'] ).
verb( 'row', 'rows', 'rowing', 'rowed', 'rowed', _, ['2A','2B','2C','6A','15A','15B'] ).
verb( 'row', 'rows', 'rowing', 'rowed', 'rowed', _, ['2A','3A','6A'] ).
verb( 'rub', 'rubs', 'rubbing', 'rubbed', 'rubbed', _, ['2A','2C','6A','15A','15B','22'] ).
verb( 'rubber', 'rubbers', 'rubbering', 'rubbered', 'rubbered', tran, ['6A'] ).
verb( 'rubber-stamp', 'rubber-stamps', 'rubber-stamping', 'rubber-stamped', 'rubber-stamped', tran, [] ).
verb( 'rubberize', 'rubberizes', 'rubberizing', 'rubberized', 'rubberized', tran, ['6A'] ).
verb( 'rubberneck', 'rubbernecks', 'rubbernecking', 'rubbernecked', 'rubbernecked', intran, [] ).
verb( 'rubbish', 'rubbishes', 'rubbishing', 'rubbished', 'rubbished', tran, ['6A'] ).
verb( 'ruck', 'rucks', 'rucking', 'rucked', 'rucked', _, ['2A','2C','6A','15B'] ).
verb( 'ruddle', 'ruddles', 'ruddling', 'ruddled', 'ruddled', tran, [] ).
verb( 'rue', 'rues', 'ruing', 'rued', 'rued', tran, ['6A'] ).
verb( 'ruff', 'ruffs', 'ruffing', 'ruffed', 'ruffed', _, [] ).
verb( 'ruffle', 'ruffles', 'ruffling', 'ruffled', 'ruffled', _, ['2A','6A','15B'] ).
verb( 'ruin', 'ruins', 'ruining', 'ruined', 'ruined', tran, ['6A'] ).
verb( 'rule', 'rules', 'ruling', 'ruled', 'ruled', _, ['2A','2B','2C','3A','6A','9','15A','15B','25'] ).
verb( 'rumble', 'rumbles', 'rumbling', 'rumbled', 'rumbled', _, ['2A','2C','15B'] ).
verb( 'ruminate', 'ruminates', 'ruminating', 'ruminated', 'ruminated', intran, ['2A','2B','2C'] ).
verb( 'rummage', 'rummages', 'rummaging', 'rummaged', 'rummaged', _, ['2A','2B','2C','3A','6A','15B'] ).
verb( 'rumour', 'rumours', 'rumouring', 'rumoured', 'rumoured', tran, [] ).
verb( 'rumple', 'rumples', 'rumpling', 'rumpled', 'rumpled', tran, ['6A'] ).
verb( 'run', 'runs', 'running', 'ran', 'run', _, ['2A','2B','2C','2D','2E','3A','4A','6A','14','15A','15B'] ).
verb( 'rupture', 'ruptures', 'rupturing', 'ruptured', 'ruptured', _, ['2A','6A'] ).
verb( 'rush', 'rushes', 'rushing', 'rushed', 'rushed', _, ['2A','2C','4A','6A','15A','15B'] ).
verb( 'rust', 'rusts', 'rusting', 'rusted', 'rusted', _, ['2A','2C','6A'] ).
verb( 'rusticate', 'rusticates', 'rusticating', 'rusticated', 'rusticated', _, ['2A','6A'] ).
verb( 'rustle', 'rustles', 'rustling', 'rustled', 'rustled', _, ['2A','2C','6A','15B'] ).
verb( 'rut', 'ruts', 'rutting', 'rutted', 'rutted', tran, ['2A'] ).
verb( 'sabotage', 'sabotages', 'sabotaging', 'sabotaged', 'sabotaged', tran, ['6A'] ).
verb( 'sabre', 'sabres', 'sabring', 'sabred', 'sabred', tran, [] ).
verb( 'sack', 'sacks', 'sacking', 'sacked', 'sacked', tran, ['6A'] ).
verb( 'sacrifice', 'sacrifices', 'sacrificing', 'sacrificed', 'sacrificed', _, ['3A','6A','14','16A'] ).
verb( 'sadden', 'saddens', 'saddening', 'saddened', 'saddened', _, ['2A','6A'] ).
verb( 'saddle', 'saddles', 'saddling', 'saddled', 'saddled', tran, ['6A','14'] ).
verb( 'safeguard', 'safeguards', 'safeguarding', 'safeguarded', 'safeguarded', tran, ['6A','14'] ).
verb( 'sag', 'sags', 'sagging', 'sagged', 'sagged', intran, ['2A'] ).
verb( 'sail', 'sails', 'sailing', 'sailed', 'sailed', _, ['2A','2B','2C','3A','6A'] ).
verb( 'salaam', 'salaams', 'salaaming', 'salaamed', 'salaamed', intran, [] ).
verb( 'salivate', 'salivates', 'salivating', 'salivated', 'salivated', intran, [] ).
verb( 'sallow', 'sallows', 'sallowing', 'sallowed', 'sallowed', _, [] ).
verb( 'sally', 'sallies', 'sallying', 'sallied', 'sallied', intran, ['2A','2C'] ).
verb( 'salt', 'salts', 'salting', 'salted', 'salted', tran, ['6A','15B'] ).
verb( 'salute', 'salutes', 'saluting', 'saluted', 'saluted', _, ['2A','6A'] ).
verb( 'salvage', 'salvages', 'salvaging', 'salvaged', 'salvaged', tran, ['6A'] ).
verb( 'salve', 'salves', 'salving', 'salved', 'salved', tran, ['6A'] ).
verb( 'sample', 'samples', 'sampling', 'sampled', 'sampled', tran, ['6A'] ).
verb( 'sanctify', 'sanctifies', 'sanctifying', 'sanctified', 'sanctified', tran, ['6A'] ).
verb( 'sanction', 'sanctions', 'sanctioning', 'sanctioned', 'sanctioned', tran, ['6A'] ).
verb( 'sand', 'sands', 'sanding', 'sanded', 'sanded', tran, ['6A'] ).
verb( 'sandblast', 'sandblasts', 'sandblasting', 'sandblasted', 'sandblasted', tran, ['6A'] ).
verb( 'sandpaper', 'sandpapers', 'sandpapering', 'sandpapered', 'sandpapered', tran, ['6A'] ).
verb( 'sandwich', 'sandwiches', 'sandwiching', 'sandwiched', 'sandwiched', tran, ['6A','14'] ).
verb( 'sap', 'saps', 'sapping', 'sapped', 'sapped', _, ['2A','6A'] ).
verb( 'sate', 'sates', 'sating', 'sated', 'sated', tran, ['6A'] ).
verb( 'satiate', 'satiates', 'satiating', 'satiated', 'satiated', tran, ['6A'] ).
verb( 'satirize', 'satirizes', 'satirizing', 'satirized', 'satirized', tran, ['6A'] ).
verb( 'satisfy', 'satisfies', 'satisfying', 'satisfied', 'satisfied', _, ['2A','6A','11','14'] ).
verb( 'saturate', 'saturates', 'saturating', 'saturated', 'saturated', tran, ['6A','14'] ).
verb( 'sauce', 'sauces', 'saucing', 'sauced', 'sauced', tran, ['6A'] ).
verb( 'saunter', 'saunters', 'sauntering', 'sauntered', 'sauntered', intran, ['2A','2C'] ).
verb( 'saut_e', 'saut_es', 'saut_eing', 'saut_eed', 'saut_eed', tran, [] ).
verb( 'savage', 'savages', 'savaging', 'savaged', 'savaged', tran, ['6A'] ).
verb( 'save', 'saves', 'saving', 'saved', 'saved', _, ['2A','2C','3A','6A','6D','12B','12C','13B','14','15B'] ).
verb( 'savour', 'savours', 'savouring', 'savoured', 'savoured', _, ['3A','6A'] ).
verb( 'savvy', 'savvies', 'savvying', 'savvyed', 'savvyed', intran, [] ).
verb( 'saw', 'saws', 'sawing', 'sawed', 'sawed', _, ['2A','2B','2C','4A','6A','8','9','10','15A','15B','16B','18A','19A','22','24A'] ).
verb( 'say', 'says', 'saying', 'said', 'said', _, ['6A','9','10','14'] ).
verb( 'scald', 'scalds', 'scalding', 'scalded', 'scalded', tran, ['6A'] ).
verb( 'scale', 'scales', 'scaling', 'scaled', 'scaled', _, ['2B','2C','6A','15A','15B'] ).
verb( 'scallop', 'scallops', 'scalloping', 'scalloped', 'scalloped', tran, ['6A'] ).
verb( 'scalp', 'scalps', 'scalping', 'scalped', 'scalped', tran, ['6A'] ).
verb( 'scamp', 'scamps', 'scamping', 'scamped', 'scamped', tran, ['6A'] ).
verb( 'scamper', 'scampers', 'scampering', 'scampered', 'scampered', intran, ['2A','2C'] ).
verb( 'scan', 'scans', 'scanning', 'scanned', 'scanned', _, ['2A','6A'] ).
verb( 'scandalize', 'scandalizes', 'scandalizing', 'scandalized', 'scandalized', tran, ['6A'] ).
verb( 'scant', 'scants', 'scanting', 'scanted', 'scanted', tran, ['6A'] ).
verb( 'scar', 'scars', 'scarring', 'scarred', 'scarred', _, ['2C','6A'] ).
verb( 'scare', 'scares', 'scaring', 'scared', 'scared', _, ['2A','6A','15A','15B'] ).
verb( 'scarify', 'scarifies', 'scarifying', 'scarified', 'scarified', tran, ['6A'] ).
verb( 'scarper', 'scarpers', 'scarpering', 'scarpered', 'scarpered', intran, ['2A'] ).
verb( 'scatter', 'scatters', 'scattering', 'scattered', 'scattered', _, ['2A','2C','6A','15A','15B'] ).
verb( 'scavenge', 'scavenges', 'scavenging', 'scavenged', 'scavenged', intran, ['2A','3A'] ).
verb( 'scent', 'scents', 'scenting', 'scented', 'scented', tran, ['6A'] ).
verb( 'schedule', 'schedules', 'scheduling', 'scheduled', 'scheduled', tran, ['6A','7A','14'] ).
verb( 'scheme', 'schemes', 'scheming', 'schemed', 'schemed', _, ['2A','3A','4A','6A'] ).
verb( 'school', 'schools', 'schooling', 'schooled', 'schooled', tran, ['6A','15A'] ).
verb( 'scintillate', 'scintillates', 'scintillating', 'scintillated', 'scintillated', intran, ['2A'] ).
verb( 'scoff', 'scoffs', 'scoffing', 'scoffed', 'scoffed', _, ['2A','3A'] ).
verb( 'scold', 'scolds', 'scolding', 'scolded', 'scolded', _, ['2A','6A','14'] ).
verb( 'scollop', 'scollops', 'scolloping', 'scolloped', 'scolloped', tran, [] ).
verb( 'scoop', 'scoops', 'scooping', 'scooped', 'scooped', tran, ['6A','15B'] ).
verb( 'scoot', 'scoots', 'scooting', 'scooted', 'scooted', intran, [] ).
verb( 'scorch', 'scorches', 'scorching', 'scorched', 'scorched', _, ['2A','2C','6A'] ).
verb( 'score', 'scores', 'scoring', 'scored', 'scored', _, ['2A','3A','6A','15A','15B'] ).
verb( 'scorn', 'scorns', 'scorning', 'scorned', 'scorned', tran, ['6A','6D','7A'] ).
verb( 'scotch', 'scotches', 'scotching', 'scotched', 'scotched', tran, ['6A'] ).
verb( 'scour', 'scours', 'scouring', 'scoured', 'scoured', _, ['2C','6A','15A','15B'] ).
verb( 'scourge', 'scourges', 'scourging', 'scourged', 'scourged', tran, ['6A'] ).
verb( 'scout', 'scouts', 'scouting', 'scouted', 'scouted', _, ['2C','6A'] ).
verb( 'scowl', 'scowls', 'scowling', 'scowled', 'scowled', intran, ['2A','3A'] ).
verb( 'scrabble', 'scrabbles', 'scrabbling', 'scrabbled', 'scrabbled', intran, ['2A','2C'] ).
verb( 'scrag', 'scrags', 'scragging', 'scragged', 'scragged', tran, [] ).
verb( 'scram', 'scrams', 'scramming', 'scrammed', 'scrammed', intran, [] ).
verb( 'scramble', 'scrambles', 'scrambling', 'scrambled', 'scrambled', _, ['2A','2C','3A','4A','6A'] ).
verb( 'scrap', 'scraps', 'scrapping', 'scrapped', 'scrapped', tran, ['6A'] ).
verb( 'scrape', 'scrapes', 'scraping', 'scraped', 'scraped', _, ['2C','3A','6A','14','15A','15B','22'] ).
verb( 'scratch', 'scratches', 'scratching', 'scratched', 'scratched', _, ['2A','2C','6A','15B'] ).
verb( 'scrawl', 'scrawls', 'scrawling', 'scrawled', 'scrawled', _, ['2A','2C','6A'] ).
verb( 'scream', 'screams', 'screaming', 'screamed', 'screamed', _, ['2A','2C','6A','15A','15B','22'] ).
verb( 'screech', 'screeches', 'screeching', 'screeched', 'screeched', _, ['2A','2C','6A','15A','15B'] ).
verb( 'screen', 'screens', 'screening', 'screened', 'screened', _, ['2C','6A','14','15A','15B'] ).
verb( 'screw', 'screws', 'screwing', 'screwed', 'screwed', _, ['2A','6A','15A','15B'] ).
verb( 'scribble', 'scribbles', 'scribbling', 'scribbled', 'scribbled', _, ['2A','6A'] ).
verb( 'scrimmage', 'scrimmages', 'scrimmaging', 'scrimmaged', 'scrimmaged', _, [] ).
verb( 'scrimp', 'scrimps', 'scrimping', 'scrimped', 'scrimped', _, [] ).
verb( 'scrimshank', 'scrimshanks', 'scrimshanking', 'scrimshanked', 'scrimshanked', intran, [] ).
verb( 'scrounge', 'scrounges', 'scrounging', 'scrounged', 'scrounged', _, ['2A','6A'] ).
verb( 'scrub', 'scrubs', 'scrubbing', 'scrubbed', 'scrubbed', _, ['2A','2C','6A','15B','22'] ).
verb( 'scrunch', 'scrunches', 'scrunching', 'scrunched', 'scrunched', tran, [] ).
verb( 'scruple', 'scruples', 'scrupling', 'scrupled', 'scrupled', intran, ['4C'] ).
verb( 'scrutinize', 'scrutinizes', 'scrutinizing', 'scrutinized', 'scrutinized', tran, ['6A'] ).
verb( 'scud', 'scuds', 'scudding', 'scudded', 'scudded', intran, ['2A','2C'] ).
verb( 'scuff', 'scuffs', 'scuffing', 'scuffed', 'scuffed', _, ['2A','2C','6A','15A','15B'] ).
verb( 'scuffle', 'scuffles', 'scuffling', 'scuffled', 'scuffled', intran, ['2A','2C'] ).
verb( 'scull', 'sculls', 'sculling', 'sculled', 'sculled', _, ['2A','6A'] ).
verb( 'sculpt', 'sculpts', 'sculpting', 'sculpted', 'sculpted', _, [] ).
verb( 'sculpture', 'sculptures', 'sculpturing', 'sculptured', 'sculptured', _, ['2A','6A'] ).
verb( 'scupper', 'scuppers', 'scuppering', 'scuppered', 'scuppered', tran, [] ).
verb( 'scurry', 'scurries', 'scurrying', 'scurried', 'scurried', intran, ['2A','2C','3A'] ).
verb( 'scuttle', 'scuttles', 'scuttling', 'scuttled', 'scuttled', _, ['2A','2C','6A'] ).
verb( 'scythe', 'scythes', 'scything', 'scythed', 'scythed', tran, ['6A','15A'] ).
verb( 'seal', 'seals', 'sealing', 'sealed', 'sealed', _, ['2A','6A','15A','15B'] ).
verb( 'seam', 'seams', 'seaming', 'seamed', 'seamed', tran, [] ).
verb( 'sear', 'sears', 'searing', 'seared', 'seared', tran, ['6A'] ).
verb( 'search', 'searches', 'searching', 'searched', 'searched', _, ['2A','3A','6A','14','15A','15B'] ).
verb( 'season', 'seasons', 'seasoning', 'seasoned', 'seasoned', _, ['2A','6A','14'] ).
verb( 'seat', 'seats', 'seating', 'seated', 'seated', tran, ['6A'] ).
verb( 'secede', 'secedes', 'seceding', 'seceded', 'seceded', intran, ['2A','3A'] ).
verb( 'seclude', 'secludes', 'secluding', 'secluded', 'secluded', tran, ['6A','14','15A'] ).
verb( 'second', 'seconds', 'seconding', 'seconded', 'seconded', tran, ['6A'] ).
verb( 'second', 'seconds', 'seconding', 'seconded', 'seconded', tran, ['6A','15A'] ).
verb( 'secrete', 'secretes', 'secreting', 'secreted', 'secreted', tran, ['6A'] ).
verb( 'secularize', 'secularizes', 'secularizing', 'secularized', 'secularized', tran, ['6A'] ).
verb( 'secure', 'secures', 'securing', 'secured', 'secured', tran, ['6A','12B','13B','14'] ).
verb( 'sedate', 'sedates', 'sedating', 'sedated', 'sedated', tran, ['6A'] ).
verb( 'seduce', 'seduces', 'seducing', 'seduced', 'seduced', tran, ['6A','14'] ).
verb( 'see', 'sees', 'seeing', 'saw', 'seen', _, ['2A','2B','2C','4A','6A','8','9','10','15A','16B','18A','19A','22','24A'] ).
verb( 'seed', 'seeds', 'seeding', 'seeded', 'seeded', _, ['2A','6A'] ).
verb( 'seek', 'seeks', 'seeking', 'sought', 'sought', tran, ['2A','3A','6A','7A','15A'] ).
verb( 'seem', 'seems', 'seeming', 'seemed', 'seemed', intran, ['2A','4D','4E'] ).
verb( 'seep', 'seeps', 'seeping', 'seeped', 'seeped', intran, ['2C'] ).
verb( 'seesaw', 'seesaws', 'seesawing', 'seesawed', 'seesawed', intran, [] ).
verb( 'seethe', 'seethes', 'seething', 'seethed', 'seethed', _, ['2A','3A','6A'] ).
verb( 'segment', 'segments', 'segmenting', 'segmented', 'segmented', _, [] ).
verb( 'segregate', 'segregates', 'segregating', 'segregated', 'segregated', tran, ['6A'] ).
verb( 'seine', 'seines', 'seining', 'seined', 'seined', _, [] ).
verb( 'seize', 'seizes', 'seizing', 'seized', 'seized', _, ['2A','2C','3A','6A','15A'] ).
verb( 'select', 'selects', 'selecting', 'selected', 'selected', tran, ['6A','15A','16A'] ).
verb( 'sell', 'sells', 'selling', 'sold', 'sold', _, ['2A','2C','6A','12A','13A','15B'] ).
verb( 'semaphore', 'semaphores', 'semaphoring', 'semaphored', 'semaphored', _, ['2A','6A'] ).
verb( 'send', 'sends', 'sending', 'sent', 'sent', _, ['2C','3A','6A','12A','13A','15A','15B','19B','22'] ).
verb( 'sense', 'senses', 'sensing', 'sensed', 'sensed', tran, ['6A','9'] ).
verb( 'sensitize', 'sensitizes', 'sensitizing', 'sensitized', 'sensitized', tran, ['6A'] ).
verb( 'sentence', 'sentences', 'sentencing', 'sentenced', 'sentenced', tran, ['6A','14','17'] ).
verb( 'sentimentalize', 'sentimentalizes', 'sentimentalizing', 'sentimentalized', 'sentimentalized', _, ['2A','6A'] ).
verb( 'separate', 'separates', 'separating', 'separated', 'separated', _, ['2A','6A','14','15B'] ).
verb( 'sequester', 'sequesters', 'sequestering', 'sequestered', 'sequestered', tran, ['6A'] ).
verb( 'sequestrate', 'sequestrates', 'sequestrating', 'sequestrated', 'sequestrated', tran, ['6A'] ).
verb( 'serenade', 'serenades', 'serenading', 'serenaded', 'serenaded', tran, ['6A'] ).
verb( 'serialize', 'serializes', 'serializing', 'serialized', 'serialized', tran, ['6A'] ).
verb( 'sermonize', 'sermonizes', 'sermonizing', 'sermonized', 'sermonized', _, ['2A','6A'] ).
verb( 'serve', 'serves', 'serving', 'served', 'served', _, ['2A','2B','2C','3A','4A','6A','14','15A','15B'] ).
verb( 'service', 'services', 'servicing', 'serviced', 'serviced', tran, ['6A'] ).
verb( 'set', 'sets', 'setting', 'set', 'set', _, ['2A','2C','3A','6A','12C','14','15A','16A','17','19B','22'] ).
verb( 'settle', 'settles', 'settling', 'settled', 'settled', _, ['2A','2C','3A','6A','7A','8','10','14','15A','15B'] ).
verb( 'sever', 'severs', 'severing', 'severed', 'severed', _, ['2A','6A','15A'] ).
verb( 'sew', 'sews', 'sewing', 'sewed', 'sewn', _, ['2A','2B','2C','6A','15B'] ).
verb( 'sex', 'sexes', 'sexing', 'sexed', 'sexed', tran, ['6A'] ).
verb( 'shack', 'shacks', 'shacking', 'shacked', 'shacked', intran, ['2C'] ).
verb( 'shackle', 'shackles', 'shackling', 'shackled', 'shackled', tran, ['6A'] ).
verb( 'shade', 'shades', 'shading', 'shaded', 'shaded', _, ['2C','6A','15A'] ).
verb( 'shadow', 'shadows', 'shadowing', 'shadowed', 'shadowed', tran, ['6A'] ).
verb( 'shag', 'shags', 'shagging', 'shagged', 'shagged', _, ['6A'] ).
verb( 'shake', 'shakes', 'shaking', 'shook', 'shaken', _, ['2A','2C','6A','14','15A','15B'] ).
verb( 'shall', 'shall', '-', '-', '-', unknown, [] ).
verb( 'shallow', 'shallows', 'shallowing', 'shallowed', 'shallowed', intran, [] ).
verb( 'sham', 'shams', 'shamming', 'shammed', 'shammed', _, ['2A','2D','6A'] ).
verb( 'shamble', 'shambles', 'shambling', 'shambled', 'shambled', intran, ['2A','2C'] ).
verb( 'shame', 'shames', 'shaming', 'shamed', 'shamed', tran, ['6A','14'] ).
verb( 'shampoo', 'shampoos', 'shampooing', 'shampooed', 'shampooed', tran, ['6A'] ).
verb( 'shanghai', 'shanghais', 'shanghaiing', 'shanghaied', 'shanghaied', tran, [] ).
verb( 'shape', 'shapes', 'shaping', 'shaped', 'shaped', _, ['2A','2C','6A','15A'] ).
verb( 'share', 'shares', 'sharing', 'shared', 'shared', _, ['3A','6A','14','15B'] ).
verb( 'sharpen', 'sharpens', 'sharpening', 'sharpened', 'sharpened', _, ['2A','6A'] ).
verb( 'shatter', 'shatters', 'shattering', 'shattered', 'shattered', _, ['2A','6A'] ).
verb( 'shave', 'shaves', 'shaving', 'shaved', 'shaved', _, ['2A','6A','15A','15B'] ).
verb( 'shear', 'shears', 'shearing', 'sheared', 'sheared', tran, ['6A'] ).
verb( 'sheathe', 'sheathes', 'sheathing', 'sheathed', 'sheathed', tran, ['6A'] ).
verb( 'shed', 'sheds', 'shedding', 'shed', 'shed', tran, ['6A'] ).
verb( 'sheer', 'sheers', 'sheering', 'sheered', 'sheered', intran, ['2C'] ).
verb( 'shell', 'shells', 'shelling', 'shelled', 'shelled', _, ['2C','6A','15B'] ).
verb( 'shellac', 'shellacs', 'shellacking', 'shellacked', 'shellacked', tran, [] ).
verb( 'shelter', 'shelters', 'sheltering', 'sheltered', 'sheltered', _, ['2A','2C','6A','14'] ).
verb( 'shelve', 'shelves', 'shelving', 'shelved', 'shelved', _, ['2A','2C','6A'] ).
verb( 'shepherd', 'shepherds', 'shepherding', 'shepherded', 'shepherded', tran, ['6A','15A'] ).
verb( 'shew', 'shews', 'shewing', 'shewed', 'shewed', _, ['2A','2C','6A','9','10','12A','13A','14','15A','15B','19B','20','21','24A','25'] ).
verb( 'shield', 'shields', 'shielding', 'shielded', 'shielded', tran, ['6A','15A'] ).
verb( 'shift', 'shifts', 'shifting', 'shifted', 'shifted', _, ['2A','2C','6A','14','15A'] ).
verb( 'shillyshally', 'shillyshallies', 'shillyshallying', 'shillyshallied', 'shillyshallied', intran, ['2A'] ).
verb( 'shimmer', 'shimmers', 'shimmering', 'shimmered', 'shimmered', intran, ['2A','2C'] ).
verb( 'shin', 'shins', 'shinning', 'shinned', 'shinned', intran, ['3A'] ).
verb( 'shine', 'shines', 'shining', 'shone', 'shone', _, ['2A','2C','6A'] ).
verb( 'shingle', 'shingles', 'shingling', 'shingled', 'shingled', tran, ['6A'] ).
verb( 'ship', 'ships', 'shipping', 'shipped', 'shipped', _, ['2C','6A','15A','15B'] ).
verb( 'shipwreck', 'shipwrecks', 'shipwrecking', 'shipwrecked', 'shipwrecked', intran, [] ).
verb( 'shirk', 'shirks', 'shirking', 'shirked', 'shirked', _, ['2A','6A','6D'] ).
verb( 'shit', 'shits', 'shitting', 'shitted', 'shitted', intran, ['2A','3A'] ).
verb( 'shiver', 'shivers', 'shivering', 'shivered', 'shivered', _, ['2A','2C','15A'] ).
verb( 'shoal', 'shoals', 'shoaling', 'shoaled', 'shoaled', intran, ['2A'] ).
verb( 'shock', 'shocks', 'shocking', 'shocked', 'shocked', tran, ['6A'] ).
verb( 'shoe', 'shoes', 'shoeing', 'shod', 'shod', tran, ['6A'] ).
verb( 'shoo', 'shoos', 'shooing', 'shooed', 'shooed', _, ['15B'] ).
verb( 'shoot', 'shoots', 'shooting', 'shot', 'shot', _, ['2A','2C','4A','6A','15A','15B'] ).
verb( 'shop', 'shops', 'shopping', 'shopped', 'shopped', intran, ['2A','2C'] ).
verb( 'shoplift', 'shoplifts', 'shoplifting', 'shoplifted', 'shoplifted', _, ['2A','6A'] ).
verb( 'shore', 'shores', 'shoring', 'shored', 'shored', tran, ['15B'] ).
verb( 'short', 'shorts', 'shorting', 'shorted', 'shorted', _, [] ).
verb( 'short-change', 'short-changes', 'short-changing', 'short-changed', 'short-changed', tran, [] ).
verb( 'short-circuit', 'short-circuits', 'short-circuiting', 'short-circuited', 'short-circuited', _, [] ).
verb( 'shorten', 'shortens', 'shortening', 'shortened', 'shortened', _, ['2A','6A'] ).
verb( 'shortlist', 'shortlists', 'shortlisting', 'shortlisted', 'shortlisted', tran, [] ).
verb( 'should', 'should', '-', '-', '-', unknown, [] ).
verb( 'shoulder', 'shoulders', 'shouldering', 'shouldered', 'shouldered', tran, ['6A','15A'] ).
verb( 'shout', 'shouts', 'shouting', 'shouted', 'shouted', _, ['2A','2B','2C','3A','4B','6A','15A','15B','22'] ).
verb( 'shove', 'shoves', 'shoving', 'shoved', 'shoved', _, ['2A','2C','6A','15A','15B'] ).
verb( 'shovel', 'shovels', 'shovelling', 'shovelled', 'shovelled', tran, ['6A','15A','15B'] ).
verb( 'show', 'shows', 'showing', 'showed', 'showed', _, ['2A','2C','6A','9','10','12A','13A','14','15A','15B','19B','20','21','24A','25'] ).
verb( 'shower', 'showers', 'showering', 'showered', 'showered', _, ['2C','14'] ).
verb( 'shred', 'shreds', 'shredding', 'shredded', 'shredded', tran, ['6A'] ).
verb( 'shriek', 'shrieks', 'shrieking', 'shrieked', 'shrieked', _, ['2A','2C','6A','15B'] ).
verb( 'shrill', 'shrills', 'shrilling', 'shrilled', 'shrilled', _, [] ).
verb( 'shrimp', 'shrimps', 'shrimping', 'shrimped', 'shrimped', intran, [] ).
verb( 'shrine', 'shrines', 'shrining', 'shrined', 'shrined', tran, [] ).
verb( 'shrink', 'shrinks', 'shrinking', 'shrank', 'shrunk', _, ['2A','2C','3A','6A'] ).
verb( 'shrive', 'shrives', 'shriving', 'shrived', 'shrived', tran, ['6A'] ).
verb( 'shrivel', 'shrivels', 'shrivelling', 'shrivelled', 'shrivelled', _, ['2A','2C','6A','15B'] ).
verb( 'shroud', 'shrouds', 'shrouding', 'shrouded', 'shrouded', tran, ['6A','15A'] ).
verb( 'shrug', 'shrugs', 'shrugging', 'shrugged', 'shrugged', tran, ['6A','15B'] ).
verb( 'shuck', 'shucks', 'shucking', 'shucked', 'shucked', tran, ['6A'] ).
verb( 'shudder', 'shudders', 'shuddering', 'shuddered', 'shuddered', intran, ['2A','2C','4C'] ).
verb( 'shuffle', 'shuffles', 'shuffling', 'shuffled', 'shuffled', _, ['2A','2C','6A','15A','15B'] ).
verb( 'shun', 'shuns', 'shunning', 'shunned', 'shunned', tran, ['6A','6D'] ).
verb( 'shunt', 'shunts', 'shunting', 'shunted', 'shunted', _, ['2A','6A','15A','15B'] ).
verb( 'shush', 'shushes', 'shushing', 'shushed', 'shushed', _, [] ).
verb( 'shut', 'shuts', 'shutting', 'shut', 'shut', _, ['2A','2C','6A','15A','15B'] ).
verb( 'shutter', 'shutters', 'shuttering', 'shuttered', 'shuttered', tran, ['6A'] ).
verb( 'shuttle', 'shuttles', 'shuttling', 'shuttled', 'shuttled', _, ['2C','6A','15A'] ).
verb( 'shy', 'shies', 'shying', 'shied', 'shied', _, ['2A','3A','6A','15A'] ).
verb( 'sick', 'sicks', 'sicking', 'sicked', 'sicked', tran, [] ).
verb( 'sicken', 'sickens', 'sickening', 'sickened', 'sickened', _, ['2A','3A','4C','6A'] ).
verb( 'side', 'sides', 'siding', 'sided', 'sided', intran, ['3A'] ).
verb( 'side-slip', 'side-slips', 'side-slipping', 'side-slipped', 'side-slipped', intran, [] ).
verb( 'sidestep', 'sidesteps', 'sidestepping', 'sidestepped', 'sidestepped', _, ['2A','6A'] ).
verb( 'sidetrack', 'sidetracks', 'sidetracking', 'sidetracked', 'sidetracked', tran, ['6A'] ).
verb( 'sidle', 'sidles', 'sidling', 'sidled', 'sidled', intran, ['2C'] ).
verb( 'sieve', 'sieves', 'sieving', 'sieved', 'sieved', tran, ['6A'] ).
verb( 'sift', 'sifts', 'sifting', 'sifted', 'sifted', _, ['2C','6A','14','15A','15B'] ).
verb( 'sigh', 'sighs', 'sighing', 'sighed', 'sighed', _, ['2A','3A','6A','15B'] ).
verb( 'sight', 'sights', 'sighting', 'sighted', 'sighted', tran, ['6A'] ).
verb( 'sign', 'signs', 'signing', 'signed', 'signed', _, ['2A','2C','3A','6A','15B'] ).
verb( 'signal', 'signals', 'signalling', 'signalled', 'signalled', _, ['2A','2C','6A','9','17'] ).
verb( 'signalize', 'signalizes', 'signalizing', 'signalized', 'signalized', tran, ['6A'] ).
verb( 'signify', 'signifies', 'signifying', 'signified', 'signified', _, ['2A','2C','6A','9','15A'] ).
verb( 'signpost', 'signposts', 'signposting', 'signposted', 'signposted', tran, [] ).
verb( 'silence', 'silences', 'silencing', 'silenced', 'silenced', tran, ['6A'] ).
verb( 'silhouette', 'silhouettes', 'silhouetting', 'silhouetted', 'silhouetted', tran, [] ).
verb( 'silt', 'silts', 'silting', 'silted', 'silted', _, ['2C','15B'] ).
verb( 'silver', 'silvers', 'silvering', 'silvered', 'silvered', _, ['2A','6A'] ).
verb( 'simmer', 'simmers', 'simmering', 'simmered', 'simmered', _, ['2A','2B','2C','6A','15A'] ).
verb( 'simper', 'simpers', 'simpering', 'simpered', 'simpered', intran, ['2A'] ).
verb( 'simplify', 'simplifies', 'simplifying', 'simplified', 'simplified', tran, ['6A'] ).
verb( 'simulate', 'simulates', 'simulating', 'simulated', 'simulated', tran, ['6A'] ).
verb( 'sin', 'sins', 'sinning', 'sinned', 'sinned', intran, ['2A','3A'] ).
verb( 'sing', 'sings', 'singing', 'sang', 'sung', _, ['2A','2C','3A','6A','12B','13B','15A'] ).
verb( 'singe', 'singes', 'singeing', 'singed', 'singed', _, ['2A','6A'] ).
verb( 'single', 'singles', 'singling', 'singled', 'singled', tran, ['15B'] ).
verb( 'singularize', 'singularizes', 'singularizing', 'singularized', 'singularized', tran, [] ).
verb( 'sink', 'sinks', 'sinking', 'sank', 'sunk', _, ['2A','2C','3A','6A','14','15A'] ).
verb( 'sip', 'sips', 'sipping', 'sipped', 'sipped', _, ['2A','6A','15B'] ).
verb( 'siphon', 'siphons', 'siphoning', 'siphoned', 'siphoned', _, ['15B'] ).
verb( 'sire', 'sires', 'siring', 'sired', 'sired', tran, ['6A'] ).
verb( 'sit', 'sits', 'siting', 'sat', 'sat', _, ['2A','2C','3A','6A','15B'] ).
verb( 'site', 'sites', 'siting', 'sited', 'sited', intran, ['6A'] ).
verb( 'size', 'sizes', 'sizing', 'sized', 'sized', tran, ['6A','15B'] ).
verb( 'sizzle', 'sizzles', 'sizzling', 'sizzled', 'sizzled', intran, ['6A'] ).
verb( 'skate', 'skates', 'skating', 'skated', 'skated', intran, ['2A','2C'] ).
verb( 'skedaddle', 'skedaddles', 'skedaddling', 'skedaddled', 'skedaddled', intran, ['2A'] ).
verb( 'sketch', 'sketches', 'sketching', 'sketched', 'sketched', _, ['2A','6A','15B'] ).
verb( 'skewer', 'skewers', 'skewering', 'skewered', 'skewered', tran, ['6A'] ).
verb( 'ski', 'skis', 'skiing', 'ski\'d', 'ski\'d', intran, [] ).
verb( 'skid', 'skids', 'skidding', 'skidded', 'skidded', intran, ['2A'] ).
verb( 'skim', 'skims', 'skimming', 'skimmed', 'skimmed', _, ['2C','3A','6A','14','15A','15B'] ).
verb( 'skimp', 'skimps', 'skimping', 'skimped', 'skimped', _, ['2A','6A'] ).
verb( 'skin', 'skins', 'skinning', 'skinned', 'skinned', _, ['2C','6A','14'] ).
verb( 'skip', 'skips', 'skipping', 'skipped', 'skipped', _, ['2A','2C','6A'] ).
verb( 'skipper', 'skippers', 'skippering', 'skippered', 'skippered', tran, [] ).
verb( 'skirmish', 'skirmishes', 'skirmishing', 'skirmished', 'skirmished', intran, ['2A','2C'] ).
verb( 'skirt', 'skirts', 'skirting', 'skirted', 'skirted', _, ['2C','6A'] ).
verb( 'skitter', 'skitters', 'skittering', 'skittered', 'skittered', intran, ['2A'] ).
verb( 'skittle', 'skittles', 'skittling', 'skittled', 'skittled', tran, ['15B'] ).
verb( 'skulk', 'skulks', 'skulking', 'skulked', 'skulked', intran, ['2A','2C'] ).
verb( 'sky', 'skies', 'skying', 'skied', 'skied', tran, [] ).
verb( 'skylark', 'skylarks', 'skylarking', 'skylarked', 'skylarked', intran, [] ).
verb( 'skyrocket', 'skyrockets', 'skyrocketing', 'skyrocketed', 'skyrocketed', intran, [] ).
verb( 'slack', 'slacks', 'slacking', 'slacked', 'slacked', intran, ['2A','2C'] ).
verb( 'slacken', 'slackens', 'slackening', 'slackened', 'slackened', _, ['2A','6A'] ).
verb( 'slake', 'slakes', 'slaking', 'slaked', 'slaked', tran, ['6A'] ).
verb( 'slam', 'slams', 'slamming', 'slammed', 'slammed', _, ['2A','2C','6A','15A','15B','22'] ).
verb( 'slander', 'slanders', 'slandering', 'slandered', 'slandered', tran, ['6A'] ).
verb( 'slang', 'slangs', 'slanging', 'slanged', 'slanged', tran, ['6A'] ).
verb( 'slant', 'slants', 'slanting', 'slanted', 'slanted', _, ['2A','2C','6A'] ).
verb( 'slap', 'slaps', 'slapping', 'slapped', 'slapped', tran, ['6A','15A','15B'] ).
verb( 'slash', 'slashes', 'slashing', 'slashed', 'slashed', _, ['2C','6A'] ).
verb( 'slate', 'slates', 'slating', 'slated', 'slated', tran, ['6A'] ).
verb( 'slaughter', 'slaughters', 'slaughtering', 'slaughtered', 'slaughtered', tran, ['6A'] ).
verb( 'slave', 'slaves', 'slaving', 'slaved', 'slaved', intran, ['2A','2B','2C','3A'] ).
verb( 'slaver', 'slavers', 'slavering', 'slavered', 'slavered', intran, ['2A','2C'] ).
verb( 'slay', 'slays', 'slaying', 'slew', 'slain', tran, ['6A'] ).
verb( 'sledge', 'sledges', 'sledging', 'sledged', 'sledged', _, [] ).
verb( 'sleek', 'sleeks', 'sleeking', 'sleeked', 'sleeked', tran, ['6A'] ).
verb( 'sleep', 'sleeps', 'sleeping', 'slept', 'slept', _, ['2A','2B','2C','6A'] ).
verb( 'sleet', 'sleets', 'sleeting', 'sleeted', 'sleeted', intran, ['2A'] ).
verb( 'sleigh', 'sleighs', 'sleighing', 'sleighed', 'sleighed', _, ['2A','6A'] ).
verb( 'slenderize', 'slenderizes', 'slenderizing', 'slenderized', 'slenderized', _, ['6A'] ).
verb( 'slew', 'slews', 'slewing', 'slewed', 'slewed', _, ['2C','6A','15B'] ).
verb( 'slice', 'slices', 'slicing', 'sliced', 'sliced', _, ['3A','6A','15A','15B','22'] ).
verb( 'slide', 'slides', 'sliding', 'slid', 'slid', _, ['2A','2C','3A','6A','15A'] ).
verb( 'slight', 'slights', 'slighting', 'slighted', 'slighted', tran, ['6A'] ).
verb( 'slim', 'slims', 'slimming', 'slimmed', 'slimmed', intran, [] ).
verb( 'sling', 'slings', 'slinging', 'slung', 'slung', _, ['6A','15A','15B'] ).
verb( 'slink', 'slinks', 'slinking', 'slunk', 'slunk', intran, ['2C'] ).
verb( 'slip', 'slips', 'slipping', 'slipped', 'slipped', _, ['2A','2C','6A','15A','15B'] ).
verb( 'slit', 'slits', 'slitting', 'slit', 'slit', tran, ['2A','2C','6A','15A','22'] ).
verb( 'slither', 'slithers', 'slithering', 'slithered', 'slithered', intran, ['2A','2C'] ).
verb( 'sliver', 'slivers', 'slivering', 'slivered', 'slivered', _, ['2A','6A'] ).
verb( 'slobber', 'slobbers', 'slobbering', 'slobbered', 'slobbered', _, ['2A','3A','6A'] ).
verb( 'slog', 'slogs', 'slogging', 'slogged', 'slogged', _, ['2C','3A','6A'] ).
verb( 'slop', 'slops', 'slopping', 'slopped', 'slopped', _, ['2A','2C','6A','15A'] ).
verb( 'slope', 'slopes', 'sloping', 'sloped', 'sloped', _, ['2A','2C','6A'] ).
verb( 'slosh', 'sloshes', 'sloshing', 'sloshed', 'sloshed', _, ['3A','6A','15A','15B'] ).
verb( 'slot', 'slots', 'slotting', 'slotted', 'slotted', tran, ['6A','15A'] ).
verb( 'slouch', 'slouches', 'slouching', 'slouched', 'slouched', intran, ['2A','2C'] ).
verb( 'slough', 'sloughs', 'sloughing', 'sloughed', 'sloughed', _, ['2A','6A','15B'] ).
verb( 'slow', 'slows', 'slowing', 'slowed', 'slowed', _, ['2C','15B'] ).
verb( 'slug', 'slugs', 'slugging', 'slugged', 'slugged', _, [] ).
verb( 'sluice', 'sluices', 'sluicing', 'sluiced', 'sluiced', _, ['3A','6A','15B'] ).
verb( 'slum', 'slums', 'slumming', 'slummed', 'slummed', intran, [] ).
verb( 'slumber', 'slumbers', 'slumbering', 'slumbered', 'slumbered', _, ['2A','15B'] ).
verb( 'slump', 'slumps', 'slumping', 'slumped', 'slumped', intran, ['2A','2C'] ).
verb( 'slur', 'slurs', 'slurring', 'slurred', 'slurred', _, ['3A','6A'] ).
verb( 'smack', 'smacks', 'smacking', 'smacked', 'smacked', _, ['3A','6A'] ).
verb( 'smart', 'smarts', 'smarting', 'smarted', 'smarted', intran, ['2A','2C','3A'] ).
verb( 'smarten', 'smartens', 'smartening', 'smartened', 'smartened', _, ['2C','6A','15B'] ).
verb( 'smash', 'smashes', 'smashing', 'smashed', 'smashed', _, ['2A','2C','6A','15A','15B','22'] ).
verb( 'smear', 'smears', 'smearing', 'smeared', 'smeared', _, ['2A','6A','14'] ).
verb( 'smell', 'smells', 'smelling', 'smelled', 'smelled', _, ['2A','2C','2D','3A','6A','15B','19A'] ).
verb( 'smelt', 'smelts', 'smelting', 'smelted', 'smelted', tran, ['2A','2C','2D','3A','6A','15B','19A'] ).
verb( 'smile', 'smiles', 'smiling', 'smiled', 'smiled', _, ['2A','2B','3A','4B','6B'] ).
verb( 'smirch', 'smirches', 'smirching', 'smirched', 'smirched', tran, ['6A'] ).
verb( 'smirk', 'smirks', 'smirking', 'smirked', 'smirked', intran, ['2A'] ).
verb( 'smite', 'smites', 'smiting', 'smote', 'smitten', _, ['2A','2C','6A'] ).
verb( 'smoke', 'smokes', 'smoking', 'smoked', 'smoked', _, ['2A','2C','6A','15B','22'] ).
verb( 'smooth', 'smooths', 'smoothing', 'smoothed', 'smoothed', _, ['2C','6A','15B'] ).
verb( 'smother', 'smothers', 'smothering', 'smothered', 'smothered', tran, ['6A','14','15B'] ).
verb( 'smoulder', 'smoulders', 'smouldering', 'smouldered', 'smouldered', intran, ['2A','2C'] ).
verb( 'smudge', 'smudges', 'smudging', 'smudged', 'smudged', _, ['2A','6A'] ).
verb( 'smuggle', 'smuggles', 'smuggling', 'smuggled', 'smuggled', tran, ['6A','14','15A','15B'] ).
verb( 'smut', 'smuts', 'smutting', 'smutted', 'smutted', tran, ['6A'] ).
verb( 'snaffle', 'snaffles', 'snaffling', 'snaffled', 'snaffled', tran, ['6A'] ).
verb( 'snag', 'snags', 'snagging', 'snagged', 'snagged', tran, [] ).
verb( 'snake', 'snakes', 'snaking', 'snaked', 'snaked', intran, ['2C'] ).
verb( 'snap', 'snaps', 'snapping', 'snapped', 'snapped', _, ['2A','2C','3A','6A','14','15A','15B','22'] ).
verb( 'snare', 'snares', 'snaring', 'snared', 'snared', tran, ['6A'] ).
verb( 'snarl', 'snarls', 'snarling', 'snarled', 'snarled', _, ['2A','2C','3A','6A','15B'] ).
verb( 'snatch', 'snatches', 'snatching', 'snatched', 'snatched', _, ['2A','6A','15A','15B'] ).
verb( 'sneak', 'sneaks', 'sneaking', 'sneaked', 'sneaked', _, ['2A','3A','6A'] ).
verb( 'sneer', 'sneers', 'sneering', 'sneered', 'sneered', intran, ['2A','3A'] ).
verb( 'sneeze', 'sneezes', 'sneezing', 'sneezed', 'sneezed', intran, ['2A'] ).
verb( 'snick', 'snicks', 'snicking', 'snicked', 'snicked', _, [] ).
verb( 'snicker', 'snickers', 'snickering', 'snickered', 'snickered', intran, [] ).
verb( 'sniff', 'sniffs', 'sniffing', 'sniffed', 'sniffed', _, ['2A','3A','6A','15B'] ).
verb( 'sniffle', 'sniffles', 'sniffling', 'sniffled', 'sniffled', intran, [] ).
verb( 'snigger', 'sniggers', 'sniggering', 'sniggered', 'sniggered', intran, ['2A','2C'] ).
verb( 'snip', 'snips', 'snipping', 'snipped', 'snipped', _, ['3A','6A','15B'] ).
verb( 'snipe', 'snipes', 'sniping', 'sniped', 'sniped', _, ['2A','3A','6A'] ).
verb( 'snitch', 'snitches', 'snitching', 'snitched', 'snitched', _, ['2A','3A','6A'] ).
verb( 'snivel', 'snivels', 'snivelling', 'snivelled', 'snivelled', intran, ['2A'] ).
verb( 'snog', 'snogs', 'snogging', 'snogged', 'snogged', intran, ['2A'] ).
verb( 'snoop', 'snoops', 'snooping', 'snooped', 'snooped', intran, ['2A','2C','3A'] ).
verb( 'snooze', 'snoozes', 'snoozing', 'snoozed', 'snoozed', intran, ['2A'] ).
verb( 'snore', 'snores', 'snoring', 'snored', 'snored', intran, ['2A','2C'] ).
verb( 'snort', 'snorts', 'snorting', 'snorted', 'snorted', _, ['2A','2C','6A','15B'] ).
verb( 'snow', 'snows', 'snowing', 'snowed', 'snowed', _, ['2A','2B','2C'] ).
verb( 'snowball', 'snowballs', 'snowballing', 'snowballed', 'snowballed', _, [] ).
verb( 'snub', 'snubs', 'snubbing', 'snubbed', 'snubbed', tran, ['6A'] ).
verb( 'snuff', 'snuffs', 'snuffing', 'snuffed', 'snuffed', _, ['2C','6A','15B'] ).
verb( 'snuffle', 'snuffles', 'snuffling', 'snuffled', 'snuffled', intran, ['2A'] ).
verb( 'snuggle', 'snuggles', 'snuggling', 'snuggled', 'snuggled', _, ['2C','14'] ).
verb( 'soak', 'soaks', 'soaking', 'soaked', 'soaked', _, ['2A','3A','6A','14','15A','15B'] ).
verb( 'soap', 'soaps', 'soaping', 'soaped', 'soaped', tran, ['6A','15B'] ).
verb( 'soar', 'soars', 'soaring', 'soared', 'soared', intran, ['2A','2C'] ).
verb( 'sob', 'sobs', 'sobbing', 'sobbed', 'sobbed', _, ['2A','2C','15A','15B'] ).
verb( 'sober', 'sobers', 'sobering', 'sobered', 'sobered', _, ['2C','6A','15B'] ).
verb( 'socialize', 'socializes', 'socializing', 'socialized', 'socialized', tran, ['6A'] ).
verb( 'sock', 'socks', 'socking', 'socked', 'socked', tran, ['6A','15A'] ).
verb( 'sod', 'sods', 'sodding', 'sodded', 'sodded', intran, [] ).
verb( 'soft-pedal', 'soft-pedals', 'soft-pedalling', 'soft-pedalled', 'soft-pedalled', _, [] ).
verb( 'soft-soap', 'soft-soaps', 'soft-soaping', 'soft-soaped', 'soft-soaped', tran, [] ).
verb( 'soft-solder', 'soft-solders', 'soft-soldering', 'soft-soldered', 'soft-soldered', tran, [] ).
verb( 'soften', 'softens', 'softening', 'softened', 'softened', _, ['2A','6A'] ).
verb( 'softland', 'softlands', 'softlanding', 'softlanded', 'softlanded', intran, [] ).
verb( 'soil', 'soils', 'soiling', 'soiled', 'soiled', _, ['2A','2C','6A'] ).
verb( 'sojourn', 'sojourns', 'sojourning', 'sojourned', 'sojourned', intran, ['2C'] ).
verb( 'solace', 'solaces', 'solacing', 'solaced', 'solaced', tran, ['6A','15A'] ).
verb( 'solder', 'solders', 'soldering', 'soldered', 'soldered', tran, ['6A','15A','15B'] ).
verb( 'soldier', 'soldiers', 'soldiering', 'soldiered', 'soldiered', intran, ['2A'] ).
verb( 'sole', 'soles', 'soling', 'soled', 'soled', tran, ['6A'] ).
verb( 'solemnize', 'solemnizes', 'solemnizing', 'solemnized', 'solemnized', tran, ['6A'] ).
verb( 'solicit', 'solicits', 'soliciting', 'solicited', 'solicited', _, ['2A','6A','14'] ).
verb( 'solidify', 'solidifies', 'solidifying', 'solidified', 'solidified', _, ['2A','6A'] ).
verb( 'soliloquize', 'soliloquizes', 'soliloquizing', 'soliloquized', 'soliloquized', intran, ['2A'] ).
verb( 'solve', 'solves', 'solving', 'solved', 'solved', tran, ['6A'] ).
verb( 'somersault', 'somersaults', 'somersaulting', 'somersaulted', 'somersaulted', intran, ['2A'] ).
verb( 'soot', 'soots', 'sooting', 'sooted', 'sooted', tran, [] ).
verb( 'soothe', 'soothes', 'soothing', 'soothed', 'soothed', tran, ['6A'] ).
verb( 'sop', 'sops', 'sopping', 'sopped', 'sopped', tran, ['6A','15B'] ).
verb( 'sorrow', 'sorrows', 'sorrowing', 'sorrowed', 'sorrowed', intran, ['2A','2C'] ).
verb( 'sort', 'sorts', 'sorting', 'sorted', 'sorted', _, ['6A','15B'] ).
verb( 'sough', 'soughs', 'soughing', 'soughed', 'soughed', intran, ['2A'] ).
verb( 'sound', 'sounds', 'sounding', 'sounded', 'sounded', _, ['2A','2C','2D','6A','15A','15B'] ).
verb( 'soundproof', 'soundproofs', 'soundproofing', 'soundproofed', 'soundproofed', tran, ['6A'] ).
verb( 'soup', 'soups', 'souping', 'souped', 'souped', tran, ['15B'] ).
verb( 'sour', 'sours', 'souring', 'soured', 'soured', _, ['2A','6A'] ).
verb( 'souse', 'souses', 'sousing', 'soused', 'soused', tran, ['6A'] ).
verb( 'sovietize', 'sovietizes', 'sovietizing', 'sovietized', 'sovietized', tran, [] ).
verb( 'sow', 'sows', 'sowing', 'sowed', 'sowed', _, ['2A','6A','15A'] ).
verb( 'space', 'spaces', 'spacing', 'spaced', 'spaced', tran, ['6A','15B'] ).
verb( 'spade', 'spades', 'spading', 'spaded', 'spaded', tran, ['6A','15B'] ).
verb( 'span', 'spans', 'spanning', 'spanned', 'spanned', tran, ['2A','2C','6A','14','15B'] ).
verb( 'spangle', 'spangles', 'spangling', 'spangled', 'spangled', tran, [] ).
verb( 'spank', 'spanks', 'spanking', 'spanked', 'spanked', _, ['2C','6A'] ).
verb( 'spar', 'spars', 'sparring', 'sparred', 'sparred', intran, ['2A','2C'] ).
verb( 'spare', 'spares', 'sparing', 'spared', 'spared', _, ['6A','12A','12B','13A','13B'] ).
verb( 'spark', 'sparks', 'sparking', 'sparked', 'sparked', _, ['2A','15B'] ).
verb( 'sparkle', 'sparkles', 'sparkling', 'sparkled', 'sparkled', intran, ['2A','2C'] ).
verb( 'spat', 'spats', 'spatting', 'spatted', 'spatted', _, ['2A','2C','3A','6A','15B'] ).
verb( 'spatchcock', 'spatchcocks', 'spatchcocking', 'spatchcocked', 'spatchcocked', tran, ['6A','14'] ).
verb( 'spatter', 'spatters', 'spattering', 'spattered', 'spattered', _, ['2C','6A','14'] ).
verb( 'spawn', 'spawns', 'spawning', 'spawned', 'spawned', _, ['2A','6A'] ).
verb( 'spay', 'spays', 'spaying', 'spayed', 'spayed', tran, ['6A'] ).
verb( 'speak', 'speaks', 'speaking', 'spoke', 'spoken', _, ['2A','2B','2C','3A','6A','15A'] ).
verb( 'spear', 'spears', 'spearing', 'speared', 'speared', tran, ['6A'] ).
verb( 'specialize', 'specializes', 'specializing', 'specialized', 'specialized', _, ['2A','3A'] ).
verb( 'specify', 'specifies', 'specifying', 'specified', 'specified', tran, ['6A','9'] ).
verb( 'speculate', 'speculates', 'speculating', 'speculated', 'speculated', intran, ['2A','2C'] ).
verb( 'speechify', 'speechifies', 'speechifying', 'speechified', 'speechified', intran, [] ).
verb( 'speed', 'speeds', 'speeding', 'speeded', 'speeded', _, ['2A','2C','6A','14','15B'] ).
verb( 'spell', 'spells', 'spelling', 'spelled', 'spelled', _, ['2A','6A','15B'] ).
verb( 'spend', 'spends', 'spending', 'spent', 'spent', _, ['2A','6A','14','19B'] ).
verb( 'spew', 'spews', 'spewing', 'spewed', 'spewed', _, ['2A','6A','15B'] ).
verb( 'spice', 'spices', 'spicing', 'spiced', 'spiced', tran, ['6A'] ).
verb( 'spiel', 'spiels', 'spieling', 'spieled', 'spieled', _, [] ).
verb( 'spike', 'spikes', 'spiking', 'spiked', 'spiked', tran, ['6A'] ).
verb( 'spill', 'spills', 'spilling', 'spilled', 'spilled', _, ['2A','2C','6A'] ).
verb( 'spin', 'spins', 'spinning', 'span', 'spun', _, ['2A','2C','6A','14','15B'] ).
verb( 'spin-dry', 'spin-dries', 'spin-drying', 'spin-dried', 'spin-dried', tran, [] ).
verb( 'spiral', 'spirals', 'spiralling', 'spiralled', 'spiralled', intran, ['2A','2C'] ).
verb( 'spirit', 'spirits', 'spiriting', 'spirited', 'spirited', tran, ['15B'] ).
verb( 'spiritualize', 'spiritualizes', 'spiritualizing', 'spiritualized', 'spiritualized', tran, ['6A'] ).
verb( 'spirt', 'spirts', 'spirting', 'spirted', 'spirted', intran, ['2A','2C'] ).
verb( 'spit', 'spits', 'spitting', 'spat', 'spat', _, ['2A','2C','3A','6A','15B'] ).
verb( 'spite', 'spites', 'spiting', 'spited', 'spited', tran, ['6A'] ).
verb( 'splash', 'splashes', 'splashing', 'splashed', 'splashed', _, ['2A','2C','6A','14','15B'] ).
verb( 'splay', 'splays', 'splaying', 'splayed', 'splayed', _, [] ).
verb( 'splice', 'splices', 'splicing', 'spliced', 'spliced', tran, ['6A'] ).
verb( 'splinter', 'splinters', 'splintering', 'splintered', 'splintered', _, ['2C','15B'] ).
verb( 'split', 'splits', 'splitting', 'split', 'split', _, ['2A','2C','2D','3A','6A','14','15A','15B','22'] ).
verb( 'splosh', 'sploshes', 'sploshing', 'sploshed', 'sploshed', tran, [] ).
verb( 'splurge', 'splurges', 'splurging', 'splurged', 'splurged', intran, [] ).
verb( 'splutter', 'splutters', 'spluttering', 'spluttered', 'spluttered', _, ['2A','2C','6A','15B'] ).
verb( 'spoil', 'spoils', 'spoiling', 'spoiled', 'spoiled', _, ['2A','6A','14'] ).
verb( 'sponge', 'sponges', 'sponging', 'sponged', 'sponged', _, ['3A','6A','14','15B'] ).
verb( 'sponsor', 'sponsors', 'sponsoring', 'sponsored', 'sponsored', tran, ['6A'] ).
verb( 'spoof', 'spoofs', 'spoofing', 'spoofed', 'spoofed', tran, [] ).
verb( 'spoon', 'spoons', 'spooning', 'spooned', 'spooned', _, ['2A','15B'] ).
verb( 'spoonfeed', 'spoonfeeds', 'spoonfeeding', 'spoonfed', 'spoonfed', tran, [] ).
verb( 'sport', 'sports', 'sporting', 'sported', 'sported', _, ['2C','6A'] ).
verb( 'spot', 'spots', 'spotting', 'spotted', 'spotted', _, ['2A','6A'] ).
verb( 'spotlight', 'spotlights', 'spotlighting', 'spotlighted', 'spotlighted', tran, ['6A'] ).
verb( 'spout', 'spouts', 'spouting', 'spouted', 'spouted', _, ['2A','2C','6A'] ).
verb( 'sprain', 'sprains', 'spraining', 'sprained', 'sprained', tran, ['6A'] ).
verb( 'sprawl', 'sprawls', 'sprawling', 'sprawled', 'sprawled', intran, ['2A','2C'] ).
verb( 'spray', 'sprays', 'spraying', 'sprayed', 'sprayed', tran, ['6A','14'] ).
verb( 'spread', 'spreads', 'spreading', 'spread', 'spread', _, ['2A','2B','2C','6A','14','15A','15B'] ).
verb( 'spreadeagle', 'spreadeagles', 'spreadeagling', 'spreadeagled', 'spreadeagled', tran, [] ).
verb( 'spring', 'springs', 'springing', 'sprang', 'sprung', _, ['2A','2C','3A','6A','14'] ).
verb( 'spring-clean', 'spring-cleans', 'spring-cleaning', 'spring-cleaned', 'spring-cleaned', tran, [] ).
verb( 'sprinkle', 'sprinkles', 'sprinkling', 'sprinkled', 'sprinkled', tran, ['6A','14'] ).
verb( 'sprint', 'sprints', 'sprinting', 'sprinted', 'sprinted', intran, ['2A','2C'] ).
verb( 'sprout', 'sprouts', 'sprouting', 'sprouted', 'sprouted', _, ['2A','2C','6A'] ).
verb( 'spruce', 'spruces', 'sprucing', 'spruced', 'spruced', _, ['2C','6A','15B'] ).
verb( 'spue', 'spues', 'spuing', 'spued', 'spued', _, [] ).
verb( 'spur', 'spurs', 'spurring', 'spurred', 'spurred', _, ['2C','6A','15B'] ).
verb( 'spurn', 'spurns', 'spurning', 'spurned', 'spurned', tran, ['6A'] ).
verb( 'spurt', 'spurts', 'spurting', 'spurted', 'spurted', intran, ['2A','2C','3A'] ).
verb( 'sputter', 'sputters', 'sputtering', 'sputtered', 'sputtered', _, ['2A','2C'] ).
verb( 'spy', 'spies', 'spying', 'spied', 'spied', _, ['2A','3A','6A','15B','19A'] ).
verb( 'squabble', 'squabbles', 'squabbling', 'squabbled', 'squabbled', intran, ['2C'] ).
verb( 'squall', 'squalls', 'squalling', 'squalled', 'squalled', intran, ['2A'] ).
verb( 'squander', 'squanders', 'squandering', 'squandered', 'squandered', tran, ['6A'] ).
verb( 'square', 'squares', 'squaring', 'squared', 'squared', _, ['2C','3A','6A','14','15B'] ).
verb( 'squash', 'squashes', 'squashing', 'squashed', 'squashed', _, ['2A','2C','6A','15A','15B','22'] ).
verb( 'squat', 'squats', 'squatting', 'squatted', 'squatted', intran, ['2A','2C','6A','15B'] ).
verb( 'squawk', 'squawks', 'squawking', 'squawked', 'squawked', intran, ['2A','2C'] ).
verb( 'squeak', 'squeaks', 'squeaking', 'squeaked', 'squeaked', _, ['2A','2C','6A','15B'] ).
verb( 'squeal', 'squeals', 'squealing', 'squealed', 'squealed', _, ['2A','2C','6A','15B'] ).
verb( 'squeegee', 'squeegees', 'squeegeeing', 'squeegeed', 'squeegeed', tran, ['6A'] ).
verb( 'squeeze', 'squeezes', 'squeezing', 'squeezed', 'squeezed', _, ['2A','2C','6A','14','15A','15B','22'] ).
verb( 'squelch', 'squelches', 'squelching', 'squelched', 'squelched', _, ['2A','2C'] ).
verb( 'squint', 'squints', 'squinting', 'squinted', 'squinted', intran, ['2A','3A'] ).
verb( 'squire', 'squires', 'squiring', 'squired', 'squired', tran, ['6A','15A'] ).
verb( 'squirm', 'squirms', 'squirming', 'squirmed', 'squirmed', intran, ['2A','2C'] ).
verb( 'squirt', 'squirts', 'squirting', 'squirted', 'squirted', _, ['2A','2C','6A','15A','15B'] ).
verb( 'stab', 'stabs', 'stabbing', 'stabbed', 'stabbed', _, ['2A','3A','6A','15A'] ).
verb( 'stabilize', 'stabilizes', 'stabilizing', 'stabilized', 'stabilized', tran, ['6A'] ).
verb( 'stable', 'stables', 'stabling', 'stabled', 'stabled', tran, [] ).
verb( 'stack', 'stacks', 'stacking', 'stacked', 'stacked', tran, ['6A','15B'] ).
verb( 'staff', 'staffs', 'staffing', 'staffed', 'staffed', tran, ['6A'] ).
verb( 'stage', 'stages', 'staging', 'staged', 'staged', _, ['2C','6A'] ).
verb( 'stagger', 'staggers', 'staggering', 'staggered', 'staggered', _, ['2A','2C','6A'] ).
verb( 'stagnate', 'stagnates', 'stagnating', 'stagnated', 'stagnated', intran, ['2A'] ).
verb( 'stain', 'stains', 'staining', 'stained', 'stained', _, ['2A','6A','22'] ).
verb( 'stake', 'stakes', 'staking', 'staked', 'staked', tran, ['6A','14','15B'] ).
verb( 'stale', 'stales', 'staling', 'staled', 'staled', intran, ['2A'] ).
verb( 'stalemate', 'stalemates', 'stalemating', 'stalemated', 'stalemated', tran, ['6A'] ).
verb( 'stalk', 'stalks', 'stalking', 'stalked', 'stalked', _, ['2C','6A'] ).
verb( 'stall', 'stalls', 'stalling', 'stalled', 'stalled', _, ['2A','6A'] ).
verb( 'stammer', 'stammers', 'stammering', 'stammered', 'stammered', _, ['2A','6A','15B'] ).
verb( 'stamp', 'stamps', 'stamping', 'stamped', 'stamped', _, ['2C','3A','6A','14','15A','15B','16B','22'] ).
verb( 'stampede', 'stampedes', 'stampeding', 'stampeded', 'stampeded', _, ['2A','6A','14'] ).
verb( 'stand', 'stands', 'standing', 'stood', 'stood', _, ['2A','2B','2C','3A','4A','6A','6C','12A','15A','15B'] ).
verb( 'standardize', 'standardizes', 'standardizing', 'standardized', 'standardized', tran, ['6A'] ).
verb( 'staple', 'staples', 'stapling', 'stapled', 'stapled', tran, ['6A'] ).
verb( 'star', 'stars', 'starring', 'starred', 'starred', _, ['3A','6A','14'] ).
verb( 'starboard', 'starboards', 'starboarding', 'starboarded', 'starboarded', tran, [] ).
verb( 'starch', 'starches', 'starching', 'starched', 'starched', tran, ['6A'] ).
verb( 'stare', 'stares', 'staring', 'stared', 'stared', _, ['2A','2B','2C','3A','15B'] ).
verb( 'start', 'starts', 'starting', 'started', 'started', _, ['2A','2C','3A','6A','6D','7A','15A','19B'] ).
verb( 'startle', 'startles', 'startling', 'startled', 'startled', tran, ['6A','15A'] ).
verb( 'starve', 'starves', 'starving', 'starved', 'starved', _, ['2A','2C','6A','15A'] ).
verb( 'stash', 'stashes', 'stashing', 'stashed', 'stashed', tran, ['15B'] ).
verb( 'state', 'states', 'stating', 'stated', 'stated', tran, ['6A','9'] ).
verb( 'station', 'stations', 'stationing', 'stationed', 'stationed', tran, ['6A','15A'] ).
verb( 'staunch', 'staunches', 'staunching', 'staunched', 'staunched', tran, ['6A'] ).
verb( 'stave', 'staves', 'staving', 'staved', 'staved', _, ['3A','15B'] ).
verb( 'stay', 'stays', 'staying', 'stayed', 'stayed', _, ['2A','2B','2C','4A','6A','15B'] ).
verb( 'steady', 'steadies', 'steadying', 'steadied', 'steadied', _, ['2A','6A'] ).
verb( 'steal', 'steals', 'stealing', 'stole', 'stolen', _, ['2A','2C','6A','14','15A'] ).
verb( 'steam', 'steams', 'steaming', 'steamed', 'steamed', _, ['2A','2C','6A','22'] ).
verb( 'steam-heat', 'steam-heats', 'steam-heating', 'steam-heated', 'steam-heated', tran, [] ).
verb( 'steamroller', 'steamrollers', 'steamrollering', 'steamrollered', 'steamrollered', tran, ['6A'] ).
verb( 'steel', 'steels', 'steeling', 'steeled', 'steeled', tran, ['6A','15A','16A'] ).
verb( 'steep', 'steeps', 'steeping', 'steeped', 'steeped', _, ['2A','6A','14'] ).
verb( 'steepen', 'steepens', 'steepening', 'steepened', 'steepened', _, ['2A','6A'] ).
verb( 'steer', 'steers', 'steering', 'steered', 'steered', _, ['2A','2C','6A'] ).
verb( 'stem', 'stems', 'stemming', 'stemmed', 'stemmed', _, ['3A','6A'] ).
verb( 'stencil', 'stencils', 'stencilling', 'stencilled', 'stencilled', tran, ['6A'] ).
verb( 'step', 'steps', 'stepping', 'stepped', 'stepped', _, ['2C','15B'] ).
verb( 'stereotype', 'stereotypes', 'stereotyping', 'stereotyped', 'stereotyped', tran, ['6A'] ).
verb( 'sterilize', 'sterilizes', 'sterilizing', 'sterilized', 'sterilized', tran, ['6A'] ).
verb( 'stew', 'stews', 'stewing', 'stewed', 'stewed', _, ['2A','2C','6A','15A'] ).
verb( 'stick', 'sticks', 'sticking', 'sticked', 'sticked', tran, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'stick', 'sticks', 'sticking', 'stuck', 'stuck', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'stiffen', 'stiffens', 'stiffening', 'stiffened', 'stiffened', _, ['2A','6A'] ).
verb( 'stifle', 'stifles', 'stifling', 'stifled', 'stifled', _, ['2A','6A'] ).
verb( 'stigmatize', 'stigmatizes', 'stigmatizing', 'stigmatized', 'stigmatized', tran, ['16B'] ).
verb( 'still', 'stills', 'stilling', 'stilled', 'stilled', tran, ['6A'] ).
verb( 'stimulate', 'stimulates', 'stimulating', 'stimulated', 'stimulated', tran, ['6A','14','17'] ).
verb( 'sting', 'stings', 'stinging', 'stang', 'stung', _, ['2A','6A','14','15A'] ).
verb( 'stink', 'stinks', 'stinking', 'stank', 'stunk', _, ['2A','3A','15B'] ).
verb( 'stint', 'stints', 'stinting', 'stinted', 'stinted', _, ['6A','14'] ).
verb( 'stipple', 'stipples', 'stippling', 'stippled', 'stippled', tran, ['6A'] ).
verb( 'stipulate', 'stipulates', 'stipulating', 'stipulated', 'stipulated', _, ['3A','6A','9'] ).
verb( 'stir', 'stirs', 'stirring', 'stirred', 'stirred', _, ['2A','2C','6A','14','15B'] ).
verb( 'stitch', 'stitches', 'stitching', 'stitched', 'stitched', _, ['2A','2C','6A','15A','15B'] ).
verb( 'stock', 'stocks', 'stocking', 'stocked', 'stocked', tran, ['6A','14'] ).
verb( 'stockade', 'stockades', 'stockading', 'stockaded', 'stockaded', tran, [] ).
verb( 'stoke', 'stokes', 'stoking', 'stoked', 'stoked', _, ['2A','2C','6A','15B'] ).
verb( 'stomach', 'stomachs', 'stomaching', 'stomached', 'stomached', tran, ['6A'] ).
verb( 'stomp', 'stomps', 'stomping', 'stomped', 'stomped', intran, ['2C'] ).
verb( 'stone', 'stones', 'stoning', 'stoned', 'stoned', tran, ['6A','15A'] ).
verb( 'stonewall', 'stonewalls', 'stonewalling', 'stonewalled', 'stonewalled', tran, [] ).
verb( 'stooge', 'stooges', 'stooging', 'stooged', 'stooged', intran, [] ).
verb( 'stoop', 'stoops', 'stooping', 'stooped', 'stooped', _, ['2A','2C','3A','4A','6A'] ).
verb( 'stop', 'stops', 'stopping', 'stopped', 'stopped', _, ['2A','2B','2C','3A','4A','6A','6B','6C','14','15A','15B'] ).
verb( 'store', 'stores', 'storing', 'stored', 'stored', tran, ['6A','15B'] ).
verb( 'storm', 'storms', 'storming', 'stormed', 'stormed', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'stow', 'stows', 'stowing', 'stowed', 'stowed', tran, ['6A','15A','15B'] ).
verb( 'straddle', 'straddles', 'straddling', 'straddled', 'straddled', _, ['2A','6A'] ).
verb( 'strafe', 'strafes', 'strafing', 'strafed', 'strafed', tran, ['6A'] ).
verb( 'straggle', 'straggles', 'straggling', 'straggled', 'straggled', intran, ['2A','2C'] ).
verb( 'straighten', 'straightens', 'straightening', 'straightened', 'straightened', _, ['2A','2C','6A','15B'] ).
verb( 'strain', 'strains', 'straining', 'strained', 'strained', _, ['2A','3A','6A','15A','15B','16A'] ).
verb( 'straiten', 'straitens', 'straitening', 'straitened', 'straitened', tran, [] ).
verb( 'strand', 'strands', 'stranding', 'stranded', 'stranded', _, ['2A','6A'] ).
verb( 'strangle', 'strangles', 'strangling', 'strangled', 'strangled', tran, ['6A'] ).
verb( 'strap', 'straps', 'strapping', 'strapped', 'strapped', tran, ['6A','15B'] ).
verb( 'stratify', 'stratifies', 'stratifying', 'stratified', 'stratified', _, ['2A','6A'] ).
verb( 'straw', 'straws', 'strawing', 'strawed', 'strawed', tran, [] ).
verb( 'stray', 'strays', 'straying', 'strayed', 'strayed', intran, ['2A','2C'] ).
verb( 'streak', 'streaks', 'streaking', 'streaked', 'streaked', _, ['2C','6A','15A'] ).
verb( 'stream', 'streams', 'streaming', 'streamed', 'streamed', intran, ['2A','2C','6A'] ).
verb( 'streamline', 'streamlines', 'streamlining', 'streamlined', 'streamlined', tran, [] ).
verb( 'strengthen', 'strengthens', 'strengthening', 'strengthened', 'strengthened', _, ['2A','6A'] ).
verb( 'stress', 'stresses', 'stressing', 'stressed', 'stressed', tran, ['6A'] ).
verb( 'stretch', 'stretches', 'stretching', 'stretched', 'stretched', _, ['2A','2B','2C','6A','15A','15B','16A','22'] ).
verb( 'strew', 'strews', 'strewing', 'strewed', 'strewed', tran, ['6A','14'] ).
verb( 'stride', 'strides', 'striding', 'strode', 'stridden', _, ['2C','6A'] ).
verb( 'stridulate', 'stridulates', 'stridulating', 'stridulated', 'stridulated', intran, ['2A'] ).
verb( 'strike', 'strikes', 'striking', 'struck', 'struck', _, ['2A','2C','2D','3A','6A','12C','14','15B','16A','22'] ).
verb( 'string', 'strings', 'stringing', 'strang', 'strung', _, ['2C','6A','15A','15B'] ).
verb( 'strip', 'strips', 'stripping', 'stripped', 'stripped', _, ['2A','2C','6A','14','15B','22'] ).
verb( 'strive', 'strives', 'striving', 'strove', 'striven', intran, ['2A','3A','4A'] ).
verb( 'stroke', 'strokes', 'stroking', 'stroked', 'stroked', tran, ['6A','15B'] ).
verb( 'stroll', 'strolls', 'strolling', 'strolled', 'strolled', intran, ['2A','2C'] ).
verb( 'strop', 'strops', 'stropping', 'stropped', 'stropped', tran, ['6A'] ).
verb( 'struggle', 'struggles', 'struggling', 'struggled', 'struggled', intran, ['2A','2B','3A','4A'] ).
verb( 'strum', 'strums', 'strumming', 'strummed', 'strummed', _, ['2A','2B','2C','3A','6A'] ).
verb( 'strut', 'struts', 'strutting', 'strutted', 'strutted', intran, ['2A','2C'] ).
verb( 'stub', 'stubs', 'stubbing', 'stubbed', 'stubbed', tran, ['6A','15B'] ).
verb( 'stucco', 'stuccos', 'stuccoing', 'stuccoed', 'stuccoed', tran, [] ).
verb( 'stud', 'studs', 'studding', 'studded', 'studded', tran, [] ).
verb( 'study', 'studies', 'studying', 'studied', 'studied', _, ['2A','2B','4A','6A','8','15A'] ).
verb( 'stuff', 'stuffs', 'stuffing', 'stuffed', 'stuffed', tran, ['2A','6A','14','15B'] ).
verb( 'stultify', 'stultifies', 'stultifying', 'stultified', 'stultified', tran, ['6A'] ).
verb( 'stumble', 'stumbles', 'stumbling', 'stumbled', 'stumbled', intran, ['2A','2C','3A'] ).
verb( 'stump', 'stumps', 'stumping', 'stumped', 'stumped', _, ['2C','6A','15B'] ).
verb( 'stun', 'stuns', 'stunning', 'stunned', 'stunned', tran, ['6A'] ).
verb( 'stunt', 'stunts', 'stunting', 'stunted', 'stunted', tran, ['6A'] ).
verb( 'stupefy', 'stupefies', 'stupefying', 'stupefied', 'stupefied', tran, ['6A'] ).
verb( 'stupify', 'stupifies', 'stupifying', 'stupified', 'stupified', tran, ['6A'] ).
verb( 'stutter', 'stutters', 'stuttering', 'stuttered', 'stuttered', _, [] ).
verb( 'style', 'styles', 'styling', 'styled', 'styled', tran, ['6A'] ).
verb( 'stylize', 'stylizes', 'stylizing', 'stylized', 'stylized', tran, [] ).
verb( 'stymie', 'stymies', 'stymying', 'stymied', 'stymied', tran, ['6A'] ).
verb( 'sub', 'subs', 'subbing', 'subbed', 'subbed', _, ['2A','3A','6A'] ).
verb( 'subcontract', 'subcontracts', 'subcontracting', 'subcontracted', 'subcontracted', _, ['2A','6A'] ).
verb( 'subdivide', 'subdivides', 'subdividing', 'subdivided', 'subdivided', _, ['2A','6A'] ).
verb( 'subdue', 'subdues', 'subduing', 'subdued', 'subdued', tran, ['6A'] ).
verb( 'subedit', 'subedits', 'subediting', 'subedited', 'subedited', tran, [] ).
verb( 'subject', 'subjects', 'subjecting', 'subjected', 'subjected', tran, ['14'] ).
verb( 'subjoin', 'subjoins', 'subjoining', 'subjoined', 'subjoined', tran, ['6A'] ).
verb( 'subjugate', 'subjugates', 'subjugating', 'subjugated', 'subjugated', tran, ['6A'] ).
verb( 'sublease', 'subleases', 'subleasing', 'subleased', 'subleased', _, ['2A','6A'] ).
verb( 'sublet', 'sublets', 'subletting', 'sublet', 'sublet', _, ['2A','6A'] ).
verb( 'sublimate', 'sublimates', 'sublimating', 'sublimated', 'sublimated', tran, ['6A'] ).
verb( 'submerge', 'submerges', 'submerging', 'submerged', 'submerged', _, ['2A','6A'] ).
verb( 'submit', 'submits', 'submitting', 'submitted', 'submitted', _, ['3A','6A','9','14'] ).
verb( 'subordinate', 'subordinates', 'subordinating', 'subordinated', 'subordinated', tran, ['6A','14'] ).
verb( 'suborn', 'suborns', 'suborning', 'suborned', 'suborned', tran, ['6A'] ).
verb( 'subpoena', 'subpoenas', 'subpoenaing', 'subpoenaed', 'subpoenaed', tran, ['6A'] ).
verb( 'subscribe', 'subscribes', 'subscribing', 'subscribed', 'subscribed', _, ['2A','3A','6A','14'] ).
verb( 'subserve', 'subserves', 'subserving', 'subserved', 'subserved', tran, ['6A'] ).
verb( 'subside', 'subsides', 'subsiding', 'subsided', 'subsided', intran, ['2A'] ).
verb( 'subsidize', 'subsidizes', 'subsidizing', 'subsidized', 'subsidized', tran, ['6A'] ).
verb( 'subsist', 'subsists', 'subsisting', 'subsisted', 'subsisted', intran, ['2A','3A'] ).
verb( 'substantiate', 'substantiates', 'substantiating', 'substantiated', 'substantiated', tran, ['6A'] ).
verb( 'substitute', 'substitutes', 'substituting', 'substituted', 'substituted', _, ['3A','6A','14'] ).
verb( 'subsume', 'subsumes', 'subsuming', 'subsumed', 'subsumed', tran, ['6A','14'] ).
verb( 'subtend', 'subtends', 'subtending', 'subtended', 'subtended', tran, [] ).
verb( 'subtract', 'subtracts', 'subtracting', 'subtracted', 'subtracted', tran, ['6A','14'] ).
verb( 'subvert', 'subverts', 'subverting', 'subverted', 'subverted', tran, ['6A'] ).
verb( 'succeed', 'succeeds', 'succeeding', 'succeeded', 'succeeded', _, ['2A','3A','6A','16B'] ).
verb( 'succour', 'succours', 'succouring', 'succoured', 'succoured', tran, ['6A'] ).
verb( 'succumb', 'succumbs', 'succumbing', 'succumbed', 'succumbed', intran, ['2A','3A'] ).
verb( 'suck', 'sucks', 'sucking', 'sucked', 'sucked', _, ['2C','6A','15A','15B','22'] ).
verb( 'suckle', 'suckles', 'suckling', 'suckled', 'suckled', tran, ['6A'] ).
verb( 'sue', 'sues', 'suing', 'sued', 'sued', _, ['3A','6A','14'] ).
verb( 'suffer', 'suffers', 'suffering', 'suffered', 'suffered', _, ['2A','3A','6A','17'] ).
verb( 'suffice', 'suffices', 'sufficing', 'sufficed', 'sufficed', _, ['2A','3A','6A'] ).
verb( 'suffocate', 'suffocates', 'suffocating', 'suffocated', 'suffocated', _, ['2A','2C','6A'] ).
verb( 'suffuse', 'suffuses', 'suffusing', 'suffused', 'suffused', tran, ['6A'] ).
verb( 'sugar', 'sugars', 'sugaring', 'sugared', 'sugared', tran, ['6A'] ).
verb( 'suggest', 'suggests', 'suggesting', 'suggested', 'suggested', tran, ['6A','6C','9','10','14'] ).
verb( 'suit', 'suits', 'suiting', 'suited', 'suited', _, ['2A','6A','6B'] ).
verb( 'sulk', 'sulks', 'sulking', 'sulked', 'sulked', intran, ['2A','2C'] ).
verb( 'sully', 'sullies', 'sullying', 'sullied', 'sullied', tran, ['6A'] ).
verb( 'sum', 'sums', 'summing', 'summed', 'summed', _, ['2C','15B'] ).
verb( 'summarize', 'summarizes', 'summarizing', 'summarized', 'summarized', tran, ['6A'] ).
verb( 'summer', 'summers', 'summering', 'summered', 'summered', intran, ['2C'] ).
verb( 'summon', 'summons', 'summoning', 'summoned', 'summoned', tran, ['6A','14','15B','17'] ).
verb( 'summons', 'summonses', 'summonsing', 'summonsed', 'summonsed', tran, ['6A','14','15B','17'] ).
verb( 'sun', 'suns', 'sunning', 'sunned', 'sunned', tran, ['6A'] ).
verb( 'sunbathe', 'sunbathes', 'sunbathing', 'sunbathed', 'sunbathed', intran, [] ).
verb( 'sunder', 'sunders', 'sundering', 'sundered', 'sundered', tran, ['6A'] ).
verb( 'sup', 'sups', 'supping', 'supped', 'supped', _, ['2A','2C','3A','6A','15B'] ).
verb( 'superannuate', 'superannuates', 'superannuating', 'superannuated', 'superannuated', tran, ['6A'] ).
verb( 'superimpose', 'superimposes', 'superimposing', 'superimposed', 'superimposed', tran, ['6A','14'] ).
verb( 'superintend', 'superintends', 'superintending', 'superintended', 'superintended', _, ['2A','6A'] ).
verb( 'supersede', 'supersedes', 'superseding', 'superseded', 'superseded', tran, ['6A'] ).
verb( 'supervene', 'supervenes', 'supervening', 'supervened', 'supervened', intran, ['2A'] ).
verb( 'supervise', 'supervises', 'supervising', 'supervised', 'supervised', _, ['2A','6A'] ).
verb( 'supplant', 'supplants', 'supplanting', 'supplanted', 'supplanted', tran, ['6A'] ).
verb( 'supplement', 'supplements', 'supplementing', 'supplemented', 'supplemented', tran, ['6A','15A'] ).
verb( 'supplicate', 'supplicates', 'supplicating', 'supplicated', 'supplicated', _, ['2C','6A','14','17'] ).
verb( 'supply', 'supplies', 'supplying', 'supplied', 'supplied', tran, ['6A','14'] ).
verb( 'support', 'supports', 'supporting', 'supported', 'supported', tran, ['6A'] ).
verb( 'suppose', 'supposes', 'supposing', 'supposed', 'supposed', tran, ['6A','9','25'] ).
verb( 'suppress', 'suppresses', 'suppressing', 'suppressed', 'suppressed', tran, ['6A'] ).
verb( 'suppurate', 'suppurates', 'suppurating', 'suppurated', 'suppurated', intran, ['2A'] ).
verb( 'surcharge', 'surcharges', 'surcharging', 'surcharged', 'surcharged', tran, ['6A','15A'] ).
verb( 'surface', 'surfaces', 'surfacing', 'surfaced', 'surfaced', _, ['2A','6A'] ).
verb( 'surfeit', 'surfeits', 'surfeiting', 'surfeited', 'surfeited', tran, ['6A','14'] ).
verb( 'surge', 'surges', 'surging', 'surged', 'surged', intran, ['2C'] ).
verb( 'surmise', 'surmises', 'surmising', 'surmised', 'surmised', _, ['2A','6A','9'] ).
verb( 'surmount', 'surmounts', 'surmounting', 'surmounted', 'surmounted', tran, ['6A'] ).
verb( 'surpass', 'surpasses', 'surpassing', 'surpassed', 'surpassed', tran, ['6A','15A'] ).
verb( 'surprise', 'surprises', 'surprising', 'surprised', 'surprised', tran, ['6A'] ).
verb( 'surrender', 'surrenders', 'surrendering', 'surrendered', 'surrendered', _, ['2A','6A','14'] ).
verb( 'surround', 'surrounds', 'surrounding', 'surrounded', 'surrounded', tran, ['6A'] ).
verb( 'surtax', 'surtaxes', 'surtaxing', 'surtaxed', 'surtaxed', tran, [] ).
verb( 'survey', 'surveys', 'surveying', 'surveyed', 'surveyed', tran, ['6A'] ).
verb( 'survive', 'survives', 'surviving', 'survived', 'survived', _, ['2A','6A'] ).
verb( 'suspect', 'suspects', 'suspecting', 'suspected', 'suspected', tran, ['6A','9','14','25'] ).
verb( 'suspend', 'suspends', 'suspending', 'suspended', 'suspended', tran, ['6A','14'] ).
verb( 'suss', 'susses', 'sussing', 'sussed', 'sussed', tran, ['15B'] ).
verb( 'sustain', 'sustains', 'sustaining', 'sustained', 'sustained', tran, ['6A'] ).
verb( 'swab', 'swabs', 'swabbing', 'swabbed', 'swabbed', tran, ['6A','15B'] ).
verb( 'swaddle', 'swaddles', 'swaddling', 'swaddled', 'swaddled', tran, ['6A'] ).
verb( 'swagger', 'swaggers', 'swaggering', 'swaggered', 'swaggered', intran, ['2A','2C'] ).
verb( 'swallow', 'swallows', 'swallowing', 'swallowed', 'swallowed', _, ['2C','6A','15B'] ).
verb( 'swamp', 'swamps', 'swamping', 'swamped', 'swamped', tran, ['6A','14'] ).
verb( 'swan', 'swans', 'swanning', 'swanned', 'swanned', intran, ['2C'] ).
verb( 'swank', 'swanks', 'swanking', 'swanked', 'swanked', intran, ['2A','2C'] ).
verb( 'swap', 'swaps', 'swapping', 'swapped', 'swapped', _, ['2A','6A','15A'] ).
verb( 'swarm', 'swarms', 'swarming', 'swarmed', 'swarmed', _, ['2A','2C','3A','6A','15B'] ).
verb( 'swat', 'swats', 'swatting', 'swatted', 'swatted', tran, [] ).
verb( 'swathe', 'swathes', 'swathing', 'swathed', 'swathed', tran, ['6A','15A'] ).
verb( 'sway', 'sways', 'swaying', 'swayed', 'swayed', _, ['2A','2C','6A','15A'] ).
verb( 'swear', 'swears', 'swearing', 'swore', 'sworn', _, ['2A','2B','2C','3A','6A','7A','9','14','15A','15B','22'] ).
verb( 'sweat', 'sweats', 'sweating', 'sweated', 'sweated', _, ['2A','2C','6A','15B'] ).
verb( 'sweep', 'sweeps', 'sweeping', 'swept', 'swept', _, ['2A','2C','6A','12A','15A','15B','22'] ).
verb( 'sweeten', 'sweetens', 'sweetening', 'sweetened', 'sweetened', _, ['2A','6A'] ).
verb( 'swell', 'swells', 'swelling', 'swelled', 'swelled', _, ['2A','2C','6A','14','15B'] ).
verb( 'swelter', 'swelters', 'sweltering', 'sweltered', 'sweltered', intran, ['2A'] ).
verb( 'swerve', 'swerves', 'swerving', 'swerved', 'swerved', _, ['2A','2C','4A','6A'] ).
verb( 'swig', 'swigs', 'swigging', 'swigged', 'swigged', _, ['2A','2C','6A','15B'] ).
verb( 'swill', 'swills', 'swilling', 'swilled', 'swilled', _, ['6A','15B'] ).
verb( 'swim', 'swims', 'swimming', 'swam', 'swum', _, ['2A','2B','2C','3A','6A'] ).
verb( 'swindle', 'swindles', 'swindling', 'swindled', 'swindled', _, ['6A','14'] ).
verb( 'swing', 'swings', 'swinging', 'swang', 'swung', _, ['2A','2B','2C','6A','6C','15A','15B','22'] ).
verb( 'swinge', 'swinges', 'swinging', 'swinged', 'swinged', tran, [] ).
verb( 'swipe', 'swipes', 'swiping', 'swiped', 'swiped', tran, ['3A','6A','15A'] ).
verb( 'swirl', 'swirls', 'swirling', 'swirled', 'swirled', _, ['2C','15B'] ).
verb( 'swish', 'swishes', 'swishing', 'swished', 'swished', _, ['2A','6A','15B'] ).
verb( 'switch', 'switches', 'switching', 'switched', 'switched', _, ['6A','15A','15B'] ).
verb( 'swivel', 'swivels', 'swivelling', 'swivelled', 'swivelled', _, ['2A','2C','6A','15B'] ).
verb( 'swob', 'swobs', 'swobbing', 'swobbed', 'swobbed', tran, [] ).
verb( 'swoon', 'swoons', 'swooning', 'swooned', 'swooned', intran, ['2A'] ).
verb( 'swoop', 'swoops', 'swooping', 'swooped', 'swooped', _, ['2A','2C','15B'] ).
verb( 'swop', 'swops', 'swopping', 'swopped', 'swopped', _, ['2A','6A','15A'] ).
verb( 'swot', 'swots', 'swotting', 'swotted', 'swotted', _, ['2A','2C','3A','15B'] ).
verb( 'syllabicate', 'syllabicates', 'syllabicating', 'syllabicated', 'syllabicated', tran, [] ).
verb( 'syllabify', 'syllabifies', 'syllabifying', 'syllabified', 'syllabified', tran, [] ).
verb( 'syllabize', 'syllabizes', 'syllabizing', 'syllabized', 'syllabized', tran, [] ).
verb( 'symbolize', 'symbolizes', 'symbolizing', 'symbolized', 'symbolized', tran, ['6A'] ).
verb( 'sympathize', 'sympathizes', 'sympathizing', 'sympathized', 'sympathized', intran, ['2A','3A'] ).
verb( 'synchronize', 'synchronizes', 'synchronizing', 'synchronized', 'synchronized', _, ['2A','6A'] ).
verb( 'syncopate', 'syncopates', 'syncopating', 'syncopated', 'syncopated', tran, ['6A'] ).
verb( 'syndicate', 'syndicates', 'syndicating', 'syndicated', 'syndicated', tran, ['6A'] ).
verb( 'synthesize', 'synthesizes', 'synthesizing', 'synthesized', 'synthesized', tran, ['6A'] ).
verb( 'syphon', 'syphons', 'syphoning', 'syphoned', 'syphoned', _, ['15B'] ).
verb( 'syringe', 'syringes', 'syringing', 'syringed', 'syringed', tran, ['6A','15B'] ).
verb( 'systematize', 'systematizes', 'systematizing', 'systematized', 'systematized', tran, ['6A'] ).
verb( 'table', 'tables', 'tabling', 'tabled', 'tabled', tran, ['6A'] ).
verb( 'taboo', 'taboos', 'tabooing', 'tabooed', 'tabooed', tran, ['6A'] ).
verb( 'tabulate', 'tabulates', 'tabulating', 'tabulated', 'tabulated', tran, ['6A'] ).
verb( 'tack', 'tacks', 'tacking', 'tacked', 'tacked', _, ['2A','2C','6A','15A','15B'] ).
verb( 'tackle', 'tackles', 'tackling', 'tackled', 'tackled', _, ['2A','6A','14'] ).
verb( 'tag', 'tags', 'tagging', 'tagged', 'tagged', _, ['2C','6A','14','15A','15B'] ).
verb( 'tail', 'tails', 'tailing', 'tailed', 'tailed', _, ['2C','3A','6A'] ).
verb( 'tailor', 'tailors', 'tailoring', 'tailored', 'tailored', tran, ['6A','15A'] ).
verb( 'taint', 'taints', 'tainting', 'tainted', 'tainted', _, ['2A','6A'] ).
verb( 'take', 'takes', 'taking', 'took', 'taken', _, ['2A','2B','2C','3A','6A','6B','12A','13A','14','15A','15B','16B','19B','22'] ).
verb( 'talk', 'talks', 'talking', 'talked', 'talked', _, ['2A','2B','2C','3A','6A','14','15B','22'] ).
verb( 'tally', 'tallies', 'tallying', 'tallied', 'tallied', intran, ['2A','3A'] ).
verb( 'tame', 'tames', 'taming', 'tamed', 'tamed', tran, ['6A'] ).
verb( 'tamp', 'tamps', 'tamping', 'tamped', 'tamped', tran, ['15B'] ).
verb( 'tamper', 'tampers', 'tampering', 'tampered', 'tampered', intran, ['3A'] ).
verb( 'tan', 'tans', 'tanning', 'tanned', 'tanned', _, ['2A','6A'] ).
verb( 'tangle', 'tangles', 'tangling', 'tangled', 'tangled', _, ['2A','2C','3A','6A','15B'] ).
verb( 'tank', 'tanks', 'tanking', 'tanked', 'tanked', intran, [] ).
verb( 'tantalize', 'tantalizes', 'tantalizing', 'tantalized', 'tantalized', tran, ['6A'] ).
verb( 'tap', 'taps', 'tapping', 'tapped', 'tapped', _, ['2A','2C','6A','14','15A','15B'] ).
verb( 'tape', 'tapes', 'taping', 'taped', 'taped', tran, ['6A'] ).
verb( 'taper', 'tapers', 'tapering', 'tapered', 'tapered', _, ['2A','2C','6A','15B'] ).
verb( 'tar', 'tars', 'tarring', 'tarred', 'tarred', tran, ['6A'] ).
verb( 'tarmac', 'tarmacs', 'tarmacking', 'tarmacked', 'tarmacked', tran, [] ).
verb( 'tarnish', 'tarnishes', 'tarnishing', 'tarnished', 'tarnished', _, ['2A','6A'] ).
verb( 'tarry', 'tarries', 'tarrying', 'tarried', 'tarried', intran, ['2A','2B','2C'] ).
verb( 'tart', 'tarts', 'tarting', 'tarted', 'tarted', tran, ['15B'] ).
verb( 'task', 'tasks', 'tasking', 'tasked', 'tasked', tran, ['6A'] ).
verb( 'taste', 'tastes', 'tasting', 'tasted', 'tasted', _, ['2A','2D','3A','6A'] ).
verb( 'tat', 'tats', 'tatting', 'tatted', 'tatted', _, [] ).
verb( 'tattle', 'tattles', 'tattling', 'tattled', 'tattled', _, ['2A'] ).
verb( 'tattoo', 'tattoos', 'tattooing', 'tattooed', 'tattooed', tran, ['6A'] ).
verb( 'taunt', 'taunts', 'taunting', 'taunted', 'taunted', tran, ['6A','14'] ).
verb( 'tax', 'taxes', 'taxing', 'taxed', 'taxed', tran, ['6A','14'] ).
verb( 'taxi', 'taxis', 'taxiing', 'taxied', 'taxied', _, ['2C','15A'] ).
verb( 'teach', 'teaches', 'teaching', 'taught', 'taught', _, ['2A','2B','2C','6A','11','12A','13A','17','20','21'] ).
verb( 'team', 'teams', 'teaming', 'teamed', 'teamed', intran, ['2C'] ).
verb( 'tear', 'tears', 'tearing', 'tore', 'torn', _, ['2A','2C','3A','6A','15A','15B','22'] ).
verb( 'tease', 'teases', 'teasing', 'teased', 'teased', tran, ['6A','15A','15B'] ).
verb( 'tee', 'tees', 'teeing', 'teed', 'teed', _, ['2A','2C','15B'] ).
verb( 'teem', 'teems', 'teeming', 'teemed', 'teemed', intran, ['2A','2C','3A'] ).
verb( 'teeter', 'teeters', 'teetering', 'teetered', 'teetered', intran, ['2C'] ).
verb( 'teethe', 'teethes', 'teething', 'teethed', 'teethed', intran, ['2A'] ).
verb( 'telecast', 'telecasts', 'telecasting', 'telecast', 'telecast', tran, [] ).
verb( 'telegraph', 'telegraphs', 'telegraphing', 'telegraphed', 'telegraphed', _, ['2A','6A','11','12A','13A'] ).
verb( 'telephone', 'telephones', 'telephoning', 'telephoned', 'telephoned', _, ['2A','4A','6A','11','12A','13A'] ).
verb( 'telescope', 'telescopes', 'telescoping', 'telescoped', 'telescoped', _, ['2A','6A'] ).
verb( 'televise', 'televises', 'televising', 'televised', 'televised', tran, ['6A'] ).
verb( 'tell', 'tells', 'telling', 'told', 'told', _, ['2A','3A','6A','8','10','11','12A','13A','14','15A','15B','17','20','21'] ).
verb( 'temper', 'tempers', 'tempering', 'tempered', 'tempered', _, ['2A','6A','15A'] ).
verb( 'temporize', 'temporizes', 'temporizing', 'temporized', 'temporized', intran, ['2A'] ).
verb( 'tempt', 'tempts', 'tempting', 'tempted', 'tempted', tran, ['6A','14','17'] ).
verb( 'tenant', 'tenants', 'tenanting', 'tenanted', 'tenanted', tran, [] ).
verb( 'tend', 'tends', 'tending', 'tended', 'tended', _, ['2C','4A','6A'] ).
verb( 'tender', 'tenders', 'tendering', 'tendered', 'tendered', _, ['2A','3A','6A','12A','13A'] ).
verb( 'tense', 'tenses', 'tensing', 'tensed', 'tensed', _, ['2A','6A'] ).
verb( 'tergiversate', 'tergiversates', 'tergiversating', 'tergiversated', 'tergiversated', intran, [] ).
verb( 'term', 'terms', 'terming', 'termed', 'termed', tran, ['23'] ).
verb( 'terminate', 'terminates', 'terminating', 'terminated', 'terminated', _, ['2A','6A','15A'] ).
verb( 'terrace', 'terraces', 'terracing', 'terraced', 'terraced', tran, ['6A'] ).
verb( 'terrify', 'terrifies', 'terrifying', 'terrified', 'terrified', tran, ['6A','15A'] ).
verb( 'terrorize', 'terrorizes', 'terrorizing', 'terrorized', 'terrorized', tran, ['6A'] ).
verb( 'test', 'tests', 'testing', 'tested', 'tested', tran, ['6A','15A'] ).
verb( 'test-drive', 'test-drives', 'test-driving', 'test-drove', 'test-driven', tran, [] ).
verb( 'testify', 'testifies', 'testifying', 'testified', 'testified', _, ['2A','3A','6A','9'] ).
verb( 'tether', 'tethers', 'tethering', 'tethered', 'tethered', tran, ['6A','15A'] ).
verb( 'thank', 'thanks', 'thanking', 'thanked', 'thanked', tran, ['6A','11','14','17'] ).
verb( 'thatch', 'thatches', 'thatching', 'thatched', 'thatched', tran, ['6A'] ).
verb( 'thaw', 'thaws', 'thawing', 'thawed', 'thawed', _, ['2A','2C','6A','15B'] ).
verb( 'theorize', 'theorizes', 'theorizing', 'theorized', 'theorized', intran, ['2A','3A'] ).
verb( 'thicken', 'thickens', 'thickening', 'thickened', 'thickened', _, [] ).
verb( 'thieve', 'thieves', 'thieving', 'thieved', 'thieved', _, ['2A','6A'] ).
verb( 'thin', 'thins', 'thinning', 'thinned', 'thinned', _, ['2A','2C','6A','15B'] ).
verb( 'think', 'thinks', 'thinking', 'thought', 'thought', _, ['2A','2B','2C','3A','6A','7A','8','9','10','15B','22','25'] ).
verb( 'thirst', 'thirsts', 'thirsting', 'thirsted', 'thirsted', tran, ['2A','3A'] ).
verb( 'thrash', 'thrashes', 'thrashing', 'thrashed', 'thrashed', _, ['2C','6A','15A','15B'] ).
verb( 'thread', 'threads', 'threading', 'threaded', 'threaded', tran, ['6A','15A'] ).
verb( 'threaten', 'threatens', 'threatening', 'threatened', 'threatened', _, ['2A','6A','14','17'] ).
verb( 'thresh', 'threshes', 'threshing', 'threshed', 'threshed', _, ['2A','6A','15A'] ).
verb( 'thrill', 'thrills', 'thrilling', 'thrilled', 'thrilled', _, ['2A','2C','6A'] ).
verb( 'thrive', 'thrives', 'thriving', 'thrived', 'thrived', intran, ['2A','3A'] ).
verb( 'throb', 'throbs', 'throbbing', 'throbbed', 'throbbed', intran, ['2A','2C'] ).
verb( 'throng', 'throngs', 'thronging', 'thronged', 'thronged', _, ['2C','4A','6A'] ).
verb( 'throttle', 'throttles', 'throttling', 'throttled', 'throttled', _, ['2C','6A','15B'] ).
verb( 'throw', 'throws', 'throwing', 'threw', 'thrown', _, ['2A','6A','12A','13A','15A','15B','22'] ).
verb( 'thrum', 'thrums', 'thrumming', 'thrummed', 'thrummed', _, ['3A','6A'] ).
verb( 'thrust', 'thrusts', 'thrusting', 'thrust', 'thrust', _, ['2A','2C','6A','15A','15B'] ).
verb( 'thud', 'thuds', 'thudding', 'thudded', 'thudded', intran, ['2C'] ).
verb( 'thumb', 'thumbs', 'thumbing', 'thumbed', 'thumbed', tran, ['6A'] ).
verb( 'thump', 'thumps', 'thumping', 'thumped', 'thumped', _, ['2A','2C','3A','6A','15A','22'] ).
verb( 'thunder', 'thunders', 'thundering', 'thundered', 'thundered', _, ['2A','2C','3A','15B'] ).
verb( 'thwack', 'thwacks', 'thwacking', 'thwacked', 'thwacked', tran, [] ).
verb( 'thwart', 'thwarts', 'thwarting', 'thwarted', 'thwarted', tran, ['6A'] ).
verb( 'tick', 'ticks', 'ticking', 'ticked', 'ticked', _, ['2A','2C','6A','10','15B'] ).
verb( 'ticket', 'tickets', 'ticketing', 'ticketed', 'ticketed', tran, ['6A'] ).
verb( 'tickle', 'tickles', 'tickling', 'tickled', 'tickled', _, ['2A','6A'] ).
verb( 'tide', 'tides', 'tiding', 'tided', 'tided', tran, ['14','15B'] ).
verb( 'tidy', 'tidies', 'tidying', 'tidied', 'tidied', _, ['2A','2C','6A','15B'] ).
verb( 'tie', 'ties', 'tying', 'tied', 'tied', _, ['2A','3A','6A','15A','15B'] ).
verb( 'tighten', 'tightens', 'tightening', 'tightened', 'tightened', _, ['2A','2C','6A','15B'] ).
verb( 'tile', 'tiles', 'tiling', 'tiled', 'tiled', tran, ['6A'] ).
verb( 'till', 'tills', 'tilling', 'tilled', 'tilled', tran, ['6A'] ).
verb( 'tilt', 'tilts', 'tilting', 'tilted', 'tilted', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'time', 'times', 'timing', 'timed', 'timed', tran, ['6A','15A'] ).
verb( 'tin', 'tins', 'tinning', 'tinned', 'tinned', tran, ['6A'] ).
verb( 'tincture', 'tinctures', 'tincturing', 'tinctured', 'tinctured', tran, [] ).
verb( 'ting', 'tings', 'tinging', 'tinged', 'tinged', _, [] ).
verb( 'tinge', 'tinges', 'tinging', 'tinged', 'tinged', tran, ['6A','14'] ).
verb( 'tingle', 'tingles', 'tingling', 'tingled', 'tingled', intran, ['2A','2C'] ).
verb( 'tinker', 'tinkers', 'tinkering', 'tinkered', 'tinkered', intran, ['2A','2C','3A'] ).
verb( 'tinkle', 'tinkles', 'tinkling', 'tinkled', 'tinkled', _, ['2A','2C','6A'] ).
verb( 'tinsel', 'tinsels', 'tinselling', 'tinselled', 'tinselled', tran, [] ).
verb( 'tint', 'tints', 'tinting', 'tinted', 'tinted', tran, ['6A','22'] ).
verb( 'tip', 'tips', 'tipping', 'tipped', 'tipped', _, ['2A','2C','6A','12C','15A','15B'] ).
verb( 'tipple', 'tipples', 'tippling', 'tippled', 'tippled', _, ['2A','6A'] ).
verb( 'tiptoe', 'tiptoes', 'tiptoeing', 'tiptoed', 'tiptoed', intran, ['2A','2C'] ).
verb( 'tire', 'tires', 'tiring', 'tired', 'tired', _, ['2A','3A','6A','15B'] ).
verb( 'titillate', 'titillates', 'titillating', 'titillated', 'titillated', tran, ['6A'] ).
verb( 'titivate', 'titivates', 'titivating', 'titivated', 'titivated', _, ['2A','6A'] ).
verb( 'titter', 'titters', 'tittering', 'tittered', 'tittered', intran, ['2A'] ).
verb( 'tittivate', 'tittivates', 'tittivating', 'tittivated', 'tittivated', _, ['2A','6A'] ).
verb( 'tittle-tattle', 'tittle-tattles', 'tittle-tattling', 'tittle-tattled', 'tittle-tattled', intran, [] ).
verb( 'toady', 'toadies', 'toadying', 'toadied', 'toadied', intran, ['2A','3A'] ).
verb( 'toast', 'toasts', 'toasting', 'toasted', 'toasted', _, ['2A','6A'] ).
verb( 'toboggan', 'toboggans', 'tobogganing', 'tobogganed', 'tobogganed', intran, ['2A','2C'] ).
verb( 'toddle', 'toddles', 'toddling', 'toddled', 'toddled', intran, ['2A','2C'] ).
verb( 'toe', 'toes', 'toeing', 'toed', 'toed', tran, ['6A'] ).
verb( 'tog', 'togs', 'togging', 'togged', 'togged', tran, ['15B'] ).
verb( 'toil', 'toils', 'toiling', 'toiled', 'toiled', intran, ['2A','2B','2C','3A','4A'] ).
verb( 'tolerate', 'tolerates', 'tolerating', 'tolerated', 'tolerated', tran, ['6A','6C'] ).
verb( 'toll', 'tolls', 'tolling', 'tolled', 'tolled', _, ['2A','6A'] ).
verb( 'tomahawk', 'tomahawks', 'tomahawking', 'tomahawked', 'tomahawked', tran, [] ).
verb( 'ton', 'tons', 'tonning', 'tonned', 'tonned', intran, ['2C'] ).
verb( 'tone', 'tones', 'toning', 'toned', 'toned', _, ['2C','6A','15B'] ).
verb( 'tonsure', 'tonsures', 'tonsuring', 'tonsured', 'tonsured', tran, [] ).
verb( 'tool', 'tools', 'tooling', 'tooled', 'tooled', tran, ['2C','6A'] ).
verb( 'toot', 'toots', 'tooting', 'tooted', 'tooted', _, ['2A','6A'] ).
verb( 'tootle', 'tootles', 'tootling', 'tootled', 'tootled', intran, [] ).
verb( 'top', 'tops', 'topping', 'topped', 'topped', tran, ['6A'] ).
verb( 'top-dress', 'top-dresses', 'top-dressing', 'top-dressed', 'top-dressed', tran, [] ).
verb( 'tope', 'topes', 'toping', 'toped', 'toped', _, ['2A','6A'] ).
verb( 'topple', 'topples', 'toppling', 'toppled', 'toppled', _, ['2A','2C','6A','15B'] ).
verb( 'torment', 'torments', 'tormenting', 'tormented', 'tormented', tran, ['6A','15A'] ).
verb( 'torpedo', 'torpedos', 'torpedoing', 'torpedoed', 'torpedoed', tran, ['6A'] ).
verb( 'torture', 'tortures', 'torturing', 'tortured', 'tortured', tran, ['6A','16A'] ).
verb( 'toss', 'tosses', 'tossing', 'tossed', 'tossed', _, ['2C','6A','12A','13A','15A','15B'] ).
verb( 'tot', 'tots', 'totting', 'totted', 'totted', _, ['2C','15B'] ).
verb( 'total', 'totals', 'totalling', 'totalled', 'totalled', _, ['2C','6A'] ).
verb( 'tote', 'totes', 'toting', 'toted', 'toted', tran, ['6A'] ).
verb( 'totter', 'totters', 'tottering', 'tottered', 'tottered', intran, ['2A','2C'] ).
verb( 'touch', 'touches', 'touching', 'touched', 'touched', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'touch-type', 'touch-types', 'touch-typing', 'touch-typed', 'touch-typed', intran, ['2A'] ).
verb( 'toughen', 'toughens', 'toughening', 'toughened', 'toughened', _, [] ).
verb( 'tour', 'tours', 'touring', 'toured', 'toured', _, ['2A','2C','6A'] ).
verb( 'tousle', 'tousles', 'tousling', 'tousled', 'tousled', tran, ['6A'] ).
verb( 'tout', 'touts', 'touting', 'touted', 'touted', intran, ['2A','3A'] ).
verb( 'tow', 'tows', 'towing', 'towed', 'towed', tran, ['6A','15A','15B'] ).
verb( 'towel', 'towels', 'towelling', 'towelled', 'towelled', tran, [] ).
verb( 'tower', 'towers', 'towering', 'towered', 'towered', intran, ['2C'] ).
verb( 'toy', 'toys', 'toying', 'toyed', 'toyed', intran, ['3A'] ).
verb( 'trace', 'traces', 'tracing', 'traced', 'traced', _, ['6A','15A','15B'] ).
verb( 'track', 'tracks', 'tracking', 'tracked', 'tracked', tran, ['6A','15A','15B'] ).
verb( 'trade', 'trades', 'trading', 'traded', 'traded', _, ['2A','2C','3A','14','15B'] ).
verb( 'traduce', 'traduces', 'traducing', 'traduced', 'traduced', tran, ['6A'] ).
verb( 'traffic', 'traffics', 'trafficking', 'trafficked', 'trafficked', intran, ['3A'] ).
verb( 'trail', 'trails', 'trailing', 'trailed', 'trailed', _, ['2A','2C','6A','15A','15B'] ).
verb( 'train', 'trains', 'training', 'trained', 'trained', _, ['2C','3A','6A','14','15A','17'] ).
verb( 'traipse', 'traipses', 'traipsing', 'traipsed', 'traipsed', intran, ['2A','2B','2C'] ).
verb( 'trammel', 'trammels', 'trammelling', 'trammelled', 'trammelled', tran, ['6A'] ).
verb( 'tramp', 'tramps', 'tramping', 'tramped', 'tramped', _, ['2A','2B','2C','6A'] ).
verb( 'trample', 'tramples', 'trampling', 'trampled', 'trampled', _, ['2C','3A','6A','15B'] ).
verb( 'tranquilize', 'tranquilizes', 'tranquilizing', 'tranquilized', 'tranquilized', tran, ['6A'] ).
verb( 'tranquillize', 'tranquillizes', 'tranquillizing', 'tranquillized', 'tranquillized', tran, ['6A'] ).
verb( 'transact', 'transacts', 'transacting', 'transacted', 'transacted', tran, ['6A','14'] ).
verb( 'transcend', 'transcends', 'transcending', 'transcended', 'transcended', tran, ['6A'] ).
verb( 'transcribe', 'transcribes', 'transcribing', 'transcribed', 'transcribed', tran, ['6A'] ).
verb( 'transfer', 'transfers', 'transferring', 'transferred', 'transferred', _, ['3A','6A','14'] ).
verb( 'transfigure', 'transfigures', 'transfiguring', 'transfigured', 'transfigured', tran, ['6A'] ).
verb( 'transfix', 'transfixes', 'transfixing', 'transfixed', 'transfixed', tran, ['6A'] ).
verb( 'transform', 'transforms', 'transforming', 'transformed', 'transformed', tran, ['6A','14'] ).
verb( 'transfuse', 'transfuses', 'transfusing', 'transfused', 'transfused', tran, ['6A'] ).
verb( 'transgress', 'transgresses', 'transgressing', 'transgressed', 'transgressed', _, ['2A','6A'] ).
verb( 'translate', 'translates', 'translating', 'translated', 'translated', tran, ['6A','14'] ).
verb( 'transliterate', 'transliterates', 'transliterating', 'transliterated', 'transliterated', tran, ['6A','14'] ).
verb( 'transmit', 'transmits', 'transmitting', 'transmitted', 'transmitted', tran, ['6A','14'] ).
verb( 'transmogrify', 'transmogrifies', 'transmogrifying', 'transmogrified', 'transmogrified', tran, ['6A'] ).
verb( 'transmute', 'transmutes', 'transmuting', 'transmuted', 'transmuted', tran, ['6A','14'] ).
verb( 'transpire', 'transpires', 'transpiring', 'transpired', 'transpired', _, ['2A','6A'] ).
verb( 'transplant', 'transplants', 'transplanting', 'transplanted', 'transplanted', _, ['2A','6A'] ).
verb( 'transport', 'transports', 'transporting', 'transported', 'transported', tran, ['6A','15A'] ).
verb( 'transpose', 'transposes', 'transposing', 'transposed', 'transposed', tran, ['6A','14'] ).
verb( 'transship', 'transships', 'transshipping', 'transshipped', 'transshipped', tran, ['6A'] ).
verb( 'trap', 'traps', 'trapping', 'trapped', 'trapped', tran, ['6A','15A'] ).
verb( 'travel', 'travels', 'travelling', 'travelled', 'travelled', _, ['2A','2B','2C','3A','4A'] ).
verb( 'traverse', 'traverses', 'traversing', 'traversed', 'traversed', tran, ['6A'] ).
verb( 'travesty', 'travesties', 'travestying', 'travestied', 'travestied', tran, ['6A'] ).
verb( 'trawl', 'trawls', 'trawling', 'trawled', 'trawled', _, ['2A','6A'] ).
verb( 'tread', 'treads', 'treading', 'trod', 'trodden', _, ['2C','3A','6A','15A','15B'] ).
verb( 'treadle', 'treadles', 'treadling', 'treadled', 'treadled', intran, ['2A'] ).
verb( 'treasure', 'treasures', 'treasuring', 'treasured', 'treasured', tran, ['6A','15B'] ).
verb( 'treat', 'treats', 'treating', 'treated', 'treated', _, ['3A','6A','14','15A','16B'] ).
verb( 'treble', 'trebles', 'trebling', 'trebled', 'trebled', _, ['2A','6A'] ).
verb( 'tree', 'trees', 'treeing', 'treed', 'treed', tran, ['6A'] ).
verb( 'trek', 'treks', 'trekking', 'trekked', 'trekked', intran, ['2A','2B','2C'] ).
verb( 'trellis', 'trellises', 'trellising', 'trellised', 'trellised', tran, ['6A'] ).
verb( 'tremble', 'trembles', 'trembling', 'trembled', 'trembled', intran, ['2A','2B','2C','4B'] ).
verb( 'trench', 'trenches', 'trenching', 'trenched', 'trenched', tran, ['6A'] ).
verb( 'trend', 'trends', 'trending', 'trended', 'trended', intran, ['2C'] ).
verb( 'trepan', 'trepans', 'trepanning', 'trepanned', 'trepanned', tran, [] ).
verb( 'trephine', 'trephines', 'trephining', 'trephined', 'trephined', tran, ['6A'] ).
verb( 'trespass', 'trespasses', 'trespassing', 'trespassed', 'trespassed', intran, ['2A','3A'] ).
verb( 'trice', 'trices', 'tricing', 'triced', 'triced', tran, ['15B'] ).
verb( 'trick', 'tricks', 'tricking', 'tricked', 'tricked', tran, ['6A','14','15B'] ).
verb( 'trickle', 'trickles', 'trickling', 'trickled', 'trickled', _, ['2A','2C','15A'] ).
verb( 'trifle', 'trifles', 'trifling', 'trifled', 'trifled', _, ['3A','15B'] ).
verb( 'trigger', 'triggers', 'triggering', 'triggered', 'triggered', tran, ['15B'] ).
verb( 'trill', 'trills', 'trilling', 'trilled', 'trilled', _, ['2A','2C','6A'] ).
verb( 'trim', 'trims', 'trimming', 'trimmed', 'trimmed', _, ['2A','6A','14','15A','22'] ).
verb( 'trip', 'trips', 'tripping', 'tripped', 'tripped', _, ['2A','2C','3A','15B'] ).
verb( 'triple', 'triples', 'tripling', 'tripled', 'tripled', _, ['2A','6A'] ).
verb( 'triplicate', 'triplicates', 'triplicating', 'triplicated', 'triplicated', tran, ['6A'] ).
verb( 'trisect', 'trisects', 'trisecting', 'trisected', 'trisected', tran, ['6A'] ).
verb( 'triumph', 'triumphs', 'triumphing', 'triumphed', 'triumphed', intran, ['2A','3A'] ).
verb( 'trivialize', 'trivializes', 'trivializing', 'trivialized', 'trivialized', tran, ['6A'] ).
verb( 'troll', 'trolls', 'trolling', 'trolled', 'trolled', _, ['2A','2C'] ).
verb( 'troop', 'troops', 'trooping', 'trooped', 'trooped', _, ['2C'] ).
verb( 'trot', 'trots', 'trotting', 'trotted', 'trotted', _, ['2A','2B','2C','15A','15B'] ).
verb( 'trouble', 'troubles', 'troubling', 'troubled', 'troubled', _, ['2A','2C','4A','6A','14','17'] ).
verb( 'trounce', 'trounces', 'trouncing', 'trounced', 'trounced', tran, ['6A'] ).
verb( 'truckle', 'truckles', 'truckling', 'truckled', 'truckled', intran, ['3A'] ).
verb( 'trudge', 'trudges', 'trudging', 'trudged', 'trudged', intran, ['2A','2B','2C'] ).
verb( 'true', 'trues', 'truing', 'trued', 'trued', tran, ['15B'] ).
verb( 'trump', 'trumps', 'trumping', 'trumped', 'trumped', _, ['2A','6A','15B'] ).
verb( 'trumpet', 'trumpets', 'trumpeting', 'trumpeted', 'trumpeted', _, ['2A','2C','6A','15B'] ).
verb( 'truncate', 'truncates', 'truncating', 'truncated', 'truncated', tran, ['6A'] ).
verb( 'trundle', 'trundles', 'trundling', 'trundled', 'trundled', _, ['2C','6A','15A','15B'] ).
verb( 'truss', 'trusses', 'trussing', 'trussed', 'trussed', tran, ['6A','15B'] ).
verb( 'trust', 'trusts', 'trusting', 'trusted', 'trusted', _, ['3A','6A','7A','9','14','15A','15B','17'] ).
verb( 'try', 'tries', 'trying', 'tried', 'tried', _, ['2A','2B','3A','6A','6B','6C','7A','10','15A','15B'] ).
verb( 'tuck', 'tucks', 'tucking', 'tucked', 'tucked', _, ['2C','3A','15A','15B'] ).
verb( 'tug', 'tugs', 'tugging', 'tugged', 'tugged', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'tumble', 'tumbles', 'tumbling', 'tumbled', 'tumbled', _, ['2A','2C','3A','6A','15A','15B'] ).
verb( 'tune', 'tunes', 'tuning', 'tuned', 'tuned', _, ['2C','6A'] ).
verb( 'tunnel', 'tunnels', 'tunnelling', 'tunnelled', 'tunnelled', _, ['2A','2C','3A','6A'] ).
verb( 'turf', 'turfs', 'turfing', 'turfed', 'turfed', tran, ['6A','15B'] ).
verb( 'turn', 'turns', 'turning', 'turned', 'turned', _, ['2A','2C','3A','4A','6A','14','15A','15B'] ).
verb( 'tussle', 'tussles', 'tussling', 'tussled', 'tussled', intran, [] ).
verb( 'tut', 'tuts', 'tutting', 'tutted', 'tutted', tran, ['6A'] ).
verb( 'tutor', 'tutors', 'tutoring', 'tutored', 'tutored', tran, ['6A','15A','16A'] ).
verb( 'twaddle', 'twaddles', 'twaddling', 'twaddled', 'twaddled', intran, ['2A'] ).
verb( 'twang', 'twangs', 'twanging', 'twanged', 'twanged', _, ['2A','6A'] ).
verb( 'tweak', 'tweaks', 'tweaking', 'tweaked', 'tweaked', tran, ['6A'] ).
verb( 'tweet', 'tweets', 'tweeting', 'tweeted', 'tweeted', intran, [] ).
verb( 'twiddle', 'twiddles', 'twiddling', 'twiddled', 'twiddled', _, ['3A','6A'] ).
verb( 'twig', 'twigs', 'twigging', 'twigged', 'twigged', _, ['2A','6A'] ).
verb( 'twin', 'twins', 'twinning', 'twinned', 'twinned', tran, ['6A','14'] ).
verb( 'twine', 'twines', 'twining', 'twined', 'twined', _, ['2A','2C','15A','15B'] ).
verb( 'twinkle', 'twinkles', 'twinkling', 'twinkled', 'twinkled', intran, ['2A','2C'] ).
verb( 'twirl', 'twirls', 'twirling', 'twirled', 'twirled', _, ['2A','2C','6A','15B','16A'] ).
verb( 'twist', 'twists', 'twisting', 'twisted', 'twisted', _, ['2A','2C','6A','15A','15B','16A'] ).
verb( 'twit', 'twits', 'twitting', 'twitted', 'twitted', tran, ['6A','14'] ).
verb( 'twitch', 'twitches', 'twitching', 'twitched', 'twitched', _, ['2A','2C','6A','15A','15B'] ).
verb( 'twitter', 'twitters', 'twittering', 'twittered', 'twittered', intran, ['15A','15B'] ).
verb( 'type', 'types', 'typing', 'typed', 'typed', _, ['2A','6A'] ).
verb( 'typecast', 'typecasts', 'typecasting', 'typecast', 'typecast', tran, ['6A'] ).
verb( 'typify', 'typifies', 'typifying', 'typified', 'typified', tran, ['6A'] ).
verb( 'tyrannize', 'tyrannizes', 'tyrannizing', 'tyrannized', 'tyrannized', _, ['3A','6A'] ).
verb( 'uglify', 'uglifies', 'uglifying', 'uglified', 'uglified', tran, [] ).
verb( 'ulcerate', 'ulcerates', 'ulcerating', 'ulcerated', 'ulcerated', _, ['2A','6A'] ).
verb( 'ululate', 'ululates', 'ululating', 'ululated', 'ululated', intran, ['2A'] ).
verb( 'umpire', 'umpires', 'umpiring', 'umpired', 'umpired', _, ['2A','6A'] ).
verb( 'unbalance', 'unbalances', 'unbalancing', 'unbalanced', 'unbalanced', tran, ['6A'] ).
verb( 'unbar', 'unbars', 'unbarring', 'unbarred', 'unbarred', tran, [] ).
verb( 'unbend', 'unbends', 'unbending', 'unbended', 'unbended', _, ['2A','6A'] ).
verb( 'unbind', 'unbinds', 'unbinding', 'unbound', 'unbound', tran, ['6A'] ).
verb( 'unblock', 'unblocks', 'unblocking', 'unblocked', 'unblocked', tran, ['6A','15B'] ).
verb( 'unbolt', 'unbolts', 'unbolting', 'unbolted', 'unbolted', _, ['2A','2C','6A','15B'] ).
verb( 'unbosom', 'unbosoms', 'unbosoming', 'unbosomed', 'unbosomed', tran, ['6A','14'] ).
verb( 'unbuckle', 'unbuckles', 'unbuckling', 'unbuckled', 'unbuckled', tran, ['6A'] ).
verb( 'unburden', 'unburdens', 'unburdening', 'unburdened', 'unburdened', tran, ['6A','15A'] ).
verb( 'unbutton', 'unbuttons', 'unbuttoning', 'unbuttoned', 'unbuttoned', tran, ['6A'] ).
verb( 'unchain', 'unchains', 'unchaining', 'unchained', 'unchained', tran, ['6A','15A','15B'] ).
verb( 'uncork', 'uncorks', 'uncorking', 'uncorked', 'uncorked', tran, ['6A'] ).
verb( 'uncouple', 'uncouples', 'uncoupling', 'uncoupled', 'uncoupled', tran, ['6A'] ).
verb( 'uncover', 'uncovers', 'uncovering', 'uncovered', 'uncovered', tran, ['2A','6A'] ).
verb( 'uncross', 'uncrosses', 'uncrossing', 'uncrossed', 'uncrossed', tran, ['6A'] ).
verb( 'uncurl', 'uncurls', 'uncurling', 'uncurled', 'uncurled', _, ['2A','2C','6A','15A','15B'] ).
verb( 'undeceive', 'undeceives', 'undeceiving', 'undeceived', 'undeceived', tran, ['6A'] ).
verb( 'underact', 'underacts', 'underacting', 'underacted', 'underacted', _, ['2A','6A'] ).
verb( 'underbid', 'underbids', 'underbidding', 'underbid', 'underbid', tran, [] ).
verb( 'undercharge', 'undercharges', 'undercharging', 'undercharged', 'undercharged', tran, ['6A'] ).
verb( 'undercut', 'undercuts', 'undercutting', 'undercut', 'undercut', tran, ['6A'] ).
verb( 'underestimate', 'underestimates', 'underestimating', 'underestimated', 'underestimated', tran, ['6A'] ).
verb( 'underexpose', 'underexposes', 'underexposing', 'underexposed', 'underexposed', tran, ['6A'] ).
verb( 'undergo', 'undergoes', 'undergoing', 'underwent', 'undergone', tran, ['6A'] ).
verb( 'underlie', 'underlies', 'underlying', 'underlay', 'underlain', tran, ['6A'] ).
verb( 'underline', 'underlines', 'underlining', 'underlined', 'underlined', tran, ['6A'] ).
verb( 'undermine', 'undermines', 'undermining', 'undermined', 'undermined', tran, ['6A'] ).
verb( 'underpay', 'underpays', 'underpaying', 'underpaid', 'underpaid', tran, ['6A'] ).
verb( 'underpin', 'underpins', 'underpinning', 'underpinned', 'underpinned', tran, [] ).
verb( 'underquote', 'underquotes', 'underquoting', 'underquoted', 'underquoted', tran, ['6A'] ).
verb( 'underrate', 'underrates', 'underrating', 'underrated', 'underrated', tran, ['6A'] ).
verb( 'underscore', 'underscores', 'underscoring', 'underscored', 'underscored', tran, ['6A'] ).
verb( 'undersell', 'undersells', 'underselling', 'undersold', 'undersold', tran, ['6A'] ).
verb( 'undershoot', 'undershoots', 'undershooting', 'undershot', 'undershot', tran, [] ).
verb( 'undersign', 'undersigns', 'undersigning', 'undersigned', 'undersigned', tran, ['6A'] ).
verb( 'underspend', 'underspends', 'underspending', 'underspent', 'underspent', _, ['2A','6A','14','19B'] ).
verb( 'understand', 'understands', 'understanding', 'understood', 'understood', _, ['2A','6A','6C','8','9','10','17','19C'] ).
verb( 'understate', 'understates', 'understating', 'understated', 'understated', tran, ['6A'] ).
verb( 'understock', 'understocks', 'understocking', 'understocked', 'understocked', tran, ['6A'] ).
verb( 'understudy', 'understudies', 'understudying', 'understudied', 'understudied', tran, ['6A'] ).
verb( 'undertake', 'undertakes', 'undertaking', 'undertook', 'undertaken', tran, ['6A','7A','9'] ).
verb( 'undervalue', 'undervalues', 'undervaluing', 'undervalued', 'undervalued', tran, ['6A'] ).
verb( 'underwrite', 'underwrites', 'underwriting', 'underwrote', 'underwritten', tran, ['6A'] ).
verb( 'undo', 'undoes', 'undoing', 'undid', 'undone', tran, ['6A'] ).
verb( 'undock', 'undocks', 'undocking', 'undocked', 'undocked', _, ['2A','6A'] ).
verb( 'undress', 'undresses', 'undressing', 'undressed', 'undressed', _, ['2A','6A'] ).
verb( 'undulate', 'undulates', 'undulating', 'undulated', 'undulated', intran, ['2A','2C'] ).
verb( 'unearth', 'unearths', 'unearthing', 'unearthed', 'unearthed', tran, ['6A'] ).
verb( 'unfasten', 'unfastens', 'unfastening', 'unfastened', 'unfastened', _, ['2A','2C','3A','6A','14','15A','15B'] ).
verb( 'unfit', 'unfits', 'unfitting', 'unfitted', 'unfitted', tran, ['14'] ).
verb( 'unfold', 'unfolds', 'unfolding', 'unfolded', 'unfolded', _, ['2A','2C','6A'] ).
verb( 'unfrock', 'unfrocks', 'unfrocking', 'unfrocked', 'unfrocked', tran, ['6A'] ).
verb( 'unfurl', 'unfurls', 'unfurling', 'unfurled', 'unfurled', _, ['2A','6A'] ).
verb( 'unhand', 'unhands', 'unhanding', 'unhanded', 'unhanded', tran, ['6A'] ).
verb( 'unhinge', 'unhinges', 'unhinging', 'unhinged', 'unhinged', tran, ['6A'] ).
verb( 'unhitch', 'unhitches', 'unhitching', 'unhitched', 'unhitched', _, ['2A','2C','6A','15A','15B'] ).
verb( 'unhook', 'unhooks', 'unhooking', 'unhooked', 'unhooked', tran, ['6A'] ).
verb( 'unhorse', 'unhorses', 'unhorsing', 'unhorsed', 'unhorsed', tran, ['6A'] ).
verb( 'unify', 'unifies', 'unifying', 'unified', 'unified', tran, ['6A'] ).
verb( 'unite', 'unites', 'uniting', 'united', 'united', _, ['2A','3A','4A','6A'] ).
verb( 'unlearn', 'unlearns', 'unlearning', 'unlearned', 'unlearned', tran, ['6A'] ).
verb( 'unleash', 'unleashes', 'unleashing', 'unleashed', 'unleashed', tran, ['6A'] ).
verb( 'unload', 'unloads', 'unloading', 'unloaded', 'unloaded', _, ['2A','6A','14'] ).
verb( 'unlock', 'unlocks', 'unlocking', 'unlocked', 'unlocked', _, ['2A','2C','6A','15A','15B'] ).
verb( 'unloose', 'unlooses', 'unloosing', 'unloosed', 'unloosed', tran, ['6A'] ).
verb( 'unman', 'unmans', 'unmanning', 'unmanned', 'unmanned', tran, ['6A'] ).
verb( 'unmask', 'unmasks', 'unmasking', 'unmasked', 'unmasked', _, ['2A','6A'] ).
verb( 'unnerve', 'unnerves', 'unnerving', 'unnerved', 'unnerved', tran, ['6A'] ).
verb( 'unpack', 'unpacks', 'unpacking', 'unpacked', 'unpacked', _, ['2A','6A'] ).
verb( 'unpick', 'unpicks', 'unpicking', 'unpicked', 'unpicked', _, ['3A','6A','15B'] ).
verb( 'unplug', 'unplugs', 'unplugging', 'unplugged', 'unplugged', _, ['2C','6A','15B'] ).
verb( 'unravel', 'unravels', 'unravelling', 'unravelled', 'unravelled', _, ['2A','6A'] ).
verb( 'unroll', 'unrolls', 'unrolling', 'unrolled', 'unrolled', _, ['2A','6A'] ).
verb( 'unsaddle', 'unsaddles', 'unsaddling', 'unsaddled', 'unsaddled', tran, ['6A'] ).
verb( 'unsay', 'unsays', 'unsaying', 'unsaid', 'unsaid', tran, ['6A'] ).
verb( 'unscramble', 'unscrambles', 'unscrambling', 'unscrambled', 'unscrambled', tran, [] ).
verb( 'unscrew', 'unscrews', 'unscrewing', 'unscrewed', 'unscrewed', _, ['2A','6A','15A','15B'] ).
verb( 'unseat', 'unseats', 'unseating', 'unseated', 'unseated', tran, ['6A'] ).
verb( 'unsettle', 'unsettles', 'unsettling', 'unsettled', 'unsettled', tran, ['6A'] ).
verb( 'unsex', 'unsexes', 'unsexing', 'unsexed', 'unsexed', tran, ['6A'] ).
verb( 'unsheathe', 'unsheathes', 'unsheathing', 'unsheathed', 'unsheathed', tran, ['6A'] ).
verb( 'untie', 'unties', 'untying', 'untied', 'untied', _, ['2A','3A','6A','15A','15B'] ).
verb( 'unveil', 'unveils', 'unveiling', 'unveiled', 'unveiled', _, ['2A','6A'] ).
verb( 'unwind', 'unwinds', 'unwinding', 'unwound', 'unwound', _, ['2A','6A','21'] ).
verb( 'unwrap', 'unwraps', 'unwrapping', 'unwrapped', 'unwrapped', _, ['6A','14','15A','15B'] ).
verb( 'unzip', 'unzips', 'unzipping', 'unzipped', 'unzipped', tran, [] ).
verb( 'up', 'ups', 'upping', 'upped', 'upped', _, ['2A','6A'] ).
verb( 'upbraid', 'upbraids', 'upbraiding', 'upbraided', 'upbraided', tran, ['6A','14'] ).
verb( 'update', 'updates', 'updating', 'updated', 'updated', tran, ['6A'] ).
verb( 'upgrade', 'upgrades', 'upgrading', 'upgraded', 'upgraded', tran, ['6A'] ).
verb( 'uphold', 'upholds', 'upholding', 'upheld', 'upheld', tran, ['6A'] ).
verb( 'upholster', 'upholsters', 'upholstering', 'upholstered', 'upholstered', tran, ['6A'] ).
verb( 'uplift', 'uplifts', 'uplifting', 'uplifted', 'uplifted', tran, ['6A'] ).
verb( 'uproot', 'uproots', 'uprooting', 'uprooted', 'uprooted', tran, ['6A'] ).
verb( 'upset', 'upsets', 'upsetting', 'upset', 'upset', _, ['2A','6A'] ).
verb( 'upstage', 'upstages', 'upstaging', 'upstaged', 'upstaged', tran, ['6A'] ).
verb( 'urbanize', 'urbanizes', 'urbanizing', 'urbanized', 'urbanized', tran, ['6A'] ).
verb( 'urge', 'urges', 'urging', 'urged', 'urged', tran, ['6A','6D','9','14','15B','17','19C'] ).
verb( 'urinate', 'urinates', 'urinating', 'urinated', 'urinated', intran, [] ).
verb( 'use', 'uses', 'using', 'used', 'used', tran, ['6A','14','15A','15B','16A'] ).
verb( 'usher', 'ushers', 'ushering', 'ushered', 'ushered', tran, ['14','15B'] ).
verb( 'usurp', 'usurps', 'usurping', 'usurped', 'usurped', tran, ['6A'] ).
verb( 'utilize', 'utilizes', 'utilizing', 'utilized', 'utilized', tran, ['6A'] ).
verb( 'utter', 'utters', 'uttering', 'uttered', 'uttered', tran, ['6A'] ).
verb( 'vacate', 'vacates', 'vacating', 'vacated', 'vacated', tran, ['6A'] ).
verb( 'vacation', 'vacations', 'vacationing', 'vacationed', 'vacationed', intran, ['3A'] ).
verb( 'vaccinate', 'vaccinates', 'vaccinating', 'vaccinated', 'vaccinated', tran, ['6A','14'] ).
verb( 'vacillate', 'vacillates', 'vacillating', 'vacillated', 'vacillated', intran, ['2A','3A'] ).
verb( 'valet', 'valets', 'valeting', 'valeted', 'valeted', tran, ['6A'] ).
verb( 'validate', 'validates', 'validating', 'validated', 'validated', tran, ['6A'] ).
verb( 'value', 'values', 'valuing', 'valued', 'valued', tran, ['6A','15A','16B'] ).
verb( 'vamoose', 'vamooses', 'vamoosing', 'vamoosed', 'vamoosed', intran, [] ).
verb( 'vamp', 'vamps', 'vamping', 'vamped', 'vamped', _, ['2A','6A','15B'] ).
verb( 'vanish', 'vanishes', 'vanishing', 'vanished', 'vanished', intran, ['2A'] ).
verb( 'vanquish', 'vanquishes', 'vanquishing', 'vanquished', 'vanquished', tran, ['6A'] ).
verb( 'vaporize', 'vaporizes', 'vaporizing', 'vaporized', 'vaporized', _, ['2A','6A'] ).
verb( 'varnish', 'varnishes', 'varnishing', 'varnished', 'varnished', tran, ['6A'] ).
verb( 'vary', 'varies', 'varying', 'varied', 'varied', _, ['2A','6A'] ).
verb( 'vault', 'vaults', 'vaulting', 'vaulted', 'vaulted', _, ['2A','2B','2C','6A'] ).
verb( 'vaunt', 'vaunts', 'vaunting', 'vaunted', 'vaunted', _, [] ).
verb( 'veer', 'veers', 'veering', 'veered', 'veered', intran, ['2A','2C'] ).
verb( 'vegetate', 'vegetates', 'vegetating', 'vegetated', 'vegetated', intran, ['2A'] ).
verb( 'veil', 'veils', 'veiling', 'veiled', 'veiled', tran, ['6A'] ).
verb( 'vend', 'vends', 'vending', 'vended', 'vended', tran, ['6A'] ).
verb( 'veneer', 'veneers', 'veneering', 'veneered', 'veneered', tran, ['6A'] ).
verb( 'venerate', 'venerates', 'venerating', 'venerated', 'venerated', tran, ['6A'] ).
verb( 'vent', 'vents', 'venting', 'vented', 'vented', tran, ['6A','14'] ).
verb( 'ventilate', 'ventilates', 'ventilating', 'ventilated', 'ventilated', tran, ['6A'] ).
verb( 'venture', 'ventures', 'venturing', 'ventured', 'ventured', _, ['3A','6A','7A','15A','16A'] ).
verb( 'verbalize', 'verbalizes', 'verbalizing', 'verbalized', 'verbalized', tran, [] ).
verb( 'verge', 'verges', 'verging', 'verged', 'verged', intran, ['3A'] ).
verb( 'verify', 'verifies', 'verifying', 'verified', 'verified', tran, ['6A'] ).
verb( 'versify', 'versifies', 'versifying', 'versified', 'versified', _, ['2A','6A'] ).
verb( 'vest', 'vests', 'vesting', 'vested', 'vested', _, ['3A','6A','14'] ).
verb( 'vesture', 'vestures', 'vesturing', 'vestured', 'vestured', tran, [] ).
verb( 'vet', 'vets', 'vetting', 'vetted', 'vetted', tran, ['6A'] ).
verb( 'veto', 'vetos', 'vetoing', 'vetoed', 'vetoed', tran, ['6A'] ).
verb( 'vex', 'vexes', 'vexing', 'vexed', 'vexed', tran, ['6A'] ).
verb( 'vibrate', 'vibrates', 'vibrating', 'vibrated', 'vibrated', _, ['2A','2C','6A'] ).
verb( 'victimize', 'victimizes', 'victimizing', 'victimized', 'victimized', tran, ['6A'] ).
verb( 'victual', 'victuals', 'victualling', 'victualled', 'victualled', _, ['2A','6A'] ).
verb( 'videotape', 'videotapes', 'videotaping', 'videotaped', 'videotaped', tran, ['6A'] ).
verb( 'vie', 'vies', 'vying', 'vied', 'vied', intran, ['3A'] ).
verb( 'view', 'views', 'viewing', 'viewed', 'viewed', tran, ['6A'] ).
verb( 'vilify', 'vilifies', 'vilifying', 'vilified', 'vilified', tran, ['6A'] ).
verb( 'vindicate', 'vindicates', 'vindicating', 'vindicated', 'vindicated', tran, ['6A'] ).
verb( 'violate', 'violates', 'violating', 'violated', 'violated', tran, ['6A'] ).
verb( 'visa', 'visas', 'visaing', 'visaed', 'visaed', tran, ['6A'] ).
verb( 'visit', 'visits', 'visiting', 'visited', 'visited', _, ['2C','3A','6A','14'] ).
verb( 'visualize', 'visualizes', 'visualizing', 'visualized', 'visualized', tran, ['6A'] ).
verb( 'vitalize', 'vitalizes', 'vitalizing', 'vitalized', 'vitalized', tran, ['6A'] ).
verb( 'vitiate', 'vitiates', 'vitiating', 'vitiated', 'vitiated', tran, ['6A'] ).
verb( 'vitrify', 'vitrifies', 'vitrifying', 'vitrified', 'vitrified', _, ['2A','6A'] ).
verb( 'vituperate', 'vituperates', 'vituperating', 'vituperated', 'vituperated', tran, ['6A'] ).
verb( 'vivisect', 'vivisects', 'vivisecting', 'vivisected', 'vivisected', tran, ['6A'] ).
verb( 'vocalize', 'vocalizes', 'vocalizing', 'vocalized', 'vocalized', tran, [] ).
verb( 'vociferate', 'vociferates', 'vociferating', 'vociferated', 'vociferated', _, ['2A','6A'] ).
verb( 'voice', 'voices', 'voicing', 'voiced', 'voiced', tran, ['6A'] ).
verb( 'void', 'voids', 'voiding', 'voided', 'voided', tran, ['6A'] ).
verb( 'volley', 'volleys', 'volleying', 'volleyed', 'volleyed', _, ['2A','2C','6A'] ).
verb( 'volunteer', 'volunteers', 'volunteering', 'volunteered', 'volunteered', _, ['2A','3A','6A','7A'] ).
verb( 'vomit', 'vomits', 'vomiting', 'vomited', 'vomited', _, ['2A','6A','15B'] ).
verb( 'vote', 'votes', 'voting', 'voted', 'voted', _, ['3A','6A','9','12B','13B','15B','25'] ).
verb( 'vouch', 'vouches', 'vouching', 'vouched', 'vouched', intran, [] ).
verb( 'vouchsafe', 'vouchsafes', 'vouchsafing', 'vouchsafed', 'vouchsafed', tran, ['6A','7A','12C'] ).
verb( 'vow', 'vows', 'vowing', 'vowed', 'vowed', tran, ['6A','7A','9'] ).
verb( 'voyage', 'voyages', 'voyaging', 'voyaged', 'voyaged', intran, ['2A','2C'] ).
verb( 'vulcanize', 'vulcanizes', 'vulcanizing', 'vulcanized', 'vulcanized', tran, ['6A'] ).
verb( 'vulgarize', 'vulgarizes', 'vulgarizing', 'vulgarized', 'vulgarized', tran, ['6A'] ).
verb( 'wad', 'wads', 'wadding', 'wadded', 'wadded', tran, [] ).
verb( 'waddle', 'waddles', 'waddling', 'waddled', 'waddled', intran, ['2A','2C'] ).
verb( 'wade', 'wades', 'wading', 'waded', 'waded', _, ['2A','2C','6A'] ).
verb( 'waffle', 'waffles', 'waffling', 'waffled', 'waffled', intran, ['2A','2C'] ).
verb( 'waft', 'wafts', 'wafting', 'wafted', 'wafted', tran, ['6A'] ).
verb( 'wag', 'wags', 'wagging', 'wagged', 'wagged', _, ['2A','2C','6A'] ).
verb( 'wage', 'wages', 'waging', 'waged', 'waged', tran, ['6A'] ).
verb( 'wager', 'wagers', 'wagering', 'wagered', 'wagered', _, ['2A','6A','11','12C','14'] ).
verb( 'waggle', 'waggles', 'waggling', 'waggled', 'waggled', _, [] ).
verb( 'wail', 'wails', 'wailing', 'wailed', 'wailed', _, ['2A','2B','2C','6A'] ).
verb( 'wait', 'waits', 'waiting', 'waited', 'waited', _, ['2A','2B','2C','3A','4A','6A','14'] ).
verb( 'waive', 'waives', 'waiving', 'waived', 'waived', tran, ['6A'] ).
verb( 'wake', 'wakes', 'waking', 'waked', 'waked', _, ['2A','2C','4B','6A','15B'] ).
verb( 'waken', 'wakens', 'wakening', 'wakened', 'wakened', _, ['2A','6A'] ).
verb( 'walk', 'walks', 'walking', 'walked', 'walked', _, ['2A','2B','2C','6A','15A','15B'] ).
verb( 'wall', 'walls', 'walling', 'walled', 'walled', tran, ['15B'] ).
verb( 'wallop', 'wallops', 'walloping', 'walloped', 'walloped', tran, [] ).
verb( 'wallow', 'wallows', 'wallowing', 'wallowed', 'wallowed', intran, ['2A','2C'] ).
verb( 'waltz', 'waltzes', 'waltzing', 'waltzed', 'waltzed', _, ['2A','2C','15A'] ).
verb( 'wander', 'wanders', 'wandering', 'wandered', 'wandered', _, ['2A','2B','2C','6A'] ).
verb( 'wane', 'wanes', 'waning', 'waned', 'waned', intran, ['2A'] ).
verb( 'wangle', 'wangles', 'wangling', 'wangled', 'wangled', tran, ['6A'] ).
verb( 'wank', 'wanks', 'wanking', 'wanked', 'wanked', intran, ['2A'] ).
verb( 'want', 'wants', 'wanting', 'wanted', 'wanted', _, ['2A','3A','6A','6E','7A','17','19B','24A'] ).
verb( 'wanton', 'wantons', 'wantoning', 'wantoned', 'wantoned', intran, ['2A','2C'] ).
verb( 'war', 'wars', 'warring', 'warred', 'warred', intran, [] ).
verb( 'warble', 'warbles', 'warbling', 'warbled', 'warbled', _, ['2A','2C','6A'] ).
verb( 'ward', 'wards', 'warding', 'warded', 'warded', tran, ['15B'] ).
verb( 'ware', 'wares', 'waring', 'wared', 'wared', tran, ['6A'] ).
verb( 'warm', 'warms', 'warming', 'warmed', 'warmed', _, ['2A','2C','6A','15B'] ).
verb( 'warn', 'warns', 'warning', 'warned', 'warned', tran, ['6A','11','14','15B','17'] ).
verb( 'warp', 'warps', 'warping', 'warped', 'warped', _, ['2A','6A'] ).
verb( 'warrant', 'warrants', 'warranting', 'warranted', 'warranted', tran, ['6A','9','25'] ).
verb( 'wash', 'washes', 'washing', 'washed', 'washed', _, ['2A','2C','6A','15A','15B','22'] ).
verb( 'waste', 'wastes', 'wasting', 'wasted', 'wasted', _, ['2A','2C','6A','14'] ).
verb( 'watch', 'watches', 'watching', 'watched', 'watched', _, ['2A','2B','2C','3A','4A','6A','8','10','15A','18A','19A'] ).
verb( 'water', 'waters', 'watering', 'watered', 'watered', _, ['2A','6A','15B'] ).
verb( 'waterproof', 'waterproofs', 'waterproofing', 'waterproofed', 'waterproofed', tran, [] ).
verb( 'wave', 'waves', 'waving', 'waved', 'waved', _, ['2A','3A','6A','12A','13A','15A','15B','16A'] ).
verb( 'waver', 'wavers', 'wavering', 'wavered', 'wavered', intran, ['2A','2C'] ).
verb( 'wax', 'waxes', 'waxing', 'waxed', 'waxed', _, ['2A','2D','6A'] ).
verb( 'waylay', 'waylays', 'waylaying', 'waylaid', 'waylaid', tran, ['6A'] ).
verb( 'weaken', 'weakens', 'weakening', 'weakened', 'weakened', _, ['2A','6A'] ).
verb( 'wean', 'weans', 'weaning', 'weaned', 'weaned', tran, ['6A','14'] ).
verb( 'wear', 'wears', 'wearing', 'wore', 'worn', _, ['2A','2B','2C','2D','6A','15A','15B','22'] ).
verb( 'weary', 'wearies', 'wearying', 'wearied', 'wearied', _, ['2A','3A','6A','14'] ).
verb( 'weather', 'weathers', 'weathering', 'weathered', 'weathered', _, ['2A','6A'] ).
verb( 'weave', 'weaves', 'weaving', 'wove', 'woven', _, ['2A','2C','6A','15A','15B'] ).
verb( 'wed', 'weds', 'wedding', 'wedded', 'wedded', _, ['2A','6A','14'] ).
verb( 'wedge', 'wedges', 'wedging', 'wedged', 'wedged', tran, ['6A','15A','22'] ).
verb( 'wee', 'wees', 'weeing', 'weed', 'weed', intran, [] ).
verb( 'weed', 'weeds', 'weeding', 'weeded', 'weeded', _, ['2A','6A','15B'] ).
verb( 'weekend', 'weekends', 'weekending', 'weekended', 'weekended', intran, [] ).
verb( 'weep', 'weeps', 'weeping', 'wept', 'wept', _, ['2A','2B','2C','3A','4B','6A'] ).
verb( 'weigh', 'weighs', 'weighing', 'weighed', 'weighed', _, ['2B','2C','3A','6A','14','15B'] ).
verb( 'weight', 'weights', 'weighting', 'weighted', 'weighted', tran, ['6A'] ).
verb( 'welcome', 'welcomes', 'welcoming', 'welcomed', 'welcomed', tran, ['6A','15A'] ).
verb( 'weld', 'welds', 'welding', 'welded', 'welded', _, ['2A','6A','15A','15B'] ).
verb( 'well', 'wells', 'welling', 'welled', 'welled', intran, ['2C'] ).
verb( 'welsh', 'welshes', 'welshing', 'welshed', 'welshed', intran, ['3A'] ).
verb( 'welter', 'welters', 'weltering', 'weltered', 'weltered', intran, ['2C'] ).
verb( 'wench', 'wenches', 'wenching', 'wenched', 'wenched', intran, ['2A'] ).
verb( 'wend', 'wends', 'wending', 'wended', 'wended', tran, [] ).
verb( 'westernize', 'westernizes', 'westernizing', 'westernized', 'westernized', tran, ['6A'] ).
verb( 'wet', 'wets', 'wetting', 'wetted', 'wetted', tran, ['6A'] ).
verb( 'whack', 'whacks', 'whacking', 'whacked', 'whacked', tran, ['6A'] ).
verb( 'whale', 'whales', 'whaling', 'whaled', 'whaled', intran, ['2A'] ).
verb( 'whang', 'whangs', 'whanging', 'whanged', 'whanged', tran, [] ).
verb( 'wheedle', 'wheedles', 'wheedling', 'wheedled', 'wheedled', tran, ['6A','14'] ).
verb( 'wheel', 'wheels', 'wheeling', 'wheeled', 'wheeled', _, ['2A','2C','6A','15A','15B'] ).
verb( 'wheeze', 'wheezes', 'wheezing', 'wheezed', 'wheezed', _, ['2A','2B','2C','15B'] ).
verb( 'whelp', 'whelps', 'whelping', 'whelped', 'whelped', intran, [] ).
verb( 'whet', 'whets', 'whetting', 'whetted', 'whetted', tran, ['6A'] ).
verb( 'while', 'whiles', 'whiling', 'whiled', 'whiled', tran, ['15B'] ).
verb( 'whimper', 'whimpers', 'whimpering', 'whimpered', 'whimpered', _, ['2A','6A'] ).
verb( 'whine', 'whines', 'whining', 'whined', 'whined', _, ['2A','2C','4A','6A','15B'] ).
verb( 'whinny', 'whinnies', 'whinnying', 'whinnied', 'whinnied', intran, [] ).
verb( 'whip', 'whips', 'whipping', 'whipped', 'whipped', _, ['2C','6A','15A','15B'] ).
verb( 'whir', 'whirs', 'whirring', 'whirred', 'whirred', intran, ['2A','2C'] ).
verb( 'whirl', 'whirls', 'whirling', 'whirled', 'whirled', _, ['2A','2C','15A','15B'] ).
verb( 'whirr', 'whirrs', 'whirring', 'whirred', 'whirred', intran, ['2A','2C'] ).
verb( 'whisk', 'whisks', 'whisking', 'whisked', 'whisked', _, ['6A','15B'] ).
verb( 'whisper', 'whispers', 'whispering', 'whispered', 'whispered', _, ['2A','2C','3A','6A','14','15B'] ).
verb( 'whistle', 'whistles', 'whistling', 'whistled', 'whistled', _, ['2A','2C','6A','15B','16A'] ).
verb( 'whiten', 'whitens', 'whitening', 'whitened', 'whitened', _, ['2A','6A'] ).
verb( 'whitewash', 'whitewashes', 'whitewashing', 'whitewashed', 'whitewashed', tran, [] ).
verb( 'whittle', 'whittles', 'whittling', 'whittled', 'whittled', _, ['2C','3A','6A','15A','15B'] ).
verb( 'whiz', 'whizzes', 'whizzing', 'whizzed', 'whizzed', intran, ['2C'] ).
verb( 'whoop', 'whoops', 'whooping', 'whooped', 'whooped', _, [] ).
verb( 'whop', 'whops', 'whopping', 'whopped', 'whopped', tran, ['6A'] ).
verb( 'widen', 'widens', 'widening', 'widened', 'widened', _, ['2A','6A'] ).
verb( 'wield', 'wields', 'wielding', 'wielded', 'wielded', tran, ['6A'] ).
verb( 'wiggle', 'wiggles', 'wiggling', 'wiggled', 'wiggled', _, ['2A','6A'] ).
verb( 'will', 'will', '-', '-', '-', unknown, ['2A','5','6A','9','12A','13A','14','15A','17'] ).
verb( 'will', 'wills', 'willing', 'willed', 'willed', _, ['2A','5','6A','9','12A','13A','14','15A','17'] ).
verb( 'wilt', 'wilts', 'wilting', 'wilted', 'wilted', _, ['2A','5','6A','9'] ).
verb( 'win', 'wins', 'winning', 'won', 'won', _, ['2A','6A','12B','13B','15A','15B','17'] ).
verb( 'wince', 'winces', 'wincing', 'winced', 'winced', tran, ['2A','2C'] ).
verb( 'winch', 'winches', 'winching', 'winched', 'winched', tran, ['6A','15B'] ).
verb( 'wind', 'winds', 'winding', 'winded', 'winded', tran, ['6A'] ).
verb( 'wind', 'winds', 'winding', 'wound', 'wound', _, ['2A','2B','2C','6A','14','15A','15B'] ).
verb( 'wine', 'wines', 'wining', 'wined', 'wined', tran, [] ).
verb( 'wing', 'wings', 'winging', 'winged', 'winged', _, ['2C','6A','15A'] ).
verb( 'wink', 'winks', 'winking', 'winked', 'winked', _, ['2A','2C','3A','15B'] ).
verb( 'winkle', 'winkles', 'winkling', 'winkled', 'winkled', tran, ['15B'] ).
verb( 'winnow', 'winnows', 'winnowing', 'winnowed', 'winnowed', tran, ['6A','14','15A','15B'] ).
verb( 'winter', 'winters', 'wintering', 'wintered', 'wintered', intran, ['2C'] ).
verb( 'wipe', 'wipes', 'wiping', 'wiped', 'wiped', _, ['6A','15A','15B','22'] ).
verb( 'wire', 'wires', 'wiring', 'wired', 'wired', _, ['6A','11','12A','13A','15A','15B','16A'] ).
verb( 'wisecrack', 'wisecracks', 'wisecracking', 'wisecracked', 'wisecracked', intran, [] ).
verb( 'wish', 'wishes', 'wishing', 'wished', 'wished', _, ['2A','3A','6A','7A','9','12A','13A','15A','17','22'] ).
verb( 'withdraw', 'withdraws', 'withdrawing', 'withdrew', 'withdrawn', _, ['2A','2C','6A','14'] ).
verb( 'wither', 'withers', 'withering', 'withered', 'withered', _, ['2A','2C','6A','15B'] ).
verb( 'withhold', 'withholds', 'withholding', 'withheld', 'withheld', tran, ['6A','14'] ).
verb( 'withstand', 'withstands', 'withstanding', 'withstood', 'withstood', tran, ['6A'] ).
verb( 'witness', 'witnesses', 'witnessing', 'witnessed', 'witnessed', _, ['3A','6A'] ).
verb( 'wive', 'wives', 'wiving', 'wived', 'wived', _, ['2A','6A'] ).
verb( 'wobble', 'wobbles', 'wobbling', 'wobbled', 'wobbled', _, ['2A','2C','6A'] ).
verb( 'wolf', 'wolfs', 'wolfing', 'wolfed', 'wolfed', tran, ['6A','15A'] ).
verb( 'womanize', 'womanizes', 'womanizing', 'womanized', 'womanized', intran, [] ).
verb( 'wonder', 'wonders', 'wondering', 'wondered', 'wondered', _, ['2A','3A','3B','4B','8','10'] ).
verb( 'woo', 'woos', 'wooing', 'wooed', 'wooed', tran, ['6A'] ).
verb( 'word', 'words', 'wording', 'worded', 'worded', tran, ['6A'] ).
verb( 'work', 'works', 'working', 'worked', 'worked', _, ['2A','2B','2C','2D','3A','4A','6A','14','15A','15B','22'] ).
verb( 'worm', 'worms', 'worming', 'wormed', 'wormed', tran, ['6A','15A','15B'] ).
verb( 'worry', 'worries', 'worrying', 'worried', 'worried', _, ['2A','2B','2C','3A','6A','14','15A','15B','17','22'] ).
verb( 'worsen', 'worsens', 'worsening', 'worsened', 'worsened', _, ['2A','6A'] ).
verb( 'worship', 'worships', 'worshipping', 'worshipped', 'worshipped', _, ['2A','2B','6A'] ).
verb( 'worst', 'worsts', 'worsting', 'worsted', 'worsted', tran, ['6A'] ).
verb( 'wound', 'wounds', 'wounding', 'wounded', 'wounded', tran, ['6A'] ).
verb( 'wrangle', 'wrangles', 'wrangling', 'wrangled', 'wrangled', intran, ['2A','3A'] ).
verb( 'wrap', 'wraps', 'wrapping', 'wrapped', 'wrapped', _, ['6A','14','15A','15B'] ).
verb( 'wreak', 'wreaks', 'wreaking', 'wreaked', 'wreaked', tran, ['6A','14'] ).
verb( 'wreathe', 'wreathes', 'wreathing', 'wreathed', 'wreathed', _, ['2A','2C','6A','14'] ).
verb( 'wreck', 'wrecks', 'wrecking', 'wrecked', 'wrecked', tran, ['6A'] ).
verb( 'wrench', 'wrenches', 'wrenching', 'wrenched', 'wrenched', tran, ['6A','15A','22'] ).
verb( 'wrest', 'wrests', 'wresting', 'wrested', 'wrested', tran, ['14'] ).
verb( 'wrestle', 'wrestles', 'wrestling', 'wrestled', 'wrestled', intran, ['2A','2C','3A'] ).
verb( 'wrick', 'wricks', 'wricking', 'wricked', 'wricked', tran, ['6A'] ).
verb( 'wriggle', 'wriggles', 'wriggling', 'wriggled', 'wriggled', _, ['2A','2C','3A','6A','15B','22'] ).
verb( 'wring', 'wrings', 'wringing', 'wrung', 'wrung', tran, ['6A','14','15B'] ).
verb( 'wrinkle', 'wrinkles', 'wrinkling', 'wrinkled', 'wrinkled', _, ['2A','2C','6A','15B'] ).
verb( 'write', 'writes', 'writing', 'wrote', 'written', _, ['2A','2B','2C','4A','6A','12A','13A','15B'] ).
verb( 'writhe', 'writhes', 'writhing', 'writhed', 'writhed', intran, ['2A','2C'] ).
verb( 'wrong', 'wrongs', 'wronging', 'wronged', 'wronged', tran, ['6A'] ).
verb( 'yacht', 'yachts', 'yachting', 'yachted', 'yachted', intran, ['2A'] ).
verb( 'yammer', 'yammers', 'yammering', 'yammered', 'yammered', intran, [] ).
verb( 'yank', 'yanks', 'yanking', 'yanked', 'yanked', tran, ['6A','15A','15B'] ).
verb( 'yap', 'yaps', 'yapping', 'yapped', 'yapped', intran, ['2A'] ).
verb( 'yarn', 'yarns', 'yarning', 'yarned', 'yarned', intran, ['2A','2C'] ).
verb( 'yaw', 'yaws', 'yawing', 'yawed', 'yawed', intran, [] ).
verb( 'yawn', 'yawns', 'yawning', 'yawned', 'yawned', intran, ['2A','2C'] ).
verb( 'yearn', 'yearns', 'yearning', 'yearned', 'yearned', intran, ['3A','4A'] ).
verb( 'yell', 'yells', 'yelling', 'yelled', 'yelled', _, ['2A','2C','6A','15B'] ).
verb( 'yellow', 'yellows', 'yellowing', 'yellowed', 'yellowed', _, ['2A','6A'] ).
verb( 'yelp', 'yelps', 'yelping', 'yelped', 'yelped', intran, ['2A'] ).
verb( 'yen', 'yens', 'yenning', 'yenned', 'yenned', intran, [] ).
verb( 'yield', 'yields', 'yielding', 'yielded', 'yielded', _, ['2A','3A','6A','15A','15B'] ).
verb( 'yodel', 'yodels', 'yodelling', 'yodelled', 'yodelled', _, [] ).
verb( 'yoke', 'yokes', 'yoking', 'yoked', 'yoked', _, ['6A','15A'] ).
verb( 'yowl', 'yowls', 'yowling', 'yowled', 'yowled', intran, [] ).
verb( 'zap', 'zaps', 'zapping', 'zapped', 'zapped', tran, ['6A'] ).
verb( 'zero', 'zeros', 'zeroing', 'zeroed', 'zeroed', intran, ['2C'] ).
verb( 'zigzag', 'zigzags', 'zigzagging', 'zigzagged', 'zigzagged', intran, [] ).
verb( 'zip', 'zips', 'zipping', 'zipped', 'zipped', tran, ['6A','15B','22'] ).
verb( 'zone', 'zones', 'zoning', 'zoned', 'zoned', tran, ['6A'] ).
verb( 'zoom', 'zooms', 'zooming', 'zoomed', 'zoomed', intran, ['2A','2C'] ).

noun( 'a-bomb', 'a-bombs', count, _ ).
noun( 'a-level', 'a-levels', count, _ ).
noun( 'aa', '-', count, _ ).
noun( 'abc', '-', count, _ ).
noun( '-', 'abcs', count, _ ).
noun( 'ad', '-', proper, _ ).
noun( 'agm', '-', count, _ ).
noun( 'aids', '-', mass, _ ).
noun( 'awol', '-', proper, _ ).
noun( 'aachen', '-', proper, loc ).
noun( 'aarhus', '-', proper, loc ).
noun( 'abe', '-', proper, per ).
noun( 'abercarn', '-', proper, loc ).
noun( 'aberdare', '-', proper, loc ).
noun( 'aberdeen', '-', proper, loc ).
noun( 'abergavenny', '-', proper, loc ).
noun( 'abergele', '-', proper, loc ).
noun( 'abertillery', '-', proper, loc ).
noun( 'aberystwyth', '-', proper, loc ).
noun( 'abingdon', '-', proper, loc ).
noun( 'abo', 'abos', count, _ ).
noun( 'aborigine', 'aborigines', count, _ ).
noun( 'abraham', '-', proper, per ).
noun( 'accra', '-', proper, loc ).
noun( 'accrington', '-', proper, loc ).
noun( 'achilles', '-', proper, per ).
noun( 'ada', '-', proper, per ).
noun( 'adam', '-', proper, per ).
noun( 'addis ababa', '-', proper, loc ).
noun( 'addressograph', 'addressographs', count, _ ).
noun( 'adelaide', '-', proper, loc ).
noun( 'adrian', '-', proper, per ).
noun( 'adventist', 'adventists', count, _ ).
noun( 'afghan', 'afghans', both, _ ).
noun( 'afghanistan', '-', proper, loc ).
noun( 'afghanistani', 'afghanistanis', count, _ ).
noun( 'africa', '-', proper, loc ).
noun( 'african', 'africans', count, _ ).
noun( 'afrikaans', '-', mass, _ ).
noun( 'afrikaner', 'afrikaners', count, _ ).
noun( 'afro-american', 'afro-americans', count, _ ).
noun( 'afro-wig', 'afro-wigs', count, _ ).
noun( 'agatha', '-', proper, per ).
noun( 'aggie', '-', proper, per ).
noun( 'agnes', '-', proper, per ).
noun( 'agra', '-', proper, loc ).
noun( 'ahmedabad', '-', proper, loc ).
noun( 'airdrie', '-', proper, loc ).
noun( 'aireborough', '-', proper, loc ).
noun( 'airedale', 'airedales', count, _ ).
noun( 'akron', '-', proper, loc ).
noun( 'al', '-', proper, per ).
noun( 'alabama', '-', proper, loc ).
noun( 'alan', '-', proper, per ).
noun( 'alaska', '-', proper, loc ).
noun( 'albania', '-', proper, loc ).
noun( 'albanian', 'albanians', both, _ ).
noun( 'albert', '-', proper, per ).
noun( 'alberta', '-', proper, loc ).
noun( 'aldershot', '-', proper, loc ).
noun( 'aleppo', '-', proper, loc ).
noun( 'alessandria', '-', proper, loc ).
noun( 'alex', '-', proper, per ).
noun( 'alexander', '-', proper, per ).
noun( 'alexandra', '-', proper, per ).
noun( 'alexandria', '-', proper, loc ).
noun( 'alf', '-', proper, per ).
noun( 'alfred', '-', proper, per ).
noun( 'alfreton', '-', proper, loc ).
noun( 'alger', '-', proper, loc ).
noun( 'algeria', '-', proper, loc ).
noun( 'algerian', 'algerians', count, _ ).
noun( 'ali', '-', proper, per ).
noun( 'alicante', '-', proper, loc ).
noun( 'alice', '-', proper, per ).
noun( 'alison', '-', proper, per ).
noun( 'allah', '-', proper, _ ).
noun( 'allahabad', '-', proper, loc ).
noun( 'allan', '-', proper, per ).
noun( 'allen', '-', proper, per ).
noun( 'alloa', '-', proper, loc ).
noun( 'alma mater', 'alma maters', count, _ ).
noun( 'alma-ata', '-', proper, loc ).
noun( 'almeria', '-', proper, loc ).
noun( 'alnwick', '-', proper, loc ).
noun( 'alsatian', 'alsatians', count, _ ).
noun( 'alton', '-', proper, loc ).
noun( 'altrincham', '-', proper, loc ).
noun( 'alvechurch', '-', proper, loc ).
noun( 'amanda', '-', proper, per ).
noun( 'amazon', 'amazons', count, _ ).
noun( 'america', '-', proper, loc ).
noun( 'american', 'americans', count, _ ).
noun( 'americanism', 'americanisms', both, _ ).
noun( 'amesbury', '-', proper, loc ).
noun( 'amharic', '-', mass, _ ).
noun( 'amiens', '-', proper, loc ).
noun( 'ammanford', '-', proper, loc ).
noun( 'amsterdam', '-', proper, loc ).
noun( 'amy', '-', proper, per ).
noun( 'ancona', '-', proper, loc ).
noun( 'andorra', '-', proper, loc ).
noun( 'andorran', 'andorrans', count, _ ).
noun( 'andover', '-', proper, loc ).
noun( 'andrew', '-', proper, per ).
noun( 'andy', '-', proper, per ).
noun( 'angela', '-', proper, per ).
noun( 'angers', '-', proper, loc ).
noun( 'angie', '-', proper, per ).
noun( 'anglican', 'anglicans', count, _ ).
noun( 'anglo-catholic', 'anglo-catholics', count, _ ).
noun( 'anglo-indian', 'anglo-indians', count, _ ).
noun( 'anglo-saxon', 'anglo-saxons', count, _ ).
noun( 'anglomania', '-', mass, _ ).
noun( 'anglophil', 'anglophils', count, _ ).
noun( 'anglophile', 'anglophiles', count, _ ).
noun( 'anglophobe', 'anglophobes', count, _ ).
noun( 'anglophobia', '-', mass, _ ).
noun( 'angola', '-', proper, loc ).
noun( 'angolan', 'angolans', count, _ ).
noun( 'anguilla', '-', proper, loc ).
noun( 'anguillan', 'anguillans', count, _ ).
noun( 'angus', '-', proper, per ).
noun( 'anita', '-', proper, per ).
noun( 'ankara', '-', proper, loc ).
noun( 'ann', '-', proper, per ).
noun( 'annabel', '-', proper, per ).
noun( 'annan', '-', proper, loc ).
noun( 'anne', '-', proper, per ).
noun( 'annfield', '-', proper, loc ).
noun( 'annfield plain', '-', proper, loc ).
noun( 'annie', '-', proper, per ).
noun( 'anon', '-', proper, _ ).
noun( 'anshan', '-', proper, loc ).
noun( 'anthea', '-', proper, per ).
noun( 'anthony', '-', proper, per ).
noun( 'antigua', '-', proper, loc ).
noun( 'antiguan', 'antiguans', count, _ ).
noun( 'antony', '-', proper, per ).
noun( 'antrim', '-', proper, loc ).
noun( 'antwerp', '-', proper, loc ).
noun( 'apeldoorn', '-', proper, loc ).
noun( 'apocrypha', 'apocrypha', mass, _ ).
noun( 'apr', '-', proper, _ ).
noun( 'april', 'aprils', count, _ ).
noun( 'april', '-', proper, per ).
noun( 'aquarius', '-', proper, _ ).
noun( 'arab', 'arabs', count, _ ).
noun( 'arabian', 'arabians', count, _ ).
noun( 'arabic', '-', mass, _ ).
noun( 'arabist', 'arabists', count, _ ).
noun( 'arbroath', '-', proper, loc ).
noun( 'arcadian', 'arcadians', count, _ ).
noun( 'ardrossan', '-', proper, loc ).
noun( 'argentina', '-', proper, loc ).
noun( 'argentine', '-', proper, loc ).
noun( 'argentinian', 'argentinians', count, _ ).
noun( 'argonaut', 'argonauts', count, _ ).
noun( 'argus', 'arguses', count, _ ).
noun( 'aries', '-', proper, _ ).
noun( 'arizona', '-', proper, loc ).
noun( 'arkansas', '-', proper, loc ).
noun( 'armadale', '-', proper, loc ).
noun( 'armageddon', '-', proper, _ ).
noun( 'armagh', '-', proper, loc ).
noun( 'arnhem', '-', proper, loc ).
noun( 'arnold', '-', proper, per ).
noun( 'art', '-', proper, per ).
noun( 'arthur', '-', proper, per ).
noun( 'artie', '-', proper, per ).
noun( 'aryan', 'aryans', count, _ ).
noun( 'ascot', '-', proper, loc ).
noun( 'ashbourne', '-', proper, loc ).
noun( 'ashby', '-', proper, loc ).
noun( 'ashby woulds', '-', proper, loc ).
noun( 'ashby de la zouch', '-', proper, loc ).
noun( 'ashford', '-', proper, loc ).
noun( 'ashington', '-', proper, loc ).
noun( 'ashton-in-makerfield', '-', proper, loc ).
noun( 'ashton-under-lyne', '-', proper, loc ).
noun( 'asian', 'asians', count, _ ).
noun( 'asiatic', 'asiatics', count, _ ).
noun( 'askern', '-', proper, loc ).
noun( 'aspull', '-', proper, loc ).
noun( 'aston', '-', proper, loc ).
noun( 'athenian', 'athenians', count, _ ).
noun( 'athens', '-', proper, loc ).
noun( 'atherstone', '-', proper, loc ).
noun( 'athlone', '-', proper, loc ).
noun( 'atlanta', '-', proper, loc ).
noun( 'atlantic', '-', proper, _ ).
noun( 'auckland', '-', proper, loc ).
noun( 'audley', '-', proper, loc ).
noun( 'audrey', '-', proper, per ).
noun( 'aug', '-', proper, _ ).
noun( 'augsburg', '-', proper, loc ).
noun( 'august', 'augusts', count, _ ).
noun( 'august', '-', proper, _ ).
noun( 'aussie', 'aussies', count, _ ).
noun( 'australia', '-', proper, loc ).
noun( 'australian', 'australians', count, _ ).
noun( 'austria', '-', proper, loc ).
noun( 'austrian', 'austrians', count, _ ).
noun( 'ave', '-', proper, _ ).
noun( 'aveley', '-', proper, loc ).
noun( 'avignon', '-', proper, loc ).
noun( 'avon', '-', proper, loc ).
noun( 'aycliffe', '-', proper, loc ).
noun( 'aylesbury', '-', proper, loc ).
noun( 'aylesford', '-', proper, loc ).
noun( 'ayr', '-', proper, loc ).
noun( 'b ed', '-', count, _ ).
noun( 'ba', '-', count, _ ).
noun( 'bbc', '-', count, _ ).
noun( 'bc', '-', proper, _ ).
noun( 'bma', '-', count, _ ).
noun( 'bmus', '-', count, _ ).
noun( 'bst', '-', proper, _ ).
noun( 'bsc', '-', count, _ ).
noun( 'bvm', '-', count, _ ).
noun( 'babs', '-', proper, per ).
noun( 'backworth', '-', proper, loc ).
noun( 'bacup', '-', proper, loc ).
noun( 'badajoz', '-', proper, loc ).
noun( 'badalona', '-', proper, loc ).
noun( 'baghdad', '-', proper, loc ).
noun( 'bahamas', '-', proper, loc ).
noun( 'bahamian', 'bahamians', count, _ ).
noun( 'bahasa', '-', mass, _ ).
noun( 'bahrain', '-', proper, loc ).
noun( 'bahraini', 'bahrainis', count, _ ).
noun( 'bailey', '-', proper, _ ).
noun( 'baku', '-', proper, loc ).
noun( 'baldock', '-', proper, loc ).
noun( 'baltimore', '-', proper, loc ).
noun( 'banbury', '-', proper, loc ).
noun( 'bandung', '-', proper, loc ).
noun( 'bangalore', '-', proper, loc ).
noun( 'bangkok', '-', proper, loc ).
noun( 'bangladesh', '-', proper, loc ).
noun( 'bangladeshi', 'bangladeshis', count, _ ).
noun( 'bangor', '-', proper, loc ).
noun( 'bannockburn', '-', proper, loc ).
noun( 'bantry', '-', proper, loc ).
noun( 'bantu', 'bantu', count, _ ).
noun( 'baptist', 'baptists', count, _ ).
noun( 'baracaldo', '-', proper, loc ).
noun( 'barbadian', 'barbadians', count, _ ).
noun( 'barbados', '-', proper, loc ).
noun( 'barbara', '-', proper, per ).
noun( 'barcelona', '-', proper, loc ).
noun( 'bargoed', '-', proper, loc ).
noun( 'bari', '-', proper, loc ).
noun( 'barking', '-', proper, loc ).
noun( 'barnard castle', '-', proper, loc ).
noun( 'barnet', '-', proper, loc ).
noun( 'barnoldswick', '-', proper, loc ).
noun( 'barnsley', '-', proper, loc ).
noun( 'barnstaple', '-', proper, loc ).
noun( 'barranquilla', '-', proper, loc ).
noun( 'barrhead', '-', proper, loc ).
noun( 'barrow-in-furness', '-', proper, loc ).
noun( 'barry', '-', proper, per ).
noun( 'bart', '-', proper, per ).
noun( 'bart', '-', proper, per ).
noun( 'bartholomew', '-', proper, per ).
noun( 'barton-upon-humber', '-', proper, loc ).
noun( 'basel', '-', proper, loc ).
noun( 'basil', '-', proper, per ).
noun( 'basildon', '-', proper, loc ).
noun( 'basingstoke', '-', proper, loc ).
noun( 'bass', '-', mass, _ ).
noun( 'bath', '-', proper, loc ).
noun( 'bath-chair', 'bath-chairs', count, _ ).
noun( 'bathgate', '-', proper, loc ).
noun( 'batley', '-', proper, loc ).
noun( 'beaconsfield', '-', proper, loc ).
noun( 'bearsden', '-', proper, loc ).
noun( 'beatrice', '-', proper, per ).
noun( 'beaujolais', '-', mass, _ ).
noun( 'beaumaris', '-', proper, loc ).
noun( 'bebington', '-', proper, loc ).
noun( 'beccles', '-', proper, loc ).
noun( 'beddau', '-', proper, loc ).
noun( 'bedford', '-', proper, loc ).
noun( 'bedfordshire', '-', proper, loc ).
noun( 'bedlington', '-', proper, loc ).
noun( 'bedouin', 'bedouin', count, _ ).
noun( 'beduin', 'beduin', count, _ ).
noun( 'bedwas', '-', proper, loc ).
noun( 'beirut', '-', proper, loc ).
noun( 'beith', '-', proper, loc ).
noun( 'belem', '-', proper, loc ).
noun( 'belfast', '-', proper, loc ).
noun( 'belgian', 'belgians', count, _ ).
noun( 'belgium', '-', proper, loc ).
noun( 'belgrade', '-', proper, loc ).
noun( 'belinda', '-', proper, per ).
noun( 'bella', '-', proper, per ).
noun( 'belle', '-', proper, per ).
noun( 'bellshill', '-', proper, loc ).
noun( 'belo horizonte', '-', proper, loc ).
noun( 'belper', '-', proper, loc ).
noun( 'ben', '-', proper, per ).
noun( 'benedictine', 'benedictines', both, _ ).
noun( 'bengali', 'bengalis', both, _ ).
noun( 'benin', '-', proper, loc ).
noun( 'beninese', 'beninese', count, _ ).
noun( 'benjamin', '-', proper, per ).
noun( 'benny', '-', proper, per ).
noun( 'benzedrine', 'benzedrines', count, _ ).
noun( 'beograd', '-', proper, loc ).
noun( 'bergamo', '-', proper, loc ).
noun( 'bergen', '-', proper, loc ).
noun( 'berkhamsted', '-', proper, loc ).
noun( 'berkshire', '-', proper, loc ).
noun( 'berlin', '-', proper, loc ).
noun( 'bermuda', '-', proper, loc ).
noun( 'bermudan', 'bermudans', count, _ ).
noun( 'bern', '-', proper, loc ).
noun( 'bernard', '-', proper, per ).
noun( 'bernie', '-', proper, per ).
noun( 'bert', '-', proper, per ).
noun( 'bertha', '-', proper, per ).
noun( 'bertie', '-', proper, per ).
noun( 'berwick-upon-tweed', '-', proper, loc ).
noun( 'beryl', '-', proper, per ).
noun( 'bess', '-', proper, per ).
noun( 'bessie', '-', proper, per ).
noun( 'bethune', '-', proper, loc ).
noun( 'betsy', '-', proper, per ).
noun( 'betty', '-', proper, per ).
noun( 'beverley', '-', proper, loc ).
noun( 'bewdley', '-', proper, loc ).
noun( 'bexhill', '-', proper, loc ).
noun( 'bexley', '-', proper, loc ).
noun( 'bhutan', '-', proper, loc ).
noun( 'bhutani', 'bhutanis', count, _ ).
noun( 'bible', 'bibles', count, _ ).
noun( 'bicester', '-', proper, loc ).
noun( 'biddulph', '-', proper, loc ).
noun( 'bideford', '-', proper, loc ).
noun( 'bielefeld', '-', proper, loc ).
noun( 'biggleswade', '-', proper, loc ).
noun( 'bilbao', '-', proper, loc ).
noun( 'bill', '-', proper, per ).
noun( 'billericay', '-', proper, loc ).
noun( 'billington', '-', proper, loc ).
noun( 'billy', '-', proper, per ).
noun( 'bingley', '-', proper, loc ).
noun( 'birkenhead', '-', proper, loc ).
noun( 'birstall', '-', proper, loc ).
noun( 'birtley', '-', proper, loc ).
noun( 'bishop auckland', '-', proper, loc ).
noun( 'bishop\'s cleeve', '-', proper, loc ).
noun( 'bishop\'s stortford', '-', proper, loc ).
noun( 'bishopbriggs', '-', proper, loc ).
noun( 'blackburn', '-', proper, loc ).
noun( 'blackhall', '-', proper, loc ).
noun( 'blackpool', '-', proper, loc ).
noun( 'blackshirt', 'blackshirts', count, _ ).
noun( 'blackwood', '-', proper, loc ).
noun( 'blaenau-ffestiniog', '-', proper, loc ).
noun( 'blaenavon', '-', proper, loc ).
noun( 'blaengwrach', '-', proper, loc ).
noun( 'blaina', '-', proper, loc ).
noun( 'blairgowrie', '-', proper, loc ).
noun( 'blantyre', '-', proper, loc ).
noun( 'bldg', '-', proper, _ ).
noun( 'bldgs', '-', proper, _ ).
noun( 'bletchley', '-', proper, loc ).
noun( 'blidworth', '-', proper, loc ).
noun( 'blighty', '-', count, _ ).
noun( 'blvd', '-', count, _ ).
noun( 'blyth', '-', proper, loc ).
noun( 'bo\'ness', '-', proper, loc ).
noun( 'bob', '-', proper, per ).
noun( 'bobby', '-', proper, per ).
noun( 'bochum', '-', proper, loc ).
noun( 'bodmin', '-', proper, loc ).
noun( 'boer', 'boers', count, _ ).
noun( 'bognor', '-', proper, loc ).
noun( 'bognor regis', '-', proper, loc ).
noun( 'bogota', '-', proper, loc ).
noun( 'boldon', '-', proper, loc ).
noun( 'bolivia', '-', proper, loc ).
noun( 'bolivian', 'bolivians', count, _ ).
noun( 'bollington', '-', proper, loc ).
noun( 'bologna', '-', proper, loc ).
noun( 'bolshevik', 'bolsheviks', count, _ ).
noun( 'bolsover', '-', proper, loc ).
noun( 'bolton', '-', proper, loc ).
noun( 'bolton-le-sands', '-', proper, loc ).
noun( 'bolzano', '-', proper, loc ).
noun( 'bombay', '-', proper, loc ).
noun( 'bonn', '-', proper, loc ).
noun( 'bonnybridge', '-', proper, loc ).
noun( 'bonnyrigg', '-', proper, loc ).
noun( 'bootle', '-', proper, loc ).
noun( 'bordeaux', '-', mass, _ ).
noun( 'bordeaux', '-', proper, loc ).
noun( 'borders', '-', proper, loc ).
noun( 'bordon', '-', proper, loc ).
noun( 'boris', '-', proper, per ).
noun( 'boston', '-', proper, loc ).
noun( 'botswana', '-', proper, loc ).
noun( 'bottrop', '-', proper, loc ).
noun( 'bourne', '-', proper, loc ).
noun( 'bournemouth', '-', proper, loc ).
noun( 'bowburn', '-', proper, loc ).
noun( 'boxing day', 'boxing days', count, _ ).
noun( 'boxing day', '-', proper, _ ).
noun( 'bracknell', '-', proper, loc ).
noun( 'bradford', '-', proper, loc ).
noun( 'bradford-on-avon', '-', proper, loc ).
noun( 'brahmin', 'brahmins', count, _ ).
noun( 'braintree', '-', proper, loc ).
noun( 'branderburgh', '-', proper, loc ).
noun( 'brasilia', '-', proper, loc ).
noun( 'bratislava', '-', proper, loc ).
noun( 'braunschweig', '-', proper, loc ).
noun( 'braunton', '-', proper, loc ).
noun( 'brazil', '-', proper, loc ).
noun( 'brazilian', 'brazilians', count, _ ).
noun( 'breaston', '-', proper, loc ).
noun( 'brecknock', '-', proper, loc ).
noun( 'breda', '-', proper, loc ).
noun( 'breedsall', '-', proper, loc ).
noun( 'bremen', '-', proper, loc ).
noun( 'bremerhaven', '-', proper, loc ).
noun( 'bren', 'brens', count, _ ).
noun( 'bren-gun', 'bren-guns', count, _ ).
noun( 'brenda', '-', proper, per ).
noun( 'brent', '-', proper, loc ).
noun( 'brentwood', '-', proper, loc ).
noun( 'brescia', '-', proper, loc ).
noun( 'brest', '-', proper, loc ).
noun( 'brian', '-', proper, per ).
noun( 'bricket wood', '-', proper, loc ).
noun( 'bridgend', '-', proper, loc ).
noun( 'bridget', '-', proper, per ).
noun( 'bridgnorth', '-', proper, loc ).
noun( 'bridgwater', '-', proper, loc ).
noun( 'bridlington', '-', proper, loc ).
noun( 'bridport', '-', proper, loc ).
noun( 'brig', '-', proper, per ).
noun( 'brigadier', 'brigadiers', count, _ ).
noun( 'brighouse', '-', proper, loc ).
noun( 'brightlingsea', '-', proper, loc ).
noun( 'brighton', '-', proper, loc ).
noun( 'brisbane', '-', proper, loc ).
noun( 'bristol', '-', proper, loc ).
noun( 'brit', 'brits', count, _ ).
noun( 'britain', '-', proper, loc ).
noun( 'britisher', 'britishers', count, _ ).
noun( 'briton', 'britons', count, _ ).
noun( 'brixham', '-', proper, loc ).
noun( 'brno', '-', proper, loc ).
noun( 'bromley', '-', proper, loc ).
noun( 'bromsgrove', '-', proper, loc ).
noun( 'bros', '-', proper, _ ).
noun( 'broughton', '-', proper, loc ).
noun( 'broxburn', '-', proper, loc ).
noun( 'bruce', '-', proper, per ).
noun( 'bruges', '-', proper, loc ).
noun( 'brunei', '-', proper, loc ).
noun( 'bruneian', 'bruneians', count, _ ).
noun( 'brunswick', '-', proper, loc ).
noun( 'brussels', '-', proper, loc ).
noun( 'bryan', '-', proper, per ).
noun( 'bt', '-', proper, per ).
noun( 'bucarest', '-', proper, loc ).
noun( 'buckhaven', '-', proper, loc ).
noun( 'buckie', '-', proper, loc ).
noun( 'buckingham', '-', proper, loc ).
noun( 'buckinghamshire', '-', proper, loc ).
noun( 'buckley', '-', proper, loc ).
noun( 'bucksburn', '-', proper, loc ).
noun( 'budapest', '-', proper, loc ).
noun( 'buddhism', '-', mass, _ ).
noun( 'buddhist', 'buddhists', count, _ ).
noun( 'buenos aires', '-', proper, loc ).
noun( 'buffalo', '-', proper, loc ).
noun( 'bulgaria', '-', proper, loc ).
noun( 'bulgarian', 'bulgarians', count, _ ).
noun( 'bulkington', '-', proper, loc ).
noun( 'bunsen', 'bunsens', count, _ ).
noun( 'burberry', 'burberries', count, _ ).
noun( 'burgess', '-', proper, loc ).
noun( 'burgos', '-', proper, loc ).
noun( 'burgundy', '-', mass, _ ).
noun( 'burley', '-', proper, loc ).
noun( 'burma', '-', proper, loc ).
noun( 'burmese', 'burmese', both, _ ).
noun( 'burnham-on-sea', '-', proper, loc ).
noun( 'burnley', '-', proper, loc ).
noun( 'burntisland', '-', proper, loc ).
noun( 'burntwood', '-', proper, loc ).
noun( 'burry port', '-', proper, loc ).
noun( 'burscough', '-', proper, loc ).
noun( 'burton-upon-trent', '-', proper, loc ).
noun( 'burundi', '-', proper, loc ).
noun( 'burundian', 'burundians', count, _ ).
noun( 'bury', '-', proper, loc ).
noun( 'bury st. edmunds', '-', proper, loc ).
noun( 'bushman', 'bushmen', count, _ ).
noun( 'buxton', '-', proper, loc ).
noun( 'c of e', '-', count, _ ).
noun( 'c-in-c', '-', proper, per ).
noun( 'cbi', '-', count, _ ).
noun( 'cd', '-', count, _ ).
noun( 'cia', '-', count, _ ).
noun( 'cid', '-', count, _ ).
noun( 'co', '-', count, _ ).
noun( 'cod', '-', proper, _ ).
noun( 'cadiz', '-', proper, loc ).
noun( 'caen', '-', proper, loc ).
noun( 'caerleon', '-', proper, loc ).
noun( 'caernarfon', '-', proper, loc ).
noun( 'caerphilly', '-', proper, loc ).
noun( 'caesar', 'caesars', count, _ ).
noun( 'cagliari', '-', proper, loc ).
noun( 'cairo', '-', proper, loc ).
noun( 'calcutta', '-', proper, loc ).
noun( 'caldicot', '-', proper, loc ).
noun( 'cali', '-', proper, loc ).
noun( 'california', '-', proper, loc ).
noun( 'californian', 'californians', count, _ ).
noun( 'calne', '-', proper, loc ).
noun( 'calor', '-', proper, _ ).
noun( 'calvary', 'calvaries', count, _ ).
noun( 'calverton', '-', proper, loc ).
noun( 'calvinism', '-', mass, _ ).
noun( 'calvinist', 'calvinists', count, _ ).
noun( 'cambodia', '-', proper, loc ).
noun( 'cambodian', 'cambodians', count, _ ).
noun( 'cambourne', '-', proper, loc ).
noun( 'cambridge', '-', proper, loc ).
noun( 'cambridgeshire', '-', proper, loc ).
noun( 'cambuslang', '-', proper, loc ).
noun( 'camden', '-', proper, loc ).
noun( 'camembert', 'camemberts', count, _ ).
noun( 'cameroon', '-', proper, loc ).
noun( 'cameroonian', 'cameroonians', count, _ ).
noun( 'campbeltown', '-', proper, loc ).
noun( 'canada', '-', proper, loc ).
noun( 'canadian', 'canadians', count, _ ).
noun( 'canberra', '-', proper, loc ).
noun( 'cancer', '-', proper, _ ).
noun( 'cannes', '-', proper, loc ).
noun( 'cannock', '-', proper, loc ).
noun( 'cantab', '-', proper, _ ).
noun( 'canterbury', '-', proper, loc ).
noun( 'canton', '-', proper, loc ).
noun( 'canuck', 'canucks', count, _ ).
noun( 'canvey', '-', proper, loc ).
noun( 'canvey island', '-', proper, loc ).
noun( 'cape town', '-', proper, loc ).
noun( 'capitol', 'capitols', count, _ ).
noun( 'capricorn', '-', proper, _ ).
noun( 'capt', '-', proper, per ).
noun( 'caracas', '-', proper, loc ).
noun( 'cardenden', '-', proper, loc ).
noun( 'cardiff', '-', proper, loc ).
noun( 'cardigan', '-', proper, loc ).
noun( 'carl', '-', proper, per ).
noun( 'carlisle', '-', proper, loc ).
noun( 'carlow', '-', proper, loc ).
noun( 'carluke', '-', proper, loc ).
noun( 'carmarthen', '-', proper, loc ).
noun( 'carmelite', 'carmelites', count, _ ).
noun( 'carnforth', '-', proper, loc ).
noun( 'carnoustie', '-', proper, loc ).
noun( 'carol', '-', proper, per ).
noun( 'carole', '-', proper, per ).
noun( 'carolina', '-', proper, loc ).
noun( 'caroline', '-', proper, per ).
noun( 'carolyn', '-', proper, per ).
noun( 'carrie', '-', proper, per ).
noun( 'carron', '-', proper, loc ).
noun( 'cartagena', '-', proper, loc ).
noun( 'casablanca', '-', proper, loc ).
noun( 'castleford', '-', proper, loc ).
noun( 'catalan', '-', mass, _ ).
noun( 'catania', '-', proper, loc ).
noun( 'cath', '-', proper, per ).
noun( 'catherine', '-', proper, per ).
noun( 'catholic', 'catholics', count, _ ).
noun( 'catholicism', '-', mass, _ ).
noun( 'cathy', '-', proper, per ).
noun( 'caucasian', 'caucasians', count, _ ).
noun( 'cavan', '-', proper, loc ).
noun( 'cdr', '-', proper, per ).
noun( 'cdre', '-', proper, per ).
noun( 'cecil', '-', proper, per ).
noun( 'cecilia', '-', proper, per ).
noun( 'cecily', '-', proper, per ).
noun( 'cedric', '-', proper, per ).
noun( 'cefn-mawr', '-', proper, loc ).
noun( 'celia', '-', proper, per ).
noun( 'celt', 'celts', count, _ ).
noun( 'celtic', 'celtics', count, _ ).
noun( 'ceylon', '-', proper, loc ).
noun( 'chad', '-', proper, loc ).
noun( 'chadian', 'chadians', count, _ ).
noun( 'chalfont', '-', proper, loc ).
noun( 'chalfont st giles', '-', proper, loc ).
noun( 'chalfont st peter', '-', proper, loc ).
noun( 'changchun', '-', proper, loc ).
noun( 'changsha', '-', proper, loc ).
noun( 'chard', '-', proper, loc ).
noun( 'charles', '-', proper, per ).
noun( 'charleston', 'charlestons', count, _ ).
noun( 'charlie', '-', proper, per ).
noun( 'charlotte', '-', proper, per ).
noun( 'chartism', '-', mass, _ ).
noun( 'chartist', 'chartists', count, _ ).
noun( 'charybdis', '-', proper, _ ).
noun( 'chas', '-', proper, per ).
noun( 'chatham', '-', proper, loc ).
noun( 'cheadle', '-', proper, loc ).
noun( 'cheddar', '-', mass, _ ).
noun( 'chelmsford', '-', proper, loc ).
noun( 'chelsea', '-', proper, loc ).
noun( 'cheltenham', '-', proper, loc ).
noun( 'chelyabinsk', '-', proper, loc ).
noun( 'chengchow', '-', proper, loc ).
noun( 'chengtu', '-', proper, loc ).
noun( 'chepstow', '-', proper, loc ).
noun( 'chesham', '-', proper, loc ).
noun( 'cheshire', '-', proper, loc ).
noun( 'chester', '-', proper, loc ).
noun( 'chester-le-street', '-', proper, loc ).
noun( 'chesterfield', '-', proper, loc ).
noun( 'chianti', '-', mass, _ ).
noun( 'chicago', '-', proper, loc ).
noun( 'chichester', '-', proper, loc ).
noun( 'chile', '-', proper, loc ).
noun( 'chilean', 'chileans', count, _ ).
noun( 'china', '-', proper, loc ).
noun( 'chinchow', '-', proper, loc ).
noun( 'chinese', 'chinese', both, _ ).
noun( 'chippendale', 'chippendales', count, _ ).
noun( 'chippenham', '-', proper, loc ).
noun( 'chloe', '-', proper, per ).
noun( 'chopwell', '-', proper, loc ).
noun( 'chorley', '-', proper, loc ).
noun( 'chris', '-', proper, per ).
noun( 'chrissie', '-', proper, per ).
noun( 'chrissy', '-', proper, per ).
noun( 'christ', '-', proper, _ ).
noun( 'christchurch', '-', proper, loc ).
noun( 'christendom', 'christendoms', count, _ ).
noun( 'christian', 'christians', count, _ ).
noun( 'christian', '-', proper, per ).
noun( 'christianity', '-', mass, _ ).
noun( 'christina', '-', proper, per ).
noun( 'christine', '-', proper, per ).
noun( 'christmas', 'christmases', count, _ ).
noun( 'christmas', '-', proper, _ ).
noun( 'christmas-box', 'christmas-boxes', count, _ ).
noun( 'christmas-tree', 'christmas-trees', count, _ ).
noun( 'christmastide', 'christmastides', count, _ ).
noun( 'christmastime', 'christmastimes', count, _ ).
noun( 'christopher', '-', proper, per ).
noun( 'chungking', '-', proper, loc ).
noun( 'cincinnati', '-', proper, loc ).
noun( 'cinderella', 'cinderellas', count, _ ).
noun( 'cinderford', '-', proper, loc ).
noun( 'cirencester', '-', proper, loc ).
noun( 'civvy street', '-', proper, _ ).
noun( 'clackmannon', '-', proper, loc ).
noun( 'clacton', '-', proper, loc ).
noun( 'clare', '-', proper, per ).
noun( 'clarkston', '-', proper, loc ).
noun( 'clarrie', '-', proper, per ).
noun( 'classics', 'classics', mass, _ ).
noun( 'claud', '-', proper, per ).
noun( 'claude', '-', proper, per ).
noun( 'clay cross', '-', proper, loc ).
noun( 'cleator moor', '-', proper, loc ).
noun( 'cleethorpes', '-', proper, loc ).
noun( 'clem', '-', proper, per ).
noun( 'clement', '-', proper, per ).
noun( 'clermont-ferrand', '-', proper, loc ).
noun( 'clevedon', '-', proper, loc ).
noun( 'cleveland', '-', proper, loc ).
noun( 'cliff', '-', proper, per ).
noun( 'clifford', '-', proper, per ).
noun( 'clitheroe', '-', proper, loc ).
noun( 'clive', '-', proper, per ).
noun( 'clowne', '-', proper, loc ).
noun( 'clwyd', '-', proper, loc ).
noun( 'clydach', '-', proper, loc ).
noun( 'clydebank', '-', proper, loc ).
noun( 'co', '-', count, _ ).
noun( 'co-op', '-', count, _ ).
noun( 'coalville', '-', proper, loc ).
noun( 'coatbridge', '-', proper, loc ).
noun( 'cobham', '-', proper, loc ).
noun( 'coca-cola', 'coca-colas', both, _ ).
noun( 'cockermouth', '-', proper, loc ).
noun( 'codsall', '-', proper, loc ).
noun( 'coimbatore', '-', proper, loc ).
noun( 'col', '-', proper, per ).
noun( 'colchester', '-', proper, loc ).
noun( 'coleshill', '-', proper, loc ).
noun( 'colin', '-', proper, per ).
noun( 'coll', '-', proper, _ ).
noun( 'colne', '-', proper, loc ).
noun( 'cologne', '-', proper, loc ).
noun( 'colombia', '-', proper, loc ).
noun( 'colombian', 'colombians', count, _ ).
noun( 'colombo', '-', proper, loc ).
noun( 'colorado', '-', proper, loc ).
noun( 'columbia', '-', proper, loc ).
noun( 'columbus', '-', proper, loc ).
noun( 'colwyn bay', '-', proper, loc ).
noun( 'confucian', 'confucians', count, _ ).
noun( 'congleton', '-', proper, loc ).
noun( 'congo', '-', proper, loc ).
noun( 'congolese', 'congolese', count, _ ).
noun( 'connah\'s quay', '-', proper, loc ).
noun( 'connaught', '-', proper, loc ).
noun( 'connecticut', '-', proper, loc ).
noun( 'connie', '-', proper, per ).
noun( 'cons', '-', proper, _ ).
noun( 'consett', '-', proper, loc ).
noun( 'constance', '-', proper, per ).
noun( 'conwy', '-', proper, loc ).
noun( 'cookham', '-', proper, loc ).
noun( 'copenhagen', '-', proper, loc ).
noun( 'coppull', '-', proper, loc ).
noun( 'copt', 'copts', count, _ ).
noun( 'coptic', 'coptics', count, _ ).
noun( 'corby', '-', proper, loc ).
noun( 'cordoba', '-', proper, loc ).
noun( 'corinthian', 'corinthians', count, _ ).
noun( 'cork', '-', proper, loc ).
noun( 'cornwall', '-', proper, loc ).
noun( 'corp', '-', proper, _ ).
noun( 'corps diplomatique', '-', count, _ ).
noun( 'corsham', '-', proper, loc ).
noun( 'corunna', '-', proper, loc ).
noun( 'cosenza', '-', proper, loc ).
noun( 'costa rica', '-', proper, loc ).
noun( 'costa rican', 'costa ricans', count, _ ).
noun( 'coventry', '-', proper, loc ).
noun( 'cowdenbeath', '-', proper, loc ).
noun( 'cowes', '-', proper, loc ).
noun( 'cpl', '-', proper, per ).
noun( 'cramlington', '-', proper, loc ).
noun( 'crawley', '-', proper, loc ).
noun( 'creole', 'creoles', both, _ ).
noun( 'cres', '-', proper, _ ).
noun( 'creswell', '-', proper, loc ).
noun( 'crewe', '-', proper, loc ).
noun( 'crieff', '-', proper, loc ).
noun( 'croesus', '-', proper, per ).
noun( 'cromer', '-', proper, loc ).
noun( 'crook', '-', proper, loc ).
noun( 'crosby', '-', proper, loc ).
noun( 'crowborough', '-', proper, loc ).
noun( 'crowthorne', '-', proper, loc ).
noun( 'croydon', '-', proper, loc ).
noun( 'cuba', '-', proper, loc ).
noun( 'cuban', 'cubans', count, _ ).
noun( 'cudworth', '-', proper, loc ).
noun( 'cuffley', '-', proper, loc ).
noun( 'culcheth', '-', proper, loc ).
noun( 'cumberland', '-', proper, loc ).
noun( 'cumbernauld', '-', proper, loc ).
noun( 'cumbria', '-', proper, loc ).
noun( 'cumnock', '-', proper, loc ).
noun( 'cupar', '-', proper, loc ).
noun( 'cupid', '-', proper, per ).
noun( 'curitiba', '-', proper, loc ).
noun( 'currie', '-', proper, loc ).
noun( 'cwmbran', '-', proper, loc ).
noun( 'cynthia', '-', proper, per ).
noun( 'cypriot', 'cypriots', count, _ ).
noun( 'cyprus', '-', proper, loc ).
noun( 'cyril', '-', proper, per ).
noun( 'czech', 'czechs', both, _ ).
noun( 'czechoslovak', 'czechoslovaks', count, _ ).
noun( 'czechoslovakia', '-', proper, loc ).
noun( 'czechoslovakian', 'czechoslovakians', count, _ ).
noun( 'd-day', '-', proper, _ ).
noun( 'ddt', '-', mass, _ ).
noun( 'des', '-', count, _ ).
noun( 'dg', '-', proper, per ).
noun( 'diy', '-', mass, _ ).
noun( 'dj', '-', count, _ ).
noun( '-', 'djs', count, _ ).
noun( 'dlitt', '-', count, _ ).
noun( 'dm', 'dm', count, _ ).
noun( 'dna', '-', mass, _ ).
noun( 'dphil', '-', count, _ ).
noun( 'dss', '-', count, _ ).
noun( 'dsc', '-', count, _ ).
noun( 'dti', '-', count, _ ).
noun( '-', 'dts', count, _ ).
noun( 'dacca', '-', proper, loc ).
noun( 'dail eireann', '-', count, _ ).
noun( 'daisy', '-', proper, per ).
noun( 'dakar', '-', proper, loc ).
noun( 'dakota', '-', proper, loc ).
noun( 'dalkeith', '-', proper, loc ).
noun( 'dallas', '-', proper, loc ).
noun( 'dalry', '-', proper, loc ).
noun( 'dalton', '-', proper, loc ).
noun( 'damascus', '-', proper, loc ).
noun( 'damocles', '-', proper, per ).
noun( 'dan', '-', proper, per ).
noun( 'dane', 'danes', count, _ ).
noun( 'daniel', 'daniels', count, _ ).
noun( 'daniel', '-', proper, per ).
noun( 'danish', '-', mass, _ ).
noun( 'danny', '-', proper, per ).
noun( 'daphne', '-', proper, per ).
noun( 'darby', '-', proper, per ).
noun( 'darenth', '-', proper, loc ).
noun( 'darfield', '-', proper, loc ).
noun( 'darkey', 'darkeys', count, _ ).
noun( 'darkie', 'darkies', count, _ ).
noun( 'darky', 'darkies', count, _ ).
noun( 'darlington', '-', proper, loc ).
noun( 'darmstadt', '-', proper, loc ).
noun( 'dartford', '-', proper, loc ).
noun( 'dartmouth', '-', proper, loc ).
noun( 'darwen', '-', proper, loc ).
noun( 'dave', '-', proper, per ).
noun( 'daventry', '-', proper, loc ).
noun( 'davey', '-', proper, per ).
noun( 'david', '-', proper, per ).
noun( 'dawley', '-', proper, loc ).
noun( 'dawlish', '-', proper, loc ).
noun( 'dawn', '-', proper, per ).
noun( 'dayton', '-', proper, loc ).
noun( 'deal', '-', proper, loc ).
noun( 'dean', '-', proper, per ).
noun( 'debbie', '-', proper, per ).
noun( 'debby', '-', proper, per ).
noun( 'deborah', '-', proper, per ).
noun( 'dec', '-', proper, _ ).
noun( 'decalogue', 'decalogues', count, _ ).
noun( 'december', 'decembers', count, _ ).
noun( 'december', '-', proper, _ ).
noun( 'deirdre', '-', proper, per ).
noun( 'delaware', '-', proper, loc ).
noun( 'delhi', '-', proper, loc ).
noun( 'denain', '-', proper, loc ).
noun( 'denbigh', '-', proper, loc ).
noun( 'denis', '-', proper, per ).
noun( 'denise', '-', proper, per ).
noun( 'denmark', '-', proper, loc ).
noun( 'dennis', '-', proper, per ).
noun( 'denny', '-', proper, loc ).
noun( 'denver', '-', proper, loc ).
noun( 'dept', '-', proper, _ ).
noun( 'derby', '-', proper, loc ).
noun( 'derbyshire', '-', proper, loc ).
noun( 'dereham', '-', proper, loc ).
noun( 'derek', '-', proper, per ).
noun( 'des', '-', proper, per ).
noun( 'desmond', '-', proper, per ).
noun( 'dessau', '-', proper, loc ).
noun( 'detroit', '-', proper, loc ).
noun( 'deutschmark', 'deutschmarks', count, _ ).
noun( 'devizes', '-', proper, loc ).
noun( 'devon', '-', proper, loc ).
noun( 'dewsbury', '-', proper, loc ).
noun( 'di', '-', proper, per ).
noun( 'diana', '-', proper, per ).
noun( 'diaspora', 'diasporas', count, _ ).
noun( 'dick', '-', proper, per ).
noun( 'dicky', '-', proper, per ).
noun( 'dictaphone', 'dictaphones', count, _ ).
noun( 'didcot', '-', proper, loc ).
noun( 'dijon', '-', proper, loc ).
noun( 'dingle', '-', proper, loc ).
noun( 'dinnington', '-', proper, loc ).
noun( 'dip', '-', count, _ ).
noun( 'dip ed', '-', count, _ ).
noun( 'dir', '-', count, _ ).
noun( 'dives', '-', proper, per ).
noun( 'djibouti', '-', proper, loc ).
noun( 'djiboutian', 'djiboutians', count, _ ).
noun( 'dnepropetrovsk', '-', proper, loc ).
noun( 'doe', '-', count, _ ).
noun( 'dolly', '-', proper, per ).
noun( 'domesday', '-', proper, _ ).
noun( 'dominic', '-', proper, per ).
noun( 'dominica', '-', proper, loc ).
noun( 'dominican', 'dominicans', count, _ ).
noun( 'don', '-', proper, per ).
noun( 'donald', '-', proper, per ).
noun( 'doncaster', '-', proper, loc ).
noun( 'donegal', '-', proper, loc ).
noun( 'donetsk', '-', proper, loc ).
noun( 'doomsday', '-', proper, _ ).
noun( 'dora', '-', proper, per ).
noun( 'dorchester', '-', proper, loc ).
noun( 'dordrecht', '-', proper, loc ).
noun( 'doreen', '-', proper, per ).
noun( 'doris', '-', proper, per ).
noun( 'dorking', '-', proper, loc ).
noun( 'dorothy', '-', proper, per ).
noun( 'dorset', '-', proper, loc ).
noun( 'dortmund', '-', proper, loc ).
noun( 'douai', '-', proper, loc ).
noun( 'doug', '-', proper, per ).
noun( 'douglas', '-', proper, per ).
noun( 'dover', '-', proper, loc ).
noun( 'down', '-', proper, loc ).
noun( 'downing street', '-', proper, _ ).
noun( 'dr', '-', proper, per ).
noun( 'drayton', '-', proper, loc ).
noun( 'dresden', '-', proper, loc ).
noun( 'driffield', '-', proper, loc ).
noun( 'drogheda', '-', proper, loc ).
noun( 'droitwich', '-', proper, loc ).
noun( 'dronfield', '-', proper, loc ).
noun( 'druid', 'druids', count, _ ).
noun( 'dublin', '-', proper, loc ).
noun( 'dubliner', 'dubliners', count, _ ).
noun( 'duce', '-', count, _ ).
noun( 'dudley', '-', proper, loc ).
noun( 'duisburg', '-', proper, loc ).
noun( 'dukinfield', '-', proper, loc ).
noun( 'dumbarton', '-', proper, loc ).
noun( 'dumfries', '-', proper, loc ).
noun( 'dun laoghaire', '-', proper, loc ).
noun( 'duncan', '-', proper, per ).
noun( 'dundalk', '-', proper, loc ).
noun( 'dundee', '-', proper, loc ).
noun( 'dunfermline', '-', proper, loc ).
noun( 'dunkirk', '-', proper, loc ).
noun( 'dunoon', '-', proper, loc ).
noun( 'dunstable', '-', proper, loc ).
noun( 'durban', '-', proper, loc ).
noun( 'durham', '-', proper, loc ).
noun( 'durrington', '-', proper, loc ).
noun( 'dursley', '-', proper, loc ).
noun( 'dusseldorf', '-', proper, loc ).
noun( 'dutch', '-', mass, _ ).
noun( 'dutchman', 'dutchmen', count, _ ).
noun( 'dyfed', '-', proper, loc ).
noun( 'ec', '-', count, _ ).
noun( 'edp', '-', mass, _ ).
noun( 'eec', '-', count, _ ).
noun( 'eeg', '-', count, _ ).
noun( 'efta', '-', count, _ ).
noun( 'esp', '-', mass, _ ).
noun( 'eaglescliffe', '-', proper, loc ).
noun( 'ealing', '-', proper, loc ).
noun( 'earl shilton', '-', proper, loc ).
noun( 'easington', '-', proper, loc ).
noun( 'east dereham', '-', proper, loc ).
noun( 'east grinstead', '-', proper, loc ).
noun( 'east kilbride', '-', proper, loc ).
noun( 'east retford', '-', proper, loc ).
noun( 'eastbourne', '-', proper, loc ).
noun( 'easter', 'easters', count, _ ).
noun( 'easter', '-', proper, _ ).
noun( 'eastleigh', '-', proper, loc ).
noun( 'ebbw vale', '-', proper, loc ).
noun( 'eccles', '-', proper, loc ).
noun( 'ecuador', '-', proper, loc ).
noun( 'ecuadorian', 'ecuadorians', count, _ ).
noun( 'ed', '-', proper, per ).
noun( 'eddie', '-', proper, per ).
noun( 'eddy', '-', proper, per ).
noun( 'eden', '-', proper, _ ).
noun( 'edenbridge', '-', proper, loc ).
noun( 'edgar', '-', proper, per ).
noun( 'edinburgh', '-', proper, loc ).
noun( 'edith', '-', proper, per ).
noun( 'edmund', '-', proper, per ).
noun( 'edward', '-', proper, per ).
noun( 'edwardian', 'edwardians', count, _ ).
noun( 'egremont', '-', proper, loc ).
noun( 'egypt', '-', proper, loc ).
noun( 'egyptian', 'egyptians', count, _ ).
noun( 'eiche', '-', proper, loc ).
noun( 'eileen', '-', proper, per ).
noun( 'eindhoven', '-', proper, loc ).
noun( 'el dorado', '-', count, _ ).
noun( 'el salvador', '-', proper, loc ).
noun( 'elaine', '-', proper, per ).
noun( 'elastoplast', '-', mass, _ ).
noun( 'elderslie', '-', proper, loc ).
noun( 'eleanor', '-', proper, per ).
noun( 'elgin', '-', proper, loc ).
noun( 'eliza', '-', proper, per ).
noun( 'elizabeth', '-', proper, per ).
noun( 'elizabethan', 'elizabethans', count, _ ).
noun( 'elland', '-', proper, loc ).
noun( 'ellen', '-', proper, per ).
noun( 'ellesmere', '-', proper, loc ).
noun( 'ellesmere port', '-', proper, loc ).
noun( 'ellie', '-', proper, per ).
noun( 'elloughton', '-', proper, loc ).
noun( 'elsie', '-', proper, per ).
noun( 'elstree', '-', proper, loc ).
noun( 'ely', '-', proper, loc ).
noun( 'elysium', '-', proper, _ ).
noun( 'emily', '-', proper, per ).
noun( 'emma', '-', proper, per ).
noun( 'emsworth', '-', proper, loc ).
noun( 'enfield', '-', proper, loc ).
noun( 'england', '-', proper, loc ).
noun( 'english', '-', mass, _ ).
noun( 'englishman', 'englishmen', count, _ ).
noun( 'englishwoman', 'englishwomen', count, _ ).
noun( 'enoch', '-', proper, per ).
noun( 'enschede', '-', proper, loc ).
noun( 'epiphany', '-', proper, _ ).
noun( 'epping', '-', proper, loc ).
noun( 'epsom', '-', proper, loc ).
noun( 'erfurt', '-', proper, loc ).
noun( 'eric', '-', proper, per ).
noun( 'erica', '-', proper, per ).
noun( 'erin', '-', proper, loc ).
noun( 'eritrea', '-', proper, loc ).
noun( 'eritrean', 'eritreans', count, _ ).
noun( 'erlangen', '-', proper, loc ).
noun( 'ernest', '-', proper, per ).
noun( 'ernie', '-', proper, per ).
noun( 'erse', '-', mass, _ ).
noun( 'esfahan', '-', proper, loc ).
noun( 'eskimo', 'eskimos', count, _ ).
noun( 'esperanto', '-', mass, _ ).
noun( 'esq', '-', proper, per ).
noun( 'esquire', 'esquires', count, _ ).
noun( 'essen', '-', proper, loc ).
noun( 'essex', '-', proper, loc ).
noun( 'esther', '-', proper, per ).
noun( 'ethel', '-', proper, per ).
noun( 'ethiopia', '-', proper, loc ).
noun( 'ethiopian', 'ethiopians', count, _ ).
noun( 'eucharist', 'eucharists', count, _ ).
noun( 'eugene', '-', proper, per ).
noun( 'eunice', '-', proper, per ).
noun( 'eurasia', '-', proper, loc ).
noun( 'eurasian', 'eurasians', count, _ ).
noun( 'eurodollar', 'eurodollars', count, _ ).
noun( 'europe', '-', proper, loc ).
noun( 'european', 'europeans', count, _ ).
noun( 'eurovision', '-', proper, _ ).
noun( 'eva', '-', proper, per ).
noun( 'eve', '-', proper, per ).
noun( 'evelyn', '-', proper, per ).
noun( 'evesham', '-', proper, loc ).
noun( 'ewell', '-', proper, loc ).
noun( 'excellency', 'excellencies', count, _ ).
noun( 'exeter', '-', proper, loc ).
noun( 'exmouth', '-', proper, loc ).
noun( 'fa', '-', count, _ ).
noun( 'fao', '-', count, _ ).
noun( 'fbi', '-', count, _ ).
noun( 'fm', '-', mass, _ ).
noun( 'fo', '-', count, _ ).
noun( 'frs', '-', count, _ ).
noun( 'fabian', 'fabians', count, _ ).
noun( 'falkirk', '-', proper, loc ).
noun( 'falmouth', '-', proper, loc ).
noun( 'fanny', '-', proper, per ).
noun( 'farnham', '-', proper, loc ).
noun( 'farnworth', '-', proper, loc ).
noun( 'farsi', '-', mass, _ ).
noun( 'fauldhouse', '-', proper, loc ).
noun( 'faversham', '-', proper, loc ).
noun( 'fawley', '-', proper, loc ).
noun( 'featherstone', '-', proper, loc ).
noun( 'feb', '-', proper, _ ).
noun( 'february', 'februaries', count, _ ).
noun( 'february', '-', proper, _ ).
noun( 'fed', '-', count, _ ).
noun( 'felicity', '-', proper, per ).
noun( 'felix', '-', proper, per ).
noun( 'felixstowe', '-', proper, loc ).
noun( 'fermanagh', '-', proper, loc ).
noun( 'ferrara', '-', proper, loc ).
noun( 'ferryhill', '-', proper, loc ).
noun( 'fife', '-', proper, loc ).
noun( 'fiji', '-', proper, loc ).
noun( 'fijian', 'fijians', count, _ ).
noun( 'filipino', 'filipinos', count, _ ).
noun( 'finland', '-', proper, loc ).
noun( 'finn', 'finns', count, _ ).
noun( 'finnish', '-', mass, _ ).
noun( 'fiona', '-', proper, per ).
noun( 'firenze', '-', proper, loc ).
noun( 'fleet', '-', proper, loc ).
noun( 'fleet street', '-', proper, _ ).
noun( 'fleetwood', '-', proper, loc ).
noun( 'flemish', '-', mass, _ ).
noun( 'flint', '-', proper, loc ).
noun( 'flora', '-', proper, per ).
noun( 'florence', '-', proper, per ).
noun( 'florida', '-', proper, loc ).
noun( 'florrie', '-', proper, per ).
noun( 'foggia', '-', proper, loc ).
noun( 'folkestone', '-', proper, loc ).
noun( 'foochow', '-', proper, loc ).
noun( 'forfar', '-', proper, loc ).
noun( 'forli', '-', proper, loc ).
noun( 'formby', '-', proper, loc ).
noun( 'formica', '-', mass, _ ).
noun( 'formosa', '-', proper, loc ).
noun( 'fort lauderdale', '-', proper, loc ).
noun( 'fort william', '-', proper, loc ).
noun( 'fort worth', '-', proper, loc ).
noun( 'fortaleza', '-', proper, loc ).
noun( 'fowey', '-', proper, loc ).
noun( 'fr', '-', proper, per ).
noun( 'fr\"aulein', 'fr\"auleins', count, _ ).
noun( 'frampton cotterell', '-', proper, loc ).
noun( 'fran', '-', proper, per ).
noun( 'france', '-', proper, loc ).
noun( 'frances', '-', proper, per ).
noun( 'francis', '-', proper, per ).
noun( 'franciscan', 'franciscans', count, _ ).
noun( 'frank', 'franks', count, _ ).
noun( 'frank', '-', proper, per ).
noun( 'frankfurt', '-', proper, loc ).
noun( 'frankie', '-', proper, per ).
noun( 'fraserburgh', '-', proper, loc ).
noun( 'frau', 'frauen', count, _ ).
noun( 'freckleton', '-', proper, loc ).
noun( 'fred', '-', proper, per ).
noun( 'freda', '-', proper, per ).
noun( 'freddie', '-', proper, per ).
noun( 'freddy', '-', proper, per ).
noun( 'frederick', '-', proper, per ).
noun( 'frederiksberg', '-', proper, loc ).
noun( 'freemason', 'freemasons', count, _ ).
noun( 'freemasonry', '-', mass, _ ).
noun( 'freiburg', '-', proper, loc ).
noun( 'french', '-', mass, _ ).
noun( 'frenchman', 'frenchmen', count, _ ).
noun( 'frenchwoman', 'frenchwomen', count, _ ).
noun( 'freshwater', '-', proper, loc ).
noun( 'fri', '-', proper, _ ).
noun( 'friday', 'fridays', count, _ ).
noun( 'friday', '-', proper, _ ).
noun( 'frinton', '-', proper, loc ).
noun( 'frodsham', '-', proper, loc ).
noun( 'frome', '-', proper, loc ).
noun( 'furth', '-', proper, loc ).
noun( 'fushun', '-', proper, loc ).
noun( 'g-man', 'g-men', count, _ ).
noun( 'gatt', '-', count, _ ).
noun( 'gb', '-', proper, _ ).
noun( 'gcse', '-', count, _ ).
noun( '-', 'gcses', count, _ ).
noun( 'ghq', '-', proper, _ ).
noun( 'gi', '-', count, _ ).
noun( '-', 'gis', count, _ ).
noun( 'gmt', '-', proper, _ ).
noun( 'gnp', '-', count, _ ).
noun( 'gp', '-', count, _ ).
noun( '-', 'gp\'s', count, _ ).
noun( 'gabon', '-', proper, loc ).
noun( 'gabonese', 'gabonese', count, _ ).
noun( 'gael', 'gaels', count, _ ).
noun( 'gaelic', 'gaelics', count, _ ).
noun( 'gainsborough', 'gainsboroughs', count, _ ).
noun( 'gainsborough', '-', proper, loc ).
noun( 'galashiels', '-', proper, loc ).
noun( 'galloway', '-', proper, loc ).
noun( 'gallup', '-', proper, _ ).
noun( 'galway', '-', proper, loc ).
noun( 'gambia', '-', proper, loc ).
noun( 'gambian', 'gambians', count, _ ).
noun( 'gareth', '-', proper, per ).
noun( 'garforth', '-', proper, loc ).
noun( 'garrowhill', '-', proper, loc ).
noun( 'gary', '-', proper, per ).
noun( 'gateshead', '-', proper, loc ).
noun( 'gaul', 'gauls', count, _ ).
noun( 'gavin', '-', proper, per ).
noun( 'gdansk', '-', proper, loc ).
noun( 'gdn', '-', proper, _ ).
noun( 'gdns', '-', proper, _ ).
noun( 'geiger', 'geigers', count, _ ).
noun( 'gelligaer', '-', proper, loc ).
noun( 'gelsenkirchen', '-', proper, loc ).
noun( 'gemini', '-', proper, _ ).
noun( 'gen', '-', proper, per ).
noun( 'gene', '-', proper, per ).
noun( 'geneva', '-', proper, loc ).
noun( 'genoa', '-', proper, loc ).
noun( 'genova', '-', proper, loc ).
noun( 'geoff', '-', proper, per ).
noun( 'geoffrey', '-', proper, per ).
noun( 'george', '-', proper, per ).
noun( 'georgia', '-', proper, loc ).
noun( 'georgian', 'georgians', count, _ ).
noun( 'georgie', '-', proper, per ).
noun( 'gerald', '-', proper, per ).
noun( 'geraldine', '-', proper, per ).
noun( 'gerard', '-', proper, per ).
noun( 'german', 'germans', both, _ ).
noun( 'germany', '-', proper, loc ).
noun( 'gerry', '-', proper, per ).
noun( 'gertie', '-', proper, per ).
noun( 'gertrude', '-', proper, per ).
noun( 'gestapo', 'gestapos', count, _ ).
noun( 'ghana', '-', proper, loc ).
noun( 'ghanaian', 'ghanaians', count, _ ).
noun( 'ghent', '-', proper, loc ).
noun( 'gibraltar', '-', proper, loc ).
noun( 'gibraltarian', 'gibraltarians', count, _ ).
noun( 'giffnock', '-', proper, loc ).
noun( 'gilbert', '-', proper, per ).
noun( 'giles', '-', proper, per ).
noun( 'gilfach goch', '-', proper, loc ).
noun( 'gill', '-', proper, per ).
noun( 'gillian', '-', proper, per ).
noun( 'gillingham', '-', proper, loc ).
noun( 'gina', '-', proper, per ).
noun( 'girvan', '-', proper, loc ).
noun( 'gladys', '-', proper, per ).
noun( 'glamorgan', '-', proper, loc ).
noun( 'glasgow', '-', proper, loc ).
noun( 'glastonbury', '-', proper, loc ).
noun( 'glaswegian', 'glaswegians', count, _ ).
noun( 'glen', '-', proper, per ).
noun( 'glengarry', 'glengarries', count, _ ).
noun( 'glenrothes', '-', proper, loc ).
noun( 'gloria', '-', proper, per ).
noun( 'glossop', '-', proper, loc ).
noun( 'gloucester', '-', proper, loc ).
noun( 'gloucestershire', '-', proper, loc ).
noun( 'glusburn', '-', proper, loc ).
noun( 'god', '-', proper, _ ).
noun( 'godalming', '-', proper, loc ).
noun( 'godfrey', '-', proper, per ).
noun( 'godmanchester', '-', proper, loc ).
noun( 'goffs oak', '-', proper, loc ).
noun( 'golbourne', '-', proper, loc ).
noun( 'goliath', 'goliaths', count, _ ).
noun( 'goole', '-', proper, loc ).
noun( 'gordon', '-', proper, per ).
noun( 'gorgon', 'gorgons', count, _ ).
noun( 'gorgonzola', '-', mass, _ ).
noun( 'gorky', '-', proper, loc ).
noun( 'gorseinon', '-', proper, loc ).
noun( 'gosport', '-', proper, loc ).
noun( 'goteborg', '-', proper, loc ).
noun( 'goth', 'goths', count, _ ).
noun( 'gothenburg', '-', proper, loc ).
noun( 'gothic', '-', mass, _ ).
noun( 'gottingen', '-', proper, loc ).
noun( 'gourock', '-', proper, loc ).
noun( 'gov', '-', proper, per ).
noun( 'governor-general', 'governor-generals', count, _ ).
noun( 'grampian', '-', proper, loc ).
noun( 'granada', '-', proper, loc ).
noun( 'grand prix', 'grands prix', count, _ ).
noun( 'grangemouth', '-', proper, loc ).
noun( 'grantham', '-', proper, loc ).
noun( 'gravenhage', '-', proper, loc ).
noun( 'gravesend', '-', proper, loc ).
noun( 'graz', '-', proper, loc ).
noun( 'great harwood', '-', proper, loc ).
noun( 'great malvern', '-', proper, loc ).
noun( 'great shelford', '-', proper, loc ).
noun( 'great yarmouth', '-', proper, loc ).
noun( 'greece', '-', proper, loc ).
noun( 'greek', 'greeks', both, _ ).
noun( 'greenock', '-', proper, loc ).
noun( 'greenwich', '-', proper, loc ).
noun( 'greg', '-', proper, per ).
noun( 'gregory', '-', proper, per ).
noun( 'grenada', '-', proper, loc ).
noun( 'grenadian', 'grenadians', count, _ ).
noun( 'grimethorpe', '-', proper, loc ).
noun( 'grimsby', '-', proper, loc ).
noun( 'grinstead', '-', proper, loc ).
noun( 'groningen', '-', proper, loc ).
noun( 'grundyism', '-', mass, _ ).
noun( 'guadalajara', '-', proper, loc ).
noun( 'guatemala', '-', proper, loc ).
noun( 'guatemala city', '-', proper, loc ).
noun( 'guatemalan', 'guatemalans', count, _ ).
noun( 'guayaquil', '-', proper, loc ).
noun( 'guild-hall', 'guild-halls', count, _ ).
noun( 'guildford', '-', proper, loc ).
noun( 'guildhall', 'guildhalls', count, _ ).
noun( 'guinea', '-', proper, loc ).
noun( 'guinean', 'guineans', count, _ ).
noun( 'guinness', 'guinnesses', count, _ ).
noun( 'guisborough', '-', proper, loc ).
noun( 'gurkha', 'gurkhas', count, _ ).
noun( 'guy', '-', proper, per ).
noun( 'guy\'s', '-', proper, _ ).
noun( 'guyana', '-', proper, loc ).
noun( 'guyanese', 'guyanese', count, _ ).
noun( 'gwen', '-', proper, per ).
noun( 'gwendoline', '-', proper, per ).
noun( 'gwent', '-', proper, loc ).
noun( 'gwersyllt', '-', proper, loc ).
noun( 'gwynedd', '-', proper, loc ).
noun( 'gypsy', 'gypsies', count, _ ).
noun( 'h-bomb', 'h-bombs', count, _ ).
noun( 'hm', '-', proper, per ).
noun( 'hmso', '-', proper, _ ).
noun( 'hp', '-', mass, _ ).
noun( 'hq', '-', proper, _ ).
noun( 'hrh', '-', proper, per ).
noun( 'haarlem', '-', proper, loc ).
noun( 'hackney', '-', proper, loc ).
noun( 'haddington', '-', proper, loc ).
noun( 'hades', '-', proper, _ ).
noun( 'hadji', 'hadjis', count, _ ).
noun( 'hagen', '-', proper, loc ).
noun( 'hagley', '-', proper, loc ).
noun( 'hailsham', '-', proper, loc ).
noun( 'haiphong', '-', proper, loc ).
noun( 'haiti', '-', proper, loc ).
noun( 'haitian', 'haitians', count, _ ).
noun( 'hal', '-', proper, per ).
noun( 'halesowen', '-', proper, loc ).
noun( 'halifax', '-', proper, loc ).
noun( 'halle', '-', proper, loc ).
noun( 'hallowe\'en', 'hallowe\'ens', count, _ ).
noun( 'hallowe\'en', '-', proper, _ ).
noun( 'halstead', '-', proper, loc ).
noun( 'hamburg', '-', proper, loc ).
noun( 'hamhung', '-', proper, loc ).
noun( 'hamilton', '-', proper, loc ).
noun( 'hammersmith', '-', proper, loc ).
noun( 'hampreston', '-', proper, loc ).
noun( 'hampshire', '-', proper, loc ).
noun( 'hangchon', '-', proper, loc ).
noun( 'hanoi', '-', proper, loc ).
noun( 'hanover', '-', proper, loc ).
noun( 'hansard', 'hansards', count, _ ).
noun( 'harbin', '-', proper, loc ).
noun( 'haringey', '-', proper, loc ).
noun( 'harley street', '-', proper, _ ).
noun( 'harlow', '-', proper, loc ).
noun( 'harold', '-', proper, per ).
noun( 'harpenden', '-', proper, loc ).
noun( 'harriet', '-', proper, per ).
noun( 'harrogate', '-', proper, loc ).
noun( 'harrow', '-', proper, loc ).
noun( 'harry', '-', proper, per ).
noun( 'hartlepool', '-', proper, loc ).
noun( 'hartley', '-', proper, loc ).
noun( 'harvey', '-', proper, per ).
noun( 'harwich', '-', proper, loc ).
noun( 'harwood', '-', proper, loc ).
noun( 'harworth', '-', proper, loc ).
noun( 'haslemere', '-', proper, loc ).
noun( 'haslingden', '-', proper, loc ).
noun( 'hastings', '-', proper, loc ).
noun( 'hatfield', '-', proper, loc ).
noun( 'havana', 'havanas', count, _ ).
noun( 'havana', '-', proper, loc ).
noun( 'haverfordwest', '-', proper, loc ).
noun( 'haverhill', '-', proper, loc ).
noun( 'havering', '-', proper, loc ).
noun( 'hawaii', '-', proper, loc ).
noun( 'hawick', '-', proper, loc ).
noun( 'hawke\'s bay', '-', proper, loc ).
noun( 'hayling', '-', proper, loc ).
noun( 'haywards heath', '-', proper, loc ).
noun( 'hazel', '-', proper, per ).
noun( 'heanor', '-', proper, loc ).
noun( 'heather', '-', proper, per ).
noun( 'heaviside', '-', proper, _ ).
noun( 'hebden royal', '-', proper, loc ).
noun( 'hebrew', 'hebrews', both, _ ).
noun( 'hedge end', '-', proper, loc ).
noun( 'hegira', 'hegiras', count, _ ).
noun( 'heidelburg', '-', proper, loc ).
noun( 'hejira', 'hejiras', count, _ ).
noun( 'helen', '-', proper, per ).
noun( 'helensburgh', '-', proper, loc ).
noun( 'hellene', 'hellenes', count, _ ).
noun( 'helsinki', '-', proper, loc ).
noun( 'helston', '-', proper, loc ).
noun( 'hemel hempstead', '-', proper, loc ).
noun( 'hemsworth', '-', proper, loc ).
noun( 'henley-on-thames', '-', proper, loc ).
noun( 'henry', '-', proper, per ).
noun( 'herb', '-', proper, per ).
noun( 'herbert', '-', proper, per ).
noun( 'hereford', '-', proper, loc ).
noun( 'herne', '-', proper, loc ).
noun( 'herne bay', '-', proper, loc ).
noun( 'herr', 'herren', count, _ ).
noun( 'hertford', '-', proper, loc ).
noun( 'hertfordshire', '-', proper, loc ).
noun( 'hetton', '-', proper, loc ).
noun( 'hexham', '-', proper, loc ).
noun( 'heysham', '-', proper, loc ).
noun( 'heywood', '-', proper, loc ).
noun( 'high spen', '-', proper, loc ).
noun( 'high wycombe', '-', proper, loc ).
noun( 'highland', '-', proper, loc ).
noun( 'highlander', 'highlanders', count, _ ).
noun( 'hilary', '-', proper, per ).
noun( 'hilda', '-', proper, per ).
noun( 'hillingdon', '-', proper, loc ).
noun( 'hinckley', '-', proper, loc ).
noun( 'hindi', '-', mass, _ ).
noun( 'hindu', 'hindus', count, _ ).
noun( 'hinduism', '-', mass, _ ).
noun( 'hindustani', 'hindustanis', count, _ ).
noun( 'hiroshima', '-', proper, loc ).
noun( 'hitchin', '-', proper, loc ).
noun( 'hobson', '-', proper, _ ).
noun( 'hockley', '-', proper, loc ).
noun( 'hoddesdon', '-', proper, loc ).
noun( 'hofei', '-', proper, loc ).
noun( 'hogmanay', 'hogmanays', count, _ ).
noun( 'hogmanay', '-', proper, _ ).
noun( 'holland', '-', proper, loc ).
noun( 'hollander', 'hollanders', count, _ ).
noun( 'hollywood', '-', proper, _ ).
noun( 'holmfirth', '-', proper, loc ).
noun( 'holyhead', '-', proper, loc ).
noun( 'hon', '-', proper, per ).
noun( 'honduran', 'hondurans', count, _ ).
noun( 'honduras', '-', proper, loc ).
noun( 'hong kong', '-', proper, loc ).
noun( 'honiton', '-', proper, loc ).
noun( 'hoo', '-', proper, loc ).
noun( 'hoover', 'hoovers', count, _ ).
noun( 'hope', '-', proper, loc ).
noun( 'horace', '-', proper, per ).
noun( 'horley', '-', proper, loc ).
noun( 'hornsea', '-', proper, loc ).
noun( 'horsham', '-', proper, loc ).
noun( 'horsley', '-', proper, loc ).
noun( 'horwich', '-', proper, loc ).
noun( 'hosp', '-', proper, _ ).
noun( 'host', '-', count, _ ).
noun( 'houghton', '-', proper, loc ).
noun( 'hounslow', '-', proper, loc ).
noun( 'houston', '-', proper, loc ).
noun( 'hove', '-', proper, loc ).
noun( 'howard', '-', proper, per ).
noun( 'hoylake', '-', proper, loc ).
noun( 'hoyland nether', '-', proper, loc ).
noun( 'hubert', '-', proper, per ).
noun( 'hucknall', '-', proper, loc ).
noun( 'huddersfield', '-', proper, loc ).
noun( 'hugh', '-', proper, per ).
noun( 'hughie', '-', proper, per ).
noun( 'huguenot', 'huguenots', count, _ ).
noun( 'huhehot', '-', proper, loc ).
noun( 'humberside', '-', proper, loc ).
noun( 'humberston', '-', proper, loc ).
noun( 'humphrey', '-', proper, per ).
noun( 'hun', 'huns', count, _ ).
noun( 'hungarian', 'hungarians', both, _ ).
noun( 'hungary', '-', proper, loc ).
noun( 'huntingdon', '-', proper, loc ).
noun( 'hurstpierpoint', '-', proper, loc ).
noun( 'hwainan', '-', proper, loc ).
noun( 'hyde', '-', proper, loc ).
noun( 'hyderabad', '-', proper, loc ).
noun( 'hythe', '-', proper, loc ).
noun( 'iba', '-', count, _ ).
noun( 'icbm', '-', count, _ ).
noun( 'ilo', '-', count, _ ).
noun( 'imf', '-', count, _ ).
noun( 'iou', '-', count, _ ).
noun( '-', 'ious', count, _ ).
noun( 'iq', '-', mass, _ ).
noun( 'ira', '-', count, _ ).
noun( 'itv', '-', count, _ ).
noun( 'iud', '-', count, _ ).
noun( 'ian', '-', proper, per ).
noun( 'ibadan', '-', proper, loc ).
noun( 'iceland', '-', proper, loc ).
noun( 'icelander', 'icelanders', count, _ ).
noun( 'icelandic', '-', mass, _ ).
noun( 'ida', '-', proper, per ).
noun( 'idaho', '-', proper, loc ).
noun( 'ilfracombe', '-', proper, loc ).
noun( 'ilkeston', '-', proper, loc ).
noun( 'ilkley', '-', proper, loc ).
noun( 'illinois', '-', proper, loc ).
noun( 'immingham', '-', proper, loc ).
noun( 'inc', '-', proper, _ ).
noun( 'inchon', '-', proper, loc ).
noun( 'india', '-', proper, loc ).
noun( 'india-rubber', 'india-rubbers', both, _ ).
noun( 'indiaman', 'indiamen', count, _ ).
noun( 'indian', 'indians', count, _ ).
noun( 'indiana', '-', proper, loc ).
noun( 'indianapolis', '-', proper, loc ).
noun( 'indies', '-', proper, loc ).
noun( 'indonesia', '-', proper, loc ).
noun( 'indonesian', 'indonesians', both, _ ).
noun( 'indore', '-', proper, loc ).
noun( 'ingatestone', '-', proper, loc ).
noun( 'ingrid', '-', proper, per ).
noun( 'innsbruck', '-', proper, loc ).
noun( 'inst', '-', proper, _ ).
noun( 'interpol', '-', proper, _ ).
noun( 'inverkeithing', '-', proper, loc ).
noun( 'inverness', '-', proper, loc ).
noun( 'inverurie', '-', proper, loc ).
noun( 'iowa', '-', proper, loc ).
noun( 'ipswich', '-', proper, loc ).
noun( 'iran', '-', proper, loc ).
noun( 'iranian', 'iranians', both, _ ).
noun( 'iraq', '-', proper, loc ).
noun( 'iraqi', 'iraqis', count, _ ).
noun( 'ireland', '-', proper, loc ).
noun( 'irene', '-', proper, per ).
noun( 'iris', '-', proper, per ).
noun( 'irish', '-', mass, _ ).
noun( 'irishman', 'irishmen', count, _ ).
noun( 'irishwoman', 'irishwomen', count, _ ).
noun( 'irthlingborough', '-', proper, loc ).
noun( 'irvine', '-', proper, loc ).
noun( 'isaac', '-', proper, per ).
noun( 'isabel', '-', proper, per ).
noun( 'isabella', '-', proper, per ).
noun( 'islam', '-', mass, _ ).
noun( 'islington', '-', proper, loc ).
noun( 'isobel', '-', proper, per ).
noun( 'israel', '-', proper, loc ).
noun( 'israeli', 'israelis', count, _ ).
noun( 'istanbul', '-', proper, loc ).
noun( 'italian', 'italians', both, _ ).
noun( 'italy', '-', proper, loc ).
noun( 'ivan', '-', proper, per ).
noun( 'iver', '-', proper, loc ).
noun( 'ivor', '-', proper, per ).
noun( 'ivy', '-', proper, per ).
noun( 'izmir', '-', proper, loc ).
noun( 'jc', '-', proper, _ ).
noun( 'jp', '-', count, _ ).
noun( 'jabalpur', '-', proper, loc ).
noun( 'jack', '-', proper, per ).
noun( 'jackie', '-', proper, per ).
noun( 'jacksonville', '-', proper, loc ).
noun( 'jacob', '-', proper, per ).
noun( 'jacobin', 'jacobins', count, _ ).
noun( 'jacobinism', '-', mass, _ ).
noun( 'jacobite', 'jacobites', count, _ ).
noun( 'jacqueline', '-', proper, per ).
noun( 'jaipur', '-', proper, loc ).
noun( 'jakarta', '-', proper, loc ).
noun( 'jake', '-', proper, per ).
noun( 'jamaica', '-', proper, loc ).
noun( 'jamaican', 'jamaicans', count, _ ).
noun( 'james', '-', proper, per ).
noun( 'jan', '-', proper, per ).
noun( 'jan', '-', proper, _ ).
noun( 'jane', '-', proper, per ).
noun( 'janet', '-', proper, per ).
noun( 'janice', '-', proper, per ).
noun( 'janie', '-', proper, per ).
noun( 'january', 'januaries', count, _ ).
noun( 'january', '-', proper, _ ).
noun( 'janus', '-', proper, _ ).
noun( 'japan', '-', proper, loc ).
noun( 'japanese', 'japanese', both, _ ).
noun( 'jarrow', '-', proper, loc ).
noun( 'jason', '-', proper, per ).
noun( 'java', '-', proper, loc ).
noun( 'javanese', 'javanese', count, _ ).
noun( 'jean', '-', proper, per ).
noun( 'jeanie', '-', proper, per ).
noun( 'jedburgh', '-', proper, loc ).
noun( 'jeff', '-', proper, per ).
noun( 'jeffrey', '-', proper, per ).
noun( 'jehovah', '-', proper, _ ).
noun( 'jekyll-and-hyde', '-', proper, _ ).
noun( 'jennifer', '-', proper, per ).
noun( 'jenny', '-', proper, per ).
noun( 'jeremy', '-', proper, per ).
noun( 'jerome', '-', proper, per ).
noun( 'jerry', '-', proper, per ).
noun( 'jersey', '-', proper, loc ).
noun( 'jess', '-', proper, per ).
noun( 'jessica', '-', proper, per ).
noun( 'jessie', '-', proper, per ).
noun( 'jesuit', 'jesuits', count, _ ).
noun( 'jesus', '-', proper, per ).
noun( 'jew', 'jews', count, _ ).
noun( 'jewess', 'jewesses', count, _ ).
noun( 'jezebel', 'jezebels', count, _ ).
noun( 'jill', '-', proper, per ).
noun( 'jim', '-', proper, per ).
noun( 'jim crow', '-', count, _ ).
noun( 'jimmy', '-', proper, per ).
noun( 'jnr', '-', proper, _ ).
noun( 'jo', '-', proper, per ).
noun( 'joan', '-', proper, per ).
noun( 'joanie', '-', proper, per ).
noun( 'joann', '-', proper, per ).
noun( 'joanna', '-', proper, per ).
noun( 'joanne', '-', proper, per ).
noun( 'job', '-', proper, per ).
noun( 'jocelyn', '-', proper, per ).
noun( 'joe', '-', proper, per ).
noun( 'joey', '-', proper, per ).
noun( 'johannesburg', '-', proper, loc ).
noun( 'john', '-', proper, per ).
noun( 'john bull', '-', count, _ ).
noun( 'johnny', '-', proper, per ).
noun( 'johnstone', '-', proper, loc ).
noun( 'jonah', '-', proper, per ).
noun( 'jonathan', '-', proper, per ).
noun( 'jordan', '-', proper, loc ).
noun( 'jordanian', 'jordanians', count, _ ).
noun( 'joseph', '-', proper, per ).
noun( 'josephine', '-', proper, per ).
noun( 'josh', '-', proper, per ).
noun( 'joshua', '-', proper, per ).
noun( 'jove', '-', proper, _ ).
noun( 'joy', '-', proper, per ).
noun( 'joyce', '-', proper, per ).
noun( 'judaism', '-', mass, _ ).
noun( 'judas', 'judases', count, _ ).
noun( 'judas', '-', proper, per ).
noun( 'judith', '-', proper, per ).
noun( 'judy', '-', proper, per ).
noun( 'jul', '-', proper, _ ).
noun( 'julia', '-', proper, per ).
noun( 'julian', '-', proper, per ).
noun( 'julie', '-', proper, per ).
noun( 'juliet', '-', proper, per ).
noun( 'july', '-', count, _ ).
noun( 'jun', '-', proper, _ ).
noun( 'june', 'junes', count, _ ).
noun( 'june', '-', proper, per ).
noun( 'jupiter', '-', proper, _ ).
noun( 'justin', '-', proper, per ).
noun( 'kgb', '-', count, _ ).
noun( 'ko', '-', count, _ ).
noun( 'kaiser', 'kaisers', count, _ ).
noun( 'kalgan', '-', proper, loc ).
noun( 'kampuchea', '-', proper, loc ).
noun( 'kampuchean', 'kampucheans', count, _ ).
noun( 'kanpur', '-', proper, loc ).
noun( 'kansas', '-', proper, loc ).
noun( 'kansas city', '-', proper, loc ).
noun( 'kaohsiung', '-', proper, loc ).
noun( 'karachi', '-', proper, loc ).
noun( 'karaganda', '-', proper, loc ).
noun( 'karen', '-', proper, per ).
noun( 'karlsruhe', '-', proper, loc ).
noun( 'kashmir', '-', proper, loc ).
noun( 'kashmiri', 'kashmiris', count, _ ).
noun( 'kassel', '-', proper, loc ).
noun( 'kate', '-', proper, per ).
noun( 'kath', '-', proper, per ).
noun( 'katherine', '-', proper, per ).
noun( 'kathy', '-', proper, per ).
noun( 'katie', '-', proper, per ).
noun( 'katowice', '-', proper, loc ).
noun( 'kay', '-', proper, per ).
noun( 'kazan', '-', proper, loc ).
noun( 'keighley', '-', proper, loc ).
noun( 'keith', '-', proper, per ).
noun( 'kelso', '-', proper, loc ).
noun( 'kelt', 'kelts', count, _ ).
noun( 'kelty', '-', proper, loc ).
noun( 'kemsing', '-', proper, loc ).
noun( 'ken', '-', proper, per ).
noun( 'kendal', '-', proper, loc ).
noun( 'kenilworth', '-', proper, loc ).
noun( 'kenneth', '-', proper, per ).
noun( 'kenny', '-', proper, per ).
noun( 'kensington', '-', proper, loc ).
noun( 'kent', '-', proper, loc ).
noun( 'kentucky', '-', proper, loc ).
noun( 'kenya', '-', proper, loc ).
noun( 'kenyan', 'kenyans', count, _ ).
noun( 'kerry', '-', proper, loc ).
noun( 'kettering', '-', proper, loc ).
noun( 'kevin', '-', proper, per ).
noun( 'keynsham', '-', proper, loc ).
noun( 'keyworth', '-', proper, loc ).
noun( 'khalka', '-', mass, _ ).
noun( 'kharkov', '-', proper, loc ).
noun( 'khartoum', '-', proper, loc ).
noun( 'khmer', '-', mass, _ ).
noun( 'kidderminster', '-', proper, loc ).
noun( 'kidlington', '-', proper, loc ).
noun( 'kidsgrove', '-', proper, loc ).
noun( 'kiev', '-', proper, loc ).
noun( 'kilbirnie', '-', proper, loc ).
noun( 'kildare', '-', proper, loc ).
noun( 'kilkenny', '-', proper, loc ).
noun( 'killarney', '-', proper, loc ).
noun( 'kilmarnock', '-', proper, loc ).
noun( 'kilsyth', '-', proper, loc ).
noun( 'kilwinning', '-', proper, loc ).
noun( 'kimberley', '-', proper, loc ).
noun( 'king\'s lynn', '-', proper, loc ).
noun( 'kingsbury', '-', proper, loc ).
noun( 'kingsteignton', '-', proper, loc ).
noun( 'kingston', '-', proper, loc ).
noun( 'kingston upon hull', '-', proper, loc ).
noun( 'kingston upon thames', '-', proper, loc ).
noun( 'kinshasa', '-', proper, loc ).
noun( 'kippax', '-', proper, loc ).
noun( 'kirin', '-', proper, loc ).
noun( 'kirkby', '-', proper, loc ).
noun( 'kirkby in ashfield', '-', proper, loc ).
noun( 'kirkcaldy', '-', proper, loc ).
noun( 'kirkham', '-', proper, loc ).
noun( 'kirkintilloch', '-', proper, loc ).
noun( 'kirkwall', '-', proper, loc ).
noun( 'kitakyushu', '-', proper, loc ).
noun( 'kitty', '-', proper, per ).
noun( 'knaresborough', '-', proper, loc ).
noun( 'knesset', 'knessets', count, _ ).
noun( 'knottingley', '-', proper, loc ).
noun( 'knutsford', '-', proper, loc ).
noun( 'kobe', '-', proper, loc ).
noun( 'koblenz', '-', proper, loc ).
noun( 'koln', '-', proper, loc ).
noun( 'koran', 'korans', count, _ ).
noun( 'korea', '-', proper, loc ).
noun( 'korean', 'koreans', both, _ ).
noun( 'krakow', '-', proper, loc ).
noun( 'krasnoyarsk', '-', proper, loc ).
noun( 'krefeld', '-', proper, loc ).
noun( 'kremlin', 'kremlins', count, _ ).
noun( 'krivoi rog', '-', proper, loc ).
noun( 'kunming', '-', proper, loc ).
noun( 'kuwait', '-', proper, loc ).
noun( 'kuwait city', '-', proper, loc ).
noun( 'kuwaiti', 'kuwaitis', count, _ ).
noun( 'kuybyshev', '-', proper, loc ).
noun( 'kweiyang', '-', proper, loc ).
noun( 'kyoto', '-', proper, loc ).
noun( 'l-plate', 'l-plates', count, _ ).
noun( 'la', '-', proper, _ ).
noun( 'lea', '-', count, _ ).
noun( 'llb', '-', count, _ ).
noun( 'lp', '-', count, _ ).
noun( '-', 'lps', count, _ ).
noun( 'lsd', '-', mass, _ ).
noun( 'la habana', '-', proper, loc ).
noun( 'la paz', '-', proper, loc ).
noun( 'la plata', '-', proper, loc ).
noun( 'la spezia', '-', proper, loc ).
noun( 'lab', '-', proper, _ ).
noun( 'labourite', 'labourites', count, _ ).
noun( 'labrador', '-', proper, loc ).
noun( 'lady-chapel', 'lady-chapels', count, _ ).
noun( 'lagos', '-', proper, loc ).
noun( 'lahore', '-', proper, loc ).
noun( 'lakenheath', '-', proper, loc ).
noun( 'lambeth', '-', proper, loc ).
noun( 'lammas', 'lammases', count, _ ).
noun( 'lammas', '-', proper, _ ).
noun( 'lanark', '-', proper, loc ).
noun( 'lancashire', '-', proper, loc ).
noun( 'lancaster', '-', proper, loc ).
noun( 'lancastrian', 'lancastrians', count, _ ).
noun( 'lanchow', '-', proper, loc ).
noun( 'lancing', '-', proper, loc ).
noun( 'landrover', 'landrovers', count, _ ).
noun( 'lao', '-', mass, _ ).
noun( 'laos', '-', proper, loc ).
noun( 'laotian', 'laotians', both, _ ).
noun( 'largs', '-', proper, loc ).
noun( 'larkhall', '-', proper, loc ).
noun( 'larne', '-', proper, loc ).
noun( 'larry', '-', proper, per ).
noun( 'las palmas', '-', proper, loc ).
noun( 'latin', 'latins', count, _ ).
noun( 'latinist', 'latinists', count, _ ).
noun( 'laura', '-', proper, per ).
noun( 'laurence', '-', proper, per ).
noun( 'lausanne', '-', proper, loc ).
noun( 'lawrence', '-', proper, per ).
noun( 'lazarus', '-', count, _ ).
noun( 'lazarus', '-', proper, per ).
noun( 'le havre', '-', proper, loc ).
noun( 'le mans', '-', proper, loc ).
noun( 'leamington spa', '-', proper, loc ).
noun( 'leatherhead', '-', proper, loc ).
noun( 'lebanese', 'lebanese', count, _ ).
noun( 'lebanon', '-', proper, loc ).
noun( 'leeds', '-', proper, loc ).
noun( 'leek', '-', proper, loc ).
noun( 'leicester', '-', proper, loc ).
noun( 'leicestershire', '-', proper, loc ).
noun( 'leiden', '-', proper, loc ).
noun( 'leigh', '-', proper, loc ).
noun( 'leighton-linslade', '-', proper, loc ).
noun( 'leinster', '-', proper, loc ).
noun( 'leipzig', '-', proper, loc ).
noun( 'leitrim', '-', proper, loc ).
noun( 'leix', '-', proper, loc ).
noun( 'len', '-', proper, per ).
noun( 'leningrad', '-', proper, loc ).
noun( 'lenny', '-', proper, per ).
noun( 'lens', '-', proper, loc ).
noun( 'lent', 'lents', count, _ ).
noun( 'lent', '-', proper, _ ).
noun( 'leo', '-', proper, per ).
noun( 'leominster', '-', proper, loc ).
noun( 'leon', '-', proper, loc ).
noun( 'leonard', '-', proper, per ).
noun( 'lerwick', '-', proper, loc ).
noun( 'les', '-', proper, per ).
noun( 'lesley', '-', proper, per ).
noun( 'leslie', '-', proper, per ).
noun( 'lesotho', '-', proper, loc ).
noun( 'letchworth', '-', proper, loc ).
noun( 'lethe', '-', proper, _ ).
noun( 'levant', '-', proper, _ ).
noun( 'levantine', 'levantines', count, _ ).
noun( 'leven', '-', proper, loc ).
noun( 'leverkusen', '-', proper, loc ).
noun( 'lewes', '-', proper, loc ).
noun( 'lewis', '-', proper, per ).
noun( 'lewisham', '-', proper, loc ).
noun( 'leyland', '-', proper, loc ).
noun( 'lib', '-', proper, _ ).
noun( 'liberia', '-', proper, loc ).
noun( 'liberian', 'liberians', count, _ ).
noun( 'libra', '-', proper, _ ).
noun( 'libya', '-', proper, loc ).
noun( 'libyan', 'libyans', count, _ ).
noun( 'lichfield', '-', proper, loc ).
noun( 'liechtenstein', '-', proper, loc ).
noun( 'liechtensteiner', 'liechtensteiners', count, _ ).
noun( 'liege', '-', proper, loc ).
noun( 'lilian', '-', proper, per ).
noun( 'lille', '-', proper, loc ).
noun( 'lilliputian', 'lilliputians', count, _ ).
noun( 'lily', '-', proper, per ).
noun( 'lima', '-', proper, loc ).
noun( 'limerick', '-', proper, loc ).
noun( 'limoges', '-', proper, loc ).
noun( 'lincoln', '-', proper, loc ).
noun( 'lincolnshire', '-', proper, loc ).
noun( 'linda', '-', proper, per ).
noun( 'linlithgow', '-', proper, loc ).
noun( 'linwood', '-', proper, loc ).
noun( 'linz', '-', proper, loc ).
noun( 'lionel', '-', proper, per ).
noun( 'lisa', '-', proper, per ).
noun( 'lisbon', '-', proper, loc ).
noun( 'liskeard', '-', proper, loc ).
noun( 'littlehampton', '-', proper, loc ).
noun( 'liverpool', '-', proper, loc ).
noun( 'liverpudlian', 'liverpudlians', count, _ ).
noun( 'livingston', '-', proper, loc ).
noun( 'livorno', '-', proper, loc ).
noun( 'liz', '-', proper, per ).
noun( 'liza', '-', proper, per ).
noun( 'lizzy', '-', proper, per ).
noun( 'ljubljana', '-', proper, loc ).
noun( 'llandudno', '-', proper, loc ).
noun( 'llanelli', '-', proper, loc ).
noun( 'llangollen', '-', proper, loc ).
noun( 'llantrisant', '-', proper, loc ).
noun( 'llantwit major', '-', proper, loc ).
noun( 'lloyd\'s', '-', proper, _ ).
noun( 'loanhead', '-', proper, loc ).
noun( 'lochgelly', '-', proper, loc ).
noun( 'lodz', '-', proper, loc ).
noun( 'loftus', '-', proper, loc ).
noun( 'lois', '-', proper, per ).
noun( 'london', '-', proper, loc ).
noun( 'londonderry', '-', proper, loc ).
noun( 'londoner', 'londoners', count, _ ).
noun( 'long eaton', '-', proper, loc ).
noun( 'longridge', '-', proper, loc ).
noun( 'longton', '-', proper, loc ).
noun( 'lorna', '-', proper, per ).
noun( 'los angeles', '-', proper, loc ).
noun( 'lossiemouth', '-', proper, loc ).
noun( 'lothian', '-', proper, loc ).
noun( 'loughborough', '-', proper, loc ).
noun( 'louis', '-', proper, per ).
noun( 'louise', '-', proper, per ).
noun( 'louisiana', '-', proper, loc ).
noun( 'louisville', '-', proper, loc ).
noun( 'louth', '-', proper, loc ).
noun( 'lowestoft', '-', proper, loc ).
noun( 'loyang', '-', proper, loc ).
noun( 'lt', '-', proper, per ).
noun( 'ltd', '-', proper, _ ).
noun( 'lubeck', '-', proper, loc ).
noun( 'lucifer', 'lucifers', count, _ ).
noun( 'lucknow', '-', proper, loc ).
noun( 'lucy', '-', proper, per ).
noun( 'luddite', 'luddites', count, _ ).
noun( 'ludlow', '-', proper, loc ).
noun( 'ludwigshafen', '-', proper, loc ).
noun( 'luke', '-', proper, per ).
noun( 'luta', '-', proper, loc ).
noun( 'lutheran', 'lutherans', count, _ ).
noun( 'luton', '-', proper, loc ).
noun( 'luxembourg', '-', proper, loc ).
noun( 'luxemburg', '-', proper, loc ).
noun( 'luxemburger', 'luxemburgers', count, _ ).
noun( 'lvov', '-', proper, loc ).
noun( 'lyallpur', '-', proper, loc ).
noun( 'lydia', '-', proper, per ).
noun( 'lydney', '-', proper, loc ).
noun( 'lymington', '-', proper, loc ).
noun( 'lymm', '-', proper, loc ).
noun( 'lynn', '-', proper, per ).
noun( 'lynne', '-', proper, per ).
noun( 'lyons', '-', proper, loc ).
noun( 'lytham', '-', proper, loc ).
noun( 'lytham st annes', '-', proper, loc ).
noun( 'ma', '-', count, _ ).
noun( 'mb', '-', count, _ ).
noun( 'mc', '-', count, _ ).
noun( 'mcc', '-', count, _ ).
noun( 'md', '-', count, _ ).
noun( 'mi5', '-', proper, _ ).
noun( 'mi6', '-', proper, _ ).
noun( 'mp', '-', count, _ ).
noun( '-', 'mp\'s', count, _ ).
noun( 'ms', '-', count, _ ).
noun( 'msc', '-', count, _ ).
noun( 'maastricht', '-', proper, loc ).
noun( 'mabel', '-', proper, per ).
noun( 'macclesfield', '-', proper, loc ).
noun( 'macedonian', 'macedonians', both, _ ).
noun( 'mach', '-', proper, _ ).
noun( 'madagascan', 'madagascans', count, _ ).
noun( 'madagascar', '-', proper, loc ).
noun( 'madame', 'madames', count, _ ).
noun( 'maddy', '-', proper, per ).
noun( 'madeira', 'madeiras', count, _ ).
noun( 'madeleine', '-', proper, per ).
noun( 'mademoiselle', 'mademoiselles', count, _ ).
noun( 'madge', '-', proper, per ).
noun( 'madonna', 'madonnas', count, _ ).
noun( 'madras', '-', proper, loc ).
noun( 'madrid', '-', proper, loc ).
noun( 'madurai', '-', proper, loc ).
noun( 'maecenas', '-', proper, _ ).
noun( 'maesteg', '-', proper, loc ).
noun( 'mafia', 'mafias', count, _ ).
noun( 'magdeburg', '-', proper, loc ).
noun( 'maggie', '-', proper, per ).
noun( 'maghull', '-', proper, loc ).
noun( 'magnificat', 'magnificats', count, _ ).
noun( 'magyar', 'magyars', both, _ ).
noun( 'maharaja', 'maharajas', count, _ ).
noun( 'maharajah', 'maharajahs', count, _ ).
noun( 'maharanee', 'maharanees', count, _ ).
noun( 'mahatma', 'mahatmas', count, _ ).
noun( 'maidenhead', '-', proper, loc ).
noun( 'maidstone', '-', proper, loc ).
noun( 'maine', '-', proper, loc ).
noun( 'mainz', '-', proper, loc ).
noun( 'maj', '-', proper, per ).
noun( 'malaga', '-', proper, loc ).
noun( 'malawi', '-', proper, loc ).
noun( 'malawian', 'malawians', count, _ ).
noun( 'malay', 'malays', both, _ ).
noun( 'malaya', '-', proper, loc ).
noun( 'malayan', 'malayans', count, _ ).
noun( 'malaysia', '-', proper, loc ).
noun( 'malaysian', 'malaysians', count, _ ).
noun( 'malcolm', '-', proper, per ).
noun( 'maldon', '-', proper, loc ).
noun( 'mali', '-', proper, loc ).
noun( 'malian', 'malians', count, _ ).
noun( 'mallow', '-', proper, loc ).
noun( 'malmo', '-', proper, loc ).
noun( 'malta', '-', proper, loc ).
noun( 'maltby', '-', proper, loc ).
noun( 'maltese', 'maltese', both, _ ).
noun( 'malvern', '-', proper, loc ).
noun( 'mamie', '-', proper, per ).
noun( 'manchester', '-', proper, loc ).
noun( 'mancunian', 'mancunians', count, _ ).
noun( 'mandy', '-', proper, per ).
noun( 'manila', '-', proper, loc ).
noun( 'manilla', 'manillas', count, _ ).
noun( 'manitoba', '-', proper, loc ).
noun( 'mannheim', '-', proper, loc ).
noun( 'mansfield', '-', proper, loc ).
noun( 'manx', '-', mass, _ ).
noun( 'maoism', '-', mass, _ ).
noun( 'maoist', 'maoists', count, _ ).
noun( 'maori', 'maoris', count, _ ).
noun( 'mar', '-', proper, _ ).
noun( 'maracaibo', '-', proper, loc ).
noun( 'march', 'marches', count, _ ).
noun( 'march', '-', proper, loc ).
noun( 'mardi gras', 'mardi gras', count, _ ).
noun( 'margaret', '-', proper, per ).
noun( 'margate', '-', proper, loc ).
noun( 'marge', '-', proper, per ).
noun( 'margery', '-', proper, per ).
noun( 'margie', '-', proper, per ).
noun( 'margrave', 'margraves', count, _ ).
noun( 'maria', '-', proper, per ).
noun( 'marian', '-', proper, per ).
noun( 'marie', '-', proper, per ).
noun( 'marilyn', '-', proper, per ).
noun( 'marion', '-', proper, per ).
noun( 'marjorie', '-', proper, per ).
noun( 'mark', '-', proper, per ).
noun( 'market drayton', '-', proper, loc ).
noun( 'market harborough', '-', proper, loc ).
noun( 'marlborough', '-', proper, loc ).
noun( 'marlene', '-', proper, per ).
noun( 'marlow', '-', proper, loc ).
noun( 'marple', '-', proper, loc ).
noun( 'mars', '-', proper, _ ).
noun( 'marsala', '-', mass, _ ).
noun( 'marseillaise', 'marseillaises', count, _ ).
noun( 'marseilles', '-', proper, loc ).
noun( 'marske', '-', proper, loc ).
noun( 'martha', '-', proper, per ).
noun( 'martian', 'martians', count, _ ).
noun( 'martin', '-', proper, per ).
noun( 'marxism', '-', mass, _ ).
noun( 'marxist', 'marxists', count, _ ).
noun( 'mary', '-', proper, per ).
noun( 'maryland', '-', proper, loc ).
noun( 'maryport', '-', proper, loc ).
noun( 'mason-dixon', '-', proper, _ ).
noun( 'mass', 'masses', both, _ ).
noun( 'massachusetts', '-', proper, loc ).
noun( 'matlock', '-', proper, loc ).
noun( 'matt', '-', proper, per ).
noun( 'matthew', '-', proper, per ).
noun( 'maud', '-', proper, per ).
noun( 'maudie', '-', proper, per ).
noun( 'maundy thursday', 'maundy thursdays', count, _ ).
noun( 'maundy thursday', '-', proper, _ ).
noun( 'maureen', '-', proper, per ).
noun( 'maurice', '-', proper, per ).
noun( 'mauritania', '-', proper, loc ).
noun( 'mauritanian', 'mauritanians', count, _ ).
noun( 'mauritian', 'mauritians', count, _ ).
noun( 'mauritius', '-', proper, loc ).
noun( 'mavis', '-', proper, per ).
noun( 'max', '-', proper, per ).
noun( 'maxine', '-', proper, per ).
noun( 'may', 'mays', count, _ ).
noun( 'may', '-', proper, per ).
noun( 'mayfair', '-', proper, _ ).
noun( 'mayfield', '-', proper, loc ).
noun( 'mayo', '-', proper, loc ).
noun( 'mccarthyism', '-', mass, _ ).
noun( 'meath', '-', proper, loc ).
noun( 'mecca', '-', proper, loc ).
noun( 'med', '-', count, _ ).
noun( 'medan', '-', proper, loc ).
noun( 'medellin', '-', proper, loc ).
noun( 'medicare', '-', mass, _ ).
noun( 'meg', '-', proper, per ).
noun( 'melbourne', '-', proper, loc ).
noun( 'melcombe', '-', proper, loc ).
noun( 'melcombe regis', '-', proper, loc ).
noun( 'melksham', '-', proper, loc ).
noun( 'meltham', '-', proper, loc ).
noun( 'melton mowbray', '-', proper, loc ).
noun( 'memphis', '-', proper, loc ).
noun( 'meopham', '-', proper, loc ).
noun( 'mercator', '-', proper, _ ).
noun( 'mercury', '-', proper, _ ).
noun( 'merthyr tydfil', '-', proper, loc ).
noun( 'merton', '-', proper, loc ).
noun( 'messiah', 'messiahs', count, _ ).
noun( 'messina', '-', proper, loc ).
noun( 'met', '-', proper, _ ).
noun( 'methodism', '-', mass, _ ).
noun( 'methodist', 'methodists', count, _ ).
noun( 'methuselah', '-', proper, _ ).
noun( 'metro', 'metros', count, _ ).
noun( 'metz', '-', proper, loc ).
noun( 'mexican', 'mexicans', count, _ ).
noun( 'mexico', '-', proper, loc ).
noun( 'mexico city', '-', proper, loc ).
noun( 'mgr', '-', proper, per ).
noun( 'miami', '-', proper, loc ).
noun( 'michael', '-', proper, per ).
noun( 'michaelmas', 'michaelmases', count, _ ).
noun( 'michaelmas', '-', proper, _ ).
noun( 'michelle', '-', proper, per ).
noun( 'michigan', '-', proper, loc ).
noun( 'mick', '-', proper, per ).
noun( 'mickey', '-', proper, per ).
noun( 'middlesex', '-', proper, loc ).
noun( 'middleton', '-', proper, loc ).
noun( 'middlewich', '-', proper, loc ).
noun( 'midlands', '-', proper, loc ).
noun( 'midwest', '-', proper, _ ).
noun( 'mike', '-', proper, per ).
noun( 'milan', '-', proper, loc ).
noun( 'milano', '-', proper, loc ).
noun( 'mildenhall', '-', proper, loc ).
noun( 'mildred', '-', proper, per ).
noun( 'miles', '-', proper, per ).
noun( 'milford', '-', proper, loc ).
noun( 'milford haven', '-', proper, loc ).
noun( 'millicent', '-', proper, per ).
noun( 'millie', '-', proper, per ).
noun( 'millom', '-', proper, loc ).
noun( 'milly', '-', proper, per ).
noun( 'milngavie', '-', proper, loc ).
noun( 'milnrow', '-', proper, loc ).
noun( 'milton keynes', '-', proper, loc ).
noun( 'milwaukee', '-', proper, loc ).
noun( 'minehead', '-', proper, loc ).
noun( 'minneapolis', '-', proper, loc ).
noun( 'minnesota', '-', proper, loc ).
noun( 'minotaur', 'minotaurs', count, _ ).
noun( 'minsk', '-', proper, loc ).
noun( 'minster-in-sheppey', '-', proper, loc ).
noun( 'miranda', '-', proper, per ).
noun( 'miriam', '-', proper, per ).
noun( 'miss', 'misses', count, _ ).
noun( 'mississippi', '-', proper, loc ).
noun( 'missouri', '-', proper, loc ).
noun( 'mlle', '-', proper, per ).
noun( 'mme', '-', proper, per ).
noun( 'mo', '-', proper, per ).
noun( 'mod', 'mods', count, _ ).
noun( 'modena', '-', proper, loc ).
noun( 'mohammedan', 'mohammedans', count, _ ).
noun( 'moira', '-', proper, per ).
noun( 'mold', '-', proper, loc ).
noun( 'molly', '-', proper, per ).
noun( 'moloch', '-', proper, _ ).
noun( 'mon', '-', proper, _ ).
noun( 'monaco', '-', proper, loc ).
noun( 'monaghan', '-', proper, loc ).
noun( 'monchengladbach', '-', proper, loc ).
noun( 'monday', 'mondays', count, _ ).
noun( 'monday', '-', proper, _ ).
noun( 'monegasque', 'monegasques', count, _ ).
noun( 'mongol', 'mongols', both, _ ).
noun( 'mongolia', '-', proper, loc ).
noun( 'mongolian', 'mongolians', both, _ ).
noun( 'monica', '-', proper, per ).
noun( 'monifieth', '-', proper, loc ).
noun( 'monmouth', '-', proper, loc ).
noun( 'monsieur', 'messieurs', count, _ ).
noun( 'monsignor', 'monsignors', count, _ ).
noun( 'montana', '-', proper, loc ).
noun( 'monterrey', '-', proper, loc ).
noun( 'montevideo', '-', proper, loc ).
noun( 'montreal', '-', proper, loc ).
noun( 'montrose', '-', proper, loc ).
noun( 'montserrat', '-', proper, loc ).
noun( 'montserratian', 'montserratians', count, _ ).
noun( 'monza', '-', proper, loc ).
noun( 'moor', 'moors', count, _ ).
noun( 'morecambe', '-', proper, loc ).
noun( 'morley', '-', proper, loc ).
noun( 'mormon', 'mormons', count, _ ).
noun( 'mormonism', '-', mass, _ ).
noun( 'moroccan', 'moroccans', count, _ ).
noun( 'morocco', '-', proper, loc ).
noun( 'morpeth', '-', proper, loc ).
noun( 'morpheus', '-', proper, _ ).
noun( 'morse', '-', mass, _ ).
noun( 'moscow', '-', proper, loc ).
noun( 'moslem', 'moslems', count, _ ).
noun( 'mossley', '-', proper, loc ).
noun( 'motherwell', '-', proper, loc ).
noun( 'mountain ash', '-', proper, loc ).
noun( 'mountie', 'mounties', count, _ ).
noun( 'mozambican', 'mozambicans', count, _ ).
noun( 'mozambique', '-', proper, loc ).
noun( 'mr', '-', proper, per ).
noun( 'mrs', '-', proper, per ).
noun( 'ms', '-', proper, per ).
noun( 'mt', '-', proper, _ ).
noun( 'muhammad', '-', proper, _ ).
noun( 'muhammadan', 'muhammadans', count, _ ).
noun( 'muhammadanism', '-', mass, _ ).
noun( 'mukden', '-', proper, loc ).
noun( 'mulheim', '-', proper, loc ).
noun( 'mulhouse', '-', proper, loc ).
noun( 'multan', '-', proper, loc ).
noun( 'munich', '-', proper, loc ).
noun( 'munster', '-', proper, loc ).
noun( 'muriel', '-', proper, per ).
noun( 'murton', '-', proper, loc ).
noun( 'muscovite', 'muscovites', count, _ ).
noun( 'muscovy', 'muscovies', count, _ ).
noun( 'muslim', 'muslims', count, _ ).
noun( 'musselburgh', '-', proper, loc ).
noun( 'myra', '-', proper, per ).
noun( 'naafi', '-', count, _ ).
noun( 'nasa', '-', proper, _ ).
noun( 'nato', '-', proper, _ ).
noun( 'nb', '-', proper, _ ).
noun( 'nco', '-', count, _ ).
noun( 'nhs', '-', count, _ ).
noun( 'nspcc', '-', count, _ ).
noun( 'nt', '-', count, _ ).
noun( 'nagoya', '-', proper, loc ).
noun( 'nagpur', '-', proper, loc ).
noun( 'nailsea', '-', proper, loc ).
noun( 'nairn', '-', proper, loc ).
noun( 'nairobi', '-', proper, loc ).
noun( 'namibia', '-', proper, loc ).
noun( 'namibian', 'namibians', count, _ ).
noun( 'nanchang', '-', proper, loc ).
noun( 'nancy', '-', proper, per ).
noun( 'nanking', '-', proper, loc ).
noun( 'nanning', '-', proper, loc ).
noun( 'nantwich', '-', proper, loc ).
noun( 'naomi', '-', proper, per ).
noun( 'naples', '-', proper, loc ).
noun( 'napoli', '-', proper, loc ).
noun( 'narborough', '-', proper, loc ).
noun( 'nat', '-', proper, per ).
noun( 'natalie', '-', proper, per ).
noun( 'nathaniel', '-', proper, per ).
noun( 'nauru', '-', proper, loc ).
noun( 'nauruan', 'nauruans', both, _ ).
noun( 'nazi', 'nazis', count, _ ).
noun( 'nazism', '-', mass, _ ).
noun( 'neapolitan', 'neapolitans', count, _ ).
noun( 'neath', '-', proper, loc ).
noun( 'nebraska', '-', proper, loc ).
noun( 'ned', '-', proper, per ).
noun( 'neddy', '-', proper, per ).
noun( 'negress', 'negresses', count, _ ).
noun( 'negro', 'negroes', count, _ ).
noun( 'negroid', 'negroids', count, _ ).
noun( 'negus', 'neguses', count, _ ).
noun( 'neil', '-', proper, per ).
noun( 'nell', '-', proper, per ).
noun( 'nelly', '-', count, _ ).
noun( 'nelly', '-', proper, per ).
noun( 'nelson', '-', proper, loc ).
noun( 'nepal', '-', proper, loc ).
noun( 'nepalese', 'nepalese', count, _ ).
noun( 'nepali', 'nepalis', both, _ ).
noun( 'neptune', '-', proper, _ ).
noun( 'neston', '-', proper, loc ).
noun( 'nestor', '-', proper, _ ).
noun( 'netherlander', 'netherlanders', count, _ ).
noun( 'netherlands', '-', proper, loc ).
noun( 'netley', '-', proper, loc ).
noun( 'neuss', '-', proper, loc ).
noun( 'nevada', '-', proper, loc ).
noun( 'neville', '-', proper, per ).
noun( 'new cumnock', '-', proper, loc ).
noun( 'new mills', '-', proper, loc ).
noun( 'new orleans', '-', proper, loc ).
noun( 'new tredegar', '-', proper, loc ).
noun( 'new windsor', '-', proper, loc ).
noun( 'new york', '-', proper, loc ).
noun( 'newark', '-', proper, loc ).
noun( 'newarthill', '-', proper, loc ).
noun( 'newbiggin', '-', proper, loc ).
noun( 'newbury', '-', proper, loc ).
noun( 'newcastle', '-', proper, loc ).
noun( 'newcastle upon tyne', '-', proper, loc ).
noun( 'newcastle-under-lyme', '-', proper, loc ).
noun( 'newfoundland', '-', proper, loc ).
noun( 'newham', '-', proper, loc ).
noun( 'newhaven', '-', proper, loc ).
noun( 'newmains', '-', proper, loc ).
noun( 'newmarket', '-', mass, _ ).
noun( 'newmarket', '-', proper, loc ).
noun( 'newport', '-', proper, loc ).
noun( 'newport pagnell', '-', proper, loc ).
noun( 'newquay', '-', proper, loc ).
noun( 'newry', '-', proper, loc ).
noun( 'newton abbot', '-', proper, loc ).
noun( 'newton aycliffe', '-', proper, loc ).
noun( 'newton mearns', '-', proper, loc ).
noun( 'newton-le-willows', '-', proper, loc ).
noun( 'newtonian', 'newtonians', count, _ ).
noun( 'newtown', '-', proper, loc ).
noun( 'nicaragua', '-', proper, loc ).
noun( 'nicaraguan', 'nicaraguans', count, _ ).
noun( 'nice', '-', proper, loc ).
noun( 'nicholas', '-', proper, per ).
noun( 'nick', '-', proper, per ).
noun( 'nicola', '-', proper, per ).
noun( 'nicole', '-', proper, per ).
noun( 'nigel', '-', proper, per ).
noun( 'niger', '-', proper, loc ).
noun( 'nigeria', '-', proper, loc ).
noun( 'nigerian', 'nigerians', count, _ ).
noun( 'nigerien', 'nigeriens', count, _ ).
noun( 'nijmegen', '-', proper, loc ).
noun( 'nimrod', '-', proper, _ ).
noun( 'niobe', '-', proper, _ ).
noun( 'nissen', '-', proper, _ ).
noun( 'noah', '-', proper, _ ).
noun( 'nobel', '-', proper, _ ).
noun( 'noel', 'noels', count, _ ).
noun( 'noel', '-', proper, per ).
noun( 'nora', '-', proper, per ).
noun( 'nordic', 'nordics', count, _ ).
noun( 'norfolk', '-', proper, loc ).
noun( 'norman', 'normans', count, _ ).
noun( 'norman', '-', proper, per ).
noun( 'normanton', '-', proper, loc ).
noun( 'normantown', '-', proper, loc ).
noun( 'norse', '-', mass, _ ).
noun( 'north walsham', '-', proper, loc ).
noun( 'northallerton', '-', proper, loc ).
noun( 'northam', '-', proper, loc ).
noun( 'northampton', '-', proper, loc ).
noun( 'northamptonshire', '-', proper, loc ).
noun( 'northman', 'northmen', count, _ ).
noun( 'northumberland', '-', proper, loc ).
noun( 'northwich', '-', proper, loc ).
noun( 'norton', '-', proper, loc ).
noun( 'norway', '-', proper, loc ).
noun( 'norwegian', 'norwegians', both, _ ).
noun( 'norwich', '-', proper, loc ).
noun( 'nottingham', '-', proper, loc ).
noun( 'nottinghamshire', '-', proper, loc ).
noun( 'nov', '-', proper, _ ).
noun( 'nova scotia', '-', proper, loc ).
noun( 'novara', '-', proper, loc ).
noun( 'november', 'novembers', count, _ ).
noun( 'november', '-', proper, _ ).
noun( 'novokuznetsk', '-', proper, loc ).
noun( 'novosibirsk', '-', proper, loc ).
noun( 'nuneaton', '-', proper, loc ).
noun( 'nurenburg', '-', proper, loc ).
noun( 'o-level', 'o-levels', count, _ ).
noun( 'oap', '-', count, _ ).
noun( '-', 'oap\'s', count, _ ).
noun( 'oau', '-', count, _ ).
noun( 'oecd', '-', count, _ ).
noun( 'oed', '-', count, _ ).
noun( 'opec', '-', proper, _ ).
noun( 'ot', '-', count, _ ).
noun( 'oakengates', '-', proper, loc ).
noun( 'oban', '-', proper, loc ).
noun( 'oberhausen', '-', proper, loc ).
noun( 'occident', '-', proper, _ ).
noun( 'occidental', 'occidentals', count, _ ).
noun( 'ockbrook', '-', proper, loc ).
noun( 'ockendon', '-', proper, loc ).
noun( 'oct', '-', proper, _ ).
noun( 'october', 'octobers', count, _ ).
noun( 'october', '-', proper, _ ).
noun( 'odessa', '-', proper, loc ).
noun( 'oedipus', '-', proper, _ ).
noun( 'offaly', '-', proper, loc ).
noun( 'offenbach', '-', proper, loc ).
noun( 'ogmore valley', '-', proper, loc ).
noun( 'ohio', '-', proper, loc ).
noun( 'oklahoma', '-', proper, loc ).
noun( 'oklahoma city', '-', proper, loc ).
noun( 'old windsor', '-', proper, loc ).
noun( 'oldenburg', '-', proper, loc ).
noun( 'oldham', '-', proper, loc ).
noun( 'oldland', '-', proper, loc ).
noun( 'olive', '-', proper, per ).
noun( 'oliver', '-', proper, per ).
noun( 'olivia', '-', proper, per ).
noun( 'ollerton', '-', proper, loc ).
noun( 'ollie', '-', proper, per ).
noun( 'olympiad', 'olympiads', count, _ ).
noun( 'olympian', 'olympians', count, _ ).
noun( 'olympics', '-', proper, _ ).
noun( 'oman', '-', proper, loc ).
noun( 'omani', 'omanis', count, _ ).
noun( 'omsk', '-', proper, loc ).
noun( 'ongar', '-', proper, loc ).
noun( 'ontario', '-', proper, loc ).
noun( 'orangeman', 'orangemen', count, _ ).
noun( 'oregon', '-', proper, loc ).
noun( 'orkney', '-', proper, loc ).
noun( 'orleans', '-', proper, loc ).
noun( 'ormskirk', '-', proper, loc ).
noun( 'osaka', '-', proper, loc ).
noun( 'oscar', 'oscars', count, _ ).
noun( 'oscar', '-', proper, per ).
noun( 'oslo', '-', proper, loc ).
noun( 'osnabruck', '-', proper, loc ).
noun( 'ossett', '-', proper, loc ).
noun( 'oswald', '-', proper, per ).
noun( 'oswestry', '-', proper, loc ).
noun( 'otago', '-', proper, loc ).
noun( 'otley', '-', proper, loc ).
noun( 'ottawa', '-', proper, loc ).
noun( 'oviedo', '-', proper, loc ).
noun( 'oxbridge', '-', proper, _ ).
noun( 'oxford', '-', proper, loc ).
noun( 'oxfordshire', '-', proper, loc ).
noun( 'oxon', '-', proper, _ ).
noun( 'oxonian', 'oxonians', count, _ ).
noun( 'oxted', '-', proper, loc ).
noun( 'ozzie', '-', proper, per ).
noun( 'pa', '-', count, _ ).
noun( 'paye', '-', proper, _ ).
noun( 'pc', '-', proper, per ).
noun( 'pdsa', '-', count, _ ).
noun( 'pe', '-', mass, _ ).
noun( 'plc', '-', proper, _ ).
noun( 'pm', '-', count, _ ).
noun( 'po', '-', count, _ ).
noun( 'pow', '-', count, _ ).
noun( 'pr', '-', mass, _ ).
noun( 'pt', '-', mass, _ ).
noun( 'pta', '-', count, _ ).
noun( 'pto', '-', proper, _ ).
noun( 'pacific', '-', proper, _ ).
noun( 'paddy', 'paddies', count, _ ).
noun( 'paddy', '-', proper, per ).
noun( 'paddy-wagon', 'paddy-wagons', count, _ ).
noun( 'padova', '-', proper, loc ).
noun( 'padua', '-', proper, loc ).
noun( 'paisley', '-', proper, loc ).
noun( 'pakistan', '-', proper, loc ).
noun( 'pakistani', 'pakistanis', count, _ ).
noun( 'palembang', '-', proper, loc ).
noun( 'palermo', '-', proper, loc ).
noun( 'palestine', '-', proper, loc ).
noun( 'palestinian', 'palestinians', count, _ ).
noun( 'pam', '-', proper, per ).
noun( 'pamela', '-', proper, per ).
noun( 'pamplona', '-', proper, loc ).
noun( 'panama', '-', proper, loc ).
noun( 'panamanian', 'panamanians', count, _ ).
noun( 'paotow', '-', proper, loc ).
noun( 'papua', '-', proper, loc ).
noun( 'papuan', 'papuans', count, _ ).
noun( 'paraguay', '-', proper, loc ).
noun( 'paraguayan', 'paraguayans', count, _ ).
noun( 'paris', '-', proper, loc ).
noun( 'parisian', 'parisians', count, _ ).
noun( 'parma', '-', proper, loc ).
noun( 'parmesan', '-', mass, _ ).
noun( 'parsee', 'parsees', count, _ ).
noun( 'pashto', '-', mass, _ ).
noun( 'passover', 'passovers', count, _ ).
noun( 'pat', '-', proper, per ).
noun( 'patience', '-', proper, per ).
noun( 'patrai', '-', proper, loc ).
noun( 'patricia', '-', proper, per ).
noun( 'patrick', '-', proper, per ).
noun( 'patty', '-', proper, per ).
noun( 'paul', '-', proper, per ).
noun( 'paula', '-', proper, per ).
noun( 'pauline', '-', proper, per ).
noun( 'pax romana', '-', count, _ ).
noun( 'peacehaven', '-', proper, loc ).
noun( 'pearl', '-', proper, per ).
noun( 'peebles', '-', proper, loc ).
noun( 'peg', '-', proper, per ).
noun( 'peggy', '-', proper, per ).
noun( 'peking', '-', proper, loc ).
noun( 'pelton', '-', proper, loc ).
noun( 'pembroke', '-', proper, loc ).
noun( 'penarth', '-', proper, loc ).
noun( 'pendlebury', '-', proper, loc ).
noun( 'penelope', '-', proper, per ).
noun( 'penicuik', '-', proper, loc ).
noun( 'penki', '-', proper, loc ).
noun( 'pennsylvania', '-', proper, loc ).
noun( 'penny', '-', proper, per ).
noun( 'penrhyn', '-', proper, loc ).
noun( 'penrith', '-', proper, loc ).
noun( 'penryn', '-', proper, loc ).
noun( 'pentateuch', 'pentateuchs', count, _ ).
noun( 'pentecost', '-', mass, _ ).
noun( 'penzance', '-', proper, loc ).
noun( 'percy', '-', proper, per ).
noun( 'perm', '-', proper, loc ).
noun( 'pershore', '-', proper, loc ).
noun( 'persia', '-', proper, loc ).
noun( 'persian', 'persians', count, _ ).
noun( 'perth', '-', proper, loc ).
noun( 'peru', '-', proper, loc ).
noun( 'perugia', '-', proper, loc ).
noun( 'peruvian', 'peruvians', count, _ ).
noun( 'pescara', '-', proper, loc ).
noun( 'pete', '-', proper, per ).
noun( 'peter', '-', proper, per ).
noun( 'peterborough', '-', proper, loc ).
noun( 'peterhead', '-', proper, loc ).
noun( 'peterlee', '-', proper, loc ).
noun( 'petersfield', '-', proper, loc ).
noun( 'phd', '-', count, _ ).
noun( 'pharaoh', 'pharaohs', count, _ ).
noun( 'pharisee', 'pharisees', count, _ ).
noun( 'phil', '-', proper, per ).
noun( 'philadelphia', '-', proper, loc ).
noun( 'philip', '-', proper, per ).
noun( 'philippa', '-', proper, per ).
noun( 'philippine', 'philippines', count, _ ).
noun( 'philippines', '-', proper, loc ).
noun( 'philistine', 'philistines', count, _ ).
noun( 'philistinism', '-', mass, _ ).
noun( 'phnom-penh', '-', proper, loc ).
noun( 'phoebe', '-', proper, per ).
noun( 'phoenix', '-', proper, loc ).
noun( 'phyllis', '-', proper, per ).
noun( 'piacenza', '-', proper, loc ).
noun( 'pill', '-', proper, loc ).
noun( 'piraeus', '-', proper, loc ).
noun( 'pisa', '-', proper, loc ).
noun( 'pisces', '-', proper, _ ).
noun( 'pittsburgh', '-', proper, loc ).
noun( 'plimsoll', 'plimsolls', count, _ ).
noun( 'pluto', '-', proper, _ ).
noun( 'plymouth', '-', proper, loc ).
noun( 'poland', '-', proper, loc ).
noun( 'polaris', '-', count, _ ).
noun( 'polaroid', '-', mass, _ ).
noun( 'pole', 'poles', count, _ ).
noun( 'polish', '-', mass, _ ).
noun( 'polly', '-', proper, per ).
noun( 'pont-llan-fraith', '-', proper, loc ).
noun( 'pontardawe', '-', proper, loc ).
noun( 'pontardulais', '-', proper, loc ).
noun( 'pontefract', '-', proper, loc ).
noun( 'pontycymmer', '-', proper, loc ).
noun( 'pontypool', '-', proper, loc ).
noun( 'poole', '-', proper, loc ).
noun( 'poona', '-', proper, loc ).
noun( 'pope', 'popes', count, _ ).
noun( 'port glasgow', '-', proper, loc ).
noun( 'port talbot', '-', proper, loc ).
noun( 'porthcawl', '-', proper, loc ).
noun( 'porthmadog', '-', proper, loc ).
noun( 'portishead', '-', proper, loc ).
noun( 'portland', '-', proper, loc ).
noun( 'porto alegre', '-', proper, loc ).
noun( 'portsmouth', '-', proper, loc ).
noun( 'portugal', '-', proper, loc ).
noun( 'portuguese', 'portuguese', both, _ ).
noun( 'potsdam', '-', proper, loc ).
noun( 'potters bar', '-', proper, loc ).
noun( 'powys', '-', proper, loc ).
noun( 'poynton', '-', proper, loc ).
noun( 'poznan', '-', proper, loc ).
noun( 'prague', '-', proper, loc ).
noun( 'prato', '-', proper, loc ).
noun( 'pre-raphaelite', 'pre-raphaelites', count, _ ).
noun( 'pres', '-', count, _ ).
noun( 'presbyterian', 'presbyterians', count, _ ).
noun( 'presbyterianism', '-', mass, _ ).
noun( 'prescot', '-', proper, loc ).
noun( 'prestatyn', '-', proper, loc ).
noun( 'preston', '-', proper, loc ).
noun( 'prestonpans', '-', proper, loc ).
noun( 'prestwich', '-', proper, loc ).
noun( 'prestwick', '-', proper, loc ).
noun( 'pretoria', '-', proper, loc ).
noun( 'princes risborough', '-', proper, loc ).
noun( 'priscilla', '-', proper, per ).
noun( 'prof', '-', proper, per ).
noun( 'protestant', 'protestants', count, _ ).
noun( 'protestantism', '-', mass, _ ).
noun( 'providence', '-', proper, loc ).
noun( 'pru', '-', proper, per ).
noun( 'prudence', '-', proper, per ).
noun( 'prudhoe', '-', proper, loc ).
noun( 'prussian', 'prussians', count, _ ).
noun( 'pte', '-', proper, per ).
noun( 'pty', '-', proper, _ ).
noun( 'pudsey', '-', proper, loc ).
noun( 'pullman', 'pullmans', count, _ ).
noun( 'punch', 'punches', count, _ ).
noun( 'pusan', '-', proper, loc ).
noun( 'pwllheli', '-', proper, loc ).
noun( 'pyle', '-', proper, loc ).
noun( 'pyongyang', '-', proper, loc ).
noun( 'qc', '-', count, _ ).
noun( 'qed', '-', proper, _ ).
noun( 'qatar', '-', proper, loc ).
noun( 'qatari', 'qataris', count, _ ).
noun( 'quai d\'orsay', '-', count, _ ).
noun( 'quaker', 'quakers', count, _ ).
noun( 'quebec', '-', proper, loc ).
noun( 'queenborough-in-sheppey', '-', proper, loc ).
noun( 'queensbury', '-', proper, loc ).
noun( 'queensferry', '-', proper, loc ).
noun( 'queensland', '-', proper, loc ).
noun( 'quentin', '-', proper, per ).
noun( 'quinquagesima', 'quinquagesimas', count, _ ).
noun( 'quinquagesima', '-', proper, _ ).
noun( 'quito', '-', proper, loc ).
noun( 'quonset', 'quonsets', count, _ ).
noun( 'r\"ontgen', 'r\"ontgens', count, _ ).
noun( 'ra', '-', count, _ ).
noun( 'rada', '-', proper, _ ).
noun( 'raf', '-', count, _ ).
noun( 'rc', '-', proper, _ ).
noun( 'rip', '-', proper, _ ).
noun( 'rm', '-', count, _ ).
noun( 'rn', '-', count, _ ).
noun( 'rspca', '-', count, _ ).
noun( 'rsvp', '-', proper, _ ).
noun( 'rachel', '-', proper, per ).
noun( 'radcliffe', '-', proper, loc ).
noun( 'radcliffe on trent', '-', proper, loc ).
noun( 'radlett', '-', proper, loc ).
noun( 'radstock', '-', proper, loc ).
noun( 'rainford', '-', proper, loc ).
noun( 'ralph', '-', proper, per ).
noun( 'ramadan', 'ramadans', count, _ ).
noun( 'ramadan', '-', proper, _ ).
noun( 'ramsbottom', '-', proper, loc ).
noun( 'ramsgate', '-', proper, loc ).
noun( 'randolph', '-', proper, per ).
noun( 'rangoon', '-', proper, loc ).
noun( 'ravenna', '-', proper, loc ).
noun( 'rawtenstall', '-', proper, loc ).
noun( 'ray', '-', proper, per ).
noun( 'raymond', '-', proper, per ).
noun( 'rd', '-', proper, _ ).
noun( 'reading', '-', proper, loc ).
noun( 'realtor', 'realtors', count, _ ).
noun( 'rebecca', '-', proper, per ).
noun( 'recife', '-', proper, loc ).
noun( 'recklinghausen', '-', proper, loc ).
noun( 'redbridge', '-', proper, loc ).
noun( 'redditch', '-', proper, loc ).
noun( 'redruth', '-', proper, loc ).
noun( 'reg', '-', proper, per ).
noun( 'regensburg', '-', proper, loc ).
noun( 'regina', '-', proper, _ ).
noun( 'reginald', '-', proper, per ).
noun( 'reich', 'reichs', count, _ ).
noun( 'reigate', '-', proper, loc ).
noun( 'remscheid', '-', proper, loc ).
noun( 'renfrew', '-', proper, loc ).
noun( 'rennes', '-', proper, loc ).
noun( 'retford', '-', proper, loc ).
noun( 'rev', '-', proper, per ).
noun( 'revd', '-', proper, per ).
noun( 'rex', '-', proper, per ).
noun( 'reykjavik', '-', proper, loc ).
noun( 'rheims', '-', proper, loc ).
noun( 'rheydt', '-', proper, loc ).
noun( 'rhine', '-', proper, _ ).
noun( 'rhinestone', 'rhinestones', count, _ ).
noun( 'rhode', '-', proper, loc ).
noun( 'rhondda', '-', proper, loc ).
noun( 'rhosllanerchrugog', '-', proper, loc ).
noun( 'rhyl', '-', proper, loc ).
noun( 'rhymney', '-', proper, loc ).
noun( 'richard', '-', proper, per ).
noun( 'richmond', '-', proper, loc ).
noun( 'richmond upon thames', '-', proper, loc ).
noun( 'rick', '-', proper, per ).
noun( 'rickmansworth', '-', proper, loc ).
noun( 'ricky', '-', proper, per ).
noun( 'riesling', 'rieslings', count, _ ).
noun( 'riga', '-', proper, loc ).
noun( 'rijeka', '-', proper, loc ).
noun( 'rimini', '-', proper, loc ).
noun( 'ringwood', '-', proper, loc ).
noun( 'rio de janeiro', '-', proper, loc ).
noun( 'ripley', '-', proper, loc ).
noun( 'ripon', '-', proper, loc ).
noun( 'risborough', '-', proper, loc ).
noun( 'risca', '-', proper, loc ).
noun( 'rishton', '-', proper, loc ).
noun( 'rita', '-', proper, per ).
noun( 'riviera', '-', proper, _ ).
noun( 'rob', '-', proper, per ).
noun( 'robert', '-', proper, per ).
noun( 'robin', '-', proper, per ).
noun( 'rochdale', '-', proper, loc ).
noun( 'rochester', '-', proper, loc ).
noun( 'rod', '-', proper, per ).
noun( 'rodney', '-', proper, per ).
noun( 'roentgen', 'roentgens', count, _ ).
noun( 'roger', '-', proper, per ).
noun( 'romaic', '-', mass, _ ).
noun( 'roman', 'romans', count, _ ).
noun( 'romanesque', '-', mass, _ ).
noun( 'romania', '-', proper, loc ).
noun( 'romanian', 'romanians', both, _ ).
noun( 'romany', 'romanies', both, _ ).
noun( 'rome', '-', proper, loc ).
noun( 'romsey', '-', proper, loc ).
noun( 'ron', '-', proper, per ).
noun( 'ronald', '-', proper, per ).
noun( 'roneo', 'roneos', count, _ ).
noun( 'ronnie', '-', proper, per ).
noun( 'roquefort', '-', mass, _ ).
noun( 'rosalie', '-', proper, per ).
noun( 'rosalind', '-', proper, per ).
noun( 'rosamund', '-', proper, per ).
noun( 'rosario', '-', proper, loc ).
noun( 'roscommon', '-', proper, loc ).
noun( 'rose', '-', proper, per ).
noun( 'rosemary', '-', proper, per ).
noun( 'rosie', '-', proper, per ).
noun( 'roslyn', '-', proper, per ).
noun( 'ross-on-wye', '-', proper, loc ).
noun( 'rossington', '-', proper, loc ).
noun( 'rosslare', '-', proper, loc ).
noun( 'rostock', '-', proper, loc ).
noun( 'rostov-na-donu', '-', proper, loc ).
noun( 'rotarian', 'rotarians', count, _ ).
noun( 'rotherham', '-', proper, loc ).
noun( 'rothesay', '-', proper, loc ).
noun( 'rothwell', '-', proper, loc ).
noun( 'rotterdam', '-', proper, loc ).
noun( 'roubaix', '-', proper, loc ).
noun( 'rouen', '-', proper, loc ).
noun( 'roundhead', 'roundheads', count, _ ).
noun( 'rowlands gill', '-', proper, loc ).
noun( 'roy', '-', proper, per ).
noun( 'royston', '-', proper, loc ).
noun( 'rt hon', '-', proper, _ ).
noun( 'rubicon', 'rubicons', count, _ ).
noun( 'ruddington', '-', proper, loc ).
noun( 'rudolf', '-', proper, per ).
noun( 'rudy', '-', proper, per ).
noun( 'rugby', '-', proper, loc ).
noun( 'rugeley', '-', proper, loc ).
noun( 'runcorn', '-', proper, loc ).
noun( 'rupert', '-', proper, per ).
noun( 'russia', '-', proper, loc ).
noun( 'russian', 'russians', both, _ ).
noun( 'ruth', '-', proper, per ).
noun( 'rutherglen', '-', proper, loc ).
noun( 'rutland', '-', proper, loc ).
noun( 'rwanda', '-', proper, loc ).
noun( 'rwandan', 'rwandans', count, _ ).
noun( 'ryde', '-', proper, loc ).
noun( 'ryehill', '-', proper, loc ).
noun( 'salt', '-', proper, _ ).
noun( 'sos', '-', count, _ ).
noun( '-', 'soss', count, _ ).
noun( 'srn', '-', count, _ ).
noun( 'ss', '-', count, _ ).
noun( 'std', '-', proper, _ ).
noun( 'saarbrucken', '-', proper, loc ).
noun( 'sabadeli', '-', proper, loc ).
noun( 'sabah', '-', proper, loc ).
noun( 'sabahan', 'sabahans', count, _ ).
noun( 'sabbath', 'sabbaths', count, _ ).
noun( 'sacramento', '-', proper, loc ).
noun( 'saffron walden', '-', proper, loc ).
noun( 'sagittarius', '-', proper, _ ).
noun( 'saigon', '-', proper, loc ).
noun( 'salamanca', '-', proper, loc ).
noun( 'sale', '-', proper, loc ).
noun( 'salerno', '-', proper, loc ).
noun( 'salford', '-', proper, loc ).
noun( 'salisbury', '-', proper, loc ).
noun( 'sally', '-', proper, per ).
noun( 'saltash', '-', proper, loc ).
noun( 'saltburn', '-', proper, loc ).
noun( 'saltcoats', '-', proper, loc ).
noun( 'salvador', '-', proper, loc ).
noun( 'salvadorean', 'salvadoreans', count, _ ).
noun( 'salzburg', '-', proper, loc ).
noun( 'sam', '-', proper, per ).
noun( 'samantha', '-', proper, per ).
noun( 'samaritan', 'samaritans', count, _ ).
noun( 'sammy', '-', proper, per ).
noun( 'samoa', '-', proper, loc ).
noun( 'samoan', 'samoans', count, _ ).
noun( 'samuel', '-', proper, per ).
noun( 'san antonio', '-', proper, loc ).
noun( 'san bernardino', '-', proper, loc ).
noun( 'san diego', '-', proper, loc ).
noun( 'san francisco', '-', proper, loc ).
noun( 'san jose', '-', proper, loc ).
noun( 'san juan', '-', proper, loc ).
noun( 'san marinese', 'san marinese', count, _ ).
noun( 'san marino', '-', proper, loc ).
noun( 'san sebastian', '-', proper, loc ).
noun( 'sandbach', '-', proper, loc ).
noun( 'sandown', '-', proper, loc ).
noun( 'sandra', '-', proper, per ).
noun( 'sandy', '-', proper, per ).
noun( 'sango', '-', mass, _ ).
noun( 'sanskrit', '-', mass, _ ).
noun( 'santa claus', 'santa clauses', count, _ ).
noun( 'santa claus', '-', proper, _ ).
noun( 'santander', '-', proper, loc ).
noun( 'santiago', '-', proper, loc ).
noun( 'santo domingo', '-', proper, loc ).
noun( 'santos', '-', proper, loc ).
noun( 'sao paolo', '-', proper, loc ).
noun( 'sapporo', '-', proper, loc ).
noun( 'sara', '-', proper, per ).
noun( 'saracen', 'saracens', count, _ ).
noun( 'sarah', '-', proper, per ).
noun( 'sarajevo', '-', proper, loc ).
noun( 'saratov', '-', proper, loc ).
noun( 'sarawak', '-', proper, loc ).
noun( 'sarawakian', 'sarawakians', count, _ ).
noun( 'sarisbury', '-', proper, loc ).
noun( 'saskatchewan', '-', proper, loc ).
noun( 'sassari', '-', proper, loc ).
noun( 'sassenach', 'sassenachs', count, _ ).
noun( 'sat', '-', proper, _ ).
noun( 'satan', 'satans', count, _ ).
noun( 'saturday', 'saturdays', count, _ ).
noun( 'saturday', '-', proper, _ ).
noun( 'saturn', '-', proper, _ ).
noun( 'sauchie', '-', proper, loc ).
noun( 'saudi arabia', '-', proper, loc ).
noun( 'saudi arabian', 'saudi arabians', count, _ ).
noun( 'sawbridgeworth', '-', proper, loc ).
noun( 'saxon', 'saxons', count, _ ).
noun( 'scandinavian', 'scandinavians', count, _ ).
noun( 'scarborough', '-', proper, loc ).
noun( 'schwerin', '-', proper, loc ).
noun( 'sci fi', '-', mass, _ ).
noun( 'scorpio', '-', proper, _ ).
noun( 'scot', 'scots', count, _ ).
noun( 'scotch', 'scotches', both, _ ).
noun( 'scotchman', 'scotchmen', count, _ ).
noun( 'scotchwoman', 'scotchwomen', count, _ ).
noun( 'scotland', '-', proper, loc ).
noun( 'scotland yard', '-', proper, _ ).
noun( 'scotsman', 'scotsmen', count, _ ).
noun( 'scotswoman', 'scotswomen', count, _ ).
noun( 'scrooge', 'scrooges', count, _ ).
noun( 'scunthorpe', '-', proper, loc ).
noun( 'scylla', 'scyllas', count, _ ).
noun( 'seaford', '-', proper, loc ).
noun( 'seaham', '-', proper, loc ).
noun( 'sealyham', 'sealyhams', count, _ ).
noun( 'seamus', '-', proper, per ).
noun( 'sean', '-', proper, per ).
noun( 'seaton burn', '-', proper, loc ).
noun( 'seaton delaval', '-', proper, loc ).
noun( 'seattle', '-', proper, loc ).
noun( 'secretary-general', 'secretary-generals', count, _ ).
noun( 'securicor', '-', proper, _ ).
noun( 'selby', '-', proper, loc ).
noun( 'selkirk', '-', proper, loc ).
noun( 'selsey', '-', proper, loc ).
noun( 'selston', '-', proper, loc ).
noun( 'semarang', '-', proper, loc ).
noun( 'semite', 'semites', count, _ ).
noun( 'senegal', '-', proper, loc ).
noun( 'senegalese', 'senegalese', count, _ ).
noun( 'seoul', '-', proper, loc ).
noun( 'sept', '-', proper, _ ).
noun( 'september', 'septembers', count, _ ).
noun( 'september', '-', proper, _ ).
noun( 'septuagint', 'septuagints', count, _ ).
noun( 'serbo-croat', '-', mass, _ ).
noun( 'serjeant-at-arms', 'serjeants-at-arms', count, _ ).
noun( 'sesotho', '-', mass, _ ).
noun( 'setswana', '-', mass, _ ).
noun( 'seven sisters', '-', proper, loc ).
noun( 'sevenoaks', '-', proper, loc ).
noun( 'sevilla', '-', proper, loc ).
noun( 'seville', '-', proper, loc ).
noun( 'seychelles', '-', proper, loc ).
noun( 'seychellois', 'seychellois', count, _ ).
noun( 'sgt', '-', proper, per ).
noun( 'shanghai', '-', proper, loc ).
noun( 'shanklin', '-', proper, loc ).
noun( 'sharon', '-', proper, per ).
noun( 'shavian', 'shavians', count, _ ).
noun( 'sheerness', '-', proper, loc ).
noun( 'sheffield', '-', proper, loc ).
noun( 'sheila', '-', proper, per ).
noun( 'shelford', '-', proper, loc ).
noun( 'shepshed', '-', proper, loc ).
noun( 'shepton mallet', '-', proper, loc ).
noun( 'sheraton', '-', mass, _ ).
noun( 'sherbourne', '-', proper, loc ).
noun( 'sheringham', '-', proper, loc ).
noun( 'shetland', '-', proper, loc ).
noun( 'shevington', '-', proper, loc ).
noun( 'shihkiachwang', '-', proper, loc ).
noun( 'shildon', '-', proper, loc ).
noun( 'shirebrook', '-', proper, loc ).
noun( 'shiremoor', '-', proper, loc ).
noun( 'shirley', '-', proper, per ).
noun( 'shotton', '-', proper, loc ).
noun( 'shotts', '-', proper, loc ).
noun( 'shrewsbury', '-', proper, loc ).
noun( 'shropshire', '-', proper, loc ).
noun( 'shrove tuesday', 'shrove tuesdays', count, _ ).
noun( 'shrove tuesday', '-', proper, _ ).
noun( 'siam', '-', proper, loc ).
noun( 'siamese', 'siamese', count, _ ).
noun( 'sian', '-', proper, loc ).
noun( 'sicilian', 'sicilians', count, _ ).
noun( 'sid', '-', proper, per ).
noun( 'sidmouth', '-', proper, loc ).
noun( 'sidney', '-', proper, per ).
noun( 'sierra leone', '-', proper, loc ).
noun( 'sierra leonian', 'sierra leonians', count, _ ).
noun( 'sikh', 'sikhs', count, _ ).
noun( 'sileby', '-', proper, loc ).
noun( 'silsden', '-', proper, loc ).
noun( 'silvia', '-', proper, per ).
noun( 'simon', '-', proper, per ).
noun( 'singapore', '-', proper, loc ).
noun( 'singaporean', 'singaporeans', count, _ ).
noun( 'sinhala', '-', mass, _ ).
noun( 'sinhalese', 'sinhalese', count, _ ).
noun( 'sining', '-', proper, loc ).
noun( 'sinn fein', '-', proper, _ ).
noun( 'sinologist', 'sinologists', count, _ ).
noun( 'sinology', '-', mass, _ ).
noun( 'sioux', 'sioux', count, _ ).
noun( 'sir roger de coverley', '-', count, _ ).
noun( 'siracusa', '-', proper, loc ).
noun( 'siswati', '-', mass, _ ).
noun( 'sittingbourne', '-', proper, loc ).
noun( 'skegness', '-', proper, loc ).
noun( 'skelmanthorpe', '-', proper, loc ).
noun( 'skelmersdale', '-', proper, loc ).
noun( 'skewen', '-', proper, loc ).
noun( 'skipton', '-', proper, loc ).
noun( 'skopje', '-', proper, loc ).
noun( 'slav', 'slavs', count, _ ).
noun( 'sleaford', '-', proper, loc ).
noun( 'sligo', '-', proper, loc ).
noun( 'slough', '-', proper, loc ).
noun( 'slovak', '-', mass, _ ).
noun( 'slovenian', 'slovenians', both, _ ).
noun( 'snr', '-', proper, _ ).
noun( 'soc', '-', count, _ ).
noun( 'sofia', '-', proper, loc ).
noun( 'soho', '-', proper, _ ).
noun( 'sol', '-', proper, _ ).
noun( 'solicitor-general', 'solicitor-generals', count, _ ).
noun( 'solihull', '-', proper, loc ).
noun( 'solingen', '-', proper, loc ).
noun( 'solon', '-', proper, _ ).
noun( 'somali', 'somalis', both, _ ).
noun( 'somalia', '-', proper, loc ).
noun( 'somalian', 'somalians', count, _ ).
noun( 'somercotes', '-', proper, loc ).
noun( 'somerset', '-', proper, loc ).
noun( 'sonia', '-', proper, per ).
noun( 'soochow', '-', proper, loc ).
noun( 'sophia', '-', proper, per ).
noun( 'sophie', '-', proper, per ).
noun( 'sotho', 'sothos', both, _ ).
noun( 'south kirkby', '-', proper, loc ).
noun( 'south normantown', '-', proper, loc ).
noun( 'south ockendon', '-', proper, loc ).
noun( 'south shields', '-', proper, loc ).
noun( 'southampton', '-', proper, loc ).
noun( 'southend-on-sea', '-', proper, loc ).
noun( 'southport', '-', proper, loc ).
noun( 'southwark', '-', proper, loc ).
noun( 'soviet', 'soviets', count, _ ).
noun( 'spain', '-', proper, loc ).
noun( 'spalding', '-', proper, loc ).
noun( 'spaniard', 'spaniards', count, _ ).
noun( 'spanish', '-', mass, _ ).
noun( 'spartan', 'spartans', count, _ ).
noun( 'spenborough', '-', proper, loc ).
noun( 'spennymoor', '-', proper, loc ).
noun( 'split', '-', proper, loc ).
noun( 'spode', '-', mass, _ ).
noun( 'springfield', '-', proper, loc ).
noun( 'sq', '-', proper, _ ).
noun( 'sr', '-', proper, per ).
noun( 'sri lanka', '-', proper, loc ).
noun( 'sri lankan', 'sri lankans', count, _ ).
noun( 'st', '-', proper, _ ).
noun( 'st albans', '-', proper, loc ).
noun( 'st andrews', '-', proper, loc ).
noun( 'st andrews major', '-', proper, loc ).
noun( 'st austell', '-', proper, loc ).
noun( 'st blazey', '-', proper, loc ).
noun( 'st helens', '-', proper, loc ).
noun( 'st ives', '-', proper, loc ).
noun( 'st louis', '-', proper, loc ).
noun( 'st neots', '-', proper, loc ).
noun( 'st paul', '-', proper, loc ).
noun( 'st petersburg', '-', proper, loc ).
noun( 'st stephen', '-', proper, loc ).
noun( 'sta', '-', proper, _ ).
noun( 'stafford', '-', proper, loc ).
noun( 'staffordshire', '-', proper, loc ).
noun( 'stakeford', '-', proper, loc ).
noun( 'stalybridge', '-', proper, loc ).
noun( 'stamford', '-', proper, loc ).
noun( 'stan', '-', proper, per ).
noun( 'stanford', '-', proper, loc ).
noun( 'stanley', '-', proper, per ).
noun( 'statehouse', 'statehouses', count, _ ).
noun( 'stella', '-', proper, per ).
noun( 'sten', '-', proper, _ ).
noun( 'stephanie', '-', proper, per ).
noun( 'stephen', '-', proper, per ).
noun( 'steve', '-', proper, per ).
noun( 'steven', '-', proper, per ).
noun( 'stevenage', '-', proper, loc ).
noun( 'stevenston', '-', proper, loc ).
noun( 'stewart', '-', proper, per ).
noun( 'steyning', '-', proper, loc ).
noun( 'stilton', 'stiltons', both, _ ).
noun( 'stirling', '-', proper, loc ).
noun( 'stockholm', '-', proper, loc ).
noun( 'stockport', '-', proper, loc ).
noun( 'stocksbridge', '-', proper, loc ).
noun( 'stoke-on-trent', '-', proper, loc ).
noun( 'stone', '-', proper, loc ).
noun( 'stonehouse', '-', proper, loc ).
noun( 'stornoway', '-', proper, loc ).
noun( 'stotfold', '-', proper, loc ).
noun( 'stourbridge', '-', proper, loc ).
noun( 'stourport-on-severn', '-', proper, loc ).
noun( 'stowmarket', '-', proper, loc ).
noun( 'stranraer', '-', proper, loc ).
noun( 'strasbourg', '-', proper, loc ).
noun( 'stratford-on-avon', '-', proper, loc ).
noun( 'strathaven', '-', proper, loc ).
noun( 'strathclyde', '-', proper, loc ).
noun( 'street', '-', proper, loc ).
noun( 'stretford', '-', proper, loc ).
noun( 'stroud', '-', proper, loc ).
noun( 'stuart', '-', proper, per ).
noun( 'studley', '-', proper, loc ).
noun( 'stuttgart', '-', proper, loc ).
noun( 'styx', '-', proper, _ ).
noun( 'suchow', '-', proper, loc ).
noun( 'sudan', '-', proper, loc ).
noun( 'sudanese', 'sudanese', count, _ ).
noun( 'sudbury', '-', proper, loc ).
noun( 'sue', '-', proper, per ).
noun( 'suffolk', '-', proper, loc ).
noun( 'sumatra', '-', proper, loc ).
noun( 'sumatran', 'sumatrans', count, _ ).
noun( 'sun', '-', proper, _ ).
noun( 'sunday', 'sundays', count, _ ).
noun( 'sunday', '-', proper, _ ).
noun( 'sunderland', '-', proper, loc ).
noun( 'supt', '-', proper, per ).
noun( 'surabaja', '-', proper, loc ).
noun( 'surrey', '-', proper, loc ).
noun( 'susan', '-', proper, per ).
noun( 'susanna', '-', proper, per ).
noun( 'susie', '-', proper, per ).
noun( 'sussex', '-', proper, loc ).
noun( 'sutton', '-', proper, loc ).
noun( 'sutton coldfield', '-', proper, loc ).
noun( 'sutton in ashfield', '-', proper, loc ).
noun( 'sutton-at-hone', '-', proper, loc ).
noun( 'suzanne', '-', proper, per ).
noun( 'sverdlovsk', '-', proper, loc ).
noun( 'swadlincote', '-', proper, loc ).
noun( 'swahili', '-', mass, _ ).
noun( 'swanage', '-', proper, loc ).
noun( 'swanley', '-', proper, loc ).
noun( 'swansea', '-', proper, loc ).
noun( 'swazi', 'swazis', both, _ ).
noun( 'swaziland', '-', proper, loc ).
noun( 'swede', 'swedes', count, _ ).
noun( 'sweden', '-', proper, loc ).
noun( 'swedish', '-', mass, _ ).
noun( 'swindon', '-', proper, loc ).
noun( 'swinton', '-', proper, loc ).
noun( 'swiss', 'swiss', count, _ ).
noun( 'switzerland', '-', proper, loc ).
noun( 'sydney', '-', proper, loc ).
noun( 'sylvia', '-', proper, per ).
noun( 'syracuse', '-', proper, loc ).
noun( 'syria', '-', proper, loc ).
noun( 'syrian', 'syrians', count, _ ).
noun( 'syston', '-', proper, loc ).
noun( 't-junction', 't-junctions', count, _ ).
noun( 't-shirt', 't-shirts', count, _ ).
noun( 't-square', 't-squares', count, _ ).
noun( 'tb', '-', mass, _ ).
noun( 'tnt', '-', mass, _ ).
noun( 'tuc', '-', count, _ ).
noun( 'tv', '-', count, _ ).
noun( '-', 'tvs', count, _ ).
noun( 'tadley', '-', proper, loc ).
noun( 'taegu', '-', proper, loc ).
noun( 'taffy', 'taffies', count, _ ).
noun( 'tahiti', '-', proper, loc ).
noun( 'tahitian', 'tahitians', count, _ ).
noun( 'taipei', '-', proper, loc ).
noun( 'taiwan', '-', proper, loc ).
noun( 'taiwanese', 'taiwanese', count, _ ).
noun( 'taiyuan', '-', proper, loc ).
noun( 'talmud', 'talmuds', count, _ ).
noun( 'tamil', 'tamils', both, _ ).
noun( 'tammany', '-', count, _ ).
noun( 'tampere', '-', proper, loc ).
noun( 'tamworth', '-', proper, loc ).
noun( 'tangshan', '-', proper, loc ).
noun( 'tanzania', '-', proper, loc ).
noun( 'tanzanian', 'tanzanians', count, _ ).
noun( 'taranaki', '-', proper, loc ).
noun( 'taranto', '-', proper, loc ).
noun( 'tarrasa', '-', proper, loc ).
noun( 'tashkent', '-', proper, loc ).
noun( 'tasmania', '-', proper, loc ).
noun( 'taunton', '-', proper, loc ).
noun( 'taurus', '-', proper, _ ).
noun( 'taverham', '-', proper, loc ).
noun( 'tavistock', '-', proper, loc ).
noun( 'tayside', '-', proper, loc ).
noun( 'tbilisi', '-', proper, loc ).
noun( 'te deum', 'te deums', count, _ ).
noun( 'tech', '-', count, _ ).
noun( 'technicolor', '-', mass, _ ).
noun( 'ted', 'teds', count, _ ).
noun( 'ted', '-', proper, per ).
noun( 'teddy', '-', proper, per ).
noun( 'teddy boy', 'teddy boys', count, _ ).
noun( 'teesside', '-', proper, loc ).
noun( 'teheran', '-', proper, loc ).
noun( 'teignmouth', '-', proper, loc ).
noun( 'tel-aviv', '-', proper, loc ).
noun( 'telford', '-', proper, loc ).
noun( 'telstar', '-', proper, _ ).
noun( 'tenby', '-', proper, loc ).
noun( 'tennessee', '-', proper, loc ).
noun( 'tenterden', '-', proper, loc ).
noun( 'terence', '-', proper, per ).
noun( 'teresa', '-', proper, per ).
noun( 'termi', '-', proper, loc ).
noun( 'terr', '-', proper, _ ).
noun( 'terry', '-', proper, per ).
noun( 'tess', '-', proper, per ).
noun( 'tessa', '-', proper, per ).
noun( 'teuton', 'teutons', count, _ ).
noun( 'tewkesbury', '-', proper, loc ).
noun( 'texan', 'texans', count, _ ).
noun( 'texas', '-', proper, loc ).
noun( 'thai', 'thais', both, _ ).
noun( 'thailand', '-', proper, loc ).
noun( 'thame', '-', proper, loc ).
noun( 'thanksgiving', 'thanksgivings', count, _ ).
noun( 'thatcham', '-', proper, loc ).
noun( 'theo', '-', proper, per ).
noun( 'theodore', '-', proper, per ).
noun( 'theresa', '-', proper, per ).
noun( 'thespian', 'thespians', count, _ ).
noun( 'thessaloniki', '-', proper, loc ).
noun( 'thetford', '-', proper, loc ).
noun( 'thionville', '-', proper, loc ).
noun( 'thirsk', '-', proper, loc ).
noun( 'thomas', '-', proper, per ).
noun( 'thornbury', '-', proper, loc ).
noun( 'thorne', '-', proper, loc ).
noun( 'thurcroft', '-', proper, loc ).
noun( 'thurmaston', '-', proper, loc ).
noun( 'thurrock', '-', proper, loc ).
noun( 'thurs', '-', proper, _ ).
noun( 'thursday', 'thursdays', count, _ ).
noun( 'thursday', '-', proper, _ ).
noun( 'thurso', '-', proper, loc ).
noun( 'tibet', '-', proper, loc ).
noun( 'tibetan', 'tibetans', both, _ ).
noun( 'tidworth', '-', proper, loc ).
noun( 'tientsin', '-', proper, loc ).
noun( 'tilburg', '-', proper, loc ).
noun( 'tilbury', '-', proper, loc ).
noun( 'tim', '-', proper, per ).
noun( 'timmy', '-', proper, per ).
noun( 'timothy', '-', proper, per ).
noun( 'tina', '-', proper, per ).
noun( 'tipperary', '-', proper, loc ).
noun( 'tiverton', '-', proper, loc ).
noun( 'tobago', '-', proper, loc ).
noun( 'tobagonian', 'tobagonians', count, _ ).
noun( 'toby', '-', proper, per ).
noun( 'todmorden', '-', proper, loc ).
noun( 'togo', '-', proper, loc ).
noun( 'togolese', 'togolese', count, _ ).
noun( 'tokay', '-', mass, _ ).
noun( 'tokyo', '-', proper, loc ).
noun( 'tom', '-', proper, per ).
noun( 'tommy', '-', proper, per ).
noun( 'tonbridge', '-', proper, loc ).
noun( 'tonga', '-', proper, loc ).
noun( 'tongan', 'tongans', both, _ ).
noun( 'tony', '-', proper, per ).
noun( 'tonyrefail', '-', proper, loc ).
noun( 'torbay', '-', proper, loc ).
noun( 'torino', '-', proper, loc ).
noun( 'toronto', '-', proper, loc ).
noun( 'tory', 'tories', count, _ ).
noun( 'totnes', '-', proper, loc ).
noun( 'toulon', '-', proper, loc ).
noun( 'toulouse', '-', proper, loc ).
noun( 'tours', '-', proper, loc ).
noun( 'tower hamlets', '-', proper, loc ).
noun( 'tracy', '-', proper, per ).
noun( 'tralee', '-', proper, loc ).
noun( 'tranent', '-', proper, loc ).
noun( 'trappist', 'trappists', count, _ ).
noun( 'tredegar', '-', proper, loc ).
noun( 'trevor', '-', proper, per ).
noun( 'trieste', '-', proper, loc ).
noun( 'tring', '-', proper, loc ).
noun( 'trinidad', '-', proper, loc ).
noun( 'trinidadian', 'trinidadians', count, _ ).
noun( 'trojan', 'trojans', count, _ ).
noun( 'troon', '-', proper, loc ).
noun( 'trowbridge', '-', proper, loc ).
noun( 'truro', '-', proper, loc ).
noun( 'tsar', 'tsars', count, _ ).
noun( 'tsarina', 'tsarinas', count, _ ).
noun( 'tsinan', '-', proper, loc ).
noun( 'tsingtao', '-', proper, loc ).
noun( 'tsitsihar', '-', proper, loc ).
noun( 'tswana', 'tswanas', both, _ ).
noun( 'tues', '-', proper, _ ).
noun( 'tuesday', 'tuesdays', count, _ ).
noun( 'tuesday', '-', proper, _ ).
noun( 'tullibody', '-', proper, loc ).
noun( 'tunbridge wells', '-', proper, loc ).
noun( 'tunis', '-', proper, loc ).
noun( 'tunisia', '-', proper, loc ).
noun( 'tunisian', 'tunisians', count, _ ).
noun( 'turin', '-', proper, loc ).
noun( 'turk', 'turks', count, _ ).
noun( 'turkey', '-', proper, loc ).
noun( 'turkish', '-', mass, _ ).
noun( 'turku', '-', proper, loc ).
noun( 'twelfth-night', 'twelfth-nights', count, _ ).
noun( 'twelfth-night', '-', proper, _ ).
noun( 'tyne', '-', proper, loc ).
noun( 'tynemouth', '-', proper, loc ).
noun( 'tyrone', '-', proper, loc ).
noun( 'tzupo', '-', proper, loc ).
noun( 'u-boat', 'u-boats', count, _ ).
noun( 'u-turn', 'u-turns', count, _ ).
noun( 'udi', '-', proper, _ ).
noun( 'ufo', '-', count, _ ).
noun( '-', 'ufo\'s', count, _ ).
noun( 'uk', '-', count, _ ).
noun( 'un', '-', count, _ ).
noun( 'unesco', '-', proper, _ ).
noun( 'unicef', '-', proper, _ ).
noun( 'us', '-', count, _ ).
noun( 'usa', '-', count, _ ).
noun( 'usaf', '-', count, _ ).
noun( 'usn', '-', count, _ ).
noun( 'uss', '-', count, _ ).
noun( 'ussr', '-', count, _ ).
noun( 'uddingston', '-', proper, loc ).
noun( 'udine', '-', proper, loc ).
noun( 'ufa', '-', proper, loc ).
noun( 'uganda', '-', proper, loc ).
noun( 'ugandan', 'ugandans', count, _ ).
noun( 'ulster', '-', proper, loc ).
noun( 'ulverston', '-', proper, loc ).
noun( 'unitarian', 'unitarians', count, _ ).
noun( 'unitarianism', '-', mass, _ ).
noun( 'univ', '-', proper, _ ).
noun( 'up holland', '-', proper, loc ).
noun( 'upton', '-', proper, loc ).
noun( 'uranus', '-', proper, _ ).
noun( 'urdu', '-', mass, _ ).
noun( 'ursula', '-', proper, per ).
noun( 'uruguay', '-', proper, loc ).
noun( 'uruguayan', 'uruguayans', count, _ ).
noun( 'urumchi', '-', proper, loc ).
noun( 'ushaw moor', '-', proper, loc ).
noun( 'utah', '-', proper, loc ).
noun( 'utopia', 'utopias', count, _ ).
noun( 'utrecht', '-', proper, loc ).
noun( 'uttoxeter', '-', proper, loc ).
noun( 'v and a', '-', count, _ ).
noun( 'vat', '-', mass, _ ).
noun( 'vc', '-', count, _ ).
noun( 'vd', '-', mass, _ ).
noun( 'vhf', '-', proper, _ ).
noun( 'vip', '-', count, _ ).
noun( '-', 'vips', count, _ ).
noun( 'vp', '-', count, _ ).
noun( 'vso', '-', proper, _ ).
noun( 'valencia', '-', proper, loc ).
noun( 'valenciennes', '-', proper, loc ).
noun( 'valladolid', '-', proper, loc ).
noun( 'vancouver', '-', proper, loc ).
noun( 'vandal', 'vandals', count, _ ).
noun( 'vanessa', '-', proper, per ).
noun( 'varanasi', '-', proper, loc ).
noun( 'vasteras', '-', proper, loc ).
noun( 'vatican', 'vaticans', count, _ ).
noun( 'venezia', '-', proper, loc ).
noun( 'venezuela', '-', proper, loc ).
noun( 'venezuelan', 'venezuelans', count, _ ).
noun( 'venice', '-', proper, loc ).
noun( 'ventnor', '-', proper, loc ).
noun( 'venus', '-', proper, _ ).
noun( 'vera', '-', proper, per ).
noun( 'vermont', '-', proper, loc ).
noun( 'vernon', '-', proper, per ).
noun( 'verona', '-', proper, loc ).
noun( 'veronica', '-', proper, per ).
noun( 'vic', '-', proper, per ).
noun( 'vicenza', '-', proper, loc ).
noun( 'vicky', '-', proper, per ).
noun( 'victor', '-', proper, per ).
noun( 'victoria', '-', proper, per ).
noun( 'victorian', 'victorians', count, _ ).
noun( 'victoriana', '-', mass, _ ).
noun( 'vienna', '-', proper, loc ).
noun( 'vietnam', '-', proper, loc ).
noun( 'vietnamese', 'vietnamese', both, _ ).
noun( 'viewpark', '-', proper, loc ).
noun( 'vigo', '-', proper, loc ).
noun( 'viking', 'vikings', count, _ ).
noun( 'vince', '-', proper, per ).
noun( 'vincent', '-', proper, per ).
noun( 'viola', '-', proper, per ).
noun( 'violet', '-', proper, per ).
noun( 'virginia', '-', mass, _ ).
noun( 'virginia', '-', proper, per ).
noun( 'virginia water', '-', proper, loc ).
noun( 'virgo', '-', proper, _ ).
noun( 'vitoria', '-', proper, loc ).
noun( 'viv', '-', proper, per ).
noun( 'vivian', '-', proper, per ).
noun( 'vivien', '-', proper, per ).
noun( 'vivienne', '-', proper, per ).
noun( 'volgograd', '-', proper, loc ).
noun( 'voronezh', '-', proper, loc ).
noun( 'vulgate', 'vulgates', count, _ ).
noun( 'wc', '-', count, _ ).
noun( '-', 'wcs', count, _ ).
noun( 'who', '-', count, _ ).
noun( 'wi', '-', count, _ ).
noun( 'wrac', '-', count, _ ).
noun( 'wraf', '-', count, _ ).
noun( 'wrns', '-', count, _ ).
noun( 'wakefield', '-', proper, loc ).
noun( 'wales', '-', proper, loc ).
noun( 'wall street', '-', proper, _ ).
noun( 'wallace', '-', proper, per ).
noun( 'wallasey', '-', proper, loc ).
noun( 'wallingford', '-', proper, loc ).
noun( 'wallsend', '-', proper, loc ).
noun( 'wally', '-', proper, per ).
noun( 'walsall', '-', proper, loc ).
noun( 'walsham', '-', proper, loc ).
noun( 'walter', '-', proper, per ).
noun( 'waltham forest', '-', proper, loc ).
noun( 'walton', '-', proper, loc ).
noun( 'wandsworth', '-', proper, loc ).
noun( 'wantage', '-', proper, loc ).
noun( 'ware', '-', proper, loc ).
noun( 'warley', '-', proper, loc ).
noun( 'warminster', '-', proper, loc ).
noun( 'warrington', '-', proper, loc ).
noun( 'warsaw', '-', proper, loc ).
noun( 'warsop', '-', proper, loc ).
noun( 'warwick', '-', proper, loc ).
noun( 'warwickshire', '-', proper, loc ).
noun( 'washington', '-', proper, loc ).
noun( 'washington dc', '-', proper, loc ).
noun( 'waterford', '-', proper, loc ).
noun( 'waterloo', '-', proper, _ ).
noun( 'watford', '-', proper, loc ).
noun( 'wayne', '-', proper, per ).
noun( 'wear', '-', proper, loc ).
noun( 'wed', '-', proper, _ ).
noun( 'wednesday', 'wednesdays', count, _ ).
noun( 'wednesday', '-', proper, _ ).
noun( 'wellingborough', '-', proper, loc ).
noun( 'wellington', '-', proper, loc ).
noun( 'wells', '-', proper, loc ).
noun( 'welsh', '-', mass, _ ).
noun( 'welshman', 'welshmen', count, _ ).
noun( 'welshpool', '-', proper, loc ).
noun( 'welwyn', '-', proper, loc ).
noun( 'welwyn garden city', '-', proper, loc ).
noun( 'wendover', '-', proper, loc ).
noun( 'wendy', '-', proper, per ).
noun( 'wesleyan', 'wesleyans', count, _ ).
noun( 'west bromwich', '-', proper, loc ).
noun( 'westbury', '-', proper, loc ).
noun( 'westhoughton', '-', proper, loc ).
noun( 'westmeath', '-', proper, loc ).
noun( 'westminster', '-', proper, loc ).
noun( 'westmoreland', '-', proper, loc ).
noun( 'weston-super-mare', '-', proper, loc ).
noun( 'wetherby', '-', proper, loc ).
noun( 'wexford', '-', proper, loc ).
noun( 'weymouth', '-', proper, loc ).
noun( 'whaley bridge', '-', proper, loc ).
noun( 'wheatley hill', '-', proper, loc ).
noun( 'whig', 'whigs', count, _ ).
noun( 'whit', 'whits', count, _ ).
noun( 'whitburn', '-', proper, loc ).
noun( 'whitby', '-', proper, loc ).
noun( 'whitchurch', '-', proper, loc ).
noun( 'whitehall', '-', proper, _ ).
noun( 'whitehaven', '-', proper, loc ).
noun( 'whitley bay', '-', proper, loc ).
noun( 'whitstable', '-', proper, loc ).
noun( 'whitsun', 'whitsuns', count, _ ).
noun( 'whitsun', '-', proper, _ ).
noun( 'whitsuntide', 'whitsuntides', count, _ ).
noun( 'whitsuntide', '-', proper, _ ).
noun( 'whittle-le-woods', '-', proper, loc ).
noun( 'whittlesey', '-', proper, loc ).
noun( 'whitworth', '-', proper, loc ).
noun( 'wick', '-', proper, loc ).
noun( 'wickford', '-', proper, loc ).
noun( 'wicklow', '-', proper, loc ).
noun( 'widnes', '-', proper, loc ).
noun( 'wiesbaden', '-', proper, loc ).
noun( 'wigan', '-', proper, loc ).
noun( 'wight', '-', proper, loc ).
noun( 'wilf', '-', proper, per ).
noun( 'wilfrid', '-', proper, per ).
noun( 'wilhelmshaven', '-', proper, loc ).
noun( 'will', '-', proper, per ).
noun( 'william', '-', proper, per ).
noun( 'willie', '-', proper, per ).
noun( 'willington', '-', proper, loc ).
noun( 'willy', '-', proper, per ).
noun( 'wilmslow', '-', proper, loc ).
noun( 'wilton', 'wiltons', count, _ ).
noun( 'wiltshire', '-', proper, loc ).
noun( 'wimbourne', '-', proper, loc ).
noun( 'wimbourne minster', '-', proper, loc ).
noun( 'winchester', '-', proper, loc ).
noun( 'windermere', '-', proper, loc ).
noun( 'windsor', '-', proper, loc ).
noun( 'wingate', '-', proper, loc ).
noun( 'winifred', '-', proper, per ).
noun( 'winnie', '-', proper, per ).
noun( 'winnipeg', '-', proper, loc ).
noun( 'winsford', '-', proper, loc ).
noun( 'wisbech', '-', proper, loc ).
noun( 'wisconsin', '-', proper, loc ).
noun( 'witham', '-', proper, loc ).
noun( 'withernsea', '-', proper, loc ).
noun( 'witney', '-', proper, loc ).
noun( 'woburn', '-', proper, loc ).
noun( 'wokingham', '-', proper, loc ).
noun( 'wolverhampton', '-', proper, loc ).
noun( 'wolverton', '-', proper, loc ).
noun( 'wombourne', '-', proper, loc ).
noun( 'woodbridge', '-', proper, loc ).
noun( 'wootton bassett', '-', proper, loc ).
noun( 'worcester', '-', proper, loc ).
noun( 'workington', '-', proper, loc ).
noun( 'worksop', '-', proper, loc ).
noun( 'worthing', '-', proper, loc ).
noun( 'wrexham', '-', proper, loc ).
noun( 'writtle', '-', proper, loc ).
noun( 'wroclaw', '-', proper, loc ).
noun( 'wroughton', '-', proper, loc ).
noun( 'wuhan', '-', proper, loc ).
noun( 'wuppertal', '-', proper, loc ).
noun( 'wurzburg', '-', proper, loc ).
noun( 'wusih', '-', proper, loc ).
noun( 'wycombe', '-', proper, loc ).
noun( 'wymondham', '-', proper, loc ).
noun( 'wyoming', '-', proper, loc ).
noun( 'wythall', '-', proper, loc ).
noun( 'x-ray', 'x-rays', count, _ ).
noun( 'xerox', 'xeroxes', count, _ ).
noun( 'xmas', 'xmases', count, _ ).
noun( 'xmas', '-', proper, _ ).
noun( 'yha', '-', count, _ ).
noun( 'ymca', '-', count, _ ).
noun( 'ywca', '-', count, _ ).
noun( 'yank', 'yanks', count, _ ).
noun( 'yankee', 'yankees', count, _ ).
noun( 'yarmouth', '-', proper, loc ).
noun( 'yaroslavi', '-', proper, loc ).
noun( 'yate', '-', proper, loc ).
noun( 'yateley', '-', proper, loc ).
noun( 'yemen', '-', proper, loc ).
noun( 'yemeni', 'yemenis', count, _ ).
noun( 'yeovil', '-', proper, loc ).
noun( 'yerevan', '-', proper, loc ).
noun( 'yiddish', '-', mass, _ ).
noun( 'yokohama', '-', proper, loc ).
noun( 'york', '-', proper, loc ).
noun( 'yorkley', '-', proper, loc ).
noun( 'yorkshire', '-', proper, loc ).
noun( 'ystradgynlais', '-', proper, loc ).
noun( 'yugoslav', 'yugoslavs', count, _ ).
noun( 'yugoslavia', '-', proper, loc ).
noun( 'yugoslavian', 'yugoslavians', count, _ ).
noun( 'yukon', '-', proper, loc ).
noun( 'yvonne', '-', proper, per ).
noun( 'zagreb', '-', proper, loc ).
noun( 'zaire', '-', proper, loc ).
noun( 'zairean', 'zaireans', count, _ ).
noun( 'zambia', '-', proper, loc ).
noun( 'zambian', 'zambians', count, _ ).
noun( 'zaporozhye', '-', proper, loc ).
noun( 'zaragoza', '-', proper, loc ).
noun( 'zealand', '-', proper, loc ).
noun( 'zealander', 'zealanders', count, _ ).
noun( 'zen', '-', mass, _ ).
noun( 'zimbabwe', '-', proper, loc ).
noun( 'zimbabwean', 'zimbabweans', count, _ ).
noun( 'zion', 'zions', count, _ ).
noun( 'zionism', '-', mass, _ ).
noun( 'zionist', 'zionists', count, _ ).
noun( 'zoe', '-', proper, per ).
noun( 'zurich', '-', proper, loc ).
noun( '_eclair', '_eclairs', count, _ ).
noun( '_eclat', '-', mass, _ ).
noun( '_elan', '-', mass, _ ).
noun( '_elite', '_elites', count, _ ).
noun( '_emigr_e', '_emigr_es', count, _ ).
noun( '_ep_ee', '_ep_ees', count, _ ).
noun( 'a', '-', count, _ ).
noun( 'abacus', 'abacuses', count, _ ).
noun( 'abandon', '-', mass, _ ).
noun( 'abandonment', '-', mass, _ ).
noun( 'abasement', '-', mass, _ ).
noun( 'abatement', '-', mass, _ ).
noun( 'abattoir', 'abattoirs', count, _ ).
noun( 'abb_e', 'abb_es', count, _ ).
noun( 'abbess', 'abbesses', count, _ ).
noun( 'abbey', 'abbeys', count, _ ).
noun( 'abbot', 'abbots', count, _ ).
noun( 'abbreviation', 'abbreviations', both, _ ).
noun( 'abdication', 'abdications', both, _ ).
noun( 'abdomen', 'abdomens', count, _ ).
noun( 'abduction', 'abductions', count, _ ).
noun( 'aberration', 'aberrations', both, _ ).
noun( 'abeyance', '-', mass, _ ).
noun( 'abhorrence', '-', mass, _ ).
noun( 'ability', 'abilities', both, _ ).
noun( 'abjection', 'abjections', count, _ ).
noun( 'abjuration', 'abjurations', both, _ ).
noun( 'ablative', 'ablatives', count, _ ).
noun( 'ablaut', 'ablauts', count, _ ).
noun( 'ablution', 'ablutions', count, _ ).
noun( 'abnegation', '-', mass, _ ).
noun( 'abnormality', 'abnormalities', both, _ ).
noun( 'abode', 'abodes', count, _ ).
noun( 'abolition', '-', mass, _ ).
noun( 'abolitionist', 'abolitionists', count, _ ).
noun( 'abomination', 'abominations', both, _ ).
noun( 'aboriginal', 'aboriginals', count, _ ).
noun( 'aborigine', 'aborigines', count, _ ).
noun( 'abortion', 'abortions', both, _ ).
noun( 'abortionist', 'abortionists', count, _ ).
noun( 'about-face', '-', count, _ ).
noun( 'abracadabra', '-', mass, _ ).
noun( 'abrasion', 'abrasions', both, _ ).
noun( 'abrasive', 'abrasives', both, _ ).
noun( 'abridgement', 'abridgements', both, _ ).
noun( 'abridgment', 'abridgments', both, _ ).
noun( 'abrogation', 'abrogations', count, _ ).
noun( 'abruptness', '-', mass, _ ).
noun( 'abscess', 'abscesses', count, _ ).
noun( 'absence', 'absences', both, _ ).
noun( 'absent-mindedness', '-', mass, _ ).
noun( 'absentee', 'absentees', count, _ ).
noun( 'absenteeism', '-', mass, _ ).
noun( 'absinth', '-', mass, _ ).
noun( 'absinthe', '-', mass, _ ).
noun( 'absolution', '-', mass, _ ).
noun( 'absolutism', '-', mass, _ ).
noun( 'absorbent', 'absorbents', count, _ ).
noun( 'absorption', '-', mass, _ ).
noun( 'abstainer', 'abstainers', count, _ ).
noun( 'abstemiousness', '-', mass, _ ).
noun( 'abstention', 'abstentions', both, _ ).
noun( 'abstinence', '-', mass, _ ).
noun( 'abstract', 'abstracts', count, _ ).
noun( 'abstraction', 'abstractions', both, _ ).
noun( 'abstruseness', '-', mass, _ ).
noun( 'absurdity', 'absurdities', both, _ ).
noun( 'abundance', '-', mass, _ ).
noun( 'abuse', 'abuses', both, _ ).
noun( 'abutment', 'abutments', count, _ ).
noun( 'abysm', 'abysms', count, _ ).
noun( 'abyss', 'abysses', count, _ ).
noun( 'acacia', 'acacias', count, _ ).
noun( 'academic', 'academics', count, _ ).
noun( 'academician', 'academicians', count, _ ).
noun( 'academy', 'academies', count, _ ).
noun( 'accelerando', 'accelerandos', count, _ ).
noun( 'acceleration', '-', mass, _ ).
noun( 'accelerator', 'accelerators', count, _ ).
noun( 'accent', 'accents', both, _ ).
noun( 'accentuation', 'accentuations', count, _ ).
noun( 'acceptability', '-', mass, _ ).
noun( 'acceptance', '-', mass, _ ).
noun( 'acceptation', 'acceptations', count, _ ).
noun( 'access', '-', mass, _ ).
noun( 'accessary', 'accessaries', count, _ ).
noun( 'accessibility', '-', mass, _ ).
noun( 'accession', 'accessions', both, _ ).
noun( 'accessory', 'accessories', count, _ ).
noun( 'accidence', '-', mass, _ ).
noun( 'accident', 'accidents', both, _ ).
noun( 'acclaim', '-', mass, _ ).
noun( 'acclamation', '-', mass, _ ).
noun( 'acclimation', '-', mass, _ ).
noun( 'acclimatization', '-', mass, _ ).
noun( 'acclivity', 'acclivities', count, _ ).
noun( 'accolade', 'accolades', count, _ ).
noun( 'accommodation', 'accommodations', both, _ ).
noun( 'accompaniment', 'accompaniments', count, _ ).
noun( 'accompanist', 'accompanists', count, _ ).
noun( 'accomplice', 'accomplices', count, _ ).
noun( 'accomplishment', 'accomplishments', both, _ ).
noun( 'accord', 'accords', both, _ ).
noun( 'accordance', 'accordances', count, _ ).
noun( 'accordion', 'accordions', count, _ ).
noun( 'accouchement', 'accouchements', count, _ ).
noun( 'account', 'accounts', both, _ ).
noun( 'accountancy', '-', mass, _ ).
noun( 'accountant', 'accountants', count, _ ).
noun( 'accretion', 'accretions', both, _ ).
noun( 'accumulation', 'accumulations', both, _ ).
noun( 'accumulator', 'accumulators', count, _ ).
noun( 'accuracy', 'accuracies', both, _ ).
noun( 'accusation', 'accusations', both, _ ).
noun( 'accusative', 'accusatives', count, _ ).
noun( 'accuser', 'accusers', count, _ ).
noun( 'ace', 'aces', count, _ ).
noun( 'acerbity', 'acerbities', both, _ ).
noun( 'acetate', 'acetates', count, _ ).
noun( 'acetylene', '-', mass, _ ).
noun( 'ache', 'aches', count, _ ).
noun( 'achievement', 'achievements', both, _ ).
noun( 'acid', 'acids', both, _ ).
noun( 'acidity', '-', mass, _ ).
noun( 'ack-ack', '-', mass, _ ).
noun( 'acknowledgement', 'acknowledgements', count, _ ).
noun( 'acme', '-', count, _ ).
noun( 'acne', '-', mass, _ ).
noun( 'acolyte', 'acolytes', count, _ ).
noun( 'aconite', 'aconites', count, _ ).
noun( 'acorn', 'acorns', count, _ ).
noun( 'acorn-cup', 'acorn-cups', count, _ ).
noun( 'acoustic', 'acoustics', count, _ ).
noun( 'acoustics', 'acoustics', mass, _ ).
noun( 'acquaintance', 'acquaintances', both, _ ).
noun( 'acquaintanceship', 'acquaintanceships', count, _ ).
noun( 'acquiescence', 'acquiescences', count, _ ).
noun( 'acquirement', 'acquirements', both, _ ).
noun( 'acquisition', 'acquisitions', both, _ ).
noun( 'acquittal', 'acquittals', both, _ ).
noun( 'acre', 'acres', count, _ ).
noun( 'acreage', '-', mass, _ ).
noun( 'acrimony', '-', mass, _ ).
noun( 'acrobat', 'acrobats', count, _ ).
noun( 'acrobatics', 'acrobatics', mass, _ ).
noun( 'acronym', 'acronyms', count, _ ).
noun( 'acropolis', 'acropolises', count, _ ).
noun( 'acrostic', 'acrostics', count, _ ).
noun( 'acrylic', 'acrylics', count, _ ).
noun( 'act', 'acts', count, _ ).
noun( 'acting', '-', mass, _ ).
noun( 'actinism', '-', mass, _ ).
noun( 'action', 'actions', both, _ ).
noun( 'activation', 'activations', count, _ ).
noun( 'activist', 'activists', count, _ ).
noun( 'activity', 'activities', both, _ ).
noun( 'actor', 'actors', count, _ ).
noun( 'actress', 'actresses', count, _ ).
noun( 'actuality', 'actualities', both, _ ).
noun( 'actuary', 'actuaries', count, _ ).
noun( 'acuity', '-', mass, _ ).
noun( 'acumen', '-', mass, _ ).
noun( 'acupuncture', '-', mass, _ ).
noun( 'acuteness', '-', mass, _ ).
noun( 'ad', 'ads', count, _ ).
noun( 'ad-man', 'ad-men', count, _ ).
noun( 'adage', 'adages', count, _ ).
noun( 'adagio', 'adagios', count, _ ).
noun( 'adamant', 'adamants', count, _ ).
noun( 'adaptability', '-', mass, _ ).
noun( 'adaptation', 'adaptations', both, _ ).
noun( 'adapter', 'adapters', count, _ ).
noun( 'adaptor', 'adaptors', count, _ ).
noun( 'addendum', 'addenda', count, _ ).
noun( 'adder', 'adders', count, _ ).
noun( 'addict', 'addicts', count, _ ).
noun( 'addiction', 'addictions', both, _ ).
noun( 'adding-machine', 'adding-machines', count, _ ).
noun( 'addition', 'additions', both, _ ).
noun( 'additive', 'additives', count, _ ).
noun( 'addle-head', 'addle-heads', count, _ ).
noun( 'address', 'addresses', both, _ ).
noun( 'addressee', 'addressees', count, _ ).
noun( 'adept', 'adepts', count, _ ).
noun( 'adequacy', '-', mass, _ ).
noun( 'adherence', 'adherences', count, _ ).
noun( 'adherent', 'adherents', count, _ ).
noun( 'adhesion', 'adhesions', both, _ ).
noun( 'adhesive', 'adhesives', both, _ ).
noun( 'adieu', 'adieus', count, _ ).
noun( 'adjective', 'adjectives', count, _ ).
noun( 'adjournment', 'adjournments', count, _ ).
noun( 'adjudication', 'adjudications', count, _ ).
noun( 'adjudicator', 'adjudicators', count, _ ).
noun( 'adjunct', 'adjuncts', count, _ ).
noun( 'adjuration', 'adjurations', both, _ ).
noun( 'adjuster', 'adjusters', count, _ ).
noun( 'adjustment', 'adjustments', both, _ ).
noun( 'adjutant', 'adjutants', count, _ ).
noun( 'admass', '-', mass, _ ).
noun( 'administration', 'administrations', both, _ ).
noun( 'administrator', 'administrators', count, _ ).
noun( 'admiral', 'admirals', count, _ ).
noun( 'admiralty', 'admiralties', count, _ ).
noun( 'admiration', '-', mass, _ ).
noun( 'admirer', 'admirers', count, _ ).
noun( 'admissibility', '-', mass, _ ).
noun( 'admission', 'admissions', both, _ ).
noun( 'admittance', '-', mass, _ ).
noun( 'admixture', 'admixtures', count, _ ).
noun( 'admonition', 'admonitions', both, _ ).
noun( 'ado', '-', mass, _ ).
noun( 'adobe', '-', mass, _ ).
noun( 'adolescence', '-', mass, _ ).
noun( 'adolescent', 'adolescents', count, _ ).
noun( 'adoption', 'adoptions', count, _ ).
noun( 'adoration', '-', mass, _ ).
noun( 'adorer', 'adorers', count, _ ).
noun( 'adornment', 'adornments', both, _ ).
noun( 'adrenalin', '-', mass, _ ).
noun( 'adroitness', '-', mass, _ ).
noun( 'adulation', '-', mass, _ ).
noun( 'adult', 'adults', count, _ ).
noun( 'adulterant', 'adulterants', count, _ ).
noun( 'adulteration', 'adulterations', count, _ ).
noun( 'adulterer', 'adulterers', count, _ ).
noun( 'adulteress', 'adulteresses', count, _ ).
noun( 'adultery', 'adulteries', both, _ ).
noun( 'adulthood', 'adulthoods', both, _ ).
noun( 'advance', 'advances', both, _ ).
noun( 'advancement', '-', mass, _ ).
noun( 'advantage', 'advantages', both, _ ).
noun( 'advent', 'advents', count, _ ).
noun( 'adventure', 'adventures', both, _ ).
noun( 'adventurer', 'adventurers', count, _ ).
noun( 'adventuress', 'adventuresses', count, _ ).
noun( 'adverb', 'adverbs', count, _ ).
noun( 'adversary', 'adversaries', count, _ ).
noun( 'adversity', 'adversities', both, _ ).
noun( 'advert', 'adverts', count, _ ).
noun( 'advertisement', 'advertisements', both, _ ).
noun( 'advertiser', 'advertisers', count, _ ).
noun( 'advice', 'advices', both, _ ).
noun( 'advisability', '-', mass, _ ).
noun( 'adviser', 'advisers', count, _ ).
noun( 'advocacy', '-', mass, _ ).
noun( 'advocate', 'advocates', count, _ ).
noun( 'advowson', 'advowsons', count, _ ).
noun( 'adz', 'adzes', count, _ ).
noun( 'adze', 'adzes', count, _ ).
noun( 'aegis', 'aegises', count, _ ).
noun( 'aeon', 'aeons', count, _ ).
noun( 'aeration', 'aerations', count, _ ).
noun( 'aerial', 'aerials', count, _ ).
noun( 'aerie', 'aeries', count, _ ).
noun( 'aerobatics', 'aerobatics', mass, _ ).
noun( 'aerodrome', 'aerodromes', count, _ ).
noun( 'aerodynamics', 'aerodynamics', mass, _ ).
noun( 'aeronaut', 'aeronauts', count, _ ).
noun( 'aeronautics', 'aeronautics', mass, _ ).
noun( 'aeroplane', 'aeroplanes', count, _ ).
noun( 'aerosol', 'aerosols', both, _ ).
noun( 'aerospace', '-', mass, _ ).
noun( 'aertex', '-', mass, _ ).
noun( 'aery', 'aeries', count, _ ).
noun( 'aesthete', 'aesthetes', count, _ ).
noun( 'aesthetic', '-', mass, _ ).
noun( 'aesthetics', 'aesthetics', mass, _ ).
noun( 'aether', '-', mass, _ ).
noun( 'aetiology', 'aetiologies', count, _ ).
noun( 'affability', '-', mass, _ ).
noun( 'affair', 'affairs', count, _ ).
noun( 'affectation', 'affectations', both, _ ).
noun( 'affection', 'affections', both, _ ).
noun( 'affidavit', 'affidavits', count, _ ).
noun( 'affiliation', 'affiliations', both, _ ).
noun( 'affinity', 'affinities', both, _ ).
noun( 'affirmation', 'affirmations', both, _ ).
noun( 'affirmative', 'affirmatives', count, _ ).
noun( 'affix', 'affixes', count, _ ).
noun( 'afflatus', '-', mass, _ ).
noun( 'affliction', 'afflictions', both, _ ).
noun( 'affluence', '-', mass, _ ).
noun( 'affluent', 'affluents', count, _ ).
noun( 'afforestation', 'afforestations', count, _ ).
noun( 'affray', 'affrays', count, _ ).
noun( 'affront', 'affronts', count, _ ).
noun( 'aftercare', '-', mass, _ ).
noun( 'afterdamp', '-', mass, _ ).
noun( 'aftereffect', 'aftereffects', count, _ ).
noun( 'afterglow', '-', count, _ ).
noun( 'aftermath', '', count, _ ).
noun( 'afternoon', 'afternoons', both, _ ).
noun( 'afterthought', 'afterthoughts', both, _ ).
noun( 'agar-agar', '-', mass, _ ).
noun( 'agate', 'agates', count, _ ).
noun( 'agave', 'agaves', count, _ ).
noun( 'age', 'ages', both, _ ).
noun( 'age-bracket', 'age-brackets', count, _ ).
noun( 'age-group', 'age-groups', count, _ ).
noun( 'ageing', '-', mass, _ ).
noun( 'agency', 'agencies', both, _ ).
noun( 'agenda', 'agendas', count, _ ).
noun( 'agent', 'agents', count, _ ).
noun( 'agent provocateur', 'agents provocateurs', count, _ ).
noun( 'agglomeration', 'agglomerations', both, _ ).
noun( 'aggrandizement', 'aggrandizements', count, _ ).
noun( 'aggravation', 'aggravations', both, _ ).
noun( 'aggregate', 'aggregates', count, _ ).
noun( 'aggregation', 'aggregations', both, _ ).
noun( 'aggression', 'aggressions', both, _ ).
noun( 'aggressiveness', '-', mass, _ ).
noun( 'aggressor', 'aggressors', count, _ ).
noun( 'aggro', '-', mass, _ ).
noun( 'agility', '-', mass, _ ).
noun( 'aging', '-', mass, _ ).
noun( 'agitation', 'agitations', both, _ ).
noun( 'agitator', 'agitators', count, _ ).
noun( 'agnail', '-', mass, _ ).
noun( 'agnostic', 'agnostics', count, _ ).
noun( 'agnosticism', '-', mass, _ ).
noun( 'agony', 'agonies', both, _ ).
noun( 'agora', 'agoras', count, _ ).
noun( 'agoraphobia', '-', mass, _ ).
noun( 'agreement', 'agreements', both, _ ).
noun( 'agriculture', '-', mass, _ ).
noun( 'ague', 'agues', count, _ ).
noun( 'aid', 'aids', both, _ ).
noun( 'aide-de-camp', 'aides-de-camp', count, _ ).
noun( 'aide-m_emoire', 'aide-m_emoires', count, _ ).
noun( 'aigret', 'aigrets', count, _ ).
noun( 'aigrette', 'aigrettes', count, _ ).
noun( 'aileron', 'ailerons', count, _ ).
noun( 'ailment', 'ailments', count, _ ).
noun( 'aim', 'aims', both, _ ).
noun( 'air', 'airs', both, _ ).
noun( 'air-bladder', 'air-bladders', count, _ ).
noun( 'air-conditioning', '-', mass, _ ).
noun( 'air-pump', 'air-pumps', count, _ ).
noun( 'air-raid', 'air-raids', count, _ ).
noun( 'air-shaft', 'air-shafts', count, _ ).
noun( 'air-sickness', '-', mass, _ ).
noun( 'airbed', 'airbeds', count, _ ).
noun( 'airbrake', 'airbrakes', count, _ ).
noun( 'aircraft', 'aircraft', count, _ ).
noun( 'aircraftman', 'aircraftmen', count, _ ).
noun( 'aircrew', 'aircrews', count, _ ).
noun( 'airdrome', 'airdromes', count, _ ).
noun( 'airfield', 'airfields', count, _ ).
noun( 'airflow', 'airflows', count, _ ).
noun( 'airframe', 'airframes', count, _ ).
noun( 'airing', 'airings', count, _ ).
noun( 'airing-cupboard', 'airing-cupboards', count, _ ).
noun( 'airline', 'airlines', count, _ ).
noun( 'airliner', 'airliners', count, _ ).
noun( 'airmail', '-', mass, _ ).
noun( 'airman', 'airmen', count, _ ).
noun( 'airplane', 'airplanes', count, _ ).
noun( 'airport', 'airports', count, _ ).
noun( 'airscrew', 'airscrews', count, _ ).
noun( 'airship', 'airships', count, _ ).
noun( 'airstrip', 'airstrips', count, _ ).
noun( 'airway', 'airways', count, _ ).
noun( 'airwoman', 'airwomen', count, _ ).
noun( 'airworthiness', '-', mass, _ ).
noun( 'aisle', 'aisles', count, _ ).
noun( 'aitch', 'aitches', count, _ ).
noun( 'aitch-bone', 'aitch-bones', count, _ ).
noun( 'alabaster', '-', mass, _ ).
noun( 'alacrity', '-', mass, _ ).
noun( 'alarm', 'alarms', both, _ ).
noun( 'alarm-clock', 'alarm-clocks', count, _ ).
noun( 'alarmist', 'alarmists', count, _ ).
noun( 'alb', 'albs', count, _ ).
noun( 'albatross', 'albatrosses', count, _ ).
noun( 'albino', 'albinos', count, _ ).
noun( 'album', 'albums', count, _ ).
noun( 'albumen', '-', mass, _ ).
noun( 'alchemist', 'alchemists', count, _ ).
noun( 'alchemy', '-', mass, _ ).
noun( 'alcohol', 'alcohols', both, _ ).
noun( 'alcoholic', 'alcoholics', count, _ ).
noun( 'alcoholism', '-', mass, _ ).
noun( 'alcove', 'alcoves', count, _ ).
noun( 'alder', 'alders', count, _ ).
noun( 'alderman', 'aldermen', count, _ ).
noun( 'ale', 'ales', both, _ ).
noun( 'ale-house', 'ale-houses', count, _ ).
noun( 'alert', 'alerts', count, _ ).
noun( 'alertness', '-', mass, _ ).
noun( 'alexandrine', 'alexandrines', count, _ ).
noun( 'alexia', '-', mass, _ ).
noun( 'alexic', 'alexics', count, _ ).
noun( 'alfalfa', '-', mass, _ ).
noun( 'alga', 'algae', count, _ ).
noun( 'algebra', 'algebras', both, _ ).
noun( 'alias', 'aliases', count, _ ).
noun( 'alibi', 'alibis', count, _ ).
noun( 'alien', 'aliens', count, _ ).
noun( 'alienation', '-', mass, _ ).
noun( 'alienist', 'alienists', count, _ ).
noun( 'alignment', 'alignments', both, _ ).
noun( 'alimony', '-', mass, _ ).
noun( 'alkali', 'alkalis', both, _ ).
noun( 'all', '-', count, _ ).
noun( 'all-rounder', 'all-rounders', count, _ ).
noun( 'allegation', 'allegations', both, _ ).
noun( 'allegiance', '-', mass, _ ).
noun( 'allegory', 'allegories', count, _ ).
noun( 'allegretto', 'allegrettos', count, _ ).
noun( 'allegro', 'allegros', count, _ ).
noun( 'allergen', 'allergens', count, _ ).
noun( 'allergy', 'allergies', count, _ ).
noun( 'alleviation', 'alleviations', count, _ ).
noun( 'alley', 'alleys', count, _ ).
noun( 'alleyway', 'alleyways', count, _ ).
noun( 'alliance', 'alliances', both, _ ).
noun( 'alligator', 'alligators', count, _ ).
noun( 'alliteration', '-', mass, _ ).
noun( 'allocation', 'allocations', both, _ ).
noun( 'allotment', 'allotments', both, _ ).
noun( 'allowance', 'allowances', count, _ ).
noun( 'alloy', 'alloys', both, _ ).
noun( 'allspice', '-', mass, _ ).
noun( 'allure', 'allures', both, _ ).
noun( 'allurement', 'allurements', both, _ ).
noun( 'allusion', 'allusions', count, _ ).
noun( 'ally', 'allies', count, _ ).
noun( 'almanac', 'almanacs', count, _ ).
noun( 'almighty', '-', count, _ ).
noun( 'almond', 'almonds', count, _ ).
noun( 'almoner', 'almoners', count, _ ).
noun( 'alms-box', 'alms-boxes', count, _ ).
noun( 'alms-giving', '-', mass, _ ).
noun( 'alms-house', 'alms-houses', count, _ ).
noun( 'aloe', 'aloes', count, _ ).
noun( 'aloofness', '-', mass, _ ).
noun( 'alp', 'alps', count, _ ).
noun( 'alpaca', 'alpacas', both, _ ).
noun( 'alpenstock', 'alpenstocks', count, _ ).
noun( 'alpha', 'alphas', count, _ ).
noun( 'alphabet', 'alphabets', count, _ ).
noun( 'alpinist', 'alpinists', count, _ ).
noun( 'alsatian', 'alsatians', count, _ ).
noun( 'also-ran', 'also-rans', count, _ ).
noun( 'altar', 'altars', count, _ ).
noun( 'altar-piece', 'altar-pieces', count, _ ).
noun( 'alter ego', 'alter egos', count, _ ).
noun( 'alteration', 'alterations', both, _ ).
noun( 'altercation', 'altercations', both, _ ).
noun( 'alternation', 'alternations', count, _ ).
noun( 'alternative', 'alternatives', count, _ ).
noun( 'altimeter', 'altimeters', count, _ ).
noun( 'altitude', 'altitudes', count, _ ).
noun( 'alto', 'altos', count, _ ).
noun( 'altruism', 'altruisms', both, _ ).
noun( 'altruist', 'altruists', count, _ ).
noun( 'alum', '-', mass, _ ).
noun( 'aluminium', '-', mass, _ ).
noun( 'alumna', 'alumnae', count, _ ).
noun( 'alumnus', 'alumni', count, _ ).
noun( 'alveolar', 'alveolars', count, _ ).
noun( 'am', '-', proper, _ ).
noun( 'amah', 'amahs', count, _ ).
noun( 'amalgam', 'amalgams', count, _ ).
noun( 'amalgamation', 'amalgamations', both, _ ).
noun( 'amanuensis', 'amanuenses', count, _ ).
noun( 'amaryllis', 'amaryllises', count, _ ).
noun( 'amateur', 'amateurs', count, _ ).
noun( 'amateurism', '-', mass, _ ).
noun( 'amazement', '-', mass, _ ).
noun( 'ambassador', 'ambassadors', count, _ ).
noun( 'ambassadress', 'ambassadresses', count, _ ).
noun( 'amber', '-', mass, _ ).
noun( 'ambergris', '-', mass, _ ).
noun( 'ambience', '-', count, _ ).
noun( 'ambiguity', 'ambiguities', both, _ ).
noun( 'ambit', 'ambits', count, _ ).
noun( 'ambition', 'ambitions', both, _ ).
noun( 'ambivalence', '-', mass, _ ).
noun( 'amble', 'ambles', count, _ ).
noun( 'ambrosia', '-', mass, _ ).
noun( 'ambulance', 'ambulances', count, _ ).
noun( 'ambuscade', 'ambuscades', count, _ ).
noun( 'ambush', 'ambushes', both, _ ).
noun( 'ameba', 'amebas', count, _ ).
noun( 'ameer', 'ameers', count, _ ).
noun( 'amelioration', 'ameliorations', count, _ ).
noun( 'amendment', 'amendments', both, _ ).
noun( 'amenity', 'amenities', count, _ ).
noun( 'amethyst', 'amethysts', count, _ ).
noun( 'amiability', '-', mass, _ ).
noun( 'amicability', '-', mass, _ ).
noun( 'amir', 'amirs', count, _ ).
noun( 'amity', '-', mass, _ ).
noun( 'ammeter', 'ammeters', count, _ ).
noun( 'ammonia', '-', mass, _ ).
noun( 'ammonite', 'ammonites', count, _ ).
noun( 'ammunition', '-', mass, _ ).
noun( 'amnesia', '-', mass, _ ).
noun( 'amnesty', 'amnesties', count, _ ).
noun( 'amoeba', 'amoebas', count, _ ).
noun( 'amortization', 'amortizations', count, _ ).
noun( 'amount', 'amounts', count, _ ).
noun( 'amour', 'amours', count, _ ).
noun( 'amour-propre', '-', mass, _ ).
noun( 'amp', 'amps', count, _ ).
noun( 'ampere', 'amperes', count, _ ).
noun( 'amphetamine', 'amphetamines', both, _ ).
noun( 'amphibian', 'amphibians', count, _ ).
noun( 'amphitheatre', 'amphitheatres', count, _ ).
noun( 'amphora', 'amphoras', count, _ ).
noun( 'amplification', 'amplifications', count, _ ).
noun( 'amplifier', 'amplifiers', count, _ ).
noun( 'amplitude', '-', mass, _ ).
noun( 'ampoule', 'ampoules', count, _ ).
noun( 'amputation', 'amputations', count, _ ).
noun( 'amulet', 'amulets', count, _ ).
noun( 'amusement', 'amusements', both, _ ).
noun( 'anachronism', 'anachronisms', count, _ ).
noun( 'anaconda', 'anacondas', count, _ ).
noun( 'anaemia', '-', mass, _ ).
noun( 'anaesthesia', '-', mass, _ ).
noun( 'anaesthetic', 'anaesthetics', count, _ ).
noun( 'anaesthetist', 'anaesthetists', count, _ ).
noun( 'anagram', 'anagrams', count, _ ).
noun( 'analgesia', '-', mass, _ ).
noun( 'analgesic', 'analgesics', count, _ ).
noun( 'analog', 'analogs', count, _ ).
noun( 'analogue', 'analogues', count, _ ).
noun( 'analogy', 'analogies', both, _ ).
noun( 'analysis', 'analyses', both, _ ).
noun( 'analyst', 'analysts', count, _ ).
noun( 'anapaest', 'anapaests', count, _ ).
noun( 'anarchism', '-', mass, _ ).
noun( 'anarchist', 'anarchists', count, _ ).
noun( 'anarchy', '-', mass, _ ).
noun( 'anathema', 'anathemas', count, _ ).
noun( 'anatomist', 'anatomists', count, _ ).
noun( 'anatomy', '-', mass, _ ).
noun( 'ancestor', 'ancestors', count, _ ).
noun( 'ancestress', 'ancestresses', count, _ ).
noun( 'ancestry', 'ancestries', count, _ ).
noun( 'anchor', 'anchors', count, _ ).
noun( 'anchorage', 'anchorages', count, _ ).
noun( 'anchorite', 'anchorites', count, _ ).
noun( 'anchorman', 'anchormen', count, _ ).
noun( 'anchovy', 'anchovies', count, _ ).
noun( 'andante', 'andantes', count, _ ).
noun( 'andiron', 'andirons', count, _ ).
noun( 'anecdote', 'anecdotes', count, _ ).
noun( 'anemometer', 'anemometers', count, _ ).
noun( 'anemone', 'anemones', count, _ ).
noun( 'aneroid', 'aneroids', count, _ ).
noun( 'anesthetic', 'anesthetics', count, _ ).
noun( 'anesthetist', 'anesthetists', count, _ ).
noun( 'angel', 'angels', count, _ ).
noun( 'angelica', '-', mass, _ ).
noun( 'angelus', 'angeluses', count, _ ).
noun( 'anger', '-', mass, _ ).
noun( 'angina', '-', mass, _ ).
noun( 'angle', 'angles', count, _ ).
noun( 'angle-dozer', 'angle-dozers', count, _ ).
noun( 'angle-iron', 'angle-irons', count, _ ).
noun( 'angler', 'anglers', count, _ ).
noun( 'anglicism', 'anglicisms', count, _ ).
noun( 'angling', '-', mass, _ ).
noun( 'angora', 'angoras', both, _ ).
noun( 'angostura', '-', mass, _ ).
noun( 'angst', '-', mass, _ ).
noun( 'anguish', '-', mass, _ ).
noun( 'angularity', 'angularities', both, _ ).
noun( 'aniline', 'anilines', both, _ ).
noun( 'animadversion', 'animadversions', count, _ ).
noun( 'animal', 'animals', count, _ ).
noun( 'animalcule', 'animalcules', count, _ ).
noun( 'animation', '-', mass, _ ).
noun( 'animism', '-', mass, _ ).
noun( 'animosity', 'animosities', both, _ ).
noun( 'animus', '-', mass, _ ).
noun( 'anise', 'anises', count, _ ).
noun( 'aniseed', '-', mass, _ ).
noun( 'ankle', 'ankles', count, _ ).
noun( 'anklet', 'anklets', count, _ ).
noun( 'anna', 'annas', count, _ ).
noun( 'annalist', 'annalists', count, _ ).
noun( 'annex', 'annexes', count, _ ).
noun( 'annexation', '-', mass, _ ).
noun( 'annexe', 'annexes', count, _ ).
noun( 'annihilation', '-', mass, _ ).
noun( 'anniversary', 'anniversaries', count, _ ).
noun( 'annotation', 'annotations', both, _ ).
noun( 'announcement', 'announcements', count, _ ).
noun( 'announcer', 'announcers', count, _ ).
noun( 'annoyance', 'annoyances', both, _ ).
noun( 'annual', 'annuals', count, _ ).
noun( 'annuitant', 'annuitants', count, _ ).
noun( 'annuity', 'annuities', count, _ ).
noun( 'annulment', 'annulments', count, _ ).
noun( 'annunciation', 'annunciations', count, _ ).
noun( 'anode', 'anodes', count, _ ).
noun( 'anodyne', 'anodynes', both, _ ).
noun( 'anointment', 'anointments', count, _ ).
noun( 'anomaly', 'anomalies', count, _ ).
noun( 'anonymity', '-', mass, _ ).
noun( 'anopheles', 'anopheles', count, _ ).
noun( 'anorak', 'anoraks', count, _ ).
noun( 'answer', 'answers', count, _ ).
noun( 'ant', 'ants', count, _ ).
noun( 'ant-eater', 'ant-eaters', count, _ ).
noun( 'ant-hill', 'ant-hills', count, _ ).
noun( 'antagonism', 'antagonisms', both, _ ).
noun( 'antagonist', 'antagonists', count, _ ).
noun( 'ante', 'antes', count, _ ).
noun( 'antecedence', 'antecedences', count, _ ).
noun( 'antecedent', 'antecedents', count, _ ).
noun( 'antechamber', 'antechambers', count, _ ).
noun( 'antediluvian', 'antediluvians', count, _ ).
noun( 'antelope', 'antelopes', count, _ ).
noun( 'antenna', 'antennae', count, _ ).
noun( 'anteroom', 'anterooms', count, _ ).
noun( 'anthem', 'anthems', count, _ ).
noun( 'anther', 'anthers', count, _ ).
noun( 'anthology', 'anthologies', count, _ ).
noun( 'anthracite', '-', mass, _ ).
noun( 'anthrax', '-', mass, _ ).
noun( 'anthropoid', 'anthropoids', count, _ ).
noun( 'anthropologist', 'anthropologists', count, _ ).
noun( 'anthropology', '-', mass, _ ).
noun( 'anti-semite', 'anti-semites', count, _ ).
noun( 'anti-semitism', '-', mass, _ ).
noun( 'anti-hero', 'anti-heroes', count, _ ).
noun( 'antibiotic', 'antibiotics', count, _ ).
noun( 'antibody', 'antibodies', count, _ ).
noun( 'antic', 'antics', count, _ ).
noun( 'anticipation', 'anticipations', both, _ ).
noun( 'anticlimax', 'anticlimaxes', count, _ ).
noun( 'anticyclone', 'anticyclones', count, _ ).
noun( 'antidote', 'antidotes', count, _ ).
noun( 'antifreeze', '-', mass, _ ).
noun( 'antiknock', '-', mass, _ ).
noun( 'antilogarithm', 'antilogarithms', count, _ ).
noun( 'antimacassar', 'antimacassars', count, _ ).
noun( 'antimony', '-', mass, _ ).
noun( 'antipathy', 'antipathies', both, _ ).
noun( 'antiquarian', 'antiquarians', count, _ ).
noun( 'antiquary', 'antiquaries', count, _ ).
noun( 'antique', 'antiques', count, _ ).
noun( 'antiquity', 'antiquities', both, _ ).
noun( 'antirrhinum', 'antirrhinums', count, _ ).
noun( 'antiseptic', 'antiseptics', count, _ ).
noun( 'antithesis', 'antitheses', both, _ ).
noun( 'antitoxin', 'antitoxins', count, _ ).
noun( 'antitrade', 'antitrades', count, _ ).
noun( 'antler', 'antlers', count, _ ).
noun( 'antonym', 'antonyms', count, _ ).
noun( 'anus', 'anuses', count, _ ).
noun( 'anvil', 'anvils', count, _ ).
noun( 'anxiety', 'anxieties', both, _ ).
noun( 'anybody', '-', count, _ ).
noun( 'anyone', '-', count, _ ).
noun( 'anything', '-', count, _ ).
noun( 'aorta', 'aortas', count, _ ).
noun( 'apache', 'apaches', count, _ ).
noun( 'apanage', '-', mass, _ ).
noun( 'apartheid', '-', mass, _ ).
noun( 'apartment', 'apartments', count, _ ).
noun( 'apathy', '-', mass, _ ).
noun( 'ape', 'apes', count, _ ).
noun( 'aperient', 'aperients', count, _ ).
noun( 'aperitif', 'aperitifs', count, _ ).
noun( 'aperture', 'apertures', count, _ ).
noun( 'apex', 'apexes', count, _ ).
noun( 'aphasia', '-', mass, _ ).
noun( 'aphid', 'aphids', count, _ ).
noun( 'aphis', 'aphides', count, _ ).
noun( 'aphorism', 'aphorisms', count, _ ).
noun( 'aphrodisiac', 'aphrodisiacs', both, _ ).
noun( 'apiarist', 'apiarists', count, _ ).
noun( 'apiary', 'apiaries', count, _ ).
noun( 'apiculture', 'apicultures', count, _ ).
noun( 'aplomb', '-', mass, _ ).
noun( 'apocalypse', 'apocalypses', count, _ ).
noun( 'apogee', 'apogees', count, _ ).
noun( 'apologetics', 'apologetics', mass, _ ).
noun( 'apologist', 'apologists', count, _ ).
noun( 'apology', 'apologies', count, _ ).
noun( 'apophthegm', 'apophthegms', count, _ ).
noun( 'apoplexy', '-', mass, _ ).
noun( 'apostasy', 'apostasies', both, _ ).
noun( 'apostate', 'apostates', count, _ ).
noun( 'apostle', 'apostles', count, _ ).
noun( 'apostrophe', 'apostrophes', count, _ ).
noun( 'apothecary', 'apothecaries', count, _ ).
noun( 'apothegm', 'apothegms', count, _ ).
noun( 'apotheosis', 'apotheoses', count, _ ).
noun( 'appanage', '-', mass, _ ).
noun( 'apparatus', 'apparatuses', both, _ ).
noun( 'apparel', '-', mass, _ ).
noun( 'apparition', 'apparitions', count, _ ).
noun( 'appeal', 'appeals', both, _ ).
noun( 'appearance', 'appearances', count, _ ).
noun( 'appeasement', '-', mass, _ ).
noun( 'appellant', 'appellants', count, _ ).
noun( 'appellation', 'appellations', count, _ ).
noun( 'appendage', 'appendages', count, _ ).
noun( 'appendectomy', 'appendectomies', count, _ ).
noun( 'appendicitis', '-', mass, _ ).
noun( 'appendix', 'appendixes', count, _ ).
noun( 'appetite', 'appetites', both, _ ).
noun( 'appetizer', 'appetizers', count, _ ).
noun( 'applause', '-', mass, _ ).
noun( 'apple', 'apples', both, _ ).
noun( 'applejack', '-', mass, _ ).
noun( 'appliance', 'appliances', count, _ ).
noun( 'applicability', '-', mass, _ ).
noun( 'applicant', 'applicants', count, _ ).
noun( 'application', 'applications', both, _ ).
noun( 'appliqu_e', '-', mass, _ ).
noun( 'appointee', 'appointees', count, _ ).
noun( 'appointment', 'appointments', both, _ ).
noun( 'apportionment', '-', mass, _ ).
noun( 'apposition', '-', mass, _ ).
noun( 'appraisal', 'appraisals', count, _ ).
noun( 'appreciation', 'appreciations', both, _ ).
noun( 'apprehension', 'apprehensions', both, _ ).
noun( 'apprentice', 'apprentices', count, _ ).
noun( 'apprenticeship', 'apprenticeships', count, _ ).
noun( 'appro', '-', mass, _ ).
noun( 'approach', 'approaches', both, _ ).
noun( 'approbation', '-', mass, _ ).
noun( 'appropriation', 'appropriations', both, _ ).
noun( 'approval', '-', mass, _ ).
noun( 'approx', '-', proper, _ ).
noun( 'approximation', 'approximations', both, _ ).
noun( 'appurtenance', 'appurtenances', count, _ ).
noun( 'apricot', 'apricots', count, _ ).
noun( 'apron', 'aprons', count, _ ).
noun( 'apse', 'apses', count, _ ).
noun( 'aptitude', 'aptitudes', both, _ ).
noun( 'aptness', '-', mass, _ ).
noun( 'aqualung', 'aqualungs', count, _ ).
noun( 'aquamarine', 'aquamarines', both, _ ).
noun( 'aquanaut', 'aquanauts', count, _ ).
noun( 'aquaplane', 'aquaplanes', count, _ ).
noun( 'aquarium', 'aquariums', count, _ ).
noun( 'aquatint', 'aquatints', both, _ ).
noun( 'aquavit', '-', mass, _ ).
noun( 'aqueduct', 'aqueducts', count, _ ).
noun( 'ar^ete', 'ar^etes', count, _ ).
noun( 'arabesque', 'arabesques', count, _ ).
noun( 'arachnid', 'arachnids', count, _ ).
noun( 'arbiter', 'arbiters', count, _ ).
noun( 'arbitrament', 'arbitraments', both, _ ).
noun( 'arbitration', 'arbitrations', both, _ ).
noun( 'arbitrator', 'arbitrators', count, _ ).
noun( 'arbour', 'arbours', count, _ ).
noun( 'arc', 'arcs', count, _ ).
noun( 'arc-lamp', 'arc-lamps', count, _ ).
noun( 'arc-light', 'arc-lights', count, _ ).
noun( 'arcade', 'arcades', count, _ ).
noun( 'arch', 'arches', count, _ ).
noun( 'archaeologist', 'archaeologists', count, _ ).
noun( 'archaeology', '-', mass, _ ).
noun( 'archaism', 'archaisms', both, _ ).
noun( 'archangel', 'archangels', count, _ ).
noun( 'archbishop', 'archbishops', count, _ ).
noun( 'archbishopric', 'archbishoprics', count, _ ).
noun( 'archdeacon', 'archdeacons', count, _ ).
noun( 'archdeaconry', 'archdeaconries', count, _ ).
noun( 'archdiocese', 'archdioceses', count, _ ).
noun( 'archduke', 'archdukes', count, _ ).
noun( 'archeology', '-', mass, _ ).
noun( 'archer', 'archers', count, _ ).
noun( 'archery', '-', mass, _ ).
noun( 'archetype', 'archetypes', count, _ ).
noun( 'archimandrite', 'archimandrites', count, _ ).
noun( 'archipelago', 'archipelagos', count, _ ).
noun( 'architect', 'architects', count, _ ).
noun( 'architecture', '-', mass, _ ).
noun( 'archivist', 'archivists', count, _ ).
noun( 'archway', 'archways', count, _ ).
noun( 'ardour', 'ardours', both, _ ).
noun( 'are', 'ares', count, _ ).
noun( 'area', 'areas', both, _ ).
noun( 'areca', 'arecas', count, _ ).
noun( 'arena', 'arenas', count, _ ).
noun( 'argent', 'argents', count, _ ).
noun( 'argon', '-', mass, _ ).
noun( 'argosy', 'argosies', count, _ ).
noun( 'argot', '-', mass, _ ).
noun( 'argument', 'arguments', both, _ ).
noun( 'argumentation', '-', mass, _ ).
noun( 'aria', 'arias', count, _ ).
noun( 'aridity', '-', mass, _ ).
noun( 'aristocracy', 'aristocracies', both, _ ).
noun( 'aristocrat', 'aristocrats', count, _ ).
noun( 'arithmetic', '-', mass, _ ).
noun( 'arithmetician', 'arithmeticians', count, _ ).
noun( 'ark', 'arks', count, _ ).
noun( 'arm', 'arms', count, _ ).
noun( 'arm-hole', 'arm-holes', count, _ ).
noun( 'armada', 'armadas', count, _ ).
noun( 'armadillo', 'armadillos', count, _ ).
noun( 'armament', 'armaments', both, _ ).
noun( 'armature', 'armatures', count, _ ).
noun( 'armband', 'armbands', count, _ ).
noun( 'armchair', 'armchairs', count, _ ).
noun( 'armful', 'armfuls', count, _ ).
noun( 'armistice', 'armistices', count, _ ).
noun( 'armlet', 'armlets', count, _ ).
noun( 'armoire', 'armoires', count, _ ).
noun( 'armour', '-', mass, _ ).
noun( 'armour-plate', 'armour-plates', count, _ ).
noun( 'armourer', 'armourers', count, _ ).
noun( 'armoury', 'armouries', count, _ ).
noun( 'armpit', 'armpits', count, _ ).
noun( 'arms-race', '-', count, _ ).
noun( 'arms-runner', 'arms-runners', count, _ ).
noun( 'army', 'armies', count, _ ).
noun( 'arnica', '-', mass, _ ).
noun( 'aroma', 'aromas', count, _ ).
noun( 'arpeggio', 'arpeggios', count, _ ).
noun( 'arquebus', 'arquebuses', count, _ ).
noun( 'arr', '-', proper, _ ).
noun( 'arrack', '-', mass, _ ).
noun( 'arraignment', 'arraignments', count, _ ).
noun( 'arrangement', 'arrangements', both, _ ).
noun( 'arras', 'arrases', count, _ ).
noun( 'array', 'arrays', count, _ ).
noun( 'arrest', 'arrests', count, _ ).
noun( 'arrester', 'arresters', count, _ ).
noun( 'arri`ere pens_ee', 'arri`ere pens_ees', count, _ ).
noun( 'arrival', 'arrivals', both, _ ).
noun( 'arrogance', '-', mass, _ ).
noun( 'arrow', 'arrows', count, _ ).
noun( 'arrowhead', 'arrowheads', count, _ ).
noun( 'arrowroot', '-', mass, _ ).
noun( 'arse', 'arses', count, _ ).
noun( 'arsehole', 'arseholes', count, _ ).
noun( 'arsenal', 'arsenals', count, _ ).
noun( 'arsenic', '-', mass, _ ).
noun( 'arson', '-', mass, _ ).
noun( 'art', 'arts', both, _ ).
noun( 'artefact', 'artefacts', count, _ ).
noun( 'arteriosclerosis', '-', mass, _ ).
noun( 'artery', 'arteries', count, _ ).
noun( 'artfulness', '-', mass, _ ).
noun( 'arthritis', '-', mass, _ ).
noun( 'artichoke', 'artichokes', count, _ ).
noun( 'article', 'articles', count, _ ).
noun( 'articulation', '-', mass, _ ).
noun( 'artifact', 'artifacts', count, _ ).
noun( 'artifice', 'artifices', both, _ ).
noun( 'artificer', 'artificers', count, _ ).
noun( 'artillery', '-', mass, _ ).
noun( 'artisan', 'artisans', count, _ ).
noun( 'artist', 'artists', count, _ ).
noun( 'artiste', 'artistes', count, _ ).
noun( 'artistry', '-', mass, _ ).
noun( 'artlessness', '-', mass, _ ).
noun( 'arum', 'arums', count, _ ).
noun( 'asap', '-', proper, _ ).
noun( 'asbestos', '-', mass, _ ).
noun( 'ascendancy', '-', mass, _ ).
noun( 'ascendant', 'ascendants', count, _ ).
noun( 'ascendency', '-', mass, _ ).
noun( 'ascendent', 'ascendents', count, _ ).
noun( 'ascension', 'ascensions', count, _ ).
noun( 'ascent', 'ascents', count, _ ).
noun( 'ascetic', 'ascetics', count, _ ).
noun( 'asceticism', '-', mass, _ ).
noun( 'ascription', 'ascriptions', count, _ ).
noun( 'asdic', 'asdics', count, _ ).
noun( 'asepsis', '-', mass, _ ).
noun( 'asexuality', '-', mass, _ ).
noun( 'ash', 'ashes', both, _ ).
noun( 'ash-bin', 'ash-bins', count, _ ).
noun( 'ash-can', 'ash-cans', count, _ ).
noun( 'ash-key', 'ash-keys', count, _ ).
noun( 'ash-pan', 'ash-pans', count, _ ).
noun( 'ashtray', 'ashtrays', count, _ ).
noun( 'aside', 'asides', count, _ ).
noun( 'asking', '-', count, _ ).
noun( 'asp', 'asps', count, _ ).
noun( 'asparagus', '-', mass, _ ).
noun( 'aspect', 'aspects', count, _ ).
noun( 'aspen', 'aspens', count, _ ).
noun( 'asperity', 'asperities', both, _ ).
noun( 'aspersion', 'aspersions', count, _ ).
noun( 'asphalt', '-', mass, _ ).
noun( 'asphodel', 'asphodels', count, _ ).
noun( 'asphyxia', '-', mass, _ ).
noun( 'asphyxiation', '-', mass, _ ).
noun( 'aspic', '-', mass, _ ).
noun( 'aspidistra', 'aspidistras', count, _ ).
noun( 'aspirant', 'aspirants', count, _ ).
noun( 'aspirate', 'aspirates', count, _ ).
noun( 'aspiration', 'aspirations', both, _ ).
noun( 'aspirin', 'aspirins', both, _ ).
noun( 'ass', 'asses', count, _ ).
noun( 'assagai', 'assagais', count, _ ).
noun( 'assailant', 'assailants', count, _ ).
noun( 'assassin', 'assassins', count, _ ).
noun( 'assassination', 'assassinations', both, _ ).
noun( 'assault', 'assaults', count, _ ).
noun( 'assay', 'assays', count, _ ).
noun( 'assegai', 'assegais', count, _ ).
noun( 'assemblage', 'assemblages', both, _ ).
noun( 'assembly', 'assemblies', count, _ ).
noun( 'assent', 'assents', count, _ ).
noun( 'assertion', 'assertions', both, _ ).
noun( 'assessment', 'assessments', both, _ ).
noun( 'assessor', 'assessors', count, _ ).
noun( 'asset', 'assets', count, _ ).
noun( 'asseveration', 'asseverations', count, _ ).
noun( 'asshole', 'assholes', count, _ ).
noun( 'assiduity', 'assiduities', both, _ ).
noun( 'assignation', 'assignations', count, _ ).
noun( 'assignment', 'assignments', both, _ ).
noun( 'assimilation', '-', mass, _ ).
noun( 'assistance', '-', mass, _ ).
noun( 'assistant', 'assistants', count, _ ).
noun( 'assize', '-', mass, _ ).
noun( 'assoc', '-', count, _ ).
noun( 'associate', 'associates', count, _ ).
noun( 'association', 'associations', both, _ ).
noun( 'assonance', 'assonances', count, _ ).
noun( 'assortment', 'assortments', count, _ ).
noun( 'asst', '-', count, _ ).
noun( 'assumption', 'assumptions', count, _ ).
noun( 'assurance', 'assurances', both, _ ).
noun( 'assuredness', '-', mass, _ ).
noun( 'aster', 'asters', count, _ ).
noun( 'asterisk', 'asterisks', count, _ ).
noun( 'asteroid', 'asteroids', count, _ ).
noun( 'asthma', '-', mass, _ ).
noun( 'astigmatism', '-', mass, _ ).
noun( 'astonishment', '-', mass, _ ).
noun( 'astrakhan', '-', mass, _ ).
noun( 'astringency', '-', mass, _ ).
noun( 'astringent', 'astringents', count, _ ).
noun( 'astrodome', 'astrodomes', count, _ ).
noun( 'astrolabe', 'astrolabes', count, _ ).
noun( 'astrologer', 'astrologers', count, _ ).
noun( 'astrology', '-', mass, _ ).
noun( 'astronaut', 'astronauts', count, _ ).
noun( 'astronautics', 'astronautics', mass, _ ).
noun( 'astronomer', 'astronomers', count, _ ).
noun( 'astronomy', '-', mass, _ ).
noun( 'astrophysics', 'astrophysics', mass, _ ).
noun( 'astuteness', '-', mass, _ ).
noun( 'asylum', 'asylums', both, _ ).
noun( 'asymmetry', '-', mass, _ ).
noun( 'asymptote', 'asymptotes', count, _ ).
noun( 'at-home', 'at-homes', count, _ ).
noun( 'atabrine', '-', mass, _ ).
noun( 'atavism', 'atavisms', count, _ ).
noun( 'atelier', 'ateliers', count, _ ).
noun( 'atheism', '-', mass, _ ).
noun( 'atheist', 'atheists', count, _ ).
noun( 'athlete', 'athletes', count, _ ).
noun( 'athletics', 'athletics', mass, _ ).
noun( 'atlas', 'atlases', count, _ ).
noun( 'atmosphere', 'atmospheres', both, _ ).
noun( 'atmospherics', 'atmospherics', mass, _ ).
noun( 'atoll', 'atolls', count, _ ).
noun( 'atom', 'atoms', count, _ ).
noun( 'atomizer', 'atomizers', count, _ ).
noun( 'atonality', '-', mass, _ ).
noun( 'atonement', '-', mass, _ ).
noun( 'atrocity', 'atrocities', both, _ ).
noun( 'atrophy', '-', mass, _ ).
noun( 'attach_e', 'attach_es', count, _ ).
noun( 'attachment', 'attachments', both, _ ).
noun( 'attack', 'attacks', both, _ ).
noun( 'attacker', 'attackers', count, _ ).
noun( 'attainder', 'attainders', count, _ ).
noun( 'attainment', 'attainments', both, _ ).
noun( 'attar', '-', mass, _ ).
noun( 'attempt', 'attempts', count, _ ).
noun( 'attendance', 'attendances', both, _ ).
noun( 'attendant', 'attendants', count, _ ).
noun( 'attention', 'attentions', both, _ ).
noun( 'attentiveness', '-', mass, _ ).
noun( 'attenuation', '-', mass, _ ).
noun( 'attic', 'attics', count, _ ).
noun( 'attire', '-', mass, _ ).
noun( 'attitude', 'attitudes', count, _ ).
noun( 'attorney', 'attorneys', count, _ ).
noun( 'attraction', 'attractions', both, _ ).
noun( 'attribute', 'attributes', count, _ ).
noun( 'attribution', 'attributions', both, _ ).
noun( 'attrition', '-', mass, _ ).
noun( 'au pair', 'au pairs', count, _ ).
noun( 'aubergine', 'aubergines', count, _ ).
noun( 'aubrietia', 'aubrietias', count, _ ).
noun( 'auction', 'auctions', both, _ ).
noun( 'auctioneer', 'auctioneers', count, _ ).
noun( 'audacity', '-', mass, _ ).
noun( 'audibility', '-', mass, _ ).
noun( 'audience', 'audiences', count, _ ).
noun( 'audit', 'audits', count, _ ).
noun( 'audition', 'auditions', both, _ ).
noun( 'auditor', 'auditors', count, _ ).
noun( 'auditorium', 'auditoriums', count, _ ).
noun( 'auger', 'augers', count, _ ).
noun( 'aught', '-', mass, _ ).
noun( 'augmentation', 'augmentations', both, _ ).
noun( 'augur', 'augurs', count, _ ).
noun( 'augury', 'auguries', count, _ ).
noun( 'auk', 'auks', count, _ ).
noun( 'auld lang syne', '-', count, _ ).
noun( 'aunt', 'aunts', count, _ ).
noun( 'auntie', 'aunties', count, _ ).
noun( 'aunty', 'aunties', count, _ ).
noun( 'aura', 'auras', count, _ ).
noun( 'aureole', 'aureoles', count, _ ).
noun( 'auricle', 'auricles', count, _ ).
noun( 'aurora', 'auroras', count, _ ).
noun( 'austerity', 'austerities', both, _ ).
noun( 'autarchy', 'autarchies', both, _ ).
noun( 'autarky', '-', mass, _ ).
noun( 'authentication', '-', mass, _ ).
noun( 'authenticity', '-', mass, _ ).
noun( 'author', 'authors', count, _ ).
noun( 'authoress', 'authoresses', count, _ ).
noun( 'authoritarian', 'authoritarians', count, _ ).
noun( 'authoritarianism', '-', mass, _ ).
noun( 'authority', 'authorities', both, _ ).
noun( 'authorization', '-', mass, _ ).
noun( 'authorship', '-', mass, _ ).
noun( 'autism', '-', mass, _ ).
noun( 'auto', 'autos', count, _ ).
noun( 'auto-changer', 'auto-changers', count, _ ).
noun( 'auto-da-f_e', '-', count, _ ).
noun( 'autobahn', 'autobahns', count, _ ).
noun( 'autobiography', 'autobiographies', both, _ ).
noun( 'autocracy', 'autocracies', both, _ ).
noun( 'autocrat', 'autocrats', count, _ ).
noun( 'autogiro', 'autogiros', count, _ ).
noun( 'autograph', 'autographs', count, _ ).
noun( 'autogyro', 'autogyros', count, _ ).
noun( 'automat', 'automats', count, _ ).
noun( 'automatic', 'automatics', count, _ ).
noun( 'automation', '-', mass, _ ).
noun( 'automaton', 'automatons', count, _ ).
noun( 'automobile', 'automobiles', count, _ ).
noun( 'autonomy', 'autonomies', both, _ ).
noun( 'autopsy', 'autopsies', count, _ ).
noun( 'autostrada', 'autostradas', count, _ ).
noun( 'autumn', 'autumns', both, _ ).
noun( 'auxiliary', 'auxiliaries', count, _ ).
noun( 'avail', '-', mass, _ ).
noun( 'availability', '-', mass, _ ).
noun( 'avalanche', 'avalanches', count, _ ).
noun( 'avant-garde', '-', count, _ ).
noun( 'avarice', '-', mass, _ ).
noun( 'avatar', 'avatars', count, _ ).
noun( 'avenger', 'avengers', count, _ ).
noun( 'avenue', 'avenues', count, _ ).
noun( 'average', 'averages', both, _ ).
noun( 'aversion', 'aversions', both, _ ).
noun( 'aviary', 'aviaries', count, _ ).
noun( 'aviation', '-', mass, _ ).
noun( 'aviator', 'aviators', count, _ ).
noun( 'avidity', '-', mass, _ ).
noun( 'avocado', 'avocados', count, _ ).
noun( 'avocation', 'avocations', count, _ ).
noun( 'avoidance', '-', mass, _ ).
noun( 'avoirdupois', '-', mass, _ ).
noun( 'avowal', 'avowals', both, _ ).
noun( 'awakening', 'awakenings', count, _ ).
noun( 'award', 'awards', count, _ ).
noun( 'awareness', '-', mass, _ ).
noun( 'awe', '-', mass, _ ).
noun( 'awkwardness', '-', mass, _ ).
noun( 'awl', 'awls', count, _ ).
noun( 'awning', 'awnings', count, _ ).
noun( 'ax', 'axes', count, _ ).
noun( 'axe', 'axes', count, _ ).
noun( 'axiom', 'axioms', count, _ ).
noun( 'axis', 'axes', count, _ ).
noun( 'axle', 'axles', count, _ ).
noun( 'ayah', 'ayahs', count, _ ).
noun( 'azalea', 'azaleas', count, _ ).
noun( 'azimuth', 'azimuths', count, _ ).
noun( 'azure', 'azures', count, _ ).
noun( 'b', '-', count, _ ).
noun( 'b^ete noire', 'b^etes noires', count, _ ).
noun( 'baa', 'baas', count, _ ).
noun( 'baa-lamb', 'baa-lambs', count, _ ).
noun( 'baas', '-', count, _ ).
noun( 'babble', '-', mass, _ ).
noun( 'babbler', 'babblers', count, _ ).
noun( 'babe', 'babes', count, _ ).
noun( 'babel', 'babels', count, _ ).
noun( 'baboo', 'baboos', count, _ ).
noun( 'baboon', 'baboons', count, _ ).
noun( 'babu', 'babus', count, _ ).
noun( 'baby', 'babies', count, _ ).
noun( 'baby-farmer', 'baby-farmers', count, _ ).
noun( 'baby-minder', 'baby-minders', count, _ ).
noun( 'baby-talk', '-', mass, _ ).
noun( 'babyhood', '-', count, _ ).
noun( 'babysitter', 'babysitters', count, _ ).
noun( 'babysitting', '-', mass, _ ).
noun( 'baccalaureate', 'baccalaureates', count, _ ).
noun( 'baccarat', '-', mass, _ ).
noun( 'bacchanal', 'bacchanals', count, _ ).
noun( 'baccy', '-', mass, _ ).
noun( 'bachelor', 'bachelors', count, _ ).
noun( 'bacillus', 'bacilli', count, _ ).
noun( 'back', 'backs', count, _ ).
noun( 'back-down', 'back-downs', count, _ ).
noun( 'back-formation', 'back-formations', both, _ ).
noun( 'back-up', 'back-ups', count, _ ).
noun( 'backache', 'backaches', both, _ ).
noun( 'backband', 'backbands', count, _ ).
noun( 'backbench', 'backbenches', count, _ ).
noun( 'backbencher', 'backbenchers', count, _ ).
noun( 'backbiter', 'backbiters', count, _ ).
noun( 'backboard', 'backboards', count, _ ).
noun( 'backbone', 'backbones', both, _ ).
noun( 'backchat', '-', mass, _ ).
noun( 'backcloth', 'backcloths', count, _ ).
noun( 'backdoor', 'backdoors', count, _ ).
noun( 'backdrop', 'backdrops', count, _ ).
noun( 'backer', 'backers', count, _ ).
noun( 'backfire', 'backfires', count, _ ).
noun( 'backgammon', '-', mass, _ ).
noun( 'background', 'backgrounds', count, _ ).
noun( 'backing', 'backings', both, _ ).
noun( 'backlash', '-', mass, _ ).
noun( 'backlog', 'backlogs', count, _ ).
noun( 'backroom', 'backrooms', count, _ ).
noun( 'backscratcher', 'backscratchers', count, _ ).
noun( 'backseat', 'backseats', count, _ ).
noun( 'backsheesh', 'backsheesh', both, _ ).
noun( 'backside', 'backsides', count, _ ).
noun( 'backstroke', 'backstrokes', both, _ ).
noun( 'backsword', 'backswords', count, _ ).
noun( 'backtalk', '-', mass, _ ).
noun( 'backwash', '-', mass, _ ).
noun( 'backwater', 'backwaters', count, _ ).
noun( 'backwoodsman', 'backwoodsmen', count, _ ).
noun( 'bacon', '-', mass, _ ).
noun( 'bacteriologist', 'bacteriologists', count, _ ).
noun( 'bacteriology', '-', mass, _ ).
noun( 'bacterium', 'bacteria', count, _ ).
noun( 'bad', '-', mass, _ ).
noun( 'badge', 'badges', count, _ ).
noun( 'badger', 'badgers', count, _ ).
noun( 'badinage', '-', mass, _ ).
noun( 'badminton', '-', mass, _ ).
noun( 'badness', '-', mass, _ ).
noun( 'baffle', 'baffles', count, _ ).
noun( 'bag', 'bags', count, _ ).
noun( 'bagatelle', 'bagatelles', both, _ ).
noun( 'baggage', '-', mass, _ ).
noun( 'bagnio', 'bagnios', count, _ ).
noun( 'bagpipe', 'bagpipes', count, _ ).
noun( 'bail', 'bails', both, _ ).
noun( 'bailee', 'bailees', count, _ ).
noun( 'bailey', 'baileys', count, _ ).
noun( 'bailiff', 'bailiffs', count, _ ).
noun( 'bailment', 'bailments', count, _ ).
noun( 'bailor', 'bailors', count, _ ).
noun( 'bairn', 'bairns', count, _ ).
noun( 'bait', 'baits', count, _ ).
noun( 'baize', '-', mass, _ ).
noun( 'bakelite', '-', mass, _ ).
noun( 'baker', 'bakers', count, _ ).
noun( 'bakery', 'bakeries', count, _ ).
noun( 'baking-powder', 'baking-powders', count, _ ).
noun( 'baksheesh', 'baksheesh', both, _ ).
noun( 'balaclava', 'balaclavas', count, _ ).
noun( 'balalaika', 'balalaikas', count, _ ).
noun( 'balance', 'balances', both, _ ).
noun( 'balance-sheet', 'balance-sheets', count, _ ).
noun( 'balance-wheel', 'balance-wheels', count, _ ).
noun( 'balcony', 'balconies', count, _ ).
noun( 'bald-head', 'bald-heads', count, _ ).
noun( 'bald-pate', 'bald-pates', count, _ ).
noun( 'balderdash', '-', mass, _ ).
noun( 'baldness', '-', mass, _ ).
noun( 'baldric', 'baldrics', count, _ ).
noun( 'bale', 'bales', count, _ ).
noun( 'balk', 'balks', count, _ ).
noun( 'ball', 'balls', count, _ ).
noun( 'ball-cartridge', 'ball-cartridges', count, _ ).
noun( 'ball-dress', 'ball-dresses', count, _ ).
noun( 'ballad', 'ballads', count, _ ).
noun( 'ballade', 'ballades', count, _ ).
noun( 'ballast', '-', mass, _ ).
noun( 'ballbearing', 'ballbearings', count, _ ).
noun( 'ballcock', 'ballcocks', count, _ ).
noun( 'ballerina', 'ballerinas', count, _ ).
noun( 'ballet', 'ballets', both, _ ).
noun( 'ballet-dancer', 'ballet-dancers', count, _ ).
noun( 'ballet-skirt', 'ballet-skirts', count, _ ).
noun( 'ballistics', 'ballistics', mass, _ ).
noun( 'ballock', 'ballocks', count, _ ).
noun( 'balloon', 'balloons', count, _ ).
noun( 'balloonist', 'balloonists', count, _ ).
noun( 'ballot', 'ballots', both, _ ).
noun( 'ballot-box', 'ballot-boxes', count, _ ).
noun( 'ballpen', 'ballpens', count, _ ).
noun( 'ballpoint', 'ballpoints', count, _ ).
noun( 'ballpoint-pen', 'ballpoint-pens', count, _ ).
noun( 'ballroom', 'ballrooms', count, _ ).
noun( 'balls-up', 'balls-ups', count, _ ).
noun( 'ballyhoo', '-', mass, _ ).
noun( 'balm', '-', mass, _ ).
noun( 'baloney', '-', mass, _ ).
noun( 'balsa', 'balsas', both, _ ).
noun( 'balsam', 'balsams', count, _ ).
noun( 'baluster', 'balusters', count, _ ).
noun( 'balustrade', 'balustrades', count, _ ).
noun( 'bambino', 'bambinos', count, _ ).
noun( 'bamboo', 'bamboos', both, _ ).
noun( 'ban', 'bans', count, _ ).
noun( 'banality', 'banalities', both, _ ).
noun( 'banana', 'bananas', count, _ ).
noun( 'band', 'bands', count, _ ).
noun( 'band-saw', 'band-saws', count, _ ).
noun( 'bandage', 'bandages', count, _ ).
noun( 'bandanna', 'bandannas', count, _ ).
noun( 'bandbox', 'bandboxes', count, _ ).
noun( 'bandeau', 'bandeaux', count, _ ).
noun( 'bandit', 'bandits', count, _ ).
noun( 'banditry', '-', mass, _ ).
noun( 'bandleader', 'bandleaders', count, _ ).
noun( 'bandmaster', 'bandmasters', count, _ ).
noun( 'bandoleer', 'bandoleers', count, _ ).
noun( 'bandolier', 'bandoliers', count, _ ).
noun( 'bandsman', 'bandsmen', count, _ ).
noun( 'bandstand', 'bandstands', count, _ ).
noun( 'bandwagon', 'bandwagons', count, _ ).
noun( 'bane', '-', mass, _ ).
noun( 'bang', 'bangs', both, _ ).
noun( 'banger', 'bangers', count, _ ).
noun( 'bangle', 'bangles', count, _ ).
noun( 'banian', 'banians', count, _ ).
noun( 'banian-tree', 'banian-trees', count, _ ).
noun( 'banishment', '-', mass, _ ).
noun( 'banister', 'banisters', count, _ ).
noun( 'banjo', 'banjos', count, _ ).
noun( 'bank', 'banks', count, _ ).
noun( 'bank-bill', 'bank-bills', count, _ ).
noun( 'bank-book', 'bank-books', count, _ ).
noun( 'bank-draft', 'bank-drafts', count, _ ).
noun( 'bank-rate', 'bank-rates', count, _ ).
noun( 'banker', 'bankers', count, _ ).
noun( 'banking', '-', mass, _ ).
noun( 'banknote', 'banknotes', count, _ ).
noun( 'bankroll', 'bankrolls', count, _ ).
noun( 'bankrupt', 'bankrupts', count, _ ).
noun( 'bankruptcy', 'bankruptcies', both, _ ).
noun( 'banner', 'banners', count, _ ).
noun( 'banning-order', 'banning-orders', count, _ ).
noun( 'bannister', 'bannisters', count, _ ).
noun( 'bannock', 'bannocks', count, _ ).
noun( 'banquet', 'banquets', count, _ ).
noun( 'banshee', 'banshees', count, _ ).
noun( 'bantam', 'bantams', count, _ ).
noun( 'banter', '-', mass, _ ).
noun( 'banting', '-', mass, _ ).
noun( 'banyan', 'banyans', count, _ ).
noun( 'baobab', 'baobabs', count, _ ).
noun( 'baptism', 'baptisms', both, _ ).
noun( 'bar', 'bars', count, _ ).
noun( 'barb', 'barbs', count, _ ).
noun( 'barbarian', 'barbarians', count, _ ).
noun( 'barbarism', 'barbarisms', both, _ ).
noun( 'barbarity', 'barbarities', both, _ ).
noun( 'barbecue', 'barbecues', count, _ ).
noun( 'barbel', 'barbels', count, _ ).
noun( 'barber', 'barbers', count, _ ).
noun( 'barbican', 'barbicans', count, _ ).
noun( 'barbitone', '-', mass, _ ).
noun( 'barbiturate', 'barbiturates', both, _ ).
noun( 'barcarole', 'barcaroles', count, _ ).
noun( 'barcarolle', 'barcarolles', count, _ ).
noun( 'bard', 'bards', count, _ ).
noun( 'bardolatry', '-', mass, _ ).
noun( 'bareness', '-', mass, _ ).
noun( 'bargain', 'bargains', count, _ ).
noun( 'barge', 'barges', count, _ ).
noun( 'bargee', 'bargees', count, _ ).
noun( 'bargepole', 'bargepoles', count, _ ).
noun( 'baritone', 'baritones', count, _ ).
noun( 'barium', '-', mass, _ ).
noun( 'bark', 'barks', both, _ ).
noun( 'barker', 'barkers', count, _ ).
noun( 'barley', '-', mass, _ ).
noun( 'barley-sugar', 'barley-sugars', both, _ ).
noun( 'barley-water', '-', mass, _ ).
noun( 'barleycorn', 'barleycorns', both, _ ).
noun( 'barm', '-', mass, _ ).
noun( 'barmaid', 'barmaids', count, _ ).
noun( 'barman', 'barmen', count, _ ).
noun( 'barn', 'barns', count, _ ).
noun( 'barn-door', 'barn-doors', count, _ ).
noun( 'barnacle', 'barnacles', count, _ ).
noun( 'barnstormer', 'barnstormers', count, _ ).
noun( 'barnyard', 'barnyards', count, _ ).
noun( 'barometer', 'barometers', count, _ ).
noun( 'baron', 'barons', count, _ ).
noun( 'baronage', 'baronages', count, _ ).
noun( 'baroness', 'baronesses', count, _ ).
noun( 'baronet', 'baronets', count, _ ).
noun( 'baronetcy', 'baronetcies', count, _ ).
noun( 'barony', 'baronies', count, _ ).
noun( 'baroque', '-', count, _ ).
noun( 'barouche', 'barouches', count, _ ).
noun( 'barque', 'barques', count, _ ).
noun( 'barrack', 'barracks', count, _ ).
noun( 'barracking', '-', mass, _ ).
noun( 'barracuda', 'barracudas', count, _ ).
noun( 'barrage', 'barrages', count, _ ).
noun( 'barrel', 'barrels', count, _ ).
noun( 'barrenness', '-', mass, _ ).
noun( 'barricade', 'barricades', count, _ ).
noun( 'barrier', 'barriers', count, _ ).
noun( 'barrister', 'barristers', count, _ ).
noun( 'barrow', 'barrows', count, _ ).
noun( 'barrow-boy', 'barrow-boys', count, _ ).
noun( 'barrow-man', 'barrow-men', count, _ ).
noun( 'bartender', 'bartenders', count, _ ).
noun( 'barter', '-', mass, _ ).
noun( 'barterer', 'barterers', count, _ ).
noun( 'bas-relief', 'bas-reliefs', both, _ ).
noun( 'basalt', '-', mass, _ ).
noun( 'bascule', 'bascules', count, _ ).
noun( 'base', 'bases', count, _ ).
noun( 'baseball', 'baseballs', both, _ ).
noun( 'baseboard', 'baseboards', count, _ ).
noun( 'basement', 'basements', count, _ ).
noun( 'bash', 'bashes', count, _ ).
noun( 'basil', '-', mass, _ ).
noun( 'basilica', 'basilicas', count, _ ).
noun( 'basilisk', 'basilisks', count, _ ).
noun( 'basin', 'basins', count, _ ).
noun( 'basis', 'bases', count, _ ).
noun( 'basket', 'baskets', count, _ ).
noun( 'basketball', 'basketballs', both, _ ).
noun( 'basketry', '-', mass, _ ).
noun( 'bass', 'bass', count, _ ).
noun( 'bass', 'basses', count, _ ).
noun( 'bassinet', 'bassinets', count, _ ).
noun( 'bassoon', 'bassoons', count, _ ).
noun( 'bast', '-', mass, _ ).
noun( 'bastard', 'bastards', count, _ ).
noun( 'bastardy', 'bastardies', count, _ ).
noun( 'bastinado', '-', count, _ ).
noun( 'bastion', 'bastions', count, _ ).
noun( 'bat', 'bats', count, _ ).
noun( 'batch', 'batches', count, _ ).
noun( 'bath', 'baths', count, _ ).
noun( 'bathe', 'bathes', count, _ ).
noun( 'bather', 'bathers', count, _ ).
noun( 'bathing', 'bathings', count, _ ).
noun( 'bathing-cap', 'bathing-caps', count, _ ).
noun( 'bathing-costume', 'bathing-costumes', count, _ ).
noun( 'bathing-machine', 'bathing-machines', count, _ ).
noun( 'bathing-suit', 'bathing-suits', count, _ ).
noun( 'bathos', '-', mass, _ ).
noun( 'bathrobe', 'bathrobes', count, _ ).
noun( 'bathroom', 'bathrooms', count, _ ).
noun( 'bathtub', 'bathtubs', count, _ ).
noun( 'bathysphere', 'bathyspheres', count, _ ).
noun( 'batik', 'batiks', both, _ ).
noun( 'batiste', '-', mass, _ ).
noun( 'batman', 'batmen', count, _ ).
noun( 'baton', 'batons', count, _ ).
noun( 'batsman', 'batsmen', count, _ ).
noun( 'battalion', 'battalions', count, _ ).
noun( 'batten', 'battens', count, _ ).
noun( 'batter', '-', mass, _ ).
noun( 'battery', 'batteries', count, _ ).
noun( 'batting', '-', mass, _ ).
noun( 'battle', 'battles', both, _ ).
noun( 'battle-axe', 'battle-axes', count, _ ).
noun( 'battle-cruiser', 'battle-cruisers', count, _ ).
noun( 'battledore', 'battledores', count, _ ).
noun( 'battledress', '-', count, _ ).
noun( 'battlefield', 'battlefields', count, _ ).
noun( 'battleground', 'battlegrounds', count, _ ).
noun( 'battleship', 'battleships', count, _ ).
noun( 'battue', 'battues', count, _ ).
noun( 'bauble', 'baubles', count, _ ).
noun( 'baulk', 'baulks', count, _ ).
noun( 'bauxite', '-', mass, _ ).
noun( 'bawbee', 'bawbees', count, _ ).
noun( 'bawd', 'bawds', count, _ ).
noun( 'bawdy', '-', mass, _ ).
noun( 'bay', 'bays', count, _ ).
noun( 'bay-wreath', 'bay-wreaths', count, _ ).
noun( 'bayonet', 'bayonets', count, _ ).
noun( 'bayou', 'bayous', count, _ ).
noun( 'bazaar', 'bazaars', count, _ ).
noun( 'bazooka', 'bazookas', count, _ ).
noun( 'beach', 'beaches', count, _ ).
noun( 'beachcomber', 'beachcombers', count, _ ).
noun( 'beachhead', 'beachheads', count, _ ).
noun( 'beachwear', '-', mass, _ ).
noun( 'beacon', 'beacons', count, _ ).
noun( 'beacon-fire', 'beacon-fires', count, _ ).
noun( 'beacon-light', 'beacon-lights', count, _ ).
noun( 'bead', 'beads', count, _ ).
noun( 'beading', '-', mass, _ ).
noun( 'beadle', 'beadles', count, _ ).
noun( 'beagle', 'beagles', count, _ ).
noun( 'beagling', '-', mass, _ ).
noun( 'beak', 'beaks', count, _ ).
noun( 'beaker', 'beakers', count, _ ).
noun( 'beam', 'beams', count, _ ).
noun( 'bean', 'beans', count, _ ).
noun( 'beanfeast', 'beanfeasts', count, _ ).
noun( 'beano', 'beanos', count, _ ).
noun( 'beanstalk', 'beanstalks', count, _ ).
noun( 'bear', 'bears', count, _ ).
noun( 'beard', 'beards', count, _ ).
noun( 'bearer', 'bearers', count, _ ).
noun( 'bearing', 'bearings', both, _ ).
noun( 'bearskin', 'bearskins', count, _ ).
noun( 'beast', 'beasts', count, _ ).
noun( 'beastliness', '-', mass, _ ).
noun( 'beat', 'beats', count, _ ).
noun( 'beater', 'beaters', count, _ ).
noun( 'beatification', 'beatifications', count, _ ).
noun( 'beating', 'beatings', count, _ ).
noun( 'beatitude', 'beatitudes', both, _ ).
noun( 'beatnik', 'beatniks', count, _ ).
noun( 'beau', 'beaux', count, _ ).
noun( 'beautician', 'beauticians', count, _ ).
noun( 'beauty', 'beauties', both, _ ).
noun( 'beauty-parlour', 'beauty-parlours', count, _ ).
noun( 'beauty-salon', 'beauty-salons', count, _ ).
noun( 'beauty-sleep', '-', mass, _ ).
noun( 'beauty-spot', 'beauty-spots', count, _ ).
noun( 'beaver', 'beavers', both, _ ).
noun( 'beck', 'becks', count, _ ).
noun( 'bed', 'beds', count, _ ).
noun( 'bedbug', 'bedbugs', count, _ ).
noun( 'bedding', '-', mass, _ ).
noun( 'bedevilment', 'bedevilments', count, _ ).
noun( 'bedfellow', 'bedfellows', count, _ ).
noun( 'bedlam', 'bedlams', count, _ ).
noun( 'bedpan', 'bedpans', count, _ ).
noun( 'bedpost', 'bedposts', count, _ ).
noun( 'bedrock', 'bedrocks', count, _ ).
noun( 'bedroll', 'bedrolls', count, _ ).
noun( 'bedroom', 'bedrooms', count, _ ).
noun( 'bedside', 'bedsides', count, _ ).
noun( 'bedsit', 'bedsits', count, _ ).
noun( 'bedsitter', 'bedsitters', count, _ ).
noun( 'bedsitting-room', 'bedsitting-rooms', count, _ ).
noun( 'bedsore', 'bedsores', count, _ ).
noun( 'bedspread', 'bedspreads', count, _ ).
noun( 'bedstead', 'bedsteads', count, _ ).
noun( 'bedtime', 'bedtimes', count, _ ).
noun( 'bee', 'bees', count, _ ).
noun( 'beech', 'beeches', both, _ ).
noun( 'beef', '-', both, _ ).
noun( 'beefeater', 'beefeaters', count, _ ).
noun( 'beefsteak', 'beefsteaks', count, _ ).
noun( 'beehive', 'beehives', count, _ ).
noun( 'beep', 'beeps', count, _ ).
noun( 'beer', 'beers', both, _ ).
noun( 'beeswax', '-', mass, _ ).
noun( 'beet', 'beets', count, _ ).
noun( 'beetle', 'beetles', count, _ ).
noun( 'beetroot', 'beetroots', count, _ ).
noun( 'begetter', 'begetters', count, _ ).
noun( 'beggar', 'beggars', count, _ ).
noun( 'beggarman', 'beggarmen', count, _ ).
noun( 'beggarwoman', 'beggarwomen', count, _ ).
noun( 'beggary', '-', mass, _ ).
noun( 'beginner', 'beginners', count, _ ).
noun( 'beginning', 'beginnings', count, _ ).
noun( 'begonia', 'begonias', count, _ ).
noun( 'begum', 'begums', count, _ ).
noun( 'behalf', 'behalves', count, _ ).
noun( 'behaviour', '-', mass, _ ).
noun( 'behaviourism', '-', mass, _ ).
noun( 'behaviourist', 'behaviourists', count, _ ).
noun( 'behest', 'behests', count, _ ).
noun( 'behind', 'behinds', count, _ ).
noun( 'beholder', 'beholders', count, _ ).
noun( 'beige', '-', mass, _ ).
noun( 'being', 'beings', both, _ ).
noun( 'belay', 'belays', count, _ ).
noun( 'belaying-pin', 'belaying-pins', count, _ ).
noun( 'belch', 'belches', count, _ ).
noun( 'beldam', 'beldams', count, _ ).
noun( 'beldame', 'beldames', count, _ ).
noun( 'belfry', 'belfries', count, _ ).
noun( 'belief', 'beliefs', both, _ ).
noun( 'believer', 'believers', count, _ ).
noun( 'believing', '-', mass, _ ).
noun( 'bell', 'bells', count, _ ).
noun( 'bell-buoy', 'bell-buoys', count, _ ).
noun( 'bell-flower', 'bell-flowers', count, _ ).
noun( 'bell-founder', 'bell-founders', count, _ ).
noun( 'bell-foundry', 'bell-foundries', count, _ ).
noun( 'bell-metal', '-', mass, _ ).
noun( 'bell-push', 'bell-pushes', count, _ ).
noun( 'bell-ringer', 'bell-ringers', count, _ ).
noun( 'bell-tent', 'bell-tents', count, _ ).
noun( 'belladonna', 'belladonnas', both, _ ).
noun( 'bellboy', 'bellboys', count, _ ).
noun( 'belle', 'belles', count, _ ).
noun( 'bellhop', 'bellhops', count, _ ).
noun( 'belligerency', '-', mass, _ ).
noun( 'belligerent', 'belligerents', count, _ ).
noun( 'bellwether', 'bellwethers', count, _ ).
noun( 'belly', 'bellies', count, _ ).
noun( 'bellyache', 'bellyaches', count, _ ).
noun( 'bellyflop', 'bellyflops', count, _ ).
noun( 'bellyful', 'bellyfuls', count, _ ).
noun( 'bellylaugh', 'bellylaughs', count, _ ).
noun( 'beloved', 'beloveds', count, _ ).
noun( 'belt', 'belts', count, _ ).
noun( 'belting', 'beltings', count, _ ).
noun( 'ben', 'bens', count, _ ).
noun( 'bench', 'benches', count, _ ).
noun( 'bend', 'bends', count, _ ).
noun( 'benedick', 'benedicks', count, _ ).
noun( 'benediction', 'benedictions', count, _ ).
noun( 'benefaction', 'benefactions', both, _ ).
noun( 'benefactor', 'benefactors', count, _ ).
noun( 'benefactress', 'benefactresses', count, _ ).
noun( 'benefice', 'benefices', count, _ ).
noun( 'beneficence', '-', mass, _ ).
noun( 'beneficiary', 'beneficiaries', count, _ ).
noun( 'benefit', 'benefits', both, _ ).
noun( 'benevolence', '-', mass, _ ).
noun( 'benignity', 'benignities', both, _ ).
noun( 'benison', 'benisons', count, _ ).
noun( 'bent', 'bents', count, _ ).
noun( 'benzene', '-', mass, _ ).
noun( 'benzine', '-', mass, _ ).
noun( 'benzol', '-', mass, _ ).
noun( 'bequest', 'bequests', both, _ ).
noun( 'bereavement', 'bereavements', both, _ ).
noun( 'beret', 'berets', count, _ ).
noun( 'berg', 'bergs', count, _ ).
noun( 'beri-beri', '-', mass, _ ).
noun( 'berry', 'berries', count, _ ).
noun( 'berth', 'berths', count, _ ).
noun( 'beryl', 'beryls', count, _ ).
noun( 'besieger', 'besiegers', count, _ ).
noun( 'besom', 'besoms', count, _ ).
noun( 'best-seller', 'best-sellers', count, _ ).
noun( 'bestiality', 'bestialities', both, _ ).
noun( 'bestiary', 'bestiaries', count, _ ).
noun( 'bestowal', 'bestowals', count, _ ).
noun( 'bet', 'bets', count, _ ).
noun( 'beta', 'betas', count, _ ).
noun( 'betel', 'betels', count, _ ).
noun( 'betel-nut', 'betel-nuts', count, _ ).
noun( 'bethel', 'bethels', count, _ ).
noun( 'betrayal', 'betrayals', both, _ ).
noun( 'betrayer', 'betrayers', count, _ ).
noun( 'betrothal', 'betrothals', count, _ ).
noun( 'betrothed', 'betrotheds', count, _ ).
noun( 'better', 'betters', count, _ ).
noun( 'betterment', '-', mass, _ ).
noun( 'bettor', 'bettors', count, _ ).
noun( 'bevel', 'bevels', count, _ ).
noun( 'beverage', 'beverages', count, _ ).
noun( 'bevy', 'bevies', count, _ ).
noun( 'bewilderment', '-', mass, _ ).
noun( 'bey', 'beys', count, _ ).
noun( 'bezique', '-', mass, _ ).
noun( 'bhang', 'bhangs', count, _ ).
noun( 'bias', 'biases', count, _ ).
noun( 'bib', 'bibs', count, _ ).
noun( 'bibliographer', 'bibliographers', count, _ ).
noun( 'bibliography', 'bibliographies', both, _ ).
noun( 'bibliophile', 'bibliophiles', count, _ ).
noun( 'bicarbonate', '-', mass, _ ).
noun( 'bicentenary', 'bicentenaries', count, _ ).
noun( 'bicentennial', 'bicentennials', count, _ ).
noun( 'biceps', 'biceps', count, _ ).
noun( 'bicycle', 'bicycles', count, _ ).
noun( 'bid', 'bids', count, _ ).
noun( 'bidder', 'bidders', count, _ ).
noun( 'bidding', '-', mass, _ ).
noun( 'bidet', 'bidets', count, _ ).
noun( 'biennial', 'biennials', count, _ ).
noun( 'bier', 'biers', count, _ ).
noun( 'biff', 'biffs', count, _ ).
noun( 'bifurcation', 'bifurcations', count, _ ).
noun( 'bigamist', 'bigamists', count, _ ).
noun( 'bigamy', '-', mass, _ ).
noun( 'bight', 'bights', count, _ ).
noun( 'bigot', 'bigots', count, _ ).
noun( 'bigotry', 'bigotries', both, _ ).
noun( 'bigwig', 'bigwigs', count, _ ).
noun( 'bijou', '-', count, _ ).
noun( 'bike', 'bikes', count, _ ).
noun( 'bikini', 'bikinis', count, _ ).
noun( 'bilabial', 'bilabials', count, _ ).
noun( 'bilateralism', 'bilateralisms', both, _ ).
noun( 'bilberry', 'bilberries', count, _ ).
noun( 'bile', '-', mass, _ ).
noun( 'bile-duct', 'bile-ducts', count, _ ).
noun( 'bilge', '-', mass, _ ).
noun( 'bilge-water', '-', mass, _ ).
noun( 'bilharzia', '-', mass, _ ).
noun( 'bilingual', 'bilinguals', count, _ ).
noun( 'biliousness', '-', mass, _ ).
noun( 'bill', 'bills', count, _ ).
noun( 'bill-poster', 'bill-posters', count, _ ).
noun( 'bill-sticker', 'bill-stickers', count, _ ).
noun( 'billboard', 'billboards', count, _ ).
noun( 'billet', 'billets', count, _ ).
noun( 'billet-doux', 'billets-doux', count, _ ).
noun( 'billfold', 'billfolds', count, _ ).
noun( 'billhook', 'billhooks', count, _ ).
noun( 'billiard-marker', 'billiard-markers', count, _ ).
noun( 'billiard-player', 'billiard-players', count, _ ).
noun( 'billiard-room', 'billiard-rooms', count, _ ).
noun( 'billiard-table', 'billiard-tables', count, _ ).
noun( 'billiards', 'billiards', mass, _ ).
noun( 'billingsgate', '-', mass, _ ).
noun( 'billion', 'billions', count, _ ).
noun( 'billionth', 'billionths', count, _ ).
noun( 'billow', 'billows', count, _ ).
noun( 'billy', 'billies', count, _ ).
noun( 'billy-goat', 'billy-goats', count, _ ).
noun( 'billy-ho', '-', count, _ ).
noun( 'billy-o', '-', mass, _ ).
noun( 'biltong', '-', mass, _ ).
noun( 'bimetallism', '-', mass, _ ).
noun( 'bin', 'bins', count, _ ).
noun( 'bind', 'binds', both, _ ).
noun( 'binder', 'binders', count, _ ).
noun( 'bindery', 'binderies', count, _ ).
noun( 'bindweed', '-', mass, _ ).
noun( 'bine', 'bines', count, _ ).
noun( 'binge', 'binges', count, _ ).
noun( 'bingo', '-', mass, _ ).
noun( 'binnacle', 'binnacles', count, _ ).
noun( 'biochemistry', '-', mass, _ ).
noun( 'biograph', 'biographs', count, _ ).
noun( 'biographer', 'biographers', count, _ ).
noun( 'biography', 'biographies', both, _ ).
noun( 'biologist', 'biologists', count, _ ).
noun( 'biology', '-', mass, _ ).
noun( 'bioscope', 'bioscopes', count, _ ).
noun( 'biped', 'bipeds', count, _ ).
noun( 'biplane', 'biplanes', count, _ ).
noun( 'birch', 'birches', both, _ ).
noun( 'birch-rod', 'birch-rods', count, _ ).
noun( 'bird', 'birds', count, _ ).
noun( 'bird-fancier', 'bird-fanciers', count, _ ).
noun( 'birdcage', 'birdcages', count, _ ).
noun( 'birdlime', 'birdlimes', count, _ ).
noun( 'birdnesting', '-', mass, _ ).
noun( 'birdwatcher', 'birdwatchers', count, _ ).
noun( 'biretta', 'birettas', count, _ ).
noun( 'biro', 'biros', count, _ ).
noun( 'birth', 'births', both, _ ).
noun( 'birth-control', '-', mass, _ ).
noun( 'birthday', 'birthdays', count, _ ).
noun( 'birthmark', 'birthmarks', count, _ ).
noun( 'birthplace', 'birthplaces', count, _ ).
noun( 'birthrate', 'birthrates', count, _ ).
noun( 'birthright', 'birthrights', count, _ ).
noun( 'biscuit', 'biscuits', count, _ ).
noun( 'bisection', '-', mass, _ ).
noun( 'bisexual', 'bisexuals', count, _ ).
noun( 'bisexuality', '-', mass, _ ).
noun( 'bishop', 'bishops', count, _ ).
noun( 'bishopric', 'bishoprics', count, _ ).
noun( 'bismuth', '-', mass, _ ).
noun( 'bison', 'bison', count, _ ).
noun( 'bistro', 'bistros', count, _ ).
noun( 'bit', 'bits', count, _ ).
noun( 'bitch', 'bitches', count, _ ).
noun( 'bite', 'bites', both, _ ).
noun( 'bitter', 'bitters', both, _ ).
noun( 'bittern', 'bitterns', count, _ ).
noun( 'bitterness', '-', mass, _ ).
noun( 'bitumen', '-', mass, _ ).
noun( 'bivalve', 'bivalves', count, _ ).
noun( 'bivouac', 'bivouacs', count, _ ).
noun( 'biz', '-', mass, _ ).
noun( 'blabbermouth', 'blabbermouths', count, _ ).
noun( 'black', 'blacks', both, _ ).
noun( 'black-beetle', 'black-beetles', count, _ ).
noun( 'black-lead', '-', mass, _ ).
noun( 'blackamoor', 'blackamoors', count, _ ).
noun( 'blackberry', 'blackberries', count, _ ).
noun( 'blackbird', 'blackbirds', count, _ ).
noun( 'blackboard', 'blackboards', count, _ ).
noun( 'blackcurrant', 'blackcurrants', count, _ ).
noun( 'blackguard', 'blackguards', count, _ ).
noun( 'blackhead', 'blackheads', count, _ ).
noun( 'blacking', '-', mass, _ ).
noun( 'blackleg', 'blacklegs', count, _ ).
noun( 'blacklist', 'blacklists', count, _ ).
noun( 'blackmail', '-', mass, _ ).
noun( 'blackmailer', 'blackmailers', count, _ ).
noun( 'blackness', '-', mass, _ ).
noun( 'blackout', 'blackouts', count, _ ).
noun( 'blacksmith', 'blacksmiths', count, _ ).
noun( 'blackthorn', 'blackthorns', count, _ ).
noun( 'bladder', 'bladders', count, _ ).
noun( 'blade', 'blades', count, _ ).
noun( 'blaeberry', 'blaeberries', count, _ ).
noun( 'blah', '-', mass, _ ).
noun( 'blame', '-', mass, _ ).
noun( 'blancmange', 'blancmanges', both, _ ).
noun( 'blandishment', 'blandishments', count, _ ).
noun( 'blandness', '-', mass, _ ).
noun( 'blank', 'blanks', both, _ ).
noun( 'blanket', 'blankets', count, _ ).
noun( 'blare', '-', mass, _ ).
noun( 'blarney', '-', mass, _ ).
noun( 'blasphemer', 'blasphemers', count, _ ).
noun( 'blasphemy', 'blasphemies', both, _ ).
noun( 'blast', 'blasts', both, _ ).
noun( 'blast-furnace', 'blast-furnaces', count, _ ).
noun( 'blast-off', 'blast-offs', count, _ ).
noun( 'blather', 'blathers', both, _ ).
noun( 'blaze', 'blazes', count, _ ).
noun( 'blazer', 'blazers', count, _ ).
noun( 'blazon', 'blazons', count, _ ).
noun( 'blazonry', 'blazonries', count, _ ).
noun( 'bleach', 'bleaches', count, _ ).
noun( 'bleaching-powder', 'bleaching-powders', count, _ ).
noun( 'bleat', 'bleats', count, _ ).
noun( 'bleep', 'bleeps', count, _ ).
noun( 'blemish', 'blemishes', both, _ ).
noun( 'blend', 'blends', count, _ ).
noun( 'blessedness', '-', mass, _ ).
noun( 'blessing', 'blessings', count, _ ).
noun( 'blether', '-', mass, _ ).
noun( 'blight', 'blights', both, _ ).
noun( 'blighter', 'blighters', count, _ ).
noun( 'blimp', 'blimps', count, _ ).
noun( 'blind', 'blinds', count, _ ).
noun( 'blindfold', 'blindfolds', count, _ ).
noun( 'blindman\'s buff', '-', mass, _ ).
noun( 'blindness', '-', mass, _ ).
noun( 'blink', 'blinks', count, _ ).
noun( 'blip', 'blips', count, _ ).
noun( 'bliss', '-', mass, _ ).
noun( 'blister', 'blisters', count, _ ).
noun( 'blitz', 'blitzes', count, _ ).
noun( 'blizzard', 'blizzards', count, _ ).
noun( 'bloater', 'bloaters', count, _ ).
noun( 'blob', 'blobs', count, _ ).
noun( 'bloc', 'blocs', count, _ ).
noun( 'block', 'blocks', count, _ ).
noun( 'blockade', 'blockades', count, _ ).
noun( 'blockade-runner', 'blockade-runners', count, _ ).
noun( 'blockage', 'blockages', count, _ ).
noun( 'blockbuster', 'blockbusters', count, _ ).
noun( 'blockhead', 'blockheads', count, _ ).
noun( 'blockhouse', 'blockhouses', count, _ ).
noun( 'bloke', 'blokes', count, _ ).
noun( 'blond', 'blonds', count, _ ).
noun( 'blonde', 'blondes', count, _ ).
noun( 'blood', 'bloods', both, _ ).
noun( 'blood-bath', 'blood-baths', count, _ ).
noun( 'blood-donor', 'blood-donors', count, _ ).
noun( 'blood-group', 'blood-groups', count, _ ).
noun( 'blood-heat', '-', mass, _ ).
noun( 'blood-letting', 'blood-lettings', both, _ ).
noun( 'blood-money', '-', mass, _ ).
noun( 'blood-poisoning', '-', mass, _ ).
noun( 'blood-pressure', 'blood-pressures', both, _ ).
noun( 'blood-relation', 'blood-relations', count, _ ).
noun( 'blood-transfusion', 'blood-transfusions', count, _ ).
noun( 'blood-type', 'blood-types', count, _ ).
noun( 'blood-vessel', 'blood-vessels', count, _ ).
noun( 'bloodhound', 'bloodhounds', count, _ ).
noun( 'bloodlust', '-', mass, _ ).
noun( 'bloodshed', 'bloodsheds', count, _ ).
noun( 'bloodstain', 'bloodstains', count, _ ).
noun( 'bloodstock', 'bloodstocks', count, _ ).
noun( 'bloodsucker', 'bloodsuckers', count, _ ).
noun( 'bloodthirstiness', '-', mass, _ ).
noun( 'bloom', 'blooms', both, _ ).
noun( 'bloomer', 'bloomers', count, _ ).
noun( 'blossom', 'blossoms', both, _ ).
noun( 'blot', 'blots', count, _ ).
noun( 'blotch', 'blotches', count, _ ).
noun( 'blotter', 'blotters', count, _ ).
noun( 'blotting-paper', 'blotting-papers', count, _ ).
noun( 'blouse', 'blouses', count, _ ).
noun( 'blow', 'blows', count, _ ).
noun( 'blow-up', 'blow-ups', count, _ ).
noun( 'blowback', 'blowbacks', count, _ ).
noun( 'blower', 'blowers', count, _ ).
noun( 'blowfly', 'blowflies', count, _ ).
noun( 'blowhole', 'blowholes', count, _ ).
noun( 'blowing-up', 'blowing-ups', count, _ ).
noun( 'blowlamp', 'blowlamps', count, _ ).
noun( 'blowout', 'blowouts', count, _ ).
noun( 'blowpipe', 'blowpipes', count, _ ).
noun( 'blowtorch', 'blowtorches', count, _ ).
noun( 'blubber', '-', mass, _ ).
noun( 'bludgeon', 'bludgeons', count, _ ).
noun( 'blue', 'blues', both, _ ).
noun( 'blue-jacket', 'blue-jackets', count, _ ).
noun( 'bluebell', 'bluebells', count, _ ).
noun( 'bluebottle', 'bluebottles', count, _ ).
noun( 'blueprint', 'blueprints', count, _ ).
noun( 'bluestocking', 'bluestockings', count, _ ).
noun( 'bluff', 'bluffs', both, _ ).
noun( 'bluffer', 'bluffers', count, _ ).
noun( 'bluffness', '-', mass, _ ).
noun( 'blunder', 'blunders', count, _ ).
noun( 'blunderbuss', 'blunderbusses', count, _ ).
noun( 'blunderer', 'blunderers', count, _ ).
noun( 'bluntness', '-', mass, _ ).
noun( 'blur', 'blurs', count, _ ).
noun( 'blurb', 'blurbs', both, _ ).
noun( 'blush', 'blushes', count, _ ).
noun( 'bluster', 'blusters', both, _ ).
noun( 'bo\'sn', 'bo\'sns', count, _ ).
noun( 'bo\'sun', 'bo\'suns', count, _ ).
noun( 'boa', 'boas', count, _ ).
noun( 'boa-constrictor', 'boa-constrictors', count, _ ).
noun( 'boar', 'boars', count, _ ).
noun( 'board', 'boards', count, _ ).
noun( 'boarder', 'boarders', count, _ ).
noun( 'boarding', '-', mass, _ ).
noun( 'boarding-card', 'boarding-cards', count, _ ).
noun( 'boarding-house', 'boarding-houses', count, _ ).
noun( 'boarding-school', 'boarding-schools', count, _ ).
noun( 'boardroom', 'boardrooms', count, _ ).
noun( 'boardwalk', 'boardwalks', count, _ ).
noun( 'boast', 'boasts', count, _ ).
noun( 'boaster', 'boasters', count, _ ).
noun( 'boat', 'boats', count, _ ).
noun( 'boat-hook', 'boat-hooks', count, _ ).
noun( 'boat-house', 'boat-houses', count, _ ).
noun( 'boat-race', 'boat-races', count, _ ).
noun( 'boat-train', 'boat-trains', count, _ ).
noun( 'boater', 'boaters', count, _ ).
noun( 'boatman', 'boatmen', count, _ ).
noun( 'boatswain', 'boatswains', count, _ ).
noun( 'bob', 'bob', count, _ ).
noun( 'bobbin', 'bobbins', count, _ ).
noun( 'bobby', 'bobbies', count, _ ).
noun( 'bobby-soxer', 'bobby-soxers', count, _ ).
noun( 'bobolink', 'bobolinks', count, _ ).
noun( 'bobsled', 'bobsleds', count, _ ).
noun( 'bobsleigh', 'bobsleighs', count, _ ).
noun( 'bobtail', 'bobtails', count, _ ).
noun( 'bodice', 'bodices', count, _ ).
noun( 'boding', '-', mass, _ ).
noun( 'bodkin', 'bodkins', count, _ ).
noun( 'body', 'bodies', both, _ ).
noun( 'body-servant', 'body-servants', count, _ ).
noun( 'body-snatcher', 'body-snatchers', count, _ ).
noun( 'bodyguard', 'bodyguards', count, _ ).
noun( 'bodywork', 'bodyworks', count, _ ).
noun( 'boffin', 'boffins', count, _ ).
noun( 'bog', 'bogs', count, _ ).
noun( 'bogey', 'bogeys', count, _ ).
noun( 'bogeyman', 'bogeymen', count, _ ).
noun( 'bogie', 'bogies', count, _ ).
noun( 'bogy', 'bogies', count, _ ).
noun( 'bohemian', 'bohemians', count, _ ).
noun( 'boil', 'boils', count, _ ).
noun( 'boiler', 'boilers', count, _ ).
noun( 'boilersuit', 'boilersuits', count, _ ).
noun( 'boiling-point', 'boiling-points', count, _ ).
noun( 'boisterousness', '-', mass, _ ).
noun( 'boldness', '-', mass, _ ).
noun( 'bole', 'boles', count, _ ).
noun( 'bolero', 'boleros', count, _ ).
noun( 'boll', 'bolls', count, _ ).
noun( 'bollard', 'bollards', count, _ ).
noun( 'bollock', 'bollocks', count, _ ).
noun( 'boloney', '-', mass, _ ).
noun( 'bolster', 'bolsters', count, _ ).
noun( 'bolt', 'bolts', count, _ ).
noun( 'bolt-hole', 'bolt-holes', count, _ ).
noun( 'bomb', 'bombs', count, _ ).
noun( 'bomb-sight', 'bomb-sights', count, _ ).
noun( 'bomb-site', 'bomb-sites', count, _ ).
noun( 'bombardier', 'bombardiers', count, _ ).
noun( 'bombardment', 'bombardments', both, _ ).
noun( 'bombast', '-', mass, _ ).
noun( 'bomber', 'bombers', count, _ ).
noun( 'bombshell', 'bombshells', count, _ ).
noun( 'bon mot', 'bons mots', count, _ ).
noun( 'bonanza', 'bonanzas', count, _ ).
noun( 'bonbon', 'bonbons', count, _ ).
noun( 'bond', 'bonds', count, _ ).
noun( 'bond-holder', 'bond-holders', count, _ ).
noun( 'bondage', '-', mass, _ ).
noun( 'bone', 'bones', both, _ ).
noun( 'bone-head', 'bone-heads', count, _ ).
noun( 'bone-setter', 'bone-setters', count, _ ).
noun( 'bonemeal', 'bonemeals', count, _ ).
noun( 'boner', 'boners', count, _ ).
noun( 'boneshaker', 'boneshakers', count, _ ).
noun( 'bonfire', 'bonfires', count, _ ).
noun( 'bongo', 'bongos', count, _ ).
noun( 'bonhomie', '-', mass, _ ).
noun( 'bonito', 'bonitos', count, _ ).
noun( 'bonnet', 'bonnets', count, _ ).
noun( 'bonus', 'bonuses', count, _ ).
noun( 'boo', 'boos', count, _ ).
noun( 'boob', 'boobs', count, _ ).
noun( 'booby', 'boobies', count, _ ).
noun( 'booby-trap', 'booby-traps', count, _ ).
noun( 'boogie', 'boogies', both, _ ).
noun( 'boogie-woogie', 'boogie-woogies', both, _ ).
noun( 'book', 'books', count, _ ).
noun( 'book-end', 'book-ends', count, _ ).
noun( 'book-keeper', 'book-keepers', count, _ ).
noun( 'bookcase', 'bookcases', count, _ ).
noun( 'bookclub', 'bookclubs', count, _ ).
noun( 'bookie', 'bookies', count, _ ).
noun( 'booking', 'bookings', count, _ ).
noun( 'bookishness', '-', mass, _ ).
noun( 'bookkeeper', 'bookkeepers', count, _ ).
noun( 'bookkeeping', '-', mass, _ ).
noun( 'booklet', 'booklets', count, _ ).
noun( 'bookmaker', 'bookmakers', count, _ ).
noun( 'bookmark', 'bookmarks', count, _ ).
noun( 'bookmarker', 'bookmarkers', count, _ ).
noun( 'bookmobile', 'bookmobiles', count, _ ).
noun( 'bookseller', 'booksellers', count, _ ).
noun( 'bookshop', 'bookshops', count, _ ).
noun( 'bookstall', 'bookstalls', count, _ ).
noun( 'bookworm', 'bookworms', count, _ ).
noun( 'boom', 'booms', count, _ ).
noun( 'boomerang', 'boomerangs', count, _ ).
noun( 'boon', 'boons', count, _ ).
noun( 'boor', 'boors', count, _ ).
noun( 'boorishness', '-', mass, _ ).
noun( 'boost', 'boosts', count, _ ).
noun( 'booster', 'boosters', count, _ ).
noun( 'boot', 'boots', count, _ ).
noun( 'bootee', 'bootees', count, _ ).
noun( 'booth', 'booths', count, _ ).
noun( 'bootlace', 'bootlaces', count, _ ).
noun( 'bootlegger', 'bootleggers', count, _ ).
noun( 'booty', '-', mass, _ ).
noun( 'booze', '-', mass, _ ).
noun( 'booze-up', 'booze-ups', count, _ ).
noun( 'boozer', 'boozers', count, _ ).
noun( 'bopeep', '-', mass, _ ).
noun( 'borage', '-', mass, _ ).
noun( 'borax', '-', mass, _ ).
noun( 'border', 'borders', count, _ ).
noun( 'borderer', 'borderers', count, _ ).
noun( 'borderland', 'borderlands', count, _ ).
noun( 'borderline', 'borderlines', count, _ ).
noun( 'bore', 'bores', count, _ ).
noun( 'bore-hole', 'bore-holes', count, _ ).
noun( 'boredom', '-', mass, _ ).
noun( 'borer', 'borers', count, _ ).
noun( 'boron', '-', mass, _ ).
noun( 'borough', 'boroughs', count, _ ).
noun( 'borrower', 'borrowers', count, _ ).
noun( 'borsch', '-', mass, _ ).
noun( 'borstal', 'borstals', count, _ ).
noun( 'bortsch', '-', mass, _ ).
noun( 'borzoi', 'borzois', count, _ ).
noun( 'bosh', '-', mass, _ ).
noun( 'bosom', 'bosoms', count, _ ).
noun( 'boss', 'bosses', count, _ ).
noun( 'botanist', 'botanists', count, _ ).
noun( 'botany', '-', mass, _ ).
noun( 'botch', 'botches', count, _ ).
noun( 'botcher', 'botchers', count, _ ).
noun( 'bother', '-', mass, _ ).
noun( 'bottle', 'bottles', both, _ ).
noun( 'bottleneck', 'bottlenecks', count, _ ).
noun( 'bottom', 'bottoms', count, _ ).
noun( 'botulism', '-', mass, _ ).
noun( 'boudoir', 'boudoirs', count, _ ).
noun( 'bougainvillea', 'bougainvilleas', count, _ ).
noun( 'bough', 'boughs', count, _ ).
noun( 'bouillon', '-', mass, _ ).
noun( 'boulder', 'boulders', count, _ ).
noun( 'boulevard', 'boulevards', count, _ ).
noun( 'bounce', 'bounces', both, _ ).
noun( 'bound', 'bounds', count, _ ).
noun( 'boundary', 'boundaries', count, _ ).
noun( 'bounder', 'bounders', count, _ ).
noun( 'bounty', 'bounties', both, _ ).
noun( 'bouquet', 'bouquets', count, _ ).
noun( 'bourbon', 'bourbons', both, _ ).
noun( 'bourgeois', 'bourgeois', count, _ ).
noun( 'bourgeoisie', 'bourgeoisies', count, _ ).
noun( 'bourn', 'bourns', count, _ ).
noun( 'bourne', 'bournes', count, _ ).
noun( 'bourse', 'bourses', count, _ ).
noun( 'bout', 'bouts', count, _ ).
noun( 'boutique', 'boutiques', count, _ ).
noun( 'bovril', '-', mass, _ ).
noun( 'bow', 'bows', count, _ ).
noun( 'bow', 'bows', count, _ ).
noun( 'bow-wow', 'bow-wows', count, _ ).
noun( 'bowel', 'bowels', count, _ ).
noun( 'bower', 'bowers', count, _ ).
noun( 'bowie knife', 'bowie knives', count, _ ).
noun( 'bowing', '-', mass, _ ).
noun( 'bowl', 'bowls', count, _ ).
noun( 'bowler', 'bowlers', count, _ ).
noun( 'bowline', 'bowlines', count, _ ).
noun( 'bowling-green', 'bowling-greens', count, _ ).
noun( 'bowls', 'bowls', mass, _ ).
noun( 'bowman', 'bowmen', count, _ ).
noun( 'bowsprit', 'bowsprits', count, _ ).
noun( 'box', 'boxes', both, _ ).
noun( 'box-kite', 'box-kites', count, _ ).
noun( 'box-number', 'box-numbers', count, _ ).
noun( 'box-office', 'box-offices', count, _ ).
noun( 'boxer', 'boxers', count, _ ).
noun( 'boxful', 'boxfuls', count, _ ).
noun( 'boxing', '-', mass, _ ).
noun( 'boxing-glove', 'boxing-gloves', count, _ ).
noun( 'boxing-match', 'boxing-matches', count, _ ).
noun( 'boxwood', '-', mass, _ ).
noun( 'boy', 'boys', count, _ ).
noun( 'boycott', 'boycotts', count, _ ).
noun( 'boyfriend', 'boyfriends', count, _ ).
noun( 'boyhood', '-', mass, _ ).
noun( 'bra', 'bras', count, _ ).
noun( 'brace', 'brace', count, _ ).
noun( 'bracelet', 'bracelets', count, _ ).
noun( 'bracken', '-', mass, _ ).
noun( 'bracket', 'brackets', count, _ ).
noun( 'bract', 'bracts', count, _ ).
noun( 'brad', 'brads', count, _ ).
noun( 'bradawl', 'bradawls', count, _ ).
noun( 'brae', 'braes', count, _ ).
noun( 'braggart', 'braggarts', count, _ ).
noun( 'bragging', '-', mass, _ ).
noun( 'braid', 'braids', both, _ ).
noun( 'braille', '-', mass, _ ).
noun( 'brain', 'brains', count, _ ).
noun( 'brain-fag', '-', mass, _ ).
noun( 'brain-teaser', 'brain-teasers', count, _ ).
noun( 'brainchild', 'brainchildren', count, _ ).
noun( 'brainstorm', 'brainstorms', count, _ ).
noun( 'brainwashing', '-', mass, _ ).
noun( 'brainwave', 'brainwaves', count, _ ).
noun( 'brake', 'brakes', count, _ ).
noun( 'brakeman', 'brakemen', count, _ ).
noun( 'bramble', 'brambles', count, _ ).
noun( 'bran', '-', mass, _ ).
noun( 'branch', 'branches', count, _ ).
noun( 'brand', 'brands', count, _ ).
noun( 'branding-iron', 'branding-irons', count, _ ).
noun( 'brandy', 'brandies', both, _ ).
noun( 'brandy-ball', 'brandy-balls', count, _ ).
noun( 'brandy-snap', 'brandy-snaps', count, _ ).
noun( 'brass', 'brasses', both, _ ).
noun( 'brassard', 'brassards', count, _ ).
noun( 'brasserie', 'brasseries', count, _ ).
noun( 'brassi`ere', 'brassi`eres', count, _ ).
noun( 'brassiere', 'brassieres', count, _ ).
noun( 'brat', 'brats', count, _ ).
noun( 'bravado', 'bravados', both, _ ).
noun( 'brave', 'braves', count, _ ).
noun( 'bravery', '-', mass, _ ).
noun( 'bravo', 'bravos', count, _ ).
noun( 'brawl', 'brawls', count, _ ).
noun( 'brawler', 'brawlers', count, _ ).
noun( 'brawn', '-', mass, _ ).
noun( 'bray', 'brays', count, _ ).
noun( 'brazier', 'braziers', count, _ ).
noun( 'breach', 'breaches', count, _ ).
noun( 'bread', '-', mass, _ ).
noun( 'breadcrumb', 'breadcrumbs', count, _ ).
noun( 'breadfruit', 'breadfruit', count, _ ).
noun( 'breadline', 'breadlines', count, _ ).
noun( 'breadth', 'breadths', both, _ ).
noun( 'breadwinner', 'breadwinners', count, _ ).
noun( 'break', 'breaks', both, _ ).
noun( 'break-in', 'break-ins', count, _ ).
noun( 'break-up', 'break-ups', count, _ ).
noun( 'breakage', 'breakages', count, _ ).
noun( 'breakaway', 'breakaways', count, _ ).
noun( 'breakaways', 'breakawayss', count, _ ).
noun( 'breakdown', 'breakdowns', count, _ ).
noun( 'breaker', 'breakers', count, _ ).
noun( 'breakfast', 'breakfasts', count, _ ).
noun( 'breakthrough', 'breakthroughs', count, _ ).
noun( 'breakwater', 'breakwaters', count, _ ).
noun( 'bream', 'bream', count, _ ).
noun( 'breast', 'breasts', count, _ ).
noun( 'breast-plate', 'breast-plates', count, _ ).
noun( 'breaststroke', '-', mass, _ ).
noun( 'breastwork', 'breastworks', count, _ ).
noun( 'breath', 'breaths', both, _ ).
noun( 'breathalyser', 'breathalysers', count, _ ).
noun( 'breather', 'breathers', count, _ ).
noun( 'breathing', '-', mass, _ ).
noun( 'breathing-space', 'breathing-spaces', count, _ ).
noun( 'breech', 'breeches', count, _ ).
noun( 'breech-block', 'breech-blocks', count, _ ).
noun( 'breeches-buoy', 'breeches-buoys', count, _ ).
noun( 'breed', 'breeds', count, _ ).
noun( 'breeder', 'breeders', count, _ ).
noun( 'breeding', '-', mass, _ ).
noun( 'breeze', 'breezes', both, _ ).
noun( 'breeziness', '-', mass, _ ).
noun( 'breve', 'breves', count, _ ).
noun( 'brevet', 'brevets', count, _ ).
noun( 'breviary', 'breviaries', count, _ ).
noun( 'brevity', '-', mass, _ ).
noun( 'brew', 'brews', count, _ ).
noun( 'brewer', 'brewers', count, _ ).
noun( 'brewery', 'breweries', count, _ ).
noun( 'briar', 'briars', both, _ ).
noun( 'bribe', 'bribes', count, _ ).
noun( 'bribery', '-', mass, _ ).
noun( 'bric-a-brac', '-', mass, _ ).
noun( 'brick', 'bricks', both, _ ).
noun( 'brick-field', 'brick-fields', count, _ ).
noun( 'brickbat', 'brickbats', count, _ ).
noun( 'brickkiln', 'brickkilns', count, _ ).
noun( 'bricklayer', 'bricklayers', count, _ ).
noun( 'brickwork', 'brickworks', count, _ ).
noun( 'bridal', 'bridals', count, _ ).
noun( 'bride', 'brides', count, _ ).
noun( 'bridecake', 'bridecakes', count, _ ).
noun( 'bridegroom', 'bridegrooms', count, _ ).
noun( 'bridesmaid', 'bridesmaids', count, _ ).
noun( 'bridge', 'bridges', both, _ ).
noun( 'bridgehead', 'bridgeheads', count, _ ).
noun( 'bridle', 'bridles', count, _ ).
noun( 'bridle-path', 'bridle-paths', count, _ ).
noun( 'bridle-road', 'bridle-roads', count, _ ).
noun( 'brief', 'briefs', count, _ ).
noun( 'briefcase', 'briefcases', count, _ ).
noun( 'briefing', 'briefings', count, _ ).
noun( 'brier', 'briers', count, _ ).
noun( 'brig', 'brigs', count, _ ).
noun( 'brigade', 'brigades', count, _ ).
noun( 'brigand', 'brigands', count, _ ).
noun( 'brigantine', 'brigantines', count, _ ).
noun( 'brightness', '-', mass, _ ).
noun( 'brill', 'brill', count, _ ).
noun( 'brilliance', '-', mass, _ ).
noun( 'brilliancy', '-', mass, _ ).
noun( 'brilliantine', '-', mass, _ ).
noun( 'brim', 'brims', count, _ ).
noun( 'brimstone', '-', mass, _ ).
noun( 'brine', 'brines', both, _ ).
noun( 'brink', 'brinks', count, _ ).
noun( 'brinkmanship', '-', mass, _ ).
noun( 'brioche', 'brioches', count, _ ).
noun( 'briquet', 'briquets', count, _ ).
noun( 'briquette', 'briquettes', count, _ ).
noun( 'brisket', '-', mass, _ ).
noun( 'bristle', 'bristles', count, _ ).
noun( 'broad', 'broads', count, _ ).
noun( 'broad-mindedness', '-', mass, _ ).
noun( 'broadcast', 'broadcasts', count, _ ).
noun( 'broadcasting', '-', mass, _ ).
noun( 'broadcloth', '-', mass, _ ).
noun( 'broadness', '-', mass, _ ).
noun( 'broadsheet', 'broadsheets', count, _ ).
noun( 'broadside', 'broadsides', count, _ ).
noun( 'brocade', 'brocades', both, _ ).
noun( 'broccoli', '-', mass, _ ).
noun( 'brochure', 'brochures', count, _ ).
noun( 'brogue', 'brogues', count, _ ).
noun( 'broiler', 'broilers', count, _ ).
noun( 'broker', 'brokers', count, _ ).
noun( 'brokerage', '-', mass, _ ).
noun( 'brolly', 'brollies', count, _ ).
noun( 'bromide', 'bromides', both, _ ).
noun( 'bromine', '-', mass, _ ).
noun( 'bronchitis', '-', mass, _ ).
noun( 'bronchus', 'bronchi', count, _ ).
noun( 'bronco', 'broncos', count, _ ).
noun( 'bronze', 'bronzes', both, _ ).
noun( 'brooch', 'brooches', count, _ ).
noun( 'brood', 'broods', count, _ ).
noun( 'brood-hen', 'brood-hens', count, _ ).
noun( 'brood-mare', 'brood-mares', count, _ ).
noun( 'brook', 'brooks', count, _ ).
noun( 'broom', 'brooms', both, _ ).
noun( 'broomstick', 'broomsticks', count, _ ).
noun( 'broth', '-', mass, _ ).
noun( 'brothel', 'brothels', count, _ ).
noun( 'brother', 'brothers', count, _ ).
noun( 'brother-in-law', 'brothers-in-law', count, _ ).
noun( 'brotherhood', 'brotherhoods', both, _ ).
noun( 'brougham', 'broughams', count, _ ).
noun( 'brouhaha', 'brouhahas', count, _ ).
noun( 'brow', 'brows', count, _ ).
noun( 'brownie', 'brownies', count, _ ).
noun( 'brownstone', 'brownstones', both, _ ).
noun( 'browse', 'browses', count, _ ).
noun( 'bruin', 'bruins', count, _ ).
noun( 'bruise', 'bruises', count, _ ).
noun( 'bruiser', 'bruisers', count, _ ).
noun( 'brunch', 'brunches', count, _ ).
noun( 'brunette', 'brunettes', count, _ ).
noun( 'brunt', 'brunts', count, _ ).
noun( 'brush', 'brushes', both, _ ).
noun( 'brush-off', 'brush-offs', count, _ ).
noun( 'brush-up', 'brush-ups', count, _ ).
noun( 'brushwood', '-', mass, _ ).
noun( 'brushwork', 'brushworks', count, _ ).
noun( 'brusqueness', '-', mass, _ ).
noun( 'brutality', 'brutalities', both, _ ).
noun( 'brute', 'brutes', count, _ ).
noun( 'bubble', 'bubbles', count, _ ).
noun( 'bubbly', '-', mass, _ ).
noun( 'buccaneer', 'buccaneers', count, _ ).
noun( 'buck', 'bucks', count, _ ).
noun( 'bucket', 'buckets', count, _ ).
noun( 'bucketful', 'bucketfuls', count, _ ).
noun( 'buckle', 'buckles', count, _ ).
noun( 'buckler', 'bucklers', count, _ ).
noun( 'buckram', '-', mass, _ ).
noun( 'buckshot', '-', mass, _ ).
noun( 'buckskin', '-', mass, _ ).
noun( 'bucktooth', 'buckteeth', count, _ ).
noun( 'buckwheat', '-', mass, _ ).
noun( 'bucolics', 'bucolics', mass, _ ).
noun( 'bud', 'buds', count, _ ).
noun( 'buddy', 'buddies', count, _ ).
noun( 'budgerigar', 'budgerigars', count, _ ).
noun( 'budget', 'budgets', count, _ ).
noun( 'budgie', 'budgies', count, _ ).
noun( 'buff', 'buffs', both, _ ).
noun( 'buffalo', 'buffalo', count, _ ).
noun( 'buffer', 'buffers', count, _ ).
noun( 'buffet', 'buffets', count, _ ).
noun( 'buffet', 'buffets', count, _ ).
noun( 'buffoon', 'buffoons', count, _ ).
noun( 'buffoonery', '-', mass, _ ).
noun( 'bug', 'bugs', count, _ ).
noun( 'bug-hunter', 'bug-hunters', count, _ ).
noun( 'bugaboo', 'bugaboos', count, _ ).
noun( 'bugbear', 'bugbears', count, _ ).
noun( 'bugger', 'buggers', count, _ ).
noun( 'bugger-all', '-', mass, _ ).
noun( 'buggery', '-', mass, _ ).
noun( 'buggy', 'buggies', count, _ ).
noun( 'bugle', 'bugles', count, _ ).
noun( 'bugler', 'buglers', count, _ ).
noun( 'buhl', '-', mass, _ ).
noun( 'build', '-', mass, _ ).
noun( 'build-up', 'build-ups', count, _ ).
noun( 'builder', 'builders', count, _ ).
noun( 'building', 'buildings', both, _ ).
noun( 'building-society', 'building-societies', count, _ ).
noun( 'bulb', 'bulbs', count, _ ).
noun( 'bulbul', 'bulbuls', count, _ ).
noun( 'bulge', 'bulges', count, _ ).
noun( 'bulk', '-', mass, _ ).
noun( 'bulkhead', 'bulkheads', count, _ ).
noun( 'bull', 'bulls', count, _ ).
noun( 'bull\'s-eye', 'bull\'s-eyes', count, _ ).
noun( 'bull-neck', 'bull-necks', count, _ ).
noun( 'bull-terrier', 'bull-terriers', count, _ ).
noun( 'bulldog', 'bulldogs', count, _ ).
noun( 'bulldozer', 'bulldozers', count, _ ).
noun( 'bullet', 'bullets', count, _ ).
noun( 'bulletin', 'bulletins', count, _ ).
noun( 'bullfight', 'bullfights', count, _ ).
noun( 'bullfighter', 'bullfighters', count, _ ).
noun( 'bullfinch', 'bullfinches', count, _ ).
noun( 'bullfrog', 'bullfrogs', count, _ ).
noun( 'bullion', '-', mass, _ ).
noun( 'bullock', 'bullocks', count, _ ).
noun( 'bullring', 'bullrings', count, _ ).
noun( 'bullshit', '-', mass, _ ).
noun( 'bully', 'bullies', both, _ ).
noun( 'bulrush', 'bulrushes', count, _ ).
noun( 'bulwark', 'bulwarks', count, _ ).
noun( 'bum', 'bums', count, _ ).
noun( 'bumblebee', 'bumblebees', count, _ ).
noun( 'bumboat', 'bumboats', count, _ ).
noun( 'bump', 'bumps', count, _ ).
noun( 'bumper', 'bumpers', count, _ ).
noun( 'bumpkin', 'bumpkins', count, _ ).
noun( 'bumptiousness', '-', mass, _ ).
noun( 'bun', 'buns', count, _ ).
noun( 'buna', '-', mass, _ ).
noun( 'bunch', 'bunches', count, _ ).
noun( 'bundle', 'bundles', count, _ ).
noun( 'bung', 'bungs', count, _ ).
noun( 'bung-hole', 'bung-holes', count, _ ).
noun( 'bungalow', 'bungalows', count, _ ).
noun( 'bungle', 'bungles', count, _ ).
noun( 'bungler', 'bunglers', count, _ ).
noun( 'bunion', 'bunions', count, _ ).
noun( 'bunk', 'bunks', both, _ ).
noun( 'bunker', 'bunkers', count, _ ).
noun( 'bunkum', '-', mass, _ ).
noun( 'bunny', 'bunnies', count, _ ).
noun( 'bunting', '-', mass, _ ).
noun( 'buoy', 'buoys', count, _ ).
noun( 'buoyancy', '-', mass, _ ).
noun( 'bur', 'burs', count, _ ).
noun( 'burden', 'burdens', both, _ ).
noun( 'burdock', 'burdocks', count, _ ).
noun( 'bureau', 'bureaux', count, _ ).
noun( 'bureaucracy', 'bureaucracies', both, _ ).
noun( 'bureaucrat', 'bureaucrats', count, _ ).
noun( 'burette', 'burettes', count, _ ).
noun( 'burg', 'burgs', count, _ ).
noun( 'burgess', 'burgesses', count, _ ).
noun( 'burgh', 'burghs', count, _ ).
noun( 'burgher', 'burghers', count, _ ).
noun( 'burglar', 'burglars', count, _ ).
noun( 'burglar-alarm', 'burglar-alarms', count, _ ).
noun( 'burglary', 'burglaries', both, _ ).
noun( 'burgomaster', 'burgomasters', count, _ ).
noun( 'burial', 'burials', both, _ ).
noun( 'burial-ground', 'burial-grounds', count, _ ).
noun( 'burlap', '-', mass, _ ).
noun( 'burlesque', 'burlesques', both, _ ).
noun( 'burn', 'burns', count, _ ).
noun( 'burn-up', 'burn-ups', count, _ ).
noun( 'burner', 'burners', count, _ ).
noun( 'burnouse', 'burnouses', count, _ ).
noun( 'burp', 'burps', count, _ ).
noun( 'burr', 'burrs', count, _ ).
noun( 'burr-drill', 'burr-drills', count, _ ).
noun( 'burrow', 'burrows', count, _ ).
noun( 'bursar', 'bursars', count, _ ).
noun( 'bursary', 'bursaries', count, _ ).
noun( 'burst', 'bursts', count, _ ).
noun( 'burthen', 'burthens', count, _ ).
noun( 'burton', 'burtons', count, _ ).
noun( 'burying-ground', 'burying-grounds', count, _ ).
noun( 'bus', 'buses', count, _ ).
noun( 'busby', 'busbies', count, _ ).
noun( 'bush', 'bushes', both, _ ).
noun( 'bushel', 'bushels', count, _ ).
noun( 'business', 'businesses', both, _ ).
noun( 'businessman', 'businessmen', count, _ ).
noun( 'busker', 'buskers', count, _ ).
noun( 'busman', 'busmen', count, _ ).
noun( 'bust', 'busts', count, _ ).
noun( 'bust-up', 'bust-ups', count, _ ).
noun( 'bustard', 'bustards', count, _ ).
noun( 'buster', 'busters', count, _ ).
noun( 'bustle', 'bustles', both, _ ).
noun( 'busybody', 'busybodies', count, _ ).
noun( 'butane', '-', mass, _ ).
noun( 'butcher', 'butchers', count, _ ).
noun( 'butchery', '-', mass, _ ).
noun( 'butler', 'butlers', count, _ ).
noun( 'butt', 'butts', count, _ ).
noun( 'butter', '-', mass, _ ).
noun( 'butterbean', 'butterbeans', count, _ ).
noun( 'buttercup', 'buttercups', count, _ ).
noun( 'butterfingers', '-', count, _ ).
noun( 'butterfly', 'butterflies', count, _ ).
noun( 'buttermilk', '-', mass, _ ).
noun( 'butterscotch', '-', mass, _ ).
noun( 'buttery', 'butteries', count, _ ).
noun( 'buttock', 'buttocks', count, _ ).
noun( 'button', 'buttons', count, _ ).
noun( 'buttonhole', 'buttonholes', count, _ ).
noun( 'buttonhook', 'buttonhooks', count, _ ).
noun( 'buttonwood', 'buttonwoods', both, _ ).
noun( 'buttress', 'buttresses', count, _ ).
noun( 'buy', 'buys', count, _ ).
noun( 'buyer', 'buyers', count, _ ).
noun( 'buzz', 'buzzes', count, _ ).
noun( 'buzzard', 'buzzards', count, _ ).
noun( 'buzzer', 'buzzers', count, _ ).
noun( 'by-election', 'by-elections', count, _ ).
noun( 'bye', 'byes', count, _ ).
noun( 'bye-bye', 'bye-byes', count, _ ).
noun( 'bye-law', 'bye-laws', count, _ ).
noun( 'bylaw', 'bylaws', count, _ ).
noun( 'bypass', 'bypasses', count, _ ).
noun( 'bypath', 'bypaths', count, _ ).
noun( 'byplay', '-', mass, _ ).
noun( 'byproduct', 'byproducts', count, _ ).
noun( 'byroad', 'byroads', count, _ ).
noun( 'bystander', 'bystanders', count, _ ).
noun( 'byway', 'byways', count, _ ).
noun( 'byword', 'bywords', count, _ ).
noun( 'c', '-', count, _ ).
noun( 'ca\'canny', '-', count, _ ).
noun( 'cab', 'cabs', count, _ ).
noun( 'cab-rank', 'cab-ranks', count, _ ).
noun( 'cabal', 'cabals', count, _ ).
noun( 'cabaret', 'cabarets', count, _ ).
noun( 'cabbage', 'cabbages', both, _ ).
noun( 'cabby', 'cabbies', count, _ ).
noun( 'caber', 'cabers', count, _ ).
noun( 'cabin', 'cabins', count, _ ).
noun( 'cabinet', 'cabinets', count, _ ).
noun( 'cabinet-maker', 'cabinet-makers', count, _ ).
noun( 'cable', 'cables', both, _ ).
noun( 'cable\'s-length', 'cable\'s-lengths', count, _ ).
noun( 'cable-car', 'cable-cars', count, _ ).
noun( 'cable-length', 'cable-lengths', count, _ ).
noun( 'cable-railway', 'cable-railways', count, _ ).
noun( 'cablegram', 'cablegrams', count, _ ).
noun( 'cabman', 'cabmen', count, _ ).
noun( 'caboodle', 'caboodles', count, _ ).
noun( 'caboose', 'cabooses', count, _ ).
noun( 'cabstand', 'cabstands', count, _ ).
noun( 'cacao', 'cacaos', count, _ ).
noun( 'cacao-bean', 'cacao-beans', count, _ ).
noun( 'cacao-tree', 'cacao-trees', count, _ ).
noun( 'cache', 'caches', count, _ ).
noun( 'cachet', 'cachets', count, _ ).
noun( 'cachou', 'cachous', count, _ ).
noun( 'cackle', 'cackles', both, _ ).
noun( 'cackler', 'cacklers', count, _ ).
noun( 'cacophony', 'cacophonies', count, _ ).
noun( 'cactus', 'cactuses', count, _ ).
noun( 'cad', 'cads', count, _ ).
noun( 'cadaver', 'cadavers', count, _ ).
noun( 'caddie', 'caddies', count, _ ).
noun( 'caddy', 'caddies', count, _ ).
noun( 'cadence', 'cadences', count, _ ).
noun( 'cadenza', 'cadenzas', count, _ ).
noun( 'cadet', 'cadets', count, _ ).
noun( 'cadger', 'cadgers', count, _ ).
noun( 'cadmium', '-', mass, _ ).
noun( 'cadre', 'cadres', count, _ ).
noun( 'caesura', 'caesuras', count, _ ).
noun( 'caf_e', 'caf_es', count, _ ).
noun( 'cafe-au-lait', 'cafe-au-laits', count, _ ).
noun( 'cafeteria', 'cafeterias', count, _ ).
noun( 'caff', 'caffs', count, _ ).
noun( 'caffeine', '-', mass, _ ).
noun( 'caftan', 'caftans', count, _ ).
noun( 'cage', 'cages', count, _ ).
noun( 'cagoule', 'cagoules', count, _ ).
noun( 'caiman', 'caimans', count, _ ).
noun( 'cairn', 'cairns', count, _ ).
noun( 'caisson', 'caissons', count, _ ).
noun( 'caitiff', 'caitiffs', count, _ ).
noun( 'cajolery', '-', mass, _ ).
noun( 'cake', 'cakes', both, _ ).
noun( 'calabash', 'calabashes', count, _ ).
noun( 'calamity', 'calamities', count, _ ).
noun( 'calcination', 'calcinations', count, _ ).
noun( 'calcium', '-', mass, _ ).
noun( 'calculation', 'calculations', both, _ ).
noun( 'calculator', 'calculators', count, _ ).
noun( 'calculus', 'calculuses', both, _ ).
noun( 'caldron', 'caldrons', count, _ ).
noun( 'calendar', 'calendars', count, _ ).
noun( 'calender', 'calenders', count, _ ).
noun( 'calf', 'calves', both, _ ).
noun( 'calf-love', '-', mass, _ ).
noun( 'calibration', 'calibrations', both, _ ).
noun( 'calibre', 'calibres', both, _ ).
noun( 'calico', '-', mass, _ ).
noun( 'calif', 'califs', count, _ ).
noun( 'caliph', 'caliphs', count, _ ).
noun( 'caliphate', 'caliphates', count, _ ).
noun( 'calisthenics', 'calisthenics', mass, _ ).
noun( 'calk', 'calks', count, _ ).
noun( 'call', 'calls', both, _ ).
noun( 'call-box', 'call-boxes', count, _ ).
noun( 'call-girl', 'call-girls', count, _ ).
noun( 'call-over', 'call-overs', count, _ ).
noun( 'call-up', 'call-ups', count, _ ).
noun( 'caller', 'callers', count, _ ).
noun( 'calligraphy', '-', mass, _ ).
noun( 'calling', 'callings', count, _ ).
noun( 'calliope', 'calliopes', count, _ ).
noun( 'callisthenics', 'callisthenics', mass, _ ).
noun( 'callosity', 'callosities', count, _ ).
noun( 'callousness', '-', mass, _ ).
noun( 'callowness', '-', mass, _ ).
noun( 'callus', 'calluses', count, _ ).
noun( 'calm', 'calms', count, _ ).
noun( 'calmness', '-', mass, _ ).
noun( 'calomel', '-', mass, _ ).
noun( 'calorie', 'calories', count, _ ).
noun( 'calumny', 'calumnies', both, _ ).
noun( 'calypso', 'calypsos', count, _ ).
noun( 'calyx', 'calyxes', count, _ ).
noun( 'cam', 'cams', count, _ ).
noun( 'camaraderie', '-', mass, _ ).
noun( 'camber', 'cambers', count, _ ).
noun( 'cambric', '-', mass, _ ).
noun( 'camel', 'camels', count, _ ).
noun( 'camel-hair', '-', mass, _ ).
noun( 'camellia', 'camellias', count, _ ).
noun( 'cameo', 'cameos', count, _ ).
noun( 'camera', 'cameras', count, _ ).
noun( 'cameraman', 'cameramen', count, _ ).
noun( 'camion', 'camions', count, _ ).
noun( 'camomile', '-', mass, _ ).
noun( 'camouflage', '-', mass, _ ).
noun( 'camp', 'camps', count, _ ).
noun( 'camp-bed', 'camp-beds', count, _ ).
noun( 'camp-chair', 'camp-chairs', count, _ ).
noun( 'camp-fire', 'camp-fires', count, _ ).
noun( 'camp-follower', 'camp-followers', count, _ ).
noun( 'camp-stool', 'camp-stools', count, _ ).
noun( 'campaign', 'campaigns', count, _ ).
noun( 'campaigner', 'campaigners', count, _ ).
noun( 'campanile', 'campaniles', count, _ ).
noun( 'campanula', 'campanulas', count, _ ).
noun( 'camper', 'campers', count, _ ).
noun( 'camphor', '-', mass, _ ).
noun( 'camping', '-', mass, _ ).
noun( 'campion', '-', mass, _ ).
noun( 'campus', 'campuses', count, _ ).
noun( 'camshaft', 'camshafts', count, _ ).
noun( 'can', 'cans', count, _ ).
noun( 'canal', 'canals', count, _ ).
noun( 'canalization', 'canalizations', count, _ ).
noun( 'canap_e', 'canap_es', count, _ ).
noun( 'canard', 'canards', count, _ ).
noun( 'canary', 'canaries', count, _ ).
noun( 'canary-bird', 'canary-birds', count, _ ).
noun( 'canary-wine', 'canary-wines', count, _ ).
noun( 'canasta', 'canastas', count, _ ).
noun( 'cancan', 'cancans', count, _ ).
noun( 'cancellation', 'cancellations', both, _ ).
noun( 'cancer', 'cancers', both, _ ).
noun( 'candelabrum', 'candelabra', count, _ ).
noun( 'candidate', 'candidates', count, _ ).
noun( 'candidature', 'candidatures', count, _ ).
noun( 'candle', 'candles', count, _ ).
noun( 'candle-power', '-', mass, _ ).
noun( 'candlelight', '-', mass, _ ).
noun( 'candlestick', 'candlesticks', count, _ ).
noun( 'candlewick', '-', mass, _ ).
noun( 'candour', '-', mass, _ ).
noun( 'candy', 'candies', both, _ ).
noun( 'candy-floss', '-', mass, _ ).
noun( 'candytuft', 'candytufts', count, _ ).
noun( 'cane', 'canes', both, _ ).
noun( 'canister', 'canisters', count, _ ).
noun( 'canker', '-', mass, _ ).
noun( 'canna', 'cannas', count, _ ).
noun( 'cannabis', '-', mass, _ ).
noun( 'cannery', 'canneries', count, _ ).
noun( 'cannibal', 'cannibals', count, _ ).
noun( 'cannibalism', '-', mass, _ ).
noun( 'cannon', 'cannons', count, _ ).
noun( 'cannon-fodder', '-', mass, _ ).
noun( 'cannonade', 'cannonades', count, _ ).
noun( 'canoe', 'canoes', count, _ ).
noun( 'canoeist', 'canoeists', count, _ ).
noun( 'canon', 'canons', count, _ ).
noun( 'canonization', 'canonizations', count, _ ).
noun( 'canopy', 'canopies', count, _ ).
noun( 'cant', 'cants', both, _ ).
noun( 'cantaloup', 'cantaloups', count, _ ).
noun( 'cantaloupe', 'cantaloupes', count, _ ).
noun( 'cantata', 'cantatas', count, _ ).
noun( 'canteen', 'canteens', count, _ ).
noun( 'canter', 'canters', count, _ ).
noun( 'canticle', 'canticles', count, _ ).
noun( 'cantilever', 'cantilevers', count, _ ).
noun( 'canto', 'cantos', count, _ ).
noun( 'canton', 'cantons', count, _ ).
noun( 'cantonment', 'cantonments', count, _ ).
noun( 'cantor', 'cantors', count, _ ).
noun( 'canvas', 'canvases', both, _ ).
noun( 'canvass', 'canvasses', count, _ ).
noun( 'canyon', 'canyons', count, _ ).
noun( 'cap', 'caps', count, _ ).
noun( 'capability', 'capabilities', both, _ ).
noun( 'capacity', 'capacities', both, _ ).
noun( 'caparison', 'caparisons', count, _ ).
noun( 'cape', 'capes', count, _ ).
noun( 'caper', 'capers', count, _ ).
noun( 'capillary', 'capillaries', count, _ ).
noun( 'capital', 'capitals', both, _ ).
noun( 'capitalism', '-', mass, _ ).
noun( 'capitalist', 'capitalists', count, _ ).
noun( 'capitalization', 'capitalizations', count, _ ).
noun( 'capitation', 'capitations', count, _ ).
noun( 'capitulation', '-', mass, _ ).
noun( 'capon', 'capons', count, _ ).
noun( 'caprice', 'caprices', count, _ ).
noun( 'capriciousness', '-', mass, _ ).
noun( 'capsicum', 'capsicums', count, _ ).
noun( 'capstan', 'capstans', count, _ ).
noun( 'capsule', 'capsules', count, _ ).
noun( 'captain', 'captains', count, _ ).
noun( 'caption', 'captions', count, _ ).
noun( 'captive', 'captives', count, _ ).
noun( 'captivity', '-', mass, _ ).
noun( 'captor', 'captors', count, _ ).
noun( 'capture', 'captures', both, _ ).
noun( 'car', 'cars', count, _ ).
noun( 'car-ferry', 'car-ferries', count, _ ).
noun( 'carafe', 'carafes', count, _ ).
noun( 'caramel', 'caramels', both, _ ).
noun( 'carapace', 'carapaces', count, _ ).
noun( 'carat', 'carats', count, _ ).
noun( 'caravan', 'caravans', count, _ ).
noun( 'caravanning', '-', mass, _ ).
noun( 'caravansary', 'caravansaries', count, _ ).
noun( 'caravanserai', 'caravanserais', count, _ ).
noun( 'caraway', 'caraways', count, _ ).
noun( 'carbide', 'carbides', count, _ ).
noun( 'carbine', 'carbines', count, _ ).
noun( 'carbohydrate', 'carbohydrates', both, _ ).
noun( 'carbon', 'carbons', both, _ ).
noun( 'carbon-paper', '-', mass, _ ).
noun( 'carbonization', 'carbonizations', both, _ ).
noun( 'carborundum', '-', mass, _ ).
noun( 'carboy', 'carboys', count, _ ).
noun( 'carbuncle', 'carbuncles', count, _ ).
noun( 'carburettor', 'carburettors', count, _ ).
noun( 'carcase', 'carcases', count, _ ).
noun( 'carcass', 'carcasses', count, _ ).
noun( 'card', 'cards', count, _ ).
noun( 'card-sharper', 'card-sharpers', count, _ ).
noun( 'cardamom', '-', mass, _ ).
noun( 'cardboard', '-', mass, _ ).
noun( 'cardigan', 'cardigans', count, _ ).
noun( 'cardinal', 'cardinals', count, _ ).
noun( 'care', 'cares', both, _ ).
noun( 'career', 'careers', both, _ ).
noun( 'careerist', 'careerists', count, _ ).
noun( 'carefulness', '-', mass, _ ).
noun( 'carelessness', '-', mass, _ ).
noun( 'caress', 'caresses', count, _ ).
noun( 'caret', 'carets', count, _ ).
noun( 'caretaker', 'caretakers', count, _ ).
noun( 'cargo', 'cargoes', both, _ ).
noun( 'caribou', 'caribou', count, _ ).
noun( 'caricature', 'caricatures', both, _ ).
noun( 'caricaturist', 'caricaturists', count, _ ).
noun( 'caries', '-', mass, _ ).
noun( 'carillon', 'carillons', count, _ ).
noun( 'carmine', 'carmines', both, _ ).
noun( 'carnage', '-', mass, _ ).
noun( 'carnation', 'carnations', count, _ ).
noun( 'carnival', 'carnivals', both, _ ).
noun( 'carnivore', 'carnivores', count, _ ).
noun( 'carol', 'carols', count, _ ).
noun( 'caroller', 'carollers', count, _ ).
noun( 'carousal', 'carousals', count, _ ).
noun( 'carousel', 'carousels', count, _ ).
noun( 'carp', 'carp', count, _ ).
noun( 'carpal', 'carpals', count, _ ).
noun( 'carpenter', 'carpenters', count, _ ).
noun( 'carpentry', '-', mass, _ ).
noun( 'carpet', 'carpets', count, _ ).
noun( 'carpet-beater', 'carpet-beaters', count, _ ).
noun( 'carpet-knight', 'carpet-knights', count, _ ).
noun( 'carpet-sweeper', 'carpet-sweepers', count, _ ).
noun( 'carpetbag', 'carpetbags', count, _ ).
noun( 'carpetbagger', 'carpetbaggers', count, _ ).
noun( 'carport', 'carports', count, _ ).
noun( 'carriage', 'carriages', both, _ ).
noun( 'carriageway', 'carriageways', count, _ ).
noun( 'carrier', 'carriers', count, _ ).
noun( 'carrier-bag', 'carrier-bags', count, _ ).
noun( 'carrier-pigeon', 'carrier-pigeons', count, _ ).
noun( 'carrion', '-', mass, _ ).
noun( 'carrion-crow', 'carrion-crows', count, _ ).
noun( 'carrot', 'carrots', count, _ ).
noun( 'carry', 'carries', count, _ ).
noun( 'carrycot', 'carrycots', count, _ ).
noun( 'carsickness', '-', mass, _ ).
noun( 'cart', 'carts', count, _ ).
noun( 'cart-track', 'cart-tracks', count, _ ).
noun( 'cartage', '-', mass, _ ).
noun( 'carte blanche', '-', count, _ ).
noun( 'cartel', 'cartels', count, _ ).
noun( 'carter', 'carters', count, _ ).
noun( 'carthorse', 'carthorses', count, _ ).
noun( 'cartilage', 'cartilages', both, _ ).
noun( 'cartload', 'cartloads', count, _ ).
noun( 'cartographer', 'cartographers', count, _ ).
noun( 'cartography', '-', mass, _ ).
noun( 'carton', 'cartons', count, _ ).
noun( 'cartoon', 'cartoons', count, _ ).
noun( 'cartoonist', 'cartoonists', count, _ ).
noun( 'cartridge', 'cartridges', count, _ ).
noun( 'cartridge-belt', 'cartridge-belts', count, _ ).
noun( 'cartridge-paper', 'cartridge-papers', count, _ ).
noun( 'cartroad', 'cartroads', count, _ ).
noun( 'cartwheel', 'cartwheels', count, _ ).
noun( 'carver', 'carvers', count, _ ).
noun( 'carving', 'carvings', both, _ ).
noun( 'carving-fork', 'carving-forks', count, _ ).
noun( 'carving-knife', 'carving-knives', count, _ ).
noun( 'caryatid', 'caryatids', count, _ ).
noun( 'cascade', 'cascades', count, _ ).
noun( 'case', 'cases', count, _ ).
noun( 'case-history', 'case-histories', count, _ ).
noun( 'case-law', 'case-laws', count, _ ).
noun( 'casebook', 'casebooks', count, _ ).
noun( 'casein', '-', mass, _ ).
noun( 'casement', 'casements', count, _ ).
noun( 'casework', 'caseworks', count, _ ).
noun( 'cash', '-', mass, _ ).
noun( 'cashew', 'cashews', count, _ ).
noun( 'cashier', 'cashiers', count, _ ).
noun( 'cashmere', '-', mass, _ ).
noun( 'casing', 'casings', count, _ ).
noun( 'casino', 'casinos', count, _ ).
noun( 'cask', 'casks', count, _ ).
noun( 'casket', 'caskets', count, _ ).
noun( 'cassava', '-', mass, _ ).
noun( 'casserole', 'casseroles', count, _ ).
noun( 'cassette', 'cassettes', count, _ ).
noun( 'cassock', 'cassocks', count, _ ).
noun( 'cassowary', 'cassowaries', count, _ ).
noun( 'cast', 'casts', count, _ ).
noun( 'castaway', 'castaways', count, _ ).
noun( 'caste', 'castes', both, _ ).
noun( 'caster', 'casters', count, _ ).
noun( 'castigation', 'castigations', both, _ ).
noun( 'casting', 'castings', count, _ ).
noun( 'castle', 'castles', count, _ ).
noun( 'castor', 'castors', count, _ ).
noun( 'castor oil', '-', mass, _ ).
noun( 'castration', 'castrations', count, _ ).
noun( 'casualty', 'casualties', count, _ ).
noun( 'casuist', 'casuists', count, _ ).
noun( 'casuistry', 'casuistries', both, _ ).
noun( 'casus belli', '-', count, _ ).
noun( 'cat', 'cats', count, _ ).
noun( 'cat-nap', 'cat-naps', count, _ ).
noun( 'cat-o\'-nine-tails', '-', count, _ ).
noun( 'cat-sleep', 'cat-sleeps', count, _ ).
noun( 'cataclysm', 'cataclysms', count, _ ).
noun( 'catafalque', 'catafalques', count, _ ).
noun( 'catalepsy', '-', mass, _ ).
noun( 'cataleptic', 'cataleptics', count, _ ).
noun( 'catalogue', 'catalogues', count, _ ).
noun( 'catalpa', 'catalpas', count, _ ).
noun( 'catalysis', '-', mass, _ ).
noun( 'catalyst', 'catalysts', count, _ ).
noun( 'catamaran', 'catamarans', count, _ ).
noun( 'catapult', 'catapults', count, _ ).
noun( 'cataract', 'cataracts', count, _ ).
noun( 'catarrh', '-', mass, _ ).
noun( 'catastrophe', 'catastrophes', count, _ ).
noun( 'catcall', 'catcalls', count, _ ).
noun( 'catch', 'catches', count, _ ).
noun( 'catch-crop', 'catch-crops', count, _ ).
noun( 'catcher', 'catchers', count, _ ).
noun( 'catchment', 'catchments', count, _ ).
noun( 'catchment-area', 'catchment-areas', count, _ ).
noun( 'catchment-basin', 'catchment-basins', count, _ ).
noun( 'catchup', 'catchups', count, _ ).
noun( 'catchword', 'catchwords', count, _ ).
noun( 'catechism', 'catechisms', both, _ ).
noun( 'category', 'categories', count, _ ).
noun( 'caterer', 'caterers', count, _ ).
noun( 'caterpillar', 'caterpillars', count, _ ).
noun( 'caterwaul', 'caterwauls', count, _ ).
noun( 'catfish', 'catfish', count, _ ).
noun( 'catgut', '-', mass, _ ).
noun( 'catharsis', 'catharses', both, _ ).
noun( 'cathartic', 'cathartics', both, _ ).
noun( 'cathedral', 'cathedrals', count, _ ).
noun( 'cathode', 'cathodes', count, _ ).
noun( 'catholicity', '-', mass, _ ).
noun( 'catkin', 'catkins', count, _ ).
noun( 'catsup', 'catsups', both, _ ).
noun( 'cattiness', '-', mass, _ ).
noun( 'cattle-cake', '-', mass, _ ).
noun( 'cattleman', 'cattlemen', count, _ ).
noun( 'catwalk', 'catwalks', count, _ ).
noun( 'caucus', 'caucuses', count, _ ).
noun( 'caul', 'cauls', count, _ ).
noun( 'cauldron', 'cauldrons', count, _ ).
noun( 'cauliflower', 'cauliflowers', both, _ ).
noun( 'causality', '-', mass, _ ).
noun( 'causation', '-', mass, _ ).
noun( 'cause', 'causes', both, _ ).
noun( 'causerie', 'causeries', count, _ ).
noun( 'causeway', 'causeways', count, _ ).
noun( 'caution', 'cautions', both, _ ).
noun( 'cavalcade', 'cavalcades', count, _ ).
noun( 'cavalier', 'cavaliers', count, _ ).
noun( 'cavalry', 'cavalries', count, _ ).
noun( 'cavalryman', 'cavalrymen', count, _ ).
noun( 'cave', 'caves', count, _ ).
noun( 'cave-dweller', 'cave-dwellers', count, _ ).
noun( 'cave-in', 'cave-ins', count, _ ).
noun( 'caveat', 'caveats', count, _ ).
noun( 'caveman', 'cavemen', count, _ ).
noun( 'cavern', 'caverns', count, _ ).
noun( 'caviar', '-', mass, _ ).
noun( 'caviare', '-', mass, _ ).
noun( 'cavity', 'cavities', count, _ ).
noun( 'caw', 'caws', count, _ ).
noun( 'cayenne', '-', mass, _ ).
noun( 'cayenne pepper', '-', mass, _ ).
noun( 'cayman', 'caymans', count, _ ).
noun( 'ca~non', 'ca~nons', count, _ ).
noun( 'cc', 'cc', count, _ ).
noun( 'cease', 'ceases', count, _ ).
noun( 'cease-fire', 'cease-fires', count, _ ).
noun( 'cedar', 'cedars', both, _ ).
noun( 'cedilla', 'cedillas', count, _ ).
noun( 'ceiling', 'ceilings', count, _ ).
noun( 'celandine', 'celandines', count, _ ).
noun( 'celebrant', 'celebrants', count, _ ).
noun( 'celebration', 'celebrations', both, _ ).
noun( 'celebrity', 'celebrities', both, _ ).
noun( 'celerity', '-', mass, _ ).
noun( 'celery', '-', mass, _ ).
noun( 'celibacy', '-', mass, _ ).
noun( 'celibate', 'celibates', count, _ ).
noun( 'cell', 'cells', count, _ ).
noun( 'cellar', 'cellars', count, _ ).
noun( 'cellarage', 'cellarages', both, _ ).
noun( 'cellist', 'cellists', count, _ ).
noun( 'cello', 'cellos', count, _ ).
noun( 'cellophane', '-', mass, _ ).
noun( 'celluloid', '-', mass, _ ).
noun( 'cellulose', '-', mass, _ ).
noun( 'cement', '-', mass, _ ).
noun( 'cement-mixer', 'cement-mixers', count, _ ).
noun( 'cemetery', 'cemeteries', count, _ ).
noun( 'cenotaph', 'cenotaphs', count, _ ).
noun( 'censer', 'censers', count, _ ).
noun( 'censor', 'censors', count, _ ).
noun( 'censorship', 'censorships', count, _ ).
noun( 'censure', 'censures', both, _ ).
noun( 'census', 'censuses', count, _ ).
noun( 'cent', 'cents', count, _ ).
noun( 'centaur', 'centaurs', count, _ ).
noun( 'centenarian', 'centenarians', count, _ ).
noun( 'centenary', 'centenaries', count, _ ).
noun( 'centennial', 'centennials', count, _ ).
noun( 'centime', 'centimes', count, _ ).
noun( 'centimetre', 'centimetres', count, _ ).
noun( 'centipede', 'centipedes', count, _ ).
noun( 'central', 'centrals', count, _ ).
noun( 'centralization', 'centralizations', both, _ ).
noun( 'centre', 'centres', count, _ ).
noun( 'centre-bit', 'centre-bits', count, _ ).
noun( 'centre-board', 'centre-boards', count, _ ).
noun( 'centrepiece', 'centrepieces', count, _ ).
noun( 'centrifuge', 'centrifuges', count, _ ).
noun( 'centurion', 'centurions', count, _ ).
noun( 'century', 'centuries', count, _ ).
noun( 'ceramics', 'ceramics', mass, _ ).
noun( 'cereal', 'cereals', both, _ ).
noun( 'cerebration', '-', mass, _ ).
noun( 'ceremonial', 'ceremonials', both, _ ).
noun( 'ceremony', 'ceremonies', both, _ ).
noun( 'cerise', 'cerises', both, _ ).
noun( 'cert', 'certs', count, _ ).
noun( 'certainty', 'certainties', both, _ ).
noun( 'certificate', 'certificates', count, _ ).
noun( 'certification', 'certifications', both, _ ).
noun( 'certitude', '-', mass, _ ).
noun( 'cervix', 'cervixes', count, _ ).
noun( 'cessation', '-', mass, _ ).
noun( 'cession', 'cessions', both, _ ).
noun( 'cesspit', 'cesspits', count, _ ).
noun( 'cesspool', 'cesspools', count, _ ).
noun( 'cf', '-', proper, _ ).
noun( 'ch^ateau', 'ch^ateaux', count, _ ).
noun( 'chafe', 'chafes', count, _ ).
noun( 'chaff', '-', mass, _ ).
noun( 'chaffinch', 'chaffinches', count, _ ).
noun( 'chafing dish', 'chafing dishes', count, _ ).
noun( 'chagrin', '-', mass, _ ).
noun( 'chain', 'chains', count, _ ).
noun( 'chain-armour', '-', mass, _ ).
noun( 'chain-gang', 'chain-gangs', count, _ ).
noun( 'chain-letter', 'chain-letters', count, _ ).
noun( 'chain-mail', '-', mass, _ ).
noun( 'chain-smoker', 'chain-smokers', count, _ ).
noun( 'chain-stitch', 'chain-stitches', count, _ ).
noun( 'chain-store', 'chain-stores', count, _ ).
noun( 'chair', 'chairs', count, _ ).
noun( 'chair-lift', 'chair-lifts', count, _ ).
noun( 'chairman', 'chairmen', count, _ ).
noun( 'chairmanship', '-', mass, _ ).
noun( 'chaise', 'chaises', count, _ ).
noun( 'chaise longue', 'chaise longues', count, _ ).
noun( 'chalet', 'chalets', count, _ ).
noun( 'chalice', 'chalices', count, _ ).
noun( 'chalk', 'chalks', both, _ ).
noun( 'chalkpit', 'chalkpits', count, _ ).
noun( 'challenge', 'challenges', count, _ ).
noun( 'challenger', 'challengers', count, _ ).
noun( 'chamber', 'chambers', count, _ ).
noun( 'chamberlain', 'chamberlains', count, _ ).
noun( 'chambermaid', 'chambermaids', count, _ ).
noun( 'chamberpot', 'chamberpots', count, _ ).
noun( 'chameleon', 'chameleons', count, _ ).
noun( 'chammy-leather', 'chammy-leathers', count, _ ).
noun( 'chamois', 'chamoises', count, _ ).
noun( 'chamois-leather', 'chamois-leathers', count, _ ).
noun( 'champ', 'champs', count, _ ).
noun( 'champagne', 'champagnes', count, _ ).
noun( 'champion', 'champions', count, _ ).
noun( 'championship', 'championships', both, _ ).
noun( 'chance', 'chances', both, _ ).
noun( 'chancel', 'chancels', count, _ ).
noun( 'chancellery', 'chancelleries', count, _ ).
noun( 'chancellor', 'chancellors', count, _ ).
noun( 'chancery', 'chanceries', count, _ ).
noun( 'chandelier', 'chandeliers', count, _ ).
noun( 'chandler', 'chandlers', count, _ ).
noun( 'change', 'changes', both, _ ).
noun( 'changeableness', '-', mass, _ ).
noun( 'changeling', 'changelings', count, _ ).
noun( 'changeover', 'changeovers', count, _ ).
noun( 'channel', 'channels', count, _ ).
noun( 'chant', 'chants', count, _ ).
noun( 'chaos', '-', mass, _ ).
noun( 'chap', 'chaps', count, _ ).
noun( 'chapel', 'chapels', count, _ ).
noun( 'chapelgoer', 'chapelgoers', count, _ ).
noun( 'chaperon', 'chaperons', count, _ ).
noun( 'chaplain', 'chaplains', count, _ ).
noun( 'chaplaincy', 'chaplaincies', count, _ ).
noun( 'chaplet', 'chaplets', count, _ ).
noun( 'chapman', 'chapmen', count, _ ).
noun( 'chapter', 'chapters', count, _ ).
noun( 'chapterhouse', 'chapterhouses', count, _ ).
noun( 'char', 'chars', both, _ ).
noun( 'char`abanc', 'char`abancs', count, _ ).
noun( 'charabanc', 'charabancs', count, _ ).
noun( 'character', 'characters', both, _ ).
noun( 'characteristic', 'characteristics', count, _ ).
noun( 'characterization', '-', mass, _ ).
noun( 'charade', 'charades', count, _ ).
noun( 'charcoal', '-', mass, _ ).
noun( 'charcoal-burner', 'charcoal-burners', count, _ ).
noun( 'chard', 'chards', count, _ ).
noun( 'charg_e d\'affaires', 'charg_es d\'affaires', count, _ ).
noun( 'charge', 'charges', both, _ ).
noun( 'charge-account', 'charge-accounts', count, _ ).
noun( 'charge-sheet', 'charge-sheets', count, _ ).
noun( 'charger', 'chargers', count, _ ).
noun( 'chariot', 'chariots', count, _ ).
noun( 'charioteer', 'charioteers', count, _ ).
noun( 'charisma', 'charismas', count, _ ).
noun( 'charity', 'charities', both, _ ).
noun( 'charivari', '-', mass, _ ).
noun( 'charlady', 'charladies', count, _ ).
noun( 'charlatan', 'charlatans', count, _ ).
noun( 'charlock', '-', mass, _ ).
noun( 'charm', 'charms', both, _ ).
noun( 'charmer', 'charmers', count, _ ).
noun( 'charnel house', 'charnel houses', count, _ ).
noun( 'chart', 'charts', count, _ ).
noun( 'charter', 'charters', count, _ ).
noun( 'charter-party', 'charter-parties', count, _ ).
noun( 'chartreuse', '-', mass, _ ).
noun( 'charwoman', 'charwomen', count, _ ).
noun( 'chase', 'chases', count, _ ).
noun( 'chaser', 'chasers', count, _ ).
noun( 'chasm', 'chasms', count, _ ).
noun( 'chassis', 'chassis', count, _ ).
noun( 'chastisement', '-', mass, _ ).
noun( 'chastity', '-', mass, _ ).
noun( 'chasuble', 'chasubles', count, _ ).
noun( 'chat', 'chats', count, _ ).
noun( 'chatelaine', 'chatelaines', count, _ ).
noun( 'chattel', 'chattels', count, _ ).
noun( 'chatter', '-', mass, _ ).
noun( 'chatterbox', 'chatterboxes', count, _ ).
noun( 'chauffeur', 'chauffeurs', count, _ ).
noun( 'chauffeuse', 'chauffeuses', count, _ ).
noun( 'chauvinism', '-', mass, _ ).
noun( 'chauvinist', 'chauvinists', count, _ ).
noun( 'chaw', 'chaws', count, _ ).
noun( 'chaw-bacon', 'chaw-bacons', count, _ ).
noun( 'cheapness', '-', mass, _ ).
noun( 'cheat', 'cheats', count, _ ).
noun( 'check', 'checks', both, _ ).
noun( 'checkbook', 'checkbooks', count, _ ).
noun( 'checker', 'checkers', count, _ ).
noun( 'checkers', 'checkers', mass, _ ).
noun( 'checklist', 'checklists', count, _ ).
noun( 'checkmate', 'checkmates', count, _ ).
noun( 'checkout', 'checkouts', count, _ ).
noun( 'checkpoint', 'checkpoints', count, _ ).
noun( 'checkroom', 'checkrooms', count, _ ).
noun( 'checkup', 'checkups', count, _ ).
noun( 'cheek', 'cheeks', both, _ ).
noun( 'cheekbone', 'cheekbones', count, _ ).
noun( 'cheep', 'cheeps', count, _ ).
noun( 'cheer', 'cheers', both, _ ).
noun( 'cheerfulness', '-', mass, _ ).
noun( 'cheering', '-', mass, _ ).
noun( 'cheerleader', 'cheerleaders', count, _ ).
noun( 'cheerlessness', '-', mass, _ ).
noun( 'cheese', 'cheeses', both, _ ).
noun( 'cheesecake', 'cheesecakes', both, _ ).
noun( 'cheesecloth', 'cheesecloths', both, _ ).
noun( 'cheetah', 'cheetahs', count, _ ).
noun( 'chef', 'chefs', count, _ ).
noun( 'chef-d\'oeuvre', 'chefs-d\'oeuvre', count, _ ).
noun( 'chemical', 'chemicals', count, _ ).
noun( 'chemise', 'chemises', count, _ ).
noun( 'chemist', 'chemists', count, _ ).
noun( 'chemistry', '-', mass, _ ).
noun( 'chemotherapy', '-', mass, _ ).
noun( 'chenille', '-', mass, _ ).
noun( 'cheque', 'cheques', count, _ ).
noun( 'chequebook', 'chequebooks', count, _ ).
noun( 'cheroot', 'cheroots', count, _ ).
noun( 'cherry', 'cherries', count, _ ).
noun( 'cherub', 'cherubs', count, _ ).
noun( 'chervil', '-', mass, _ ).
noun( 'chess', '-', mass, _ ).
noun( 'chessboard', 'chessboards', count, _ ).
noun( 'chessman', 'chessmen', count, _ ).
noun( 'chest', 'chests', count, _ ).
noun( 'chesterfield', 'chesterfields', count, _ ).
noun( 'chestnut', 'chestnuts', both, _ ).
noun( 'cheval glass', 'cheval glasses', count, _ ).
noun( 'chevron', 'chevrons', count, _ ).
noun( 'chew', 'chews', count, _ ).
noun( 'chewing-gum', '-', mass, _ ).
noun( 'chiaroscuro', '-', mass, _ ).
noun( 'chic', '-', mass, _ ).
noun( 'chicanery', 'chicaneries', both, _ ).
noun( 'chick', 'chicks', count, _ ).
noun( 'chicken', 'chickens', both, _ ).
noun( 'chicken-run', 'chicken-runs', count, _ ).
noun( 'chickenfeed', '-', mass, _ ).
noun( 'chickenpox', '-', mass, _ ).
noun( 'chickpea', 'chickpeas', count, _ ).
noun( 'chickweed', '-', mass, _ ).
noun( 'chicle', '-', mass, _ ).
noun( 'chicory', '-', mass, _ ).
noun( 'chief', 'chiefs', count, _ ).
noun( 'chieftain', 'chieftains', count, _ ).
noun( 'chieftaincy', 'chieftaincies', count, _ ).
noun( 'chiffon', '-', mass, _ ).
noun( 'chiffonier', 'chiffoniers', count, _ ).
noun( 'chignon', 'chignons', count, _ ).
noun( 'chilblain', 'chilblains', count, _ ).
noun( 'child', 'children', count, _ ).
noun( 'child\'s-play', '-', mass, _ ).
noun( 'child-bearing', '-', mass, _ ).
noun( 'childbirth', '-', mass, _ ).
noun( 'childhood', '-', mass, _ ).
noun( 'chill', 'chills', count, _ ).
noun( 'chilli', 'chillies', count, _ ).
noun( 'chilly', 'chillies', count, _ ).
noun( 'chimaera', 'chimaeras', count, _ ).
noun( 'chime', 'chimes', count, _ ).
noun( 'chimera', 'chimeras', count, _ ).
noun( 'chimney', 'chimneys', count, _ ).
noun( 'chimney-sweep', 'chimney-sweeps', count, _ ).
noun( 'chimneybreast', 'chimneybreasts', count, _ ).
noun( 'chimneypiece', 'chimneypieces', count, _ ).
noun( 'chimneypot', 'chimneypots', count, _ ).
noun( 'chimneystack', 'chimneystacks', count, _ ).
noun( 'chimneysweep', 'chimneysweeps', count, _ ).
noun( 'chimneysweeper', 'chimneysweepers', count, _ ).
noun( 'chimp', 'chimps', count, _ ).
noun( 'chimpanzee', 'chimpanzees', count, _ ).
noun( 'chin', 'chins', count, _ ).
noun( 'chin-strap', 'chin-straps', count, _ ).
noun( 'chin-wagging', '-', mass, _ ).
noun( 'china', '-', mass, _ ).
noun( 'china-closet', 'china-closets', count, _ ).
noun( 'chinaware', '-', mass, _ ).
noun( 'chinchilla', 'chinchillas', both, _ ).
noun( 'chine', 'chines', count, _ ).
noun( 'chink', 'chinks', count, _ ).
noun( 'chintz', '-', mass, _ ).
noun( 'chip', 'chips', count, _ ).
noun( 'chipboard', '-', mass, _ ).
noun( 'chipmunk', 'chipmunks', count, _ ).
noun( 'chiropodist', 'chiropodists', count, _ ).
noun( 'chiropody', '-', mass, _ ).
noun( 'chiropractor', 'chiropractors', count, _ ).
noun( 'chirp', 'chirps', count, _ ).
noun( 'chirpiness', '-', mass, _ ).
noun( 'chirrup', 'chirrups', count, _ ).
noun( 'chisel', 'chisels', count, _ ).
noun( 'chiseller', 'chisellers', count, _ ).
noun( 'chit', 'chits', count, _ ).
noun( 'chit-chat', '-', mass, _ ).
noun( 'chivalry', '-', mass, _ ).
noun( 'chive', 'chives', both, _ ).
noun( 'chloride', '-', mass, _ ).
noun( 'chlorination', '-', mass, _ ).
noun( 'chlorine', '-', mass, _ ).
noun( 'chloroform', '-', mass, _ ).
noun( 'chlorophyll', '-', mass, _ ).
noun( 'choc', 'chocs', count, _ ).
noun( 'choc-ice', 'choc-ices', count, _ ).
noun( 'chock', 'chocks', count, _ ).
noun( 'chocolate', 'chocolates', both, _ ).
noun( 'choice', 'choices', both, _ ).
noun( 'choir', 'choirs', count, _ ).
noun( 'choir-school', 'choir-schools', count, _ ).
noun( 'choirboy', 'choirboys', count, _ ).
noun( 'choke', 'chokes', count, _ ).
noun( 'choke-damp', '-', mass, _ ).
noun( 'choker', 'chokers', count, _ ).
noun( 'chokey', 'chokeys', count, _ ).
noun( 'choky', 'chokies', count, _ ).
noun( 'choler', 'cholers', count, _ ).
noun( 'cholera', '-', mass, _ ).
noun( 'chop', 'chops', count, _ ).
noun( 'chop suey', '-', mass, _ ).
noun( 'chop-house', 'chop-houses', count, _ ).
noun( 'chopper', 'choppers', count, _ ).
noun( 'chorale', 'chorales', count, _ ).
noun( 'chord', 'chords', count, _ ).
noun( 'chore', 'chores', count, _ ).
noun( 'choreographer', 'choreographers', count, _ ).
noun( 'choreography', '-', mass, _ ).
noun( 'chorister', 'choristers', count, _ ).
noun( 'chortle', 'chortles', count, _ ).
noun( 'chorus', 'choruses', count, _ ).
noun( 'chorus-girl', 'chorus-girls', count, _ ).
noun( 'chow', 'chows', count, _ ).
noun( 'chowder', 'chowders', count, _ ).
noun( 'christening', 'christenings', count, _ ).
noun( 'chrome', '-', mass, _ ).
noun( 'chromium', '-', mass, _ ).
noun( 'chromosome', 'chromosomes', count, _ ).
noun( 'chronicle', 'chronicles', count, _ ).
noun( 'chronicler', 'chroniclers', count, _ ).
noun( 'chronology', 'chronologies', both, _ ).
noun( 'chronometer', 'chronometers', count, _ ).
noun( 'chrysalis', 'chrysalises', count, _ ).
noun( 'chrysanthemum', 'chrysanthemums', count, _ ).
noun( 'chuck', 'chucks', count, _ ).
noun( 'chucker-out', 'chuckers-out', count, _ ).
noun( 'chuckle', 'chuckles', count, _ ).
noun( 'chug', 'chugs', count, _ ).
noun( 'chukker', 'chukkers', count, _ ).
noun( 'chum', 'chums', count, _ ).
noun( 'chump', 'chumps', count, _ ).
noun( 'chunk', 'chunks', count, _ ).
noun( 'church', 'churches', both, _ ).
noun( 'churchgoer', 'churchgoers', count, _ ).
noun( 'churchman', 'churchmen', count, _ ).
noun( 'churchwarden', 'churchwardens', count, _ ).
noun( 'churchyard', 'churchyards', count, _ ).
noun( 'churl', 'churls', count, _ ).
noun( 'churn', 'churns', count, _ ).
noun( 'chute', 'chutes', count, _ ).
noun( 'chutney', 'chutneys', both, _ ).
noun( 'cicada', 'cicadas', count, _ ).
noun( 'cicala', 'cicalas', count, _ ).
noun( 'cicatrice', 'cicatrices', count, _ ).
noun( 'cicatrix', 'cicatrices', count, _ ).
noun( 'cicerone', '-', count, _ ).
noun( 'cider', 'ciders', both, _ ).
noun( 'ciderpress', 'ciderpresses', count, _ ).
noun( 'cif', '-', proper, _ ).
noun( 'cigar', 'cigars', count, _ ).
noun( 'cigarette', 'cigarettes', count, _ ).
noun( 'cigarette-case', 'cigarette-cases', count, _ ).
noun( 'cigarette-holder', 'cigarette-holders', count, _ ).
noun( 'cigarette-paper', 'cigarette-papers', count, _ ).
noun( 'cinch', '-', count, _ ).
noun( 'cinchona', 'cinchonas', count, _ ).
noun( 'cincture', 'cinctures', count, _ ).
noun( 'cinder', 'cinders', count, _ ).
noun( 'cinder-track', 'cinder-tracks', count, _ ).
noun( 'cine-camera', 'cine-cameras', count, _ ).
noun( 'cine-film', 'cine-films', count, _ ).
noun( 'cine-projector', 'cine-projectors', count, _ ).
noun( 'cinema', 'cinemas', count, _ ).
noun( 'cinematography', '-', mass, _ ).
noun( 'cinnamon', '-', mass, _ ).
noun( 'cinquefoil', 'cinquefoils', count, _ ).
noun( 'cipher', 'ciphers', count, _ ).
noun( 'circle', 'circles', count, _ ).
noun( 'circlet', 'circlets', count, _ ).
noun( 'circuit', 'circuits', count, _ ).
noun( 'circular', 'circulars', count, _ ).
noun( 'circularity', '-', mass, _ ).
noun( 'circulation', 'circulations', both, _ ).
noun( 'circumcision', 'circumcisions', count, _ ).
noun( 'circumference', 'circumferences', count, _ ).
noun( 'circumflex', 'circumflexes', count, _ ).
noun( 'circumlocution', 'circumlocutions', both, _ ).
noun( 'circumnavigation', 'circumnavigations', count, _ ).
noun( 'circumscription', 'circumscriptions', both, _ ).
noun( 'circumspection', '-', mass, _ ).
noun( 'circumstance', 'circumstances', count, _ ).
noun( 'circumvention', 'circumventions', count, _ ).
noun( 'circus', 'circuses', count, _ ).
noun( 'cirrhosis', '-', mass, _ ).
noun( 'cirrus', '-', mass, _ ).
noun( 'cissy', 'cissies', count, _ ).
noun( 'cistern', 'cisterns', count, _ ).
noun( 'citadel', 'citadels', count, _ ).
noun( 'citation', 'citations', both, _ ).
noun( 'citizen', 'citizens', count, _ ).
noun( 'citizenship', 'citizenships', count, _ ).
noun( 'citron', 'citrons', count, _ ).
noun( 'citrus', 'citruses', count, _ ).
noun( 'city', 'cities', count, _ ).
noun( 'civet', 'civets', both, _ ).
noun( 'civet-cat', 'civet-cats', count, _ ).
noun( 'civics', 'civics', mass, _ ).
noun( 'civilian', 'civilians', count, _ ).
noun( 'civility', 'civilities', both, _ ).
noun( 'civilization', 'civilizations', both, _ ).
noun( 'clack', 'clacks', count, _ ).
noun( 'claim', 'claims', both, _ ).
noun( 'claimant', 'claimants', count, _ ).
noun( 'clairvoyance', '-', mass, _ ).
noun( 'clairvoyant', 'clairvoyants', count, _ ).
noun( 'clam', 'clams', count, _ ).
noun( 'clambake', 'clambakes', count, _ ).
noun( 'clamber', 'clambers', count, _ ).
noun( 'clamour', 'clamours', both, _ ).
noun( 'clamp', 'clamps', count, _ ).
noun( 'clamp-down', 'clamp-downs', count, _ ).
noun( 'clan', 'clans', count, _ ).
noun( 'clang', 'clangs', count, _ ).
noun( 'clanger', 'clangers', count, _ ).
noun( 'clangour', '-', mass, _ ).
noun( 'clank', 'clanks', count, _ ).
noun( 'clansman', 'clansmen', count, _ ).
noun( 'clap', 'claps', both, _ ).
noun( 'clapboard', 'clapboards', count, _ ).
noun( 'clapper', 'clappers', count, _ ).
noun( 'clapperboard', 'clapperboards', count, _ ).
noun( 'claptrap', '-', mass, _ ).
noun( 'claque', 'claques', count, _ ).
noun( 'claret', 'clarets', both, _ ).
noun( 'clarification', '-', mass, _ ).
noun( 'clarinet', 'clarinets', count, _ ).
noun( 'clarinetist', 'clarinetists', count, _ ).
noun( 'clarinettist', 'clarinettists', count, _ ).
noun( 'clarion', 'clarions', count, _ ).
noun( 'clarity', '-', mass, _ ).
noun( 'clash', 'clashes', count, _ ).
noun( 'clasp', 'clasps', count, _ ).
noun( 'clasp-knife', 'clasp-knives', count, _ ).
noun( 'class', 'classes', both, _ ).
noun( 'class-feeling', '-', mass, _ ).
noun( 'class-fellow', 'class-fellows', count, _ ).
noun( 'class-list', 'class-lists', count, _ ).
noun( 'class-warfare', '-', mass, _ ).
noun( 'classic', 'classics', count, _ ).
noun( 'classicism', '-', mass, _ ).
noun( 'classicist', 'classicists', count, _ ).
noun( 'classics', 'classics', mass, _ ).
noun( 'classification', 'classifications', both, _ ).
noun( 'classmate', 'classmates', count, _ ).
noun( 'classroom', 'classrooms', count, _ ).
noun( 'clatter', '-', count, _ ).
noun( 'clause', 'clauses', count, _ ).
noun( 'claustrophobia', '-', mass, _ ).
noun( 'clavichord', 'clavichords', count, _ ).
noun( 'clavicle', 'clavicles', count, _ ).
noun( 'claw', 'claws', count, _ ).
noun( 'clawback', '-', mass, _ ).
noun( 'clawhammer', 'clawhammers', count, _ ).
noun( 'clay', '-', mass, _ ).
noun( 'clean', 'cleans', count, _ ).
noun( 'clean-up', 'clean-ups', count, _ ).
noun( 'cleaner', 'cleaners', count, _ ).
noun( 'cleanliness', '-', mass, _ ).
noun( 'cleanser', 'cleansers', both, _ ).
noun( 'clear', 'clears', count, _ ).
noun( 'clearance', 'clearances', both, _ ).
noun( 'clearing', 'clearings', count, _ ).
noun( 'clearing-house', 'clearing-houses', count, _ ).
noun( 'clearness', '-', mass, _ ).
noun( 'clearway', 'clearways', count, _ ).
noun( 'cleat', 'cleats', count, _ ).
noun( 'cleavage', 'cleavages', count, _ ).
noun( 'cleaver', 'cleavers', count, _ ).
noun( 'clef', 'clefs', count, _ ).
noun( 'cleft', 'clefts', count, _ ).
noun( 'clematis', '-', mass, _ ).
noun( 'clemency', '-', mass, _ ).
noun( 'clerestory', 'clerestories', count, _ ).
noun( 'clergy', 'clergies', count, _ ).
noun( 'clergyman', 'clergymen', count, _ ).
noun( 'cleric', 'clerics', count, _ ).
noun( 'clerihew', 'clerihews', count, _ ).
noun( 'clerk', 'clerks', count, _ ).
noun( 'cleverness', '-', mass, _ ).
noun( 'clew', 'clews', count, _ ).
noun( 'clich_e', 'clich_es', count, _ ).
noun( 'click', 'clicks', count, _ ).
noun( 'client', 'clients', count, _ ).
noun( 'clientele', 'clienteles', count, _ ).
noun( 'cliff', 'cliffs', count, _ ).
noun( 'cliff-hanger', 'cliff-hangers', count, _ ).
noun( 'climacteric', 'climacterics', count, _ ).
noun( 'climate', 'climates', count, _ ).
noun( 'climatology', '-', mass, _ ).
noun( 'climax', 'climaxes', count, _ ).
noun( 'climb', 'climbs', count, _ ).
noun( 'climb-down', 'climb-downs', count, _ ).
noun( 'climber', 'climbers', count, _ ).
noun( 'clime', 'climes', count, _ ).
noun( 'clinch', 'clinches', count, _ ).
noun( 'clincher', 'clinchers', count, _ ).
noun( 'clinic', 'clinics', count, _ ).
noun( 'clink', 'clinks', count, _ ).
noun( 'clinker', 'clinkers', both, _ ).
noun( 'clip', 'clips', count, _ ).
noun( 'clip-joint', 'clip-joints', count, _ ).
noun( 'clipper', 'clippers', count, _ ).
noun( 'clipping', 'clippings', count, _ ).
noun( 'clique', 'cliques', count, _ ).
noun( 'clitoris', 'clitorises', count, _ ).
noun( 'cloak', 'cloaks', count, _ ).
noun( 'cloakroom', 'cloakrooms', count, _ ).
noun( 'clobber', 'clobbers', count, _ ).
noun( 'cloche', 'cloches', count, _ ).
noun( 'clock', 'clocks', count, _ ).
noun( 'clock-dial', 'clock-dials', count, _ ).
noun( 'clock-face', 'clock-faces', count, _ ).
noun( 'clock-golf', '-', mass, _ ).
noun( 'clock-tower', 'clock-towers', count, _ ).
noun( 'clock-watching', '-', mass, _ ).
noun( 'clockwork', 'clockworks', count, _ ).
noun( 'clod', 'clods', count, _ ).
noun( 'clodhopper', 'clodhoppers', count, _ ).
noun( 'clog', 'clogs', count, _ ).
noun( 'clog-dance', 'clog-dances', count, _ ).
noun( 'cloisonn_e', '-', mass, _ ).
noun( 'cloister', 'cloisters', count, _ ).
noun( 'clone', 'clones', count, _ ).
noun( 'close', 'closes', count, _ ).
noun( 'close', '-', count, _ ).
noun( 'close-down', 'close-downs', count, _ ).
noun( 'close-up', 'close-ups', count, _ ).
noun( 'closeness', '-', mass, _ ).
noun( 'closet', 'closets', count, _ ).
noun( 'closure', 'closures', both, _ ).
noun( 'clot', 'clots', count, _ ).
noun( 'cloth', 'cloths', both, _ ).
noun( 'clothes-basket', 'clothes-baskets', count, _ ).
noun( 'clothes-hanger', 'clothes-hangers', count, _ ).
noun( 'clothes-moth', 'clothes-moths', count, _ ).
noun( 'clothes-peg', 'clothes-pegs', count, _ ).
noun( 'clothes-pin', 'clothes-pins', count, _ ).
noun( 'clotheshorse', 'clotheshorses', count, _ ).
noun( 'clothesline', 'clotheslines', count, _ ).
noun( 'clothier', 'clothiers', count, _ ).
noun( 'clothing', '-', mass, _ ).
noun( 'cloud', 'clouds', both, _ ).
noun( 'cloud-bank', 'cloud-banks', count, _ ).
noun( 'cloud-cuckoo-land', '-', count, _ ).
noun( 'cloudburst', 'cloudbursts', count, _ ).
noun( 'clout', 'clouts', count, _ ).
noun( 'clove', 'cloves', count, _ ).
noun( 'clove hitch', 'clove hitches', count, _ ).
noun( 'clover', '-', mass, _ ).
noun( 'cloverleaf', 'cloverleaves', count, _ ).
noun( 'clown', 'clowns', count, _ ).
noun( 'club', 'clubs', count, _ ).
noun( 'clubfoot', 'clubfeet', count, _ ).
noun( 'clubhouse', 'clubhouses', count, _ ).
noun( 'cluck', 'clucks', count, _ ).
noun( 'clue', 'clues', count, _ ).
noun( 'clump', 'clumps', count, _ ).
noun( 'clumsiness', '-', mass, _ ).
noun( 'clunk', 'clunks', count, _ ).
noun( 'cluster', 'clusters', count, _ ).
noun( 'clutch', 'clutches', count, _ ).
noun( 'clutter', 'clutters', both, _ ).
noun( 'cm', 'cm', count, _ ).
noun( 'co-ed', 'co-eds', count, _ ).
noun( 'co-op', 'co-ops', count, _ ).
noun( 'co-respondent', 'co-respondents', count, _ ).
noun( 'co-star', 'co-stars', count, _ ).
noun( 'coach', 'coaches', count, _ ).
noun( 'coach-builder', 'coach-builders', count, _ ).
noun( 'coachman', 'coachmen', count, _ ).
noun( 'coagulation', 'coagulations', both, _ ).
noun( 'coal', 'coals', both, _ ).
noun( 'coal-gas', '-', mass, _ ).
noun( 'coal-hole', 'coal-holes', count, _ ).
noun( 'coal-house', 'coal-houses', count, _ ).
noun( 'coal-scuttle', 'coal-scuttles', count, _ ).
noun( 'coal-seam', 'coal-seams', count, _ ).
noun( 'coal-tar', '-', mass, _ ).
noun( 'coalescence', 'coalescences', both, _ ).
noun( 'coalface', 'coalfaces', count, _ ).
noun( 'coalfield', 'coalfields', count, _ ).
noun( 'coaling-station', 'coaling-stations', count, _ ).
noun( 'coalition', 'coalitions', both, _ ).
noun( 'coalman', 'coalmen', count, _ ).
noun( 'coalmine', 'coalmines', count, _ ).
noun( 'coalpit', 'coalpits', count, _ ).
noun( 'coaming', 'coamings', count, _ ).
noun( 'coarseness', '-', mass, _ ).
noun( 'coast', 'coasts', count, _ ).
noun( 'coaster', 'coasters', count, _ ).
noun( 'coastguard', 'coastguards', count, _ ).
noun( 'coastline', 'coastlines', count, _ ).
noun( 'coat', 'coats', count, _ ).
noun( 'coat-hanger', 'coat-hangers', count, _ ).
noun( 'coatee', 'coatees', count, _ ).
noun( 'coating', 'coatings', both, _ ).
noun( 'coaxing', 'coaxings', both, _ ).
noun( 'cob', 'cobs', count, _ ).
noun( 'cob-nut', 'cob-nuts', count, _ ).
noun( 'cobalt', '-', mass, _ ).
noun( 'cobber', 'cobbers', count, _ ).
noun( 'cobble', 'cobbles', count, _ ).
noun( 'cobbler', 'cobblers', count, _ ).
noun( 'cobblestone', 'cobblestones', count, _ ).
noun( 'cobra', 'cobras', count, _ ).
noun( 'cobweb', 'cobwebs', count, _ ).
noun( 'cocaine', '-', mass, _ ).
noun( 'cochineal', '-', mass, _ ).
noun( 'cochlea', 'cochleas', count, _ ).
noun( 'cock', 'cocks', count, _ ).
noun( 'cock-a-doodle-doo', 'cock-a-doodle-doos', count, _ ).
noun( 'cock-crow', 'cock-crows', count, _ ).
noun( 'cockade', 'cockades', count, _ ).
noun( 'cockatoo', 'cockatoos', count, _ ).
noun( 'cockchafer', 'cockchafers', count, _ ).
noun( 'cocker', 'cockers', count, _ ).
noun( 'cockerel', 'cockerels', count, _ ).
noun( 'cockfighting', '-', mass, _ ).
noun( 'cockhorse', 'cockhorses', count, _ ).
noun( 'cockle', 'cockles', count, _ ).
noun( 'cockleshell', 'cockleshells', count, _ ).
noun( 'cockney', 'cockneys', count, _ ).
noun( 'cockpit', 'cockpits', count, _ ).
noun( 'cockroach', 'cockroaches', count, _ ).
noun( 'cockscomb', 'cockscombs', count, _ ).
noun( 'cocktail', 'cocktails', count, _ ).
noun( 'cockup', 'cockups', count, _ ).
noun( 'coco', 'cocos', count, _ ).
noun( 'cocoa', '-', mass, _ ).
noun( 'coconut', 'coconuts', count, _ ).
noun( 'cocoon', 'cocoons', count, _ ).
noun( 'cocotte', 'cocottes', count, _ ).
noun( 'cod', 'cods', both, _ ).
noun( 'cod-liver oil', '-', mass, _ ).
noun( 'coda', 'codas', count, _ ).
noun( 'code', 'codes', both, _ ).
noun( 'codeine', '-', mass, _ ).
noun( 'codex', 'codices', count, _ ).
noun( 'codfish', 'codfish', count, _ ).
noun( 'codger', 'codgers', count, _ ).
noun( 'codicil', 'codicils', count, _ ).
noun( 'codification', 'codifications', count, _ ).
noun( 'codling', 'codlings', count, _ ).
noun( 'codpiece', 'codpieces', count, _ ).
noun( 'coeducation', '-', mass, _ ).
noun( 'coefficient', 'coefficients', count, _ ).
noun( 'coercion', '-', mass, _ ).
noun( 'coeval', 'coevals', count, _ ).
noun( 'coexistence', '-', mass, _ ).
noun( 'coffee', 'coffees', both, _ ).
noun( 'coffee-house', 'coffee-houses', count, _ ).
noun( 'coffee-mill', 'coffee-mills', count, _ ).
noun( 'coffee-stall', 'coffee-stalls', count, _ ).
noun( 'coffer', 'coffers', count, _ ).
noun( 'coffer-dam', 'coffer-dams', count, _ ).
noun( 'coffin', 'coffins', count, _ ).
noun( 'cog', 'cogs', count, _ ).
noun( 'cogency', '-', mass, _ ).
noun( 'cogitation', '-', mass, _ ).
noun( 'cognac', '-', mass, _ ).
noun( 'cognate', 'cognates', count, _ ).
noun( 'cognition', '-', mass, _ ).
noun( 'cognizance', '-', mass, _ ).
noun( 'cognomen', 'cognomens', count, _ ).
noun( 'cogwheel', 'cogwheels', count, _ ).
noun( 'cohabitation', 'cohabitations', both, _ ).
noun( 'coherence', '-', mass, _ ).
noun( 'coherency', '-', mass, _ ).
noun( 'cohesion', '-', mass, _ ).
noun( 'cohort', 'cohorts', count, _ ).
noun( 'coif', 'coifs', count, _ ).
noun( 'coiffeur', 'coiffeurs', count, _ ).
noun( 'coiffure', 'coiffures', count, _ ).
noun( 'coign', 'coigns', count, _ ).
noun( 'coil', 'coils', count, _ ).
noun( 'coin', 'coins', both, _ ).
noun( 'coinage', 'coinages', both, _ ).
noun( 'coincidence', 'coincidences', both, _ ).
noun( 'coiner', 'coiners', count, _ ).
noun( 'coir', '-', mass, _ ).
noun( 'coition', '-', mass, _ ).
noun( 'coitus', '-', mass, _ ).
noun( 'coke', 'cokes', both, _ ).
noun( 'col', 'cols', count, _ ).
noun( 'cola', 'colas', count, _ ).
noun( 'colander', 'colanders', count, _ ).
noun( 'cold', 'colds', both, _ ).
noun( 'coldness', '-', mass, _ ).
noun( 'coleslaw', '-', mass, _ ).
noun( 'colic', '-', mass, _ ).
noun( 'colitis', '-', mass, _ ).
noun( 'collaboration', '-', mass, _ ).
noun( 'collaborationist', 'collaborationists', count, _ ).
noun( 'collaborator', 'collaborators', count, _ ).
noun( 'collage', 'collages', both, _ ).
noun( 'collapse', 'collapses', count, _ ).
noun( 'collar', 'collars', count, _ ).
noun( 'collarbone', 'collarbones', count, _ ).
noun( 'collateral', '-', mass, _ ).
noun( 'collation', 'collations', count, _ ).
noun( 'colleague', 'colleagues', count, _ ).
noun( 'collect', 'collects', count, _ ).
noun( 'collection', 'collections', both, _ ).
noun( 'collectivization', 'collectivizations', count, _ ).
noun( 'collector', 'collectors', count, _ ).
noun( 'colleen', 'colleens', count, _ ).
noun( 'college', 'colleges', both, _ ).
noun( 'collie', 'collies', count, _ ).
noun( 'collier', 'colliers', count, _ ).
noun( 'colliery', 'collieries', count, _ ).
noun( 'collision', 'collisions', both, _ ).
noun( 'collocation', 'collocations', both, _ ).
noun( 'colloquialism', 'colloquialisms', count, _ ).
noun( 'colloquy', 'colloquies', both, _ ).
noun( 'collusion', '-', mass, _ ).
noun( 'colon', 'colons', count, _ ).
noun( 'colonel', 'colonels', count, _ ).
noun( 'colonial', 'colonials', count, _ ).
noun( 'colonialism', '-', mass, _ ).
noun( 'colonialist', 'colonialists', count, _ ).
noun( 'colonist', 'colonists', count, _ ).
noun( 'colonization', '-', mass, _ ).
noun( 'colonizer', 'colonizers', count, _ ).
noun( 'colonnade', 'colonnades', count, _ ).
noun( 'colony', 'colonies', count, _ ).
noun( 'coloratura', '-', mass, _ ).
noun( 'colossus', 'colossi', count, _ ).
noun( 'colour', 'colours', both, _ ).
noun( 'colour-bar', 'colour-bars', count, _ ).
noun( 'colour-wash', 'colour-washes', count, _ ).
noun( 'colouring', '-', mass, _ ).
noun( 'colt', 'colts', count, _ ).
noun( 'columbine', 'columbines', count, _ ).
noun( 'column', 'columns', count, _ ).
noun( 'columnist', 'columnists', count, _ ).
noun( 'coma', 'comas', count, _ ).
noun( 'comb', 'combs', count, _ ).
noun( 'comb-out', 'comb-outs', count, _ ).
noun( 'combat', 'combats', count, _ ).
noun( 'combatant', 'combatants', count, _ ).
noun( 'combination', 'combinations', both, _ ).
noun( 'combination-lock', 'combination-locks', count, _ ).
noun( 'combine', 'combines', count, _ ).
noun( 'combustible', 'combustibles', count, _ ).
noun( 'combustion', '-', mass, _ ).
noun( 'come-on', 'come-ons', count, _ ).
noun( 'comeback', 'comebacks', count, _ ).
noun( 'comedian', 'comedians', count, _ ).
noun( 'comedienne', 'comediennes', count, _ ).
noun( 'comedown', 'comedowns', count, _ ).
noun( 'comedy', 'comedies', both, _ ).
noun( 'comeliness', '-', mass, _ ).
noun( 'comer', 'comers', count, _ ).
noun( 'comestible', 'comestibles', count, _ ).
noun( 'comet', 'comets', count, _ ).
noun( 'comfit', 'comfits', count, _ ).
noun( 'comfort', 'comforts', both, _ ).
noun( 'comforter', 'comforters', count, _ ).
noun( 'comfrey', '-', mass, _ ).
noun( 'comic', 'comics', count, _ ).
noun( 'coming', 'comings', count, _ ).
noun( 'comity', '-', mass, _ ).
noun( 'comma', 'commas', count, _ ).
noun( 'command', 'commands', both, _ ).
noun( 'commandant', 'commandants', count, _ ).
noun( 'commander', 'commanders', count, _ ).
noun( 'commandment', 'commandments', count, _ ).
noun( 'commando', 'commandos', count, _ ).
noun( 'commemoration', 'commemorations', both, _ ).
noun( 'commencement', 'commencements', count, _ ).
noun( 'commendation', 'commendations', both, _ ).
noun( 'comment', 'comments', both, _ ).
noun( 'commentary', 'commentaries', count, _ ).
noun( 'commentator', 'commentators', count, _ ).
noun( 'commerce', '-', mass, _ ).
noun( 'commercial', 'commercials', count, _ ).
noun( 'commercialism', '-', mass, _ ).
noun( 'commination', 'comminations', both, _ ).
noun( 'commiseration', 'commiserations', both, _ ).
noun( 'commissar', 'commissars', count, _ ).
noun( 'commissariat', 'commissariats', count, _ ).
noun( 'commissary', 'commissaries', count, _ ).
noun( 'commission', 'commissions', both, _ ).
noun( 'commissionaire', 'commissionaires', count, _ ).
noun( 'commissioner', 'commissioners', count, _ ).
noun( 'commital', 'commitals', count, _ ).
noun( 'commitment', 'commitments', both, _ ).
noun( 'committee', 'committees', count, _ ).
noun( 'commode', 'commodes', count, _ ).
noun( 'commodity', 'commodities', count, _ ).
noun( 'commodore', 'commodores', count, _ ).
noun( 'common', 'commons', count, _ ).
noun( 'common-room', 'common-rooms', count, _ ).
noun( 'commonalty', 'commonalties', count, _ ).
noun( 'commoner', 'commoners', count, _ ).
noun( 'commonplace', 'commonplaces', count, _ ).
noun( 'commonwealth', 'commonwealths', count, _ ).
noun( 'commotion', 'commotions', both, _ ).
noun( 'commune', 'communes', count, _ ).
noun( 'communicant', 'communicants', count, _ ).
noun( 'communication', 'communications', both, _ ).
noun( 'communion', 'communions', both, _ ).
noun( 'communiqu_e', 'communiqu_es', count, _ ).
noun( 'communism', '-', mass, _ ).
noun( 'communist', 'communists', count, _ ).
noun( 'community', 'communities', both, _ ).
noun( 'commutation', 'commutations', both, _ ).
noun( 'commutator', 'commutators', count, _ ).
noun( 'commuter', 'commuters', count, _ ).
noun( 'comp`ere', 'comp`eres', count, _ ).
noun( 'compact', 'compacts', count, _ ).
noun( 'compactness', '-', mass, _ ).
noun( 'companion', 'companions', count, _ ).
noun( 'companionship', '-', mass, _ ).
noun( 'companionway', 'companionways', count, _ ).
noun( 'company', 'companies', both, _ ).
noun( 'comparability', '-', mass, _ ).
noun( 'comparative', 'comparatives', count, _ ).
noun( 'compare', 'compares', count, _ ).
noun( 'comparison', 'comparisons', both, _ ).
noun( 'compartment', 'compartments', count, _ ).
noun( 'compass', 'compasses', count, _ ).
noun( 'compassion', '-', mass, _ ).
noun( 'compatibility', '-', mass, _ ).
noun( 'compatriot', 'compatriots', count, _ ).
noun( 'compeer', 'compeers', count, _ ).
noun( 'compendium', 'compendiums', count, _ ).
noun( 'compensation', 'compensations', both, _ ).
noun( 'competence', '-', mass, _ ).
noun( 'competition', 'competitions', both, _ ).
noun( 'competitiveness', '-', mass, _ ).
noun( 'competitor', 'competitors', count, _ ).
noun( 'compilation', 'compilations', both, _ ).
noun( 'compiler', 'compilers', count, _ ).
noun( 'complacence', '-', mass, _ ).
noun( 'complacency', '-', mass, _ ).
noun( 'complainant', 'complainants', count, _ ).
noun( 'complaint', 'complaints', both, _ ).
noun( 'complaisance', '-', mass, _ ).
noun( 'complement', 'complements', count, _ ).
noun( 'completeness', '-', mass, _ ).
noun( 'completion', '-', mass, _ ).
noun( 'complex', 'complexes', count, _ ).
noun( 'complexion', 'complexions', count, _ ).
noun( 'complexity', 'complexities', both, _ ).
noun( 'compliance', '-', mass, _ ).
noun( 'complication', 'complications', count, _ ).
noun( 'complicity', '-', mass, _ ).
noun( 'compliment', 'compliments', count, _ ).
noun( 'complin', 'complins', count, _ ).
noun( 'compline', 'complines', count, _ ).
noun( 'component', 'components', count, _ ).
noun( 'comportment', 'comportments', count, _ ).
noun( 'composer', 'composers', count, _ ).
noun( 'composition', 'compositions', both, _ ).
noun( 'compositor', 'compositors', count, _ ).
noun( 'compost', '-', mass, _ ).
noun( 'composure', '-', mass, _ ).
noun( 'compote', 'compotes', both, _ ).
noun( 'compound', 'compounds', count, _ ).
noun( 'comprehensibility', '-', mass, _ ).
noun( 'comprehension', 'comprehensions', both, _ ).
noun( 'comprehensiveness', '-', mass, _ ).
noun( 'compress', 'compresses', count, _ ).
noun( 'compression', '-', mass, _ ).
noun( 'compressor', 'compressors', count, _ ).
noun( 'compromise', 'compromises', both, _ ).
noun( 'comptroller', 'comptrollers', count, _ ).
noun( 'compulsion', '-', mass, _ ).
noun( 'compunction', '-', mass, _ ).
noun( 'computation', 'computations', both, _ ).
noun( 'computer', 'computers', count, _ ).
noun( 'comrade', 'comrades', count, _ ).
noun( 'comradeship', 'comradeships', count, _ ).
noun( 'con', 'cons', count, _ ).
noun( 'con-man', 'con-men', count, _ ).
noun( 'concatenation', 'concatenations', both, _ ).
noun( 'concavity', 'concavities', both, _ ).
noun( 'concealment', '-', mass, _ ).
noun( 'conceit', 'conceits', both, _ ).
noun( 'concentrate', 'concentrates', count, _ ).
noun( 'concentration', 'concentrations', both, _ ).
noun( 'concept', 'concepts', count, _ ).
noun( 'conception', 'conceptions', both, _ ).
noun( 'concern', 'concerns', both, _ ).
noun( 'concert', 'concerts', both, _ ).
noun( 'concert-hall', 'concert-halls', count, _ ).
noun( 'concertina', 'concertinas', count, _ ).
noun( 'concerto', 'concertos', count, _ ).
noun( 'concession', 'concessions', both, _ ).
noun( 'concessionaire', 'concessionaires', count, _ ).
noun( 'conch', 'conches', count, _ ).
noun( 'conchology', '-', mass, _ ).
noun( 'concierge', 'concierges', count, _ ).
noun( 'conciliation', '-', mass, _ ).
noun( 'conciseness', '-', mass, _ ).
noun( 'conclave', 'conclaves', count, _ ).
noun( 'conclusion', 'conclusions', count, _ ).
noun( 'concoction', 'concoctions', both, _ ).
noun( 'concomitant', 'concomitants', count, _ ).
noun( 'concord', 'concords', both, _ ).
noun( 'concordance', 'concordances', both, _ ).
noun( 'concordat', 'concordats', count, _ ).
noun( 'concourse', 'concourses', count, _ ).
noun( 'concrete', '-', mass, _ ).
noun( 'concretion', 'concretions', both, _ ).
noun( 'concubine', 'concubines', count, _ ).
noun( 'concupiscence', '-', mass, _ ).
noun( 'concurrence', 'concurrences', both, _ ).
noun( 'concussion', 'concussions', both, _ ).
noun( 'condemnation', 'condemnations', both, _ ).
noun( 'condensation', 'condensations', both, _ ).
noun( 'condenser', 'condensers', count, _ ).
noun( 'condescension', 'condescensions', both, _ ).
noun( 'condiment', 'condiments', both, _ ).
noun( 'condition', 'conditions', count, _ ).
noun( 'conditioner', 'conditioners', count, _ ).
noun( 'condolence', 'condolences', count, _ ).
noun( 'condominium', 'condominiums', count, _ ).
noun( 'condonation', 'condonations', count, _ ).
noun( 'condor', 'condors', count, _ ).
noun( 'conduct', '-', mass, _ ).
noun( 'conduction', '-', mass, _ ).
noun( 'conductivity', 'conductivities', both, _ ).
noun( 'conductor', 'conductors', count, _ ).
noun( 'conductress', 'conductresses', count, _ ).
noun( 'conduit', 'conduits', count, _ ).
noun( 'cone', 'cones', count, _ ).
noun( 'coney', 'coneys', count, _ ).
noun( 'confab', 'confabs', count, _ ).
noun( 'confabulation', 'confabulations', count, _ ).
noun( 'confection', 'confections', both, _ ).
noun( 'confectioner', 'confectioners', count, _ ).
noun( 'confectionery', 'confectioneries', both, _ ).
noun( 'confederacy', 'confederacies', count, _ ).
noun( 'confederate', 'confederates', count, _ ).
noun( 'confederation', 'confederations', both, _ ).
noun( 'conference', 'conferences', both, _ ).
noun( 'conferment', 'conferments', count, _ ).
noun( 'confession', 'confessions', both, _ ).
noun( 'confessional', 'confessionals', count, _ ).
noun( 'confessor', 'confessors', count, _ ).
noun( 'confetti', 'confetti', mass, _ ).
noun( 'confidant', 'confidants', count, _ ).
noun( 'confidence', 'confidences', both, _ ).
noun( 'confidentiality', 'confidentialities', both, _ ).
noun( 'configuration', 'configurations', count, _ ).
noun( 'confinement', 'confinements', both, _ ).
noun( 'confirmation', 'confirmations', both, _ ).
noun( 'confiscation', 'confiscations', both, _ ).
noun( 'conflagration', 'conflagrations', count, _ ).
noun( 'conflict', 'conflicts', both, _ ).
noun( 'confluence', 'confluences', count, _ ).
noun( 'conformation', 'conformations', count, _ ).
noun( 'conformist', 'conformists', count, _ ).
noun( 'conformity', '-', mass, _ ).
noun( 'confr`ere', 'confr`eres', count, _ ).
noun( 'confrontation', 'confrontations', both, _ ).
noun( 'confusion', '-', mass, _ ).
noun( 'confutation', 'confutations', count, _ ).
noun( 'cong_e', 'cong_es', count, _ ).
noun( 'conger', 'congers', count, _ ).
noun( 'conger-eel', 'conger-eels', count, _ ).
noun( 'congestion', '-', mass, _ ).
noun( 'conglomerate', 'conglomerates', count, _ ).
noun( 'conglomeration', 'conglomerations', both, _ ).
noun( 'congratulation', 'congratulations', count, _ ).
noun( 'congregation', 'congregations', both, _ ).
noun( 'congress', 'congresses', count, _ ).
noun( 'congressman', 'congressmen', count, _ ).
noun( 'congresswoman', 'congresswomen', count, _ ).
noun( 'conifer', 'conifers', count, _ ).
noun( 'conjecture', 'conjectures', both, _ ).
noun( 'conjugation', 'conjugations', both, _ ).
noun( 'conjunction', 'conjunctions', both, _ ).
noun( 'conjunctive', 'conjunctives', count, _ ).
noun( 'conjuncture', 'conjunctures', count, _ ).
noun( 'conjuration', 'conjurations', count, _ ).
noun( 'conjurer', 'conjurers', count, _ ).
noun( 'conjuror', 'conjurors', count, _ ).
noun( 'conk', 'conks', count, _ ).
noun( 'conker', 'conkers', count, _ ).
noun( 'connection', 'connections', both, _ ).
noun( 'connective', 'connectives', count, _ ).
noun( 'connexion', 'connexions', both, _ ).
noun( 'connivance', '-', mass, _ ).
noun( 'connoisseur', 'connoisseurs', count, _ ).
noun( 'connotation', 'connotations', count, _ ).
noun( 'conqueror', 'conquerors', count, _ ).
noun( 'conquest', 'conquests', both, _ ).
noun( 'conquistador', 'conquistadors', count, _ ).
noun( 'consanguinity', '-', mass, _ ).
noun( 'conscience', 'consciences', both, _ ).
noun( 'conscientiousness', '-', mass, _ ).
noun( 'consciousness', '-', mass, _ ).
noun( 'conscript', 'conscripts', count, _ ).
noun( 'conscription', '-', mass, _ ).
noun( 'consecration', 'consecrations', both, _ ).
noun( 'consensus', '-', both, _ ).
noun( 'consent', '-', mass, _ ).
noun( 'consequence', 'consequences', both, _ ).
noun( 'conservancy', 'conservancies', both, _ ).
noun( 'conservation', '-', mass, _ ).
noun( 'conservatism', '-', mass, _ ).
noun( 'conservative', 'conservatives', count, _ ).
noun( 'conservatoire', 'conservatoires', count, _ ).
noun( 'conservatory', 'conservatories', count, _ ).
noun( 'conserve', 'conserves', count, _ ).
noun( 'considerateness', '-', mass, _ ).
noun( 'consideration', 'considerations', both, _ ).
noun( 'consignee', 'consignees', count, _ ).
noun( 'consigner', 'consigners', count, _ ).
noun( 'consignment', 'consignments', both, _ ).
noun( 'consignor', 'consignors', count, _ ).
noun( 'consistence', '-', mass, _ ).
noun( 'consistency', 'consistencies', both, _ ).
noun( 'consistory', 'consistories', count, _ ).
noun( 'consolation', 'consolations', both, _ ).
noun( 'console', 'consoles', count, _ ).
noun( 'consolidation', 'consolidations', both, _ ).
noun( 'consomm_e', 'consomm_es', both, _ ).
noun( 'consonance', '-', mass, _ ).
noun( 'consonant', 'consonants', count, _ ).
noun( 'consort', 'consorts', count, _ ).
noun( 'consortium', 'consortia', count, _ ).
noun( 'conspectus', 'conspectuses', count, _ ).
noun( 'conspicuousness', '-', mass, _ ).
noun( 'conspiracy', 'conspiracies', both, _ ).
noun( 'conspirator', 'conspirators', count, _ ).
noun( 'constable', 'constables', count, _ ).
noun( 'constabulary', 'constabularies', count, _ ).
noun( 'constancy', '-', mass, _ ).
noun( 'constant', 'constants', count, _ ).
noun( 'constellation', 'constellations', count, _ ).
noun( 'consternation', '-', mass, _ ).
noun( 'constipation', '-', mass, _ ).
noun( 'constituency', 'constituencies', count, _ ).
noun( 'constituent', 'constituents', count, _ ).
noun( 'constitution', 'constitutions', both, _ ).
noun( 'constitutional', 'constitutionals', count, _ ).
noun( 'constitutionalism', '-', mass, _ ).
noun( 'constitutionalist', 'constitutionalists', count, _ ).
noun( 'constraint', 'constraints', both, _ ).
noun( 'constriction', 'constrictions', both, _ ).
noun( 'construction', 'constructions', both, _ ).
noun( 'constructor', 'constructors', count, _ ).
noun( 'consubstantiation', '-', mass, _ ).
noun( 'consul', 'consuls', count, _ ).
noun( 'consulate', 'consulates', count, _ ).
noun( 'consulship', 'consulships', count, _ ).
noun( 'consultant', 'consultants', count, _ ).
noun( 'consultation', 'consultations', both, _ ).
noun( 'consumer', 'consumers', count, _ ).
noun( 'consummation', 'consummations', both, _ ).
noun( 'consumption', '-', mass, _ ).
noun( 'consumptive', 'consumptives', count, _ ).
noun( 'cont', '-', proper, _ ).
noun( 'contact', 'contacts', both, _ ).
noun( 'contagion', 'contagions', both, _ ).
noun( 'container', 'containers', count, _ ).
noun( 'containment', '-', mass, _ ).
noun( 'contaminant', 'contaminants', count, _ ).
noun( 'contamination', 'contaminations', both, _ ).
noun( 'contemplation', '-', mass, _ ).
noun( 'contemporary', 'contemporaries', count, _ ).
noun( 'contempt', '-', mass, _ ).
noun( 'contender', 'contenders', count, _ ).
noun( 'content', 'contents', count, _ ).
noun( 'content', '-', mass, _ ).
noun( 'contention', 'contentions', both, _ ).
noun( 'contentment', '-', mass, _ ).
noun( 'contest', 'contests', count, _ ).
noun( 'contestant', 'contestants', count, _ ).
noun( 'context', 'contexts', both, _ ).
noun( 'contiguity', '-', mass, _ ).
noun( 'continence', '-', mass, _ ).
noun( 'continent', 'continents', count, _ ).
noun( 'contingency', 'contingencies', both, _ ).
noun( 'contingent', 'contingents', count, _ ).
noun( 'continuance', '-', count, _ ).
noun( 'continuation', 'continuations', both, _ ).
noun( 'continuity', '-', mass, _ ).
noun( 'contortion', 'contortions', both, _ ).
noun( 'contortionist', 'contortionists', count, _ ).
noun( 'contour', 'contours', count, _ ).
noun( 'contraband', '-', mass, _ ).
noun( 'contrabass', 'contrabasses', count, _ ).
noun( 'contraception', '-', mass, _ ).
noun( 'contraceptive', 'contraceptives', count, _ ).
noun( 'contract', 'contracts', both, _ ).
noun( 'contraction', 'contractions', both, _ ).
noun( 'contractor', 'contractors', count, _ ).
noun( 'contradiction', 'contradictions', both, _ ).
noun( 'contradistinction', 'contradistinctions', count, _ ).
noun( 'contralto', 'contraltos', count, _ ).
noun( 'contraption', 'contraptions', count, _ ).
noun( 'contrariety', 'contrarieties', both, _ ).
noun( 'contrariness', '-', mass, _ ).
noun( 'contrary', 'contraries', both, _ ).
noun( 'contrast', 'contrasts', both, _ ).
noun( 'contravention', 'contraventions', both, _ ).
noun( 'contretemps', 'contretemps', count, _ ).
noun( 'contribution', 'contributions', both, _ ).
noun( 'contributor', 'contributors', count, _ ).
noun( 'contrition', '-', mass, _ ).
noun( 'contrivance', 'contrivances', both, _ ).
noun( 'contriver', 'contrivers', count, _ ).
noun( 'control', 'controls', both, _ ).
noun( 'controller', 'controllers', count, _ ).
noun( 'controversialist', 'controversialists', count, _ ).
noun( 'controversy', 'controversies', both, _ ).
noun( 'contumacy', 'contumacies', both, _ ).
noun( 'contumely', 'contumelies', both, _ ).
noun( 'contusion', 'contusions', count, _ ).
noun( 'conundrum', 'conundrums', count, _ ).
noun( 'conurbation', 'conurbations', count, _ ).
noun( 'convalescence', '-', mass, _ ).
noun( 'convalescent', 'convalescents', count, _ ).
noun( 'convection', '-', mass, _ ).
noun( 'convector', 'convectors', count, _ ).
noun( 'convener', 'conveners', count, _ ).
noun( 'convenience', 'conveniences', both, _ ).
noun( 'convent', 'convents', count, _ ).
noun( 'conventicle', 'conventicles', count, _ ).
noun( 'convention', 'conventions', both, _ ).
noun( 'conventionality', 'conventionalities', both, _ ).
noun( 'convergence', 'convergences', count, _ ).
noun( 'conversation', 'conversations', both, _ ).
noun( 'conversationalist', 'conversationalists', count, _ ).
noun( 'converse', '-', both, _ ).
noun( 'conversion', 'conversions', both, _ ).
noun( 'convert', 'converts', count, _ ).
noun( 'converter', 'converters', count, _ ).
noun( 'convertibility', '-', mass, _ ).
noun( 'convertible', 'convertibles', count, _ ).
noun( 'convexity', '-', mass, _ ).
noun( 'conveyance', 'conveyances', both, _ ).
noun( 'conveyancer', 'conveyancers', count, _ ).
noun( 'conveyer', 'conveyers', count, _ ).
noun( 'conveyer-belt', 'conveyer-belts', count, _ ).
noun( 'conveyor', 'conveyors', count, _ ).
noun( 'convict', 'convicts', count, _ ).
noun( 'conviction', 'convictions', both, _ ).
noun( 'conviviality', 'convivialities', both, _ ).
noun( 'convocation', 'convocations', both, _ ).
noun( 'convolution', 'convolutions', count, _ ).
noun( 'convolvulus', 'convolvuluses', count, _ ).
noun( 'convoy', 'convoys', both, _ ).
noun( 'convulsion', 'convulsions', count, _ ).
noun( 'cony', 'conies', count, _ ).
noun( 'coo', 'coos', count, _ ).
noun( 'cook', 'cooks', count, _ ).
noun( 'cookbook', 'cookbooks', count, _ ).
noun( 'cooker', 'cookers', count, _ ).
noun( 'cookery', '-', mass, _ ).
noun( 'cookery-book', 'cookery-books', count, _ ).
noun( 'cookhouse', 'cookhouses', count, _ ).
noun( 'cookie', 'cookies', count, _ ).
noun( 'cooking', '-', mass, _ ).
noun( 'cooky', 'cookies', count, _ ).
noun( 'cool', '-', mass, _ ).
noun( 'coolant', 'coolants', both, _ ).
noun( 'cooler', 'coolers', count, _ ).
noun( 'coolie', 'coolies', count, _ ).
noun( 'cooling-tower', 'cooling-towers', count, _ ).
noun( 'coolness', '-', mass, _ ).
noun( 'coon', 'coons', count, _ ).
noun( 'coop', 'coops', count, _ ).
noun( 'cooper', 'coopers', count, _ ).
noun( 'cooperation', '-', mass, _ ).
noun( 'cooperative', 'cooperatives', count, _ ).
noun( 'cooperator', 'cooperators', count, _ ).
noun( 'coordinate', 'coordinates', count, _ ).
noun( 'coordination', 'coordinations', both, _ ).
noun( 'coordinator', 'coordinators', count, _ ).
noun( 'coot', 'coots', count, _ ).
noun( 'cop', 'cops', count, _ ).
noun( 'cop-out', 'cop-outs', count, _ ).
noun( 'copartner', 'copartners', count, _ ).
noun( 'copartnership', 'copartnerships', count, _ ).
noun( 'cope', 'copes', count, _ ).
noun( 'copeck', 'copecks', count, _ ).
noun( 'coping', 'copings', count, _ ).
noun( 'coping-stone', 'coping-stones', count, _ ).
noun( 'copper', 'coppers', both, _ ).
noun( 'copperhead', 'copperheads', count, _ ).
noun( 'copperplate', '-', mass, _ ).
noun( 'coppersmith', 'coppersmiths', count, _ ).
noun( 'coppice', 'coppices', count, _ ).
noun( 'copra', '-', mass, _ ).
noun( 'copse', 'copses', count, _ ).
noun( 'copula', 'copulas', count, _ ).
noun( 'copulation', 'copulations', count, _ ).
noun( 'copulative', 'copulatives', count, _ ).
noun( 'copy', 'copies', both, _ ).
noun( 'copybook', 'copybooks', count, _ ).
noun( 'copycat', 'copycats', count, _ ).
noun( 'copyhold', '-', mass, _ ).
noun( 'copyholder', 'copyholders', count, _ ).
noun( 'copyist', 'copyists', count, _ ).
noun( 'copyright', 'copyrights', both, _ ).
noun( 'copywriter', 'copywriters', count, _ ).
noun( 'coquetry', 'coquetries', both, _ ).
noun( 'coquette', 'coquettes', count, _ ).
noun( 'cor anglais', 'cor anglais', count, _ ).
noun( 'coracle', 'coracles', count, _ ).
noun( 'coral', 'corals', both, _ ).
noun( 'coral-reef', 'coral-reefs', count, _ ).
noun( 'corbel', 'corbels', count, _ ).
noun( 'cord', 'cords', both, _ ).
noun( 'cordage', '-', mass, _ ).
noun( 'cordial', 'cordials', both, _ ).
noun( 'cordiality', 'cordialities', both, _ ).
noun( 'cordite', '-', mass, _ ).
noun( 'cordon', 'cordons', count, _ ).
noun( 'corduroy', '-', mass, _ ).
noun( 'core', 'cores', count, _ ).
noun( 'coreligionist', 'coreligionists', count, _ ).
noun( 'corgi', 'corgis', count, _ ).
noun( 'cork', 'corks', both, _ ).
noun( 'corkage', '-', mass, _ ).
noun( 'corker', 'corkers', count, _ ).
noun( 'corkscrew', 'corkscrews', count, _ ).
noun( 'corm', 'corms', count, _ ).
noun( 'cormorant', 'cormorants', count, _ ).
noun( 'corn', 'corns', both, _ ).
noun( 'corn-exchange', 'corn-exchanges', count, _ ).
noun( 'corncob', 'corncobs', count, _ ).
noun( 'corncrake', 'corncrakes', count, _ ).
noun( 'cornea', 'corneas', count, _ ).
noun( 'cornelian', 'cornelians', count, _ ).
noun( 'corner', 'corners', count, _ ).
noun( 'corner-kick', 'corner-kicks', count, _ ).
noun( 'cornerstone', 'cornerstones', count, _ ).
noun( 'cornet', 'cornets', count, _ ).
noun( 'cornflake', 'cornflakes', count, _ ).
noun( 'cornflour', '-', mass, _ ).
noun( 'cornflower', 'cornflowers', count, _ ).
noun( 'cornice', 'cornices', count, _ ).
noun( 'cornpone', '-', mass, _ ).
noun( 'cornstarch', '-', mass, _ ).
noun( 'cornucopia', 'cornucopias', count, _ ).
noun( 'corolla', 'corollas', count, _ ).
noun( 'corollary', 'corollaries', count, _ ).
noun( 'corona', 'coronas', count, _ ).
noun( 'coronary', 'coronaries', count, _ ).
noun( 'coronation', 'coronations', count, _ ).
noun( 'coroner', 'coroners', count, _ ).
noun( 'coronet', 'coronets', count, _ ).
noun( 'corporal', 'corporals', count, _ ).
noun( 'corporation', 'corporations', count, _ ).
noun( 'corps', '-', count, _ ).
noun( 'corps de ballet', '-', count, _ ).
noun( 'corpse', 'corpses', count, _ ).
noun( 'corpulence', '-', mass, _ ).
noun( 'corpus', 'corpora', count, _ ).
noun( 'corpuscle', 'corpuscles', count, _ ).
noun( 'corral', 'corrals', count, _ ).
noun( 'correction', 'corrections', both, _ ).
noun( 'correctitude', '-', mass, _ ).
noun( 'corrective', 'correctives', count, _ ).
noun( 'correctness', '-', mass, _ ).
noun( 'correlation', 'correlations', count, _ ).
noun( 'correlative', 'correlatives', count, _ ).
noun( 'correspondence', 'correspondences', both, _ ).
noun( 'correspondent', 'correspondents', count, _ ).
noun( 'corridor', 'corridors', count, _ ).
noun( 'corrie', 'corries', count, _ ).
noun( 'corrigendum', 'corrigenda', count, _ ).
noun( 'corroboration', '-', mass, _ ).
noun( 'corrosion', '-', mass, _ ).
noun( 'corrosive', 'corrosives', count, _ ).
noun( 'corrugation', 'corrugations', both, _ ).
noun( 'corruptibility', '-', mass, _ ).
noun( 'corruption', '-', mass, _ ).
noun( 'corruptness', '-', mass, _ ).
noun( 'corsage', 'corsages', count, _ ).
noun( 'corsair', 'corsairs', count, _ ).
noun( 'corse', 'corses', count, _ ).
noun( 'corselet', 'corselets', count, _ ).
noun( 'corset', 'corsets', count, _ ).
noun( 'corslet', 'corslets', count, _ ).
noun( 'cort`ege', 'cort`eges', count, _ ).
noun( 'cortege', 'corteges', count, _ ).
noun( 'cortex', 'cortices', count, _ ).
noun( 'cortisone', '-', mass, _ ).
noun( 'corundum', '-', mass, _ ).
noun( 'coruscation', 'coruscations', count, _ ).
noun( 'corv_ee', 'corv_ees', count, _ ).
noun( 'corvette', 'corvettes', count, _ ).
noun( 'cos', 'cos', count, _ ).
noun( 'cosh', 'coshes', count, _ ).
noun( 'cosignatory', 'cosignatories', count, _ ).
noun( 'cosine', 'cosines', count, _ ).
noun( 'cosiness', '-', mass, _ ).
noun( 'cosmetic', 'cosmetics', count, _ ).
noun( 'cosmetician', 'cosmeticians', count, _ ).
noun( 'cosmogony', 'cosmogonies', count, _ ).
noun( 'cosmonaut', 'cosmonauts', count, _ ).
noun( 'cosmopolitan', 'cosmopolitans', count, _ ).
noun( 'cosmos', '-', count, _ ).
noun( 'cost', 'costs', both, _ ).
noun( 'costermonger', 'costermongers', count, _ ).
noun( 'costing', 'costings', both, _ ).
noun( 'costliness', '-', mass, _ ).
noun( 'costume', 'costumes', both, _ ).
noun( 'costumier', 'costumiers', count, _ ).
noun( 'cosy', 'cosies', count, _ ).
noun( 'cot', 'cots', count, _ ).
noun( 'cote', 'cotes', count, _ ).
noun( 'cotenant', 'cotenants', count, _ ).
noun( 'coterie', 'coteries', count, _ ).
noun( 'cotilion', 'cotilions', count, _ ).
noun( 'cotillion', 'cotillions', count, _ ).
noun( 'cottage', 'cottages', count, _ ).
noun( 'cottar', 'cottars', count, _ ).
noun( 'cotter', 'cotters', count, _ ).
noun( 'cotton', '-', mass, _ ).
noun( 'cotton-cake', '-', mass, _ ).
noun( 'cotton-plant', 'cotton-plants', count, _ ).
noun( 'cotton-wool', '-', mass, _ ).
noun( 'cottontail', 'cottontails', count, _ ).
noun( 'cotyledon', 'cotyledons', count, _ ).
noun( 'couch', 'couches', both, _ ).
noun( 'couch-grass', '-', mass, _ ).
noun( 'couchette', 'couchettes', count, _ ).
noun( 'cougar', 'cougars', count, _ ).
noun( 'cough', 'coughs', count, _ ).
noun( 'coulter', 'coulters', count, _ ).
noun( 'council', 'councils', count, _ ).
noun( 'council-board', 'council-boards', count, _ ).
noun( 'council-chamber', 'council-chambers', count, _ ).
noun( 'councillor', 'councillors', count, _ ).
noun( 'counsel', '-', mass, _ ).
noun( 'counsellor', 'counsellors', count, _ ).
noun( 'count', 'counts', both, _ ).
noun( 'countdown', 'countdowns', count, _ ).
noun( 'countenance', 'countenances', both, _ ).
noun( 'counter', 'counters', count, _ ).
noun( 'counter-example', 'counter-examples', count, _ ).
noun( 'counter-revolution', 'counter-revolutions', both, _ ).
noun( 'counter-revolutionary', 'counter-revolutionaries', count, _ ).
noun( 'counteraction', 'counteractions', count, _ ).
noun( 'counterattack', 'counterattacks', count, _ ).
noun( 'counterattraction', 'counterattractions', count, _ ).
noun( 'counterbalance', 'counterbalances', count, _ ).
noun( 'counterblast', 'counterblasts', count, _ ).
noun( 'counterclaim', 'counterclaims', count, _ ).
noun( 'counterespionage', '-', mass, _ ).
noun( 'counterfeit', 'counterfeits', count, _ ).
noun( 'counterfeiter', 'counterfeiters', count, _ ).
noun( 'counterfoil', 'counterfoils', count, _ ).
noun( 'counterintelligence', '-', mass, _ ).
noun( 'counterirritant', 'counterirritants', count, _ ).
noun( 'countermine', 'countermines', count, _ ).
noun( 'counteroffer', 'counteroffers', count, _ ).
noun( 'counterpane', 'counterpanes', count, _ ).
noun( 'counterpart', 'counterparts', count, _ ).
noun( 'counterplot', 'counterplots', count, _ ).
noun( 'counterpoint', 'counterpoints', both, _ ).
noun( 'counterpoise', 'counterpoises', both, _ ).
noun( 'countersign', 'countersigns', count, _ ).
noun( 'countertenor', 'countertenors', count, _ ).
noun( 'countess', 'countesses', count, _ ).
noun( 'counting-house', 'counting-houses', count, _ ).
noun( 'country', 'countries', both, _ ).
noun( 'country-house', 'country-houses', count, _ ).
noun( 'country-seat', 'country-seats', count, _ ).
noun( 'countryman', 'countrymen', count, _ ).
noun( 'countryside', '-', mass, _ ).
noun( 'countrywoman', 'countrywomen', count, _ ).
noun( 'county', 'counties', count, _ ).
noun( 'coup', 'coups', count, _ ).
noun( 'coup d\'etat', 'coups d\'etat', count, _ ).
noun( 'coup de grace', '-', count, _ ).
noun( 'coup_e', 'coup_es', count, _ ).
noun( 'couple', 'couples', count, _ ).
noun( 'couplet', 'couplets', count, _ ).
noun( 'coupling', 'couplings', both, _ ).
noun( 'coupon', 'coupons', count, _ ).
noun( 'courage', '-', mass, _ ).
noun( 'courgette', 'courgettes', count, _ ).
noun( 'courier', 'couriers', count, _ ).
noun( 'course', 'courses', both, _ ).
noun( 'courser', 'coursers', count, _ ).
noun( 'coursing', 'coursings', count, _ ).
noun( 'court', 'courts', both, _ ).
noun( 'court-card', 'court-cards', count, _ ).
noun( 'court-martial', 'courts-martial', count, _ ).
noun( 'courtesan', 'courtesans', count, _ ).
noun( 'courtesy', 'courtesies', both, _ ).
noun( 'courtier', 'courtiers', count, _ ).
noun( 'courtliness', '-', mass, _ ).
noun( 'courtroom', 'courtrooms', count, _ ).
noun( 'courtship', 'courtships', both, _ ).
noun( 'courtyard', 'courtyards', count, _ ).
noun( 'cousin', 'cousins', count, _ ).
noun( 'cove', 'coves', count, _ ).
noun( 'coven', 'covens', count, _ ).
noun( 'covenant', 'covenants', count, _ ).
noun( 'cover', 'covers', both, _ ).
noun( 'cover-up', 'cover-ups', count, _ ).
noun( 'coverage', '-', mass, _ ).
noun( 'covering', 'coverings', count, _ ).
noun( 'coverlet', 'coverlets', count, _ ).
noun( 'covert', 'coverts', count, _ ).
noun( 'covetousness', '-', mass, _ ).
noun( 'covey', 'coveys', count, _ ).
noun( 'cow', 'cows', count, _ ).
noun( 'coward', 'cowards', count, _ ).
noun( 'cowardice', '-', mass, _ ).
noun( 'cowbell', 'cowbells', count, _ ).
noun( 'cowboy', 'cowboys', count, _ ).
noun( 'cowcatcher', 'cowcatchers', count, _ ).
noun( 'cowhand', 'cowhands', count, _ ).
noun( 'cowherd', 'cowherds', count, _ ).
noun( 'cowhide', 'cowhides', both, _ ).
noun( 'cowhouse', 'cowhouses', count, _ ).
noun( 'cowl', 'cowls', count, _ ).
noun( 'cowling', 'cowlings', count, _ ).
noun( 'cowman', 'cowmen', count, _ ).
noun( 'cowpox', 'cowpoxes', count, _ ).
noun( 'cowrie', 'cowries', count, _ ).
noun( 'cowshed', 'cowsheds', count, _ ).
noun( 'cowskin', 'cowskins', count, _ ).
noun( 'cowslip', 'cowslips', count, _ ).
noun( 'cox', 'coxes', count, _ ).
noun( 'coxcomb', 'coxcombs', count, _ ).
noun( 'coxswain', 'coxswains', count, _ ).
noun( 'coyness', '-', mass, _ ).
noun( 'coyote', 'coyotes', count, _ ).
noun( 'coypu', 'coypus', count, _ ).
noun( 'cp', '-', proper, _ ).
noun( 'cr^epe', '-', mass, _ ).
noun( 'cr`eche', 'cr`eches', count, _ ).
noun( 'cr`eme de menthe', '-', mass, _ ).
noun( 'crab', 'crabs', both, _ ).
noun( 'crab-apple', 'crab-apples', count, _ ).
noun( 'crack', 'cracks', count, _ ).
noun( 'crack-down', 'crack-downs', count, _ ).
noun( 'crack-up', 'crack-ups', count, _ ).
noun( 'cracker', 'crackers', count, _ ).
noun( 'crackle', '-', mass, _ ).
noun( 'crackle-china', '-', mass, _ ).
noun( 'crackleware', '-', mass, _ ).
noun( 'crackling', '-', mass, _ ).
noun( 'crackpot', 'crackpots', count, _ ).
noun( 'cracksman', 'cracksmen', count, _ ).
noun( 'cradle', 'cradles', count, _ ).
noun( 'craft', 'crafts', both, _ ).
noun( 'craftiness', '-', mass, _ ).
noun( 'craftsman', 'craftsmen', count, _ ).
noun( 'craftsmanship', '-', mass, _ ).
noun( 'crag', 'crags', count, _ ).
noun( 'cragsman', 'cragsmen', count, _ ).
noun( 'crake', 'crakes', count, _ ).
noun( 'crammer', 'crammers', count, _ ).
noun( 'cramp', 'cramps', both, _ ).
noun( 'cramp-iron', 'cramp-irons', count, _ ).
noun( 'crampon', 'crampons', count, _ ).
noun( 'cranberry', 'cranberries', count, _ ).
noun( 'crane', 'cranes', count, _ ).
noun( 'crane-fly', 'crane-flies', count, _ ).
noun( 'cranium', 'craniums', count, _ ).
noun( 'crank', 'cranks', count, _ ).
noun( 'crankshaft', 'crankshafts', count, _ ).
noun( 'cranny', 'crannies', count, _ ).
noun( 'crap', 'craps', both, _ ).
noun( 'crap-shooting', 'crap-shootings', count, _ ).
noun( 'crape', '-', mass, _ ).
noun( 'craps', 'craps', mass, _ ).
noun( 'crash', 'crashes', both, _ ).
noun( 'crash-dive', 'crash-dives', count, _ ).
noun( 'crash-helmet', 'crash-helmets', count, _ ).
noun( 'crash-landing', 'crash-landings', count, _ ).
noun( 'crate', 'crates', count, _ ).
noun( 'crater', 'craters', count, _ ).
noun( 'cravat', 'cravats', count, _ ).
noun( 'craven', 'cravens', count, _ ).
noun( 'craving', 'cravings', count, _ ).
noun( 'crawfish', 'crawfishes', count, _ ).
noun( 'crawl', '-', count, _ ).
noun( 'crawler', 'crawlers', count, _ ).
noun( 'crayfish', 'crayfishes', count, _ ).
noun( 'crayon', 'crayons', count, _ ).
noun( 'craze', 'crazes', count, _ ).
noun( 'craziness', '-', mass, _ ).
noun( 'creak', 'creaks', count, _ ).
noun( 'cream', 'creams', both, _ ).
noun( 'creamery', 'creameries', count, _ ).
noun( 'crease', 'creases', count, _ ).
noun( 'creation', 'creations', both, _ ).
noun( 'creativeness', '-', mass, _ ).
noun( 'creator', 'creators', count, _ ).
noun( 'creature', 'creatures', count, _ ).
noun( 'credence', '-', mass, _ ).
noun( 'credibility', '-', mass, _ ).
noun( 'credit', 'credits', both, _ ).
noun( 'credit-side', 'credit-sides', count, _ ).
noun( 'credit-worthiness', '-', mass, _ ).
noun( 'creditor', 'creditors', count, _ ).
noun( 'credo', 'credos', count, _ ).
noun( 'credulity', 'credulities', both, _ ).
noun( 'creed', 'creeds', count, _ ).
noun( 'creek', 'creeks', count, _ ).
noun( 'creel', 'creels', count, _ ).
noun( 'creep', 'creeps', count, _ ).
noun( 'creeper', 'creepers', count, _ ).
noun( 'cremation', 'cremations', both, _ ).
noun( 'crematorium', 'crematoriums', count, _ ).
noun( 'crematory', 'crematories', count, _ ).
noun( 'creosote', '-', mass, _ ).
noun( 'crepe', '-', mass, _ ).
noun( 'crepitation', 'crepitations', count, _ ).
noun( 'crescendo', 'crescendos', count, _ ).
noun( 'crescent', 'crescents', count, _ ).
noun( 'cress', '-', mass, _ ).
noun( 'crest', 'crests', count, _ ).
noun( 'cretin', 'cretins', count, _ ).
noun( 'cretonne', '-', mass, _ ).
noun( 'crevasse', 'crevasses', count, _ ).
noun( 'crevice', 'crevices', count, _ ).
noun( 'crew', 'crews', count, _ ).
noun( 'crew-cut', 'crew-cuts', count, _ ).
noun( 'crew-neck', 'crew-necks', count, _ ).
noun( 'crib', 'cribs', count, _ ).
noun( 'cribbage', '-', mass, _ ).
noun( 'cribbage-board', 'cribbage-boards', count, _ ).
noun( 'crick', 'cricks', count, _ ).
noun( 'cricket', 'crickets', both, _ ).
noun( 'cricketer', 'cricketers', count, _ ).
noun( 'crier', 'criers', count, _ ).
noun( 'crime', 'crimes', both, _ ).
noun( 'criminal', 'criminals', count, _ ).
noun( 'criminology', '-', mass, _ ).
noun( 'crimson', 'crimsons', both, _ ).
noun( 'crinkle', 'crinkles', count, _ ).
noun( 'crinoline', 'crinolines', both, _ ).
noun( 'cripple', 'cripples', count, _ ).
noun( 'crisis', 'crises', count, _ ).
noun( 'crisp', 'crisps', count, _ ).
noun( 'crispness', '-', mass, _ ).
noun( 'criterion', 'criterions', count, _ ).
noun( 'critic', 'critics', count, _ ).
noun( 'criticism', 'criticisms', both, _ ).
noun( 'critique', 'critiques', count, _ ).
noun( 'croak', 'croaks', count, _ ).
noun( 'crochet', '-', mass, _ ).
noun( 'crochet-hook', 'crochet-hooks', count, _ ).
noun( 'crock', 'crocks', count, _ ).
noun( 'crockery', '-', mass, _ ).
noun( 'crocodile', 'crocodiles', count, _ ).
noun( 'crocus', 'crocuses', count, _ ).
noun( 'croft', 'crofts', count, _ ).
noun( 'crofter', 'crofters', count, _ ).
noun( 'cromlech', 'cromlechs', count, _ ).
noun( 'crone', 'crones', count, _ ).
noun( 'crony', 'cronies', count, _ ).
noun( 'crook', 'crooks', count, _ ).
noun( 'crookedness', '-', mass, _ ).
noun( 'crooner', 'crooners', count, _ ).
noun( 'crop', 'crops', count, _ ).
noun( 'crop-dusting', '-', mass, _ ).
noun( 'cropper', 'croppers', count, _ ).
noun( 'croquet', '-', mass, _ ).
noun( 'croquette', 'croquettes', count, _ ).
noun( 'crore', 'crores', count, _ ).
noun( 'crosier', 'crosiers', count, _ ).
noun( 'cross', 'crosses', count, _ ).
noun( 'cross-bench', 'cross-benches', count, _ ).
noun( 'cross-bencher', 'cross-benchers', count, _ ).
noun( 'cross-division', 'cross-divisions', both, _ ).
noun( 'cross-examination', 'cross-examinations', count, _ ).
noun( 'cross-examiner', 'cross-examiners', count, _ ).
noun( 'cross-fertilization', 'cross-fertilizations', count, _ ).
noun( 'cross-heading', 'cross-headings', count, _ ).
noun( 'cross-index', 'cross-indexes', count, _ ).
noun( 'cross-reference', 'cross-references', count, _ ).
noun( 'cross-section', 'cross-sections', count, _ ).
noun( 'cross-stitch', 'cross-stitches', both, _ ).
noun( 'crossbar', 'crossbars', count, _ ).
noun( 'crossbeam', 'crossbeams', count, _ ).
noun( 'crossbow', 'crossbows', count, _ ).
noun( 'crossbreed', 'crossbreeds', count, _ ).
noun( 'crosscheck', 'crosschecks', count, _ ).
noun( 'crosscurrent', 'crosscurrents', count, _ ).
noun( 'crosscut', 'crosscuts', count, _ ).
noun( 'crosse', 'crosses', count, _ ).
noun( 'crossfire', 'crossfires', both, _ ).
noun( 'crossing', 'crossings', both, _ ).
noun( 'crossness', '-', mass, _ ).
noun( 'crosspatch', 'crosspatches', count, _ ).
noun( 'crosspiece', 'crosspieces', count, _ ).
noun( 'crossroad', 'crossroads', count, _ ).
noun( 'crosstalk', '-', mass, _ ).
noun( 'crosswalk', 'crosswalks', count, _ ).
noun( 'crosswind', 'crosswinds', count, _ ).
noun( 'crossword', 'crosswords', count, _ ).
noun( 'crotch', 'crotches', count, _ ).
noun( 'crotchet', 'crotchets', count, _ ).
noun( 'crouch', 'crouches', count, _ ).
noun( 'croup', 'croups', both, _ ).
noun( 'croupier', 'croupiers', count, _ ).
noun( 'crow', 'crows', count, _ ).
noun( 'crow\'s-nest', 'crow\'s-nests', count, _ ).
noun( 'crowbar', 'crowbars', count, _ ).
noun( 'crowd', 'crowds', count, _ ).
noun( 'crown', 'crowns', count, _ ).
noun( 'crown-land', 'crown-lands', count, _ ).
noun( 'crozier', 'croziers', count, _ ).
noun( 'crucible', 'crucibles', count, _ ).
noun( 'crucifix', 'crucifixes', count, _ ).
noun( 'crucifixion', 'crucifixions', both, _ ).
noun( 'crudeness', '-', mass, _ ).
noun( 'crudity', 'crudities', both, _ ).
noun( 'cruelty', 'cruelties', both, _ ).
noun( 'cruet', 'cruets', count, _ ).
noun( 'cruet-stand', 'cruet-stands', count, _ ).
noun( 'cruise', 'cruises', count, _ ).
noun( 'cruiser', 'cruisers', count, _ ).
noun( 'crumb', 'crumbs', both, _ ).
noun( 'crumpet', 'crumpets', count, _ ).
noun( 'crunch', 'crunches', count, _ ).
noun( 'crupper', 'cruppers', count, _ ).
noun( 'crusade', 'crusades', count, _ ).
noun( 'crusader', 'crusaders', count, _ ).
noun( 'cruse', 'cruses', count, _ ).
noun( 'crush', '-', mass, _ ).
noun( 'crust', 'crusts', both, _ ).
noun( 'crustacean', 'crustaceans', count, _ ).
noun( 'crutch', 'crutches', count, _ ).
noun( 'crux', 'cruxes', count, _ ).
noun( 'cruzeiro', 'cruzeiros', count, _ ).
noun( 'cry', 'cries', count, _ ).
noun( 'crybaby', 'crybabies', count, _ ).
noun( 'crypt', 'crypts', count, _ ).
noun( 'cryptogram', 'cryptograms', count, _ ).
noun( 'crystal', 'crystals', both, _ ).
noun( 'crystal-gazing', '-', mass, _ ).
noun( 'crystallization', 'crystallizations', both, _ ).
noun( 'crystallography', '-', mass, _ ).
noun( 'cub', 'cubs', count, _ ).
noun( 'cubbyhole', 'cubbyholes', count, _ ).
noun( 'cube', 'cubes', count, _ ).
noun( 'cubicle', 'cubicles', count, _ ).
noun( 'cubism', '-', mass, _ ).
noun( 'cubist', 'cubists', count, _ ).
noun( 'cubit', 'cubits', count, _ ).
noun( 'cuckold', 'cuckolds', count, _ ).
noun( 'cuckoo', 'cuckoos', count, _ ).
noun( 'cuckoo-clock', 'cuckoo-clocks', count, _ ).
noun( 'cucumber', 'cucumbers', both, _ ).
noun( 'cud', '-', mass, _ ).
noun( 'cuddle', 'cuddles', count, _ ).
noun( 'cudgel', 'cudgels', count, _ ).
noun( 'cue', 'cues', count, _ ).
noun( 'cuff', 'cuffs', count, _ ).
noun( 'cuirass', 'cuirasses', count, _ ).
noun( 'cuirassier', 'cuirassiers', count, _ ).
noun( 'cuisine', '-', mass, _ ).
noun( 'cul-de-sac', 'cul-de-sacs', count, _ ).
noun( 'cull', 'culls', count, _ ).
noun( 'cullender', 'cullenders', count, _ ).
noun( 'culmination', 'culminations', count, _ ).
noun( 'culpability', '-', mass, _ ).
noun( 'culprit', 'culprits', count, _ ).
noun( 'cult', 'cults', count, _ ).
noun( 'cultivation', '-', mass, _ ).
noun( 'cultivator', 'cultivators', count, _ ).
noun( 'culture', 'cultures', both, _ ).
noun( 'culvert', 'culverts', count, _ ).
noun( 'cummerbund', 'cummerbunds', count, _ ).
noun( 'cumulus', 'cumuli', count, _ ).
noun( 'cunning', '-', mass, _ ).
noun( 'cunt', 'cunts', count, _ ).
noun( 'cup', 'cups', count, _ ).
noun( 'cup-bearer', 'cup-bearers', count, _ ).
noun( 'cup-final', 'cup-finals', count, _ ).
noun( 'cup-tie', 'cup-ties', count, _ ).
noun( 'cupboard', 'cupboards', count, _ ).
noun( 'cupboard-love', '-', mass, _ ).
noun( 'cupful', 'cupfuls', count, _ ).
noun( 'cupidity', '-', mass, _ ).
noun( 'cupola', 'cupolas', count, _ ).
noun( 'cuppa', 'cuppas', count, _ ).
noun( 'cupping', '-', mass, _ ).
noun( 'cupro-nickel', '-', mass, _ ).
noun( 'cur', 'curs', count, _ ).
noun( 'cur_e', 'cur_es', count, _ ).
noun( 'cura<cao', '-', mass, _ ).
noun( 'cura<coa', '-', mass, _ ).
noun( 'curability', '-', mass, _ ).
noun( 'curacy', 'curacies', count, _ ).
noun( 'curate', 'curates', count, _ ).
noun( 'curator', 'curators', count, _ ).
noun( 'curb', 'curbs', count, _ ).
noun( 'curd', 'curds', both, _ ).
noun( 'cure', 'cures', count, _ ).
noun( 'cure-all', 'cure-alls', count, _ ).
noun( 'curfew', 'curfews', count, _ ).
noun( 'curio', 'curios', count, _ ).
noun( 'curiosity', 'curiosities', both, _ ).
noun( 'curl', 'curls', both, _ ).
noun( 'curler', 'curlers', count, _ ).
noun( 'curlew', 'curlews', count, _ ).
noun( 'curling', '-', mass, _ ).
noun( 'curmudgeon', 'curmudgeons', count, _ ).
noun( 'currant', 'currants', count, _ ).
noun( 'currency', 'currencies', both, _ ).
noun( 'current', 'currents', count, _ ).
noun( 'curriculum', 'curriculums', count, _ ).
noun( 'curriculum vitae', '-', count, _ ).
noun( 'curry', 'curries', both, _ ).
noun( 'curry-powder', 'curry-powders', count, _ ).
noun( 'curse', 'curses', count, _ ).
noun( 'curtailment', 'curtailments', both, _ ).
noun( 'curtain', 'curtains', count, _ ).
noun( 'curtain-call', 'curtain-calls', count, _ ).
noun( 'curtain-lecture', 'curtain-lectures', count, _ ).
noun( 'curtain-raiser', 'curtain-raisers', count, _ ).
noun( 'curtness', '-', mass, _ ).
noun( 'curtsey', 'curtseys', count, _ ).
noun( 'curtsy', 'curtsies', count, _ ).
noun( 'curvature', '-', mass, _ ).
noun( 'curve', 'curves', count, _ ).
noun( 'cushion', 'cushions', count, _ ).
noun( 'cusp', 'cusps', count, _ ).
noun( 'cuspidor', 'cuspidors', count, _ ).
noun( 'cuss', 'cusses', count, _ ).
noun( 'cussedness', '-', mass, _ ).
noun( 'custard', 'custards', both, _ ).
noun( 'custodian', 'custodians', count, _ ).
noun( 'custody', '-', mass, _ ).
noun( 'custom', 'customs', both, _ ).
noun( 'customer', 'customers', count, _ ).
noun( 'cut', 'cuts', count, _ ).
noun( 'cut-out', 'cut-outs', count, _ ).
noun( 'cut-throat', 'cut-throats', count, _ ).
noun( 'cutback', 'cutbacks', count, _ ).
noun( 'cuteness', '-', mass, _ ).
noun( 'cuticle', 'cuticles', count, _ ).
noun( 'cutlass', 'cutlasses', count, _ ).
noun( 'cutler', 'cutlers', count, _ ).
noun( 'cutlery', '-', mass, _ ).
noun( 'cutlet', 'cutlets', count, _ ).
noun( 'cutpurse', 'cutpurses', count, _ ).
noun( 'cutter', 'cutters', count, _ ).
noun( 'cutting', 'cuttings', both, _ ).
noun( 'cutting-room', 'cutting-rooms', count, _ ).
noun( 'cuttlefish', 'cuttlefishes', both, _ ).
noun( 'cutworm', 'cutworms', count, _ ).
noun( 'cwt', 'cwt', count, _ ).
noun( 'cyanide', '-', mass, _ ).
noun( 'cybernetics', 'cybernetics', mass, _ ).
noun( 'cyclamen', 'cyclamens', count, _ ).
noun( 'cycle', 'cycles', count, _ ).
noun( 'cyclist', 'cyclists', count, _ ).
noun( 'cyclone', 'cyclones', count, _ ).
noun( 'cyclopaedia', 'cyclopaedias', count, _ ).
noun( 'cyclostyle', 'cyclostyles', count, _ ).
noun( 'cyclotron', 'cyclotrons', count, _ ).
noun( 'cyder', 'cyders', both, _ ).
noun( 'cygnet', 'cygnets', count, _ ).
noun( 'cylinder', 'cylinders', count, _ ).
noun( 'cymbal', 'cymbals', count, _ ).
noun( 'cynic', 'cynics', count, _ ).
noun( 'cynicism', 'cynicisms', both, _ ).
noun( 'cynosure', 'cynosures', count, _ ).
noun( 'cypher', 'cyphers', count, _ ).
noun( 'cypress', 'cypresses', count, _ ).
noun( 'cyst', 'cysts', count, _ ).
noun( 'czar', 'czars', count, _ ).
noun( 'czarina', 'czarinas', count, _ ).
noun( 'd', '-', count, _ ).
noun( 'd_eb^acle', 'd_eb^acles', count, _ ).
noun( 'd_ebris', '-', mass, _ ).
noun( 'd_ebut', 'd_ebuts', count, _ ).
noun( 'd_ebutante', 'd_ebutantes', count, _ ).
noun( 'd_ecor', '-', count, _ ).
noun( 'd_emarche', 'd_emarches', count, _ ).
noun( 'd_enouement', 'd_enouements', count, _ ).
noun( 'd_eshabill_e', '-', mass, _ ).
noun( 'd_etente', '-', mass, _ ).
noun( 'dab', 'dabs', count, _ ).
noun( 'dabbler', 'dabblers', count, _ ).
noun( 'dace', 'dace', count, _ ).
noun( 'dacha', 'dachas', count, _ ).
noun( 'dachshund', 'dachshunds', count, _ ).
noun( 'dacoit', 'dacoits', count, _ ).
noun( 'dacoity', 'dacoities', count, _ ).
noun( 'dactyl', 'dactyls', count, _ ).
noun( 'dad', 'dads', count, _ ).
noun( 'daddy', 'daddies', count, _ ).
noun( 'daddy-longlegs', 'daddy-longlegs', count, _ ).
noun( 'dado', 'dados', count, _ ).
noun( 'daemon', 'daemons', count, _ ).
noun( 'daffodil', 'daffodils', count, _ ).
noun( 'dagger', 'daggers', count, _ ).
noun( 'dago', 'dagos', count, _ ).
noun( 'daguerreotype', 'daguerreotypes', count, _ ).
noun( 'dahlia', 'dahlias', count, _ ).
noun( 'daily', 'dailies', count, _ ).
noun( 'daintiness', '-', mass, _ ).
noun( 'dainty', 'dainties', count, _ ).
noun( 'dairy', 'dairies', count, _ ).
noun( 'dairy-farm', 'dairy-farms', count, _ ).
noun( 'dairy-farming', '-', mass, _ ).
noun( 'dairying', '-', mass, _ ).
noun( 'dairymaid', 'dairymaids', count, _ ).
noun( 'dairyman', 'dairymen', count, _ ).
noun( 'dais', 'daises', count, _ ).
noun( 'daisy', 'daisies', count, _ ).
noun( 'dale', 'dales', count, _ ).
noun( 'dalesman', 'dalesmen', count, _ ).
noun( 'dalliance', '-', mass, _ ).
noun( 'dalmatian', 'dalmatians', count, _ ).
noun( 'dam', 'dams', count, _ ).
noun( 'damage', 'damages', both, _ ).
noun( 'damask', '-', mass, _ ).
noun( 'dame', 'dames', count, _ ).
noun( 'damn', '-', count, _ ).
noun( 'damnation', '-', mass, _ ).
noun( 'damp', '-', mass, _ ).
noun( 'damper', 'dampers', count, _ ).
noun( 'dampness', '-', mass, _ ).
noun( 'damsel', 'damsels', count, _ ).
noun( 'damson', 'damsons', count, _ ).
noun( 'dance', 'dances', count, _ ).
noun( 'dance-band', 'dance-bands', count, _ ).
noun( 'dance-hall', 'dance-halls', count, _ ).
noun( 'dance-orchestra', 'dance-orchestras', count, _ ).
noun( 'dancer', 'dancers', count, _ ).
noun( 'dancing', '-', mass, _ ).
noun( 'dandelion', 'dandelions', count, _ ).
noun( 'dander', '-', count, _ ).
noun( 'dandruff', '-', mass, _ ).
noun( 'dandy', 'dandies', count, _ ).
noun( 'danger', 'dangers', both, _ ).
noun( 'danse macabre', '-', count, _ ).
noun( 'daphne', 'daphnes', count, _ ).
noun( 'dapple-grey', 'dapple-greys', count, _ ).
noun( 'dare', 'dares', count, _ ).
noun( 'daredevil', 'daredevils', count, _ ).
noun( 'daring', '-', mass, _ ).
noun( 'dark', '-', mass, _ ).
noun( 'darkness', '-', mass, _ ).
noun( 'darkroom', 'darkrooms', count, _ ).
noun( 'darling', 'darlings', count, _ ).
noun( 'darn', 'darns', count, _ ).
noun( 'darning', '-', mass, _ ).
noun( 'darning-needle', 'darning-needles', count, _ ).
noun( 'dart', 'darts', count, _ ).
noun( 'dash', 'dashes', both, _ ).
noun( 'dashboard', 'dashboards', count, _ ).
noun( 'dastard', 'dastards', count, _ ).
noun( 'data', 'data', both, _ ).
noun( 'date', 'dates', both, _ ).
noun( 'dateline', 'datelines', count, _ ).
noun( 'dative', 'datives', count, _ ).
noun( 'datum', 'data', count, _ ).
noun( 'daub', 'daubs', both, _ ).
noun( 'dauber', 'daubers', count, _ ).
noun( 'daughter', 'daughters', count, _ ).
noun( 'daughter-in-law', 'daughters-in-law', count, _ ).
noun( 'dauphin', 'dauphins', count, _ ).
noun( 'davenport', 'davenports', count, _ ).
noun( 'davit', 'davits', count, _ ).
noun( 'daw', 'daws', count, _ ).
noun( 'dawdler', 'dawdlers', count, _ ).
noun( 'dawn', 'dawns', both, _ ).
noun( 'day', 'days', both, _ ).
noun( 'day-boarder', 'day-boarders', count, _ ).
noun( 'day-labourer', 'day-labourers', count, _ ).
noun( 'day-return', 'day-returns', count, _ ).
noun( 'day-school', 'day-schools', count, _ ).
noun( 'daybook', 'daybooks', count, _ ).
noun( 'dayboy', 'dayboys', count, _ ).
noun( 'daybreak', 'daybreaks', count, _ ).
noun( 'daydream', 'daydreams', count, _ ).
noun( 'daygirl', 'daygirls', count, _ ).
noun( 'daylight', '-', mass, _ ).
noun( 'daylight-saving', '-', mass, _ ).
noun( 'dayspring', 'daysprings', count, _ ).
noun( 'daytime', 'daytimes', count, _ ).
noun( 'daze', '-', count, _ ).
noun( 'dazzle', '-', mass, _ ).
noun( 'de-escalation', 'de-escalations', count, _ ).
noun( 'deacon', 'deacons', count, _ ).
noun( 'deaconess', 'deaconesses', count, _ ).
noun( 'deadline', 'deadlines', count, _ ).
noun( 'deadlock', 'deadlocks', both, _ ).
noun( 'deaf mute', 'deaf mutes', count, _ ).
noun( 'deaf-aid', 'deaf-aids', count, _ ).
noun( 'deafness', '-', mass, _ ).
noun( 'deal', 'deals', count, _ ).
noun( 'dealer', 'dealers', count, _ ).
noun( 'dealing', 'dealings', both, _ ).
noun( 'dean', 'deans', count, _ ).
noun( 'deanery', 'deaneries', count, _ ).
noun( 'dear', 'dears', count, _ ).
noun( 'dearest', '-', count, _ ).
noun( 'dearie', 'dearies', count, _ ).
noun( 'dearness', '-', mass, _ ).
noun( 'dearth', 'dearths', count, _ ).
noun( 'deary', 'dearies', count, _ ).
noun( 'death', 'deaths', both, _ ).
noun( 'death-mask', 'death-masks', count, _ ).
noun( 'death-rate', 'death-rates', count, _ ).
noun( 'death-roll', 'death-rolls', count, _ ).
noun( 'death-warrant', 'death-warrants', count, _ ).
noun( 'deathbed', 'deathbeds', count, _ ).
noun( 'deathblow', 'deathblows', count, _ ).
noun( 'deathtrap', 'deathtraps', count, _ ).
noun( 'deb', 'debs', count, _ ).
noun( 'debarkation', 'debarkations', count, _ ).
noun( 'debasement', 'debasements', count, _ ).
noun( 'debate', 'debates', both, _ ).
noun( 'debater', 'debaters', count, _ ).
noun( 'debauch', 'debauches', count, _ ).
noun( 'debauchee', 'debauchees', count, _ ).
noun( 'debauchery', 'debaucheries', both, _ ).
noun( 'debenture', 'debentures', count, _ ).
noun( 'debility', '-', mass, _ ).
noun( 'debit', 'debits', count, _ ).
noun( 'debit-side', 'debit-sides', count, _ ).
noun( 'debris', '-', mass, _ ).
noun( 'debt', 'debts', both, _ ).
noun( 'debtor', 'debtors', count, _ ).
noun( 'debut', 'debuts', count, _ ).
noun( 'debutante', 'debutantes', count, _ ).
noun( 'decade', 'decades', count, _ ).
noun( 'decadence', '-', mass, _ ).
noun( 'decadent', 'decadents', count, _ ).
noun( 'decanter', 'decanters', count, _ ).
noun( 'decapitation', 'decapitations', count, _ ).
noun( 'decasyllable', 'decasyllables', count, _ ).
noun( 'decay', '-', mass, _ ).
noun( 'decease', '-', mass, _ ).
noun( 'deceit', 'deceits', both, _ ).
noun( 'deceitfulness', '-', mass, _ ).
noun( 'deceiver', 'deceivers', count, _ ).
noun( 'decency', 'decencies', both, _ ).
noun( 'decentralization', '-', mass, _ ).
noun( 'deception', 'deceptions', both, _ ).
noun( 'decibel', 'decibels', count, _ ).
noun( 'decimalization', '-', mass, _ ).
noun( 'decision', 'decisions', both, _ ).
noun( 'deck', 'decks', count, _ ).
noun( 'decker', 'deckers', count, _ ).
noun( 'declamation', 'declamations', both, _ ).
noun( 'declaration', 'declarations', both, _ ).
noun( 'declassification', 'declassifications', both, _ ).
noun( 'declension', 'declensions', both, _ ).
noun( 'declination', 'declinations', count, _ ).
noun( 'decline', 'declines', count, _ ).
noun( 'declivity', 'declivities', count, _ ).
noun( 'decoder', 'decoders', count, _ ).
noun( 'decolonization', '-', mass, _ ).
noun( 'decomposition', 'decompositions', both, _ ).
noun( 'decompression', 'decompressions', count, _ ).
noun( 'decontamination', '-', mass, _ ).
noun( 'decoration', 'decorations', both, _ ).
noun( 'decorator', 'decorators', count, _ ).
noun( 'decorum', '-', mass, _ ).
noun( 'decoy', 'decoys', count, _ ).
noun( 'decrease', 'decreases', both, _ ).
noun( 'decree', 'decrees', count, _ ).
noun( 'decree nisi', '-', count, _ ).
noun( 'decrepitude', '-', mass, _ ).
noun( 'dedication', 'dedications', both, _ ).
noun( 'deduction', 'deductions', both, _ ).
noun( 'deed', 'deeds', count, _ ).
noun( 'deed-box', 'deed-boxes', count, _ ).
noun( 'deedpoll', 'deedpolls', both, _ ).
noun( 'deep', 'deeps', count, _ ).
noun( 'deep-freeze', 'deep-freezes', count, _ ).
noun( 'deepness', '-', mass, _ ).
noun( 'deer', 'deer', count, _ ).
noun( 'deerskin', 'deerskins', count, _ ).
noun( 'deerstalker', 'deerstalkers', count, _ ).
noun( 'deerstalking', '-', mass, _ ).
noun( 'defacement', 'defacements', both, _ ).
noun( 'defalcation', 'defalcations', both, _ ).
noun( 'defamation', '-', mass, _ ).
noun( 'default', '-', mass, _ ).
noun( 'defaulter', 'defaulters', count, _ ).
noun( 'defeat', 'defeats', both, _ ).
noun( 'defeatism', '-', mass, _ ).
noun( 'defeatist', 'defeatists', count, _ ).
noun( 'defecation', 'defecations', both, _ ).
noun( 'defect', 'defects', count, _ ).
noun( 'defection', 'defections', both, _ ).
noun( 'defectiveness', '-', mass, _ ).
noun( 'defector', 'defectors', count, _ ).
noun( 'defence', 'defences', both, _ ).
noun( 'defencelessness', '-', mass, _ ).
noun( 'defendant', 'defendants', count, _ ).
noun( 'defender', 'defenders', count, _ ).
noun( 'defensive', '-', count, _ ).
noun( 'deference', '-', mass, _ ).
noun( 'deferment', 'deferments', count, _ ).
noun( 'defiance', '-', mass, _ ).
noun( 'deficiency', 'deficiencies', both, _ ).
noun( 'deficit', 'deficits', count, _ ).
noun( 'defile', 'defiles', count, _ ).
noun( 'defilement', '-', mass, _ ).
noun( 'definition', 'definitions', both, _ ).
noun( 'deflation', '-', mass, _ ).
noun( 'deflection', 'deflections', count, _ ).
noun( 'defoliant', 'defoliants', count, _ ).
noun( 'defoliation', 'defoliations', count, _ ).
noun( 'deformity', 'deformities', both, _ ).
noun( 'defrayal', 'defrayals', count, _ ).
noun( 'defrayment', 'defrayments', count, _ ).
noun( 'defroster', 'defrosters', count, _ ).
noun( 'deftness', '-', mass, _ ).
noun( 'degeneracy', '-', mass, _ ).
noun( 'degenerate', 'degenerates', count, _ ).
noun( 'degeneration', '-', mass, _ ).
noun( 'degradation', 'degradations', both, _ ).
noun( 'degree', 'degrees', both, _ ).
noun( 'deification', '-', mass, _ ).
noun( 'deism', '-', mass, _ ).
noun( 'deist', 'deists', count, _ ).
noun( 'deity', 'deities', both, _ ).
noun( 'dejection', '-', mass, _ ).
noun( 'dekko', '-', count, _ ).
noun( 'delay', 'delays', both, _ ).
noun( 'delayed-action', '-', mass, _ ).
noun( 'delectation', '-', mass, _ ).
noun( 'delegacy', 'delegacies', count, _ ).
noun( 'delegate', 'delegates', count, _ ).
noun( 'delegation', 'delegations', both, _ ).
noun( 'deletion', 'deletions', both, _ ).
noun( 'delf', '-', mass, _ ).
noun( 'delft', '-', mass, _ ).
noun( 'deliberation', 'deliberations', both, _ ).
noun( 'delicacy', 'delicacies', both, _ ).
noun( 'delicatessen', 'delicatessens', both, _ ).
noun( 'delight', 'delights', both, _ ).
noun( 'delimitation', 'delimitations', both, _ ).
noun( 'delineation', 'delineations', both, _ ).
noun( 'delinquency', 'delinquencies', both, _ ).
noun( 'delinquent', 'delinquents', count, _ ).
noun( 'delirium', '-', mass, _ ).
noun( 'delirium tremens', '-', count, _ ).
noun( 'deliverance', 'deliverances', both, _ ).
noun( 'deliverer', 'deliverers', count, _ ).
noun( 'delivery', 'deliveries', both, _ ).
noun( 'dell', 'dells', count, _ ).
noun( 'delphinium', 'delphiniums', count, _ ).
noun( 'delta', 'deltas', count, _ ).
noun( 'deluge', 'deluges', count, _ ).
noun( 'delusion', 'delusions', both, _ ).
noun( 'demagnetization', '-', mass, _ ).
noun( 'demagogue', 'demagogues', count, _ ).
noun( 'demagogy', '-', mass, _ ).
noun( 'demand', 'demands', both, _ ).
noun( 'demarcation', '-', mass, _ ).
noun( 'demeanour', '-', mass, _ ).
noun( 'demerara', '-', mass, _ ).
noun( 'demerit', 'demerits', count, _ ).
noun( 'demesne', 'demesnes', both, _ ).
noun( 'demigod', 'demigods', count, _ ).
noun( 'demijohn', 'demijohns', count, _ ).
noun( 'demimondaine', 'demimondaines', count, _ ).
noun( 'demimonde', '-', count, _ ).
noun( 'demise', '-', count, _ ).
noun( 'demister', 'demisters', count, _ ).
noun( 'demo', 'demos', count, _ ).
noun( 'demobilization', '-', mass, _ ).
noun( 'democracy', 'democracies', both, _ ).
noun( 'democrat', 'democrats', count, _ ).
noun( 'democratization', '-', mass, _ ).
noun( 'demography', '-', mass, _ ).
noun( 'demolition', 'demolitions', both, _ ).
noun( 'demon', 'demons', count, _ ).
noun( 'demonetization', 'demonetizations', both, _ ).
noun( 'demoniac', 'demoniacs', count, _ ).
noun( 'demonstrability', '-', mass, _ ).
noun( 'demonstration', 'demonstrations', both, _ ).
noun( 'demonstrator', 'demonstrators', count, _ ).
noun( 'demoralization', '-', mass, _ ).
noun( 'demotion', 'demotions', both, _ ).
noun( 'demur', '-', mass, _ ).
noun( 'demureness', '-', mass, _ ).
noun( 'den', 'dens', count, _ ).
noun( 'denationalization', '-', mass, _ ).
noun( 'denial', 'denials', both, _ ).
noun( 'denier', 'denier', count, _ ).
noun( 'denigration', '-', mass, _ ).
noun( 'denim', '-', mass, _ ).
noun( 'denizen', 'denizens', count, _ ).
noun( 'denomination', 'denominations', count, _ ).
noun( 'denominator', 'denominators', count, _ ).
noun( 'denseness', '-', mass, _ ).
noun( 'density', 'densities', both, _ ).
noun( 'dent', 'dents', count, _ ).
noun( 'dentifrice', '-', mass, _ ).
noun( 'dentist', 'dentists', count, _ ).
noun( 'dentistry', '-', mass, _ ).
noun( 'denture', 'dentures', count, _ ).
noun( 'denudation', '-', mass, _ ).
noun( 'denunciation', 'denunciations', both, _ ).
noun( 'deodar', 'deodars', count, _ ).
noun( 'deodorant', 'deodorants', count, _ ).
noun( 'dep', '-', proper, _ ).
noun( 'departed', 'departed', count, _ ).
noun( 'department', 'departments', count, _ ).
noun( 'departure', 'departures', both, _ ).
noun( 'dependant', 'dependants', count, _ ).
noun( 'dependence', '-', mass, _ ).
noun( 'dependency', 'dependencies', count, _ ).
noun( 'dependent', 'dependents', count, _ ).
noun( 'depiction', 'depictions', count, _ ).
noun( 'depilatory', 'depilatories', both, _ ).
noun( 'depletion', '-', mass, _ ).
noun( 'deployment', 'deployments', count, _ ).
noun( 'deponent', 'deponents', count, _ ).
noun( 'depopulation', '-', mass, _ ).
noun( 'deportation', 'deportations', both, _ ).
noun( 'deportee', 'deportees', count, _ ).
noun( 'deportment', '-', mass, _ ).
noun( 'deposit', 'deposits', count, _ ).
noun( 'deposition', 'depositions', both, _ ).
noun( 'depositor', 'depositors', count, _ ).
noun( 'depository', 'depositories', count, _ ).
noun( 'depot', 'depots', count, _ ).
noun( 'depravity', 'depravities', both, _ ).
noun( 'deprecation', 'deprecations', count, _ ).
noun( 'depreciation', '-', mass, _ ).
noun( 'depredation', 'depredations', count, _ ).
noun( 'depression', 'depressions', both, _ ).
noun( 'depressive', 'depressives', count, _ ).
noun( 'deprivation', 'deprivations', both, _ ).
noun( 'depth', 'depths', both, _ ).
noun( 'depth-bomb', 'depth-bombs', count, _ ).
noun( 'depth-charge', 'depth-charges', count, _ ).
noun( 'deputation', 'deputations', count, _ ).
noun( 'deputy', 'deputies', count, _ ).
noun( 'derailment', 'derailments', count, _ ).
noun( 'derangement', 'derangements', count, _ ).
noun( 'derby', 'derbies', count, _ ).
noun( 'derby', 'derbies', count, _ ).
noun( 'dereliction', '-', mass, _ ).
noun( 'derision', '-', mass, _ ).
noun( 'derivation', 'derivations', both, _ ).
noun( 'derivative', 'derivatives', count, _ ).
noun( 'dermatitis', '-', mass, _ ).
noun( 'dermatologist', 'dermatologists', count, _ ).
noun( 'dermatology', '-', mass, _ ).
noun( 'derogation', '-', mass, _ ).
noun( 'derrick', 'derricks', count, _ ).
noun( 'derring-do', '-', mass, _ ).
noun( 'derv', '-', mass, _ ).
noun( 'dervish', 'dervishes', count, _ ).
noun( 'desalination', '-', mass, _ ).
noun( 'desalinization', '-', mass, _ ).
noun( 'descant', 'descants', count, _ ).
noun( 'descendant', 'descendants', count, _ ).
noun( 'descent', 'descents', both, _ ).
noun( 'description', 'descriptions', both, _ ).
noun( 'desecration', '-', mass, _ ).
noun( 'desegregation', '-', mass, _ ).
noun( 'desensitization', '-', mass, _ ).
noun( 'desert', 'deserts', both, _ ).
noun( 'deserter', 'deserters', count, _ ).
noun( 'desertion', 'desertions', both, _ ).
noun( 'desiccant', 'desiccants', count, _ ).
noun( 'desideratum', 'desiderata', count, _ ).
noun( 'design', 'designs', both, _ ).
noun( 'designation', 'designations', both, _ ).
noun( 'designer', 'designers', count, _ ).
noun( 'designing', '-', mass, _ ).
noun( 'desirability', '-', mass, _ ).
noun( 'desire', 'desires', both, _ ).
noun( 'desk', 'desks', count, _ ).
noun( 'desolation', '-', mass, _ ).
noun( 'despair', '-', mass, _ ).
noun( 'despatch', 'despatches', both, _ ).
noun( 'desperado', 'desperadoes', count, _ ).
noun( 'desperation', '-', mass, _ ).
noun( 'despite', '-', mass, _ ).
noun( 'despondency', '-', mass, _ ).
noun( 'despot', 'despots', count, _ ).
noun( 'despotism', 'despotisms', both, _ ).
noun( 'dessert', 'desserts', count, _ ).
noun( 'dessertspoon', 'dessertspoons', count, _ ).
noun( 'dessertspoonful', 'dessertspoonfuls', count, _ ).
noun( 'destination', 'destinations', count, _ ).
noun( 'destiny', 'destinies', both, _ ).
noun( 'destitution', '-', mass, _ ).
noun( 'destroyer', 'destroyers', count, _ ).
noun( 'destructibility', '-', mass, _ ).
noun( 'destruction', '-', mass, _ ).
noun( 'destructiveness', '-', mass, _ ).
noun( 'desuetude', '-', mass, _ ).
noun( 'detachment', 'detachments', both, _ ).
noun( 'detail', 'details', both, _ ).
noun( 'detainee', 'detainees', count, _ ).
noun( 'detection', '-', mass, _ ).
noun( 'detective', 'detectives', count, _ ).
noun( 'detector', 'detectors', count, _ ).
noun( 'detention', 'detentions', both, _ ).
noun( 'detergent', 'detergents', count, _ ).
noun( 'deterioration', 'deteriorations', both, _ ).
noun( 'determinant', 'determinants', count, _ ).
noun( 'determination', '-', mass, _ ).
noun( 'determinative', 'determinatives', count, _ ).
noun( 'determiner', 'determiners', count, _ ).
noun( 'deterrent', 'deterrents', count, _ ).
noun( 'detestation', 'detestations', both, _ ).
noun( 'dethronement', 'dethronements', count, _ ).
noun( 'detonation', 'detonations', count, _ ).
noun( 'detonator', 'detonators', count, _ ).
noun( 'detour', 'detours', count, _ ).
noun( 'detraction', 'detractions', both, _ ).
noun( 'detractor', 'detractors', count, _ ).
noun( 'detribalization', '-', mass, _ ).
noun( 'detriment', '-', mass, _ ).
noun( 'detritus', '-', mass, _ ).
noun( 'deuce', 'deuces', count, _ ).
noun( 'devaluation', 'devaluations', both, _ ).
noun( 'devastation', '-', mass, _ ).
noun( 'developer', 'developers', count, _ ).
noun( 'development', 'developments', both, _ ).
noun( 'deviant', 'deviants', count, _ ).
noun( 'deviation', 'deviations', both, _ ).
noun( 'deviationism', '-', mass, _ ).
noun( 'deviationist', 'deviationists', count, _ ).
noun( 'device', 'devices', count, _ ).
noun( 'devil', 'devils', count, _ ).
noun( 'devilment', 'devilments', both, _ ).
noun( 'devilry', 'devilries', both, _ ).
noun( 'deviousness', '-', mass, _ ).
noun( 'devitalization', '-', mass, _ ).
noun( 'devolution', '-', mass, _ ).
noun( 'devotee', 'devotees', count, _ ).
noun( 'devotion', 'devotions', both, _ ).
noun( 'devoutness', '-', mass, _ ).
noun( 'dew', '-', mass, _ ).
noun( 'dewlap', 'dewlaps', count, _ ).
noun( 'dexterity', '-', mass, _ ).
noun( 'dextrose', '-', mass, _ ).
noun( 'dhoti', 'dhotis', count, _ ).
noun( 'dhow', 'dhows', count, _ ).
noun( 'diabetes', '-', mass, _ ).
noun( 'diabetic', 'diabetics', count, _ ).
noun( 'diacritic', 'diacritics', count, _ ).
noun( 'diadem', 'diadems', count, _ ).
noun( 'diaeresis', 'diaereses', count, _ ).
noun( 'diagnosis', 'diagnoses', both, _ ).
noun( 'diagonal', 'diagonals', count, _ ).
noun( 'diagram', 'diagrams', count, _ ).
noun( 'dial', 'dials', count, _ ).
noun( 'dialect', 'dialects', both, _ ).
noun( 'dialectic', 'dialectics', count, _ ).
noun( 'dialectician', 'dialecticians', count, _ ).
noun( 'dialogue', 'dialogues', both, _ ).
noun( 'diameter', 'diameters', count, _ ).
noun( 'diamond', 'diamonds', count, _ ).
noun( 'diaper', 'diapers', count, _ ).
noun( 'diaphragm', 'diaphragms', count, _ ).
noun( 'diarchy', 'diarchies', count, _ ).
noun( 'diarist', 'diarists', count, _ ).
noun( 'diarrhea', '-', mass, _ ).
noun( 'diarrhoea', '-', mass, _ ).
noun( 'diary', 'diaries', count, _ ).
noun( 'diatribe', 'diatribes', count, _ ).
noun( 'dibber', 'dibbers', count, _ ).
noun( 'dibble', 'dibbles', count, _ ).
noun( 'dice', 'dice', count, _ ).
noun( 'dice-box', 'dice-boxes', count, _ ).
noun( 'dichotomy', 'dichotomies', count, _ ).
noun( 'dickens', '-', count, _ ).
noun( 'dickey', 'dickeys', count, _ ).
noun( 'dicky', 'dickies', count, _ ).
noun( 'dicky-seat', 'dicky-seats', count, _ ).
noun( 'dickybird', 'dickybirds', count, _ ).
noun( 'dictate', 'dictates', count, _ ).
noun( 'dictation', 'dictations', both, _ ).
noun( 'dictator', 'dictators', count, _ ).
noun( 'dictatorship', 'dictatorships', both, _ ).
noun( 'diction', '-', mass, _ ).
noun( 'dictionary', 'dictionaries', count, _ ).
noun( 'dictum', 'dictums', count, _ ).
noun( 'die', 'dies', count, _ ).
noun( 'die-hard', 'die-hards', count, _ ).
noun( 'dieresis', 'diereses', count, _ ).
noun( 'diesel', 'diesels', both, _ ).
noun( 'diet', 'diets', count, _ ).
noun( 'dietetics', 'dietetics', mass, _ ).
noun( 'dietician', 'dieticians', count, _ ).
noun( 'dietitian', 'dietitians', count, _ ).
noun( 'difference', 'differences', both, _ ).
noun( 'differential', 'differentials', count, _ ).
noun( 'differentiation', 'differentiations', both, _ ).
noun( 'difficulty', 'difficulties', both, _ ).
noun( 'diffidence', '-', mass, _ ).
noun( 'diffraction', 'diffractions', count, _ ).
noun( 'diffuseness', '-', mass, _ ).
noun( 'diffusion', '-', mass, _ ).
noun( 'dig', 'digs', count, _ ).
noun( 'digest', 'digests', count, _ ).
noun( 'digestibility', '-', mass, _ ).
noun( 'digestion', 'digestions', both, _ ).
noun( 'digger', 'diggers', count, _ ).
noun( 'digging', 'diggings', both, _ ).
noun( 'digit', 'digits', count, _ ).
noun( 'dignitary', 'dignitaries', count, _ ).
noun( 'dignity', 'dignities', both, _ ).
noun( 'digraph', 'digraphs', count, _ ).
noun( 'digression', 'digressions', both, _ ).
noun( 'dike', 'dikes', count, _ ).
noun( 'dilapidation', '-', mass, _ ).
noun( 'dilation', '-', mass, _ ).
noun( 'dilemma', 'dilemmas', count, _ ).
noun( 'dilettante', 'dilettantes', count, _ ).
noun( 'diligence', '-', mass, _ ).
noun( 'dill', '-', mass, _ ).
noun( 'dilution', 'dilutions', both, _ ).
noun( 'dime', 'dimes', count, _ ).
noun( 'dimension', 'dimensions', both, _ ).
noun( 'diminuendo', 'diminuendos', count, _ ).
noun( 'diminution', 'diminutions', both, _ ).
noun( 'diminutive', 'diminutives', count, _ ).
noun( 'dimity', '-', mass, _ ).
noun( 'dimness', '-', mass, _ ).
noun( 'dimple', 'dimples', count, _ ).
noun( 'din', '-', mass, _ ).
noun( 'dinar', 'dinars', count, _ ).
noun( 'diner', 'diners', count, _ ).
noun( 'ding-dong', 'ding-dongs', count, _ ).
noun( 'dinghy', 'dinghies', count, _ ).
noun( 'dinginess', '-', mass, _ ).
noun( 'dingle', 'dingles', count, _ ).
noun( 'dining-car', 'dining-cars', count, _ ).
noun( 'dining-room', 'dining-rooms', count, _ ).
noun( 'dining-table', 'dining-tables', count, _ ).
noun( 'dinner', 'dinners', count, _ ).
noun( 'dinner-jacket', 'dinner-jackets', count, _ ).
noun( 'dinner-party', 'dinner-parties', count, _ ).
noun( 'dinner-service', 'dinner-services', count, _ ).
noun( 'dinner-set', 'dinner-sets', count, _ ).
noun( 'dinosaur', 'dinosaurs', count, _ ).
noun( 'dint', 'dints', count, _ ).
noun( 'diocesan', 'diocesans', count, _ ).
noun( 'diocese', 'dioceses', count, _ ).
noun( 'dioxide', 'dioxides', count, _ ).
noun( 'dip', 'dips', both, _ ).
noun( 'diphtheria', '-', mass, _ ).
noun( 'diphthong', 'diphthongs', count, _ ).
noun( 'diploma', 'diplomas', count, _ ).
noun( 'diplomacy', '-', mass, _ ).
noun( 'diplomat', 'diplomats', count, _ ).
noun( 'diplomatist', 'diplomatists', count, _ ).
noun( 'dipper', 'dippers', count, _ ).
noun( 'dipsomania', '-', mass, _ ).
noun( 'dipsomaniac', 'dipsomaniacs', count, _ ).
noun( 'dipstick', 'dipsticks', count, _ ).
noun( 'diptych', 'diptychs', count, _ ).
noun( 'direction', 'directions', both, _ ).
noun( 'direction-finder', 'direction-finders', count, _ ).
noun( 'directive', 'directives', count, _ ).
noun( 'directness', '-', mass, _ ).
noun( 'director', 'directors', count, _ ).
noun( 'directorate', 'directorates', count, _ ).
noun( 'directorship', 'directorships', count, _ ).
noun( 'directory', 'directories', count, _ ).
noun( 'dirge', 'dirges', count, _ ).
noun( 'dirigible', 'dirigibles', count, _ ).
noun( 'dirk', 'dirks', count, _ ).
noun( 'dirndl', 'dirndls', count, _ ).
noun( 'dirt', '-', mass, _ ).
noun( 'dirt-track', 'dirt-tracks', count, _ ).
noun( 'disability', 'disabilities', both, _ ).
noun( 'disablement', 'disablements', both, _ ).
noun( 'disadvantage', 'disadvantages', both, _ ).
noun( 'disaffection', '-', mass, _ ).
noun( 'disagreeableness', '-', mass, _ ).
noun( 'disagreement', 'disagreements', both, _ ).
noun( 'disappearance', 'disappearances', count, _ ).
noun( 'disappointment', 'disappointments', both, _ ).
noun( 'disapprobation', '-', mass, _ ).
noun( 'disapproval', '-', mass, _ ).
noun( 'disarmament', '-', mass, _ ).
noun( 'disarrangement', 'disarrangements', count, _ ).
noun( 'disarray', '-', mass, _ ).
noun( 'disaster', 'disasters', both, _ ).
noun( 'disavowal', 'disavowals', count, _ ).
noun( 'disbandment', 'disbandments', count, _ ).
noun( 'disbelief', '-', mass, _ ).
noun( 'disbursement', 'disbursements', both, _ ).
noun( 'disc', 'discs', count, _ ).
noun( 'discard', 'discards', count, _ ).
noun( 'discernment', '-', mass, _ ).
noun( 'discharge', 'discharges', both, _ ).
noun( 'disciple', 'disciples', count, _ ).
noun( 'disciplinarian', 'disciplinarians', count, _ ).
noun( 'discipline', 'disciplines', both, _ ).
noun( 'disclaimer', 'disclaimers', count, _ ).
noun( 'disclosure', 'disclosures', both, _ ).
noun( 'disco', 'discos', count, _ ).
noun( 'discolouration', 'discolourations', both, _ ).
noun( 'discomfiture', '-', mass, _ ).
noun( 'discomfort', 'discomforts', both, _ ).
noun( 'discomposure', '-', mass, _ ).
noun( 'discontent', 'discontents', both, _ ).
noun( 'discontinuance', 'discontinuances', count, _ ).
noun( 'discontinuity', 'discontinuities', both, _ ).
noun( 'discord', 'discords', both, _ ).
noun( 'discordance', '-', mass, _ ).
noun( 'discotheque', 'discotheques', count, _ ).
noun( 'discount', 'discounts', both, _ ).
noun( 'discouragement', 'discouragements', both, _ ).
noun( 'discourse', 'discourses', count, _ ).
noun( 'discourtesy', 'discourtesies', both, _ ).
noun( 'discoverer', 'discoverers', count, _ ).
noun( 'discovery', 'discoveries', both, _ ).
noun( 'discredit', '-', mass, _ ).
noun( 'discrepancy', 'discrepancies', both, _ ).
noun( 'discreteness', '-', mass, _ ).
noun( 'discretion', '-', mass, _ ).
noun( 'discrimination', '-', mass, _ ).
noun( 'discursiveness', '-', mass, _ ).
noun( 'discus', 'discuses', count, _ ).
noun( 'discussion', 'discussions', both, _ ).
noun( 'disdain', '-', mass, _ ).
noun( 'disease', 'diseases', both, _ ).
noun( 'disembarkation', 'disembarkations', count, _ ).
noun( 'disembarrassment', 'disembarrassments', count, _ ).
noun( 'disenchantment', 'disenchantments', count, _ ).
noun( 'disengagement', 'disengagements', count, _ ).
noun( 'disentanglement', 'disentanglements', count, _ ).
noun( 'disequilibrium', '-', mass, _ ).
noun( 'disestablishment', '-', mass, _ ).
noun( 'disfavour', '-', mass, _ ).
noun( 'disfigurement', 'disfigurements', both, _ ).
noun( 'disfranchisement', '-', mass, _ ).
noun( 'disgrace', '-', mass, _ ).
noun( 'disguise', 'disguises', both, _ ).
noun( 'disgust', '-', mass, _ ).
noun( 'dish', 'dishes', count, _ ).
noun( 'dishabille', '-', mass, _ ).
noun( 'disharmony', '-', mass, _ ).
noun( 'dishcloth', 'dishcloths', count, _ ).
noun( 'dishful', 'dishfuls', count, _ ).
noun( 'dishonesty', 'dishonesties', both, _ ).
noun( 'dishonour', '-', mass, _ ).
noun( 'dishwasher', 'dishwashers', count, _ ).
noun( 'dishwater', '-', mass, _ ).
noun( 'disillusion', '-', mass, _ ).
noun( 'disillusionment', 'disillusionments', both, _ ).
noun( 'disincentive', 'disincentives', count, _ ).
noun( 'disinclination', 'disinclinations', both, _ ).
noun( 'disinfectant', 'disinfectants', count, _ ).
noun( 'disinfection', '-', mass, _ ).
noun( 'disinfestation', '-', mass, _ ).
noun( 'disinflation', '-', mass, _ ).
noun( 'disingenuousness', '-', mass, _ ).
noun( 'disinheritance', '-', mass, _ ).
noun( 'disintegration', 'disintegrations', both, _ ).
noun( 'disinterestedness', '-', mass, _ ).
noun( 'disinterment', 'disinterments', count, _ ).
noun( 'disjointedness', '-', mass, _ ).
noun( 'disk', 'disks', count, _ ).
noun( 'dislike', 'dislikes', both, _ ).
noun( 'dislocation', 'dislocations', both, _ ).
noun( 'dislodgement', 'dislodgements', count, _ ).
noun( 'disloyalty', 'disloyalties', both, _ ).
noun( 'dismantlement', '-', mass, _ ).
noun( 'dismay', '-', mass, _ ).
noun( 'dismemberment', 'dismemberments', count, _ ).
noun( 'dismissal', 'dismissals', both, _ ).
noun( 'disobedience', '-', mass, _ ).
noun( 'disorder', 'disorders', both, _ ).
noun( 'disorganization', '-', mass, _ ).
noun( 'disparagement', 'disparagements', count, _ ).
noun( 'disparity', 'disparities', both, _ ).
noun( 'dispassionateness', '-', mass, _ ).
noun( 'dispatch', 'dispatches', both, _ ).
noun( 'dispatch-box', 'dispatch-boxes', count, _ ).
noun( 'dispatch-rider', 'dispatch-riders', count, _ ).
noun( 'dispensary', 'dispensaries', count, _ ).
noun( 'dispensation', 'dispensations', both, _ ).
noun( 'dispenser', 'dispensers', count, _ ).
noun( 'dispersal', 'dispersals', both, _ ).
noun( 'dispersion', 'dispersions', count, _ ).
noun( 'displacement', 'displacements', both, _ ).
noun( 'display', 'displays', both, _ ).
noun( 'displeasure', '-', mass, _ ).
noun( 'disposal', '-', mass, _ ).
noun( 'disposition', 'dispositions', count, _ ).
noun( 'dispossession', 'dispossessions', count, _ ).
noun( 'disproof', 'disproofs', both, _ ).
noun( 'disproportion', '-', mass, _ ).
noun( 'disputant', 'disputants', count, _ ).
noun( 'disputation', 'disputations', both, _ ).
noun( 'dispute', 'disputes', both, _ ).
noun( 'disqualification', 'disqualifications', both, _ ).
noun( 'disquiet', '-', mass, _ ).
noun( 'disquietude', '-', mass, _ ).
noun( 'disquisition', 'disquisitions', count, _ ).
noun( 'disregard', '-', mass, _ ).
noun( 'disrepair', '-', mass, _ ).
noun( 'disrepute', '-', mass, _ ).
noun( 'disrespect', '-', mass, _ ).
noun( 'disruption', 'disruptions', both, _ ).
noun( 'dissatisfaction', '-', mass, _ ).
noun( 'dissection', 'dissections', both, _ ).
noun( 'dissembler', 'dissemblers', count, _ ).
noun( 'dissemination', '-', mass, _ ).
noun( 'dissension', 'dissensions', both, _ ).
noun( 'dissent', '-', mass, _ ).
noun( 'dissenter', 'dissenters', count, _ ).
noun( 'dissertation', 'dissertations', count, _ ).
noun( 'disservice', 'disservices', both, _ ).
noun( 'dissidence', '-', mass, _ ).
noun( 'dissident', 'dissidents', count, _ ).
noun( 'dissimilarity', 'dissimilarities', both, _ ).
noun( 'dissimilitude', '-', mass, _ ).
noun( 'dissimulation', 'dissimulations', count, _ ).
noun( 'dissipation', '-', mass, _ ).
noun( 'dissociation', '-', mass, _ ).
noun( 'dissolubility', '-', mass, _ ).
noun( 'dissolution', 'dissolutions', both, _ ).
noun( 'dissonance', 'dissonances', both, _ ).
noun( 'dissuasion', '-', mass, _ ).
noun( 'dissyllable', 'dissyllables', count, _ ).
noun( 'distaff', 'distaffs', count, _ ).
noun( 'distance', 'distances', both, _ ).
noun( 'distaste', 'distastes', both, _ ).
noun( 'distastefulness', '-', mass, _ ).
noun( 'distemper', '-', mass, _ ).
noun( 'distension', 'distensions', count, _ ).
noun( 'distillation', 'distillations', both, _ ).
noun( 'distiller', 'distillers', count, _ ).
noun( 'distillery', 'distilleries', count, _ ).
noun( 'distinction', 'distinctions', both, _ ).
noun( 'distinctiveness', '-', mass, _ ).
noun( 'distinctness', '-', mass, _ ).
noun( 'distortion', 'distortions', both, _ ).
noun( 'distraction', 'distractions', both, _ ).
noun( 'distraint', 'distraints', count, _ ).
noun( 'distress', '-', mass, _ ).
noun( 'distribution', 'distributions', both, _ ).
noun( 'distributor', 'distributors', count, _ ).
noun( 'district', 'districts', count, _ ).
noun( 'distrust', '-', both, _ ).
noun( 'distrustfulness', '-', mass, _ ).
noun( 'disturbance', 'disturbances', both, _ ).
noun( 'disunion', '-', mass, _ ).
noun( 'disunity', '-', mass, _ ).
noun( 'disuse', '-', mass, _ ).
noun( 'disyllable', 'disyllables', count, _ ).
noun( 'ditch', 'ditches', count, _ ).
noun( 'dither', 'dithers', count, _ ).
noun( 'ditto', '-', count, _ ).
noun( 'ditty', 'ditties', count, _ ).
noun( 'divagation', 'divagations', both, _ ).
noun( 'divan', 'divans', count, _ ).
noun( 'divan-bed', 'divan-beds', count, _ ).
noun( 'dive', 'dives', count, _ ).
noun( 'dive-bomber', 'dive-bombers', count, _ ).
noun( 'diver', 'divers', count, _ ).
noun( 'divergence', 'divergences', both, _ ).
noun( 'divergency', 'divergencies', both, _ ).
noun( 'diversification', '-', mass, _ ).
noun( 'diversion', 'diversions', both, _ ).
noun( 'diversionist', 'diversionists', count, _ ).
noun( 'diversity', '-', mass, _ ).
noun( 'divide', 'divides', count, _ ).
noun( 'dividend', 'dividends', count, _ ).
noun( 'dividend-warrant', 'dividend-warrants', count, _ ).
noun( 'divination', 'divinations', both, _ ).
noun( 'divine', 'divines', count, _ ).
noun( 'diviner', 'diviners', count, _ ).
noun( 'diving-bell', 'diving-bells', count, _ ).
noun( 'diving-board', 'diving-boards', count, _ ).
noun( 'diving-dress', '-', mass, _ ).
noun( 'diving-suit', 'diving-suits', count, _ ).
noun( 'divinity', 'divinities', both, _ ).
noun( 'division', 'divisions', both, _ ).
noun( 'divisor', 'divisors', count, _ ).
noun( 'divorce', 'divorces', both, _ ).
noun( 'divorcee', 'divorcees', count, _ ).
noun( 'divot', 'divots', count, _ ).
noun( 'divulgence', 'divulgences', count, _ ).
noun( 'divvy', 'divvies', count, _ ).
noun( 'dixie', 'dixies', count, _ ).
noun( 'dizziness', '-', mass, _ ).
noun( 'djinn', 'djinns', count, _ ).
noun( 'do', '-', count, _ ).
noun( 'do', 'dos', count, _ ).
noun( 'do-gooder', 'do-gooders', count, _ ).
noun( 'dobbin', 'dobbins', count, _ ).
noun( 'docility', '-', mass, _ ).
noun( 'dock', 'docks', count, _ ).
noun( 'docker', 'dockers', count, _ ).
noun( 'docket', 'dockets', count, _ ).
noun( 'dockyard', 'dockyards', count, _ ).
noun( 'doctor', 'doctors', count, _ ).
noun( 'doctorate', 'doctorates', count, _ ).
noun( 'doctrinaire', 'doctrinaires', count, _ ).
noun( 'doctrine', 'doctrines', both, _ ).
noun( 'document', 'documents', count, _ ).
noun( 'documentation', '-', mass, _ ).
noun( 'dodderer', 'dodderers', count, _ ).
noun( 'dodge', 'dodges', count, _ ).
noun( 'dodgem', 'dodgems', count, _ ).
noun( 'dodger', 'dodgers', count, _ ).
noun( 'dodo', 'dodos', count, _ ).
noun( 'doe', 'does', count, _ ).
noun( 'doer', 'doers', count, _ ).
noun( 'dog', 'dogs', count, _ ).
noun( 'dog\'s-tooth', '-', mass, _ ).
noun( 'dog-biscuit', 'dog-biscuits', count, _ ).
noun( 'dog-cart', 'dog-carts', count, _ ).
noun( 'dog-collar', 'dog-collars', count, _ ).
noun( 'doge', 'doges', count, _ ).
noun( 'dogfish', 'dogfish', count, _ ).
noun( 'doggedness', '-', mass, _ ).
noun( 'doggerel', '-', mass, _ ).
noun( 'doggie', 'doggies', count, _ ).
noun( 'doggy', 'doggies', count, _ ).
noun( 'doghouse', 'doghouses', count, _ ).
noun( 'dogma', 'dogmas', both, _ ).
noun( 'dogmatism', '-', mass, _ ).
noun( 'dogsbody', 'dogsbodies', count, _ ).
noun( 'dogtooth', '-', mass, _ ).
noun( 'dogtrot', '-', count, _ ).
noun( 'dogwatch', 'dogwatches', count, _ ).
noun( 'dogwood', 'dogwoods', count, _ ).
noun( 'doh', '-', count, _ ).
noun( 'doily', 'doilies', count, _ ).
noun( 'dole', 'doles', both, _ ).
noun( 'doll', 'dolls', count, _ ).
noun( 'dollar', 'dollars', count, _ ).
noun( 'dollop', 'dollops', count, _ ).
noun( 'dolly', 'dollies', count, _ ).
noun( 'dolmen', 'dolmens', count, _ ).
noun( 'dolour', 'dolours', count, _ ).
noun( 'dolphin', 'dolphins', count, _ ).
noun( 'dolt', 'dolts', count, _ ).
noun( 'domain', 'domains', count, _ ).
noun( 'dome', 'domes', count, _ ).
noun( 'domestication', '-', mass, _ ).
noun( 'domesticity', '-', mass, _ ).
noun( 'domicile', 'domiciles', count, _ ).
noun( 'dominance', '-', mass, _ ).
noun( 'domination', '-', mass, _ ).
noun( 'dominie', 'dominies', count, _ ).
noun( 'dominion', 'dominions', both, _ ).
noun( 'domino', 'dominos', count, _ ).
noun( 'don', 'dons', count, _ ).
noun( 'don\'t-know', 'don\'t-knows', count, _ ).
noun( 'donation', 'donations', both, _ ).
noun( 'donjon', 'donjons', count, _ ).
noun( 'donkey', 'donkeys', count, _ ).
noun( 'donkey-jacket', 'donkey-jackets', count, _ ).
noun( 'donkey-work', '-', mass, _ ).
noun( 'donor', 'donors', count, _ ).
noun( 'doodlebug', 'doodlebugs', count, _ ).
noun( 'doom', 'dooms', count, _ ).
noun( 'door', 'doors', count, _ ).
noun( 'doorbell', 'doorbells', count, _ ).
noun( 'doorcase', 'doorcases', count, _ ).
noun( 'doorframe', 'doorframes', count, _ ).
noun( 'doorhandle', 'doorhandles', count, _ ).
noun( 'doorkeeper', 'doorkeepers', count, _ ).
noun( 'doorknob', 'doorknobs', count, _ ).
noun( 'doorknocker', 'doorknockers', count, _ ).
noun( 'doorman', 'doormen', count, _ ).
noun( 'doormat', 'doormats', count, _ ).
noun( 'doornail', 'doornails', count, _ ).
noun( 'doorplate', 'doorplates', count, _ ).
noun( 'doorpost', 'doorposts', count, _ ).
noun( 'doorstep', 'doorsteps', count, _ ).
noun( 'doorstopper', 'doorstoppers', count, _ ).
noun( 'doorway', 'doorways', count, _ ).
noun( 'dope', 'dopes', both, _ ).
noun( 'dormer', 'dormers', count, _ ).
noun( 'dormer-window', 'dormer-windows', count, _ ).
noun( 'dormitory', 'dormitories', count, _ ).
noun( 'dormouse', 'dormice', count, _ ).
noun( 'dory', 'dories', count, _ ).
noun( 'dosage', 'dosages', both, _ ).
noun( 'dose', 'doses', count, _ ).
noun( 'doss-house', 'doss-houses', count, _ ).
noun( 'dosser', 'dossers', count, _ ).
noun( 'dossier', 'dossiers', count, _ ).
noun( 'dot', 'dots', count, _ ).
noun( 'dotage', '-', mass, _ ).
noun( 'dotard', 'dotards', count, _ ).
noun( 'dottle', 'dottles', count, _ ).
noun( 'double', 'doubles', count, _ ).
noun( 'double-bass', 'double-basses', count, _ ).
noun( 'double-cross', 'double-crosses', count, _ ).
noun( 'double-dealer', 'double-dealers', count, _ ).
noun( 'double-dealing', '-', mass, _ ).
noun( 'double-decker', 'double-deckers', count, _ ).
noun( 'double-dutch', '-', mass, _ ).
noun( 'double-entry', 'double-entries', count, _ ).
noun( 'double-first', 'double-firsts', count, _ ).
noun( 'double-spacing', '-', mass, _ ).
noun( 'double-talk', '-', mass, _ ).
noun( 'double-think', '-', mass, _ ).
noun( 'doublet', 'doublets', count, _ ).
noun( 'doubloon', 'doubloons', count, _ ).
noun( 'doubt', 'doubts', both, _ ).
noun( 'douche', 'douches', count, _ ).
noun( 'dough', '-', mass, _ ).
noun( 'doughnut', 'doughnuts', count, _ ).
noun( 'dove', 'doves', count, _ ).
noun( 'dovecote', 'dovecotes', count, _ ).
noun( 'dovetail', 'dovetails', count, _ ).
noun( 'dowager', 'dowagers', count, _ ).
noun( 'dowdiness', '-', mass, _ ).
noun( 'dowel', 'dowels', count, _ ).
noun( 'dower', 'dowers', count, _ ).
noun( 'down', 'downs', both, _ ).
noun( 'down-and-out', 'down-and-outs', count, _ ).
noun( 'downbeat', 'downbeats', count, _ ).
noun( 'downfall', '-', count, _ ).
noun( 'downpour', 'downpours', count, _ ).
noun( 'downrightness', '-', mass, _ ).
noun( 'dowry', 'dowries', count, _ ).
noun( 'dowser', 'dowsers', count, _ ).
noun( 'dowsing', '-', mass, _ ).
noun( 'doxology', 'doxologies', count, _ ).
noun( 'doyen', 'doyens', count, _ ).
noun( 'doyley', 'doyleys', count, _ ).
noun( 'doyly', 'doylies', count, _ ).
noun( 'doz', 'doz', count, _ ).
noun( 'doze', 'dozes', count, _ ).
noun( 'dozen', 'dozens', count, _ ).
noun( 'drabness', '-', mass, _ ).
noun( 'drachm', 'drachms', count, _ ).
noun( 'drachma', 'drachmas', count, _ ).
noun( 'draft', 'drafts', count, _ ).
noun( 'draftee', 'draftees', count, _ ).
noun( 'drafting', 'draftings', count, _ ).
noun( 'draftsman', 'draftsmen', count, _ ).
noun( 'drag', '-', mass, _ ).
noun( 'dragnet', 'dragnets', count, _ ).
noun( 'dragoman', 'dragomans', count, _ ).
noun( 'dragon', 'dragons', count, _ ).
noun( 'dragonfly', 'dragonflies', count, _ ).
noun( 'dragoon', 'dragoons', count, _ ).
noun( 'drain', 'drains', count, _ ).
noun( 'drainage', '-', mass, _ ).
noun( 'drainage-basin', 'drainage-basins', count, _ ).
noun( 'draining-board', 'draining-boards', count, _ ).
noun( 'drainpipe', 'drainpipes', count, _ ).
noun( 'drake', 'drakes', count, _ ).
noun( 'dram', 'drams', count, _ ).
noun( 'drama', 'dramas', both, _ ).
noun( 'dramatics', 'dramatics', mass, _ ).
noun( 'dramatist', 'dramatists', count, _ ).
noun( 'dramatization', 'dramatizations', both, _ ).
noun( 'drape', 'drapes', count, _ ).
noun( 'draper', 'drapers', count, _ ).
noun( 'drapery', 'draperies', both, _ ).
noun( 'draught', 'draughts', both, _ ).
noun( 'draught-horse', 'draught-horses', count, _ ).
noun( 'draughts', 'draughts', mass, _ ).
noun( 'draughtsman', 'draughtsmen', count, _ ).
noun( 'draw', 'draws', count, _ ).
noun( 'drawback', 'drawbacks', both, _ ).
noun( 'drawbridge', 'drawbridges', count, _ ).
noun( 'drawer', 'drawers', count, _ ).
noun( 'drawing', 'drawings', both, _ ).
noun( 'drawing-board', 'drawing-boards', count, _ ).
noun( 'drawing-pin', 'drawing-pins', count, _ ).
noun( 'drawing-room', 'drawing-rooms', count, _ ).
noun( 'drawl', 'drawls', count, _ ).
noun( 'dray', 'drays', count, _ ).
noun( 'dread', '-', mass, _ ).
noun( 'dreadfulness', '-', mass, _ ).
noun( 'dreadnought', 'dreadnoughts', count, _ ).
noun( 'dream', 'dreams', count, _ ).
noun( 'dreamer', 'dreamers', count, _ ).
noun( 'dreamland', 'dreamlands', count, _ ).
noun( 'dreamworld', 'dreamworlds', count, _ ).
noun( 'dreariness', '-', mass, _ ).
noun( 'dredge', 'dredges', count, _ ).
noun( 'dredger', 'dredgers', count, _ ).
noun( 'drenching', 'drenchings', count, _ ).
noun( 'dress', 'dresses', both, _ ).
noun( 'dress-hanger', 'dress-hangers', count, _ ).
noun( 'dressage', '-', mass, _ ).
noun( 'dresser', 'dressers', count, _ ).
noun( 'dressing', 'dressings', both, _ ).
noun( 'dressing-case', 'dressing-cases', count, _ ).
noun( 'dressing-down', '-', count, _ ).
noun( 'dressing-gown', 'dressing-gowns', count, _ ).
noun( 'dressing-table', 'dressing-tables', count, _ ).
noun( 'dressmaker', 'dressmakers', count, _ ).
noun( 'dressmaking', '-', mass, _ ).
noun( 'dribbler', 'dribblers', count, _ ).
noun( 'driblet', 'driblets', count, _ ).
noun( 'drier', 'driers', count, _ ).
noun( 'drift', 'drifts', both, _ ).
noun( 'drift-ice', '-', mass, _ ).
noun( 'drift-net', 'drift-nets', count, _ ).
noun( 'drift-wood', '-', mass, _ ).
noun( 'driftage', '-', mass, _ ).
noun( 'drifter', 'drifters', count, _ ).
noun( 'drill', 'drills', both, _ ).
noun( 'drink', 'drinks', both, _ ).
noun( 'drinker', 'drinkers', count, _ ).
noun( 'drinking', '-', mass, _ ).
noun( 'drinking-bout', 'drinking-bouts', count, _ ).
noun( 'drinking-fountain', 'drinking-fountains', count, _ ).
noun( 'drinking-song', 'drinking-songs', count, _ ).
noun( 'drinking-water', '-', mass, _ ).
noun( 'drip', 'drips', count, _ ).
noun( 'dripping', '-', mass, _ ).
noun( 'dripping-pan', 'dripping-pans', count, _ ).
noun( 'drive', 'drives', both, _ ).
noun( 'drive-in', 'drive-ins', count, _ ).
noun( 'drivel', '-', mass, _ ).
noun( 'driveller', 'drivellers', count, _ ).
noun( 'driver', 'drivers', count, _ ).
noun( 'driveway', 'driveways', count, _ ).
noun( 'driving-belt', 'driving-belts', count, _ ).
noun( 'driving-wheel', 'driving-wheels', count, _ ).
noun( 'drizzle', '-', mass, _ ).
noun( 'drogue', 'drogues', count, _ ).
noun( 'drollery', 'drolleries', both, _ ).
noun( 'dromedary', 'dromedaries', count, _ ).
noun( 'drone', 'drones', both, _ ).
noun( 'droop', '-', count, _ ).
noun( 'drop', 'drops', count, _ ).
noun( 'drop-curtain', 'drop-curtains', count, _ ).
noun( 'drop-kick', 'drop-kicks', count, _ ).
noun( 'dropout', 'dropouts', count, _ ).
noun( 'dropping-zone', 'dropping-zones', count, _ ).
noun( 'dropsy', '-', mass, _ ).
noun( 'droshky', 'droshkies', count, _ ).
noun( 'dross', '-', mass, _ ).
noun( 'drought', 'droughts', both, _ ).
noun( 'drove', 'droves', count, _ ).
noun( 'drover', 'drovers', count, _ ).
noun( 'drowse', '-', count, _ ).
noun( 'drowsiness', '-', mass, _ ).
noun( 'drubbing', '-', count, _ ).
noun( 'drudge', 'drudges', count, _ ).
noun( 'drudgery', '-', mass, _ ).
noun( 'drug', 'drugs', count, _ ).
noun( 'drugget', 'druggets', both, _ ).
noun( 'druggist', 'druggists', count, _ ).
noun( 'drugstore', 'drugstores', count, _ ).
noun( 'druid', 'druids', count, _ ).
noun( 'drum', 'drums', count, _ ).
noun( 'drum-major', 'drum-majors', count, _ ).
noun( 'drum-majorette', 'drum-majorettes', count, _ ).
noun( 'drumfire', '-', mass, _ ).
noun( 'drummer', 'drummers', count, _ ).
noun( 'drumstick', 'drumsticks', count, _ ).
noun( 'drunk', 'drunks', count, _ ).
noun( 'drunkard', 'drunkards', count, _ ).
noun( 'drunkenness', '-', mass, _ ).
noun( 'drupe', 'drupes', count, _ ).
noun( 'dry-cleaner', 'dry-cleaners', count, _ ).
noun( 'dry-cleaning', '-', mass, _ ).
noun( 'dry-walling', '-', mass, _ ).
noun( 'dryad', 'dryads', count, _ ).
noun( 'dryer', 'dryers', count, _ ).
noun( 'dryness', '-', mass, _ ).
noun( 'dubbin', '-', mass, _ ).
noun( 'dubiety', 'dubieties', both, _ ).
noun( 'dubiousness', '-', mass, _ ).
noun( 'ducat', 'ducats', count, _ ).
noun( 'duchess', 'duchesses', count, _ ).
noun( 'duchy', 'duchies', count, _ ).
noun( 'duck', 'duck', both, _ ).
noun( 'ducking', 'duckings', count, _ ).
noun( 'ducking-stool', 'ducking-stools', count, _ ).
noun( 'duckling', 'ducklings', count, _ ).
noun( 'duckweed', '-', mass, _ ).
noun( 'ducky', 'duckies', count, _ ).
noun( 'duct', 'ducts', count, _ ).
noun( 'ductility', '-', mass, _ ).
noun( 'dud', 'duds', count, _ ).
noun( 'dude', 'dudes', count, _ ).
noun( 'dudgeon', '-', mass, _ ).
noun( 'due', 'dues', count, _ ).
noun( 'duel', 'duels', count, _ ).
noun( 'duelist', 'duelists', count, _ ).
noun( 'duellist', 'duellists', count, _ ).
noun( 'duenna', 'duennas', count, _ ).
noun( 'duet', 'duets', count, _ ).
noun( 'duffel', '-', mass, _ ).
noun( 'duffer', 'duffers', count, _ ).
noun( 'duffle', '-', mass, _ ).
noun( 'dug', 'dugs', count, _ ).
noun( 'dugong', 'dugongs', count, _ ).
noun( 'dugout', 'dugouts', count, _ ).
noun( 'duke', 'dukes', count, _ ).
noun( 'dukedom', 'dukedoms', count, _ ).
noun( 'dulcimer', 'dulcimers', count, _ ).
noun( 'dullard', 'dullards', count, _ ).
noun( 'dullness', '-', mass, _ ).
noun( 'dumbbell', 'dumbbells', count, _ ).
noun( 'dumbness', '-', mass, _ ).
noun( 'dumbwaiter', 'dumbwaiters', count, _ ).
noun( 'dumdum', 'dumdums', count, _ ).
noun( 'dummy', 'dummies', count, _ ).
noun( 'dump', 'dumps', count, _ ).
noun( 'dumper', 'dumpers', count, _ ).
noun( 'dumpling', 'dumplings', count, _ ).
noun( 'dun', 'duns', count, _ ).
noun( 'dunce', 'dunces', count, _ ).
noun( 'dunderhead', 'dunderheads', count, _ ).
noun( 'dune', 'dunes', count, _ ).
noun( 'dung', '-', mass, _ ).
noun( 'dungeon', 'dungeons', count, _ ).
noun( 'dunghill', 'dunghills', count, _ ).
noun( 'duodenum', 'duodenums', count, _ ).
noun( 'duologue', 'duologues', count, _ ).
noun( 'dupe', 'dupes', count, _ ).
noun( 'dupl', '-', proper, _ ).
noun( 'duplicate', 'duplicates', count, _ ).
noun( 'duplication', 'duplications', both, _ ).
noun( 'duplicator', 'duplicators', count, _ ).
noun( 'duplicity', '-', mass, _ ).
noun( 'durability', '-', mass, _ ).
noun( 'durable', 'durables', count, _ ).
noun( 'durance', '-', mass, _ ).
noun( 'duration', 'durations', both, _ ).
noun( 'durbar', 'durbars', count, _ ).
noun( 'duress', '-', mass, _ ).
noun( 'dusk', '-', mass, _ ).
noun( 'dust', '-', mass, _ ).
noun( 'dust-bowl', 'dust-bowls', count, _ ).
noun( 'dust-coat', 'dust-coats', count, _ ).
noun( 'dust-jacket', 'dust-jackets', count, _ ).
noun( 'dust-sheet', 'dust-sheets', count, _ ).
noun( 'dust-up', 'dust-ups', count, _ ).
noun( 'dust-wrapper', 'dust-wrappers', count, _ ).
noun( 'dustbin', 'dustbins', count, _ ).
noun( 'dustcart', 'dustcarts', count, _ ).
noun( 'duster', 'dusters', count, _ ).
noun( 'dustman', 'dustmen', count, _ ).
noun( 'dustpan', 'dustpans', count, _ ).
noun( 'duty', 'duties', both, _ ).
noun( 'duvet', 'duvets', count, _ ).
noun( 'dwarf', 'dwarfs', count, _ ).
noun( 'dweller', 'dwellers', count, _ ).
noun( 'dwelling', 'dwellings', count, _ ).
noun( 'dwelling-house', 'dwelling-houses', count, _ ).
noun( 'dyarchy', 'dyarchies', count, _ ).
noun( 'dye', 'dyes', both, _ ).
noun( 'dye-works', 'dye-works', count, _ ).
noun( 'dyer', 'dyers', count, _ ).
noun( 'dyestuff', 'dyestuffs', count, _ ).
noun( 'dyke', 'dykes', count, _ ).
noun( 'dynamic', 'dynamics', count, _ ).
noun( 'dynamism', '-', mass, _ ).
noun( 'dynamite', '-', mass, _ ).
noun( 'dynamo', 'dynamos', count, _ ).
noun( 'dynast', 'dynasts', count, _ ).
noun( 'dynasty', 'dynasties', count, _ ).
noun( 'dyne', 'dynes', count, _ ).
noun( 'dysentery', '-', mass, _ ).
noun( 'dyslexia', '-', mass, _ ).
noun( 'dyspepsia', '-', mass, _ ).
noun( 'dyspeptic', 'dyspeptics', count, _ ).
noun( 'e', '-', count, _ ).
noun( 'eagerness', '-', mass, _ ).
noun( 'eagle', 'eagles', count, _ ).
noun( 'eaglet', 'eaglets', count, _ ).
noun( 'ear', 'ears', count, _ ).
noun( 'ear-trumpet', 'ear-trumpets', count, _ ).
noun( 'earache', 'earaches', both, _ ).
noun( 'eardrop', 'eardrops', count, _ ).
noun( 'eardrum', 'eardrums', count, _ ).
noun( 'earful', 'earfuls', count, _ ).
noun( 'earl', 'earls', count, _ ).
noun( 'earldom', 'earldoms', count, _ ).
noun( 'earmark', 'earmarks', count, _ ).
noun( 'earnest', '-', mass, _ ).
noun( 'earnest-money', '-', mass, _ ).
noun( 'earnestness', '-', mass, _ ).
noun( 'earphone', 'earphones', count, _ ).
noun( 'earpiece', 'earpieces', count, _ ).
noun( 'earring', 'earrings', count, _ ).
noun( 'earshot', '-', mass, _ ).
noun( 'earth', '-', mass, _ ).
noun( 'earth-closet', 'earth-closets', count, _ ).
noun( 'earthenware', '-', mass, _ ).
noun( 'earthnut', 'earthnuts', count, _ ).
noun( 'earthquake', 'earthquakes', count, _ ).
noun( 'earthwork', 'earthworks', count, _ ).
noun( 'earthworm', 'earthworms', count, _ ).
noun( 'earwax', '-', mass, _ ).
noun( 'earwig', 'earwigs', count, _ ).
noun( 'ease', '-', mass, _ ).
noun( 'easel', 'easels', count, _ ).
noun( 'east', '-', mass, _ ).
noun( 'eatable', 'eatables', count, _ ).
noun( 'eater', 'eaters', count, _ ).
noun( 'eating-apple', 'eating-apples', count, _ ).
noun( 'eating-house', 'eating-houses', count, _ ).
noun( 'eau de cologne', '-', mass, _ ).
noun( 'eau-de-vie', '-', mass, _ ).
noun( 'eavesdropper', 'eavesdroppers', count, _ ).
noun( 'ebb', 'ebbs', count, _ ).
noun( 'ebbtide', 'ebbtides', count, _ ).
noun( 'ebonite', '-', mass, _ ).
noun( 'ebony', '-', mass, _ ).
noun( 'ebullience', '-', mass, _ ).
noun( 'eccentric', 'eccentrics', count, _ ).
noun( 'eccentricity', 'eccentricities', both, _ ).
noun( 'ecclesiastic', 'ecclesiastics', count, _ ).
noun( 'echelon', 'echelons', count, _ ).
noun( 'echo', 'echoes', both, _ ).
noun( 'echo-sounder', 'echo-sounders', count, _ ).
noun( 'echo-sounding', 'echo-soundings', count, _ ).
noun( 'eclecticism', '-', mass, _ ).
noun( 'eclipse', 'eclipses', count, _ ).
noun( 'ecliptic', '-', count, _ ).
noun( 'ecologist', 'ecologists', count, _ ).
noun( 'ecology', '-', mass, _ ).
noun( 'economics', 'economics', mass, _ ).
noun( 'economist', 'economists', count, _ ).
noun( 'economy', 'economies', both, _ ).
noun( 'ecosystem', 'ecosystems', count, _ ).
noun( 'ecstasy', 'ecstasies', both, _ ).
noun( 'ectoplasm', '-', mass, _ ).
noun( 'eczema', '-', mass, _ ).
noun( 'ed', '-', proper, _ ).
noun( 'eddy', 'eddies', count, _ ).
noun( 'edelweiss', '-', mass, _ ).
noun( 'edge', 'edges', count, _ ).
noun( 'edging', 'edgings', both, _ ).
noun( 'edibility', '-', mass, _ ).
noun( 'edible', 'edibles', count, _ ).
noun( 'edict', 'edicts', count, _ ).
noun( 'edification', '-', mass, _ ).
noun( 'edifice', 'edifices', count, _ ).
noun( 'edition', 'editions', count, _ ).
noun( 'editor', 'editors', count, _ ).
noun( 'editorial', 'editorials', count, _ ).
noun( 'education', '-', mass, _ ).
noun( 'educationalist', 'educationalists', count, _ ).
noun( 'educationist', 'educationists', count, _ ).
noun( 'educator', 'educators', count, _ ).
noun( 'eel', 'eels', count, _ ).
noun( 'eeriness', '-', mass, _ ).
noun( 'effacement', '-', mass, _ ).
noun( 'effect', 'effects', both, _ ).
noun( 'effectiveness', '-', mass, _ ).
noun( 'effectuality', '-', mass, _ ).
noun( 'effectualness', '-', mass, _ ).
noun( 'effeminacy', '-', mass, _ ).
noun( 'effendi', 'effendis', count, _ ).
noun( 'effervescence', '-', mass, _ ).
noun( 'effeteness', '-', mass, _ ).
noun( 'efficacy', '-', mass, _ ).
noun( 'efficiency', '-', mass, _ ).
noun( 'effigy', 'effigies', count, _ ).
noun( 'efflorescence', '-', mass, _ ).
noun( 'effluent', 'effluents', both, _ ).
noun( 'efflux', 'effluxes', both, _ ).
noun( 'effort', 'efforts', both, _ ).
noun( 'effrontery', 'effronteries', both, _ ).
noun( 'effulgence', '-', mass, _ ).
noun( 'effusion', 'effusions', both, _ ).
noun( 'effusiveness', '-', mass, _ ).
noun( 'eft', 'efts', count, _ ).
noun( 'eg', '-', proper, _ ).
noun( 'egalitarian', 'egalitarians', count, _ ).
noun( 'egalitarianism', '-', mass, _ ).
noun( 'egg', 'eggs', both, _ ).
noun( 'egg-beater', 'egg-beaters', count, _ ).
noun( 'egg-cup', 'egg-cups', count, _ ).
noun( 'egg-whisk', 'egg-whisks', count, _ ).
noun( 'egghead', 'eggheads', count, _ ).
noun( 'eggplant', 'eggplants', both, _ ).
noun( 'eggshake', 'eggshakes', count, _ ).
noun( 'eggshell', 'eggshells', count, _ ).
noun( 'eglantine', '-', mass, _ ).
noun( 'ego', 'egos', count, _ ).
noun( 'egoism', '-', mass, _ ).
noun( 'egoist', 'egoists', count, _ ).
noun( 'egotism', '-', mass, _ ).
noun( 'egotist', 'egotists', count, _ ).
noun( 'egotrip', 'egotrips', count, _ ).
noun( 'egress', 'egresses', both, _ ).
noun( 'egret', 'egrets', count, _ ).
noun( 'eiderdown', 'eiderdowns', count, _ ).
noun( 'eight', 'eights', count, _ ).
noun( 'eighteen', 'eighteens', count, _ ).
noun( 'eighteenth', 'eighteenths', count, _ ).
noun( 'eighth', 'eighths', count, _ ).
noun( 'eightieth', 'eightieths', count, _ ).
noun( 'eightpence', 'eightpences', count, _ ).
noun( 'eightsome', 'eightsomes', count, _ ).
noun( 'eighty', 'eighties', count, _ ).
noun( 'eisteddfod', 'eisteddfods', count, _ ).
noun( 'ejaculation', 'ejaculations', count, _ ).
noun( 'ejection', 'ejections', count, _ ).
noun( 'ejector', 'ejectors', count, _ ).
noun( 'ejector-seat', 'ejector-seats', count, _ ).
noun( 'elaborateness', '-', mass, _ ).
noun( 'elaboration', 'elaborations', both, _ ).
noun( 'eland', 'elands', count, _ ).
noun( 'elastic', '-', mass, _ ).
noun( 'elasticity', 'elasticities', both, _ ).
noun( 'elation', '-', mass, _ ).
noun( 'elbow', 'elbows', count, _ ).
noun( 'elder', 'elders', count, _ ).
noun( 'election', 'elections', both, _ ).
noun( 'electioneering', '-', mass, _ ).
noun( 'elector', 'electors', count, _ ).
noun( 'electorate', 'electorates', count, _ ).
noun( 'electrician', 'electricians', count, _ ).
noun( 'electricity', '-', mass, _ ).
noun( 'electrification', '-', mass, _ ).
noun( 'electrocardiogram', 'electrocardiograms', count, _ ).
noun( 'electrocardiograph', 'electrocardiographs', count, _ ).
noun( 'electrochemistry', '-', mass, _ ).
noun( 'electrocution', 'electrocutions', both, _ ).
noun( 'electrode', 'electrodes', count, _ ).
noun( 'electrolysis', '-', mass, _ ).
noun( 'electrolyte', 'electrolytes', count, _ ).
noun( 'electromagnet', 'electromagnets', count, _ ).
noun( 'electromagnetism', '-', mass, _ ).
noun( 'electron', 'electrons', count, _ ).
noun( 'electronics', 'electronics', mass, _ ).
noun( 'electroplate', '-', mass, _ ).
noun( 'elegance', '-', mass, _ ).
noun( 'elegy', 'elegies', count, _ ).
noun( 'element', 'elements', count, _ ).
noun( 'elephant', 'elephants', count, _ ).
noun( 'elephantiasis', '-', mass, _ ).
noun( 'elevation', 'elevations', both, _ ).
noun( 'elevator', 'elevators', count, _ ).
noun( 'eleven', 'elevens', count, _ ).
noun( 'eleventh', 'elevenths', count, _ ).
noun( 'elf', 'elves', count, _ ).
noun( 'elicitation', 'elicitations', count, _ ).
noun( 'eligibility', '-', mass, _ ).
noun( 'elimination', 'eliminations', both, _ ).
noun( 'elision', 'elisions', both, _ ).
noun( 'elitism', '-', mass, _ ).
noun( 'elitist', 'elitists', count, _ ).
noun( 'elixir', 'elixirs', count, _ ).
noun( 'elk', 'elks', count, _ ).
noun( 'ellipse', 'ellipses', count, _ ).
noun( 'ellipsis', 'ellipses', both, _ ).
noun( 'elm', 'elms', both, _ ).
noun( 'elocution', '-', mass, _ ).
noun( 'elocutionist', 'elocutionists', count, _ ).
noun( 'elongation', 'elongations', both, _ ).
noun( 'elopement', 'elopements', count, _ ).
noun( 'eloquence', '-', mass, _ ).
noun( 'elucidation', 'elucidations', count, _ ).
noun( 'elver', 'elvers', count, _ ).
noun( 'emaciation', '-', mass, _ ).
noun( 'emanation', 'emanations', both, _ ).
noun( 'emancipation', '-', mass, _ ).
noun( 'emasculation', 'emasculations', both, _ ).
noun( 'embalmment', 'embalmments', count, _ ).
noun( 'embankment', 'embankments', count, _ ).
noun( 'embargo', 'embargoes', count, _ ).
noun( 'embarkation', 'embarkations', both, _ ).
noun( 'embarrassment', 'embarrassments', both, _ ).
noun( 'embassy', 'embassies', count, _ ).
noun( 'embellishment', 'embellishments', both, _ ).
noun( 'ember', 'embers', count, _ ).
noun( 'embezzlement', 'embezzlements', both, _ ).
noun( 'embitterment', 'embitterments', count, _ ).
noun( 'emblem', 'emblems', count, _ ).
noun( 'embodiment', 'embodiments', count, _ ).
noun( 'embonpoint', '-', mass, _ ).
noun( 'embrace', 'embraces', count, _ ).
noun( 'embrasure', 'embrasures', count, _ ).
noun( 'embrocation', '-', mass, _ ).
noun( 'embroidery', '-', mass, _ ).
noun( 'embryo', 'embryos', count, _ ).
noun( 'emeer', 'emeers', count, _ ).
noun( 'emendation', 'emendations', both, _ ).
noun( 'emerald', 'emeralds', count, _ ).
noun( 'emergence', '-', mass, _ ).
noun( 'emergency', 'emergencies', both, _ ).
noun( 'emery', '-', mass, _ ).
noun( 'emetic', 'emetics', count, _ ).
noun( 'emigrant', 'emigrants', count, _ ).
noun( 'emigration', 'emigrations', both, _ ).
noun( 'eminence', 'eminences', both, _ ).
noun( 'emir', 'emirs', count, _ ).
noun( 'emirate', 'emirates', count, _ ).
noun( 'emissary', 'emissaries', count, _ ).
noun( 'emission', 'emissions', both, _ ).
noun( 'emolument', 'emoluments', count, _ ).
noun( 'emotion', 'emotions', both, _ ).
noun( 'empathy', '-', mass, _ ).
noun( 'emperor', 'emperors', count, _ ).
noun( 'emphasis', 'emphases', both, _ ).
noun( 'empire', 'empires', both, _ ).
noun( 'empiricism', '-', mass, _ ).
noun( 'empiricist', 'empiricists', count, _ ).
noun( 'emplacement', 'emplacements', count, _ ).
noun( 'employ', 'employs', count, _ ).
noun( 'employee', 'employees', count, _ ).
noun( 'employer', 'employers', count, _ ).
noun( 'employment', '-', mass, _ ).
noun( 'emporium', 'emporiums', count, _ ).
noun( 'empress', 'empresses', count, _ ).
noun( 'emptiness', '-', mass, _ ).
noun( 'empty', 'empties', count, _ ).
noun( 'empyrean', '-', count, _ ).
noun( 'emu', 'emus', count, _ ).
noun( 'emulation', '-', mass, _ ).
noun( 'emulsion', 'emulsions', both, _ ).
noun( 'enactment', 'enactments', both, _ ).
noun( 'enamel', '-', mass, _ ).
noun( 'enc', '-', proper, _ ).
noun( 'encampment', 'encampments', count, _ ).
noun( 'encephalitis', '-', mass, _ ).
noun( 'enchanter', 'enchanters', count, _ ).
noun( 'enchantment', 'enchantments', both, _ ).
noun( 'enchantress', 'enchantresses', count, _ ).
noun( 'encirclement', 'encirclements', count, _ ).
noun( 'enclave', 'enclaves', count, _ ).
noun( 'enclosure', 'enclosures', both, _ ).
noun( 'encomium', 'encomiums', count, _ ).
noun( 'encore', 'encores', count, _ ).
noun( 'encounter', 'encounters', count, _ ).
noun( 'encouragement', 'encouragements', both, _ ).
noun( 'encroachment', 'encroachments', both, _ ).
noun( 'encumbrance', 'encumbrances', count, _ ).
noun( 'encyclical', 'encyclicals', count, _ ).
noun( 'encyclopaedia', 'encyclopaedias', count, _ ).
noun( 'encyclopedia', 'encyclopedias', count, _ ).
noun( 'end', 'ends', count, _ ).
noun( 'end-all', '-', count, _ ).
noun( 'endearment', 'endearments', both, _ ).
noun( 'endeavour', 'endeavours', count, _ ).
noun( 'endemic', 'endemics', count, _ ).
noun( 'ending', 'endings', count, _ ).
noun( 'endive', 'endives', both, _ ).
noun( 'endorsement', 'endorsements', both, _ ).
noun( 'endowment', 'endowments', both, _ ).
noun( 'endurance', '-', mass, _ ).
noun( 'enema', 'enemas', count, _ ).
noun( 'enemy', 'enemies', count, _ ).
noun( 'energy', 'energies', both, _ ).
noun( 'enfant terrible', 'enfants terribles', count, _ ).
noun( 'enforcement', '-', mass, _ ).
noun( 'enfranchisement', 'enfranchisements', count, _ ).
noun( 'engagement', 'engagements', both, _ ).
noun( 'engine', 'engines', count, _ ).
noun( 'engine-driver', 'engine-drivers', count, _ ).
noun( 'engineer', 'engineers', count, _ ).
noun( 'engineering', '-', mass, _ ).
noun( 'engraver', 'engravers', count, _ ).
noun( 'engraving', 'engravings', both, _ ).
noun( 'enhancement', 'enhancements', both, _ ).
noun( 'enigma', 'enigmas', count, _ ).
noun( 'enjoyment', 'enjoyments', both, _ ).
noun( 'enlargement', 'enlargements', both, _ ).
noun( 'enlightenment', '-', mass, _ ).
noun( 'enlistment', 'enlistments', both, _ ).
noun( 'enmity', 'enmities', both, _ ).
noun( 'ennoblement', 'ennoblements', count, _ ).
noun( 'ennui', 'ennuis', both, _ ).
noun( 'enormity', 'enormities', both, _ ).
noun( 'enormousness', '-', mass, _ ).
noun( 'enough', 'enough', mass, _ ).
noun( 'enquirer', 'enquirers', count, _ ).
noun( 'enquiry', 'enquiries', both, _ ).
noun( 'enrichment', '-', mass, _ ).
noun( 'enrolment', 'enrolments', both, _ ).
noun( 'ensemble', 'ensembles', count, _ ).
noun( 'ensign', 'ensigns', count, _ ).
noun( 'ensilage', '-', mass, _ ).
noun( 'enslavement', 'enslavements', count, _ ).
noun( 'entail', 'entails', both, _ ).
noun( 'entanglement', 'entanglements', both, _ ).
noun( 'entente', 'ententes', count, _ ).
noun( 'entente cordiale', '-', count, _ ).
noun( 'enteritis', '-', mass, _ ).
noun( 'enterprise', 'enterprises', both, _ ).
noun( 'entertainer', 'entertainers', count, _ ).
noun( 'entertainment', 'entertainments', both, _ ).
noun( 'enthronement', 'enthronements', count, _ ).
noun( 'enthusiasm', '-', mass, _ ).
noun( 'enthusiast', 'enthusiasts', count, _ ).
noun( 'enticement', 'enticements', both, _ ).
noun( 'entirety', '-', mass, _ ).
noun( 'entitlement', 'entitlements', count, _ ).
noun( 'entity', 'entities', both, _ ).
noun( 'entomologist', 'entomologists', count, _ ).
noun( 'entomology', '-', mass, _ ).
noun( 'entourage', '-', count, _ ).
noun( 'entr\'acte', 'entr\'actes', count, _ ).
noun( 'entr_ee', 'entr_ees', both, _ ).
noun( 'entrance', 'entrances', both, _ ).
noun( 'entrance-fee', 'entrance-fees', count, _ ).
noun( 'entrance-money', '-', mass, _ ).
noun( 'entrant', 'entrants', count, _ ).
noun( 'entreaty', 'entreaties', both, _ ).
noun( 'entrenchment', 'entrenchments', count, _ ).
noun( 'entrepot', 'entrepots', count, _ ).
noun( 'entrepreneur', 'entrepreneurs', count, _ ).
noun( 'entry', 'entries', both, _ ).
noun( 'enumeration', 'enumerations', both, _ ).
noun( 'enunciation', 'enunciations', count, _ ).
noun( 'envelope', 'envelopes', count, _ ).
noun( 'envelopment', 'envelopments', count, _ ).
noun( 'environment', 'environments', both, _ ).
noun( 'envoi', 'envois', count, _ ).
noun( 'envoy', 'envoys', count, _ ).
noun( 'envy', '-', mass, _ ).
noun( 'enzyme', 'enzymes', count, _ ).
noun( 'eon', 'eons', count, _ ).
noun( 'epaulet', 'epaulets', count, _ ).
noun( 'epaulette', 'epaulettes', count, _ ).
noun( 'epic', 'epics', count, _ ).
noun( 'epicentre', 'epicentres', count, _ ).
noun( 'epicure', 'epicures', count, _ ).
noun( 'epicurean', 'epicureans', count, _ ).
noun( 'epidemic', 'epidemics', count, _ ).
noun( 'epidemiologist', 'epidemiologists', count, _ ).
noun( 'epidemiology', '-', mass, _ ).
noun( 'epidermis', '-', mass, _ ).
noun( 'epidiascope', 'epidiascopes', count, _ ).
noun( 'epiglottis', 'epiglottises', count, _ ).
noun( 'epigram', 'epigrams', count, _ ).
noun( 'epilepsy', '-', mass, _ ).
noun( 'epileptic', 'epileptics', count, _ ).
noun( 'epilogue', 'epilogues', count, _ ).
noun( 'episcopalian', 'episcopalians', count, _ ).
noun( 'episode', 'episodes', count, _ ).
noun( 'epistle', 'epistles', count, _ ).
noun( 'epitaph', 'epitaphs', count, _ ).
noun( 'epithet', 'epithets', count, _ ).
noun( 'epitome', 'epitomes', count, _ ).
noun( 'epoch', 'epochs', count, _ ).
noun( 'equal', 'equals', count, _ ).
noun( 'equalitarian', 'equalitarians', count, _ ).
noun( 'equality', '-', mass, _ ).
noun( 'equalization', 'equalizations', count, _ ).
noun( 'equalizer', 'equalizers', count, _ ).
noun( 'equanimity', '-', mass, _ ).
noun( 'equation', 'equations', both, _ ).
noun( 'equator', 'equators', count, _ ).
noun( 'equerry', 'equerries', count, _ ).
noun( 'equestrian', 'equestrians', count, _ ).
noun( 'equilibrium', '-', mass, _ ).
noun( 'equinox', 'equinoxes', count, _ ).
noun( 'equipage', 'equipages', count, _ ).
noun( 'equipment', '-', mass, _ ).
noun( 'equipoise', 'equipoises', both, _ ).
noun( 'equity', 'equities', both, _ ).
noun( 'equivalence', 'equivalences', both, _ ).
noun( 'equivalent', 'equivalents', count, _ ).
noun( 'equivocation', 'equivocations', both, _ ).
noun( 'era', 'eras', count, _ ).
noun( 'eradication', 'eradications', count, _ ).
noun( 'eraser', 'erasers', count, _ ).
noun( 'erasure', 'erasures', both, _ ).
noun( 'erection', 'erections', both, _ ).
noun( 'erectness', '-', mass, _ ).
noun( 'eremite', 'eremites', count, _ ).
noun( 'erg', 'ergs', count, _ ).
noun( 'ergonomics', 'ergonomics', mass, _ ).
noun( 'ermine', '-', mass, _ ).
noun( 'erosion', '-', mass, _ ).
noun( 'eroticism', '-', mass, _ ).
noun( 'errand', 'errands', count, _ ).
noun( 'erratum', 'errata', count, _ ).
noun( 'error', 'errors', both, _ ).
noun( 'eructation', 'eructations', both, _ ).
noun( 'erudition', '-', mass, _ ).
noun( 'eruption', 'eruptions', both, _ ).
noun( 'erysipelas', '-', mass, _ ).
noun( 'escalation', 'escalations', both, _ ).
noun( 'escalator', 'escalators', count, _ ).
noun( 'escalope', 'escalopes', count, _ ).
noun( 'escapade', 'escapades', count, _ ).
noun( 'escape', 'escapes', both, _ ).
noun( 'escapee', 'escapees', count, _ ).
noun( 'escapement', 'escapements', count, _ ).
noun( 'escapism', '-', mass, _ ).
noun( 'escapist', 'escapists', count, _ ).
noun( 'escapologist', 'escapologists', count, _ ).
noun( 'escarpment', 'escarpments', count, _ ).
noun( 'eschatology', '-', mass, _ ).
noun( 'escort', 'escorts', count, _ ).
noun( 'escritoire', 'escritoires', count, _ ).
noun( 'escudo', 'escudos', count, _ ).
noun( 'escutcheon', 'escutcheons', count, _ ).
noun( 'esophagus', 'esophaguses', count, _ ).
noun( 'espalier', 'espaliers', count, _ ).
noun( 'espionage', '-', mass, _ ).
noun( 'esplanade', 'esplanades', count, _ ).
noun( 'espousal', 'espousals', count, _ ).
noun( 'espresso', 'espressos', count, _ ).
noun( 'esprit', '-', mass, _ ).
noun( 'esprit de corps', '-', count, _ ).
noun( 'essay', 'essays', count, _ ).
noun( 'essayist', 'essayists', count, _ ).
noun( 'essence', 'essences', both, _ ).
noun( 'essential', 'essentials', count, _ ).
noun( 'establishment', 'establishments', both, _ ).
noun( 'estaminet', 'estaminets', count, _ ).
noun( 'estate', 'estates', both, _ ).
noun( 'esteem', '-', mass, _ ).
noun( 'esthete', 'esthetes', count, _ ).
noun( 'esthetic', '-', mass, _ ).
noun( 'esthetics', 'esthetics', mass, _ ).
noun( 'estimate', 'estimates', count, _ ).
noun( 'estimation', '-', mass, _ ).
noun( 'estrangement', 'estrangements', both, _ ).
noun( 'estuary', 'estuaries', count, _ ).
noun( 'et al', '-', proper, _ ).
noun( 'et seq', '-', proper, _ ).
noun( 'eta', '-', count, _ ).
noun( 'etc', '-', proper, _ ).
noun( 'etcher', 'etchers', count, _ ).
noun( 'etching', 'etchings', both, _ ).
noun( 'etd', '-', count, _ ).
noun( 'eternity', 'eternities', both, _ ).
noun( 'ether', '-', mass, _ ).
noun( 'ethic', '-', count, _ ).
noun( 'ethics', 'ethics', mass, _ ).
noun( 'ethnographer', 'ethnographers', count, _ ).
noun( 'ethnography', '-', mass, _ ).
noun( 'ethnologist', 'ethnologists', count, _ ).
noun( 'ethnology', '-', mass, _ ).
noun( 'ethos', '-', count, _ ).
noun( 'ethyl', 'ethyls', both, _ ).
noun( 'etiology', 'etiologies', count, _ ).
noun( 'etiquette', '-', mass, _ ).
noun( 'etymologist', 'etymologists', count, _ ).
noun( 'etymology', 'etymologies', both, _ ).
noun( 'eucalyptus', 'eucalyptuses', count, _ ).
noun( 'eugenics', 'eugenics', mass, _ ).
noun( 'eulogist', 'eulogists', count, _ ).
noun( 'eulogy', 'eulogies', both, _ ).
noun( 'eunuch', 'eunuchs', count, _ ).
noun( 'euphemism', 'euphemisms', both, _ ).
noun( 'euphonium', 'euphoniums', count, _ ).
noun( 'euphony', 'euphonies', both, _ ).
noun( 'euphoria', '-', mass, _ ).
noun( 'euphuism', 'euphuisms', count, _ ).
noun( 'eurhythmics', 'eurhythmics', mass, _ ).
noun( 'eurythmics', 'eurythmics', mass, _ ).
noun( 'euthanasia', '-', mass, _ ).
noun( 'evacuation', 'evacuations', both, _ ).
noun( 'evacuee', 'evacuees', count, _ ).
noun( 'evaluation', 'evaluations', count, _ ).
noun( 'evanescence', '-', mass, _ ).
noun( 'evangelicalism', '-', mass, _ ).
noun( 'evangelism', '-', mass, _ ).
noun( 'evangelist', 'evangelists', count, _ ).
noun( 'evaporation', 'evaporations', both, _ ).
noun( 'evasion', 'evasions', both, _ ).
noun( 'evasiveness', '-', mass, _ ).
noun( 'eve', 'eves', count, _ ).
noun( 'even', 'evens', count, _ ).
noun( 'evening', 'evenings', both, _ ).
noun( 'evenness', '-', mass, _ ).
noun( 'evensong', 'evensongs', count, _ ).
noun( 'event', 'events', count, _ ).
noun( 'eventide', '-', mass, _ ).
noun( 'eventuality', 'eventualities', count, _ ).
noun( 'evergreen', 'evergreens', count, _ ).
noun( 'eviction', 'evictions', both, _ ).
noun( 'evidence', '-', mass, _ ).
noun( 'evil', 'evils', both, _ ).
noun( 'evil-doer', 'evil-doers', count, _ ).
noun( 'evocation', 'evocations', count, _ ).
noun( 'evolution', 'evolutions', both, _ ).
noun( 'ewe', 'ewes', count, _ ).
noun( 'ewer', 'ewers', count, _ ).
noun( 'ex-serviceman', 'ex-servicemen', count, _ ).
noun( 'exacerbation', 'exacerbations', count, _ ).
noun( 'exaction', 'exactions', both, _ ).
noun( 'exactitude', '-', mass, _ ).
noun( 'exactness', '-', mass, _ ).
noun( 'exaggeration', 'exaggerations', both, _ ).
noun( 'exaltation', '-', mass, _ ).
noun( 'exam', 'exams', count, _ ).
noun( 'examination', 'examinations', both, _ ).
noun( 'examiner', 'examiners', count, _ ).
noun( 'example', 'examples', both, _ ).
noun( 'exasperation', '-', mass, _ ).
noun( 'excavation', 'excavations', both, _ ).
noun( 'excavator', 'excavators', count, _ ).
noun( 'excellence', 'excellences', both, _ ).
noun( 'excelsior', '-', mass, _ ).
noun( 'exception', 'exceptions', both, _ ).
noun( 'excerpt', 'excerpts', count, _ ).
noun( 'excess', 'excesses', both, _ ).
noun( 'exchange', 'exchanges', both, _ ).
noun( 'exchanger', 'exchangers', count, _ ).
noun( 'exchequer', 'exchequers', count, _ ).
noun( 'excise', '-', mass, _ ).
noun( 'exciseman', 'excisemen', count, _ ).
noun( 'excision', 'excisions', both, _ ).
noun( 'excitability', '-', mass, _ ).
noun( 'excitement', 'excitements', both, _ ).
noun( 'excl', '-', proper, _ ).
noun( 'exclamation', 'exclamations', both, _ ).
noun( 'exclusion', '-', mass, _ ).
noun( 'excogitation', 'excogitations', both, _ ).
noun( 'excommunication', 'excommunications', both, _ ).
noun( 'excoriation', 'excoriations', count, _ ).
noun( 'excrement', '-', mass, _ ).
noun( 'excrescence', 'excrescences', count, _ ).
noun( 'excretion', 'excretions', both, _ ).
noun( 'excursion', 'excursions', count, _ ).
noun( 'excursionist', 'excursionists', count, _ ).
noun( 'excuse', 'excuses', count, _ ).
noun( 'execration', 'execrations', count, _ ).
noun( 'executant', 'executants', count, _ ).
noun( 'execution', 'executions', both, _ ).
noun( 'executioner', 'executioners', count, _ ).
noun( 'executive', 'executives', count, _ ).
noun( 'executor', 'executors', count, _ ).
noun( 'executrix', 'executrixes', count, _ ).
noun( 'exegesis', '-', mass, _ ).
noun( 'exemplification', 'exemplifications', both, _ ).
noun( 'exemption', 'exemptions', both, _ ).
noun( 'exercise', 'exercises', both, _ ).
noun( 'exertion', 'exertions', both, _ ).
noun( 'exhalation', 'exhalations', both, _ ).
noun( 'exhaust', 'exhausts', both, _ ).
noun( 'exhaust-pipe', 'exhaust-pipes', count, _ ).
noun( 'exhaustion', '-', mass, _ ).
noun( 'exhibit', 'exhibits', count, _ ).
noun( 'exhibition', 'exhibitions', count, _ ).
noun( 'exhibitioner', 'exhibitioners', count, _ ).
noun( 'exhibitionism', '-', mass, _ ).
noun( 'exhibitionist', 'exhibitionists', count, _ ).
noun( 'exhibitor', 'exhibitors', count, _ ).
noun( 'exhilaration', '-', mass, _ ).
noun( 'exhortation', 'exhortations', both, _ ).
noun( 'exhumation', 'exhumations', both, _ ).
noun( 'exigency', 'exigencies', count, _ ).
noun( 'exile', 'exiles', both, _ ).
noun( 'existence', 'existences', both, _ ).
noun( 'existentialism', '-', mass, _ ).
noun( 'existentialist', 'existentialists', count, _ ).
noun( 'exit', 'exits', count, _ ).
noun( 'exodus', 'exoduses', count, _ ).
noun( 'exoneration', 'exonerations', count, _ ).
noun( 'exorbitance', '-', mass, _ ).
noun( 'expanse', 'expanses', count, _ ).
noun( 'expansion', '-', mass, _ ).
noun( 'expansiveness', '-', mass, _ ).
noun( 'expatriate', 'expatriates', count, _ ).
noun( 'expectancy', 'expectancies', both, _ ).
noun( 'expectation', 'expectations', both, _ ).
noun( 'expectorant', 'expectorants', count, _ ).
noun( 'expedience', '-', mass, _ ).
noun( 'expediency', '-', mass, _ ).
noun( 'expedient', 'expedients', count, _ ).
noun( 'expedition', 'expeditions', both, _ ).
noun( 'expenditure', 'expenditures', both, _ ).
noun( 'expense', 'expenses', both, _ ).
noun( 'experience', 'experiences', both, _ ).
noun( 'experiment', 'experiments', both, _ ).
noun( 'experimentation', '-', mass, _ ).
noun( 'experimenter', 'experimenters', count, _ ).
noun( 'expert', 'experts', count, _ ).
noun( 'expertise', '-', mass, _ ).
noun( 'expertness', '-', mass, _ ).
noun( 'expiation', '-', mass, _ ).
noun( 'expiration', '-', mass, _ ).
noun( 'expiry', 'expiries', count, _ ).
noun( 'explanation', 'explanations', both, _ ).
noun( 'expletive', 'expletives', count, _ ).
noun( 'explicitness', '-', mass, _ ).
noun( 'exploit', 'exploits', count, _ ).
noun( 'exploitation', '-', mass, _ ).
noun( 'exploration', 'explorations', both, _ ).
noun( 'explorer', 'explorers', count, _ ).
noun( 'explosion', 'explosions', count, _ ).
noun( 'explosive', 'explosives', count, _ ).
noun( 'expo', 'expos', count, _ ).
noun( 'exponent', 'exponents', count, _ ).
noun( 'exponential', 'exponentials', count, _ ).
noun( 'export', 'exports', both, _ ).
noun( 'exportation', '-', mass, _ ).
noun( 'exporter', 'exporters', count, _ ).
noun( 'expos_e', 'expos_es', count, _ ).
noun( 'exposition', 'expositions', both, _ ).
noun( 'expostulation', 'expostulations', both, _ ).
noun( 'exposure', 'exposures', both, _ ).
noun( 'express', 'expresses', both, _ ).
noun( 'expression', 'expressions', both, _ ).
noun( 'expressionism', '-', mass, _ ).
noun( 'expressionist', 'expressionists', count, _ ).
noun( 'expressway', 'expressways', count, _ ).
noun( 'expropriation', '-', mass, _ ).
noun( 'expulsion', 'expulsions', both, _ ).
noun( 'expurgation', 'expurgations', count, _ ).
noun( 'exquisiteness', '-', mass, _ ).
noun( 'extension', 'extensions', both, _ ).
noun( 'extent', '-', mass, _ ).
noun( 'extenuation', 'extenuations', both, _ ).
noun( 'exterior', 'exteriors', count, _ ).
noun( 'extermination', 'exterminations', count, _ ).
noun( 'external', 'externals', count, _ ).
noun( 'extinction', '-', mass, _ ).
noun( 'extinguisher', 'extinguishers', count, _ ).
noun( 'extirpation', '-', mass, _ ).
noun( 'extortion', 'extortions', both, _ ).
noun( 'extra', 'extras', count, _ ).
noun( 'extract', 'extracts', both, _ ).
noun( 'extraction', 'extractions', both, _ ).
noun( 'extradition', 'extraditions', both, _ ).
noun( 'extrapolation', '-', mass, _ ).
noun( 'extravagance', 'extravagances', both, _ ).
noun( 'extravaganza', 'extravaganzas', count, _ ).
noun( 'extreme', 'extremes', count, _ ).
noun( 'extremist', 'extremists', count, _ ).
noun( 'extremity', 'extremities', count, _ ).
noun( 'extrication', '-', mass, _ ).
noun( 'extroversion', '-', mass, _ ).
noun( 'extrovert', 'extroverts', count, _ ).
noun( 'extrusion', 'extrusions', both, _ ).
noun( 'exuberance', '-', mass, _ ).
noun( 'exultation', '-', mass, _ ).
noun( 'eye', 'eyes', count, _ ).
noun( 'eye-opener', 'eye-openers', count, _ ).
noun( 'eye-shadow', '-', mass, _ ).
noun( 'eyeball', 'eyeballs', count, _ ).
noun( 'eyebath', 'eyebaths', count, _ ).
noun( 'eyebrow', 'eyebrows', count, _ ).
noun( 'eyecup', 'eyecups', count, _ ).
noun( 'eyeful', 'eyefuls', count, _ ).
noun( 'eyeglass', 'eyeglasses', count, _ ).
noun( 'eyelash', 'eyelashes', count, _ ).
noun( 'eyelet', 'eyelets', count, _ ).
noun( 'eyelid', 'eyelids', count, _ ).
noun( 'eyepiece', 'eyepieces', count, _ ).
noun( 'eyeshot', '-', mass, _ ).
noun( 'eyesight', '-', mass, _ ).
noun( 'eyesore', 'eyesores', count, _ ).
noun( 'eyestrain', '-', mass, _ ).
noun( 'eyetooth', 'eyeteeth', count, _ ).
noun( 'eyewash', '-', mass, _ ).
noun( 'eyewitness', 'eyewitnesses', count, _ ).
noun( 'eyrie', 'eyries', count, _ ).
noun( 'eyry', 'eyries', count, _ ).
noun( 'f', '-', count, _ ).
noun( 'f\"uhrer', 'f\"uhrers', count, _ ).
noun( 'f^ete', 'f^etes', count, _ ).
noun( 'fa', '-', count, _ ).
noun( 'fa<cade', 'fa<cades', count, _ ).
noun( 'fable', 'fables', both, _ ).
noun( 'fabric', 'fabrics', both, _ ).
noun( 'fabrication', 'fabrications', both, _ ).
noun( 'face', 'faces', count, _ ).
noun( 'face-ache', '-', mass, _ ).
noun( 'face-card', 'face-cards', count, _ ).
noun( 'face-cloth', 'face-cloths', count, _ ).
noun( 'face-cream', 'face-creams', both, _ ).
noun( 'face-lift', 'face-lifts', count, _ ).
noun( 'face-lifting', 'face-liftings', count, _ ).
noun( 'face-pack', 'face-packs', count, _ ).
noun( 'face-powder', 'face-powders', both, _ ).
noun( 'face-saver', 'face-savers', count, _ ).
noun( 'face-saving', '-', mass, _ ).
noun( 'facer', 'facers', count, _ ).
noun( 'facet', 'facets', count, _ ).
noun( 'facetiousness', '-', mass, _ ).
noun( 'facia', 'facias', count, _ ).
noun( 'facial', 'facials', count, _ ).
noun( 'facility', 'facilities', both, _ ).
noun( 'facing', 'facings', count, _ ).
noun( 'facsimile', 'facsimiles', count, _ ).
noun( 'fact', 'facts', both, _ ).
noun( 'faction', 'factions', both, _ ).
noun( 'factor', 'factors', count, _ ).
noun( 'factory', 'factories', count, _ ).
noun( 'factotum', 'factotums', count, _ ).
noun( 'faculty', 'faculties', count, _ ).
noun( 'fad', 'fads', count, _ ).
noun( 'faerie', 'faeries', count, _ ).
noun( 'faery', 'faeries', count, _ ).
noun( 'fag', 'fags', both, _ ).
noun( 'fag-end', 'fag-ends', count, _ ).
noun( 'faggot', 'faggots', count, _ ).
noun( 'faience', '-', mass, _ ).
noun( 'fail', '-', mass, _ ).
noun( 'failing', 'failings', count, _ ).
noun( 'failure', 'failures', both, _ ).
noun( 'faint', 'faints', count, _ ).
noun( 'faintness', '-', mass, _ ).
noun( 'fair', 'fairs', count, _ ).
noun( 'fairground', 'fairgrounds', count, _ ).
noun( 'fairness', '-', mass, _ ).
noun( 'fairway', 'fairways', count, _ ).
noun( 'fairy', 'fairies', count, _ ).
noun( 'fairyland', 'fairylands', count, _ ).
noun( 'fairytale', 'fairytales', count, _ ).
noun( 'fait accompli', 'faits accomplis', count, _ ).
noun( 'faith', 'faiths', both, _ ).
noun( 'faith-healing', '-', mass, _ ).
noun( 'faithfulness', '-', mass, _ ).
noun( 'faithlessness', '-', mass, _ ).
noun( 'fake', 'fakes', count, _ ).
noun( 'fakir', 'fakirs', count, _ ).
noun( 'falcon', 'falcons', count, _ ).
noun( 'falconry', '-', mass, _ ).
noun( 'fall', 'falls', count, _ ).
noun( 'fallacy', 'fallacies', both, _ ).
noun( 'fallibility', '-', mass, _ ).
noun( 'fallout', '-', mass, _ ).
noun( 'fallow', '-', mass, _ ).
noun( 'fallow-deer', 'fallow-deer', count, _ ).
noun( 'falsehood', 'falsehoods', both, _ ).
noun( 'falseness', '-', mass, _ ).
noun( 'falsetto', 'falsettos', count, _ ).
noun( 'falsification', 'falsifications', both, _ ).
noun( 'falsity', 'falsities', both, _ ).
noun( 'fame', '-', mass, _ ).
noun( 'familiar', 'familiars', count, _ ).
noun( 'familiarity', 'familiarities', both, _ ).
noun( 'family', 'families', both, _ ).
noun( 'famine', 'famines', both, _ ).
noun( 'fan', 'fans', count, _ ).
noun( 'fan-belt', 'fan-belts', count, _ ).
noun( 'fanatic', 'fanatics', count, _ ).
noun( 'fanaticism', 'fanaticisms', both, _ ).
noun( 'fancier', 'fanciers', count, _ ).
noun( 'fancy', 'fancies', both, _ ).
noun( 'fandango', 'fandangos', count, _ ).
noun( 'fanfare', 'fanfares', count, _ ).
noun( 'fang', 'fangs', count, _ ).
noun( 'fanlight', 'fanlights', count, _ ).
noun( 'fanny', 'fannies', count, _ ).
noun( 'fantan', '-', mass, _ ).
noun( 'fantasia', 'fantasias', count, _ ).
noun( 'fantasy', 'fantasies', both, _ ).
noun( 'farce', 'farces', both, _ ).
noun( 'fare', 'fares', both, _ ).
noun( 'fare-stage', 'fare-stages', count, _ ).
noun( 'farewell', 'farewells', count, _ ).
noun( 'farm', 'farms', count, _ ).
noun( 'farmer', 'farmers', count, _ ).
noun( 'farmhand', 'farmhands', count, _ ).
noun( 'farmhouse', 'farmhouses', count, _ ).
noun( 'farmstead', 'farmsteads', count, _ ).
noun( 'farmyard', 'farmyards', count, _ ).
noun( 'farrago', 'farragos', count, _ ).
noun( 'farrier', 'farriers', count, _ ).
noun( 'farrow', 'farrows', count, _ ).
noun( 'fart', 'farts', count, _ ).
noun( 'farthing', 'farthings', count, _ ).
noun( 'fascia', 'fascias', count, _ ).
noun( 'fascination', 'fascinations', both, _ ).
noun( 'fascism', '-', mass, _ ).
noun( 'fascist', 'fascists', count, _ ).
noun( 'fashion', 'fashions', both, _ ).
noun( 'fast', 'fasts', count, _ ).
noun( 'fastener', 'fasteners', count, _ ).
noun( 'fastening', 'fastenings', count, _ ).
noun( 'fastidiousness', '-', mass, _ ).
noun( 'fastness', 'fastnesses', both, _ ).
noun( 'fat', 'fats', both, _ ).
noun( 'fatalism', '-', mass, _ ).
noun( 'fatalist', 'fatalists', count, _ ).
noun( 'fatality', 'fatalities', both, _ ).
noun( 'fate', 'fates', both, _ ).
noun( 'fathead', 'fatheads', count, _ ).
noun( 'father', 'fathers', count, _ ).
noun( 'father-in-law', 'fathers-in-law', count, _ ).
noun( 'fatherhood', '-', mass, _ ).
noun( 'fatherland', 'fatherlands', count, _ ).
noun( 'fathom', 'fathoms', count, _ ).
noun( 'fatigue', 'fatigues', both, _ ).
noun( 'fatigue-party', 'fatigue-parties', count, _ ).
noun( 'fatness', '-', mass, _ ).
noun( 'fatuity', 'fatuities', both, _ ).
noun( 'fatuousness', '-', mass, _ ).
noun( 'faucet', 'faucets', count, _ ).
noun( 'fault', 'faults', count, _ ).
noun( 'fault-finder', 'fault-finders', count, _ ).
noun( 'fault-finding', '-', mass, _ ).
noun( 'faun', 'fauns', count, _ ).
noun( 'faux pas', 'faux pas', count, _ ).
noun( 'favour', 'favours', both, _ ).
noun( 'favourite', 'favourites', count, _ ).
noun( 'favouritism', '-', mass, _ ).
noun( 'fawn', 'fawns', count, _ ).
noun( 'fealty', 'fealties', count, _ ).
noun( 'fear', 'fears', both, _ ).
noun( 'fearfulness', '-', mass, _ ).
noun( 'fearlessness', '-', mass, _ ).
noun( 'feasibility', '-', mass, _ ).
noun( 'feast', 'feasts', count, _ ).
noun( 'feast-day', 'feast-days', count, _ ).
noun( 'feat', 'feats', count, _ ).
noun( 'feather', 'feathers', count, _ ).
noun( 'feather-boa', 'feather-boas', count, _ ).
noun( 'featherbed', 'featherbeds', count, _ ).
noun( 'featherweight', 'featherweights', count, _ ).
noun( 'feature', 'features', count, _ ).
noun( 'fecklessness', '-', mass, _ ).
noun( 'fecundity', '-', mass, _ ).
noun( 'federalism', '-', mass, _ ).
noun( 'federalist', 'federalists', count, _ ).
noun( 'federation', 'federations', both, _ ).
noun( 'fee', 'fees', both, _ ).
noun( 'feebleness', '-', mass, _ ).
noun( 'feed', 'feeds', both, _ ).
noun( 'feedback', '-', mass, _ ).
noun( 'feeder', 'feeders', count, _ ).
noun( 'feeding-bottle', 'feeding-bottles', count, _ ).
noun( 'feel', '-', count, _ ).
noun( 'feeler', 'feelers', count, _ ).
noun( 'feeling', 'feelings', both, _ ).
noun( 'feint', 'feints', count, _ ).
noun( 'feldspar', '-', mass, _ ).
noun( 'felicitation', 'felicitations', count, _ ).
noun( 'felicity', 'felicities', both, _ ).
noun( 'fellah', '-', count, _ ).
noun( 'fellow', 'fellows', count, _ ).
noun( 'fellow-feeling', '-', mass, _ ).
noun( 'fellow-traveller', 'fellow-travellers', count, _ ).
noun( 'fellowship', 'fellowships', both, _ ).
noun( 'felon', 'felons', count, _ ).
noun( 'felony', 'felonies', both, _ ).
noun( 'felspar', '-', mass, _ ).
noun( 'felt', '-', mass, _ ).
noun( 'felucca', 'feluccas', count, _ ).
noun( 'fem', '-', proper, _ ).
noun( 'female', 'females', count, _ ).
noun( 'femininity', '-', mass, _ ).
noun( 'feminism', '-', mass, _ ).
noun( 'feminist', 'feminists', count, _ ).
noun( 'femur', 'femurs', count, _ ).
noun( 'fen', 'fens', count, _ ).
noun( 'fence', 'fences', count, _ ).
noun( 'fencer', 'fencers', count, _ ).
noun( 'fencing', '-', mass, _ ).
noun( 'fender', 'fenders', count, _ ).
noun( 'fennel', '-', mass, _ ).
noun( 'feoff', 'feoffs', count, _ ).
noun( 'ferment', 'ferments', count, _ ).
noun( 'fermentation', '-', mass, _ ).
noun( 'fern', 'ferns', count, _ ).
noun( 'ferocity', 'ferocities', both, _ ).
noun( 'ferret', 'ferrets', count, _ ).
noun( 'ferroconcrete', '-', mass, _ ).
noun( 'ferrule', 'ferrules', count, _ ).
noun( 'ferry', 'ferries', count, _ ).
noun( 'ferryboat', 'ferryboats', count, _ ).
noun( 'ferryman', 'ferrymen', count, _ ).
noun( 'fertility', '-', mass, _ ).
noun( 'fertilization', '-', mass, _ ).
noun( 'fertilizer', 'fertilizers', both, _ ).
noun( 'ferule', 'ferules', count, _ ).
noun( 'fervency', '-', mass, _ ).
noun( 'fervour', '-', mass, _ ).
noun( 'festival', 'festivals', count, _ ).
noun( 'festivity', 'festivities', both, _ ).
noun( 'festoon', 'festoons', count, _ ).
noun( 'fete-day', 'fete-days', count, _ ).
noun( 'fetish', 'fetishes', count, _ ).
noun( 'fetlock', 'fetlocks', count, _ ).
noun( 'fetter', 'fetters', count, _ ).
noun( 'fettle', '-', mass, _ ).
noun( 'feud', 'feuds', count, _ ).
noun( 'feudalism', '-', mass, _ ).
noun( 'feudatory', 'feudatories', count, _ ).
noun( 'fever', 'fevers', both, _ ).
noun( 'fewness', '-', mass, _ ).
noun( 'fez', 'fezes', count, _ ).
noun( 'fianc_e', 'fianc_es', count, _ ).
noun( 'fianc_ee', 'fianc_ees', count, _ ).
noun( 'fiasco', 'fiascos', count, _ ).
noun( 'fiat', 'fiats', count, _ ).
noun( 'fib', 'fibs', count, _ ).
noun( 'fibber', 'fibbers', count, _ ).
noun( 'fibbing', '-', mass, _ ).
noun( 'fibre', 'fibres', both, _ ).
noun( 'fibreboard', '-', mass, _ ).
noun( 'fibreglass', '-', mass, _ ).
noun( 'fibrositis', '-', mass, _ ).
noun( 'fibula', 'fibulas', count, _ ).
noun( 'fickleness', '-', mass, _ ).
noun( 'fiction', 'fictions', both, _ ).
noun( 'fiddle', 'fiddles', count, _ ).
noun( 'fiddler', 'fiddlers', count, _ ).
noun( 'fiddlestick', 'fiddlesticks', count, _ ).
noun( 'fidelity', '-', mass, _ ).
noun( 'fidget', 'fidgets', count, _ ).
noun( 'fief', 'fiefs', count, _ ).
noun( 'field', 'fields', count, _ ).
noun( 'field-hospital', 'field-hospitals', count, _ ).
noun( 'field-officer', 'field-officers', count, _ ).
noun( 'fielder', 'fielders', count, _ ).
noun( 'fieldsman', 'fieldsmen', count, _ ).
noun( 'fieldwork', 'fieldworks', both, _ ).
noun( 'fiend', 'fiends', count, _ ).
noun( 'fierceness', '-', mass, _ ).
noun( 'fieriness', '-', mass, _ ).
noun( 'fiesta', 'fiestas', count, _ ).
noun( 'fife', 'fifes', count, _ ).
noun( 'fifteen', 'fifteens', count, _ ).
noun( 'fifteenth', 'fifteenths', count, _ ).
noun( 'fifth', 'fifths', count, _ ).
noun( 'fiftieth', 'fiftieths', count, _ ).
noun( 'fifty', 'fifties', count, _ ).
noun( 'fig', 'figs', count, _ ).
noun( 'fig', '-', count, _ ).
noun( 'fig-leaf', 'fig-leaves', count, _ ).
noun( 'fight', 'fights', both, _ ).
noun( 'fighter', 'fighters', count, _ ).
noun( 'fighting', '-', mass, _ ).
noun( 'figment', 'figments', count, _ ).
noun( 'figure', 'figures', count, _ ).
noun( 'figurehead', 'figureheads', count, _ ).
noun( 'filament', 'filaments', count, _ ).
noun( 'filature', 'filatures', count, _ ).
noun( 'filbert', 'filberts', count, _ ).
noun( 'file', 'files', count, _ ).
noun( 'filibuster', 'filibusters', count, _ ).
noun( 'filigree', '-', mass, _ ).
noun( 'fill', 'fills', both, _ ).
noun( 'fillet', 'fillets', count, _ ).
noun( 'filling', 'fillings', both, _ ).
noun( 'fillip', 'fillips', count, _ ).
noun( 'filly', 'fillies', count, _ ).
noun( 'film', 'films', both, _ ).
noun( 'film-star', 'film-stars', count, _ ).
noun( 'filter', 'filters', count, _ ).
noun( 'filth', '-', mass, _ ).
noun( 'filthiness', '-', mass, _ ).
noun( 'filtrate', 'filtrates', count, _ ).
noun( 'filtration', '-', mass, _ ).
noun( 'fin', 'fins', count, _ ).
noun( 'final', 'finals', count, _ ).
noun( 'finale', 'finales', count, _ ).
noun( 'finalist', 'finalists', count, _ ).
noun( 'finality', '-', mass, _ ).
noun( 'finance', 'finances', both, _ ).
noun( 'financier', 'financiers', count, _ ).
noun( 'finch', 'finches', count, _ ).
noun( 'find', 'finds', count, _ ).
noun( 'finder', 'finders', count, _ ).
noun( 'finding', 'findings', count, _ ).
noun( 'fine', 'fines', count, _ ).
noun( 'fineness', '-', mass, _ ).
noun( 'finery', '-', mass, _ ).
noun( 'finesse', 'finesses', both, _ ).
noun( 'finger', 'fingers', count, _ ).
noun( 'finger-alphabet', 'finger-alphabets', count, _ ).
noun( 'finger-bowl', 'finger-bowls', count, _ ).
noun( 'finger-plate', 'finger-plates', count, _ ).
noun( 'finger-post', 'finger-posts', count, _ ).
noun( 'fingerboard', 'fingerboards', count, _ ).
noun( 'fingermark', 'fingermarks', count, _ ).
noun( 'fingernail', 'fingernails', count, _ ).
noun( 'fingerprint', 'fingerprints', count, _ ).
noun( 'fingerstall', 'fingerstalls', count, _ ).
noun( 'fingertip', 'fingertips', count, _ ).
noun( 'finis', '-', count, _ ).
noun( 'finish', 'finishes', both, _ ).
noun( 'finnan', '-', mass, _ ).
noun( 'finnan haddie', '-', mass, _ ).
noun( 'finnan haddock', '-', mass, _ ).
noun( 'fiord', 'fiords', count, _ ).
noun( 'fir', 'firs', both, _ ).
noun( 'fir-cone', 'fir-cones', count, _ ).
noun( 'fire', 'fires', both, _ ).
noun( 'fire-alarm', 'fire-alarms', count, _ ).
noun( 'fire-brigade', 'fire-brigades', count, _ ).
noun( 'fire-control', '-', mass, _ ).
noun( 'fire-eater', 'fire-eaters', count, _ ).
noun( 'fire-engine', 'fire-engines', count, _ ).
noun( 'fire-escape', 'fire-escapes', count, _ ).
noun( 'fire-extinguisher', 'fire-extinguishers', count, _ ).
noun( 'fire-fighter', 'fire-fighters', count, _ ).
noun( 'fire-hose', 'fire-hoses', count, _ ).
noun( 'fire-power', '-', mass, _ ).
noun( 'fire-raising', '-', mass, _ ).
noun( 'fire-walker', 'fire-walkers', count, _ ).
noun( 'fire-walking', '-', mass, _ ).
noun( 'fire-watcher', 'fire-watchers', count, _ ).
noun( 'fire-watching', '-', mass, _ ).
noun( 'firearm', 'firearms', count, _ ).
noun( 'fireball', 'fireballs', count, _ ).
noun( 'firebird', 'firebirds', count, _ ).
noun( 'firebomb', 'firebombs', count, _ ).
noun( 'firebox', 'fireboxes', count, _ ).
noun( 'firebrand', 'firebrands', count, _ ).
noun( 'firebreak', 'firebreaks', count, _ ).
noun( 'firebrick', 'firebricks', count, _ ).
noun( 'firebug', 'firebugs', count, _ ).
noun( 'fireclay', '-', mass, _ ).
noun( 'firecracker', 'firecrackers', count, _ ).
noun( 'firedamp', '-', mass, _ ).
noun( 'firedog', 'firedogs', count, _ ).
noun( 'firefly', 'fireflies', count, _ ).
noun( 'fireguard', 'fireguards', count, _ ).
noun( 'firelight', 'firelights', count, _ ).
noun( 'firelighter', 'firelighters', count, _ ).
noun( 'fireman', 'firemen', count, _ ).
noun( 'fireplace', 'fireplaces', count, _ ).
noun( 'fireside', 'firesides', count, _ ).
noun( 'firestone', '-', mass, _ ).
noun( 'firewater', '-', mass, _ ).
noun( 'firewood', '-', mass, _ ).
noun( 'firework', 'fireworks', count, _ ).
noun( 'firing-line', 'firing-lines', count, _ ).
noun( 'firing-party', 'firing-parties', count, _ ).
noun( 'firing-squad', 'firing-squads', count, _ ).
noun( 'firkin', 'firkins', count, _ ).
noun( 'firm', 'firms', count, _ ).
noun( 'firmament', 'firmaments', count, _ ).
noun( 'firmness', '-', mass, _ ).
noun( 'first', 'firsts', count, _ ).
noun( 'first-nighter', 'first-nighters', count, _ ).
noun( 'firstborn', 'firstborns', count, _ ).
noun( 'firth', 'firths', count, _ ).
noun( 'fish', 'fish', both, _ ).
noun( 'fish-hook', 'fish-hooks', count, _ ).
noun( 'fish-knife', 'fish-knives', count, _ ).
noun( 'fish-slice', 'fish-slices', count, _ ).
noun( 'fishball', 'fishballs', count, _ ).
noun( 'fishbone', 'fishbones', count, _ ).
noun( 'fishcake', 'fishcakes', count, _ ).
noun( 'fisher', 'fishers', count, _ ).
noun( 'fisherman', 'fishermen', count, _ ).
noun( 'fishery', 'fisheries', count, _ ).
noun( 'fishing', '-', mass, _ ).
noun( 'fishing-line', 'fishing-lines', count, _ ).
noun( 'fishing-rod', 'fishing-rods', count, _ ).
noun( 'fishing-tackle', '-', mass, _ ).
noun( 'fishmonger', 'fishmongers', count, _ ).
noun( 'fishpaste', 'fishpastes', both, _ ).
noun( 'fishplate', 'fishplates', count, _ ).
noun( 'fishwife', 'fishwives', count, _ ).
noun( 'fission', '-', mass, _ ).
noun( 'fissure', 'fissures', count, _ ).
noun( 'fist', 'fists', count, _ ).
noun( 'fistula', 'fistulas', count, _ ).
noun( 'fit', 'fits', count, _ ).
noun( 'fitment', 'fitments', count, _ ).
noun( 'fitness', '-', mass, _ ).
noun( 'fitter', 'fitters', count, _ ).
noun( 'fitting', 'fittings', count, _ ).
noun( 'five', 'fives', count, _ ).
noun( 'fivepence', 'fivepences', count, _ ).
noun( 'fiver', 'fivers', count, _ ).
noun( 'fives', 'fives', mass, _ ).
noun( 'fix', 'fixes', count, _ ).
noun( 'fixation', 'fixations', both, _ ).
noun( 'fixative', 'fixatives', count, _ ).
noun( 'fixture', 'fixtures', count, _ ).
noun( 'fizz', '-', mass, _ ).
noun( 'fjord', 'fjords', count, _ ).
noun( 'flabbiness', '-', mass, _ ).
noun( 'flaccidity', '-', mass, _ ).
noun( 'flag', 'flags', count, _ ).
noun( 'flag-captain', 'flag-captains', count, _ ).
noun( 'flag-day', 'flag-days', count, _ ).
noun( 'flagellant', 'flagellants', count, _ ).
noun( 'flagellation', 'flagellations', count, _ ).
noun( 'flageolet', 'flageolets', count, _ ).
noun( 'flagon', 'flagons', count, _ ).
noun( 'flagpole', 'flagpoles', count, _ ).
noun( 'flagship', 'flagships', count, _ ).
noun( 'flagstaff', 'flagstaffs', count, _ ).
noun( 'flagstone', 'flagstones', count, _ ).
noun( 'flail', 'flails', count, _ ).
noun( 'flair', 'flairs', both, _ ).
noun( 'flak', '-', mass, _ ).
noun( 'flake', 'flakes', count, _ ).
noun( 'flakiness', '-', mass, _ ).
noun( 'flambeau', 'flambeaus', count, _ ).
noun( 'flamboyance', '-', mass, _ ).
noun( 'flame', 'flames', both, _ ).
noun( 'flamethrower', 'flamethrowers', count, _ ).
noun( 'flamingo', 'flamingos', count, _ ).
noun( 'flan', 'flans', count, _ ).
noun( 'flange', 'flanges', count, _ ).
noun( 'flank', 'flanks', count, _ ).
noun( 'flannel', 'flannels', both, _ ).
noun( 'flannelette', '-', mass, _ ).
noun( 'flap', 'flaps', count, _ ).
noun( 'flapjack', 'flapjacks', both, _ ).
noun( 'flapper', 'flappers', count, _ ).
noun( 'flare', 'flares', both, _ ).
noun( 'flare-path', 'flare-paths', count, _ ).
noun( 'flare-up', 'flare-ups', count, _ ).
noun( 'flash', 'flashes', count, _ ).
noun( 'flashback', 'flashbacks', count, _ ).
noun( 'flashbulb', 'flashbulbs', count, _ ).
noun( 'flashgun', 'flashguns', count, _ ).
noun( 'flashlight', 'flashlights', count, _ ).
noun( 'flashpoint', 'flashpoints', count, _ ).
noun( 'flask', 'flasks', count, _ ).
noun( 'flat', 'flats', count, _ ).
noun( 'flat-car', 'flat-cars', count, _ ).
noun( 'flat-iron', 'flat-irons', count, _ ).
noun( 'flatfish', 'flatfish', count, _ ).
noun( 'flatlet', 'flatlets', count, _ ).
noun( 'flatness', '-', mass, _ ).
noun( 'flatterer', 'flatterers', count, _ ).
noun( 'flattery', 'flatteries', both, _ ).
noun( 'flattop', 'flattops', count, _ ).
noun( 'flatulence', '-', mass, _ ).
noun( 'flautist', 'flautists', count, _ ).
noun( 'flavour', 'flavours', both, _ ).
noun( 'flavouring', 'flavourings', both, _ ).
noun( 'flaw', 'flaws', count, _ ).
noun( 'flax', '-', mass, _ ).
noun( 'flea', 'fleas', count, _ ).
noun( 'flea-bite', 'flea-bites', count, _ ).
noun( 'fleapit', 'fleapits', count, _ ).
noun( 'fleck', 'flecks', count, _ ).
noun( 'fledgeling', 'fledgelings', count, _ ).
noun( 'fledgling', 'fledglings', count, _ ).
noun( 'fleece', 'fleeces', both, _ ).
noun( 'fleet', 'fleets', count, _ ).
noun( 'fleetness', '-', mass, _ ).
noun( 'flesh', '-', mass, _ ).
noun( 'flesh-wound', 'flesh-wounds', count, _ ).
noun( 'fleur-de-lis', 'fleurs-de-lis', count, _ ).
noun( 'fleur-de-lys', 'fleurs-de-lys', count, _ ).
noun( 'flex', 'flexes', both, _ ).
noun( 'flexibility', '-', mass, _ ).
noun( 'flibbertigibbet', 'flibbertigibbets', count, _ ).
noun( 'flick', 'flicks', count, _ ).
noun( 'flick-knife', 'flick-knives', count, _ ).
noun( 'flicker', 'flickers', count, _ ).
noun( 'flier', 'fliers', count, _ ).
noun( 'flight', 'flights', both, _ ).
noun( 'flimsiness', '-', mass, _ ).
noun( 'flimsy', '-', mass, _ ).
noun( 'fling', 'flings', count, _ ).
noun( 'flint', 'flints', both, _ ).
noun( 'flintstone', '-', mass, _ ).
noun( 'flip', 'flips', count, _ ).
noun( 'flippancy', 'flippancies', both, _ ).
noun( 'flipper', 'flippers', count, _ ).
noun( 'flirt', 'flirts', count, _ ).
noun( 'flirtation', 'flirtations', both, _ ).
noun( 'flit', 'flits', count, _ ).
noun( 'float', 'floats', count, _ ).
noun( 'floatation', 'floatations', both, _ ).
noun( 'flock', 'flocks', count, _ ).
noun( 'floe', 'floes', count, _ ).
noun( 'flogging', 'floggings', both, _ ).
noun( 'flood', 'floods', count, _ ).
noun( 'flood-tide', 'flood-tides', count, _ ).
noun( 'floodgate', 'floodgates', count, _ ).
noun( 'floor', 'floors', count, _ ).
noun( 'floor-walker', 'floor-walkers', count, _ ).
noun( 'floorboard', 'floorboards', count, _ ).
noun( 'flooring', '-', mass, _ ).
noun( 'floozie', 'floozies', count, _ ).
noun( 'floozy', 'floozies', count, _ ).
noun( 'flop', 'flops', count, _ ).
noun( 'floriculture', '-', mass, _ ).
noun( 'florin', 'florins', count, _ ).
noun( 'florist', 'florists', count, _ ).
noun( 'floss', '-', mass, _ ).
noun( 'flotation', 'flotations', both, _ ).
noun( 'flotilla', 'flotillas', count, _ ).
noun( 'flotsam', '-', mass, _ ).
noun( 'flounce', 'flounces', count, _ ).
noun( 'flounder', 'flounders', count, _ ).
noun( 'flour', '-', mass, _ ).
noun( 'flourish', 'flourishes', count, _ ).
noun( 'flow', '-', count, _ ).
noun( 'flower', 'flowers', count, _ ).
noun( 'flower-girl', 'flower-girls', count, _ ).
noun( 'flowerbed', 'flowerbeds', count, _ ).
noun( 'flowerpot', 'flowerpots', count, _ ).
noun( 'flu', '-', mass, _ ).
noun( 'fluctuation', 'fluctuations', both, _ ).
noun( 'flue', 'flues', count, _ ).
noun( 'fluency', '-', mass, _ ).
noun( 'fluff', '-', mass, _ ).
noun( 'fluid', 'fluids', both, _ ).
noun( 'fluidity', 'fluidities', both, _ ).
noun( 'fluke', 'flukes', count, _ ).
noun( 'flume', 'flumes', count, _ ).
noun( 'flunkey', 'flunkeys', count, _ ).
noun( 'flunky', 'flunkies', count, _ ).
noun( 'fluorescence', 'fluorescences', both, _ ).
noun( 'fluoridation', 'fluoridations', count, _ ).
noun( 'fluoride', 'fluorides', both, _ ).
noun( 'fluoridization', 'fluoridizations', count, _ ).
noun( 'fluorine', '-', mass, _ ).
noun( 'flurry', 'flurries', count, _ ).
noun( 'flush', 'flushes', both, _ ).
noun( 'fluster', 'flusters', count, _ ).
noun( 'flute', 'flutes', count, _ ).
noun( 'fluting', '-', mass, _ ).
noun( 'flutist', 'flutists', count, _ ).
noun( 'flutter', 'flutters', both, _ ).
noun( 'flux', 'fluxes', both, _ ).
noun( 'fly', 'flies', count, _ ).
noun( 'fly-fishing', '-', mass, _ ).
noun( 'fly-swat', 'fly-swats', count, _ ).
noun( 'fly-swatter', 'fly-swatters', count, _ ).
noun( 'flycatcher', 'flycatchers', count, _ ).
noun( 'flyer', 'flyers', count, _ ).
noun( 'flying-bomb', 'flying-bombs', count, _ ).
noun( 'flying-fish', 'flying-fish', count, _ ).
noun( 'flying-fox', 'flying-foxes', count, _ ).
noun( 'flying-squad', 'flying-squads', count, _ ).
noun( 'flyleaf', 'flyleaves', count, _ ).
noun( 'flyover', 'flyovers', count, _ ).
noun( 'flypaper', 'flypapers', count, _ ).
noun( 'flypast', 'flypasts', count, _ ).
noun( 'flytrap', 'flytraps', count, _ ).
noun( 'flyweight', 'flyweights', count, _ ).
noun( 'flywheel', 'flywheels', count, _ ).
noun( 'fo\'c\'sle', 'fo\'c\'sles', count, _ ).
noun( 'foal', 'foals', count, _ ).
noun( 'foam', 'foams', both, _ ).
noun( 'foam-rubber', '-', mass, _ ).
noun( 'focus', 'focuses', count, _ ).
noun( 'fodder', '-', mass, _ ).
noun( 'foe', 'foes', count, _ ).
noun( 'foetus', 'foetuses', count, _ ).
noun( 'fog', 'fogs', both, _ ).
noun( 'fogbank', 'fogbanks', count, _ ).
noun( 'fogey', 'fogeys', count, _ ).
noun( 'foghorn', 'foghorns', count, _ ).
noun( 'foglamp', 'foglamps', count, _ ).
noun( 'fogsignal', 'fogsignals', count, _ ).
noun( 'foible', 'foibles', count, _ ).
noun( 'foil', 'foils', both, _ ).
noun( 'fold', 'folds', count, _ ).
noun( 'folder', 'folders', count, _ ).
noun( 'foliage', '-', mass, _ ).
noun( 'folio', 'folios', count, _ ).
noun( 'folk', 'folks', count, _ ).
noun( 'folk-dance', 'folk-dances', count, _ ).
noun( 'folklore', '-', mass, _ ).
noun( 'folksong', 'folksongs', count, _ ).
noun( 'folktale', 'folktales', count, _ ).
noun( 'follow-on', 'follow-ons', count, _ ).
noun( 'follow-through', 'follow-throughs', count, _ ).
noun( 'follow-up', 'follow-ups', count, _ ).
noun( 'follower', 'followers', count, _ ).
noun( 'following', 'followings', count, _ ).
noun( 'folly', 'follies', both, _ ).
noun( 'fomentation', 'fomentations', both, _ ).
noun( 'fondant', 'fondants', count, _ ).
noun( 'fondness', '-', mass, _ ).
noun( 'font', 'fonts', count, _ ).
noun( 'food', 'foods', both, _ ).
noun( 'foodstuff', 'foodstuffs', count, _ ).
noun( 'fool', 'fools', count, _ ).
noun( 'foolery', '-', mass, _ ).
noun( 'foolhardiness', '-', mass, _ ).
noun( 'foolishness', '-', mass, _ ).
noun( 'foolscap', '-', mass, _ ).
noun( 'foot', 'feet', count, _ ).
noun( 'foot-and-mouth', '-', mass, _ ).
noun( 'foot-bath', 'foot-baths', count, _ ).
noun( 'foot-pound', 'foot-pounds', count, _ ).
noun( 'foot-race', 'foot-races', count, _ ).
noun( 'foot-rot', '-', mass, _ ).
noun( 'footage', '-', mass, _ ).
noun( 'football', 'footballs', both, _ ).
noun( 'footballer', 'footballers', count, _ ).
noun( 'footboard', 'footboards', count, _ ).
noun( 'footbridge', 'footbridges', count, _ ).
noun( 'footer', 'footers', count, _ ).
noun( 'footfall', 'footfalls', count, _ ).
noun( 'footfault', 'footfaults', count, _ ).
noun( 'foothold', 'footholds', count, _ ).
noun( 'footing', '-', count, _ ).
noun( 'footman', 'footmen', count, _ ).
noun( 'footmark', 'footmarks', count, _ ).
noun( 'footnote', 'footnotes', count, _ ).
noun( 'footpath', 'footpaths', count, _ ).
noun( 'footplate', 'footplates', count, _ ).
noun( 'footprint', 'footprints', count, _ ).
noun( 'footslogger', 'footsloggers', count, _ ).
noun( 'footstep', 'footsteps', count, _ ).
noun( 'footstool', 'footstools', count, _ ).
noun( 'footwear', '-', mass, _ ).
noun( 'footwork', '-', mass, _ ).
noun( 'fop', 'fops', count, _ ).
noun( 'forage', '-', mass, _ ).
noun( 'foray', 'forays', count, _ ).
noun( 'forbear', 'forbears', count, _ ).
noun( 'forbearance', '-', mass, _ ).
noun( 'force', 'forces', both, _ ).
noun( 'force majeure', '-', mass, _ ).
noun( 'forcefulness', '-', mass, _ ).
noun( 'forcemeat', '-', mass, _ ).
noun( 'ford', 'fords', count, _ ).
noun( 'fore', '-', mass, _ ).
noun( 'forearm', 'forearms', count, _ ).
noun( 'foreboding', 'forebodings', both, _ ).
noun( 'forecast', 'forecasts', count, _ ).
noun( 'forecaster', 'forecasters', count, _ ).
noun( 'forecastle', 'forecastles', count, _ ).
noun( 'foreclosure', 'foreclosures', both, _ ).
noun( 'forecourt', 'forecourts', count, _ ).
noun( 'forefather', 'forefathers', count, _ ).
noun( 'forefinger', 'forefingers', count, _ ).
noun( 'forefoot', 'forefeet', count, _ ).
noun( 'forefront', 'forefronts', count, _ ).
noun( 'foreground', 'foregrounds', count, _ ).
noun( 'forehead', 'foreheads', count, _ ).
noun( 'foreigner', 'foreigners', count, _ ).
noun( 'foreknowledge', '-', mass, _ ).
noun( 'foreland', 'forelands', count, _ ).
noun( 'foreleg', 'forelegs', count, _ ).
noun( 'forelock', 'forelocks', count, _ ).
noun( 'foreman', 'foremen', count, _ ).
noun( 'foremast', 'foremasts', count, _ ).
noun( 'forename', 'forenames', count, _ ).
noun( 'forenoon', 'forenoons', count, _ ).
noun( 'forerunner', 'forerunners', count, _ ).
noun( 'foresail', 'foresails', count, _ ).
noun( 'foreshore', 'foreshores', count, _ ).
noun( 'foresight', '-', mass, _ ).
noun( 'foreskin', 'foreskins', count, _ ).
noun( 'forest', 'forests', both, _ ).
noun( 'forester', 'foresters', count, _ ).
noun( 'forestry', '-', mass, _ ).
noun( 'foretaste', 'foretastes', count, _ ).
noun( 'forethought', '-', mass, _ ).
noun( 'foretop', 'foretops', count, _ ).
noun( 'forewoman', 'forewomen', count, _ ).
noun( 'foreword', 'forewords', count, _ ).
noun( 'forfeit', 'forfeits', count, _ ).
noun( 'forfeiture', '-', mass, _ ).
noun( 'forge', 'forges', count, _ ).
noun( 'forger', 'forgers', count, _ ).
noun( 'forgery', 'forgeries', both, _ ).
noun( 'forget-me-not', 'forget-me-nots', count, _ ).
noun( 'forgetfulness', '-', mass, _ ).
noun( 'forging', 'forgings', count, _ ).
noun( 'forgiveness', '-', mass, _ ).
noun( 'fork', 'forks', count, _ ).
noun( 'forlornness', '-', mass, _ ).
noun( 'form', 'forms', both, _ ).
noun( 'formaldehyde', '-', mass, _ ).
noun( 'formalin', '-', mass, _ ).
noun( 'formalism', '-', mass, _ ).
noun( 'formality', 'formalities', both, _ ).
noun( 'format', 'formats', count, _ ).
noun( 'formation', 'formations', both, _ ).
noun( 'formula', 'formulas', count, _ ).
noun( 'formulation', 'formulations', both, _ ).
noun( 'fornication', '-', mass, _ ).
noun( 'forsythia', '-', mass, _ ).
noun( 'fort', 'forts', count, _ ).
noun( 'forte', 'fortes', count, _ ).
noun( 'fortieth', 'fortieths', count, _ ).
noun( 'fortification', 'fortifications', both, _ ).
noun( 'fortitude', '-', mass, _ ).
noun( 'fortnight', 'fortnights', count, _ ).
noun( 'fortress', 'fortresses', count, _ ).
noun( 'fortune', 'fortunes', both, _ ).
noun( 'forty', 'forties', count, _ ).
noun( 'forum', 'forums', count, _ ).
noun( 'forward', 'forwards', count, _ ).
noun( 'forwardness', '-', mass, _ ).
noun( 'fosse', 'fosses', count, _ ).
noun( 'fossil', 'fossils', count, _ ).
noun( 'fossilization', 'fossilizations', both, _ ).
noun( 'foster-brother', 'foster-brothers', count, _ ).
noun( 'foster-child', 'foster-children', count, _ ).
noun( 'foster-father', 'foster-fathers', count, _ ).
noun( 'foster-mother', 'foster-mothers', count, _ ).
noun( 'foster-parent', 'foster-parents', count, _ ).
noun( 'foster-sister', 'foster-sisters', count, _ ).
noun( 'foul', 'fouls', both, _ ).
noun( 'foulness', '-', mass, _ ).
noun( 'foundation', 'foundations', both, _ ).
noun( 'foundation-stone', 'foundation-stones', count, _ ).
noun( 'founder', 'founders', count, _ ).
noun( 'foundling', 'foundlings', count, _ ).
noun( 'foundress', 'foundresses', count, _ ).
noun( 'foundry', 'foundries', count, _ ).
noun( 'fount', 'founts', count, _ ).
noun( 'fountain', 'fountains', count, _ ).
noun( 'fountain-head', 'fountain-heads', count, _ ).
noun( 'fountain-pen', 'fountain-pens', count, _ ).
noun( 'four', 'fours', count, _ ).
noun( 'four-in-hand', 'four-in-hands', count, _ ).
noun( 'four-poster', 'four-posters', count, _ ).
noun( 'four-pounder', 'four-pounders', count, _ ).
noun( 'four-wheeler', 'four-wheelers', count, _ ).
noun( 'fourpence', 'fourpences', count, _ ).
noun( 'fourscore', '-', count, _ ).
noun( 'foursome', 'foursomes', count, _ ).
noun( 'fourteen', 'fourteens', count, _ ).
noun( 'fourteenth', 'fourteenths', count, _ ).
noun( 'fourth', 'fourths', count, _ ).
noun( 'fowl', 'fowls', both, _ ).
noun( 'fowl-run', 'fowl-runs', count, _ ).
noun( 'fowler', 'fowlers', count, _ ).
noun( 'fowlingpiece', 'fowlingpieces', count, _ ).
noun( 'fowlpest', '-', mass, _ ).
noun( 'fox', 'foxes', count, _ ).
noun( 'fox-terrier', 'fox-terriers', count, _ ).
noun( 'foxglove', 'foxgloves', count, _ ).
noun( 'foxhole', 'foxholes', count, _ ).
noun( 'foxhound', 'foxhounds', count, _ ).
noun( 'foxhunt', 'foxhunts', count, _ ).
noun( 'foxhunter', 'foxhunters', count, _ ).
noun( 'foxtrot', 'foxtrots', count, _ ).
noun( 'foyer', 'foyers', count, _ ).
noun( 'fracas', '-', count, _ ).
noun( 'fraction', 'fractions', count, _ ).
noun( 'fractiousness', '-', mass, _ ).
noun( 'fracture', 'fractures', both, _ ).
noun( 'fragility', '-', mass, _ ).
noun( 'fragment', 'fragments', count, _ ).
noun( 'fragmentation', 'fragmentations', both, _ ).
noun( 'fragrance', 'fragrances', both, _ ).
noun( 'frailty', 'frailties', both, _ ).
noun( 'frame', 'frames', count, _ ).
noun( 'frame-up', 'frame-ups', count, _ ).
noun( 'framework', 'frameworks', count, _ ).
noun( 'franc', 'francs', count, _ ).
noun( 'franchise', 'franchises', both, _ ).
noun( 'frankfurter', 'frankfurters', count, _ ).
noun( 'frankincense', '-', mass, _ ).
noun( 'franking-machine', 'franking-machines', count, _ ).
noun( 'franklin', 'franklins', count, _ ).
noun( 'frankness', '-', mass, _ ).
noun( 'fraternity', 'fraternities', both, _ ).
noun( 'fraternization', '-', mass, _ ).
noun( 'fratricide', 'fratricides', both, _ ).
noun( 'fraud', 'frauds', both, _ ).
noun( 'fray', 'frays', count, _ ).
noun( 'frazzle', 'frazzles', count, _ ).
noun( 'freak', 'freaks', count, _ ).
noun( 'freak-out', 'freak-outs', count, _ ).
noun( 'freakishness', '-', mass, _ ).
noun( 'freckle', 'freckles', count, _ ).
noun( 'free-for-all', 'free-for-alls', count, _ ).
noun( 'free-list', 'free-lists', count, _ ).
noun( 'free-liver', 'free-livers', count, _ ).
noun( 'free-living', '-', mass, _ ).
noun( 'free-thinker', 'free-thinkers', count, _ ).
noun( 'free-thought', '-', mass, _ ).
noun( 'free-trader', 'free-traders', count, _ ).
noun( 'freebooter', 'freebooters', count, _ ).
noun( 'freedman', 'freedmen', count, _ ).
noun( 'freedom', 'freedoms', both, _ ).
noun( 'freehold', 'freeholds', count, _ ).
noun( 'freeholder', 'freeholders', count, _ ).
noun( 'freelance', 'freelances', count, _ ).
noun( 'freeman', 'freemen', count, _ ).
noun( 'freesia', 'freesias', count, _ ).
noun( 'freestone', '-', mass, _ ).
noun( 'freestyle', '-', mass, _ ).
noun( 'freeway', 'freeways', count, _ ).
noun( 'freeze', 'freezes', count, _ ).
noun( 'freezer', 'freezers', count, _ ).
noun( 'freezing-mixture', 'freezing-mixtures', count, _ ).
noun( 'freezing-point', 'freezing-points', count, _ ).
noun( 'freight', '-', mass, _ ).
noun( 'freight-train', 'freight-trains', count, _ ).
noun( 'freighter', 'freighters', count, _ ).
noun( 'freightliner', 'freightliners', count, _ ).
noun( 'frenzy', '-', mass, _ ).
noun( 'frequency', 'frequencies', both, _ ).
noun( 'fresco', 'frescos', both, _ ).
noun( 'fresher', 'freshers', count, _ ).
noun( 'freshman', 'freshmen', count, _ ).
noun( 'freshness', '-', mass, _ ).
noun( 'fret', 'frets', count, _ ).
noun( 'fretsaw', 'fretsaws', count, _ ).
noun( 'fretwork', '-', mass, _ ).
noun( 'friability', '-', mass, _ ).
noun( 'friar', 'friars', count, _ ).
noun( 'fricassee', 'fricassees', both, _ ).
noun( 'fricative', 'fricatives', count, _ ).
noun( 'friction', 'frictions', both, _ ).
noun( 'fridge', 'fridges', count, _ ).
noun( 'friend', 'friends', count, _ ).
noun( 'friendlessness', '-', mass, _ ).
noun( 'friendliness', '-', mass, _ ).
noun( 'friendship', 'friendships', both, _ ).
noun( 'frier', 'friers', count, _ ).
noun( 'frieze', 'friezes', count, _ ).
noun( 'frigate', 'frigates', count, _ ).
noun( 'fright', 'frights', both, _ ).
noun( 'frightfulness', '-', mass, _ ).
noun( 'frigidity', '-', mass, _ ).
noun( 'frill', 'frills', count, _ ).
noun( 'fringe', 'fringes', count, _ ).
noun( 'frippery', 'fripperies', both, _ ).
noun( 'frisson', 'frissons', count, _ ).
noun( 'fritter', 'fritters', count, _ ).
noun( 'frivolity', 'frivolities', both, _ ).
noun( 'frock', 'frocks', count, _ ).
noun( 'frock-coat', 'frock-coats', count, _ ).
noun( 'frog', 'frogs', count, _ ).
noun( 'frogman', 'frogmen', count, _ ).
noun( 'frolic', 'frolics', count, _ ).
noun( 'frond', 'fronds', count, _ ).
noun( 'front', 'fronts', both, _ ).
noun( 'front-bench', 'front-benches', count, _ ).
noun( 'front-bencher', 'front-benchers', count, _ ).
noun( 'frontage', 'frontages', count, _ ).
noun( 'frontier', 'frontiers', count, _ ).
noun( 'frontiersman', 'frontiersmen', count, _ ).
noun( 'frontispiece', 'frontispieces', count, _ ).
noun( 'frost', 'frosts', both, _ ).
noun( 'frostbite', '-', mass, _ ).
noun( 'frostiness', '-', mass, _ ).
noun( 'frosting', '-', mass, _ ).
noun( 'froth', '-', mass, _ ).
noun( 'frothiness', '-', mass, _ ).
noun( 'frown', 'frowns', count, _ ).
noun( 'fructification', '-', mass, _ ).
noun( 'frugality', 'frugalities', both, _ ).
noun( 'fruit', 'fruits', both, _ ).
noun( 'fruit-fly', 'fruit-flies', count, _ ).
noun( 'fruitcake', 'fruitcakes', count, _ ).
noun( 'fruiterer', 'fruiterers', count, _ ).
noun( 'fruitfulness', '-', mass, _ ).
noun( 'fruition', '-', mass, _ ).
noun( 'fruitlessness', '-', mass, _ ).
noun( 'frump', 'frumps', count, _ ).
noun( 'frustration', 'frustrations', both, _ ).
noun( 'fry', 'fry', count, _ ).
noun( 'fry-pan', 'fry-pans', count, _ ).
noun( 'fryer', 'fryers', count, _ ).
noun( 'frying-pan', 'frying-pans', count, _ ).
noun( 'ft', 'ft', count, _ ).
noun( 'fuchsia', 'fuchsias', count, _ ).
noun( 'fuck-all', '-', mass, _ ).
noun( 'fucker', 'fuckers', count, _ ).
noun( 'fuddy-duddy', 'fuddy-duddies', count, _ ).
noun( 'fudge', 'fudges', both, _ ).
noun( 'fuel', 'fuels', both, _ ).
noun( 'fug', 'fugs', count, _ ).
noun( 'fugitive', 'fugitives', count, _ ).
noun( 'fugue', 'fugues', count, _ ).
noun( 'fulcrum', 'fulcrums', count, _ ).
noun( 'fulfilment', 'fulfilments', count, _ ).
noun( 'fullback', 'fullbacks', count, _ ).
noun( 'fuller', 'fullers', count, _ ).
noun( 'fullness', '-', mass, _ ).
noun( 'fulmar', 'fulmars', count, _ ).
noun( 'fulmination', 'fulminations', both, _ ).
noun( 'fulsomeness', '-', mass, _ ).
noun( 'fumbler', 'fumblers', count, _ ).
noun( 'fume', 'fumes', count, _ ).
noun( 'fumigation', 'fumigations', count, _ ).
noun( 'fun', '-', mass, _ ).
noun( 'function', 'functions', count, _ ).
noun( 'functionalism', '-', mass, _ ).
noun( 'functionalist', 'functionalists', count, _ ).
noun( 'functionary', 'functionaries', count, _ ).
noun( 'fund', 'funds', count, _ ).
noun( 'fundamental', 'fundamentals', count, _ ).
noun( 'fundamentalism', '-', mass, _ ).
noun( 'fundamentalist', 'fundamentalists', count, _ ).
noun( 'funeral', 'funerals', count, _ ).
noun( 'funfair', 'funfairs', count, _ ).
noun( 'fungicide', 'fungicides', both, _ ).
noun( 'fungus', 'funguses', both, _ ).
noun( 'funicular', 'funiculars', count, _ ).
noun( 'funk', 'funks', count, _ ).
noun( 'funnel', 'funnels', count, _ ).
noun( 'funniness', '-', mass, _ ).
noun( 'funny-bone', 'funny-bones', count, _ ).
noun( 'fur', 'furs', both, _ ).
noun( 'furbelow', 'furbelows', count, _ ).
noun( 'furlong', 'furlongs', count, _ ).
noun( 'furlough', 'furloughs', both, _ ).
noun( 'furnace', 'furnaces', count, _ ).
noun( 'furniture', '-', mass, _ ).
noun( 'furore', 'furores', count, _ ).
noun( 'furrier', 'furriers', count, _ ).
noun( 'furrow', 'furrows', count, _ ).
noun( 'furtherance', '-', mass, _ ).
noun( 'furtiveness', '-', mass, _ ).
noun( 'fury', 'furies', both, _ ).
noun( 'furze', '-', mass, _ ).
noun( 'fuse', 'fuses', count, _ ).
noun( 'fuselage', 'fuselages', count, _ ).
noun( 'fusilier', 'fusiliers', count, _ ).
noun( 'fusillade', 'fusillades', count, _ ).
noun( 'fusion', 'fusions', both, _ ).
noun( 'fuss', 'fusses', both, _ ).
noun( 'fussiness', '-', mass, _ ).
noun( 'fusspot', 'fusspots', count, _ ).
noun( 'fustian', '-', mass, _ ).
noun( 'futility', 'futilities', both, _ ).
noun( 'future', 'futures', both, _ ).
noun( 'futurism', '-', mass, _ ).
noun( 'futurist', 'futurists', count, _ ).
noun( 'futurity', 'futurities', both, _ ).
noun( 'fuze', 'fuzes', count, _ ).
noun( 'fuzz', '-', mass, _ ).
noun( 'fwd', '-', proper, _ ).
noun( 'g', '-', count, _ ).
noun( 'g^ateau', 'g^ateaus', count, _ ).
noun( 'gab', '-', mass, _ ).
noun( 'gabardine', '-', mass, _ ).
noun( 'gabble', '-', mass, _ ).
noun( 'gaberdine', '-', mass, _ ).
noun( 'gable', 'gables', count, _ ).
noun( 'gadabout', 'gadabouts', count, _ ).
noun( 'gadfly', 'gadflies', count, _ ).
noun( 'gadget', 'gadgets', count, _ ).
noun( 'gadgetry', '-', mass, _ ).
noun( 'gaff', 'gaffs', count, _ ).
noun( 'gaffe', 'gaffes', count, _ ).
noun( 'gaffer', 'gaffers', count, _ ).
noun( 'gag', 'gags', count, _ ).
noun( 'gage', 'gages', count, _ ).
noun( 'gaggle', 'gaggles', count, _ ).
noun( 'gaiety', '-', mass, _ ).
noun( 'gain', 'gains', both, _ ).
noun( 'gait', 'gaits', count, _ ).
noun( 'gaiter', 'gaiters', count, _ ).
noun( 'gal', 'gals', count, _ ).
noun( 'gala', 'galas', count, _ ).
noun( 'galantine', 'galantines', count, _ ).
noun( 'galaxy', 'galaxies', count, _ ).
noun( 'gale', 'gales', count, _ ).
noun( 'gall', 'galls', both, _ ).
noun( 'gallant', 'gallants', count, _ ).
noun( 'gallantry', 'gallantries', both, _ ).
noun( 'galleon', 'galleons', count, _ ).
noun( 'gallery', 'galleries', count, _ ).
noun( 'galley', 'galleys', count, _ ).
noun( 'galley-proof', 'galley-proofs', count, _ ).
noun( 'galley-slave', 'galley-slaves', count, _ ).
noun( 'gallicism', 'gallicisms', count, _ ).
noun( 'gallon', 'gallons', count, _ ).
noun( 'gallop', 'gallops', count, _ ).
noun( 'gallows-bird', 'gallows-birds', count, _ ).
noun( 'gallstone', 'gallstones', count, _ ).
noun( 'galosh', 'galoshes', count, _ ).
noun( 'galvanism', '-', mass, _ ).
noun( 'gambit', 'gambits', count, _ ).
noun( 'gamble', 'gambles', count, _ ).
noun( 'gambler', 'gamblers', count, _ ).
noun( 'gambling', '-', mass, _ ).
noun( 'gambling-den', 'gambling-dens', count, _ ).
noun( 'gamboge', '-', mass, _ ).
noun( 'gambol', 'gambols', count, _ ).
noun( 'game', 'games', both, _ ).
noun( 'game-bag', 'game-bags', count, _ ).
noun( 'game-bird', 'game-birds', count, _ ).
noun( 'game-licence', 'game-licences', count, _ ).
noun( 'gamecock', 'gamecocks', count, _ ).
noun( 'gamekeeper', 'gamekeepers', count, _ ).
noun( 'games-master', 'games-masters', count, _ ).
noun( 'games-mistress', 'games-mistresses', count, _ ).
noun( 'gamesmanship', '-', mass, _ ).
noun( 'gaming-house', 'gaming-houses', count, _ ).
noun( 'gaming-table', 'gaming-tables', count, _ ).
noun( 'gamma', 'gammas', count, _ ).
noun( 'gammon', 'gammons', both, _ ).
noun( 'gamp', 'gamps', count, _ ).
noun( 'gamut', 'gamuts', count, _ ).
noun( 'gander', 'ganders', count, _ ).
noun( 'gang', 'gangs', count, _ ).
noun( 'ganger', 'gangers', count, _ ).
noun( 'ganglion', 'ganglions', count, _ ).
noun( 'gangplank', 'gangplanks', count, _ ).
noun( 'gangrene', '-', mass, _ ).
noun( 'gangster', 'gangsters', count, _ ).
noun( 'gangway', 'gangways', count, _ ).
noun( 'gannet', 'gannets', count, _ ).
noun( 'gantry', 'gantries', count, _ ).
noun( 'gaol', 'gaols', both, _ ).
noun( 'gaolbird', 'gaolbirds', count, _ ).
noun( 'gaolbreak', 'gaolbreaks', count, _ ).
noun( 'gaoler', 'gaolers', count, _ ).
noun( 'gap', 'gaps', count, _ ).
noun( 'gape', 'gapes', count, _ ).
noun( 'garage', 'garages', count, _ ).
noun( 'garb', '-', mass, _ ).
noun( 'garbage', '-', mass, _ ).
noun( 'garbage-can', 'garbage-cans', count, _ ).
noun( 'garden', 'gardens', both, _ ).
noun( 'garden-truck', 'garden-trucks', count, _ ).
noun( 'gardener', 'gardeners', count, _ ).
noun( 'gardenia', 'gardenias', count, _ ).
noun( 'gardening', '-', mass, _ ).
noun( 'gargle', 'gargles', count, _ ).
noun( 'gargoyle', 'gargoyles', count, _ ).
noun( 'garland', 'garlands', count, _ ).
noun( 'garlic', '-', mass, _ ).
noun( 'garment', 'garments', count, _ ).
noun( 'garner', 'garners', count, _ ).
noun( 'garnet', 'garnets', count, _ ).
noun( 'garnish', 'garnishes', count, _ ).
noun( 'garotte', 'garottes', count, _ ).
noun( 'garret', 'garrets', count, _ ).
noun( 'garrison', 'garrisons', count, _ ).
noun( 'garrotte', 'garrottes', count, _ ).
noun( 'garrulity', '-', mass, _ ).
noun( 'garter', 'garters', count, _ ).
noun( 'gas', 'gases', both, _ ).
noun( 'gas-bracket', 'gas-brackets', count, _ ).
noun( 'gas-cooker', 'gas-cookers', count, _ ).
noun( 'gas-engine', 'gas-engines', count, _ ).
noun( 'gas-fitter', 'gas-fitters', count, _ ).
noun( 'gas-helmet', 'gas-helmets', count, _ ).
noun( 'gas-holder', 'gas-holders', count, _ ).
noun( 'gas-mask', 'gas-masks', count, _ ).
noun( 'gas-meter', 'gas-meters', count, _ ).
noun( 'gas-oven', 'gas-ovens', count, _ ).
noun( 'gas-ring', 'gas-rings', count, _ ).
noun( 'gas-station', 'gas-stations', count, _ ).
noun( 'gas-stove', 'gas-stoves', count, _ ).
noun( 'gasbag', 'gasbags', count, _ ).
noun( 'gash', 'gashes', count, _ ).
noun( 'gasification', 'gasifications', both, _ ).
noun( 'gasket', 'gaskets', count, _ ).
noun( 'gaslight', '-', mass, _ ).
noun( 'gasmask', 'gasmasks', count, _ ).
noun( 'gasolene', '-', mass, _ ).
noun( 'gasoline', '-', mass, _ ).
noun( 'gasometer', 'gasometers', count, _ ).
noun( 'gasp', 'gasps', count, _ ).
noun( 'gastritis', '-', mass, _ ).
noun( 'gastronomy', '-', mass, _ ).
noun( 'gasworks', 'gasworks', count, _ ).
noun( 'gate', 'gates', count, _ ).
noun( 'gatecrasher', 'gatecrashers', count, _ ).
noun( 'gatehouse', 'gatehouses', count, _ ).
noun( 'gatepost', 'gateposts', count, _ ).
noun( 'gateway', 'gateways', count, _ ).
noun( 'gatherer', 'gatherers', count, _ ).
noun( 'gathering', 'gatherings', count, _ ).
noun( 'gaucherie', 'gaucheries', both, _ ).
noun( 'gaucho', 'gauchos', count, _ ).
noun( 'gaud', 'gauds', count, _ ).
noun( 'gaudy', 'gaudies', count, _ ).
noun( 'gauge', 'gauges', both, _ ).
noun( 'gauntlet', 'gauntlets', count, _ ).
noun( 'gauntness', '-', mass, _ ).
noun( 'gauze', '-', mass, _ ).
noun( 'gavel', 'gavels', count, _ ).
noun( 'gavotte', 'gavottes', count, _ ).
noun( 'gawk', 'gawks', count, _ ).
noun( 'gawkiness', '-', mass, _ ).
noun( 'gay', 'gays', count, _ ).
noun( 'gayness', '-', mass, _ ).
noun( 'gaze', '-', count, _ ).
noun( 'gazelle', 'gazelles', count, _ ).
noun( 'gazette', 'gazettes', count, _ ).
noun( 'gazetteer', 'gazetteers', count, _ ).
noun( 'gear', 'gears', both, _ ).
noun( 'gear-case', 'gear-cases', count, _ ).
noun( 'gearbox', 'gearboxes', count, _ ).
noun( 'gearshift', 'gearshifts', count, _ ).
noun( 'gecko', 'geckos', count, _ ).
noun( 'gee-gee', 'gee-gees', count, _ ).
noun( 'geezer', 'geezers', count, _ ).
noun( 'geisha', 'geishas', count, _ ).
noun( 'gel', 'gels', count, _ ).
noun( 'gelatin', '-', mass, _ ).
noun( 'gelatine', '-', mass, _ ).
noun( 'gelding', 'geldings', count, _ ).
noun( 'gelignite', '-', mass, _ ).
noun( 'gem', 'gems', count, _ ).
noun( 'gen', '-', mass, _ ).
noun( 'gendarme', 'gendarmes', count, _ ).
noun( 'gendarmerie', 'gendarmeries', count, _ ).
noun( 'gender', 'genders', count, _ ).
noun( 'gene', 'genes', count, _ ).
noun( 'genealogist', 'genealogists', count, _ ).
noun( 'genealogy', 'genealogies', both, _ ).
noun( 'general', 'generals', count, _ ).
noun( 'generalissimo', 'generalissimos', count, _ ).
noun( 'generality', 'generalities', both, _ ).
noun( 'generalization', 'generalizations', both, _ ).
noun( 'generation', 'generations', both, _ ).
noun( 'generator', 'generators', count, _ ).
noun( 'generosity', 'generosities', both, _ ).
noun( 'genesis', 'geneses', count, _ ).
noun( 'geneticist', 'geneticists', count, _ ).
noun( 'genetics', 'genetics', mass, _ ).
noun( 'geniality', 'genialities', both, _ ).
noun( 'genie', 'genies', count, _ ).
noun( 'genius', 'geniuses', both, _ ).
noun( 'genius loci', '-', count, _ ).
noun( 'genocide', '-', mass, _ ).
noun( 'genre', 'genres', count, _ ).
noun( 'genre-painting', 'genre-paintings', count, _ ).
noun( 'gent', 'gents', count, _ ).
noun( 'gentian', 'gentians', count, _ ).
noun( 'gentile', 'gentiles', count, _ ).
noun( 'gentility', '-', mass, _ ).
noun( 'gentleman', 'gentlemen', count, _ ).
noun( 'gentleman-at-arms', 'gentlemen-at-arms', count, _ ).
noun( 'gentleness', '-', mass, _ ).
noun( 'gentlewoman', 'gentlewomen', count, _ ).
noun( 'gentry', 'gentry', count, _ ).
noun( 'genuflection', 'genuflections', count, _ ).
noun( 'genuflexion', 'genuflexions', count, _ ).
noun( 'genuineness', '-', mass, _ ).
noun( 'genus', 'geni', count, _ ).
noun( 'geographer', 'geographers', count, _ ).
noun( 'geography', '-', mass, _ ).
noun( 'geologist', 'geologists', count, _ ).
noun( 'geology', '-', mass, _ ).
noun( 'geometry', '-', mass, _ ).
noun( 'geophysics', 'geophysics', mass, _ ).
noun( 'geopolitics', 'geopolitics', mass, _ ).
noun( 'georgette', '-', mass, _ ).
noun( 'geranium', 'geraniums', count, _ ).
noun( 'geriatrician', 'geriatricians', count, _ ).
noun( 'geriatrics', 'geriatrics', mass, _ ).
noun( 'germ', 'germs', count, _ ).
noun( 'germicide', 'germicides', count, _ ).
noun( 'germination', '-', mass, _ ).
noun( 'gerontology', '-', mass, _ ).
noun( 'gerrymander', 'gerrymanders', count, _ ).
noun( 'gerund', 'gerunds', count, _ ).
noun( 'gestation', 'gestations', count, _ ).
noun( 'gesticulation', 'gesticulations', both, _ ).
noun( 'gesture', 'gestures', both, _ ).
noun( 'get-together', 'get-togethers', count, _ ).
noun( 'get-up', 'get-ups', count, _ ).
noun( 'getaway', 'getaways', count, _ ).
noun( 'geum', 'geums', count, _ ).
noun( 'geyser', 'geysers', count, _ ).
noun( 'gharry', 'gharries', count, _ ).
noun( 'ghat', 'ghats', count, _ ).
noun( 'ghee', '-', mass, _ ).
noun( 'gherkin', 'gherkins', count, _ ).
noun( 'ghetto', 'ghettos', count, _ ).
noun( 'ghost', 'ghosts', count, _ ).
noun( 'ghost-writer', 'ghost-writers', count, _ ).
noun( 'ghostliness', '-', mass, _ ).
noun( 'ghoul', 'ghouls', count, _ ).
noun( 'giant', 'giants', count, _ ).
noun( 'giantess', 'giantesses', count, _ ).
noun( 'gibberish', '-', mass, _ ).
noun( 'gibbet', 'gibbets', count, _ ).
noun( 'gibbon', 'gibbons', count, _ ).
noun( 'gibe', 'gibes', count, _ ).
noun( 'giddiness', '-', mass, _ ).
noun( 'gift', 'gifts', both, _ ).
noun( 'gig', 'gigs', count, _ ).
noun( 'giggle', 'giggles', count, _ ).
noun( 'gigolo', 'gigolos', count, _ ).
noun( 'gild', 'gilds', count, _ ).
noun( 'gilder', 'gilders', count, _ ).
noun( 'gilding', '-', mass, _ ).
noun( 'gill', 'gills', count, _ ).
noun( 'gill', 'gills', count, _ ).
noun( 'gillie', 'gillies', count, _ ).
noun( 'gilt', '-', mass, _ ).
noun( 'gimlet', 'gimlets', count, _ ).
noun( 'gimmick', 'gimmicks', count, _ ).
noun( 'gin', 'gins', both, _ ).
noun( 'ginger', '-', mass, _ ).
noun( 'gingerbread', 'gingerbreads', both, _ ).
noun( 'gingham', '-', mass, _ ).
noun( 'gingko', 'gingkos', count, _ ).
noun( 'ginseng', '-', mass, _ ).
noun( 'gipsy', 'gipsies', count, _ ).
noun( 'giraffe', 'giraffes', count, _ ).
noun( 'girder', 'girders', count, _ ).
noun( 'girdle', 'girdles', count, _ ).
noun( 'girl', 'girls', count, _ ).
noun( 'girlfriend', 'girlfriends', count, _ ).
noun( 'girlhood', '-', mass, _ ).
noun( 'girlishness', '-', mass, _ ).
noun( 'giro', '-', mass, _ ).
noun( 'girth', 'girths', count, _ ).
noun( 'gist', 'gists', count, _ ).
noun( 'give', '-', mass, _ ).
noun( 'giveaway', 'giveaways', count, _ ).
noun( 'giver', 'givers', count, _ ).
noun( 'gizzard', 'gizzards', count, _ ).
noun( 'glacier', 'glaciers', count, _ ).
noun( 'glade', 'glades', count, _ ).
noun( 'gladiator', 'gladiators', count, _ ).
noun( 'gladiolus', 'gladioluses', count, _ ).
noun( 'gladness', '-', mass, _ ).
noun( 'glamorization', '-', mass, _ ).
noun( 'glamour', '-', mass, _ ).
noun( 'glance', 'glances', count, _ ).
noun( 'gland', 'glands', count, _ ).
noun( 'glanders', '-', mass, _ ).
noun( 'glare', 'glares', both, _ ).
noun( 'glass', 'glasses', both, _ ).
noun( 'glass-blower', 'glass-blowers', count, _ ).
noun( 'glass-cutter', 'glass-cutters', count, _ ).
noun( 'glass-wool', '-', mass, _ ).
noun( 'glassful', 'glassfuls', count, _ ).
noun( 'glasshouse', 'glasshouses', count, _ ).
noun( 'glassware', '-', mass, _ ).
noun( 'glassworks', 'glassworks', count, _ ).
noun( 'glaucoma', '-', mass, _ ).
noun( 'glaze', 'glazes', both, _ ).
noun( 'glazier', 'glaziers', count, _ ).
noun( 'gleam', 'gleams', count, _ ).
noun( 'gleaner', 'gleaners', count, _ ).
noun( 'glebe', 'glebes', count, _ ).
noun( 'glee', 'glees', both, _ ).
noun( 'glen', 'glens', count, _ ).
noun( 'glibness', '-', mass, _ ).
noun( 'glide', 'glides', count, _ ).
noun( 'glider', 'gliders', count, _ ).
noun( 'gliding', '-', mass, _ ).
noun( 'glimmer', 'glimmers', count, _ ).
noun( 'glimpse', 'glimpses', count, _ ).
noun( 'glint', 'glints', count, _ ).
noun( 'glissade', 'glissades', count, _ ).
noun( 'glister', '-', count, _ ).
noun( 'glitter', '-', mass, _ ).
noun( 'gloaming', '-', count, _ ).
noun( 'globe', 'globes', count, _ ).
noun( 'globetrotter', 'globetrotters', count, _ ).
noun( 'globule', 'globules', count, _ ).
noun( 'glockenspiel', 'glockenspiels', count, _ ).
noun( 'gloom', 'glooms', both, _ ).
noun( 'glorification', '-', mass, _ ).
noun( 'glory', 'glories', both, _ ).
noun( 'glory-hole', 'glory-holes', count, _ ).
noun( 'gloss', 'glosses', count, _ ).
noun( 'glossary', 'glossaries', count, _ ).
noun( 'glossiness', '-', mass, _ ).
noun( 'glottis', 'glottises', count, _ ).
noun( 'glove', 'gloves', count, _ ).
noun( 'glove-compartment', 'glove-compartments', count, _ ).
noun( 'glow', '-', count, _ ).
noun( 'glow-worm', 'glow-worms', count, _ ).
noun( 'glucose', '-', mass, _ ).
noun( 'glue', 'glues', both, _ ).
noun( 'glumness', '-', mass, _ ).
noun( 'glut', 'gluts', count, _ ).
noun( 'gluten', '-', mass, _ ).
noun( 'glutton', 'gluttons', count, _ ).
noun( 'gluttony', '-', mass, _ ).
noun( 'glycerine', '-', mass, _ ).
noun( 'gnat', 'gnats', count, _ ).
noun( 'gnome', 'gnomes', count, _ ).
noun( 'gnu', 'gnus', count, _ ).
noun( 'go', 'goes', count, _ ).
noun( 'go-ahead', '-', count, _ ).
noun( 'go-between', 'go-betweens', count, _ ).
noun( 'go-by', '-', count, _ ).
noun( 'go-cart', 'go-carts', count, _ ).
noun( 'go-getter', 'go-getters', count, _ ).
noun( 'go-kart', 'go-karts', count, _ ).
noun( 'go-slow', 'go-slows', count, _ ).
noun( 'goad', 'goads', count, _ ).
noun( 'goal', 'goals', count, _ ).
noun( 'goal-kick', 'goal-kicks', count, _ ).
noun( 'goal-line', 'goal-lines', count, _ ).
noun( 'goalie', 'goalies', count, _ ).
noun( 'goalkeeper', 'goalkeepers', count, _ ).
noun( 'goalmouth', 'goalmouths', count, _ ).
noun( 'goat', 'goats', count, _ ).
noun( 'goatee', 'goatees', count, _ ).
noun( 'goatherd', 'goatherds', count, _ ).
noun( 'goatskin', 'goatskins', both, _ ).
noun( 'gob', 'gobs', count, _ ).
noun( 'gobbet', 'gobbets', count, _ ).
noun( 'gobble', 'gobbles', count, _ ).
noun( 'gobbledygook', '-', mass, _ ).
noun( 'gobbler', 'gobblers', count, _ ).
noun( 'goblet', 'goblets', count, _ ).
noun( 'goblin', 'goblins', count, _ ).
noun( 'god', 'gods', count, _ ).
noun( 'godchild', 'godchildren', count, _ ).
noun( 'goddaughter', 'goddaughters', count, _ ).
noun( 'goddess', 'goddesses', count, _ ).
noun( 'godfather', 'godfathers', count, _ ).
noun( 'godhead', '-', mass, _ ).
noun( 'godlessness', '-', mass, _ ).
noun( 'godliness', '-', mass, _ ).
noun( 'godmother', 'godmothers', count, _ ).
noun( 'godown', 'godowns', count, _ ).
noun( 'godparent', 'godparents', count, _ ).
noun( 'godsend', 'godsends', count, _ ).
noun( 'godson', 'godsons', count, _ ).
noun( 'godspeed', '-', mass, _ ).
noun( 'goggle-box', 'goggle-boxes', count, _ ).
noun( 'going', 'goings', both, _ ).
noun( 'going-over', 'going-overs', count, _ ).
noun( 'goitre', 'goitres', count, _ ).
noun( 'gold', '-', mass, _ ).
noun( 'gold-beater', 'gold-beaters', count, _ ).
noun( 'gold-digger', 'gold-diggers', count, _ ).
noun( 'gold-dust', '-', mass, _ ).
noun( 'gold-foil', '-', mass, _ ).
noun( 'gold-leaf', '-', mass, _ ).
noun( 'gold-plate', '-', mass, _ ).
noun( 'gold-rush', 'gold-rushes', count, _ ).
noun( 'goldfield', 'goldfields', count, _ ).
noun( 'goldfinch', 'goldfinches', count, _ ).
noun( 'goldfish', 'goldfish', count, _ ).
noun( 'goldmine', 'goldmines', count, _ ).
noun( 'goldsmith', 'goldsmiths', count, _ ).
noun( 'golf', '-', mass, _ ).
noun( 'golf-ball', 'golf-balls', count, _ ).
noun( 'golf-club', 'golf-clubs', count, _ ).
noun( 'golf-course', 'golf-courses', count, _ ).
noun( 'golf-links', 'golf-links', count, _ ).
noun( 'golfer', 'golfers', count, _ ).
noun( 'golliwog', 'golliwogs', count, _ ).
noun( 'golosh', 'goloshes', count, _ ).
noun( 'gondola', 'gondolas', count, _ ).
noun( 'gondolier', 'gondoliers', count, _ ).
noun( 'goner', 'goners', count, _ ).
noun( 'gong', 'gongs', count, _ ).
noun( 'gonorrhea', '-', mass, _ ).
noun( 'gonorrhoea', '-', mass, _ ).
noun( 'goo', '-', mass, _ ).
noun( 'good', '-', mass, _ ).
noun( 'good-fellowship', '-', mass, _ ).
noun( 'good-for-naught', 'good-for-naughts', count, _ ).
noun( 'good-for-nothing', 'good-for-nothings', count, _ ).
noun( 'good-neighbourliness', '-', mass, _ ).
noun( 'goodbye', 'goodbyes', count, _ ).
noun( 'goodness', '-', mass, _ ).
noun( 'goodwill', '-', mass, _ ).
noun( 'goody', 'goodies', count, _ ).
noun( 'goody-goody', 'goody-goodies', count, _ ).
noun( 'goof', 'goofs', count, _ ).
noun( 'googly', 'googlies', count, _ ).
noun( 'goon', 'goons', count, _ ).
noun( 'goose', 'geese', count, _ ).
noun( 'goose-flesh', '-', mass, _ ).
noun( 'goose-step', 'goose-steps', count, _ ).
noun( 'gooseberry', 'gooseberries', count, _ ).
noun( 'gopher', 'gophers', count, _ ).
noun( 'gore', '-', mass, _ ).
noun( 'gorge', 'gorges', count, _ ).
noun( 'gorilla', 'gorillas', count, _ ).
noun( 'gorse', '-', mass, _ ).
noun( 'gosling', 'goslings', count, _ ).
noun( 'gospel', 'gospels', both, _ ).
noun( 'gossamer', 'gossamers', both, _ ).
noun( 'gossip', 'gossips', both, _ ).
noun( 'gouache', '-', mass, _ ).
noun( 'gouge', 'gouges', count, _ ).
noun( 'goulash', 'goulashes', both, _ ).
noun( 'gourd', 'gourds', count, _ ).
noun( 'gourmand', 'gourmands', count, _ ).
noun( 'gourmet', 'gourmets', count, _ ).
noun( 'gout', '-', mass, _ ).
noun( 'governance', '-', mass, _ ).
noun( 'governess', 'governesses', count, _ ).
noun( 'government', 'governments', both, _ ).
noun( 'governor', 'governors', count, _ ).
noun( 'govt', '-', count, _ ).
noun( 'gown', 'gowns', count, _ ).
noun( 'grab', 'grabs', count, _ ).
noun( 'grabber', 'grabbers', count, _ ).
noun( 'grace', 'graces', both, _ ).
noun( 'graciousness', '-', mass, _ ).
noun( 'gradation', 'gradations', both, _ ).
noun( 'grade', 'grades', count, _ ).
noun( 'gradient', 'gradients', count, _ ).
noun( 'gradualness', '-', mass, _ ).
noun( 'graduate', 'graduates', count, _ ).
noun( 'graduation', 'graduations', count, _ ).
noun( 'graffito', 'graffiti', count, _ ).
noun( 'graft', 'grafts', both, _ ).
noun( 'grail', 'grails', count, _ ).
noun( 'grain', 'grains', both, _ ).
noun( 'gram', 'grams', count, _ ).
noun( 'grammar', 'grammars', both, _ ).
noun( 'grammarian', 'grammarians', count, _ ).
noun( 'gramme', 'grammes', count, _ ).
noun( 'gramophone', 'gramophones', count, _ ).
noun( 'grampus', 'grampuses', count, _ ).
noun( 'granary', 'granaries', count, _ ).
noun( 'grandad', 'grandads', count, _ ).
noun( 'grandaunt', 'grandaunts', count, _ ).
noun( 'grandchild', 'grandchildren', count, _ ).
noun( 'granddad', 'granddads', count, _ ).
noun( 'granddaughter', 'granddaughters', count, _ ).
noun( 'grandee', 'grandees', count, _ ).
noun( 'grandeur', '-', mass, _ ).
noun( 'grandfather', 'grandfathers', count, _ ).
noun( 'grandiloquence', '-', mass, _ ).
noun( 'grandma', 'grandmas', count, _ ).
noun( 'grandmother', 'grandmothers', count, _ ).
noun( 'grandnephew', 'grandnephews', count, _ ).
noun( 'grandniece', 'grandnieces', count, _ ).
noun( 'grandpa', 'grandpas', count, _ ).
noun( 'grandparent', 'grandparents', count, _ ).
noun( 'grandson', 'grandsons', count, _ ).
noun( 'grandstand', 'grandstands', count, _ ).
noun( 'granduncle', 'granduncles', count, _ ).
noun( 'grange', 'granges', count, _ ).
noun( 'granite', '-', mass, _ ).
noun( 'grannie', 'grannies', count, _ ).
noun( 'granny', 'grannies', count, _ ).
noun( 'grant', 'grants', count, _ ).
noun( 'granularity', '-', mass, _ ).
noun( 'granule', 'granules', count, _ ).
noun( 'grape', 'grapes', count, _ ).
noun( 'grape-sugar', '-', mass, _ ).
noun( 'grapefruit', 'grapefruit', both, _ ).
noun( 'grapeshot', '-', mass, _ ).
noun( 'grapevine', 'grapevines', count, _ ).
noun( 'graph', 'graphs', count, _ ).
noun( 'graphics', 'graphics', mass, _ ).
noun( 'graphite', '-', mass, _ ).
noun( 'grapnel', 'grapnels', count, _ ).
noun( 'grappling-iron', 'grappling-irons', count, _ ).
noun( 'grasp', 'grasps', count, _ ).
noun( 'grass', 'grasses', both, _ ).
noun( 'grasshopper', 'grasshoppers', count, _ ).
noun( 'grassland', 'grasslands', count, _ ).
noun( 'grate', 'grates', count, _ ).
noun( 'gratefulness', '-', mass, _ ).
noun( 'grater', 'graters', count, _ ).
noun( 'gratification', 'gratifications', both, _ ).
noun( 'grating', 'gratings', count, _ ).
noun( 'gratitude', '-', mass, _ ).
noun( 'gratuity', 'gratuities', count, _ ).
noun( 'grave', 'graves', count, _ ).
noun( 'grave', 'graves', count, _ ).
noun( 'gravel', '-', mass, _ ).
noun( 'gravestone', 'gravestones', count, _ ).
noun( 'graveyard', 'graveyards', count, _ ).
noun( 'graving dock', 'graving docks', count, _ ).
noun( 'gravitation', '-', mass, _ ).
noun( 'gravity', '-', mass, _ ).
noun( 'gravure', 'gravures', count, _ ).
noun( 'gravy', '-', mass, _ ).
noun( 'gravy-boat', 'gravy-boats', count, _ ).
noun( 'gray', 'grays', both, _ ).
noun( 'graze', 'grazes', count, _ ).
noun( 'grazier', 'graziers', count, _ ).
noun( 'grazing-land', 'grazing-lands', count, _ ).
noun( 'grease', '-', mass, _ ).
noun( 'grease-gun', 'grease-guns', count, _ ).
noun( 'greasepaint', '-', mass, _ ).
noun( 'greaser', 'greasers', count, _ ).
noun( 'greasiness', '-', mass, _ ).
noun( 'greatcoat', 'greatcoats', count, _ ).
noun( 'greatness', '-', mass, _ ).
noun( 'grebe', 'grebes', count, _ ).
noun( 'greed', '-', mass, _ ).
noun( 'greediness', '-', mass, _ ).
noun( 'green', 'greens', both, _ ).
noun( 'greenback', 'greenbacks', count, _ ).
noun( 'greenery', '-', mass, _ ).
noun( 'greenfly', '-', mass, _ ).
noun( 'greengage', 'greengages', count, _ ).
noun( 'greengrocer', 'greengrocers', count, _ ).
noun( 'greengrocery', 'greengroceries', count, _ ).
noun( 'greenhorn', 'greenhorns', count, _ ).
noun( 'greenhouse', 'greenhouses', count, _ ).
noun( 'greenishness', '-', mass, _ ).
noun( 'greensward', '-', mass, _ ).
noun( 'greenwood', 'greenwoods', count, _ ).
noun( 'greeting', 'greetings', count, _ ).
noun( 'gregariousness', '-', mass, _ ).
noun( 'gremlin', 'gremlins', count, _ ).
noun( 'grenade', 'grenades', count, _ ).
noun( 'grenadier', 'grenadiers', count, _ ).
noun( 'grey', 'greys', both, _ ).
noun( 'greybeard', 'greybeards', count, _ ).
noun( 'greyhound', 'greyhounds', count, _ ).
noun( 'grid', 'grids', count, _ ).
noun( 'griddle', 'griddles', count, _ ).
noun( 'gridiron', 'gridirons', count, _ ).
noun( 'grief', 'griefs', both, _ ).
noun( 'grievance', 'grievances', count, _ ).
noun( 'griffin', 'griffins', count, _ ).
noun( 'griffon', 'griffons', count, _ ).
noun( 'grill', 'grills', count, _ ).
noun( 'grille', 'grilles', count, _ ).
noun( 'grillroom', 'grillrooms', count, _ ).
noun( 'grimace', 'grimaces', count, _ ).
noun( 'grime', '-', mass, _ ).
noun( 'grimness', '-', mass, _ ).
noun( 'grin', 'grins', count, _ ).
noun( 'grind', 'grinds', count, _ ).
noun( 'grinder', 'grinders', count, _ ).
noun( 'grindstone', 'grindstones', count, _ ).
noun( 'grip', 'grips', count, _ ).
noun( 'grippe', 'grippes', count, _ ).
noun( 'gripsack', 'gripsacks', count, _ ).
noun( 'grist', '-', mass, _ ).
noun( 'gristle', '-', mass, _ ).
noun( 'grit', '-', mass, _ ).
noun( 'grits', 'grits', mass, _ ).
noun( 'grizzly', 'grizzlies', count, _ ).
noun( 'groan', 'groans', count, _ ).
noun( 'groat', 'groats', count, _ ).
noun( 'grocer', 'grocers', count, _ ).
noun( 'grocery', 'groceries', both, _ ).
noun( 'grog', '-', mass, _ ).
noun( 'groin', 'groins', count, _ ).
noun( 'groom', 'grooms', count, _ ).
noun( 'groove', 'grooves', count, _ ).
noun( 'groover', 'groovers', count, _ ).
noun( 'gross', 'gross', count, _ ).
noun( 'grossness', '-', mass, _ ).
noun( 'grot', 'grots', both, _ ).
noun( 'grotesque', 'grotesques', count, _ ).
noun( 'grotesqueness', '-', mass, _ ).
noun( 'grotto', 'grottos', count, _ ).
noun( 'grouch', 'grouches', count, _ ).
noun( 'ground', 'grounds', both, _ ).
noun( 'ground-bait', '-', mass, _ ).
noun( 'ground-fish', 'ground-fish', count, _ ).
noun( 'ground-plan', 'ground-plans', count, _ ).
noun( 'ground-rent', 'ground-rents', both, _ ).
noun( 'grounding', 'groundings', count, _ ).
noun( 'groundnut', 'groundnuts', count, _ ).
noun( 'groundsel', '-', mass, _ ).
noun( 'groundsheet', 'groundsheets', count, _ ).
noun( 'groundsman', 'groundsmen', count, _ ).
noun( 'groundwork', '-', mass, _ ).
noun( 'group', 'groups', count, _ ).
noun( 'grouping', 'groupings', count, _ ).
noun( 'grouse', 'grouse', count, _ ).
noun( 'grove', 'groves', count, _ ).
noun( 'groveller', 'grovellers', count, _ ).
noun( 'grower', 'growers', count, _ ).
noun( 'growl', 'growls', count, _ ).
noun( 'growler', 'growlers', count, _ ).
noun( 'grown-up', 'grown-ups', count, _ ).
noun( 'growth', 'growths', both, _ ).
noun( 'groyne', 'groynes', count, _ ).
noun( 'grub', 'grubs', both, _ ).
noun( 'grubbiness', '-', mass, _ ).
noun( 'grudge', 'grudges', count, _ ).
noun( 'gruel', '-', mass, _ ).
noun( 'gruesomeness', '-', mass, _ ).
noun( 'gruffness', '-', mass, _ ).
noun( 'grumble', 'grumbles', count, _ ).
noun( 'grumbler', 'grumblers', count, _ ).
noun( 'grumpiness', '-', mass, _ ).
noun( 'grunt', 'grunts', count, _ ).
noun( 'gryphon', 'gryphons', count, _ ).
noun( 'guano', 'guanos', both, _ ).
noun( 'guarantee', 'guarantees', count, _ ).
noun( 'guarantor', 'guarantors', count, _ ).
noun( 'guaranty', 'guaranties', count, _ ).
noun( 'guard', 'guards', both, _ ).
noun( 'guard-boat', 'guard-boats', count, _ ).
noun( 'guardhouse', 'guardhouses', count, _ ).
noun( 'guardian', 'guardians', count, _ ).
noun( 'guardianship', 'guardianships', count, _ ).
noun( 'guardrail', 'guardrails', count, _ ).
noun( 'guardroom', 'guardrooms', count, _ ).
noun( 'guardship', 'guardships', count, _ ).
noun( 'guardsman', 'guardsmen', count, _ ).
noun( 'guava', 'guavas', both, _ ).
noun( 'gudgeon', 'gudgeons', count, _ ).
noun( 'guelder rose', 'guelder roses', count, _ ).
noun( 'guerilla', 'guerillas', count, _ ).
noun( 'guerrilla', 'guerrillas', count, _ ).
noun( 'guess', 'guesses', count, _ ).
noun( 'guesstimate', 'guesstimates', count, _ ).
noun( 'guesswork', '-', mass, _ ).
noun( 'guest', 'guests', count, _ ).
noun( 'guest-night', 'guest-nights', count, _ ).
noun( 'guesthouse', 'guesthouses', count, _ ).
noun( 'guestroom', 'guestrooms', count, _ ).
noun( 'guffaw', 'guffaws', count, _ ).
noun( 'guidance', '-', mass, _ ).
noun( 'guide', 'guides', count, _ ).
noun( 'guidebook', 'guidebooks', count, _ ).
noun( 'guideline', 'guidelines', count, _ ).
noun( 'guild', 'guilds', count, _ ).
noun( 'guilder', 'guilders', count, _ ).
noun( 'guildhall', 'guildhalls', count, _ ).
noun( 'guile', '-', mass, _ ).
noun( 'guillemot', 'guillemots', count, _ ).
noun( 'guillotine', 'guillotines', count, _ ).
noun( 'guilt', '-', mass, _ ).
noun( 'guiltiness', '-', mass, _ ).
noun( 'guinea', 'guineas', count, _ ).
noun( 'guinea-fowl', 'guinea-fowl', count, _ ).
noun( 'guinea-pig', 'guinea-pigs', count, _ ).
noun( 'guise', 'guises', count, _ ).
noun( 'guitar', 'guitars', count, _ ).
noun( 'guitarist', 'guitarists', count, _ ).
noun( 'gulch', 'gulches', count, _ ).
noun( 'gulden', 'guldens', count, _ ).
noun( 'gulf', 'gulfs', count, _ ).
noun( 'gull', 'gulls', count, _ ).
noun( 'gullet', 'gullets', count, _ ).
noun( 'gullibility', '-', mass, _ ).
noun( 'gully', 'gullies', count, _ ).
noun( 'gulp', 'gulps', count, _ ).
noun( 'gum', 'gums', both, _ ).
noun( 'gumbo', 'gumbos', count, _ ).
noun( 'gumboil', 'gumboils', count, _ ).
noun( 'gumboot', 'gumboots', count, _ ).
noun( 'gumption', '-', mass, _ ).
noun( 'gumshoe', 'gumshoes', count, _ ).
noun( 'gun', 'guns', count, _ ).
noun( 'gun-carriage', 'gun-carriages', count, _ ).
noun( 'gunboat', 'gunboats', count, _ ).
noun( 'guncotton', '-', mass, _ ).
noun( 'gundog', 'gundogs', count, _ ).
noun( 'gunfire', '-', mass, _ ).
noun( 'gunman', 'gunmen', count, _ ).
noun( 'gunmetal', '-', mass, _ ).
noun( 'gunner', 'gunners', count, _ ).
noun( 'gunnery', '-', mass, _ ).
noun( 'gunny', '-', mass, _ ).
noun( 'gunplay', '-', mass, _ ).
noun( 'gunpoint', '-', count, _ ).
noun( 'gunpowder', '-', mass, _ ).
noun( 'gunroom', 'gunrooms', count, _ ).
noun( 'gunrunner', 'gunrunners', count, _ ).
noun( 'gunrunning', '-', mass, _ ).
noun( 'gunshot', 'gunshots', both, _ ).
noun( 'gunsmith', 'gunsmiths', count, _ ).
noun( 'gunwale', 'gunwales', count, _ ).
noun( 'gurgle', 'gurgles', both, _ ).
noun( 'guru', 'gurus', count, _ ).
noun( 'gush', 'gushes', count, _ ).
noun( 'gusher', 'gushers', count, _ ).
noun( 'gusset', 'gussets', count, _ ).
noun( 'gust', 'gusts', count, _ ).
noun( 'gustation', '-', mass, _ ).
noun( 'gusto', '-', mass, _ ).
noun( 'gut', 'guts', both, _ ).
noun( 'gutta-percha', '-', mass, _ ).
noun( 'gutter', 'gutters', count, _ ).
noun( 'guttersnipe', 'guttersnipes', count, _ ).
noun( 'guttural', 'gutturals', count, _ ).
noun( 'guvnor', 'guvnors', count, _ ).
noun( 'guy', 'guys', count, _ ).
noun( 'guzzler', 'guzzlers', count, _ ).
noun( 'gym', 'gyms', count, _ ).
noun( 'gymkhana', 'gymkhanas', count, _ ).
noun( 'gymnasium', 'gymnasiums', count, _ ).
noun( 'gymnast', 'gymnasts', count, _ ).
noun( 'gymnastics', 'gymnastics', mass, _ ).
noun( 'gymslip', 'gymslips', count, _ ).
noun( 'gynaecologist', 'gynaecologists', count, _ ).
noun( 'gynaecology', '-', mass, _ ).
noun( 'gynecologist', 'gynecologists', count, _ ).
noun( 'gyp', 'gyps', count, _ ).
noun( 'gypsum', '-', mass, _ ).
noun( 'gyration', 'gyrations', both, _ ).
noun( 'gyro', 'gyros', count, _ ).
noun( 'gyroscope', 'gyroscopes', count, _ ).
noun( 'h', '-', count, _ ).
noun( 'ha\'p\'orth', 'ha\'p\'orths', count, _ ).
noun( 'ha\'penny', 'ha\'pennies', count, _ ).
noun( 'habeas corpus', '-', mass, _ ).
noun( 'haberdasher', 'haberdashers', count, _ ).
noun( 'haberdashery', '-', mass, _ ).
noun( 'habit', 'habits', both, _ ).
noun( 'habitat', 'habitats', count, _ ).
noun( 'habitation', 'habitations', both, _ ).
noun( 'habitu_e', 'habitu_es', count, _ ).
noun( 'habituation', '-', mass, _ ).
noun( 'habitude', '-', mass, _ ).
noun( 'hacienda', 'haciendas', count, _ ).
noun( 'hack', 'hacks', count, _ ).
noun( 'hackney', 'hackneys', count, _ ).
noun( 'hacksaw', 'hacksaws', count, _ ).
noun( 'haddock', 'haddock', both, _ ).
noun( 'haematite', 'haematites', count, _ ).
noun( 'haemoglobin', '-', mass, _ ).
noun( 'haemophilia', '-', mass, _ ).
noun( 'haemophiliac', 'haemophiliacs', count, _ ).
noun( 'haemorrhage', 'haemorrhages', both, _ ).
noun( 'haft', 'hafts', count, _ ).
noun( 'hag', 'hags', count, _ ).
noun( 'haggis', 'haggises', count, _ ).
noun( 'hagiology', 'hagiologies', count, _ ).
noun( 'haha', 'hahas', count, _ ).
noun( 'hail', 'hails', both, _ ).
noun( 'hailstone', 'hailstones', count, _ ).
noun( 'hailstorm', 'hailstorms', count, _ ).
noun( 'hair', 'hairs', both, _ ).
noun( 'hair\'s-breadth', 'hair\'s-breadths', count, _ ).
noun( 'hair-breadth', 'hair-breadths', count, _ ).
noun( 'hair-dye', 'hair-dyes', count, _ ).
noun( 'hair-oil', 'hair-oils', both, _ ).
noun( 'hair-shirt', 'hair-shirts', count, _ ).
noun( 'hair-slide', 'hair-slides', count, _ ).
noun( 'hair-trigger', 'hair-triggers', count, _ ).
noun( 'hairbrush', 'hairbrushes', count, _ ).
noun( 'haircloth', 'haircloths', count, _ ).
noun( 'haircut', 'haircuts', count, _ ).
noun( 'hairdo', 'hairdos', count, _ ).
noun( 'hairdresser', 'hairdressers', count, _ ).
noun( 'hairdressing', '-', mass, _ ).
noun( 'hairiness', '-', mass, _ ).
noun( 'hairline', 'hairlines', count, _ ).
noun( 'hairnet', 'hairnets', count, _ ).
noun( 'hairpiece', 'hairpieces', count, _ ).
noun( 'hairpin', 'hairpins', count, _ ).
noun( 'hairsplitting', '-', mass, _ ).
noun( 'hairspring', 'hairsprings', count, _ ).
noun( 'hairstyle', 'hairstyles', count, _ ).
noun( 'hairstylist', 'hairstylists', count, _ ).
noun( 'hake', 'hake', both, _ ).
noun( 'halberd', 'halberds', count, _ ).
noun( 'halberdier', 'halberdiers', count, _ ).
noun( 'half', 'halves', count, _ ).
noun( 'half-blood', 'half-bloods', count, _ ).
noun( 'half-breed', 'half-breeds', count, _ ).
noun( 'half-brother', 'half-brothers', count, _ ).
noun( 'half-caste', 'half-castes', count, _ ).
noun( 'half-crown', 'half-crowns', count, _ ).
noun( 'half-holiday', 'half-holidays', count, _ ).
noun( 'half-hour', 'half-hours', count, _ ).
noun( 'half-pay', '-', mass, _ ).
noun( 'half-sister', 'half-sisters', count, _ ).
noun( 'half-time', '-', mass, _ ).
noun( 'half-track', 'half-tracks', count, _ ).
noun( 'half-truth', 'half-truths', count, _ ).
noun( 'half-volley', 'half-volleys', count, _ ).
noun( 'halfback', 'halfbacks', count, _ ).
noun( 'halfpenny', 'halfpennies', count, _ ).
noun( 'halfpennyworth', 'halfpennyworths', count, _ ).
noun( 'halftone', 'halftones', count, _ ).
noun( 'halfwit', 'halfwits', count, _ ).
noun( 'halibut', 'halibut', both, _ ).
noun( 'halitosis', '-', mass, _ ).
noun( 'hall', 'halls', count, _ ).
noun( 'hall-stand', 'hall-stands', count, _ ).
noun( 'hallelujah', 'hallelujahs', count, _ ).
noun( 'halliard', 'halliards', count, _ ).
noun( 'hallmark', 'hallmarks', count, _ ).
noun( 'halloo', 'halloos', count, _ ).
noun( 'hallow', 'hallows', count, _ ).
noun( 'hallucination', 'hallucinations', both, _ ).
noun( 'halma', '-', mass, _ ).
noun( 'halo', 'halos', count, _ ).
noun( 'halt', 'halts', count, _ ).
noun( 'halter', 'halters', count, _ ).
noun( 'halyard', 'halyards', count, _ ).
noun( 'ham', 'hams', both, _ ).
noun( 'hamadryad', 'hamadryads', count, _ ).
noun( 'hamburger', 'hamburgers', count, _ ).
noun( 'hamlet', 'hamlets', count, _ ).
noun( 'hammer', 'hammers', count, _ ).
noun( 'hammock', 'hammocks', count, _ ).
noun( 'hamper', 'hampers', count, _ ).
noun( 'hamster', 'hamsters', count, _ ).
noun( 'hamstring', 'hamstrings', count, _ ).
noun( 'hand', 'hands', count, _ ).
noun( 'hand-barrow', 'hand-barrows', count, _ ).
noun( 'hand-grenade', 'hand-grenades', count, _ ).
noun( 'hand-luggage', '-', mass, _ ).
noun( 'hand-me-down', 'hand-me-downs', count, _ ).
noun( 'hand-organ', 'hand-organs', count, _ ).
noun( 'hand-out', 'hand-outs', count, _ ).
noun( 'handbag', 'handbags', count, _ ).
noun( 'handbill', 'handbills', count, _ ).
noun( 'handbook', 'handbooks', count, _ ).
noun( 'handbrake', 'handbrakes', count, _ ).
noun( 'handcart', 'handcarts', count, _ ).
noun( 'handclap', 'handclaps', count, _ ).
noun( 'handcuff', 'handcuffs', count, _ ).
noun( 'handful', 'handfuls', count, _ ).
noun( 'handhold', 'handholds', count, _ ).
noun( 'handicap', 'handicaps', count, _ ).
noun( 'handicraft', 'handicrafts', both, _ ).
noun( 'handiness', '-', mass, _ ).
noun( 'handiwork', 'handiworks', both, _ ).
noun( 'handkerchief', 'handkerchiefs', count, _ ).
noun( 'handle', 'handles', count, _ ).
noun( 'handlebar', 'handlebars', count, _ ).
noun( 'handler', 'handlers', count, _ ).
noun( 'handmaid', 'handmaids', count, _ ).
noun( 'handover', 'handovers', count, _ ).
noun( 'handrail', 'handrails', count, _ ).
noun( 'handsaw', 'handsaws', count, _ ).
noun( 'handshake', 'handshakes', count, _ ).
noun( 'handshaking', '-', mass, _ ).
noun( 'handstand', 'handstands', count, _ ).
noun( 'handwork', '-', mass, _ ).
noun( 'handwriting', '-', mass, _ ).
noun( 'handyman', 'handymen', count, _ ).
noun( 'hang', '-', count, _ ).
noun( 'hang-up', 'hang-ups', count, _ ).
noun( 'hangar', 'hangars', count, _ ).
noun( 'hanger', 'hangers', count, _ ).
noun( 'hanger-on', 'hangers-on', count, _ ).
noun( 'hanging', 'hangings', both, _ ).
noun( 'hangman', 'hangmen', count, _ ).
noun( 'hangnail', 'hangnails', count, _ ).
noun( 'hangover', 'hangovers', count, _ ).
noun( 'hank', 'hanks', count, _ ).
noun( 'hankering', 'hankerings', count, _ ).
noun( 'hanky', 'hankies', count, _ ).
noun( 'hanky-panky', '-', mass, _ ).
noun( 'hansom', 'hansoms', count, _ ).
noun( 'hap', 'haps', count, _ ).
noun( 'happening', 'happenings', count, _ ).
noun( 'happiness', '-', mass, _ ).
noun( 'hara-kiri', '-', mass, _ ).
noun( 'harangue', 'harangues', count, _ ).
noun( 'harassment', '-', mass, _ ).
noun( 'harbinger', 'harbingers', count, _ ).
noun( 'harbour', 'harbours', count, _ ).
noun( 'harbourage', 'harbourages', count, _ ).
noun( 'hardback', 'hardbacks', count, _ ).
noun( 'hardboard', '-', mass, _ ).
noun( 'hardcover', 'hardcovers', count, _ ).
noun( 'hardihood', '-', mass, _ ).
noun( 'hardiness', '-', mass, _ ).
noun( 'hardliner', 'hardliners', count, _ ).
noun( 'hardness', '-', mass, _ ).
noun( 'hardship', 'hardships', both, _ ).
noun( 'hardtop', 'hardtops', count, _ ).
noun( 'hardware', '-', mass, _ ).
noun( 'hardwood', '-', mass, _ ).
noun( 'hare', 'hares', count, _ ).
noun( 'harebell', 'harebells', count, _ ).
noun( 'harelip', 'harelips', count, _ ).
noun( 'harem', 'harems', count, _ ).
noun( 'haricot', 'haricots', count, _ ).
noun( 'harlequin', 'harlequins', count, _ ).
noun( 'harlequinade', 'harlequinades', count, _ ).
noun( 'harlot', 'harlots', count, _ ).
noun( 'harm', '-', mass, _ ).
noun( 'harmattan', 'harmattans', count, _ ).
noun( 'harmonic', 'harmonics', count, _ ).
noun( 'harmonica', 'harmonicas', count, _ ).
noun( 'harmonium', 'harmoniums', count, _ ).
noun( 'harmonization', 'harmonizations', count, _ ).
noun( 'harmony', 'harmonies', both, _ ).
noun( 'harness', 'harnesses', count, _ ).
noun( 'harp', 'harps', count, _ ).
noun( 'harper', 'harpers', count, _ ).
noun( 'harpist', 'harpists', count, _ ).
noun( 'harpoon', 'harpoons', count, _ ).
noun( 'harpsichord', 'harpsichords', count, _ ).
noun( 'harpsichordist', 'harpsichordists', count, _ ).
noun( 'harpy', 'harpies', count, _ ).
noun( 'harridan', 'harridans', count, _ ).
noun( 'harrier', 'harriers', count, _ ).
noun( 'harrow', 'harrows', count, _ ).
noun( 'harshness', '-', mass, _ ).
noun( 'hart', 'harts', count, _ ).
noun( 'harum-scarum', 'harum-scarums', count, _ ).
noun( 'harvest', 'harvests', count, _ ).
noun( 'harvester', 'harvesters', count, _ ).
noun( 'has-been', 'has-beens', count, _ ).
noun( 'hash', '-', mass, _ ).
noun( 'hashish', '-', mass, _ ).
noun( 'hasp', 'hasps', count, _ ).
noun( 'hassle', 'hassles', count, _ ).
noun( 'hassock', 'hassocks', count, _ ).
noun( 'haste', '-', mass, _ ).
noun( 'hastiness', '-', mass, _ ).
noun( 'hat', 'hats', count, _ ).
noun( 'hatband', 'hatbands', count, _ ).
noun( 'hatch', 'hatches', count, _ ).
noun( 'hatchery', 'hatcheries', count, _ ).
noun( 'hatchet', 'hatchets', count, _ ).
noun( 'hatching', '-', mass, _ ).
noun( 'hatchway', 'hatchways', count, _ ).
noun( 'hate', 'hates', both, _ ).
noun( 'hatful', 'hatfuls', count, _ ).
noun( 'hatpin', 'hatpins', count, _ ).
noun( 'hatred', 'hatreds', both, _ ).
noun( 'hatter', 'hatters', count, _ ).
noun( 'hauberk', 'hauberks', count, _ ).
noun( 'haughtiness', '-', mass, _ ).
noun( 'haul', 'hauls', count, _ ).
noun( 'haulage', '-', mass, _ ).
noun( 'haulier', 'hauliers', count, _ ).
noun( 'haulm', '-', mass, _ ).
noun( 'haunch', 'haunches', count, _ ).
noun( 'haunt', 'haunts', count, _ ).
noun( 'hautboy', 'hautboys', count, _ ).
noun( 'hauteur', '-', mass, _ ).
noun( 'haven', 'havens', count, _ ).
noun( 'haversack', 'haversacks', count, _ ).
noun( 'havoc', '-', mass, _ ).
noun( 'haw', 'haws', count, _ ).
noun( 'haw-haw', 'haw-haws', count, _ ).
noun( 'hawk', 'hawks', count, _ ).
noun( 'hawker', 'hawkers', count, _ ).
noun( 'hawser', 'hawsers', count, _ ).
noun( 'hawthorn', 'hawthorns', count, _ ).
noun( 'hay', '-', mass, _ ).
noun( 'haycock', 'haycocks', count, _ ).
noun( 'hayfork', 'hayforks', count, _ ).
noun( 'haymaker', 'haymakers', count, _ ).
noun( 'haymaking', '-', mass, _ ).
noun( 'hayrick', 'hayricks', count, _ ).
noun( 'haystack', 'haystacks', count, _ ).
noun( 'haywire', '-', mass, _ ).
noun( 'hazard', 'hazards', both, _ ).
noun( 'haze', 'hazes', both, _ ).
noun( 'hazel', 'hazels', both, _ ).
noun( 'haziness', '-', mass, _ ).
noun( 'he-goat', 'he-goats', count, _ ).
noun( 'he-man', 'he-men', count, _ ).
noun( 'head', 'heads', count, _ ).
noun( 'head-hunter', 'head-hunters', count, _ ).
noun( 'headache', 'headaches', both, _ ).
noun( 'headband', 'headbands', count, _ ).
noun( 'headdress', 'headdresses', count, _ ).
noun( 'header', 'headers', count, _ ).
noun( 'headgear', '-', mass, _ ).
noun( 'heading', 'headings', count, _ ).
noun( 'headlamp', 'headlamps', count, _ ).
noun( 'headland', 'headlands', count, _ ).
noun( 'headlight', 'headlights', count, _ ).
noun( 'headline', 'headlines', count, _ ).
noun( 'headman', 'headmen', count, _ ).
noun( 'headmaster', 'headmasters', count, _ ).
noun( 'headmistress', 'headmistresses', count, _ ).
noun( 'headpiece', 'headpieces', count, _ ).
noun( 'headrest', 'headrests', count, _ ).
noun( 'headroom', '-', mass, _ ).
noun( 'headset', 'headsets', count, _ ).
noun( 'headship', 'headships', count, _ ).
noun( 'headstall', 'headstalls', count, _ ).
noun( 'headstone', 'headstones', count, _ ).
noun( 'headway', '-', mass, _ ).
noun( 'headwind', 'headwinds', count, _ ).
noun( 'headword', 'headwords', count, _ ).
noun( 'healer', 'healers', count, _ ).
noun( 'health', '-', mass, _ ).
noun( 'heap', 'heaps', count, _ ).
noun( 'hearer', 'hearers', count, _ ).
noun( 'hearing', 'hearings', both, _ ).
noun( 'hearing-aid', 'hearing-aids', count, _ ).
noun( 'hearsay', '-', mass, _ ).
noun( 'hearse', 'hearses', count, _ ).
noun( 'heart', 'hearts', both, _ ).
noun( 'heart\'s-ease', '-', mass, _ ).
noun( 'heart-disease', 'heart-diseases', both, _ ).
noun( 'heart-failure', '-', mass, _ ).
noun( 'heartache', 'heartaches', both, _ ).
noun( 'heartbeat', 'heartbeats', count, _ ).
noun( 'heartbreak', '-', mass, _ ).
noun( 'heartburn', '-', mass, _ ).
noun( 'heartburning', '-', mass, _ ).
noun( 'hearth', 'hearths', count, _ ).
noun( 'hearthrug', 'hearthrugs', count, _ ).
noun( 'heartlessness', '-', mass, _ ).
noun( 'heat', 'heats', both, _ ).
noun( 'heat-flash', 'heat-flashes', count, _ ).
noun( 'heater', 'heaters', count, _ ).
noun( 'heath', 'heaths', both, _ ).
noun( 'heathen', 'heathens', count, _ ).
noun( 'heather', 'heathers', both, _ ).
noun( 'heather-mixture', '-', mass, _ ).
noun( 'heating', '-', mass, _ ).
noun( 'heatspot', 'heatspots', count, _ ).
noun( 'heatstroke', '-', mass, _ ).
noun( 'heatwave', 'heatwaves', count, _ ).
noun( 'heave', 'heaves', count, _ ).
noun( 'heaven', 'heavens', count, _ ).
noun( 'heaviness', '-', mass, _ ).
noun( 'heavyweight', 'heavyweights', count, _ ).
noun( 'hecatomb', 'hecatombs', count, _ ).
noun( 'heck', 'hecks', count, _ ).
noun( 'heckler', 'hecklers', count, _ ).
noun( 'hectare', 'hectares', count, _ ).
noun( 'hedge', 'hedges', count, _ ).
noun( 'hedge-sparrow', 'hedge-sparrows', count, _ ).
noun( 'hedgehog', 'hedgehogs', count, _ ).
noun( 'hedgerow', 'hedgerows', count, _ ).
noun( 'hedonism', '-', mass, _ ).
noun( 'hedonist', 'hedonists', count, _ ).
noun( 'heed', '-', mass, _ ).
noun( 'heehaw', 'heehaws', count, _ ).
noun( 'heel', 'heels', count, _ ).
noun( 'hegemony', 'hegemonies', both, _ ).
noun( 'heifer', 'heifers', count, _ ).
noun( 'height', 'heights', both, _ ).
noun( 'heinousness', '-', mass, _ ).
noun( 'heir', 'heirs', count, _ ).
noun( 'heiress', 'heiresses', count, _ ).
noun( 'heirloom', 'heirlooms', count, _ ).
noun( 'helicopter', 'helicopters', count, _ ).
noun( 'heliograph', 'heliographs', count, _ ).
noun( 'heliotrope', 'heliotropes', count, _ ).
noun( 'heliport', 'heliports', count, _ ).
noun( 'helium', '-', mass, _ ).
noun( 'hell', 'hells', count, _ ).
noun( 'hellcat', 'hellcats', count, _ ).
noun( 'helm', 'helms', count, _ ).
noun( 'helmet', 'helmets', count, _ ).
noun( 'helmsman', 'helmsmen', count, _ ).
noun( 'helot', 'helots', count, _ ).
noun( 'help', 'helps', both, _ ).
noun( 'helper', 'helpers', count, _ ).
noun( 'helpfulness', '-', mass, _ ).
noun( 'helping', 'helpings', count, _ ).
noun( 'helplessness', '-', mass, _ ).
noun( 'helpmate', 'helpmates', count, _ ).
noun( 'helpmeet', 'helpmeets', count, _ ).
noun( 'helter-skelter', 'helter-skelters', count, _ ).
noun( 'helve', 'helves', count, _ ).
noun( 'hem', 'hems', count, _ ).
noun( 'hematite', 'hematites', count, _ ).
noun( 'hemisphere', 'hemispheres', count, _ ).
noun( 'hemline', 'hemlines', count, _ ).
noun( 'hemlock', 'hemlocks', both, _ ).
noun( 'hemming-stitch', 'hemming-stitches', count, _ ).
noun( 'hemoglobin', '-', mass, _ ).
noun( 'hemophilia', '-', mass, _ ).
noun( 'hemophiliac', 'hemophiliacs', count, _ ).
noun( 'hemorrhage', 'hemorrhages', both, _ ).
noun( 'hemp', '-', mass, _ ).
noun( 'hemstitch', 'hemstitches', count, _ ).
noun( 'hen', 'hens', count, _ ).
noun( 'hen-party', 'hen-parties', count, _ ).
noun( 'henbane', '-', mass, _ ).
noun( 'henchman', 'henchmen', count, _ ).
noun( 'hencoop', 'hencoops', count, _ ).
noun( 'henhouse', 'henhouses', count, _ ).
noun( 'henna', '-', mass, _ ).
noun( 'henroost', 'henroosts', count, _ ).
noun( 'hepatitis', '-', mass, _ ).
noun( 'heptagon', 'heptagons', count, _ ).
noun( 'herald', 'heralds', count, _ ).
noun( 'heraldry', '-', mass, _ ).
noun( 'herb', 'herbs', count, _ ).
noun( 'herbage', '-', mass, _ ).
noun( 'herbalist', 'herbalists', count, _ ).
noun( 'herd', 'herds', count, _ ).
noun( 'herdsman', 'herdsmen', count, _ ).
noun( 'hereafter', '-', count, _ ).
noun( 'hereditament', 'hereditaments', count, _ ).
noun( 'heredity', '-', mass, _ ).
noun( 'heresy', 'heresies', both, _ ).
noun( 'heretic', 'heretics', count, _ ).
noun( 'heritage', '-', count, _ ).
noun( 'hermaphrodite', 'hermaphrodites', count, _ ).
noun( 'hermit', 'hermits', count, _ ).
noun( 'hermitage', 'hermitages', count, _ ).
noun( 'hernia', 'hernias', both, _ ).
noun( 'hero', 'heroes', count, _ ).
noun( 'heroics', 'heroics', mass, _ ).
noun( 'heroin', '-', mass, _ ).
noun( 'heroine', 'heroines', count, _ ).
noun( 'heroism', '-', mass, _ ).
noun( 'heron', 'herons', count, _ ).
noun( 'heronry', 'heronries', count, _ ).
noun( 'herring', 'herring', both, _ ).
noun( 'herringbone', '-', mass, _ ).
noun( 'hertz', 'hertz', count, _ ).
noun( 'hesitance', '-', both, _ ).
noun( 'hesitancy', '-', mass, _ ).
noun( 'hesitation', 'hesitations', both, _ ).
noun( 'hessian', '-', mass, _ ).
noun( 'heterodoxy', '-', mass, _ ).
noun( 'heterosexual', 'heterosexuals', count, _ ).
noun( 'heterosexuality', '-', mass, _ ).
noun( 'heuristics', 'heuristics', mass, _ ).
noun( 'hewer', 'hewers', count, _ ).
noun( 'hexagon', 'hexagons', count, _ ).
noun( 'hexameter', 'hexameters', count, _ ).
noun( 'heyday', '-', count, _ ).
noun( 'hi-fi', 'hi-fis', count, _ ).
noun( 'hiatus', 'hiatuses', count, _ ).
noun( 'hibernation', '-', mass, _ ).
noun( 'hibiscus', '-', mass, _ ).
noun( 'hiccough', 'hiccoughs', count, _ ).
noun( 'hiccup', 'hiccups', count, _ ).
noun( 'hick', 'hicks', count, _ ).
noun( 'hickory', 'hickories', both, _ ).
noun( 'hide', 'hides', both, _ ).
noun( 'hide-and-seek', '-', mass, _ ).
noun( 'hide-out', 'hide-outs', count, _ ).
noun( 'hideaway', 'hideaways', count, _ ).
noun( 'hiding', 'hidings', both, _ ).
noun( 'hiding-place', 'hiding-places', count, _ ).
noun( 'hierarchy', 'hierarchies', count, _ ).
noun( 'hieroglyph', 'hieroglyphs', count, _ ).
noun( 'hieroglyphics', 'hieroglyphics', mass, _ ).
noun( 'high-fidelity', '-', mass, _ ).
noun( 'high-frequency', 'high-frequencies', both, _ ).
noun( 'high-mindedness', '-', mass, _ ).
noun( 'high-pressure', '-', mass, _ ).
noun( 'high-up', 'high-ups', count, _ ).
noun( 'highball', 'highballs', count, _ ).
noun( 'highboy', 'highboys', count, _ ).
noun( 'highbrow', 'highbrows', count, _ ).
noun( 'highflier', 'highfliers', count, _ ).
noun( 'highflyer', 'highflyers', count, _ ).
noun( 'highjack', 'highjacks', count, _ ).
noun( 'highland', 'highlands', count, _ ).
noun( 'highlight', 'highlights', count, _ ).
noun( 'highness', 'highnesses', both, _ ).
noun( 'highroad', 'highroads', count, _ ).
noun( 'highway', 'highways', count, _ ).
noun( 'highwayman', 'highwaymen', count, _ ).
noun( 'hijack', 'hijacks', count, _ ).
noun( 'hijacker', 'hijackers', count, _ ).
noun( 'hike', 'hikes', count, _ ).
noun( 'hiker', 'hikers', count, _ ).
noun( 'hilarity', '-', mass, _ ).
noun( 'hill', 'hills', count, _ ).
noun( 'hill-billy', 'hill-billies', count, _ ).
noun( 'hillock', 'hillocks', count, _ ).
noun( 'hillside', 'hillsides', count, _ ).
noun( 'hilt', 'hilts', count, _ ).
noun( 'hind', 'hinds', count, _ ).
noun( 'hindrance', 'hindrances', count, _ ).
noun( 'hindsight', '-', mass, _ ).
noun( 'hinge', 'hinges', count, _ ).
noun( 'hint', 'hints', count, _ ).
noun( 'hinterland', 'hinterlands', count, _ ).
noun( 'hip', 'hips', count, _ ).
noun( 'hip-bath', 'hip-baths', count, _ ).
noun( 'hip-flask', 'hip-flasks', count, _ ).
noun( 'hip-pocket', 'hip-pockets', count, _ ).
noun( 'hippie', 'hippies', count, _ ).
noun( 'hippo', 'hippos', count, _ ).
noun( 'hippodrome', 'hippodromes', count, _ ).
noun( 'hippopotamus', 'hippopotamuses', count, _ ).
noun( 'hippy', 'hippies', count, _ ).
noun( 'hire', '-', mass, _ ).
noun( 'hireling', 'hirelings', count, _ ).
noun( 'hiss', 'hisses', count, _ ).
noun( 'histogram', 'histograms', count, _ ).
noun( 'historian', 'historians', count, _ ).
noun( 'history', 'histories', both, _ ).
noun( 'histrionics', 'histrionics', mass, _ ).
noun( 'hit', 'hits', count, _ ).
noun( 'hitch', 'hitches', count, _ ).
noun( 'hitchhiker', 'hitchhikers', count, _ ).
noun( 'hive', 'hives', count, _ ).
noun( 'hoard', 'hoards', count, _ ).
noun( 'hoarder', 'hoarders', count, _ ).
noun( 'hoarding', 'hoardings', count, _ ).
noun( 'hoarfrost', '-', mass, _ ).
noun( 'hoariness', '-', mass, _ ).
noun( 'hoarseness', '-', mass, _ ).
noun( 'hoax', 'hoaxes', count, _ ).
noun( 'hoaxer', 'hoaxers', count, _ ).
noun( 'hob', 'hobs', count, _ ).
noun( 'hobble', 'hobbles', count, _ ).
noun( 'hobble-skirt', 'hobble-skirts', count, _ ).
noun( 'hobbledehoy', 'hobbledehoys', count, _ ).
noun( 'hobby', 'hobbies', count, _ ).
noun( 'hobbyhorse', 'hobbyhorses', count, _ ).
noun( 'hobgoblin', 'hobgoblins', count, _ ).
noun( 'hobnail', 'hobnails', count, _ ).
noun( 'hobo', 'hobos', count, _ ).
noun( 'hock', 'hocks', both, _ ).
noun( 'hockey', '-', mass, _ ).
noun( 'hocus-pocus', '-', mass, _ ).
noun( 'hod', 'hods', count, _ ).
noun( 'hodgepodge', '-', mass, _ ).
noun( 'hoe', 'hoes', count, _ ).
noun( 'hog', 'hogs', count, _ ).
noun( 'hogshead', 'hogsheads', count, _ ).
noun( 'hogwash', '-', mass, _ ).
noun( 'hoist', 'hoists', count, _ ).
noun( 'hold', 'holds', both, _ ).
noun( 'hold-up', 'hold-ups', count, _ ).
noun( 'holdall', 'holdalls', count, _ ).
noun( 'holder', 'holders', count, _ ).
noun( 'holding', 'holdings', count, _ ).
noun( 'holdup', 'holdups', count, _ ).
noun( 'hole', 'holes', count, _ ).
noun( 'holiday', 'holidays', count, _ ).
noun( 'holiday-maker', 'holiday-makers', count, _ ).
noun( 'holiness', '-', mass, _ ).
noun( 'holloa', 'holloas', count, _ ).
noun( 'hollow', 'hollows', count, _ ).
noun( 'holly', '-', mass, _ ).
noun( 'hollyhock', 'hollyhocks', count, _ ).
noun( 'holm-oak', 'holm-oaks', count, _ ).
noun( 'holocaust', 'holocausts', count, _ ).
noun( 'holograph', 'holographs', count, _ ).
noun( 'holster', 'holsters', count, _ ).
noun( 'holy', 'holies', count, _ ).
noun( 'holystone', '-', mass, _ ).
noun( 'homage', '-', mass, _ ).
noun( 'home', 'homes', count, _ ).
noun( 'home-farm', 'home-farms', count, _ ).
noun( 'homecoming', 'homecomings', count, _ ).
noun( 'homeland', 'homelands', count, _ ).
noun( 'homeliness', '-', mass, _ ).
noun( 'homeopath', 'homeopaths', count, _ ).
noun( 'homesickness', '-', mass, _ ).
noun( 'homespun', '-', mass, _ ).
noun( 'homestead', 'homesteads', count, _ ).
noun( 'homework', '-', mass, _ ).
noun( 'homicide', 'homicides', both, _ ).
noun( 'homiletics', 'homiletics', mass, _ ).
noun( 'homily', 'homilies', count, _ ).
noun( 'hominy', '-', mass, _ ).
noun( 'homo', 'homos', count, _ ).
noun( 'homo sapiens', '-', count, _ ).
noun( 'homoeopath', 'homoeopaths', count, _ ).
noun( 'homoeopathy', '-', mass, _ ).
noun( 'homogeneity', '-', mass, _ ).
noun( 'homograph', 'homographs', count, _ ).
noun( 'homonym', 'homonyms', count, _ ).
noun( 'homophone', 'homophones', count, _ ).
noun( 'homosexual', 'homosexuals', count, _ ).
noun( 'homosexuality', '-', mass, _ ).
noun( 'hone', 'hones', count, _ ).
noun( 'honesty', '-', mass, _ ).
noun( 'honey', 'honeys', both, _ ).
noun( 'honeybee', 'honeybees', count, _ ).
noun( 'honeycomb', 'honeycombs', both, _ ).
noun( 'honeydew', '-', mass, _ ).
noun( 'honeymoon', 'honeymoons', count, _ ).
noun( 'honeysuckle', '-', mass, _ ).
noun( 'honk', 'honks', count, _ ).
noun( 'honorarium', 'honorariums', count, _ ).
noun( 'honorific', 'honorifics', count, _ ).
noun( 'honour', 'honours', both, _ ).
noun( 'hooch', '-', mass, _ ).
noun( 'hood', 'hoods', count, _ ).
noun( 'hoodlum', 'hoodlums', count, _ ).
noun( 'hoodoo', 'hoodoos', count, _ ).
noun( 'hooey', '-', mass, _ ).
noun( 'hoof', 'hoofs', count, _ ).
noun( 'hook', 'hooks', count, _ ).
noun( 'hook-up', 'hook-ups', count, _ ).
noun( 'hookah', 'hookahs', count, _ ).
noun( 'hooker', 'hookers', count, _ ).
noun( 'hookworm', 'hookworms', both, _ ).
noun( 'hooky', '-', count, _ ).
noun( 'hooligan', 'hooligans', count, _ ).
noun( 'hooliganism', '-', mass, _ ).
noun( 'hoop', 'hoops', count, _ ).
noun( 'hoop-la', '-', mass, _ ).
noun( 'hoot', 'hoots', count, _ ).
noun( 'hooter', 'hooters', count, _ ).
noun( 'hop', 'hops', count, _ ).
noun( 'hop-field', 'hop-fields', count, _ ).
noun( 'hop-garden', 'hop-gardens', count, _ ).
noun( 'hop-picker', 'hop-pickers', count, _ ).
noun( 'hop-pole', 'hop-poles', count, _ ).
noun( 'hope', 'hopes', both, _ ).
noun( 'hopefulness', '-', mass, _ ).
noun( 'hopelessness', '-', mass, _ ).
noun( 'hopper', 'hoppers', count, _ ).
noun( 'hopscotch', '-', mass, _ ).
noun( 'horde', 'hordes', count, _ ).
noun( 'horizon', 'horizons', count, _ ).
noun( 'hormone', 'hormones', count, _ ).
noun( 'horn', 'horns', both, _ ).
noun( 'hornbeam', 'hornbeams', count, _ ).
noun( 'hornbill', 'hornbills', count, _ ).
noun( 'hornet', 'hornets', count, _ ).
noun( 'hornpipe', 'hornpipes', count, _ ).
noun( 'horology', '-', mass, _ ).
noun( 'horoscope', 'horoscopes', count, _ ).
noun( 'horridness', '-', mass, _ ).
noun( 'horror', 'horrors', both, _ ).
noun( 'horse', 'horses', count, _ ).
noun( 'horse-chestnut', 'horse-chestnuts', count, _ ).
noun( 'horse-laugh', 'horse-laughs', count, _ ).
noun( 'horse-pond', 'horse-ponds', count, _ ).
noun( 'horse-sense', '-', mass, _ ).
noun( 'horseback', '-', mass, _ ).
noun( 'horsebox', 'horseboxes', count, _ ).
noun( 'horseflesh', '-', mass, _ ).
noun( 'horsefly', 'horseflies', count, _ ).
noun( 'horsehair', '-', mass, _ ).
noun( 'horseman', 'horsemen', count, _ ).
noun( 'horsemanship', '-', mass, _ ).
noun( 'horsemeat', '-', mass, _ ).
noun( 'horseplay', '-', mass, _ ).
noun( 'horsepower', 'horsepower', both, _ ).
noun( 'horserace', 'horseraces', count, _ ).
noun( 'horseracing', '-', mass, _ ).
noun( 'horseradish', '-', mass, _ ).
noun( 'horseshoe', 'horseshoes', count, _ ).
noun( 'horsewhip', 'horsewhips', count, _ ).
noun( 'horsewoman', 'horsewomen', count, _ ).
noun( 'horticulture', '-', mass, _ ).
noun( 'horticulturist', 'horticulturists', count, _ ).
noun( 'hosanna', 'hosannas', count, _ ).
noun( 'hose', 'hoses', both, _ ).
noun( 'hosepipe', 'hosepipes', count, _ ).
noun( 'hosier', 'hosiers', count, _ ).
noun( 'hosiery', '-', mass, _ ).
noun( 'hospice', 'hospices', count, _ ).
noun( 'hospital', 'hospitals', count, _ ).
noun( 'hospitality', '-', mass, _ ).
noun( 'hospitalization', 'hospitalizations', both, _ ).
noun( 'host', 'hosts', count, _ ).
noun( 'hostage', 'hostages', count, _ ).
noun( 'hostel', 'hostels', count, _ ).
noun( 'hosteller', 'hostellers', count, _ ).
noun( 'hostelry', 'hostelries', count, _ ).
noun( 'hostess', 'hostesses', count, _ ).
noun( 'hostility', 'hostilities', both, _ ).
noun( 'hot-water-bottle', 'hot-water-bottles', count, _ ).
noun( 'hotbed', 'hotbeds', count, _ ).
noun( 'hotchpotch', 'hotchpotches', count, _ ).
noun( 'hotel', 'hotels', count, _ ).
noun( 'hotelier', 'hoteliers', count, _ ).
noun( 'hothead', 'hotheads', count, _ ).
noun( 'hothouse', 'hothouses', count, _ ).
noun( 'hotplate', 'hotplates', count, _ ).
noun( 'hound', 'hounds', count, _ ).
noun( 'hour', 'hours', count, _ ).
noun( 'hourglass', 'hourglasses', count, _ ).
noun( 'houri', 'houris', count, _ ).
noun( 'house', 'houses', count, _ ).
noun( 'house-party', 'house-parties', count, _ ).
noun( 'house-warming', 'house-warmings', count, _ ).
noun( 'houseboat', 'houseboats', count, _ ).
noun( 'housebreaker', 'housebreakers', count, _ ).
noun( 'housecoat', 'housecoats', count, _ ).
noun( 'housecraft', '-', mass, _ ).
noun( 'housedog', 'housedogs', count, _ ).
noun( 'housefather', 'housefathers', count, _ ).
noun( 'housefly', 'houseflies', count, _ ).
noun( 'houseful', 'housefuls', count, _ ).
noun( 'household', 'households', count, _ ).
noun( 'householder', 'householders', count, _ ).
noun( 'housekeeper', 'housekeepers', count, _ ).
noun( 'housemaid', 'housemaids', count, _ ).
noun( 'houseman', 'housemen', count, _ ).
noun( 'housemaster', 'housemasters', count, _ ).
noun( 'housemother', 'housemothers', count, _ ).
noun( 'houseroom', '-', mass, _ ).
noun( 'housetop', 'housetops', count, _ ).
noun( 'housewife', 'housewives', count, _ ).
noun( 'housewifery', '-', mass, _ ).
noun( 'housework', '-', mass, _ ).
noun( 'housewrecker', 'housewreckers', count, _ ).
noun( 'housing', '-', mass, _ ).
noun( 'hovel', 'hovels', count, _ ).
noun( 'hovercraft', 'hovercraft', count, _ ).
noun( 'how-d\'ye-do', 'how-d\'ye-dos', count, _ ).
noun( 'howdah', 'howdahs', count, _ ).
noun( 'howitzer', 'howitzers', count, _ ).
noun( 'howl', 'howls', count, _ ).
noun( 'howler', 'howlers', count, _ ).
noun( 'hoyden', 'hoydens', count, _ ).
noun( 'hub', 'hubs', count, _ ).
noun( 'hubble-bubble', 'hubble-bubbles', count, _ ).
noun( 'hubbub', '-', mass, _ ).
noun( 'hubby', 'hubbies', count, _ ).
noun( 'hubris', '-', mass, _ ).
noun( 'huckaback', '-', mass, _ ).
noun( 'huckleberry', 'huckleberries', count, _ ).
noun( 'huckster', 'hucksters', count, _ ).
noun( 'huddle', 'huddles', count, _ ).
noun( 'hue', 'hues', count, _ ).
noun( 'huff', 'huffs', count, _ ).
noun( 'hug', 'hugs', count, _ ).
noun( 'hugger-mugger', '-', count, _ ).
noun( 'hula', 'hulas', count, _ ).
noun( 'hulk', 'hulks', count, _ ).
noun( 'hull', 'hulls', count, _ ).
noun( 'hullabaloo', 'hullabaloos', count, _ ).
noun( 'hum', 'hums', count, _ ).
noun( 'human', 'humans', count, _ ).
noun( 'humanism', '-', mass, _ ).
noun( 'humanist', 'humanists', count, _ ).
noun( 'humanitarian', 'humanitarians', count, _ ).
noun( 'humanitarianism', '-', mass, _ ).
noun( 'humanity', 'humanities', both, _ ).
noun( 'humankind', '-', mass, _ ).
noun( 'humbug', 'humbugs', both, _ ).
noun( 'humdinger', 'humdingers', count, _ ).
noun( 'humerus', 'humeruses', count, _ ).
noun( 'humidity', '-', mass, _ ).
noun( 'humiliation', 'humiliations', both, _ ).
noun( 'humility', '-', mass, _ ).
noun( 'humming-top', 'humming-tops', count, _ ).
noun( 'hummingbird', 'hummingbirds', count, _ ).
noun( 'hummock', 'hummocks', count, _ ).
noun( 'humorist', 'humorists', count, _ ).
noun( 'humour', 'humours', both, _ ).
noun( 'humourist', 'humourists', count, _ ).
noun( 'hump', 'humps', count, _ ).
noun( 'humpback', 'humpbacks', count, _ ).
noun( 'humus', '-', mass, _ ).
noun( 'hunch', 'hunches', count, _ ).
noun( 'hunchback', 'hunchbacks', count, _ ).
noun( 'hundred', 'hundreds', count, _ ).
noun( 'hundredth', 'hundredths', count, _ ).
noun( 'hundredweight', 'hundredweights', count, _ ).
noun( 'hunger', '-', mass, _ ).
noun( 'hunger-march', 'hunger-marches', count, _ ).
noun( 'hunger-marcher', 'hunger-marchers', count, _ ).
noun( 'hunk', 'hunks', count, _ ).
noun( 'hunt', 'hunts', count, _ ).
noun( 'hunter', 'hunters', count, _ ).
noun( 'hunting', '-', mass, _ ).
noun( 'hunting-crop', 'hunting-crops', count, _ ).
noun( 'huntress', 'huntresses', count, _ ).
noun( 'huntsman', 'huntsmen', count, _ ).
noun( 'hurdle', 'hurdles', count, _ ).
noun( 'hurdler', 'hurdlers', count, _ ).
noun( 'hurdy-gurdy', 'hurdy-gurdies', count, _ ).
noun( 'hurl', 'hurls', count, _ ).
noun( 'hurling', '-', mass, _ ).
noun( 'hurly-burly', '-', mass, _ ).
noun( 'hurricane', 'hurricanes', count, _ ).
noun( 'hurry', 'hurries', both, _ ).
noun( 'hurt', '-', mass, _ ).
noun( 'husband', 'husbands', count, _ ).
noun( 'husbandman', 'husbandmen', count, _ ).
noun( 'husbandry', '-', mass, _ ).
noun( 'hush', 'hushes', count, _ ).
noun( 'hush-money', '-', mass, _ ).
noun( 'husk', 'husks', count, _ ).
noun( 'huskiness', '-', mass, _ ).
noun( 'husky', 'huskies', count, _ ).
noun( 'hussar', 'hussars', count, _ ).
noun( 'hussy', 'hussies', count, _ ).
noun( 'hustle', '-', count, _ ).
noun( 'hustler', 'hustlers', count, _ ).
noun( 'hut', 'huts', count, _ ).
noun( 'hutch', 'hutches', count, _ ).
noun( 'hutment', 'hutments', count, _ ).
noun( 'hyacinth', 'hyacinths', count, _ ).
noun( 'hyaena', 'hyaenas', count, _ ).
noun( 'hybrid', 'hybrids', count, _ ).
noun( 'hydra', 'hydras', count, _ ).
noun( 'hydrangea', 'hydrangeas', count, _ ).
noun( 'hydrant', 'hydrants', count, _ ).
noun( 'hydrate', 'hydrates', count, _ ).
noun( 'hydraulics', 'hydraulics', mass, _ ).
noun( 'hydrocarbon', 'hydrocarbons', count, _ ).
noun( 'hydrofoil', 'hydrofoils', count, _ ).
noun( 'hydrogen', '-', mass, _ ).
noun( 'hydropathy', '-', mass, _ ).
noun( 'hydrophobia', '-', mass, _ ).
noun( 'hydroplane', 'hydroplanes', count, _ ).
noun( 'hydroponics', 'hydroponics', mass, _ ).
noun( 'hyena', 'hyenas', count, _ ).
noun( 'hygiene', '-', mass, _ ).
noun( 'hymen', 'hymens', count, _ ).
noun( 'hymn', 'hymns', count, _ ).
noun( 'hymnal', 'hymnals', count, _ ).
noun( 'hyperbola', 'hyperbolas', count, _ ).
noun( 'hyperbole', 'hyperboles', both, _ ).
noun( 'hypermarket', 'hypermarkets', count, _ ).
noun( 'hyphen', 'hyphens', count, _ ).
noun( 'hypnosis', 'hypnoses', both, _ ).
noun( 'hypnotism', '-', mass, _ ).
noun( 'hypnotist', 'hypnotists', count, _ ).
noun( 'hypo', '-', mass, _ ).
noun( 'hypochondria', 'hypochondrias', both, _ ).
noun( 'hypochondriac', 'hypochondriacs', count, _ ).
noun( 'hypocrisy', 'hypocrisies', both, _ ).
noun( 'hypocrite', 'hypocrites', count, _ ).
noun( 'hypodermic', 'hypodermics', count, _ ).
noun( 'hypotenuse', 'hypotenuses', count, _ ).
noun( 'hypothesis', 'hypotheses', count, _ ).
noun( 'hyssop', '-', mass, _ ).
noun( 'hysteria', '-', mass, _ ).
noun( 'hysterics', 'hysterics', mass, _ ).
noun( 'i', '-', count, _ ).
noun( 'iamb', 'iambs', count, _ ).
noun( 'iambics', 'iambics', mass, _ ).
noun( 'iambus', 'iambuses', count, _ ).
noun( 'ibex', 'ibexes', count, _ ).
noun( 'ibid', '-', proper, _ ).
noun( 'ibis', 'ibises', count, _ ).
noun( 'ice', 'ices', both, _ ).
noun( 'ice-axe', 'ice-axes', count, _ ).
noun( 'ice-cream', 'ice-creams', both, _ ).
noun( 'ice-lolly', 'ice-lollies', count, _ ).
noun( 'ice-show', 'ice-shows', count, _ ).
noun( 'ice-skate', 'ice-skates', count, _ ).
noun( 'ice-skating', '-', mass, _ ).
noun( 'ice-tray', 'ice-trays', count, _ ).
noun( 'iceberg', 'icebergs', count, _ ).
noun( 'iceboat', 'iceboats', count, _ ).
noun( 'icebox', 'iceboxes', count, _ ).
noun( 'icebreaker', 'icebreakers', count, _ ).
noun( 'icecap', 'icecaps', count, _ ).
noun( 'icecube', 'icecubes', count, _ ).
noun( 'icefall', 'icefalls', count, _ ).
noun( 'icefield', 'icefields', count, _ ).
noun( 'icefloe', 'icefloes', count, _ ).
noun( 'icehouse', 'icehouses', count, _ ).
noun( 'iceman', 'icemen', count, _ ).
noun( 'icepack', 'icepacks', count, _ ).
noun( 'icepick', 'icepicks', count, _ ).
noun( 'icerink', 'icerinks', count, _ ).
noun( 'ichneumon', 'ichneumons', count, _ ).
noun( 'ichneumon-fly', 'ichneumon-flies', count, _ ).
noun( 'icicle', 'icicles', count, _ ).
noun( 'icing', '-', mass, _ ).
noun( 'icon', 'icons', count, _ ).
noun( 'iconoclast', 'iconoclasts', count, _ ).
noun( 'id', '-', count, _ ).
noun( 'idea', 'ideas', count, _ ).
noun( 'ideal', 'ideals', count, _ ).
noun( 'idealism', '-', mass, _ ).
noun( 'idealist', 'idealists', count, _ ).
noun( 'idealization', 'idealizations', count, _ ).
noun( 'idem', '-', proper, _ ).
noun( 'identification', '-', mass, _ ).
noun( 'identikit', 'identikits', count, _ ).
noun( 'identity', 'identities', both, _ ).
noun( 'ideogram', 'ideograms', count, _ ).
noun( 'ideograph', 'ideographs', count, _ ).
noun( 'ideology', 'ideologies', both, _ ).
noun( 'idiocy', 'idiocies', both, _ ).
noun( 'idiolect', 'idiolects', count, _ ).
noun( 'idiom', 'idioms', both, _ ).
noun( 'idiosyncrasy', 'idiosyncrasies', count, _ ).
noun( 'idiot', 'idiots', count, _ ).
noun( 'idleness', '-', mass, _ ).
noun( 'idler', 'idlers', count, _ ).
noun( 'idol', 'idols', count, _ ).
noun( 'idolater', 'idolaters', count, _ ).
noun( 'idolatress', 'idolatresses', count, _ ).
noun( 'idolatry', 'idolatries', both, _ ).
noun( 'idolization', 'idolizations', count, _ ).
noun( 'idyll', 'idylls', count, _ ).
noun( 'ie', '-', proper, _ ).
noun( 'igloo', 'igloos', count, _ ).
noun( 'ignis fatuus', 'ignes fatui', count, _ ).
noun( 'ignition', '-', mass, _ ).
noun( 'ignominy', 'ignominies', both, _ ).
noun( 'ignoramus', 'ignoramuses', count, _ ).
noun( 'ignorance', '-', mass, _ ).
noun( 'iguana', 'iguanas', count, _ ).
noun( 'ikon', 'ikons', count, _ ).
noun( 'ilex', 'ilexes', count, _ ).
noun( 'ilk', '-', mass, _ ).
noun( 'ill', 'ills', both, _ ).
noun( 'ill-breeding', '-', mass, _ ).
noun( 'ill-treatment', '-', mass, _ ).
noun( 'ill-usage', '-', mass, _ ).
noun( 'illegality', 'illegalities', both, _ ).
noun( 'illegibility', '-', mass, _ ).
noun( 'illegitimacy', '-', mass, _ ).
noun( 'illegitimate', 'illegitimate', count, _ ).
noun( 'illiberality', '-', mass, _ ).
noun( 'illiteracy', 'illiteracies', both, _ ).
noun( 'illiterate', 'illiterates', count, _ ).
noun( 'illness', 'illnesses', both, _ ).
noun( 'illogicality', 'illogicalities', both, _ ).
noun( 'illogicalness', '-', mass, _ ).
noun( 'illumination', 'illuminations', both, _ ).
noun( 'illusion', 'illusions', both, _ ).
noun( 'illusionist', 'illusionists', count, _ ).
noun( 'illustration', 'illustrations', both, _ ).
noun( 'illustrator', 'illustrators', count, _ ).
noun( 'image', 'images', count, _ ).
noun( 'imagery', '-', mass, _ ).
noun( 'imagination', 'imaginations', both, _ ).
noun( 'imam', 'imams', count, _ ).
noun( 'imbalance', 'imbalances', count, _ ).
noun( 'imbecile', 'imbeciles', count, _ ).
noun( 'imbecility', 'imbecilities', both, _ ).
noun( 'imbroglio', 'imbroglios', count, _ ).
noun( 'imitation', 'imitations', both, _ ).
noun( 'imitator', 'imitators', count, _ ).
noun( 'immanence', 'immanences', count, _ ).
noun( 'immaturity', '-', mass, _ ).
noun( 'immediacy', '-', mass, _ ).
noun( 'immensity', 'immensities', both, _ ).
noun( 'immersion', 'immersions', count, _ ).
noun( 'immigrant', 'immigrants', count, _ ).
noun( 'immigration', 'immigrations', both, _ ).
noun( 'imminence', '-', mass, _ ).
noun( 'immobility', '-', mass, _ ).
noun( 'immobilization', '-', mass, _ ).
noun( 'immodesty', 'immodesties', both, _ ).
noun( 'immolation', 'immolations', both, _ ).
noun( 'immorality', 'immoralities', both, _ ).
noun( 'immortal', 'immortals', count, _ ).
noun( 'immortality', '-', mass, _ ).
noun( 'immunity', '-', mass, _ ).
noun( 'immunization', '-', mass, _ ).
noun( 'immunology', '-', mass, _ ).
noun( 'immutability', '-', mass, _ ).
noun( 'imp', 'imps', count, _ ).
noun( 'impact', 'impacts', both, _ ).
noun( 'impairment', 'impairments', count, _ ).
noun( 'impala', 'impalas', count, _ ).
noun( 'impalement', 'impalements', count, _ ).
noun( 'impartiality', '-', mass, _ ).
noun( 'impasse', 'impasses', count, _ ).
noun( 'impassiveness', '-', mass, _ ).
noun( 'impassivity', '-', mass, _ ).
noun( 'impatience', '-', mass, _ ).
noun( 'impeachment', 'impeachments', both, _ ).
noun( 'impediment', 'impediments', count, _ ).
noun( 'impeller', 'impellers', count, _ ).
noun( 'impenitence', '-', mass, _ ).
noun( 'imperfect', '-', count, _ ).
noun( 'imperfection', 'imperfections', both, _ ).
noun( 'imperial', 'imperials', count, _ ).
noun( 'imperialism', '-', mass, _ ).
noun( 'imperialist', 'imperialists', count, _ ).
noun( 'imperiousness', '-', mass, _ ).
noun( 'impermanence', '-', mass, _ ).
noun( 'impersonation', 'impersonations', both, _ ).
noun( 'impersonator', 'impersonators', count, _ ).
noun( 'impertinence', 'impertinences', both, _ ).
noun( 'imperturbability', '-', mass, _ ).
noun( 'impetigo', '-', mass, _ ).
noun( 'impetuosity', 'impetuosities', both, _ ).
noun( 'impetus', '-', both, _ ).
noun( 'impiety', 'impieties', both, _ ).
noun( 'impingement', 'impingements', count, _ ).
noun( 'impishness', '-', mass, _ ).
noun( 'implausibility', '-', mass, _ ).
noun( 'implement', 'implements', count, _ ).
noun( 'implementation', 'implementations', count, _ ).
noun( 'implication', 'implications', both, _ ).
noun( 'implosion', 'implosions', both, _ ).
noun( 'impoliteness', '-', mass, _ ).
noun( 'imponderable', 'imponderables', count, _ ).
noun( 'import', 'imports', both, _ ).
noun( 'importance', '-', mass, _ ).
noun( 'importation', 'importations', both, _ ).
noun( 'importer', 'importers', count, _ ).
noun( 'importunity', '-', mass, _ ).
noun( 'imposition', 'impositions', both, _ ).
noun( 'impossibility', 'impossibilities', both, _ ).
noun( 'impostor', 'impostors', count, _ ).
noun( 'imposture', 'impostures', both, _ ).
noun( 'impotence', '-', mass, _ ).
noun( 'impoverishment', 'impoverishments', count, _ ).
noun( 'impracticability', '-', mass, _ ).
noun( 'impracticableness', '-', mass, _ ).
noun( 'imprecation', 'imprecations', count, _ ).
noun( 'imprecision', '-', mass, _ ).
noun( 'impregnability', '-', mass, _ ).
noun( 'impresario', 'impresarios', count, _ ).
noun( 'impress', 'impresses', count, _ ).
noun( 'impression', 'impressions', count, _ ).
noun( 'impressionism', '-', mass, _ ).
noun( 'impressionist', 'impressionists', count, _ ).
noun( 'impressiveness', '-', mass, _ ).
noun( 'imprimatur', 'imprimaturs', count, _ ).
noun( 'imprint', 'imprints', count, _ ).
noun( 'imprisonment', '-', mass, _ ).
noun( 'improbability', 'improbabilities', both, _ ).
noun( 'impromptu', 'impromptus', count, _ ).
noun( 'impropriety', 'improprieties', both, _ ).
noun( 'improvement', 'improvements', both, _ ).
noun( 'improver', 'improvers', count, _ ).
noun( 'improvidence', '-', mass, _ ).
noun( 'improvisation', 'improvisations', both, _ ).
noun( 'imprudence', 'imprudences', both, _ ).
noun( 'impudence', '-', mass, _ ).
noun( 'impulse', 'impulses', both, _ ).
noun( 'impulsion', 'impulsions', both, _ ).
noun( 'impulsiveness', '-', mass, _ ).
noun( 'impunity', '-', mass, _ ).
noun( 'impurity', 'impurities', both, _ ).
noun( 'imputation', 'imputations', both, _ ).
noun( 'in', 'ins', count, _ ).
noun( 'in-fighting', '-', mass, _ ).
noun( 'in-patient', 'in-patients', count, _ ).
noun( 'in-tray', 'in-trays', count, _ ).
noun( 'inability', '-', mass, _ ).
noun( 'inaccessibility', '-', mass, _ ).
noun( 'inaccuracy', 'inaccuracies', both, _ ).
noun( 'inaction', '-', mass, _ ).
noun( 'inactivity', '-', mass, _ ).
noun( 'inadequacy', 'inadequacies', both, _ ).
noun( 'inadvertence', 'inadvertences', both, _ ).
noun( 'inadvisability', '-', mass, _ ).
noun( 'inanition', '-', mass, _ ).
noun( 'inanity', 'inanities', both, _ ).
noun( 'inappropriateness', '-', mass, _ ).
noun( 'inaptitude', '-', mass, _ ).
noun( 'inattention', '-', mass, _ ).
noun( 'inaudibility', '-', mass, _ ).
noun( 'inaugural', 'inaugurals', count, _ ).
noun( 'inauguration', 'inaugurations', both, _ ).
noun( 'inbreeding', '-', mass, _ ).
noun( 'incandescence', '-', mass, _ ).
noun( 'incantation', 'incantations', both, _ ).
noun( 'incapability', '-', mass, _ ).
noun( 'incapacity', '-', mass, _ ).
noun( 'incarceration', 'incarcerations', count, _ ).
noun( 'incarnation', 'incarnations', count, _ ).
noun( 'incendiarism', '-', mass, _ ).
noun( 'incendiary', 'incendiaries', count, _ ).
noun( 'incense', '-', mass, _ ).
noun( 'incentive', 'incentives', both, _ ).
noun( 'inception', 'inceptions', count, _ ).
noun( 'incertitude', '-', mass, _ ).
noun( 'incest', '-', mass, _ ).
noun( 'inch', 'inches', count, _ ).
noun( 'incidence', 'incidences', count, _ ).
noun( 'incident', 'incidents', count, _ ).
noun( 'incineration', '-', mass, _ ).
noun( 'incinerator', 'incinerators', count, _ ).
noun( 'incision', 'incisions', both, _ ).
noun( 'incisor', 'incisors', count, _ ).
noun( 'incitement', 'incitements', both, _ ).
noun( 'incivility', 'incivilities', both, _ ).
noun( 'incl', '-', proper, _ ).
noun( 'inclemency', '-', mass, _ ).
noun( 'inclination', 'inclinations', both, _ ).
noun( 'incline', 'inclines', count, _ ).
noun( 'inclosure', 'inclosures', both, _ ).
noun( 'inclusion', '-', mass, _ ).
noun( 'incoherence', '-', mass, _ ).
noun( 'income', 'incomes', count, _ ).
noun( 'income-tax', 'income-taxes', both, _ ).
noun( 'incompatibility', 'incompatibilities', both, _ ).
noun( 'incompetence', '-', mass, _ ).
noun( 'incompetency', '-', mass, _ ).
noun( 'incomprehensibility', '-', mass, _ ).
noun( 'incomprehension', '-', mass, _ ).
noun( 'incongruity', 'incongruities', both, _ ).
noun( 'inconsistency', 'inconsistencies', both, _ ).
noun( 'inconstancy', 'inconstancies', both, _ ).
noun( 'incontinence', '-', mass, _ ).
noun( 'inconvenience', 'inconveniences', both, _ ).
noun( 'inconvertibility', '-', mass, _ ).
noun( 'incorporation', '-', mass, _ ).
noun( 'incorrectness', '-', mass, _ ).
noun( 'incorruptibility', '-', mass, _ ).
noun( 'increase', 'increases', both, _ ).
noun( 'incredibility', '-', mass, _ ).
noun( 'incredulity', '-', mass, _ ).
noun( 'increment', 'increments', both, _ ).
noun( 'incrustation', 'incrustations', both, _ ).
noun( 'incubation', 'incubations', count, _ ).
noun( 'incubator', 'incubators', count, _ ).
noun( 'incubus', 'incubuses', count, _ ).
noun( 'incumbency', 'incumbencies', count, _ ).
noun( 'incumbent', 'incumbents', count, _ ).
noun( 'incurable', 'incurables', count, _ ).
noun( 'incursion', 'incursions', count, _ ).
noun( 'indebtedness', '-', mass, _ ).
noun( 'indecency', 'indecencies', both, _ ).
noun( 'indecision', '-', mass, _ ).
noun( 'indecorum', '-', mass, _ ).
noun( 'indelicacy', 'indelicacies', both, _ ).
noun( 'indemnification', 'indemnifications', both, _ ).
noun( 'indemnity', 'indemnities', both, _ ).
noun( 'indent', 'indents', count, _ ).
noun( 'indentation', 'indentations', both, _ ).
noun( 'indenture', 'indentures', count, _ ).
noun( 'independence', '-', mass, _ ).
noun( 'independent', 'independents', count, _ ).
noun( 'indestructibility', '-', mass, _ ).
noun( 'indeterminacy', '-', mass, _ ).
noun( 'index', 'indexes', count, _ ).
noun( 'indexer', 'indexers', count, _ ).
noun( 'indication', 'indications', both, _ ).
noun( 'indicator', 'indicators', count, _ ).
noun( 'indictment', 'indictments', both, _ ).
noun( 'indifference', '-', mass, _ ).
noun( 'indigence', '-', mass, _ ).
noun( 'indigestion', '-', mass, _ ).
noun( 'indignation', '-', mass, _ ).
noun( 'indignity', 'indignities', both, _ ).
noun( 'indigo', '-', mass, _ ).
noun( 'indirectness', '-', mass, _ ).
noun( 'indiscipline', '-', mass, _ ).
noun( 'indiscretion', 'indiscretions', both, _ ).
noun( 'indispensability', '-', mass, _ ).
noun( 'indisposition', 'indispositions', both, _ ).
noun( 'indistinctness', '-', mass, _ ).
noun( 'individual', 'individuals', count, _ ).
noun( 'individualism', '-', mass, _ ).
noun( 'individualist', 'individualists', count, _ ).
noun( 'individuality', 'individualities', both, _ ).
noun( 'indoctrination', '-', mass, _ ).
noun( 'indolence', '-', mass, _ ).
noun( 'inducement', 'inducements', both, _ ).
noun( 'induction', 'inductions', count, _ ).
noun( 'indulgence', 'indulgences', both, _ ).
noun( 'industrialism', '-', mass, _ ).
noun( 'industrialist', 'industrialists', count, _ ).
noun( 'industrialization', '-', mass, _ ).
noun( 'industry', 'industries', both, _ ).
noun( 'inebriate', 'inebriates', count, _ ).
noun( 'inebriation', 'inebriations', count, _ ).
noun( 'inebriety', '-', mass, _ ).
noun( 'ineffectiveness', '-', mass, _ ).
noun( 'ineffectuality', '-', mass, _ ).
noun( 'inefficiency', 'inefficiencies', both, _ ).
noun( 'inelegance', 'inelegances', count, _ ).
noun( 'ineligibility', '-', mass, _ ).
noun( 'ineptitude', 'ineptitudes', both, _ ).
noun( 'inequality', 'inequalities', both, _ ).
noun( 'inequity', 'inequities', both, _ ).
noun( 'inertia', '-', mass, _ ).
noun( 'inessential', 'inessentials', count, _ ).
noun( 'inevitability', '-', mass, _ ).
noun( 'inexactitude', 'inexactitudes', both, _ ).
noun( 'inexpediency', '-', mass, _ ).
noun( 'inexperience', '-', mass, _ ).
noun( 'infallibility', '-', mass, _ ).
noun( 'infamy', 'infamies', both, _ ).
noun( 'infancy', 'infancies', both, _ ).
noun( 'infant', 'infants', count, _ ).
noun( 'infanticide', '-', mass, _ ).
noun( 'infantilism', '-', mass, _ ).
noun( 'infantry', 'infantries', both, _ ).
noun( 'infantryman', 'infantrymen', count, _ ).
noun( 'infatuation', 'infatuations', both, _ ).
noun( 'infection', 'infections', both, _ ).
noun( 'inference', 'inferences', both, _ ).
noun( 'inferior', 'inferiors', count, _ ).
noun( 'inferiority', '-', mass, _ ).
noun( 'inferno', 'infernos', count, _ ).
noun( 'infertility', '-', mass, _ ).
noun( 'infestation', 'infestations', both, _ ).
noun( 'infidel', 'infidels', count, _ ).
noun( 'infidelity', 'infidelities', both, _ ).
noun( 'infield', '-', mass, _ ).
noun( 'infiltration', '-', mass, _ ).
noun( 'infinitive', 'infinitives', count, _ ).
noun( 'infinitude', 'infinitudes', both, _ ).
noun( 'infinity', '-', mass, _ ).
noun( 'infirmary', 'infirmaries', count, _ ).
noun( 'infirmity', 'infirmities', both, _ ).
noun( 'inflammation', 'inflammations', both, _ ).
noun( 'inflation', '-', mass, _ ).
noun( 'inflection', 'inflections', both, _ ).
noun( 'inflexibility', '-', mass, _ ).
noun( 'inflexion', 'inflexions', both, _ ).
noun( 'infliction', 'inflictions', both, _ ).
noun( 'inflorescence', '-', mass, _ ).
noun( 'inflow', 'inflows', both, _ ).
noun( 'influence', 'influences', both, _ ).
noun( 'influenza', '-', mass, _ ).
noun( 'influx', 'influxes', both, _ ).
noun( 'info', '-', mass, _ ).
noun( 'informality', 'informalities', both, _ ).
noun( 'informant', 'informants', count, _ ).
noun( 'information', '-', mass, _ ).
noun( 'informer', 'informers', count, _ ).
noun( 'infraction', 'infractions', both, _ ).
noun( 'infrastructure', 'infrastructures', count, _ ).
noun( 'infrequency', '-', mass, _ ).
noun( 'infringement', 'infringements', both, _ ).
noun( 'infusion', 'infusions', both, _ ).
noun( 'ing_enue', 'ing_enues', count, _ ).
noun( 'ingathering', 'ingatherings', count, _ ).
noun( 'ingenuity', '-', mass, _ ).
noun( 'ingenuousness', '-', mass, _ ).
noun( 'ingle-nook', 'ingle-nooks', count, _ ).
noun( 'ingot', 'ingots', count, _ ).
noun( 'ingratitude', '-', mass, _ ).
noun( 'ingredient', 'ingredients', count, _ ).
noun( 'ingress', '-', mass, _ ).
noun( 'inhabitant', 'inhabitants', count, _ ).
noun( 'inhalation', 'inhalations', count, _ ).
noun( 'inhaler', 'inhalers', count, _ ).
noun( 'inheritance', 'inheritances', both, _ ).
noun( 'inheritor', 'inheritors', count, _ ).
noun( 'inhibition', 'inhibitions', both, _ ).
noun( 'inhibitor', 'inhibitors', count, _ ).
noun( 'inhumanity', 'inhumanities', both, _ ).
noun( 'iniquity', 'iniquities', both, _ ).
noun( 'initial', 'initials', count, _ ).
noun( 'initiate', 'initiates', count, _ ).
noun( 'initiation', '-', mass, _ ).
noun( 'initiative', 'initiatives', both, _ ).
noun( 'injection', 'injections', both, _ ).
noun( 'injunction', 'injunctions', count, _ ).
noun( 'injury', 'injuries', both, _ ).
noun( 'injustice', 'injustices', both, _ ).
noun( 'ink', 'inks', both, _ ).
noun( 'ink-bottle', 'ink-bottles', count, _ ).
noun( 'ink-pad', 'ink-pads', count, _ ).
noun( 'ink-pot', 'ink-pots', count, _ ).
noun( 'inkling', 'inklings', count, _ ).
noun( 'inkstand', 'inkstands', count, _ ).
noun( 'inkwell', 'inkwells', count, _ ).
noun( 'inlay', 'inlays', both, _ ).
noun( 'inlet', 'inlets', count, _ ).
noun( 'inmate', 'inmates', count, _ ).
noun( 'inn', 'inns', count, _ ).
noun( 'inning', 'innings', count, _ ).
noun( 'innings', 'innings', count, _ ).
noun( 'innkeeper', 'innkeepers', count, _ ).
noun( 'innocence', '-', mass, _ ).
noun( 'innocent', 'innocents', count, _ ).
noun( 'innovation', 'innovations', both, _ ).
noun( 'innovator', 'innovators', count, _ ).
noun( 'innuendo', 'innuendoes', count, _ ).
noun( 'inoculation', 'inoculations', both, _ ).
noun( 'inpouring', 'inpourings', count, _ ).
noun( 'input', 'inputs', both, _ ).
noun( 'inquest', 'inquests', count, _ ).
noun( 'inquietude', '-', mass, _ ).
noun( 'inquirer', 'inquirers', count, _ ).
noun( 'inquiry', 'inquiries', both, _ ).
noun( 'inquisition', 'inquisitions', both, _ ).
noun( 'inquisitiveness', '-', mass, _ ).
noun( 'inquisitor', 'inquisitors', count, _ ).
noun( 'inroad', 'inroads', count, _ ).
noun( 'inrush', 'inrushes', count, _ ).
noun( '-', 'ins', count, _ ).
noun( 'insanity', '-', mass, _ ).
noun( 'inscription', 'inscriptions', count, _ ).
noun( 'insect', 'insects', count, _ ).
noun( 'insect-powder', 'insect-powders', count, _ ).
noun( 'insecticide', 'insecticides', count, _ ).
noun( 'insecurity', '-', mass, _ ).
noun( 'insemination', '-', mass, _ ).
noun( 'insensibility', '-', mass, _ ).
noun( 'insensitivity', '-', mass, _ ).
noun( 'insert', 'inserts', count, _ ).
noun( 'insertion', 'insertions', both, _ ).
noun( 'inset', 'insets', count, _ ).
noun( 'inside', 'insides', count, _ ).
noun( 'insider', 'insiders', count, _ ).
noun( 'insidiousness', '-', mass, _ ).
noun( 'insight', 'insights', both, _ ).
noun( 'insignificance', '-', mass, _ ).
noun( 'insincerity', '-', mass, _ ).
noun( 'insinuation', 'insinuations', both, _ ).
noun( 'insipidity', '-', mass, _ ).
noun( 'insipidness', '-', mass, _ ).
noun( 'insistence', '-', mass, _ ).
noun( 'insole', 'insoles', count, _ ).
noun( 'insolence', '-', mass, _ ).
noun( 'insolvency', '-', mass, _ ).
noun( 'insolvent', 'insolvents', count, _ ).
noun( 'insomnia', '-', mass, _ ).
noun( 'insomniac', 'insomniacs', count, _ ).
noun( 'insouciance', '-', mass, _ ).
noun( 'inspection', 'inspections', both, _ ).
noun( 'inspector', 'inspectors', count, _ ).
noun( 'inspectorate', 'inspectorates', count, _ ).
noun( 'inspiration', 'inspirations', both, _ ).
noun( 'instability', 'instabilities', both, _ ).
noun( 'installation', 'installations', both, _ ).
noun( 'instalment', 'instalments', count, _ ).
noun( 'instance', 'instances', count, _ ).
noun( 'instant', 'instants', count, _ ).
noun( 'instep', 'insteps', count, _ ).
noun( 'instigation', 'instigations', count, _ ).
noun( 'instigator', 'instigators', count, _ ).
noun( 'instillation', 'instillations', count, _ ).
noun( 'instinct', 'instincts', both, _ ).
noun( 'institute', 'institutes', count, _ ).
noun( 'institution', 'institutions', both, _ ).
noun( 'instruction', 'instructions', both, _ ).
noun( 'instructor', 'instructors', count, _ ).
noun( 'instructress', 'instructresses', count, _ ).
noun( 'instrument', 'instruments', count, _ ).
noun( 'instrumentalist', 'instrumentalists', count, _ ).
noun( 'instrumentality', '-', mass, _ ).
noun( 'instrumentation', '-', mass, _ ).
noun( 'insubordination', 'insubordinations', both, _ ).
noun( 'insufficiency', '-', mass, _ ).
noun( 'insularism', '-', mass, _ ).
noun( 'insularity', '-', mass, _ ).
noun( 'insulation', '-', mass, _ ).
noun( 'insulator', 'insulators', count, _ ).
noun( 'insulin', '-', mass, _ ).
noun( 'insult', 'insults', both, _ ).
noun( 'insurance', 'insurances', both, _ ).
noun( 'insurgent', 'insurgents', count, _ ).
noun( 'insurrection', 'insurrections', both, _ ).
noun( 'intaglio', 'intaglios', both, _ ).
noun( 'intake', 'intakes', both, _ ).
noun( 'intangibility', '-', mass, _ ).
noun( 'intangibles', 'intangibless', count, _ ).
noun( 'integer', 'integers', count, _ ).
noun( 'integration', '-', mass, _ ).
noun( 'integrity', '-', mass, _ ).
noun( 'integument', 'integuments', count, _ ).
noun( 'intellect', 'intellects', both, _ ).
noun( 'intellectual', 'intellectuals', count, _ ).
noun( 'intelligence', '-', mass, _ ).
noun( 'intelligentsia', 'intelligentsias', count, _ ).
noun( 'intelligibility', '-', mass, _ ).
noun( 'intemperance', '-', mass, _ ).
noun( 'intensification', 'intensifications', both, _ ).
noun( 'intensity', 'intensities', both, _ ).
noun( 'intent', 'intents', both, _ ).
noun( 'intention', 'intentions', both, _ ).
noun( 'intentness', '-', mass, _ ).
noun( 'interaction', 'interactions', count, _ ).
noun( 'interception', 'interceptions', both, _ ).
noun( 'interceptor', 'interceptors', count, _ ).
noun( 'intercession', 'intercessions', both, _ ).
noun( 'interchange', 'interchanges', both, _ ).
noun( 'intercom', 'intercoms', count, _ ).
noun( 'intercommunication', '-', mass, _ ).
noun( 'intercommunion', 'intercommunions', count, _ ).
noun( 'interconnectedness', '-', mass, _ ).
noun( 'interconnection', 'interconnections', both, _ ).
noun( 'intercourse', '-', mass, _ ).
noun( 'interdependence', 'interdependences', count, _ ).
noun( 'interdict', 'interdicts', count, _ ).
noun( 'interdiction', 'interdictions', both, _ ).
noun( 'interest', 'interests', both, _ ).
noun( 'interface', 'interfaces', count, _ ).
noun( 'interference', '-', mass, _ ).
noun( 'interim', '-', mass, _ ).
noun( 'interior', 'interiors', count, _ ).
noun( 'interjection', 'interjections', count, _ ).
noun( 'interlocutor', 'interlocutors', count, _ ).
noun( 'interloper', 'interlopers', count, _ ).
noun( 'interlude', 'interludes', count, _ ).
noun( 'intermarriage', 'intermarriages', both, _ ).
noun( 'intermediary', 'intermediaries', count, _ ).
noun( 'intermediate', 'intermediates', count, _ ).
noun( 'interment', 'interments', both, _ ).
noun( 'intermezzo', 'intermezzos', count, _ ).
noun( 'intermission', 'intermissions', count, _ ).
noun( 'intermixture', 'intermixtures', count, _ ).
noun( 'intern', 'interns', count, _ ).
noun( 'international', 'internationals', count, _ ).
noun( 'internationale', 'internationales', count, _ ).
noun( 'internationalism', '-', mass, _ ).
noun( 'internationalist', 'internationalists', count, _ ).
noun( 'internationalization', 'internationalizations', both, _ ).
noun( 'interne', 'internes', count, _ ).
noun( 'internee', 'internees', count, _ ).
noun( 'internment', '-', mass, _ ).
noun( 'interpellation', 'interpellations', both, _ ).
noun( 'interphone', 'interphones', count, _ ).
noun( 'interplay', '-', mass, _ ).
noun( 'interpolation', 'interpolations', both, _ ).
noun( 'interposition', 'interpositions', both, _ ).
noun( 'interpretation', 'interpretations', both, _ ).
noun( 'interpreter', 'interpreters', count, _ ).
noun( 'interregnum', 'interregnums', count, _ ).
noun( 'interrelation', 'interrelations', both, _ ).
noun( 'interrelationship', 'interrelationships', both, _ ).
noun( 'interrogation', 'interrogations', both, _ ).
noun( 'interrogative', 'interrogatives', count, _ ).
noun( 'interrogator', 'interrogators', count, _ ).
noun( 'interrupter', 'interrupters', count, _ ).
noun( 'interruption', 'interruptions', both, _ ).
noun( 'intersection', 'intersections', both, _ ).
noun( 'interstice', 'interstices', count, _ ).
noun( 'interval', 'intervals', count, _ ).
noun( 'intervention', 'interventions', both, _ ).
noun( 'interview', 'interviews', count, _ ).
noun( 'interviewer', 'interviewers', count, _ ).
noun( 'intestine', 'intestines', count, _ ).
noun( 'intimacy', 'intimacies', both, _ ).
noun( 'intimate', 'intimates', count, _ ).
noun( 'intimation', 'intimations', both, _ ).
noun( 'intimidation', '-', mass, _ ).
noun( 'intolerance', '-', mass, _ ).
noun( 'intonation', '-', mass, _ ).
noun( 'intoxicant', 'intoxicants', count, _ ).
noun( 'intoxication', '-', mass, _ ).
noun( 'intractability', '-', mass, _ ).
noun( 'intransigence', '-', mass, _ ).
noun( 'intrepidity', 'intrepidities', both, _ ).
noun( 'intricacy', 'intricacies', both, _ ).
noun( 'intrigue', 'intrigues', both, _ ).
noun( 'intro', '-', count, _ ).
noun( 'introduction', 'introductions', both, _ ).
noun( 'introspection', '-', mass, _ ).
noun( 'introversion', '-', mass, _ ).
noun( 'introvert', 'introverts', count, _ ).
noun( 'intruder', 'intruders', count, _ ).
noun( 'intrusion', 'intrusions', both, _ ).
noun( 'intuition', 'intuitions', both, _ ).
noun( 'intumescence', '-', mass, _ ).
noun( 'inundation', 'inundations', both, _ ).
noun( 'invader', 'invaders', count, _ ).
noun( 'invalid', 'invalids', count, _ ).
noun( 'invalidation', 'invalidations', both, _ ).
noun( 'invalidism', '-', mass, _ ).
noun( 'invalidity', '-', mass, _ ).
noun( 'invasion', 'invasions', both, _ ).
noun( 'invective', '-', mass, _ ).
noun( 'invention', 'inventions', both, _ ).
noun( 'inventor', 'inventors', count, _ ).
noun( 'inventory', 'inventories', count, _ ).
noun( 'inverse', '-', mass, _ ).
noun( 'inversion', 'inversions', both, _ ).
noun( 'invertebrate', 'invertebrates', count, _ ).
noun( 'investigation', 'investigations', both, _ ).
noun( 'investigator', 'investigators', count, _ ).
noun( 'investiture', 'investitures', count, _ ).
noun( 'investment', 'investments', both, _ ).
noun( 'investor', 'investors', count, _ ).
noun( 'invigilation', 'invigilations', both, _ ).
noun( 'invigilator', 'invigilators', count, _ ).
noun( 'invincibility', '-', mass, _ ).
noun( 'invisibility', '-', mass, _ ).
noun( 'invitation', 'invitations', both, _ ).
noun( 'invite', 'invites', count, _ ).
noun( 'invocation', 'invocations', both, _ ).
noun( 'invoice', 'invoices', count, _ ).
noun( 'involution', 'involutions', count, _ ).
noun( 'involvement', 'involvements', count, _ ).
noun( 'inwardness', '-', mass, _ ).
noun( 'iodine', '-', mass, _ ).
noun( 'ion', 'ions', count, _ ).
noun( 'ionization', '-', mass, _ ).
noun( 'ionosphere', 'ionospheres', count, _ ).
noun( 'iota', 'iotas', count, _ ).
noun( 'ipse dixit', 'ipse dixits', count, _ ).
noun( 'irascibility', '-', mass, _ ).
noun( 'ire', '-', mass, _ ).
noun( 'iridescence', '-', mass, _ ).
noun( 'iridium', '-', mass, _ ).
noun( 'iris', 'irises', count, _ ).
noun( 'iron', 'irons', both, _ ).
noun( 'iron-foundry', 'iron-foundries', count, _ ).
noun( 'iron-grey', '-', mass, _ ).
noun( 'ironing-board', 'ironing-boards', count, _ ).
noun( 'ironmonger', 'ironmongers', count, _ ).
noun( 'ironmongery', 'ironmongeries', both, _ ).
noun( 'ironmould', 'ironmoulds', count, _ ).
noun( 'ironside', 'ironsides', count, _ ).
noun( 'ironware', '-', mass, _ ).
noun( 'ironwork', 'ironworks', count, _ ).
noun( 'ironworks', 'ironworks', count, _ ).
noun( 'irony', 'ironies', both, _ ).
noun( 'irredentist', 'irredentists', count, _ ).
noun( 'irregular', 'irregulars', count, _ ).
noun( 'irregularity', 'irregularities', both, _ ).
noun( 'irrelevance', 'irrelevances', both, _ ).
noun( 'irrelevancy', 'irrelevancies', both, _ ).
noun( 'irresolution', '-', mass, _ ).
noun( 'irresponsibility', '-', mass, _ ).
noun( 'irreverence', '-', mass, _ ).
noun( 'irreversibility', '-', mass, _ ).
noun( 'irridentism', '-', mass, _ ).
noun( 'irrigation', '-', mass, _ ).
noun( 'irritability', '-', mass, _ ).
noun( 'irritant', 'irritants', count, _ ).
noun( 'irritation', 'irritations', both, _ ).
noun( 'irruption', 'irruptions', count, _ ).
noun( 'isinglass', '-', mass, _ ).
noun( 'island', 'islands', count, _ ).
noun( 'islander', 'islanders', count, _ ).
noun( 'isle', 'isles', count, _ ).
noun( 'islet', 'islets', count, _ ).
noun( 'ism', 'isms', count, _ ).
noun( 'isobar', 'isobars', count, _ ).
noun( 'isolation', '-', mass, _ ).
noun( 'isolationism', '-', mass, _ ).
noun( 'isolationist', 'isolationists', count, _ ).
noun( 'isotherm', 'isotherms', count, _ ).
noun( 'isotope', 'isotopes', count, _ ).
noun( 'issue', 'issues', both, _ ).
noun( 'isthmus', 'isthmuses', count, _ ).
noun( 'itch', 'itches', count, _ ).
noun( 'item', 'items', count, _ ).
noun( 'iteration', 'iterations', both, _ ).
noun( 'itinerary', 'itineraries', count, _ ).
noun( 'ivory', '-', mass, _ ).
noun( 'ivy', 'ivies', both, _ ).
noun( 'j', '-', count, _ ).
noun( 'jab', 'jabs', count, _ ).
noun( 'jabber', '-', mass, _ ).
noun( 'jabberer', 'jabberers', count, _ ).
noun( 'jabot', 'jabots', count, _ ).
noun( 'jack', 'jacks', count, _ ).
noun( 'jack-in-the-box', 'jack-in-the-boxes', count, _ ).
noun( 'jack-knife', 'jack-knives', count, _ ).
noun( 'jack-o\'-lantern', 'jack-o\'-lanterns', count, _ ).
noun( 'jack-plane', 'jack-planes', count, _ ).
noun( 'jackal', 'jackals', count, _ ).
noun( 'jackanapes', '-', count, _ ).
noun( 'jackass', 'jackasses', count, _ ).
noun( 'jackboot', 'jackboots', count, _ ).
noun( 'jackdaw', 'jackdaws', count, _ ).
noun( 'jacket', 'jackets', count, _ ).
noun( 'jackpot', 'jackpots', count, _ ).
noun( 'jade', 'jades', both, _ ).
noun( 'jag', 'jags', count, _ ).
noun( 'jaguar', 'jaguars', count, _ ).
noun( 'jail', 'jails', both, _ ).
noun( 'jailer', 'jailers', count, _ ).
noun( 'jailor', 'jailors', count, _ ).
noun( 'jakes', '-', count, _ ).
noun( 'jalopy', 'jalopies', count, _ ).
noun( 'jam', 'jams', both, _ ).
noun( 'jamb', 'jambs', count, _ ).
noun( 'jamboree', 'jamborees', count, _ ).
noun( 'jamjar', 'jamjars', count, _ ).
noun( 'jampot', 'jampots', count, _ ).
noun( 'jangle', '-', mass, _ ).
noun( 'janitor', 'janitors', count, _ ).
noun( 'jape', 'japes', count, _ ).
noun( 'japonica', '-', mass, _ ).
noun( 'jar', 'jars', count, _ ).
noun( 'jarful', 'jarfuls', count, _ ).
noun( 'jargon', '-', mass, _ ).
noun( 'jasmine', '-', mass, _ ).
noun( 'jasper', '-', mass, _ ).
noun( 'jaundice', '-', mass, _ ).
noun( 'jaunt', 'jaunts', count, _ ).
noun( 'jauntiness', '-', mass, _ ).
noun( 'jaunting-car', 'jaunting-cars', count, _ ).
noun( 'javelin', 'javelins', count, _ ).
noun( 'jaw', 'jaws', count, _ ).
noun( 'jawbone', 'jawbones', count, _ ).
noun( 'jawbreaker', 'jawbreakers', count, _ ).
noun( 'jay', 'jays', count, _ ).
noun( 'jaywalker', 'jaywalkers', count, _ ).
noun( 'jazz', '-', mass, _ ).
noun( 'jealousy', 'jealousies', both, _ ).
noun( 'jean', '-', mass, _ ).
noun( 'jeep', 'jeeps', count, _ ).
noun( 'jeer', 'jeers', count, _ ).
noun( 'jejuneness', '-', mass, _ ).
noun( 'jellaba', 'jellabas', count, _ ).
noun( 'jelly', 'jellies', both, _ ).
noun( 'jellyfish', 'jellyfish', count, _ ).
noun( 'jemmy', 'jemmies', count, _ ).
noun( 'jenny', 'jennies', count, _ ).
noun( 'jeopardy', '-', mass, _ ).
noun( 'jerboa', 'jerboas', count, _ ).
noun( 'jeremiad', 'jeremiads', count, _ ).
noun( 'jerk', 'jerks', count, _ ).
noun( 'jerkin', 'jerkins', count, _ ).
noun( 'jerkiness', '-', mass, _ ).
noun( 'jerry', 'jerries', count, _ ).
noun( 'jerry-builder', 'jerry-builders', count, _ ).
noun( 'jerry-building', '-', mass, _ ).
noun( 'jersey', 'jerseys', both, _ ).
noun( 'jest', 'jests', count, _ ).
noun( 'jester', 'jesters', count, _ ).
noun( 'jet', 'jets', both, _ ).
noun( 'jetsam', '-', mass, _ ).
noun( 'jetty', 'jetties', count, _ ).
noun( 'jewel', 'jewels', count, _ ).
noun( 'jeweller', 'jewellers', count, _ ).
noun( 'jewellery', '-', mass, _ ).
noun( 'jewelry', '-', mass, _ ).
noun( 'jib', 'jibs', count, _ ).
noun( 'jib-boom', 'jib-booms', count, _ ).
noun( 'jibe', 'jibes', count, _ ).
noun( 'jiffy', 'jiffies', count, _ ).
noun( 'jig', 'jigs', count, _ ).
noun( 'jigger', 'jiggers', count, _ ).
noun( 'jiggery-pokery', '-', mass, _ ).
noun( 'jiggle', 'jiggles', count, _ ).
noun( 'jigsaw', 'jigsaws', count, _ ).
noun( 'jihad', 'jihads', count, _ ).
noun( 'jilt', 'jilts', count, _ ).
noun( 'jimmy', 'jimmies', count, _ ).
noun( 'jingle', 'jingles', count, _ ).
noun( 'jingo', 'jingoes', count, _ ).
noun( 'jingoism', '-', mass, _ ).
noun( 'jingoist', 'jingoists', count, _ ).
noun( 'jinks', '-', mass, _ ).
noun( 'jinn', 'jinns', count, _ ).
noun( 'jinx', 'jinxes', count, _ ).
noun( 'jitney', 'jitneys', count, _ ).
noun( 'jitterbug', 'jitterbugs', count, _ ).
noun( 'jive', 'jives', count, _ ).
noun( 'job', 'jobs', count, _ ).
noun( 'jobber', 'jobbers', count, _ ).
noun( 'jobbery', '-', mass, _ ).
noun( 'jockey', 'jockeys', count, _ ).
noun( 'jocoseness', '-', mass, _ ).
noun( 'jocosity', '-', mass, _ ).
noun( 'jocularity', 'jocularities', both, _ ).
noun( 'jocundity', 'jocundities', both, _ ).
noun( 'jog', 'jogs', count, _ ).
noun( 'jog-trot', 'jog-trots', count, _ ).
noun( 'jogger', 'joggers', count, _ ).
noun( 'jogging', '-', mass, _ ).
noun( 'joggle', 'joggles', count, _ ).
noun( 'john', '-', count, _ ).
noun( 'joie de vivre', '-', mass, _ ).
noun( 'join', 'joins', count, _ ).
noun( 'joiner', 'joiners', count, _ ).
noun( 'joinery', '-', mass, _ ).
noun( 'joint', 'joints', count, _ ).
noun( 'jointure', 'jointures', count, _ ).
noun( 'joist', 'joists', count, _ ).
noun( 'joke', 'jokes', count, _ ).
noun( 'joker', 'jokers', count, _ ).
noun( 'jollification', 'jollifications', both, _ ).
noun( 'jollity', 'jollities', both, _ ).
noun( 'jollyboat', 'jollyboats', count, _ ).
noun( 'jolt', 'jolts', count, _ ).
noun( 'jonquil', 'jonquils', count, _ ).
noun( 'joss', 'josses', count, _ ).
noun( 'joss-house', 'joss-houses', count, _ ).
noun( 'joss-stick', 'joss-sticks', count, _ ).
noun( 'jot', 'jots', count, _ ).
noun( 'jotter', 'jotters', count, _ ).
noun( 'joule', 'joules', count, _ ).
noun( 'journal', 'journals', count, _ ).
noun( 'journalese', '-', mass, _ ).
noun( 'journalism', '-', mass, _ ).
noun( 'journalist', 'journalists', count, _ ).
noun( 'journey', 'journeys', count, _ ).
noun( 'journeyman', 'journeymen', count, _ ).
noun( 'joust', 'jousts', count, _ ).
noun( 'joviality', 'jovialities', both, _ ).
noun( 'jowl', 'jowls', count, _ ).
noun( 'joy', 'joys', both, _ ).
noun( 'joy-ride', 'joy-rides', count, _ ).
noun( 'joy-stick', 'joy-sticks', count, _ ).
noun( 'joyfulness', '-', mass, _ ).
noun( 'joylessness', '-', mass, _ ).
noun( 'joyousness', '-', mass, _ ).
noun( 'ju-jitsu', '-', mass, _ ).
noun( 'jubilation', 'jubilations', both, _ ).
noun( 'jubilee', 'jubilees', count, _ ).
noun( 'judge', 'judges', count, _ ).
noun( 'judgement', 'judgements', both, _ ).
noun( 'judicature', 'judicatures', both, _ ).
noun( 'judiciary', 'judiciaries', count, _ ).
noun( 'judiciousness', '-', mass, _ ).
noun( 'judo', '-', mass, _ ).
noun( 'jug', 'jugs', count, _ ).
noun( 'jugful', 'jugfuls', count, _ ).
noun( 'juggernaut', 'juggernauts', count, _ ).
noun( 'juggler', 'jugglers', count, _ ).
noun( 'juice', 'juices', both, _ ).
noun( 'juiciness', '-', mass, _ ).
noun( 'juju', 'jujus', count, _ ).
noun( 'jujube', 'jujubes', count, _ ).
noun( 'jukebox', 'jukeboxes', count, _ ).
noun( 'julep', 'juleps', count, _ ).
noun( 'jumble', 'jumbles', both, _ ).
noun( 'jumble-sale', 'jumble-sales', count, _ ).
noun( 'jump', 'jumps', count, _ ).
noun( 'jumper', 'jumpers', count, _ ).
noun( 'jumpiness', '-', mass, _ ).
noun( 'junction', 'junctions', both, _ ).
noun( 'juncture', 'junctures', count, _ ).
noun( 'jungle', 'jungles', count, _ ).
noun( 'junior', 'juniors', count, _ ).
noun( 'juniper', 'junipers', count, _ ).
noun( 'junk', 'junks', both, _ ).
noun( 'junk-shop', 'junk-shops', count, _ ).
noun( 'junket', 'junkets', both, _ ).
noun( 'junketing', '-', mass, _ ).
noun( 'junkie', 'junkies', count, _ ).
noun( 'junky', 'junkies', count, _ ).
noun( 'junta', 'juntas', count, _ ).
noun( 'jurisdiction', '-', mass, _ ).
noun( 'jurisprudence', '-', mass, _ ).
noun( 'jurist', 'jurists', count, _ ).
noun( 'juror', 'jurors', count, _ ).
noun( 'jury', 'juries', count, _ ).
noun( 'jury-box', 'jury-boxes', count, _ ).
noun( 'jury-mast', 'jury-masts', count, _ ).
noun( 'juryman', 'jurymen', count, _ ).
noun( 'justice', 'justices', both, _ ).
noun( 'justiciary', 'justiciaries', count, _ ).
noun( 'justification', 'justifications', both, _ ).
noun( 'justness', '-', mass, _ ).
noun( 'jute', '-', mass, _ ).
noun( 'juvenile', 'juveniles', count, _ ).
noun( 'juxtaposition', '-', mass, _ ).
noun( 'k', '-', count, _ ).
noun( 'k\"ummel', '-', mass, _ ).
noun( 'kaffir', 'kaffirs', count, _ ).
noun( 'kail', '-', mass, _ ).
noun( 'kakemono', 'kakemonos', count, _ ).
noun( 'kale', '-', mass, _ ).
noun( 'kaleidoscope', 'kaleidoscopes', count, _ ).
noun( 'kampong', 'kampongs', count, _ ).
noun( 'kangaroo', 'kangaroos', count, _ ).
noun( 'kaolin', '-', mass, _ ).
noun( 'kapok', '-', mass, _ ).
noun( 'karat', 'karats', count, _ ).
noun( 'karate', '-', mass, _ ).
noun( 'karma', 'karmas', count, _ ).
noun( 'kava', '-', mass, _ ).
noun( 'kayak', 'kayaks', count, _ ).
noun( 'kebab', 'kebabs', count, _ ).
noun( 'kedgeree', '-', mass, _ ).
noun( 'keel', 'keels', count, _ ).
noun( 'keen', 'keens', count, _ ).
noun( 'keenness', '-', mass, _ ).
noun( 'keep', 'keeps', both, _ ).
noun( 'keeper', 'keepers', count, _ ).
noun( 'keeping', '-', mass, _ ).
noun( 'keepsake', 'keepsakes', count, _ ).
noun( 'keg', 'kegs', count, _ ).
noun( 'kelp', '-', mass, _ ).
noun( 'ken', '-', mass, _ ).
noun( 'kennel', 'kennels', count, _ ).
noun( 'kepi', 'kepis', count, _ ).
noun( 'kerb', 'kerbs', count, _ ).
noun( 'kerbstone', 'kerbstones', count, _ ).
noun( 'kerchief', 'kerchiefs', count, _ ).
noun( 'kernel', 'kernels', count, _ ).
noun( 'kerosene', '-', mass, _ ).
noun( 'kestrel', 'kestrels', count, _ ).
noun( 'ketch', 'ketches', count, _ ).
noun( 'ketchup', '-', mass, _ ).
noun( 'kettle', 'kettles', count, _ ).
noun( 'kettledrum', 'kettledrums', count, _ ).
noun( 'key', 'keys', count, _ ).
noun( 'keyboard', 'keyboards', count, _ ).
noun( 'keyhole', 'keyholes', count, _ ).
noun( 'keynote', 'keynotes', count, _ ).
noun( 'keyring', 'keyrings', count, _ ).
noun( 'keystone', 'keystones', count, _ ).
noun( 'keyword', 'keywords', count, _ ).
noun( 'kg', 'kg', count, _ ).
noun( 'khaki', '-', mass, _ ).
noun( 'khan', 'khans', count, _ ).
noun( 'kibbutz', 'kibbutzim', count, _ ).
noun( 'kibbutznik', 'kibbutzniks', count, _ ).
noun( 'kick', 'kicks', both, _ ).
noun( 'kick-start', 'kick-starts', count, _ ).
noun( 'kick-starter', 'kick-starters', count, _ ).
noun( 'kickback', 'kickbacks', count, _ ).
noun( 'kickoff', 'kickoffs', count, _ ).
noun( 'kid', 'kids', both, _ ).
noun( 'kiddy', 'kiddies', count, _ ).
noun( 'kidnapper', 'kidnappers', count, _ ).
noun( 'kidney', 'kidneys', count, _ ).
noun( 'kidney-bean', 'kidney-beans', count, _ ).
noun( 'kill', '-', count, _ ).
noun( 'killer', 'killers', count, _ ).
noun( 'killing', 'killings', both, _ ).
noun( 'killjoy', 'killjoys', count, _ ).
noun( 'kiln', 'kilns', count, _ ).
noun( 'kilo', 'kilos', count, _ ).
noun( 'kilocycle', 'kilocycles', count, _ ).
noun( 'kilogram', 'kilograms', count, _ ).
noun( 'kilolitre', 'kilolitres', count, _ ).
noun( 'kilometre', 'kilometres', count, _ ).
noun( 'kilowatt', 'kilowatts', count, _ ).
noun( 'kilt', 'kilts', count, _ ).
noun( 'kimono', 'kimonos', count, _ ).
noun( 'kind', 'kinds', both, _ ).
noun( 'kindergarten', 'kindergartens', count, _ ).
noun( 'kindling', '-', mass, _ ).
noun( 'kindness', 'kindnesses', both, _ ).
noun( 'kindred', '-', mass, _ ).
noun( 'kinetics', 'kinetics', mass, _ ).
noun( 'king', 'kings', count, _ ).
noun( 'kingcup', 'kingcups', count, _ ).
noun( 'kingdom', 'kingdoms', count, _ ).
noun( 'kingfisher', 'kingfishers', count, _ ).
noun( 'kingpin', 'kingpins', count, _ ).
noun( 'kingship', '-', mass, _ ).
noun( 'kink', 'kinks', count, _ ).
noun( 'kinship', '-', mass, _ ).
noun( 'kinsman', 'kinsmen', count, _ ).
noun( 'kinswoman', 'kinswomen', count, _ ).
noun( 'kiosk', 'kiosks', count, _ ).
noun( 'kip', 'kips', count, _ ).
noun( 'kipper', 'kippers', count, _ ).
noun( 'kirk', 'kirks', count, _ ).
noun( 'kirsch', '-', mass, _ ).
noun( 'kirtle', 'kirtles', count, _ ).
noun( 'kismet', '-', mass, _ ).
noun( 'kiss', 'kisses', count, _ ).
noun( 'kisser', 'kissers', count, _ ).
noun( 'kit', 'kits', both, _ ).
noun( 'kitbag', 'kitbags', count, _ ).
noun( 'kitchen', 'kitchens', count, _ ).
noun( 'kitchenette', 'kitchenettes', count, _ ).
noun( 'kite', 'kites', count, _ ).
noun( 'kite-balloon', 'kite-balloons', count, _ ).
noun( 'kitten', 'kittens', count, _ ).
noun( 'kitty', 'kitties', count, _ ).
noun( 'kiwi', 'kiwis', count, _ ).
noun( 'klaxon', 'klaxons', count, _ ).
noun( 'kleenex', 'kleenexes', both, _ ).
noun( 'kleptomania', '-', mass, _ ).
noun( 'kleptomaniac', 'kleptomaniacs', count, _ ).
noun( 'km', 'km', count, _ ).
noun( 'knack', 'knacks', count, _ ).
noun( 'knacker', 'knackers', count, _ ).
noun( 'knapsack', 'knapsacks', count, _ ).
noun( 'knave', 'knaves', count, _ ).
noun( 'knavery', 'knaveries', both, _ ).
noun( 'knee', 'knees', count, _ ).
noun( 'kneecap', 'kneecaps', count, _ ).
noun( 'knell', '-', count, _ ).
noun( 'knick-knack', 'knick-knacks', count, _ ).
noun( 'knife', 'knives', count, _ ).
noun( 'knife-edge', 'knife-edges', count, _ ).
noun( 'knight', 'knights', count, _ ).
noun( 'knight-errant', 'knights-errant', count, _ ).
noun( 'knighthood', 'knighthoods', both, _ ).
noun( 'knitter', 'knitters', count, _ ).
noun( 'knitting', '-', mass, _ ).
noun( 'knitting-machine', 'knitting-machines', count, _ ).
noun( 'knitting-needle', 'knitting-needles', count, _ ).
noun( 'knitwear', '-', mass, _ ).
noun( 'knob', 'knobs', count, _ ).
noun( 'knobble', 'knobbles', count, _ ).
noun( 'knobkerrie', 'knobkerries', count, _ ).
noun( 'knock', 'knocks', count, _ ).
noun( 'knock-on', 'knock-ons', count, _ ).
noun( 'knocker', 'knockers', count, _ ).
noun( 'knockout', 'knockouts', count, _ ).
noun( 'knoll', 'knolls', count, _ ).
noun( 'knot', 'knots', count, _ ).
noun( 'knothole', 'knotholes', count, _ ).
noun( 'knout', 'knouts', count, _ ).
noun( 'know', '-', count, _ ).
noun( 'know-all', 'know-alls', count, _ ).
noun( 'know-how', '-', mass, _ ).
noun( 'knowledge', '-', mass, _ ).
noun( 'knuckle', 'knuckles', count, _ ).
noun( 'koala', 'koalas', count, _ ).
noun( 'kobo', 'kobos', count, _ ).
noun( 'kohl', '-', mass, _ ).
noun( 'kohlrabi', 'kohlrabis', both, _ ).
noun( 'kola', 'kolas', count, _ ).
noun( 'kola-nut', 'kola-nuts', count, _ ).
noun( 'kookaburra', 'kookaburras', count, _ ).
noun( 'kopeck', 'kopecks', count, _ ).
noun( 'kopje', 'kopjes', count, _ ).
noun( 'koppie', 'koppies', count, _ ).
noun( 'kosher', '-', mass, _ ).
noun( 'kotow', 'kotows', count, _ ).
noun( 'koumiss', '-', mass, _ ).
noun( 'kowtow', 'kowtows', count, _ ).
noun( 'kraal', 'kraals', count, _ ).
noun( 'krona', 'kronor', count, _ ).
noun( 'krone', 'kroner', count, _ ).
noun( 'kudos', '-', mass, _ ).
noun( 'kumis', '-', mass, _ ).
noun( 'kung fu', '-', mass, _ ).
noun( 'kvass', '-', mass, _ ).
noun( 'kw', 'kw', count, _ ).
noun( 'kwacha', 'kwacha', count, _ ).
noun( 'kwela', '-', mass, _ ).
noun( 'l', '-', count, _ ).
noun( 'la', '-', count, _ ).
noun( 'laager', 'laagers', count, _ ).
noun( 'lab', 'labs', count, _ ).
noun( 'label', 'labels', count, _ ).
noun( 'laboratory', 'laboratories', count, _ ).
noun( 'labour', 'labours', both, _ ).
noun( 'labourer', 'labourers', count, _ ).
noun( 'laburnum', 'laburnums', count, _ ).
noun( 'labyrinth', 'labyrinths', count, _ ).
noun( 'lace', 'laces', both, _ ).
noun( 'laceration', 'lacerations', both, _ ).
noun( 'lack', '-', mass, _ ).
noun( 'lackey', 'lackeys', count, _ ).
noun( 'laconicism', 'laconicisms', both, _ ).
noun( 'laconism', 'laconisms', both, _ ).
noun( 'lacquer', 'lacquers', both, _ ).
noun( 'lacrosse', '-', mass, _ ).
noun( 'lacuna', 'lacunas', count, _ ).
noun( 'lad', 'lads', count, _ ).
noun( 'ladder', 'ladders', count, _ ).
noun( 'laddie', 'laddies', count, _ ).
noun( 'lading', '-', mass, _ ).
noun( 'ladle', 'ladles', count, _ ).
noun( 'lady', 'ladies', count, _ ).
noun( 'lady\'s-maid', 'lady\'s-maids', count, _ ).
noun( 'lady-in-waiting', 'ladies-in-waiting', count, _ ).
noun( 'lady-killer', 'lady-killers', count, _ ).
noun( 'ladybird', 'ladybirds', count, _ ).
noun( 'ladyship', 'ladyships', count, _ ).
noun( 'lag', 'lags', count, _ ).
noun( 'lager', 'lagers', both, _ ).
noun( 'laggard', 'laggards', count, _ ).
noun( 'lagging', '-', mass, _ ).
noun( 'lagoon', 'lagoons', count, _ ).
noun( 'lair', 'lairs', count, _ ).
noun( 'laird', 'lairds', count, _ ).
noun( 'laissez-faire', '-', mass, _ ).
noun( 'laity', 'laities', count, _ ).
noun( 'lake', 'lakes', count, _ ).
noun( 'lakh', 'lakhs', count, _ ).
noun( 'lam_e', '-', mass, _ ).
noun( 'lama', 'lamas', count, _ ).
noun( 'lamasery', 'lamaseries', count, _ ).
noun( 'lamb', 'lambs', both, _ ).
noun( 'lambency', '-', mass, _ ).
noun( 'lambkin', 'lambkins', count, _ ).
noun( 'lambskin', 'lambskins', both, _ ).
noun( 'lameness', '-', mass, _ ).
noun( 'lament', 'laments', count, _ ).
noun( 'lamentation', 'lamentations', both, _ ).
noun( 'lamp', 'lamps', count, _ ).
noun( 'lamp-black', '-', mass, _ ).
noun( 'lamplight', '-', mass, _ ).
noun( 'lamplighter', 'lamplighters', count, _ ).
noun( 'lampoon', 'lampoons', count, _ ).
noun( 'lamppost', 'lampposts', count, _ ).
noun( 'lamprey', 'lampreys', count, _ ).
noun( 'lampshade', 'lampshades', count, _ ).
noun( 'lance', 'lances', count, _ ).
noun( 'lance-corporal', 'lance-corporals', count, _ ).
noun( 'lancer', 'lancers', count, _ ).
noun( 'lancet', 'lancets', count, _ ).
noun( 'land', 'lands', both, _ ).
noun( 'land-agent', 'land-agents', count, _ ).
noun( 'landau', 'landaus', count, _ ).
noun( 'landfall', 'landfalls', count, _ ).
noun( 'landgrave', 'landgraves', count, _ ).
noun( 'landholder', 'landholders', count, _ ).
noun( 'landing', 'landings', count, _ ).
noun( 'landing-craft', 'landing-craft', count, _ ).
noun( 'landing-field', 'landing-fields', count, _ ).
noun( 'landing-gear', '-', mass, _ ).
noun( 'landing-net', 'landing-nets', count, _ ).
noun( 'landing-party', 'landing-parties', count, _ ).
noun( 'landing-place', 'landing-places', count, _ ).
noun( 'landing-stage', 'landing-stages', count, _ ).
noun( 'landing-strip', 'landing-strips', count, _ ).
noun( 'landlady', 'landladies', count, _ ).
noun( 'landlord', 'landlords', count, _ ).
noun( 'landlubber', 'landlubbers', count, _ ).
noun( 'landmark', 'landmarks', count, _ ).
noun( 'landmine', 'landmines', count, _ ).
noun( 'landowner', 'landowners', count, _ ).
noun( 'landscape', 'landscapes', both, _ ).
noun( 'landslide', 'landslides', count, _ ).
noun( 'landslip', 'landslips', count, _ ).
noun( 'landsman', 'landsmen', count, _ ).
noun( 'lane', 'lanes', count, _ ).
noun( 'langsyne', '-', mass, _ ).
noun( 'language', 'languages', both, _ ).
noun( 'languor', '-', mass, _ ).
noun( 'langur', 'langurs', count, _ ).
noun( 'lanolin', '-', mass, _ ).
noun( 'lantern', 'lanterns', count, _ ).
noun( 'lanyard', 'lanyards', count, _ ).
noun( 'lap', 'laps', both, _ ).
noun( 'lap-dog', 'lap-dogs', count, _ ).
noun( 'lapel', 'lapels', count, _ ).
noun( 'lapidary', 'lapidaries', count, _ ).
noun( 'lapis lazuli', 'lapis lazulis', both, _ ).
noun( 'lapse', 'lapses', count, _ ).
noun( 'lapwing', 'lapwings', count, _ ).
noun( 'larboard', '-', mass, _ ).
noun( 'larceny', 'larcenies', both, _ ).
noun( 'larch', 'larches', both, _ ).
noun( 'lard', '-', mass, _ ).
noun( 'larder', 'larders', count, _ ).
noun( 'large', '-', mass, _ ).
noun( 'largeness', '-', mass, _ ).
noun( 'largesse', '-', mass, _ ).
noun( 'largo', 'largos', count, _ ).
noun( 'lariat', 'lariats', count, _ ).
noun( 'lark', 'larks', count, _ ).
noun( 'larkspur', 'larkspurs', count, _ ).
noun( 'larva', 'larvae', count, _ ).
noun( 'laryngitis', '-', mass, _ ).
noun( 'larynx', 'larynxes', count, _ ).
noun( 'lascar', 'lascars', count, _ ).
noun( 'lasciviousness', '-', mass, _ ).
noun( 'laser', 'lasers', count, _ ).
noun( 'lash', 'lashes', count, _ ).
noun( 'lash-up', 'lash-ups', count, _ ).
noun( 'lashing', 'lashings', count, _ ).
noun( 'lass', 'lasses', count, _ ).
noun( 'lassie', 'lassies', count, _ ).
noun( 'lassitude', '-', mass, _ ).
noun( 'lasso', 'lassos', count, _ ).
noun( 'last', 'lasts', count, _ ).
noun( 'lat', '-', count, _ ).
noun( 'latch', 'latches', count, _ ).
noun( 'latchkey', 'latchkeys', count, _ ).
noun( 'laterite', '-', mass, _ ).
noun( 'latex', '-', mass, _ ).
noun( 'lath', 'laths', count, _ ).
noun( 'lathe', 'lathes', count, _ ).
noun( 'lather', '-', mass, _ ).
noun( 'lathi', 'lathis', count, _ ).
noun( 'latitude', 'latitudes', both, _ ).
noun( 'latitudes', 'latitudess', both, _ ).
noun( 'latitudinarian', 'latitudinarians', count, _ ).
noun( 'latrine', 'latrines', count, _ ).
noun( 'lattice', 'lattices', count, _ ).
noun( 'laudanum', '-', mass, _ ).
noun( 'laugh', 'laughs', count, _ ).
noun( 'laughing-gas', '-', mass, _ ).
noun( 'laughing-stock', 'laughing-stocks', count, _ ).
noun( 'laughter', '-', mass, _ ).
noun( 'launch', 'launches', count, _ ).
noun( 'launcher', 'launchers', count, _ ).
noun( 'launching-pad', 'launching-pads', count, _ ).
noun( 'launching-site', 'launching-sites', count, _ ).
noun( 'launderette', 'launderettes', count, _ ).
noun( 'laundress', 'laundresses', count, _ ).
noun( 'laundry', 'laundries', count, _ ).
noun( 'laundryman', 'laundrymen', count, _ ).
noun( 'laundrywoman', 'laundrywomen', count, _ ).
noun( 'laureate', 'laureates', count, _ ).
noun( 'laurel', 'laurels', count, _ ).
noun( 'lav', 'lavs', count, _ ).
noun( 'lava', '-', mass, _ ).
noun( 'lavatory', 'lavatories', count, _ ).
noun( 'lavender', '-', mass, _ ).
noun( 'law', 'laws', both, _ ).
noun( 'law-officer', 'law-officers', count, _ ).
noun( 'lawbreaker', 'lawbreakers', count, _ ).
noun( 'lawgiver', 'lawgivers', count, _ ).
noun( 'lawlessness', '-', mass, _ ).
noun( 'lawmaker', 'lawmakers', count, _ ).
noun( 'lawmaking', '-', mass, _ ).
noun( 'lawn', 'lawns', both, _ ).
noun( 'lawn-mower', 'lawn-mowers', count, _ ).
noun( 'lawyer', 'lawyers', count, _ ).
noun( 'laxative', 'laxatives', both, _ ).
noun( 'laxity', 'laxities', both, _ ).
noun( 'lay', 'lays', count, _ ).
noun( 'lay figure', 'lay figures', count, _ ).
noun( 'lay-off', 'lay-offs', count, _ ).
noun( 'lay-out', 'lay-outs', count, _ ).
noun( 'layabout', 'layabouts', count, _ ).
noun( 'layby', 'laybys', count, _ ).
noun( 'layer', 'layers', count, _ ).
noun( 'layer-cake', 'layer-cakes', count, _ ).
noun( 'layette', 'layettes', count, _ ).
noun( 'layman', 'laymen', count, _ ).
noun( 'layover', 'layovers', count, _ ).
noun( 'lazar', 'lazars', count, _ ).
noun( 'lazaret', 'lazarets', count, _ ).
noun( 'lazarette', 'lazarettes', count, _ ).
noun( 'lazaretto', 'lazarettos', count, _ ).
noun( 'laziness', '-', mass, _ ).
noun( 'lazybones', '-', count, _ ).
noun( 'lb', 'lb', count, _ ).
noun( '-', 'lbs', count, _ ).
noun( 'lbw', '-', proper, _ ).
noun( 'lea', 'leas', count, _ ).
noun( 'lead', 'leads', both, _ ).
noun( 'lead', 'leads', both, _ ).
noun( 'lead-in', 'lead-ins', count, _ ).
noun( 'lead-ore', '-', mass, _ ).
noun( 'leader', 'leaders', count, _ ).
noun( 'leadership', '-', mass, _ ).
noun( 'leading', '-', mass, _ ).
noun( 'leading', '-', mass, _ ).
noun( 'leading-rein', 'leading-reins', count, _ ).
noun( 'leaf', 'leaves', count, _ ).
noun( 'leaf-bud', 'leaf-buds', count, _ ).
noun( 'leaf-mould', '-', mass, _ ).
noun( 'leaflet', 'leaflets', count, _ ).
noun( 'league', 'leagues', count, _ ).
noun( 'leak', 'leaks', count, _ ).
noun( 'leakage', 'leakages', both, _ ).
noun( 'lean', '-', mass, _ ).
noun( 'lean-to', 'lean-tos', count, _ ).
noun( 'leaning', 'leanings', count, _ ).
noun( 'leanness', '-', mass, _ ).
noun( 'leap', 'leaps', count, _ ).
noun( 'leap-year', 'leap-years', count, _ ).
noun( 'leapfrog', '-', mass, _ ).
noun( 'learner', 'learners', count, _ ).
noun( 'learning', '-', mass, _ ).
noun( 'lease', 'leases', count, _ ).
noun( 'lease-lend', '-', mass, _ ).
noun( 'leasehold', 'leaseholds', count, _ ).
noun( 'leaseholder', 'leaseholders', count, _ ).
noun( 'leash', 'leashes', count, _ ).
noun( 'least', '-', count, _ ).
noun( 'leather', 'leathers', both, _ ).
noun( 'leather-jacket', 'leather-jackets', count, _ ).
noun( 'leatherette', '-', mass, _ ).
noun( 'leatherneck', 'leathernecks', count, _ ).
noun( 'leatherwork', '-', mass, _ ).
noun( 'leave', 'leaves', both, _ ).
noun( 'leave-taking', 'leave-takings', count, _ ).
noun( 'leaven', '-', mass, _ ).
noun( 'lecher', 'lechers', count, _ ).
noun( 'lechery', 'lecheries', both, _ ).
noun( 'lectern', 'lecterns', count, _ ).
noun( 'lecture', 'lectures', count, _ ).
noun( 'lecturer', 'lecturers', count, _ ).
noun( 'lectureship', 'lectureships', count, _ ).
noun( 'ledge', 'ledges', count, _ ).
noun( 'ledger', 'ledgers', count, _ ).
noun( 'lee', '-', mass, _ ).
noun( 'leech', 'leeches', count, _ ).
noun( 'leek', 'leeks', count, _ ).
noun( 'leer', 'leers', count, _ ).
noun( 'leeway', '-', mass, _ ).
noun( 'left', '-', mass, _ ).
noun( 'left-wing', 'left-wings', count, _ ).
noun( 'left-winger', 'left-wingers', count, _ ).
noun( 'leftist', 'leftists', count, _ ).
noun( 'leg', 'legs', both, _ ).
noun( 'leg-pull', 'leg-pulls', count, _ ).
noun( 'leg-pulling', '-', mass, _ ).
noun( 'legacy', 'legacies', count, _ ).
noun( 'legalism', '-', mass, _ ).
noun( 'legality', '-', mass, _ ).
noun( 'legalization', '-', mass, _ ).
noun( 'legate', 'legates', count, _ ).
noun( 'legatee', 'legatees', count, _ ).
noun( 'legation', 'legations', count, _ ).
noun( 'legend', 'legends', both, _ ).
noun( 'leger line', 'leger lines', count, _ ).
noun( 'legerdemain', '-', mass, _ ).
noun( 'legging', 'leggings', count, _ ).
noun( 'leghorn', 'leghorns', count, _ ).
noun( 'legibility', '-', mass, _ ).
noun( 'legion', 'legions', count, _ ).
noun( 'legionary', 'legionaries', count, _ ).
noun( 'legislation', '-', mass, _ ).
noun( 'legislator', 'legislators', count, _ ).
noun( 'legislature', 'legislatures', count, _ ).
noun( 'legitimacy', '-', mass, _ ).
noun( 'lei', 'leis', count, _ ).
noun( 'leisure', '-', mass, _ ).
noun( 'lemming', 'lemmings', count, _ ).
noun( 'lemon', 'lemons', count, _ ).
noun( 'lemonade', '-', mass, _ ).
noun( 'lemur', 'lemurs', count, _ ).
noun( 'lender', 'lenders', count, _ ).
noun( 'lending-library', 'lending-libraries', count, _ ).
noun( 'length', 'lengths', both, _ ).
noun( 'lenience', '-', mass, _ ).
noun( 'leniency', '-', mass, _ ).
noun( 'lenity', '-', mass, _ ).
noun( 'lens', 'lenses', count, _ ).
noun( 'lentil', 'lentils', count, _ ).
noun( 'leopard', 'leopards', count, _ ).
noun( 'leopardess', 'leopardesses', count, _ ).
noun( 'leper', 'lepers', count, _ ).
noun( 'leprechaun', 'leprechauns', count, _ ).
noun( 'leprosy', '-', mass, _ ).
noun( 'lesbian', 'lesbians', count, _ ).
noun( 'lesbianism', '-', mass, _ ).
noun( 'lese majesty', '-', mass, _ ).
noun( 'lesion', 'lesions', count, _ ).
noun( 'less', '-', mass, _ ).
noun( 'lessee', 'lessees', count, _ ).
noun( 'lesson', 'lessons', count, _ ).
noun( 'lessor', 'lessors', count, _ ).
noun( 'let', 'lets', count, _ ).
noun( 'let-down', 'let-downs', count, _ ).
noun( 'let-up', 'let-ups', count, _ ).
noun( 'lethargy', '-', mass, _ ).
noun( 'letter', 'letters', count, _ ).
noun( 'letter-box', 'letter-boxes', count, _ ).
noun( 'letter-card', 'letter-cards', count, _ ).
noun( 'letter-case', 'letter-cases', count, _ ).
noun( 'letterhead', 'letterheads', count, _ ).
noun( 'lettering', '-', mass, _ ).
noun( 'letterpress', '-', mass, _ ).
noun( 'letting', 'lettings', count, _ ).
noun( 'lettuce', 'lettuces', both, _ ).
noun( 'leucocyte', 'leucocytes', count, _ ).
noun( 'leukaemia', '-', mass, _ ).
noun( 'levee', 'levees', count, _ ).
noun( 'level', 'levels', both, _ ).
noun( 'leveller', 'levellers', count, _ ).
noun( 'lever', 'levers', count, _ ).
noun( 'leverage', '-', mass, _ ).
noun( 'leveret', 'leverets', count, _ ).
noun( 'leviathan', 'leviathans', count, _ ).
noun( 'levitation', 'levitations', both, _ ).
noun( 'levity', 'levities', both, _ ).
noun( 'levy', 'levies', count, _ ).
noun( 'lewdness', '-', mass, _ ).
noun( 'lexicographer', 'lexicographers', count, _ ).
noun( 'lexicography', '-', mass, _ ).
noun( 'lexicon', 'lexicons', count, _ ).
noun( 'lexis', '-', mass, _ ).
noun( 'ley', 'leys', count, _ ).
noun( 'liability', 'liabilities', both, _ ).
noun( 'liaison', 'liaisons', both, _ ).
noun( 'liana', 'lianas', count, _ ).
noun( 'liar', 'liars', count, _ ).
noun( 'lib', '-', count, _ ).
noun( 'libation', 'libations', count, _ ).
noun( 'libel', 'libels', both, _ ).
noun( 'liberal', 'liberals', count, _ ).
noun( 'liberalism', '-', mass, _ ).
noun( 'liberality', 'liberalities', both, _ ).
noun( 'liberalization', 'liberalizations', count, _ ).
noun( 'liberation', 'liberations', both, _ ).
noun( 'liberator', 'liberators', count, _ ).
noun( 'libertine', 'libertines', count, _ ).
noun( 'liberty', 'liberties', both, _ ).
noun( 'libido', 'libidos', both, _ ).
noun( 'librarian', 'librarians', count, _ ).
noun( 'librarianship', '-', mass, _ ).
noun( 'library', 'libraries', count, _ ).
noun( 'librettist', 'librettists', count, _ ).
noun( 'libretto', 'librettos', count, _ ).
noun( 'licence', 'licences', both, _ ).
noun( 'licensee', 'licensees', count, _ ).
noun( 'licentiate', 'licentiates', count, _ ).
noun( 'licentiousness', '-', mass, _ ).
noun( 'lichee', 'lichees', count, _ ).
noun( 'lichen', '-', mass, _ ).
noun( 'lichgate', 'lichgates', count, _ ).
noun( 'lichi', 'lichis', count, _ ).
noun( 'lick', 'licks', count, _ ).
noun( 'licking', '-', count, _ ).
noun( 'lid', 'lids', count, _ ).
noun( 'lido', 'lidos', count, _ ).
noun( 'lie', 'lies', count, _ ).
noun( 'lie-abed', 'lie-abeds', count, _ ).
noun( 'lie-detector', 'lie-detectors', count, _ ).
noun( 'lie-in', 'lie-ins', count, _ ).
noun( 'lied', 'lieder', count, _ ).
noun( 'lieder-singer', 'lieder-singers', count, _ ).
noun( 'liege', 'lieges', count, _ ).
noun( 'liegeman', 'liegemen', count, _ ).
noun( 'lien', 'liens', count, _ ).
noun( 'lieu', '-', mass, _ ).
noun( 'lieutenancy', 'lieutenancies', count, _ ).
noun( 'lieutenant', 'lieutenants', count, _ ).
noun( 'life', 'lives', both, _ ).
noun( 'life-buoy', 'life-buoys', count, _ ).
noun( 'life-jacket', 'life-jackets', count, _ ).
noun( 'life-office', 'life-offices', count, _ ).
noun( 'life-preserver', 'life-preservers', count, _ ).
noun( 'life-raft', 'life-rafts', count, _ ).
noun( 'life-saver', 'life-savers', count, _ ).
noun( 'life-span', 'life-spans', count, _ ).
noun( 'life-work', 'life-works', count, _ ).
noun( 'lifebelt', 'lifebelts', count, _ ).
noun( 'lifeblood', '-', mass, _ ).
noun( 'lifeboat', 'lifeboats', count, _ ).
noun( 'lifebuoy', 'lifebuoys', count, _ ).
noun( 'lifeguard', 'lifeguards', count, _ ).
noun( 'lifeline', 'lifelines', count, _ ).
noun( 'lifer', 'lifers', count, _ ).
noun( 'lifetime', 'lifetimes', count, _ ).
noun( 'lift', 'lifts', count, _ ).
noun( 'lift-off', 'lift-offs', count, _ ).
noun( 'liftman', 'liftmen', count, _ ).
noun( 'ligament', 'ligaments', count, _ ).
noun( 'ligature', 'ligatures', count, _ ).
noun( 'light', 'lights', both, _ ).
noun( 'light-headedness', '-', mass, _ ).
noun( 'light-heartedness', '-', mass, _ ).
noun( 'light-heavyweight', 'light-heavyweights', count, _ ).
noun( 'light-mindedness', '-', mass, _ ).
noun( 'light-o\'-love', 'light-o\'-loves', count, _ ).
noun( 'lighter', 'lighters', count, _ ).
noun( 'lighterage', '-', mass, _ ).
noun( 'lighthouse', 'lighthouses', count, _ ).
noun( 'lightness', '-', mass, _ ).
noun( 'lightning', '-', mass, _ ).
noun( 'lightning-conductor', 'lightning-conductors', count, _ ).
noun( 'lightning-rod', 'lightning-rods', count, _ ).
noun( 'lightship', 'lightships', count, _ ).
noun( 'lightsomeness', '-', mass, _ ).
noun( 'lightweight', 'lightweights', count, _ ).
noun( 'lignite', '-', mass, _ ).
noun( 'like', 'likes', count, _ ).
noun( 'likelihood', '-', mass, _ ).
noun( 'likeness', 'likenesses', both, _ ).
noun( 'liking', 'likings', count, _ ).
noun( 'lilac', 'lilacs', count, _ ).
noun( 'lilt', 'lilts', count, _ ).
noun( 'lily', 'lilies', count, _ ).
noun( 'limb', 'limbs', count, _ ).
noun( 'limbo', 'limbos', both, _ ).
noun( 'lime', 'limes', both, _ ).
noun( 'lime-tree', 'lime-trees', count, _ ).
noun( 'limejuice', '-', mass, _ ).
noun( 'limekiln', 'limekilns', count, _ ).
noun( 'limelight', '-', mass, _ ).
noun( 'limerick', 'limericks', count, _ ).
noun( 'limestone', '-', mass, _ ).
noun( 'limey', 'limeys', count, _ ).
noun( 'limit', 'limits', count, _ ).
noun( 'limitation', 'limitations', both, _ ).
noun( 'limousine', 'limousines', count, _ ).
noun( 'limp', 'limps', count, _ ).
noun( 'limpet', 'limpets', count, _ ).
noun( 'limpidity', '-', mass, _ ).
noun( 'limpness', '-', mass, _ ).
noun( 'linchpin', 'linchpins', count, _ ).
noun( 'linden', 'lindens', count, _ ).
noun( 'linden-tree', 'linden-trees', count, _ ).
noun( 'line', 'lines', both, _ ).
noun( 'line-shooter', 'line-shooters', count, _ ).
noun( 'line-shooting', '-', mass, _ ).
noun( 'line-up', 'line-ups', count, _ ).
noun( 'lineage', '-', mass, _ ).
noun( 'lineament', 'lineaments', count, _ ).
noun( 'lineman', 'linemen', count, _ ).
noun( 'linen', '-', mass, _ ).
noun( 'linen-draper', 'linen-drapers', count, _ ).
noun( 'liner', 'liners', count, _ ).
noun( 'liner-train', 'liner-trains', count, _ ).
noun( 'linesman', 'linesmen', count, _ ).
noun( 'ling', 'lings', both, _ ).
noun( 'lingam', 'lingams', count, _ ).
noun( 'lingerer', 'lingerers', count, _ ).
noun( 'lingerie', '-', mass, _ ).
noun( 'lingo', 'lingoes', count, _ ).
noun( 'lingua franca', '-', count, _ ).
noun( 'linguist', 'linguists', count, _ ).
noun( 'linguistics', 'linguistics', mass, _ ).
noun( 'liniment', 'liniments', both, _ ).
noun( 'lining', 'linings', both, _ ).
noun( 'link', 'links', count, _ ).
noun( 'link-up', 'link-ups', count, _ ).
noun( 'linkage', 'linkages', both, _ ).
noun( 'linkboy', 'linkboys', count, _ ).
noun( 'linkman', 'linkmen', count, _ ).
noun( 'links', 'links', count, _ ).
noun( 'linnet', 'linnets', count, _ ).
noun( 'lino', '-', mass, _ ).
noun( 'linocut', 'linocuts', count, _ ).
noun( 'linoleum', '-', mass, _ ).
noun( 'linotype', 'linotypes', count, _ ).
noun( 'linseed', '-', mass, _ ).
noun( 'linsey-woolsey', '-', mass, _ ).
noun( 'lint', '-', mass, _ ).
noun( 'lintel', 'lintels', count, _ ).
noun( 'lion', 'lions', count, _ ).
noun( 'lion-hunter', 'lion-hunters', count, _ ).
noun( 'lioness', 'lionesses', count, _ ).
noun( 'lip', 'lips', both, _ ).
noun( 'lip-reading', '-', mass, _ ).
noun( 'lipstick', 'lipsticks', both, _ ).
noun( 'liquefaction', '-', mass, _ ).
noun( 'liqueur', 'liqueurs', count, _ ).
noun( 'liquid', 'liquids', both, _ ).
noun( 'liquidation', '-', mass, _ ).
noun( 'liquidator', 'liquidators', count, _ ).
noun( 'liquidity', '-', mass, _ ).
noun( 'liquidizer', 'liquidizers', count, _ ).
noun( 'liquor', 'liquors', both, _ ).
noun( 'liquorice', '-', mass, _ ).
noun( 'lira', 'liras', count, _ ).
noun( 'lisle', '-', mass, _ ).
noun( 'lisp', 'lisps', count, _ ).
noun( 'lissomness', '-', mass, _ ).
noun( 'list', 'lists', count, _ ).
noun( 'list-price', 'list-prices', count, _ ).
noun( 'listener', 'listeners', count, _ ).
noun( 'listing', 'listings', count, _ ).
noun( 'listlessness', '-', mass, _ ).
noun( 'litany', 'litanies', count, _ ).
noun( 'litchee', 'litchees', count, _ ).
noun( 'litchi', 'litchis', count, _ ).
noun( 'literacy', '-', mass, _ ).
noun( 'literal', 'literals', count, _ ).
noun( 'literate', 'literates', count, _ ).
noun( 'literature', '-', mass, _ ).
noun( 'lithograph', 'lithographs', count, _ ).
noun( 'lithography', '-', mass, _ ).
noun( 'litigant', 'litigants', count, _ ).
noun( 'litigation', '-', mass, _ ).
noun( 'litmus', '-', mass, _ ).
noun( 'litmus-paper', 'litmus-papers', both, _ ).
noun( 'litotes', 'litotes', count, _ ).
noun( 'litre', 'litres', count, _ ).
noun( 'litter', 'litters', both, _ ).
noun( 'litter-basket', 'litter-baskets', count, _ ).
noun( 'litter-lout', 'litter-louts', count, _ ).
noun( 'litterbin', 'litterbins', count, _ ).
noun( 'little', '-', mass, _ ).
noun( 'littleness', '-', mass, _ ).
noun( 'littoral', 'littorals', count, _ ).
noun( 'liturgy', 'liturgies', both, _ ).
noun( 'live-birth', 'live-births', count, _ ).
noun( 'livelihood', 'livelihoods', count, _ ).
noun( 'liveliness', '-', mass, _ ).
noun( 'liver', 'livers', both, _ ).
noun( 'liverwurst', '-', mass, _ ).
noun( 'livery', 'liveries', count, _ ).
noun( 'liveryman', 'liverymen', count, _ ).
noun( 'livestock', '-', mass, _ ).
noun( 'living', 'livings', both, _ ).
noun( 'living-room', 'living-rooms', count, _ ).
noun( 'living-space', 'living-spaces', count, _ ).
noun( 'lizard', 'lizards', count, _ ).
noun( '-', 'll', count, _ ).
noun( 'llama', 'llamas', count, _ ).
noun( 'load', 'loads', count, _ ).
noun( 'load-line', 'load-lines', count, _ ).
noun( 'load-shedding', '-', mass, _ ).
noun( 'loader', 'loaders', count, _ ).
noun( 'loading', 'loadings', count, _ ).
noun( 'loadstar', 'loadstars', count, _ ).
noun( 'loadstone', '-', mass, _ ).
noun( 'loaf', 'loaves', both, _ ).
noun( 'loaf-sugar', '-', mass, _ ).
noun( 'loafer', 'loafers', count, _ ).
noun( 'loam', '-', mass, _ ).
noun( 'loan', 'loans', both, _ ).
noun( 'loan-collection', 'loan-collections', count, _ ).
noun( 'loan-office', 'loan-offices', count, _ ).
noun( 'loanword', 'loanwords', count, _ ).
noun( 'loathing', '-', mass, _ ).
noun( 'lob', 'lobs', count, _ ).
noun( 'lobby', 'lobbies', count, _ ).
noun( 'lobbyist', 'lobbyists', count, _ ).
noun( 'lobe', 'lobes', count, _ ).
noun( 'lobster', 'lobsters', both, _ ).
noun( 'lobster-pot', 'lobster-pots', count, _ ).
noun( 'loc cit', '-', proper, _ ).
noun( 'local', 'locals', count, _ ).
noun( 'locale', 'locales', count, _ ).
noun( 'localism', 'localisms', both, _ ).
noun( 'locality', 'localities', both, _ ).
noun( 'localization', 'localizations', both, _ ).
noun( 'location', 'locations', both, _ ).
noun( 'loch', 'lochs', count, _ ).
noun( 'lock', 'locks', both, _ ).
noun( 'lock-gate', 'lock-gates', count, _ ).
noun( 'lock-keeper', 'lock-keepers', count, _ ).
noun( 'locker', 'lockers', count, _ ).
noun( 'locket', 'lockets', count, _ ).
noun( 'lockjaw', '-', mass, _ ).
noun( 'locknut', 'locknuts', count, _ ).
noun( 'lockout', 'lockouts', count, _ ).
noun( 'locksmith', 'locksmiths', count, _ ).
noun( 'lockstitch', 'lockstitches', count, _ ).
noun( 'lockup', 'lockups', count, _ ).
noun( 'locomotion', '-', mass, _ ).
noun( 'locomotive', 'locomotives', count, _ ).
noun( 'locum', 'locums', count, _ ).
noun( 'locum tenens', '-', count, _ ).
noun( 'locus', 'loci', count, _ ).
noun( 'locus classicus', '-', count, _ ).
noun( 'locust', 'locusts', count, _ ).
noun( 'locust-tree', 'locust-trees', count, _ ).
noun( 'locution', 'locutions', both, _ ).
noun( 'lode', 'lodes', count, _ ).
noun( 'lodestar', 'lodestars', count, _ ).
noun( 'lodestone', '-', mass, _ ).
noun( 'lodge', 'lodges', count, _ ).
noun( 'lodgement', 'lodgements', both, _ ).
noun( 'lodger', 'lodgers', count, _ ).
noun( 'lodging', 'lodgings', count, _ ).
noun( 'lodging-house', 'lodging-houses', count, _ ).
noun( 'lodgment', 'lodgments', both, _ ).
noun( 'loess', '-', mass, _ ).
noun( 'loft', 'lofts', count, _ ).
noun( 'loftiness', '-', mass, _ ).
noun( 'log', 'logs', count, _ ).
noun( 'log-cabin', 'log-cabins', count, _ ).
noun( 'log-jam', 'log-jams', count, _ ).
noun( 'log-rolling', '-', mass, _ ).
noun( 'loganberry', 'loganberries', count, _ ).
noun( 'logarithm', 'logarithms', count, _ ).
noun( 'logbook', 'logbooks', count, _ ).
noun( 'loggerheads', '-', mass, _ ).
noun( 'loggia', 'loggias', count, _ ).
noun( 'logging', '-', mass, _ ).
noun( 'logic', 'logics', both, _ ).
noun( 'logicality', '-', mass, _ ).
noun( 'logician', 'logicians', count, _ ).
noun( 'logistics', 'logistics', mass, _ ).
noun( 'loin', 'loins', count, _ ).
noun( 'loincloth', 'loincloths', count, _ ).
noun( 'loiterer', 'loiterers', count, _ ).
noun( 'lollipop', 'lollipops', count, _ ).
noun( 'lolly', 'lollies', count, _ ).
noun( 'loneliness', '-', mass, _ ).
noun( 'long', 'longs', both, _ ).
noun( 'long', '-', count, _ ).
noun( 'long-windedness', '-', mass, _ ).
noun( 'longboat', 'longboats', count, _ ).
noun( 'longbow', 'longbows', count, _ ).
noun( 'longevity', '-', mass, _ ).
noun( 'longhand', '-', mass, _ ).
noun( 'longing', 'longings', both, _ ).
noun( 'longitude', 'longitudes', count, _ ).
noun( 'longshoreman', 'longshoremen', count, _ ).
noun( 'loo', 'loos', count, _ ).
noun( 'loofa', 'loofas', count, _ ).
noun( 'loofah', 'loofahs', count, _ ).
noun( 'look', 'looks', count, _ ).
noun( 'look-over', '-', count, _ ).
noun( 'looker', 'lookers', count, _ ).
noun( 'looker-on', 'lookers-on', count, _ ).
noun( 'looking-glass', 'looking-glasses', count, _ ).
noun( 'lookout', 'lookouts', count, _ ).
noun( 'loom', 'looms', count, _ ).
noun( 'loon', 'loons', count, _ ).
noun( 'loony', 'loonies', count, _ ).
noun( 'loonybin', 'loonybins', count, _ ).
noun( 'loop', 'loops', count, _ ).
noun( 'loop-line', 'loop-lines', count, _ ).
noun( 'loophole', 'loopholes', count, _ ).
noun( 'loot', '-', mass, _ ).
noun( 'looter', 'looters', count, _ ).
noun( 'lope', '-', count, _ ).
noun( 'loquaciousness', '-', mass, _ ).
noun( 'loquacity', '-', mass, _ ).
noun( 'loquat', 'loquats', count, _ ).
noun( 'lord', 'lords', count, _ ).
noun( 'lordliness', '-', mass, _ ).
noun( 'lordship', 'lordships', both, _ ).
noun( 'lore', '-', mass, _ ).
noun( 'lorgnette', 'lorgnettes', count, _ ).
noun( 'lorry', 'lorries', count, _ ).
noun( 'loser', 'losers', count, _ ).
noun( 'loss', 'losses', both, _ ).
noun( 'loss-leader', 'loss-leaders', count, _ ).
noun( 'lot', 'lots', both, _ ).
noun( 'lotion', 'lotions', both, _ ).
noun( 'lottery', 'lotteries', count, _ ).
noun( 'lotto', '-', mass, _ ).
noun( 'lotus', 'lotuses', count, _ ).
noun( 'lotus-eater', 'lotus-eaters', count, _ ).
noun( 'loud-hailer', 'loud-hailers', count, _ ).
noun( 'loudness', '-', mass, _ ).
noun( 'loudspeaker', 'loudspeakers', count, _ ).
noun( 'lough', 'loughs', count, _ ).
noun( 'lounge', 'lounges', count, _ ).
noun( 'lounge-chair', 'lounge-chairs', count, _ ).
noun( 'lounge-lizard', 'lounge-lizards', count, _ ).
noun( 'lounge-suit', 'lounge-suits', count, _ ).
noun( 'lounger', 'loungers', count, _ ).
noun( 'louse', 'lice', count, _ ).
noun( 'lout', 'louts', count, _ ).
noun( 'louvre', 'louvres', count, _ ).
noun( 'love', 'loves', both, _ ).
noun( 'love-affair', 'love-affairs', count, _ ).
noun( 'love-child', 'love-children', count, _ ).
noun( 'love-feast', 'love-feasts', count, _ ).
noun( 'love-knot', 'love-knots', count, _ ).
noun( 'love-letter', 'love-letters', count, _ ).
noun( 'love-match', 'love-matches', count, _ ).
noun( 'love-philtre', 'love-philtres', count, _ ).
noun( 'love-potion', 'love-potions', count, _ ).
noun( 'love-seat', 'love-seats', count, _ ).
noun( 'love-song', 'love-songs', count, _ ).
noun( 'love-story', 'love-stories', count, _ ).
noun( 'love-token', 'love-tokens', count, _ ).
noun( 'lovebird', 'lovebirds', count, _ ).
noun( 'loveliness', '-', mass, _ ).
noun( 'lovemaking', '-', mass, _ ).
noun( 'lover', 'lovers', count, _ ).
noun( 'loving-cup', 'loving-cups', count, _ ).
noun( 'loving-kindness', '-', mass, _ ).
noun( 'low', 'lows', count, _ ).
noun( 'low-relief', 'low-reliefs', both, _ ).
noun( 'lowbrow', 'lowbrows', count, _ ).
noun( 'lowlander', 'lowlanders', count, _ ).
noun( 'lowliness', '-', mass, _ ).
noun( 'lowness', '-', mass, _ ).
noun( 'loyalist', 'loyalists', count, _ ).
noun( 'loyalty', 'loyalties', both, _ ).
noun( 'lozenge', 'lozenges', count, _ ).
noun( 'lubber', 'lubbers', count, _ ).
noun( 'lubricant', 'lubricants', both, _ ).
noun( 'lubrication', 'lubrications', both, _ ).
noun( 'lucerne', '-', mass, _ ).
noun( 'lucidity', '-', mass, _ ).
noun( 'luck', '-', mass, _ ).
noun( 'lucre', '-', mass, _ ).
noun( 'ludo', '-', mass, _ ).
noun( 'lug', 'lugs', count, _ ).
noun( 'luge', 'luges', count, _ ).
noun( 'luggage', '-', mass, _ ).
noun( 'luggage-carrier', 'luggage-carriers', count, _ ).
noun( 'luggage-rack', 'luggage-racks', count, _ ).
noun( 'luggage-van', 'luggage-vans', count, _ ).
noun( 'lugger', 'luggers', count, _ ).
noun( 'lugsail', 'lugsails', count, _ ).
noun( 'lugubriousness', '-', mass, _ ).
noun( 'lukewarmness', '-', mass, _ ).
noun( 'lull', 'lulls', count, _ ).
noun( 'lullaby', 'lullabies', count, _ ).
noun( 'lumbago', '-', mass, _ ).
noun( 'lumber', '-', mass, _ ).
noun( 'lumber-mill', 'lumber-mills', count, _ ).
noun( 'lumberjack', 'lumberjacks', count, _ ).
noun( 'lumberman', 'lumbermen', count, _ ).
noun( 'lumberroom', 'lumberrooms', count, _ ).
noun( 'lumberyard', 'lumberyards', count, _ ).
noun( 'luminary', 'luminaries', count, _ ).
noun( 'luminosity', '-', mass, _ ).
noun( 'lump', 'lumps', count, _ ).
noun( 'lunacy', 'lunacies', both, _ ).
noun( 'lunatic', 'lunatics', count, _ ).
noun( 'lunch', 'lunches', count, _ ).
noun( 'luncheon', 'luncheons', count, _ ).
noun( 'lunchtime', 'lunchtimes', both, _ ).
noun( 'lung', 'lungs', count, _ ).
noun( 'lung-power', '-', mass, _ ).
noun( 'lunge', 'lunges', count, _ ).
noun( 'lupin', 'lupins', count, _ ).
noun( 'lurch', 'lurches', count, _ ).
noun( 'lurcher', 'lurchers', count, _ ).
noun( 'lure', 'lures', count, _ ).
noun( 'luridness', '-', mass, _ ).
noun( 'lurking-place', 'lurking-places', count, _ ).
noun( 'lusciousness', '-', mass, _ ).
noun( 'lush', 'lushes', count, _ ).
noun( 'lust', 'lusts', both, _ ).
noun( 'lustre', '-', mass, _ ).
noun( 'lutanist', 'lutanists', count, _ ).
noun( 'lute', 'lutes', count, _ ).
noun( 'lutenist', 'lutenists', count, _ ).
noun( 'luxuriance', '-', mass, _ ).
noun( 'luxury', 'luxuries', both, _ ).
noun( 'lyc_ee', 'lyc_ees', count, _ ).
noun( 'lyceum', 'lyceums', count, _ ).
noun( 'lychee', 'lychees', count, _ ).
noun( 'lychgate', 'lychgates', count, _ ).
noun( 'lye', 'lyes', count, _ ).
noun( 'lymph', '-', mass, _ ).
noun( 'lynch', '-', mass, _ ).
noun( 'lynchpin', 'lynchpins', count, _ ).
noun( 'lynx', 'lynxes', count, _ ).
noun( 'lyre', 'lyres', count, _ ).
noun( 'lyre-bird', 'lyre-birds', count, _ ).
noun( 'lyric', 'lyrics', count, _ ).
noun( 'lyricism', '-', mass, _ ).
noun( 'lyricist', 'lyricists', count, _ ).
noun( 'lysol', '-', mass, _ ).
noun( 'm', '-', count, _ ).
noun( 'm^el_ee', 'm^el_ees', count, _ ).
noun( 'm_elange', 'm_elanges', count, _ ).
noun( 'm_enage', 'm_enages', count, _ ).
noun( 'm_esalliance', 'm_esalliances', count, _ ).
noun( 'm_etier', 'm_etiers', count, _ ).
noun( 'ma', '-', count, _ ).
noun( 'ma\'am', '-', count, _ ).
noun( 'mac', 'macs', count, _ ).
noun( 'macadam', '-', mass, _ ).
noun( 'macaroni', '-', mass, _ ).
noun( 'macaroon', 'macaroons', count, _ ).
noun( 'macaw', 'macaws', count, _ ).
noun( 'mace', 'maces', both, _ ).
noun( 'mace-bearer', 'mace-bearers', count, _ ).
noun( 'machete', 'machetes', count, _ ).
noun( 'machination', 'machinations', both, _ ).
noun( 'machine', 'machines', count, _ ).
noun( 'machine-gun', 'machine-guns', count, _ ).
noun( 'machinery', '-', mass, _ ).
noun( 'machinist', 'machinists', count, _ ).
noun( 'machismo', '-', mass, _ ).
noun( 'mackerel', 'mackerel', both, _ ).
noun( 'mackintosh', 'mackintoshes', count, _ ).
noun( 'macrocosm', 'macrocosms', count, _ ).
noun( 'madam', 'madams', count, _ ).
noun( 'madcap', 'madcaps', count, _ ).
noun( 'madder', '-', mass, _ ).
noun( 'madhouse', 'madhouses', count, _ ).
noun( 'madman', 'madmen', count, _ ).
noun( 'madness', '-', mass, _ ).
noun( 'madrigal', 'madrigals', count, _ ).
noun( 'madwoman', 'madwomen', count, _ ).
noun( 'maelstrom', 'maelstroms', count, _ ).
noun( 'maenad', 'maenads', count, _ ).
noun( 'maestro', 'maestros', count, _ ).
noun( 'mag', 'mags', count, _ ).
noun( 'magazine', 'magazines', count, _ ).
noun( 'magenta', '-', mass, _ ).
noun( 'maggot', 'maggots', count, _ ).
noun( 'magic', '-', mass, _ ).
noun( 'magician', 'magicians', count, _ ).
noun( 'magistracy', 'magistracies', count, _ ).
noun( 'magistrate', 'magistrates', count, _ ).
noun( 'magnanimity', 'magnanimities', both, _ ).
noun( 'magnate', 'magnates', count, _ ).
noun( 'magnesia', '-', mass, _ ).
noun( 'magnesium', '-', mass, _ ).
noun( 'magnet', 'magnets', count, _ ).
noun( 'magnetism', '-', mass, _ ).
noun( 'magnetization', '-', mass, _ ).
noun( 'magneto', 'magnetos', count, _ ).
noun( 'magnification', 'magnifications', both, _ ).
noun( 'magnificence', '-', mass, _ ).
noun( 'magnifier', 'magnifiers', count, _ ).
noun( 'magniloquence', '-', mass, _ ).
noun( 'magnitude', '-', mass, _ ).
noun( 'magnolia', 'magnolias', count, _ ).
noun( 'magnum', 'magnums', count, _ ).
noun( 'magnum opus', '-', count, _ ).
noun( 'magpie', 'magpies', count, _ ).
noun( 'mahjong', '-', mass, _ ).
noun( 'mahogany', 'mahoganies', both, _ ).
noun( 'mahout', 'mahouts', count, _ ).
noun( 'maid', 'maids', count, _ ).
noun( 'maiden', 'maidens', count, _ ).
noun( 'maidenhair', 'maidenhairs', count, _ ).
noun( 'maidenhead', '-', mass, _ ).
noun( 'maidenhood', 'maidenhoods', count, _ ).
noun( 'maidservant', 'maidservants', count, _ ).
noun( 'mail', 'mails', both, _ ).
noun( 'mail-train', 'mail-trains', count, _ ).
noun( 'mailbag', 'mailbags', count, _ ).
noun( 'mailboat', 'mailboats', count, _ ).
noun( 'mailbox', 'mailboxes', count, _ ).
noun( 'mailing-card', 'mailing-cards', count, _ ).
noun( 'mailing-list', 'mailing-lists', count, _ ).
noun( 'mailman', 'mailmen', count, _ ).
noun( 'main', 'mains', count, _ ).
noun( 'mainland', 'mainlands', count, _ ).
noun( 'mainmast', 'mainmasts', count, _ ).
noun( 'mainspring', 'mainsprings', count, _ ).
noun( 'mainstay', 'mainstays', count, _ ).
noun( 'mainstream', '-', count, _ ).
noun( 'maintenance', '-', mass, _ ).
noun( 'maisonnette', 'maisonnettes', count, _ ).
noun( 'maize', '-', mass, _ ).
noun( 'majesty', 'majesties', both, _ ).
noun( 'majolica', '-', mass, _ ).
noun( 'major', 'majors', count, _ ).
noun( 'major-domo', 'major-domos', count, _ ).
noun( 'major-general', 'major-generals', count, _ ).
noun( 'majority', 'majorities', count, _ ).
noun( 'make', 'makes', both, _ ).
noun( 'make-believe', 'make-believes', both, _ ).
noun( 'make-up', 'make-ups', both, _ ).
noun( 'maker', 'makers', count, _ ).
noun( 'makeshift', 'makeshifts', count, _ ).
noun( 'makeweight', 'makeweights', count, _ ).
noun( 'making', 'makings', count, _ ).
noun( 'malacca', '-', mass, _ ).
noun( 'malachite', '-', mass, _ ).
noun( 'maladjustment', '-', mass, _ ).
noun( 'maladroitness', '-', mass, _ ).
noun( 'malady', 'maladies', count, _ ).
noun( 'malaise', 'malaises', both, _ ).
noun( 'malapropism', 'malapropisms', count, _ ).
noun( 'malaria', '-', mass, _ ).
noun( 'malcontent', 'malcontents', count, _ ).
noun( 'male', 'males', count, _ ).
noun( 'malediction', 'maledictions', count, _ ).
noun( 'malefactor', 'malefactors', count, _ ).
noun( 'malevolence', '-', mass, _ ).
noun( 'malfeasance', 'malfeasances', both, _ ).
noun( 'malformation', 'malformations', both, _ ).
noun( 'malfunction', 'malfunctions', both, _ ).
noun( 'malice', '-', mass, _ ).
noun( 'malignancy', '-', mass, _ ).
noun( 'malignity', 'malignities', both, _ ).
noun( 'malingerer', 'malingerers', count, _ ).
noun( 'mallard', 'mallards', count, _ ).
noun( 'malleability', '-', mass, _ ).
noun( 'mallet', 'mallets', count, _ ).
noun( 'mallow', 'mallows', count, _ ).
noun( 'malmsey', '-', mass, _ ).
noun( 'malnutrition', '-', mass, _ ).
noun( 'malpractice', 'malpractices', both, _ ).
noun( 'malt', '-', mass, _ ).
noun( 'maltreatment', '-', mass, _ ).
noun( 'maltster', 'maltsters', count, _ ).
noun( 'malversation', '-', mass, _ ).
noun( 'mama', 'mamas', count, _ ).
noun( 'mamba', 'mambas', count, _ ).
noun( 'mamma', 'mammas', count, _ ).
noun( 'mammal', 'mammals', count, _ ).
noun( 'mammon', '-', mass, _ ).
noun( 'mammoth', 'mammoths', count, _ ).
noun( 'mammy', 'mammies', count, _ ).
noun( 'man', 'men', count, _ ).
noun( 'man-at-arms', 'men-at-arms', count, _ ).
noun( 'man-eater', 'man-eaters', count, _ ).
noun( 'man-hour', 'man-hours', count, _ ).
noun( 'man-of-war', 'men-of-war', count, _ ).
noun( 'manacle', 'manacles', count, _ ).
noun( 'manageability', '-', mass, _ ).
noun( 'management', 'managements', both, _ ).
noun( 'manager', 'managers', count, _ ).
noun( 'manageress', 'manageresses', count, _ ).
noun( 'manatee', 'manatees', count, _ ).
noun( 'mandarin', 'mandarins', count, _ ).
noun( 'mandatary', 'mandataries', count, _ ).
noun( 'mandate', 'mandates', count, _ ).
noun( 'mandatory', 'mandatories', count, _ ).
noun( 'mandible', 'mandibles', count, _ ).
noun( 'mandolin', 'mandolins', count, _ ).
noun( 'mandragora', '-', mass, _ ).
noun( 'mandrake', 'mandrakes', count, _ ).
noun( 'mandrill', 'mandrills', count, _ ).
noun( 'mane', 'manes', count, _ ).
noun( 'manganese', '-', mass, _ ).
noun( 'mange', '-', mass, _ ).
noun( 'mangel-wurzel', 'mangel-wurzels', count, _ ).
noun( 'manger', 'mangers', count, _ ).
noun( 'mangle', 'mangles', count, _ ).
noun( 'mango', 'mangos', count, _ ).
noun( 'mangosteen', 'mangosteens', count, _ ).
noun( 'mangrove', 'mangroves', count, _ ).
noun( 'manhattan', 'manhattans', count, _ ).
noun( 'manhole', 'manholes', count, _ ).
noun( 'manhood', '-', mass, _ ).
noun( 'mania', 'manias', both, _ ).
noun( 'maniac', 'maniacs', count, _ ).
noun( 'manic-depressive', 'manic-depressives', count, _ ).
noun( 'manicure', 'manicures', both, _ ).
noun( 'manicurist', 'manicurists', count, _ ).
noun( 'manifest', 'manifests', count, _ ).
noun( 'manifestation', 'manifestations', both, _ ).
noun( 'manifesto', 'manifestos', count, _ ).
noun( 'manifold', 'manifolds', count, _ ).
noun( 'manikin', 'manikins', count, _ ).
noun( 'manipulation', 'manipulations', both, _ ).
noun( 'mankind', '-', mass, _ ).
noun( 'mankind', '-', mass, _ ).
noun( 'manliness', '-', mass, _ ).
noun( 'manna', '-', mass, _ ).
noun( 'mannequin', 'mannequins', count, _ ).
noun( 'manner', 'manners', count, _ ).
noun( 'mannerism', 'mannerisms', count, _ ).
noun( 'manoeuvrability', '-', mass, _ ).
noun( 'manoeuvre', 'manoeuvres', count, _ ).
noun( 'manoeuvrer', 'manoeuvrers', count, _ ).
noun( 'manor', 'manors', count, _ ).
noun( 'manor-house', 'manor-houses', count, _ ).
noun( 'manpower', '-', mass, _ ).
noun( 'mansard', 'mansards', count, _ ).
noun( 'manse', 'manses', count, _ ).
noun( 'manservant', 'manservants', count, _ ).
noun( 'mansion', 'mansions', count, _ ).
noun( 'manslaughter', '-', mass, _ ).
noun( 'mantel', 'mantels', count, _ ).
noun( 'mantelpiece', 'mantelpieces', count, _ ).
noun( 'mantilla', 'mantillas', count, _ ).
noun( 'mantis', 'mantises', count, _ ).
noun( 'mantle', 'mantles', count, _ ).
noun( 'mantrap', 'mantraps', count, _ ).
noun( 'manual', 'manuals', count, _ ).
noun( 'manufacture', '-', mass, _ ).
noun( 'manufacturer', 'manufacturers', count, _ ).
noun( 'manumission', 'manumissions', both, _ ).
noun( 'manure', '-', mass, _ ).
noun( 'manuscript', 'manuscripts', count, _ ).
noun( 'map', 'maps', count, _ ).
noun( 'map-reader', 'map-readers', count, _ ).
noun( 'maple', 'maples', both, _ ).
noun( 'maple-leaf', 'maple-leaves', count, _ ).
noun( 'maquis', '-', count, _ ).
noun( 'marabou', 'marabous', count, _ ).
noun( 'maraschino', 'maraschinos', count, _ ).
noun( 'marathon', 'marathons', count, _ ).
noun( 'marauder', 'marauders', count, _ ).
noun( 'marble', 'marbles', both, _ ).
noun( 'marbling', '-', mass, _ ).
noun( 'march', 'marches', both, _ ).
noun( 'marcher', 'marchers', count, _ ).
noun( 'marchioness', 'marchionesses', count, _ ).
noun( 'mare', 'mares', count, _ ).
noun( 'margarine', '-', mass, _ ).
noun( 'marge', '-', mass, _ ).
noun( 'margin', 'margins', count, _ ).
noun( 'marguerite', 'marguerites', count, _ ).
noun( 'marigold', 'marigolds', count, _ ).
noun( 'marihuana', '-', mass, _ ).
noun( 'marijuana', '-', mass, _ ).
noun( 'marimba', 'marimbas', count, _ ).
noun( 'marina', 'marinas', count, _ ).
noun( 'marinade', '-', mass, _ ).
noun( 'marine', 'marines', count, _ ).
noun( 'mariner', 'mariners', count, _ ).
noun( 'marionette', 'marionettes', count, _ ).
noun( 'marjoram', '-', mass, _ ).
noun( 'mark', 'marks', both, _ ).
noun( 'mark-up', 'mark-ups', count, _ ).
noun( 'marker', 'markers', count, _ ).
noun( 'market', 'markets', count, _ ).
noun( 'market-cross', 'market-crosses', count, _ ).
noun( 'market-day', 'market-days', count, _ ).
noun( 'market-garden', 'market-gardens', count, _ ).
noun( 'market-gardening', '-', mass, _ ).
noun( 'market-square', 'market-squares', count, _ ).
noun( 'market-town', 'market-towns', count, _ ).
noun( 'marketing', 'marketings', both, _ ).
noun( 'marketplace', 'marketplaces', count, _ ).
noun( 'marking', 'markings', count, _ ).
noun( 'marking-ink', 'marking-inks', both, _ ).
noun( 'marking-inks', 'marking-inkss', both, _ ).
noun( 'marksman', 'marksmen', count, _ ).
noun( 'marksmanship', '-', mass, _ ).
noun( 'marl', '-', mass, _ ).
noun( 'marlinespike', 'marlinespikes', count, _ ).
noun( 'marmalade', '-', mass, _ ).
noun( 'marmoset', 'marmosets', count, _ ).
noun( 'marmot', 'marmots', count, _ ).
noun( 'marocain', '-', mass, _ ).
noun( 'maroon', 'maroons', count, _ ).
noun( 'marque', 'marques', count, _ ).
noun( 'marquee', 'marquees', count, _ ).
noun( 'marquess', 'marquesses', count, _ ).
noun( 'marquetry', '-', mass, _ ).
noun( 'marquis', 'marquises', count, _ ).
noun( 'marriage', 'marriages', both, _ ).
noun( 'marriageability', '-', mass, _ ).
noun( 'marrow', 'marrows', both, _ ).
noun( 'marrowbone', 'marrowbones', count, _ ).
noun( 'marsh', 'marshes', both, _ ).
noun( 'marshal', 'marshals', count, _ ).
noun( 'marshalling-yard', 'marshalling-yards', count, _ ).
noun( 'marshmallow', 'marshmallows', both, _ ).
noun( 'marsupial', 'marsupials', count, _ ).
noun( 'mart', 'marts', count, _ ).
noun( 'marten', 'martens', both, _ ).
noun( 'martin', 'martins', count, _ ).
noun( 'martinet', 'martinets', count, _ ).
noun( 'martini', 'martinis', count, _ ).
noun( 'martyr', 'martyrs', count, _ ).
noun( 'martyrdom', 'martyrdoms', count, _ ).
noun( 'marvel', 'marvels', count, _ ).
noun( 'marzipan', 'marzipans', both, _ ).
noun( 'masc', '-', proper, _ ).
noun( 'mascara', '-', mass, _ ).
noun( 'mascot', 'mascots', count, _ ).
noun( 'masculinity', '-', mass, _ ).
noun( 'maser', 'masers', count, _ ).
noun( 'mash', 'mashes', both, _ ).
noun( 'masher', 'mashers', count, _ ).
noun( 'mask', 'masks', count, _ ).
noun( 'masochism', '-', mass, _ ).
noun( 'masochist', 'masochists', count, _ ).
noun( 'mason', 'masons', count, _ ).
noun( 'masonry', '-', mass, _ ).
noun( 'masque', 'masques', count, _ ).
noun( 'masquerade', 'masquerades', count, _ ).
noun( 'mass', 'masses', both, _ ).
noun( 'massacre', 'massacres', count, _ ).
noun( 'massage', 'massages', both, _ ).
noun( 'masseur', 'masseurs', count, _ ).
noun( 'masseuse', 'masseuses', count, _ ).
noun( 'massif', 'massifs', count, _ ).
noun( 'massiveness', '-', mass, _ ).
noun( 'mast', 'masts', both, _ ).
noun( 'master', 'masters', count, _ ).
noun( 'master-at-arms', 'masters-at-arms', count, _ ).
noun( 'master-key', 'master-keys', count, _ ).
noun( 'mastermind', 'masterminds', count, _ ).
noun( 'masterpiece', 'masterpieces', count, _ ).
noun( 'mastership', 'masterships', both, _ ).
noun( 'masterstroke', 'masterstrokes', count, _ ).
noun( 'mastery', '-', mass, _ ).
noun( 'masthead', 'mastheads', count, _ ).
noun( 'mastication', '-', mass, _ ).
noun( 'mastiff', 'mastiffs', count, _ ).
noun( 'mastodon', 'mastodons', count, _ ).
noun( 'mastoid', 'mastoids', count, _ ).
noun( 'mastoiditis', '-', mass, _ ).
noun( 'masturbation', '-', mass, _ ).
noun( 'mat', 'mats', count, _ ).
noun( 'mat_e', '-', mass, _ ).
noun( 'matador', 'matadors', count, _ ).
noun( 'match', 'matches', count, _ ).
noun( 'match-point', 'match-points', count, _ ).
noun( 'matchbox', 'matchboxes', count, _ ).
noun( 'matchet', 'matchets', count, _ ).
noun( 'matchlock', 'matchlocks', count, _ ).
noun( 'matchmaker', 'matchmakers', count, _ ).
noun( 'matchwood', '-', mass, _ ).
noun( 'mate', 'mates', count, _ ).
noun( 'material', 'materials', both, _ ).
noun( 'materialism', '-', mass, _ ).
noun( 'materialist', 'materialists', count, _ ).
noun( 'materialization', 'materializations', count, _ ).
noun( 'maternity', '-', mass, _ ).
noun( 'mathematician', 'mathematicians', count, _ ).
noun( 'mathematics', 'mathematics', mass, _ ).
noun( 'maths', 'maths', mass, _ ).
noun( 'matin_ee', 'matin_ees', count, _ ).
noun( 'matriarch', 'matriarchs', count, _ ).
noun( 'matriarchy', 'matriarchies', count, _ ).
noun( 'matric', 'matrics', count, _ ).
noun( 'matricide', 'matricides', both, _ ).
noun( 'matriculation', 'matriculations', both, _ ).
noun( 'matrimony', '-', mass, _ ).
noun( 'matrix', 'matrixes', count, _ ).
noun( 'matron', 'matrons', count, _ ).
noun( 'matter', 'matters', both, _ ).
noun( 'matting', '-', mass, _ ).
noun( 'mattock', 'mattocks', count, _ ).
noun( 'mattress', 'mattresses', count, _ ).
noun( 'maturation', '-', mass, _ ).
noun( 'maturity', '-', mass, _ ).
noun( 'maulstick', 'maulsticks', count, _ ).
noun( 'mausoleum', 'mausoleums', count, _ ).
noun( 'mauve', 'mauves', both, _ ).
noun( 'maverick', 'mavericks', count, _ ).
noun( 'mavis', 'mavises', count, _ ).
noun( 'maw', 'maws', count, _ ).
noun( 'mawkishness', '-', mass, _ ).
noun( 'max', '-', count, _ ).
noun( 'maxim', 'maxims', count, _ ).
noun( 'maximization', 'maximizations', both, _ ).
noun( 'maximum', 'maximums', count, _ ).
noun( 'may-beetle', 'may-beetles', count, _ ).
noun( 'may-bug', 'may-bugs', count, _ ).
noun( 'mayday', 'maydays', count, _ ).
noun( 'mayfly', 'mayflies', count, _ ).
noun( 'mayhem', '-', mass, _ ).
noun( 'mayonnaise', '-', mass, _ ).
noun( 'mayor', 'mayors', count, _ ).
noun( 'mayoralty', 'mayoralties', count, _ ).
noun( 'mayoress', 'mayoresses', count, _ ).
noun( 'maypole', 'maypoles', count, _ ).
noun( 'maze', 'mazes', count, _ ).
noun( 'mazurka', 'mazurkas', count, _ ).
noun( 'mead', 'meads', both, _ ).
noun( 'meadow', 'meadows', both, _ ).
noun( 'meagreness', '-', mass, _ ).
noun( 'meal', 'meals', both, _ ).
noun( 'mealie', 'mealies', count, _ ).
noun( 'mealtime', 'mealtimes', count, _ ).
noun( 'mealy-bug', 'mealy-bugs', count, _ ).
noun( 'mean', 'means', count, _ ).
noun( 'meanie', 'meanies', count, _ ).
noun( 'meaning', 'meanings', both, _ ).
noun( 'meanness', '-', mass, _ ).
noun( 'meantime', '-', mass, _ ).
noun( 'meany', 'meanies', count, _ ).
noun( 'measles', 'measles', mass, _ ).
noun( 'measure', 'measures', both, _ ).
noun( 'measurement', 'measurements', both, _ ).
noun( 'meat', 'meats', both, _ ).
noun( 'meat-safe', 'meat-safes', count, _ ).
noun( 'meatball', 'meatballs', count, _ ).
noun( 'mechanic', 'mechanics', count, _ ).
noun( 'mechanics', 'mechanics', mass, _ ).
noun( 'mechanism', 'mechanisms', both, _ ).
noun( 'mechanization', 'mechanizations', both, _ ).
noun( 'medal', 'medals', count, _ ).
noun( 'medalist', 'medalists', count, _ ).
noun( 'medallion', 'medallions', count, _ ).
noun( 'medallist', 'medallists', count, _ ).
noun( 'meddler', 'meddlers', count, _ ).
noun( 'median', 'medians', count, _ ).
noun( 'mediation', '-', mass, _ ).
noun( 'mediator', 'mediators', count, _ ).
noun( 'medic', 'medics', count, _ ).
noun( 'medical', 'medicals', count, _ ).
noun( 'medicament', 'medicaments', count, _ ).
noun( 'medication', 'medications', both, _ ).
noun( 'medicine', 'medicines', both, _ ).
noun( 'medicine-ball', 'medicine-balls', count, _ ).
noun( 'medicine-chest', 'medicine-chests', count, _ ).
noun( 'medicine-man', 'medicine-men', count, _ ).
noun( 'medico', 'medicos', count, _ ).
noun( 'mediocrity', 'mediocrities', both, _ ).
noun( 'meditation', 'meditations', both, _ ).
noun( 'medium', 'mediums', count, _ ).
noun( 'medlar', 'medlars', count, _ ).
noun( 'medley', 'medleys', count, _ ).
noun( 'meed', 'meeds', count, _ ).
noun( 'meekness', '-', mass, _ ).
noun( 'meerschaum', 'meerschaums', both, _ ).
noun( 'meet', 'meets', count, _ ).
noun( 'meeting', 'meetings', count, _ ).
noun( 'meeting-house', 'meeting-houses', count, _ ).
noun( 'meeting-place', 'meeting-places', count, _ ).
noun( 'megacycle', 'megacycles', count, _ ).
noun( 'megadeath', 'megadeaths', count, _ ).
noun( 'megalith', 'megaliths', count, _ ).
noun( 'megalomania', '-', mass, _ ).
noun( 'megalomaniac', 'megalomaniacs', count, _ ).
noun( 'megaphone', 'megaphones', count, _ ).
noun( 'megaton', 'megatons', count, _ ).
noun( 'megrim', 'megrims', count, _ ).
noun( 'meiosis', '-', mass, _ ).
noun( 'melancholia', '-', mass, _ ).
noun( 'melancholy', '-', mass, _ ).
noun( 'melioration', 'meliorations', both, _ ).
noun( 'meliorism', '-', mass, _ ).
noun( 'mellowness', '-', mass, _ ).
noun( 'melodiousness', '-', mass, _ ).
noun( 'melodrama', 'melodramas', both, _ ).
noun( 'melody', 'melodies', both, _ ).
noun( 'melon', 'melons', count, _ ).
noun( 'melting-point', 'melting-points', count, _ ).
noun( 'melting-pot', 'melting-pots', count, _ ).
noun( 'member', 'members', count, _ ).
noun( 'membership', '-', mass, _ ).
noun( 'membrane', 'membranes', both, _ ).
noun( 'memento', 'mementos', count, _ ).
noun( 'memo', 'memos', count, _ ).
noun( 'memoir', 'memoirs', count, _ ).
noun( 'memorandum', 'memorandums', count, _ ).
noun( 'memorial', 'memorials', count, _ ).
noun( 'memory', 'memories', both, _ ).
noun( 'memsahib', 'memsahibs', count, _ ).
noun( 'menace', 'menaces', both, _ ).
noun( 'menagerie', 'menageries', count, _ ).
noun( 'mend', 'mends', count, _ ).
noun( 'mendacity', 'mendacities', both, _ ).
noun( 'mender', 'menders', count, _ ).
noun( 'mendicant', 'mendicants', count, _ ).
noun( 'mending', '-', mass, _ ).
noun( 'menial', 'menials', count, _ ).
noun( 'meningitis', '-', mass, _ ).
noun( 'menopause', 'menopauses', count, _ ).
noun( 'menstruation', '-', mass, _ ).
noun( 'mensuration', 'mensurations', both, _ ).
noun( 'mentality', 'mentalities', both, _ ).
noun( 'menthol', '-', mass, _ ).
noun( 'mention', 'mentions', both, _ ).
noun( 'mentor', 'mentors', count, _ ).
noun( 'menu', 'menus', count, _ ).
noun( 'mercenary', 'mercenaries', count, _ ).
noun( 'mercer', 'mercers', count, _ ).
noun( 'merchandise', '-', mass, _ ).
noun( 'merchant', 'merchants', count, _ ).
noun( 'merchantman', 'merchantmen', count, _ ).
noun( 'mercury', '-', mass, _ ).
noun( 'mercy', 'mercies', both, _ ).
noun( 'mere', 'meres', count, _ ).
noun( 'meretriciousness', '-', mass, _ ).
noun( 'merger', 'mergers', both, _ ).
noun( 'meridian', 'meridians', count, _ ).
noun( 'meringue', 'meringues', both, _ ).
noun( 'merino', '-', mass, _ ).
noun( 'merino-sheep', 'merino-sheep', count, _ ).
noun( 'merit', 'merits', both, _ ).
noun( 'meritocracy', 'meritocracies', count, _ ).
noun( 'mermaid', 'mermaids', count, _ ).
noun( 'merman', 'mermen', count, _ ).
noun( 'merriment', '-', mass, _ ).
noun( 'merry-go-round', 'merry-go-rounds', count, _ ).
noun( 'merrymaker', 'merrymakers', count, _ ).
noun( 'merrymaking', '-', mass, _ ).
noun( 'mescal', 'mescals', both, _ ).
noun( 'mescaline', '-', mass, _ ).
noun( 'mesh', 'meshes', count, _ ).
noun( 'mesmerism', '-', mass, _ ).
noun( 'mesmerist', 'mesmerists', count, _ ).
noun( 'meson', 'mesons', count, _ ).
noun( 'mess', 'messes', both, _ ).
noun( 'mess-jacket', 'mess-jackets', count, _ ).
noun( 'mess-up', 'mess-ups', count, _ ).
noun( 'message', 'messages', count, _ ).
noun( 'messenger', 'messengers', count, _ ).
noun( 'messiness', '-', mass, _ ).
noun( 'messmate', 'messmates', count, _ ).
noun( 'messuage', 'messuages', count, _ ).
noun( 'metabolism', '-', mass, _ ).
noun( 'metacarpal', 'metacarpals', count, _ ).
noun( 'metal', 'metals', both, _ ).
noun( 'metallurgist', 'metallurgists', count, _ ).
noun( 'metallurgy', '-', mass, _ ).
noun( 'metalwork', 'metalworks', count, _ ).
noun( 'metalworker', 'metalworkers', count, _ ).
noun( 'metamorphosis', 'metamorphoses', count, _ ).
noun( 'metaphor', 'metaphors', both, _ ).
noun( 'metaphysics', 'metaphysics', mass, _ ).
noun( 'metatarsal', 'metatarsals', count, _ ).
noun( 'meteor', 'meteors', count, _ ).
noun( 'meteorite', 'meteorites', count, _ ).
noun( 'meteorologist', 'meteorologists', count, _ ).
noun( 'meteorology', '-', mass, _ ).
noun( 'meter', 'meters', count, _ ).
noun( 'methane', '-', mass, _ ).
noun( 'method', 'methods', both, _ ).
noun( 'methodology', 'methodologies', both, _ ).
noun( 'methyl', 'methyls', both, _ ).
noun( 'meticulousness', '-', mass, _ ).
noun( 'metre', 'metres', both, _ ).
noun( 'metrication', 'metrications', both, _ ).
noun( 'metronome', 'metronomes', count, _ ).
noun( 'metropolis', 'metropolises', count, _ ).
noun( 'metropolitan', 'metropolitans', count, _ ).
noun( 'mettle', '-', mass, _ ).
noun( 'mew', 'mews', count, _ ).
noun( 'mews', 'mews', count, _ ).
noun( 'mezzanine', 'mezzanines', count, _ ).
noun( 'mezzo-soprano', 'mezzo-sopranos', count, _ ).
noun( 'mezzotint', 'mezzotints', both, _ ).
noun( 'mg', 'mg', count, _ ).
noun( 'mi', '-', count, _ ).
noun( 'miaou', 'miaous', count, _ ).
noun( 'miaow', 'miaows', count, _ ).
noun( 'miasma', 'miasmas', count, _ ).
noun( 'mica', '-', mass, _ ).
noun( 'mickey', '-', count, _ ).
noun( 'mickle', '-', count, _ ).
noun( 'micro-organism', 'micro-organisms', count, _ ).
noun( 'microbe', 'microbes', count, _ ).
noun( 'microbiology', '-', mass, _ ).
noun( 'microcosm', 'microcosms', count, _ ).
noun( 'microdot', 'microdots', count, _ ).
noun( 'microelectronics', 'microelectronics', mass, _ ).
noun( 'microfiche', 'microfiches', both, _ ).
noun( 'microfilm', 'microfilms', both, _ ).
noun( 'micrometer', 'micrometers', count, _ ).
noun( 'micron', 'microns', count, _ ).
noun( 'microphone', 'microphones', count, _ ).
noun( 'microscope', 'microscopes', count, _ ).
noun( 'microscopy', '-', mass, _ ).
noun( 'microwave', 'microwaves', count, _ ).
noun( 'mid-off', '-', count, _ ).
noun( 'mid-on', '-', count, _ ).
noun( 'midday', '-', mass, _ ).
noun( 'midden', 'middens', count, _ ).
noun( 'middle', 'middles', count, _ ).
noun( 'middleman', 'middlemen', count, _ ).
noun( 'middleweight', 'middleweights', count, _ ).
noun( 'middling', 'middlings', count, _ ).
noun( 'middy', 'middies', count, _ ).
noun( 'midfield', '-', count, _ ).
noun( 'midge', 'midges', count, _ ).
noun( 'midget', 'midgets', count, _ ).
noun( 'midinette', 'midinettes', count, _ ).
noun( 'midland', 'midlands', count, _ ).
noun( 'midnight', '-', mass, _ ).
noun( 'midriff', 'midriffs', count, _ ).
noun( 'midshipman', 'midshipmen', count, _ ).
noun( 'midst', '-', count, _ ).
noun( 'midstream', '-', mass, _ ).
noun( 'midsummer', '-', mass, _ ).
noun( 'midwife', 'midwives', count, _ ).
noun( 'midwifery', '-', mass, _ ).
noun( 'mien', 'miens', count, _ ).
noun( 'might', '-', mass, _ ).
noun( 'might-have-been', 'might-have-beens', count, _ ).
noun( 'mignonette', '-', mass, _ ).
noun( 'migraine', 'migraines', count, _ ).
noun( 'migrant', 'migrants', count, _ ).
noun( 'migration', 'migrations', both, _ ).
noun( 'mikado', 'mikados', count, _ ).
noun( 'mike', 'mikes', count, _ ).
noun( 'milady', 'miladies', count, _ ).
noun( 'milage', 'milages', count, _ ).
noun( 'mildew', '-', mass, _ ).
noun( 'mildness', '-', mass, _ ).
noun( 'mile', 'miles', count, _ ).
noun( 'mileage', 'mileages', count, _ ).
noun( 'mileometer', 'mileometers', count, _ ).
noun( 'miler', 'milers', count, _ ).
noun( 'milestone', 'milestones', count, _ ).
noun( 'milieu', 'milieus', count, _ ).
noun( 'militancy', '-', mass, _ ).
noun( 'militant', 'militants', count, _ ).
noun( 'militarism', '-', mass, _ ).
noun( 'militarist', 'militarists', count, _ ).
noun( 'military', '-', mass, _ ).
noun( 'militia', 'militias', count, _ ).
noun( 'militiaman', 'militiamen', count, _ ).
noun( 'milk', '-', mass, _ ).
noun( 'milk-churn', 'milk-churns', count, _ ).
noun( 'milk-powder', 'milk-powders', both, _ ).
noun( 'milk-shake', 'milk-shakes', count, _ ).
noun( 'milk-tooth', 'milk-teeth', count, _ ).
noun( 'milkbar', 'milkbars', count, _ ).
noun( 'milking-machine', 'milking-machines', count, _ ).
noun( 'milkmaid', 'milkmaids', count, _ ).
noun( 'milkman', 'milkmen', count, _ ).
noun( 'milkshake', 'milkshakes', count, _ ).
noun( 'milksop', 'milksops', count, _ ).
noun( 'milkweed', 'milkweeds', both, _ ).
noun( 'mill', 'mills', count, _ ).
noun( 'mill-dam', 'mill-dams', count, _ ).
noun( 'mill-girl', 'mill-girls', count, _ ).
noun( 'mill-hand', 'mill-hands', count, _ ).
noun( 'millboard', '-', mass, _ ).
noun( 'millenarian', 'millenarians', count, _ ).
noun( 'millennium', 'millennia', count, _ ).
noun( 'millepede', 'millepedes', count, _ ).
noun( 'miller', 'millers', count, _ ).
noun( 'millet', '-', mass, _ ).
noun( 'milliard', 'milliards', count, _ ).
noun( 'millibar', 'millibars', count, _ ).
noun( 'milligram', 'milligrams', count, _ ).
noun( 'millimetre', 'millimetres', count, _ ).
noun( 'milliner', 'milliners', count, _ ).
noun( 'millinery', '-', mass, _ ).
noun( 'million', 'millions', count, _ ).
noun( 'millionaire', 'millionaires', count, _ ).
noun( 'millionairess', 'millionairesss', count, _ ).
noun( 'millionth', 'millionths', count, _ ).
noun( 'millipede', 'millipedes', count, _ ).
noun( 'millpond', 'millponds', count, _ ).
noun( 'millrace', 'millraces', count, _ ).
noun( 'millstone', 'millstones', count, _ ).
noun( 'millwheel', 'millwheels', count, _ ).
noun( 'millwright', 'millwrights', count, _ ).
noun( 'milometer', 'milometers', count, _ ).
noun( 'milord', 'milords', count, _ ).
noun( 'milt', '-', mass, _ ).
noun( 'mime', 'mimes', both, _ ).
noun( 'mimeograph', 'mimeographs', count, _ ).
noun( 'mimic', 'mimics', count, _ ).
noun( 'mimicry', '-', mass, _ ).
noun( 'mimosa', 'mimosas', both, _ ).
noun( 'min', '-', count, _ ).
noun( 'minaret', 'minarets', count, _ ).
noun( 'mince', '-', mass, _ ).
noun( 'mince-pie', 'mince-pies', count, _ ).
noun( 'mincemeat', '-', mass, _ ).
noun( 'mincer', 'mincers', count, _ ).
noun( 'mind', 'minds', both, _ ).
noun( 'mind-reader', 'mind-readers', count, _ ).
noun( 'minder', 'minders', count, _ ).
noun( 'mindfulness', '-', mass, _ ).
noun( 'mindlessness', '-', mass, _ ).
noun( 'mine', 'mines', count, _ ).
noun( 'mine-detector', 'mine-detectors', count, _ ).
noun( 'mine-disposal', 'mine-disposals', both, _ ).
noun( 'minefield', 'minefields', count, _ ).
noun( 'minelayer', 'minelayers', count, _ ).
noun( 'minelaying', '-', mass, _ ).
noun( 'miner', 'miners', count, _ ).
noun( 'mineral', 'minerals', count, _ ).
noun( 'mineralogist', 'mineralogists', count, _ ).
noun( 'mineralogy', '-', mass, _ ).
noun( 'minestrone', '-', mass, _ ).
noun( 'minesweeper', 'minesweepers', count, _ ).
noun( 'minesweeping', '-', mass, _ ).
noun( 'mineworker', 'mineworkers', count, _ ).
noun( 'miniature', 'miniatures', both, _ ).
noun( 'miniaturist', 'miniaturists', count, _ ).
noun( 'miniaturization', '-', mass, _ ).
noun( 'minim', 'minims', count, _ ).
noun( 'minimum', 'minimums', count, _ ).
noun( 'mining', '-', mass, _ ).
noun( 'minion', 'minions', count, _ ).
noun( 'minister', 'ministers', count, _ ).
noun( 'ministrant', 'ministrants', count, _ ).
noun( 'ministration', 'ministrations', both, _ ).
noun( 'ministry', 'ministries', count, _ ).
noun( 'miniver', '-', mass, _ ).
noun( 'mink', 'minks', both, _ ).
noun( 'minnow', 'minnows', count, _ ).
noun( 'minor', 'minors', count, _ ).
noun( 'minority', 'minorities', both, _ ).
noun( 'minster', 'minsters', count, _ ).
noun( 'minstrel', 'minstrels', count, _ ).
noun( 'minstrelsy', '-', mass, _ ).
noun( 'mint', 'mints', both, _ ).
noun( 'minuet', 'minuets', count, _ ).
noun( 'minus', 'minuses', count, _ ).
noun( 'minute', 'minutes', count, _ ).
noun( 'minute-book', 'minute-books', count, _ ).
noun( 'minute-gun', 'minute-guns', count, _ ).
noun( 'minute-hand', 'minute-hands', count, _ ).
noun( 'minuteman', 'minutemen', count, _ ).
noun( 'minuteness', '-', mass, _ ).
noun( 'minx', 'minxes', count, _ ).
noun( 'miracle', 'miracles', count, _ ).
noun( 'mirage', 'mirages', count, _ ).
noun( 'mire', '-', mass, _ ).
noun( 'mirror', 'mirrors', count, _ ).
noun( 'mirth', '-', mass, _ ).
noun( 'misadventure', 'misadventures', both, _ ).
noun( 'misalliance', 'misalliances', count, _ ).
noun( 'misanthrope', 'misanthropes', count, _ ).
noun( 'misanthropy', '-', mass, _ ).
noun( 'misapplication', 'misapplications', count, _ ).
noun( 'misapprehension', 'misapprehensions', both, _ ).
noun( 'misappropriation', 'misappropriations', both, _ ).
noun( 'misbehaviour', '-', mass, _ ).
noun( 'misc', '-', proper, _ ).
noun( 'miscalculation', 'miscalculations', both, _ ).
noun( 'miscarriage', 'miscarriages', both, _ ).
noun( 'miscegenation', '-', mass, _ ).
noun( 'miscellany', 'miscellanies', count, _ ).
noun( 'mischance', 'mischances', both, _ ).
noun( 'mischief', 'mischiefs', both, _ ).
noun( 'mischief-maker', 'mischief-makers', count, _ ).
noun( 'mischief-making', '-', mass, _ ).
noun( 'mischievousness', '-', mass, _ ).
noun( 'misconception', 'misconceptions', both, _ ).
noun( 'misconduct', '-', mass, _ ).
noun( 'misconstruction', 'misconstructions', both, _ ).
noun( 'miscount', 'miscounts', count, _ ).
noun( 'miscreant', 'miscreants', count, _ ).
noun( 'misdeal', 'misdeals', count, _ ).
noun( 'misdeed', 'misdeeds', count, _ ).
noun( 'misdemeanour', 'misdemeanours', count, _ ).
noun( 'misdirection', 'misdirections', count, _ ).
noun( 'misdoing', 'misdoings', count, _ ).
noun( 'mise en sc`ene', '-', mass, _ ).
noun( 'miser', 'misers', count, _ ).
noun( 'miserliness', '-', mass, _ ).
noun( 'misery', 'miseries', both, _ ).
noun( 'misfire', 'misfires', count, _ ).
noun( 'misfit', 'misfits', count, _ ).
noun( 'misfortune', 'misfortunes', both, _ ).
noun( 'misgiving', 'misgivings', both, _ ).
noun( 'misgovernment', '-', mass, _ ).
noun( 'mishap', 'mishaps', both, _ ).
noun( 'mishmash', '-', mass, _ ).
noun( 'misinformation', '-', mass, _ ).
noun( 'misinterpretation', 'misinterpretations', both, _ ).
noun( 'mismanagement', '-', mass, _ ).
noun( 'misnomer', 'misnomers', count, _ ).
noun( 'misogynist', 'misogynists', count, _ ).
noun( 'misprint', 'misprints', count, _ ).
noun( 'mispronunciation', 'mispronunciations', count, _ ).
noun( 'misquotation', 'misquotations', both, _ ).
noun( 'misrepresentation', 'misrepresentations', both, _ ).
noun( 'misrule', '-', mass, _ ).
noun( 'miss', 'misses', count, _ ).
noun( 'missal', 'missals', count, _ ).
noun( 'missile', 'missiles', count, _ ).
noun( 'mission', 'missions', count, _ ).
noun( 'missionary', 'missionaries', count, _ ).
noun( 'missis', '-', count, _ ).
noun( 'missive', 'missives', count, _ ).
noun( 'misspelling', 'misspellings', count, _ ).
noun( 'misstatement', 'misstatements', count, _ ).
noun( 'missus', '-', count, _ ).
noun( 'missy', 'missies', count, _ ).
noun( 'mist', 'mists', both, _ ).
noun( 'mistake', 'mistakes', count, _ ).
noun( 'mister', '-', count, _ ).
noun( 'mistiness', '-', mass, _ ).
noun( 'mistletoe', '-', mass, _ ).
noun( 'mistral', 'mistrals', count, _ ).
noun( 'mistranslation', 'mistranslations', both, _ ).
noun( 'mistress', 'mistresses', count, _ ).
noun( 'mistrial', 'mistrials', count, _ ).
noun( 'mistrust', '-', mass, _ ).
noun( 'misunderstanding', 'misunderstandings', both, _ ).
noun( 'misuse', 'misuses', both, _ ).
noun( 'mite', 'mites', count, _ ).
noun( 'mitigation', '-', mass, _ ).
noun( 'mitre', 'mitres', count, _ ).
noun( 'mitre-joint', 'mitre-joints', count, _ ).
noun( 'mitt', 'mitts', count, _ ).
noun( 'mitten', 'mittens', count, _ ).
noun( 'mix', 'mixes', count, _ ).
noun( 'mix-up', 'mix-ups', count, _ ).
noun( 'mixer', 'mixers', count, _ ).
noun( 'mixture', 'mixtures', both, _ ).
noun( 'mizen', 'mizens', count, _ ).
noun( 'mizzen', 'mizzens', count, _ ).
noun( 'mizzen-mast', 'mizzen-masts', count, _ ).
noun( 'ml', 'ml', count, _ ).
noun( 'mm', 'mm', count, _ ).
noun( 'mnemonics', 'mnemonics', mass, _ ).
noun( 'mo', '-', count, _ ).
noun( 'moan', 'moans', count, _ ).
noun( 'moat', 'moats', count, _ ).
noun( 'mob', 'mobs', count, _ ).
noun( 'mobcap', 'mobcaps', count, _ ).
noun( 'mobile', 'mobiles', count, _ ).
noun( 'mobility', '-', mass, _ ).
noun( 'mobilization', 'mobilizations', both, _ ).
noun( 'mobster', 'mobsters', count, _ ).
noun( 'moccasin', 'moccasins', both, _ ).
noun( 'mocha', '-', mass, _ ).
noun( 'mock', 'mocks', count, _ ).
noun( 'mock-up', 'mock-ups', count, _ ).
noun( 'mocker', 'mockers', count, _ ).
noun( 'mockery', 'mockeries', both, _ ).
noun( 'mod', 'mods', count, _ ).
noun( '-', 'mod cons', count, _ ).
noun( 'modality', 'modalities', both, _ ).
noun( 'mode', 'modes', count, _ ).
noun( 'model', 'models', count, _ ).
noun( 'modeler', 'modelers', count, _ ).
noun( 'modeller', 'modellers', count, _ ).
noun( 'modelling', '-', mass, _ ).
noun( 'moderate', 'moderates', count, _ ).
noun( 'moderation', '-', mass, _ ).
noun( 'moderator', 'moderators', count, _ ).
noun( 'modern', 'moderns', count, _ ).
noun( 'modernism', '-', mass, _ ).
noun( 'modernist', 'modernists', count, _ ).
noun( 'modernity', '-', mass, _ ).
noun( 'modernization', '-', mass, _ ).
noun( 'modesty', '-', mass, _ ).
noun( 'modicum', 'modica', count, _ ).
noun( 'modification', 'modifications', both, _ ).
noun( 'modifier', 'modifiers', count, _ ).
noun( 'modiste', 'modistes', count, _ ).
noun( 'modulation', 'modulations', both, _ ).
noun( 'module', 'modules', count, _ ).
noun( 'modus operandi', '-', count, _ ).
noun( 'modus vivendi', '-', count, _ ).
noun( 'mogul', 'moguls', count, _ ).
noun( 'mohair', '-', mass, _ ).
noun( 'moiety', 'moieties', count, _ ).
noun( 'moisture', '-', mass, _ ).
noun( 'moke', 'mokes', count, _ ).
noun( 'molar', 'molars', count, _ ).
noun( 'molasses', '-', mass, _ ).
noun( 'mole', 'moles', count, _ ).
noun( 'molecule', 'molecules', count, _ ).
noun( 'molehill', 'molehills', count, _ ).
noun( 'moleskin', 'moleskins', count, _ ).
noun( 'molestation', '-', mass, _ ).
noun( 'moll', 'molls', count, _ ).
noun( 'mollification', '-', mass, _ ).
noun( 'mollusc', 'molluscs', count, _ ).
noun( 'mollycoddle', 'mollycoddles', count, _ ).
noun( 'molybdenum', '-', mass, _ ).
noun( 'moment', 'moments', both, _ ).
noun( 'momentousness', '-', mass, _ ).
noun( 'momentum', '-', mass, _ ).
noun( 'monarch', 'monarchs', count, _ ).
noun( 'monarchism', '-', mass, _ ).
noun( 'monarchist', 'monarchists', count, _ ).
noun( 'monarchy', 'monarchies', both, _ ).
noun( 'monastery', 'monasteries', count, _ ).
noun( 'monasticism', '-', mass, _ ).
noun( 'money', '-', mass, _ ).
noun( 'money-grubber', 'money-grubbers', count, _ ).
noun( 'money-order', 'money-orders', count, _ ).
noun( 'money-spinner', 'money-spinners', count, _ ).
noun( 'moneybox', 'moneyboxes', count, _ ).
noun( 'moneychanger', 'moneychangers', count, _ ).
noun( 'moneylender', 'moneylenders', count, _ ).
noun( 'monger', 'mongers', count, _ ).
noun( 'mongol', 'mongols', count, _ ).
noun( 'mongolism', '-', mass, _ ).
noun( 'mongoose', 'mongooses', count, _ ).
noun( 'mongrel', 'mongrels', count, _ ).
noun( 'monitor', 'monitors', count, _ ).
noun( 'monk', 'monks', count, _ ).
noun( 'monkey', 'monkeys', count, _ ).
noun( 'monkey-jacket', 'monkey-jackets', count, _ ).
noun( 'monkey-nut', 'monkey-nuts', count, _ ).
noun( 'monkey-puzzle', 'monkey-puzzles', count, _ ).
noun( 'monkey-wrench', 'monkey-wrenches', count, _ ).
noun( 'monochrome', 'monochromes', count, _ ).
noun( 'monocle', 'monocles', count, _ ).
noun( 'monogamist', 'monogamists', count, _ ).
noun( 'monogamy', '-', mass, _ ).
noun( 'monogram', 'monograms', count, _ ).
noun( 'monograph', 'monographs', count, _ ).
noun( 'monolith', 'monoliths', count, _ ).
noun( 'monologue', 'monologues', count, _ ).
noun( 'monomania', 'monomanias', both, _ ).
noun( 'monomaniac', 'monomaniacs', count, _ ).
noun( 'monoplane', 'monoplanes', count, _ ).
noun( 'monopolist', 'monopolists', count, _ ).
noun( 'monopolization', 'monopolizations', both, _ ).
noun( 'monopoly', 'monopolies', count, _ ).
noun( 'monorail', 'monorails', count, _ ).
noun( 'monosyllable', 'monosyllables', count, _ ).
noun( 'monotheism', '-', mass, _ ).
noun( 'monotheist', 'monotheists', count, _ ).
noun( 'monotone', 'monotones', count, _ ).
noun( 'monotony', '-', mass, _ ).
noun( 'monotype', 'monotypes', count, _ ).
noun( 'monoxide', 'monoxides', both, _ ).
noun( 'monsoon', 'monsoons', count, _ ).
noun( 'monster', 'monsters', count, _ ).
noun( 'monstrance', 'monstrances', count, _ ).
noun( 'monstrosity', 'monstrosities', both, _ ).
noun( 'montage', '-', mass, _ ).
noun( 'month', 'months', count, _ ).
noun( 'monthly', 'monthlies', count, _ ).
noun( 'monument', 'monuments', count, _ ).
noun( 'moo', 'moos', count, _ ).
noun( 'moo-cow', 'moo-cows', count, _ ).
noun( 'mood', 'moods', count, _ ).
noun( 'moodiness', '-', mass, _ ).
noun( 'moon', 'moons', count, _ ).
noun( 'moonbeam', 'moonbeams', count, _ ).
noun( 'moonflower', 'moonflowers', count, _ ).
noun( 'moonlight', '-', mass, _ ).
noun( 'moonshine', '-', mass, _ ).
noun( 'moonstone', 'moonstones', count, _ ).
noun( 'moor', 'moors', both, _ ).
noun( 'moorcock', 'moorcocks', count, _ ).
noun( 'moorfowl', 'moorfowls', count, _ ).
noun( 'moorgame', 'moorgame', count, _ ).
noun( 'moorhen', 'moorhens', count, _ ).
noun( 'mooring-mast', 'mooring-masts', count, _ ).
noun( 'moorland', 'moorlands', both, _ ).
noun( 'moose', 'moose', count, _ ).
noun( 'mop', 'mops', count, _ ).
noun( 'mope', 'mopes', count, _ ).
noun( 'moped', 'mopeds', count, _ ).
noun( 'moquette', '-', mass, _ ).
noun( 'moraine', 'moraines', count, _ ).
noun( 'moral', 'morals', count, _ ).
noun( 'morale', '-', mass, _ ).
noun( 'moralist', 'moralists', count, _ ).
noun( 'morality', 'moralities', both, _ ).
noun( 'morass', 'morasses', count, _ ).
noun( 'moratorium', 'moratoriums', count, _ ).
noun( 'morbidity', 'morbidities', both, _ ).
noun( 'morbidness', '-', mass, _ ).
noun( 'more', '-', mass, _ ).
noun( 'morello', 'morellos', count, _ ).
noun( 'morgue', 'morgues', count, _ ).
noun( 'morn', 'morns', count, _ ).
noun( 'morning', 'mornings', both, _ ).
noun( 'morning-glory', 'morning-glories', count, _ ).
noun( 'morning-room', 'morning-rooms', count, _ ).
noun( 'morocco', '-', mass, _ ).
noun( 'moron', 'morons', count, _ ).
noun( 'moroseness', '-', mass, _ ).
noun( 'morpheme', 'morphemes', count, _ ).
noun( 'morphia', '-', mass, _ ).
noun( 'morphine', '-', mass, _ ).
noun( 'morphology', '-', mass, _ ).
noun( 'morris dance', 'morris dances', count, _ ).
noun( 'morrow', 'morrows', count, _ ).
noun( 'morsel', 'morsels', count, _ ).
noun( 'mortal', 'mortals', count, _ ).
noun( 'mortality', '-', mass, _ ).
noun( 'mortar', 'mortars', both, _ ).
noun( 'mortarboard', 'mortarboards', count, _ ).
noun( 'mortgage', 'mortgages', count, _ ).
noun( 'mortgagee', 'mortgagees', count, _ ).
noun( 'mortgagor', 'mortgagors', count, _ ).
noun( 'mortice', 'mortices', count, _ ).
noun( 'mortician', 'morticians', count, _ ).
noun( 'mortification', '-', mass, _ ).
noun( 'mortise', 'mortises', count, _ ).
noun( 'mortuary', 'mortuaries', count, _ ).
noun( 'mosaic', 'mosaics', count, _ ).
noun( 'moselle', 'moselles', both, _ ).
noun( 'mosque', 'mosques', count, _ ).
noun( 'mosquito', 'mosquitoes', count, _ ).
noun( 'mosquito-craft', 'mosquito-craft', count, _ ).
noun( 'mosquito-net', 'mosquito-nets', count, _ ).
noun( 'moss', 'mosses', both, _ ).
noun( 'most', 'most', count, _ ).
noun( 'mote', 'motes', count, _ ).
noun( 'motel', 'motels', count, _ ).
noun( 'moth', 'moths', count, _ ).
noun( 'mothball', 'mothballs', count, _ ).
noun( 'mother', 'mothers', count, _ ).
noun( 'mother-in-law', 'mothers-in-law', count, _ ).
noun( 'mother-of-pearl', '-', mass, _ ).
noun( 'motherhood', '-', mass, _ ).
noun( 'motherland', 'motherlands', count, _ ).
noun( 'motherliness', '-', mass, _ ).
noun( 'motif', 'motifs', count, _ ).
noun( 'motion', 'motions', both, _ ).
noun( 'motivation', 'motivations', both, _ ).
noun( 'motive', 'motives', count, _ ).
noun( 'motley', '-', mass, _ ).
noun( 'motor', 'motors', count, _ ).
noun( 'motorbike', 'motorbikes', count, _ ).
noun( 'motorboat', 'motorboats', count, _ ).
noun( 'motorcade', 'motorcades', count, _ ).
noun( 'motorcar', 'motorcars', count, _ ).
noun( 'motorcoach', 'motorcoaches', count, _ ).
noun( 'motorcycle', 'motorcycles', count, _ ).
noun( 'motorist', 'motorists', count, _ ).
noun( 'motorman', 'motormen', count, _ ).
noun( 'motorway', 'motorways', count, _ ).
noun( 'motto', 'mottos', count, _ ).
noun( 'moujik', 'moujiks', count, _ ).
noun( 'mould', 'moulds', both, _ ).
noun( 'moulding', 'mouldings', both, _ ).
noun( 'moult', 'moults', count, _ ).
noun( 'mound', 'mounds', count, _ ).
noun( 'mount', 'mounts', count, _ ).
noun( 'mountain', 'mountains', count, _ ).
noun( 'mountaineer', 'mountaineers', count, _ ).
noun( 'mountaineering', '-', mass, _ ).
noun( 'mountebank', 'mountebanks', count, _ ).
noun( 'mourner', 'mourners', count, _ ).
noun( 'mourning', '-', mass, _ ).
noun( 'mourning-band', 'mourning-bands', count, _ ).
noun( 'mourning-ring', 'mourning-rings', count, _ ).
noun( 'mouse', 'mice', count, _ ).
noun( 'mouser', 'mousers', count, _ ).
noun( 'mousetrap', 'mousetraps', count, _ ).
noun( 'mousse', 'mousses', both, _ ).
noun( 'moustache', 'moustaches', count, _ ).
noun( 'mouth', 'mouths', count, _ ).
noun( 'mouth-organ', 'mouth-organs', count, _ ).
noun( 'mouthful', 'mouthfuls', count, _ ).
noun( 'mouthpiece', 'mouthpieces', count, _ ).
noun( 'move', 'moves', count, _ ).
noun( 'movement', 'movements', both, _ ).
noun( 'mover', 'movers', count, _ ).
noun( 'movie', 'movies', count, _ ).
noun( 'mow', 'mows', count, _ ).
noun( 'mower', 'mowers', count, _ ).
noun( 'mpg', 'mpg', count, _ ).
noun( 'mph', 'mph', count, _ ).
noun( 'much', '-', count, _ ).
noun( 'muchness', '-', mass, _ ).
noun( 'mucilage', '-', mass, _ ).
noun( 'muck', '-', mass, _ ).
noun( 'muck-heap', 'muck-heaps', count, _ ).
noun( 'muckle', '-', count, _ ).
noun( 'muckraker', 'muckrakers', count, _ ).
noun( 'muckraking', '-', mass, _ ).
noun( 'mucus', '-', mass, _ ).
noun( 'mud', '-', mass, _ ).
noun( 'mud-bath', 'mud-baths', count, _ ).
noun( 'muddle', 'muddles', count, _ ).
noun( 'mudguard', 'mudguards', count, _ ).
noun( 'mudslinger', 'mudslingers', count, _ ).
noun( 'muesli', '-', mass, _ ).
noun( 'muezzin', 'muezzins', count, _ ).
noun( 'muff', 'muffs', count, _ ).
noun( 'muffin', 'muffins', count, _ ).
noun( 'muffin-man', 'muffin-men', count, _ ).
noun( 'muffler', 'mufflers', count, _ ).
noun( 'mufti', '-', mass, _ ).
noun( 'mug', 'mugs', count, _ ).
noun( 'mugger', 'muggers', count, _ ).
noun( 'mugginess', '-', mass, _ ).
noun( 'mugging', 'muggings', both, _ ).
noun( 'muggins', 'mugginses', count, _ ).
noun( 'mugwump', 'mugwumps', count, _ ).
noun( 'mulatto', 'mulattos', count, _ ).
noun( 'mulberry', 'mulberries', count, _ ).
noun( 'mulch', 'mulches', count, _ ).
noun( 'mule', 'mules', count, _ ).
noun( 'muleteer', 'muleteers', count, _ ).
noun( 'mulishness', '-', mass, _ ).
noun( 'mull', 'mulls', count, _ ).
noun( 'mullah', 'mullahs', count, _ ).
noun( 'mullein', 'mulleins', count, _ ).
noun( 'mullet', 'mullets', count, _ ).
noun( 'mulligatawny', '-', mass, _ ).
noun( 'mullion', 'mullions', count, _ ).
noun( 'multiple', 'multiples', count, _ ).
noun( 'multiplication', 'multiplications', both, _ ).
noun( 'multiplicity', '-', mass, _ ).
noun( 'multitude', 'multitudes', both, _ ).
noun( 'multum in parvo', '-', mass, _ ).
noun( 'mum', 'mums', count, _ ).
noun( 'mumbo-jumbo', '-', mass, _ ).
noun( 'mummer', 'mummers', count, _ ).
noun( 'mummery', 'mummeries', both, _ ).
noun( 'mummification', 'mummifications', both, _ ).
noun( 'mummy', 'mummies', count, _ ).
noun( 'mumps', 'mumps', mass, _ ).
noun( 'municipality', 'municipalities', count, _ ).
noun( 'munificence', '-', mass, _ ).
noun( 'munition', 'munitions', count, _ ).
noun( 'mural', 'murals', count, _ ).
noun( 'murder', 'murders', both, _ ).
noun( 'murderer', 'murderers', count, _ ).
noun( 'murderess', 'murderesses', count, _ ).
noun( 'murk', '-', mass, _ ).
noun( 'murmur', 'murmurs', count, _ ).
noun( 'murphy', 'murphies', count, _ ).
noun( 'murrain', '-', mass, _ ).
noun( 'muscatel', '-', mass, _ ).
noun( 'muscle', 'muscles', both, _ ).
noun( 'muscleman', 'musclemen', count, _ ).
noun( 'muse', 'muses', count, _ ).
noun( 'museum', 'museums', count, _ ).
noun( 'mush', '-', mass, _ ).
noun( 'mushroom', 'mushrooms', count, _ ).
noun( 'music', '-', mass, _ ).
noun( 'music-box', 'music-boxes', count, _ ).
noun( 'music-hall', 'music-halls', count, _ ).
noun( 'music-stand', 'music-stands', count, _ ).
noun( 'music-stool', 'music-stools', count, _ ).
noun( 'musical', 'musicals', count, _ ).
noun( 'musical-box', 'musical-boxes', count, _ ).
noun( 'musician', 'musicians', count, _ ).
noun( 'musicianship', '-', mass, _ ).
noun( 'musk', '-', mass, _ ).
noun( 'musk-deer', 'musk-deer', count, _ ).
noun( 'musk-rose', 'musk-roses', count, _ ).
noun( 'musket', 'muskets', count, _ ).
noun( 'musketeer', 'musketeers', count, _ ).
noun( 'musketry', '-', mass, _ ).
noun( 'muskrat', 'muskrats', count, _ ).
noun( 'muslin', '-', mass, _ ).
noun( 'musquash', '-', mass, _ ).
noun( 'muss', 'musses', both, _ ).
noun( 'mussel', 'mussels', count, _ ).
noun( 'must', 'musts', both, _ ).
noun( 'mustachio', 'mustachios', count, _ ).
noun( 'mustang', 'mustangs', count, _ ).
noun( 'mustard', '-', mass, _ ).
noun( 'muster', 'musters', count, _ ).
noun( 'mustiness', '-', mass, _ ).
noun( 'mutability', '-', mass, _ ).
noun( 'mutation', 'mutations', both, _ ).
noun( 'mute', 'mutes', count, _ ).
noun( 'mutilation', 'mutilations', both, _ ).
noun( 'mutineer', 'mutineers', count, _ ).
noun( 'mutiny', 'mutinies', both, _ ).
noun( 'mutt', 'mutts', count, _ ).
noun( 'mutter', 'mutters', count, _ ).
noun( 'mutterer', 'mutterers', count, _ ).
noun( 'mutton', '-', mass, _ ).
noun( 'mutton-head', 'mutton-heads', count, _ ).
noun( 'muzzle', 'muzzles', count, _ ).
noun( 'muzzle-velocity', 'muzzle-velocities', count, _ ).
noun( 'mycology', '-', mass, _ ).
noun( 'myelitis', '-', mass, _ ).
noun( 'myna', 'mynas', count, _ ).
noun( 'mynah', 'mynahs', count, _ ).
noun( 'myopia', '-', mass, _ ).
noun( 'myriad', 'myriads', count, _ ).
noun( 'myrmidon', 'myrmidons', count, _ ).
noun( 'myrrh', '-', mass, _ ).
noun( 'myrtle', 'myrtles', count, _ ).
noun( 'mystery', 'mysteries', both, _ ).
noun( 'mystic', 'mystics', count, _ ).
noun( 'mysticism', '-', mass, _ ).
noun( 'mystification', 'mystifications', both, _ ).
noun( 'mystique', 'mystiques', count, _ ).
noun( 'myth', 'myths', both, _ ).
noun( 'mythologist', 'mythologists', count, _ ).
noun( 'mythology', 'mythologies', both, _ ).
noun( 'myxomatosis', '-', mass, _ ).
noun( 'n', '-', count, _ ).
noun( 'n_eglig_e', 'n_eglig_es', both, _ ).
noun( 'nabob', 'nabobs', count, _ ).
noun( 'nacelle', 'nacelles', count, _ ).
noun( 'nacre', '-', mass, _ ).
noun( 'nadir', 'nadirs', count, _ ).
noun( 'nag', 'nags', count, _ ).
noun( 'nagger', 'naggers', count, _ ).
noun( 'naiad', 'naiads', count, _ ).
noun( 'nail', 'nails', count, _ ).
noun( 'nail-polish', 'nail-polishes', both, _ ).
noun( 'nail-varnish', 'nail-varnishes', both, _ ).
noun( 'nailbrush', 'nailbrushes', count, _ ).
noun( 'nailfile', 'nailfiles', count, _ ).
noun( 'nainsook', '-', mass, _ ).
noun( 'naira', 'nairas', count, _ ).
noun( 'naivet_e', 'naivet_es', both, _ ).
noun( 'naivety', 'naiveties', both, _ ).
noun( 'nakedness', '-', mass, _ ).
noun( 'namby-pamby', 'namby-pambies', count, _ ).
noun( 'name', 'names', count, _ ).
noun( 'name-day', 'name-days', count, _ ).
noun( 'name-dropping', '-', mass, _ ).
noun( 'name-part', 'name-parts', count, _ ).
noun( 'nameplate', 'nameplates', count, _ ).
noun( 'namesake', 'namesakes', count, _ ).
noun( 'nankeen', '-', mass, _ ).
noun( 'nanny', 'nannies', count, _ ).
noun( 'nanny-goat', 'nanny-goats', count, _ ).
noun( 'nap', 'naps', both, _ ).
noun( 'napalm', '-', mass, _ ).
noun( 'nape', 'napes', count, _ ).
noun( 'napery', '-', mass, _ ).
noun( 'naphtha', '-', mass, _ ).
noun( 'naphthalene', '-', mass, _ ).
noun( 'napkin', 'napkins', count, _ ).
noun( 'napkin-ring', 'napkin-rings', count, _ ).
noun( 'nappy', 'nappies', count, _ ).
noun( 'narcissism', '-', mass, _ ).
noun( 'narcissus', 'narcissuses', count, _ ).
noun( 'narcotic', 'narcotics', count, _ ).
noun( 'nark', 'narks', count, _ ).
noun( 'narration', 'narrations', both, _ ).
noun( 'narrative', 'narratives', both, _ ).
noun( 'narrator', 'narrators', count, _ ).
noun( 'narrow', 'narrows', count, _ ).
noun( 'narrow-mindedness', '-', mass, _ ).
noun( 'narrowness', '-', mass, _ ).
noun( 'narwhal', 'narwhals', count, _ ).
noun( 'nasal', 'nasals', count, _ ).
noun( 'nastiness', '-', mass, _ ).
noun( 'nasturtium', 'nasturtiums', count, _ ).
noun( 'nation', 'nations', count, _ ).
noun( 'national', 'nationals', count, _ ).
noun( 'nationalism', '-', mass, _ ).
noun( 'nationalist', 'nationalists', count, _ ).
noun( 'nationality', 'nationalities', both, _ ).
noun( 'nationalization', 'nationalizations', both, _ ).
noun( 'native', 'natives', count, _ ).
noun( 'nativity', 'nativities', count, _ ).
noun( 'natural', 'naturals', count, _ ).
noun( 'naturalism', '-', mass, _ ).
noun( 'naturalist', 'naturalists', count, _ ).
noun( 'naturalization', '-', mass, _ ).
noun( 'naturalness', '-', mass, _ ).
noun( 'nature', 'natures', both, _ ).
noun( 'naturism', '-', mass, _ ).
noun( 'naturist', 'naturists', count, _ ).
noun( 'naught', 'naughts', both, _ ).
noun( 'naughtiness', '-', mass, _ ).
noun( 'nausea', '-', mass, _ ).
noun( 'nautch', 'nautches', count, _ ).
noun( 'nautch-girl', 'nautch-girls', count, _ ).
noun( 'nautilus', 'nautiluses', count, _ ).
noun( 'nave', 'naves', count, _ ).
noun( 'navel', 'navels', count, _ ).
noun( 'navigability', '-', mass, _ ).
noun( 'navigation', '-', mass, _ ).
noun( 'navigator', 'navigators', count, _ ).
noun( 'navvy', 'navvies', count, _ ).
noun( 'navy', 'navies', count, _ ).
noun( 'ne plus ultra', '-', count, _ ).
noun( 'ne\'er-do-well', 'ne\'er-do-wells', count, _ ).
noun( 'neap', 'neaps', count, _ ).
noun( 'neap-tide', 'neap-tides', count, _ ).
noun( 'nearness', '-', mass, _ ).
noun( 'nearside', 'nearsides', count, _ ).
noun( 'neatness', '-', mass, _ ).
noun( 'nebula', 'nebulas', count, _ ).
noun( 'necessary', 'necessaries', count, _ ).
noun( 'necessity', 'necessities', both, _ ).
noun( 'neck', 'necks', count, _ ).
noun( 'neckband', 'neckbands', count, _ ).
noun( 'neckcloth', 'neckcloths', count, _ ).
noun( 'neckerchief', 'neckerchiefs', count, _ ).
noun( 'necklace', 'necklaces', count, _ ).
noun( 'necklet', 'necklets', count, _ ).
noun( 'neckline', 'necklines', count, _ ).
noun( 'necktie', 'neckties', count, _ ).
noun( 'neckwear', '-', mass, _ ).
noun( 'necromancer', 'necromancers', count, _ ).
noun( 'necromancy', '-', mass, _ ).
noun( 'necropolis', 'necropolises', count, _ ).
noun( 'nectar', '-', mass, _ ).
noun( 'nectarine', 'nectarines', count, _ ).
noun( 'need', 'needs', both, _ ).
noun( 'needle', 'needles', count, _ ).
noun( 'needlecraft', '-', mass, _ ).
noun( 'needlewoman', 'needlewomen', count, _ ).
noun( 'needlework', '-', mass, _ ).
noun( 'nefariousness', '-', mass, _ ).
noun( 'negation', '-', mass, _ ).
noun( 'negative', 'negatives', count, _ ).
noun( 'neglect', '-', mass, _ ).
noun( 'neglectfulness', '-', mass, _ ).
noun( 'negligee', 'negligees', both, _ ).
noun( 'negligence', '-', mass, _ ).
noun( 'negotiation', 'negotiations', both, _ ).
noun( 'negotiator', 'negotiators', count, _ ).
noun( 'negus', '-', mass, _ ).
noun( 'neigh', 'neighs', count, _ ).
noun( 'neighbour', 'neighbours', count, _ ).
noun( 'neighbourhood', 'neighbourhoods', count, _ ).
noun( 'neighbourliness', '-', mass, _ ).
noun( 'nemesis', 'nemeses', count, _ ).
noun( 'neocolonialism', '-', mass, _ ).
noun( 'neologism', 'neologisms', both, _ ).
noun( 'neon', '-', mass, _ ).
noun( 'neophyte', 'neophytes', count, _ ).
noun( 'neoplasm', 'neoplasms', count, _ ).
noun( 'nephew', 'nephews', count, _ ).
noun( 'nephritis', '-', mass, _ ).
noun( 'nepotism', '-', mass, _ ).
noun( 'nereid', 'nereids', count, _ ).
noun( 'nerve', 'nerves', both, _ ).
noun( 'nerve-cell', 'nerve-cells', count, _ ).
noun( 'nerve-centre', 'nerve-centres', count, _ ).
noun( 'nervousness', '-', mass, _ ).
noun( 'nescience', '-', mass, _ ).
noun( 'ness', 'nesses', count, _ ).
noun( 'nest', 'nests', count, _ ).
noun( 'nest-egg', 'nest-eggs', count, _ ).
noun( 'nestling', 'nestlings', count, _ ).
noun( 'net', 'nets', both, _ ).
noun( 'netball', '-', mass, _ ).
noun( 'netting', '-', mass, _ ).
noun( 'nettle', 'nettles', count, _ ).
noun( 'nettlerash', '-', mass, _ ).
noun( 'network', 'networks', count, _ ).
noun( 'neuralgia', '-', mass, _ ).
noun( 'neurasthenia', '-', mass, _ ).
noun( 'neurasthenic', 'neurasthenics', count, _ ).
noun( 'neuritis', '-', mass, _ ).
noun( 'neurologist', 'neurologists', count, _ ).
noun( 'neurology', '-', mass, _ ).
noun( 'neurosis', 'neuroses', count, _ ).
noun( 'neurotic', 'neurotics', count, _ ).
noun( 'neuter', 'neuters', count, _ ).
noun( 'neutral', 'neutrals', count, _ ).
noun( 'neutrality', '-', mass, _ ).
noun( 'neutralization', 'neutralizations', both, _ ).
noun( 'neutron', 'neutrons', count, _ ).
noun( 'newcomer', 'newcomers', count, _ ).
noun( 'newel', 'newels', count, _ ).
noun( 'newlywed', 'newlyweds', count, _ ).
noun( 'newness', '-', mass, _ ).
noun( 'news', '-', mass, _ ).
noun( 'newsagent', 'newsagents', count, _ ).
noun( 'newsboy', 'newsboys', count, _ ).
noun( 'newscast', 'newscasts', count, _ ).
noun( 'newscaster', 'newscasters', count, _ ).
noun( 'newsdealer', 'newsdealers', count, _ ).
noun( 'newsflash', 'newsflashes', count, _ ).
noun( 'newsletter', 'newsletters', count, _ ).
noun( 'newsmonger', 'newsmongers', count, _ ).
noun( 'newspaper', 'newspapers', count, _ ).
noun( 'newspaperman', 'newspapermen', count, _ ).
noun( 'newsprint', '-', mass, _ ).
noun( 'newsreel', 'newsreels', count, _ ).
noun( 'newsroom', 'newsrooms', count, _ ).
noun( 'newssheet', 'newssheets', count, _ ).
noun( 'newsstand', 'newsstands', count, _ ).
noun( 'newsvendor', 'newsvendors', count, _ ).
noun( 'newt', 'newts', count, _ ).
noun( 'next', 'next', count, _ ).
noun( 'nexus', 'nexuses', count, _ ).
noun( 'nib', 'nibs', count, _ ).
noun( 'nibble', 'nibbles', count, _ ).
noun( 'niceness', '-', mass, _ ).
noun( 'nicety', 'niceties', both, _ ).
noun( 'niche', 'niches', count, _ ).
noun( 'nick', 'nicks', count, _ ).
noun( 'nickel', 'nickels', both, _ ).
noun( 'nicknack', 'nicknacks', count, _ ).
noun( 'nickname', 'nicknames', count, _ ).
noun( 'nicotine', '-', mass, _ ).
noun( 'niece', 'nieces', count, _ ).
noun( 'niff', 'niffs', count, _ ).
noun( 'niggard', 'niggards', count, _ ).
noun( 'niggardliness', '-', mass, _ ).
noun( 'nigger', 'niggers', count, _ ).
noun( 'night', 'nights', both, _ ).
noun( 'night-bell', 'night-bells', count, _ ).
noun( 'night-bird', 'night-birds', count, _ ).
noun( 'night-light', 'night-lights', count, _ ).
noun( 'night-line', 'night-lines', count, _ ).
noun( 'night-porter', 'night-porters', count, _ ).
noun( 'night-soil', '-', mass, _ ).
noun( 'night-stop', 'night-stops', count, _ ).
noun( 'night-time', '-', mass, _ ).
noun( 'night-watch', 'night-watches', count, _ ).
noun( 'night-watchman', 'night-watchmen', count, _ ).
noun( 'nightcap', 'nightcaps', count, _ ).
noun( 'nightclub', 'nightclubs', count, _ ).
noun( 'nightdress', 'nightdresses', count, _ ).
noun( 'nightfall', '-', mass, _ ).
noun( 'nightgown', 'nightgowns', count, _ ).
noun( 'nightie', 'nighties', count, _ ).
noun( 'nightingale', 'nightingales', count, _ ).
noun( 'nightjar', 'nightjars', count, _ ).
noun( 'nightmare', 'nightmares', count, _ ).
noun( 'nightshade', '-', mass, _ ).
noun( 'nightshirt', 'nightshirts', count, _ ).
noun( 'nightwork', '-', mass, _ ).
noun( 'nihilism', '-', mass, _ ).
noun( 'nihilist', 'nihilists', count, _ ).
noun( 'nil', '-', mass, _ ).
noun( 'nimbleness', '-', mass, _ ).
noun( 'nimbus', 'nimbuses', count, _ ).
noun( 'nincompoop', 'nincompoops', count, _ ).
noun( 'nine', 'nines', count, _ ).
noun( 'ninepence', 'ninepences', count, _ ).
noun( 'ninepins', 'ninepins', mass, _ ).
noun( 'nineteen', 'nineteens', count, _ ).
noun( 'nineteenth', 'nineteenths', count, _ ).
noun( 'ninetieth', 'ninetieths', count, _ ).
noun( 'ninety', 'nineties', count, _ ).
noun( 'ninny', 'ninnies', count, _ ).
noun( 'ninth', 'ninths', count, _ ).
noun( 'nip', 'nips', count, _ ).
noun( 'nipper', 'nippers', count, _ ).
noun( 'nipple', 'nipples', count, _ ).
noun( 'nirvana', 'nirvanas', count, _ ).
noun( 'nit', 'nits', count, _ ).
noun( 'nitrate', 'nitrates', count, _ ).
noun( 'nitre', '-', mass, _ ).
noun( 'nitrochalk', '-', mass, _ ).
noun( 'nitrogen', '-', mass, _ ).
noun( 'nitroglycerin', '-', mass, _ ).
noun( 'nitroglycerine', '-', mass, _ ).
noun( 'nitty-gritty', '-', mass, _ ).
noun( 'nitwit', 'nitwits', count, _ ).
noun( 'nix', '-', mass, _ ).
noun( 'no', '-', count, _ ).
noun( 'no', 'nos', count, _ ).
noun( 'no-ball', 'no-balls', count, _ ).
noun( 'no-man\'s-land', '-', count, _ ).
noun( 'nob', 'nobs', count, _ ).
noun( 'nobility', '-', mass, _ ).
noun( 'noble', 'nobles', count, _ ).
noun( 'noble-mindedness', '-', mass, _ ).
noun( 'nobleman', 'noblemen', count, _ ).
noun( 'noblesse', '-', count, _ ).
noun( 'noblesse oblige', '-', mass, _ ).
noun( 'nobody', 'nobodies', count, _ ).
noun( 'noctambulist', 'noctambulists', count, _ ).
noun( 'nocturne', 'nocturnes', count, _ ).
noun( 'nod', 'nods', count, _ ).
noun( 'noddle', 'noddles', count, _ ).
noun( 'node', 'nodes', count, _ ).
noun( 'nodule', 'nodules', count, _ ).
noun( 'noggin', 'noggins', count, _ ).
noun( 'noise', 'noises', both, _ ).
noun( 'noiselessness', '-', mass, _ ).
noun( 'noisiness', '-', mass, _ ).
noun( 'nom de plume', 'noms de plume', count, _ ).
noun( 'nomad', 'nomads', count, _ ).
noun( 'nomenclature', 'nomenclatures', count, _ ).
noun( 'nomination', 'nominations', both, _ ).
noun( 'nominative', 'nominatives', count, _ ).
noun( 'nominee', 'nominees', count, _ ).
noun( 'non sequitur', 'non sequiturs', count, _ ).
noun( 'non-compliance', '-', mass, _ ).
noun( 'non-interference', '-', mass, _ ).
noun( 'nonage', '-', mass, _ ).
noun( 'nonagenarian', 'nonagenarians', count, _ ).
noun( 'nonaggression', '-', mass, _ ).
noun( 'nonalignment', '-', mass, _ ).
noun( 'nonce', 'nonces', count, _ ).
noun( 'nonce-word', 'nonce-words', count, _ ).
noun( 'nonchalance', '-', mass, _ ).
noun( 'noncombatant', 'noncombatants', count, _ ).
noun( 'nonconductor', 'nonconductors', count, _ ).
noun( 'nonconformist', 'nonconformists', count, _ ).
noun( 'nonconformity', '-', mass, _ ).
noun( 'nondescript', 'nondescripts', count, _ ).
noun( 'nonentity', 'nonentities', count, _ ).
noun( 'nonesuch', '-', count, _ ).
noun( 'nonevent', 'nonevents', count, _ ).
noun( 'nonfiction', '-', mass, _ ).
noun( 'nonintervention', '-', mass, _ ).
noun( 'nonobservance', '-', mass, _ ).
noun( 'nonpareil', 'nonpareils', count, _ ).
noun( 'nonpayment', 'nonpayments', both, _ ).
noun( 'nonsense', 'nonsenses', both, _ ).
noun( 'nonsmoker', 'nonsmokers', count, _ ).
noun( 'nonstarter', 'nonstarters', count, _ ).
noun( 'nonsuch', '-', count, _ ).
noun( 'nonviolence', '-', mass, _ ).
noun( 'noodle', 'noodles', count, _ ).
noun( 'nook', 'nooks', count, _ ).
noun( 'noon', '-', mass, _ ).
noun( 'noontide', '-', mass, _ ).
noun( 'noose', 'nooses', count, _ ).
noun( 'nor\'-east', '-', mass, _ ).
noun( 'nor\'-nor\'-east', '-', mass, _ ).
noun( 'nor\'-nor\'-west', '-', mass, _ ).
noun( 'nor\'-west', '-', mass, _ ).
noun( 'norm', 'norms', count, _ ).
noun( 'normal', '-', mass, _ ).
noun( 'normalcy', '-', mass, _ ).
noun( 'normality', '-', mass, _ ).
noun( 'normalization', 'normalizations', both, _ ).
noun( 'north', '-', mass, _ ).
noun( 'north-northeast', '-', mass, _ ).
noun( 'north-northwest', '-', mass, _ ).
noun( 'northeast', '-', mass, _ ).
noun( 'northeaster', 'northeasters', count, _ ).
noun( 'northerner', 'northerners', count, _ ).
noun( 'northwest', '-', mass, _ ).
noun( 'northwester', 'northwesters', count, _ ).
noun( '-', 'nos', count, _ ).
noun( 'nose', 'noses', count, _ ).
noun( 'nose-flute', 'nose-flutes', count, _ ).
noun( 'nose-wheel', 'nose-wheels', count, _ ).
noun( 'nosebag', 'nosebags', count, _ ).
noun( 'nosebleed', 'nosebleeds', count, _ ).
noun( 'nosecone', 'nosecones', count, _ ).
noun( 'nosedive', 'nosedives', count, _ ).
noun( 'nosegay', 'nosegays', count, _ ).
noun( 'nosering', 'noserings', count, _ ).
noun( 'nosh', '-', mass, _ ).
noun( 'nosh-up', 'nosh-ups', count, _ ).
noun( 'nostalgia', '-', mass, _ ).
noun( 'nostril', 'nostrils', count, _ ).
noun( 'nostrum', 'nostrums', count, _ ).
noun( 'notability', 'notabilities', both, _ ).
noun( 'notable', 'notables', count, _ ).
noun( 'notary', 'notaries', count, _ ).
noun( 'notation', 'notations', both, _ ).
noun( 'notch', 'notches', count, _ ).
noun( 'note', 'notes', both, _ ).
noun( 'notebook', 'notebooks', count, _ ).
noun( 'notecase', 'notecases', count, _ ).
noun( 'notepaper', '-', mass, _ ).
noun( 'nothing', 'nothings', count, _ ).
noun( 'nothingness', '-', mass, _ ).
noun( 'notice', 'notices', both, _ ).
noun( 'notice-board', 'notice-boards', count, _ ).
noun( 'notification', 'notifications', both, _ ).
noun( 'notion', 'notions', count, _ ).
noun( 'notoriety', '-', mass, _ ).
noun( 'nougat', 'nougats', both, _ ).
noun( 'nought', 'noughts', count, _ ).
noun( 'noun', 'nouns', count, _ ).
noun( 'nourishment', '-', mass, _ ).
noun( 'nous', '-', mass, _ ).
noun( 'nouveau riche', 'nouveaux riches', count, _ ).
noun( 'nova', 'novas', count, _ ).
noun( 'novel', 'novels', count, _ ).
noun( 'novelette', 'novelettes', count, _ ).
noun( 'novelist', 'novelists', count, _ ).
noun( 'novelty', 'novelties', both, _ ).
noun( 'novice', 'novices', count, _ ).
noun( 'noviciate', 'noviciates', count, _ ).
noun( 'novitiate', 'novitiates', count, _ ).
noun( 'noxiousness', '-', mass, _ ).
noun( 'nozzle', 'nozzles', count, _ ).
noun( 'nr', '-', proper, _ ).
noun( 'nuance', 'nuances', count, _ ).
noun( 'nub', 'nubs', count, _ ).
noun( 'nucleus', 'nuclei', count, _ ).
noun( 'nude', 'nudes', count, _ ).
noun( 'nudge', 'nudges', count, _ ).
noun( 'nudism', '-', mass, _ ).
noun( 'nudist', 'nudists', count, _ ).
noun( 'nudity', '-', mass, _ ).
noun( 'nugget', 'nuggets', count, _ ).
noun( 'nuisance', 'nuisances', count, _ ).
noun( 'nullification', '-', mass, _ ).
noun( 'nullity', '-', mass, _ ).
noun( 'number', 'numbers', count, _ ).
noun( 'numberplate', 'numberplates', count, _ ).
noun( 'numbness', '-', mass, _ ).
noun( 'numeracy', '-', mass, _ ).
noun( 'numeral', 'numerals', count, _ ).
noun( 'numeration', 'numerations', both, _ ).
noun( 'numerator', 'numerators', count, _ ).
noun( 'numismatics', 'numismatics', mass, _ ).
noun( 'numismatist', 'numismatists', count, _ ).
noun( 'numskull', 'numskulls', count, _ ).
noun( 'nun', 'nuns', count, _ ).
noun( 'nuncio', 'nuncios', count, _ ).
noun( 'nunnery', 'nunneries', count, _ ).
noun( 'nurse', 'nurses', both, _ ).
noun( 'nurseling', 'nurselings', count, _ ).
noun( 'nursemaid', 'nursemaids', count, _ ).
noun( 'nursery', 'nurseries', count, _ ).
noun( 'nurseryman', 'nurserymen', count, _ ).
noun( 'nursing-home', 'nursing-homes', count, _ ).
noun( 'nursling', 'nurslings', count, _ ).
noun( 'nurture', '-', mass, _ ).
noun( 'nut', 'nuts', count, _ ).
noun( 'nut-butter', 'nut-butters', both, _ ).
noun( 'nuthouse', 'nuthouses', count, _ ).
noun( 'nutmeg', 'nutmegs', both, _ ).
noun( 'nutria', '-', mass, _ ).
noun( 'nutrient', 'nutrients', count, _ ).
noun( 'nutriment', 'nutriments', count, _ ).
noun( 'nutrition', '-', mass, _ ).
noun( 'nutshell', 'nutshells', count, _ ).
noun( 'nylon', 'nylons', both, _ ).
noun( 'nymph', 'nymphs', count, _ ).
noun( 'nymphet', 'nymphets', count, _ ).
noun( 'nympho', 'nymphos', count, _ ).
noun( 'nymphomania', '-', mass, _ ).
noun( 'nymphomaniac', 'nymphomaniacs', count, _ ).
noun( 'o', '-', count, _ ).
noun( 'oaf', 'oafs', count, _ ).
noun( 'oak', 'oaks', both, _ ).
noun( 'oak-apple', 'oak-apples', count, _ ).
noun( 'oakum', '-', mass, _ ).
noun( 'oar', 'oars', count, _ ).
noun( 'oarsman', 'oarsmen', count, _ ).
noun( 'oarsmanship', '-', mass, _ ).
noun( 'oarswoman', 'oarswomen', count, _ ).
noun( 'oasis', 'oases', count, _ ).
noun( 'oast', 'oasts', count, _ ).
noun( 'oasthouse', 'oasthouses', count, _ ).
noun( 'oat', 'oats', count, _ ).
noun( 'oatcake', 'oatcakes', count, _ ).
noun( 'oath', 'oaths', count, _ ).
noun( 'oatmeal', '-', mass, _ ).
noun( 'ob', '-', proper, _ ).
noun( 'obbligato', 'obbligatos', count, _ ).
noun( 'obduracy', '-', mass, _ ).
noun( 'obedience', '-', mass, _ ).
noun( 'obeisance', 'obeisances', both, _ ).
noun( 'obelisk', 'obelisks', count, _ ).
noun( 'obesity', '-', mass, _ ).
noun( 'obi', 'obis', count, _ ).
noun( 'obiter dictum', '-', count, _ ).
noun( 'obituary', 'obituaries', count, _ ).
noun( 'object', 'objects', count, _ ).
noun( 'objection', 'objections', both, _ ).
noun( 'objective', 'objectives', count, _ ).
noun( 'objectivity', '-', mass, _ ).
noun( 'objector', 'objectors', count, _ ).
noun( 'objurgation', 'objurgations', both, _ ).
noun( 'oblation', 'oblations', count, _ ).
noun( 'obligation', 'obligations', count, _ ).
noun( 'obliquity', 'obliquities', both, _ ).
noun( 'obliteration', '-', mass, _ ).
noun( 'oblivion', '-', mass, _ ).
noun( 'oblong', 'oblongs', count, _ ).
noun( 'obloquy', '-', mass, _ ).
noun( 'obnoxiousness', '-', mass, _ ).
noun( 'oboe', 'oboes', count, _ ).
noun( 'oboist', 'oboists', count, _ ).
noun( 'obscenity', 'obscenities', both, _ ).
noun( 'obscurantism', '-', mass, _ ).
noun( 'obscurantist', 'obscurantists', count, _ ).
noun( 'obscurity', 'obscurities', both, _ ).
noun( 'obsequiousness', '-', mass, _ ).
noun( 'observance', 'observances', both, _ ).
noun( 'observation', 'observations', both, _ ).
noun( 'observatory', 'observatories', count, _ ).
noun( 'observer', 'observers', count, _ ).
noun( 'obsession', 'obsessions', both, _ ).
noun( 'obsidian', '-', mass, _ ).
noun( 'obsolescence', '-', mass, _ ).
noun( 'obstacle', 'obstacles', count, _ ).
noun( 'obstetrician', 'obstetricians', count, _ ).
noun( 'obstetrics', 'obstetrics', mass, _ ).
noun( 'obstinacy', '-', mass, _ ).
noun( 'obstreperousness', '-', mass, _ ).
noun( 'obstruction', 'obstructions', both, _ ).
noun( 'obstructionism', '-', mass, _ ).
noun( 'obstructionist', 'obstructionists', count, _ ).
noun( 'obtuseness', '-', mass, _ ).
noun( 'obverse', 'obverses', count, _ ).
noun( 'obviousness', '-', mass, _ ).
noun( 'ocarina', 'ocarinas', count, _ ).
noun( 'occasion', 'occasions', both, _ ).
noun( 'occult', '-', count, _ ).
noun( 'occupancy', 'occupancies', count, _ ).
noun( 'occupant', 'occupants', count, _ ).
noun( 'occupation', 'occupations', both, _ ).
noun( 'occupier', 'occupiers', count, _ ).
noun( 'occurrence', 'occurrences', both, _ ).
noun( 'ocean', 'oceans', count, _ ).
noun( 'ochre', '-', mass, _ ).
noun( 'octagon', 'octagons', count, _ ).
noun( 'octane', '-', mass, _ ).
noun( 'octave', 'octaves', count, _ ).
noun( 'octavo', 'octavos', count, _ ).
noun( 'octet', 'octets', count, _ ).
noun( 'octette', 'octettes', count, _ ).
noun( 'octogenarian', 'octogenarians', count, _ ).
noun( 'octopus', 'octopuses', count, _ ).
noun( 'octroi', 'octrois', count, _ ).
noun( 'oculist', 'oculists', count, _ ).
noun( 'odalisque', 'odalisques', count, _ ).
noun( 'oddity', 'oddities', both, _ ).
noun( 'oddment', 'oddments', count, _ ).
noun( 'ode', 'odes', count, _ ).
noun( 'odium', '-', mass, _ ).
noun( 'odour', 'odours', both, _ ).
noun( 'odyssey', 'odysseys', count, _ ).
noun( 'oesophagus', 'oesophaguses', count, _ ).
noun( 'off-day', 'off-days', count, _ ).
noun( 'off-licence', 'off-licences', count, _ ).
noun( 'offal', '-', mass, _ ).
noun( 'offence', 'offences', both, _ ).
noun( 'offender', 'offenders', count, _ ).
noun( 'offensive', 'offensives', count, _ ).
noun( 'offensiveness', '-', mass, _ ).
noun( 'offer', 'offers', count, _ ).
noun( 'offering', 'offerings', both, _ ).
noun( 'offertory', 'offertories', count, _ ).
noun( 'office', 'offices', count, _ ).
noun( 'office-bearer', 'office-bearers', count, _ ).
noun( 'office-block', 'office-blocks', count, _ ).
noun( 'office-boy', 'office-boys', count, _ ).
noun( 'office-holder', 'office-holders', count, _ ).
noun( 'officer', 'officers', count, _ ).
noun( 'official', 'officials', count, _ ).
noun( 'officialdom', 'officialdoms', count, _ ).
noun( 'officialese', '-', mass, _ ).
noun( 'officiousness', '-', mass, _ ).
noun( 'offing', '-', count, _ ).
noun( 'offprint', 'offprints', count, _ ).
noun( 'offset', '-', mass, _ ).
noun( 'offshoot', 'offshoots', count, _ ).
noun( 'offspring', 'offspring', count, _ ).
noun( 'ogre', 'ogres', count, _ ).
noun( 'ogress', 'ogresses', count, _ ).
noun( 'ohm', 'ohms', count, _ ).
noun( 'oil', 'oils', both, _ ).
noun( 'oil-burner', 'oil-burners', count, _ ).
noun( 'oil-cake', '-', mass, _ ).
noun( 'oil-painting', 'oil-paintings', both, _ ).
noun( 'oil-palm', 'oil-palms', count, _ ).
noun( 'oil-paper', '-', mass, _ ).
noun( 'oil-rig', 'oil-rigs', count, _ ).
noun( 'oil-silk', '-', mass, _ ).
noun( 'oil-slick', 'oil-slicks', count, _ ).
noun( 'oil-tanker', 'oil-tankers', count, _ ).
noun( 'oil-well', 'oil-wells', count, _ ).
noun( 'oilcan', 'oilcans', count, _ ).
noun( 'oilcloth', '-', mass, _ ).
noun( 'oiler', 'oilers', count, _ ).
noun( 'oilfield', 'oilfields', count, _ ).
noun( 'oiliness', '-', mass, _ ).
noun( 'oilskin', 'oilskins', both, _ ).
noun( 'ointment', 'ointments', both, _ ).
noun( 'okapi', 'okapis', count, _ ).
noun( 'okay', '-', count, _ ).
noun( 'okra', '-', mass, _ ).
noun( 'old', '-', mass, _ ).
noun( 'old-timer', 'old-timers', count, _ ).
noun( 'oldster', 'oldsters', count, _ ).
noun( 'oleander', 'oleanders', count, _ ).
noun( 'oligarch', 'oligarchs', count, _ ).
noun( 'oligarchy', 'oligarchies', both, _ ).
noun( 'olive', 'olives', count, _ ).
noun( 'olive-tree', 'olive-trees', count, _ ).
noun( 'ombudsman', 'ombudsmen', count, _ ).
noun( 'omega', 'omegas', count, _ ).
noun( 'omelet', 'omelets', count, _ ).
noun( 'omelette', 'omelettes', count, _ ).
noun( 'omen', 'omens', both, _ ).
noun( 'omission', 'omissions', both, _ ).
noun( 'omnibus', 'omnibuses', count, _ ).
noun( 'omnipotence', '-', mass, _ ).
noun( 'omniscience', '-', mass, _ ).
noun( 'on-licence', 'on-licences', count, _ ).
noun( 'oncoming', '-', count, _ ).
noun( 'one', 'ones', count, _ ).
noun( 'one-step', 'one-steps', count, _ ).
noun( 'one-upmanship', '-', mass, _ ).
noun( 'oneness', '-', mass, _ ).
noun( 'onion', 'onions', both, _ ).
noun( 'onlooker', 'onlookers', count, _ ).
noun( 'onomatopoeia', '-', mass, _ ).
noun( 'onrush', 'onrushes', count, _ ).
noun( 'onset', 'onsets', count, _ ).
noun( 'onslaught', 'onslaughts', count, _ ).
noun( 'ontology', 'ontologies', both, _ ).
noun( 'onus', 'onera', count, _ ).
noun( 'onyx', '-', mass, _ ).
noun( 'oomph', '-', mass, _ ).
noun( 'ooze', '-', mass, _ ).
noun( 'op', '-', proper, _ ).
noun( 'op art', '-', mass, _ ).
noun( 'op cit', '-', proper, _ ).
noun( 'opacity', '-', mass, _ ).
noun( 'opal', 'opals', count, _ ).
noun( 'opaqueness', '-', mass, _ ).
noun( 'open', '-', mass, _ ).
noun( 'opener', 'openers', count, _ ).
noun( 'opening', 'openings', count, _ ).
noun( 'openness', '-', mass, _ ).
noun( 'openwork', '-', mass, _ ).
noun( 'opera', 'operas', both, _ ).
noun( 'opera-cloak', 'opera-cloaks', count, _ ).
noun( 'opera-hat', 'opera-hats', count, _ ).
noun( 'opera-house', 'opera-houses', count, _ ).
noun( 'operating-table', 'operating-tables', count, _ ).
noun( 'operating-theatre', 'operating-theatres', count, _ ).
noun( 'operation', 'operations', both, _ ).
noun( 'operative', 'operatives', count, _ ).
noun( 'operator', 'operators', count, _ ).
noun( 'operetta', 'operettas', count, _ ).
noun( 'ophthalmia', '-', mass, _ ).
noun( 'ophthalmoscope', 'ophthalmoscopes', count, _ ).
noun( 'opiate', 'opiates', count, _ ).
noun( 'opinion', 'opinions', both, _ ).
noun( 'opium', '-', mass, _ ).
noun( 'opium-den', 'opium-dens', count, _ ).
noun( 'opossum', 'opossums', count, _ ).
noun( 'opp', '-', proper, _ ).
noun( 'opponent', 'opponents', count, _ ).
noun( 'opportunism', '-', mass, _ ).
noun( 'opportunist', 'opportunists', count, _ ).
noun( 'opportunity', 'opportunities', both, _ ).
noun( 'opposite', 'opposites', count, _ ).
noun( 'opposition', '-', mass, _ ).
noun( 'oppression', 'oppressions', both, _ ).
noun( 'oppressor', 'oppressors', count, _ ).
noun( 'opprobrium', '-', mass, _ ).
noun( 'optative', 'optatives', count, _ ).
noun( 'optician', 'opticians', count, _ ).
noun( 'optics', 'optics', mass, _ ).
noun( 'optimism', '-', mass, _ ).
noun( 'optimist', 'optimists', count, _ ).
noun( 'optimum', 'optimums', count, _ ).
noun( 'option', 'options', both, _ ).
noun( 'opulence', '-', mass, _ ).
noun( 'opus', 'opera', count, _ ).
noun( 'oracle', 'oracles', count, _ ).
noun( 'oral', 'orals', count, _ ).
noun( 'orang-outan', 'orang-outans', count, _ ).
noun( 'orang-outang', 'orang-outangs', count, _ ).
noun( 'orang-utan', 'orang-utans', count, _ ).
noun( 'orange', 'oranges', both, _ ).
noun( 'orangeade', '-', mass, _ ).
noun( 'oration', 'orations', count, _ ).
noun( 'orator', 'orators', count, _ ).
noun( 'oratorio', 'oratorios', both, _ ).
noun( 'oratory', 'oratories', both, _ ).
noun( 'orb', 'orbs', count, _ ).
noun( 'orbit', 'orbits', count, _ ).
noun( 'orchard', 'orchards', count, _ ).
noun( 'orchestra', 'orchestras', count, _ ).
noun( 'orchestration', 'orchestrations', count, _ ).
noun( 'orchid', 'orchids', count, _ ).
noun( 'orchis', 'orchises', count, _ ).
noun( 'ordeal', 'ordeals', count, _ ).
noun( 'order', 'orders', both, _ ).
noun( 'order-book', 'order-books', count, _ ).
noun( 'order-form', 'order-forms', count, _ ).
noun( 'order-paper', 'order-papers', count, _ ).
noun( 'ordering', 'orderings', count, _ ).
noun( 'orderliness', '-', mass, _ ).
noun( 'orderly', 'orderlies', count, _ ).
noun( 'ordinal', 'ordinals', count, _ ).
noun( 'ordinance', 'ordinances', count, _ ).
noun( 'ordinand', 'ordinands', count, _ ).
noun( 'ordination', 'ordinations', both, _ ).
noun( 'ordnance', '-', mass, _ ).
noun( 'ordure', '-', mass, _ ).
noun( 'ore', 'ores', both, _ ).
noun( 'organ', 'organs', count, _ ).
noun( 'organ-blower', 'organ-blowers', count, _ ).
noun( 'organ-grinder', 'organ-grinders', count, _ ).
noun( 'organ-loft', 'organ-lofts', count, _ ).
noun( 'organdie', '-', mass, _ ).
noun( 'organism', 'organisms', count, _ ).
noun( 'organist', 'organists', count, _ ).
noun( 'organization', 'organizations', both, _ ).
noun( 'organizer', 'organizers', count, _ ).
noun( 'orgasm', 'orgasms', count, _ ).
noun( 'orgy', 'orgies', count, _ ).
noun( 'oriel', 'oriels', count, _ ).
noun( 'orient', '-', count, _ ).
noun( 'oriental', 'orientals', count, _ ).
noun( 'orientalist', 'orientalists', count, _ ).
noun( 'orientation', '-', mass, _ ).
noun( 'orifice', 'orifices', count, _ ).
noun( 'origin', 'origins', both, _ ).
noun( 'original', 'originals', count, _ ).
noun( 'originality', '-', mass, _ ).
noun( 'originator', 'originators', count, _ ).
noun( 'oriole', 'orioles', count, _ ).
noun( 'orison', 'orisons', count, _ ).
noun( 'orlop', 'orlops', count, _ ).
noun( 'ormolu', 'ormolus', both, _ ).
noun( 'ornament', 'ornaments', both, _ ).
noun( 'ornamentation', '-', mass, _ ).
noun( 'ornateness', '-', mass, _ ).
noun( 'ornithologist', 'ornithologists', count, _ ).
noun( 'ornithology', '-', mass, _ ).
noun( 'orphan', 'orphans', count, _ ).
noun( 'orphanage', 'orphanages', count, _ ).
noun( 'orrisroot', '-', mass, _ ).
noun( 'orthodontics', 'orthodontics', mass, _ ).
noun( 'orthodontist', 'orthodontists', count, _ ).
noun( 'orthodoxy', 'orthodoxies', both, _ ).
noun( 'orthography', 'orthographies', both, _ ).
noun( 'orthopaedics', 'orthopaedics', mass, _ ).
noun( 'orthopedics', 'orthopedics', mass, _ ).
noun( 'ortolan', 'ortolans', count, _ ).
noun( 'oryx', 'oryxes', count, _ ).
noun( 'oscillation', 'oscillations', both, _ ).
noun( 'oscillator', 'oscillators', count, _ ).
noun( 'oscillograph', 'oscillographs', count, _ ).
noun( 'oscilloscope', 'oscilloscopes', count, _ ).
noun( 'osier', 'osiers', count, _ ).
noun( 'osprey', 'ospreys', count, _ ).
noun( 'ossification', '-', mass, _ ).
noun( 'ostentation', '-', mass, _ ).
noun( 'osteopath', 'osteopaths', count, _ ).
noun( 'osteopathy', '-', mass, _ ).
noun( 'ostler', 'ostlers', count, _ ).
noun( 'ostracism', '-', mass, _ ).
noun( 'ostrich', 'ostriches', count, _ ).
noun( 'other', 'others', count, _ ).
noun( 'otter', 'otters', both, _ ).
noun( 'ottoman', 'ottomans', count, _ ).
noun( 'oubliette', 'oubliettes', count, _ ).
noun( 'ouija', 'ouijas', count, _ ).
noun( 'ouija-board', 'ouija-boards', count, _ ).
noun( 'ounce', 'ounces', count, _ ).
noun( 'out-tray', 'out-trays', count, _ ).
noun( 'outback', '-', count, _ ).
noun( 'outbreak', 'outbreaks', count, _ ).
noun( 'outbuilding', 'outbuildings', count, _ ).
noun( 'outburst', 'outbursts', count, _ ).
noun( 'outcast', 'outcasts', count, _ ).
noun( 'outcaste', 'outcastes', count, _ ).
noun( 'outcome', 'outcomes', count, _ ).
noun( 'outcrop', 'outcrops', count, _ ).
noun( 'outcry', 'outcries', both, _ ).
noun( 'outfall', 'outfalls', count, _ ).
noun( 'outfield', 'outfields', count, _ ).
noun( 'outfielder', 'outfielders', count, _ ).
noun( 'outfit', 'outfits', count, _ ).
noun( 'outfitter', 'outfitters', count, _ ).
noun( 'outflow', 'outflows', count, _ ).
noun( 'outgo', 'outgoes', count, _ ).
noun( 'outgrowth', 'outgrowths', count, _ ).
noun( 'outhouse', 'outhouses', count, _ ).
noun( 'outing', 'outings', count, _ ).
noun( 'outlandishness', '-', mass, _ ).
noun( 'outlaw', 'outlaws', count, _ ).
noun( 'outlawry', '-', mass, _ ).
noun( 'outlay', 'outlays', both, _ ).
noun( 'outlet', 'outlets', count, _ ).
noun( 'outlier', 'outliers', count, _ ).
noun( 'outline', 'outlines', count, _ ).
noun( 'outlook', 'outlooks', count, _ ).
noun( 'outpatient', 'outpatients', count, _ ).
noun( 'outport', 'outports', count, _ ).
noun( 'outpost', 'outposts', count, _ ).
noun( 'outpouring', 'outpourings', count, _ ).
noun( 'output', '-', count, _ ).
noun( 'outrage', 'outrages', both, _ ).
noun( 'outrider', 'outriders', count, _ ).
noun( 'outrigger', 'outriggers', count, _ ).
noun( 'outset', 'outsets', count, _ ).
noun( 'outside', 'outsides', count, _ ).
noun( 'outsider', 'outsiders', count, _ ).
noun( 'outspokenness', '-', mass, _ ).
noun( 'outstation', 'outstations', count, _ ).
noun( 'outwork', 'outworks', count, _ ).
noun( 'ouzel', 'ouzels', count, _ ).
noun( 'ouzo', '-', mass, _ ).
noun( 'oval', 'ovals', count, _ ).
noun( 'ovary', 'ovaries', count, _ ).
noun( 'ovation', 'ovations', count, _ ).
noun( 'oven', 'ovens', count, _ ).
noun( 'ovenware', '-', mass, _ ).
noun( 'over', 'overs', count, _ ).
noun( 'over-abundance', '-', mass, _ ).
noun( 'overall', 'overalls', count, _ ).
noun( 'overanxiety', '-', mass, _ ).
noun( 'overbid', 'overbids', count, _ ).
noun( 'overburden', '-', mass, _ ).
noun( 'overcapitalization', '-', mass, _ ).
noun( 'overcast', '-', count, _ ).
noun( 'overcharge', 'overcharges', count, _ ).
noun( 'overcoat', 'overcoats', count, _ ).
noun( 'overconfidence', '-', mass, _ ).
noun( 'overcredulity', '-', mass, _ ).
noun( 'overdraft', 'overdrafts', count, _ ).
noun( 'overdrive', 'overdrives', both, _ ).
noun( 'overemphasis', '-', mass, _ ).
noun( 'overexertion', 'overexertions', both, _ ).
noun( 'overexposure', 'overexposures', both, _ ).
noun( 'overfeeding', '-', mass, _ ).
noun( 'overflow', 'overflows', count, _ ).
noun( 'overgrowth', 'overgrowths', both, _ ).
noun( 'overhang', 'overhangs', count, _ ).
noun( 'overhaul', 'overhauls', count, _ ).
noun( 'overindulgence', 'overindulgences', both, _ ).
noun( 'overkill', '-', mass, _ ).
noun( 'overlap', 'overlaps', both, _ ).
noun( 'overlay', 'overlays', count, _ ).
noun( 'overlord', 'overlords', count, _ ).
noun( 'overmantel', 'overmantels', count, _ ).
noun( 'overpass', 'overpasses', count, _ ).
noun( 'overpayment', 'overpayments', both, _ ).
noun( 'overplus', 'overpluses', count, _ ).
noun( 'overpopulation', '-', mass, _ ).
noun( 'overprint', 'overprints', count, _ ).
noun( 'overproduction', '-', mass, _ ).
noun( 'overseer', 'overseers', count, _ ).
noun( 'overshoe', 'overshoes', count, _ ).
noun( 'oversight', 'oversights', both, _ ).
noun( 'overskirt', 'overskirts', count, _ ).
noun( 'overspill', 'overspills', count, _ ).
noun( 'overstatement', 'overstatements', both, _ ).
noun( 'overstrain', '-', mass, _ ).
noun( 'overthrow', 'overthrows', count, _ ).
noun( 'overtime', '-', mass, _ ).
noun( 'overtolerance', '-', mass, _ ).
noun( 'overtone', 'overtones', count, _ ).
noun( 'overture', 'overtures', count, _ ).
noun( 'overweight', '-', mass, _ ).
noun( 'overwork', '-', mass, _ ).
noun( 'oviduct', 'oviducts', count, _ ).
noun( 'ovoid', 'ovoids', count, _ ).
noun( 'ovum', 'ova', count, _ ).
noun( 'owl', 'owls', count, _ ).
noun( 'owlet', 'owlets', count, _ ).
noun( 'owner', 'owners', count, _ ).
noun( 'owner-driver', 'owner-drivers', count, _ ).
noun( 'owner-occupier', 'owner-occupiers', count, _ ).
noun( 'ownership', '-', mass, _ ).
noun( 'ox', 'oxen', count, _ ).
noun( 'oxeye', 'oxeyes', count, _ ).
noun( 'oxidation', '-', mass, _ ).
noun( 'oxide', 'oxides', both, _ ).
noun( 'oxidization', 'oxidizations', both, _ ).
noun( 'oxtail', 'oxtails', count, _ ).
noun( 'oxyacetylene', '-', mass, _ ).
noun( 'oxygen', '-', mass, _ ).
noun( 'oyster', 'oysters', count, _ ).
noun( 'oyster-bank', 'oyster-banks', count, _ ).
noun( 'oyster-bar', 'oyster-bars', count, _ ).
noun( 'oyster-bed', 'oyster-beds', count, _ ).
noun( 'oyster-catcher', 'oyster-catchers', count, _ ).
noun( 'oz', 'oz', count, _ ).
noun( 'ozone', '-', mass, _ ).
noun( 'p', '-', count, _ ).
noun( 'p^at_e', 'p^at_es', both, _ ).
noun( 'pa', '-', count, _ ).
noun( 'pabulum', '-', mass, _ ).
noun( 'pace', 'paces', count, _ ).
noun( 'pacemaker', 'pacemakers', count, _ ).
noun( 'pacesetter', 'pacesetters', count, _ ).
noun( 'pachyderm', 'pachyderms', count, _ ).
noun( 'pacification', '-', mass, _ ).
noun( 'pacifism', '-', mass, _ ).
noun( 'pacifist', 'pacifists', count, _ ).
noun( 'pack', 'packs', count, _ ).
noun( 'pack-animal', 'pack-animals', count, _ ).
noun( 'pack-saddle', 'pack-saddles', count, _ ).
noun( 'pack-thread', '-', mass, _ ).
noun( 'package', 'packages', count, _ ).
noun( 'packer', 'packers', count, _ ).
noun( 'packet', 'packets', count, _ ).
noun( 'packet-boat', 'packet-boats', count, _ ).
noun( 'packhorse', 'packhorses', count, _ ).
noun( 'packing', '-', mass, _ ).
noun( 'packing-case', 'packing-cases', count, _ ).
noun( 'packing-needle', 'packing-needles', count, _ ).
noun( 'pact', 'pacts', count, _ ).
noun( 'pad', 'pads', count, _ ).
noun( 'padding', '-', mass, _ ).
noun( 'paddle', 'paddles', count, _ ).
noun( 'paddle-box', 'paddle-boxes', count, _ ).
noun( 'paddle-steamer', 'paddle-steamers', count, _ ).
noun( 'paddle-wheel', 'paddle-wheels', count, _ ).
noun( 'paddock', 'paddocks', count, _ ).
noun( 'paddy', 'paddies', both, _ ).
noun( 'paddy-field', 'paddy-fields', count, _ ).
noun( 'padlock', 'padlocks', count, _ ).
noun( 'padre', 'padres', count, _ ).
noun( 'paean', 'paeans', count, _ ).
noun( 'paederasty', '-', mass, _ ).
noun( 'paediatrics', 'paediatrics', mass, _ ).
noun( 'paeony', 'paeonies', count, _ ).
noun( 'pagan', 'pagans', count, _ ).
noun( 'paganism', '-', mass, _ ).
noun( 'page', 'pages', count, _ ).
noun( 'pageant', 'pageants', count, _ ).
noun( 'pageantry', '-', mass, _ ).
noun( 'pagination', '-', mass, _ ).
noun( 'pagoda', 'pagodas', count, _ ).
noun( 'pail', 'pails', count, _ ).
noun( 'pailful', 'pailfuls', count, _ ).
noun( 'paillasse', 'paillasses', count, _ ).
noun( 'pailliasse', 'pailliasses', count, _ ).
noun( 'pain', 'pains', both, _ ).
noun( 'painkiller', 'painkillers', count, _ ).
noun( 'paint', 'paints', both, _ ).
noun( 'paintbox', 'paintboxes', count, _ ).
noun( 'paintbrush', 'paintbrushes', count, _ ).
noun( 'painter', 'painters', count, _ ).
noun( 'painting', 'paintings', both, _ ).
noun( 'pair', 'pairs', count, _ ).
noun( 'paisley', '-', mass, _ ).
noun( 'pal', 'pals', count, _ ).
noun( 'palace', 'palaces', count, _ ).
noun( 'paladin', 'paladins', count, _ ).
noun( 'palaeontologist', 'palaeontologists', count, _ ).
noun( 'palaeontology', '-', mass, _ ).
noun( 'palankeen', 'palankeens', count, _ ).
noun( 'palanquin', 'palanquins', count, _ ).
noun( 'palatal', 'palatals', count, _ ).
noun( 'palate', 'palates', count, _ ).
noun( 'palatinate', 'palatinates', count, _ ).
noun( 'palaver', 'palavers', both, _ ).
noun( 'pale', 'pales', count, _ ).
noun( 'paleface', 'palefaces', count, _ ).
noun( 'paleness', '-', mass, _ ).
noun( 'paleontologist', 'paleontologists', count, _ ).
noun( 'paleontology', '-', mass, _ ).
noun( 'palette', 'palettes', count, _ ).
noun( 'palette-knife', 'palette-knives', count, _ ).
noun( 'palfrey', 'palfreys', count, _ ).
noun( 'palimpsest', 'palimpsests', count, _ ).
noun( 'palindrome', 'palindromes', count, _ ).
noun( 'paling', 'palings', count, _ ).
noun( 'palisade', 'palisades', count, _ ).
noun( 'pall', 'palls', count, _ ).
noun( 'pallbearer', 'pallbearers', count, _ ).
noun( 'pallet', 'pallets', count, _ ).
noun( 'palliasse', 'palliasses', count, _ ).
noun( 'palliation', 'palliations', both, _ ).
noun( 'palliative', 'palliatives', count, _ ).
noun( 'pallidness', '-', mass, _ ).
noun( 'pallor', '-', mass, _ ).
noun( 'palm', 'palms', count, _ ).
noun( 'palm-oil', '-', mass, _ ).
noun( 'palmer', 'palmers', count, _ ).
noun( 'palmetto', 'palmettos', count, _ ).
noun( 'palmist', 'palmists', count, _ ).
noun( 'palmistry', '-', mass, _ ).
noun( 'palpitation', 'palpitations', count, _ ).
noun( 'palsy', '-', mass, _ ).
noun( 'pampas-grass', '-', mass, _ ).
noun( 'pamphlet', 'pamphlets', count, _ ).
noun( 'pamphleteer', 'pamphleteers', count, _ ).
noun( 'pan', 'pans', both, _ ).
noun( 'panacea', 'panaceas', count, _ ).
noun( 'panache', '-', mass, _ ).
noun( 'panama', 'panamas', count, _ ).
noun( 'panatella', 'panatellas', count, _ ).
noun( 'pancake', 'pancakes', count, _ ).
noun( 'pancreas', 'pancreases', count, _ ).
noun( 'panda', 'pandas', count, _ ).
noun( 'pandemic', 'pandemics', count, _ ).
noun( 'pandemonium', 'pandemoniums', both, _ ).
noun( 'pander', 'panders', count, _ ).
noun( 'pane', 'panes', count, _ ).
noun( 'panegyric', 'panegyrics', count, _ ).
noun( 'panel', 'panels', count, _ ).
noun( 'panelling', '-', mass, _ ).
noun( 'pang', 'pangs', count, _ ).
noun( 'panga', 'pangas', count, _ ).
noun( 'panhandle', 'panhandles', count, _ ).
noun( 'panic', 'panics', both, _ ).
noun( 'panjandrum', 'panjandrums', count, _ ).
noun( 'pannier', 'panniers', count, _ ).
noun( 'pannikin', 'pannikins', count, _ ).
noun( 'panoply', 'panoplies', count, _ ).
noun( 'panorama', 'panoramas', count, _ ).
noun( 'pansy', 'pansies', count, _ ).
noun( 'pant', 'pants', count, _ ).
noun( 'pantaloon', 'pantaloons', count, _ ).
noun( 'pantechnicon', 'pantechnicons', count, _ ).
noun( 'pantheism', '-', mass, _ ).
noun( 'pantheist', 'pantheists', count, _ ).
noun( 'pantheon', 'pantheons', count, _ ).
noun( 'panther', 'panthers', count, _ ).
noun( 'pantile', 'pantiles', count, _ ).
noun( 'panto', 'pantos', count, _ ).
noun( 'pantograph', 'pantographs', count, _ ).
noun( 'pantomime', 'pantomimes', both, _ ).
noun( 'pantry', 'pantries', count, _ ).
noun( 'pantryman', 'pantrymen', count, _ ).
noun( 'panty-hose', '-', mass, _ ).
noun( 'pap', '-', mass, _ ).
noun( 'papa', 'papas', count, _ ).
noun( 'papacy', 'papacies', count, _ ).
noun( 'papaw', 'papaws', count, _ ).
noun( 'papaya', 'papayas', count, _ ).
noun( 'paper', 'papers', both, _ ).
noun( 'paper-chase', 'paper-chases', count, _ ).
noun( 'paperback', 'paperbacks', count, _ ).
noun( 'paperclip', 'paperclips', count, _ ).
noun( 'paperhanger', 'paperhangers', count, _ ).
noun( 'paperknife', 'paperknives', count, _ ).
noun( 'papermill', 'papermills', count, _ ).
noun( 'paperweight', 'paperweights', count, _ ).
noun( 'paperwork', '-', mass, _ ).
noun( 'papier-m^ach_e', '-', mass, _ ).
noun( 'papist', 'papists', count, _ ).
noun( 'papoose', 'papooses', count, _ ).
noun( 'paprika', '-', mass, _ ).
noun( 'papyrus', 'papyruses', both, _ ).
noun( 'par', 'pars', both, _ ).
noun( 'para', '-', count, _ ).
noun( 'parable', 'parables', count, _ ).
noun( 'parabola', 'parabolas', count, _ ).
noun( 'parachute', 'parachutes', count, _ ).
noun( 'parachutist', 'parachutists', count, _ ).
noun( 'parade', 'parades', both, _ ).
noun( 'parade-ground', 'parade-grounds', count, _ ).
noun( 'paradigm', 'paradigms', count, _ ).
noun( 'paradise', 'paradises', both, _ ).
noun( 'paradox', 'paradoxes', count, _ ).
noun( 'paraffin', '-', mass, _ ).
noun( 'paragon', 'paragons', count, _ ).
noun( 'paragraph', 'paragraphs', count, _ ).
noun( 'parakeet', 'parakeets', count, _ ).
noun( 'parallel', 'parallels', both, _ ).
noun( 'parallelism', 'parallelisms', both, _ ).
noun( 'parallelogram', 'parallelograms', count, _ ).
noun( 'paralysis', '-', mass, _ ).
noun( 'paralytic', 'paralytics', count, _ ).
noun( 'parameter', 'parameters', count, _ ).
noun( 'paramountcy', '-', mass, _ ).
noun( 'paramour', 'paramours', count, _ ).
noun( 'paranoia', '-', mass, _ ).
noun( 'paranoiac', 'paranoiacs', count, _ ).
noun( 'paranoid', 'paranoids', count, _ ).
noun( 'parapet', 'parapets', count, _ ).
noun( 'paraphernalia', '-', mass, _ ).
noun( 'paraplegia', '-', mass, _ ).
noun( 'paraplegic', 'paraplegics', count, _ ).
noun( 'parasite', 'parasites', count, _ ).
noun( 'parasol', 'parasols', count, _ ).
noun( 'paratrooper', 'paratroopers', count, _ ).
noun( 'paratyphoid', '-', mass, _ ).
noun( 'parcel', 'parcels', count, _ ).
noun( 'parchment', 'parchments', both, _ ).
noun( 'pardon', 'pardons', both, _ ).
noun( 'pardoner', 'pardoners', count, _ ).
noun( 'paregoric', '-', mass, _ ).
noun( 'parent', 'parents', count, _ ).
noun( 'parentage', '-', mass, _ ).
noun( 'parenthesis', 'parentheses', count, _ ).
noun( 'pari-mutuel', 'pari-mutuels', count, _ ).
noun( 'pariah', 'pariahs', count, _ ).
noun( 'pariah-dog', 'pariah-dogs', count, _ ).
noun( 'parish', 'parishes', count, _ ).
noun( 'parishioner', 'parishioners', count, _ ).
noun( 'parity', '-', mass, _ ).
noun( 'park', 'parks', count, _ ).
noun( 'parka', 'parkas', count, _ ).
noun( 'parking', '-', mass, _ ).
noun( 'parlance', 'parlances', count, _ ).
noun( 'parley', 'parleys', count, _ ).
noun( 'parliament', 'parliaments', count, _ ).
noun( 'parliamentarian', 'parliamentarians', count, _ ).
noun( 'parlour', 'parlours', count, _ ).
noun( 'parlour-car', 'parlour-cars', count, _ ).
noun( 'parochialism', '-', mass, _ ).
noun( 'parodist', 'parodists', count, _ ).
noun( 'parody', 'parodies', both, _ ).
noun( 'parole', '-', mass, _ ).
noun( 'paroquet', 'paroquets', count, _ ).
noun( 'paroxysm', 'paroxysms', count, _ ).
noun( 'parquet', '-', mass, _ ).
noun( 'parr', 'parrs', count, _ ).
noun( 'parricide', 'parricides', both, _ ).
noun( 'parrot', 'parrots', count, _ ).
noun( 'parry', 'parries', count, _ ).
noun( 'parsimony', '-', mass, _ ).
noun( 'parsley', '-', mass, _ ).
noun( 'parsnip', 'parsnips', count, _ ).
noun( 'parson', 'parsons', count, _ ).
noun( 'parsonage', 'parsonages', count, _ ).
noun( 'part', 'parts', count, _ ).
noun( 'part-owner', 'part-owners', count, _ ).
noun( 'part-singing', '-', mass, _ ).
noun( 'part-song', 'part-songs', count, _ ).
noun( 'part-timer', 'part-timers', count, _ ).
noun( 'parterre', 'parterres', count, _ ).
noun( 'parthenogenesis', '-', mass, _ ).
noun( 'partiality', 'partialities', both, _ ).
noun( 'participant', 'participants', count, _ ).
noun( 'participation', '-', mass, _ ).
noun( 'participle', 'participles', count, _ ).
noun( 'particle', 'particles', count, _ ).
noun( 'particular', 'particulars', count, _ ).
noun( 'particularity', 'particularities', both, _ ).
noun( 'parting', 'partings', both, _ ).
noun( 'partisan', 'partisans', count, _ ).
noun( 'partisanship', '-', mass, _ ).
noun( 'partition', 'partitions', both, _ ).
noun( 'partitive', 'partitives', count, _ ).
noun( 'partner', 'partners', count, _ ).
noun( 'partnership', 'partnerships', both, _ ).
noun( 'partridge', 'partridges', both, _ ).
noun( 'parturition', '-', mass, _ ).
noun( 'party', 'parties', both, _ ).
noun( 'party-spirit', '-', mass, _ ).
noun( 'party-wall', 'party-walls', count, _ ).
noun( 'parvenu', 'parvenus', count, _ ).
noun( 'pasha', 'pashas', count, _ ).
noun( 'pass', 'passes', count, _ ).
noun( 'passage', 'passages', both, _ ).
noun( 'passageway', 'passageways', count, _ ).
noun( 'passbook', 'passbooks', count, _ ).
noun( 'passenger', 'passengers', count, _ ).
noun( 'passepartout', 'passepartouts', both, _ ).
noun( 'passer-by', 'passers-by', count, _ ).
noun( 'passing', '-', mass, _ ).
noun( 'passion', 'passions', both, _ ).
noun( 'passion-flower', 'passion-flowers', count, _ ).
noun( 'passive', 'passives', count, _ ).
noun( 'passiveness', '-', mass, _ ).
noun( 'passivity', '-', mass, _ ).
noun( 'passkey', 'passkeys', count, _ ).
noun( 'passport', 'passports', count, _ ).
noun( 'password', 'passwords', count, _ ).
noun( 'past', 'pasts', count, _ ).
noun( 'pasta', '-', mass, _ ).
noun( 'paste', 'pastes', both, _ ).
noun( 'paste-up', 'paste-ups', count, _ ).
noun( 'pasteboard', '-', mass, _ ).
noun( 'pastel', 'pastels', count, _ ).
noun( 'pastern', 'pasterns', count, _ ).
noun( 'pasteurization', '-', mass, _ ).
noun( 'pastiche', 'pastiches', count, _ ).
noun( 'pastille', 'pastilles', count, _ ).
noun( 'pastime', 'pastimes', count, _ ).
noun( 'pasting', 'pastings', count, _ ).
noun( 'pastor', 'pastors', count, _ ).
noun( 'pastoral', 'pastorals', count, _ ).
noun( 'pastorate', 'pastorates', count, _ ).
noun( 'pastry', 'pastries', both, _ ).
noun( 'pastry-cook', 'pastry-cooks', count, _ ).
noun( 'pasturage', '-', mass, _ ).
noun( 'pasture', 'pastures', both, _ ).
noun( 'pasty', 'pasties', count, _ ).
noun( 'pat', 'pats', count, _ ).
noun( 'patch', 'patches', count, _ ).
noun( 'patch-pocket', 'patch-pockets', count, _ ).
noun( 'patchiness', '-', mass, _ ).
noun( 'patchouli', '-', mass, _ ).
noun( 'patchwork', '-', mass, _ ).
noun( 'pate', 'pates', count, _ ).
noun( 'pate de foie gras', '-', mass, _ ).
noun( 'patella', 'patellas', count, _ ).
noun( 'patent', 'patents', count, _ ).
noun( 'patentee', 'patentees', count, _ ).
noun( 'paterfamilias', '-', count, _ ).
noun( 'paternalism', '-', mass, _ ).
noun( 'paternity', '-', mass, _ ).
noun( 'paternoster', 'paternosters', count, _ ).
noun( 'path', 'paths', count, _ ).
noun( 'path-finder', 'path-finders', count, _ ).
noun( 'pathogen', 'pathogens', count, _ ).
noun( 'pathologist', 'pathologists', count, _ ).
noun( 'pathology', '-', mass, _ ).
noun( 'pathos', '-', mass, _ ).
noun( 'pathway', 'pathways', count, _ ).
noun( 'patience', '-', mass, _ ).
noun( 'patient', 'patients', count, _ ).
noun( 'patina', 'patinas', count, _ ).
noun( 'patio', 'patios', count, _ ).
noun( 'patisserie', 'patisseries', count, _ ).
noun( 'patois', '-', count, _ ).
noun( 'patrial', 'patrials', count, _ ).
noun( 'patriarch', 'patriarchs', count, _ ).
noun( 'patriarchate', 'patriarchates', count, _ ).
noun( 'patriarchy', 'patriarchies', count, _ ).
noun( 'patrician', 'patricians', count, _ ).
noun( 'patricide', 'patricides', both, _ ).
noun( 'patrimony', 'patrimonies', count, _ ).
noun( 'patriot', 'patriots', count, _ ).
noun( 'patriotism', '-', mass, _ ).
noun( 'patrol', 'patrols', both, _ ).
noun( 'patrolman', 'patrolmen', count, _ ).
noun( 'patron', 'patrons', count, _ ).
noun( 'patronage', '-', mass, _ ).
noun( 'patroness', 'patronesses', count, _ ).
noun( 'patronymic', 'patronymics', count, _ ).
noun( 'patten', 'pattens', count, _ ).
noun( 'patter', '-', mass, _ ).
noun( 'pattern', 'patterns', count, _ ).
noun( 'patty', 'patties', count, _ ).
noun( 'patty-pan', 'patty-pans', count, _ ).
noun( 'paucity', '-', mass, _ ).
noun( 'paunch', 'paunches', count, _ ).
noun( 'paunchiness', '-', mass, _ ).
noun( 'pauper', 'paupers', count, _ ).
noun( 'pauperism', '-', mass, _ ).
noun( 'pauperization', '-', mass, _ ).
noun( 'pause', 'pauses', count, _ ).
noun( 'pavement', 'pavements', count, _ ).
noun( 'pavilion', 'pavilions', count, _ ).
noun( 'paving-stone', 'paving-stones', count, _ ).
noun( 'paw', 'paws', count, _ ).
noun( 'pawl', 'pawls', count, _ ).
noun( 'pawn', 'pawns', both, _ ).
noun( 'pawn-ticket', 'pawn-tickets', count, _ ).
noun( 'pawnbroker', 'pawnbrokers', count, _ ).
noun( 'pawnshop', 'pawnshops', count, _ ).
noun( 'pawpaw', 'pawpaws', count, _ ).
noun( 'pax', 'paxes', count, _ ).
noun( 'pay', '-', mass, _ ).
noun( 'pay-as-you-earn', '-', mass, _ ).
noun( 'pay-claim', 'pay-claims', count, _ ).
noun( 'pay-packet', 'pay-packets', count, _ ).
noun( 'pay-station', 'pay-stations', count, _ ).
noun( 'payday', 'paydays', count, _ ).
noun( 'paydirt', '-', mass, _ ).
noun( 'payee', 'payees', count, _ ).
noun( 'payer', 'payers', count, _ ).
noun( 'payload', 'payloads', count, _ ).
noun( 'paymaster', 'paymasters', count, _ ).
noun( 'payment', 'payments', both, _ ).
noun( 'paynim', 'paynims', count, _ ).
noun( 'payoff', 'payoffs', count, _ ).
noun( 'payphone', 'payphones', count, _ ).
noun( 'payroll', 'payrolls', count, _ ).
noun( 'paysheet', 'paysheets', count, _ ).
noun( 'payslip', 'payslips', count, _ ).
noun( 'pea', 'peas', count, _ ).
noun( 'pea-chick', 'pea-chicks', count, _ ).
noun( 'pea-flour', '-', mass, _ ).
noun( 'pea-green', 'pea-greens', both, _ ).
noun( 'pea-jacket', 'pea-jackets', count, _ ).
noun( 'pea-soup', 'pea-soups', both, _ ).
noun( 'peace', '-', mass, _ ).
noun( 'peace-offering', 'peace-offerings', count, _ ).
noun( 'peacefulness', '-', mass, _ ).
noun( 'peacemaker', 'peacemakers', count, _ ).
noun( 'peacetime', '-', mass, _ ).
noun( 'peach', 'peaches', count, _ ).
noun( 'peacock', 'peacocks', count, _ ).
noun( 'peacock-blue', 'peacock-blues', both, _ ).
noun( 'peafowl', 'peafowls', count, _ ).
noun( 'peahen', 'peahens', count, _ ).
noun( 'peak', 'peaks', count, _ ).
noun( 'peal', 'peals', count, _ ).
noun( 'peanut', 'peanuts', count, _ ).
noun( 'pear', 'pears', both, _ ).
noun( 'pearl', 'pearls', count, _ ).
noun( 'pearl-barley', '-', mass, _ ).
noun( 'pearl-diver', 'pearl-divers', count, _ ).
noun( 'pearl-fishery', 'pearl-fisheries', count, _ ).
noun( 'pearl-oyster', 'pearl-oysters', count, _ ).
noun( 'pearl-sago', '-', mass, _ ).
noun( 'pearmain', 'pearmains', count, _ ).
noun( 'peasant', 'peasants', count, _ ).
noun( 'peasantry', 'peasantries', count, _ ).
noun( 'pease', '-', mass, _ ).
noun( 'pease-pudding', 'pease-puddings', both, _ ).
noun( 'peashooter', 'peashooters', count, _ ).
noun( 'peasouper', 'peasoupers', count, _ ).
noun( 'peat', '-', mass, _ ).
noun( 'pebble', 'pebbles', count, _ ).
noun( 'pecan', 'pecans', count, _ ).
noun( 'peccadillo', 'peccadillos', count, _ ).
noun( 'peccary', 'peccaries', count, _ ).
noun( 'peck', 'pecks', count, _ ).
noun( 'pecker', 'peckers', count, _ ).
noun( 'pectin', '-', mass, _ ).
noun( 'peculation', 'peculations', both, _ ).
noun( 'peculiarity', 'peculiarities', both, _ ).
noun( 'pedagogue', 'pedagogues', count, _ ).
noun( 'pedagogy', '-', mass, _ ).
noun( 'pedal', 'pedals', count, _ ).
noun( 'pedant', 'pedants', count, _ ).
noun( 'pedantry', 'pedantries', both, _ ).
noun( 'peddler', 'peddlers', count, _ ).
noun( 'pederast', 'pederasts', count, _ ).
noun( 'pederasty', '-', mass, _ ).
noun( 'pedestal', 'pedestals', count, _ ).
noun( 'pedestrian', 'pedestrians', count, _ ).
noun( 'pediatrician', 'pediatricians', count, _ ).
noun( 'pediatrics', 'pediatrics', mass, _ ).
noun( 'pedicab', 'pedicabs', count, _ ).
noun( 'pedicure', 'pedicures', count, _ ).
noun( 'pedigree', 'pedigrees', both, _ ).
noun( 'pediment', 'pediments', count, _ ).
noun( 'pedlar', 'pedlars', count, _ ).
noun( 'pedometer', 'pedometers', count, _ ).
noun( 'pee', 'pees', both, _ ).
noun( 'peek', 'peeks', count, _ ).
noun( 'peek-a-boo', '-', mass, _ ).
noun( 'peel', '-', mass, _ ).
noun( 'peeler', 'peelers', count, _ ).
noun( 'peep', 'peeps', count, _ ).
noun( 'peeper', 'peepers', count, _ ).
noun( 'peephole', 'peepholes', count, _ ).
noun( 'peepshow', 'peepshows', count, _ ).
noun( 'peepul', 'peepuls', count, _ ).
noun( 'peer', 'peers', count, _ ).
noun( 'peerage', 'peerages', count, _ ).
noun( 'peeress', 'peeresses', count, _ ).
noun( 'peevishness', '-', mass, _ ).
noun( 'peewit', 'peewits', count, _ ).
noun( 'peg', 'pegs', count, _ ).
noun( 'peignoir', 'peignoirs', count, _ ).
noun( 'peke', 'pekes', count, _ ).
noun( 'pekinese', 'pekinese', count, _ ).
noun( 'pekoe', '-', mass, _ ).
noun( 'pelf', '-', mass, _ ).
noun( 'pelican', 'pelicans', count, _ ).
noun( 'pelisse', 'pelisses', count, _ ).
noun( 'pellet', 'pellets', count, _ ).
noun( 'pelmet', 'pelmets', count, _ ).
noun( 'pelota', '-', mass, _ ).
noun( 'pelt', 'pelts', count, _ ).
noun( 'pelvis', 'pelvises', count, _ ).
noun( 'pemmican', '-', mass, _ ).
noun( 'pen', 'pens', count, _ ).
noun( 'pen-and-ink', '-', mass, _ ).
noun( 'pen-friend', 'pen-friends', count, _ ).
noun( 'pen-name', 'pen-names', count, _ ).
noun( 'pen-pusher', 'pen-pushers', count, _ ).
noun( 'penalization', 'penalizations', both, _ ).
noun( 'penalty', 'penalties', both, _ ).
noun( 'penance', '-', mass, _ ).
noun( 'penchant', 'penchants', count, _ ).
noun( 'pencil', 'pencils', count, _ ).
noun( 'pendant', 'pendants', count, _ ).
noun( 'pendulum', 'pendulums', count, _ ).
noun( 'penetrability', '-', mass, _ ).
noun( 'penetration', 'penetrations', both, _ ).
noun( 'penguin', 'penguins', count, _ ).
noun( 'penicillin', '-', mass, _ ).
noun( 'peninsula', 'peninsulas', count, _ ).
noun( 'penis', 'penises', count, _ ).
noun( 'penitence', '-', mass, _ ).
noun( 'penitent', 'penitents', count, _ ).
noun( 'penitentiary', 'penitentiaries', count, _ ).
noun( 'penknife', 'penknives', count, _ ).
noun( 'penmanship', '-', mass, _ ).
noun( 'penn\'orth', 'penn\'orths', count, _ ).
noun( 'pennant', 'pennants', count, _ ).
noun( 'pennon', 'pennons', count, _ ).
noun( 'penny', 'pennies', count, _ ).
noun( 'pennyweight', 'pennyweights', count, _ ).
noun( 'pennyworth', 'pennyworths', count, _ ).
noun( 'penology', '-', mass, _ ).
noun( 'pension', 'pensions', count, _ ).
noun( 'pension', 'pensions', count, _ ).
noun( 'pensioner', 'pensioners', count, _ ).
noun( 'pensiveness', '-', mass, _ ).
noun( 'penstock', 'penstocks', count, _ ).
noun( 'pentagon', 'pentagons', count, _ ).
noun( 'pentameter', 'pentameters', count, _ ).
noun( 'pentathlon', 'pentathlons', count, _ ).
noun( 'penthouse', 'penthouses', count, _ ).
noun( 'penultimate', 'penultimates', count, _ ).
noun( 'penumbra', 'penumbras', count, _ ).
noun( 'penuriousness', '-', mass, _ ).
noun( 'penury', '-', mass, _ ).
noun( 'peon', 'peons', count, _ ).
noun( 'peonage', '-', mass, _ ).
noun( 'peony', 'peonies', count, _ ).
noun( 'people', 'peoples', both, _ ).
noun( 'pep', '-', mass, _ ).
noun( 'pepper', 'peppers', both, _ ).
noun( 'pepper-and-salt', '-', mass, _ ).
noun( 'pepper-mill', 'pepper-mills', count, _ ).
noun( 'pepper-pot', 'pepper-pots', count, _ ).
noun( 'peppercorn', 'peppercorns', count, _ ).
noun( 'peppermint', 'peppermints', both, _ ).
noun( 'pepsin', '-', mass, _ ).
noun( 'perambulation', 'perambulations', count, _ ).
noun( 'perambulator', 'perambulators', count, _ ).
noun( 'perceiver', 'perceivers', count, _ ).
noun( 'percent', 'percent', count, _ ).
noun( 'percentage', 'percentages', count, _ ).
noun( 'perceptibility', '-', mass, _ ).
noun( 'perception', 'perceptions', both, _ ).
noun( 'perch', 'perch', count, _ ).
noun( 'percolator', 'percolators', count, _ ).
noun( 'percussion', 'percussions', both, _ ).
noun( 'percussionist', 'percussionists', count, _ ).
noun( 'perdition', '-', mass, _ ).
noun( 'peregrination', 'peregrinations', both, _ ).
noun( 'perennial', 'perennials', count, _ ).
noun( 'perfectibility', '-', mass, _ ).
noun( 'perfection', '-', mass, _ ).
noun( 'perfectionist', 'perfectionists', count, _ ).
noun( 'perfidiousness', '-', mass, _ ).
noun( 'perfidy', 'perfidies', both, _ ).
noun( 'perforation', 'perforations', both, _ ).
noun( 'performance', 'performances', both, _ ).
noun( 'performer', 'performers', count, _ ).
noun( 'perfume', 'perfumes', both, _ ).
noun( 'perfumer', 'perfumers', count, _ ).
noun( 'pergola', 'pergolas', count, _ ).
noun( 'peri', 'peris', count, _ ).
noun( 'perigee', 'perigees', count, _ ).
noun( 'perihelion', 'perihelions', count, _ ).
noun( 'peril', 'perils', both, _ ).
noun( 'perimeter', 'perimeters', count, _ ).
noun( 'period', 'periods', count, _ ).
noun( 'periodical', 'periodicals', count, _ ).
noun( 'periphery', 'peripheries', count, _ ).
noun( 'periphrasis', 'periphrases', count, _ ).
noun( 'periscope', 'periscopes', count, _ ).
noun( 'perisher', 'perishers', count, _ ).
noun( 'peristyle', 'peristyles', count, _ ).
noun( 'peritonitis', '-', mass, _ ).
noun( 'periwig', 'periwigs', count, _ ).
noun( 'periwinkle', 'periwinkles', count, _ ).
noun( 'perjurer', 'perjurers', count, _ ).
noun( 'perjury', 'perjuries', both, _ ).
noun( 'perk', 'perks', count, _ ).
noun( 'perkiness', '-', mass, _ ).
noun( 'perm', 'perms', count, _ ).
noun( 'permafrost', '-', mass, _ ).
noun( 'permanence', '-', mass, _ ).
noun( 'permanency', 'permanencies', both, _ ).
noun( 'permanganate', 'permanganates', both, _ ).
noun( 'permeability', '-', mass, _ ).
noun( 'permeation', '-', mass, _ ).
noun( 'permission', '-', mass, _ ).
noun( 'permissiveness', '-', mass, _ ).
noun( 'permit', 'permits', count, _ ).
noun( 'permutation', 'permutations', both, _ ).
noun( 'perniciousness', '-', mass, _ ).
noun( 'peroration', 'perorations', count, _ ).
noun( 'peroxide', 'peroxides', both, _ ).
noun( 'perpendicular', 'perpendiculars', both, _ ).
noun( 'perpetration', 'perpetrations', count, _ ).
noun( 'perpetrator', 'perpetrators', count, _ ).
noun( 'perpetuation', 'perpetuations', count, _ ).
noun( 'perpetuity', 'perpetuities', both, _ ).
noun( 'perplexity', 'perplexities', both, _ ).
noun( 'perquisite', 'perquisites', count, _ ).
noun( 'perry', '-', mass, _ ).
noun( 'persecution', 'persecutions', both, _ ).
noun( 'persecutor', 'persecutors', count, _ ).
noun( 'perseverance', '-', mass, _ ).
noun( 'persiflage', '-', mass, _ ).
noun( 'persimmon', 'persimmons', count, _ ).
noun( 'persistence', '-', mass, _ ).
noun( 'person', 'persons', count, _ ).
noun( 'persona', 'personas', count, _ ).
noun( 'persona grata', 'personae gratae', count, _ ).
noun( 'persona non grata', 'personae non gratae', count, _ ).
noun( 'personage', 'personages', count, _ ).
noun( 'personal', 'personals', count, _ ).
noun( 'personality', 'personalities', both, _ ).
noun( 'personalty', '-', mass, _ ).
noun( 'personation', 'personations', count, _ ).
noun( 'personification', 'personifications', both, _ ).
noun( 'perspective', 'perspectives', both, _ ).
noun( 'perspex', '-', mass, _ ).
noun( 'perspicacity', '-', mass, _ ).
noun( 'perspicuity', '-', mass, _ ).
noun( 'perspicuousness', '-', mass, _ ).
noun( 'perspiration', '-', mass, _ ).
noun( 'persuasion', 'persuasions', both, _ ).
noun( 'persuasiveness', '-', mass, _ ).
noun( 'pertinacity', '-', mass, _ ).
noun( 'pertinence', '-', mass, _ ).
noun( 'pertness', '-', mass, _ ).
noun( 'perturbation', 'perturbations', both, _ ).
noun( 'peruke', 'perukes', count, _ ).
noun( 'perusal', 'perusals', both, _ ).
noun( 'pervasion', '-', mass, _ ).
noun( 'pervasiveness', '-', mass, _ ).
noun( 'perverseness', '-', mass, _ ).
noun( 'perversion', 'perversions', both, _ ).
noun( 'perversity', 'perversities', both, _ ).
noun( 'pervert', 'perverts', count, _ ).
noun( 'peseta', 'pesetas', count, _ ).
noun( 'peso', 'pesos', count, _ ).
noun( 'pessary', 'pessaries', count, _ ).
noun( 'pessimism', '-', mass, _ ).
noun( 'pessimist', 'pessimists', count, _ ).
noun( 'pest', 'pests', count, _ ).
noun( 'pesticide', 'pesticides', both, _ ).
noun( 'pestilence', 'pestilences', both, _ ).
noun( 'pestle', 'pestles', count, _ ).
noun( 'pet', 'pets', count, _ ).
noun( 'petal', 'petals', count, _ ).
noun( 'petard', 'petards', count, _ ).
noun( 'petit bourgeois', 'petits bourgeois', count, _ ).
noun( 'petition', 'petitions', count, _ ).
noun( 'petitioner', 'petitioners', count, _ ).
noun( 'petrel', 'petrels', count, _ ).
noun( 'petrifaction', 'petrifactions', both, _ ).
noun( 'petro-chemical', 'petro-chemicals', count, _ ).
noun( 'petrol', '-', mass, _ ).
noun( 'petroleum', '-', mass, _ ).
noun( 'petrology', '-', mass, _ ).
noun( 'petticoat', 'petticoats', count, _ ).
noun( 'pettiness', 'pettinesses', both, _ ).
noun( 'pettishness', '-', mass, _ ).
noun( 'petulance', '-', mass, _ ).
noun( 'petunia', 'petunias', count, _ ).
noun( 'pew', 'pews', count, _ ).
noun( 'pew-opener', 'pew-openers', count, _ ).
noun( 'pewit', 'pewits', count, _ ).
noun( 'pewter', '-', mass, _ ).
noun( 'peyote', 'peyotes', count, _ ).
noun( 'pfennig', 'pfennigs', count, _ ).
noun( 'phaeton', 'phaetons', count, _ ).
noun( 'phagocyte', 'phagocytes', count, _ ).
noun( 'phalanx', 'phalanxes', count, _ ).
noun( 'phallus', 'phalluses', count, _ ).
noun( 'phantasm', 'phantasms', count, _ ).
noun( 'phantasmagoria', 'phantasmagorias', count, _ ).
noun( 'phantasy', 'phantasies', count, _ ).
noun( 'phantom', 'phantoms', count, _ ).
noun( 'pharmacist', 'pharmacists', count, _ ).
noun( 'pharmacologist', 'pharmacologists', count, _ ).
noun( 'pharmacology', '-', mass, _ ).
noun( 'pharmacopoeia', 'pharmacopoeias', count, _ ).
noun( 'pharmacy', 'pharmacies', both, _ ).
noun( 'pharos', '-', count, _ ).
noun( 'pharyngitis', '-', mass, _ ).
noun( 'pharynx', 'pharynxes', count, _ ).
noun( 'phase', 'phases', count, _ ).
noun( 'pheasant', 'pheasants', both, _ ).
noun( 'phenobarbitone', '-', mass, _ ).
noun( 'phenol', '-', mass, _ ).
noun( 'phenomenon', 'phenomena', count, _ ).
noun( 'phial', 'phials', count, _ ).
noun( 'philanderer', 'philanderers', count, _ ).
noun( 'philanthropist', 'philanthropists', count, _ ).
noun( 'philanthropy', '-', mass, _ ).
noun( 'philatelist', 'philatelists', count, _ ).
noun( 'philately', '-', mass, _ ).
noun( 'philhellene', 'philhellenes', count, _ ).
noun( 'philologist', 'philologists', count, _ ).
noun( 'philology', '-', mass, _ ).
noun( 'philosopher', 'philosophers', count, _ ).
noun( 'philosophy', 'philosophies', both, _ ).
noun( 'philtre', 'philtres', count, _ ).
noun( 'phlebitis', '-', mass, _ ).
noun( 'phlegm', '-', mass, _ ).
noun( 'phlox', '-', mass, _ ).
noun( 'phobia', 'phobias', count, _ ).
noun( 'phoenix', 'phoenixes', count, _ ).
noun( 'phone', 'phones', count, _ ).
noun( 'phone-in', 'phone-ins', count, _ ).
noun( 'phonebooth', 'phonebooths', count, _ ).
noun( 'phonecall', 'phonecalls', count, _ ).
noun( 'phoneme', 'phonemes', count, _ ).
noun( 'phonemics', 'phonemics', mass, _ ).
noun( 'phonetician', 'phoneticians', count, _ ).
noun( 'phonetics', 'phonetics', mass, _ ).
noun( 'phoney', 'phoneys', count, _ ).
noun( 'phonics', 'phonics', mass, _ ).
noun( 'phonograph', 'phonographs', count, _ ).
noun( 'phonology', '-', mass, _ ).
noun( 'phony', 'phonies', count, _ ).
noun( 'phosgene', '-', mass, _ ).
noun( 'phosphate', 'phosphates', both, _ ).
noun( 'phosphorescence', '-', mass, _ ).
noun( 'phosphorus', '-', mass, _ ).
noun( 'photo', 'photos', count, _ ).
noun( 'photocopier', 'photocopiers', count, _ ).
noun( 'photocopy', 'photocopies', count, _ ).
noun( 'photoflash', 'photoflashes', count, _ ).
noun( 'photograph', 'photographs', count, _ ).
noun( 'photographer', 'photographers', count, _ ).
noun( 'photography', '-', mass, _ ).
noun( 'photogravure', 'photogravures', both, _ ).
noun( 'photolithography', '-', mass, _ ).
noun( 'photometer', 'photometers', count, _ ).
noun( 'photon', 'photons', count, _ ).
noun( 'photostat', 'photostats', count, _ ).
noun( 'photosynthesis', '-', mass, _ ).
noun( 'phrase', 'phrases', count, _ ).
noun( 'phrase-book', 'phrase-books', count, _ ).
noun( 'phraseology', '-', mass, _ ).
noun( 'phrenologist', 'phrenologists', count, _ ).
noun( 'phrenology', '-', mass, _ ).
noun( 'phthisis', '-', mass, _ ).
noun( 'phylum', 'phylums', count, _ ).
noun( 'physic', 'physics', count, _ ).
noun( 'physician', 'physicians', count, _ ).
noun( 'physicist', 'physicists', count, _ ).
noun( 'physics', 'physics', mass, _ ).
noun( 'physiognomy', 'physiognomies', both, _ ).
noun( 'physiologist', 'physiologists', count, _ ).
noun( 'physiology', '-', mass, _ ).
noun( 'physiotherapist', 'physiotherapists', count, _ ).
noun( 'physiotherapy', '-', mass, _ ).
noun( 'physique', 'physiques', both, _ ).
noun( 'pi', 'pis', count, _ ).
noun( 'pianist', 'pianists', count, _ ).
noun( 'piano', 'pianos', count, _ ).
noun( 'pianoforte', 'pianofortes', count, _ ).
noun( 'pianola', 'pianolas', count, _ ).
noun( 'piastre', 'piastres', count, _ ).
noun( 'piazza', 'piazzas', count, _ ).
noun( 'pibroch', 'pibrochs', count, _ ).
noun( 'pica', 'pica', count, _ ).
noun( 'picador', 'picadors', count, _ ).
noun( 'piccalilli', '-', mass, _ ).
noun( 'piccaninny', 'piccaninnies', count, _ ).
noun( 'piccolo', 'piccolos', count, _ ).
noun( 'pick', 'picks', count, _ ).
noun( 'pick-me-up', 'pick-me-ups', count, _ ).
noun( 'pick-up', 'pick-ups', count, _ ).
noun( 'pickaxe', 'pickaxes', count, _ ).
noun( 'picker', 'pickers', count, _ ).
noun( 'pickerel', 'pickerel', count, _ ).
noun( 'picket', 'pickets', count, _ ).
noun( 'picking', 'pickings', both, _ ).
noun( 'pickle', 'pickles', both, _ ).
noun( 'pickpocket', 'pickpockets', count, _ ).
noun( 'picnic', 'picnics', count, _ ).
noun( 'picnicker', 'picnickers', count, _ ).
noun( 'pictorial', 'pictorials', count, _ ).
noun( 'picture', 'pictures', count, _ ).
noun( 'picture-book', 'picture-books', count, _ ).
noun( 'picture-card', 'picture-cards', count, _ ).
noun( 'picture-gallery', 'picture-galleries', count, _ ).
noun( 'picturesqueness', '-', mass, _ ).
noun( 'piddle', 'piddles', count, _ ).
noun( 'pidgin', 'pidgins', count, _ ).
noun( 'pie', 'pies', both, _ ).
noun( 'pie-crust', '-', mass, _ ).
noun( 'piece', 'pieces', count, _ ).
noun( 'piecework', '-', mass, _ ).
noun( 'pied-`a-terre', 'pied-`a-terres', count, _ ).
noun( 'pier', 'piers', count, _ ).
noun( 'pier-glass', 'pier-glasses', count, _ ).
noun( 'pierrot', 'pierrots', count, _ ).
noun( 'piet`a', 'piet`as', count, _ ).
noun( 'piety', 'pieties', both, _ ).
noun( 'piffle', '-', mass, _ ).
noun( 'pig', 'pigs', both, _ ).
noun( 'pig-headedness', '-', mass, _ ).
noun( 'pig-iron', '-', mass, _ ).
noun( 'pig-sticking', '-', mass, _ ).
noun( 'pigboat', 'pigboats', count, _ ).
noun( 'pigeon', 'pigeons', count, _ ).
noun( 'pigeonhole', 'pigeonholes', count, _ ).
noun( 'piggery', 'piggeries', count, _ ).
noun( 'piggishness', '-', mass, _ ).
noun( 'piggy', 'piggies', count, _ ).
noun( 'piggyback', 'piggybacks', count, _ ).
noun( 'piglet', 'piglets', count, _ ).
noun( 'pigment', 'pigments', both, _ ).
noun( 'pigmentation', 'pigmentations', count, _ ).
noun( 'pigmy', 'pigmies', count, _ ).
noun( 'pigskin', '-', mass, _ ).
noun( 'pigsty', 'pigsties', count, _ ).
noun( 'pigswill', '-', mass, _ ).
noun( 'pigtail', 'pigtails', count, _ ).
noun( 'pigwash', '-', mass, _ ).
noun( 'pike', 'pike', count, _ ).
noun( 'pikestaff', 'pikestaffs', count, _ ).
noun( 'pilaf', '-', mass, _ ).
noun( 'pilaff', '-', mass, _ ).
noun( 'pilaster', 'pilasters', count, _ ).
noun( 'pilau', '-', mass, _ ).
noun( 'pilchard', 'pilchards', both, _ ).
noun( 'pile', 'piles', both, _ ).
noun( 'pile-driver', 'pile-drivers', count, _ ).
noun( 'pile-dwelling', 'pile-dwellings', count, _ ).
noun( 'pile-up', 'pile-ups', count, _ ).
noun( 'piles', 'piles', mass, _ ).
noun( 'pilferage', '-', mass, _ ).
noun( 'pilferer', 'pilferers', count, _ ).
noun( 'pilgrim', 'pilgrims', count, _ ).
noun( 'pilgrimage', 'pilgrimages', count, _ ).
noun( 'pill', 'pills', count, _ ).
noun( 'pillage', 'pillages', count, _ ).
noun( 'pillager', 'pillagers', count, _ ).
noun( 'pillar', 'pillars', count, _ ).
noun( 'pillar-box', 'pillar-boxes', count, _ ).
noun( 'pillbox', 'pillboxes', count, _ ).
noun( 'pillion', 'pillions', count, _ ).
noun( 'pillory', 'pillories', count, _ ).
noun( 'pillow', 'pillows', count, _ ).
noun( 'pillow-fight', 'pillow-fights', count, _ ).
noun( 'pillowcase', 'pillowcases', count, _ ).
noun( 'pillowslip', 'pillowslips', count, _ ).
noun( 'pilot', 'pilots', count, _ ).
noun( 'pilot-boat', 'pilot-boats', count, _ ).
noun( 'pilot-burner', 'pilot-burners', count, _ ).
noun( 'pilot-cloth', '-', mass, _ ).
noun( 'pilot-engine', 'pilot-engines', count, _ ).
noun( 'pilot-fish', 'pilot-fish', count, _ ).
noun( 'pilot-light', 'pilot-lights', count, _ ).
noun( 'pimento', 'pimentos', both, _ ).
noun( 'pimp', 'pimps', count, _ ).
noun( 'pimpernel', 'pimpernels', count, _ ).
noun( 'pimple', 'pimples', count, _ ).
noun( 'pin', 'pins', count, _ ).
noun( 'pin-money', '-', mass, _ ).
noun( 'pin-table', 'pin-tables', count, _ ).
noun( 'pin-up', 'pin-ups', count, _ ).
noun( 'pinafore', 'pinafores', count, _ ).
noun( 'pinball', '-', mass, _ ).
noun( 'pince-nez', 'pince-nez', count, _ ).
noun( 'pincer', 'pincers', count, _ ).
noun( 'pinch', 'pinches', count, _ ).
noun( 'pinchbeck', 'pinchbecks', count, _ ).
noun( 'pincushion', 'pincushions', count, _ ).
noun( 'pine', 'pines', both, _ ).
noun( 'pineapple', 'pineapples', both, _ ).
noun( 'ping', 'pings', count, _ ).
noun( 'pingpong', '-', mass, _ ).
noun( 'pinhead', 'pinheads', count, _ ).
noun( 'pinion', 'pinions', count, _ ).
noun( 'pink', 'pinks', both, _ ).
noun( 'pinnace', 'pinnaces', count, _ ).
noun( 'pinnacle', 'pinnacles', count, _ ).
noun( 'pinny', 'pinnies', count, _ ).
noun( 'pinpoint', 'pinpoints', count, _ ).
noun( 'pinprick', 'pinpricks', count, _ ).
noun( 'pint', 'pints', count, _ ).
noun( 'pioneer', 'pioneers', count, _ ).
noun( 'pip', 'pips', count, _ ).
noun( 'pipal', 'pipals', count, _ ).
noun( 'pipe', 'pipes', count, _ ).
noun( 'pipe-organ', 'pipe-organs', count, _ ).
noun( 'pipe-rack', 'pipe-racks', count, _ ).
noun( 'pipeclay', '-', mass, _ ).
noun( 'pipedream', 'pipedreams', count, _ ).
noun( 'pipeful', 'pipefuls', count, _ ).
noun( 'pipeline', 'pipelines', count, _ ).
noun( 'piper', 'pipers', count, _ ).
noun( 'pipette', 'pipettes', count, _ ).
noun( 'pipework', '-', mass, _ ).
noun( 'piping', '-', mass, _ ).
noun( 'pippin', 'pippins', count, _ ).
noun( 'pipsqueak', 'pipsqueaks', count, _ ).
noun( 'piquancy', '-', mass, _ ).
noun( 'pique', 'piques', both, _ ).
noun( 'piquet', '-', mass, _ ).
noun( 'piracy', 'piracies', both, _ ).
noun( 'piranha', 'piranhas', count, _ ).
noun( 'pirate', 'pirates', count, _ ).
noun( 'pirouette', 'pirouettes', count, _ ).
noun( 'pis aller', '-', count, _ ).
noun( 'piss', 'pisses', count, _ ).
noun( 'pistachio', 'pistachios', count, _ ).
noun( 'pistil', 'pistils', count, _ ).
noun( 'pistol', 'pistols', count, _ ).
noun( 'piston', 'pistons', count, _ ).
noun( 'pit', 'pits', count, _ ).
noun( 'pit-prop', 'pit-props', count, _ ).
noun( 'pitch', 'pitches', both, _ ).
noun( 'pitchblende', '-', mass, _ ).
noun( 'pitcher', 'pitchers', count, _ ).
noun( 'pitchfork', 'pitchforks', count, _ ).
noun( 'pitfall', 'pitfalls', count, _ ).
noun( 'pith', '-', mass, _ ).
noun( 'pithead', 'pitheads', count, _ ).
noun( 'pitman', 'pitmen', count, _ ).
noun( 'piton', 'pitons', count, _ ).
noun( 'pitsaw', 'pitsaws', count, _ ).
noun( 'pittance', 'pittances', count, _ ).
noun( 'pitter-patter', '-', mass, _ ).
noun( 'pituitary', 'pituitaries', count, _ ).
noun( 'pity', 'pities', both, _ ).
noun( 'pivot', 'pivots', count, _ ).
noun( 'pixie', 'pixies', count, _ ).
noun( 'pixy', 'pixies', count, _ ).
noun( 'pizza', 'pizzas', count, _ ).
noun( 'placard', 'placards', count, _ ).
noun( 'place', 'places', count, _ ).
noun( 'place-bet', 'place-bets', count, _ ).
noun( 'place-name', 'place-names', count, _ ).
noun( 'placebo', 'placebos', count, _ ).
noun( 'placeman', 'placemen', count, _ ).
noun( 'placement', 'placements', count, _ ).
noun( 'placenta', 'placentas', count, _ ).
noun( 'placeseeker', 'placeseekers', count, _ ).
noun( 'placidity', '-', mass, _ ).
noun( 'placing', 'placings', count, _ ).
noun( 'placket', 'plackets', count, _ ).
noun( 'plage', '-', count, _ ).
noun( 'plagiarism', 'plagiarisms', both, _ ).
noun( 'plagiarist', 'plagiarists', count, _ ).
noun( 'plague', 'plagues', count, _ ).
noun( 'plague-spot', 'plague-spots', count, _ ).
noun( 'plaice', 'plaice', both, _ ).
noun( 'plaid', 'plaids', both, _ ).
noun( 'plain', 'plains', count, _ ).
noun( 'plainness', '-', mass, _ ).
noun( 'plainsman', 'plainsmen', count, _ ).
noun( 'plaint', 'plaints', count, _ ).
noun( 'plaintiff', 'plaintiffs', count, _ ).
noun( 'plaintiveness', '-', mass, _ ).
noun( 'plait', 'plaits', count, _ ).
noun( 'plan', 'plans', count, _ ).
noun( 'planchette', 'planchettes', count, _ ).
noun( 'plane', 'planes', count, _ ).
noun( 'plane-tree', 'plane-trees', count, _ ).
noun( 'planet', 'planets', count, _ ).
noun( 'planetarium', 'planetariums', count, _ ).
noun( 'plank', 'planks', count, _ ).
noun( 'plank-bed', 'plank-beds', count, _ ).
noun( 'planking', '-', mass, _ ).
noun( 'plankton', '-', mass, _ ).
noun( 'planner', 'planners', count, _ ).
noun( 'plant', 'plants', both, _ ).
noun( 'plant-louse', 'plant-lice', count, _ ).
noun( 'plantain', 'plantains', count, _ ).
noun( 'plantation', 'plantations', count, _ ).
noun( 'planter', 'planters', count, _ ).
noun( 'plaque', 'plaques', both, _ ).
noun( 'plash', '-', count, _ ).
noun( 'plasm', '-', mass, _ ).
noun( 'plasma', '-', mass, _ ).
noun( 'plaster', 'plasters', both, _ ).
noun( 'plasterboard', '-', mass, _ ).
noun( 'plasterer', 'plasterers', count, _ ).
noun( 'plastic', 'plastics', count, _ ).
noun( 'plastic-bomb', 'plastic-bombs', count, _ ).
noun( 'plasticine', '-', mass, _ ).
noun( 'plasticity', '-', mass, _ ).
noun( 'plastics', 'plastics', mass, _ ).
noun( 'plate', 'plates', both, _ ).
noun( 'plate-glass', '-', mass, _ ).
noun( 'plate-powder', '-', mass, _ ).
noun( 'plate-rack', 'plate-racks', count, _ ).
noun( 'plateau', 'plateaus', count, _ ).
noun( 'plateful', 'platefuls', count, _ ).
noun( 'platelayer', 'platelayers', count, _ ).
noun( 'platform', 'platforms', count, _ ).
noun( 'plating', '-', mass, _ ).
noun( 'platinum', '-', mass, _ ).
noun( 'platitude', 'platitudes', both, _ ).
noun( 'platoon', 'platoons', count, _ ).
noun( 'platter', 'platters', count, _ ).
noun( 'platypus', 'platypuses', count, _ ).
noun( 'plaudit', 'plaudits', count, _ ).
noun( 'plausibility', 'plausibilities', both, _ ).
noun( 'play', 'plays', both, _ ).
noun( 'play-acting', '-', mass, _ ).
noun( 'play-actor', 'play-actors', count, _ ).
noun( 'play-box', 'play-boxes', count, _ ).
noun( 'play-off', 'play-offs', count, _ ).
noun( 'play-reading', 'play-readings', count, _ ).
noun( 'playback', 'playbacks', count, _ ).
noun( 'playbill', 'playbills', count, _ ).
noun( 'playboy', 'playboys', count, _ ).
noun( 'player', 'players', count, _ ).
noun( 'player-piano', 'player-pianos', count, _ ).
noun( 'playfellow', 'playfellows', count, _ ).
noun( 'playfulness', '-', mass, _ ).
noun( 'playgoer', 'playgoers', count, _ ).
noun( 'playground', 'playgrounds', count, _ ).
noun( 'playgroup', 'playgroups', count, _ ).
noun( 'playhouse', 'playhouses', count, _ ).
noun( 'playing', 'playings', count, _ ).
noun( 'playing-card', 'playing-cards', count, _ ).
noun( 'playing-field', 'playing-fields', count, _ ).
noun( 'playlet', 'playlets', count, _ ).
noun( 'playmate', 'playmates', count, _ ).
noun( 'playpen', 'playpens', count, _ ).
noun( 'playroom', 'playrooms', count, _ ).
noun( 'playschool', 'playschools', count, _ ).
noun( 'playsuit', 'playsuits', count, _ ).
noun( 'plaything', 'playthings', count, _ ).
noun( 'playtime', 'playtimes', count, _ ).
noun( 'playwright', 'playwrights', count, _ ).
noun( 'plaza', 'plazas', count, _ ).
noun( 'plea', 'pleas', count, _ ).
noun( 'pleasance', 'pleasances', count, _ ).
noun( 'pleasantness', '-', mass, _ ).
noun( 'pleasantry', 'pleasantries', both, _ ).
noun( 'pleasure', 'pleasures', both, _ ).
noun( 'pleasure-boat', 'pleasure-boats', count, _ ).
noun( 'pleasure-craft', 'pleasure-crafts', count, _ ).
noun( 'pleasure-ground', 'pleasure-grounds', count, _ ).
noun( 'pleat', 'pleats', count, _ ).
noun( 'pleb', 'plebs', count, _ ).
noun( 'plebeian', 'plebeians', count, _ ).
noun( 'plebiscite', 'plebiscites', count, _ ).
noun( 'plectrum', 'plectrums', count, _ ).
noun( 'pledge', 'pledges', both, _ ).
noun( 'plenipotentiary', 'plenipotentiaries', count, _ ).
noun( 'plenitude', '-', count, _ ).
noun( 'plenty', '-', mass, _ ).
noun( 'pleonasm', 'pleonasms', both, _ ).
noun( 'plethora', 'plethoras', count, _ ).
noun( 'pleurisy', '-', mass, _ ).
noun( 'plexus', 'plexus', count, _ ).
noun( 'pliability', '-', mass, _ ).
noun( 'pliancy', '-', mass, _ ).
noun( 'plight', 'plights', count, _ ).
noun( 'plinth', 'plinths', count, _ ).
noun( 'plodder', 'plodders', count, _ ).
noun( 'plonk', 'plonks', both, _ ).
noun( 'plop', 'plops', count, _ ).
noun( 'plosive', 'plosives', count, _ ).
noun( 'plot', 'plots', count, _ ).
noun( 'plotter', 'plotters', count, _ ).
noun( 'plough', 'ploughs', both, _ ).
noun( 'ploughboy', 'ploughboys', count, _ ).
noun( 'ploughman', 'ploughmen', count, _ ).
noun( 'ploughshare', 'ploughshares', count, _ ).
noun( 'plover', 'plovers', count, _ ).
noun( 'ploy', 'ploys', count, _ ).
noun( 'pluck', 'plucks', both, _ ).
noun( 'plug', 'plugs', count, _ ).
noun( 'plughole', 'plugholes', count, _ ).
noun( 'plum', 'plums', count, _ ).
noun( 'plum-pudding', 'plum-puddings', both, _ ).
noun( 'plumage', '-', mass, _ ).
noun( 'plumb', 'plumbs', count, _ ).
noun( 'plumbago', 'plumbagos', both, _ ).
noun( 'plumber', 'plumbers', count, _ ).
noun( 'plumbing', '-', mass, _ ).
noun( 'plume', 'plumes', count, _ ).
noun( 'plummet', 'plummets', count, _ ).
noun( 'plump', 'plumps', count, _ ).
noun( 'plunder', '-', mass, _ ).
noun( 'plunderer', 'plunderers', count, _ ).
noun( 'plunge', 'plunges', count, _ ).
noun( 'plunger', 'plungers', count, _ ).
noun( 'plunk', 'plunks', count, _ ).
noun( 'pluperfect', 'pluperfects', count, _ ).
noun( 'plural', 'plurals', count, _ ).
noun( 'pluralism', '-', mass, _ ).
noun( 'pluralist', 'pluralists', count, _ ).
noun( 'plurality', 'pluralities', both, _ ).
noun( 'plus', 'pluses', count, _ ).
noun( 'plush', '-', mass, _ ).
noun( 'plutocracy', 'plutocracies', both, _ ).
noun( 'plutocrat', 'plutocrats', count, _ ).
noun( 'plutonium', '-', mass, _ ).
noun( 'ply', 'ply', count, _ ).
noun( 'plywood', '-', mass, _ ).
noun( 'pm', '-', proper, _ ).
noun( 'pneumonia', '-', mass, _ ).
noun( 'poacher', 'poachers', count, _ ).
noun( 'pock', 'pocks', count, _ ).
noun( 'pocket', 'pockets', count, _ ).
noun( 'pocket-book', 'pocket-books', count, _ ).
noun( 'pocket-handkerchief', 'pocket-handkerchiefs', count, _ ).
noun( 'pocket-knife', 'pocket-knives', count, _ ).
noun( 'pocket-money', '-', mass, _ ).
noun( 'pocketful', 'pocketfuls', count, _ ).
noun( 'pod', 'pods', count, _ ).
noun( 'podiatry', '-', mass, _ ).
noun( 'podium', 'podiums', count, _ ).
noun( 'poem', 'poems', count, _ ).
noun( 'poesy', '-', mass, _ ).
noun( 'poet', 'poets', count, _ ).
noun( 'poetess', 'poetesses', count, _ ).
noun( 'poetry', '-', mass, _ ).
noun( 'pogrom', 'pogroms', count, _ ).
noun( 'poignancy', '-', mass, _ ).
noun( 'poinsettia', 'poinsettias', count, _ ).
noun( 'point', 'points', both, _ ).
noun( 'point-duty', '-', mass, _ ).
noun( 'pointer', 'pointers', count, _ ).
noun( 'pointsman', 'pointsmen', count, _ ).
noun( 'poise', '-', mass, _ ).
noun( 'poison', 'poisons', both, _ ).
noun( 'poison-gas', 'poison-gases', both, _ ).
noun( 'poison-ivy', 'poison-ivies', both, _ ).
noun( 'poisoner', 'poisoners', count, _ ).
noun( 'poke', 'pokes', count, _ ).
noun( 'poke-bonnet', 'poke-bonnets', count, _ ).
noun( 'poker', 'pokers', both, _ ).
noun( 'poker-face', 'poker-faces', count, _ ).
noun( 'polarity', 'polarities', both, _ ).
noun( 'polarization', 'polarizations', count, _ ).
noun( 'pole', 'poles', count, _ ).
noun( 'pole-jumping', '-', mass, _ ).
noun( 'pole-star', 'pole-stars', count, _ ).
noun( 'pole-vault', 'pole-vaults', count, _ ).
noun( 'poleax', 'poleaxes', count, _ ).
noun( 'poleaxe', 'poleaxes', count, _ ).
noun( 'polecat', 'polecats', count, _ ).
noun( 'polemic', 'polemics', count, _ ).
noun( 'police-office', 'police-offices', count, _ ).
noun( 'police-officer', 'police-officers', count, _ ).
noun( 'police-station', 'police-stations', count, _ ).
noun( 'policeman', 'policemen', count, _ ).
noun( 'policewoman', 'policewomen', count, _ ).
noun( 'policy', 'policies', both, _ ).
noun( 'polio', '-', mass, _ ).
noun( 'poliomyelitis', '-', mass, _ ).
noun( 'polish', 'polishes', both, _ ).
noun( 'polisher', 'polishers', count, _ ).
noun( 'politburo', 'politburos', count, _ ).
noun( 'politeness', '-', mass, _ ).
noun( 'politician', 'politicians', count, _ ).
noun( 'politics', 'politics', mass, _ ).
noun( 'polity', 'polities', both, _ ).
noun( 'polka', 'polkas', count, _ ).
noun( 'poll', 'polls', count, _ ).
noun( 'poll', 'polls', count, _ ).
noun( 'poll-tax', 'poll-taxes', count, _ ).
noun( 'pollard', 'pollards', count, _ ).
noun( 'pollen', '-', mass, _ ).
noun( 'pollination', '-', mass, _ ).
noun( 'polling-booth', 'polling-booths', count, _ ).
noun( 'polling-day', 'polling-days', count, _ ).
noun( 'polling-station', 'polling-stations', count, _ ).
noun( 'pollster', 'pollsters', count, _ ).
noun( 'pollutant', 'pollutants', count, _ ).
noun( 'pollution', '-', mass, _ ).
noun( 'polo', '-', mass, _ ).
noun( 'polonaise', 'polonaises', count, _ ).
noun( 'polony', '-', mass, _ ).
noun( 'poltergeist', 'poltergeists', count, _ ).
noun( 'poltroon', 'poltroons', count, _ ).
noun( 'poltroonery', '-', mass, _ ).
noun( 'poly', 'polys', count, _ ).
noun( 'polyandry', '-', mass, _ ).
noun( 'polyanthus', 'polyanthuses', count, _ ).
noun( 'polygamist', 'polygamists', count, _ ).
noun( 'polygamy', '-', mass, _ ).
noun( 'polyglot', 'polyglots', count, _ ).
noun( 'polygon', 'polygons', count, _ ).
noun( 'polynomial', 'polynomials', count, _ ).
noun( 'polyp', 'polyps', count, _ ).
noun( 'polyphony', '-', mass, _ ).
noun( 'polypus', 'polypuses', count, _ ).
noun( 'polysyllable', 'polysyllables', count, _ ).
noun( 'polytechnic', 'polytechnics', count, _ ).
noun( 'polytheism', '-', mass, _ ).
noun( 'polythene', '-', mass, _ ).
noun( 'pom', 'poms', count, _ ).
noun( 'pomade', '-', mass, _ ).
noun( 'pomegranate', 'pomegranates', count, _ ).
noun( 'pomelo', 'pomelos', count, _ ).
noun( 'pommel', 'pommels', count, _ ).
noun( 'pommy', 'pommies', count, _ ).
noun( 'pomp', '-', mass, _ ).
noun( 'pompon', 'pompons', count, _ ).
noun( 'pomposity', 'pomposities', both, _ ).
noun( 'ponce', 'ponces', count, _ ).
noun( 'poncho', 'ponchos', count, _ ).
noun( 'pond', 'ponds', count, _ ).
noun( 'pone', 'pones', count, _ ).
noun( 'pongee', '-', mass, _ ).
noun( 'poniard', 'poniards', count, _ ).
noun( 'pontiff', 'pontiffs', count, _ ).
noun( 'pontificate', 'pontificates', count, _ ).
noun( 'pontoon', 'pontoons', both, _ ).
noun( 'pony', 'ponies', count, _ ).
noun( 'pony-trekking', '-', mass, _ ).
noun( 'ponytail', 'ponytails', count, _ ).
noun( 'poodle', 'poodles', count, _ ).
noun( 'poof', 'poofs', count, _ ).
noun( 'pool', 'pools', both, _ ).
noun( 'poolroom', 'poolrooms', count, _ ).
noun( 'poop', 'poops', count, _ ).
noun( 'poor-box', 'poor-boxes', count, _ ).
noun( 'poor-rate', 'poor-rates', count, _ ).
noun( 'poorhouse', 'poorhouses', count, _ ).
noun( 'poorness', '-', mass, _ ).
noun( 'pop', 'pops', both, _ ).
noun( 'pop', '-', count, _ ).
noun( 'popcorn', 'popcorns', count, _ ).
noun( 'popery', '-', mass, _ ).
noun( 'popgun', 'popguns', count, _ ).
noun( 'popinjay', 'popinjays', count, _ ).
noun( 'poplar', 'poplars', both, _ ).
noun( 'poplin', '-', mass, _ ).
noun( 'poppa', 'poppas', count, _ ).
noun( 'poppet', 'poppets', count, _ ).
noun( 'poppy', 'poppies', count, _ ).
noun( 'poppycock', '-', mass, _ ).
noun( 'populace', 'populaces', count, _ ).
noun( 'popularity', '-', mass, _ ).
noun( 'popularization', 'popularizations', both, _ ).
noun( 'population', 'populations', count, _ ).
noun( 'populism', '-', mass, _ ).
noun( 'populist', 'populists', count, _ ).
noun( 'porcelain', '-', mass, _ ).
noun( 'porch', 'porches', count, _ ).
noun( 'porcupine', 'porcupines', count, _ ).
noun( 'pore', 'pores', count, _ ).
noun( 'pork', '-', mass, _ ).
noun( 'pork-barrel', 'pork-barrels', count, _ ).
noun( 'pork-butcher', 'pork-butchers', count, _ ).
noun( 'porker', 'porkers', count, _ ).
noun( 'porn', '-', mass, _ ).
noun( 'pornographer', 'pornographers', count, _ ).
noun( 'pornography', '-', mass, _ ).
noun( 'porosity', 'porosities', both, _ ).
noun( 'porousness', '-', mass, _ ).
noun( 'porphyry', '-', mass, _ ).
noun( 'porpoise', 'porpoises', count, _ ).
noun( 'porridge', '-', mass, _ ).
noun( 'porringer', 'porringers', count, _ ).
noun( 'port', 'ports', both, _ ).
noun( 'portability', '-', mass, _ ).
noun( 'portage', 'portages', both, _ ).
noun( 'portal', 'portals', count, _ ).
noun( 'portcullis', 'portcullises', count, _ ).
noun( 'porte-coch`ere', 'porte-coch`eres', count, _ ).
noun( 'portent', 'portents', count, _ ).
noun( 'porter', 'porters', both, _ ).
noun( 'porterage', '-', mass, _ ).
noun( 'porterhouse', 'porterhouses', count, _ ).
noun( 'portfolio', 'portfolios', count, _ ).
noun( 'porthole', 'portholes', count, _ ).
noun( 'porti`ere', 'porti`eres', count, _ ).
noun( 'portico', 'porticos', count, _ ).
noun( 'portion', 'portions', count, _ ).
noun( 'portmanteau', 'portmanteaus', count, _ ).
noun( 'portrait', 'portraits', count, _ ).
noun( 'portraitist', 'portraitists', count, _ ).
noun( 'portraiture', '-', mass, _ ).
noun( 'portrayal', 'portrayals', both, _ ).
noun( 'pose', 'poses', count, _ ).
noun( 'poser', 'posers', count, _ ).
noun( 'poseur', 'poseurs', count, _ ).
noun( 'poseuse', 'poseuses', count, _ ).
noun( 'position', 'positions', both, _ ).
noun( 'positive', 'positives', count, _ ).
noun( 'positiveness', '-', mass, _ ).
noun( 'positivism', '-', mass, _ ).
noun( 'positivist', 'positivists', count, _ ).
noun( 'posse', 'posses', count, _ ).
noun( 'possession', 'possessions', both, _ ).
noun( 'possessiveness', '-', mass, _ ).
noun( 'possessor', 'possessors', count, _ ).
noun( 'posset', 'possets', count, _ ).
noun( 'possibility', 'possibilities', both, _ ).
noun( 'possible', 'possibles', count, _ ).
noun( 'possum', 'possums', count, _ ).
noun( 'post', 'posts', both, _ ).
noun( 'post-chaise', 'post-chaises', count, _ ).
noun( 'post-horse', 'post-horses', count, _ ).
noun( 'post-mortem', 'post-mortems', count, _ ).
noun( 'postage', '-', mass, _ ).
noun( 'postage-stamp', 'postage-stamps', count, _ ).
noun( 'postbag', 'postbags', count, _ ).
noun( 'postbox', 'postboxes', count, _ ).
noun( 'postcard', 'postcards', count, _ ).
noun( 'postcode', 'postcodes', count, _ ).
noun( 'poste restante', '-', mass, _ ).
noun( 'poster', 'posters', count, _ ).
noun( 'posterior', 'posteriors', count, _ ).
noun( 'posterity', '-', mass, _ ).
noun( 'postern', 'posterns', count, _ ).
noun( 'postgraduate', 'postgraduates', count, _ ).
noun( 'postilion', 'postilions', count, _ ).
noun( 'postillion', 'postillions', count, _ ).
noun( 'postman', 'postmen', count, _ ).
noun( 'postmark', 'postmarks', count, _ ).
noun( 'postmaster', 'postmasters', count, _ ).
noun( 'postmistress', 'postmistresses', count, _ ).
noun( 'postponement', 'postponements', both, _ ).
noun( 'postscript', 'postscripts', count, _ ).
noun( 'postulant', 'postulants', count, _ ).
noun( 'postulate', 'postulates', count, _ ).
noun( 'posture', 'postures', both, _ ).
noun( 'posturing', 'posturings', both, _ ).
noun( 'posy', 'posies', count, _ ).
noun( 'pot', 'pots', count, _ ).
noun( 'pot-shot', 'pot-shots', count, _ ).
noun( 'potash', '-', mass, _ ).
noun( 'potassium', '-', mass, _ ).
noun( 'potation', 'potations', count, _ ).
noun( 'potato', 'potatoes', count, _ ).
noun( 'potbelly', 'potbellies', count, _ ).
noun( 'potboiler', 'potboilers', count, _ ).
noun( 'potboy', 'potboys', count, _ ).
noun( 'poteen', '-', mass, _ ).
noun( 'potency', 'potencies', both, _ ).
noun( 'potentate', 'potentates', count, _ ).
noun( 'potential', 'potentials', both, _ ).
noun( 'potentiality', 'potentialities', both, _ ).
noun( 'pothead', 'potheads', count, _ ).
noun( 'pother', 'pothers', count, _ ).
noun( 'potherb', 'potherbs', count, _ ).
noun( 'pothole', 'potholes', count, _ ).
noun( 'potholer', 'potholers', count, _ ).
noun( 'pothook', 'pothooks', count, _ ).
noun( 'pothouse', 'pothouses', count, _ ).
noun( 'pothunter', 'pothunters', count, _ ).
noun( 'potion', 'potions', count, _ ).
noun( 'potman', 'potmen', count, _ ).
noun( 'potpourri', 'potpourris', count, _ ).
noun( 'potsherd', 'potsherds', count, _ ).
noun( 'pottage', 'pottages', both, _ ).
noun( 'potter', 'potters', count, _ ).
noun( 'potterer', 'potterers', count, _ ).
noun( 'pottery', 'potteries', both, _ ).
noun( 'potty', 'potties', count, _ ).
noun( 'pouch', 'pouches', count, _ ).
noun( 'pouf', 'poufs', count, _ ).
noun( 'pouffe', 'pouffes', count, _ ).
noun( 'poulterer', 'poulterers', count, _ ).
noun( 'poultice', 'poultices', count, _ ).
noun( 'pounce', 'pounces', count, _ ).
noun( 'pound', 'pounds', count, _ ).
noun( 'poundage', '-', mass, _ ).
noun( 'pounder', 'pounders', count, _ ).
noun( 'pout', 'pouts', count, _ ).
noun( 'poverty', '-', mass, _ ).
noun( 'powder', 'powders', both, _ ).
noun( 'powder-flask', 'powder-flasks', count, _ ).
noun( 'powder-horn', 'powder-horns', count, _ ).
noun( 'powder-magazine', 'powder-magazines', count, _ ).
noun( 'powder-puff', 'powder-puffs', count, _ ).
noun( 'powder-room', 'powder-rooms', count, _ ).
noun( 'power', 'powers', both, _ ).
noun( 'power-dive', 'power-dives', count, _ ).
noun( 'power-point', 'power-points', count, _ ).
noun( 'power-station', 'power-stations', count, _ ).
noun( 'powerboat', 'powerboats', count, _ ).
noun( 'powerhouse', 'powerhouses', count, _ ).
noun( 'powwow', 'powwows', count, _ ).
noun( 'pox', '-', count, _ ).
noun( '-', 'pp', count, _ ).
noun( 'pr_ecis', 'pr_ecis', count, _ ).
noun( 'practicability', '-', mass, _ ).
noun( 'practicality', 'practicalities', both, _ ).
noun( 'practice', 'practices', both, _ ).
noun( 'practician', 'practicians', count, _ ).
noun( 'practitioner', 'practitioners', count, _ ).
noun( 'praesidium', 'praesidiums', count, _ ).
noun( 'praetor', 'praetors', count, _ ).
noun( 'pragmatism', '-', mass, _ ).
noun( 'pragmatist', 'pragmatists', count, _ ).
noun( 'prairie', 'prairies', count, _ ).
noun( 'praise', 'praises', both, _ ).
noun( 'praiseworthiness', '-', mass, _ ).
noun( 'pram', 'prams', count, _ ).
noun( 'prance', 'prances', count, _ ).
noun( 'prank', 'pranks', count, _ ).
noun( 'prattle', '-', mass, _ ).
noun( 'prattler', 'prattlers', count, _ ).
noun( 'prawn', 'prawns', count, _ ).
noun( 'prayer', 'prayers', both, _ ).
noun( 'prayer-book', 'prayer-books', count, _ ).
noun( 'prayer-mat', 'prayer-mats', count, _ ).
noun( 'prayer-meeting', 'prayer-meetings', count, _ ).
noun( 'prayer-rug', 'prayer-rugs', count, _ ).
noun( 'prayer-wheel', 'prayer-wheels', count, _ ).
noun( 'pre-eminence', '-', mass, _ ).
noun( 'pre-emption', '-', mass, _ ).
noun( 'pre-existence', 'pre-existences', count, _ ).
noun( 'preacher', 'preachers', count, _ ).
noun( 'preamble', 'preambles', count, _ ).
noun( 'prearrangement', 'prearrangements', count, _ ).
noun( 'prebend', 'prebends', count, _ ).
noun( 'prebendary', 'prebendaries', count, _ ).
noun( 'precariousness', '-', mass, _ ).
noun( 'precaution', 'precautions', both, _ ).
noun( 'precedence', '-', mass, _ ).
noun( 'precedent', 'precedents', count, _ ).
noun( 'precentor', 'precentors', count, _ ).
noun( 'precept', 'precepts', both, _ ).
noun( 'preceptor', 'preceptors', count, _ ).
noun( 'precession', 'precessions', count, _ ).
noun( 'precinct', 'precincts', count, _ ).
noun( 'preciosity', 'preciosities', both, _ ).
noun( 'preciousness', '-', mass, _ ).
noun( 'precipice', 'precipices', count, _ ).
noun( 'precipitate', 'precipitates', count, _ ).
noun( 'precipitation', '-', mass, _ ).
noun( 'preciseness', '-', mass, _ ).
noun( 'precision', '-', mass, _ ).
noun( 'preclusion', 'preclusions', count, _ ).
noun( 'precociousness', '-', mass, _ ).
noun( 'precocity', '-', mass, _ ).
noun( 'precognition', '-', mass, _ ).
noun( 'preconception', 'preconceptions', both, _ ).
noun( 'precondition', 'preconditions', count, _ ).
noun( 'precursor', 'precursors', count, _ ).
noun( 'predator', 'predators', count, _ ).
noun( 'predecessor', 'predecessors', count, _ ).
noun( 'predestination', 'predestinations', count, _ ).
noun( 'predetermination', 'predeterminations', count, _ ).
noun( 'predicament', 'predicaments', count, _ ).
noun( 'predicate', 'predicates', count, _ ).
noun( 'predictability', '-', mass, _ ).
noun( 'prediction', 'predictions', both, _ ).
noun( 'predictor', 'predictors', count, _ ).
noun( 'predilection', 'predilections', count, _ ).
noun( 'predisposition', 'predispositions', count, _ ).
noun( 'predominance', '-', mass, _ ).
noun( 'prefab', 'prefabs', count, _ ).
noun( 'prefabrication', 'prefabrications', count, _ ).
noun( 'preface', 'prefaces', count, _ ).
noun( 'prefect', 'prefects', count, _ ).
noun( 'prefecture', 'prefectures', count, _ ).
noun( 'preference', 'preferences', both, _ ).
noun( 'preferment', '-', mass, _ ).
noun( 'prefix', 'prefixes', count, _ ).
noun( 'pregnancy', 'pregnancies', both, _ ).
noun( 'prehistory', 'prehistories', both, _ ).
noun( 'prejudgement', 'prejudgements', count, _ ).
noun( 'prejudice', 'prejudices', both, _ ).
noun( 'prelacy', 'prelacies', count, _ ).
noun( 'prelate', 'prelates', count, _ ).
noun( 'prelim', 'prelims', count, _ ).
noun( 'preliminary', 'preliminaries', count, _ ).
noun( 'prelude', 'preludes', count, _ ).
noun( 'premeditation', '-', mass, _ ).
noun( 'premi`ere', 'premi`eres', count, _ ).
noun( 'premier', 'premiers', count, _ ).
noun( 'premiership', 'premierships', count, _ ).
noun( 'premise', 'premises', count, _ ).
noun( 'premiss', 'premisses', count, _ ).
noun( 'premium', 'premiums', count, _ ).
noun( 'premonition', 'premonitions', count, _ ).
noun( 'prentice', 'prentices', count, _ ).
noun( 'preoccupation', 'preoccupations', both, _ ).
noun( 'prep', 'preps', both, _ ).
noun( 'preparation', 'preparations', both, _ ).
noun( 'preparedness', '-', mass, _ ).
noun( 'preponderance', 'preponderances', count, _ ).
noun( 'preposition', 'prepositions', count, _ ).
noun( 'prepossession', 'prepossessions', count, _ ).
noun( 'prepuce', 'prepuces', count, _ ).
noun( 'prerequisite', 'prerequisites', count, _ ).
noun( 'prerogative', 'prerogatives', count, _ ).
noun( 'presage', 'presages', count, _ ).
noun( 'presbyter', 'presbyters', count, _ ).
noun( 'presbytery', 'presbyteries', count, _ ).
noun( 'prescience', '-', mass, _ ).
noun( 'prescript', 'prescripts', count, _ ).
noun( 'prescription', 'prescriptions', both, _ ).
noun( 'presence', '-', mass, _ ).
noun( 'present', 'presents', count, _ ).
noun( 'present', '-', count, _ ).
noun( 'presentation', 'presentations', both, _ ).
noun( 'presentiment', 'presentiments', count, _ ).
noun( 'preservation', '-', mass, _ ).
noun( 'preservative', 'preservatives', count, _ ).
noun( 'preserve', 'preserves', count, _ ).
noun( 'preserver', 'preservers', count, _ ).
noun( 'presidency', 'presidencies', count, _ ).
noun( 'president', 'presidents', count, _ ).
noun( 'presidium', 'presidiums', count, _ ).
noun( 'press', 'presses', count, _ ).
noun( 'press-agency', 'press-agencies', count, _ ).
noun( 'press-agent', 'press-agents', count, _ ).
noun( 'press-box', 'press-boxes', count, _ ).
noun( 'press-clipping', 'press-clippings', count, _ ).
noun( 'press-cutting', 'press-cuttings', count, _ ).
noun( 'press-gallery', 'press-galleries', count, _ ).
noun( 'press-gang', 'press-gangs', count, _ ).
noun( 'press-lord', 'press-lords', count, _ ).
noun( 'press-photographer', 'press-photographers', count, _ ).
noun( 'press-stud', 'press-studs', count, _ ).
noun( 'press-up', 'press-ups', count, _ ).
noun( 'pressing', 'pressings', count, _ ).
noun( 'pressman', 'pressmen', count, _ ).
noun( 'pressmark', 'pressmarks', count, _ ).
noun( 'pressure', 'pressures', both, _ ).
noun( 'pressure-cooker', 'pressure-cookers', count, _ ).
noun( 'pressure-gauge', 'pressure-gauges', count, _ ).
noun( 'prestidigitation', 'prestidigitations', both, _ ).
noun( 'prestidigitator', 'prestidigitators', count, _ ).
noun( 'prestige', '-', mass, _ ).
noun( 'presumption', 'presumptions', both, _ ).
noun( 'presupposition', 'presuppositions', both, _ ).
noun( 'pretence', 'pretences', both, _ ).
noun( 'pretender', 'pretenders', count, _ ).
noun( 'pretension', 'pretensions', both, _ ).
noun( 'pretentiousness', '-', mass, _ ).
noun( 'preterit', 'preterits', count, _ ).
noun( 'preterite', 'preterites', count, _ ).
noun( 'pretext', 'pretexts', count, _ ).
noun( 'pretor', 'pretors', count, _ ).
noun( 'prettiness', '-', mass, _ ).
noun( 'pretty', 'pretties', count, _ ).
noun( 'pretzel', 'pretzels', count, _ ).
noun( 'prevalence', '-', mass, _ ).
noun( 'prevarication', 'prevarications', both, _ ).
noun( 'preventative', 'preventatives', count, _ ).
noun( 'prevention', '-', mass, _ ).
noun( 'preview', 'previews', count, _ ).
noun( 'prevision', 'previsions', both, _ ).
noun( 'prey', '-', count, _ ).
noun( 'price', 'prices', both, _ ).
noun( 'price-control', 'price-controls', count, _ ).
noun( 'pricelist', 'pricelists', count, _ ).
noun( 'prick', 'pricks', count, _ ).
noun( 'pricker', 'prickers', count, _ ).
noun( 'pricking', 'prickings', count, _ ).
noun( 'prickle', 'prickles', count, _ ).
noun( 'pride', 'prides', both, _ ).
noun( 'prie-dieu', 'prie-dieus', count, _ ).
noun( 'priest', 'priests', count, _ ).
noun( 'priestcraft', '-', mass, _ ).
noun( 'priestess', 'priestesses', count, _ ).
noun( 'priesthood', 'priesthoods', count, _ ).
noun( 'prig', 'prigs', count, _ ).
noun( 'priggishness', '-', mass, _ ).
noun( 'prima ballerina', 'prima ballerinas', count, _ ).
noun( 'prima donna', 'prima donnas', count, _ ).
noun( 'primacy', 'primacies', count, _ ).
noun( 'primary', 'primaries', count, _ ).
noun( 'primate', 'primates', count, _ ).
noun( 'prime', '-', mass, _ ).
noun( 'primer', 'primers', count, _ ).
noun( 'priming', 'primings', count, _ ).
noun( 'primitive', 'primitives', count, _ ).
noun( 'primitiveness', '-', mass, _ ).
noun( 'primness', '-', mass, _ ).
noun( 'primogeniture', '-', mass, _ ).
noun( 'primrose', 'primroses', count, _ ).
noun( 'primula', 'primulas', count, _ ).
noun( 'primus', 'primuses', count, _ ).
noun( 'prince', 'princes', count, _ ).
noun( 'princedom', 'princedoms', count, _ ).
noun( 'princess', 'princesses', count, _ ).
noun( 'principal', 'principals', count, _ ).
noun( 'principality', 'principalities', count, _ ).
noun( 'principle', 'principles', count, _ ).
noun( 'print', 'prints', both, _ ).
noun( 'print-seller', 'print-sellers', count, _ ).
noun( 'print-shop', 'print-shops', count, _ ).
noun( 'printer', 'printers', count, _ ).
noun( 'printing', 'printings', both, _ ).
noun( 'printing-ink', 'printing-inks', count, _ ).
noun( 'printing-press', 'printing-presses', count, _ ).
noun( 'printout', 'printouts', count, _ ).
noun( 'prior', 'priors', count, _ ).
noun( 'prioress', 'prioresses', count, _ ).
noun( 'priority', 'priorities', both, _ ).
noun( 'priory', 'priories', count, _ ).
noun( 'prism', 'prisms', count, _ ).
noun( 'prison', 'prisons', both, _ ).
noun( 'prison-breaking', 'prison-breakings', both, _ ).
noun( 'prisoner', 'prisoners', count, _ ).
noun( 'privacy', '-', mass, _ ).
noun( 'private', 'privates', count, _ ).
noun( 'privateer', 'privateers', count, _ ).
noun( 'privation', 'privations', both, _ ).
noun( 'privet', '-', mass, _ ).
noun( 'privilege', 'privileges', both, _ ).
noun( 'privy', 'privies', count, _ ).
noun( 'prize', 'prizes', count, _ ).
noun( 'prize-fight', 'prize-fights', count, _ ).
noun( 'prize-money', '-', mass, _ ).
noun( 'prize-ring', 'prize-rings', count, _ ).
noun( 'pro', 'pros', count, _ ).
noun( 'probability', 'probabilities', both, _ ).
noun( 'probable', 'probables', count, _ ).
noun( 'probate', 'probates', both, _ ).
noun( 'probation', '-', mass, _ ).
noun( 'probationer', 'probationers', count, _ ).
noun( 'probe', 'probes', count, _ ).
noun( 'probity', '-', mass, _ ).
noun( 'problem', 'problems', count, _ ).
noun( 'proboscis', 'proboscises', count, _ ).
noun( 'procedure', 'procedures', both, _ ).
noun( 'proceeding', 'proceedings', both, _ ).
noun( 'process', 'processes', both, _ ).
noun( 'process-server', 'process-servers', count, _ ).
noun( 'procession', 'processions', both, _ ).
noun( 'processor', 'processors', count, _ ).
noun( 'proclamation', 'proclamations', both, _ ).
noun( 'proclivity', 'proclivities', count, _ ).
noun( 'proconsul', 'proconsuls', count, _ ).
noun( 'proconsulate', 'proconsulates', count, _ ).
noun( 'procrastination', '-', mass, _ ).
noun( 'procreation', 'procreations', count, _ ).
noun( 'proctor', 'proctors', count, _ ).
noun( 'procurator', 'procurators', count, _ ).
noun( 'procurement', 'procurements', count, _ ).
noun( 'procurer', 'procurers', count, _ ).
noun( 'procuress', 'procuresses', count, _ ).
noun( 'prod', 'prods', count, _ ).
noun( 'prodigal', 'prodigals', count, _ ).
noun( 'prodigality', '-', mass, _ ).
noun( 'prodigy', 'prodigies', count, _ ).
noun( 'produce', '-', mass, _ ).
noun( 'producer', 'producers', count, _ ).
noun( 'product', 'products', count, _ ).
noun( 'production', 'productions', both, _ ).
noun( 'productivity', '-', mass, _ ).
noun( 'profanation', 'profanations', both, _ ).
noun( 'profaneness', '-', mass, _ ).
noun( 'profanity', 'profanities', both, _ ).
noun( 'profession', 'professions', count, _ ).
noun( 'professional', 'professionals', count, _ ).
noun( 'professionalism', '-', mass, _ ).
noun( 'professor', 'professors', count, _ ).
noun( 'professorship', 'professorships', count, _ ).
noun( 'proffer', 'proffers', count, _ ).
noun( 'proficiency', '-', mass, _ ).
noun( 'profile', 'profiles', both, _ ).
noun( 'profit', 'profits', both, _ ).
noun( 'profit-margin', 'profit-margins', count, _ ).
noun( 'profit-sharing', '-', mass, _ ).
noun( 'profiteer', 'profiteers', count, _ ).
noun( 'profligacy', '-', mass, _ ).
noun( 'profligate', 'profligates', count, _ ).
noun( 'profundity', 'profundities', both, _ ).
noun( 'profuseness', '-', mass, _ ).
noun( 'profusion', '-', mass, _ ).
noun( 'progenitor', 'progenitors', count, _ ).
noun( 'prognosis', 'prognoses', count, _ ).
noun( 'prognostic', 'prognostics', count, _ ).
noun( 'prognostication', 'prognostications', both, _ ).
noun( 'program', 'programs', count, _ ).
noun( 'programme', 'programmes', count, _ ).
noun( 'programmer', 'programmers', count, _ ).
noun( 'progress', 'progresses', both, _ ).
noun( 'progression', '-', mass, _ ).
noun( 'progressive', 'progressives', count, _ ).
noun( 'progressiveness', '-', mass, _ ).
noun( 'prohibition', 'prohibitions', both, _ ).
noun( 'prohibitionist', 'prohibitionists', count, _ ).
noun( 'project', 'projects', count, _ ).
noun( 'projectile', 'projectiles', count, _ ).
noun( 'projection', 'projections', both, _ ).
noun( 'projectionist', 'projectionists', count, _ ).
noun( 'projector', 'projectors', count, _ ).
noun( 'prolapse', 'prolapses', count, _ ).
noun( 'prole', 'proles', count, _ ).
noun( 'proletarian', 'proletarians', count, _ ).
noun( 'proletariat', 'proletariats', count, _ ).
noun( 'proliferation', 'proliferations', count, _ ).
noun( 'prolixity', '-', mass, _ ).
noun( 'prologue', 'prologues', count, _ ).
noun( 'prolongation', 'prolongations', both, _ ).
noun( 'prom', 'proms', count, _ ).
noun( 'promenade', 'promenades', count, _ ).
noun( 'prominence', 'prominences', both, _ ).
noun( 'promiscuity', '-', mass, _ ).
noun( 'promise', 'promises', both, _ ).
noun( 'promontory', 'promontories', count, _ ).
noun( 'promoter', 'promoters', count, _ ).
noun( 'promotion', 'promotions', both, _ ).
noun( 'prompt', 'prompts', count, _ ).
noun( 'prompt-box', 'prompt-boxes', count, _ ).
noun( 'prompt-copy', 'prompt-copies', count, _ ).
noun( 'prompter', 'prompters', count, _ ).
noun( 'prompting', 'promptings', count, _ ).
noun( 'promptitude', '-', mass, _ ).
noun( 'promptness', '-', mass, _ ).
noun( 'promulgation', '-', mass, _ ).
noun( 'proneness', '-', mass, _ ).
noun( 'prong', 'prongs', count, _ ).
noun( 'pronoun', 'pronouns', count, _ ).
noun( 'pronouncement', 'pronouncements', count, _ ).
noun( 'pronunciamento', 'pronunciamentos', count, _ ).
noun( 'pronunciation', 'pronunciations', both, _ ).
noun( 'proof', 'proofs', both, _ ).
noun( 'proofreader', 'proofreaders', count, _ ).
noun( 'prop', 'props', count, _ ).
noun( 'propaganda', '-', mass, _ ).
noun( 'propagandist', 'propagandists', count, _ ).
noun( 'propagation', '-', mass, _ ).
noun( 'propagator', 'propagators', count, _ ).
noun( 'propane', '-', mass, _ ).
noun( 'propellant', 'propellants', both, _ ).
noun( 'propellent', 'propellents', both, _ ).
noun( 'propeller', 'propellers', count, _ ).
noun( 'propensity', 'propensities', both, _ ).
noun( 'property', 'properties', both, _ ).
noun( 'property-man', 'property-men', count, _ ).
noun( 'property-master', 'property-masters', count, _ ).
noun( 'prophecy', 'prophecies', both, _ ).
noun( 'prophet', 'prophets', count, _ ).
noun( 'prophetess', 'prophetesses', count, _ ).
noun( 'prophylactic', 'prophylactics', count, _ ).
noun( 'prophylaxis', '-', mass, _ ).
noun( 'propinquity', '-', mass, _ ).
noun( 'propitiation', '-', mass, _ ).
noun( 'proponent', 'proponents', count, _ ).
noun( 'proportion', 'proportions', both, _ ).
noun( 'proportionality', '-', mass, _ ).
noun( 'proposal', 'proposals', both, _ ).
noun( 'proposer', 'proposers', count, _ ).
noun( 'proposition', 'propositions', count, _ ).
noun( 'proprietor', 'proprietors', count, _ ).
noun( 'proprietress', 'proprietresses', count, _ ).
noun( 'propriety', 'proprieties', both, _ ).
noun( 'propulsion', '-', mass, _ ).
noun( 'prorogation', 'prorogations', count, _ ).
noun( 'proscenium', 'prosceniums', count, _ ).
noun( 'proscription', 'proscriptions', both, _ ).
noun( 'prose', '-', mass, _ ).
noun( 'prosecution', 'prosecutions', both, _ ).
noun( 'prosecutor', 'prosecutors', count, _ ).
noun( 'proselyte', 'proselytes', count, _ ).
noun( 'prosiness', '-', mass, _ ).
noun( 'prosody', '-', mass, _ ).
noun( 'prospect', 'prospects', both, _ ).
noun( 'prospector', 'prospectors', count, _ ).
noun( 'prospectus', 'prospectuses', count, _ ).
noun( 'prosperity', '-', mass, _ ).
noun( 'prostate', 'prostates', count, _ ).
noun( 'prostitute', 'prostitutes', count, _ ).
noun( 'prostitution', '-', mass, _ ).
noun( 'prostration', 'prostrations', both, _ ).
noun( 'prot_eg_e', 'prot_eg_es', count, _ ).
noun( 'prot_eg_ee', 'prot_eg_ees', count, _ ).
noun( 'protagonist', 'protagonists', count, _ ).
noun( 'protection', 'protections', both, _ ).
noun( 'protectionism', '-', mass, _ ).
noun( 'protectionist', 'protectionists', count, _ ).
noun( 'protector', 'protectors', count, _ ).
noun( 'protectorate', 'protectorates', count, _ ).
noun( 'protein', 'proteins', both, _ ).
noun( 'protest', 'protests', both, _ ).
noun( 'protestation', 'protestations', count, _ ).
noun( 'protester', 'protesters', count, _ ).
noun( 'protocol', 'protocols', both, _ ).
noun( 'proton', 'protons', count, _ ).
noun( 'protoplasm', '-', mass, _ ).
noun( 'prototype', 'prototypes', count, _ ).
noun( 'protraction', 'protractions', both, _ ).
noun( 'protractor', 'protractors', count, _ ).
noun( 'protrusion', 'protrusions', both, _ ).
noun( 'protuberance', 'protuberances', both, _ ).
noun( 'provenance', '-', mass, _ ).
noun( 'provender', '-', mass, _ ).
noun( 'proverb', 'proverbs', count, _ ).
noun( 'providence', '-', mass, _ ).
noun( 'provider', 'providers', count, _ ).
noun( 'province', 'provinces', count, _ ).
noun( 'provincial', 'provincials', count, _ ).
noun( 'provincialism', 'provincialisms', both, _ ).
noun( 'provision', 'provisions', both, _ ).
noun( 'proviso', 'provisos', count, _ ).
noun( 'provocation', 'provocations', both, _ ).
noun( 'provost', 'provosts', count, _ ).
noun( 'prow', 'prows', count, _ ).
noun( 'prowess', '-', mass, _ ).
noun( 'prowl', 'prowls', count, _ ).
noun( 'prowler', 'prowlers', count, _ ).
noun( 'proximity', '-', mass, _ ).
noun( 'proxy', 'proxies', both, _ ).
noun( 'prude', 'prudes', count, _ ).
noun( 'prudence', '-', mass, _ ).
noun( 'prudery', 'pruderies', both, _ ).
noun( 'prune', 'prunes', count, _ ).
noun( 'pruning', '-', mass, _ ).
noun( 'pruning-hook', 'pruning-hooks', count, _ ).
noun( 'pruning-knife', 'pruning-knives', count, _ ).
noun( 'pruning-saw', 'pruning-saws', count, _ ).
noun( 'prurience', '-', mass, _ ).
noun( 'pruriency', '-', mass, _ ).
noun( 'psalm', 'psalms', count, _ ).
noun( 'psalmist', 'psalmists', count, _ ).
noun( 'psalmody', 'psalmodies', both, _ ).
noun( 'psalter', 'psalters', count, _ ).
noun( 'psaltery', 'psalteries', count, _ ).
noun( 'psephologist', 'psephologists', count, _ ).
noun( 'psephology', '-', mass, _ ).
noun( 'pseud', 'pseuds', count, _ ).
noun( 'pseudo', 'pseudos', count, _ ).
noun( 'pseudonym', 'pseudonyms', count, _ ).
noun( 'psittacosis', '-', mass, _ ).
noun( 'psyche', 'psyches', count, _ ).
noun( 'psychiatrist', 'psychiatrists', count, _ ).
noun( 'psychiatry', '-', mass, _ ).
noun( 'psychic', 'psychics', count, _ ).
noun( 'psychoanalysis', '-', mass, _ ).
noun( 'psychoanalyst', 'psychoanalysts', count, _ ).
noun( 'psychologist', 'psychologists', count, _ ).
noun( 'psychology', 'psychologies', both, _ ).
noun( 'psychopath', 'psychopaths', count, _ ).
noun( 'psychosis', 'psychoses', count, _ ).
noun( 'psychotherapy', '-', mass, _ ).
noun( 'psychotic', 'psychotics', count, _ ).
noun( 'pt', 'pt', count, _ ).
noun( 'ptarmigan', 'ptarmigans', count, _ ).
noun( 'pterodactyl', 'pterodactyls', count, _ ).
noun( 'ptomaine', 'ptomaines', count, _ ).
noun( 'pub', 'pubs', count, _ ).
noun( 'pub-crawl', 'pub-crawls', count, _ ).
noun( 'puberty', '-', mass, _ ).
noun( 'public', 'publics', count, _ ).
noun( 'publican', 'publicans', count, _ ).
noun( 'publication', 'publications', both, _ ).
noun( 'publicist', 'publicists', count, _ ).
noun( 'publicity', '-', mass, _ ).
noun( 'publisher', 'publishers', count, _ ).
noun( 'puce', '-', mass, _ ).
noun( 'puck', 'pucks', count, _ ).
noun( 'pucker', 'puckers', count, _ ).
noun( 'pud', '-', mass, _ ).
noun( 'pudden', 'puddens', count, _ ).
noun( 'pudden-head', 'pudden-heads', count, _ ).
noun( 'pudding', 'puddings', both, _ ).
noun( 'pudding-face', 'pudding-faces', count, _ ).
noun( 'puddle', 'puddles', both, _ ).
noun( 'puddler', 'puddlers', count, _ ).
noun( 'pueblo', 'pueblos', count, _ ).
noun( 'puerility', 'puerilities', both, _ ).
noun( 'puff', 'puffs', count, _ ).
noun( 'puffball', 'puffballs', count, _ ).
noun( 'puffin', 'puffins', count, _ ).
noun( 'puffiness', '-', mass, _ ).
noun( 'pug', 'pugs', count, _ ).
noun( 'pug-dog', 'pug-dogs', count, _ ).
noun( 'pug-nose', 'pug-noses', count, _ ).
noun( 'pugilism', '-', mass, _ ).
noun( 'pugilist', 'pugilists', count, _ ).
noun( 'pugnacity', '-', mass, _ ).
noun( 'puissance', '-', mass, _ ).
noun( 'puke', '-', mass, _ ).
noun( 'pulchritude', '-', mass, _ ).
noun( 'pull', 'pulls', both, _ ).
noun( 'pull-in', 'pull-ins', count, _ ).
noun( 'pull-off', 'pull-offs', count, _ ).
noun( 'pull-out', 'pull-outs', count, _ ).
noun( 'pull-through', 'pull-throughs', count, _ ).
noun( 'pull-up', 'pull-ups', count, _ ).
noun( 'pullet', 'pullets', count, _ ).
noun( 'pulley', 'pulleys', count, _ ).
noun( 'pulley-block', 'pulley-blocks', count, _ ).
noun( 'pullover', 'pullovers', count, _ ).
noun( 'pulp', 'pulps', both, _ ).
noun( 'pulpit', 'pulpits', count, _ ).
noun( 'pulque', '-', mass, _ ).
noun( 'pulsar', 'pulsars', count, _ ).
noun( 'pulsation', 'pulsations', both, _ ).
noun( 'pulse', 'pulses', both, _ ).
noun( 'puma', 'pumas', count, _ ).
noun( 'pumice', '-', mass, _ ).
noun( 'pumice-stone', 'pumice-stones', count, _ ).
noun( 'pump', 'pumps', count, _ ).
noun( 'pump-room', 'pump-rooms', count, _ ).
noun( 'pumpernickel', '-', mass, _ ).
noun( 'pumpkin', 'pumpkins', count, _ ).
noun( 'pun', 'puns', count, _ ).
noun( 'punch', 'punches', both, _ ).
noun( 'punch-up', 'punch-ups', count, _ ).
noun( 'punchball', 'punchballs', count, _ ).
noun( 'punchbowl', 'punchbowls', count, _ ).
noun( 'punching-ball', 'punching-balls', count, _ ).
noun( 'punctilio', 'punctilios', both, _ ).
noun( 'punctiliousness', '-', mass, _ ).
noun( 'punctuality', '-', mass, _ ).
noun( 'punctuation', '-', mass, _ ).
noun( 'puncture', 'punctures', count, _ ).
noun( 'pundit', 'pundits', count, _ ).
noun( 'pungency', '-', mass, _ ).
noun( 'punishment', 'punishments', both, _ ).
noun( 'punk', 'punks', both, _ ).
noun( 'punkah', 'punkahs', count, _ ).
noun( 'punnet', 'punnets', count, _ ).
noun( 'punster', 'punsters', count, _ ).
noun( 'punt', 'punts', count, _ ).
noun( 'punter', 'punters', count, _ ).
noun( 'pup', 'pups', count, _ ).
noun( 'pupa', 'pupas', count, _ ).
noun( 'pupil', 'pupils', count, _ ).
noun( 'puppet', 'puppets', count, _ ).
noun( 'puppeteer', 'puppeteers', count, _ ).
noun( 'puppy', 'puppies', count, _ ).
noun( 'pur_ee', 'pur_ees', count, _ ).
noun( 'purchase', 'purchases', both, _ ).
noun( 'purchaser', 'purchasers', count, _ ).
noun( 'purdah', '-', mass, _ ).
noun( 'pureness', '-', mass, _ ).
noun( 'purgation', '-', mass, _ ).
noun( 'purgative', 'purgatives', count, _ ).
noun( 'purgatory', 'purgatories', count, _ ).
noun( 'purge', 'purges', count, _ ).
noun( 'purification', 'purifications', both, _ ).
noun( 'purifier', 'purifiers', count, _ ).
noun( 'purist', 'purists', count, _ ).
noun( 'puritan', 'puritans', count, _ ).
noun( 'puritanism', '-', mass, _ ).
noun( 'purity', '-', mass, _ ).
noun( 'purl', 'purls', count, _ ).
noun( 'purple', 'purples', both, _ ).
noun( 'purport', '-', mass, _ ).
noun( 'purpose', 'purposes', both, _ ).
noun( 'purr', 'purrs', count, _ ).
noun( 'purse', 'purses', count, _ ).
noun( 'purser', 'pursers', count, _ ).
noun( 'pursuance', 'pursuances', count, _ ).
noun( 'pursuer', 'pursuers', count, _ ).
noun( 'pursuit', 'pursuits', both, _ ).
noun( 'purulence', '-', mass, _ ).
noun( 'purveyance', 'purveyances', both, _ ).
noun( 'purveyor', 'purveyors', count, _ ).
noun( 'purview', 'purviews', count, _ ).
noun( 'pus', '-', mass, _ ).
noun( 'push', 'pushes', both, _ ).
noun( 'push-bike', 'push-bikes', count, _ ).
noun( 'pushcart', 'pushcarts', count, _ ).
noun( 'pushchair', 'pushchairs', count, _ ).
noun( 'pusher', 'pushers', count, _ ).
noun( 'pushover', 'pushovers', count, _ ).
noun( 'pusillanimity', '-', mass, _ ).
noun( 'puss', '-', count, _ ).
noun( 'pussy', 'pussies', count, _ ).
noun( 'pussycat', 'pussycats', count, _ ).
noun( 'pustule', 'pustules', count, _ ).
noun( 'put', 'puts', count, _ ).
noun( 'put-down', 'put-downs', count, _ ).
noun( 'put-on', 'put-ons', count, _ ).
noun( 'putrefaction', 'putrefactions', count, _ ).
noun( 'putrescence', 'putrescences', count, _ ).
noun( 'putridity', '-', mass, _ ).
noun( 'putsch', 'putsches', count, _ ).
noun( 'putt', 'putts', count, _ ).
noun( 'puttee', 'puttees', count, _ ).
noun( 'putting-green', 'putting-greens', count, _ ).
noun( 'putting-iron', 'putting-irons', count, _ ).
noun( 'putty', '-', mass, _ ).
noun( 'puzzle', 'puzzles', count, _ ).
noun( 'puzzlement', 'puzzlements', count, _ ).
noun( 'puzzler', 'puzzlers', count, _ ).
noun( 'pygmy', 'pygmies', count, _ ).
noun( 'pylon', 'pylons', count, _ ).
noun( 'pyorrhoea', '-', mass, _ ).
noun( 'pyramid', 'pyramids', count, _ ).
noun( 'pyre', 'pyres', count, _ ).
noun( 'pyrites', '-', mass, _ ).
noun( 'python', 'pythons', count, _ ).
noun( 'pyx', 'pyxes', count, _ ).
noun( 'q', '-', count, _ ).
noun( 'qt', 'qt', count, _ ).
noun( 'quack', 'quacks', count, _ ).
noun( 'quack-quack', 'quack-quacks', count, _ ).
noun( 'quackery', '-', mass, _ ).
noun( 'quad', 'quads', count, _ ).
noun( 'quadrangle', 'quadrangles', count, _ ).
noun( 'quadrant', 'quadrants', count, _ ).
noun( 'quadrilateral', 'quadrilaterals', count, _ ).
noun( 'quadrille', 'quadrilles', count, _ ).
noun( 'quadrillion', 'quadrillions', count, _ ).
noun( 'quadrophony', '-', mass, _ ).
noun( 'quadruped', 'quadrupeds', count, _ ).
noun( 'quadruple', 'quadruples', count, _ ).
noun( 'quadruplet', 'quadruplets', count, _ ).
noun( 'quadruplicate', 'quadruplicates', count, _ ).
noun( 'quagga', 'quaggas', count, _ ).
noun( 'quagmire', 'quagmires', count, _ ).
noun( 'quail', 'quails', count, _ ).
noun( 'quaintness', '-', mass, _ ).
noun( 'quake', 'quakes', count, _ ).
noun( 'qualification', 'qualifications', both, _ ).
noun( 'qualifier', 'qualifiers', count, _ ).
noun( 'quality', 'qualities', both, _ ).
noun( 'qualm', 'qualms', count, _ ).
noun( 'quandary', 'quandaries', count, _ ).
noun( 'quango', 'quangos', count, _ ).
noun( 'quantity', 'quantities', both, _ ).
noun( 'quantum', 'quanta', count, _ ).
noun( 'quarantine', '-', mass, _ ).
noun( 'quark', 'quarks', count, _ ).
noun( 'quarrel', 'quarrels', count, _ ).
noun( 'quarry', 'quarries', count, _ ).
noun( 'quarryman', 'quarrymen', count, _ ).
noun( 'quart', 'quarts', count, _ ).
noun( 'quarter', 'quarters', both, _ ).
noun( 'quarter-day', 'quarter-days', count, _ ).
noun( 'quarter-deck', 'quarter-decks', count, _ ).
noun( 'quarter-plate', 'quarter-plates', count, _ ).
noun( 'quarterfinal', 'quarterfinals', count, _ ).
noun( 'quartering', 'quarterings', count, _ ).
noun( 'quarterlight', 'quarterlights', count, _ ).
noun( 'quarterly', 'quarterlies', count, _ ).
noun( 'quartermaster', 'quartermasters', count, _ ).
noun( 'quartermaster-general', 'quartermaster-generals', count, _ ).
noun( 'quarterstaff', 'quarterstaffs', count, _ ).
noun( 'quartet', 'quartets', count, _ ).
noun( 'quarto', 'quartos', count, _ ).
noun( 'quartz', '-', mass, _ ).
noun( 'quasar', 'quasars', count, _ ).
noun( 'quassia', '-', mass, _ ).
noun( 'quatercentenary', 'quatercentenaries', count, _ ).
noun( 'quatrain', 'quatrains', count, _ ).
noun( 'quattrocento', 'quattrocentos', count, _ ).
noun( 'quaver', 'quavers', count, _ ).
noun( 'quay', 'quays', count, _ ).
noun( 'queasiness', '-', mass, _ ).
noun( 'queen', 'queens', count, _ ).
noun( 'queer', 'queers', count, _ ).
noun( 'queerness', '-', mass, _ ).
noun( 'quern', 'querns', count, _ ).
noun( 'querulousness', '-', mass, _ ).
noun( 'query', 'queries', count, _ ).
noun( 'quest', 'quests', count, _ ).
noun( 'question', 'questions', both, _ ).
noun( 'question-mark', 'question-marks', count, _ ).
noun( 'question-master', 'question-masters', count, _ ).
noun( 'questioner', 'questioners', count, _ ).
noun( 'questionnaire', 'questionnaires', count, _ ).
noun( 'quetzal', 'quetzals', count, _ ).
noun( 'queue', 'queues', count, _ ).
noun( 'qui vive', '-', mass, _ ).
noun( 'quibble', 'quibbles', count, _ ).
noun( 'quibbler', 'quibblers', count, _ ).
noun( 'quiche', 'quiches', count, _ ).
noun( 'quick', '-', mass, _ ).
noun( 'quickie', 'quickies', count, _ ).
noun( 'quicklime', '-', mass, _ ).
noun( 'quickness', '-', mass, _ ).
noun( 'quicksand', 'quicksands', both, _ ).
noun( 'quicksilver', '-', mass, _ ).
noun( 'quickstep', 'quicksteps', count, _ ).
noun( 'quid', 'quid', count, _ ).
noun( 'quid pro quo', '-', count, _ ).
noun( 'quiescence', '-', count, _ ).
noun( 'quiet', '-', mass, _ ).
noun( 'quietism', '-', mass, _ ).
noun( 'quietist', 'quietists', count, _ ).
noun( 'quietness', '-', mass, _ ).
noun( 'quietude', '-', count, _ ).
noun( 'quietus', 'quietuses', count, _ ).
noun( 'quiff', 'quiffs', count, _ ).
noun( 'quill', 'quills', count, _ ).
noun( 'quill-feather', 'quill-feathers', count, _ ).
noun( 'quilt', 'quilts', count, _ ).
noun( 'quin', 'quins', count, _ ).
noun( 'quince', 'quinces', count, _ ).
noun( 'quincentenary', 'quincentenaries', count, _ ).
noun( 'quinine', '-', mass, _ ).
noun( 'quinsy', '-', mass, _ ).
noun( 'quintal', 'quintals', count, _ ).
noun( 'quintessence', 'quintessences', count, _ ).
noun( 'quintet', 'quintets', count, _ ).
noun( 'quintuplet', 'quintuplets', count, _ ).
noun( 'quip', 'quips', count, _ ).
noun( 'quire', 'quires', count, _ ).
noun( 'quirk', 'quirks', count, _ ).
noun( 'quisling', 'quislings', count, _ ).
noun( 'quittance', 'quittances', count, _ ).
noun( 'quitter', 'quitters', count, _ ).
noun( 'quiver', 'quivers', count, _ ).
noun( 'quiz', 'quizes', count, _ ).
noun( 'quizmaster', 'quizmasters', count, _ ).
noun( 'quoin', 'quoins', count, _ ).
noun( 'quoit', 'quoits', count, _ ).
noun( 'quorum', 'quorums', count, _ ).
noun( 'quota', 'quotas', count, _ ).
noun( 'quotability', '-', mass, _ ).
noun( 'quotation', 'quotations', both, _ ).
noun( 'quote', 'quotes', count, _ ).
noun( 'quotient', 'quotients', count, _ ).
noun( 'qv', '-', proper, _ ).
noun( 'r', '-', count, _ ).
noun( 'r^ole', 'r^oles', count, _ ).
noun( 'r_echauff_e', 'r_echauff_es', count, _ ).
noun( 'r_egime', 'r_egimes', count, _ ).
noun( 'r_esum_e', 'r_esum_es', count, _ ).
noun( 'rabbi', 'rabbis', count, _ ).
noun( 'rabbit', 'rabbits', count, _ ).
noun( 'rabbit-burrow', 'rabbit-burrows', count, _ ).
noun( 'rabbit-hole', 'rabbit-holes', count, _ ).
noun( 'rabbit-hutch', 'rabbit-hutches', count, _ ).
noun( 'rabbit-punch', 'rabbit-punches', count, _ ).
noun( 'rabbit-warren', 'rabbit-warrens', count, _ ).
noun( 'rabble', 'rabbles', count, _ ).
noun( 'rabies', '-', mass, _ ).
noun( 'raccoon', 'raccoons', count, _ ).
noun( 'race', 'races', both, _ ).
noun( 'race-meeting', 'race-meetings', count, _ ).
noun( 'racecard', 'racecards', count, _ ).
noun( 'racecourse', 'racecourses', count, _ ).
noun( 'racehorse', 'racehorses', count, _ ).
noun( 'raceme', 'racemes', count, _ ).
noun( 'racer', 'racers', count, _ ).
noun( 'racialism', '-', mass, _ ).
noun( 'racialist', 'racialists', count, _ ).
noun( 'raciness', '-', mass, _ ).
noun( 'racing', '-', mass, _ ).
noun( 'racism', '-', mass, _ ).
noun( 'racist', 'racists', count, _ ).
noun( 'rack', 'racks', both, _ ).
noun( 'rack-railway', 'rack-railways', count, _ ).
noun( 'rack-rent', 'rack-rents', count, _ ).
noun( 'racket', 'rackets', both, _ ).
noun( 'racketeer', 'racketeers', count, _ ).
noun( 'racketeering', '-', mass, _ ).
noun( 'raconteur', 'raconteurs', count, _ ).
noun( 'racoon', 'racoons', count, _ ).
noun( 'racquet', 'racquets', count, _ ).
noun( 'radar', '-', mass, _ ).
noun( 'radial', 'radials', count, _ ).
noun( 'radiance', '-', mass, _ ).
noun( 'radiation', 'radiations', both, _ ).
noun( 'radiator', 'radiators', count, _ ).
noun( 'radical', 'radicals', count, _ ).
noun( 'radicalism', '-', mass, _ ).
noun( 'radicle', 'radicles', count, _ ).
noun( 'radio', 'radios', both, _ ).
noun( 'radio-gramophone', 'radio-gramophones', count, _ ).
noun( 'radio-location', 'radio-locations', count, _ ).
noun( 'radio-set', 'radio-sets', count, _ ).
noun( 'radio-telescope', 'radio-telescopes', count, _ ).
noun( 'radioactivity', '-', mass, _ ).
noun( 'radiogram', 'radiograms', count, _ ).
noun( 'radiograph', 'radiographs', count, _ ).
noun( 'radiographer', 'radiographers', count, _ ).
noun( 'radiography', '-', mass, _ ).
noun( 'radioisotope', 'radioisotopes', count, _ ).
noun( 'radiologist', 'radiologists', count, _ ).
noun( 'radiology', '-', mass, _ ).
noun( 'radiotherapist', 'radiotherapists', count, _ ).
noun( 'radiotherapy', '-', mass, _ ).
noun( 'radish', 'radishes', count, _ ).
noun( 'radium', '-', mass, _ ).
noun( 'radius', 'radii', count, _ ).
noun( 'raffia', '-', mass, _ ).
noun( 'raffle', 'raffles', count, _ ).
noun( 'raft', 'rafts', count, _ ).
noun( 'rafter', 'rafters', count, _ ).
noun( 'raftsman', 'raftsmen', count, _ ).
noun( 'rag', 'rags', count, _ ).
noun( 'rag-day', 'rag-days', count, _ ).
noun( 'ragamuffin', 'ragamuffins', count, _ ).
noun( 'ragbag', 'ragbags', count, _ ).
noun( 'rage', 'rages', both, _ ).
noun( 'raggedness', '-', mass, _ ).
noun( 'raglan', 'raglans', count, _ ).
noun( 'ragout', 'ragouts', both, _ ).
noun( 'ragtag', 'ragtags', count, _ ).
noun( 'ragtime', '-', mass, _ ).
noun( 'raid', 'raids', count, _ ).
noun( 'raider', 'raiders', count, _ ).
noun( 'rail', 'rails', count, _ ).
noun( 'railcar', 'railcars', count, _ ).
noun( 'railhead', 'railheads', count, _ ).
noun( 'railing', 'railings', both, _ ).
noun( 'raillery', 'railleries', both, _ ).
noun( 'railroad', 'railroads', count, _ ).
noun( 'railway', 'railways', count, _ ).
noun( 'railwayman', 'railwaymen', count, _ ).
noun( 'raiment', '-', mass, _ ).
noun( 'rain', 'rains', both, _ ).
noun( 'rain-gauge', 'rain-gauges', count, _ ).
noun( 'rainbow', 'rainbows', count, _ ).
noun( 'raincoat', 'raincoats', count, _ ).
noun( 'raindrop', 'raindrops', count, _ ).
noun( 'rainfall', 'rainfalls', both, _ ).
noun( 'rainwater', '-', mass, _ ).
noun( 'raise', 'raises', count, _ ).
noun( 'raiser', 'raisers', count, _ ).
noun( 'raisin', 'raisins', count, _ ).
noun( 'raison d\'^etre', 'raisons d\'^etre', count, _ ).
noun( 'raj', '-', count, _ ).
noun( 'rajah', 'rajahs', count, _ ).
noun( 'rake', 'rakes', count, _ ).
noun( 'rake-off', 'rake-offs', count, _ ).
noun( 'rakishness', '-', mass, _ ).
noun( 'rally', 'rallies', count, _ ).
noun( 'ram', 'rams', count, _ ).
noun( 'ramble', 'rambles', count, _ ).
noun( 'rambler', 'ramblers', count, _ ).
noun( 'ramification', 'ramifications', count, _ ).
noun( 'ramjet', 'ramjets', count, _ ).
noun( 'ramp', 'ramps', count, _ ).
noun( 'rampage', 'rampages', count, _ ).
noun( 'rampart', 'ramparts', count, _ ).
noun( 'ramrod', 'ramrods', count, _ ).
noun( 'ranch', 'ranches', count, _ ).
noun( 'rancher', 'ranchers', count, _ ).
noun( 'rancour', '-', mass, _ ).
noun( 'rand', 'rands', count, _ ).
noun( 'random', '-', count, _ ).
noun( 'randomness', '-', mass, _ ).
noun( 'ranee', 'ranees', count, _ ).
noun( 'range', 'ranges', count, _ ).
noun( 'rangefinder', 'rangefinders', count, _ ).
noun( 'ranger', 'rangers', count, _ ).
noun( 'rani', 'ranis', count, _ ).
noun( 'rank', 'ranks', both, _ ).
noun( 'ranker', 'rankers', count, _ ).
noun( 'ranking', 'rankings', count, _ ).
noun( 'rankness', '-', mass, _ ).
noun( 'ransom', 'ransoms', both, _ ).
noun( 'rant', 'rants', count, _ ).
noun( 'ranter', 'ranters', count, _ ).
noun( 'rap', 'raps', count, _ ).
noun( 'rapacity', '-', mass, _ ).
noun( 'rape', 'rapes', both, _ ).
noun( 'rapid', 'rapids', count, _ ).
noun( 'rapidity', '-', mass, _ ).
noun( 'rapier', 'rapiers', count, _ ).
noun( 'rapier-thrust', 'rapier-thrusts', count, _ ).
noun( 'rapine', '-', mass, _ ).
noun( 'rapist', 'rapists', count, _ ).
noun( 'rapport', 'rapports', both, _ ).
noun( 'rapprochement', 'rapprochements', count, _ ).
noun( 'rapscallion', 'rapscallions', count, _ ).
noun( 'rapture', 'raptures', both, _ ).
noun( 'rarebit', 'rarebits', both, _ ).
noun( 'rarefaction', '-', mass, _ ).
noun( 'rareness', '-', mass, _ ).
noun( 'rarity', 'rarities', both, _ ).
noun( 'rascal', 'rascals', count, _ ).
noun( 'rash', 'rashes', count, _ ).
noun( 'rasher', 'rashers', count, _ ).
noun( 'rashness', '-', mass, _ ).
noun( 'rasp', 'rasps', count, _ ).
noun( 'raspberry', 'raspberries', count, _ ).
noun( 'rat', 'rats', count, _ ).
noun( 'rat-a-tat-tat', 'rat-a-tat-tats', count, _ ).
noun( 'rat-tat', 'rat-tats', count, _ ).
noun( 'ratability', '-', mass, _ ).
noun( 'ratan', 'ratans', both, _ ).
noun( 'ratch', 'ratches', count, _ ).
noun( 'ratchet', 'ratchets', count, _ ).
noun( 'rate', 'rates', count, _ ).
noun( 'rateability', '-', mass, _ ).
noun( 'ratepayer', 'ratepayers', count, _ ).
noun( 'ratification', 'ratifications', count, _ ).
noun( 'rating', 'ratings', count, _ ).
noun( 'ratio', 'ratios', count, _ ).
noun( 'ratiocination', '-', mass, _ ).
noun( 'ration', 'rations', count, _ ).
noun( 'rationale', 'rationales', count, _ ).
noun( 'rationalism', '-', mass, _ ).
noun( 'rationalist', 'rationalists', count, _ ).
noun( 'rationality', 'rationalities', both, _ ).
noun( 'rationalization', 'rationalizations', both, _ ).
noun( 'ratlin', 'ratlins', count, _ ).
noun( 'ratline', 'ratlines', count, _ ).
noun( 'rattan', 'rattans', both, _ ).
noun( 'ratter', 'ratters', count, _ ).
noun( 'rattle', 'rattles', both, _ ).
noun( 'rattler', 'rattlers', count, _ ).
noun( 'rattlesnake', 'rattlesnakes', count, _ ).
noun( 'ravage', 'ravages', both, _ ).
noun( 'rave', 'raves', count, _ ).
noun( 'rave-up', 'rave-ups', count, _ ).
noun( 'raven', 'ravens', count, _ ).
noun( 'raver', 'ravers', count, _ ).
noun( 'ravine', 'ravines', count, _ ).
noun( 'ravioli', 'raviolis', both, _ ).
noun( 'ravishment', 'ravishments', both, _ ).
noun( 'raw', '-', count, _ ).
noun( 'ray', 'rays', count, _ ).
noun( 'rayon', '-', mass, _ ).
noun( 'razor', 'razors', count, _ ).
noun( 'razor-edge', 'razor-edges', count, _ ).
noun( 'razorback', 'razorbacks', count, _ ).
noun( 'razorblade', 'razorblades', count, _ ).
noun( 'razzle', '-', count, _ ).
noun( 'razzle-dazzle', '-', count, _ ).
noun( 're', '-', count, _ ).
noun( 're-afforestation', 're-afforestations', both, _ ).
noun( 're-count', 're-counts', count, _ ).
noun( 're-echo', 're-echoes', count, _ ).
noun( 're-entry', 're-entries', count, _ ).
noun( 're-formation', 're-formations', both, _ ).
noun( 'reach', 'reaches', both, _ ).
noun( 'reaction', 'reactions', both, _ ).
noun( 'reactionary', 'reactionaries', count, _ ).
noun( 'reactor', 'reactors', count, _ ).
noun( 'read', '-', count, _ ).
noun( 'readability', '-', mass, _ ).
noun( 'reader', 'readers', count, _ ).
noun( 'readership', 'readerships', count, _ ).
noun( 'readiness', '-', mass, _ ).
noun( 'reading', 'readings', both, _ ).
noun( 'reading-lamp', 'reading-lamps', count, _ ).
noun( 'reading-room', 'reading-rooms', count, _ ).
noun( 'readjustment', 'readjustments', count, _ ).
noun( 'readmission', 'readmissions', both, _ ).
noun( 'ready', 'readies', count, _ ).
noun( 'reaffiliation', 'reaffiliations', both, _ ).
noun( 'reagent', 'reagents', count, _ ).
noun( 'real', 'reals', count, _ ).
noun( 'realism', '-', mass, _ ).
noun( 'realist', 'realists', count, _ ).
noun( 'reality', 'realities', both, _ ).
noun( 'realization', '-', mass, _ ).
noun( 'reallocation', 'reallocations', both, _ ).
noun( 'realm', 'realms', count, _ ).
noun( 'realty', 'realties', count, _ ).
noun( 'ream', 'reams', count, _ ).
noun( 'reaper', 'reapers', count, _ ).
noun( 'reaping-hook', 'reaping-hooks', count, _ ).
noun( 'reappearance', 'reappearances', count, _ ).
noun( 'reappraisal', 'reappraisals', count, _ ).
noun( 'rear', 'rears', count, _ ).
noun( 'rear-admiral', 'rear-admirals', count, _ ).
noun( 'rearguard', 'rearguards', count, _ ).
noun( 'rearmament', 'rearmaments', both, _ ).
noun( 'rearrangement', 'rearrangements', both, _ ).
noun( 'rearward', 'rearwards', count, _ ).
noun( 'reason', 'reasons', both, _ ).
noun( 'reasonableness', '-', mass, _ ).
noun( 'reasoning', '-', mass, _ ).
noun( 'reassessment', 'reassessments', both, _ ).
noun( 'reassurance', 'reassurances', both, _ ).
noun( 'rebate', 'rebates', both, _ ).
noun( 'rebel', 'rebels', count, _ ).
noun( 'rebellion', 'rebellions', both, _ ).
noun( 'rebelliousness', '-', mass, _ ).
noun( 'rebirth', 'rebirths', count, _ ).
noun( 'rebound', 'rebounds', count, _ ).
noun( 'rebuff', 'rebuffs', count, _ ).
noun( 'rebuke', 'rebukes', count, _ ).
noun( 'rebus', 'rebuses', count, _ ).
noun( 'rebuttal', 'rebuttals', count, _ ).
noun( 'rec', '-', proper, _ ).
noun( 'recalcitrance', '-', mass, _ ).
noun( 'recalcitrancy', '-', mass, _ ).
noun( 'recall', 'recalls', both, _ ).
noun( 'recantation', 'recantations', both, _ ).
noun( 'recap', 'recaps', count, _ ).
noun( 'recapitulation', 'recapitulations', both, _ ).
noun( 'recce', 'recces', count, _ ).
noun( 'recd', '-', proper, _ ).
noun( 'receipt', 'receipts', both, _ ).
noun( 'receiver', 'receivers', count, _ ).
noun( 'receivership', 'receiverships', count, _ ).
noun( 'receiving-set', 'receiving-sets', count, _ ).
noun( 'receptacle', 'receptacles', count, _ ).
noun( 'reception', 'receptions', both, _ ).
noun( 'reception-desk', 'reception-desks', count, _ ).
noun( 'receptionist', 'receptionists', count, _ ).
noun( 'receptiveness', '-', mass, _ ).
noun( 'receptivity', '-', mass, _ ).
noun( 'recess', 'recesses', count, _ ).
noun( 'recession', 'recessions', both, _ ).
noun( 'recessional', 'recessionals', count, _ ).
noun( 'recidivism', '-', mass, _ ).
noun( 'recidivist', 'recidivists', count, _ ).
noun( 'recipe', 'recipes', count, _ ).
noun( 'recipient', 'recipients', count, _ ).
noun( 'reciprocal', 'reciprocals', count, _ ).
noun( 'reciprocation', '-', mass, _ ).
noun( 'reciprocity', '-', mass, _ ).
noun( 'recital', 'recitals', count, _ ).
noun( 'recitalist', 'recitalists', count, _ ).
noun( 'recitation', 'recitations', both, _ ).
noun( 'recitative', 'recitatives', both, _ ).
noun( 'recklessness', '-', mass, _ ).
noun( 'reckoner', 'reckoners', count, _ ).
noun( 'reckoning', 'reckonings', both, _ ).
noun( 'reclamation', '-', mass, _ ).
noun( 'recluse', 'recluses', count, _ ).
noun( 'recognition', '-', mass, _ ).
noun( 'recognizance', 'recognizances', count, _ ).
noun( 'recoil', 'recoils', count, _ ).
noun( 'recollection', 'recollections', both, _ ).
noun( 'recommendation', 'recommendations', both, _ ).
noun( 'recompense', 'recompenses', both, _ ).
noun( 'reconciliation', 'reconciliations', both, _ ).
noun( 'reconnaissance', 'reconnaissances', both, _ ).
noun( 'reconstruction', 'reconstructions', both, _ ).
noun( 'record', 'records', both, _ ).
noun( 'record-player', 'record-players', count, _ ).
noun( 'recorder', 'recorders', count, _ ).
noun( 'recording', 'recordings', count, _ ).
noun( 'recourse', '-', mass, _ ).
noun( 'recovery', 'recoveries', both, _ ).
noun( 'recreant', 'recreants', count, _ ).
noun( 'recreation', 'recreations', both, _ ).
noun( 'recrimination', 'recriminations', both, _ ).
noun( 'recrudescence', 'recrudescences', both, _ ).
noun( 'recruit', 'recruits', count, _ ).
noun( 'recruitment', 'recruitments', count, _ ).
noun( 'rectangle', 'rectangles', count, _ ).
noun( 'rectification', 'rectifications', both, _ ).
noun( 'rectifier', 'rectifiers', count, _ ).
noun( 'rectitude', '-', mass, _ ).
noun( 'recto', 'rectos', count, _ ).
noun( 'rector', 'rectors', count, _ ).
noun( 'rectory', 'rectories', count, _ ).
noun( 'rectum', 'rectums', count, _ ).
noun( 'recuperation', 'recuperations', both, _ ).
noun( 'recurrence', 'recurrences', both, _ ).
noun( 'recusancy', '-', mass, _ ).
noun( 'recusant', 'recusants', count, _ ).
noun( 'red', 'reds', both, _ ).
noun( 'redaction', 'redactions', both, _ ).
noun( 'redbreast', 'redbreasts', count, _ ).
noun( 'redcap', 'redcaps', count, _ ).
noun( 'redcoat', 'redcoats', count, _ ).
noun( 'redeemer', 'redeemers', count, _ ).
noun( 'redemption', '-', mass, _ ).
noun( 'redeployment', 'redeployments', both, _ ).
noun( 'redevelopment', 'redevelopments', both, _ ).
noun( 'redhead', 'redheads', count, _ ).
noun( 'rediffusion', '-', mass, _ ).
noun( 'rediscovery', 'rediscoveries', both, _ ).
noun( 'redisposition', 'redispositions', count, _ ).
noun( 'redistribution', 'redistributions', both, _ ).
noun( 'redness', '-', mass, _ ).
noun( 'redolence', '-', mass, _ ).
noun( 'redoubt', 'redoubts', count, _ ).
noun( 'redress', '-', mass, _ ).
noun( 'redskin', 'redskins', count, _ ).
noun( 'reductio ad absurdum', '-', both, _ ).
noun( 'reduction', 'reductions', both, _ ).
noun( 'redundance', 'redundances', count, _ ).
noun( 'redundancy', 'redundancies', both, _ ).
noun( 'reduplication', 'reduplications', both, _ ).
noun( 'redwing', 'redwings', count, _ ).
noun( 'redwood', 'redwoods', count, _ ).
noun( 'reed', 'reeds', both, _ ).
noun( 'reef', 'reefs', count, _ ).
noun( 'reef-knot', 'reef-knots', count, _ ).
noun( 'reefer', 'reefers', count, _ ).
noun( 'reek', '-', mass, _ ).
noun( 'reel', 'reels', count, _ ).
noun( 'reeve', 'reeves', count, _ ).
noun( 'ref', '-', count, _ ).
noun( 'refashion', 'refashions', both, _ ).
noun( 'refection', 'refections', both, _ ).
noun( 'refectory', 'refectories', count, _ ).
noun( 'referee', 'referees', count, _ ).
noun( 'reference', 'references', both, _ ).
noun( 'referendum', 'referendums', count, _ ).
noun( 'refill', 'refills', count, _ ).
noun( 'refinement', 'refinements', both, _ ).
noun( 'refiner', 'refiners', count, _ ).
noun( 'refinery', 'refineries', count, _ ).
noun( 'refit', 'refits', count, _ ).
noun( 'reflation', '-', mass, _ ).
noun( 'reflection', 'reflections', both, _ ).
noun( 'reflector', 'reflectors', count, _ ).
noun( 'reflex', 'reflexes', count, _ ).
noun( 'reflexion', 'reflexions', both, _ ).
noun( 'reflexive', 'reflexives', count, _ ).
noun( 'reflux', 'refluxes', count, _ ).
noun( 'reforestation', 'reforestations', both, _ ).
noun( 'reform', 'reforms', both, _ ).
noun( 'reformation', 'reformations', both, _ ).
noun( 'reformatory', 'reformatories', count, _ ).
noun( 'reformer', 'reformers', count, _ ).
noun( 'refraction', 'refractions', both, _ ).
noun( 'refrain', 'refrains', count, _ ).
noun( 'refresher', 'refreshers', count, _ ).
noun( 'refreshment', 'refreshments', both, _ ).
noun( 'refrigerant', 'refrigerants', count, _ ).
noun( 'refrigeration', 'refrigerations', both, _ ).
noun( 'refrigerator', 'refrigerators', count, _ ).
noun( 'refuge', 'refuges', both, _ ).
noun( 'refugee', 'refugees', count, _ ).
noun( 'refulgence', '-', mass, _ ).
noun( 'refund', 'refunds', both, _ ).
noun( 'refusal', 'refusals', both, _ ).
noun( 'refuse', '-', mass, _ ).
noun( 'refuse-collector', 'refuse-collectors', count, _ ).
noun( 'refutation', 'refutations', both, _ ).
noun( 'regard', 'regards', both, _ ).
noun( 'regatta', 'regattas', count, _ ).
noun( 'regency', 'regencies', count, _ ).
noun( 'regeneration', '-', mass, _ ).
noun( 'regent', 'regents', count, _ ).
noun( 'reggae', '-', mass, _ ).
noun( 'regicide', 'regicides', both, _ ).
noun( 'regime', 'regimes', count, _ ).
noun( 'regimen', 'regimens', count, _ ).
noun( 'regiment', 'regiments', count, _ ).
noun( 'regimentation', '-', mass, _ ).
noun( 'region', 'regions', count, _ ).
noun( 'register', 'registers', count, _ ).
noun( 'registrar', 'registrars', count, _ ).
noun( 'registration', 'registrations', both, _ ).
noun( 'registry', 'registries', both, _ ).
noun( 'regression', 'regressions', both, _ ).
noun( 'regret', 'regrets', both, _ ).
noun( 'regular', 'regulars', count, _ ).
noun( 'regularity', 'regularities', both, _ ).
noun( 'regularization', 'regularizations', both, _ ).
noun( 'regulation', 'regulations', both, _ ).
noun( 'regulator', 'regulators', count, _ ).
noun( 'rehabilitation', 'rehabilitations', both, _ ).
noun( 'rehash', 'rehashes', count, _ ).
noun( 'rehearing', 'rehearings', count, _ ).
noun( 'rehearsal', 'rehearsals', both, _ ).
noun( 'reign', 'reigns', count, _ ).
noun( 'reimbursement', 'reimbursements', both, _ ).
noun( 'reimposition', 'reimpositions', both, _ ).
noun( 'rein', 'reins', count, _ ).
noun( 'reincarnation', 'reincarnations', both, _ ).
noun( 'reindeer', 'reindeer', count, _ ).
noun( 'reinforcement', 'reinforcements', both, _ ).
noun( 'reinstatement', 'reinstatements', count, _ ).
noun( 'reinsurance', 'reinsurances', both, _ ).
noun( 'reinterpretation', 'reinterpretations', both, _ ).
noun( 'reissue', 'reissues', count, _ ).
noun( 'reiteration', 'reiterations', both, _ ).
noun( 'reject', 'rejects', count, _ ).
noun( 'rejection', 'rejections', both, _ ).
noun( 'rejoicing', '-', mass, _ ).
noun( 'rejoinder', 'rejoinders', count, _ ).
noun( 'rejuvenation', 'rejuvenations', count, _ ).
noun( 'relapse', 'relapses', count, _ ).
noun( 'relation', 'relations', both, _ ).
noun( 'relationship', 'relationships', both, _ ).
noun( 'relative', 'relatives', count, _ ).
noun( 'relativity', '-', mass, _ ).
noun( 'relaxation', 'relaxations', both, _ ).
noun( 'relay', 'relays', count, _ ).
noun( 'release', 'releases', both, _ ).
noun( 'relegation', '-', mass, _ ).
noun( 'relevance', 'relevances', count, _ ).
noun( 'relevancy', '-', mass, _ ).
noun( 'reliability', '-', mass, _ ).
noun( 'reliance', '-', mass, _ ).
noun( 'relic', 'relics', count, _ ).
noun( 'relict', 'relicts', count, _ ).
noun( 'relief', 'reliefs', both, _ ).
noun( 'religion', 'religions', both, _ ).
noun( 'religious', 'religious', count, _ ).
noun( 'reliquary', 'reliquaries', count, _ ).
noun( 'relish', 'relishes', both, _ ).
noun( 'relocation', '-', mass, _ ).
noun( 'reluctance', '-', mass, _ ).
noun( 'remainder', 'remainders', count, _ ).
noun( 'remake', 'remakes', count, _ ).
noun( 'remand', '-', mass, _ ).
noun( 'remark', 'remarks', both, _ ).
noun( 'remarriage', 'remarriages', count, _ ).
noun( 'remedy', 'remedies', both, _ ).
noun( 'remembrance', 'remembrances', both, _ ).
noun( 'remilitarization', 'remilitarizations', both, _ ).
noun( 'reminder', 'reminders', count, _ ).
noun( 'reminiscence', 'reminiscences', both, _ ).
noun( 'remission', 'remissions', both, _ ).
noun( 'remissness', '-', mass, _ ).
noun( 'remittance', 'remittances', both, _ ).
noun( 'remnant', 'remnants', count, _ ).
noun( 'remonstrance', 'remonstrances', both, _ ).
noun( 'remorse', '-', mass, _ ).
noun( 'remoteness', '-', mass, _ ).
noun( 'remount', 'remounts', count, _ ).
noun( 'removal', 'removals', both, _ ).
noun( 'remove', 'removes', count, _ ).
noun( 'remover', 'removers', count, _ ).
noun( 'remuneration', '-', mass, _ ).
noun( 'renaissance', 'renaissances', count, _ ).
noun( 'renascence', 'renascences', count, _ ).
noun( 'rendering', 'renderings', both, _ ).
noun( 'rendezvous', 'rendezvous', count, _ ).
noun( 'rendition', 'renditions', count, _ ).
noun( 'renegade', 'renegades', count, _ ).
noun( 'renewal', 'renewals', both, _ ).
noun( 'rennet', '-', mass, _ ).
noun( 'renovation', 'renovations', both, _ ).
noun( 'renovator', 'renovators', count, _ ).
noun( 'renown', '-', mass, _ ).
noun( 'rent', 'rents', both, _ ).
noun( 'rent-collector', 'rent-collectors', count, _ ).
noun( 'rent-rebate', 'rent-rebates', both, _ ).
noun( 'rent-roll', 'rent-rolls', count, _ ).
noun( 'rental', 'rentals', count, _ ).
noun( 'rentier', 'rentiers', count, _ ).
noun( 'renunciation', '-', mass, _ ).
noun( 'reorganization', 'reorganizations', both, _ ).
noun( 'reorientation', '-', mass, _ ).
noun( 'rep', 'reps', both, _ ).
noun( 'repair', 'repairs', both, _ ).
noun( 'repairer', 'repairers', count, _ ).
noun( 'reparation', 'reparations', both, _ ).
noun( 'repartee', 'repartees', both, _ ).
noun( 'repast', 'repasts', count, _ ).
noun( 'repatriate', 'repatriates', count, _ ).
noun( 'repatriation', 'repatriations', both, _ ).
noun( 'repayment', 'repayments', both, _ ).
noun( 'repeal', 'repeals', count, _ ).
noun( 'repeat', 'repeats', count, _ ).
noun( 'repeater', 'repeaters', count, _ ).
noun( 'repellent', '-', mass, _ ).
noun( 'repentance', '-', mass, _ ).
noun( 'repercussion', 'repercussions', both, _ ).
noun( 'repertoire', 'repertoires', count, _ ).
noun( 'repertory', 'repertories', count, _ ).
noun( 'repetition', 'repetitions', both, _ ).
noun( 'replacement', 'replacements', both, _ ).
noun( 'replay', 'replays', count, _ ).
noun( 'replenishment', 'replenishments', both, _ ).
noun( 'repletion', '-', mass, _ ).
noun( 'replica', 'replicas', count, _ ).
noun( 'reply', 'replies', count, _ ).
noun( 'report', 'reports', both, _ ).
noun( 'reportage', '-', mass, _ ).
noun( 'reporter', 'reporters', count, _ ).
noun( 'repose', '-', mass, _ ).
noun( 'repository', 'repositories', count, _ ).
noun( 'repp', '-', mass, _ ).
noun( 'representation', 'representations', both, _ ).
noun( 'representative', 'representatives', count, _ ).
noun( 'repression', 'repressions', both, _ ).
noun( 'reprieve', 'reprieves', count, _ ).
noun( 'reprimand', 'reprimands', count, _ ).
noun( 'reprint', 'reprints', count, _ ).
noun( 'reprisal', 'reprisals', both, _ ).
noun( 'reproach', 'reproaches', both, _ ).
noun( 'reprobate', 'reprobates', count, _ ).
noun( 'reprobation', '-', mass, _ ).
noun( 'reproducer', 'reproducers', count, _ ).
noun( 'reproduction', 'reproductions', both, _ ).
noun( 'reproof', 'reproofs', both, _ ).
noun( 'reps', '-', mass, _ ).
noun( 'reptile', 'reptiles', count, _ ).
noun( 'reptilian', 'reptilians', count, _ ).
noun( 'republic', 'republics', count, _ ).
noun( 'republican', 'republicans', count, _ ).
noun( 'republicanism', '-', mass, _ ).
noun( 'repudiation', 'repudiations', count, _ ).
noun( 'repugnance', '-', mass, _ ).
noun( 'repulse', 'repulses', count, _ ).
noun( 'repulsion', '-', mass, _ ).
noun( 'reputation', 'reputations', both, _ ).
noun( 'repute', '-', mass, _ ).
noun( 'request', 'requests', both, _ ).
noun( 'requiem', 'requiems', count, _ ).
noun( 'requirement', 'requirements', count, _ ).
noun( 'requisite', 'requisites', count, _ ).
noun( 'requisition', 'requisitions', both, _ ).
noun( 'requital', '-', mass, _ ).
noun( 'reredos', 'reredoses', count, _ ).
noun( 'rerun', 'reruns', count, _ ).
noun( 'res', '-', count, _ ).
noun( 'rescript', 'rescripts', count, _ ).
noun( 'rescue', 'rescues', both, _ ).
noun( 'rescuer', 'rescuers', count, _ ).
noun( 'research', 'researches', both, _ ).
noun( 'researcher', 'researchers', count, _ ).
noun( 'resemblance', 'resemblances', both, _ ).
noun( 'resentment', '-', mass, _ ).
noun( 'reservation', 'reservations', both, _ ).
noun( 'reserve', 'reserves', both, _ ).
noun( 'reservist', 'reservists', count, _ ).
noun( 'reservoir', 'reservoirs', count, _ ).
noun( 'resettlement', 'resettlements', count, _ ).
noun( 'reshuffle', 'reshuffles', count, _ ).
noun( 'residence', 'residences', both, _ ).
noun( 'residency', 'residencies', count, _ ).
noun( 'resident', 'residents', count, _ ).
noun( 'residue', 'residues', count, _ ).
noun( 'resignation', 'resignations', both, _ ).
noun( 'resilience', '-', mass, _ ).
noun( 'resiliency', '-', mass, _ ).
noun( 'resin', 'resins', both, _ ).
noun( 'resistance', 'resistances', both, _ ).
noun( 'resister', 'resisters', count, _ ).
noun( 'resistivity', '-', mass, _ ).
noun( 'resistor', 'resistors', count, _ ).
noun( 'resoluteness', '-', mass, _ ).
noun( 'resolution', 'resolutions', both, _ ).
noun( 'resolve', 'resolves', count, _ ).
noun( 'resonance', '-', mass, _ ).
noun( 'resonator', 'resonators', count, _ ).
noun( 'resort', 'resorts', both, _ ).
noun( 'resource', 'resources', both, _ ).
noun( 'respect', 'respects', both, _ ).
noun( 'respectability', 'respectabilities', both, _ ).
noun( 'respecter', 'respecters', count, _ ).
noun( 'respiration', 'respirations', both, _ ).
noun( 'respirator', 'respirators', count, _ ).
noun( 'respite', 'respites', both, _ ).
noun( 'resplendence', '-', mass, _ ).
noun( 'resplendency', '-', mass, _ ).
noun( 'respondent', 'respondents', count, _ ).
noun( 'response', 'responses', both, _ ).
noun( 'responsibility', 'responsibilities', both, _ ).
noun( 'responsiveness', '-', mass, _ ).
noun( 'rest', 'rests', both, _ ).
noun( 'rest-cure', 'rest-cures', count, _ ).
noun( 'rest-day', 'rest-days', count, _ ).
noun( 'rest-home', 'rest-homes', count, _ ).
noun( 'rest-house', 'rest-houses', count, _ ).
noun( 'restatement', 'restatements', count, _ ).
noun( 'restaurant', 'restaurants', count, _ ).
noun( 'restauranteur', 'restauranteurs', count, _ ).
noun( 'restaurateur', 'restaurateurs', count, _ ).
noun( 'restfulness', '-', mass, _ ).
noun( 'restitution', '-', mass, _ ).
noun( 'restiveness', '-', mass, _ ).
noun( 'restlessness', '-', mass, _ ).
noun( 'restoration', 'restorations', both, _ ).
noun( 'restorative', 'restoratives', both, _ ).
noun( 'restorer', 'restorers', count, _ ).
noun( 'restraint', 'restraints', both, _ ).
noun( 'restriction', 'restrictions', both, _ ).
noun( 'restrictiveness', '-', mass, _ ).
noun( 'result', 'results', both, _ ).
noun( 'resumption', 'resumptions', both, _ ).
noun( 'resurgence', 'resurgences', count, _ ).
noun( 'resurrection', '-', mass, _ ).
noun( 'resuscitation', 'resuscitations', both, _ ).
noun( 'ret', '-', proper, _ ).
noun( 'retail', 'retails', count, _ ).
noun( 'retailer', 'retailers', count, _ ).
noun( 'retainer', 'retainers', count, _ ).
noun( 'retake', 'retakes', count, _ ).
noun( 'retaliation', '-', mass, _ ).
noun( 'retardation', 'retardations', both, _ ).
noun( 'retd', '-', proper, _ ).
noun( 'retention', '-', mass, _ ).
noun( 'retentiveness', '-', mass, _ ).
noun( 'rethink', 'rethinks', count, _ ).
noun( 'reticence', 'reticences', both, _ ).
noun( 'reticulation', 'reticulations', count, _ ).
noun( 'reticule', 'reticules', count, _ ).
noun( 'retina', 'retinas', count, _ ).
noun( 'retinue', 'retinues', count, _ ).
noun( 'retire', '-', count, _ ).
noun( 'retirement', 'retirements', both, _ ).
noun( 'retort', 'retorts', both, _ ).
noun( 'retraction', 'retractions', both, _ ).
noun( 'retread', 'retreads', count, _ ).
noun( 'retreat', 'retreats', both, _ ).
noun( 'retrenchment', 'retrenchments', both, _ ).
noun( 'retrial', 'retrials', count, _ ).
noun( 'retribution', '-', mass, _ ).
noun( 'retrieval', '-', mass, _ ).
noun( 'retriever', 'retrievers', count, _ ).
noun( 'retrogression', 'retrogressions', both, _ ).
noun( 'retrorocket', 'retrorockets', count, _ ).
noun( 'retrospect', '-', mass, _ ).
noun( 'retrospection', 'retrospections', both, _ ).
noun( 'retroversion', 'retroversions', both, _ ).
noun( 'retsina', '-', mass, _ ).
noun( 'return', 'returns', both, _ ).
noun( 'reunification', '-', mass, _ ).
noun( 'reunion', 'reunions', both, _ ).
noun( 'rev', 'revs', count, _ ).
noun( 'revaluation', 'revaluations', both, _ ).
noun( 'reveille', 'reveilles', count, _ ).
noun( 'revel', 'revels', both, _ ).
noun( 'revelation', 'revelations', both, _ ).
noun( 'reveller', 'revellers', count, _ ).
noun( 'revelry', 'revelries', both, _ ).
noun( 'revenge', '-', mass, _ ).
noun( 'revenue', 'revenues', both, _ ).
noun( 'reverberation', 'reverberations', both, _ ).
noun( 'reverence', '-', mass, _ ).
noun( 'reverend', 'reverends', count, _ ).
noun( 'reverie', 'reveries', both, _ ).
noun( 'revers', '-', count, _ ).
noun( 'reversal', 'reversals', both, _ ).
noun( 'reverse', 'reverses', both, _ ).
noun( 'reversibility', '-', mass, _ ).
noun( 'reversion', 'reversions', both, _ ).
noun( 'revetment', 'revetments', count, _ ).
noun( 'review', 'reviews', both, _ ).
noun( 'reviewer', 'reviewers', count, _ ).
noun( 'revise', 'revises', count, _ ).
noun( 'reviser', 'revisers', count, _ ).
noun( 'revision', 'revisions', both, _ ).
noun( 'revisionism', '-', mass, _ ).
noun( 'revisionist', 'revisionists', count, _ ).
noun( 'revitalization', 'revitalizations', both, _ ).
noun( 'revival', 'revivals', both, _ ).
noun( 'revivalist', 'revivalists', count, _ ).
noun( 'revocation', 'revocations', both, _ ).
noun( 'revoke', 'revokes', count, _ ).
noun( 'revolt', 'revolts', both, _ ).
noun( 'revolution', 'revolutions', both, _ ).
noun( 'revolutionary', 'revolutionarys', count, _ ).
noun( 'revolver', 'revolvers', count, _ ).
noun( 'revue', 'revues', both, _ ).
noun( 'revulsion', '-', mass, _ ).
noun( 'reward', 'rewards', both, _ ).
noun( 'rewrite', 'rewrites', count, _ ).
noun( 'rhapsody', 'rhapsodies', count, _ ).
noun( 'rhea', 'rheas', count, _ ).
noun( 'rheostat', 'rheostats', count, _ ).
noun( 'rhesus', 'rhesuses', count, _ ).
noun( 'rhetoric', '-', mass, _ ).
noun( 'rhetorician', 'rhetoricians', count, _ ).
noun( 'rheum', '-', mass, _ ).
noun( 'rheumatic', 'rheumatics', count, _ ).
noun( 'rheumatism', '-', mass, _ ).
noun( 'rhino', 'rhinos', count, _ ).
noun( 'rhinoceros', 'rhinoceros', count, _ ).
noun( 'rhizome', 'rhizomes', count, _ ).
noun( 'rhododendron', 'rhododendrons', count, _ ).
noun( 'rhomb', 'rhombs', count, _ ).
noun( 'rhomboid', 'rhomboids', count, _ ).
noun( 'rhombus', 'rhombuses', count, _ ).
noun( 'rhubarb', '-', mass, _ ).
noun( 'rhyme', 'rhymes', both, _ ).
noun( 'rhymester', 'rhymesters', count, _ ).
noun( 'rhythm', 'rhythms', both, _ ).
noun( 'rib', 'ribs', count, _ ).
noun( 'ribald', 'ribalds', count, _ ).
noun( 'ribaldry', '-', mass, _ ).
noun( 'riband', 'ribands', count, _ ).
noun( 'ribbon', 'ribbons', both, _ ).
noun( 'riboflavin', '-', mass, _ ).
noun( 'rice', '-', mass, _ ).
noun( 'rice-paper', '-', mass, _ ).
noun( 'richness', '-', mass, _ ).
noun( 'rick', 'ricks', count, _ ).
noun( 'rickets', 'rickets', mass, _ ).
noun( 'rickshaw', 'rickshaws', count, _ ).
noun( 'ricochet', 'ricochets', both, _ ).
noun( 'riddance', '-', mass, _ ).
noun( 'riddle', 'riddles', count, _ ).
noun( 'ride', 'rides', count, _ ).
noun( 'rider', 'riders', count, _ ).
noun( 'ridge', 'ridges', count, _ ).
noun( 'ridge-tile', 'ridge-tiles', count, _ ).
noun( 'ridgepole', 'ridgepoles', count, _ ).
noun( 'ridicule', 'ridicules', count, _ ).
noun( 'riding', 'ridings', both, _ ).
noun( 'riding-habit', 'riding-habits', count, _ ).
noun( 'riding-lamp', 'riding-lamps', count, _ ).
noun( 'riding-light', 'riding-lights', count, _ ).
noun( 'riding-master', 'riding-masters', count, _ ).
noun( 'riding-school', 'riding-schools', count, _ ).
noun( 'riff', 'riffs', count, _ ).
noun( 'riff-raff', '-', both, _ ).
noun( 'rifle', 'rifles', count, _ ).
noun( 'rifle-range', 'rifle-ranges', count, _ ).
noun( 'rifle-shot', 'rifle-shots', count, _ ).
noun( 'rifleman', 'riflemen', count, _ ).
noun( 'rift', 'rifts', count, _ ).
noun( 'rift-valley', 'rift-valleys', count, _ ).
noun( 'rig', 'rigs', count, _ ).
noun( 'rigger', 'riggers', count, _ ).
noun( 'rigging', '-', mass, _ ).
noun( 'right', 'rights', both, _ ).
noun( 'right-hander', 'right-handers', count, _ ).
noun( 'right-turn', 'right-turns', count, _ ).
noun( 'right-wing', 'right-wings', count, _ ).
noun( 'right-winger', 'right-wingers', count, _ ).
noun( 'righteousness', '-', mass, _ ).
noun( 'rightfulness', '-', mass, _ ).
noun( 'rightist', 'rightists', count, _ ).
noun( 'rightness', '-', mass, _ ).
noun( 'rigidity', 'rigidities', both, _ ).
noun( 'rigmarole', 'rigmaroles', both, _ ).
noun( 'rigor mortis', '-', mass, _ ).
noun( 'rigour', 'rigours', both, _ ).
noun( 'rigout', 'rigouts', count, _ ).
noun( 'rill', 'rills', count, _ ).
noun( 'rim', 'rims', count, _ ).
noun( 'rime', '-', mass, _ ).
noun( 'rind', 'rinds', both, _ ).
noun( 'rinderpest', '-', mass, _ ).
noun( 'ring', 'rings', count, _ ).
noun( 'ring-armour', '-', mass, _ ).
noun( 'ring-finger', 'ring-fingers', count, _ ).
noun( 'ring-mail', '-', mass, _ ).
noun( 'ring-road', 'ring-roads', count, _ ).
noun( 'ringer', 'ringers', count, _ ).
noun( 'ringleader', 'ringleaders', count, _ ).
noun( 'ringlet', 'ringlets', count, _ ).
noun( 'ringmaster', 'ringmasters', count, _ ).
noun( 'ringside', 'ringsides', count, _ ).
noun( 'ringworm', '-', mass, _ ).
noun( 'rink', 'rinks', count, _ ).
noun( 'rinse', 'rinses', count, _ ).
noun( 'riot', 'riots', both, _ ).
noun( 'rioter', 'rioters', count, _ ).
noun( 'rip', 'rips', count, _ ).
noun( 'rip-off', 'rip-offs', count, _ ).
noun( 'ripcord', 'ripcords', count, _ ).
noun( 'ripeness', '-', mass, _ ).
noun( 'riposte', 'ripostes', count, _ ).
noun( 'ripple', 'ripples', count, _ ).
noun( 'ripsaw', 'ripsaws', count, _ ).
noun( 'riptide', 'riptides', count, _ ).
noun( 'rise', 'rises', count, _ ).
noun( 'riser', 'risers', count, _ ).
noun( 'risibility', '-', mass, _ ).
noun( 'rising', 'risings', count, _ ).
noun( 'risk', 'risks', both, _ ).
noun( 'riskiness', '-', mass, _ ).
noun( 'risotto', 'risottos', both, _ ).
noun( 'rissole', 'rissoles', count, _ ).
noun( 'rite', 'rites', count, _ ).
noun( 'ritual', 'rituals', both, _ ).
noun( 'ritualism', '-', mass, _ ).
noun( 'ritualist', 'ritualists', count, _ ).
noun( 'rival', 'rivals', count, _ ).
noun( 'rivalry', 'rivalries', both, _ ).
noun( 'river', 'rivers', count, _ ).
noun( 'river-basin', 'river-basins', count, _ ).
noun( 'river-bed', 'river-beds', count, _ ).
noun( 'riverside', 'riversides', count, _ ).
noun( 'rivet', 'rivets', count, _ ).
noun( 'riveter', 'riveters', count, _ ).
noun( 'rivulet', 'rivulets', count, _ ).
noun( 'rly', '-', count, _ ).
noun( 'roach', 'roach', count, _ ).
noun( 'road', 'roads', count, _ ).
noun( 'road-book', 'road-books', count, _ ).
noun( 'road-hog', 'road-hogs', count, _ ).
noun( 'road-metal', '-', mass, _ ).
noun( 'road-sense', '-', mass, _ ).
noun( 'roadbed', 'roadbeds', count, _ ).
noun( 'roadblock', 'roadblocks', count, _ ).
noun( 'roadhouse', 'roadhouses', count, _ ).
noun( 'roadman', 'roadmen', count, _ ).
noun( 'roadmender', 'roadmenders', count, _ ).
noun( 'roadside', 'roadsides', count, _ ).
noun( 'roadstead', 'roadsteads', count, _ ).
noun( 'roadster', 'roadsters', count, _ ).
noun( 'roadway', 'roadways', count, _ ).
noun( 'roan', 'roans', both, _ ).
noun( 'roar', 'roars', count, _ ).
noun( 'roast', 'roasts', both, _ ).
noun( 'roaster', 'roasters', count, _ ).
noun( 'roasting', 'roastings', both, _ ).
noun( 'robber', 'robbers', count, _ ).
noun( 'robbery', 'robberies', both, _ ).
noun( 'robe', 'robes', count, _ ).
noun( 'robin', 'robins', count, _ ).
noun( 'robot', 'robots', count, _ ).
noun( 'robustness', '-', mass, _ ).
noun( 'roc', 'rocs', count, _ ).
noun( 'rock', 'rocks', both, _ ).
noun( 'rock-\'n-roll', '-', mass, _ ).
noun( 'rock-bottom', '-', mass, _ ).
noun( 'rock-cake', 'rock-cakes', count, _ ).
noun( 'rock-climbing', '-', mass, _ ).
noun( 'rock-crystal', 'rock-crystals', count, _ ).
noun( 'rock-garden', 'rock-gardens', count, _ ).
noun( 'rock-plant', 'rock-plants', count, _ ).
noun( 'rock-salmon', 'rock-salmon', count, _ ).
noun( 'rock-salt', '-', mass, _ ).
noun( 'rocker', 'rockers', count, _ ).
noun( 'rockery', 'rockeries', count, _ ).
noun( 'rocket', 'rockets', count, _ ).
noun( 'rocket-base', 'rocket-bases', count, _ ).
noun( 'rocket-range', 'rocket-ranges', count, _ ).
noun( 'rocketry', '-', mass, _ ).
noun( 'rocking-chair', 'rocking-chairs', count, _ ).
noun( 'rocking-horse', 'rocking-horses', count, _ ).
noun( 'rod', 'rods', count, _ ).
noun( 'rodent', 'rodents', count, _ ).
noun( 'rodeo', 'rodeos', count, _ ).
noun( 'rodomontade', '-', mass, _ ).
noun( 'roe', 'roe', both, _ ).
noun( 'roebuck', 'roebucks', count, _ ).
noun( 'rogation', 'rogations', count, _ ).
noun( 'rogue', 'rogues', count, _ ).
noun( 'rogue-elephant', 'rogue-elephants', count, _ ).
noun( 'roguery', 'rogueries', both, _ ).
noun( 'roguishness', '-', mass, _ ).
noun( 'roisterer', 'roisterers', count, _ ).
noun( 'role', 'roles', count, _ ).
noun( 'roll', 'rolls', count, _ ).
noun( 'roll-call', 'roll-calls', count, _ ).
noun( 'roll-on', 'roll-ons', count, _ ).
noun( 'roller', 'rollers', count, _ ).
noun( 'roller-skate', 'roller-skates', count, _ ).
noun( 'rolling', '-', mass, _ ).
noun( 'rolling-mill', 'rolling-mills', count, _ ).
noun( 'rolling-pin', 'rolling-pins', count, _ ).
noun( 'rolling-stock', '-', mass, _ ).
noun( 'roly-poly', 'roly-polies', count, _ ).
noun( 'romance', 'romances', both, _ ).
noun( 'romantic', 'romantics', count, _ ).
noun( 'romanticism', '-', mass, _ ).
noun( 'romanticist', 'romanticists', count, _ ).
noun( 'romp', 'romps', count, _ ).
noun( 'romper', 'rompers', count, _ ).
noun( 'rondeau', 'rondeaus', count, _ ).
noun( 'rondel', 'rondels', count, _ ).
noun( 'rondo', 'rondos', count, _ ).
noun( 'rood', 'roods', count, _ ).
noun( 'rood-tree', 'rood-trees', count, _ ).
noun( 'roof', 'roofs', count, _ ).
noun( 'roof-garden', 'roof-gardens', count, _ ).
noun( 'roof-tree', 'roof-trees', count, _ ).
noun( 'roofing', '-', mass, _ ).
noun( 'rook', 'rooks', count, _ ).
noun( 'rookery', 'rookeries', count, _ ).
noun( 'rookie', 'rookies', count, _ ).
noun( 'room', 'rooms', both, _ ).
noun( 'room-mate', 'room-mates', count, _ ).
noun( 'roomer', 'roomers', count, _ ).
noun( 'roomful', 'roomfuls', count, _ ).
noun( 'roost', 'roosts', count, _ ).
noun( 'rooster', 'roosters', count, _ ).
noun( 'root', 'roots', count, _ ).
noun( 'rope', 'ropes', both, _ ).
noun( 'rope-dancer', 'rope-dancers', count, _ ).
noun( 'rope-ladder', 'rope-ladders', count, _ ).
noun( 'rope-yard', 'rope-yards', count, _ ).
noun( 'rope-yarn', '-', mass, _ ).
noun( 'ropewalk', 'ropewalks', count, _ ).
noun( 'ropewalker', 'ropewalkers', count, _ ).
noun( 'ropeway', 'ropeways', count, _ ).
noun( 'rosary', 'rosaries', count, _ ).
noun( 'rose', 'roses', both, _ ).
noun( 'rose-bed', 'rose-beds', count, _ ).
noun( 'rose-leaf', 'rose-leaves', count, _ ).
noun( 'rose-water', '-', mass, _ ).
noun( 'rosebud', 'rosebuds', count, _ ).
noun( 'rosemary', '-', mass, _ ).
noun( 'rosette', 'rosettes', count, _ ).
noun( 'rosewood', '-', mass, _ ).
noun( 'rosin', '-', mass, _ ).
noun( 'roster', 'rosters', count, _ ).
noun( 'rostrum', 'rostrums', count, _ ).
noun( 'rot', '-', mass, _ ).
noun( 'rota', 'rotas', count, _ ).
noun( 'rotary', 'rotaries', count, _ ).
noun( 'rotation', 'rotations', both, _ ).
noun( 'rote', '-', count, _ ).
noun( 'rotgut', '-', mass, _ ).
noun( 'rotisserie', 'rotisseries', count, _ ).
noun( 'rotogravure', 'rotogravures', both, _ ).
noun( 'rotor', 'rotors', count, _ ).
noun( 'rottenness', '-', mass, _ ).
noun( 'rotter', 'rotters', count, _ ).
noun( 'rotunda', 'rotundas', count, _ ).
noun( 'rotundity', '-', mass, _ ).
noun( 'rou_e', 'rou_es', count, _ ).
noun( 'rouble', 'roubles', count, _ ).
noun( 'rouge', '-', mass, _ ).
noun( 'rough', 'roughs', both, _ ).
noun( 'rough-and-tumble', 'rough-and-tumbles', count, _ ).
noun( 'roughage', '-', mass, _ ).
noun( 'roughcast', '-', mass, _ ).
noun( 'roughneck', 'roughnecks', count, _ ).
noun( 'roughness', '-', mass, _ ).
noun( 'roughrider', 'roughriders', count, _ ).
noun( 'roulette', '-', mass, _ ).
noun( 'round', 'rounds', both, _ ).
noun( 'round-hand', '-', mass, _ ).
noun( 'round-shot', 'round-shots', count, _ ).
noun( 'roundabout', 'roundabouts', count, _ ).
noun( 'roundel', 'roundels', count, _ ).
noun( 'roundelay', 'roundelays', count, _ ).
noun( 'roundhouse', 'roundhouses', count, _ ).
noun( 'roundness', '-', mass, _ ).
noun( 'roundsman', 'roundsmen', count, _ ).
noun( 'roundup', 'roundups', count, _ ).
noun( 'rout', 'routs', count, _ ).
noun( 'route', 'routes', both, _ ).
noun( 'routemarch', 'routemarches', count, _ ).
noun( 'routine', 'routines', both, _ ).
noun( 'rover', 'rovers', count, _ ).
noun( 'row', 'rows', count, _ ).
noun( 'row', 'rows', both, _ ).
noun( 'rowan', 'rowans', count, _ ).
noun( 'rowan-berry', 'rowan-berries', count, _ ).
noun( 'rowan-tree', 'rowan-trees', count, _ ).
noun( 'rowboat', 'rowboats', count, _ ).
noun( 'rowdiness', '-', mass, _ ).
noun( 'rowdy', 'rowdies', count, _ ).
noun( 'rowdyism', '-', mass, _ ).
noun( 'rowel', 'rowels', count, _ ).
noun( 'rower', 'rowers', count, _ ).
noun( 'rowing', '-', mass, _ ).
noun( 'rowing-boat', 'rowing-boats', count, _ ).
noun( 'rowing-club', 'rowing-clubs', count, _ ).
noun( 'rowlock', 'rowlocks', count, _ ).
noun( 'royalist', 'royalists', count, _ ).
noun( 'royalty', 'royalties', both, _ ).
noun( 'rpm', 'rpm', count, _ ).
noun( 'rub', 'rubs', count, _ ).
noun( 'rub-a-dub', '-', mass, _ ).
noun( 'rub-down', 'rub-downs', count, _ ).
noun( 'rub-up', 'rub-ups', count, _ ).
noun( 'rubber', 'rubbers', both, _ ).
noun( 'rubberneck', 'rubbernecks', count, _ ).
noun( 'rubbing', 'rubbings', both, _ ).
noun( 'rubbish', '-', mass, _ ).
noun( 'rubbishing', 'rubbishings', count, _ ).
noun( 'rubble', '-', mass, _ ).
noun( 'rubric', 'rubrics', count, _ ).
noun( 'ruby', 'rubies', count, _ ).
noun( 'ruck', 'rucks', count, _ ).
noun( 'rucksack', 'rucksacks', count, _ ).
noun( 'ruckus', 'ruckuss', count, _ ).
noun( 'rudder', 'rudders', count, _ ).
noun( 'ruddiness', '-', mass, _ ).
noun( 'ruddle', '-', mass, _ ).
noun( 'rudeness', 'rudenesses', both, _ ).
noun( 'rudiment', 'rudiments', count, _ ).
noun( 'rue', '-', mass, _ ).
noun( 'ruff', 'ruffs', count, _ ).
noun( 'ruffian', 'ruffians', count, _ ).
noun( 'ruffianism', '-', mass, _ ).
noun( 'ruffle', 'ruffles', count, _ ).
noun( 'rug', 'rugs', count, _ ).
noun( 'rugby', '-', mass, _ ).
noun( 'ruggedness', '-', mass, _ ).
noun( 'rugger', '-', mass, _ ).
noun( 'ruin', 'ruins', both, _ ).
noun( 'ruination', '-', mass, _ ).
noun( 'rule', 'rules', both, _ ).
noun( 'ruler', 'rulers', count, _ ).
noun( 'ruling', 'rulings', count, _ ).
noun( 'rum', '-', mass, _ ).
noun( 'rum-runner', 'rum-runners', count, _ ).
noun( 'rumba', 'rumbas', count, _ ).
noun( 'rumble', 'rumbles', both, _ ).
noun( 'rumbling', 'rumblings', count, _ ).
noun( 'ruminant', 'ruminants', count, _ ).
noun( 'rumination', '-', mass, _ ).
noun( 'rummage', '-', mass, _ ).
noun( 'rummy', '-', mass, _ ).
noun( 'rumour', 'rumours', both, _ ).
noun( 'rumour-monger', 'rumour-mongers', count, _ ).
noun( 'rump', 'rumps', count, _ ).
noun( 'rump-steak', 'rump-steaks', both, _ ).
noun( 'rumpus', '-', count, _ ).
noun( 'run', 'runs', count, _ ).
noun( 'run-off', 'run-offs', count, _ ).
noun( 'run-through', 'run-throughs', count, _ ).
noun( 'run-up', 'run-ups', count, _ ).
noun( 'runaway', 'runaways', count, _ ).
noun( 'rundown', 'rundowns', count, _ ).
noun( 'rune', 'runes', count, _ ).
noun( 'rung', 'rungs', count, _ ).
noun( 'runnel', 'runnels', count, _ ).
noun( 'runner', 'runners', count, _ ).
noun( 'runner-up', 'runners-up', count, _ ).
noun( 'running', '-', mass, _ ).
noun( 'running-board', 'running-boards', count, _ ).
noun( 'runt', 'runts', count, _ ).
noun( 'runway', 'runways', count, _ ).
noun( 'rupee', 'rupees', count, _ ).
noun( 'rupiah', 'rupiahs', count, _ ).
noun( 'rupture', 'ruptures', both, _ ).
noun( 'ruse', 'ruses', count, _ ).
noun( 'rush', 'rushes', both, _ ).
noun( 'rushlight', 'rushlights', count, _ ).
noun( 'rusk', 'rusks', count, _ ).
noun( 'russet', 'russets', both, _ ).
noun( 'rust', '-', mass, _ ).
noun( 'rustic', 'rustics', count, _ ).
noun( 'rusticity', '-', mass, _ ).
noun( 'rustiness', '-', mass, _ ).
noun( 'rustle', '-', mass, _ ).
noun( 'rustler', 'rustlers', count, _ ).
noun( 'rustling', 'rustlings', both, _ ).
noun( 'rut', 'ruts', both, _ ).
noun( 'ruthlessness', '-', mass, _ ).
noun( 'rye', '-', mass, _ ).
noun( 'rye-bread', '-', mass, _ ).
noun( 's', '-', count, _ ).
noun( 's_eance', 's_eances', count, _ ).
noun( 'sabbatarian', 'sabbatarians', count, _ ).
noun( 'sabbatical', 'sabbaticals', count, _ ).
noun( 'sable', 'sables', both, _ ).
noun( 'sabot', 'sabots', count, _ ).
noun( 'sabotage', '-', mass, _ ).
noun( 'saboteur', 'saboteurs', count, _ ).
noun( 'sabre', 'sabres', count, _ ).
noun( 'sabre-rattling', '-', mass, _ ).
noun( 'sac', 'sacs', count, _ ).
noun( 'saccharin', '-', mass, _ ).
noun( 'sacerdotalism', '-', mass, _ ).
noun( 'sachet', 'sachets', count, _ ).
noun( 'sack', 'sacks', both, _ ).
noun( 'sack-race', 'sack-races', count, _ ).
noun( 'sackbut', 'sackbuts', count, _ ).
noun( 'sackcloth', '-', mass, _ ).
noun( 'sacking', '-', mass, _ ).
noun( 'sacrament', 'sacraments', count, _ ).
noun( 'sacredness', '-', mass, _ ).
noun( 'sacrifice', 'sacrifices', both, _ ).
noun( 'sacrilege', '-', mass, _ ).
noun( 'sacristan', 'sacristans', count, _ ).
noun( 'sacristy', 'sacristies', count, _ ).
noun( 'saddle', 'saddles', count, _ ).
noun( 'saddlebag', 'saddlebags', count, _ ).
noun( 'saddler', 'saddlers', count, _ ).
noun( 'saddlery', 'saddleries', both, _ ).
noun( 'sadhu', 'sadhus', count, _ ).
noun( 'sadism', '-', mass, _ ).
noun( 'sadist', 'sadists', count, _ ).
noun( 'sadness', '-', mass, _ ).
noun( 'sado-masochist', 'sado-masochists', count, _ ).
noun( 'sadomasochism', '-', mass, _ ).
noun( 'sae', '-', count, _ ).
noun( 'safari', 'safaris', both, _ ).
noun( 'safe', 'safes', count, _ ).
noun( 'safe-conduct', '-', mass, _ ).
noun( 'safe-deposit', 'safe-deposits', count, _ ).
noun( 'safecracker', 'safecrackers', count, _ ).
noun( 'safeguard', 'safeguards', count, _ ).
noun( 'safekeeping', '-', mass, _ ).
noun( 'safeness', '-', mass, _ ).
noun( 'safety', '-', mass, _ ).
noun( 'safety-belt', 'safety-belts', count, _ ).
noun( 'safety-bolt', 'safety-bolts', count, _ ).
noun( 'safety-catch', 'safety-catches', count, _ ).
noun( 'safety-curtain', 'safety-curtains', count, _ ).
noun( 'safety-factor', 'safety-factors', count, _ ).
noun( 'safety-lamp', 'safety-lamps', count, _ ).
noun( 'safety-lock', 'safety-locks', count, _ ).
noun( 'safety-match', 'safety-matches', count, _ ).
noun( 'safety-pin', 'safety-pins', count, _ ).
noun( 'safety-razor', 'safety-razors', count, _ ).
noun( 'safety-valve', 'safety-valves', count, _ ).
noun( 'saffron', 'saffrons', both, _ ).
noun( 'sag', 'sags', count, _ ).
noun( 'saga', 'sagas', count, _ ).
noun( 'sagacity', '-', mass, _ ).
noun( 'sage', 'sages', both, _ ).
noun( 'sage-green', '-', mass, _ ).
noun( 'sago', '-', mass, _ ).
noun( 'sahib', 'sahibs', count, _ ).
noun( 'sail', 'sails', both, _ ).
noun( 'sailcloth', '-', mass, _ ).
noun( 'sailing', 'sailings', count, _ ).
noun( 'sailing-boat', 'sailing-boats', count, _ ).
noun( 'sailing-master', 'sailing-masters', count, _ ).
noun( 'sailing-ship', 'sailing-ships', count, _ ).
noun( 'sailing-vessel', 'sailing-vessels', count, _ ).
noun( 'sailor', 'sailors', count, _ ).
noun( 'saint', 'saints', count, _ ).
noun( 'saint\'s-day', 'saint\'s-days', count, _ ).
noun( 'sainthood', '-', mass, _ ).
noun( 'saintliness', '-', mass, _ ).
noun( 'sak_e', '-', mass, _ ).
noun( 'sake', 'sakes', count, _ ).
noun( 'sal volatile', '-', mass, _ ).
noun( 'salaam', 'salaams', count, _ ).
noun( 'salaciousness', '-', mass, _ ).
noun( 'salacity', '-', mass, _ ).
noun( 'salad', 'salads', both, _ ).
noun( 'salad-dressing', 'salad-dressings', both, _ ).
noun( 'salad-oil', 'salad-oils', both, _ ).
noun( 'salamander', 'salamanders', count, _ ).
noun( 'salami', '-', mass, _ ).
noun( 'salary', 'salaries', count, _ ).
noun( 'sale', 'sales', both, _ ).
noun( 'saleroom', 'salerooms', count, _ ).
noun( 'salesman', 'salesmen', count, _ ).
noun( 'salesmanship', '-', mass, _ ).
noun( 'saleswoman', 'saleswomen', count, _ ).
noun( 'salience', '-', mass, _ ).
noun( 'salient', 'salients', count, _ ).
noun( 'saline', 'salines', both, _ ).
noun( 'salinity', '-', mass, _ ).
noun( 'saliva', '-', mass, _ ).
noun( 'sally', 'sallies', count, _ ).
noun( 'salmon', 'salmon', both, _ ).
noun( 'salon', 'salons', count, _ ).
noun( 'saloon', 'saloons', count, _ ).
noun( 'salsify', '-', mass, _ ).
noun( 'salt', 'salts', both, _ ).
noun( 'salt-cellar', 'salt-cellars', count, _ ).
noun( 'salt-lick', 'salt-licks', count, _ ).
noun( 'saltiness', '-', mass, _ ).
noun( 'saltpan', 'saltpans', count, _ ).
noun( 'saltpetre', '-', mass, _ ).
noun( 'saltwater', '-', mass, _ ).
noun( 'saltworks', 'saltworks', count, _ ).
noun( 'salubrity', '-', mass, _ ).
noun( 'salutation', 'salutations', both, _ ).
noun( 'salute', 'salutes', count, _ ).
noun( 'salvage', '-', mass, _ ).
noun( 'salvation', '-', mass, _ ).
noun( 'salve', 'salves', both, _ ).
noun( 'salver', 'salvers', count, _ ).
noun( 'salvia', 'salvias', count, _ ).
noun( 'salvo', 'salvos', count, _ ).
noun( 'samba', 'sambas', count, _ ).
noun( 'sameness', '-', mass, _ ).
noun( 'samovar', 'samovars', count, _ ).
noun( 'sampan', 'sampans', count, _ ).
noun( 'sample', 'samples', count, _ ).
noun( 'sampler', 'samplers', count, _ ).
noun( 'samurai', 'samurai', count, _ ).
noun( 'sanatorium', 'sanatoriums', count, _ ).
noun( 'sanctification', 'sanctifications', both, _ ).
noun( 'sanction', 'sanctions', both, _ ).
noun( 'sanctity', 'sanctities', both, _ ).
noun( 'sanctuary', 'sanctuaries', both, _ ).
noun( 'sanctum', 'sanctums', count, _ ).
noun( 'sand', 'sands', both, _ ).
noun( 'sand-bar', 'sand-bars', count, _ ).
noun( 'sandal', 'sandals', count, _ ).
noun( 'sandalwood', '-', mass, _ ).
noun( 'sandbag', 'sandbags', count, _ ).
noun( 'sandbank', 'sandbanks', count, _ ).
noun( 'sandboy', 'sandboys', count, _ ).
noun( 'sandfly', 'sandflies', count, _ ).
noun( 'sandglass', 'sandglasses', count, _ ).
noun( 'sandiness', '-', mass, _ ).
noun( 'sandpaper', 'sandpapers', both, _ ).
noun( 'sandpiper', 'sandpipers', count, _ ).
noun( 'sandpit', 'sandpits', count, _ ).
noun( 'sandstone', '-', mass, _ ).
noun( 'sandstorm', 'sandstorms', count, _ ).
noun( 'sandwich', 'sandwiches', count, _ ).
noun( 'sandwich-board', 'sandwich-boards', count, _ ).
noun( 'sandwichman', 'sandwichmen', count, _ ).
noun( 'sang froid', '-', mass, _ ).
noun( 'sanitation', '-', mass, _ ).
noun( 'sanity', '-', mass, _ ).
noun( 'sap', 'saps', both, _ ).
noun( 'saphead', 'sapheads', count, _ ).
noun( 'sapience', '-', mass, _ ).
noun( 'sapling', 'saplings', count, _ ).
noun( 'sapper', 'sappers', count, _ ).
noun( 'sapphire', 'sapphires', both, _ ).
noun( 'sapwood', '-', mass, _ ).
noun( 'saraband', 'sarabands', count, _ ).
noun( 'sarcasm', 'sarcasms', both, _ ).
noun( 'sarcophagus', 'sarcophagi', count, _ ).
noun( 'sardine', 'sardines', both, _ ).
noun( 'sari', 'saris', count, _ ).
noun( 'sarong', 'sarongs', count, _ ).
noun( 'sarsaparilla', '-', mass, _ ).
noun( 'sash', 'sashes', count, _ ).
noun( 'sash-cord', 'sash-cords', count, _ ).
noun( 'sash-line', 'sash-lines', count, _ ).
noun( 'satchel', 'satchels', count, _ ).
noun( 'sateen', '-', mass, _ ).
noun( 'satellite', 'satellites', count, _ ).
noun( 'satiety', '-', mass, _ ).
noun( 'satin', '-', mass, _ ).
noun( 'satinwood', '-', mass, _ ).
noun( 'satire', 'satires', both, _ ).
noun( 'satirist', 'satirists', count, _ ).
noun( 'satisfaction', 'satisfactions', both, _ ).
noun( 'satrap', 'satraps', count, _ ).
noun( 'satsuma', 'satsumas', count, _ ).
noun( 'saturation', '-', mass, _ ).
noun( 'saturnalia', 'saturnalias', count, _ ).
noun( 'satyr', 'satyrs', count, _ ).
noun( 'sauce', 'sauces', both, _ ).
noun( 'sauce-boat', 'sauce-boats', count, _ ).
noun( 'saucepan', 'saucepans', count, _ ).
noun( 'saucer', 'saucers', count, _ ).
noun( 'sauciness', '-', mass, _ ).
noun( 'sauerkraut', '-', mass, _ ).
noun( 'sauna', 'saunas', count, _ ).
noun( 'saunter', 'saunters', count, _ ).
noun( 'saunterer', 'saunterers', count, _ ).
noun( 'saurian', 'saurians', count, _ ).
noun( 'sausage', 'sausages', both, _ ).
noun( 'sausage-dog', 'sausage-dogs', count, _ ).
noun( 'sausage-meat', '-', mass, _ ).
noun( 'sausage-roll', 'sausage-rolls', count, _ ).
noun( 'savage', 'savages', count, _ ).
noun( 'savageness', '-', mass, _ ).
noun( 'savagery', '-', mass, _ ).
noun( 'savanna', 'savannas', count, _ ).
noun( 'savannah', 'savannahs', count, _ ).
noun( 'savant', 'savants', count, _ ).
noun( 'save', 'saves', count, _ ).
noun( 'saveloy', 'saveloys', both, _ ).
noun( 'saver', 'savers', count, _ ).
noun( 'saving', 'savings', count, _ ).
noun( 'savings-bank', 'savings-banks', count, _ ).
noun( 'saviour', 'saviours', count, _ ).
noun( 'savoir-faire', '-', mass, _ ).
noun( 'savory', '-', mass, _ ).
noun( 'savour', 'savours', both, _ ).
noun( 'savoury', 'savouries', count, _ ).
noun( 'savoy', 'savoys', both, _ ).
noun( 'savvy', '-', mass, _ ).
noun( 'saw', 'saws', count, _ ).
noun( 'saw-pit', 'saw-pits', count, _ ).
noun( 'sawdust', '-', mass, _ ).
noun( 'sawhorse', 'sawhorses', count, _ ).
noun( 'sawmill', 'sawmills', count, _ ).
noun( 'sawyer', 'sawyers', count, _ ).
noun( 'sax', 'saxes', count, _ ).
noun( 'saxhorn', 'saxhorns', count, _ ).
noun( 'saxifrage', '-', mass, _ ).
noun( 'saxophone', 'saxophones', count, _ ).
noun( 'saxophonist', 'saxophonists', count, _ ).
noun( 'say', '-', count, _ ).
noun( 'saying', 'sayings', count, _ ).
noun( 'scab', 'scabs', both, _ ).
noun( 'scabbard', 'scabbards', count, _ ).
noun( 'scabies', '-', mass, _ ).
noun( 'scabious', '-', mass, _ ).
noun( 'scaffold', 'scaffolds', count, _ ).
noun( 'scaffolding', '-', mass, _ ).
noun( 'scalawag', 'scalawags', count, _ ).
noun( 'scald', 'scalds', count, _ ).
noun( 'scale', 'scales', both, _ ).
noun( 'scaling-ladder', 'scaling-ladders', count, _ ).
noun( 'scallop', 'scallops', count, _ ).
noun( 'scallop-shell', 'scallop-shells', count, _ ).
noun( 'scallywag', 'scallywags', count, _ ).
noun( 'scalp', 'scalps', count, _ ).
noun( 'scalpel', 'scalpels', count, _ ).
noun( 'scamp', 'scamps', count, _ ).
noun( 'scamper', 'scampers', count, _ ).
noun( 'scandal', 'scandals', both, _ ).
noun( 'scandalmonger', 'scandalmongers', count, _ ).
noun( 'scandalmongering', '-', mass, _ ).
noun( 'scanner', 'scanners', count, _ ).
noun( 'scansion', '-', mass, _ ).
noun( 'scantiness', '-', mass, _ ).
noun( 'scantling', 'scantlings', count, _ ).
noun( 'scapegoat', 'scapegoats', count, _ ).
noun( 'scapegrace', 'scapegraces', count, _ ).
noun( 'scapula', 'scapulas', count, _ ).
noun( 'scar', 'scars', count, _ ).
noun( 'scarab', 'scarabs', count, _ ).
noun( 'scarcity', 'scarcities', both, _ ).
noun( 'scare', 'scares', count, _ ).
noun( 'scarecrow', 'scarecrows', count, _ ).
noun( 'scaremonger', 'scaremongers', count, _ ).
noun( 'scarf', 'scarfs', count, _ ).
noun( 'scarf-pin', 'scarf-pins', count, _ ).
noun( 'scarlet', 'scarlets', both, _ ).
noun( 'scarp', 'scarps', count, _ ).
noun( 'scatter', 'scatters', count, _ ).
noun( 'scatterbrain', 'scatterbrains', count, _ ).
noun( 'scavenger', 'scavengers', count, _ ).
noun( 'scenario', 'scenarios', count, _ ).
noun( 'scenarist', 'scenarists', count, _ ).
noun( 'scene', 'scenes', count, _ ).
noun( 'scene-painter', 'scene-painters', count, _ ).
noun( 'scene-shifter', 'scene-shifters', count, _ ).
noun( 'scenery', '-', mass, _ ).
noun( 'scent', 'scents', both, _ ).
noun( 'sceptic', 'sceptics', count, _ ).
noun( 'scepticism', '-', mass, _ ).
noun( 'sceptre', 'sceptres', count, _ ).
noun( 'schedule', 'schedules', count, _ ).
noun( 'schema', 'schemas', count, _ ).
noun( 'scheme', 'schemes', count, _ ).
noun( 'schemer', 'schemers', count, _ ).
noun( 'scherzo', 'scherzos', count, _ ).
noun( 'schism', 'schisms', both, _ ).
noun( 'schist', 'schists', count, _ ).
noun( 'schizophrenia', '-', mass, _ ).
noun( 'schizophrenic', 'schizophrenics', count, _ ).
noun( 'schmaltz', '-', mass, _ ).
noun( 'schmalz', '-', mass, _ ).
noun( 'schnapps', '-', mass, _ ).
noun( 'schnitzel', 'schnitzels', both, _ ).
noun( 'schnorkel', 'schnorkels', count, _ ).
noun( 'scholar', 'scholars', count, _ ).
noun( 'scholarship', 'scholarships', both, _ ).
noun( 'scholasticism', '-', mass, _ ).
noun( 'school', 'schools', both, _ ).
noun( 'school-board', 'school-boards', count, _ ).
noun( 'schoolbook', 'schoolbooks', count, _ ).
noun( 'schoolboy', 'schoolboys', count, _ ).
noun( 'schoolchild', 'schoolchildren', count, _ ).
noun( 'schoolfellow', 'schoolfellows', count, _ ).
noun( 'schoolfriend', 'schoolfriends', count, _ ).
noun( 'schoolgirl', 'schoolgirls', count, _ ).
noun( 'schoolhouse', 'schoolhouses', count, _ ).
noun( 'schooling', '-', mass, _ ).
noun( 'schoolman', 'schoolmen', count, _ ).
noun( 'schoolmaster', 'schoolmasters', count, _ ).
noun( 'schoolmate', 'schoolmates', count, _ ).
noun( 'schoolmistress', 'schoolmistresses', count, _ ).
noun( 'schoolroom', 'schoolrooms', count, _ ).
noun( 'schoolteacher', 'schoolteachers', count, _ ).
noun( 'schooltime', 'schooltimes', both, _ ).
noun( 'schooner', 'schooners', count, _ ).
noun( 'schottische', 'schottisches', count, _ ).
noun( 'schwa', 'schwas', count, _ ).
noun( 'sciatica', '-', mass, _ ).
noun( 'science', 'sciences', both, _ ).
noun( 'scientist', 'scientists', count, _ ).
noun( 'scimitar', 'scimitars', count, _ ).
noun( 'scintilla', 'scintillas', count, _ ).
noun( 'scintillation', 'scintillations', count, _ ).
noun( 'scion', 'scions', count, _ ).
noun( 'sclerosis', '-', mass, _ ).
noun( 'scoff', 'scoffs', count, _ ).
noun( 'scoffer', 'scoffers', count, _ ).
noun( 'scold', 'scolds', count, _ ).
noun( 'scolding', 'scoldings', count, _ ).
noun( 'scollop', 'scollops', count, _ ).
noun( 'sconce', 'sconces', count, _ ).
noun( 'scone', 'scones', count, _ ).
noun( 'scoop', 'scoops', count, _ ).
noun( 'scoopful', 'scoopfuls', count, _ ).
noun( 'scooter', 'scooters', count, _ ).
noun( 'scope', '-', mass, _ ).
noun( 'scorch', 'scorches', count, _ ).
noun( 'scorcher', 'scorchers', count, _ ).
noun( 'score', 'scores', count, _ ).
noun( 'scoreboard', 'scoreboards', count, _ ).
noun( 'scorebook', 'scorebooks', count, _ ).
noun( 'scorecard', 'scorecards', count, _ ).
noun( 'scorer', 'scorers', count, _ ).
noun( 'scorn', '-', mass, _ ).
noun( 'scorpion', 'scorpions', count, _ ).
noun( 'scot', 'scots', count, _ ).
noun( 'scotch', 'scotches', both, _ ).
noun( 'scoundrel', 'scoundrels', count, _ ).
noun( 'scour', 'scours', count, _ ).
noun( 'scourer', 'scourers', count, _ ).
noun( 'scourge', 'scourges', count, _ ).
noun( 'scout', 'scouts', count, _ ).
noun( 'scoutmaster', 'scoutmasters', count, _ ).
noun( 'scow', 'scows', count, _ ).
noun( 'scowl', 'scowls', count, _ ).
noun( 'scrabble', 'scrabbles', both, _ ).
noun( 'scrag', 'scrags', count, _ ).
noun( 'scrag-end', 'scrag-ends', count, _ ).
noun( 'scramble', 'scrambles', count, _ ).
noun( 'scrambler', 'scramblers', count, _ ).
noun( 'scrap', 'scraps', both, _ ).
noun( 'scrap-iron', '-', mass, _ ).
noun( 'scrapbook', 'scrapbooks', count, _ ).
noun( 'scrape', 'scrapes', count, _ ).
noun( 'scraper', 'scrapers', count, _ ).
noun( 'scrapheap', 'scrapheaps', count, _ ).
noun( 'scraping', 'scrapings', count, _ ).
noun( 'scrappiness', '-', mass, _ ).
noun( 'scratch', 'scratches', count, _ ).
noun( 'scratch-pad', 'scratch-pads', count, _ ).
noun( 'scratch-race', 'scratch-races', count, _ ).
noun( 'scrawl', 'scrawls', count, _ ).
noun( 'scream', 'screams', count, _ ).
noun( 'scree', 'screes', both, _ ).
noun( 'screech', 'screeches', count, _ ).
noun( 'screech-owl', 'screech-owls', count, _ ).
noun( 'screed', 'screeds', count, _ ).
noun( 'screen', 'screens', count, _ ).
noun( 'screw', 'screws', both, _ ).
noun( 'screwball', 'screwballs', count, _ ).
noun( 'screwdriver', 'screwdrivers', count, _ ).
noun( 'scribble', 'scribbles', both, _ ).
noun( 'scribbler', 'scribblers', count, _ ).
noun( 'scribbling-block', 'scribbling-blocks', count, _ ).
noun( 'scribe', 'scribes', count, _ ).
noun( 'scrimmage', 'scrimmages', count, _ ).
noun( 'scrimshanker', 'scrimshankers', count, _ ).
noun( 'scrip', 'scrips', both, _ ).
noun( 'script', 'scripts', both, _ ).
noun( 'scripture', 'scriptures', count, _ ).
noun( 'scriptwriter', 'scriptwriters', count, _ ).
noun( 'scrivener', 'scriveners', count, _ ).
noun( 'scrofula', '-', mass, _ ).
noun( 'scroll', 'scrolls', count, _ ).
noun( 'scrotum', 'scrotums', count, _ ).
noun( 'scrounger', 'scroungers', count, _ ).
noun( 'scrub', 'scrubs', both, _ ).
noun( 'scrubbing-brush', 'scrubbing-brushes', count, _ ).
noun( 'scruff', 'scruffs', count, _ ).
noun( 'scrum', 'scrums', count, _ ).
noun( 'scrummage', 'scrummages', count, _ ).
noun( 'scrunch', 'scrunches', count, _ ).
noun( 'scruple', 'scruples', both, _ ).
noun( 'scrutineer', 'scrutineers', count, _ ).
noun( 'scrutiny', 'scrutinies', both, _ ).
noun( 'scud', 'scuds', both, _ ).
noun( 'scull', 'sculls', count, _ ).
noun( 'sculler', 'scullers', count, _ ).
noun( 'scullery', 'sculleries', count, _ ).
noun( 'scullion', 'scullions', count, _ ).
noun( 'sculptor', 'sculptors', count, _ ).
noun( 'sculptress', 'sculptresses', count, _ ).
noun( 'sculpture', 'sculptures', both, _ ).
noun( 'scum', '-', mass, _ ).
noun( 'scupper', 'scuppers', count, _ ).
noun( 'scurf', '-', mass, _ ).
noun( 'scurrility', '-', mass, _ ).
noun( 'scurry', 'scurries', both, _ ).
noun( 'scurvy', '-', mass, _ ).
noun( 'scut', 'scuts', count, _ ).
noun( 'scutcheon', 'scutcheons', count, _ ).
noun( 'scuttle', 'scuttles', count, _ ).
noun( 'scythe', 'scythes', count, _ ).
noun( 'sea', 'seas', both, _ ).
noun( 'sea-anemone', 'sea-anemones', count, _ ).
noun( 'sea-animal', 'sea-animals', count, _ ).
noun( 'sea-bathing', '-', mass, _ ).
noun( 'sea-boat', 'sea-boats', count, _ ).
noun( 'sea-bream', 'sea-bream', count, _ ).
noun( 'sea-breeze', 'sea-breezes', count, _ ).
noun( 'sea-coal', '-', mass, _ ).
noun( 'sea-cow', 'sea-cows', count, _ ).
noun( 'sea-dog', 'sea-dogs', count, _ ).
noun( 'sea-fish', 'sea-fish', count, _ ).
noun( 'sea-god', 'sea-gods', count, _ ).
noun( 'sea-green', '-', mass, _ ).
noun( 'sea-horse', 'sea-horses', count, _ ).
noun( 'sea-level', '-', count, _ ).
noun( 'sea-lion', 'sea-lions', count, _ ).
noun( 'sea-power', '-', mass, _ ).
noun( 'sea-rover', 'sea-rovers', count, _ ).
noun( 'sea-snake', 'sea-snakes', count, _ ).
noun( 'sea-urchin', 'sea-urchins', count, _ ).
noun( 'sea-wall', 'sea-walls', count, _ ).
noun( 'sea-water', '-', mass, _ ).
noun( 'seabed', 'seabeds', count, _ ).
noun( 'seabird', 'seabirds', count, _ ).
noun( 'seaboard', 'seaboards', count, _ ).
noun( 'seafarer', 'seafarers', count, _ ).
noun( 'seafood', 'seafoods', both, _ ).
noun( 'seafront', 'seafronts', count, _ ).
noun( 'seagull', 'seagulls', count, _ ).
noun( 'seakale', '-', mass, _ ).
noun( 'seal', 'seals', count, _ ).
noun( 'seal-ring', 'seal-rings', count, _ ).
noun( 'sealer', 'sealers', count, _ ).
noun( 'sealing-wax', '-', mass, _ ).
noun( 'sealskin', 'sealskins', both, _ ).
noun( 'seam', 'seams', count, _ ).
noun( 'seaman', 'seamen', count, _ ).
noun( 'seamanship', '-', mass, _ ).
noun( 'seamstress', 'seamstresses', count, _ ).
noun( 'seaplane', 'seaplanes', count, _ ).
noun( 'seaport', 'seaports', count, _ ).
noun( 'search', 'searches', both, _ ).
noun( 'search-party', 'search-parties', count, _ ).
noun( 'search-warrant', 'search-warrants', count, _ ).
noun( 'searcher', 'searchers', count, _ ).
noun( 'searchlight', 'searchlights', count, _ ).
noun( 'searing-iron', 'searing-irons', count, _ ).
noun( 'seascape', 'seascapes', count, _ ).
noun( 'seashell', 'seashells', count, _ ).
noun( 'seashore', 'seashores', count, _ ).
noun( 'seasickness', '-', mass, _ ).
noun( 'seaside', 'seasides', count, _ ).
noun( 'season', 'seasons', count, _ ).
noun( 'season-ticket', 'season-tickets', count, _ ).
noun( 'seasoning', 'seasonings', both, _ ).
noun( 'seat', 'seats', count, _ ).
noun( 'seat-belt', 'seat-belts', count, _ ).
noun( 'seating-room', '-', mass, _ ).
noun( 'seaway', 'seaways', count, _ ).
noun( 'seaweed', 'seaweeds', both, _ ).
noun( 'seaworthiness', '-', mass, _ ).
noun( 'sec', 'secs', count, _ ).
noun( 'secession', 'secessions', both, _ ).
noun( 'secessionist', 'secessionists', count, _ ).
noun( 'seclusion', '-', mass, _ ).
noun( 'second', 'seconds', count, _ ).
noun( 'second-best', '-', count, _ ).
noun( 'second-class', '-', mass, _ ).
noun( 'second-hand', 'second-hands', count, _ ).
noun( 'second-rater', 'second-raters', count, _ ).
noun( 'seconder', 'seconders', count, _ ).
noun( 'secondment', 'secondments', count, _ ).
noun( 'secrecy', '-', mass, _ ).
noun( 'secret', 'secrets', both, _ ).
noun( 'secretariat', 'secretariats', count, _ ).
noun( 'secretary', 'secretaries', count, _ ).
noun( 'secretion', 'secretions', both, _ ).
noun( 'secretiveness', '-', mass, _ ).
noun( 'sect', 'sects', count, _ ).
noun( 'sectarian', 'sectarians', count, _ ).
noun( 'sectarianism', '-', mass, _ ).
noun( 'section', 'sections', count, _ ).
noun( 'sectionalism', '-', mass, _ ).
noun( 'sector', 'sectors', count, _ ).
noun( 'secularism', '-', mass, _ ).
noun( 'secularist', 'secularists', count, _ ).
noun( 'security', 'securities', both, _ ).
noun( 'sedan', 'sedans', count, _ ).
noun( 'sedan-chair', 'sedan-chairs', count, _ ).
noun( 'sedateness', '-', mass, _ ).
noun( 'sedation', '-', mass, _ ).
noun( 'sedative', 'sedatives', both, _ ).
noun( 'sedge', 'sedges', both, _ ).
noun( 'sediment', '-', mass, _ ).
noun( 'sedition', '-', mass, _ ).
noun( 'seducer', 'seducers', count, _ ).
noun( 'seduction', 'seductions', both, _ ).
noun( 'see', 'sees', count, _ ).
noun( 'seed', 'seed', both, _ ).
noun( 'seed-corn', '-', mass, _ ).
noun( 'seedbed', 'seedbeds', count, _ ).
noun( 'seedcake', 'seedcakes', count, _ ).
noun( 'seediness', '-', mass, _ ).
noun( 'seedling', 'seedlings', count, _ ).
noun( 'seedsman', 'seedsmen', count, _ ).
noun( 'seedtime', 'seedtimes', count, _ ).
noun( 'seeker', 'seekers', count, _ ).
noun( 'seemliness', '-', mass, _ ).
noun( 'seepage', '-', mass, _ ).
noun( 'seer', 'seers', count, _ ).
noun( 'seersucker', '-', mass, _ ).
noun( 'seesaw', 'seesaws', both, _ ).
noun( 'segment', 'segments', count, _ ).
noun( 'segmentation', 'segmentations', both, _ ).
noun( 'segregation', 'segregations', count, _ ).
noun( 'seignior', 'seigniors', count, _ ).
noun( 'seine', 'seines', count, _ ).
noun( 'seismograph', 'seismographs', count, _ ).
noun( 'seismologist', 'seismologists', count, _ ).
noun( 'seismology', '-', mass, _ ).
noun( 'seizure', 'seizures', both, _ ).
noun( 'selection', 'selections', both, _ ).
noun( 'selectivity', '-', mass, _ ).
noun( 'selector', 'selectors', count, _ ).
noun( 'selenium', '-', mass, _ ).
noun( 'self', 'selves', both, _ ).
noun( 'self-abasement', '-', mass, _ ).
noun( 'self-abnegation', '-', mass, _ ).
noun( 'self-assertion', '-', mass, _ ).
noun( 'self-assurance', '-', mass, _ ).
noun( 'self-command', '-', mass, _ ).
noun( 'self-communion', '-', mass, _ ).
noun( 'self-complacency', '-', mass, _ ).
noun( 'self-confidence', '-', mass, _ ).
noun( 'self-consciousness', '-', mass, _ ).
noun( 'self-control', '-', mass, _ ).
noun( 'self-defence', '-', mass, _ ).
noun( 'self-denial', '-', mass, _ ).
noun( 'self-determination', '-', mass, _ ).
noun( 'self-esteem', '-', mass, _ ).
noun( 'self-examination', 'self-examinations', both, _ ).
noun( 'self-help', '-', mass, _ ).
noun( 'self-importance', '-', mass, _ ).
noun( 'self-indulgence', '-', mass, _ ).
noun( 'self-interest', '-', mass, _ ).
noun( 'self-pity', '-', mass, _ ).
noun( 'self-possession', '-', mass, _ ).
noun( 'self-preservation', '-', mass, _ ).
noun( 'self-reliance', '-', mass, _ ).
noun( 'self-respect', '-', mass, _ ).
noun( 'self-rule', '-', mass, _ ).
noun( 'self-sacrifice', 'self-sacrifices', both, _ ).
noun( 'self-seeker', 'self-seekers', count, _ ).
noun( 'self-seeking', '-', mass, _ ).
noun( 'self-service', '-', mass, _ ).
noun( 'self-starter', 'self-starters', count, _ ).
noun( 'self-sufficiency', '-', mass, _ ).
noun( 'self-will', '-', mass, _ ).
noun( 'selfishness', '-', mass, _ ).
noun( 'sell', '-', count, _ ).
noun( 'seller', 'sellers', count, _ ).
noun( 'selling', '-', mass, _ ).
noun( 'sellout', 'sellouts', count, _ ).
noun( 'selvage', 'selvages', count, _ ).
noun( 'selvedge', 'selvedges', count, _ ).
noun( 'semantics', 'semantics', mass, _ ).
noun( 'semaphore', '-', mass, _ ).
noun( 'semblance', 'semblances', both, _ ).
noun( 'semen', '-', mass, _ ).
noun( 'semester', 'semesters', count, _ ).
noun( 'semibreve', 'semibreves', count, _ ).
noun( 'semicircle', 'semicircles', count, _ ).
noun( 'semicolon', 'semicolons', count, _ ).
noun( 'semiconductor', 'semiconductors', count, _ ).
noun( 'semifinal', 'semifinals', count, _ ).
noun( 'semifinalist', 'semifinalists', count, _ ).
noun( 'seminar', 'seminars', count, _ ).
noun( 'seminarist', 'seminarists', count, _ ).
noun( 'seminary', 'seminaries', count, _ ).
noun( 'semiquaver', 'semiquavers', count, _ ).
noun( 'semitone', 'semitones', count, _ ).
noun( 'semivowel', 'semivowels', count, _ ).
noun( 'semolina', '-', mass, _ ).
noun( 'sempstress', 'sempstresses', count, _ ).
noun( 'senate', 'senates', count, _ ).
noun( 'senator', 'senators', count, _ ).
noun( 'send-up', 'send-ups', count, _ ).
noun( 'sender', 'senders', count, _ ).
noun( 'sendoff', 'sendoffs', count, _ ).
noun( 'senescence', '-', mass, _ ).
noun( 'seneschal', 'seneschals', count, _ ).
noun( 'senility', '-', mass, _ ).
noun( 'senior', 'seniors', count, _ ).
noun( 'seniority', '-', mass, _ ).
noun( 'senna', '-', mass, _ ).
noun( 'senora', 'senoras', count, _ ).
noun( 'senorita', 'senoritas', count, _ ).
noun( 'sensation', 'sensations', both, _ ).
noun( 'sensationalism', '-', mass, _ ).
noun( 'sensationalist', 'sensationalists', count, _ ).
noun( 'sense', 'senses', both, _ ).
noun( 'sense-organ', 'sense-organs', count, _ ).
noun( 'senselessness', '-', mass, _ ).
noun( 'sensibility', 'sensibilities', both, _ ).
noun( 'sensitivity', 'sensitivities', both, _ ).
noun( 'sensitization', '-', mass, _ ).
noun( 'sensualism', '-', mass, _ ).
noun( 'sensualist', 'sensualists', count, _ ).
noun( 'sensuality', '-', mass, _ ).
noun( 'sensuousness', '-', mass, _ ).
noun( 'sentence', 'sentences', count, _ ).
noun( 'sentiment', 'sentiments', both, _ ).
noun( 'sentimentalist', 'sentimentalists', count, _ ).
noun( 'sentimentality', '-', mass, _ ).
noun( 'sentinel', 'sentinels', count, _ ).
noun( 'sentry', 'sentries', count, _ ).
noun( 'sentry-box', 'sentry-boxes', count, _ ).
noun( 'sentry-go', '-', count, _ ).
noun( 'sepal', 'sepals', count, _ ).
noun( 'separability', '-', mass, _ ).
noun( 'separate', 'separates', count, _ ).
noun( 'separation', 'separations', both, _ ).
noun( 'separatist', 'separatists', count, _ ).
noun( 'separator', 'separators', count, _ ).
noun( 'sepia', '-', mass, _ ).
noun( 'sepsis', '-', mass, _ ).
noun( 'septet', 'septets', count, _ ).
noun( 'septicaemia', '-', mass, _ ).
noun( 'septicemia', '-', mass, _ ).
noun( 'septuagenarian', 'septuagenarians', count, _ ).
noun( 'sepulchre', 'sepulchres', count, _ ).
noun( 'sepulture', '-', mass, _ ).
noun( 'sequel', 'sequels', count, _ ).
noun( 'sequence', 'sequences', both, _ ).
noun( 'sequestration', 'sequestrations', both, _ ).
noun( 'sequin', 'sequins', count, _ ).
noun( 'sequoia', 'sequoias', count, _ ).
noun( 'seraglio', 'seraglios', count, _ ).
noun( 'seraph', 'seraphs', count, _ ).
noun( 'serenade', 'serenades', count, _ ).
noun( 'serendipity', '-', mass, _ ).
noun( 'serenity', '-', mass, _ ).
noun( 'serf', 'serfs', count, _ ).
noun( 'serfdom', '-', mass, _ ).
noun( 'serge', '-', mass, _ ).
noun( 'sergeant', 'sergeants', count, _ ).
noun( 'sergeant-major', 'sergeant-majors', count, _ ).
noun( 'serial', 'serials', count, _ ).
noun( 'sericulture', 'sericultures', count, _ ).
noun( 'sericulturist', 'sericulturists', count, _ ).
noun( 'series', 'series', count, _ ).
noun( 'seriousness', '-', mass, _ ).
noun( 'serjeant', 'serjeants', count, _ ).
noun( 'sermon', 'sermons', count, _ ).
noun( 'serpent', 'serpents', count, _ ).
noun( 'serum', '-', mass, _ ).
noun( 'servant', 'servants', count, _ ).
noun( 'serve', 'serves', count, _ ).
noun( 'server', 'servers', count, _ ).
noun( 'service', 'services', both, _ ).
noun( 'serviceman', 'servicemen', count, _ ).
noun( 'serviette', 'serviettes', count, _ ).
noun( 'servility', '-', mass, _ ).
noun( 'serving', 'servings', count, _ ).
noun( 'servitor', 'servitors', count, _ ).
noun( 'servitude', '-', mass, _ ).
noun( 'sesame', '-', mass, _ ).
noun( 'session', 'sessions', count, _ ).
noun( 'set', 'sets', both, _ ).
noun( 'set-square', 'set-squares', count, _ ).
noun( 'set-to', 'set-tos', count, _ ).
noun( 'set-up', 'set-ups', count, _ ).
noun( 'setback', 'setbacks', count, _ ).
noun( 'sett', 'setts', count, _ ).
noun( 'settee', 'settees', count, _ ).
noun( 'setter', 'setters', count, _ ).
noun( 'setting', 'settings', count, _ ).
noun( 'settle', 'settles', count, _ ).
noun( 'settlement', 'settlements', both, _ ).
noun( 'settler', 'settlers', count, _ ).
noun( 'seven', 'sevens', count, _ ).
noun( 'seventeen', 'seventeens', count, _ ).
noun( 'seventeenth', 'seventeenths', count, _ ).
noun( 'seventh', 'sevenths', count, _ ).
noun( 'seventieth', 'seventieths', count, _ ).
noun( 'seventy', 'seventies', count, _ ).
noun( 'severance', '-', mass, _ ).
noun( 'severity', 'severities', both, _ ).
noun( 'sewage', '-', mass, _ ).
noun( 'sewage-farm', 'sewage-farms', count, _ ).
noun( 'sewage-works', 'sewage-works', count, _ ).
noun( 'sewer', 'sewers', count, _ ).
noun( 'sewer', 'sewers', count, _ ).
noun( 'sewer-gas', '-', mass, _ ).
noun( 'sewer-rat', 'sewer-rats', count, _ ).
noun( 'sewerage', 'sewerages', both, _ ).
noun( 'sewing', '-', mass, _ ).
noun( 'sewing-machine', 'sewing-machines', count, _ ).
noun( 'sex', 'sexes', both, _ ).
noun( 'sexagenarian', 'sexagenarians', count, _ ).
noun( 'sexism', '-', mass, _ ).
noun( 'sexist', 'sexists', count, _ ).
noun( 'sextant', 'sextants', count, _ ).
noun( 'sextet', 'sextets', count, _ ).
noun( 'sextette', 'sextettes', count, _ ).
noun( 'sexton', 'sextons', count, _ ).
noun( 'sexuality', '-', mass, _ ).
noun( 'se~nor', 'se~nores', count, _ ).
noun( 'sgd', '-', proper, _ ).
noun( 'shabbiness', '-', mass, _ ).
noun( 'shack', 'shacks', count, _ ).
noun( 'shackle', 'shackles', count, _ ).
noun( 'shad', 'shad', count, _ ).
noun( 'shaddock', 'shaddocks', count, _ ).
noun( 'shade', 'shades', both, _ ).
noun( 'shade-tree', 'shade-trees', count, _ ).
noun( 'shading', 'shadings', both, _ ).
noun( 'shadow', 'shadows', both, _ ).
noun( 'shadow-boxing', '-', mass, _ ).
noun( 'shaft', 'shafts', count, _ ).
noun( 'shag', '-', mass, _ ).
noun( 'shagginess', '-', mass, _ ).
noun( 'shagging', '-', mass, _ ).
noun( 'shah', 'shahs', count, _ ).
noun( 'shake', 'shakes', count, _ ).
noun( 'shake-up', 'shake-ups', count, _ ).
noun( 'shakedown', 'shakedowns', count, _ ).
noun( 'shakeout', 'shakeouts', count, _ ).
noun( 'shaker', 'shakers', count, _ ).
noun( 'shakiness', '-', mass, _ ).
noun( 'shaking', 'shakings', count, _ ).
noun( 'shale', '-', mass, _ ).
noun( 'shale-oil', '-', mass, _ ).
noun( 'shallot', 'shallots', count, _ ).
noun( 'shallow', 'shallows', count, _ ).
noun( 'sham', '-', both, _ ).
noun( 'shamble', 'shambles', count, _ ).
noun( 'shambles', '-', count, _ ).
noun( 'shame', '-', mass, _ ).
noun( 'shamelessness', '-', mass, _ ).
noun( 'shammy', 'shammies', count, _ ).
noun( 'shampoo', 'shampoos', both, _ ).
noun( 'shamrock', 'shamrocks', count, _ ).
noun( 'shandy', 'shandies', both, _ ).
noun( 'shank', 'shanks', count, _ ).
noun( 'shantung', '-', mass, _ ).
noun( 'shanty', 'shanties', count, _ ).
noun( 'shantytown', 'shantytowns', count, _ ).
noun( 'shape', 'shapes', both, _ ).
noun( 'shapelessness', '-', mass, _ ).
noun( 'shard', 'shards', count, _ ).
noun( 'share', 'shares', both, _ ).
noun( 'share-out', 'share-outs', count, _ ).
noun( 'sharecropper', 'sharecroppers', count, _ ).
noun( 'shareholder', 'shareholders', count, _ ).
noun( 'shareholding', 'shareholdings', count, _ ).
noun( 'shark', 'sharks', count, _ ).
noun( 'sharkskin', 'sharkskins', both, _ ).
noun( 'sharp', 'sharps', count, _ ).
noun( 'sharpener', 'sharpeners', count, _ ).
noun( 'sharper', 'sharpers', count, _ ).
noun( 'sharpness', '-', mass, _ ).
noun( 'sharpshooter', 'sharpshooters', count, _ ).
noun( 'shave', 'shaves', count, _ ).
noun( 'shaver', 'shavers', count, _ ).
noun( 'shaving-brush', 'shaving-brushes', count, _ ).
noun( 'shawl', 'shawls', count, _ ).
noun( 'she-goat', 'she-goats', count, _ ).
noun( 'sheaf', 'sheaves', count, _ ).
noun( 'sheath', 'sheaths', count, _ ).
noun( 'sheath-knife', 'sheath-knives', count, _ ).
noun( 'sheathing', 'sheathings', count, _ ).
noun( 'shebang', '-', count, _ ).
noun( 'shebeen', 'shebeens', count, _ ).
noun( 'shed', 'sheds', count, _ ).
noun( 'sheen', '-', mass, _ ).
noun( 'sheep', 'sheep', count, _ ).
noun( 'sheepdog', 'sheepdogs', count, _ ).
noun( 'sheepfold', 'sheepfolds', count, _ ).
noun( 'sheepishness', '-', mass, _ ).
noun( 'sheeprun', 'sheepruns', count, _ ).
noun( 'sheepskin', 'sheepskins', count, _ ).
noun( 'sheet', 'sheets', count, _ ).
noun( 'sheet-anchor', 'sheet-anchors', count, _ ).
noun( 'sheet-lightning', '-', mass, _ ).
noun( 'sheeting', '-', mass, _ ).
noun( 'sheik', 'sheiks', count, _ ).
noun( 'sheikdom', 'sheikdoms', count, _ ).
noun( 'sheikh', 'sheikhs', count, _ ).
noun( 'sheikhdom', 'sheikhdoms', count, _ ).
noun( 'shekel', 'shekels', count, _ ).
noun( 'sheldrake', 'sheldrakes', count, _ ).
noun( 'shelf', 'shelves', count, _ ).
noun( 'shell', 'shells', count, _ ).
noun( 'shell-shock', '-', mass, _ ).
noun( 'shellac', '-', mass, _ ).
noun( 'shellfire', '-', mass, _ ).
noun( 'shellfish', 'shellfish', count, _ ).
noun( 'shelter', 'shelters', both, _ ).
noun( 'shepherd', 'shepherds', count, _ ).
noun( 'shepherdess', 'shepherdesses', count, _ ).
noun( 'sherbet', 'sherbets', both, _ ).
noun( 'sheriff', 'sheriffs', count, _ ).
noun( 'sherry', 'sherries', both, _ ).
noun( 'shibboleth', 'shibboleths', count, _ ).
noun( 'shield', 'shields', count, _ ).
noun( 'shift', 'shifts', count, _ ).
noun( 'shiftiness', '-', mass, _ ).
noun( 'shillelagh', 'shillelaghs', count, _ ).
noun( 'shilling', 'shillings', count, _ ).
noun( 'shillyshally', '-', mass, _ ).
noun( 'shimmer', '-', mass, _ ).
noun( 'shin', 'shins', count, _ ).
noun( 'shinbone', 'shinbones', count, _ ).
noun( 'shindig', 'shindigs', count, _ ).
noun( 'shindy', 'shindies', count, _ ).
noun( 'shine', '-', mass, _ ).
noun( 'shingle', 'shingles', both, _ ).
noun( 'shingles', 'shingles', mass, _ ).
noun( 'shinguard', 'shinguards', count, _ ).
noun( 'ship', 'ships', count, _ ).
noun( 'ship\'s-chandler', 'ship\'s-chandlers', count, _ ).
noun( 'ship-breaker', 'ship-breakers', count, _ ).
noun( 'ship-canal', 'ship-canals', count, _ ).
noun( 'shipbroker', 'shipbrokers', count, _ ).
noun( 'shipbuilder', 'shipbuilders', count, _ ).
noun( 'shipbuilding', '-', mass, _ ).
noun( 'shipload', 'shiploads', count, _ ).
noun( 'shipmate', 'shipmates', count, _ ).
noun( 'shipment', 'shipments', both, _ ).
noun( 'shipowner', 'shipowners', count, _ ).
noun( 'shipper', 'shippers', count, _ ).
noun( 'shipping', '-', mass, _ ).
noun( 'shipping-agent', 'shipping-agents', count, _ ).
noun( 'shipping-office', 'shipping-offices', count, _ ).
noun( 'shipway', 'shipways', count, _ ).
noun( 'shipwreck', 'shipwrecks', both, _ ).
noun( 'shipwright', 'shipwrights', count, _ ).
noun( 'shipyard', 'shipyards', count, _ ).
noun( 'shire', 'shires', count, _ ).
noun( 'shirker', 'shirkers', count, _ ).
noun( 'shirt', 'shirts', count, _ ).
noun( 'shirt-front', 'shirt-fronts', count, _ ).
noun( 'shirting', '-', mass, _ ).
noun( 'shirtwaist', 'shirtwaists', count, _ ).
noun( 'shirtwaister', 'shirtwaisters', count, _ ).
noun( 'shish kebab', 'shish kebabs', count, _ ).
noun( 'shit', '-', mass, _ ).
noun( 'shiver', 'shivers', count, _ ).
noun( 'shoal', 'shoals', count, _ ).
noun( 'shock', 'shocks', both, _ ).
noun( 'shock-brigade', 'shock-brigades', count, _ ).
noun( 'shock-worker', 'shock-workers', count, _ ).
noun( 'shocker', 'shockers', count, _ ).
noun( 'shoddiness', '-', mass, _ ).
noun( 'shoddy', '-', mass, _ ).
noun( 'shoe', 'shoes', count, _ ).
noun( 'shoe-leather', 'shoe-leathers', both, _ ).
noun( 'shoeblack', '-', mass, _ ).
noun( 'shoehorn', 'shoehorns', count, _ ).
noun( 'shoelace', 'shoelaces', count, _ ).
noun( 'shoemaker', 'shoemakers', count, _ ).
noun( 'shoemaking', '-', mass, _ ).
noun( 'shoestring', 'shoestrings', count, _ ).
noun( 'shoetree', 'shoetrees', count, _ ).
noun( 'shogun', 'shoguns', count, _ ).
noun( 'shoot', 'shoots', count, _ ).
noun( 'shooter', 'shooters', count, _ ).
noun( 'shooting', '-', mass, _ ).
noun( 'shooting-box', 'shooting-boxes', count, _ ).
noun( 'shooting-brake', 'shooting-brakes', count, _ ).
noun( 'shooting-gallery', 'shooting-galleries', count, _ ).
noun( 'shooting-range', 'shooting-ranges', count, _ ).
noun( 'shooting-stick', 'shooting-sticks', count, _ ).
noun( 'shop', 'shops', both, _ ).
noun( 'shop-assistant', 'shop-assistants', count, _ ).
noun( 'shop-bell', 'shop-bells', count, _ ).
noun( 'shop-boy', 'shop-boys', count, _ ).
noun( 'shop-front', 'shop-fronts', count, _ ).
noun( 'shop-girl', 'shop-girls', count, _ ).
noun( 'shop-steward', 'shop-stewards', count, _ ).
noun( 'shopkeeper', 'shopkeepers', count, _ ).
noun( 'shoplifter', 'shoplifters', count, _ ).
noun( 'shoplifting', '-', mass, _ ).
noun( 'shopper', 'shoppers', count, _ ).
noun( 'shopping', '-', mass, _ ).
noun( 'shopwalker', 'shopwalkers', count, _ ).
noun( 'shopwindow', 'shopwindows', count, _ ).
noun( 'shore', 'shores', both, _ ).
noun( 'short', 'shorts', count, _ ).
noun( 'short-circuit', 'short-circuits', count, _ ).
noun( 'shortage', 'shortages', both, _ ).
noun( 'shortbread', '-', mass, _ ).
noun( 'shortcake', '-', mass, _ ).
noun( 'shortcoming', 'shortcomings', count, _ ).
noun( 'shortening', '-', mass, _ ).
noun( 'shortfall', 'shortfalls', count, _ ).
noun( 'shorthand', '-', mass, _ ).
noun( 'shorthorn', 'shorthorns', count, _ ).
noun( 'shortlist', 'shortlists', count, _ ).
noun( 'shortness', '-', mass, _ ).
noun( 'shot', 'shots', both, _ ).
noun( 'shot-put', 'shot-puts', both, _ ).
noun( 'shot-tower', 'shot-towers', count, _ ).
noun( 'shotgun', 'shotguns', count, _ ).
noun( 'shoulder', 'shoulders', count, _ ).
noun( 'shoulder-blade', 'shoulder-blades', count, _ ).
noun( 'shoulder-flash', 'shoulder-flashes', count, _ ).
noun( 'shoulder-strap', 'shoulder-straps', count, _ ).
noun( 'shout', 'shouts', count, _ ).
noun( 'shouting', '-', mass, _ ).
noun( 'shove', 'shoves', count, _ ).
noun( 'shove-ha\'penny', '-', mass, _ ).
noun( 'shovel', 'shovels', count, _ ).
noun( 'shovel-board', '-', mass, _ ).
noun( 'shovelful', 'shovelfuls', count, _ ).
noun( 'show', 'shows', both, _ ).
noun( 'show-business', '-', mass, _ ).
noun( 'show-off', 'show-offs', count, _ ).
noun( 'show-window', 'show-windows', count, _ ).
noun( 'showbiz', '-', mass, _ ).
noun( 'showboat', 'showboats', count, _ ).
noun( 'showcase', 'showcases', count, _ ).
noun( 'showdown', 'showdowns', count, _ ).
noun( 'shower', 'showers', count, _ ).
noun( 'shower-bath', 'shower-baths', count, _ ).
noun( 'showgirl', 'showgirls', count, _ ).
noun( 'showiness', '-', mass, _ ).
noun( 'showing', 'showings', count, _ ).
noun( 'showjumping', '-', mass, _ ).
noun( 'showman', 'showmen', count, _ ).
noun( 'showmanship', '-', mass, _ ).
noun( 'showplace', 'showplaces', count, _ ).
noun( 'showroom', 'showrooms', count, _ ).
noun( 'shrapnel', '-', mass, _ ).
noun( 'shred', 'shreds', count, _ ).
noun( 'shrew', 'shrews', count, _ ).
noun( 'shrew-mouse', 'shrew-mice', count, _ ).
noun( 'shrewdness', '-', mass, _ ).
noun( 'shrewishness', '-', mass, _ ).
noun( 'shriek', 'shrieks', count, _ ).
noun( 'shrift', '-', mass, _ ).
noun( 'shrike', 'shrikes', count, _ ).
noun( 'shrillness', '-', mass, _ ).
noun( 'shrimp', 'shrimps', count, _ ).
noun( 'shrine', 'shrines', count, _ ).
noun( 'shrink', 'shrinks', count, _ ).
noun( 'shrinkage', '-', mass, _ ).
noun( 'shroud', 'shrouds', count, _ ).
noun( 'shrub', 'shrubs', count, _ ).
noun( 'shrubbery', 'shrubberies', count, _ ).
noun( 'shrug', 'shrugs', count, _ ).
noun( 'shuck', 'shucks', count, _ ).
noun( 'shudder', 'shudders', count, _ ).
noun( 'shuffle', 'shuffles', count, _ ).
noun( 'shuffler', 'shufflers', count, _ ).
noun( 'shunter', 'shunters', count, _ ).
noun( 'shutdown', 'shutdowns', count, _ ).
noun( 'shuteye', '-', mass, _ ).
noun( 'shutter', 'shutters', count, _ ).
noun( 'shuttle', 'shuttles', count, _ ).
noun( 'shuttlecock', 'shuttlecocks', count, _ ).
noun( 'shy', 'shies', count, _ ).
noun( 'shyness', '-', mass, _ ).
noun( 'shyster', 'shysters', count, _ ).
noun( 'sibilant', 'sibilants', count, _ ).
noun( 'sibling', 'siblings', count, _ ).
noun( 'sibyl', 'sibyls', count, _ ).
noun( 'sick-benefit', 'sick-benefits', both, _ ).
noun( 'sick-berth', 'sick-berths', count, _ ).
noun( 'sick-headache', 'sick-headaches', count, _ ).
noun( 'sick-leave', '-', mass, _ ).
noun( 'sick-list', 'sick-lists', count, _ ).
noun( 'sick-parade', 'sick-parades', count, _ ).
noun( 'sick-pay', '-', mass, _ ).
noun( 'sick-room', 'sick-rooms', count, _ ).
noun( 'sickbay', 'sickbays', count, _ ).
noun( 'sickbed', 'sickbeds', count, _ ).
noun( 'sickle', 'sickles', count, _ ).
noun( 'sickness', 'sicknesses', both, _ ).
noun( 'side', 'sides', both, _ ).
noun( 'side-chapel', 'side-chapels', count, _ ).
noun( 'side-dish', 'side-dishes', count, _ ).
noun( 'side-drum', 'side-drums', count, _ ).
noun( 'side-glance', 'side-glances', count, _ ).
noun( 'side-road', 'side-roads', count, _ ).
noun( 'side-saddle', 'side-saddles', count, _ ).
noun( 'side-slip', 'side-slips', count, _ ).
noun( 'side-stroke', '-', mass, _ ).
noun( 'side-view', 'side-views', count, _ ).
noun( 'sideboard', 'sideboards', count, _ ).
noun( 'sidecar', 'sidecars', count, _ ).
noun( 'sidelight', 'sidelights', count, _ ).
noun( 'sideline', 'sidelines', count, _ ).
noun( 'sideshow', 'sideshows', count, _ ).
noun( 'sidesman', 'sidesmen', count, _ ).
noun( 'sidestep', 'sidesteps', count, _ ).
noun( 'sidetrack', 'sidetracks', count, _ ).
noun( 'sidewalk', 'sidewalks', count, _ ).
noun( 'siding', 'sidings', count, _ ).
noun( 'siege', 'sieges', both, _ ).
noun( 'sienna', '-', mass, _ ).
noun( 'sierra', 'sierras', count, _ ).
noun( 'siesta', 'siestas', count, _ ).
noun( 'sieve', 'sieves', count, _ ).
noun( 'sifter', 'sifters', count, _ ).
noun( 'sigh', 'sighs', count, _ ).
noun( 'sight', 'sights', both, _ ).
noun( 'sighting', 'sightings', count, _ ).
noun( 'sightseeing', '-', mass, _ ).
noun( 'sightseer', 'sightseers', count, _ ).
noun( 'sign', 'signs', count, _ ).
noun( 'sign-painter', 'sign-painters', count, _ ).
noun( 'signal', 'signals', count, _ ).
noun( 'signal-box', 'signal-boxes', count, _ ).
noun( 'signaller', 'signallers', count, _ ).
noun( 'signalman', 'signalmen', count, _ ).
noun( 'signatory', 'signatories', count, _ ).
noun( 'signature', 'signatures', count, _ ).
noun( 'signet', 'signets', count, _ ).
noun( 'signet-ring', 'signet-rings', count, _ ).
noun( 'significance', '-', mass, _ ).
noun( 'signification', 'significations', count, _ ).
noun( 'signor', 'signors', count, _ ).
noun( 'signora', 'signoras', count, _ ).
noun( 'signorina', 'signorinas', count, _ ).
noun( 'signpost', 'signposts', count, _ ).
noun( 'silage', '-', mass, _ ).
noun( 'silence', 'silences', both, _ ).
noun( 'silencer', 'silencers', count, _ ).
noun( 'silhouette', 'silhouettes', count, _ ).
noun( 'silica', '-', mass, _ ).
noun( 'silicate', '-', mass, _ ).
noun( 'silicon', '-', mass, _ ).
noun( 'silicone', '-', mass, _ ).
noun( 'silicosis', '-', mass, _ ).
noun( 'silk', 'silks', both, _ ).
noun( 'silkiness', '-', mass, _ ).
noun( 'silkworm', 'silkworms', count, _ ).
noun( 'sill', 'sills', count, _ ).
noun( 'sillabub', 'sillabubs', both, _ ).
noun( 'silliness', '-', mass, _ ).
noun( 'silly', 'sillies', count, _ ).
noun( 'silo', 'silos', count, _ ).
noun( 'silt', '-', mass, _ ).
noun( 'silver', '-', mass, _ ).
noun( 'silver-fish', 'silver-fish', count, _ ).
noun( 'silverside', '-', mass, _ ).
noun( 'silversmith', 'silversmiths', count, _ ).
noun( 'simian', 'simians', count, _ ).
noun( 'similarity', 'similarities', both, _ ).
noun( 'simile', 'similes', both, _ ).
noun( 'similitude', 'similitudes', both, _ ).
noun( 'simmer', '-', count, _ ).
noun( 'simony', '-', mass, _ ).
noun( 'simoom', 'simooms', count, _ ).
noun( 'simoon', 'simoons', count, _ ).
noun( 'simple', 'simples', count, _ ).
noun( 'simpleton', 'simpletons', count, _ ).
noun( 'simplicity', '-', mass, _ ).
noun( 'simplification', 'simplifications', both, _ ).
noun( 'simulacrum', 'simulacra', count, _ ).
noun( 'simulation', '-', mass, _ ).
noun( 'simulator', 'simulators', count, _ ).
noun( 'simultaneity', '-', mass, _ ).
noun( 'simultaneousness', '-', mass, _ ).
noun( 'sin', 'sins', both, _ ).
noun( 'sincerity', '-', mass, _ ).
noun( 'sine', 'sines', count, _ ).
noun( 'sine qua non', '-', count, _ ).
noun( 'sinecure', 'sinecures', count, _ ).
noun( 'sinew', 'sinews', count, _ ).
noun( 'sinfulness', '-', mass, _ ).
noun( 'singe', 'singes', count, _ ).
noun( 'singer', 'singers', count, _ ).
noun( 'singing', '-', mass, _ ).
noun( 'single', 'singles', count, _ ).
noun( 'single-spacing', '-', mass, _ ).
noun( 'singleness', '-', mass, _ ).
noun( 'singlestick', 'singlesticks', both, _ ).
noun( 'singlet', 'singlets', count, _ ).
noun( 'singleton', 'singletons', count, _ ).
noun( 'singsong', 'singsongs', count, _ ).
noun( 'singular', 'singulars', count, _ ).
noun( 'singularity', 'singularities', both, _ ).
noun( 'sink', 'sinks', count, _ ).
noun( 'sinker', 'sinkers', count, _ ).
noun( 'sinking', 'sinkings', count, _ ).
noun( 'sinking-fund', 'sinking-funds', count, _ ).
noun( 'sinlessness', '-', mass, _ ).
noun( 'sinner', 'sinners', count, _ ).
noun( 'sinuosity', 'sinuosities', both, _ ).
noun( 'sinus', 'sinuses', count, _ ).
noun( 'sinusitis', '-', mass, _ ).
noun( 'sip', 'sips', count, _ ).
noun( 'siphon', 'siphons', count, _ ).
noun( 'sir', 'sirs', count, _ ).
noun( 'sirdar', 'sirdars', count, _ ).
noun( 'sire', 'sires', count, _ ).
noun( 'siren', 'sirens', count, _ ).
noun( 'sirloin', 'sirloins', both, _ ).
noun( 'sirocco', 'siroccos', count, _ ).
noun( 'sirrah', 'sirrahs', count, _ ).
noun( 'sirup', 'sirups', both, _ ).
noun( 'sisal', '-', mass, _ ).
noun( 'sissy', 'sissies', count, _ ).
noun( 'sister', 'sisters', count, _ ).
noun( 'sister-in-law', 'sisters-in-law', count, _ ).
noun( 'sisterhood', '-', mass, _ ).
noun( 'sit-in', 'sit-ins', count, _ ).
noun( 'sitar', 'sitars', count, _ ).
noun( 'site', 'sites', count, _ ).
noun( 'sitter', 'sitters', count, _ ).
noun( 'sitting', 'sittings', count, _ ).
noun( 'sitting-room', 'sitting-rooms', count, _ ).
noun( 'situation', 'situations', count, _ ).
noun( 'six', 'sixes', count, _ ).
noun( 'six-footer', 'six-footers', count, _ ).
noun( 'six-shooter', 'six-shooters', count, _ ).
noun( 'sixpence', 'sixpences', count, _ ).
noun( 'sixteen', 'sixteens', count, _ ).
noun( 'sixteenth', 'sixteenths', count, _ ).
noun( 'sixth', 'sixths', count, _ ).
noun( 'sixth-former', 'sixth-formers', count, _ ).
noun( 'sixtieth', 'sixtieths', count, _ ).
noun( 'sixty', 'sixties', count, _ ).
noun( 'size', 'sizes', both, _ ).
noun( 'skate', 'skates', count, _ ).
noun( 'skateboard', 'skateboards', count, _ ).
noun( 'skateboarder', 'skateboarders', count, _ ).
noun( 'skateboarding', '-', mass, _ ).
noun( 'skater', 'skaters', count, _ ).
noun( 'skating', '-', mass, _ ).
noun( 'skating-rink', 'skating-rinks', count, _ ).
noun( 'skeet', 'skeets', count, _ ).
noun( 'skein', 'skeins', count, _ ).
noun( 'skeleton', 'skeletons', count, _ ).
noun( 'skep', 'skeps', count, _ ).
noun( 'skepticism', '-', mass, _ ).
noun( 'sketch', 'sketches', count, _ ).
noun( 'sketch-block', 'sketch-blocks', count, _ ).
noun( 'sketch-book', 'sketch-books', count, _ ).
noun( 'sketch-map', 'sketch-maps', count, _ ).
noun( 'sketcher', 'sketchers', count, _ ).
noun( 'sketchiness', '-', mass, _ ).
noun( 'skewer', 'skewers', count, _ ).
noun( 'ski', 'skis', count, _ ).
noun( 'ski-bob', 'ski-bobs', count, _ ).
noun( 'ski-jump', 'ski-jumps', count, _ ).
noun( 'ski-lift', 'ski-lifts', count, _ ).
noun( 'ski-plane', 'ski-planes', count, _ ).
noun( 'skid', 'skids', count, _ ).
noun( 'skidpan', 'skidpans', count, _ ).
noun( 'skier', 'skiers', count, _ ).
noun( 'skiff', 'skiffs', count, _ ).
noun( 'skiffle', '-', mass, _ ).
noun( 'skiffle-group', 'skiffle-groups', count, _ ).
noun( 'skill', 'skills', both, _ ).
noun( 'skillet', 'skillets', count, _ ).
noun( 'skilly', '-', mass, _ ).
noun( 'skimmed-milk', '-', mass, _ ).
noun( 'skimmer', 'skimmers', count, _ ).
noun( 'skin', 'skins', both, _ ).
noun( 'skin-diving', '-', mass, _ ).
noun( 'skin-graft', 'skin-grafts', count, _ ).
noun( 'skinflint', 'skinflints', count, _ ).
noun( 'skinhead', 'skinheads', count, _ ).
noun( 'skip', 'skips', count, _ ).
noun( 'skipper', 'skippers', count, _ ).
noun( 'skipping-rope', 'skipping-ropes', count, _ ).
noun( 'skirl', 'skirls', count, _ ).
noun( 'skirmish', 'skirmishes', count, _ ).
noun( 'skirmisher', 'skirmishers', count, _ ).
noun( 'skirt', 'skirts', count, _ ).
noun( 'skirting-board', 'skirting-boards', count, _ ).
noun( 'skit', 'skits', count, _ ).
noun( 'skittishness', '-', mass, _ ).
noun( 'skittle', 'skittles', count, _ ).
noun( 'skittle-pin', 'skittle-pins', count, _ ).
noun( 'skittles', 'skittles', mass, _ ).
noun( 'skivvy', 'skivvies', count, _ ).
noun( 'skua', 'skuas', count, _ ).
noun( 'skulker', 'skulkers', count, _ ).
noun( 'skull', 'skulls', count, _ ).
noun( 'skullcap', 'skullcaps', count, _ ).
noun( 'skullduggery', '-', count, _ ).
noun( 'skunk', 'skunks', both, _ ).
noun( 'sky', 'skies', count, _ ).
noun( 'sky-blue', '-', mass, _ ).
noun( 'skylark', 'skylarks', count, _ ).
noun( 'skylight', 'skylights', count, _ ).
noun( 'skyline', 'skylines', count, _ ).
noun( 'skyscraper', 'skyscrapers', count, _ ).
noun( 'skywriting', '-', mass, _ ).
noun( 'slab', 'slabs', count, _ ).
noun( 'slack', 'slacks', both, _ ).
noun( 'slacker', 'slackers', count, _ ).
noun( 'slackness', '-', mass, _ ).
noun( 'slag', '-', mass, _ ).
noun( 'slag-heap', 'slag-heaps', count, _ ).
noun( 'slalom', 'slaloms', count, _ ).
noun( 'slam', 'slams', count, _ ).
noun( 'slander', 'slanders', both, _ ).
noun( 'slanderer', 'slanderers', count, _ ).
noun( 'slang', '-', mass, _ ).
noun( 'slanginess', '-', mass, _ ).
noun( 'slant', 'slants', count, _ ).
noun( 'slap', 'slaps', count, _ ).
noun( 'slapstick', '-', mass, _ ).
noun( 'slash', 'slashes', count, _ ).
noun( 'slat', 'slats', count, _ ).
noun( 'slate', 'slates', both, _ ).
noun( 'slate-club', 'slate-clubs', count, _ ).
noun( 'slate-pencil', 'slate-pencils', count, _ ).
noun( 'slating', 'slatings', count, _ ).
noun( 'slattern', 'slatterns', count, _ ).
noun( 'slatternliness', '-', mass, _ ).
noun( 'slaughter', '-', mass, _ ).
noun( 'slaughterer', 'slaughterers', count, _ ).
noun( 'slaughterhouse', 'slaughterhouses', count, _ ).
noun( 'slave', 'slaves', count, _ ).
noun( 'slave-driver', 'slave-drivers', count, _ ).
noun( 'slave-trade', '-', mass, _ ).
noun( 'slave-traffic', '-', mass, _ ).
noun( 'slaver', '-', mass, _ ).
noun( 'slavery', '-', mass, _ ).
noun( 'slavey', 'slaveys', count, _ ).
noun( 'slaw', '-', mass, _ ).
noun( 'slayer', 'slayers', count, _ ).
noun( 'sled', 'sleds', count, _ ).
noun( 'sledge', 'sledges', count, _ ).
noun( 'sledgehammer', 'sledgehammers', count, _ ).
noun( 'sleekness', '-', mass, _ ).
noun( 'sleep', '-', mass, _ ).
noun( 'sleeper', 'sleepers', count, _ ).
noun( 'sleepiness', '-', mass, _ ).
noun( 'sleeping', '-', mass, _ ).
noun( 'sleeping-bag', 'sleeping-bags', count, _ ).
noun( 'sleeping-car', 'sleeping-cars', count, _ ).
noun( 'sleeping-draught', 'sleeping-draughts', count, _ ).
noun( 'sleeping-pill', 'sleeping-pills', count, _ ).
noun( 'sleeping-sickness', '-', mass, _ ).
noun( 'sleeplessness', '-', mass, _ ).
noun( 'sleepwalker', 'sleepwalkers', count, _ ).
noun( 'sleepy-head', 'sleepy-heads', count, _ ).
noun( 'sleet', '-', mass, _ ).
noun( 'sleeve', 'sleeves', count, _ ).
noun( 'sleigh', 'sleighs', count, _ ).
noun( 'sleigh-bell', 'sleigh-bells', count, _ ).
noun( 'sleight', 'sleights', count, _ ).
noun( 'slenderness', '-', mass, _ ).
noun( 'sleuth', 'sleuths', count, _ ).
noun( 'sleuth-hound', 'sleuth-hounds', count, _ ).
noun( 'slice', 'slices', count, _ ).
noun( 'slick', 'slicks', count, _ ).
noun( 'slicker', 'slickers', count, _ ).
noun( 'slide', 'slides', count, _ ).
noun( 'slide-rule', 'slide-rules', count, _ ).
noun( 'slight', 'slights', count, _ ).
noun( 'slightness', '-', mass, _ ).
noun( 'slime', '-', mass, _ ).
noun( 'sliminess', '-', mass, _ ).
noun( 'slimness', '-', mass, _ ).
noun( 'sling', 'slings', count, _ ).
noun( 'slinger', 'slingers', count, _ ).
noun( 'slip', 'slips', both, _ ).
noun( 'slip-carriage', 'slip-carriages', count, _ ).
noun( 'slip-coach', 'slip-coaches', count, _ ).
noun( 'slip-road', 'slip-roads', count, _ ).
noun( 'slip-up', 'slip-ups', count, _ ).
noun( 'slipcover', 'slipcovers', count, _ ).
noun( 'slipknot', 'slipknots', count, _ ).
noun( 'slipon', 'slipons', count, _ ).
noun( 'slipover', 'slipovers', count, _ ).
noun( 'slipper', 'slippers', count, _ ).
noun( 'slipperiness', '-', mass, _ ).
noun( 'slipstream', 'slipstreams', count, _ ).
noun( 'slipway', 'slipways', count, _ ).
noun( 'slit', 'slits', count, _ ).
noun( 'sliver', 'slivers', count, _ ).
noun( 'slob', 'slobs', count, _ ).
noun( 'slobber', '-', mass, _ ).
noun( 'sloe', 'sloes', count, _ ).
noun( 'sloe-gin', '-', mass, _ ).
noun( 'slogan', 'slogans', count, _ ).
noun( 'slogger', 'sloggers', count, _ ).
noun( 'sloop', 'sloops', count, _ ).
noun( 'slop', 'slops', count, _ ).
noun( 'slop-basin', 'slop-basins', count, _ ).
noun( 'slop-pail', 'slop-pails', count, _ ).
noun( 'slop-shop', 'slop-shops', count, _ ).
noun( 'slope', 'slopes', both, _ ).
noun( 'sloppiness', '-', mass, _ ).
noun( 'slot', 'slots', count, _ ).
noun( 'slot-machine', 'slot-machines', count, _ ).
noun( 'sloth', 'sloths', both, _ ).
noun( 'slouch', 'slouches', count, _ ).
noun( 'slouch-hat', 'slouch-hats', count, _ ).
noun( 'slough', 'sloughs', count, _ ).
noun( 'slough', 'sloughs', count, _ ).
noun( 'sloven', 'slovens', count, _ ).
noun( 'slovenliness', '-', mass, _ ).
noun( 'slow-worm', 'slow-worms', count, _ ).
noun( 'slowcoach', 'slowcoaches', count, _ ).
noun( 'slowdown', 'slowdowns', count, _ ).
noun( 'slowness', '-', mass, _ ).
noun( 'sludge', '-', mass, _ ).
noun( 'slug', 'slugs', count, _ ).
noun( 'sluggard', 'sluggards', count, _ ).
noun( 'sluggishness', '-', mass, _ ).
noun( 'sluice', 'sluices', count, _ ).
noun( 'sluice-valve', 'sluice-valves', count, _ ).
noun( 'sluicegate', 'sluicegates', count, _ ).
noun( 'slum', 'slums', count, _ ).
noun( 'slumber', 'slumbers', count, _ ).
noun( 'slumberer', 'slumberers', count, _ ).
noun( 'slump', 'slumps', count, _ ).
noun( 'slur', 'slurs', both, _ ).
noun( 'slurry', '-', mass, _ ).
noun( 'slush', '-', mass, _ ).
noun( 'slut', 'sluts', count, _ ).
noun( 'slyness', '-', mass, _ ).
noun( 'smack', 'smacks', count, _ ).
noun( 'smacker', 'smackers', count, _ ).
noun( 'smacking', 'smackings', both, _ ).
noun( 'small', 'smalls', count, _ ).
noun( 'smallholder', 'smallholders', count, _ ).
noun( 'smallholding', 'smallholdings', count, _ ).
noun( 'smallness', '-', mass, _ ).
noun( 'smallpox', '-', mass, _ ).
noun( 'smart', '-', mass, _ ).
noun( 'smartness', '-', mass, _ ).
noun( 'smash', 'smashes', count, _ ).
noun( 'smash-up', 'smash-ups', count, _ ).
noun( 'smasher', 'smashers', count, _ ).
noun( 'smattering', 'smatterings', count, _ ).
noun( 'smear', 'smears', count, _ ).
noun( 'smear-word', 'smear-words', count, _ ).
noun( 'smell', 'smells', both, _ ).
noun( 'smelling-bottle', 'smelling-bottles', count, _ ).
noun( 'smelt', 'smelts', count, _ ).
noun( 'smilax', '-', mass, _ ).
noun( 'smile', 'smiles', count, _ ).
noun( 'smirch', 'smirches', count, _ ).
noun( 'smirk', 'smirks', count, _ ).
noun( 'smith', 'smiths', count, _ ).
noun( 'smithy', 'smithies', count, _ ).
noun( 'smock', 'smocks', count, _ ).
noun( 'smocking', '-', mass, _ ).
noun( 'smog', 'smogs', both, _ ).
noun( 'smoke', 'smokes', both, _ ).
noun( 'smoke-bomb', 'smoke-bombs', count, _ ).
noun( 'smoke-screen', 'smoke-screens', count, _ ).
noun( 'smoker', 'smokers', count, _ ).
noun( 'smokestack', 'smokestacks', count, _ ).
noun( 'smoking', '-', mass, _ ).
noun( 'smoking-car', 'smoking-cars', count, _ ).
noun( 'smoking-carriage', 'smoking-carriages', count, _ ).
noun( 'smoking-compartment', 'smoking-compartments', count, _ ).
noun( 'smoking-mixture', '-', mass, _ ).
noun( 'smoking-room', 'smoking-rooms', count, _ ).
noun( 'smooth', '-', count, _ ).
noun( 'smoothing-iron', 'smoothing-irons', count, _ ).
noun( 'smoothing-plane', 'smoothing-planes', count, _ ).
noun( 'smoothness', '-', mass, _ ).
noun( 'smorgasbord', '-', mass, _ ).
noun( 'smother', '-', count, _ ).
noun( 'smoulder', '-', mass, _ ).
noun( 'smudge', 'smudges', count, _ ).
noun( 'smuggler', 'smugglers', count, _ ).
noun( 'smugness', '-', mass, _ ).
noun( 'smut', 'smuts', both, _ ).
noun( 'smuttiness', '-', mass, _ ).
noun( 'snack', 'snacks', count, _ ).
noun( 'snack-bar', 'snack-bars', count, _ ).
noun( 'snack-counter', 'snack-counters', count, _ ).
noun( 'snaffle', 'snaffles', count, _ ).
noun( 'snaffle-bit', 'snaffle-bits', count, _ ).
noun( 'snag', 'snags', count, _ ).
noun( 'snail', 'snails', count, _ ).
noun( 'snake', 'snakes', count, _ ).
noun( 'snake-charmer', 'snake-charmers', count, _ ).
noun( 'snap', 'snaps', both, _ ).
noun( 'snap-fastener', 'snap-fasteners', count, _ ).
noun( 'snapdragon', 'snapdragons', count, _ ).
noun( 'snappishness', '-', mass, _ ).
noun( 'snapshot', 'snapshots', count, _ ).
noun( 'snare', 'snares', count, _ ).
noun( 'snare-drum', 'snare-drums', count, _ ).
noun( 'snarl', 'snarls', count, _ ).
noun( 'snarl-up', 'snarl-ups', count, _ ).
noun( 'snatch', 'snatches', count, _ ).
noun( 'snatcher', 'snatchers', count, _ ).
noun( 'sneak', 'sneaks', count, _ ).
noun( 'sneak-thief', 'sneak-thieves', count, _ ).
noun( 'sneer', 'sneers', count, _ ).
noun( 'sneeze', 'sneezes', count, _ ).
noun( 'snick', 'snicks', count, _ ).
noun( 'snicker', 'snickers', count, _ ).
noun( 'sniff', 'sniffs', count, _ ).
noun( 'snifter', 'snifters', count, _ ).
noun( 'snigger', 'sniggers', count, _ ).
noun( 'snip', 'snips', count, _ ).
noun( 'snipe', 'snipe', count, _ ).
noun( 'sniper', 'snipers', count, _ ).
noun( 'snippet', 'snippets', count, _ ).
noun( 'snipping', 'snippings', count, _ ).
noun( 'sniveller', 'snivellers', count, _ ).
noun( 'snob', 'snobs', count, _ ).
noun( 'snobbery', '-', mass, _ ).
noun( 'snobbishness', '-', mass, _ ).
noun( 'snogging', '-', mass, _ ).
noun( 'snood', 'snoods', count, _ ).
noun( 'snook', 'snooks', count, _ ).
noun( 'snooker', 'snookers', both, _ ).
noun( 'snooper', 'snoopers', count, _ ).
noun( 'snooze', 'snoozes', count, _ ).
noun( 'snore', 'snores', count, _ ).
noun( 'snorer', 'snorers', count, _ ).
noun( 'snorkel', 'snorkels', count, _ ).
noun( 'snort', 'snorts', count, _ ).
noun( 'snorter', 'snorters', count, _ ).
noun( 'snot', '-', mass, _ ).
noun( 'snout', 'snouts', count, _ ).
noun( 'snow', 'snows', both, _ ).
noun( 'snow-line', 'snow-lines', count, _ ).
noun( 'snowball', 'snowballs', count, _ ).
noun( 'snowberry', 'snowberries', count, _ ).
noun( 'snowblindness', '-', mass, _ ).
noun( 'snowdrift', 'snowdrifts', count, _ ).
noun( 'snowdrop', 'snowdrops', count, _ ).
noun( 'snowfall', 'snowfalls', count, _ ).
noun( 'snowfield', 'snowfields', count, _ ).
noun( 'snowflake', 'snowflakes', count, _ ).
noun( 'snowman', 'snowmen', count, _ ).
noun( 'snowplough', 'snowploughs', count, _ ).
noun( 'snowstorm', 'snowstorms', count, _ ).
noun( 'snub', 'snubs', count, _ ).
noun( 'snuff', 'snuffs', both, _ ).
noun( 'snuff-colour', '-', mass, _ ).
noun( 'snuffbox', 'snuffboxes', count, _ ).
noun( 'snuffle', 'snuffles', count, _ ).
noun( 'snug', 'snugs', count, _ ).
noun( 'snuggery', 'snuggeries', count, _ ).
noun( 'snugness', '-', mass, _ ).
noun( 'so', '-', count, _ ).
noun( 'so-and-so', 'so-and-sos', count, _ ).
noun( 'soak', 'soaks', count, _ ).
noun( 'soaker', 'soakers', count, _ ).
noun( 'soap', 'soaps', both, _ ).
noun( 'soap-bubble', 'soap-bubbles', count, _ ).
noun( 'soap-opera', 'soap-operas', both, _ ).
noun( 'soapbox', 'soapboxes', count, _ ).
noun( 'sob', 'sobs', count, _ ).
noun( 'sob-stuff', '-', mass, _ ).
noun( 'sober-sides', '-', count, _ ).
noun( 'sobriety', '-', mass, _ ).
noun( 'sobriquet', 'sobriquets', count, _ ).
noun( 'soccer', '-', mass, _ ).
noun( 'sociability', '-', mass, _ ).
noun( 'social', 'socials', count, _ ).
noun( 'social-work', '-', mass, _ ).
noun( 'socialism', '-', mass, _ ).
noun( 'socialist', 'socialists', count, _ ).
noun( 'socialite', 'socialites', count, _ ).
noun( 'socialization', 'socializations', both, _ ).
noun( 'society', 'societies', both, _ ).
noun( 'sociologist', 'sociologists', count, _ ).
noun( 'sociology', '-', mass, _ ).
noun( 'sock', 'socks', count, _ ).
noun( 'socket', 'sockets', count, _ ).
noun( 'sod', 'sods', both, _ ).
noun( 'soda', '-', mass, _ ).
noun( 'soda-biscuit', 'soda-biscuits', count, _ ).
noun( 'soda-cracker', 'soda-crackers', count, _ ).
noun( 'soda-fountain', 'soda-fountains', count, _ ).
noun( 'soda-water', 'soda-waters', both, _ ).
noun( 'sodium', '-', mass, _ ).
noun( 'sodomite', 'sodomites', count, _ ).
noun( 'sodomy', '-', mass, _ ).
noun( 'sofa', 'sofas', count, _ ).
noun( 'soft-solder', '-', mass, _ ).
noun( 'softener', 'softeners', count, _ ).
noun( 'softie', 'softies', count, _ ).
noun( 'softness', '-', mass, _ ).
noun( 'software', '-', mass, _ ).
noun( 'softwood', 'softwoods', both, _ ).
noun( 'softy', 'softies', count, _ ).
noun( 'sogginess', '-', mass, _ ).
noun( 'soh', '-', count, _ ).
noun( 'soil', 'soils', both, _ ).
noun( 'soil-pipe', 'soil-pipes', count, _ ).
noun( 'soir_ee', 'soir_ees', count, _ ).
noun( 'sojourn', 'sojourns', count, _ ).
noun( 'sojourner', 'sojourners', count, _ ).
noun( 'sol-fa', '-', count, _ ).
noun( 'solace', 'solaces', both, _ ).
noun( 'solar plexus', '-', count, _ ).
noun( 'solarium', 'solaria', count, _ ).
noun( 'solder', '-', mass, _ ).
noun( 'soldering-iron', 'soldering-irons', count, _ ).
noun( 'soldier', 'soldiers', count, _ ).
noun( 'soldiery', '-', count, _ ).
noun( 'sole', 'soles', count, _ ).
noun( 'solecism', 'solecisms', count, _ ).
noun( 'solemnity', 'solemnities', both, _ ).
noun( 'solemnization', '-', mass, _ ).
noun( 'solemnness', '-', mass, _ ).
noun( 'solicitation', 'solicitations', both, _ ).
noun( 'solicitor', 'solicitors', count, _ ).
noun( 'solicitude', '-', mass, _ ).
noun( 'solid', 'solids', count, _ ).
noun( 'solidarity', '-', mass, _ ).
noun( 'solidification', 'solidifications', both, _ ).
noun( 'solidity', '-', mass, _ ).
noun( 'solidness', '-', mass, _ ).
noun( 'soliloquy', 'soliloquies', both, _ ).
noun( 'solipsism', '-', mass, _ ).
noun( 'solitaire', 'solitaires', both, _ ).
noun( 'solitude', 'solitudes', both, _ ).
noun( 'solo', 'solos', both, _ ).
noun( 'soloist', 'soloists', count, _ ).
noun( 'solstice', 'solstices', count, _ ).
noun( 'solubility', '-', mass, _ ).
noun( 'solution', 'solutions', both, _ ).
noun( 'solvency', '-', mass, _ ).
noun( 'solvent', 'solvents', count, _ ).
noun( 'sombreness', '-', mass, _ ).
noun( 'sombrero', 'sombreros', count, _ ).
noun( 'someone', '-', count, _ ).
noun( 'somersault', 'somersaults', count, _ ).
noun( 'somnambulism', '-', mass, _ ).
noun( 'somnambulist', 'somnambulists', count, _ ).
noun( 'somnolence', '-', mass, _ ).
noun( 'son', 'sons', count, _ ).
noun( 'son-in-law', 'sons-in-law', count, _ ).
noun( 'sonar', 'sonars', count, _ ).
noun( 'sonata', 'sonatas', count, _ ).
noun( 'song', 'songs', both, _ ).
noun( 'songbird', 'songbirds', count, _ ).
noun( 'songbook', 'songbooks', count, _ ).
noun( 'songster', 'songsters', count, _ ).
noun( 'songstress', 'songstresses', count, _ ).
noun( 'sonnet', 'sonnets', count, _ ).
noun( 'sonneteer', 'sonneteers', count, _ ).
noun( 'sonny', 'sonnies', count, _ ).
noun( 'sonority', 'sonorities', both, _ ).
noun( 'soot', '-', mass, _ ).
noun( 'sooth', 'sooths', count, _ ).
noun( 'soothsayer', 'soothsayers', count, _ ).
noun( 'sop', 'sops', count, _ ).
noun( 'sophism', 'sophisms', both, _ ).
noun( 'sophist', 'sophists', count, _ ).
noun( 'sophistication', '-', mass, _ ).
noun( 'sophistry', 'sophistries', both, _ ).
noun( 'sophomore', 'sophomores', count, _ ).
noun( 'soporific', 'soporifics', count, _ ).
noun( 'soprano', 'sopranos', count, _ ).
noun( 'sorbet', 'sorbets', count, _ ).
noun( 'sorcerer', 'sorcerers', count, _ ).
noun( 'sorceress', 'sorceresses', count, _ ).
noun( 'sorcery', 'sorceries', both, _ ).
noun( 'sordidness', '-', mass, _ ).
noun( 'sore', 'sores', count, _ ).
noun( 'soreness', '-', mass, _ ).
noun( 'sorghum', '-', mass, _ ).
noun( 'sorority', 'sororities', count, _ ).
noun( 'sorrel', 'sorrels', both, _ ).
noun( 'sorrow', 'sorrows', both, _ ).
noun( 'sort', 'sorts', count, _ ).
noun( 'sorter', 'sorters', count, _ ).
noun( 'sortie', 'sorties', count, _ ).
noun( 'sot', 'sots', count, _ ).
noun( 'sottishness', '-', mass, _ ).
noun( 'sou', 'sous', count, _ ).
noun( 'sou\'-east', '-', mass, _ ).
noun( 'sou\'-sou\'-east', '-', mass, _ ).
noun( 'sou\'-sou\'-west', '-', mass, _ ).
noun( 'sou\'-west', '-', mass, _ ).
noun( 'sou\'-wester', 'sou\'-westers', count, _ ).
noun( 'soubrette', 'soubrettes', count, _ ).
noun( 'soubriquet', 'soubriquets', count, _ ).
noun( 'souffl_e', 'souffl_es', count, _ ).
noun( 'soul', 'souls', count, _ ).
noun( 'sound', 'sounds', both, _ ).
noun( 'sound-film', 'sound-films', count, _ ).
noun( 'sound-recording', 'sound-recordings', both, _ ).
noun( 'sound-wave', 'sound-waves', count, _ ).
noun( 'soundbox', 'soundboxes', count, _ ).
noun( 'sounding-board', 'sounding-boards', count, _ ).
noun( 'soundness', '-', mass, _ ).
noun( 'soundtrack', 'soundtracks', count, _ ).
noun( 'soup', 'soups', both, _ ).
noun( 'soup-kitchen', 'soup-kitchens', count, _ ).
noun( 'soup<con', 'soup<cons', count, _ ).
noun( 'source', 'sources', count, _ ).
noun( 'sourness', '-', mass, _ ).
noun( 'soutane', 'soutanes', count, _ ).
noun( 'south', '-', mass, _ ).
noun( 'south-southeast', '-', mass, _ ).
noun( 'south-southwest', '-', mass, _ ).
noun( 'southeast', '-', mass, _ ).
noun( 'southeaster', 'southeasters', count, _ ).
noun( 'southerner', 'southerners', count, _ ).
noun( 'southpaw', 'southpaws', count, _ ).
noun( 'southwest', '-', mass, _ ).
noun( 'southwester', 'southwesters', count, _ ).
noun( 'souvenir', 'souvenirs', count, _ ).
noun( 'sovereign', 'sovereigns', count, _ ).
noun( 'sovereignty', '-', mass, _ ).
noun( 'soviet', 'soviets', count, _ ).
noun( 'sow', 'sows', count, _ ).
noun( 'sower', 'sowers', count, _ ).
noun( 'soy', '-', mass, _ ).
noun( 'soya', '-', mass, _ ).
noun( 'spa', 'spas', count, _ ).
noun( 'space', 'spaces', both, _ ).
noun( 'space-bar', 'space-bars', count, _ ).
noun( 'space-capsule', 'space-capsules', count, _ ).
noun( 'space-heater', 'space-heaters', count, _ ).
noun( 'space-helmet', 'space-helmets', count, _ ).
noun( 'space-rocket', 'space-rockets', count, _ ).
noun( 'space-time', '-', mass, _ ).
noun( 'space-vehicle', 'space-vehicles', count, _ ).
noun( 'spacecraft', 'spacecraft', count, _ ).
noun( 'spaceship', 'spaceships', count, _ ).
noun( 'spacesuit', 'spacesuits', count, _ ).
noun( 'spacing', 'spacings', both, _ ).
noun( 'spaciousness', '-', mass, _ ).
noun( 'spade', 'spades', count, _ ).
noun( 'spadeful', 'spadefuls', count, _ ).
noun( 'spadework', '-', mass, _ ).
noun( 'spaghetti', '-', mass, _ ).
noun( 'spam', '-', mass, _ ).
noun( 'span', 'spans', count, _ ).
noun( 'spangle', 'spangles', count, _ ).
noun( 'spaniel', 'spaniels', count, _ ).
noun( 'spanking', 'spankings', count, _ ).
noun( 'spanner', 'spanners', count, _ ).
noun( 'spar', 'spars', count, _ ).
noun( 'spare', 'spares', count, _ ).
noun( 'spare-rib', 'spare-ribs', count, _ ).
noun( 'spareness', '-', mass, _ ).
noun( 'spark', 'sparks', count, _ ).
noun( 'spark-plug', 'spark-plugs', count, _ ).
noun( 'sparking-plug', 'sparking-plugs', count, _ ).
noun( 'sparkle', 'sparkles', count, _ ).
noun( 'sparkler', 'sparklers', count, _ ).
noun( 'sparring-match', 'sparring-matches', count, _ ).
noun( 'sparring-partner', 'sparring-partners', count, _ ).
noun( 'sparrow', 'sparrows', count, _ ).
noun( 'sparseness', '-', mass, _ ).
noun( 'sparsity', '-', mass, _ ).
noun( 'spasm', 'spasms', count, _ ).
noun( 'spastic', 'spastics', count, _ ).
noun( 'spat', 'spats', count, _ ).
noun( 'spatchcock', 'spatchcocks', count, _ ).
noun( 'spate', 'spates', count, _ ).
noun( 'spatter', 'spatters', count, _ ).
noun( 'spatula', 'spatulas', count, _ ).
noun( 'spavin', '-', mass, _ ).
noun( 'spawn', '-', mass, _ ).
noun( 'speaker', 'speakers', count, _ ).
noun( 'speakership', 'speakerships', count, _ ).
noun( 'speaking-trumpet', 'speaking-trumpets', count, _ ).
noun( 'speaking-tube', 'speaking-tubes', count, _ ).
noun( 'spear', 'spears', count, _ ).
noun( 'spearhead', 'spearheads', count, _ ).
noun( 'spearmint', '-', mass, _ ).
noun( 'spec', 'specs', count, _ ).
noun( 'special', 'specials', count, _ ).
noun( 'specialism', 'specialisms', count, _ ).
noun( 'specialist', 'specialists', count, _ ).
noun( 'speciality', 'specialities', count, _ ).
noun( 'specialization', 'specializations', both, _ ).
noun( 'specialty', 'specialties', count, _ ).
noun( 'specie', '-', mass, _ ).
noun( 'species', 'species', count, _ ).
noun( 'specific', 'specifics', count, _ ).
noun( 'specification', 'specifications', both, _ ).
noun( 'specificity', '-', mass, _ ).
noun( 'specimen', 'specimens', count, _ ).
noun( 'speciousness', '-', mass, _ ).
noun( 'speck', 'specks', count, _ ).
noun( 'speckle', 'speckles', count, _ ).
noun( 'spectacle', 'spectacles', count, _ ).
noun( 'spectacular', 'spectaculars', count, _ ).
noun( 'spectator', 'spectators', count, _ ).
noun( 'spectre', 'spectres', count, _ ).
noun( 'spectroscope', 'spectroscopes', count, _ ).
noun( 'spectrum', 'spectra', count, _ ).
noun( 'speculation', 'speculations', both, _ ).
noun( 'speculator', 'speculators', count, _ ).
noun( 'speech', 'speeches', both, _ ).
noun( 'speech-day', 'speech-days', count, _ ).
noun( 'speed', 'speeds', both, _ ).
noun( 'speed-cop', 'speed-cops', count, _ ).
noun( 'speed-indicator', 'speed-indicators', count, _ ).
noun( 'speed-limit', 'speed-limits', count, _ ).
noun( 'speed-up', 'speed-ups', count, _ ).
noun( 'speedboat', 'speedboats', count, _ ).
noun( 'speeding', '-', mass, _ ).
noun( 'speedometer', 'speedometers', count, _ ).
noun( 'speedway', 'speedways', both, _ ).
noun( 'speedwell', 'speedwells', both, _ ).
noun( 'spelaeologist', 'spelaeologists', count, _ ).
noun( 'spelaeology', '-', mass, _ ).
noun( 'speleologist', 'speleologists', count, _ ).
noun( 'speleology', '-', mass, _ ).
noun( 'spell', 'spells', count, _ ).
noun( 'spellbinder', 'spellbinders', count, _ ).
noun( 'speller', 'spellers', count, _ ).
noun( 'spelling', 'spellings', both, _ ).
noun( 'spelt', '-', mass, _ ).
noun( 'spender', 'spenders', count, _ ).
noun( 'spendthrift', 'spendthrifts', count, _ ).
noun( 'sperm', 'sperms', both, _ ).
noun( 'sperm-whale', 'sperm-whales', count, _ ).
noun( 'spermaceti', '-', mass, _ ).
noun( 'spermatozoon', 'spermatozoa', count, _ ).
noun( 'spermicide', 'spermicides', both, _ ).
noun( 'sphagnum', 'sphagnums', both, _ ).
noun( 'sphere', 'spheres', count, _ ).
noun( 'spheroid', 'spheroids', count, _ ).
noun( 'sphinx', 'sphinxes', count, _ ).
noun( 'spice', 'spices', both, _ ).
noun( 'spiciness', '-', mass, _ ).
noun( 'spider', 'spiders', count, _ ).
noun( 'spiel', 'spiels', count, _ ).
noun( 'spigot', 'spigots', count, _ ).
noun( 'spike', 'spikes', count, _ ).
noun( 'spikenard', '-', mass, _ ).
noun( 'spill', 'spills', count, _ ).
noun( 'spillage', 'spillages', count, _ ).
noun( 'spillover', 'spillovers', count, _ ).
noun( 'spillway', 'spillways', count, _ ).
noun( 'spin', 'spins', both, _ ).
noun( 'spin-drier', 'spin-driers', count, _ ).
noun( 'spin-off', 'spin-offs', count, _ ).
noun( 'spinach', '-', mass, _ ).
noun( 'spindle', 'spindles', count, _ ).
noun( 'spindle-berry', 'spindle-berries', count, _ ).
noun( 'spindle-shanks', '-', count, _ ).
noun( 'spindle-tree', 'spindle-trees', count, _ ).
noun( 'spindrift', '-', mass, _ ).
noun( 'spine', 'spines', count, _ ).
noun( 'spinet', 'spinets', count, _ ).
noun( 'spinnaker', 'spinnakers', count, _ ).
noun( 'spinney', 'spinneys', count, _ ).
noun( 'spinning-wheel', 'spinning-wheels', count, _ ).
noun( 'spinster', 'spinsters', count, _ ).
noun( 'spinsterhood', '-', mass, _ ).
noun( 'spiral', 'spirals', count, _ ).
noun( 'spire', 'spires', count, _ ).
noun( 'spirit', 'spirits', both, _ ).
noun( 'spirit-lamp', 'spirit-lamps', count, _ ).
noun( 'spirit-level', 'spirit-levels', count, _ ).
noun( 'spirit-rapper', 'spirit-rappers', count, _ ).
noun( 'spirit-stove', 'spirit-stoves', count, _ ).
noun( 'spiritual', 'spirituals', count, _ ).
noun( 'spiritualism', '-', mass, _ ).
noun( 'spiritualist', 'spiritualists', count, _ ).
noun( 'spirituality', '-', mass, _ ).
noun( 'spiritualization', '-', mass, _ ).
noun( 'spirt', 'spirts', count, _ ).
noun( 'spit', 'spits', both, _ ).
noun( 'spite', '-', mass, _ ).
noun( 'spitefulness', '-', mass, _ ).
noun( 'spitfire', 'spitfires', count, _ ).
noun( 'spitting', '-', mass, _ ).
noun( 'spittle', '-', mass, _ ).
noun( 'spittoon', 'spittoons', count, _ ).
noun( 'spiv', 'spivs', count, _ ).
noun( 'splash', 'splashes', count, _ ).
noun( 'splashdown', 'splashdowns', count, _ ).
noun( 'splay', 'splays', count, _ ).
noun( 'splayfoot', 'splayfeet', count, _ ).
noun( 'spleen', 'spleens', both, _ ).
noun( 'splendour', 'splendours', both, _ ).
noun( 'splice', 'splices', count, _ ).
noun( 'splicer', 'splicers', count, _ ).
noun( 'splint', 'splints', count, _ ).
noun( 'splinter', 'splinters', count, _ ).
noun( 'split', 'splits', count, _ ).
noun( 'splodge', 'splodges', count, _ ).
noun( 'splotch', 'splotches', count, _ ).
noun( 'splurge', 'splurges', count, _ ).
noun( 'splutter', '-', mass, _ ).
noun( 'spoil', 'spoils', both, _ ).
noun( 'spoilsport', 'spoilsports', count, _ ).
noun( 'spoke', 'spokes', count, _ ).
noun( 'spokesman', 'spokesmen', count, _ ).
noun( 'spoliation', '-', mass, _ ).
noun( 'spondee', 'spondees', count, _ ).
noun( 'sponge', 'sponges', count, _ ).
noun( 'sponge-cake', 'sponge-cakes', both, _ ).
noun( 'sponger', 'spongers', count, _ ).
noun( 'sponginess', '-', mass, _ ).
noun( 'sponsor', 'sponsors', count, _ ).
noun( 'sponsorship', 'sponsorships', both, _ ).
noun( 'spontaneity', '-', mass, _ ).
noun( 'spontaneousness', '-', mass, _ ).
noun( 'spoof', 'spoofs', count, _ ).
noun( 'spook', 'spooks', count, _ ).
noun( 'spool', 'spools', count, _ ).
noun( 'spoon', 'spoons', count, _ ).
noun( 'spoonerism', 'spoonerisms', count, _ ).
noun( 'spoonfeeding', '-', mass, _ ).
noun( 'spoonful', 'spoonfuls', count, _ ).
noun( 'spoor', 'spoors', count, _ ).
noun( 'spore', 'spores', count, _ ).
noun( 'sporran', 'sporrans', count, _ ).
noun( 'sport', 'sports', both, _ ).
noun( 'sportiveness', '-', mass, _ ).
noun( 'sports-car', 'sports-cars', count, _ ).
noun( 'sports-coat', 'sports-coats', count, _ ).
noun( 'sports-editor', 'sports-editors', count, _ ).
noun( 'sports-jacket', 'sports-jackets', count, _ ).
noun( 'sportsman', 'sportsmen', count, _ ).
noun( 'sportsmanship', '-', mass, _ ).
noun( 'spot', 'spots', count, _ ).
noun( 'spotlight', 'spotlights', count, _ ).
noun( 'spotter', 'spotters', count, _ ).
noun( 'spouse', 'spouses', count, _ ).
noun( 'spout', 'spouts', count, _ ).
noun( 'sprain', 'sprains', count, _ ).
noun( 'sprat', 'sprats', count, _ ).
noun( 'sprawl', 'sprawls', both, _ ).
noun( 'spray', 'sprays', both, _ ).
noun( 'spray-gun', 'spray-guns', count, _ ).
noun( 'sprayer', 'sprayers', count, _ ).
noun( 'spread', 'spreads', count, _ ).
noun( 'spread-over', '-', mass, _ ).
noun( 'spreadeagle', 'spreadeagles', count, _ ).
noun( 'spreader', 'spreaders', count, _ ).
noun( 'spree', 'sprees', count, _ ).
noun( 'sprig', 'sprigs', count, _ ).
noun( 'sprightliness', '-', mass, _ ).
noun( 'spring', 'springs', both, _ ).
noun( 'spring-balance', 'spring-balances', count, _ ).
noun( 'spring-clean', 'spring-cleans', count, _ ).
noun( 'spring-cleaning', '-', mass, _ ).
noun( 'spring-gun', 'spring-guns', count, _ ).
noun( 'spring-mattress', 'spring-mattresses', count, _ ).
noun( 'springboard', 'springboards', count, _ ).
noun( 'springbok', 'springboks', count, _ ).
noun( 'springtide', 'springtides', count, _ ).
noun( 'springtime', 'springtimes', both, _ ).
noun( 'sprinkler', 'sprinklers', count, _ ).
noun( 'sprinkling', 'sprinklings', count, _ ).
noun( 'sprint', 'sprints', count, _ ).
noun( 'sprinter', 'sprinters', count, _ ).
noun( 'sprit', 'sprits', count, _ ).
noun( 'sprite', 'sprites', count, _ ).
noun( 'spritsail', 'spritsails', count, _ ).
noun( 'sprocket', 'sprockets', count, _ ).
noun( 'sprocket-wheel', 'sprocket-wheels', count, _ ).
noun( 'sprout', 'sprouts', count, _ ).
noun( 'spruce', 'spruces', both, _ ).
noun( 'spruceness', '-', mass, _ ).
noun( 'spud', 'spuds', count, _ ).
noun( 'spume', '-', mass, _ ).
noun( 'spunk', '-', mass, _ ).
noun( 'spur', 'spurs', count, _ ).
noun( 'spuriousness', '-', mass, _ ).
noun( 'spurt', 'spurts', count, _ ).
noun( 'sputnik', 'sputniks', count, _ ).
noun( 'sputum', '-', mass, _ ).
noun( 'spy', 'spies', count, _ ).
noun( 'spy-hole', 'spy-holes', count, _ ).
noun( 'spyglass', 'spyglasses', count, _ ).
noun( 'squab', 'squabs', count, _ ).
noun( 'squabble', 'squabbles', count, _ ).
noun( 'squad', 'squads', count, _ ).
noun( 'squadron', 'squadrons', count, _ ).
noun( 'squall', 'squalls', count, _ ).
noun( 'squalor', '-', mass, _ ).
noun( 'squandermania', '-', mass, _ ).
noun( 'square', 'squares', count, _ ).
noun( 'square-bashing', '-', mass, _ ).
noun( 'square-toes', '-', count, _ ).
noun( 'squareness', '-', mass, _ ).
noun( 'squash', 'squash', both, _ ).
noun( 'squatter', 'squatters', count, _ ).
noun( 'squaw', 'squaws', count, _ ).
noun( 'squawk', 'squawks', count, _ ).
noun( 'squawker', 'squawkers', count, _ ).
noun( 'squeak', 'squeaks', count, _ ).
noun( 'squeaker', 'squeakers', count, _ ).
noun( 'squeal', 'squeals', count, _ ).
noun( 'squealer', 'squealers', count, _ ).
noun( 'squeamishness', '-', mass, _ ).
noun( 'squeegee', 'squeegees', count, _ ).
noun( 'squeeze', 'squeezes', both, _ ).
noun( 'squeezer', 'squeezers', count, _ ).
noun( 'squelch', 'squelches', count, _ ).
noun( 'squib', 'squibs', count, _ ).
noun( 'squid', 'squids', count, _ ).
noun( 'squiggle', 'squiggles', count, _ ).
noun( 'squint', 'squints', count, _ ).
noun( 'squire', 'squires', count, _ ).
noun( 'squirearchy', 'squirearchies', count, _ ).
noun( 'squirm', 'squirms', count, _ ).
noun( 'squirrel', 'squirrels', count, _ ).
noun( 'squirt', 'squirts', count, _ ).
noun( 'stab', 'stabs', count, _ ).
noun( 'stabber', 'stabbers', count, _ ).
noun( 'stability', '-', mass, _ ).
noun( 'stabilization', 'stabilizations', both, _ ).
noun( 'stabilizer', 'stabilizers', count, _ ).
noun( 'stable', 'stables', count, _ ).
noun( 'stable-companion', 'stable-companions', count, _ ).
noun( 'stableboy', 'stableboys', count, _ ).
noun( 'stableman', 'stablemen', count, _ ).
noun( 'stablemate', 'stablemates', count, _ ).
noun( 'stabling', '-', mass, _ ).
noun( 'stack', 'stacks', count, _ ).
noun( 'stadium', 'stadiums', count, _ ).
noun( 'staff', 'staffs', count, _ ).
noun( 'staff-office', 'staff-offices', count, _ ).
noun( 'stag', 'stags', count, _ ).
noun( 'stag-party', 'stag-parties', count, _ ).
noun( 'stage', 'stages', count, _ ).
noun( 'stage-whisper', 'stage-whispers', count, _ ).
noun( 'stagecoach', 'stagecoaches', count, _ ).
noun( 'stagecraft', 'stagecrafts', both, _ ).
noun( 'stager', 'stagers', count, _ ).
noun( 'stagflation', '-', mass, _ ).
noun( 'stagger', 'staggers', count, _ ).
noun( 'staggerer', 'staggerers', count, _ ).
noun( 'staginess', '-', mass, _ ).
noun( 'staging', 'stagings', both, _ ).
noun( 'stagnancy', '-', mass, _ ).
noun( 'stagnation', '-', mass, _ ).
noun( 'staidness', '-', mass, _ ).
noun( 'stain', 'stains', both, _ ).
noun( 'stair', 'stairs', count, _ ).
noun( 'stair-carpet', 'stair-carpets', count, _ ).
noun( 'stair-rod', 'stair-rods', count, _ ).
noun( 'staircase', 'staircases', count, _ ).
noun( 'stairway', 'stairways', count, _ ).
noun( 'stake', 'stakes', count, _ ).
noun( 'stake-holder', 'stake-holders', count, _ ).
noun( 'stalactite', 'stalactites', count, _ ).
noun( 'stalagmite', 'stalagmites', count, _ ).
noun( 'stalemate', 'stalemates', both, _ ).
noun( 'staleness', '-', mass, _ ).
noun( 'stalk', 'stalks', count, _ ).
noun( 'stalker', 'stalkers', count, _ ).
noun( 'stalking-horse', 'stalking-horses', count, _ ).
noun( 'stall', 'stalls', count, _ ).
noun( 'stallion', 'stallions', count, _ ).
noun( 'stalwart', 'stalwarts', count, _ ).
noun( 'stamen', 'stamens', count, _ ).
noun( 'stamina', '-', mass, _ ).
noun( 'stammer', 'stammers', count, _ ).
noun( 'stammerer', 'stammerers', count, _ ).
noun( 'stamp', 'stamps', count, _ ).
noun( 'stamp-album', 'stamp-albums', count, _ ).
noun( 'stamp-collector', 'stamp-collectors', count, _ ).
noun( 'stamp-dealer', 'stamp-dealers', count, _ ).
noun( 'stamp-duty', 'stamp-duties', both, _ ).
noun( 'stampede', 'stampedes', count, _ ).
noun( 'stamping-ground', 'stamping-grounds', count, _ ).
noun( 'stance', 'stances', count, _ ).
noun( 'stanchion', 'stanchions', count, _ ).
noun( 'stand', 'stands', count, _ ).
noun( 'stand-in', 'stand-ins', count, _ ).
noun( 'stand-to', '-', count, _ ).
noun( 'standard', 'standards', count, _ ).
noun( 'standard-bearer', 'standard-bearers', count, _ ).
noun( 'standardization', 'standardizations', both, _ ).
noun( 'standby', 'standbys', both, _ ).
noun( 'standing', '-', mass, _ ).
noun( 'standoffishness', '-', mass, _ ).
noun( 'standpipe', 'standpipes', count, _ ).
noun( 'standpoint', 'standpoints', count, _ ).
noun( 'standstill', '-', count, _ ).
noun( 'stanza', 'stanzas', count, _ ).
noun( 'staple', 'staples', both, _ ).
noun( 'stapler', 'staplers', count, _ ).
noun( 'stapling-machine', 'stapling-machines', count, _ ).
noun( 'star', 'stars', count, _ ).
noun( 'starboard', 'starboards', count, _ ).
noun( 'starch', '-', mass, _ ).
noun( 'stardom', '-', mass, _ ).
noun( 'stardust', '-', mass, _ ).
noun( 'stare', 'stares', count, _ ).
noun( 'starfish', 'starfish', count, _ ).
noun( 'stargazer', 'stargazers', count, _ ).
noun( 'starlet', 'starlets', count, _ ).
noun( 'starlight', '-', mass, _ ).
noun( 'starling', 'starlings', count, _ ).
noun( 'start', 'starts', both, _ ).
noun( 'starter', 'starters', count, _ ).
noun( 'starting-gate', 'starting-gates', count, _ ).
noun( 'starting-point', 'starting-points', count, _ ).
noun( 'starting-post', 'starting-posts', count, _ ).
noun( 'starvation', '-', mass, _ ).
noun( 'starveling', 'starvelings', count, _ ).
noun( 'state', 'states', both, _ ).
noun( 'statecraft', '-', mass, _ ).
noun( 'stateliness', '-', mass, _ ).
noun( 'statement', 'statements', both, _ ).
noun( 'statesman', 'statesmen', count, _ ).
noun( 'statesmanship', '-', mass, _ ).
noun( 'statics', 'statics', mass, _ ).
noun( 'station', 'stations', both, _ ).
noun( 'station-waggon', 'station-waggons', count, _ ).
noun( 'stationer', 'stationers', count, _ ).
noun( 'stationery', '-', mass, _ ).
noun( 'stationmaster', 'stationmasters', count, _ ).
noun( 'statistic', 'statistics', count, _ ).
noun( 'statistician', 'statisticians', count, _ ).
noun( 'statistics', 'statistics', mass, _ ).
noun( 'statuary', '-', mass, _ ).
noun( 'statue', 'statues', count, _ ).
noun( 'statuette', 'statuettes', count, _ ).
noun( 'stature', '-', mass, _ ).
noun( 'status', '-', mass, _ ).
noun( 'status quo', '-', count, _ ).
noun( 'statute', 'statutes', count, _ ).
noun( 'statute-book', 'statute-books', count, _ ).
noun( 'staunchness', '-', mass, _ ).
noun( 'stave', 'staves', count, _ ).
noun( 'stay', 'stays', count, _ ).
noun( 'stay-at-home', 'stay-at-homes', count, _ ).
noun( 'stayer', 'stayers', count, _ ).
noun( 'stead', '-', mass, _ ).
noun( 'steadfastness', '-', mass, _ ).
noun( 'steadiness', '-', mass, _ ).
noun( 'steady', 'steadies', count, _ ).
noun( 'steak', 'steaks', both, _ ).
noun( 'stealth', '-', mass, _ ).
noun( 'steam', '-', mass, _ ).
noun( 'steam-boiler', 'steam-boilers', count, _ ).
noun( 'steam-coal', '-', mass, _ ).
noun( 'steam-engine', 'steam-engines', count, _ ).
noun( 'steam-heat', '-', mass, _ ).
noun( 'steamboat', 'steamboats', count, _ ).
noun( 'steamer', 'steamers', count, _ ).
noun( 'steamroller', 'steamrollers', count, _ ).
noun( 'steamship', 'steamships', count, _ ).
noun( 'steed', 'steeds', count, _ ).
noun( 'steel', '-', mass, _ ).
noun( 'steelworks', 'steelworks', count, _ ).
noun( 'steelyard', 'steelyards', count, _ ).
noun( 'steenbok', 'steenboks', count, _ ).
noun( 'steeple', 'steeples', count, _ ).
noun( 'steeplechase', 'steeplechases', count, _ ).
noun( 'steeplechaser', 'steeplechasers', count, _ ).
noun( 'steeplejack', 'steeplejacks', count, _ ).
noun( 'steepness', '-', mass, _ ).
noun( 'steer', 'steers', count, _ ).
noun( 'steerage', 'steerages', both, _ ).
noun( 'steerageway', '-', mass, _ ).
noun( 'steering-gear', '-', mass, _ ).
noun( 'steering-wheel', 'steering-wheels', count, _ ).
noun( 'steersman', 'steersmen', count, _ ).
noun( 'stele', 'stelae', count, _ ).
noun( 'stem', 'stems', count, _ ).
noun( 'stench', 'stenches', count, _ ).
noun( 'stencil', 'stencils', count, _ ).
noun( 'stenographer', 'stenographers', count, _ ).
noun( 'stenography', '-', mass, _ ).
noun( 'step', 'steps', count, _ ).
noun( 'stepbrother', 'stepbrothers', count, _ ).
noun( 'stepchild', 'stepchildren', count, _ ).
noun( 'stepdaughter', 'stepdaughters', count, _ ).
noun( 'stepfather', 'stepfathers', count, _ ).
noun( 'stepladder', 'stepladders', count, _ ).
noun( 'stepmother', 'stepmothers', count, _ ).
noun( 'stepparent', 'stepparents', count, _ ).
noun( 'steppe', 'steppes', count, _ ).
noun( 'stepping-stone', 'stepping-stones', count, _ ).
noun( 'stepsister', 'stepsisters', count, _ ).
noun( 'stepson', 'stepsons', count, _ ).
noun( 'stereo', 'stereos', count, _ ).
noun( 'stereoscope', 'stereoscopes', count, _ ).
noun( 'stereotype', 'stereotypes', both, _ ).
noun( 'sterility', '-', mass, _ ).
noun( 'sterilization', '-', mass, _ ).
noun( 'sterling', '-', mass, _ ).
noun( 'stern', 'sterns', count, _ ).
noun( 'sternness', '-', mass, _ ).
noun( 'sternum', 'sternums', count, _ ).
noun( 'sternwheeler', 'sternwheelers', count, _ ).
noun( 'stethoscope', 'stethoscopes', count, _ ).
noun( 'stetson', 'stetsons', count, _ ).
noun( 'stevedore', 'stevedores', count, _ ).
noun( 'stew', 'stews', both, _ ).
noun( 'steward', 'stewards', count, _ ).
noun( 'stewardess', 'stewardesses', count, _ ).
noun( 'stewardship', '-', mass, _ ).
noun( 'stick', 'sticks', count, _ ).
noun( 'stick-in-the-mud', '-', count, _ ).
noun( 'stick-up', 'stick-ups', count, _ ).
noun( 'sticker', 'stickers', count, _ ).
noun( 'stickiness', '-', mass, _ ).
noun( 'sticking-plaster', 'sticking-plasters', both, _ ).
noun( 'stickler', 'sticklers', count, _ ).
noun( 'stiff', 'stiffs', count, _ ).
noun( 'stiffener', 'stiffeners', count, _ ).
noun( 'stiffening', 'stiffenings', both, _ ).
noun( 'stiffness', '-', mass, _ ).
noun( 'stigma', 'stigmas', count, _ ).
noun( 'stile', 'stiles', count, _ ).
noun( 'stiletto', 'stilettos', count, _ ).
noun( 'still', 'stills', count, _ ).
noun( 'still-life', 'still-lifes', both, _ ).
noun( 'still-room', 'still-rooms', count, _ ).
noun( 'stillbirth', 'stillbirths', count, _ ).
noun( 'stillness', '-', mass, _ ).
noun( 'stilt', 'stilts', count, _ ).
noun( 'stimulant', 'stimulants', count, _ ).
noun( 'stimulation', 'stimulations', both, _ ).
noun( 'stimulus', 'stimuli', count, _ ).
noun( 'sting', 'stings', both, _ ).
noun( 'stinger', 'stingers', count, _ ).
noun( 'stinginess', '-', mass, _ ).
noun( 'stingray', 'stingrays', count, _ ).
noun( 'stink', 'stinks', count, _ ).
noun( 'stinker', 'stinkers', count, _ ).
noun( 'stint', 'stints', count, _ ).
noun( 'stipend', 'stipends', count, _ ).
noun( 'stipendiary', 'stipendiaries', count, _ ).
noun( 'stipulation', 'stipulations', count, _ ).
noun( 'stir', 'stirs', count, _ ).
noun( 'stirrup', 'stirrups', count, _ ).
noun( 'stirrup-cup', 'stirrup-cups', count, _ ).
noun( 'stitch', 'stitches', count, _ ).
noun( 'stoat', 'stoats', count, _ ).
noun( 'stock', 'stocks', both, _ ).
noun( 'stock-cube', 'stock-cubes', count, _ ).
noun( 'stock-farmer', 'stock-farmers', count, _ ).
noun( 'stock-in-trade', 'stock-in-trades', both, _ ).
noun( 'stock-list', 'stock-lists', count, _ ).
noun( 'stockade', 'stockades', count, _ ).
noun( 'stockbreeder', 'stockbreeders', count, _ ).
noun( 'stockbroker', 'stockbrokers', count, _ ).
noun( 'stockcar', 'stockcars', count, _ ).
noun( 'stockfish', 'stockfish', count, _ ).
noun( 'stockholder', 'stockholders', count, _ ).
noun( 'stockholding', 'stockholdings', count, _ ).
noun( 'stockinette', '-', mass, _ ).
noun( 'stocking', 'stockings', count, _ ).
noun( 'stockist', 'stockists', count, _ ).
noun( 'stockjobber', 'stockjobbers', count, _ ).
noun( 'stockpile', 'stockpiles', count, _ ).
noun( 'stockpiling', '-', mass, _ ).
noun( 'stockpot', 'stockpots', count, _ ).
noun( 'stockroom', 'stockrooms', count, _ ).
noun( 'stocktaking', 'stocktakings', both, _ ).
noun( 'stockyard', 'stockyards', count, _ ).
noun( 'stodge', '-', mass, _ ).
noun( 'stodginess', '-', mass, _ ).
noun( 'stoep', 'stoeps', count, _ ).
noun( 'stoic', 'stoics', count, _ ).
noun( 'stoicism', '-', mass, _ ).
noun( 'stokehold', 'stokeholds', count, _ ).
noun( 'stokehole', 'stokeholes', count, _ ).
noun( 'stoker', 'stokers', count, _ ).
noun( 'stole', 'stoles', count, _ ).
noun( 'stolidity', '-', mass, _ ).
noun( 'stolidness', '-', mass, _ ).
noun( 'stomach', 'stomachs', both, _ ).
noun( 'stomach-ache', 'stomach-aches', count, _ ).
noun( 'stomach-pump', 'stomach-pumps', count, _ ).
noun( 'stomp', 'stomps', count, _ ).
noun( 'stone', 'stones', both, _ ).
noun( 'stone-fruit', 'stone-fruits', both, _ ).
noun( 'stone-pit', 'stone-pits', count, _ ).
noun( 'stonebreaker', 'stonebreakers', count, _ ).
noun( 'stonemason', 'stonemasons', count, _ ).
noun( 'stonewaller', 'stonewallers', count, _ ).
noun( 'stonewalling', '-', mass, _ ).
noun( 'stoneware', '-', mass, _ ).
noun( 'stonework', '-', mass, _ ).
noun( 'stooge', 'stooges', count, _ ).
noun( 'stool', 'stools', count, _ ).
noun( 'stoop', 'stoops', count, _ ).
noun( 'stop', 'stops', count, _ ).
noun( 'stopcock', 'stopcocks', count, _ ).
noun( 'stopgap', 'stopgaps', count, _ ).
noun( 'stopover', 'stopovers', count, _ ).
noun( 'stoppage', 'stoppages', count, _ ).
noun( 'stopper', 'stoppers', count, _ ).
noun( 'stopping', 'stoppings', both, _ ).
noun( 'stopwatch', 'stopwatches', count, _ ).
noun( 'storage', '-', mass, _ ).
noun( 'store', 'stores', both, _ ).
noun( 'storehouse', 'storehouses', count, _ ).
noun( 'storeroom', 'storerooms', count, _ ).
noun( 'storey', 'storeys', count, _ ).
noun( 'stork', 'storks', count, _ ).
noun( 'storm', 'storms', count, _ ).
noun( 'storm-centre', 'storm-centres', count, _ ).
noun( 'storm-cloud', 'storm-clouds', count, _ ).
noun( 'storm-cone', 'storm-cones', count, _ ).
noun( 'storm-lantern', 'storm-lanterns', count, _ ).
noun( 'storm-signal', 'storm-signals', count, _ ).
noun( 'storm-trooper', 'storm-troopers', count, _ ).
noun( 'story', 'stories', count, _ ).
noun( 'storyteller', 'storytellers', count, _ ).
noun( 'stoup', 'stoups', count, _ ).
noun( 'stout', '-', mass, _ ).
noun( 'stoutness', '-', mass, _ ).
noun( 'stove', 'stoves', count, _ ).
noun( 'stovepipe', 'stovepipes', count, _ ).
noun( 'stowaway', 'stowaways', count, _ ).
noun( 'straggler', 'stragglers', count, _ ).
noun( 'straight', 'straights', count, _ ).
noun( 'straightness', '-', mass, _ ).
noun( 'strain', 'strains', both, _ ).
noun( 'strainer', 'strainers', count, _ ).
noun( 'strait', 'straits', count, _ ).
noun( 'straitjacket', 'straitjackets', count, _ ).
noun( 'strand', 'strands', count, _ ).
noun( 'strangeness', '-', mass, _ ).
noun( 'stranger', 'strangers', count, _ ).
noun( 'stranglehold', 'strangleholds', count, _ ).
noun( 'strangulation', '-', mass, _ ).
noun( 'strap', 'straps', both, _ ).
noun( 'straphanger', 'straphangers', count, _ ).
noun( 'strapping', '-', count, _ ).
noun( 'stratagem', 'stratagems', both, _ ).
noun( 'strategics', 'strategics', mass, _ ).
noun( 'strategist', 'strategists', count, _ ).
noun( 'strategy', 'strategies', both, _ ).
noun( 'stratification', 'stratifications', both, _ ).
noun( 'stratosphere', 'stratospheres', count, _ ).
noun( 'stratum', 'strata', count, _ ).
noun( 'straw', 'straws', both, _ ).
noun( 'strawberry', 'strawberries', count, _ ).
noun( 'strawboard', '-', mass, _ ).
noun( 'stray', 'strays', count, _ ).
noun( 'streak', 'streaks', count, _ ).
noun( 'stream', 'streams', count, _ ).
noun( 'streamer', 'streamers', count, _ ).
noun( 'streamlet', 'streamlets', count, _ ).
noun( 'street', 'streets', count, _ ).
noun( 'street-girl', 'street-girls', count, _ ).
noun( 'street-urchin', 'street-urchins', count, _ ).
noun( 'streetcar', 'streetcars', count, _ ).
noun( 'streetwalker', 'streetwalkers', count, _ ).
noun( 'strength', 'strengths', both, _ ).
noun( 'strenuousness', '-', mass, _ ).
noun( 'streptococcus', 'streptococci', count, _ ).
noun( 'streptomycin', '-', mass, _ ).
noun( 'stress', 'stresses', both, _ ).
noun( 'stress-mark', 'stress-marks', count, _ ).
noun( 'stretch', 'stretches', both, _ ).
noun( 'stretcher', 'stretchers', count, _ ).
noun( 'stretcher-bearer', 'stretcher-bearers', count, _ ).
noun( 'stretcher-party', 'stretcher-parties', count, _ ).
noun( 'strictness', '-', mass, _ ).
noun( 'stricture', 'strictures', count, _ ).
noun( 'stride', 'strides', count, _ ).
noun( 'stridulation', 'stridulations', both, _ ).
noun( 'strife', '-', mass, _ ).
noun( 'strike', 'strikes', count, _ ).
noun( 'strike-leader', 'strike-leaders', count, _ ).
noun( 'strike-pay', '-', mass, _ ).
noun( 'strikebreaker', 'strikebreakers', count, _ ).
noun( 'striker', 'strikers', count, _ ).
noun( 'string', 'strings', both, _ ).
noun( 'stringency', 'stringencies', both, _ ).
noun( 'strip', 'strips', count, _ ).
noun( 'strip-lighting', '-', mass, _ ).
noun( 'strip-poker', '-', mass, _ ).
noun( 'strip-show', 'strip-shows', count, _ ).
noun( 'stripe', 'stripes', count, _ ).
noun( 'stripling', 'striplings', count, _ ).
noun( 'stripper', 'strippers', count, _ ).
noun( 'striptease', '-', mass, _ ).
noun( 'striver', 'strivers', count, _ ).
noun( 'stroboscope', 'stroboscopes', count, _ ).
noun( 'stroke', 'strokes', count, _ ).
noun( 'stroll', 'strolls', count, _ ).
noun( 'stroller', 'strollers', count, _ ).
noun( 'strongbox', 'strongboxes', count, _ ).
noun( 'stronghold', 'strongholds', count, _ ).
noun( 'strongroom', 'strongrooms', count, _ ).
noun( 'strontium', '-', mass, _ ).
noun( 'strop', 'strops', count, _ ).
noun( 'strophe', 'strophes', count, _ ).
noun( 'structure', 'structures', both, _ ).
noun( 'strudel', 'strudels', both, _ ).
noun( 'struggle', 'struggles', count, _ ).
noun( 'strum', 'strums', count, _ ).
noun( 'strumpet', 'strumpets', count, _ ).
noun( 'strut', 'struts', count, _ ).
noun( 'strychnine', '-', mass, _ ).
noun( 'stub', 'stubs', count, _ ).
noun( 'stubble', '-', mass, _ ).
noun( 'stubbornness', '-', mass, _ ).
noun( 'stucco', 'stuccos', both, _ ).
noun( 'stud', 'studs', count, _ ).
noun( 'stud-farm', 'stud-farms', count, _ ).
noun( 'stud-mare', 'stud-mares', count, _ ).
noun( 'studbook', 'studbooks', count, _ ).
noun( 'student', 'students', count, _ ).
noun( 'studio', 'studios', count, _ ).
noun( 'studiousness', '-', mass, _ ).
noun( 'study', 'studies', both, _ ).
noun( 'stuff', 'stuffs', both, _ ).
noun( 'stuffiness', '-', mass, _ ).
noun( 'stuffing', '-', mass, _ ).
noun( 'stultification', 'stultifications', both, _ ).
noun( 'stumble', 'stumbles', count, _ ).
noun( 'stumbling-block', 'stumbling-blocks', count, _ ).
noun( 'stump', 'stumps', count, _ ).
noun( 'stumper', 'stumpers', count, _ ).
noun( 'stunner', 'stunners', count, _ ).
noun( 'stunt', 'stunts', count, _ ).
noun( 'stupefaction', '-', mass, _ ).
noun( 'stupid', '-', count, _ ).
noun( 'stupidity', 'stupidities', both, _ ).
noun( 'stupor', 'stupors', both, _ ).
noun( 'sturdiness', '-', mass, _ ).
noun( 'sturgeon', 'sturgeons', both, _ ).
noun( 'stutter', 'stutters', count, _ ).
noun( 'stutterer', 'stutterers', count, _ ).
noun( 'sty', 'sties', count, _ ).
noun( 'stye', 'styes', count, _ ).
noun( 'style', 'styles', both, _ ).
noun( 'stylishness', '-', mass, _ ).
noun( 'stylist', 'stylists', count, _ ).
noun( 'stylization', 'stylizations', both, _ ).
noun( 'stylus', 'styluses', count, _ ).
noun( 'stymie', 'stymies', count, _ ).
noun( 'styptic', 'styptics', count, _ ).
noun( 'suasion', '-', mass, _ ).
noun( 'suavity', '-', mass, _ ).
noun( 'sub', 'subs', count, _ ).
noun( 'subaltern', 'subalterns', count, _ ).
noun( 'subcommittee', 'subcommittees', count, _ ).
noun( 'subconscious', '-', count, _ ).
noun( 'subconsciousness', '-', mass, _ ).
noun( 'subcontinent', 'subcontinents', count, _ ).
noun( 'subcontract', 'subcontracts', count, _ ).
noun( 'subcontractor', 'subcontractors', count, _ ).
noun( 'subdivision', 'subdivisions', both, _ ).
noun( 'subeditor', 'subeditors', count, _ ).
noun( 'subgroup', 'subgroups', count, _ ).
noun( 'subheading', 'subheadings', count, _ ).
noun( 'subject', 'subjects', count, _ ).
noun( 'subjection', '-', mass, _ ).
noun( 'subjectivity', '-', mass, _ ).
noun( 'subjugation', 'subjugations', both, _ ).
noun( 'subjunctive', 'subjunctives', both, _ ).
noun( 'sublease', 'subleases', count, _ ).
noun( 'sublieutenant', 'sublieutenants', count, _ ).
noun( 'sublimate', 'sublimates', count, _ ).
noun( 'sublimation', 'sublimations', both, _ ).
noun( 'sublime', '-', count, _ ).
noun( 'sublimity', '-', mass, _ ).
noun( 'submarine', 'submarines', count, _ ).
noun( 'submariner', 'submariners', count, _ ).
noun( 'submergence', '-', mass, _ ).
noun( 'submersion', '-', mass, _ ).
noun( 'submission', 'submissions', both, _ ).
noun( 'submissiveness', '-', mass, _ ).
noun( 'subnormal', 'subnormals', count, _ ).
noun( 'subordinate', 'subordinates', count, _ ).
noun( 'subordination', 'subordinations', both, _ ).
noun( 'subornation', '-', mass, _ ).
noun( 'subpoena', 'subpoenas', count, _ ).
noun( 'subscriber', 'subscribers', count, _ ).
noun( 'subscript', 'subscripts', count, _ ).
noun( 'subscription', 'subscriptions', both, _ ).
noun( 'subsection', 'subsections', count, _ ).
noun( 'subservience', '-', mass, _ ).
noun( 'subsidence', 'subsidences', both, _ ).
noun( 'subsidiary', 'subsidiaries', count, _ ).
noun( 'subsidization', '-', mass, _ ).
noun( 'subsidy', 'subsidies', count, _ ).
noun( 'subsistence', '-', mass, _ ).
noun( 'subsoil', '-', mass, _ ).
noun( 'substance', 'substances', both, _ ).
noun( 'substantiation', 'substantiations', both, _ ).
noun( 'substantive', 'substantives', count, _ ).
noun( 'substation', 'substations', count, _ ).
noun( 'substitute', 'substitutes', count, _ ).
noun( 'substitution', 'substitutions', both, _ ).
noun( 'substrate', 'substrates', count, _ ).
noun( 'substratum', 'substrata', count, _ ).
noun( 'substructure', 'substructures', count, _ ).
noun( 'subterfuge', 'subterfuges', both, _ ).
noun( 'subtitle', 'subtitles', count, _ ).
noun( 'subtlety', 'subtleties', both, _ ).
noun( 'subtopia', '-', mass, _ ).
noun( 'subtraction', 'subtractions', both, _ ).
noun( 'suburb', 'suburbs', count, _ ).
noun( 'suburbia', '-', mass, _ ).
noun( 'subvention', 'subventions', count, _ ).
noun( 'subversion', '-', mass, _ ).
noun( 'subversive', 'subversives', count, _ ).
noun( 'subway', 'subways', count, _ ).
noun( 'success', 'successes', both, _ ).
noun( 'succession', 'successions', both, _ ).
noun( 'successor', 'successors', count, _ ).
noun( 'succinctness', '-', mass, _ ).
noun( 'succour', '-', mass, _ ).
noun( 'succubus', 'succubuses', count, _ ).
noun( 'succulence', '-', mass, _ ).
noun( 'succulent', 'succulents', count, _ ).
noun( 'suck', 'sucks', count, _ ).
noun( 'sucker', 'suckers', count, _ ).
noun( 'sucking-pig', 'sucking-pigs', both, _ ).
noun( 'suckling', 'sucklings', count, _ ).
noun( 'suction', '-', mass, _ ).
noun( 'sudden', '-', count, _ ).
noun( 'suddenness', '-', mass, _ ).
noun( 'suede', '-', mass, _ ).
noun( 'suet', '-', mass, _ ).
noun( 'sufferance', '-', mass, _ ).
noun( 'sufferer', 'sufferers', count, _ ).
noun( 'suffering', 'sufferings', both, _ ).
noun( 'sufficiency', '-', mass, _ ).
noun( 'suffix', 'suffixes', count, _ ).
noun( 'suffocation', '-', mass, _ ).
noun( 'suffragan', 'suffragans', count, _ ).
noun( 'suffrage', 'suffrages', both, _ ).
noun( 'suffragette', 'suffragettes', count, _ ).
noun( 'suffusion', '-', mass, _ ).
noun( 'sugar', 'sugars', both, _ ).
noun( 'sugar-beet', 'sugar-beet', both, _ ).
noun( 'sugar-candy', 'sugar-candies', both, _ ).
noun( 'sugar-cane', '-', mass, _ ).
noun( 'sugar-daddy', 'sugar-daddies', count, _ ).
noun( 'sugar-loaf', '-', mass, _ ).
noun( 'sugar-refinery', 'sugar-refineries', count, _ ).
noun( 'sugarlump', 'sugarlumps', count, _ ).
noun( 'suggestibility', '-', mass, _ ).
noun( 'suggestion', 'suggestions', both, _ ).
noun( 'suicide', 'suicides', both, _ ).
noun( 'suit', 'suits', count, _ ).
noun( 'suitability', '-', mass, _ ).
noun( 'suitableness', '-', mass, _ ).
noun( 'suitcase', 'suitcases', count, _ ).
noun( 'suite', 'suites', count, _ ).
noun( 'suiting', '-', mass, _ ).
noun( 'suitor', 'suitors', count, _ ).
noun( 'sulk', 'sulks', count, _ ).
noun( 'sulkiness', '-', mass, _ ).
noun( 'sulky', 'sulkies', count, _ ).
noun( 'sullenness', '-', mass, _ ).
noun( 'sulpha', '-', mass, _ ).
noun( 'sulphate', 'sulphates', both, _ ).
noun( 'sulphide', 'sulphides', both, _ ).
noun( 'sulphur', '-', mass, _ ).
noun( 'sultan', 'sultans', count, _ ).
noun( 'sultana', 'sultanas', count, _ ).
noun( 'sultanate', 'sultanates', count, _ ).
noun( 'sultriness', '-', mass, _ ).
noun( 'sum', 'sums', count, _ ).
noun( 'sumac', '-', mass, _ ).
noun( 'sumach', '-', mass, _ ).
noun( 'summary', 'summaries', count, _ ).
noun( 'summation', 'summations', count, _ ).
noun( 'summer', 'summers', both, _ ).
noun( 'summerhouse', 'summerhouses', count, _ ).
noun( 'summertime', 'summertimes', both, _ ).
noun( 'summing-up', 'summing-ups', count, _ ).
noun( 'summit', 'summits', count, _ ).
noun( 'summons', 'summonses', count, _ ).
noun( 'sump', 'sumps', count, _ ).
noun( 'sumpter', 'sumpters', count, _ ).
noun( 'sumptuousness', '-', mass, _ ).
noun( 'sun', 'suns', both, _ ).
noun( 'sun-god', 'sun-gods', count, _ ).
noun( 'sun-helmet', 'sun-helmets', count, _ ).
noun( 'sun-lounge', 'sun-lounges', count, _ ).
noun( 'sun-parlour', 'sun-parlours', count, _ ).
noun( 'sun-porch', 'sun-porches', count, _ ).
noun( 'sun-up', '-', mass, _ ).
noun( 'sun-visor', 'sun-visors', count, _ ).
noun( 'sun-worship', '-', mass, _ ).
noun( 'sunbathe', '-', count, _ ).
noun( 'sunbeam', 'sunbeams', count, _ ).
noun( 'sunblind', 'sunblinds', count, _ ).
noun( 'sunbonnet', 'sunbonnets', count, _ ).
noun( 'sunburn', 'sunburns', both, _ ).
noun( 'sunburst', 'sunbursts', count, _ ).
noun( 'sundae', 'sundaes', count, _ ).
noun( 'sunder', '-', mass, _ ).
noun( 'sundial', 'sundials', count, _ ).
noun( 'sundown', '-', mass, _ ).
noun( 'sundowner', 'sundowners', count, _ ).
noun( 'sunfish', 'sunfish', count, _ ).
noun( 'sunflower', 'sunflowers', count, _ ).
noun( 'sunhat', 'sunhats', count, _ ).
noun( 'sunlamp', 'sunlamps', count, _ ).
noun( 'sunlight', '-', mass, _ ).
noun( 'sunrise', 'sunrises', both, _ ).
noun( 'sunroof', 'sunroofs', count, _ ).
noun( 'sunset', 'sunsets', both, _ ).
noun( 'sunshade', 'sunshades', count, _ ).
noun( 'sunshine', '-', mass, _ ).
noun( 'sunshine-roof', 'sunshine-roofs', count, _ ).
noun( 'sunspot', 'sunspots', count, _ ).
noun( 'sunstroke', '-', mass, _ ).
noun( 'suntan', 'suntans', both, _ ).
noun( 'suntrap', 'suntraps', count, _ ).
noun( 'sup', 'sups', count, _ ).
noun( 'super', 'supers', count, _ ).
noun( 'superabundance', '-', count, _ ).
noun( 'superannuation', 'superannuations', both, _ ).
noun( 'supercargo', 'supercargoes', count, _ ).
noun( 'supercharger', 'superchargers', count, _ ).
noun( 'superciliousness', '-', mass, _ ).
noun( 'superego', '-', mass, _ ).
noun( 'supererogation', '-', mass, _ ).
noun( 'superficiality', 'superficialities', both, _ ).
noun( 'superficies', 'superficies', count, _ ).
noun( 'superfluity', 'superfluities', both, _ ).
noun( 'superintendence', '-', mass, _ ).
noun( 'superintendent', 'superintendents', count, _ ).
noun( 'superior', 'superiors', count, _ ).
noun( 'superiority', '-', mass, _ ).
noun( 'superlative', 'superlatives', count, _ ).
noun( 'superman', 'supermen', count, _ ).
noun( 'supermarket', 'supermarkets', count, _ ).
noun( 'supernumerary', 'supernumeraries', count, _ ).
noun( 'superscription', 'superscriptions', count, _ ).
noun( 'supersession', '-', mass, _ ).
noun( 'superstition', 'superstitions', both, _ ).
noun( 'superstructure', 'superstructures', count, _ ).
noun( 'supertax', 'supertaxes', both, _ ).
noun( 'supervision', 'supervisions', both, _ ).
noun( 'supervisor', 'supervisors', count, _ ).
noun( 'supper', 'suppers', both, _ ).
noun( 'supplanter', 'supplanters', count, _ ).
noun( 'supplement', 'supplements', count, _ ).
noun( 'suppleness', '-', mass, _ ).
noun( 'suppliant', 'suppliants', count, _ ).
noun( 'supplicant', 'supplicants', count, _ ).
noun( 'supplication', 'supplications', both, _ ).
noun( 'supplier', 'suppliers', count, _ ).
noun( 'supply', 'supplies', both, _ ).
noun( 'support', 'supports', both, _ ).
noun( 'supporter', 'supporters', count, _ ).
noun( 'supposition', 'suppositions', both, _ ).
noun( 'suppository', 'suppositories', count, _ ).
noun( 'suppression', 'suppressions', count, _ ).
noun( 'suppressor', 'suppressors', count, _ ).
noun( 'suppuration', 'suppurations', both, _ ).
noun( 'supremacy', '-', mass, _ ).
noun( 'surcharge', 'surcharges', count, _ ).
noun( 'surd', 'surds', count, _ ).
noun( 'sureness', '-', mass, _ ).
noun( 'surety', 'sureties', both, _ ).
noun( 'surf', '-', mass, _ ).
noun( 'surface', 'surfaces', count, _ ).
noun( 'surfboard', 'surfboards', count, _ ).
noun( 'surfboat', 'surfboats', count, _ ).
noun( 'surfeit', 'surfeits', count, _ ).
noun( 'surfing', '-', mass, _ ).
noun( 'surfriding', '-', mass, _ ).
noun( 'surge', 'surges', count, _ ).
noun( 'surgeon', 'surgeons', count, _ ).
noun( 'surgery', 'surgeries', both, _ ).
noun( 'surliness', '-', mass, _ ).
noun( 'surmise', 'surmises', count, _ ).
noun( 'surname', 'surnames', count, _ ).
noun( 'surplice', 'surplices', count, _ ).
noun( 'surplus', 'surpluses', count, _ ).
noun( 'surprise', 'surprises', both, _ ).
noun( 'surrealism', '-', mass, _ ).
noun( 'surrealist', 'surrealists', count, _ ).
noun( 'surrender', 'surrenders', count, _ ).
noun( 'surrogate', 'surrogates', count, _ ).
noun( 'surround', 'surrounds', count, _ ).
noun( 'surtax', 'surtaxes', both, _ ).
noun( 'surveillance', '-', mass, _ ).
noun( 'survey', 'surveys', count, _ ).
noun( 'surveying', '-', mass, _ ).
noun( 'surveyor', 'surveyors', count, _ ).
noun( 'survival', 'survivals', both, _ ).
noun( 'survivor', 'survivors', count, _ ).
noun( 'susceptibility', 'susceptibilities', both, _ ).
noun( 'suspect', 'suspects', count, _ ).
noun( 'suspender', 'suspenders', count, _ ).
noun( 'suspense', '-', mass, _ ).
noun( 'suspension', '-', mass, _ ).
noun( 'suspicion', 'suspicions', both, _ ).
noun( 'sustenance', '-', mass, _ ).
noun( 'suttee', 'suttees', both, _ ).
noun( 'suture', 'sutures', count, _ ).
noun( 'suzerain', 'suzerains', count, _ ).
noun( 'suzerainty', 'suzerainties', count, _ ).
noun( 'swab', 'swabs', count, _ ).
noun( 'swag', '-', mass, _ ).
noun( 'swagger', 'swaggers', count, _ ).
noun( 'swaggerer', 'swaggerers', count, _ ).
noun( 'swain', 'swains', count, _ ).
noun( 'swallow', 'swallows', count, _ ).
noun( 'swami', 'swamis', count, _ ).
noun( 'swamp', 'swamps', both, _ ).
noun( 'swan', 'swans', count, _ ).
noun( 'swan\'s-down', '-', mass, _ ).
noun( 'swan-song', 'swan-songs', count, _ ).
noun( 'swank', 'swanks', both, _ ).
noun( 'swap', 'swaps', count, _ ).
noun( 'sward', '-', mass, _ ).
noun( 'swarm', 'swarms', count, _ ).
noun( 'swashbuckler', 'swashbucklers', count, _ ).
noun( 'swashbuckling', '-', mass, _ ).
noun( 'swastika', 'swastikas', count, _ ).
noun( 'swat', 'swats', count, _ ).
noun( 'swath', 'swaths', count, _ ).
noun( 'swathe', 'swathes', count, _ ).
noun( 'sway', '-', mass, _ ).
noun( 'swearer', 'swearers', count, _ ).
noun( 'swearing', '-', mass, _ ).
noun( 'swearword', 'swearwords', count, _ ).
noun( 'sweat', 'sweats', both, _ ).
noun( 'sweatband', 'sweatbands', count, _ ).
noun( 'sweater', 'sweaters', count, _ ).
noun( 'sweatshop', 'sweatshops', count, _ ).
noun( 'swede', 'swedes', count, _ ).
noun( 'sweep', 'sweeps', count, _ ).
noun( 'sweeper', 'sweepers', count, _ ).
noun( 'sweeping', 'sweepings', both, _ ).
noun( 'sweet', 'sweets', both, _ ).
noun( 'sweetbread', 'sweetbreads', count, _ ).
noun( 'sweetbriar', '-', mass, _ ).
noun( 'sweetbrier', '-', mass, _ ).
noun( 'sweetening', 'sweetenings', both, _ ).
noun( 'sweetheart', 'sweethearts', count, _ ).
noun( 'sweetie', 'sweeties', count, _ ).
noun( 'sweetmeat', 'sweetmeats', count, _ ).
noun( 'sweetness', '-', mass, _ ).
noun( 'swell', 'swells', count, _ ).
noun( 'swelling', 'swellings', both, _ ).
noun( 'swerve', 'swerves', count, _ ).
noun( 'swift', 'swifts', count, _ ).
noun( 'swiftness', '-', mass, _ ).
noun( 'swig', 'swigs', count, _ ).
noun( 'swill', 'swills', both, _ ).
noun( 'swim', 'swims', count, _ ).
noun( 'swimmer', 'swimmers', count, _ ).
noun( 'swimming', '-', mass, _ ).
noun( 'swimming-bath', 'swimming-baths', count, _ ).
noun( 'swimming-costume', 'swimming-costumes', count, _ ).
noun( 'swimming-pool', 'swimming-pools', count, _ ).
noun( 'swimsuit', 'swimsuits', count, _ ).
noun( 'swindle', 'swindles', count, _ ).
noun( 'swindler', 'swindlers', count, _ ).
noun( 'swine', 'swine', count, _ ).
noun( 'swineherd', 'swineherds', count, _ ).
noun( 'swing', 'swings', count, _ ).
noun( 'swipe', 'swipes', count, _ ).
noun( 'swirl', 'swirls', count, _ ).
noun( 'swish', 'swishes', count, _ ).
noun( 'switch', 'switches', count, _ ).
noun( 'switchboard', 'switchboards', count, _ ).
noun( 'switchman', 'switchmen', count, _ ).
noun( 'swivel', 'swivels', count, _ ).
noun( 'swiz', '-', count, _ ).
noun( 'swizzle', 'swizzles', count, _ ).
noun( 'swizzle-stick', 'swizzle-sticks', count, _ ).
noun( 'swob', 'swobs', count, _ ).
noun( 'swoon', 'swoons', count, _ ).
noun( 'swoop', 'swoops', count, _ ).
noun( 'swop', 'swops', count, _ ).
noun( 'sword', 'swords', count, _ ).
noun( 'sword-cane', 'sword-canes', count, _ ).
noun( 'sword-cut', 'sword-cuts', count, _ ).
noun( 'sword-dance', 'sword-dances', count, _ ).
noun( 'swordfish', 'swordfish', count, _ ).
noun( 'swordplay', '-', mass, _ ).
noun( 'swordsman', 'swordsmen', count, _ ).
noun( 'swordsmanship', '-', mass, _ ).
noun( 'swordstick', 'swordsticks', count, _ ).
noun( 'swot', 'swots', count, _ ).
noun( 'sybarite', 'sybarites', count, _ ).
noun( 'sycamore', 'sycamores', both, _ ).
noun( 'sycophancy', '-', mass, _ ).
noun( 'sycophant', 'sycophants', count, _ ).
noun( 'syllabary', 'syllabaries', count, _ ).
noun( 'syllabication', '-', mass, _ ).
noun( 'syllabification', '-', mass, _ ).
noun( 'syllable', 'syllables', count, _ ).
noun( 'syllabus', 'syllabuses', count, _ ).
noun( 'syllogism', 'syllogisms', count, _ ).
noun( 'sylph', 'sylphs', count, _ ).
noun( 'symbiosis', '-', mass, _ ).
noun( 'symbol', 'symbols', count, _ ).
noun( 'symbolism', 'symbolisms', both, _ ).
noun( 'symbolization', 'symbolizations', both, _ ).
noun( 'symmetry', '-', mass, _ ).
noun( 'sympathizer', 'sympathizers', count, _ ).
noun( 'sympathy', 'sympathies', both, _ ).
noun( 'symphony', 'symphonies', count, _ ).
noun( 'symposium', 'symposiums', count, _ ).
noun( 'symptom', 'symptoms', count, _ ).
noun( 'synagogue', 'synagogues', count, _ ).
noun( 'synchroflash', '-', count, _ ).
noun( 'synchromesh', '-', mass, _ ).
noun( 'synchronization', 'synchronizations', both, _ ).
noun( 'synchrony', '-', mass, _ ).
noun( 'synchrotron', 'synchrotrons', count, _ ).
noun( 'syncopation', 'syncopations', both, _ ).
noun( 'syncope', '-', mass, _ ).
noun( 'syndic', 'syndics', count, _ ).
noun( 'syndicalism', '-', mass, _ ).
noun( 'syndicalist', 'syndicalists', count, _ ).
noun( 'syndicate', 'syndicates', count, _ ).
noun( 'syndication', 'syndications', count, _ ).
noun( 'syndrome', 'syndromes', count, _ ).
noun( 'synod', 'synods', count, _ ).
noun( 'synonym', 'synonyms', count, _ ).
noun( 'synopsis', 'synopses', count, _ ).
noun( 'syntax', '-', mass, _ ).
noun( 'synthesis', 'syntheses', both, _ ).
noun( 'synthetic', 'synthetics', count, _ ).
noun( 'syphilis', '-', mass, _ ).
noun( 'syphilitic', 'syphilitics', count, _ ).
noun( 'syphon', 'syphons', count, _ ).
noun( 'syringa', 'syringas', both, _ ).
noun( 'syringe', 'syringes', count, _ ).
noun( 'syrup', 'syrups', both, _ ).
noun( 'system', 'systems', both, _ ).
noun( 'systematization', '-', mass, _ ).
noun( 't', '-', count, _ ).
noun( 't^ete-`a-t^ete', 't^ete-`a-t^etes', count, _ ).
noun( 'tab', 'tabs', count, _ ).
noun( 'tabard', 'tabards', count, _ ).
noun( 'tabby', 'tabbies', count, _ ).
noun( 'tabby-cat', 'tabby-cats', count, _ ).
noun( 'tabernacle', 'tabernacles', count, _ ).
noun( 'table', 'tables', count, _ ).
noun( 'table-knife', 'table-knives', count, _ ).
noun( 'table-lifting', '-', mass, _ ).
noun( 'table-linen', '-', mass, _ ).
noun( 'table-rapping', '-', mass, _ ).
noun( 'table-talk', '-', mass, _ ).
noun( 'table-turning', '-', mass, _ ).
noun( 'tableau', 'tableaux', count, _ ).
noun( 'tableau vivant', 'tableaux vivants', count, _ ).
noun( 'tablecloth', 'tablecloths', count, _ ).
noun( 'tablemat', 'tablemats', count, _ ).
noun( 'tablespoon', 'tablespoons', count, _ ).
noun( 'tablespoonful', 'tablespoonfuls', count, _ ).
noun( 'tablet', 'tablets', count, _ ).
noun( 'tableware', '-', mass, _ ).
noun( 'tabloid', 'tabloids', count, _ ).
noun( 'taboo', 'taboos', both, _ ).
noun( 'tabor', 'tabors', count, _ ).
noun( 'tabulation', 'tabulations', both, _ ).
noun( 'tabulator', 'tabulators', count, _ ).
noun( 'tachograph', 'tachographs', count, _ ).
noun( 'taciturnity', '-', mass, _ ).
noun( 'tack', 'tacks', both, _ ).
noun( 'tackle', 'tackles', both, _ ).
noun( 'tact', '-', mass, _ ).
noun( 'tactic', 'tactics', count, _ ).
noun( 'tactician', 'tacticians', count, _ ).
noun( 'tactlessness', '-', mass, _ ).
noun( 'tadpole', 'tadpoles', count, _ ).
noun( 'taffeta', '-', mass, _ ).
noun( 'taffrail', 'taffrails', count, _ ).
noun( 'taffy', 'taffies', count, _ ).
noun( 'tag', 'tags', both, _ ).
noun( 'tail', 'tails', count, _ ).
noun( 'tail-coat', 'tail-coats', count, _ ).
noun( 'tail-end', 'tail-ends', count, _ ).
noun( 'tail-light', 'tail-lights', count, _ ).
noun( 'tailboard', 'tailboards', count, _ ).
noun( 'tailgate', 'tailgates', count, _ ).
noun( 'tailor', 'tailors', count, _ ).
noun( 'tailpiece', 'tailpieces', count, _ ).
noun( 'tailplane', 'tailplanes', count, _ ).
noun( 'tailspin', 'tailspins', count, _ ).
noun( 'taint', 'taints', both, _ ).
noun( 'take', 'takes', count, _ ).
noun( 'take-off', 'take-offs', count, _ ).
noun( 'take-up', '-', mass, _ ).
noun( 'takeover', 'takeovers', count, _ ).
noun( 'taker', 'takers', count, _ ).
noun( 'talc', '-', mass, _ ).
noun( 'talcum', '-', mass, _ ).
noun( 'tale', 'tales', count, _ ).
noun( 'tale-bearer', 'tale-bearers', count, _ ).
noun( 'tale-teller', 'tale-tellers', count, _ ).
noun( 'talent', 'talents', both, _ ).
noun( 'talisman', 'talismans', count, _ ).
noun( 'talk', 'talks', both, _ ).
noun( 'talker', 'talkers', count, _ ).
noun( 'talkie', 'talkies', count, _ ).
noun( 'talking-point', 'talking-points', count, _ ).
noun( 'talking-to', 'talking-tos', count, _ ).
noun( 'tallboy', 'tallboys', count, _ ).
noun( 'tallow', '-', mass, _ ).
noun( 'tally', 'tallies', count, _ ).
noun( 'tally-clerk', 'tally-clerks', count, _ ).
noun( 'tallyman', 'tallymen', count, _ ).
noun( 'talon', 'talons', count, _ ).
noun( 'talus', 'taluses', count, _ ).
noun( 'tam-o\'-shanter', 'tam-o\'-shanters', count, _ ).
noun( 'tamale', 'tamales', both, _ ).
noun( 'tamarind', 'tamarinds', both, _ ).
noun( 'tamarisk', 'tamarisks', count, _ ).
noun( 'tambour', 'tambours', count, _ ).
noun( 'tambourine', 'tambourines', count, _ ).
noun( 'tameness', '-', mass, _ ).
noun( 'tamer', 'tamers', count, _ ).
noun( 'tammy', 'tammies', count, _ ).
noun( 'tan', 'tans', count, _ ).
noun( 'tandem', 'tandems', count, _ ).
noun( 'tang', 'tangs', count, _ ).
noun( 'tangent', 'tangents', count, _ ).
noun( 'tangerine', 'tangerines', both, _ ).
noun( 'tangibility', '-', mass, _ ).
noun( 'tangle', 'tangles', both, _ ).
noun( 'tango', 'tangos', count, _ ).
noun( 'tank', 'tanks', count, _ ).
noun( 'tank-car', 'tank-cars', count, _ ).
noun( 'tankard', 'tankards', count, _ ).
noun( 'tanker', 'tankers', count, _ ).
noun( 'tanner', 'tanners', count, _ ).
noun( 'tannery', 'tanneries', count, _ ).
noun( 'tannin', '-', mass, _ ).
noun( 'tannoy', 'tannoys', count, _ ).
noun( 'tansy', 'tansies', count, _ ).
noun( 'tantrum', 'tantrums', count, _ ).
noun( 'tap', 'taps', count, _ ).
noun( 'tap-dancing', '-', mass, _ ).
noun( 'tape', 'tapes', both, _ ).
noun( 'tape-measure', 'tape-measures', count, _ ).
noun( 'tape-recorder', 'tape-recorders', count, _ ).
noun( 'taper', 'tapers', count, _ ).
noun( 'tapestry', 'tapestries', both, _ ).
noun( 'tapeworm', 'tapeworms', count, _ ).
noun( 'tapioca', '-', mass, _ ).
noun( 'tapir', 'tapirs', count, _ ).
noun( 'taproom', 'taprooms', count, _ ).
noun( 'taproot', 'taproots', count, _ ).
noun( 'tapster', 'tapsters', count, _ ).
noun( 'tar', 'tars', both, _ ).
noun( 'tar-macadam', '-', mass, _ ).
noun( 'taradiddle', 'taradiddles', count, _ ).
noun( 'tarantella', 'tarantellas', count, _ ).
noun( 'tarantelle', 'tarantelles', count, _ ).
noun( 'tarantula', 'tarantulas', count, _ ).
noun( 'tarboosh', 'tarbooshes', count, _ ).
noun( 'tardiness', '-', mass, _ ).
noun( 'tare', 'tares', count, _ ).
noun( 'target', 'targets', count, _ ).
noun( 'tariff', 'tariffs', count, _ ).
noun( 'tarmac', '-', mass, _ ).
noun( 'tarn', 'tarns', count, _ ).
noun( 'tarnish', '-', mass, _ ).
noun( 'taro', 'taros', both, _ ).
noun( 'tarpaulin', 'tarpaulins', both, _ ).
noun( 'tarpon', 'tarpons', count, _ ).
noun( 'tarradiddle', 'tarradiddles', count, _ ).
noun( 'tarragon', '-', mass, _ ).
noun( 'tarsal', 'tarsals', count, _ ).
noun( 'tarsus', 'tarsi', count, _ ).
noun( 'tart', 'tarts', count, _ ).
noun( 'tartan', 'tartans', both, _ ).
noun( 'tartar', 'tartars', both, _ ).
noun( 'tartness', '-', mass, _ ).
noun( 'task', 'tasks', count, _ ).
noun( 'task-force', 'task-forces', count, _ ).
noun( 'taskmaster', 'taskmasters', count, _ ).
noun( 'tassel', 'tassels', count, _ ).
noun( 'taste', 'tastes', both, _ ).
noun( 'taster', 'tasters', count, _ ).
noun( 'tat', '-', mass, _ ).
noun( 'tatter', 'tatters', count, _ ).
noun( 'tatterdemalion', 'tatterdemalions', count, _ ).
noun( 'tatting', '-', mass, _ ).
noun( 'tattle', '-', mass, _ ).
noun( 'tattler', 'tattlers', count, _ ).
noun( 'tattoo', 'tattoos', count, _ ).
noun( 'taunt', 'taunts', count, _ ).
noun( 'tautness', '-', mass, _ ).
noun( 'tautology', 'tautologies', both, _ ).
noun( 'tavern', 'taverns', count, _ ).
noun( 'tawdriness', '-', mass, _ ).
noun( 'tawse', 'tawses', count, _ ).
noun( 'tax', 'taxes', both, _ ).
noun( 'tax-collector', 'tax-collectors', count, _ ).
noun( 'taxability', '-', mass, _ ).
noun( 'taxation', '-', mass, _ ).
noun( 'taxi', 'taxis', count, _ ).
noun( 'taxicab', 'taxicabs', count, _ ).
noun( 'taxidermist', 'taxidermists', count, _ ).
noun( 'taxidermy', '-', mass, _ ).
noun( 'taximeter', 'taximeters', count, _ ).
noun( 'taxonomy', 'taxonomies', both, _ ).
noun( 'taxpayer', 'taxpayers', count, _ ).
noun( 'tea', 'teas', both, _ ).
noun( 'tea-bag', 'tea-bags', count, _ ).
noun( 'tea-break', 'tea-breaks', count, _ ).
noun( 'tea-caddy', 'tea-caddies', count, _ ).
noun( 'tea-chest', 'tea-chests', count, _ ).
noun( 'tea-cloth', 'tea-cloths', count, _ ).
noun( 'tea-cosy', 'tea-cosies', count, _ ).
noun( 'tea-garden', 'tea-gardens', count, _ ).
noun( 'tea-kettle', 'tea-kettles', count, _ ).
noun( 'tea-leaf', 'tea-leaves', count, _ ).
noun( 'tea-party', 'tea-parties', count, _ ).
noun( 'tea-service', 'tea-services', count, _ ).
noun( 'tea-set', 'tea-sets', count, _ ).
noun( 'tea-strainer', 'tea-strainers', count, _ ).
noun( 'tea-table', 'tea-tables', count, _ ).
noun( 'tea-time', '-', mass, _ ).
noun( 'tea-towel', 'tea-towels', count, _ ).
noun( 'tea-tray', 'tea-trays', count, _ ).
noun( 'tea-trolley', 'tea-trolleys', count, _ ).
noun( 'tea-urn', 'tea-urns', count, _ ).
noun( 'tea-wagon', 'tea-wagons', count, _ ).
noun( 'teacake', 'teacakes', count, _ ).
noun( 'teach-in', 'teach-ins', count, _ ).
noun( 'teacher', 'teachers', count, _ ).
noun( 'teaching', 'teachings', both, _ ).
noun( 'teacup', 'teacups', count, _ ).
noun( 'teahouse', 'teahouses', count, _ ).
noun( 'teak', '-', mass, _ ).
noun( 'teal', 'teal', count, _ ).
noun( 'team', 'teams', count, _ ).
noun( 'teamster', 'teamsters', count, _ ).
noun( 'teamwork', '-', mass, _ ).
noun( 'teapot', 'teapots', count, _ ).
noun( 'tear', 'tears', count, _ ).
noun( 'tear', 'tears', count, _ ).
noun( 'tear-drop', 'tear-drops', count, _ ).
noun( 'tear-gas', '-', mass, _ ).
noun( 'tearaway', 'tearaways', count, _ ).
noun( 'tearing', '-', mass, _ ).
noun( 'tearoom', 'tearooms', count, _ ).
noun( 'tease', 'teases', count, _ ).
noun( 'teasel', 'teasels', count, _ ).
noun( 'teaser', 'teasers', count, _ ).
noun( 'teashop', 'teashops', count, _ ).
noun( 'teaspoon', 'teaspoons', count, _ ).
noun( 'teaspoonful', 'teaspoonfuls', count, _ ).
noun( 'teat', 'teats', count, _ ).
noun( 'teazel', 'teazels', count, _ ).
noun( 'teazle', 'teazles', count, _ ).
noun( 'tec', 'tecs', count, _ ).
noun( 'tech', 'techs', count, _ ).
noun( 'technicality', 'technicalities', both, _ ).
noun( 'technician', 'technicians', count, _ ).
noun( 'technique', 'techniques', both, _ ).
noun( 'technocracy', 'technocracies', both, _ ).
noun( 'technocrat', 'technocrats', count, _ ).
noun( 'technologist', 'technologists', count, _ ).
noun( 'technology', 'technologies', both, _ ).
noun( 'teddy', 'teddies', count, _ ).
noun( 'tediousness', '-', mass, _ ).
noun( 'tedium', '-', mass, _ ).
noun( 'tee', 'tees', count, _ ).
noun( 'tee-shirt', 'tee-shirts', count, _ ).
noun( 'teenager', 'teenagers', count, _ ).
noun( 'teetotaller', 'teetotallers', count, _ ).
noun( 'teetotum', 'teetotums', count, _ ).
noun( 'teg', 'tegs', count, _ ).
noun( 'tegument', 'teguments', count, _ ).
noun( 'tel', '-', proper, _ ).
noun( 'telecast', 'telecasts', count, _ ).
noun( 'telecommunication', 'telecommunications', both, _ ).
noun( 'telegram', 'telegrams', count, _ ).
noun( 'telegraph', 'telegraphs', count, _ ).
noun( 'telegraph-line', 'telegraph-lines', count, _ ).
noun( 'telegraph-pole', 'telegraph-poles', count, _ ).
noun( 'telegraph-post', 'telegraph-posts', count, _ ).
noun( 'telegraph-wire', 'telegraph-wires', count, _ ).
noun( 'telegrapher', 'telegraphers', count, _ ).
noun( 'telegraphese', '-', mass, _ ).
noun( 'telegraphist', 'telegraphists', count, _ ).
noun( 'telegraphy', '-', mass, _ ).
noun( 'telemetry', '-', mass, _ ).
noun( 'teleologist', 'teleologists', count, _ ).
noun( 'teleology', 'teleologies', both, _ ).
noun( 'telepathist', 'telepathists', count, _ ).
noun( 'telepathy', '-', mass, _ ).
noun( 'telephone', 'telephones', both, _ ).
noun( 'telephonist', 'telephonists', count, _ ).
noun( 'telephony', '-', mass, _ ).
noun( 'telephoto', '-', mass, _ ).
noun( 'telephotograph', 'telephotographs', count, _ ).
noun( 'telephotography', '-', mass, _ ).
noun( 'teleprinter', 'teleprinters', count, _ ).
noun( 'teleprompter', 'teleprompters', count, _ ).
noun( 'telescope', 'telescopes', count, _ ).
noun( 'teletypewriter', 'teletypewriters', count, _ ).
noun( 'television', 'televisions', both, _ ).
noun( 'telex', 'telexes', count, _ ).
noun( 'telfer', 'telfers', count, _ ).
noun( 'teller', 'tellers', count, _ ).
noun( 'telltale', 'telltales', count, _ ).
noun( 'telly', 'tellies', count, _ ).
noun( 'telpher', 'telphers', count, _ ).
noun( 'temerity', '-', mass, _ ).
noun( 'temp', 'temps', count, _ ).
noun( 'temper', 'tempers', both, _ ).
noun( 'tempera', '-', mass, _ ).
noun( 'temperament', 'temperaments', both, _ ).
noun( 'temperance', '-', mass, _ ).
noun( 'temperateness', '-', mass, _ ).
noun( 'temperature', 'temperatures', both, _ ).
noun( 'tempest', 'tempests', count, _ ).
noun( 'template', 'templates', count, _ ).
noun( 'temple', 'temples', count, _ ).
noun( 'templet', 'templets', count, _ ).
noun( 'tempo', 'tempos', count, _ ).
noun( 'temporality', '-', mass, _ ).
noun( 'temporalty', '-', mass, _ ).
noun( 'temporariness', '-', mass, _ ).
noun( 'temptation', 'temptations', both, _ ).
noun( 'tempter', 'tempters', count, _ ).
noun( 'temptress', 'temptresses', count, _ ).
noun( 'ten', 'tens', count, _ ).
noun( 'tenability', '-', mass, _ ).
noun( 'tenaciousness', '-', mass, _ ).
noun( 'tenacity', '-', mass, _ ).
noun( 'tenancy', 'tenancies', both, _ ).
noun( 'tenant', 'tenants', count, _ ).
noun( 'tenantry', 'tenantries', count, _ ).
noun( 'tench', 'tench', count, _ ).
noun( 'tendency', 'tendencies', both, _ ).
noun( 'tendentiousness', '-', mass, _ ).
noun( 'tender', 'tenders', count, _ ).
noun( 'tenderfoot', 'tenderfoots', count, _ ).
noun( 'tenderloin', '-', mass, _ ).
noun( 'tenderness', '-', mass, _ ).
noun( 'tendon', 'tendons', count, _ ).
noun( 'tendril', 'tendrils', count, _ ).
noun( 'tenement', 'tenements', count, _ ).
noun( 'tenement-house', 'tenement-houses', count, _ ).
noun( 'tenet', 'tenets', count, _ ).
noun( 'tenner', 'tenners', count, _ ).
noun( 'tennis', '-', mass, _ ).
noun( 'tennis-court', 'tennis-courts', count, _ ).
noun( 'tennis-elbow', '-', mass, _ ).
noun( 'tenon', 'tenons', count, _ ).
noun( 'tenor', 'tenors', count, _ ).
noun( 'tenpence', 'tenpences', count, _ ).
noun( 'tenpin', 'tenpins', count, _ ).
noun( 'tense', 'tenses', count, _ ).
noun( 'tenseness', '-', mass, _ ).
noun( 'tension', 'tensions', both, _ ).
noun( 'tensity', '-', mass, _ ).
noun( 'tent', 'tents', count, _ ).
noun( 'tent-peg', 'tent-pegs', count, _ ).
noun( 'tentacle', 'tentacles', count, _ ).
noun( 'tenth', 'tenths', count, _ ).
noun( 'tenuity', '-', mass, _ ).
noun( 'tenure', 'tenures', both, _ ).
noun( 'tepee', 'tepees', count, _ ).
noun( 'tepidity', '-', mass, _ ).
noun( 'tepidness', '-', mass, _ ).
noun( 'tercentenary', 'tercentenaries', count, _ ).
noun( 'tercentennial', 'tercentennials', count, _ ).
noun( 'tergiversation', '-', mass, _ ).
noun( 'term', 'terms', count, _ ).
noun( 'termagant', 'termagants', count, _ ).
noun( 'terminal', 'terminals', count, _ ).
noun( 'termination', 'terminations', both, _ ).
noun( 'terminology', 'terminologies', both, _ ).
noun( 'terminus', 'terminuses', count, _ ).
noun( 'termite', 'termites', count, _ ).
noun( 'tern', 'terns', count, _ ).
noun( 'terra firma', '-', mass, _ ).
noun( 'terra incognita', '-', mass, _ ).
noun( 'terra-cotta', '-', mass, _ ).
noun( 'terrace', 'terraces', count, _ ).
noun( 'terrain', '-', mass, _ ).
noun( 'terrapin', 'terrapins', count, _ ).
noun( 'terrier', 'terriers', count, _ ).
noun( 'territorial', 'territorials', count, _ ).
noun( 'territory', 'territories', both, _ ).
noun( 'terror', 'terrors', both, _ ).
noun( 'terrorism', '-', mass, _ ).
noun( 'terrorist', 'terrorists', count, _ ).
noun( 'terseness', '-', mass, _ ).
noun( 'terylene', '-', mass, _ ).
noun( 'test', 'tests', count, _ ).
noun( 'test-drive', 'test-drives', count, _ ).
noun( 'test-tube', 'test-tubes', count, _ ).
noun( 'testament', 'testaments', count, _ ).
noun( 'testate', 'testates', count, _ ).
noun( 'testator', 'testators', count, _ ).
noun( 'testatrix', 'testatrixes', count, _ ).
noun( 'testbed', 'testbeds', count, _ ).
noun( 'testicle', 'testicles', count, _ ).
noun( 'testimonial', 'testimonials', count, _ ).
noun( 'testimony', 'testimonies', both, _ ).
noun( 'testiness', '-', mass, _ ).
noun( 'testis', 'testes', count, _ ).
noun( 'tetanus', '-', mass, _ ).
noun( 'tetchiness', '-', mass, _ ).
noun( 'tether', 'tethers', count, _ ).
noun( 'text', 'texts', both, _ ).
noun( 'textbook', 'textbooks', count, _ ).
noun( 'textile', 'textiles', count, _ ).
noun( 'texture', 'textures', both, _ ).
noun( 'thalidomide', '-', mass, _ ).
noun( 'thane', 'thanes', count, _ ).
noun( 'thank-offering', 'thank-offerings', count, _ ).
noun( 'thankfulness', '-', mass, _ ).
noun( 'thanksgiving', 'thanksgivings', both, _ ).
noun( 'thatch', '-', mass, _ ).
noun( 'thatcher', 'thatchers', count, _ ).
noun( 'thaw', 'thaws', count, _ ).
noun( 'theatre', 'theatres', both, _ ).
noun( 'theatregoer', 'theatregoers', count, _ ).
noun( 'theatrical', 'theatricals', count, _ ).
noun( 'theft', 'thefts', both, _ ).
noun( 'theism', '-', mass, _ ).
noun( 'theist', 'theists', count, _ ).
noun( 'theme', 'themes', count, _ ).
noun( 'theocracy', 'theocracies', both, _ ).
noun( 'theodolite', 'theodolites', count, _ ).
noun( 'theologian', 'theologians', count, _ ).
noun( 'theology', 'theologies', both, _ ).
noun( 'theorem', 'theorems', count, _ ).
noun( 'theoretician', 'theoreticians', count, _ ).
noun( 'theorist', 'theorists', count, _ ).
noun( 'theory', 'theories', both, _ ).
noun( 'theosophist', 'theosophists', count, _ ).
noun( 'theosophy', '-', mass, _ ).
noun( 'therapeutics', 'therapeutics', mass, _ ).
noun( 'therapist', 'therapists', count, _ ).
noun( 'therapy', 'therapies', both, _ ).
noun( 'therm', 'therms', count, _ ).
noun( 'thermal', 'thermals', count, _ ).
noun( 'thermodynamics', 'thermodynamics', mass, _ ).
noun( 'thermometer', 'thermometers', count, _ ).
noun( 'thermoplastic', 'thermoplastics', count, _ ).
noun( 'thermos', 'thermoses', count, _ ).
noun( 'thermostat', 'thermostats', count, _ ).
noun( 'thesaurus', 'thesauruses', count, _ ).
noun( 'thesis', 'theses', count, _ ).
noun( 'thick', '-', mass, _ ).
noun( 'thickening', '-', mass, _ ).
noun( 'thicket', 'thickets', count, _ ).
noun( 'thickness', 'thicknesses', both, _ ).
noun( 'thief', 'thieves', count, _ ).
noun( 'thievery', '-', mass, _ ).
noun( 'thigh', 'thighs', count, _ ).
noun( 'thighbone', 'thighbones', count, _ ).
noun( 'thimble', 'thimbles', count, _ ).
noun( 'thimbleful', 'thimblefuls', count, _ ).
noun( 'thing', 'things', count, _ ).
noun( 'thingmabob', 'thingmabobs', count, _ ).
noun( 'thingmajig', 'thingmajigs', count, _ ).
noun( 'thingumabob', 'thingumabobs', count, _ ).
noun( 'thingumajig', 'thingumajigs', count, _ ).
noun( 'thingummy', 'thingummies', count, _ ).
noun( 'think', 'thinks', count, _ ).
noun( 'think-tank', 'think-tanks', count, _ ).
noun( 'thinker', 'thinkers', count, _ ).
noun( 'thinking', '-', mass, _ ).
noun( 'thinness', '-', mass, _ ).
noun( 'third', 'thirds', count, _ ).
noun( 'third-rater', 'third-raters', count, _ ).
noun( 'thirst', '-', mass, _ ).
noun( 'thirteen', 'thirteens', count, _ ).
noun( 'thirteenth', 'thirteenths', count, _ ).
noun( 'thirtieth', 'thirtieths', count, _ ).
noun( 'thirty', 'thirties', count, _ ).
noun( 'thistle', 'thistles', count, _ ).
noun( 'thistledown', '-', mass, _ ).
noun( 'thole', 'tholes', count, _ ).
noun( 'tholepin', 'tholepins', count, _ ).
noun( 'thong', 'thongs', count, _ ).
noun( 'thorax', 'thoraxes', count, _ ).
noun( 'thorn', 'thorns', both, _ ).
noun( 'thoroughbred', 'thoroughbreds', count, _ ).
noun( 'thoroughfare', 'thoroughfares', count, _ ).
noun( 'thoroughness', '-', mass, _ ).
noun( 'thought', 'thoughts', both, _ ).
noun( 'thought-reader', 'thought-readers', count, _ ).
noun( 'thoughtfulness', '-', mass, _ ).
noun( 'thoughtlessness', '-', mass, _ ).
noun( 'thousand', 'thousands', count, _ ).
noun( 'thousandth', 'thousandths', count, _ ).
noun( 'thraldom', '-', mass, _ ).
noun( 'thrall', 'thralls', both, _ ).
noun( 'thrashing', 'thrashings', count, _ ).
noun( 'thread', 'threads', both, _ ).
noun( 'threat', 'threats', count, _ ).
noun( 'three', 'threes', count, _ ).
noun( 'three-d', '-', mass, _ ).
noun( 'three-decker', 'three-deckers', count, _ ).
noun( 'three-quarter', 'three-quarters', count, _ ).
noun( 'three-score', '-', mass, _ ).
noun( 'threepence', 'threepences', count, _ ).
noun( 'threesome', 'threesomes', count, _ ).
noun( 'threnody', 'threnodies', count, _ ).
noun( 'thresher', 'threshers', count, _ ).
noun( 'threshing-floor', 'threshing-floors', count, _ ).
noun( 'threshing-machine', 'threshing-machines', count, _ ).
noun( 'threshold', 'thresholds', count, _ ).
noun( 'thrift', '-', mass, _ ).
noun( 'thriftlessness', '-', mass, _ ).
noun( 'thrill', 'thrills', count, _ ).
noun( 'thriller', 'thrillers', count, _ ).
noun( 'throat', 'throats', count, _ ).
noun( 'throb', 'throbs', count, _ ).
noun( 'throe', 'throes', count, _ ).
noun( 'thrombosis', '-', mass, _ ).
noun( 'throne', 'thrones', count, _ ).
noun( 'throng', 'throngs', count, _ ).
noun( 'throstle', 'throstles', count, _ ).
noun( 'throttle', 'throttles', both, _ ).
noun( 'throttle-valve', 'throttle-valves', count, _ ).
noun( 'throughput', 'throughputs', both, _ ).
noun( 'throughway', 'throughways', count, _ ).
noun( 'throw', 'throws', count, _ ).
noun( 'throw-in', 'throw-ins', count, _ ).
noun( 'throwaway', 'throwaways', count, _ ).
noun( 'throwback', 'throwbacks', count, _ ).
noun( 'thrush', 'thrushes', count, _ ).
noun( 'thrust', 'thrusts', both, _ ).
noun( 'thruster', 'thrusters', count, _ ).
noun( 'thud', 'thuds', count, _ ).
noun( 'thug', 'thugs', count, _ ).
noun( 'thuggery', '-', mass, _ ).
noun( 'thumb', 'thumbs', count, _ ).
noun( 'thumbnut', 'thumbnuts', count, _ ).
noun( 'thumbscrew', 'thumbscrews', count, _ ).
noun( 'thumbstall', 'thumbstalls', count, _ ).
noun( 'thumbtack', 'thumbtacks', count, _ ).
noun( 'thump', 'thumps', count, _ ).
noun( 'thunder', 'thunders', both, _ ).
noun( 'thunderbolt', 'thunderbolts', count, _ ).
noun( 'thunderclap', 'thunderclaps', count, _ ).
noun( 'thunderstorm', 'thunderstorms', count, _ ).
noun( 'thurible', 'thuribles', count, _ ).
noun( 'thwack', 'thwacks', count, _ ).
noun( 'thwart', 'thwarts', count, _ ).
noun( 'thyme', '-', mass, _ ).
noun( 'thyroid', 'thyroids', count, _ ).
noun( 'ti', '-', count, _ ).
noun( 'tiara', 'tiaras', count, _ ).
noun( 'tibia', 'tibiae', count, _ ).
noun( 'tic', 'tics', count, _ ).
noun( 'tick', 'ticks', both, _ ).
noun( 'tick-tock', 'tick-tocks', count, _ ).
noun( 'ticker', 'tickers', count, _ ).
noun( 'ticker-tape', '-', mass, _ ).
noun( 'ticket', 'tickets', count, _ ).
noun( 'ticket-collector', 'ticket-collectors', count, _ ).
noun( 'ticking', '-', mass, _ ).
noun( 'tickler', 'ticklers', count, _ ).
noun( 'tidbit', 'tidbits', count, _ ).
noun( 'tiddler', 'tiddlers', count, _ ).
noun( 'tiddlywinks', 'tiddlywinks', mass, _ ).
noun( 'tide', 'tides', both, _ ).
noun( 'tidemark', 'tidemarks', count, _ ).
noun( 'tideway', 'tideways', count, _ ).
noun( 'tidiness', '-', mass, _ ).
noun( 'tidy', 'tidies', count, _ ).
noun( 'tie', 'ties', count, _ ).
noun( 'tie-up', 'tie-ups', count, _ ).
noun( 'tier', 'tiers', count, _ ).
noun( 'tiff', 'tiffs', count, _ ).
noun( 'tiger', 'tigers', count, _ ).
noun( 'tiger-lily', 'tiger-lilies', count, _ ).
noun( 'tight-wad', 'tight-wads', count, _ ).
noun( 'tightness', '-', mass, _ ).
noun( 'tightrope', 'tightropes', count, _ ).
noun( 'tigress', 'tigresses', count, _ ).
noun( 'tike', 'tikes', count, _ ).
noun( 'tilde', 'tildes', count, _ ).
noun( 'tilde', 'tildes', count, _ ).
noun( 'tile', 'tiles', count, _ ).
noun( 'till', 'tills', count, _ ).
noun( 'tillage', '-', mass, _ ).
noun( 'tiller', 'tillers', count, _ ).
noun( 'tilt', 'tilts', count, _ ).
noun( 'tilth', 'tilths', count, _ ).
noun( 'tiltyard', 'tiltyards', count, _ ).
noun( 'timber', 'timbers', both, _ ).
noun( 'timbre', 'timbres', count, _ ).
noun( 'timbrel', 'timbrels', count, _ ).
noun( 'time', 'times', both, _ ).
noun( 'time-ball', 'time-balls', count, _ ).
noun( 'time-bomb', 'time-bombs', count, _ ).
noun( 'time-exposure', 'time-exposures', count, _ ).
noun( 'time-fuse', 'time-fuses', count, _ ).
noun( 'time-lag', 'time-lags', count, _ ).
noun( 'time-limit', 'time-limits', count, _ ).
noun( 'time-sheet', 'time-sheets', count, _ ).
noun( 'time-signal', 'time-signals', count, _ ).
noun( 'time-switch', 'time-switches', count, _ ).
noun( 'timecard', 'timecards', count, _ ).
noun( 'timekeeper', 'timekeepers', count, _ ).
noun( 'timekeeping', '-', mass, _ ).
noun( 'timeliness', '-', mass, _ ).
noun( 'timepiece', 'timepieces', count, _ ).
noun( 'timeserver', 'timeservers', count, _ ).
noun( 'timetable', 'timetables', count, _ ).
noun( 'timework', '-', mass, _ ).
noun( 'timidity', '-', mass, _ ).
noun( 'timidness', '-', mass, _ ).
noun( 'timing', 'timings', both, _ ).
noun( 'timothy', '-', mass, _ ).
noun( 'timpanist', 'timpanists', count, _ ).
noun( 'tin', 'tins', both, _ ).
noun( 'tin pan alley', 'tin pan alleys', count, _ ).
noun( 'tin-opener', 'tin-openers', count, _ ).
noun( 'tin-plate', '-', mass, _ ).
noun( 'tincture', 'tinctures', count, _ ).
noun( 'tinder', '-', mass, _ ).
noun( 'tinderbox', 'tinderboxes', count, _ ).
noun( 'tine', 'tines', count, _ ).
noun( 'tinfoil', '-', mass, _ ).
noun( 'ting', 'tings', count, _ ).
noun( 'tinge', 'tinges', count, _ ).
noun( 'tingle', 'tingles', count, _ ).
noun( 'tinker', 'tinkers', count, _ ).
noun( 'tinkle', '-', count, _ ).
noun( 'tinsel', '-', mass, _ ).
noun( 'tinsmith', 'tinsmiths', count, _ ).
noun( 'tint', 'tints', count, _ ).
noun( 'tintack', 'tintacks', count, _ ).
noun( 'tintinnabulation', '-', mass, _ ).
noun( 'tip', 'tips', count, _ ).
noun( 'tip-off', 'tip-offs', count, _ ).
noun( 'tippet', 'tippets', count, _ ).
noun( 'tipple', '-', mass, _ ).
noun( 'tippler', 'tipplers', count, _ ).
noun( 'tipstaff', 'tipstaffs', count, _ ).
noun( 'tipster', 'tipsters', count, _ ).
noun( 'tirade', 'tirades', count, _ ).
noun( 'tire', 'tires', count, _ ).
noun( 'tiredness', '-', mass, _ ).
noun( 'tiro', 'tiros', count, _ ).
noun( 'tissue', 'tissues', both, _ ).
noun( 'tit', 'tits', count, _ ).
noun( 'titan', 'titans', count, _ ).
noun( 'titbit', 'titbits', count, _ ).
noun( 'tithe', 'tithes', count, _ ).
noun( 'tithe-barn', 'tithe-barns', count, _ ).
noun( 'titillation', 'titillations', count, _ ).
noun( 'titlark', 'titlarks', count, _ ).
noun( 'title', 'titles', both, _ ).
noun( 'title-deed', 'title-deeds', count, _ ).
noun( 'title-page', 'title-pages', count, _ ).
noun( 'title-role', 'title-roles', count, _ ).
noun( 'titmouse', 'titmice', count, _ ).
noun( 'tittle', '-', count, _ ).
noun( 'tittle-tattle', '-', mass, _ ).
noun( 'tizzy', 'tizzies', count, _ ).
noun( 'to-do', 'to-dos', count, _ ).
noun( 'toad', 'toads', count, _ ).
noun( 'toad-in-the-hole', '-', both, _ ).
noun( 'toadstool', 'toadstools', count, _ ).
noun( 'toady', 'toadies', count, _ ).
noun( 'toast', 'toasts', both, _ ).
noun( 'toaster', 'toasters', count, _ ).
noun( 'toasting-fork', 'toasting-forks', count, _ ).
noun( 'toastmaster', 'toastmasters', count, _ ).
noun( 'toastrack', 'toastracks', count, _ ).
noun( 'tobacco', 'tobaccos', both, _ ).
noun( 'tobacconist', 'tobacconists', count, _ ).
noun( 'toboggan', 'toboggans', count, _ ).
noun( 'toby-jug', 'toby-jugs', count, _ ).
noun( 'toccata', 'toccatas', count, _ ).
noun( 'tocsin', 'tocsins', count, _ ).
noun( 'today', '-', mass, _ ).
noun( 'toddler', 'toddlers', count, _ ).
noun( 'toddy', 'toddies', both, _ ).
noun( 'toe', 'toes', count, _ ).
noun( 'toecap', 'toecaps', count, _ ).
noun( 'toehold', 'toeholds', count, _ ).
noun( 'toenail', 'toenails', count, _ ).
noun( 'toff', 'toffs', count, _ ).
noun( 'toffee', 'toffees', both, _ ).
noun( 'toga', 'togas', count, _ ).
noun( 'togetherness', '-', mass, _ ).
noun( 'toggle', 'toggles', count, _ ).
noun( 'toil', 'toils', both, _ ).
noun( 'toiler', 'toilers', count, _ ).
noun( 'toilet', 'toilets', count, _ ).
noun( 'toilet-paper', 'toilet-papers', count, _ ).
noun( 'toilet-powder', '-', mass, _ ).
noun( 'toilet-roll', 'toilet-rolls', count, _ ).
noun( 'toilet-table', 'toilet-tables', count, _ ).
noun( 'token', 'tokens', count, _ ).
noun( 'tolerance', 'tolerances', both, _ ).
noun( 'toleration', '-', mass, _ ).
noun( 'toll', 'tolls', count, _ ).
noun( 'tollbar', 'tollbars', count, _ ).
noun( 'tollbooth', 'tollbooths', count, _ ).
noun( 'tollgate', 'tollgates', count, _ ).
noun( 'tollhouse', 'tollhouses', count, _ ).
noun( 'tomahawk', 'tomahawks', count, _ ).
noun( 'tomato', 'tomatoes', count, _ ).
noun( 'tomb', 'tombs', count, _ ).
noun( 'tombola', 'tombolas', count, _ ).
noun( 'tomboy', 'tomboys', count, _ ).
noun( 'tombstone', 'tombstones', count, _ ).
noun( 'tomcat', 'tomcats', count, _ ).
noun( 'tome', 'tomes', count, _ ).
noun( 'tomfool', 'tomfools', count, _ ).
noun( 'tomfoolery', 'tomfooleries', both, _ ).
noun( 'tommy-gun', 'tommy-guns', count, _ ).
noun( 'tommy-rot', '-', mass, _ ).
noun( 'tomorrow', 'tomorrows', both, _ ).
noun( 'tomtit', 'tomtits', count, _ ).
noun( 'tomtom', 'tomtoms', count, _ ).
noun( 'ton', 'tons', count, _ ).
noun( 'tonality', 'tonalities', both, _ ).
noun( 'tone', 'tones', both, _ ).
noun( 'tone-poem', 'tone-poems', count, _ ).
noun( 'tongue', 'tongues', both, _ ).
noun( 'tongue-twister', 'tongue-twisters', count, _ ).
noun( 'tonic', 'tonics', count, _ ).
noun( 'tonic sol-fa', '-', mass, _ ).
noun( 'tonight', '-', mass, _ ).
noun( 'tonnage', 'tonnages', count, _ ).
noun( 'tonne', 'tonnes', count, _ ).
noun( 'tonsil', 'tonsils', count, _ ).
noun( 'tonsillitis', '-', mass, _ ).
noun( 'tonsure', 'tonsures', count, _ ).
noun( 'tontine', 'tontines', count, _ ).
noun( 'tool', 'tools', count, _ ).
noun( 'toot', 'toots', count, _ ).
noun( 'tooth', 'teeth', count, _ ).
noun( 'toothache', '-', both, _ ).
noun( 'toothbrush', 'toothbrushes', count, _ ).
noun( 'toothpaste', 'toothpastes', both, _ ).
noun( 'toothpick', 'toothpicks', count, _ ).
noun( 'toothpowder', '-', mass, _ ).
noun( 'tootle', 'tootles', count, _ ).
noun( 'top', 'tops', count, _ ).
noun( 'top-boot', 'top-boots', count, _ ).
noun( 'top-dressing', 'top-dressings', both, _ ).
noun( 'topaz', 'topazes', both, _ ).
noun( 'topcoat', 'topcoats', count, _ ).
noun( 'toper', 'topers', count, _ ).
noun( 'topgallant', 'topgallants', count, _ ).
noun( 'topi', 'topis', count, _ ).
noun( 'topiary', '-', mass, _ ).
noun( 'topic', 'topics', count, _ ).
noun( 'topknot', 'topknots', count, _ ).
noun( 'topmast', 'topmasts', count, _ ).
noun( 'topography', '-', mass, _ ).
noun( 'topper', 'toppers', count, _ ).
noun( 'topsail', 'topsails', count, _ ).
noun( 'topsy-turvydom', 'topsy-turvydoms', count, _ ).
noun( 'toque', 'toques', count, _ ).
noun( 'tor', 'tors', count, _ ).
noun( 'torch', 'torches', count, _ ).
noun( 'torch-race', 'torch-races', count, _ ).
noun( 'torch-singer', 'torch-singers', count, _ ).
noun( 'torchlight', '-', mass, _ ).
noun( 'toreador', 'toreadors', count, _ ).
noun( 'torment', 'torments', both, _ ).
noun( 'tormentor', 'tormentors', count, _ ).
noun( 'tornado', 'tornadoes', count, _ ).
noun( 'torpedo', 'torpedoes', count, _ ).
noun( 'torpedo-boat', 'torpedo-boats', count, _ ).
noun( 'torpedo-tube', 'torpedo-tubes', count, _ ).
noun( 'torpidity', '-', mass, _ ).
noun( 'torpidness', '-', mass, _ ).
noun( 'torpor', 'torpors', both, _ ).
noun( 'torque', 'torques', both, _ ).
noun( 'torrent', 'torrents', count, _ ).
noun( 'torridity', '-', mass, _ ).
noun( 'torsion', '-', mass, _ ).
noun( 'torso', 'torsos', count, _ ).
noun( 'tort', 'torts', count, _ ).
noun( 'tortilla', 'tortillas', count, _ ).
noun( 'tortoise', 'tortoises', count, _ ).
noun( 'tortoiseshell', '-', mass, _ ).
noun( 'torture', 'tortures', both, _ ).
noun( 'torturer', 'torturers', count, _ ).
noun( 'tosh', 'toshes', count, _ ).
noun( 'toss', 'tosses', count, _ ).
noun( 'toss-up', 'toss-ups', count, _ ).
noun( 'tot', 'tots', count, _ ).
noun( 'total', 'totals', count, _ ).
noun( 'totalitarianism', '-', mass, _ ).
noun( 'totality', '-', mass, _ ).
noun( 'totalizator', 'totalizators', count, _ ).
noun( 'tote', 'totes', count, _ ).
noun( 'totem', 'totems', count, _ ).
noun( 'totem-pole', 'totem-poles', count, _ ).
noun( 'toucan', 'toucans', count, _ ).
noun( 'touch', 'touches', both, _ ).
noun( 'touchdown', 'touchdowns', count, _ ).
noun( 'touchiness', '-', mass, _ ).
noun( 'touchline', 'touchlines', count, _ ).
noun( 'touchstone', 'touchstones', count, _ ).
noun( 'tough', 'toughs', count, _ ).
noun( 'toughie', 'toughies', count, _ ).
noun( 'toughness', '-', mass, _ ).
noun( 'toupee', 'toupees', count, _ ).
noun( 'tour', 'tours', count, _ ).
noun( 'tour de force', 'tours de force', count, _ ).
noun( 'touring', '-', mass, _ ).
noun( 'tourism', '-', mass, _ ).
noun( 'tourist', 'tourists', count, _ ).
noun( 'tournament', 'tournaments', count, _ ).
noun( 'tourney', 'tourneys', count, _ ).
noun( 'tourniquet', 'tourniquets', count, _ ).
noun( 'tout', 'touts', count, _ ).
noun( 'tout ensemble', '-', count, _ ).
noun( 'tow', 'tows', both, _ ).
noun( 'towel', 'towels', count, _ ).
noun( 'towel-horse', 'towel-horses', count, _ ).
noun( 'towel-rack', 'towel-racks', count, _ ).
noun( 'towel-rail', 'towel-rails', count, _ ).
noun( 'towelling', '-', mass, _ ).
noun( 'tower', 'towers', count, _ ).
noun( 'tower-block', 'tower-blocks', count, _ ).
noun( 'towing-line', 'towing-lines', count, _ ).
noun( 'towing-path', 'towing-paths', count, _ ).
noun( 'towing-rope', 'towing-ropes', count, _ ).
noun( 'towline', 'towlines', count, _ ).
noun( 'town', 'towns', count, _ ).
noun( 'town-crier', 'town-criers', count, _ ).
noun( 'town-gas', '-', mass, _ ).
noun( 'townee', 'townees', count, _ ).
noun( 'township', 'townships', count, _ ).
noun( 'townsman', 'townsmen', count, _ ).
noun( 'towpath', 'towpaths', count, _ ).
noun( 'towrope', 'towropes', count, _ ).
noun( 'toxaemia', '-', mass, _ ).
noun( 'toxicity', '-', mass, _ ).
noun( 'toxicologist', 'toxicologists', count, _ ).
noun( 'toxicology', '-', mass, _ ).
noun( 'toxin', 'toxins', count, _ ).
noun( 'toy', 'toys', count, _ ).
noun( 'toyshop', 'toyshops', count, _ ).
noun( 'trace', 'traces', both, _ ).
noun( 'tracer', 'tracers', count, _ ).
noun( 'tracery', 'traceries', both, _ ).
noun( 'trachea', 'tracheae', count, _ ).
noun( 'trachoma', '-', mass, _ ).
noun( 'tracing', 'tracings', count, _ ).
noun( 'tracing-paper', '-', mass, _ ).
noun( 'track', 'tracks', count, _ ).
noun( 'tracker', 'trackers', count, _ ).
noun( 'tract', 'tracts', count, _ ).
noun( 'tractability', 'tractabilities', both, _ ).
noun( 'traction', '-', mass, _ ).
noun( 'traction-engine', 'traction-engines', count, _ ).
noun( 'tractor', 'tractors', count, _ ).
noun( 'trad', '-', mass, _ ).
noun( 'trade', 'trades', both, _ ).
noun( 'trade-in', 'trade-ins', count, _ ).
noun( 'trade-union', 'trade-unions', count, _ ).
noun( 'trade-unionism', '-', mass, _ ).
noun( 'trade-unionist', 'trade-unionists', count, _ ).
noun( 'trade-wind', 'trade-winds', count, _ ).
noun( 'trademark', 'trademarks', count, _ ).
noun( 'trader', 'traders', count, _ ).
noun( 'trades-union', 'trades-unions', count, _ ).
noun( 'tradesman', 'tradesmen', count, _ ).
noun( 'tradition', 'traditions', both, _ ).
noun( 'traditionalism', '-', mass, _ ).
noun( 'traditionalist', 'traditionalists', count, _ ).
noun( 'traducer', 'traducers', count, _ ).
noun( 'traffic', '-', mass, _ ).
noun( 'trafficator', 'trafficators', count, _ ).
noun( 'trafficker', 'traffickers', count, _ ).
noun( 'tragedian', 'tragedians', count, _ ).
noun( 'tragedienne', 'tragediennes', count, _ ).
noun( 'tragedy', 'tragedies', both, _ ).
noun( 'tragicomedy', 'tragicomedies', count, _ ).
noun( 'trail', 'trails', count, _ ).
noun( 'trailer', 'trailers', count, _ ).
noun( 'train', 'trains', count, _ ).
noun( 'trainbearer', 'trainbearers', count, _ ).
noun( 'trainee', 'trainees', count, _ ).
noun( 'trainer', 'trainers', count, _ ).
noun( 'training', '-', mass, _ ).
noun( 'training-college', 'training-colleges', count, _ ).
noun( 'training-ship', 'training-ships', count, _ ).
noun( 'trainload', 'trainloads', count, _ ).
noun( 'trainman', 'trainmen', count, _ ).
noun( 'trait', 'traits', count, _ ).
noun( 'traitor', 'traitors', count, _ ).
noun( 'traitress', 'traitresses', count, _ ).
noun( 'trajectory', 'trajectories', count, _ ).
noun( 'tram', 'trams', count, _ ).
noun( 'tram-car', 'tram-cars', count, _ ).
noun( 'tramline', 'tramlines', count, _ ).
noun( 'tramp', 'tramps', count, _ ).
noun( 'tramp-steamer', 'tramp-steamers', count, _ ).
noun( 'trample', 'tramples', count, _ ).
noun( 'trampoline', 'trampolines', count, _ ).
noun( 'tramway', 'tramways', count, _ ).
noun( 'trance', 'trances', count, _ ).
noun( 'tranquility', '-', mass, _ ).
noun( 'tranquillity', '-', mass, _ ).
noun( 'tranquillizer', 'tranquillizers', count, _ ).
noun( 'trans', '-', proper, _ ).
noun( 'transaction', 'transactions', both, _ ).
noun( 'transalpine', 'transalpines', count, _ ).
noun( 'transcendence', '-', mass, _ ).
noun( 'transcendency', '-', mass, _ ).
noun( 'transcendentalism', '-', mass, _ ).
noun( 'transcendentalist', 'transcendentalists', count, _ ).
noun( 'transcript', 'transcripts', count, _ ).
noun( 'transcription', 'transcriptions', both, _ ).
noun( 'transept', 'transepts', count, _ ).
noun( 'transfer', 'transfers', both, _ ).
noun( 'transferability', '-', mass, _ ).
noun( 'transference', 'transferences', both, _ ).
noun( 'transfiguration', 'transfigurations', both, _ ).
noun( 'transformation', 'transformations', both, _ ).
noun( 'transformer', 'transformers', count, _ ).
noun( 'transfusion', 'transfusions', both, _ ).
noun( 'transgression', 'transgressions', both, _ ).
noun( 'transgressor', 'transgressors', count, _ ).
noun( 'transience', '-', mass, _ ).
noun( 'transiency', '-', mass, _ ).
noun( 'transient', 'transients', count, _ ).
noun( 'transistor', 'transistors', count, _ ).
noun( 'transit', '-', mass, _ ).
noun( 'transition', 'transitions', both, _ ).
noun( 'translation', 'translations', both, _ ).
noun( 'translator', 'translators', count, _ ).
noun( 'transliteration', 'transliterations', both, _ ).
noun( 'translucence', '-', mass, _ ).
noun( 'translucency', '-', mass, _ ).
noun( 'transmigration', '-', mass, _ ).
noun( 'transmission', 'transmissions', both, _ ).
noun( 'transmitter', 'transmitters', count, _ ).
noun( 'transmogrification', 'transmogrifications', both, _ ).
noun( 'transmutation', 'transmutations', both, _ ).
noun( 'transom', 'transoms', count, _ ).
noun( 'transom-window', 'transom-windows', count, _ ).
noun( 'transparence', '-', mass, _ ).
noun( 'transparency', 'transparencies', both, _ ).
noun( 'transpiration', '-', mass, _ ).
noun( 'transplant', 'transplants', count, _ ).
noun( 'transplantation', 'transplantations', both, _ ).
noun( 'transport', 'transports', both, _ ).
noun( 'transportation', '-', mass, _ ).
noun( 'transporter', 'transporters', count, _ ).
noun( 'transposition', 'transpositions', both, _ ).
noun( 'transsexual', 'transsexuals', count, _ ).
noun( 'transshipment', 'transshipments', count, _ ).
noun( 'transubstantiation', '-', mass, _ ).
noun( 'transvestism', '-', mass, _ ).
noun( 'transvestite', 'transvestites', count, _ ).
noun( 'trap', 'traps', count, _ ).
noun( 'trap-door', 'trap-doors', count, _ ).
noun( 'trap-shooting', '-', mass, _ ).
noun( 'trapeze', 'trapezes', count, _ ).
noun( 'trapezium', 'trapeziums', count, _ ).
noun( 'trapezoid', 'trapezoids', count, _ ).
noun( 'trapper', 'trappers', count, _ ).
noun( 'trash', '-', mass, _ ).
noun( 'trauma', 'traumas', count, _ ).
noun( 'travail', 'travails', both, _ ).
noun( 'travel', 'travels', both, _ ).
noun( 'traveller', 'travellers', count, _ ).
noun( 'travelling', '-', mass, _ ).
noun( 'travelogue', 'travelogues', count, _ ).
noun( 'traverse', 'traverses', count, _ ).
noun( 'travesty', 'travesties', count, _ ).
noun( 'trawl', 'trawls', count, _ ).
noun( 'trawl-net', 'trawl-nets', count, _ ).
noun( 'trawler', 'trawlers', count, _ ).
noun( 'tray', 'trays', count, _ ).
noun( 'tray-cloth', 'tray-cloths', count, _ ).
noun( 'treachery', 'treacheries', both, _ ).
noun( 'treacle', '-', mass, _ ).
noun( 'tread', 'treads', count, _ ).
noun( 'treadle', 'treadles', count, _ ).
noun( 'treadmill', 'treadmills', count, _ ).
noun( 'treas', '-', count, _ ).
noun( 'treason', '-', mass, _ ).
noun( 'treasure', 'treasures', both, _ ).
noun( 'treasure-house', 'treasure-houses', count, _ ).
noun( 'treasure-trove', '-', mass, _ ).
noun( 'treasurer', 'treasurers', count, _ ).
noun( 'treasury', 'treasuries', count, _ ).
noun( 'treat', 'treats', count, _ ).
noun( 'treatise', 'treatises', count, _ ).
noun( 'treatment', 'treatments', both, _ ).
noun( 'treaty', 'treaties', both, _ ).
noun( 'treble', 'trebles', count, _ ).
noun( 'tree', 'trees', count, _ ).
noun( 'tree-fern', 'tree-ferns', count, _ ).
noun( 'trefoil', 'trefoils', count, _ ).
noun( 'trek', 'treks', count, _ ).
noun( 'trellis', 'trellises', count, _ ).
noun( 'tremble', 'trembles', count, _ ).
noun( 'tremolo', 'tremolos', count, _ ).
noun( 'tremor', 'tremors', count, _ ).
noun( 'trench', 'trenches', count, _ ).
noun( 'trenchancy', '-', mass, _ ).
noun( 'trencher', 'trenchers', count, _ ).
noun( 'trencherman', 'trenchermen', count, _ ).
noun( 'trend', 'trends', count, _ ).
noun( 'trend-setter', 'trend-setters', count, _ ).
noun( 'trend-setting', '-', mass, _ ).
noun( 'trephine', 'trephines', count, _ ).
noun( 'trepidation', '-', mass, _ ).
noun( 'trespass', 'trespasses', both, _ ).
noun( 'trespasser', 'trespassers', count, _ ).
noun( 'tress', 'tresses', count, _ ).
noun( 'trestle', 'trestles', count, _ ).
noun( 'trestle-bridge', 'trestle-bridges', count, _ ).
noun( 'trestle-table', 'trestle-tables', count, _ ).
noun( 'triad', 'triads', count, _ ).
noun( 'trial', 'trials', both, _ ).
noun( 'triangle', 'triangles', count, _ ).
noun( 'tribalism', '-', mass, _ ).
noun( 'tribe', 'tribes', count, _ ).
noun( 'tribesman', 'tribesmen', count, _ ).
noun( 'tribulation', 'tribulations', both, _ ).
noun( 'tribunal', 'tribunals', count, _ ).
noun( 'tribune', 'tribunes', count, _ ).
noun( 'tributary', 'tributaries', count, _ ).
noun( 'tribute', 'tributes', both, _ ).
noun( 'trice', 'trices', count, _ ).
noun( 'trick', 'tricks', count, _ ).
noun( 'trickery', '-', mass, _ ).
noun( 'trickiness', '-', mass, _ ).
noun( 'trickle', 'trickles', count, _ ).
noun( 'trickster', 'tricksters', count, _ ).
noun( 'tricolour', 'tricolours', count, _ ).
noun( 'tricycle', 'tricycles', count, _ ).
noun( 'trident', 'tridents', count, _ ).
noun( 'triennial', 'triennials', count, _ ).
noun( 'trier', 'triers', count, _ ).
noun( 'trifle', 'trifles', both, _ ).
noun( 'trifler', 'triflers', count, _ ).
noun( 'trigger', 'triggers', count, _ ).
noun( 'trigonometry', '-', mass, _ ).
noun( 'trilby', 'trilbies', count, _ ).
noun( 'trill', 'trills', count, _ ).
noun( 'trillion', 'trillions', count, _ ).
noun( 'trillionth', 'trillionths', count, _ ).
noun( 'trilogy', 'trilogies', count, _ ).
noun( 'trim', '-', mass, _ ).
noun( 'trimaran', 'trimarans', count, _ ).
noun( 'trimmer', 'trimmers', count, _ ).
noun( 'trimming', 'trimmings', both, _ ).
noun( 'trinitrotoluene', '-', mass, _ ).
noun( 'trinity', 'trinities', count, _ ).
noun( 'trinket', 'trinkets', count, _ ).
noun( 'trio', 'trios', count, _ ).
noun( 'trip', 'trips', count, _ ).
noun( 'tripe', '-', mass, _ ).
noun( 'triplet', 'triplets', count, _ ).
noun( 'triplicate', 'triplicates', count, _ ).
noun( 'tripod', 'tripods', count, _ ).
noun( 'tripos', 'triposes', count, _ ).
noun( 'tripper', 'trippers', count, _ ).
noun( 'triptych', 'triptychs', count, _ ).
noun( 'trireme', 'triremes', count, _ ).
noun( 'triteness', '-', mass, _ ).
noun( 'triumph', 'triumphs', both, _ ).
noun( 'triumvir', 'triumvirs', count, _ ).
noun( 'triumvirate', 'triumvirates', count, _ ).
noun( 'trivet', 'trivets', count, _ ).
noun( 'triviality', 'trivialities', both, _ ).
noun( 'trochee', 'trochees', count, _ ).
noun( 'troglodyte', 'troglodytes', count, _ ).
noun( 'troika', 'troikas', count, _ ).
noun( 'troll', 'trolls', count, _ ).
noun( 'trolley', 'trolleys', count, _ ).
noun( 'trolley-car', 'trolley-cars', count, _ ).
noun( 'trollop', 'trollops', count, _ ).
noun( 'trombone', 'trombones', count, _ ).
noun( 'trombonist', 'trombonists', count, _ ).
noun( 'troop', 'troops', count, _ ).
noun( 'troop-carrier', 'troop-carriers', count, _ ).
noun( 'trooper', 'troopers', count, _ ).
noun( 'troopship', 'troopships', count, _ ).
noun( 'trope', 'tropes', count, _ ).
noun( 'trophy', 'trophies', count, _ ).
noun( 'tropic', 'tropics', count, _ ).
noun( 'trot', 'trots', count, _ ).
noun( 'troth', '-', mass, _ ).
noun( 'trotter', 'trotters', count, _ ).
noun( 'troubadour', 'troubadours', count, _ ).
noun( 'trouble', 'troubles', both, _ ).
noun( 'troublemaker', 'troublemakers', count, _ ).
noun( 'troubleshooter', 'troubleshooters', count, _ ).
noun( 'trough', 'troughs', count, _ ).
noun( 'trouncing', 'trouncings', count, _ ).
noun( 'troupe', 'troupes', count, _ ).
noun( 'trouper', 'troupers', count, _ ).
noun( 'trouser', 'trousers', count, _ ).
noun( 'trousseau', 'trousseaus', count, _ ).
noun( 'trout', 'trout', both, _ ).
noun( 'trove', 'troves', count, _ ).
noun( 'trowel', 'trowels', count, _ ).
noun( 'troy', '-', mass, _ ).
noun( 'truancy', 'truancies', both, _ ).
noun( 'truant', 'truants', count, _ ).
noun( 'truce', 'truces', count, _ ).
noun( 'truck', 'trucks', both, _ ).
noun( 'truckle', 'truckles', count, _ ).
noun( 'truckle-bed', 'truckle-beds', count, _ ).
noun( 'truculence', '-', mass, _ ).
noun( 'truculency', '-', mass, _ ).
noun( 'trudge', 'trudges', count, _ ).
noun( 'true', 'trues', count, _ ).
noun( 'true-blue', 'true-blues', count, _ ).
noun( 'truelove', 'trueloves', count, _ ).
noun( 'truffle', 'truffles', count, _ ).
noun( 'truism', 'truisms', count, _ ).
noun( 'trump', 'trumps', count, _ ).
noun( 'trumpet', 'trumpets', count, _ ).
noun( 'trumpeter', 'trumpeters', count, _ ).
noun( 'truncheon', 'truncheons', count, _ ).
noun( 'trunk', 'trunks', count, _ ).
noun( 'trunk-call', 'trunk-calls', count, _ ).
noun( 'trunk-line', 'trunk-lines', count, _ ).
noun( 'trunk-road', 'trunk-roads', count, _ ).
noun( 'trunking', '-', mass, _ ).
noun( 'truss', 'trusses', count, _ ).
noun( 'trust', 'trusts', both, _ ).
noun( 'trustee', 'trustees', count, _ ).
noun( 'trusteeship', 'trusteeships', count, _ ).
noun( 'trustworthiness', '-', mass, _ ).
noun( 'truth', 'truths', both, _ ).
noun( 'truthfulness', '-', mass, _ ).
noun( 'try', 'tries', count, _ ).
noun( 'try-on', 'try-ons', count, _ ).
noun( 'try-out', 'try-outs', count, _ ).
noun( 'tryst', 'trysts', count, _ ).
noun( 'tsetse', 'tsetses', count, _ ).
noun( 'tsetse-fly', 'tsetse-flies', count, _ ).
noun( 'tub', 'tubs', count, _ ).
noun( 'tub-thumper', 'tub-thumpers', count, _ ).
noun( 'tuba', 'tubas', count, _ ).
noun( 'tube', 'tubes', both, _ ).
noun( 'tube-well', 'tube-wells', count, _ ).
noun( 'tuber', 'tubers', count, _ ).
noun( 'tuberculosis', '-', mass, _ ).
noun( 'tubful', 'tubfuls', count, _ ).
noun( 'tubing', '-', mass, _ ).
noun( 'tuck', 'tucks', both, _ ).
noun( 'tuck-in', 'tuck-ins', count, _ ).
noun( 'tuck-shop', 'tuck-shops', count, _ ).
noun( 'tucker', 'tuckers', count, _ ).
noun( 'tuft', 'tufts', count, _ ).
noun( 'tug', 'tugs', count, _ ).
noun( 'tugboat', 'tugboats', count, _ ).
noun( 'tuition', '-', mass, _ ).
noun( 'tulip', 'tulips', count, _ ).
noun( 'tulle', '-', mass, _ ).
noun( 'tumble', 'tumbles', count, _ ).
noun( 'tumbler', 'tumblers', count, _ ).
noun( 'tumbleweed', '-', mass, _ ).
noun( 'tumbrel', 'tumbrels', count, _ ).
noun( 'tumbril', 'tumbrils', count, _ ).
noun( 'tumescence', 'tumescences', both, _ ).
noun( 'tumidity', '-', mass, _ ).
noun( 'tummy', 'tummies', count, _ ).
noun( 'tumour', 'tumours', count, _ ).
noun( 'tumult', 'tumults', both, _ ).
noun( 'tumulus', 'tumuli', count, _ ).
noun( 'tun', 'tuns', count, _ ).
noun( 'tuna', 'tuna', both, _ ).
noun( 'tundra', 'tundras', both, _ ).
noun( 'tune', 'tunes', both, _ ).
noun( 'tunefulness', '-', mass, _ ).
noun( 'tuner', 'tuners', count, _ ).
noun( 'tung-oil', '-', mass, _ ).
noun( 'tungsten', '-', mass, _ ).
noun( 'tunic', 'tunics', count, _ ).
noun( 'tuning-fork', 'tuning-forks', count, _ ).
noun( 'tunnel', 'tunnels', count, _ ).
noun( 'tunny', 'tunny', count, _ ).
noun( 'tup', 'tups', count, _ ).
noun( 'tuppence', 'tuppences', count, _ ).
noun( 'turban', 'turbans', count, _ ).
noun( 'turbidity', '-', mass, _ ).
noun( 'turbidness', '-', mass, _ ).
noun( 'turbine', 'turbines', count, _ ).
noun( 'turbojet', 'turbojets', count, _ ).
noun( 'turboprop', 'turboprops', count, _ ).
noun( 'turbot', 'turbot', both, _ ).
noun( 'turbulence', 'turbulences', both, _ ).
noun( 'turd', 'turds', count, _ ).
noun( 'tureen', 'tureens', count, _ ).
noun( 'turf', 'turfs', both, _ ).
noun( 'turgidity', '-', mass, _ ).
noun( 'turkey', 'turkeys', both, _ ).
noun( 'turmeric', '-', mass, _ ).
noun( 'turmoil', 'turmoils', both, _ ).
noun( 'turn', 'turns', count, _ ).
noun( 'turn-off', 'turn-offs', count, _ ).
noun( 'turn-on', 'turn-ons', count, _ ).
noun( 'turn-out', 'turn-outs', count, _ ).
noun( 'turn-round', 'turn-rounds', count, _ ).
noun( 'turn-up', 'turn-ups', count, _ ).
noun( 'turncoat', 'turncoats', count, _ ).
noun( 'turncock', 'turncocks', count, _ ).
noun( 'turner', 'turners', count, _ ).
noun( 'turning', 'turnings', count, _ ).
noun( 'turning-point', 'turning-points', count, _ ).
noun( 'turnip', 'turnips', count, _ ).
noun( 'turnkey', 'turnkeys', count, _ ).
noun( 'turnover', 'turnovers', count, _ ).
noun( 'turnpike', 'turnpikes', count, _ ).
noun( 'turnspit', 'turnspits', count, _ ).
noun( 'turnstile', 'turnstiles', count, _ ).
noun( 'turntable', 'turntables', count, _ ).
noun( 'turpentine', '-', mass, _ ).
noun( 'turpitude', '-', mass, _ ).
noun( 'turps', '-', mass, _ ).
noun( 'turquoise', 'turquoises', both, _ ).
noun( 'turret', 'turrets', count, _ ).
noun( 'turtle', 'turtles', count, _ ).
noun( 'turtledove', 'turtledoves', count, _ ).
noun( 'tusk', 'tusks', count, _ ).
noun( 'tussle', 'tussles', count, _ ).
noun( 'tussock', 'tussocks', count, _ ).
noun( 'tutelage', '-', mass, _ ).
noun( 'tutor', 'tutors', count, _ ).
noun( 'tutorial', 'tutorials', count, _ ).
noun( 'tutorship', 'tutorships', count, _ ).
noun( 'tutti-frutti', 'tutti-fruttis', both, _ ).
noun( 'tutu', 'tutus', count, _ ).
noun( 'tuxedo', 'tuxedos', count, _ ).
noun( 'twaddle', '-', mass, _ ).
noun( 'twain', 'twains', count, _ ).
noun( 'twang', 'twangs', count, _ ).
noun( 'tweak', 'tweaks', count, _ ).
noun( 'tweed', 'tweeds', both, _ ).
noun( 'tweet', 'tweets', count, _ ).
noun( 'tweeter', 'tweeters', count, _ ).
noun( 'twelfth', 'twelfths', count, _ ).
noun( 'twelve', 'twelves', count, _ ).
noun( 'twelvemonth', 'twelvemonths', count, _ ).
noun( 'twentieth', 'twentieths', count, _ ).
noun( 'twenty', 'twenties', count, _ ).
noun( 'twerp', 'twerps', count, _ ).
noun( 'twiddle', 'twiddles', count, _ ).
noun( 'twig', 'twigs', count, _ ).
noun( 'twilight', '-', mass, _ ).
noun( 'twill', '-', mass, _ ).
noun( 'twin', 'twins', count, _ ).
noun( 'twine', '-', mass, _ ).
noun( 'twinge', 'twinges', count, _ ).
noun( 'twinkle', '-', mass, _ ).
noun( 'twinkling', '-', count, _ ).
noun( 'twirl', 'twirls', count, _ ).
noun( 'twist', 'twists', both, _ ).
noun( 'twister', 'twisters', count, _ ).
noun( 'twit', 'twits', count, _ ).
noun( 'twitch', 'twitches', count, _ ).
noun( 'twitter', 'twitters', count, _ ).
noun( 'two', 'twos', count, _ ).
noun( 'two-piece', '-', count, _ ).
noun( 'two-seater', 'two-seaters', count, _ ).
noun( 'two-step', 'two-steps', count, _ ).
noun( 'twopence', 'twopences', count, _ ).
noun( 'tycoon', 'tycoons', count, _ ).
noun( 'tyke', 'tykes', count, _ ).
noun( 'tympanum', 'tympanums', count, _ ).
noun( 'type', 'types', both, _ ).
noun( 'typeface', 'typefaces', count, _ ).
noun( 'typescript', 'typescripts', count, _ ).
noun( 'typesetter', 'typesetters', count, _ ).
noun( 'typewriter', 'typewriters', count, _ ).
noun( 'typhoid', '-', mass, _ ).
noun( 'typhoon', 'typhoons', count, _ ).
noun( 'typhus', '-', mass, _ ).
noun( 'typist', 'typists', count, _ ).
noun( 'typographer', 'typographers', count, _ ).
noun( 'typography', '-', mass, _ ).
noun( 'tyranny', 'tyrannies', both, _ ).
noun( 'tyrant', 'tyrants', count, _ ).
noun( 'tyre', 'tyres', count, _ ).
noun( 'tyro', 'tyros', count, _ ).
noun( 'tzar', 'tzars', count, _ ).
noun( 'tzarina', 'tzarinas', count, _ ).
noun( 'u', '-', count, _ ).
noun( 'ubiquity', '-', mass, _ ).
noun( 'udder', 'udders', count, _ ).
noun( 'ugliness', '-', mass, _ ).
noun( 'ukase', 'ukases', count, _ ).
noun( 'ukulele', 'ukuleles', count, _ ).
noun( 'ulcer', 'ulcers', count, _ ).
noun( 'ulceration', 'ulcerations', count, _ ).
noun( 'ulna', 'ulnae', count, _ ).
noun( 'ulster', 'ulsters', count, _ ).
noun( 'ultimatum', 'ultimatums', count, _ ).
noun( 'ultramarine', 'ultramarines', count, _ ).
noun( 'ululation', 'ululations', count, _ ).
noun( 'umber', 'umbers', count, _ ).
noun( 'umbrage', '-', mass, _ ).
noun( 'umbrella', 'umbrellas', count, _ ).
noun( 'umlaut', 'umlauts', count, _ ).
noun( 'umpire', 'umpires', count, _ ).
noun( 'unanimity', '-', mass, _ ).
noun( 'unbelief', '-', mass, _ ).
noun( 'unbeliever', 'unbelievers', count, _ ).
noun( 'unceremoniousness', '-', mass, _ ).
noun( 'uncertainty', 'uncertainties', both, _ ).
noun( 'uncle', 'uncles', count, _ ).
noun( 'unconcern', '-', mass, _ ).
noun( 'unconscious', 'unconsciouses', count, _ ).
noun( 'unconsciousness', '-', mass, _ ).
noun( 'unconventionality', '-', mass, _ ).
noun( 'uncouthness', '-', mass, _ ).
noun( 'unction', '-', mass, _ ).
noun( 'underbelly', 'underbellies', count, _ ).
noun( 'underbrush', '-', mass, _ ).
noun( 'undercarriage', 'undercarriages', count, _ ).
noun( 'undercharge', 'undercharges', count, _ ).
noun( 'underclothing', '-', mass, _ ).
noun( 'undercoat', 'undercoats', count, _ ).
noun( 'undercurrent', 'undercurrents', count, _ ).
noun( 'undercut', '-', mass, _ ).
noun( 'underdevelopment', '-', mass, _ ).
noun( 'underdog', 'underdogs', count, _ ).
noun( 'underestimate', 'underestimates', count, _ ).
noun( 'underestimation', 'underestimations', both, _ ).
noun( 'underexposure', 'underexposures', count, _ ).
noun( 'underfelt', '-', mass, _ ).
noun( 'undergarment', 'undergarments', count, _ ).
noun( 'undergraduate', 'undergraduates', count, _ ).
noun( 'underground', 'undergrounds', count, _ ).
noun( 'undergrowth', '-', mass, _ ).
noun( 'underlay', '-', mass, _ ).
noun( 'underline', 'underlines', count, _ ).
noun( 'underling', 'underlings', count, _ ).
noun( 'undernourishment', '-', mass, _ ).
noun( 'underpass', 'underpasses', count, _ ).
noun( 'underpayment', 'underpayments', both, _ ).
noun( 'underproduction', '-', mass, _ ).
noun( 'underseal', '-', mass, _ ).
noun( 'undersecretary', 'undersecretaries', count, _ ).
noun( 'underside', 'undersides', count, _ ).
noun( 'underskirt', 'underskirts', count, _ ).
noun( 'understanding', 'understandings', both, _ ).
noun( 'understatement', 'understatements', both, _ ).
noun( 'understudy', 'understudies', count, _ ).
noun( 'undersurface', 'undersurfaces', count, _ ).
noun( 'undertaker', 'undertakers', count, _ ).
noun( 'undertaking', 'undertakings', both, _ ).
noun( 'undertaking', 'undertakings', count, _ ).
noun( 'undertone', 'undertones', count, _ ).
noun( 'undertow', 'undertows', count, _ ).
noun( 'undervaluation', 'undervaluations', both, _ ).
noun( 'underwear', '-', mass, _ ).
noun( 'underworld', 'underworlds', count, _ ).
noun( 'underwriter', 'underwriters', count, _ ).
noun( 'undesirable', 'undesirables', count, _ ).
noun( 'undoing', 'undoings', count, _ ).
noun( 'undress', '-', mass, _ ).
noun( 'undulation', 'undulations', both, _ ).
noun( 'unease', '-', mass, _ ).
noun( 'uneasiness', '-', mass, _ ).
noun( 'unemployment', '-', mass, _ ).
noun( 'unexpectedness', '-', mass, _ ).
noun( 'unfairness', '-', mass, _ ).
noun( 'unfaithfulness', '-', mass, _ ).
noun( 'unfamiliarity', '-', mass, _ ).
noun( 'unfortunate', 'unfortunates', count, _ ).
noun( 'unfriendliness', '-', mass, _ ).
noun( 'ungratefulness', '-', mass, _ ).
noun( 'unguent', 'unguents', both, _ ).
noun( 'unhappiness', '-', mass, _ ).
noun( 'unicorn', 'unicorns', count, _ ).
noun( 'unification', '-', mass, _ ).
noun( 'uniform', 'uniforms', both, _ ).
noun( 'uniformity', '-', mass, _ ).
noun( 'unilateralism', '-', mass, _ ).
noun( 'unilateralist', 'unilateralists', count, _ ).
noun( 'union', 'unions', both, _ ).
noun( 'unionist', 'unionists', count, _ ).
noun( 'uniqueness', '-', mass, _ ).
noun( 'unison', '-', mass, _ ).
noun( 'unit', 'units', count, _ ).
noun( 'unity', 'unities', both, _ ).
noun( 'universality', '-', mass, _ ).
noun( 'universe', 'universes', count, _ ).
noun( 'university', 'universities', count, _ ).
noun( 'unknown', 'unknowns', count, _ ).
noun( 'unpleasantness', 'unpleasantnesses', both, _ ).
noun( 'unpopularity', '-', mass, _ ).
noun( 'unpredictability', '-', mass, _ ).
noun( 'unreality', '-', mass, _ ).
noun( 'unrest', '-', mass, _ ).
noun( 'unseen', 'unseens', count, _ ).
noun( 'unselfishness', '-', mass, _ ).
noun( 'unsightliness', '-', mass, _ ).
noun( 'unsuitability', '-', mass, _ ).
noun( 'untidiness', '-', mass, _ ).
noun( 'untouchable', 'untouchables', count, _ ).
noun( 'untruth', 'untruths', both, _ ).
noun( 'unwieldiness', '-', mass, _ ).
noun( 'unwillingness', '-', mass, _ ).
noun( 'unworthiness', '-', mass, _ ).
noun( 'up-beat', 'up-beats', count, _ ).
noun( 'upbraiding', 'upbraidings', count, _ ).
noun( 'upbringing', '-', mass, _ ).
noun( 'upgrade', 'upgrades', count, _ ).
noun( 'upheaval', 'upheavals', count, _ ).
noun( 'upholsterer', 'upholsterers', count, _ ).
noun( 'upholstery', '-', mass, _ ).
noun( 'upkeep', '-', mass, _ ).
noun( 'upland', 'uplands', count, _ ).
noun( 'uplift', '-', mass, _ ).
noun( 'upper', 'uppers', count, _ ).
noun( 'uppercut', 'uppercuts', count, _ ).
noun( 'uppishness', '-', mass, _ ).
noun( 'upright', 'uprights', count, _ ).
noun( 'uprightness', '-', mass, _ ).
noun( 'uprising', 'uprisings', count, _ ).
noun( 'uproar', 'uproars', both, _ ).
noun( 'upset', 'upsets', count, _ ).
noun( 'upshot', 'upshots', count, _ ).
noun( 'upstart', 'upstarts', count, _ ).
noun( 'upsurge', 'upsurges', count, _ ).
noun( 'uptake', 'uptakes', both, _ ).
noun( 'upturn', 'upturns', count, _ ).
noun( 'uranium', '-', mass, _ ).
noun( 'urbanity', '-', mass, _ ).
noun( 'urbanization', '-', mass, _ ).
noun( 'urchin', 'urchins', count, _ ).
noun( 'urge', 'urges', both, _ ).
noun( 'urgency', '-', mass, _ ).
noun( 'urging', 'urgings', count, _ ).
noun( 'urinal', 'urinals', count, _ ).
noun( 'urine', '-', mass, _ ).
noun( 'urn', 'urns', count, _ ).
noun( 'usage', 'usages', both, _ ).
noun( 'use', 'uses', both, _ ).
noun( 'usefulness', '-', mass, _ ).
noun( 'uselessness', '-', mass, _ ).
noun( 'user', 'users', count, _ ).
noun( 'usher', 'ushers', count, _ ).
noun( 'usherette', 'usherettes', count, _ ).
noun( 'usurer', 'usurers', count, _ ).
noun( 'usurpation', 'usurpations', both, _ ).
noun( 'usurper', 'usurpers', count, _ ).
noun( 'usury', '-', mass, _ ).
noun( 'utensil', 'utensils', count, _ ).
noun( 'uterus', 'uteruses', count, _ ).
noun( 'utilitarian', 'utilitarians', count, _ ).
noun( 'utilitarianism', '-', mass, _ ).
noun( 'utility', 'utilities', both, _ ).
noun( 'utilization', '-', mass, _ ).
noun( 'utmost', 'utmosts', count, _ ).
noun( 'utterance', 'utterances', both, _ ).
noun( 'uttermost', 'uttermosts', count, _ ).
noun( 'uvula', 'uvulas', count, _ ).
noun( 'uxoriousness', '-', mass, _ ).
noun( 'v', '-', count, _ ).
noun( 'vac', 'vacs', count, _ ).
noun( 'vacancy', 'vacancies', both, _ ).
noun( 'vacation', 'vacations', both, _ ).
noun( 'vacationist', 'vacationists', count, _ ).
noun( 'vaccination', 'vaccinations', both, _ ).
noun( 'vaccine', 'vaccines', both, _ ).
noun( 'vacillation', 'vacillations', both, _ ).
noun( 'vacuity', 'vacuities', both, _ ).
noun( 'vacuum', 'vacuums', count, _ ).
noun( 'vade-mecum', 'vade-mecums', count, _ ).
noun( 'vagabond', 'vagabonds', count, _ ).
noun( 'vagary', 'vagaries', count, _ ).
noun( 'vagina', 'vaginas', count, _ ).
noun( 'vagrancy', '-', mass, _ ).
noun( 'vagrant', 'vagrants', count, _ ).
noun( 'vagueness', '-', mass, _ ).
noun( 'vainglory', '-', mass, _ ).
noun( 'valance', 'valances', count, _ ).
noun( 'vale', 'vales', count, _ ).
noun( 'valediction', 'valedictions', count, _ ).
noun( 'valence', 'valences', count, _ ).
noun( 'valence', 'valences', both, _ ).
noun( 'valency', 'valencies', count, _ ).
noun( 'valentine', 'valentines', count, _ ).
noun( 'valerian', '-', mass, _ ).
noun( 'valet', 'valets', count, _ ).
noun( 'valetudinarian', 'valetudinarians', count, _ ).
noun( 'validity', '-', mass, _ ).
noun( 'valise', 'valises', count, _ ).
noun( 'valley', 'valleys', count, _ ).
noun( 'valour', '-', mass, _ ).
noun( 'valuable', 'valuables', count, _ ).
noun( 'valuation', 'valuations', both, _ ).
noun( 'value', 'values', both, _ ).
noun( 'valuer', 'valuers', count, _ ).
noun( 'valve', 'valves', count, _ ).
noun( 'vamp', 'vamps', count, _ ).
noun( 'vampire', 'vampires', count, _ ).
noun( 'van', 'vans', count, _ ).
noun( 'vandal', 'vandals', count, _ ).
noun( 'vandalism', '-', mass, _ ).
noun( 'vane', 'vanes', count, _ ).
noun( 'vanguard', 'vanguards', count, _ ).
noun( 'vanilla', 'vanillas', both, _ ).
noun( 'vanity', 'vanities', both, _ ).
noun( 'vantage', 'vantages', both, _ ).
noun( 'vantage-point', 'vantage-points', count, _ ).
noun( 'vapidity', '-', mass, _ ).
noun( 'vapidness', '-', mass, _ ).
noun( 'vaporization', 'vaporizations', both, _ ).
noun( 'vapour', 'vapours', both, _ ).
noun( 'vapour-bath', 'vapour-baths', count, _ ).
noun( 'variability', '-', mass, _ ).
noun( 'variable', 'variables', count, _ ).
noun( 'variableness', '-', mass, _ ).
noun( 'variance', 'variances', both, _ ).
noun( 'variant', 'variants', count, _ ).
noun( 'variation', 'variations', both, _ ).
noun( 'variegation', 'variegations', both, _ ).
noun( 'variety', 'varieties', both, _ ).
noun( 'varlet', 'varlets', count, _ ).
noun( 'varnish', 'varnishes', both, _ ).
noun( 'varsity', 'varsities', count, _ ).
noun( 'vase', 'vases', count, _ ).
noun( 'vasectomy', 'vasectomies', count, _ ).
noun( 'vaseline', '-', mass, _ ).
noun( 'vassal', 'vassals', count, _ ).
noun( 'vassalage', '-', mass, _ ).
noun( 'vastness', '-', mass, _ ).
noun( 'vat', 'vats', count, _ ).
noun( 'vaudeville', '-', mass, _ ).
noun( 'vault', 'vaults', count, _ ).
noun( 'vaulter', 'vaulters', count, _ ).
noun( 'vaulting-horse', 'vaulting-horses', count, _ ).
noun( 'vaunt', 'vaunts', count, _ ).
noun( 'vaunter', 'vaunters', count, _ ).
noun( 'veal', '-', mass, _ ).
noun( 'vector', 'vectors', count, _ ).
noun( 'vegetable', 'vegetables', count, _ ).
noun( 'vegetarian', 'vegetarians', count, _ ).
noun( 'vegetation', '-', mass, _ ).
noun( 'vehemence', '-', mass, _ ).
noun( 'vehicle', 'vehicles', count, _ ).
noun( 'veil', 'veils', count, _ ).
noun( 'veiling', '-', mass, _ ).
noun( 'vein', 'veins', count, _ ).
noun( 'veld', '-', mass, _ ).
noun( 'vellum', '-', mass, _ ).
noun( 'velocipede', 'velocipedes', count, _ ).
noun( 'velocity', 'velocities', both, _ ).
noun( 'velour', '-', mass, _ ).
noun( 'velvet', '-', mass, _ ).
noun( 'velveteen', '-', mass, _ ).
noun( 'venality', '-', mass, _ ).
noun( 'vendee', 'vendees', count, _ ).
noun( 'vender', 'venders', count, _ ).
noun( 'vendetta', 'vendettas', count, _ ).
noun( 'vendor', 'vendors', count, _ ).
noun( 'veneer', 'veneers', both, _ ).
noun( 'veneration', 'venerations', both, _ ).
noun( 'vengeance', '-', mass, _ ).
noun( 'venison', '-', mass, _ ).
noun( 'venom', '-', mass, _ ).
noun( 'vent', 'vents', count, _ ).
noun( 'vent-hole', 'vent-holes', count, _ ).
noun( 'ventilation', '-', mass, _ ).
noun( 'ventilator', 'ventilators', count, _ ).
noun( 'ventricle', 'ventricles', count, _ ).
noun( 'ventriloquism', '-', mass, _ ).
noun( 'ventriloquist', 'ventriloquists', count, _ ).
noun( 'venture', 'ventures', both, _ ).
noun( 'venue', 'venues', count, _ ).
noun( 'veracity', '-', mass, _ ).
noun( 'veranda', 'verandas', count, _ ).
noun( 'verandah', 'verandahs', count, _ ).
noun( 'verb', 'verbs', count, _ ).
noun( 'verbena', 'verbenas', count, _ ).
noun( 'verbiage', '-', mass, _ ).
noun( 'verboseness', '-', mass, _ ).
noun( 'verbosity', '-', mass, _ ).
noun( 'verdancy', '-', mass, _ ).
noun( 'verdict', 'verdicts', count, _ ).
noun( 'verdigris', '-', mass, _ ).
noun( 'verdure', '-', mass, _ ).
noun( 'verge', 'verges', count, _ ).
noun( 'verger', 'vergers', count, _ ).
noun( 'verification', 'verifications', both, _ ).
noun( 'verisimilitude', 'verisimilitudes', both, _ ).
noun( 'verity', 'verities', both, _ ).
noun( 'vermicelli', '-', mass, _ ).
noun( 'vermilion', 'vermilions', both, _ ).
noun( 'vermin', '-', mass, _ ).
noun( 'vermouth', 'vermouths', both, _ ).
noun( 'vernacular', 'vernaculars', count, _ ).
noun( 'veronica', 'veronicas', count, _ ).
noun( 'verruca', 'verrucas', count, _ ).
noun( 'versatility', '-', mass, _ ).
noun( 'verse', 'verses', both, _ ).
noun( 'versification', '-', mass, _ ).
noun( 'versifier', 'versifiers', count, _ ).
noun( 'version', 'versions', count, _ ).
noun( 'verso', 'versos', count, _ ).
noun( 'vertebra', 'vertebrae', count, _ ).
noun( 'vertebrate', 'vertebrates', count, _ ).
noun( 'vertex', 'vertices', count, _ ).
noun( 'vertical', 'verticals', count, _ ).
noun( 'vertigo', '-', mass, _ ).
noun( 'verve', '-', mass, _ ).
noun( 'vesicle', 'vesicles', count, _ ).
noun( 'vessel', 'vessels', count, _ ).
noun( 'vest', 'vests', count, _ ).
noun( 'vestal', 'vestals', count, _ ).
noun( 'vestibule', 'vestibules', count, _ ).
noun( 'vestige', 'vestiges', both, _ ).
noun( 'vestment', 'vestments', count, _ ).
noun( 'vestry', 'vestries', count, _ ).
noun( 'vestryman', 'vestrymen', count, _ ).
noun( 'vesture', 'vestures', count, _ ).
noun( 'vet', 'vets', count, _ ).
noun( 'vetch', 'vetches', both, _ ).
noun( 'veteran', 'veterans', count, _ ).
noun( 'veto', 'vetoes', count, _ ).
noun( 'vexation', 'vexations', both, _ ).
noun( 'via media', '-', count, _ ).
noun( 'viability', '-', mass, _ ).
noun( 'viaduct', 'viaducts', count, _ ).
noun( 'vial', 'vials', count, _ ).
noun( 'vibraphone', 'vibraphones', count, _ ).
noun( 'vibration', 'vibrations', both, _ ).
noun( 'vibrato', 'vibratos', both, _ ).
noun( 'vibrator', 'vibrators', count, _ ).
noun( 'vicar', 'vicars', count, _ ).
noun( 'vicarage', 'vicarages', count, _ ).
noun( 'vice', 'vices', both, _ ).
noun( 'vicereine', 'vicereines', count, _ ).
noun( 'viceroy', 'viceroys', count, _ ).
noun( 'vicinity', 'vicinities', both, _ ).
noun( 'viciousness', '-', mass, _ ).
noun( 'vicissitude', 'vicissitudes', count, _ ).
noun( 'victim', 'victims', count, _ ).
noun( 'victimization', 'victimizations', both, _ ).
noun( 'victor', 'victors', count, _ ).
noun( 'victoria', 'victorias', count, _ ).
noun( 'victory', 'victories', both, _ ).
noun( 'victual', 'victuals', count, _ ).
noun( 'victualler', 'victuallers', count, _ ).
noun( 'vicu~na', 'vicu~nas', count, _ ).
noun( 'video', 'videos', count, _ ).
noun( 'videotape', '-', mass, _ ).
noun( 'view', 'views', both, _ ).
noun( 'viewer', 'viewers', count, _ ).
noun( 'viewfinder', 'viewfinders', count, _ ).
noun( 'viewpoint', 'viewpoints', count, _ ).
noun( 'vigil', 'vigils', both, _ ).
noun( 'vigilance', '-', mass, _ ).
noun( 'vigilante', 'vigilantes', count, _ ).
noun( 'vignette', 'vignettes', count, _ ).
noun( 'vigour', '-', mass, _ ).
noun( 'vileness', '-', mass, _ ).
noun( 'vilification', 'vilifications', both, _ ).
noun( 'villa', 'villas', count, _ ).
noun( 'village', 'villages', count, _ ).
noun( 'villager', 'villagers', count, _ ).
noun( 'villain', 'villains', count, _ ).
noun( 'villainess', 'villainesss', count, _ ).
noun( 'villainy', 'villainies', both, _ ).
noun( 'villein', 'villeins', count, _ ).
noun( 'villeinage', '-', mass, _ ).
noun( 'vim', '-', mass, _ ).
noun( 'vinaigrette', '-', mass, _ ).
noun( 'vindication', 'vindications', both, _ ).
noun( 'vindictiveness', '-', mass, _ ).
noun( 'vine', 'vines', count, _ ).
noun( 'vinegar', 'vinegars', both, _ ).
noun( 'vinery', 'vineries', count, _ ).
noun( 'vineyard', 'vineyards', count, _ ).
noun( 'vino', 'vinoes', both, _ ).
noun( 'vintage', 'vintages', both, _ ).
noun( 'vintner', 'vintners', count, _ ).
noun( 'vinyl', 'vinyls', both, _ ).
noun( 'viol', 'viols', count, _ ).
noun( 'viola', 'violas', count, _ ).
noun( 'viola', 'violas', count, _ ).
noun( 'violation', 'violations', both, _ ).
noun( 'violence', '-', mass, _ ).
noun( 'violet', 'violets', both, _ ).
noun( 'violin', 'violins', count, _ ).
noun( 'violinist', 'violinists', count, _ ).
noun( 'viper', 'vipers', count, _ ).
noun( 'virago', 'viragos', count, _ ).
noun( 'virgin', 'virgins', count, _ ).
noun( 'virginal', 'virginals', count, _ ).
noun( 'virginity', '-', mass, _ ).
noun( 'virgule', 'virgules', count, _ ).
noun( 'virility', '-', mass, _ ).
noun( 'virology', '-', mass, _ ).
noun( 'virtu', '-', mass, _ ).
noun( 'virtue', 'virtues', both, _ ).
noun( 'virtuosity', '-', mass, _ ).
noun( 'virtuoso', 'virtuosos', count, _ ).
noun( 'virulence', '-', mass, _ ).
noun( 'virus', 'viruses', count, _ ).
noun( 'visa', 'visas', count, _ ).
noun( 'visage', 'visages', count, _ ).
noun( 'viscosity', '-', mass, _ ).
noun( 'viscount', 'viscounts', count, _ ).
noun( 'viscountcy', 'viscountcies', count, _ ).
noun( 'viscountess', 'viscountesses', count, _ ).
noun( 'visibility', '-', mass, _ ).
noun( 'vision', 'visions', both, _ ).
noun( 'visionary', 'visionaries', count, _ ).
noun( 'visit', 'visits', count, _ ).
noun( 'visitant', 'visitants', count, _ ).
noun( 'visitation', 'visitations', count, _ ).
noun( 'visiting', '-', mass, _ ).
noun( 'visitor', 'visitors', count, _ ).
noun( 'visor', 'visors', count, _ ).
noun( 'vista', 'vistas', count, _ ).
noun( 'visualization', 'visualizations', both, _ ).
noun( 'vitalism', '-', mass, _ ).
noun( 'vitalist', 'vitalists', count, _ ).
noun( 'vitality', '-', mass, _ ).
noun( 'vitamin', 'vitamins', count, _ ).
noun( 'vitriol', '-', mass, _ ).
noun( 'vituperation', '-', mass, _ ).
noun( 'viva', 'vivas', count, _ ).
noun( 'viva voce', 'viva voces', count, _ ).
noun( 'vivacity', '-', mass, _ ).
noun( 'vividness', '-', mass, _ ).
noun( 'vivisection', 'vivisections', both, _ ).
noun( 'vivisectionist', 'vivisectionists', count, _ ).
noun( 'vixen', 'vixens', count, _ ).
noun( 'viz', '-', proper, _ ).
noun( 'vizier', 'viziers', count, _ ).
noun( 'vocabulary', 'vocabularies', both, _ ).
noun( 'vocalist', 'vocalists', count, _ ).
noun( 'vocation', 'vocations', both, _ ).
noun( 'vocative', 'vocatives', count, _ ).
noun( 'vociferation', '-', mass, _ ).
noun( 'vodka', 'vodkas', both, _ ).
noun( 'vogue', 'vogues', count, _ ).
noun( 'voice', 'voices', both, _ ).
noun( 'void', 'voids', count, _ ).
noun( 'voile', '-', mass, _ ).
noun( 'vol', '-', count, _ ).
noun( 'volatility', 'volatilities', both, _ ).
noun( 'volcano', 'volcanos', count, _ ).
noun( 'vole', 'voles', count, _ ).
noun( 'volition', '-', mass, _ ).
noun( 'volley', 'volleys', count, _ ).
noun( 'volleyball', 'volleyballs', both, _ ).
noun( '-', 'vols', count, _ ).
noun( 'volt', 'volts', count, _ ).
noun( 'voltage', 'voltages', both, _ ).
noun( 'volte-face', '-', count, _ ).
noun( 'volubility', '-', mass, _ ).
noun( 'volume', 'volumes', both, _ ).
noun( 'voluntary', 'voluntaries', count, _ ).
noun( 'volunteer', 'volunteers', count, _ ).
noun( 'voluptuary', 'voluptuaries', count, _ ).
noun( 'voluptuousness', '-', mass, _ ).
noun( 'volute', 'volutes', count, _ ).
noun( 'vomit', '-', mass, _ ).
noun( 'voodoo', '-', mass, _ ).
noun( 'voodooism', '-', mass, _ ).
noun( 'voracity', '-', mass, _ ).
noun( 'vortex', 'vortexes', count, _ ).
noun( 'votary', 'votaries', count, _ ).
noun( 'vote', 'votes', count, _ ).
noun( 'voter', 'voters', count, _ ).
noun( 'voucher', 'vouchers', count, _ ).
noun( 'vow', 'vows', count, _ ).
noun( 'vowel', 'vowels', count, _ ).
noun( 'vox', '-', count, _ ).
noun( 'vox populi', '-', count, _ ).
noun( 'voyage', 'voyages', count, _ ).
noun( 'voyager', 'voyagers', count, _ ).
noun( 'voyeur', 'voyeurs', count, _ ).
noun( 'voyeurism', '-', mass, _ ).
noun( 'vs', '-', proper, _ ).
noun( 'vulcanite', '-', mass, _ ).
noun( 'vulcanization', 'vulcanizations', both, _ ).
noun( 'vulgarian', 'vulgarians', count, _ ).
noun( 'vulgarism', 'vulgarisms', both, _ ).
noun( 'vulgarity', 'vulgarities', both, _ ).
noun( 'vulgarization', '-', mass, _ ).
noun( 'vulnerability', 'vulnerabilities', both, _ ).
noun( 'vulture', 'vultures', count, _ ).
noun( 'vulva', 'vulvas', count, _ ).
noun( 'w', '-', count, _ ).
noun( 'wad', 'wads', count, _ ).
noun( 'wadding', '-', mass, _ ).
noun( 'waddle', '-', count, _ ).
noun( 'wader', 'waders', count, _ ).
noun( 'wadi', 'wadis', count, _ ).
noun( 'wafer', 'wafers', count, _ ).
noun( 'waffle', 'waffles', both, _ ).
noun( 'waft', 'wafts', count, _ ).
noun( 'wag', 'wags', count, _ ).
noun( 'wage', 'wages', count, _ ).
noun( 'wage-claim', 'wage-claims', count, _ ).
noun( 'wage-earner', 'wage-earners', count, _ ).
noun( 'wage-freeze', 'wage-freezes', count, _ ).
noun( 'wager', 'wagers', count, _ ).
noun( 'waggery', 'waggeries', both, _ ).
noun( 'waggishness', '-', mass, _ ).
noun( 'waggon', 'waggons', count, _ ).
noun( 'waggoner', 'waggoners', count, _ ).
noun( 'wagon-lit', 'wagons-lits', count, _ ).
noun( 'wagtail', 'wagtails', count, _ ).
noun( 'waif', 'waifs', count, _ ).
noun( 'wail', 'wails', count, _ ).
noun( 'wain', 'wains', count, _ ).
noun( 'wainscot', 'wainscots', count, _ ).
noun( 'waist', 'waists', count, _ ).
noun( 'waistband', 'waistbands', count, _ ).
noun( 'waistcoat', 'waistcoats', count, _ ).
noun( 'waistline', 'waistlines', count, _ ).
noun( 'wait', 'waits', both, _ ).
noun( 'waiter', 'waiters', count, _ ).
noun( 'waiting-list', 'waiting-lists', count, _ ).
noun( 'waiting-room', 'waiting-rooms', count, _ ).
noun( 'waitress', 'waitresses', count, _ ).
noun( 'waiver', 'waivers', count, _ ).
noun( 'wake', 'wakes', count, _ ).
noun( 'wakefulness', '-', mass, _ ).
noun( 'wale', 'wales', count, _ ).
noun( 'walk', 'walks', count, _ ).
noun( 'walkabout', 'walkabouts', count, _ ).
noun( 'walkaway', 'walkaways', count, _ ).
noun( 'walker', 'walkers', count, _ ).
noun( 'walkie-talkie', 'walkie-talkies', count, _ ).
noun( 'walking', '-', mass, _ ).
noun( 'walkout', 'walkouts', count, _ ).
noun( 'walkover', 'walkovers', count, _ ).
noun( 'wall', 'walls', count, _ ).
noun( 'wall-painting', 'wall-paintings', count, _ ).
noun( 'wallaby', 'wallabies', count, _ ).
noun( 'wallah', 'wallahs', count, _ ).
noun( 'wallet', 'wallets', count, _ ).
noun( 'wallflower', 'wallflowers', count, _ ).
noun( 'wallop', 'wallops', count, _ ).
noun( 'wallow', 'wallows', count, _ ).
noun( 'wallpaper', '-', mass, _ ).
noun( 'walnut', 'walnuts', both, _ ).
noun( 'walrus', 'walruses', count, _ ).
noun( 'waltz', 'waltzes', count, _ ).
noun( 'wampum', '-', mass, _ ).
noun( 'wand', 'wands', count, _ ).
noun( 'wanderer', 'wanderers', count, _ ).
noun( 'wanderlust', '-', mass, _ ).
noun( 'wane', 'wanes', count, _ ).
noun( 'wangle', 'wangles', count, _ ).
noun( 'wank', 'wanks', count, _ ).
noun( 'wanness', '-', mass, _ ).
noun( 'want', 'wants', both, _ ).
noun( 'want-ad', 'want-ads', count, _ ).
noun( 'wanton', 'wantons', count, _ ).
noun( 'wantonness', '-', mass, _ ).
noun( 'war', 'wars', both, _ ).
noun( 'war-baby', 'war-babies', count, _ ).
noun( 'war-bride', 'war-brides', count, _ ).
noun( 'war-cloud', 'war-clouds', count, _ ).
noun( 'war-cry', 'war-cries', count, _ ).
noun( 'war-dance', 'war-dances', count, _ ).
noun( 'war-god', 'war-gods', count, _ ).
noun( 'war-widow', 'war-widows', count, _ ).
noun( 'warble', 'warbles', count, _ ).
noun( 'warbler', 'warblers', count, _ ).
noun( 'ward', 'wards', both, _ ).
noun( 'warden', 'wardens', count, _ ).
noun( 'warder', 'warders', count, _ ).
noun( 'wardress', 'wardresses', count, _ ).
noun( 'wardrobe', 'wardrobes', count, _ ).
noun( 'wardroom', 'wardrooms', count, _ ).
noun( 'ware', 'wares', count, _ ).
noun( 'warehouse', 'warehouses', count, _ ).
noun( 'warehousing', '-', mass, _ ).
noun( 'warfare', '-', mass, _ ).
noun( 'warhead', 'warheads', count, _ ).
noun( 'warhorse', 'warhorses', count, _ ).
noun( 'wariness', '-', mass, _ ).
noun( 'warlord', 'warlords', count, _ ).
noun( 'warmer', 'warmers', count, _ ).
noun( 'warming-pan', 'warming-pans', count, _ ).
noun( 'warmonger', 'warmongers', count, _ ).
noun( 'warmth', '-', mass, _ ).
noun( 'warning', 'warnings', both, _ ).
noun( 'warp', 'warps', count, _ ).
noun( 'warpaint', '-', mass, _ ).
noun( 'warpath', 'warpaths', count, _ ).
noun( 'warrant', 'warrants', both, _ ).
noun( 'warrantee', 'warrantees', count, _ ).
noun( 'warrantor', 'warrantors', count, _ ).
noun( 'warranty', 'warranties', count, _ ).
noun( 'warren', 'warrens', count, _ ).
noun( 'warrior', 'warriors', count, _ ).
noun( 'warship', 'warships', count, _ ).
noun( 'wart', 'warts', count, _ ).
noun( 'warthog', 'warthogs', count, _ ).
noun( 'wartime', '-', mass, _ ).
noun( 'wash', 'washes', both, _ ).
noun( 'wash-drawing', 'wash-drawings', count, _ ).
noun( 'wash-hand-basin', 'wash-hand-basins', count, _ ).
noun( 'wash-hand-stand', 'wash-hand-stands', count, _ ).
noun( 'wash-house', 'wash-houses', count, _ ).
noun( 'wash-leather', 'wash-leathers', both, _ ).
noun( 'washbasin', 'washbasins', count, _ ).
noun( 'washboard', 'washboards', count, _ ).
noun( 'washbowl', 'washbowls', count, _ ).
noun( 'washcloth', 'washcloths', count, _ ).
noun( 'washday', 'washdays', count, _ ).
noun( 'washer', 'washers', count, _ ).
noun( 'washerwoman', 'washerwomen', count, _ ).
noun( 'washing', '-', mass, _ ).
noun( 'washing-day', 'washing-days', count, _ ).
noun( 'washing-machine', 'washing-machines', count, _ ).
noun( 'washing-up', 'washing-ups', both, _ ).
noun( 'washout', 'washouts', count, _ ).
noun( 'washroom', 'washrooms', count, _ ).
noun( 'washstand', 'washstands', count, _ ).
noun( 'washtub', 'washtubs', count, _ ).
noun( 'wasp', 'wasps', count, _ ).
noun( 'wassail', 'wassails', count, _ ).
noun( 'wastage', '-', mass, _ ).
noun( 'waste', 'wastes', both, _ ).
noun( 'waste-paper-basket', 'waste-paper-baskets', count, _ ).
noun( 'waste-pipe', 'waste-pipes', count, _ ).
noun( 'wastebasket', 'wastebaskets', count, _ ).
noun( 'wastebin', 'wastebins', count, _ ).
noun( 'wasteland', 'wastelands', count, _ ).
noun( 'waster', 'wasters', count, _ ).
noun( 'wastrel', 'wastrels', count, _ ).
noun( 'watch', 'watches', both, _ ).
noun( 'watch-chain', 'watch-chains', count, _ ).
noun( 'watch-glass', 'watch-glasses', count, _ ).
noun( 'watch-guard', 'watch-guards', count, _ ).
noun( 'watch-key', 'watch-keys', count, _ ).
noun( 'watchdog', 'watchdogs', count, _ ).
noun( 'watcher', 'watchers', count, _ ).
noun( 'watchfulness', '-', mass, _ ).
noun( 'watchmaker', 'watchmakers', count, _ ).
noun( 'watchman', 'watchmen', count, _ ).
noun( 'watchtower', 'watchtowers', count, _ ).
noun( 'watchword', 'watchwords', count, _ ).
noun( 'water', 'waters', both, _ ).
noun( 'water-biscuit', 'water-biscuits', count, _ ).
noun( 'water-blister', 'water-blisters', count, _ ).
noun( 'water-bottle', 'water-bottles', count, _ ).
noun( 'water-buffalo', 'water-buffalos', count, _ ).
noun( 'water-butt', 'water-butts', count, _ ).
noun( 'water-cart', 'water-carts', count, _ ).
noun( 'water-closet', 'water-closets', count, _ ).
noun( 'water-finder', 'water-finders', count, _ ).
noun( 'water-glass', '-', mass, _ ).
noun( 'water-hole', 'water-holes', count, _ ).
noun( 'water-ice', 'water-ices', both, _ ).
noun( 'water-jacket', 'water-jackets', count, _ ).
noun( 'water-level', 'water-levels', count, _ ).
noun( 'water-lily', 'water-lilies', count, _ ).
noun( 'water-line', 'water-lines', count, _ ).
noun( 'water-main', 'water-mains', count, _ ).
noun( 'water-nymph', 'water-nymphs', count, _ ).
noun( 'water-polo', '-', mass, _ ).
noun( 'water-power', '-', mass, _ ).
noun( 'water-rat', 'water-rats', count, _ ).
noun( 'water-rate', 'water-rates', count, _ ).
noun( 'water-skiing', '-', mass, _ ).
noun( 'water-skin', 'water-skins', count, _ ).
noun( 'water-softener', 'water-softeners', both, _ ).
noun( 'water-spaniel', 'water-spaniels', count, _ ).
noun( 'water-supply', 'water-supplies', count, _ ).
noun( 'water-tower', 'water-towers', count, _ ).
noun( 'water-vole', 'water-voles', count, _ ).
noun( 'water-waggon', 'water-waggons', count, _ ).
noun( 'water-wagon', 'water-wagons', count, _ ).
noun( 'water-wheel', 'water-wheels', count, _ ).
noun( 'waterbird', 'waterbirds', count, _ ).
noun( 'watercannon', 'watercannons', count, _ ).
noun( 'waterchute', 'waterchutes', count, _ ).
noun( 'watercolour', 'watercolours', both, _ ).
noun( 'watercolourist', 'watercolourists', count, _ ).
noun( 'watercourse', 'watercourses', count, _ ).
noun( 'watercress', '-', mass, _ ).
noun( 'waterfall', 'waterfalls', count, _ ).
noun( 'waterfowl', 'waterfowls', count, _ ).
noun( 'waterfront', 'waterfronts', count, _ ).
noun( 'waterhen', 'waterhens', count, _ ).
noun( 'waterhyacinth', 'waterhyacinths', count, _ ).
noun( 'watering-can', 'watering-cans', count, _ ).
noun( 'watering-cart', 'watering-carts', count, _ ).
noun( 'waterman', 'watermen', count, _ ).
noun( 'watermark', 'watermarks', count, _ ).
noun( 'watermelon', 'watermelons', count, _ ).
noun( 'watermill', 'watermills', count, _ ).
noun( 'waterproof', 'waterproofs', count, _ ).
noun( 'watershed', 'watersheds', count, _ ).
noun( 'waterside', 'watersides', count, _ ).
noun( 'waterspout', 'waterspouts', count, _ ).
noun( 'watertable', 'watertables', count, _ ).
noun( 'waterway', 'waterways', count, _ ).
noun( 'waterworks', 'waterworks', count, _ ).
noun( 'watt', 'watts', count, _ ).
noun( 'wattage', 'wattages', both, _ ).
noun( 'wattle', 'wattles', both, _ ).
noun( 'wave', 'waves', count, _ ).
noun( 'wavelength', 'wavelengths', count, _ ).
noun( 'waverer', 'waverers', count, _ ).
noun( 'wax', 'waxes', both, _ ).
noun( 'wax-chandler', 'wax-chandlers', count, _ ).
noun( 'wax-paper', 'wax-papers', count, _ ).
noun( 'waxwork', 'waxworks', count, _ ).
noun( 'way', 'ways', both, _ ).
noun( 'waybill', 'waybills', count, _ ).
noun( 'wayfarer', 'wayfarers', count, _ ).
noun( 'wayside', 'waysides', count, _ ).
noun( 'weakling', 'weaklings', count, _ ).
noun( 'weakness', 'weaknesses', both, _ ).
noun( 'weal', 'weals', both, _ ).
noun( 'weald', 'wealds', count, _ ).
noun( 'wealth', '-', mass, _ ).
noun( 'weapon', 'weapons', count, _ ).
noun( 'wear', '-', mass, _ ).
noun( 'wearer', 'wearers', count, _ ).
noun( 'weariness', '-', mass, _ ).
noun( 'weasel', 'weasels', count, _ ).
noun( 'weather', 'weathers', both, _ ).
noun( 'weather-bureau', 'weather-bureaus', count, _ ).
noun( 'weather-chart', 'weather-charts', count, _ ).
noun( 'weather-glass', 'weather-glasses', count, _ ).
noun( 'weather-map', 'weather-maps', count, _ ).
noun( 'weather-ship', 'weather-ships', count, _ ).
noun( 'weather-station', 'weather-stations', count, _ ).
noun( 'weather-vane', 'weather-vanes', count, _ ).
noun( 'weatherboarding', '-', mass, _ ).
noun( 'weathercock', 'weathercocks', count, _ ).
noun( 'weatherman', 'weathermen', count, _ ).
noun( 'weave', 'weaves', count, _ ).
noun( 'weaver', 'weavers', count, _ ).
noun( 'weaverbird', 'weaverbirds', count, _ ).
noun( 'web', 'webs', count, _ ).
noun( 'webbing', '-', mass, _ ).
noun( 'wedding', 'weddings', count, _ ).
noun( 'wedding-cake', 'wedding-cakes', both, _ ).
noun( 'wedding-ring', 'wedding-rings', count, _ ).
noun( 'wedge', 'wedges', count, _ ).
noun( 'wedlock', '-', mass, _ ).
noun( 'wee', 'wees', both, _ ).
noun( 'wee-wee', 'wee-wees', both, _ ).
noun( 'weed', 'weeds', count, _ ).
noun( 'weedkiller', 'weedkillers', both, _ ).
noun( 'week', 'weeks', count, _ ).
noun( 'weekday', 'weekdays', count, _ ).
noun( 'weekend', 'weekends', count, _ ).
noun( 'weekender', 'weekenders', count, _ ).
noun( 'weekly', 'weeklies', count, _ ).
noun( 'weevil', 'weevils', count, _ ).
noun( 'weft', 'wefts', count, _ ).
noun( 'weighbridge', 'weighbridges', count, _ ).
noun( 'weighing-machine', 'weighing-machines', count, _ ).
noun( 'weight', 'weights', both, _ ).
noun( 'weightiness', '-', mass, _ ).
noun( 'weightlessness', '-', mass, _ ).
noun( 'weightlifting', '-', mass, _ ).
noun( 'weir', 'weirs', count, _ ).
noun( 'weirdie', 'weirdies', count, _ ).
noun( 'weirdness', '-', mass, _ ).
noun( 'welcome', 'welcomes', count, _ ).
noun( 'weld', 'welds', count, _ ).
noun( 'welder', 'welders', count, _ ).
noun( 'welfare', '-', mass, _ ).
noun( 'welkin', '-', count, _ ).
noun( 'well', 'wells', count, _ ).
noun( 'well-being', '-', mass, _ ).
noun( 'well-doer', 'well-doers', count, _ ).
noun( 'well-doing', '-', mass, _ ).
noun( 'well-water', '-', mass, _ ).
noun( 'well-wisher', 'well-wishers', count, _ ).
noun( 'wellhead', 'wellheads', count, _ ).
noun( 'wellington', 'wellingtons', count, _ ).
noun( 'welsher', 'welshers', count, _ ).
noun( 'welt', 'welts', count, _ ).
noun( 'welter', '-', count, _ ).
noun( 'welterweight', 'welterweights', count, _ ).
noun( 'wen', 'wens', count, _ ).
noun( 'wench', 'wenches', count, _ ).
noun( 'werewolf', 'werewolves', count, _ ).
noun( 'west', '-', mass, _ ).
noun( 'western', 'westerns', count, _ ).
noun( 'westerner', 'westerners', count, _ ).
noun( 'westernization', '-', mass, _ ).
noun( 'wet', '-', mass, _ ).
noun( 'wet-nurse', 'wet-nurses', count, _ ).
noun( 'wether', 'wethers', count, _ ).
noun( 'wetting', 'wettings', count, _ ).
noun( 'whack', 'whacks', count, _ ).
noun( 'whacker', 'whackers', count, _ ).
noun( 'whacking', 'whackings', count, _ ).
noun( 'whale', 'whales', count, _ ).
noun( 'whalebone', '-', mass, _ ).
noun( 'whaler', 'whalers', count, _ ).
noun( 'whaling-gun', 'whaling-guns', count, _ ).
noun( 'whang', 'whangs', count, _ ).
noun( 'wharf', 'wharfs', count, _ ).
noun( 'wharfage', '-', mass, _ ).
noun( 'what-for', '-', mass, _ ).
noun( 'whatnot', '-', mass, _ ).
noun( 'wheat', '-', mass, _ ).
noun( 'wheel', 'wheels', count, _ ).
noun( 'wheelbarrow', 'wheelbarrows', count, _ ).
noun( 'wheelbase', 'wheelbases', count, _ ).
noun( 'wheelchair', 'wheelchairs', count, _ ).
noun( 'wheelhouse', '-', count, _ ).
noun( 'wheelwright', 'wheelwrights', count, _ ).
noun( 'wheeze', 'wheezes', count, _ ).
noun( 'wheeziness', '-', mass, _ ).
noun( 'whelk', 'whelks', count, _ ).
noun( 'whelp', 'whelps', count, _ ).
noun( 'wherewithal', '-', mass, _ ).
noun( 'wherry', 'wherries', count, _ ).
noun( 'whetstone', 'whetstones', count, _ ).
noun( 'whey', '-', mass, _ ).
noun( 'whiff', 'whiffs', count, _ ).
noun( 'while', '-', count, _ ).
noun( 'whim', 'whims', count, _ ).
noun( 'whimper', 'whimpers', count, _ ).
noun( 'whimsey', 'whimseys', both, _ ).
noun( 'whimsicality', 'whimsicalities', both, _ ).
noun( 'whimsy', 'whimsies', both, _ ).
noun( 'whin', '-', mass, _ ).
noun( 'whine', 'whines', count, _ ).
noun( 'whiner', 'whiners', count, _ ).
noun( 'whinny', 'whinnies', count, _ ).
noun( 'whip', 'whips', count, _ ).
noun( 'whip-round', 'whip-rounds', count, _ ).
noun( 'whipcord', '-', mass, _ ).
noun( 'whipper-in', 'whippers-in', count, _ ).
noun( 'whippersnapper', 'whippersnappers', count, _ ).
noun( 'whippet', 'whippets', count, _ ).
noun( 'whipping', 'whippings', both, _ ).
noun( 'whipping-boy', 'whipping-boys', count, _ ).
noun( 'whipping-post', 'whipping-posts', count, _ ).
noun( 'whipping-top', 'whipping-tops', count, _ ).
noun( 'whippoorwill', 'whippoorwills', count, _ ).
noun( 'whir', '-', count, _ ).
noun( 'whirl', '-', count, _ ).
noun( 'whirligig', 'whirligigs', count, _ ).
noun( 'whirlpool', 'whirlpools', count, _ ).
noun( 'whirlwind', 'whirlwinds', count, _ ).
noun( 'whirr', '-', count, _ ).
noun( 'whisk', 'whisks', count, _ ).
noun( 'whisker', 'whiskers', count, _ ).
noun( 'whiskey', 'whiskeys', both, _ ).
noun( 'whisky', 'whiskies', both, _ ).
noun( 'whisper', 'whispers', count, _ ).
noun( 'whisperer', 'whisperers', count, _ ).
noun( 'whispering', 'whisperings', count, _ ).
noun( 'whispering-gallery', 'whispering-galleries', count, _ ).
noun( 'whist', '-', mass, _ ).
noun( 'whist-drive', 'whist-drives', count, _ ).
noun( 'whistle', 'whistles', count, _ ).
noun( 'whistle-stop', 'whistle-stops', count, _ ).
noun( 'whit', '-', count, _ ).
noun( 'white', 'whites', both, _ ).
noun( 'whitebait', '-', mass, _ ).
noun( 'whiteness', '-', mass, _ ).
noun( 'whitening', '-', mass, _ ).
noun( 'whitethorn', 'whitethorns', both, _ ).
noun( 'whitewash', '-', mass, _ ).
noun( 'whiting', 'whiting', both, _ ).
noun( 'whitlow', 'whitlows', count, _ ).
noun( 'whiz', '-', mass, _ ).
noun( 'whizz-kid', 'whizz-kids', count, _ ).
noun( 'whodunit', 'whodunits', count, _ ).
noun( 'whole', 'wholes', count, _ ).
noun( 'whole-wheat', '-', mass, _ ).
noun( 'wholeheartedness', '-', mass, _ ).
noun( 'wholemeal', '-', mass, _ ).
noun( 'wholesale', '-', mass, _ ).
noun( 'wholesaler', 'wholesalers', count, _ ).
noun( 'whoop', 'whoops', count, _ ).
noun( 'whoopee', 'whoopees', count, _ ).
noun( 'whooping-cough', '-', mass, _ ).
noun( 'whopper', 'whoppers', count, _ ).
noun( 'whore', 'whores', count, _ ).
noun( 'whoremonger', 'whoremongers', count, _ ).
noun( 'whorl', 'whorls', count, _ ).
noun( 'why', 'whys', count, _ ).
noun( 'wick', 'wicks', both, _ ).
noun( 'wickedness', '-', mass, _ ).
noun( 'wicker', '-', mass, _ ).
noun( 'wickerwork', '-', mass, _ ).
noun( 'wicket', 'wickets', count, _ ).
noun( 'wicket-door', 'wicket-doors', count, _ ).
noun( 'wicket-gate', 'wicket-gates', count, _ ).
noun( 'wicket-keeper', 'wicket-keepers', count, _ ).
noun( 'widgeon', 'widgeons', count, _ ).
noun( 'widow', 'widows', count, _ ).
noun( 'widower', 'widowers', count, _ ).
noun( 'widowhood', '-', mass, _ ).
noun( 'width', 'widths', both, _ ).
noun( 'wife', 'wives', count, _ ).
noun( 'wig', 'wigs', count, _ ).
noun( 'wigging', 'wiggings', count, _ ).
noun( 'wiggle', 'wiggles', count, _ ).
noun( 'wight', 'wights', count, _ ).
noun( 'wigwam', 'wigwams', count, _ ).
noun( 'wildebeest', 'wildebeests', count, _ ).
noun( 'wilderness', 'wildernesses', count, _ ).
noun( 'wildfire', '-', mass, _ ).
noun( 'wildness', '-', mass, _ ).
noun( 'wile', 'wiles', count, _ ).
noun( 'wilfulness', '-', mass, _ ).
noun( 'will', 'wills', both, _ ).
noun( 'will-o\'-the-wisp', 'will-o\'-the-wisps', count, _ ).
noun( 'willingness', '-', mass, _ ).
noun( 'willow', 'willows', both, _ ).
noun( 'willow-pattern', '-', mass, _ ).
noun( 'willow-tree', 'willow-trees', count, _ ).
noun( 'willpower', '-', mass, _ ).
noun( 'wimple', 'wimples', count, _ ).
noun( 'win', 'wins', count, _ ).
noun( 'wince', 'winces', count, _ ).
noun( 'winceyette', '-', mass, _ ).
noun( 'winch', 'winches', count, _ ).
noun( 'wind', 'winds', both, _ ).
noun( 'wind', 'winds', count, _ ).
noun( 'wind-gauge', 'wind-gauges', count, _ ).
noun( 'wind-tunnel', 'wind-tunnels', count, _ ).
noun( 'windbag', 'windbags', count, _ ).
noun( 'windbreak', 'windbreaks', count, _ ).
noun( 'windbreaker', 'windbreakers', count, _ ).
noun( 'windcheater', 'windcheaters', count, _ ).
noun( 'windfall', 'windfalls', count, _ ).
noun( 'windflower', 'windflowers', count, _ ).
noun( 'windiness', '-', mass, _ ).
noun( 'winding-sheet', 'winding-sheets', count, _ ).
noun( 'windjammer', 'windjammers', count, _ ).
noun( 'windlass', 'windlasses', count, _ ).
noun( 'windmill', 'windmills', count, _ ).
noun( 'window', 'windows', count, _ ).
noun( 'window-box', 'window-boxes', count, _ ).
noun( 'window-dressing', '-', mass, _ ).
noun( 'windowpane', 'windowpanes', count, _ ).
noun( 'windowsill', 'windowsills', count, _ ).
noun( 'windpipe', 'windpipes', count, _ ).
noun( 'windscreen', 'windscreens', count, _ ).
noun( 'windscreen-wiper', 'windscreen-wipers', count, _ ).
noun( 'windshield', 'windshields', count, _ ).
noun( 'windsock', 'windsocks', count, _ ).
noun( 'windward', '-', mass, _ ).
noun( 'wine', 'wines', both, _ ).
noun( 'wineglass', 'wineglasses', count, _ ).
noun( 'winepress', 'winepresses', count, _ ).
noun( 'wineskin', 'wineskins', count, _ ).
noun( 'wing', 'wings', count, _ ).
noun( 'wing-commander', 'wing-commanders', count, _ ).
noun( 'wing-nut', 'wing-nuts', count, _ ).
noun( 'wing-screw', 'wing-screws', count, _ ).
noun( 'winger', 'wingers', count, _ ).
noun( 'wingspan', 'wingspans', count, _ ).
noun( 'wingspread', 'wingspreads', count, _ ).
noun( 'wink', 'winks', count, _ ).
noun( 'winkle', 'winkles', count, _ ).
noun( 'winner', 'winners', count, _ ).
noun( 'winning-post', 'winning-posts', count, _ ).
noun( 'winsomeness', '-', mass, _ ).
noun( 'winter', 'winters', both, _ ).
noun( 'wipe', 'wipes', count, _ ).
noun( 'wiper', 'wipers', count, _ ).
noun( 'wire', 'wires', both, _ ).
noun( 'wireless', 'wirelesses', both, _ ).
noun( 'wirepuller', 'wirepullers', count, _ ).
noun( 'wireworm', 'wireworms', count, _ ).
noun( 'wiring', '-', mass, _ ).
noun( 'wisdom', '-', mass, _ ).
noun( 'wisdom-tooth', 'wisdom-teeth', count, _ ).
noun( 'wise', '-', count, _ ).
noun( 'wiseacre', 'wiseacres', count, _ ).
noun( 'wisecrack', 'wisecracks', count, _ ).
noun( 'wish', 'wishes', both, _ ).
noun( 'wishbone', 'wishbones', count, _ ).
noun( 'wishing-cap', 'wishing-caps', count, _ ).
noun( 'wisp', 'wisps', count, _ ).
noun( 'wisteria', 'wisterias', both, _ ).
noun( 'wit', 'wits', both, _ ).
noun( 'witch', 'witches', count, _ ).
noun( 'witch-doctor', 'witch-doctors', count, _ ).
noun( 'witch-elm', 'witch-elms', count, _ ).
noun( 'witch-hazel', 'witch-hazels', both, _ ).
noun( 'witch-hunt', 'witch-hunts', count, _ ).
noun( 'witchcraft', '-', mass, _ ).
noun( 'witchery', '-', mass, _ ).
noun( 'withdrawal', 'withdrawals', both, _ ).
noun( 'withe', 'withes', count, _ ).
noun( 'withy', 'withies', count, _ ).
noun( 'witness', 'witnesses', both, _ ).
noun( 'witness-box', 'witness-boxes', count, _ ).
noun( 'witness-stand', 'witness-stands', count, _ ).
noun( 'witticism', 'witticisms', count, _ ).
noun( 'wizard', 'wizards', count, _ ).
noun( 'wizardry', '-', mass, _ ).
noun( 'wk', 'wk', count, _ ).
noun( 'woad', '-', mass, _ ).
noun( 'wobbler', 'wobblers', count, _ ).
noun( 'woe', 'woes', both, _ ).
noun( 'wold', 'wolds', both, _ ).
noun( 'wolf', 'wolves', count, _ ).
noun( 'wolf\'s-bane', '-', mass, _ ).
noun( 'wolf-cub', 'wolf-cubs', count, _ ).
noun( 'wolfhound', 'wolfhounds', count, _ ).
noun( 'wolfram', '-', mass, _ ).
noun( 'woman', 'women', count, _ ).
noun( 'womanhood', '-', mass, _ ).
noun( 'womanizer', 'womanizers', count, _ ).
noun( 'womankind', '-', mass, _ ).
noun( 'womb', 'wombs', count, _ ).
noun( 'wombat', 'wombats', count, _ ).
noun( 'wonder', 'wonders', both, _ ).
noun( 'wonderland', 'wonderlands', count, _ ).
noun( 'wonderment', '-', mass, _ ).
noun( 'wont', '-', mass, _ ).
noun( 'wood', 'woods', both, _ ).
noun( 'wood-block', 'wood-blocks', count, _ ).
noun( 'wood-pulp', '-', mass, _ ).
noun( 'woodbine', 'woodbines', both, _ ).
noun( 'woodcock', 'woodcocks', count, _ ).
noun( 'woodcraft', 'woodcrafts', both, _ ).
noun( 'woodcut', 'woodcuts', count, _ ).
noun( 'woodcutter', 'woodcutters', count, _ ).
noun( 'woodland', 'woodlands', both, _ ).
noun( 'woodlouse', 'woodlice', count, _ ).
noun( 'woodman', 'woodmen', count, _ ).
noun( 'woodpecker', 'woodpeckers', count, _ ).
noun( 'woodpile', 'woodpiles', count, _ ).
noun( 'woodshed', 'woodsheds', count, _ ).
noun( 'woodsman', 'woodsmen', count, _ ).
noun( 'woodwind', '-', mass, _ ).
noun( 'woodwork', '-', mass, _ ).
noun( 'woodworm', '-', mass, _ ).
noun( 'wooer', 'wooers', count, _ ).
noun( 'woof', 'woofs', count, _ ).
noun( 'woofer', 'woofers', count, _ ).
noun( 'wool', 'wools', both, _ ).
noun( 'woolgathering', '-', mass, _ ).
noun( 'wooly', 'woolies', count, _ ).
noun( 'word', 'words', count, _ ).
noun( 'word-division', '-', mass, _ ).
noun( 'word-painter', 'word-painters', count, _ ).
noun( 'word-picture', 'word-pictures', count, _ ).
noun( 'word-splitting', '-', mass, _ ).
noun( 'wordbook', 'wordbooks', count, _ ).
noun( 'wordiness', '-', mass, _ ).
noun( 'wording', '-', count, _ ).
noun( 'work', 'works', both, _ ).
noun( 'work-in', 'work-ins', count, _ ).
noun( 'work-out', 'work-outs', count, _ ).
noun( 'work-study', 'work-studies', count, _ ).
noun( 'workbag', 'workbags', count, _ ).
noun( 'workbasket', 'workbaskets', count, _ ).
noun( 'workbench', 'workbenches', count, _ ).
noun( 'workbook', 'workbooks', count, _ ).
noun( 'workbox', 'workboxes', count, _ ).
noun( 'workday', 'workdays', count, _ ).
noun( 'worker', 'workers', count, _ ).
noun( 'workhouse', '-', count, _ ).
noun( 'working', 'workings', count, _ ).
noun( 'working-out', '-', mass, _ ).
noun( 'workman', 'workmen', count, _ ).
noun( 'workmanship', '-', mass, _ ).
noun( 'workmate', 'workmates', count, _ ).
noun( 'workroom', 'workrooms', count, _ ).
noun( 'workshop', 'workshops', count, _ ).
noun( 'worktable', 'worktables', count, _ ).
noun( 'world', 'worlds', count, _ ).
noun( 'worldliness', '-', mass, _ ).
noun( 'worm', 'worms', count, _ ).
noun( 'worm-gear', 'worm-gears', count, _ ).
noun( 'wormcast', 'wormcasts', count, _ ).
noun( 'wormhole', 'wormholes', count, _ ).
noun( 'wormwood', '-', mass, _ ).
noun( 'worry', 'worries', both, _ ).
noun( 'worse', '-', mass, _ ).
noun( 'worship', '-', mass, _ ).
noun( 'worshipper', 'worshippers', count, _ ).
noun( 'worst', '-', mass, _ ).
noun( 'worsted', '-', mass, _ ).
noun( 'worth', '-', mass, _ ).
noun( 'worthiness', '-', mass, _ ).
noun( 'worthlessness', '-', mass, _ ).
noun( 'worthy', 'worthies', count, _ ).
noun( 'wound', 'wounds', count, _ ).
noun( 'wow', 'wows', both, _ ).
noun( 'wpb', '-', count, _ ).
noun( 'wpm', 'wpm', count, _ ).
noun( 'wrack', '-', mass, _ ).
noun( 'wraith', 'wraiths', count, _ ).
noun( 'wrangle', 'wrangles', count, _ ).
noun( 'wrap', 'wraps', count, _ ).
noun( 'wrapper', 'wrappers', count, _ ).
noun( 'wrapping', 'wrappings', both, _ ).
noun( 'wrath', '-', mass, _ ).
noun( 'wreath', 'wreaths', count, _ ).
noun( 'wreck', 'wrecks', both, _ ).
noun( 'wreckage', '-', mass, _ ).
noun( 'wrecker', 'wreckers', count, _ ).
noun( 'wren', 'wrens', count, _ ).
noun( 'wrench', 'wrenches', count, _ ).
noun( 'wrestle', 'wrestles', count, _ ).
noun( 'wrestler', 'wrestlers', count, _ ).
noun( 'wretch', 'wretches', count, _ ).
noun( 'wretchedness', '-', mass, _ ).
noun( 'wrick', 'wricks', count, _ ).
noun( 'wriggle', 'wriggles', count, _ ).
noun( 'wriggler', 'wrigglers', count, _ ).
noun( 'wright', 'wrights', count, _ ).
noun( 'wring', 'wrings', count, _ ).
noun( 'wringer', 'wringers', count, _ ).
noun( 'wrinkle', 'wrinkles', count, _ ).
noun( 'wrist', 'wrists', count, _ ).
noun( 'wristband', 'wristbands', count, _ ).
noun( 'wristlet', 'wristlets', count, _ ).
noun( 'wristwatch', 'wristwatches', count, _ ).
noun( 'writ', 'writs', count, _ ).
noun( 'write-off', 'write-offs', count, _ ).
noun( 'write-up', 'write-ups', count, _ ).
noun( 'writer', 'writers', count, _ ).
noun( 'writing', 'writings', both, _ ).
noun( 'writing-desk', 'writing-desks', count, _ ).
noun( 'writing-ink', 'writing-inks', count, _ ).
noun( 'writing-paper', '-', mass, _ ).
noun( 'wrong', 'wrongs', both, _ ).
noun( 'wrongdoer', 'wrongdoers', count, _ ).
noun( 'wrongdoing', 'wrongdoings', both, _ ).
noun( 'wt', '-', mass, _ ).
noun( 'x', '-', count, _ ).
noun( 'xenophobia', '-', mass, _ ).
noun( 'xylophone', 'xylophones', count, _ ).
noun( 'y', '-', count, _ ).
noun( 'yacht', 'yachts', count, _ ).
noun( 'yacht-club', 'yacht-clubs', count, _ ).
noun( 'yachting', '-', mass, _ ).
noun( 'yachtsman', 'yachtsmen', count, _ ).
noun( 'yahoo', 'yahoos', count, _ ).
noun( 'yak', 'yaks', count, _ ).
noun( 'yam', 'yams', count, _ ).
noun( 'yank', 'yanks', count, _ ).
noun( 'yap', 'yaps', count, _ ).
noun( 'yard', 'yards', count, _ ).
noun( 'yard-measure', 'yard-measures', count, _ ).
noun( 'yardarm', 'yardarms', count, _ ).
noun( 'yardstick', 'yardsticks', count, _ ).
noun( 'yarn', 'yarns', both, _ ).
noun( 'yarrow', '-', mass, _ ).
noun( 'yashmak', 'yashmaks', count, _ ).
noun( 'yaw', 'yaws', count, _ ).
noun( 'yawl', 'yawls', count, _ ).
noun( 'yawn', 'yawns', count, _ ).
noun( 'yea', 'yeas', count, _ ).
noun( 'year', 'years', count, _ ).
noun( 'yearbook', 'yearbooks', count, _ ).
noun( 'yearling', 'yearlings', count, _ ).
noun( 'yearning', 'yearnings', count, _ ).
noun( 'yeast', '-', mass, _ ).
noun( 'yell', 'yells', count, _ ).
noun( 'yellow', 'yellows', both, _ ).
noun( 'yellow-flag', 'yellow-flags', count, _ ).
noun( 'yellowness', '-', mass, _ ).
noun( 'yen', 'yen', count, _ ).
noun( 'yeoman', 'yeomen', count, _ ).
noun( 'yeomanry', 'yeomanries', count, _ ).
noun( 'yes', 'yeses', count, _ ).
noun( 'yesterday', 'yesterdays', count, _ ).
noun( 'yeti', 'yetis', count, _ ).
noun( 'yew', 'yews', both, _ ).
noun( 'yew-tree', 'yew-trees', count, _ ).
noun( 'yield', 'yields', both, _ ).
noun( 'yo-yo', 'yo-yos', count, _ ).
noun( 'yob', 'yobs', count, _ ).
noun( 'yobo', 'yobos', count, _ ).
noun( 'yodel', 'yodels', count, _ ).
noun( 'yodeller', 'yodellers', count, _ ).
noun( 'yoga', '-', mass, _ ).
noun( 'yoghourt', 'yoghourts', both, _ ).
noun( 'yoghurt', 'yoghurts', both, _ ).
noun( 'yogi', 'yogis', count, _ ).
noun( 'yogurt', 'yogurts', both, _ ).
noun( 'yoke', 'yokes', count, _ ).
noun( 'yokel', 'yokels', count, _ ).
noun( 'yolk', 'yolks', both, _ ).
noun( 'yore', '-', mass, _ ).
noun( 'young', '-', mass, _ ).
noun( 'youngster', 'youngsters', count, _ ).
noun( 'youth', 'youths', both, _ ).
noun( 'youthfulness', '-', mass, _ ).
noun( 'yr', '-', count, _ ).
noun( 'yule', '-', mass, _ ).
noun( 'yule-log', 'yule-logs', count, _ ).
noun( 'yuletide', 'yuletides', both, _ ).
noun( 'z', '-', count, _ ).
noun( 'zany', 'zanies', count, _ ).
noun( 'zeal', '-', mass, _ ).
noun( 'zealot', 'zealots', count, _ ).
noun( 'zealotry', '-', mass, _ ).
noun( 'zebra', 'zebras', count, _ ).
noun( 'zebu', 'zebus', count, _ ).
noun( 'zee', 'zees', count, _ ).
noun( 'zenith', 'zeniths', count, _ ).
noun( 'zephyr', 'zephyrs', count, _ ).
noun( 'zeppelin', 'zeppelins', count, _ ).
noun( 'zero', 'zeros', count, _ ).
noun( 'zest', '-', mass, _ ).
noun( 'zigzag', 'zigzags', count, _ ).
noun( 'zinc', '-', mass, _ ).
noun( 'zing', '-', mass, _ ).
noun( 'zinnia', 'zinnias', count, _ ).
noun( 'zip', 'zips', count, _ ).
noun( 'zip code', 'zip codes', count, _ ).
noun( 'zip-fastener', 'zip-fasteners', count, _ ).
noun( 'zipper', 'zippers', count, _ ).
noun( 'zither', 'zithers', count, _ ).
noun( 'zloty', 'zlotys', count, _ ).
noun( 'zodiac', 'zodiacs', count, _ ).
noun( 'zombie', 'zombies', count, _ ).
noun( 'zone', 'zones', count, _ ).
noun( 'zoning', '-', mass, _ ).
noun( 'zoo', 'zoos', count, _ ).
noun( 'zoologist', 'zoologists', count, _ ).
noun( 'zoology', '-', mass, _ ).
noun( 'zoom', '-', mass, _ ).
noun( 'zoophyte', 'zoophytes', count, _ ).
noun( 'zoot suit', 'zoot suits', count, _ ).
noun( 'zucchini', 'zucchini', both, _ ).

adj( 'afghan', '-', '-', normal ).
adj( 'afghanistani', '-', '-', normal ).
adj( 'african', '-', '-', normal ).
adj( 'afrikaner', '-', '-', normal ).
adj( 'afro-asian', '-', '-', normal ).
adj( 'albanian', '-', '-', normal ).
adj( 'algerian', '-', '-', normal ).
adj( 'american', '-', '-', normal ).
adj( 'andorran', '-', '-', normal ).
adj( 'anglican', '-', '-', normal ).
adj( 'anglo-catholic', '-', '-', normal ).
adj( 'anglo-indian', '-', '-', normal ).
adj( 'anglo-saxon', '-', '-', normal ).
adj( 'angolan', '-', '-', normal ).
adj( 'anguillan', '-', '-', normal ).
adj( 'antiguan', '-', '-', normal ).
adj( 'arabian', '-', '-', normal ).
adj( 'arabic', '-', '-', normal ).
adj( 'arcadian', '-', '-', normal ).
adj( 'argentinian', '-', '-', normal ).
adj( 'argus-eyed', '-', '-', normal ).
adj( 'aryan', '-', '-', normal ).
adj( 'asian', '-', '-', normal ).
adj( 'asiatic', '-', '-', normal ).
adj( 'athenian', '-', '-', normal ).
adj( 'attic', '-', '-', normal ).
adj( 'augustan', '-', '-', normal ).
adj( 'australian', '-', '-', normal ).
adj( 'austrian', '-', '-', normal ).
adj( 'bahamian', '-', '-', normal ).
adj( 'bahraini', '-', '-', normal ).
adj( 'bangladeshi', '-', '-', normal ).
adj( 'bantu', '-', '-', normal ).
adj( 'baptist', '-', '-', normal ).
adj( 'barbadian', '-', '-', normal ).
adj( 'belgian', '-', '-', normal ).
adj( 'benedictine', '-', '-', normal ).
adj( 'bengali', '-', '-', normal ).
adj( 'beninese', '-', '-', normal ).
adj( 'bermudan', '-', '-', normal ).
adj( 'bhutani', '-', '-', normal ).
adj( 'boche', '-', '-', normal ).
adj( 'boer', '-', '-', normal ).
adj( 'bolivian', '-', '-', normal ).
adj( 'brazilian', '-', '-', normal ).
adj( 'britannic', '-', '-', normal ).
adj( 'british', '-', '-', normal ).
adj( 'briton', '-', '-', normal ).
adj( 'bruneian', '-', '-', normal ).
adj( 'bulgarian', '-', '-', normal ).
adj( 'burmese', '-', '-', normal ).
adj( 'burundian', '-', '-', normal ).
adj( 'caesarian', '-', '-', normal ).
adj( 'californian', '-', '-', normal ).
adj( 'cambodian', '-', '-', normal ).
adj( 'cameroonian', '-', '-', normal ).
adj( 'canadian', '-', '-', normal ).
adj( 'carmelite', '-', '-', normal ).
adj( 'catalan', '-', '-', normal ).
adj( 'catholic', '-', '-', normal ).
adj( 'caucasian', '-', '-', normal ).
adj( 'celtic', '-', '-', normal ).
adj( 'cesarean', '-', '-', normal ).
adj( 'chadian', '-', '-', normal ).
adj( 'chilean', '-', '-', normal ).
adj( 'chinese', '-', '-', normal ).
adj( 'christian', '-', '-', normal ).
adj( 'christlike', '-', '-', normal ).
adj( 'colombian', '-', '-', normal ).
adj( 'confucian', '-', '-', normal ).
adj( 'congolese', '-', '-', normal ).
adj( 'congregational', '-', '-', normal ).
adj( 'copernican', '-', '-', normal ).
adj( 'copt', '-', '-', normal ).
adj( 'corinthian', '-', '-', normal ).
adj( 'costa rican', '-', '-', normal ).
adj( 'creole', '-', '-', normal ).
adj( 'cuban', '-', '-', normal ).
adj( 'cyclopean', '-', '-', normal ).
adj( 'cyprian', '-', '-', normal ).
adj( 'cypriot', '-', '-', normal ).
adj( 'cyrillic', '-', '-', normal ).
adj( 'czech', '-', '-', normal ).
adj( 'czechoslovak', '-', '-', normal ).
adj( 'czechoslovakian', '-', '-', normal ).
adj( 'danish', '-', '-', normal ).
adj( 'delphic', '-', '-', normal ).
adj( 'djiboutian', '-', '-', normal ).
adj( 'dominican', '-', '-', normal ).
adj( 'doric', '-', '-', normal ).
adj( 'dutch', '-', '-', normal ).
adj( 'ecuadorian', '-', '-', normal ).
adj( 'edwardian', '-', '-', normal ).
adj( 'egyptian', '-', '-', normal ).
adj( 'elizabethan', '-', '-', normal ).
adj( 'elysian', '-', '-', normal ).
adj( 'english', '-', '-', normal ).
adj( 'eritrean', '-', '-', normal ).
adj( 'ethiopian', '-', '-', normal ).
adj( 'euclidean', '-', '-', normal ).
adj( 'eurasian', '-', '-', normal ).
adj( 'european', '-', '-', normal ).
adj( 'eustachian', '-', '-', attr ).
adj( 'fabian', '-', '-', normal ).
adj( 'fallopian', '-', '-', attr ).
adj( 'fijian', '-', '-', normal ).
adj( 'filipino', '-', '-', normal ).
adj( 'finnish', '-', '-', normal ).
adj( 'flemish', '-', '-', normal ).
adj( 'franciscan', '-', '-', normal ).
adj( 'french', '-', '-', normal ).
adj( 'freudian', '-', '-', normal ).
adj( 'gabonese', '-', '-', normal ).
adj( 'gaelic', '-', '-', normal ).
adj( 'gallic', '-', '-', normal ).
adj( 'gambian', '-', '-', normal ).
adj( 'georgian', '-', '-', normal ).
adj( 'georgian', '-', '-', normal ).
adj( 'german', '-', '-', normal ).
adj( 'germanic', '-', '-', normal ).
adj( 'ghanaian', '-', '-', normal ).
adj( 'gibraltarian', '-', '-', normal ).
adj( 'gilbertian', '-', '-', normal ).
adj( 'glaswegian', '-', '-', normal ).
adj( 'gordian', '-', '-', normal ).
adj( 'gothic', '-', '-', normal ).
adj( 'grecian', '-', '-', normal ).
adj( 'greek', '-', '-', normal ).
adj( 'gregorian', '-', '-', normal ).
adj( 'grenadian', '-', '-', normal ).
adj( 'guatemalan', '-', '-', normal ).
adj( 'guinean', '-', '-', normal ).
adj( 'guyanese', '-', '-', normal ).
adj( 'haitian', '-', '-', normal ).
adj( 'hebraic', '-', '-', normal ).
adj( 'hebrew', '-', '-', normal ).
adj( 'hellenic', '-', '-', normal ).
adj( 'hertzian', '-', '-', normal ).
adj( 'hindi', '-', '-', normal ).
adj( 'hindu', '-', '-', normal ).
adj( 'hindustani', '-', '-', normal ).
adj( 'hippocratic', '-', '-', normal ).
adj( 'hollander', '-', '-', normal ).
adj( 'homeric', '-', '-', normal ).
adj( 'honduran', '-', '-', normal ).
adj( 'hungarian', '-', '-', normal ).
adj( 'icelander', '-', '-', normal ).
adj( 'icelandic', '-', '-', normal ).
adj( 'indian', '-', '-', normal ).
adj( 'indo-european', '-', '-', normal ).
adj( 'indonesian', '-', '-', normal ).
adj( 'ionic', '-', '-', normal ).
adj( 'iranian', '-', '-', normal ).
adj( 'iraqi', '-', '-', normal ).
adj( 'irish', '-', '-', normal ).
adj( 'islamic', '-', '-', normal ).
adj( 'israeli', '-', '-', normal ).
adj( 'italian', '-', '-', normal ).
adj( 'jacobean', '-', '-', normal ).
adj( 'jacobin', '-', '-', normal ).
adj( 'jamaican', '-', '-', normal ).
adj( 'japanese', '-', '-', normal ).
adj( 'javanese', '-', '-', normal ).
adj( 'jesuitical', '-', '-', normal ).
adj( 'jewish', '-', '-', normal ).
adj( 'jordanian', '-', '-', normal ).
adj( 'judaic', '-', '-', normal ).
adj( 'julian', '-', '-', normal ).
adj( 'junoesque', '-', '-', normal ).
adj( 'kampuchean', '-', '-', normal ).
adj( 'kashmiri', '-', '-', normal ).
adj( 'kenyan', '-', '-', normal ).
adj( 'koranic', '-', '-', normal ).
adj( 'korean', '-', '-', normal ).
adj( 'kuwaiti', '-', '-', normal ).
adj( 'lancastrian', '-', '-', normal ).
adj( 'laotian', '-', '-', normal ).
adj( 'latin', '-', '-', normal ).
adj( 'lebanese', '-', '-', normal ).
adj( 'lenten', '-', '-', normal ).
adj( 'levantine', '-', '-', normal ).
adj( 'liberian', '-', '-', normal ).
adj( 'libyan', '-', '-', normal ).
adj( 'liechtensteiner', '-', '-', normal ).
adj( 'lilliputian', '-', '-', normal ).
adj( 'liverpudlian', '-', '-', normal ).
adj( 'lutheran', '-', '-', normal ).
adj( 'luxemburger', '-', '-', normal ).
adj( 'macedonian', '-', '-', normal ).
adj( 'madagascan', '-', '-', normal ).
adj( 'magyar', '-', '-', normal ).
adj( 'malawian', '-', '-', normal ).
adj( 'malay', '-', '-', normal ).
adj( 'malayan', '-', '-', normal ).
adj( 'malaysian', '-', '-', normal ).
adj( 'malian', '-', '-', normal ).
adj( 'maltese', '-', '-', normal ).
adj( 'malthusian', '-', '-', normal ).
adj( 'mancunian', '-', '-', normal ).
adj( 'manx', '-', '-', normal ).
adj( 'martian', '-', '-', normal ).
adj( 'mauritanian', '-', '-', normal ).
adj( 'mauritian', '-', '-', normal ).
adj( 'mediterranean', '-', '-', normal ).
adj( 'mendelian', '-', '-', normal ).
adj( 'mephistophelian', '-', '-', normal ).
adj( 'methodist', '-', '-', normal ).
adj( 'mexican', '-', '-', normal ).
adj( 'midwestern', '-', '-', normal ).
adj( 'monegasque', '-', '-', normal ).
adj( 'mongol', '-', '-', normal ).
adj( 'mongolian', '-', '-', normal ).
adj( 'montserratian', '-', '-', normal ).
adj( 'moorish', '-', '-', normal ).
adj( 'moresque', '-', '-', normal ).
adj( 'mormon', '-', '-', normal ).
adj( 'moroccan', '-', '-', normal ).
adj( 'mosaic', '-', '-', normal ).
adj( 'moslem', '-', '-', normal ).
adj( 'mozambican', '-', '-', normal ).
adj( 'muhammadan', '-', '-', normal ).
adj( 'muscovite', '-', '-', normal ).
adj( 'namibian', '-', '-', normal ).
adj( 'napoleonic', '-', '-', normal ).
adj( 'nauruan', '-', '-', normal ).
adj( 'nazi', '-', '-', normal ).
adj( 'neanderthal', '-', '-', normal ).
adj( 'neapolitan', '-', '-', normal ).
adj( 'negroid', '-', '-', normal ).
adj( 'nepalese', '-', '-', normal ).
adj( 'nepali', '-', '-', normal ).
adj( 'newtonian', '-', '-', normal ).
adj( 'nicaraguan', '-', '-', normal ).
adj( 'nigerian', '-', '-', normal ).
adj( 'nigerien', '-', '-', normal ).
adj( 'nilotic', '-', '-', normal ).
adj( 'nipponese', '-', '-', normal ).
adj( 'nordic', '-', '-', normal ).
adj( 'norman', '-', '-', normal ).
adj( 'norse', '-', '-', normal ).
adj( 'norwegian', '-', '-', normal ).
adj( 'occidental', '-', '-', normal ).
adj( 'olympian', '-', '-', normal ).
adj( 'olympic', '-', '-', normal ).
adj( 'omani', '-', '-', normal ).
adj( 'oxonian', '-', '-', normal ).
adj( 'pakistani', '-', '-', normal ).
adj( 'palestinian', '-', '-', normal ).
adj( 'panamanian', '-', '-', normal ).
adj( 'papuan', '-', '-', normal ).
adj( 'paraguayan', '-', '-', normal ).
adj( 'parisian', '-', '-', normal ).
adj( 'parkinson\'s', '-', '-', normal ).
adj( 'parthian', '-', '-', normal ).
adj( 'pentecostal', '-', '-', normal ).
adj( 'persian', '-', '-', normal ).
adj( 'peruvian', '-', '-', normal ).
adj( 'philippine', '-', '-', normal ).
adj( 'platonic', '-', '-', normal ).
adj( 'polish', '-', '-', normal ).
adj( 'portuguese', '-', '-', normal ).
adj( 'pre-raphaelite', '-', '-', normal ).
adj( 'presbyterian', '-', '-', normal ).
adj( 'protestant', '-', '-', normal ).
adj( 'prussian', '-', '-', normal ).
adj( 'punic', '-', '-', normal ).
adj( 'pyrrhic', '-', '-', normal ).
adj( 'qatari', '-', '-', normal ).
adj( 'rabelaisian', '-', '-', normal ).
adj( 'redbrick', '-', '-', normal ).
adj( 'rhenish', '-', '-', normal ).
adj( 'romaic', '-', '-', normal ).
adj( 'roman', '-', '-', normal ).
adj( 'romance', '-', '-', normal ).
adj( 'romanian', '-', '-', normal ).
adj( 'romany', '-', '-', normal ).
adj( 'romish', '-', '-', normal ).
adj( 'ruritanian', '-', '-', normal ).
adj( 'russian', '-', '-', normal ).
adj( 'rwandan', '-', '-', normal ).
adj( 'sabahan', '-', '-', normal ).
adj( 'salvadorean', '-', '-', normal ).
adj( 'samoan', '-', '-', normal ).
adj( 'san marinese', '-', '-', normal ).
adj( 'sapphic', '-', '-', normal ).
adj( 'sarawakian', '-', '-', normal ).
adj( 'satanic', '-', '-', normal ).
adj( 'saudi arabian', '-', '-', normal ).
adj( 'saxon', '-', '-', normal ).
adj( 'scandinavian', '-', '-', normal ).
adj( 'scotch', '-', '-', normal ).
adj( 'scots', '-', '-', normal ).
adj( 'scottish', '-', '-', normal ).
adj( 'semite', '-', '-', normal ).
adj( 'semitic', '-', '-', normal ).
adj( 'senegalese', '-', '-', normal ).
adj( 'seychellois', '-', '-', normal ).
adj( 'shakespearian', '-', '-', normal ).
adj( 'shavian', '-', '-', normal ).
adj( 'siamese', '-', '-', normal ).
adj( 'siberian', '-', '-', normal ).
adj( 'sicilian', '-', '-', normal ).
adj( 'sierra leonian', '-', '-', normal ).
adj( 'singaporean', '-', '-', normal ).
adj( 'singhalese', '-', '-', normal ).
adj( 'sinhalese', '-', '-', normal ).
adj( 'slav', '-', '-', normal ).
adj( 'slavonic', '-', '-', normal ).
adj( 'slovenian', '-', '-', normal ).
adj( 'socratic', '-', '-', normal ).
adj( 'somali', '-', '-', normal ).
adj( 'somalian', '-', '-', normal ).
adj( 'sotho', '-', '-', normal ).
adj( 'soviet', '-', '-', normal ).
adj( 'spanish', '-', '-', normal ).
adj( 'spartan', '-', '-', normal ).
adj( 'sri lankan', '-', '-', normal ).
adj( 'stygian', '-', '-', normal ).
adj( 'sudanese', '-', '-', normal ).
adj( 'sumatran', '-', '-', normal ).
adj( 'swazi', '-', '-', normal ).
adj( 'swede', '-', '-', normal ).
adj( 'swedish', '-', '-', normal ).
adj( 'swiss', '-', '-', normal ).
adj( 'syrian', '-', '-', normal ).
adj( 'tahitian', '-', '-', normal ).
adj( 'taiwanese', '-', '-', normal ).
adj( 'tamil', '-', '-', normal ).
adj( 'tanzanian', '-', '-', normal ).
adj( 'terpsichorean', '-', '-', normal ).
adj( 'teutonic', '-', '-', normal ).
adj( 'texan', '-', '-', normal ).
adj( 'thai', '-', '-', normal ).
adj( 'thespian', '-', '-', normal ).
adj( 'tibetan', '-', '-', normal ).
adj( 'tobagonian', '-', '-', normal ).
adj( 'togolese', '-', '-', normal ).
adj( 'tongan', '-', '-', normal ).
adj( 'trinidadian', '-', '-', normal ).
adj( 'trojan', '-', '-', normal ).
adj( 'tswana', '-', '-', normal ).
adj( 'tunisian', '-', '-', normal ).
adj( 'turkish', '-', '-', normal ).
adj( 'ugandan', '-', '-', normal ).
adj( 'unitarian', '-', '-', normal ).
adj( 'urdu', '-', '-', normal ).
adj( 'uruguayan', '-', '-', normal ).
adj( 'utopian', '-', '-', normal ).
adj( 'venetian', '-', '-', normal ).
adj( 'venezuelan', '-', '-', normal ).
adj( 'verey', '-', '-', normal ).
adj( 'very', '-', '-', normal ).
adj( 'victorian', '-', '-', normal ).
adj( 'vietnamese', '-', '-', normal ).
adj( 'welsh', '-', '-', normal ).
adj( 'wesleyan', '-', '-', normal ).
adj( 'yemeni', '-', '-', normal ).
adj( 'yugoslav', '-', '-', normal ).
adj( 'yugoslavian', '-', '-', normal ).
adj( 'zairean', '-', '-', normal ).
adj( 'zambian', '-', '-', normal ).
adj( 'zealander', '-', '-', normal ).
adj( 'zimbabwean', '-', '-', normal ).
adj( 'zionist', '-', '-', normal ).
adj( 'a posteriori', '-', '-', normal ).
adj( 'a priori', '-', '-', normal ).
adj( 'abandoned', '-', '-', normal ).
adj( 'abdominal', '-', '-', normal ).
adj( 'aberrant', '-', '-', normal ).
adj( 'abhorrent', '-', '-', normal ).
adj( 'abiding', '-', '-', normal ).
adj( 'abject', '-', '-', normal ).
adj( 'ablative', '-', '-', normal ).
adj( 'ablaze', '-', '-', pred ).
adj( 'able', 'abler', 'ablest', normal ).
adj( 'able-bodied', '-', '-', normal ).
adj( 'abnormal', '-', '-', normal ).
adj( 'abominable', '-', '-', normal ).
adj( 'aboriginal', '-', '-', normal ).
adj( 'abortive', '-', '-', normal ).
adj( 'above board', '-', '-', pred ).
adj( 'above-mentioned', '-', '-', normal ).
adj( 'above-named', '-', '-', normal ).
adj( 'abrasive', '-', '-', normal ).
adj( 'abrupt', '-', '-', normal ).
adj( 'absent', '-', '-', normal ).
adj( 'absent-minded', '-', '-', normal ).
adj( 'absolute', '-', '-', normal ).
adj( 'absorbent', '-', '-', normal ).
adj( 'abstemious', '-', '-', normal ).
adj( 'abstract', '-', '-', normal ).
adj( 'abstracted', '-', '-', normal ).
adj( 'abstruse', '-', '-', normal ).
adj( 'absurd', '-', '-', normal ).
adj( 'abundant', '-', '-', normal ).
adj( 'abusive', '-', '-', normal ).
adj( 'abysmal', '-', '-', normal ).
adj( 'academic', '-', '-', normal ).
adj( 'accelerando', '-', '-', normal ).
adj( 'acceptable', '-', '-', normal ).
adj( 'accessible', '-', '-', normal ).
adj( 'accident-prone', '-', '-', normal ).
adj( 'accidental', '-', '-', normal ).
adj( 'accommodating', '-', '-', normal ).
adj( 'accomplished', '-', '-', normal ).
adj( 'accountable', '-', '-', normal ).
adj( 'accredited', '-', '-', normal ).
adj( 'accumulative', '-', '-', normal ).
adj( 'accurate', '-', '-', normal ).
adj( 'accursed', '-', '-', normal ).
adj( 'accurst', '-', '-', normal ).
adj( 'accusative', '-', '-', normal ).
adj( 'accustomed', '-', '-', normal ).
adj( 'acetic', '-', '-', normal ).
adj( 'achievable', '-', '-', normal ).
adj( 'acid', '-', '-', normal ).
adj( 'acidic', '-', '-', normal ).
adj( 'acidulated', '-', '-', normal ).
adj( 'acidulous', '-', '-', normal ).
adj( 'acoustic', '-', '-', normal ).
adj( 'acquiescent', '-', '-', normal ).
adj( 'acquisitive', '-', '-', normal ).
adj( 'acrid', '-', '-', normal ).
adj( 'acrimonious', '-', '-', normal ).
adj( 'acrobatic', '-', '-', normal ).
adj( 'acting', '-', '-', normal ).
adj( 'actinic', '-', '-', normal ).
adj( 'actionable', '-', '-', normal ).
adj( 'active', '-', '-', normal ).
adj( 'actual', '-', '-', normal ).
adj( 'actuarial', '-', '-', normal ).
adj( 'acute', '-', '-', normal ).
adj( 'ad hoc', '-', '-', normal ).
adj( 'ad-lib', '-', '-', normal ).
adj( 'adagio', '-', '-', normal ).
adj( 'adamant', '-', '-', pred ).
adj( 'adamantine', '-', '-', normal ).
adj( 'adaptable', '-', '-', normal ).
adj( 'addictive', '-', '-', normal ).
adj( 'additional', '-', '-', normal ).
adj( 'addle', '-', '-', normal ).
adj( 'addle-brained', '-', '-', normal ).
adj( 'addle-pated', '-', '-', normal ).
adj( 'adenoidal', '-', '-', normal ).
adj( 'adept', '-', '-', normal ).
adj( 'adequate', '-', '-', normal ).
adj( 'adhesive', '-', '-', normal ).
adj( 'adipose', '-', '-', normal ).
adj( 'adjacent', '-', '-', normal ).
adj( 'adjectival', '-', '-', normal ).
adj( 'adjoining', '-', '-', normal ).
adj( 'adjustable', '-', '-', normal ).
adj( 'administrative', '-', '-', normal ).
adj( 'admirable', '-', '-', normal ).
adj( 'admiring', '-', '-', normal ).
adj( 'admissible', '-', '-', normal ).
adj( 'admonitory', '-', '-', normal ).
adj( 'adolescent', '-', '-', normal ).
adj( 'adoptive', '-', '-', normal ).
adj( 'adorable', '-', '-', normal ).
adj( 'adoring', '-', '-', normal ).
adj( 'adrenal', '-', '-', normal ).
adj( 'adrift', '-', '-', pred ).
adj( 'adroit', '-', '-', normal ).
adj( 'adult', '-', '-', normal ).
adj( 'adulterous', '-', '-', normal ).
adj( 'advanced', '-', '-', normal ).
adj( 'advantageous', '-', '-', normal ).
adj( 'adventitious', '-', '-', normal ).
adj( 'adventuresome', '-', '-', normal ).
adj( 'adventurous', '-', '-', normal ).
adj( 'adverbial', '-', '-', normal ).
adj( 'adverse', '-', '-', normal ).
adj( 'advisable', '-', '-', normal ).
adj( 'advised', '-', '-', normal ).
adj( 'advisory', '-', '-', normal ).
adj( 'aerial', '-', '-', normal ).
adj( 'aerodynamic', '-', '-', normal ).
adj( 'aesthetic', '-', '-', normal ).
adj( 'aesthetical', '-', '-', normal ).
adj( 'affable', '-', '-', normal ).
adj( 'affected', '-', '-', normal ).
adj( 'affecting', '-', '-', normal ).
adj( 'affectionate', '-', '-', normal ).
adj( 'affirmative', '-', '-', normal ).
adj( 'affluent', '-', '-', normal ).
adj( 'afire', '-', '-', pred ).
adj( 'aflame', '-', '-', pred ).
adj( 'afloat', '-', '-', pred ).
adj( 'afoot', '-', '-', pred ).
adj( 'aforesaid', '-', '-', normal ).
adj( 'afraid', '-', '-', pred ).
adj( 'after', '-', '-', normal ).
adj( 'agape', '-', '-', pred ).
adj( 'age-long', '-', '-', normal ).
adj( 'age-old', '-', '-', normal ).
adj( 'aged', '-', '-', attr ).
adj( 'aged', '-', '-', pred ).
adj( 'ageless', '-', '-', normal ).
adj( 'agglomerate', '-', '-', normal ).
adj( 'agglutinative', '-', '-', normal ).
adj( 'aggressive', '-', '-', normal ).
adj( 'aghast', '-', '-', pred ).
adj( 'agile', '-', '-', normal ).
adj( 'agitated', '-', '-', normal ).
adj( 'agitating', '-', '-', normal ).
adj( 'aglow', '-', '-', pred ).
adj( 'agnostic', '-', '-', normal ).
adj( 'agog', '-', '-', pred ).
adj( 'agonized', '-', '-', normal ).
adj( 'agonizing', '-', '-', normal ).
adj( 'agrarian', '-', '-', normal ).
adj( 'agreeable', '-', '-', normal ).
adj( 'agricultural', '-', '-', normal ).
adj( 'aground', '-', '-', pred ).
adj( 'aimless', '-', '-', normal ).
adj( 'air-conditioned', '-', '-', normal ).
adj( 'air-cooled', '-', '-', normal ).
adj( 'air-minded', '-', '-', normal ).
adj( 'air-sick', '-', '-', normal ).
adj( 'air-to-air', '-', '-', normal ).
adj( 'air-to-ground', '-', '-', normal ).
adj( 'airborne', '-', '-', normal ).
adj( 'airless', '-', '-', normal ).
adj( 'airtight', '-', '-', normal ).
adj( 'airworthy', '-', '-', normal ).
adj( 'airy', '-', '-', normal ).
adj( 'ajar', '-', '-', pred ).
adj( 'akin', '-', '-', pred ).
adj( 'alabaster', '-', '-', normal ).
adj( 'alarming', '-', '-', normal ).
adj( 'alcoholic', '-', '-', normal ).
adj( 'aldermanic', '-', '-', normal ).
adj( 'alee', '-', '-', pred ).
adj( 'alert', '-', '-', normal ).
adj( 'alexic', '-', '-', normal ).
adj( 'alfresco', '-', '-', normal ).
adj( 'algebraic', '-', '-', normal ).
adj( 'algebraical', '-', '-', normal ).
adj( 'alien', '-', '-', normal ).
adj( 'alight', '-', '-', pred ).
adj( 'alike', '-', '-', pred ).
adj( 'alimentary', '-', '-', normal ).
adj( 'alive', '-', '-', pred ).
adj( 'alkaline', '-', '-', normal ).
adj( 'all', '-', '-', normal ).
adj( 'all-mains', '-', '-', attr ).
adj( 'all-round', '-', '-', normal ).
adj( 'allegoric', '-', '-', normal ).
adj( 'allegorical', '-', '-', normal ).
adj( 'allegretto', '-', '-', normal ).
adj( 'allegro', '-', '-', normal ).
adj( 'allergic', '-', '-', normal ).
adj( 'alliterative', '-', '-', normal ).
adj( 'allowable', '-', '-', normal ).
adj( 'alluring', '-', '-', normal ).
adj( 'allusive', '-', '-', normal ).
adj( 'alluvial', '-', '-', normal ).
adj( 'almighty', '-', '-', normal ).
adj( 'almond-eyed', '-', '-', normal ).
adj( 'alone', '-', '-', pred ).
adj( 'aloof', '-', '-', normal ).
adj( 'alphabetical', '-', '-', normal ).
adj( 'alpine', '-', '-', normal ).
adj( 'alright', '-', '-', normal ).
adj( 'alterable', '-', '-', normal ).
adj( 'alternate', '-', '-', normal ).
adj( 'alternative', '-', '-', normal ).
adj( 'altruistic', '-', '-', normal ).
adj( 'alveolar', '-', '-', normal ).
adj( 'amateurish', '-', '-', normal ).
adj( 'amatory', '-', '-', normal ).
adj( 'amazing', '-', '-', normal ).
adj( 'ambassadorial', '-', '-', normal ).
adj( 'ambidextrous', '-', '-', normal ).
adj( 'ambient', '-', '-', normal ).
adj( 'ambiguous', '-', '-', normal ).
adj( 'ambitious', '-', '-', normal ).
adj( 'ambivalent', '-', '-', normal ).
adj( 'amenable', '-', '-', normal ).
adj( 'amendable', '-', '-', normal ).
adj( 'amiable', '-', '-', normal ).
adj( 'amicable', '-', '-', normal ).
adj( 'amiss', '-', '-', pred ).
adj( 'ammoniated', '-', '-', normal ).
adj( 'amoebic', '-', '-', normal ).
adj( 'amoral', '-', '-', normal ).
adj( 'amorous', '-', '-', normal ).
adj( 'amorphous', '-', '-', normal ).
adj( 'amphibious', '-', '-', normal ).
adj( 'ample', 'ampler', 'amplest', normal ).
adj( 'amusing', '-', '-', normal ).
adj( 'anachronistic', '-', '-', normal ).
adj( 'anaemic', '-', '-', normal ).
adj( 'anaesthetic', '-', '-', normal ).
adj( 'anal', '-', '-', normal ).
adj( 'analogous', '-', '-', normal ).
adj( 'analytic', '-', '-', normal ).
adj( 'analytical', '-', '-', normal ).
adj( 'anapaestic', '-', '-', normal ).
adj( 'anarchic', '-', '-', normal ).
adj( 'anatomical', '-', '-', normal ).
adj( 'ancestral', '-', '-', normal ).
adj( 'ancient', '-', '-', normal ).
adj( 'ancillary', '-', '-', normal ).
adj( 'andante', '-', '-', normal ).
adj( 'anecdotal', '-', '-', normal ).
adj( 'aneroid', '-', '-', normal ).
adj( 'anesthetic', '-', '-', normal ).
adj( 'angelic', '-', '-', normal ).
adj( 'angry', 'angrier', 'angriest', normal ).
adj( 'anguished', '-', '-', normal ).
adj( 'angular', '-', '-', normal ).
adj( 'animate', '-', '-', normal ).
adj( 'animatedly', '-', '-', normal ).
adj( 'annoying', '-', '-', normal ).
adj( 'annual', '-', '-', normal ).
adj( 'annular', '-', '-', normal ).
adj( 'anodyne', '-', '-', normal ).
adj( 'anomalous', '-', '-', normal ).
adj( 'anonymous', '-', '-', normal ).
adj( 'another', '-', '-', normal ).
adj( 'answerable', '-', '-', normal ).
adj( 'antagonistic', '-', '-', normal ).
adj( 'antarctic', '-', '-', normal ).
adj( 'antecedent', '-', '-', normal ).
adj( 'antediluvian', '-', '-', normal ).
adj( 'antenatal', '-', '-', normal ).
adj( 'antenuptial', '-', '-', normal ).
adj( 'antepenultimate', '-', '-', normal ).
adj( 'anterior', '-', '-', normal ).
adj( 'anthropoid', '-', '-', normal ).
adj( 'anthropological', '-', '-', normal ).
adj( 'anti-semite', '-', '-', normal ).
adj( 'anti-semitic', '-', '-', normal ).
adj( 'anti-aircraft', '-', '-', normal ).
adj( 'anti-personnel', '-', '-', normal ).
adj( 'antibiotic', '-', '-', normal ).
adj( 'anticipatory', '-', '-', normal ).
adj( 'antipathetic', '-', '-', normal ).
adj( 'antiquarian', '-', '-', normal ).
adj( 'antiquated', '-', '-', normal ).
adj( 'antique', '-', '-', normal ).
adj( 'antiseptic', '-', '-', normal ).
adj( 'antisocial', '-', '-', normal ).
adj( 'antitank', '-', '-', attr ).
adj( 'antithetic', '-', '-', normal ).
adj( 'antithetical', '-', '-', normal ).
adj( 'antitrade', '-', '-', normal ).
adj( 'anxious', '-', '-', normal ).
adj( 'any', '-', '-', normal ).
adj( 'apathetic', '-', '-', normal ).
adj( 'aperient', '-', '-', normal ).
adj( 'aphrodisiac', '-', '-', normal ).
adj( 'apish', '-', '-', normal ).
adj( 'apocalyptic', '-', '-', normal ).
adj( 'apocryphal', '-', '-', normal ).
adj( 'apologetic', '-', '-', normal ).
adj( 'apoplectic', '-', '-', normal ).
adj( 'apostate', '-', '-', normal ).
adj( 'apostolic', '-', '-', normal ).
adj( 'appalling', '-', '-', normal ).
adj( 'apparent', '-', '-', normal ).
adj( 'appealing', '-', '-', normal ).
adj( 'appellant', '-', '-', normal ).
adj( 'appetizing', '-', '-', normal ).
adj( 'applicable', '-', '-', normal ).
adj( 'applied', '-', '-', normal ).
adj( 'apposite', '-', '-', normal ).
adj( 'appreciable', '-', '-', normal ).
adj( 'appreciative', '-', '-', normal ).
adj( 'apprehensible', '-', '-', normal ).
adj( 'apprehensive', '-', '-', normal ).
adj( 'approachable', '-', '-', normal ).
adj( 'appropriate', '-', '-', normal ).
adj( 'approximate', '-', '-', normal ).
adj( 'apr`es-ski', '-', '-', attr ).
adj( 'apropos', '-', '-', pred ).
adj( 'apt', 'apter', 'aptest', normal ).
adj( 'aquatic', '-', '-', normal ).
adj( 'aqueous', '-', '-', normal ).
adj( 'aquiline', '-', '-', normal ).
adj( 'arable', '-', '-', normal ).
adj( 'arbitrary', '-', '-', normal ).
adj( 'arboreal', '-', '-', normal ).
adj( 'arcane', '-', '-', normal ).
adj( 'arch', '-', '-', attr ).
adj( 'archaeological', '-', '-', normal ).
adj( 'archaic', '-', '-', normal ).
adj( 'archetypal', '-', '-', normal ).
adj( 'architectural', '-', '-', normal ).
adj( 'arctic', '-', '-', normal ).
adj( 'ardent', '-', '-', normal ).
adj( 'arduous', '-', '-', normal ).
adj( 'argent', '-', '-', normal ).
adj( 'arguable', '-', '-', normal ).
adj( 'argumentative', '-', '-', normal ).
adj( 'arid', '-', '-', normal ).
adj( 'aristocratic', '-', '-', normal ).
adj( 'arithmetical', '-', '-', normal ).
adj( 'armorial', '-', '-', normal ).
adj( 'armoured', '-', '-', normal ).
adj( 'aromatic', '-', '-', normal ).
adj( 'arrant', '-', '-', normal ).
adj( 'arresting', '-', '-', normal ).
adj( 'arrogant', '-', '-', normal ).
adj( 'arterial', '-', '-', normal ).
adj( 'artesian', '-', '-', normal ).
adj( 'artful', '-', '-', normal ).
adj( 'arthritic', '-', '-', normal ).
adj( 'articulate', '-', '-', normal ).
adj( 'artificial', '-', '-', normal ).
adj( 'artistic', '-', '-', normal ).
adj( 'artless', '-', '-', normal ).
adj( 'arty', '-', '-', normal ).
adj( 'arty-crafty', '-', '-', normal ).
adj( 'ascertainable', '-', '-', normal ).
adj( 'ascetic', '-', '-', normal ).
adj( 'ascorbic', '-', '-', normal ).
adj( 'ascribable', '-', '-', normal ).
adj( 'aseptic', '-', '-', normal ).
adj( 'asexual', '-', '-', normal ).
adj( 'ashamed', '-', '-', pred ).
adj( 'ashen', '-', '-', normal ).
adj( 'ashy', '-', '-', normal ).
adj( 'asinine', '-', '-', normal ).
adj( 'askew', '-', '-', pred ).
adj( 'asleep', '-', '-', pred ).
adj( 'aspectual', '-', '-', normal ).
adj( 'assailable', '-', '-', normal ).
adj( 'assertive', '-', '-', normal ).
adj( 'assiduous', '-', '-', normal ).
adj( 'assignable', '-', '-', normal ).
adj( 'associate', '-', '-', normal ).
adj( 'assorted', '-', '-', normal ).
adj( 'assured', '-', '-', normal ).
adj( 'asthmatic', '-', '-', normal ).
adj( 'astigmatic', '-', '-', normal ).
adj( 'astir', '-', '-', pred ).
adj( 'astonishing', '-', '-', normal ).
adj( 'astral', '-', '-', normal ).
adj( 'astray', '-', '-', pred ).
adj( 'astride', '-', '-', pred ).
adj( 'astringent', '-', '-', normal ).
adj( 'astrological', '-', '-', normal ).
adj( 'astronomical', '-', '-', normal ).
adj( 'astute', '-', '-', normal ).
adj( 'asymmetric', '-', '-', normal ).
adj( 'asymmetrical', '-', '-', normal ).
adj( 'asymptotic', '-', '-', normal ).
adj( 'atavistic', '-', '-', normal ).
adj( 'atheistic', '-', '-', normal ).
adj( 'athirst', '-', '-', pred ).
adj( 'athletic', '-', '-', normal ).
adj( 'atmospheric', '-', '-', normal ).
adj( 'atomic', '-', '-', normal ).
adj( 'atonal', '-', '-', normal ).
adj( 'atrabilious', '-', '-', normal ).
adj( 'atrocious', '-', '-', normal ).
adj( 'attainable', '-', '-', normal ).
adj( 'attendant', '-', '-', normal ).
adj( 'attentive', '-', '-', normal ).
adj( 'attractive', '-', '-', normal ).
adj( 'attributable', '-', '-', normal ).
adj( 'attributive', '-', '-', normal ).
adj( 'atypical', '-', '-', normal ).
adj( 'au fait', '-', '-', pred ).
adj( 'auburn', '-', '-', normal ).
adj( 'audacious', '-', '-', normal ).
adj( 'audible', '-', '-', normal ).
adj( 'audio-lingual', '-', '-', normal ).
adj( 'auditory', '-', '-', normal ).
adj( 'august', '-', '-', normal ).
adj( 'aural', '-', '-', normal ).
adj( 'auricular', '-', '-', normal ).
adj( 'auriferous', '-', '-', normal ).
adj( 'auspicious', '-', '-', normal ).
adj( 'austere', '-', '-', normal ).
adj( 'authentic', '-', '-', normal ).
adj( 'authoritarian', '-', '-', normal ).
adj( 'authoritative', '-', '-', normal ).
adj( 'autistic', '-', '-', normal ).
adj( 'autobiographic', '-', '-', normal ).
adj( 'autobiographical', '-', '-', normal ).
adj( 'autocratic', '-', '-', normal ).
adj( 'automatic', '-', '-', normal ).
adj( 'autonomous', '-', '-', normal ).
adj( 'autumnal', '-', '-', normal ).
adj( 'auxiliary', '-', '-', normal ).
adj( 'available', '-', '-', normal ).
adj( 'avaricious', '-', '-', normal ).
adj( 'average', '-', '-', normal ).
adj( 'averse', '-', '-', normal ).
adj( 'avid', '-', '-', normal ).
adj( 'avoidable', '-', '-', normal ).
adj( 'avuncular', '-', '-', normal ).
adj( 'awake', '-', '-', pred ).
adj( 'aware', '-', '-', pred ).
adj( 'awash', '-', '-', pred ).
adj( 'away', '-', '-', normal ).
adj( 'awe-inspiring', '-', '-', normal ).
adj( 'awe-stricken', '-', '-', normal ).
adj( 'awe-struck', '-', '-', normal ).
adj( 'awesome', '-', '-', normal ).
adj( 'awful', '-', '-', normal ).
adj( 'awkward', '-', '-', normal ).
adj( 'awry', '-', '-', pred ).
adj( 'axiomatic', '-', '-', normal ).
adj( 'azure', '-', '-', normal ).
adj( 'baby-faced', '-', '-', normal ).
adj( 'babyish', '-', '-', normal ).
adj( 'bacchanal', '-', '-', normal ).
adj( 'bacchanalian', '-', '-', normal ).
adj( 'back-breaking', '-', '-', normal ).
adj( 'backhand', '-', '-', normal ).
adj( 'backhanded', '-', '-', normal ).
adj( 'backless', '-', '-', normal ).
adj( 'backmost', '-', '-', normal ).
adj( 'backstair', '-', '-', normal ).
adj( 'backward', '-', '-', normal ).
adj( 'backwards', '-', '-', normal ).
adj( 'bacterial', '-', '-', normal ).
adj( 'bad', 'worse', 'worst', normal ).
adj( 'badly-behaved', '-', '-', normal ).
adj( 'baggy', '-', '-', normal ).
adj( 'baking-hot', '-', '-', normal ).
adj( 'balconied', '-', '-', normal ).
adj( 'bald', 'balder', 'baldest', normal ).
adj( 'baleful', '-', '-', normal ).
adj( 'ballistic', '-', '-', normal ).
adj( 'bally', '-', '-', normal ).
adj( 'balmy', 'balmier', 'balmiest', normal ).
adj( 'banal', '-', '-', normal ).
adj( 'bandy', 'bandier', 'bandiest', normal ).
adj( 'bandy-legged', '-', '-', normal ).
adj( 'baneful', '-', '-', normal ).
adj( 'bankrupt', '-', '-', normal ).
adj( 'bantering', '-', '-', normal ).
adj( 'baptismal', '-', '-', normal ).
adj( 'barbarian', '-', '-', normal ).
adj( 'barbaric', '-', '-', normal ).
adj( 'barbarous', '-', '-', normal ).
adj( 'barbed', '-', '-', normal ).
adj( 'bardic', '-', '-', normal ).
adj( 'bare', 'barer', 'barest', normal ).
adj( 'barebacked', '-', '-', normal ).
adj( 'barefaced', '-', '-', normal ).
adj( 'barefooted', '-', '-', normal ).
adj( 'bareheaded', '-', '-', normal ).
adj( 'barelegged', '-', '-', normal ).
adj( 'barmy', '-', '-', normal ).
adj( 'barometric', '-', '-', normal ).
adj( 'baronial', '-', '-', normal ).
adj( 'baroque', '-', '-', normal ).
adj( 'barrelled', '-', '-', normal ).
adj( 'barren', '-', '-', normal ).
adj( 'basal', '-', '-', normal ).
adj( 'base', 'baser', 'basest', normal ).
adj( 'baseless', '-', '-', normal ).
adj( 'bashful', '-', '-', normal ).
adj( 'basic', '-', '-', normal ).
adj( 'bass', '-', '-', normal ).
adj( 'bats', '-', '-', pred ).
adj( 'batty', '-', '-', normal ).
adj( 'bawdy', 'bawdier', 'bawdiest', normal ).
adj( 'bay', '-', '-', normal ).
adj( 'beady', '-', '-', normal ).
adj( 'bearable', '-', '-', normal ).
adj( 'bearded', '-', '-', normal ).
adj( 'beardless', '-', '-', normal ).
adj( 'bearish', '-', '-', normal ).
adj( 'beastly', 'beastlier', 'beastliest', normal ).
adj( 'beat', '-', '-', attr ).
adj( 'beaten', '-', '-', normal ).
adj( 'beatific', '-', '-', normal ).
adj( 'beauteous', '-', '-', normal ).
adj( 'beautiful', '-', '-', normal ).
adj( 'becalmed', '-', '-', pred ).
adj( 'becoming', '-', '-', normal ).
adj( 'bedaubed', '-', '-', pred ).
adj( 'bedded', '-', '-', affix ).
adj( 'bedecked', '-', '-', pred ).
adj( 'bedewed', '-', '-', pred ).
adj( 'bedimmed', '-', '-', pred ).
adj( 'bedraggled', '-', '-', pred ).
adj( 'bedridden', '-', '-', normal ).
adj( 'bedroomed', '-', '-', normal ).
adj( 'beefy', 'beefier', 'beefiest', normal ).
adj( 'beery', 'beerier', 'beeriest', normal ).
adj( 'beetle-browed', '-', '-', normal ).
adj( 'befitting', '-', '-', normal ).
adj( 'befogged', '-', '-', pred ).
adj( 'beforehand', '-', '-', pred ).
adj( 'beggarly', '-', '-', normal ).
adj( 'begrimed', '-', '-', pred ).
adj( 'behaved', '-', '-', affix ).
adj( 'behindhand', '-', '-', pred ).
adj( 'beholden', '-', '-', pred ).
adj( 'bejewelled', '-', '-', normal ).
adj( 'belated', '-', '-', normal ).
adj( 'believable', '-', '-', normal ).
adj( 'bell-bottomed', '-', '-', normal ).
adj( 'bellicose', '-', '-', normal ).
adj( 'bellied', '-', '-', affix ).
adj( 'belligerent', '-', '-', normal ).
adj( 'beloved', '-', '-', normal ).
adj( 'bemused', '-', '-', pred ).
adj( 'beneficed', '-', '-', normal ).
adj( 'beneficent', '-', '-', normal ).
adj( 'beneficial', '-', '-', normal ).
adj( 'benevolent', '-', '-', normal ).
adj( 'benighted', '-', '-', normal ).
adj( 'benign', '-', '-', normal ).
adj( 'benignant', '-', '-', normal ).
adj( 'bent', '-', '-', pred ).
adj( 'benumbed', '-', '-', pred ).
adj( 'berserk', '-', '-', pred ).
adj( 'beseeching', '-', '-', normal ).
adj( 'besotted', '-', '-', normal ).
adj( 'bespangled', '-', '-', pred ).
adj( 'bespattered', '-', '-', pred ).
adj( 'bespectacled', '-', '-', normal ).
adj( 'bestial', '-', '-', normal ).
adj( 'bewildering', '-', '-', normal ).
adj( 'bewitching', '-', '-', normal ).
adj( 'biblical', '-', '-', normal ).
adj( 'bibulous', '-', '-', normal ).
adj( 'bicameral', '-', '-', normal ).
adj( 'bicentennial', '-', '-', normal ).
adj( 'biddable', '-', '-', normal ).
adj( 'biennial', '-', '-', normal ).
adj( 'bifocal', '-', '-', normal ).
adj( 'bifurcate', '-', '-', normal ).
adj( 'bifurcated', '-', '-', normal ).
adj( 'big', 'bigger', 'biggest', normal ).
adj( 'big-boned', '-', '-', normal ).
adj( 'bigamous', '-', '-', normal ).
adj( 'bigoted', '-', '-', normal ).
adj( 'bijou', '-', '-', normal ).
adj( 'bilabial', '-', '-', normal ).
adj( 'bilateral', '-', '-', normal ).
adj( 'bilingual', '-', '-', normal ).
adj( 'bilious', '-', '-', normal ).
adj( 'billion', '-', '-', normal ).
adj( 'billionth', '-', '-', normal ).
adj( 'billowy', '-', '-', normal ).
adj( 'bimetallic', '-', '-', normal ).
adj( 'binary', '-', '-', normal ).
adj( 'binding', '-', '-', normal ).
adj( 'binomial', '-', '-', normal ).
adj( 'biodegradable', '-', '-', normal ).
adj( 'biographic', '-', '-', normal ).
adj( 'biographical', '-', '-', normal ).
adj( 'biological', '-', '-', normal ).
adj( 'bipartisan', '-', '-', normal ).
adj( 'bird\'s-eye', '-', '-', attr ).
adj( 'bisexual', '-', '-', normal ).
adj( 'bitchy', 'bitchier', 'bitchiest', normal ).
adj( 'biting', '-', '-', normal ).
adj( 'bitter', '-', '-', normal ).
adj( 'bitter-sweet', '-', '-', normal ).
adj( 'bituminous', '-', '-', normal ).
adj( 'bizarre', '-', '-', normal ).
adj( 'bizonal', '-', '-', normal ).
adj( 'black', 'blacker', 'blackest', normal ).
adj( 'blackguardly', '-', '-', normal ).
adj( 'blackish', '-', '-', normal ).
adj( 'blackwater', '-', '-', attr ).
adj( 'blameless', '-', '-', normal ).
adj( 'blameworthy', '-', '-', normal ).
adj( 'bland', 'blander', 'blandest', normal ).
adj( 'blank', '-', '-', normal ).
adj( 'blas_e', '-', '-', normal ).
adj( 'blasphemous', '-', '-', normal ).
adj( 'blasted', '-', '-', attr ).
adj( 'blatant', '-', '-', normal ).
adj( 'blazing', '-', '-', normal ).
adj( 'bleak', 'bleaker', 'bleakest', normal ).
adj( 'bleary', '-', '-', normal ).
adj( 'bleary-eyed', '-', '-', normal ).
adj( 'blessed', '-', '-', normal ).
adj( 'blind', '-', '-', normal ).
adj( 'blindfold', '-', '-', normal ).
adj( 'blinking', '-', '-', normal ).
adj( 'blissful', '-', '-', normal ).
adj( 'blithe', '-', '-', normal ).
adj( 'blithering', '-', '-', normal ).
adj( 'blithesome', '-', '-', normal ).
adj( 'bloated', '-', '-', normal ).
adj( 'blond', 'blonder', 'blondest', normal ).
adj( 'blonde', '-', '-', normal ).
adj( 'bloodcurdling', '-', '-', normal ).
adj( 'bloodless', '-', '-', normal ).
adj( 'bloodshot', '-', '-', normal ).
adj( 'bloodstained', '-', '-', normal ).
adj( 'bloodthirsty', '-', '-', normal ).
adj( 'bloody', 'bloodier', 'bloodiest', normal ).
adj( 'bloody-minded', '-', '-', normal ).
adj( 'blooming', '-', '-', normal ).
adj( 'blotto', '-', '-', pred ).
adj( 'blowzy', '-', '-', normal ).
adj( 'blue', 'bluer', 'bluest', normal ).
adj( 'blue-collar', '-', '-', normal ).
adj( 'bluff', '-', '-', normal ).
adj( 'bluish', '-', '-', normal ).
adj( 'blunt', 'blunter', 'bluntest', normal ).
adj( 'blushing', '-', '-', normal ).
adj( 'blustery', '-', '-', normal ).
adj( 'boastful', '-', '-', normal ).
adj( 'bodied', '-', '-', affix ).
adj( 'bodily', '-', '-', normal ).
adj( 'boggy', 'boggier', 'boggiest', normal ).
adj( 'bogus', '-', '-', normal ).
adj( 'bohemian', '-', '-', normal ).
adj( 'boisterous', '-', '-', normal ).
adj( 'bold', 'bolder', 'boldest', normal ).
adj( 'bolshy', '-', '-', normal ).
adj( 'bomb-proof', '-', '-', normal ).
adj( 'bombastic', '-', '-', normal ).
adj( 'bona fide', '-', '-', normal ).
adj( 'bone-dry', '-', '-', normal ).
adj( 'bone-idle', '-', '-', normal ).
adj( 'bone-lazy', '-', '-', normal ).
adj( 'boneless', '-', '-', normal ).
adj( 'bonkers', '-', '-', pred ).
adj( 'bonny', 'bonnier', 'bonniest', normal ).
adj( 'bony', 'bonier', 'boniest', normal ).
adj( 'bookable', '-', '-', normal ).
adj( 'bookish', '-', '-', normal ).
adj( 'boon', '-', '-', normal ).
adj( 'boorish', '-', '-', normal ).
adj( 'booted', '-', '-', normal ).
adj( 'bootless', '-', '-', normal ).
adj( 'boozy', 'boozier', 'booziest', normal ).
adj( 'boracic', '-', '-', normal ).
adj( 'boric', '-', '-', normal ).
adj( 'boring', '-', '-', normal ).
adj( 'bosky', '-', '-', normal ).
adj( 'boss-eyed', '-', '-', normal ).
adj( 'bossy', 'bossier', 'bossiest', normal ).
adj( 'botanical', '-', '-', normal ).
adj( 'both', '-', '-', normal ).
adj( 'bothersome', '-', '-', normal ).
adj( 'bottle-fed', '-', '-', normal ).
adj( 'bottle-green', '-', '-', normal ).
adj( 'bottomless', '-', '-', normal ).
adj( 'bouncing', '-', '-', normal ).
adj( 'bouncy', 'bouncier', 'bounciest', normal ).
adj( 'bound', '-', '-', normal ).
adj( 'bounden', '-', '-', normal ).
adj( 'boundless', '-', '-', normal ).
adj( 'bounteous', '-', '-', normal ).
adj( 'bountiful', '-', '-', normal ).
adj( 'bourgeois', '-', '-', normal ).
adj( 'bovine', '-', '-', normal ).
adj( 'boyish', '-', '-', normal ).
adj( 'brackish', '-', '-', normal ).
adj( 'brainless', '-', '-', normal ).
adj( 'brainwashed', '-', '-', normal ).
adj( 'brainy', 'brainier', 'brainiest', normal ).
adj( 'bran-new', '-', '-', normal ).
adj( 'branchy', 'branchier', 'branchiest', normal ).
adj( 'brand-new', '-', '-', normal ).
adj( 'brash', 'brasher', 'brashest', normal ).
adj( 'brassy', 'brassier', 'brassiest', normal ).
adj( 'brave', 'braver', 'bravest', normal ).
adj( 'brawny', 'brawnier', 'brawniest', normal ).
adj( 'brazen', '-', '-', normal ).
adj( 'brazen-faced', '-', '-', normal ).
adj( 'breakable', '-', '-', normal ).
adj( 'breakneck', '-', '-', normal ).
adj( 'breast-fed', '-', '-', normal ).
adj( 'breathless', '-', '-', normal ).
adj( 'breathtaking', '-', '-', normal ).
adj( 'breezy', 'breezier', 'breeziest', normal ).
adj( 'bribable', '-', '-', normal ).
adj( 'bridal', '-', '-', attr ).
adj( 'brief', 'briefer', 'briefest', normal ).
adj( 'bright', 'brighter', 'brightest', normal ).
adj( 'brilliant', '-', '-', normal ).
adj( 'brimful', '-', '-', normal ).
adj( 'brimfull', '-', '-', normal ).
adj( 'brindled', '-', '-', normal ).
adj( 'briny', 'brinier', 'briniest', normal ).
adj( 'brisk', 'brisker', 'briskest', normal ).
adj( 'bristly', 'bristlier', 'bristliest', normal ).
adj( 'brittle', '-', '-', normal ).
adj( 'broad', 'broader', 'broadest', normal ).
adj( 'broad-minded', '-', '-', normal ).
adj( 'broadcasting', '-', '-', normal ).
adj( 'broke', '-', '-', pred ).
adj( 'broken-hearted', '-', '-', normal ).
adj( 'bronchial', '-', '-', normal ).
adj( 'bronchitic', '-', '-', normal ).
adj( 'broody', 'broodier', 'broodiest', normal ).
adj( 'brotherly', '-', '-', normal ).
adj( 'brown', 'browner', 'brownest', normal ).
adj( 'brownish', '-', '-', normal ).
adj( 'brusque', '-', '-', normal ).
adj( 'brutal', '-', '-', normal ).
adj( 'brutish', '-', '-', normal ).
adj( 'bubbly', 'bubblier', 'bubbliest', normal ).
adj( 'bubonic', '-', '-', normal ).
adj( 'bucolic', '-', '-', normal ).
adj( 'budding', '-', '-', normal ).
adj( 'budgetary', '-', '-', normal ).
adj( 'built-in', '-', '-', normal ).
adj( 'built-up', '-', '-', normal ).
adj( 'bulbous', '-', '-', normal ).
adj( 'bulky', 'bulkier', 'bulkiest', normal ).
adj( 'bull-headed', '-', '-', normal ).
adj( 'bullet-headed', '-', '-', normal ).
adj( 'bulletproof', '-', '-', normal ).
adj( 'bully', '-', '-', normal ).
adj( 'bum', '-', '-', normal ).
adj( 'bumptious', '-', '-', normal ).
adj( 'bumpy', 'bumpier', 'bumpiest', normal ).
adj( 'bungaloid', '-', '-', normal ).
adj( 'buoyant', '-', '-', normal ).
adj( 'burdensome', '-', '-', normal ).
adj( 'bureaucratic', '-', '-', normal ).
adj( 'burglar-proof', '-', '-', normal ).
adj( 'burglarious', '-', '-', normal ).
adj( 'burly', 'burlier', 'burliest', normal ).
adj( 'burning', '-', '-', normal ).
adj( 'bushy', '-', '-', normal ).
adj( 'businesslike', '-', '-', normal ).
adj( 'busy', 'busier', 'busiest', normal ).
adj( 'butch', '-', '-', normal ).
adj( 'buttoned-up', '-', '-', normal ).
adj( 'buxom', '-', '-', normal ).
adj( 'bygone', '-', '-', normal ).
adj( 'cacophonous', '-', '-', normal ).
adj( 'cadaverous', '-', '-', normal ).
adj( 'caddish', '-', '-', normal ).
adj( 'cagey', '-', '-', normal ).
adj( 'calamitous', '-', '-', normal ).
adj( 'calculable', '-', '-', normal ).
adj( 'calculating', '-', '-', normal ).
adj( 'callous', '-', '-', normal ).
adj( 'callow', '-', '-', normal ).
adj( 'calm', 'calmer', 'calmest', normal ).
adj( 'calorific', '-', '-', normal ).
adj( 'camp', '-', '-', normal ).
adj( 'camphorated', '-', '-', normal ).
adj( 'cancerous', '-', '-', normal ).
adj( 'candid', '-', '-', normal ).
adj( 'candied', '-', '-', normal ).
adj( 'canine', '-', '-', normal ).
adj( 'cankerous', '-', '-', normal ).
adj( 'canned', '-', '-', normal ).
adj( 'cannibalistic', '-', '-', normal ).
adj( 'canny', 'cannier', 'canniest', normal ).
adj( 'canonical', '-', '-', normal ).
adj( 'canopied', '-', '-', normal ).
adj( 'cantankerous', '-', '-', normal ).
adj( 'capable', '-', '-', normal ).
adj( 'capacious', '-', '-', normal ).
adj( 'capital', '-', '-', normal ).
adj( 'capitalism', '-', '-', normal ).
adj( 'capitalistic', '-', '-', normal ).
adj( 'capricious', '-', '-', normal ).
adj( 'captious', '-', '-', normal ).
adj( 'captive', '-', '-', normal ).
adj( 'carbolic', '-', '-', normal ).
adj( 'carbonated', '-', '-', normal ).
adj( 'carbonic', '-', '-', normal ).
adj( 'carboniferous', '-', '-', normal ).
adj( 'cardiac', '-', '-', normal ).
adj( 'cardinal', '-', '-', normal ).
adj( 'carefree', '-', '-', normal ).
adj( 'careful', '-', '-', normal ).
adj( 'careladen', '-', '-', normal ).
adj( 'careless', '-', '-', normal ).
adj( 'caressing', '-', '-', normal ).
adj( 'careworn', '-', '-', normal ).
adj( 'carious', '-', '-', normal ).
adj( 'carmine', '-', '-', normal ).
adj( 'carnal', '-', '-', normal ).
adj( 'carnivorous', '-', '-', normal ).
adj( 'carpal', '-', '-', normal ).
adj( 'carroty', '-', '-', normal ).
adj( 'carsick', '-', '-', normal ).
adj( 'cartilaginous', '-', '-', normal ).
adj( 'case-hardened', '-', '-', normal ).
adj( 'cashable', '-', '-', normal ).
adj( 'cast-iron', '-', '-', normal ).
adj( 'cast-off', '-', '-', attr ).
adj( 'castellated', '-', '-', normal ).
adj( 'casual', '-', '-', normal ).
adj( 'casuistic', '-', '-', normal ).
adj( 'casuistical', '-', '-', normal ).
adj( 'cataclysmic', '-', '-', normal ).
adj( 'cataleptic', '-', '-', normal ).
adj( 'catalytic', '-', '-', normal ).
adj( 'catastrophic', '-', '-', normal ).
adj( 'catching', '-', '-', normal ).
adj( 'catchpenny', '-', '-', normal ).
adj( 'catchy', 'catchier', 'catchiest', normal ).
adj( 'categorical', '-', '-', normal ).
adj( 'catholic', '-', '-', normal ).
adj( 'cattish', '-', '-', normal ).
adj( 'catty', 'cattier', 'cattiest', normal ).
adj( 'causal', '-', '-', normal ).
adj( 'causative', '-', '-', normal ).
adj( 'causeless', '-', '-', normal ).
adj( 'caustic', '-', '-', normal ).
adj( 'cautionary', '-', '-', normal ).
adj( 'cautious', '-', '-', normal ).
adj( 'cavalier', '-', '-', normal ).
adj( 'cavernous', '-', '-', normal ).
adj( 'ceaseless', '-', '-', normal ).
adj( 'celebrated', '-', '-', normal ).
adj( 'celestial', '-', '-', normal ).
adj( 'cellular', '-', '-', normal ).
adj( 'censorious', '-', '-', normal ).
adj( 'centenarian', '-', '-', normal ).
adj( 'centenary', '-', '-', normal ).
adj( 'centennial', '-', '-', normal ).
adj( 'centigrade', '-', '-', normal ).
adj( 'central', '-', '-', normal ).
adj( 'centrifugal', '-', '-', normal ).
adj( 'centripetal', '-', '-', normal ).
adj( 'ceramic', '-', '-', normal ).
adj( 'cerebral', '-', '-', normal ).
adj( 'ceremonial', '-', '-', normal ).
adj( 'ceremonious', '-', '-', normal ).
adj( 'cerise', '-', '-', normal ).
adj( 'certain', '-', '-', normal ).
adj( 'certifiable', '-', '-', normal ).
adj( 'certificated', '-', '-', normal ).
adj( 'cerulean', '-', '-', normal ).
adj( 'cervical', '-', '-', normal ).
adj( 'chalky', 'chalkier', 'chalkiest', normal ).
adj( 'champion', '-', '-', normal ).
adj( 'chancy', '-', '-', normal ).
adj( 'changeable', '-', '-', normal ).
adj( 'changeful', '-', '-', normal ).
adj( 'changeless', '-', '-', normal ).
adj( 'chaotic', '-', '-', normal ).
adj( 'chapfallen', '-', '-', normal ).
adj( 'characteristic', '-', '-', normal ).
adj( 'characterless', '-', '-', normal ).
adj( 'chargeable', '-', '-', normal ).
adj( 'charismatic', '-', '-', normal ).
adj( 'charitable', '-', '-', normal ).
adj( 'charming', '-', '-', normal ).
adj( 'chary', '-', '-', normal ).
adj( 'chaste', '-', '-', normal ).
adj( 'chatty', 'chattier', 'chattiest', normal ).
adj( 'chauvinistic', '-', '-', normal ).
adj( 'cheap', 'cheaper', 'cheapest', normal ).
adj( 'cheapjack', '-', '-', normal ).
adj( 'checked', '-', '-', normal ).
adj( 'cheeked', '-', '-', affix ).
adj( 'cheeky', 'cheekier', 'cheekiest', normal ).
adj( 'cheerful', '-', '-', normal ).
adj( 'cheering', '-', '-', normal ).
adj( 'cheerless', '-', '-', normal ).
adj( 'cheery', '-', '-', normal ).
adj( 'cheeseparing', '-', '-', normal ).
adj( 'chemical', '-', '-', normal ).
adj( 'cherry', '-', '-', normal ).
adj( 'cherubic', '-', '-', normal ).
adj( 'chic', '-', '-', normal ).
adj( 'chichi', '-', '-', normal ).
adj( 'chicken-hearted', '-', '-', normal ).
adj( 'chief', '-', '-', normal ).
adj( 'chilblained', '-', '-', normal ).
adj( 'childish', '-', '-', normal ).
adj( 'childless', '-', '-', normal ).
adj( 'childlike', '-', '-', normal ).
adj( 'chill', '-', '-', normal ).
adj( 'chilly', 'chillier', 'chilliest', normal ).
adj( 'chimerical', '-', '-', normal ).
adj( 'chirpy', '-', '-', normal ).
adj( 'chivalrous', '-', '-', normal ).
adj( 'chock-a-block', '-', '-', normal ).
adj( 'chock-full', '-', '-', normal ).
adj( 'choice', '-', '-', normal ).
adj( 'choleric', '-', '-', normal ).
adj( 'choosey', '-', '-', normal ).
adj( 'choosy', 'choosier', 'choosiest', normal ).
adj( 'choppy', 'choppier', 'choppiest', normal ).
adj( 'choral', '-', '-', normal ).
adj( 'chromatic', '-', '-', normal ).
adj( 'chronic', '-', '-', normal ).
adj( 'chronological', '-', '-', normal ).
adj( 'chubby', 'chubbier', 'chubbiest', normal ).
adj( 'chummy', 'chummier', 'chummiest', normal ).
adj( 'chunky', 'chunkier', 'chunkiest', normal ).
adj( 'churlish', '-', '-', normal ).
adj( 'cigar-shaped', '-', '-', normal ).
adj( 'cinematic', '-', '-', normal ).
adj( 'circuitous', '-', '-', normal ).
adj( 'circular', '-', '-', normal ).
adj( 'circumspect', '-', '-', normal ).
adj( 'circumstantial', '-', '-', normal ).
adj( 'cissy', '-', '-', normal ).
adj( 'citric', '-', '-', normal ).
adj( 'citrous', '-', '-', normal ).
adj( 'civic', '-', '-', normal ).
adj( 'civil', '-', '-', normal ).
adj( 'civilian', '-', '-', normal ).
adj( 'clammy', 'clammier', 'clammiest', normal ).
adj( 'clamorous', '-', '-', normal ).
adj( 'clandestine', '-', '-', normal ).
adj( 'clangorous', '-', '-', normal ).
adj( 'clannish', '-', '-', normal ).
adj( 'class-conscious', '-', '-', normal ).
adj( 'classic', '-', '-', normal ).
adj( 'classical', '-', '-', normal ).
adj( 'classifiable', '-', '-', normal ).
adj( 'classified', '-', '-', normal ).
adj( 'classless', '-', '-', normal ).
adj( 'classy', 'classier', 'classiest', normal ).
adj( 'claustrophobic', '-', '-', normal ).
adj( 'clayey', '-', '-', normal ).
adj( 'clean', 'cleaner', 'cleanest', normal ).
adj( 'clean-bowled', '-', '-', normal ).
adj( 'clean-cut', '-', '-', normal ).
adj( 'clean-limbed', '-', '-', normal ).
adj( 'clean-living', '-', '-', normal ).
adj( 'clean-shaven', '-', '-', normal ).
adj( 'cleanly', 'cleanlier', 'cleanliest', normal ).
adj( 'clear', 'clearer', 'clearest', normal ).
adj( 'clear-headed', '-', '-', normal ).
adj( 'clear-sighted', '-', '-', normal ).
adj( 'clement', '-', '-', normal ).
adj( 'clerical', '-', '-', normal ).
adj( 'clever', 'cleverer', 'cleverest', normal ).
adj( 'climactic', '-', '-', normal ).
adj( 'climatic', '-', '-', normal ).
adj( 'clinical', '-', '-', normal ).
adj( 'clinker-built', '-', '-', normal ).
adj( 'clip-on', '-', '-', normal ).
adj( 'cliquish', '-', '-', normal ).
adj( 'cloggy', 'cloggier', 'cloggiest', normal ).
adj( 'close', 'closer', 'closest', normal ).
adj( 'close-cropped', '-', '-', normal ).
adj( 'close-cut', '-', '-', normal ).
adj( 'close-fisted', '-', '-', normal ).
adj( 'close-fitting', '-', '-', normal ).
adj( 'close-grained', '-', '-', normal ).
adj( 'close-hauled', '-', '-', normal ).
adj( 'close-set', '-', '-', normal ).
adj( 'closet', '-', '-', attr ).
adj( 'cloud-capped', '-', '-', normal ).
adj( 'cloudless', '-', '-', normal ).
adj( 'cloudy', 'cloudier', 'cloudiest', normal ).
adj( 'clownish', '-', '-', normal ).
adj( 'club-footed', '-', '-', normal ).
adj( 'clubbable', '-', '-', normal ).
adj( 'clumsy', 'clumsier', 'clumsiest', normal ).
adj( 'coarse', 'coarser', 'coarsest', normal ).
adj( 'coastal', '-', '-', normal ).
adj( 'coastwise', '-', '-', normal ).
adj( 'cock-a-hoop', '-', '-', normal ).
adj( 'cockeyed', '-', '-', normal ).
adj( 'cockney', '-', '-', normal ).
adj( 'cocksure', '-', '-', normal ).
adj( 'cocky', 'cockier', 'cockiest', normal ).
adj( 'coeducational', '-', '-', normal ).
adj( 'coercive', '-', '-', normal ).
adj( 'coeval', '-', '-', normal ).
adj( 'cogent', '-', '-', normal ).
adj( 'cognate', '-', '-', normal ).
adj( 'cognizant', '-', '-', normal ).
adj( 'coherent', '-', '-', normal ).
adj( 'cohesive', '-', '-', normal ).
adj( 'coincident', '-', '-', normal ).
adj( 'coincidental', '-', '-', normal ).
adj( 'cold', 'colder', 'coldest', normal ).
adj( 'cold-blooded', '-', '-', normal ).
adj( 'cold-hearted', '-', '-', normal ).
adj( 'collapsable', '-', '-', normal ).
adj( 'collapsible', '-', '-', normal ).
adj( 'collateral', '-', '-', normal ).
adj( 'collect', '-', '-', normal ).
adj( 'collected', '-', '-', normal ).
adj( 'collective', '-', '-', normal ).
adj( 'collegiate', '-', '-', normal ).
adj( 'colloquial', '-', '-', normal ).
adj( 'collusive', '-', '-', normal ).
adj( 'colonial', '-', '-', normal ).
adj( 'colonnaded', '-', '-', normal ).
adj( 'colossal', '-', '-', normal ).
adj( 'colour-blind', '-', '-', normal ).
adj( 'coloured', '-', '-', normal ).
adj( 'colourful', '-', '-', normal ).
adj( 'colourless', '-', '-', normal ).
adj( 'coltish', '-', '-', normal ).
adj( 'comatose', '-', '-', normal ).
adj( 'combatant', '-', '-', normal ).
adj( 'combative', '-', '-', normal ).
adj( 'combustible', '-', '-', normal ).
adj( 'come-at-able', '-', '-', normal ).
adj( 'comely', 'comelier', 'comeliest', normal ).
adj( 'comfortable', '-', '-', normal ).
adj( 'comfortless', '-', '-', normal ).
adj( 'comfy', 'comfier', 'comfiest', normal ).
adj( 'comic', '-', '-', normal ).
adj( 'comical', '-', '-', normal ).
adj( 'coming', '-', '-', normal ).
adj( 'commanding', '-', '-', normal ).
adj( 'commemorative', '-', '-', normal ).
adj( 'commendable', '-', '-', normal ).
adj( 'commensurable', '-', '-', normal ).
adj( 'commensurate', '-', '-', normal ).
adj( 'commercial', '-', '-', normal ).
adj( 'comminatory', '-', '-', normal ).
adj( 'commissioned', '-', '-', normal ).
adj( 'commodious', '-', '-', normal ).
adj( 'common', 'commoner', 'commonest', normal ).
adj( 'commonplace', '-', '-', normal ).
adj( 'communal', '-', '-', normal ).
adj( 'communicable', '-', '-', normal ).
adj( 'communicative', '-', '-', normal ).
adj( 'communist', '-', '-', normal ).
adj( 'commutable', '-', '-', normal ).
adj( 'compact', '-', '-', normal ).
adj( 'companionable', '-', '-', normal ).
adj( 'comparable', '-', '-', normal ).
adj( 'comparative', '-', '-', normal ).
adj( 'compassionate', '-', '-', normal ).
adj( 'compatible', '-', '-', normal ).
adj( 'compendious', '-', '-', normal ).
adj( 'compensatory', '-', '-', normal ).
adj( 'competent', '-', '-', normal ).
adj( 'competitive', '-', '-', normal ).
adj( 'complacent', '-', '-', normal ).
adj( 'complaisant', '-', '-', normal ).
adj( 'complementary', '-', '-', normal ).
adj( 'complete', '-', '-', normal ).
adj( 'complex', '-', '-', normal ).
adj( 'compliant', '-', '-', normal ).
adj( 'complicated', '-', '-', normal ).
adj( 'complimentary', '-', '-', normal ).
adj( 'component', '-', '-', normal ).
adj( 'compos mentis', '-', '-', normal ).
adj( 'composed', '-', '-', normal ).
adj( 'composite', '-', '-', normal ).
adj( 'compound', '-', '-', normal ).
adj( 'comprehensible', '-', '-', normal ).
adj( 'comprehensive', '-', '-', normal ).
adj( 'compulsive', '-', '-', normal ).
adj( 'compulsory', '-', '-', normal ).
adj( 'computational', '-', '-', normal ).
adj( 'concave', '-', '-', normal ).
adj( 'conceited', '-', '-', normal ).
adj( 'conceivable', '-', '-', normal ).
adj( 'concentrated', '-', '-', normal ).
adj( 'concentric', '-', '-', normal ).
adj( 'conceptual', '-', '-', normal ).
adj( 'concerned', '-', '-', normal ).
adj( 'concerted', '-', '-', normal ).
adj( 'concessive', '-', '-', normal ).
adj( 'conciliatory', '-', '-', normal ).
adj( 'concise', '-', '-', normal ).
adj( 'conclusive', '-', '-', normal ).
adj( 'concomitant', '-', '-', normal ).
adj( 'concordant', '-', '-', normal ).
adj( 'concrete', '-', '-', normal ).
adj( 'concurrent', '-', '-', normal ).
adj( 'condescending', '-', '-', normal ).
adj( 'condign', '-', '-', normal ).
adj( 'conditional', '-', '-', normal ).
adj( 'conditioned', '-', '-', normal ).
adj( 'conducive', '-', '-', normal ).
adj( 'conductive', '-', '-', normal ).
adj( 'confederate', '-', '-', normal ).
adj( 'confident', '-', '-', normal ).
adj( 'confidential', '-', '-', normal ).
adj( 'confiding', '-', '-', normal ).
adj( 'configured', '-', '-', normal ).
adj( 'confined', '-', '-', normal ).
adj( 'confirmed', '-', '-', normal ).
adj( 'conflicting', '-', '-', normal ).
adj( 'confluent', '-', '-', normal ).
adj( 'conformable', '-', '-', normal ).
adj( 'confounded', '-', '-', normal ).
adj( 'congenial', '-', '-', normal ).
adj( 'congenital', '-', '-', normal ).
adj( 'congested', '-', '-', normal ).
adj( 'conglomerate', '-', '-', normal ).
adj( 'congratulatory', '-', '-', normal ).
adj( 'congregational', '-', '-', normal ).
adj( 'congressional', '-', '-', normal ).
adj( 'congruent', '-', '-', normal ).
adj( 'congruous', '-', '-', normal ).
adj( 'conic', '-', '-', normal ).
adj( 'conical', '-', '-', normal ).
adj( 'coniferous', '-', '-', normal ).
adj( 'conjectural', '-', '-', normal ).
adj( 'conjoint', '-', '-', normal ).
adj( 'conjugal', '-', '-', normal ).
adj( 'conjunctive', '-', '-', normal ).
adj( 'connective', '-', '-', normal ).
adj( 'conning', '-', '-', attr ).
adj( 'connubial', '-', '-', normal ).
adj( 'conscience-smitten', '-', '-', normal ).
adj( 'conscientious', '-', '-', normal ).
adj( 'conscious', '-', '-', normal ).
adj( 'consecutive', '-', '-', normal ).
adj( 'consequent', '-', '-', normal ).
adj( 'consequential', '-', '-', normal ).
adj( 'conservative', '-', '-', normal ).
adj( 'considerable', '-', '-', normal ).
adj( 'considerate', '-', '-', normal ).
adj( 'consistent', '-', '-', normal ).
adj( 'consolable', '-', '-', normal ).
adj( 'consolatory', '-', '-', normal ).
adj( 'consonant', '-', '-', normal ).
adj( 'conspicuous', '-', '-', normal ).
adj( 'conspiratorial', '-', '-', normal ).
adj( 'constant', '-', '-', normal ).
adj( 'constipated', '-', '-', normal ).
adj( 'constituent', '-', '-', normal ).
adj( 'constitutional', '-', '-', normal ).
adj( 'constitutive', '-', '-', normal ).
adj( 'constrained', '-', '-', normal ).
adj( 'constructional', '-', '-', normal ).
adj( 'constructive', '-', '-', normal ).
adj( 'consular', '-', '-', normal ).
adj( 'consultative', '-', '-', normal ).
adj( 'consuming', '-', '-', normal ).
adj( 'consummate', '-', '-', normal ).
adj( 'consumptive', '-', '-', normal ).
adj( 'contagious', '-', '-', normal ).
adj( 'contemplative', '-', '-', normal ).
adj( 'contemporaneous', '-', '-', normal ).
adj( 'contemporary', '-', '-', normal ).
adj( 'contemptible', '-', '-', normal ).
adj( 'contemptuous', '-', '-', normal ).
adj( 'content', '-', '-', normal ).
adj( 'contented', '-', '-', normal ).
adj( 'contentious', '-', '-', normal ).
adj( 'conterminous', '-', '-', normal ).
adj( 'contextual', '-', '-', normal ).
adj( 'contiguous', '-', '-', normal ).
adj( 'continent', '-', '-', normal ).
adj( 'continental', '-', '-', normal ).
adj( 'contingent', '-', '-', normal ).
adj( 'continual', '-', '-', normal ).
adj( 'continuous', '-', '-', normal ).
adj( 'contraceptive', '-', '-', normal ).
adj( 'contractible', '-', '-', normal ).
adj( 'contractile', '-', '-', normal ).
adj( 'contractual', '-', '-', normal ).
adj( 'contradictory', '-', '-', normal ).
adj( 'contrapuntal', '-', '-', normal ).
adj( 'contrary', '-', '-', normal ).
adj( 'contrary', '-', '-', normal ).
adj( 'contributory', '-', '-', normal ).
adj( 'contrite', '-', '-', normal ).
adj( 'controllable', '-', '-', normal ).
adj( 'controversial', '-', '-', normal ).
adj( 'contumacious', '-', '-', normal ).
adj( 'contumelious', '-', '-', normal ).
adj( 'convalescent', '-', '-', normal ).
adj( 'convenient', '-', '-', normal ).
adj( 'conventional', '-', '-', normal ).
adj( 'convergent', '-', '-', normal ).
adj( 'conversant', '-', '-', normal ).
adj( 'conversational', '-', '-', normal ).
adj( 'converse', '-', '-', normal ).
adj( 'converted', '-', '-', normal ).
adj( 'convertible', '-', '-', normal ).
adj( 'convex', '-', '-', normal ).
adj( 'convincible', '-', '-', normal ).
adj( 'convincing', '-', '-', normal ).
adj( 'convivial', '-', '-', normal ).
adj( 'convoluted', '-', '-', normal ).
adj( 'convulsive', '-', '-', normal ).
adj( 'cool', 'cooler', 'coolest', normal ).
adj( 'cool-headed', '-', '-', normal ).
adj( 'cooperative', '-', '-', normal ).
adj( 'coordinate', '-', '-', normal ).
adj( 'copious', '-', '-', normal ).
adj( 'copper-bottomed', '-', '-', normal ).
adj( 'copulative', '-', '-', normal ).
adj( 'coquettish', '-', '-', normal ).
adj( 'coral', '-', '-', normal ).
adj( 'cordial', '-', '-', normal ).
adj( 'cordon bleu', '-', '-', normal ).
adj( 'corked', '-', '-', normal ).
adj( 'cornered', '-', '-', normal ).
adj( 'corny', 'cornier', 'corniest', normal ).
adj( 'coronary', '-', '-', normal ).
adj( 'corporal', '-', '-', normal ).
adj( 'corporate', '-', '-', normal ).
adj( 'corporeal', '-', '-', normal ).
adj( 'corpulent', '-', '-', normal ).
adj( 'correct', '-', '-', normal ).
adj( 'corrective', '-', '-', normal ).
adj( 'correlative', '-', '-', normal ).
adj( 'corresponding', '-', '-', normal ).
adj( 'corrigible', '-', '-', normal ).
adj( 'corroborative', '-', '-', normal ).
adj( 'corrosive', '-', '-', normal ).
adj( 'corrupt', '-', '-', normal ).
adj( 'corruptible', '-', '-', normal ).
adj( 'cortical', '-', '-', normal ).
adj( 'cosher', '-', '-', normal ).
adj( 'cosignatory', '-', '-', normal ).
adj( 'cosmetic', '-', '-', normal ).
adj( 'cosmic', '-', '-', normal ).
adj( 'cosmopolitan', '-', '-', normal ).
adj( 'costive', '-', '-', normal ).
adj( 'costly', 'costlier', 'costliest', normal ).
adj( 'cosy', 'cosier', 'cosiest', normal ).
adj( 'coterminous', '-', '-', normal ).
adj( 'couchant', '-', '-', normal ).
adj( 'countable', '-', '-', normal ).
adj( 'counter-revolutionary', '-', '-', normal ).
adj( 'counterfeit', '-', '-', normal ).
adj( 'countless', '-', '-', normal ).
adj( 'countrified', '-', '-', normal ).
adj( 'courageous', '-', '-', normal ).
adj( 'courteous', '-', '-', normal ).
adj( 'courtly', 'courtlier', 'courtliest', normal ).
adj( 'cousinly', '-', '-', normal ).
adj( 'covering', '-', '-', normal ).
adj( 'covert', '-', '-', normal ).
adj( 'covetous', '-', '-', normal ).
adj( 'cowardly', '-', '-', normal ).
adj( 'coy', 'coyer', 'coyest', normal ).
adj( 'cozy', 'cozier', 'coziest', normal ).
adj( 'crabbed', '-', '-', normal ).
adj( 'crackers', '-', '-', pred ).
adj( 'crafty', 'craftier', 'craftiest', normal ).
adj( 'cragged', '-', '-', normal ).
adj( 'craggy', 'craggier', 'craggiest', normal ).
adj( 'cram-full', '-', '-', normal ).
adj( 'cramped', '-', '-', normal ).
adj( 'cranial', '-', '-', normal ).
adj( 'cranky', 'crankier', 'crankiest', normal ).
adj( 'crannied', '-', '-', normal ).
adj( 'crass', '-', '-', normal ).
adj( 'craven', '-', '-', normal ).
adj( 'crazed', '-', '-', normal ).
adj( 'crazy', 'crazier', 'craziest', normal ).
adj( 'creaky', 'creakier', 'creakiest', normal ).
adj( 'creamy', 'creamier', 'creamiest', normal ).
adj( 'creative', '-', '-', normal ).
adj( 'credible', '-', '-', normal ).
adj( 'credit-worthy', '-', '-', normal ).
adj( 'creditable', '-', '-', normal ).
adj( 'credulous', '-', '-', normal ).
adj( 'creepy', 'creepier', 'creepiest', normal ).
adj( 'crenellated', '-', '-', normal ).
adj( 'crepuscular', '-', '-', normal ).
adj( 'crested', '-', '-', normal ).
adj( 'crestfallen', '-', '-', normal ).
adj( 'cretaceous', '-', '-', normal ).
adj( 'cretinous', '-', '-', normal ).
adj( 'criminal', '-', '-', normal ).
adj( 'crimson', '-', '-', normal ).
adj( 'crinkly', 'crinklier', 'crinkliest', normal ).
adj( 'crisp', 'crisper', 'crispest', normal ).
adj( 'crisscross', '-', '-', normal ).
adj( 'critical', '-', '-', normal ).
adj( 'crook-back', '-', '-', normal ).
adj( 'crook-backed', '-', '-', normal ).
adj( 'crooked', '-', '-', normal ).
adj( 'cross', '-', '-', normal ).
adj( 'cross-grained', '-', '-', normal ).
adj( 'crossbred', '-', '-', normal ).
adj( 'crosscountry', '-', '-', normal ).
adj( 'crosscut', '-', '-', normal ).
adj( 'crosseyed', '-', '-', normal ).
adj( 'crotchety', '-', '-', normal ).
adj( 'crowded', '-', '-', normal ).
adj( 'crowning', '-', '-', normal ).
adj( 'crucial', '-', '-', normal ).
adj( 'cruciform', '-', '-', normal ).
adj( 'crude', 'cruder', 'crudest', normal ).
adj( 'cruel', 'crueller', 'cruellest', normal ).
adj( 'crumbly', 'crumblier', 'crumbliest', normal ).
adj( 'crushing', '-', '-', normal ).
adj( 'crusted', '-', '-', normal ).
adj( 'crusty', 'crustier', 'crustiest', normal ).
adj( 'crying', '-', '-', attr ).
adj( 'cryptic', '-', '-', normal ).
adj( 'crystalline', '-', '-', normal ).
adj( 'cubic', '-', '-', normal ).
adj( 'cubical', '-', '-', normal ).
adj( 'cuddlesome', '-', '-', normal ).
adj( 'cuddly', 'cuddlier', 'cuddliest', normal ).
adj( 'culinary', '-', '-', normal ).
adj( 'culpable', '-', '-', normal ).
adj( 'cultivable', '-', '-', normal ).
adj( 'cultivated', '-', '-', normal ).
adj( 'cultural', '-', '-', normal ).
adj( 'cultured', '-', '-', normal ).
adj( 'cumbersome', '-', '-', normal ).
adj( 'cumbrous', '-', '-', normal ).
adj( 'cumulative', '-', '-', normal ).
adj( 'cuneiform', '-', '-', normal ).
adj( 'cunning', '-', '-', normal ).
adj( 'cupric', '-', '-', normal ).
adj( 'curable', '-', '-', normal ).
adj( 'curative', '-', '-', normal ).
adj( 'curious', '-', '-', normal ).
adj( 'curly', 'curlier', 'curliest', normal ).
adj( 'current', '-', '-', normal ).
adj( 'currish', '-', '-', normal ).
adj( 'cursed', '-', '-', normal ).
adj( 'cursive', '-', '-', normal ).
adj( 'cursory', '-', '-', normal ).
adj( 'curst', '-', '-', normal ).
adj( 'curt', '-', '-', normal ).
adj( 'cushy', 'cushier', 'cushiest', normal ).
adj( 'cussed', '-', '-', normal ).
adj( 'custodial', '-', '-', normal ).
adj( 'custom-built', '-', '-', normal ).
adj( 'custom-made', '-', '-', normal ).
adj( 'customary', '-', '-', normal ).
adj( 'cut-price', '-', '-', attr ).
adj( 'cut-rate', '-', '-', attr ).
adj( 'cut-throat', '-', '-', attr ).
adj( 'cute', 'cuter', 'cutest', normal ).
adj( 'cutting', '-', '-', normal ).
adj( 'cybernetic', '-', '-', normal ).
adj( 'cyclic', '-', '-', normal ).
adj( 'cyclical', '-', '-', normal ).
adj( 'cyclonic', '-', '-', normal ).
adj( 'cylindrical', '-', '-', normal ).
adj( 'cynical', '-', '-', normal ).
adj( 'd_ecollet_e', '-', '-', normal ).
adj( 'd_emod_e', '-', '-', normal ).
adj( 'dactylic', '-', '-', normal ).
adj( 'daft', 'dafter', 'daftest', normal ).
adj( 'daily', '-', '-', normal ).
adj( 'dainty', 'daintier', 'daintiest', normal ).
adj( 'damascene', '-', '-', normal ).
adj( 'damn', '-', '-', normal ).
adj( 'damnable', '-', '-', normal ).
adj( 'damned', '-', '-', normal ).
adj( 'damp', 'damper', 'dampest', normal ).
adj( 'dampish', '-', '-', normal ).
adj( 'dancing', '-', '-', normal ).
adj( 'dandified', '-', '-', normal ).
adj( 'dandy', '-', '-', normal ).
adj( 'dangerous', '-', '-', normal ).
adj( 'dank', 'danker', 'dankest', normal ).
adj( 'dapper', '-', '-', normal ).
adj( 'dapple-grey', '-', '-', normal ).
adj( 'daring', '-', '-', normal ).
adj( 'dark', 'darker', 'darkest', normal ).
adj( 'dashing', '-', '-', normal ).
adj( 'dastardly', '-', '-', normal ).
adj( 'datable', '-', '-', normal ).
adj( 'dated', '-', '-', normal ).
adj( 'dateless', '-', '-', normal ).
adj( 'dative', '-', '-', normal ).
adj( 'daughterly', '-', '-', normal ).
adj( 'dauntless', '-', '-', normal ).
adj( 'daylight-saving', '-', '-', normal ).
adj( 'daylong', '-', '-', normal ).
adj( 'de facto', '-', '-', normal ).
adj( 'de jure', '-', '-', normal ).
adj( 'de luxe', '-', '-', normal ).
adj( 'de rigeur', '-', '-', pred ).
adj( 'de trop', '-', '-', pred ).
adj( 'dead', '-', '-', normal ).
adj( 'deadlocked', '-', '-', normal ).
adj( 'deadly', 'deadlier', 'deadliest', normal ).
adj( 'deadpan', '-', '-', normal ).
adj( 'deaf', 'deafer', 'deafest', normal ).
adj( 'dear', 'dearer', 'dearest', normal ).
adj( 'deathless', '-', '-', normal ).
adj( 'deathlike', '-', '-', normal ).
adj( 'deathly', '-', '-', normal ).
adj( 'debatable', '-', '-', normal ).
adj( 'debonair', '-', '-', normal ).
adj( 'decadent', '-', '-', normal ).
adj( 'decasyllabic', '-', '-', normal ).
adj( 'deceitful', '-', '-', normal ).
adj( 'decent', '-', '-', normal ).
adj( 'deceptive', '-', '-', normal ).
adj( 'decided', '-', '-', normal ).
adj( 'deciduous', '-', '-', normal ).
adj( 'decimal', '-', '-', normal ).
adj( 'decipherable', '-', '-', normal ).
adj( 'decisive', '-', '-', normal ).
adj( 'deckle-edged', '-', '-', normal ).
adj( 'declamatory', '-', '-', normal ).
adj( 'declarable', '-', '-', normal ).
adj( 'decorative', '-', '-', normal ).
adj( 'decorous', '-', '-', normal ).
adj( 'decrepit', '-', '-', normal ).
adj( 'deductible', '-', '-', normal ).
adj( 'deductive', '-', '-', normal ).
adj( 'deep', 'deeper', 'deepest', normal ).
adj( 'deep-laid', '-', '-', normal ).
adj( 'deep-mined', '-', '-', normal ).
adj( 'deep-rooted', '-', '-', normal ).
adj( 'deep-sea', '-', '-', attr ).
adj( 'deep-seated', '-', '-', normal ).
adj( 'deep-water', '-', '-', attr ).
adj( 'defamatory', '-', '-', normal ).
adj( 'defective', '-', '-', normal ).
adj( 'defenceless', '-', '-', normal ).
adj( 'defensible', '-', '-', normal ).
adj( 'defensive', '-', '-', normal ).
adj( 'deferential', '-', '-', normal ).
adj( 'defiant', '-', '-', normal ).
adj( 'deficient', '-', '-', normal ).
adj( 'definable', '-', '-', normal ).
adj( 'definite', '-', '-', normal ).
adj( 'definitive', '-', '-', normal ).
adj( 'deflationary', '-', '-', normal ).
adj( 'deformed', '-', '-', normal ).
adj( 'deft', '-', '-', normal ).
adj( 'defunct', '-', '-', normal ).
adj( 'degenerate', '-', '-', normal ).
adj( 'delayed-action', '-', '-', normal ).
adj( 'delectable', '-', '-', normal ).
adj( 'deleterious', '-', '-', normal ).
adj( 'deliberate', '-', '-', normal ).
adj( 'deliberative', '-', '-', normal ).
adj( 'delicate', '-', '-', normal ).
adj( 'delicious', '-', '-', normal ).
adj( 'delightful', '-', '-', normal ).
adj( 'delinquent', '-', '-', normal ).
adj( 'deliquescent', '-', '-', normal ).
adj( 'delirious', '-', '-', normal ).
adj( 'delusive', '-', '-', normal ).
adj( 'demagogic', '-', '-', normal ).
adj( 'demented', '-', '-', normal ).
adj( 'democratic', '-', '-', normal ).
adj( 'demographic', '-', '-', normal ).
adj( 'demoniac', '-', '-', normal ).
adj( 'demoniacal', '-', '-', normal ).
adj( 'demonic', '-', '-', normal ).
adj( 'demonstrable', '-', '-', normal ).
adj( 'demonstrative', '-', '-', normal ).
adj( 'demotic', '-', '-', normal ).
adj( 'demure', '-', '-', normal ).
adj( 'denary', '-', '-', normal ).
adj( 'denatured', '-', '-', normal ).
adj( 'deniable', '-', '-', normal ).
adj( 'denominational', '-', '-', normal ).
adj( 'dense', 'denser', 'densest', normal ).
adj( 'dental', '-', '-', normal ).
adj( 'departed', '-', '-', normal ).
adj( 'departmental', '-', '-', normal ).
adj( 'dependable', '-', '-', normal ).
adj( 'dependent', '-', '-', normal ).
adj( 'depilatory', '-', '-', normal ).
adj( 'deplorable', '-', '-', normal ).
adj( 'depreciatory', '-', '-', normal ).
adj( 'depressive', '-', '-', normal ).
adj( 'deprived', '-', '-', normal ).
adj( 'derelict', '-', '-', normal ).
adj( 'derisive', '-', '-', normal ).
adj( 'derisory', '-', '-', normal ).
adj( 'derivative', '-', '-', normal ).
adj( 'derogatory', '-', '-', normal ).
adj( 'descriptive', '-', '-', normal ).
adj( 'desert', '-', '-', normal ).
adj( 'deserved', '-', '-', normal ).
adj( 'deserving', '-', '-', normal ).
adj( 'designate', '-', '-', normal ).
adj( 'designing', '-', '-', normal ).
adj( 'desirable', '-', '-', normal ).
adj( 'desirous', '-', '-', normal ).
adj( 'desolate', '-', '-', normal ).
adj( 'desperate', '-', '-', normal ).
adj( 'despicable', '-', '-', normal ).
adj( 'despiteful', '-', '-', normal ).
adj( 'despondent', '-', '-', normal ).
adj( 'despotic', '-', '-', normal ).
adj( 'destitute', '-', '-', normal ).
adj( 'destructible', '-', '-', normal ).
adj( 'destructive', '-', '-', normal ).
adj( 'desultory', '-', '-', normal ).
adj( 'detachable', '-', '-', normal ).
adj( 'detached', '-', '-', normal ).
adj( 'detectable', '-', '-', normal ).
adj( 'detergent', '-', '-', normal ).
adj( 'determinable', '-', '-', normal ).
adj( 'determinant', '-', '-', normal ).
adj( 'determinate', '-', '-', normal ).
adj( 'determinative', '-', '-', normal ).
adj( 'deterrent', '-', '-', normal ).
adj( 'detestable', '-', '-', normal ).
adj( 'detrimental', '-', '-', normal ).
adj( 'deuced', '-', '-', normal ).
adj( 'deviant', '-', '-', normal ).
adj( 'devil-may-care', '-', '-', normal ).
adj( 'devilish', '-', '-', normal ).
adj( 'devious', '-', '-', normal ).
adj( 'devoid', '-', '-', normal ).
adj( 'devoted', '-', '-', normal ).
adj( 'devotional', '-', '-', normal ).
adj( 'devout', '-', '-', normal ).
adj( 'dewy', 'dewier', 'dewiest', normal ).
adj( 'dexterous', '-', '-', normal ).
adj( 'dextrous', '-', '-', normal ).
adj( 'diabetic', '-', '-', normal ).
adj( 'diabolic', '-', '-', normal ).
adj( 'diabolical', '-', '-', normal ).
adj( 'diacritic', '-', '-', normal ).
adj( 'diacritical', '-', '-', normal ).
adj( 'diagnostic', '-', '-', normal ).
adj( 'diagonal', '-', '-', normal ).
adj( 'diagrammatic', '-', '-', normal ).
adj( 'diagrammatical', '-', '-', normal ).
adj( 'dialectal', '-', '-', normal ).
adj( 'dialectical', '-', '-', normal ).
adj( 'diaphanous', '-', '-', normal ).
adj( 'diatonic', '-', '-', normal ).
adj( 'dicey', '-', '-', normal ).
adj( 'dicky', '-', '-', normal ).
adj( 'dictatorial', '-', '-', normal ).
adj( 'didactic', '-', '-', normal ).
adj( 'die-cast', '-', '-', normal ).
adj( 'dietary', '-', '-', normal ).
adj( 'different', '-', '-', normal ).
adj( 'differential', '-', '-', normal ).
adj( 'difficult', '-', '-', normal ).
adj( 'diffident', '-', '-', normal ).
adj( 'diffuse', '-', '-', normal ).
adj( 'digestible', '-', '-', normal ).
adj( 'digestive', '-', '-', normal ).
adj( 'digital', '-', '-', normal ).
adj( 'dignified', '-', '-', normal ).
adj( 'dilapidated', '-', '-', normal ).
adj( 'dilatory', '-', '-', normal ).
adj( 'diligent', '-', '-', normal ).
adj( 'dilute', '-', '-', normal ).
adj( 'dim', 'dimmer', 'dimmest', normal ).
adj( 'dimensional', '-', '-', normal ).
adj( 'diminutive', '-', '-', normal ).
adj( 'dingy', 'dingier', 'dingiest', normal ).
adj( 'dinky', 'dinkier', 'dinkiest', normal ).
adj( 'diocesan', '-', '-', normal ).
adj( 'diplomatic', '-', '-', normal ).
adj( 'dire', 'direr', 'direst', normal ).
adj( 'direct', '-', '-', normal ).
adj( 'directional', '-', '-', normal ).
adj( 'direful', '-', '-', normal ).
adj( 'dirt-cheap', '-', '-', normal ).
adj( 'dirty', 'dirtier', 'dirtiest', normal ).
adj( 'disadvantageous', '-', '-', normal ).
adj( 'disaffected', '-', '-', normal ).
adj( 'disagreeable', '-', '-', normal ).
adj( 'disappointed', '-', '-', normal ).
adj( 'disappointing', '-', '-', normal ).
adj( 'disastrous', '-', '-', normal ).
adj( 'discernible', '-', '-', normal ).
adj( 'discerning', '-', '-', normal ).
adj( 'disciplinary', '-', '-', normal ).
adj( 'disconnected', '-', '-', normal ).
adj( 'disconsolate', '-', '-', normal ).
adj( 'discontinuous', '-', '-', normal ).
adj( 'discordant', '-', '-', normal ).
adj( 'discourteous', '-', '-', normal ).
adj( 'discreditable', '-', '-', normal ).
adj( 'discreet', '-', '-', normal ).
adj( 'discrete', '-', '-', normal ).
adj( 'discretionary', '-', '-', normal ).
adj( 'discriminating', '-', '-', normal ).
adj( 'discriminatory', '-', '-', normal ).
adj( 'discursive', '-', '-', normal ).
adj( 'disdainful', '-', '-', normal ).
adj( 'diseased', '-', '-', normal ).
adj( 'disgraceful', '-', '-', normal ).
adj( 'disgruntled', '-', '-', normal ).
adj( 'disgusting', '-', '-', normal ).
adj( 'disharmonious', '-', '-', normal ).
adj( 'dishevelled', '-', '-', normal ).
adj( 'dishonest', '-', '-', normal ).
adj( 'dishonourable', '-', '-', normal ).
adj( 'dishy', 'dishier', 'dishiest', normal ).
adj( 'disinfectant', '-', '-', normal ).
adj( 'disingenuous', '-', '-', normal ).
adj( 'disinterested', '-', '-', normal ).
adj( 'disjointed', '-', '-', normal ).
adj( 'disjunctive', '-', '-', normal ).
adj( 'disloyal', '-', '-', normal ).
adj( 'dismal', '-', '-', normal ).
adj( 'dismounted', '-', '-', normal ).
adj( 'disobedient', '-', '-', normal ).
adj( 'disorderly', '-', '-', normal ).
adj( 'disparate', '-', '-', normal ).
adj( 'dispassionate', '-', '-', normal ).
adj( 'dispensable', '-', '-', normal ).
adj( 'displeasing', '-', '-', normal ).
adj( 'disposable', '-', '-', normal ).
adj( 'disproportionate', '-', '-', normal ).
adj( 'disputable', '-', '-', normal ).
adj( 'disputatious', '-', '-', normal ).
adj( 'disquieting', '-', '-', normal ).
adj( 'disreputable', '-', '-', normal ).
adj( 'disrespectful', '-', '-', normal ).
adj( 'disruptive', '-', '-', normal ).
adj( 'dissident', '-', '-', normal ).
adj( 'dissimilar', '-', '-', normal ).
adj( 'dissipated', '-', '-', normal ).
adj( 'dissoluble', '-', '-', normal ).
adj( 'dissolute', '-', '-', normal ).
adj( 'dissonant', '-', '-', normal ).
adj( 'distant', '-', '-', normal ).
adj( 'distasteful', '-', '-', normal ).
adj( 'distinct', '-', '-', normal ).
adj( 'distinctive', '-', '-', normal ).
adj( 'distinguishable', '-', '-', normal ).
adj( 'distinguished', '-', '-', normal ).
adj( 'distracted', '-', '-', normal ).
adj( 'distrait', '-', '-', normal ).
adj( 'distraught', '-', '-', normal ).
adj( 'distressful', '-', '-', normal ).
adj( 'distressing', '-', '-', normal ).
adj( 'distributive', '-', '-', normal ).
adj( 'distrustful', '-', '-', normal ).
adj( 'disused', '-', '-', normal ).
adj( 'disyllabic', '-', '-', normal ).
adj( 'diurnal', '-', '-', normal ).
adj( 'divergent', '-', '-', normal ).
adj( 'divers', '-', '-', normal ).
adj( 'diverse', '-', '-', normal ).
adj( 'diversionary', '-', '-', normal ).
adj( 'diverting', '-', '-', normal ).
adj( 'divine', '-', '-', normal ).
adj( 'divisible', '-', '-', normal ).
adj( 'divisional', '-', '-', normal ).
adj( 'dizzy', 'dizzier', 'dizziest', normal ).
adj( 'docile', '-', '-', normal ).
adj( 'doctrinaire', '-', '-', normal ).
adj( 'doctrinal', '-', '-', normal ).
adj( 'documentary', '-', '-', normal ).
adj( 'doddering', '-', '-', normal ).
adj( 'doddery', '-', '-', normal ).
adj( 'dodgy', '-', '-', normal ).
adj( 'dog-eared', '-', '-', normal ).
adj( 'dog-like', '-', '-', normal ).
adj( 'dogged', '-', '-', normal ).
adj( 'dogmatic', '-', '-', normal ).
adj( 'doleful', '-', '-', normal ).
adj( 'dolourous', '-', '-', normal ).
adj( 'doltish', '-', '-', normal ).
adj( 'domed', '-', '-', normal ).
adj( 'domestic', '-', '-', normal ).
adj( 'domiciliary', '-', '-', normal ).
adj( 'dominant', '-', '-', normal ).
adj( 'domineering', '-', '-', normal ).
adj( 'donnish', '-', '-', normal ).
adj( 'door-to-door', '-', '-', normal ).
adj( 'dopey', '-', '-', normal ).
adj( 'dormant', '-', '-', normal ).
adj( 'dorsal', '-', '-', normal ).
adj( 'dotty', 'dottier', 'dottiest', normal ).
adj( 'double', '-', '-', normal ).
adj( 'double-barrelled', '-', '-', normal ).
adj( 'double-bedded', '-', '-', normal ).
adj( 'double-breasted', '-', '-', normal ).
adj( 'double-dealing', '-', '-', normal ).
adj( 'double-dyed', '-', '-', normal ).
adj( 'double-edged', '-', '-', normal ).
adj( 'double-faced', '-', '-', normal ).
adj( 'double-jointed', '-', '-', normal ).
adj( 'double-quick', '-', '-', normal ).
adj( 'doubtful', '-', '-', normal ).
adj( 'doughty', '-', '-', normal ).
adj( 'doughy', 'doughier', 'doughiest', normal ).
adj( 'dour', '-', '-', normal ).
adj( 'dowdy', 'dowdier', 'dowdiest', normal ).
adj( 'down-market', '-', '-', normal ).
adj( 'down-to-earth', '-', '-', normal ).
adj( 'downcast', '-', '-', normal ).
adj( 'downhearted', '-', '-', normal ).
adj( 'downright', '-', '-', normal ).
adj( 'downstair', '-', '-', normal ).
adj( 'downstairs', '-', '-', normal ).
adj( 'downstream', '-', '-', normal ).
adj( 'downtrodden', '-', '-', normal ).
adj( 'downward', '-', '-', normal ).
adj( 'downy', '-', '-', normal ).
adj( 'drab', 'drabber', 'drabbest', normal ).
adj( 'draconian', '-', '-', normal ).
adj( 'draggled', '-', '-', normal ).
adj( 'dramatic', '-', '-', normal ).
adj( 'drastic', '-', '-', normal ).
adj( 'draughty', 'draughtier', 'draughtiest', normal ).
adj( 'dreaded', '-', '-', normal ).
adj( 'dreadful', '-', '-', normal ).
adj( 'dreamless', '-', '-', normal ).
adj( 'dreamlike', '-', '-', normal ).
adj( 'dreamy', 'dreamier', 'dreamiest', normal ).
adj( 'drear', '-', '-', normal ).
adj( 'dreary', 'drearier', 'dreariest', normal ).
adj( 'dressy', 'dressier', 'dressiest', normal ).
adj( 'drinkable', '-', '-', normal ).
adj( 'drip-dry', '-', '-', normal ).
adj( 'drizzly', '-', '-', normal ).
adj( 'droll', '-', '-', normal ).
adj( 'dropsical', '-', '-', normal ).
adj( 'drowsy', 'drowsier', 'drowsiest', normal ).
adj( 'drumhead', '-', '-', normal ).
adj( 'drunk', 'drunker', 'drunkest', normal ).
adj( 'drunken', '-', '-', normal ).
adj( 'dry', 'drier', 'driest', normal ).
adj( 'dry-shod', '-', '-', normal ).
adj( 'dual', '-', '-', normal ).
adj( 'dubious', '-', '-', normal ).
adj( 'ducal', '-', '-', normal ).
adj( 'duckbilled', '-', '-', normal ).
adj( 'ductile', '-', '-', normal ).
adj( 'dud', '-', '-', normal ).
adj( 'due', '-', '-', normal ).
adj( 'dulcet', '-', '-', normal ).
adj( 'dull', 'duller', 'dullest', normal ).
adj( 'dumb', 'dumber', 'dumbest', normal ).
adj( 'dumpy', 'dumpier', 'dumpiest', normal ).
adj( 'dun', '-', '-', normal ).
adj( 'duodecimal', '-', '-', normal ).
adj( 'duodenal', '-', '-', normal ).
adj( 'duplex', '-', '-', normal ).
adj( 'duplicate', '-', '-', normal ).
adj( 'durable', '-', '-', normal ).
adj( 'dusky', 'duskier', 'duskiest', normal ).
adj( 'dusty', 'dustier', 'dustiest', normal ).
adj( 'duteous', '-', '-', normal ).
adj( 'dutiable', '-', '-', normal ).
adj( 'dutiful', '-', '-', normal ).
adj( 'duty-free', '-', '-', normal ).
adj( 'dwarfish', '-', '-', normal ).
adj( 'dyed-in-the-wool', '-', '-', normal ).
adj( 'dynamic', '-', '-', normal ).
adj( 'dynastic', '-', '-', normal ).
adj( 'dyslexic', '-', '-', normal ).
adj( 'dyspeptic', '-', '-', normal ).
adj( 'each', '-', '-', normal ).
adj( 'eager', '-', '-', normal ).
adj( 'eagle-eyed', '-', '-', normal ).
adj( 'early', 'earlier', 'earliest', normal ).
adj( 'early-warning', '-', '-', normal ).
adj( 'earnest', '-', '-', normal ).
adj( 'earthen', '-', '-', normal ).
adj( 'earthly', '-', '-', normal ).
adj( 'earthy', 'earthier', 'earthiest', normal ).
adj( 'easterly', '-', '-', normal ).
adj( 'eastern', '-', '-', normal ).
adj( 'easternmost', '-', '-', normal ).
adj( 'eastward', '-', '-', normal ).
adj( 'easy', 'easier', 'easiest', normal ).
adj( 'easygoing', '-', '-', normal ).
adj( 'eatable', '-', '-', normal ).
adj( 'ebony', '-', '-', normal ).
adj( 'ebullient', '-', '-', normal ).
adj( 'eccentric', '-', '-', normal ).
adj( 'ecclesiastical', '-', '-', normal ).
adj( 'eclectic', '-', '-', normal ).
adj( 'ecological', '-', '-', normal ).
adj( 'economic', '-', '-', normal ).
adj( 'economical', '-', '-', normal ).
adj( 'ecstatic', '-', '-', normal ).
adj( 'ecumenical', '-', '-', normal ).
adj( 'edgy', 'edgier', 'edgiest', normal ).
adj( 'edible', '-', '-', normal ).
adj( 'editorial', '-', '-', normal ).
adj( 'educational', '-', '-', normal ).
adj( 'eerie', 'eerier', 'eeriest', normal ).
adj( 'eery', 'eerier', 'eeriest', normal ).
adj( 'effective', '-', '-', normal ).
adj( 'effectual', '-', '-', normal ).
adj( 'effeminate', '-', '-', normal ).
adj( 'effervescent', '-', '-', normal ).
adj( 'effete', '-', '-', normal ).
adj( 'efficacious', '-', '-', normal ).
adj( 'efficient', '-', '-', normal ).
adj( 'efflorescent', '-', '-', normal ).
adj( 'effortless', '-', '-', normal ).
adj( 'effulgent', '-', '-', normal ).
adj( 'effusive', '-', '-', normal ).
adj( 'egalitarian', '-', '-', normal ).
adj( 'egocentric', '-', '-', normal ).
adj( 'egoistic', '-', '-', normal ).
adj( 'egoistical', '-', '-', normal ).
adj( 'egotistic', '-', '-', normal ).
adj( 'egregious', '-', '-', normal ).
adj( 'eight', '-', '-', normal ).
adj( 'eighteen', '-', '-', normal ).
adj( 'eighteenth', '-', '-', normal ).
adj( 'eighth', '-', '-', normal ).
adj( 'eightieth', '-', '-', normal ).
adj( 'eightpenny', '-', '-', normal ).
adj( 'eighty', '-', '-', normal ).
adj( 'either', '-', '-', normal ).
adj( 'elaborate', '-', '-', normal ).
adj( 'elastic', '-', '-', normal ).
adj( 'elasticized', '-', '-', normal ).
adj( 'elderly', '-', '-', normal ).
adj( 'elect', '-', '-', normal ).
adj( 'elective', '-', '-', normal ).
adj( 'electoral', '-', '-', normal ).
adj( 'electric', '-', '-', normal ).
adj( 'electrical', '-', '-', normal ).
adj( 'electromagnetic', '-', '-', normal ).
adj( 'electronic', '-', '-', normal ).
adj( 'electrostatic', '-', '-', normal ).
adj( 'eleemosynary', '-', '-', normal ).
adj( 'elegant', '-', '-', normal ).
adj( 'elegiac', '-', '-', normal ).
adj( 'elemental', '-', '-', normal ).
adj( 'elementary', '-', '-', normal ).
adj( 'elephantine', '-', '-', normal ).
adj( 'eleven', '-', '-', normal ).
adj( 'eleventh', '-', '-', normal ).
adj( 'elfin', '-', '-', normal ).
adj( 'elfish', '-', '-', normal ).
adj( 'eligible', '-', '-', normal ).
adj( 'elliptic', '-', '-', normal ).
adj( 'elliptical', '-', '-', normal ).
adj( 'elocutionary', '-', '-', normal ).
adj( 'eloquent', '-', '-', normal ).
adj( 'elusive', '-', '-', normal ).
adj( 'elvish', '-', '-', normal ).
adj( 'embarrassing', '-', '-', normal ).
adj( 'embattled', '-', '-', normal ).
adj( 'emblematic', '-', '-', normal ).
adj( 'embonpoint', '-', '-', pred ).
adj( 'embryonic', '-', '-', normal ).
adj( 'emergent', '-', '-', normal ).
adj( 'emeritus', '-', '-', normal ).
adj( 'eminent', '-', '-', normal ).
adj( 'emotional', '-', '-', normal ).
adj( 'emotionless', '-', '-', normal ).
adj( 'emotive', '-', '-', normal ).
adj( 'emphatic', '-', '-', normal ).
adj( 'empiric', '-', '-', normal ).
adj( 'empirical', '-', '-', normal ).
adj( 'employable', '-', '-', normal ).
adj( 'empty', 'emptier', 'emptiest', normal ).
adj( 'empty-handed', '-', '-', normal ).
adj( 'empty-headed', '-', '-', normal ).
adj( 'empurpled', '-', '-', normal ).
adj( 'empyrean', '-', '-', normal ).
adj( 'emulous', '-', '-', normal ).
adj( 'enabling', '-', '-', normal ).
adj( 'encaustic', '-', '-', normal ).
adj( 'encyclical', '-', '-', normal ).
adj( 'encyclopaedic', '-', '-', normal ).
adj( 'encyclopedic', '-', '-', normal ).
adj( 'endemic', '-', '-', normal ).
adj( 'endless', '-', '-', normal ).
adj( 'endurable', '-', '-', normal ).
adj( 'enduring', '-', '-', normal ).
adj( 'energetic', '-', '-', normal ).
adj( 'enforceable', '-', '-', normal ).
adj( 'engaging', '-', '-', normal ).
adj( 'enigmatic', '-', '-', normal ).
adj( 'enjoyable', '-', '-', normal ).
adj( 'enlightened', '-', '-', normal ).
adj( 'enormous', '-', '-', normal ).
adj( 'enough', '-', '-', normal ).
adj( 'enteric', '-', '-', normal ).
adj( 'enterprising', '-', '-', normal ).
adj( 'entertaining', '-', '-', normal ).
adj( 'enthusiastic', '-', '-', normal ).
adj( 'entire', '-', '-', normal ).
adj( 'entomological', '-', '-', normal ).
adj( 'entrepreneurial', '-', '-', normal ).
adj( 'enviable', '-', '-', normal ).
adj( 'envious', '-', '-', normal ).
adj( 'environmental', '-', '-', normal ).
adj( 'ephemeral', '-', '-', normal ).
adj( 'epic', '-', '-', normal ).
adj( 'epicurean', '-', '-', normal ).
adj( 'epidemic', '-', '-', normal ).
adj( 'epigrammatic', '-', '-', normal ).
adj( 'epileptic', '-', '-', normal ).
adj( 'episcopal', '-', '-', normal ).
adj( 'episcopalian', '-', '-', normal ).
adj( 'episodic', '-', '-', normal ).
adj( 'epistolary', '-', '-', normal ).
adj( 'epoch-making', '-', '-', normal ).
adj( 'equable', '-', '-', normal ).
adj( 'equal', '-', '-', normal ).
adj( 'equatorial', '-', '-', normal ).
adj( 'equestrian', '-', '-', normal ).
adj( 'equidistant', '-', '-', normal ).
adj( 'equilateral', '-', '-', normal ).
adj( 'equine', '-', '-', normal ).
adj( 'equinoctial', '-', '-', normal ).
adj( 'equitable', '-', '-', normal ).
adj( 'equivalent', '-', '-', normal ).
adj( 'equivocal', '-', '-', normal ).
adj( 'erect', '-', '-', normal ).
adj( 'erectile', '-', '-', normal ).
adj( 'erogenous', '-', '-', normal ).
adj( 'erosive', '-', '-', normal ).
adj( 'erotic', '-', '-', normal ).
adj( 'errant', '-', '-', normal ).
adj( 'erratic', '-', '-', normal ).
adj( 'erroneous', '-', '-', normal ).
adj( 'erudite', '-', '-', normal ).
adj( 'esoteric', '-', '-', normal ).
adj( 'especial', '-', '-', normal ).
adj( 'essential', '-', '-', normal ).
adj( 'esthetic', '-', '-', normal ).
adj( 'esthetical', '-', '-', normal ).
adj( 'estimable', '-', '-', normal ).
adj( 'eternal', '-', '-', normal ).
adj( 'ethereal', '-', '-', normal ).
adj( 'ethical', '-', '-', normal ).
adj( 'ethnic', '-', '-', normal ).
adj( 'ethnographic', '-', '-', normal ).
adj( 'ethnological', '-', '-', normal ).
adj( 'etymological', '-', '-', normal ).
adj( 'eulogistic', '-', '-', normal ).
adj( 'euphemistic', '-', '-', normal ).
adj( 'euphoric', '-', '-', normal ).
adj( 'evaluative', '-', '-', normal ).
adj( 'evanescent', '-', '-', normal ).
adj( 'evangelical', '-', '-', normal ).
adj( 'evangelistic', '-', '-', normal ).
adj( 'evasive', '-', '-', normal ).
adj( 'even', '-', '-', normal ).
adj( 'even-handed', '-', '-', normal ).
adj( 'eventful', '-', '-', normal ).
adj( 'eventual', '-', '-', normal ).
adj( 'evergreen', '-', '-', normal ).
adj( 'everlasting', '-', '-', normal ).
adj( 'every', '-', '-', normal ).
adj( 'everyday', '-', '-', normal ).
adj( 'evident', '-', '-', normal ).
adj( 'evil', '-', '-', normal ).
adj( 'evil-minded', '-', '-', normal ).
adj( 'evocative', '-', '-', normal ).
adj( 'evolutionary', '-', '-', normal ).
adj( 'ex gratia', '-', '-', normal ).
adj( 'ex officio', '-', '-', normal ).
adj( 'ex-directory', '-', '-', normal ).
adj( 'ex-service', '-', '-', normal ).
adj( 'exact', '-', '-', normal ).
adj( 'exacting', '-', '-', normal ).
adj( 'exalted', '-', '-', normal ).
adj( 'excellent', '-', '-', normal ).
adj( 'exceptionable', '-', '-', normal ).
adj( 'exceptional', '-', '-', normal ).
adj( 'excess', '-', '-', normal ).
adj( 'excessive', '-', '-', normal ).
adj( 'exchangeable', '-', '-', normal ).
adj( 'excitable', '-', '-', normal ).
adj( 'exclamatory', '-', '-', normal ).
adj( 'exclusive', '-', '-', normal ).
adj( 'excruciating', '-', '-', normal ).
adj( 'excusable', '-', '-', normal ).
adj( 'execrable', '-', '-', normal ).
adj( 'executive', '-', '-', normal ).
adj( 'exemplary', '-', '-', normal ).
adj( 'exempt', '-', '-', normal ).
adj( 'exhaustive', '-', '-', normal ).
adj( 'exigent', '-', '-', normal ).
adj( 'exiguous', '-', '-', normal ).
adj( 'existent', '-', '-', normal ).
adj( 'exorbitant', '-', '-', normal ).
adj( 'exotic', '-', '-', normal ).
adj( 'expansive', '-', '-', normal ).
adj( 'expectant', '-', '-', normal ).
adj( 'expected', '-', '-', normal ).
adj( 'expedient', '-', '-', normal ).
adj( 'expeditionary', '-', '-', normal ).
adj( 'expeditious', '-', '-', normal ).
adj( 'expendable', '-', '-', normal ).
adj( 'expensive', '-', '-', normal ).
adj( 'experienced', '-', '-', normal ).
adj( 'experimental', '-', '-', normal ).
adj( 'expert', '-', '-', normal ).
adj( 'explanatory', '-', '-', normal ).
adj( 'explicable', '-', '-', normal ).
adj( 'explicit', '-', '-', normal ).
adj( 'exploratory', '-', '-', normal ).
adj( 'explosive', '-', '-', normal ).
adj( 'exponential', '-', '-', normal ).
adj( 'exportable', '-', '-', normal ).
adj( 'express', '-', '-', normal ).
adj( 'expressionless', '-', '-', normal ).
adj( 'expressive', '-', '-', normal ).
adj( 'exquisite', '-', '-', normal ).
adj( 'extant', '-', '-', normal ).
adj( 'extemporaneous', '-', '-', normal ).
adj( 'extemporary', '-', '-', normal ).
adj( 'extempore', '-', '-', normal ).
adj( 'extensive', '-', '-', normal ).
adj( 'exterior', '-', '-', normal ).
adj( 'external', '-', '-', normal ).
adj( 'exterritorial', '-', '-', normal ).
adj( 'extinct', '-', '-', normal ).
adj( 'extortionate', '-', '-', normal ).
adj( 'extra', '-', '-', normal ).
adj( 'extracurricular', '-', '-', normal ).
adj( 'extrajudicial', '-', '-', normal ).
adj( 'extramarital', '-', '-', normal ).
adj( 'extramural', '-', '-', normal ).
adj( 'extraneous', '-', '-', normal ).
adj( 'extraordinary', '-', '-', normal ).
adj( 'extrasensory', '-', '-', normal ).
adj( 'extraterritorial', '-', '-', normal ).
adj( 'extravagant', '-', '-', normal ).
adj( 'extreme', '-', '-', normal ).
adj( 'extricable', '-', '-', normal ).
adj( 'extrinsic', '-', '-', normal ).
adj( 'exuberant', '-', '-', normal ).
adj( 'exultant', '-', '-', normal ).
adj( 'eye-catching', '-', '-', normal ).
adj( 'eyecatching', '-', '-', normal ).
adj( 'eyed', '-', '-', affix ).
adj( 'eyeless', '-', '-', normal ).
adj( 'fab', '-', '-', normal ).
adj( 'fabled', '-', '-', normal ).
adj( 'fabulous', '-', '-', normal ).
adj( 'face-saving', '-', '-', normal ).
adj( 'faceless', '-', '-', normal ).
adj( 'facetious', '-', '-', normal ).
adj( 'facial', '-', '-', normal ).
adj( 'facile', '-', '-', normal ).
adj( 'fact-finding', '-', '-', normal ).
adj( 'factious', '-', '-', normal ).
adj( 'factitious', '-', '-', normal ).
adj( 'factual', '-', '-', normal ).
adj( 'faddy', 'faddier', 'faddiest', normal ).
adj( 'fail-safe', '-', '-', normal ).
adj( 'faint', 'fainter', 'faintest', normal ).
adj( 'faint-hearted', '-', '-', normal ).
adj( 'fair', 'fairer', 'fairest', normal ).
adj( 'fair-minded', '-', '-', normal ).
adj( 'fairish', '-', '-', normal ).
adj( 'faithful', '-', '-', normal ).
adj( 'faithless', '-', '-', normal ).
adj( 'fallacious', '-', '-', normal ).
adj( 'fallible', '-', '-', normal ).
adj( 'fallow', '-', '-', normal ).
adj( 'false', '-', '-', normal ).
adj( 'falsetto', '-', '-', normal ).
adj( 'famed', '-', '-', normal ).
adj( 'familiar', '-', '-', normal ).
adj( 'famous', '-', '-', normal ).
adj( 'fanatic', '-', '-', normal ).
adj( 'fanatical', '-', '-', normal ).
adj( 'fanciful', '-', '-', normal ).
adj( 'fancy', 'fancier', 'fanciest', normal ).
adj( 'fancy-free', '-', '-', normal ).
adj( 'fantastic', '-', '-', normal ).
adj( 'far', '-', '-', normal ).
adj( 'far-famed', '-', '-', normal ).
adj( 'far-fetched', '-', '-', normal ).
adj( 'far-flung', '-', '-', normal ).
adj( 'far-off', '-', '-', normal ).
adj( 'far-reaching', '-', '-', normal ).
adj( 'far-seeing', '-', '-', normal ).
adj( 'far-sighted', '-', '-', normal ).
adj( 'faraway', '-', '-', normal ).
adj( 'farcical', '-', '-', normal ).
adj( 'farinaceous', '-', '-', normal ).
adj( 'fascinating', '-', '-', normal ).
adj( 'fascist', '-', '-', normal ).
adj( 'fashionable', '-', '-', normal ).
adj( 'fast', 'faster', 'fastest', normal ).
adj( 'fastidious', '-', '-', normal ).
adj( 'fat', 'fatter', 'fattest', normal ).
adj( 'fatal', '-', '-', normal ).
adj( 'fatalistic', '-', '-', normal ).
adj( 'fateful', '-', '-', normal ).
adj( 'fatherless', '-', '-', normal ).
adj( 'fatherly', '-', '-', normal ).
adj( 'fathomless', '-', '-', normal ).
adj( 'fatless', '-', '-', normal ).
adj( 'fattish', '-', '-', normal ).
adj( 'fatty', 'fattier', 'fattiest', normal ).
adj( 'fatuous', '-', '-', normal ).
adj( 'faultless', '-', '-', normal ).
adj( 'faulty', 'faultier', 'faultiest', normal ).
adj( 'favourable', '-', '-', normal ).
adj( 'favourite', '-', '-', attr ).
adj( 'fearful', '-', '-', normal ).
adj( 'fearless', '-', '-', normal ).
adj( 'fearsome', '-', '-', normal ).
adj( 'feasible', '-', '-', normal ).
adj( 'featherbrained', '-', '-', normal ).
adj( 'feathery', '-', '-', normal ).
adj( 'featureless', '-', '-', normal ).
adj( 'febrile', '-', '-', normal ).
adj( 'feckless', '-', '-', normal ).
adj( 'fecund', '-', '-', normal ).
adj( 'federal', '-', '-', normal ).
adj( 'feeble', 'feebler', 'feeblest', normal ).
adj( 'feeble-minded', '-', '-', normal ).
adj( 'feeling', '-', '-', normal ).
adj( 'felicitous', '-', '-', normal ).
adj( 'feline', '-', '-', normal ).
adj( 'fell', '-', '-', normal ).
adj( 'felonious', '-', '-', normal ).
adj( 'female', '-', '-', normal ).
adj( 'feminine', '-', '-', normal ).
adj( 'feral', '-', '-', normal ).
adj( 'ferny', 'fernier', 'ferniest', normal ).
adj( 'ferocious', '-', '-', normal ).
adj( 'ferrous', '-', '-', normal ).
adj( 'fertile', '-', '-', normal ).
adj( 'fervent', '-', '-', normal ).
adj( 'fervid', '-', '-', normal ).
adj( 'festal', '-', '-', normal ).
adj( 'festive', '-', '-', normal ).
adj( 'fetal', '-', '-', normal ).
adj( 'fetching', '-', '-', normal ).
adj( 'fetid', '-', '-', normal ).
adj( 'feudal', '-', '-', normal ).
adj( 'feudatory', '-', '-', normal ).
adj( 'fevered', '-', '-', normal ).
adj( 'feverish', '-', '-', normal ).
adj( 'few', 'fewer', 'fewest', normal ).
adj( 'fey', '-', '-', normal ).
adj( 'fibrous', '-', '-', normal ).
adj( 'fickle', '-', '-', normal ).
adj( 'fictional', '-', '-', normal ).
adj( 'fictitious', '-', '-', normal ).
adj( 'fiddling', '-', '-', normal ).
adj( 'fidgety', '-', '-', normal ).
adj( 'fiendish', '-', '-', normal ).
adj( 'fierce', 'fiercer', 'fiercest', normal ).
adj( 'fiery', '-', '-', normal ).
adj( 'fifteen', '-', '-', normal ).
adj( 'fifteenth', '-', '-', normal ).
adj( 'fifth', '-', '-', normal ).
adj( 'fiftieth', '-', '-', normal ).
adj( 'fifty', '-', '-', normal ).
adj( 'figurative', '-', '-', normal ).
adj( 'figured', '-', '-', normal ).
adj( 'filial', '-', '-', normal ).
adj( 'filmable', '-', '-', normal ).
adj( 'filmy', 'filmier', 'filmiest', normal ).
adj( 'filtertipped', '-', '-', normal ).
adj( 'filthy', 'filthier', 'filthiest', normal ).
adj( 'finable', '-', '-', normal ).
adj( 'final', '-', '-', normal ).
adj( 'financial', '-', '-', normal ).
adj( 'fine', 'finer', 'finest', normal ).
adj( 'fine-tooth', '-', '-', attr ).
adj( 'fineable', '-', '-', normal ).
adj( 'finical', '-', '-', normal ).
adj( 'finicky', '-', '-', normal ).
adj( 'finite', '-', '-', normal ).
adj( 'fireproof', '-', '-', normal ).
adj( 'firm', 'firmer', 'firmest', normal ).
adj( 'first', '-', '-', normal ).
adj( 'first-class', '-', '-', normal ).
adj( 'first-hand', '-', '-', normal ).
adj( 'first-rate', '-', '-', normal ).
adj( 'firstborn', '-', '-', normal ).
adj( 'fiscal', '-', '-', normal ).
adj( 'fishy', 'fishier', 'fishiest', normal ).
adj( 'fissile', '-', '-', normal ).
adj( 'fissionable', '-', '-', normal ).
adj( 'fissiparous', '-', '-', normal ).
adj( 'fit', 'fitter', 'fittest', normal ).
adj( 'fitful', '-', '-', normal ).
adj( 'fitting', '-', '-', normal ).
adj( 'five', '-', '-', normal ).
adj( 'fivefold', '-', '-', normal ).
adj( 'fivepenny', '-', '-', normal ).
adj( 'fixed', '-', '-', normal ).
adj( 'fizzy', 'fizzier', 'fizziest', normal ).
adj( 'flabby', 'flabbier', 'flabbiest', normal ).
adj( 'flaccid', '-', '-', normal ).
adj( 'flagrant', '-', '-', normal ).
adj( 'flaky', 'flakier', 'flakiest', normal ).
adj( 'flamboyant', '-', '-', normal ).
adj( 'flaming', '-', '-', normal ).
adj( 'flammable', '-', '-', normal ).
adj( 'flashy', 'flashier', 'flashiest', normal ).
adj( 'flat', 'flatter', 'flattest', normal ).
adj( 'flat-bottomed', '-', '-', normal ).
adj( 'flat-footed', '-', '-', normal ).
adj( 'flavourless', '-', '-', normal ).
adj( 'flawed', '-', '-', normal ).
adj( 'flawless', '-', '-', normal ).
adj( 'flaxen', '-', '-', normal ).
adj( 'flea-bitten', '-', '-', normal ).
adj( 'fledged', '-', '-', normal ).
adj( 'fleecy', 'fleecier', 'fleeciest', normal ).
adj( 'fleet', 'fleeter', 'fleetest', normal ).
adj( 'fleeting', '-', '-', normal ).
adj( 'fleshly', '-', '-', normal ).
adj( 'fleshy', 'fleshier', 'fleshiest', normal ).
adj( 'flexible', '-', '-', normal ).
adj( 'flightless', '-', '-', normal ).
adj( 'flighty', '-', '-', normal ).
adj( 'flimsy', 'flimsier', 'flimsiest', normal ).
adj( 'flinty', 'flintier', 'flintiest', normal ).
adj( 'flip', '-', '-', normal ).
adj( 'flippant', '-', '-', normal ).
adj( 'flirtatious', '-', '-', normal ).
adj( 'floating', '-', '-', normal ).
adj( 'floppy', 'floppier', 'floppiest', normal ).
adj( 'floral', '-', '-', normal ).
adj( 'florid', '-', '-', normal ).
adj( 'floury', 'flourier', 'flouriest', normal ).
adj( 'flowered', '-', '-', normal ).
adj( 'flowerless', '-', '-', normal ).
adj( 'flowery', 'flowerier', 'floweriest', normal ).
adj( 'fluent', '-', '-', normal ).
adj( 'fluffy', 'fluffier', 'fluffiest', normal ).
adj( 'fluid', '-', '-', normal ).
adj( 'fluorescent', '-', '-', normal ).
adj( 'flush', '-', '-', normal ).
adj( 'fluvial', '-', '-', normal ).
adj( 'fly', '-', '-', normal ).
adj( 'flyblown', '-', '-', normal ).
adj( 'flying', '-', '-', normal ).
adj( 'foamy', 'foamier', 'foamiest', normal ).
adj( 'focal', '-', '-', normal ).
adj( 'foetal', '-', '-', normal ).
adj( 'fogbound', '-', '-', normal ).
adj( 'foggy', 'foggier', 'foggiest', normal ).
adj( 'folksy', '-', '-', normal ).
adj( 'following', '-', '-', normal ).
adj( 'fond', 'fonder', 'fondest', normal ).
adj( 'foodless', '-', '-', normal ).
adj( 'foolhardy', '-', '-', normal ).
adj( 'foolish', '-', '-', normal ).
adj( 'foolproof', '-', '-', normal ).
adj( 'foot-and-mouth', '-', '-', normal ).
adj( 'footed', '-', '-', affix ).
adj( 'footling', '-', '-', normal ).
adj( 'footloose', '-', '-', normal ).
adj( 'footsore', '-', '-', normal ).
adj( 'footsure', '-', '-', normal ).
adj( 'foppish', '-', '-', normal ).
adj( 'forbidding', '-', '-', normal ).
adj( 'forceful', '-', '-', normal ).
adj( 'forcible', '-', '-', normal ).
adj( 'fordable', '-', '-', normal ).
adj( 'fore', '-', '-', normal ).
adj( 'foregoing', '-', '-', normal ).
adj( 'foregone', '-', '-', normal ).
adj( 'forehand', '-', '-', normal ).
adj( 'foreign', '-', '-', normal ).
adj( 'foremost', '-', '-', normal ).
adj( 'forensic', '-', '-', normal ).
adj( 'foreseeable', '-', '-', normal ).
adj( 'forgetful', '-', '-', normal ).
adj( 'forgivable', '-', '-', normal ).
adj( 'forgiving', '-', '-', normal ).
adj( 'forked', '-', '-', normal ).
adj( 'forlorn', '-', '-', normal ).
adj( 'formal', '-', '-', normal ).
adj( 'formative', '-', '-', normal ).
adj( 'former', '-', '-', normal ).
adj( 'formic', '-', '-', normal ).
adj( 'formidable', '-', '-', normal ).
adj( 'formless', '-', '-', normal ).
adj( 'forte', '-', '-', normal ).
adj( 'forthcoming', '-', '-', normal ).
adj( 'forthright', '-', '-', normal ).
adj( 'fortieth', '-', '-', normal ).
adj( 'fortissimo', '-', '-', normal ).
adj( 'fortnightly', '-', '-', normal ).
adj( 'fortuitous', '-', '-', normal ).
adj( 'fortunate', '-', '-', normal ).
adj( 'forty', '-', '-', normal ).
adj( 'fortyish', '-', '-', normal ).
adj( 'forward', '-', '-', normal ).
adj( 'foul', 'fouler', 'foulest', normal ).
adj( 'foul-mouthed', '-', '-', normal ).
adj( 'foul-spoken', '-', '-', normal ).
adj( 'four', '-', '-', normal ).
adj( 'four-part', '-', '-', normal ).
adj( 'four-ply', '-', '-', normal ).
adj( 'fourfold', '-', '-', normal ).
adj( 'fourpenny', '-', '-', normal ).
adj( 'fourscore', '-', '-', normal ).
adj( 'foursquare', '-', '-', normal ).
adj( 'fourteen', '-', '-', normal ).
adj( 'fourteenth', '-', '-', normal ).
adj( 'fourth', '-', '-', normal ).
adj( 'foxy', 'foxier', 'foxiest', normal ).
adj( 'fractional', '-', '-', normal ).
adj( 'fractious', '-', '-', normal ).
adj( 'fragile', '-', '-', normal ).
adj( 'fragmentary', '-', '-', normal ).
adj( 'fragrant', '-', '-', normal ).
adj( 'frail', 'frailer', 'frailest', normal ).
adj( 'frank', 'franker', 'frankest', normal ).
adj( 'frantic', '-', '-', normal ).
adj( 'fraternal', '-', '-', normal ).
adj( 'fraudulent', '-', '-', normal ).
adj( 'fraught', '-', '-', pred ).
adj( 'freakish', '-', '-', normal ).
adj( 'freaky', 'freakier', 'freakiest', normal ).
adj( 'free', 'freer', 'freest', normal ).
adj( 'free-and-easy', '-', '-', normal ).
adj( 'free-handed', '-', '-', normal ).
adj( 'free-range', '-', '-', normal ).
adj( 'free-spoken', '-', '-', normal ).
adj( 'free-standing', '-', '-', normal ).
adj( 'free-thinking', '-', '-', normal ).
adj( 'freeborn', '-', '-', normal ).
adj( 'freehand', '-', '-', normal ).
adj( 'freewill', '-', '-', normal ).
adj( 'frenetic', '-', '-', normal ).
adj( 'frenzied', '-', '-', normal ).
adj( 'frequent', '-', '-', normal ).
adj( 'fresh', 'fresher', 'freshest', normal ).
adj( 'freshwater', '-', '-', normal ).
adj( 'fretful', '-', '-', normal ).
adj( 'friable', '-', '-', normal ).
adj( 'fricative', '-', '-', normal ).
adj( 'friendless', '-', '-', normal ).
adj( 'friendly', 'friendlier', 'friendliest', normal ).
adj( 'frightened', '-', '-', normal ).
adj( 'frightening', '-', '-', normal ).
adj( 'frightful', '-', '-', normal ).
adj( 'frigid', '-', '-', normal ).
adj( 'frilled', '-', '-', normal ).
adj( 'frilly', 'frillier', 'frilliest', normal ).
adj( 'frisky', 'friskier', 'friskiest', normal ).
adj( 'frivolous', '-', '-', normal ).
adj( 'frizzy', 'frizzier', 'frizziest', normal ).
adj( 'frolicsome', '-', '-', normal ).
adj( 'frontal', '-', '-', normal ).
adj( 'frost-bound', '-', '-', normal ).
adj( 'frostbitten', '-', '-', normal ).
adj( 'frosty', 'frostier', 'frostiest', normal ).
adj( 'frothy', 'frothier', 'frothiest', normal ).
adj( 'froward', '-', '-', normal ).
adj( 'frowsty', '-', '-', normal ).
adj( 'frowzy', '-', '-', normal ).
adj( 'frugal', '-', '-', normal ).
adj( 'fruitful', '-', '-', normal ).
adj( 'fruitless', '-', '-', normal ).
adj( 'fruity', 'fruitier', 'fruitiest', normal ).
adj( 'frumpish', '-', '-', normal ).
adj( 'frumpy', 'frumpier', 'frumpiest', normal ).
adj( 'fucking', '-', '-', normal ).
adj( 'fuggy', 'fuggier', 'fuggiest', normal ).
adj( 'fugitive', '-', '-', normal ).
adj( 'full', 'fuller', 'fullest', normal ).
adj( 'full-blooded', '-', '-', normal ).
adj( 'full-blown', '-', '-', normal ).
adj( 'full-dress', '-', '-', normal ).
adj( 'full-fashioned', '-', '-', normal ).
adj( 'full-fledged', '-', '-', normal ).
adj( 'full-grown', '-', '-', normal ).
adj( 'full-length', '-', '-', normal ).
adj( 'full-page', '-', '-', normal ).
adj( 'full-scale', '-', '-', normal ).
adj( 'full-time', '-', '-', normal ).
adj( 'fully-fashioned', '-', '-', normal ).
adj( 'fully-fledged', '-', '-', normal ).
adj( 'fully-grown', '-', '-', normal ).
adj( 'fulsome', '-', '-', normal ).
adj( 'functional', '-', '-', normal ).
adj( 'fundamental', '-', '-', normal ).
adj( 'funereal', '-', '-', normal ).
adj( 'fungoid', '-', '-', normal ).
adj( 'fungous', '-', '-', normal ).
adj( 'funky', 'funkier', 'funkiest', normal ).
adj( 'funny', 'funnier', 'funniest', normal ).
adj( 'furious', '-', '-', normal ).
adj( 'furry', 'furrier', 'furriest', normal ).
adj( 'furthermost', '-', '-', normal ).
adj( 'furthest', '-', '-', normal ).
adj( 'furtive', '-', '-', normal ).
adj( 'fussy', 'fussier', 'fussiest', normal ).
adj( 'fusty', '-', '-', normal ).
adj( 'futile', '-', '-', normal ).
adj( 'future', '-', '-', normal ).
adj( 'futureless', '-', '-', normal ).
adj( 'fuzzy', 'fuzzier', 'fuzziest', normal ).
adj( 'gabled', '-', '-', normal ).
adj( 'gaga', '-', '-', normal ).
adj( 'gainful', '-', '-', normal ).
adj( 'galactic', '-', '-', normal ).
adj( 'gallant', '-', '-', normal ).
adj( 'galvanic', '-', '-', normal ).
adj( 'game', '-', '-', normal ).
adj( 'gammy', '-', '-', normal ).
adj( 'gamy', 'gamier', 'gamiest', normal ).
adj( 'gangling', '-', '-', normal ).
adj( 'gangrenous', '-', '-', normal ).
adj( 'gap-toothed', '-', '-', normal ).
adj( 'gargantuan', '-', '-', normal ).
adj( 'garish', '-', '-', normal ).
adj( 'garrulous', '-', '-', normal ).
adj( 'gaseous', '-', '-', normal ).
adj( 'gassy', 'gassier', 'gassiest', normal ).
adj( 'gastric', '-', '-', normal ).
adj( 'gastronomic', '-', '-', normal ).
adj( 'gauche', '-', '-', normal ).
adj( 'gaudy', 'gaudier', 'gaudiest', normal ).
adj( 'gaunt', '-', '-', normal ).
adj( 'gauzy', 'gauzier', 'gauziest', normal ).
adj( 'gawky', 'gawkier', 'gawkiest', normal ).
adj( 'gay', 'gayer', 'gayest', normal ).
adj( 'gelatinous', '-', '-', normal ).
adj( 'gemmed', '-', '-', normal ).
adj( 'genealogical', '-', '-', normal ).
adj( 'general', '-', '-', normal ).
adj( 'generative', '-', '-', normal ).
adj( 'generic', '-', '-', normal ).
adj( 'generous', '-', '-', normal ).
adj( 'genetic', '-', '-', normal ).
adj( 'genial', '-', '-', normal ).
adj( 'genital', '-', '-', normal ).
adj( 'genitive', '-', '-', normal ).
adj( 'genteel', '-', '-', normal ).
adj( 'gentile', '-', '-', normal ).
adj( 'gentle', 'gentler', 'gentlest', normal ).
adj( 'gentlemanly', '-', '-', normal ).
adj( 'genuine', '-', '-', normal ).
adj( 'geocentric', '-', '-', normal ).
adj( 'geographical', '-', '-', normal ).
adj( 'geological', '-', '-', normal ).
adj( 'geometric', '-', '-', normal ).
adj( 'geometrical', '-', '-', normal ).
adj( 'geophysical', '-', '-', normal ).
adj( 'geriatric', '-', '-', normal ).
adj( 'germane', '-', '-', normal ).
adj( 'get-at-able', '-', '-', normal ).
adj( 'ghastly', 'ghastlier', 'ghastliest', normal ).
adj( 'ghostly', '-', '-', normal ).
adj( 'ghoulish', '-', '-', normal ).
adj( 'gibbous', '-', '-', normal ).
adj( 'giddy', 'giddier', 'giddiest', normal ).
adj( 'gifted', '-', '-', normal ).
adj( 'gigantic', '-', '-', normal ).
adj( 'gilt-edged', '-', '-', normal ).
adj( 'gimcrack', '-', '-', normal ).
adj( 'gingerly', '-', '-', normal ).
adj( 'girlish', '-', '-', normal ).
adj( 'glac_e', '-', '-', normal ).
adj( 'glacial', '-', '-', normal ).
adj( 'glad', 'gladder', 'gladdest', normal ).
adj( 'gladiatorial', '-', '-', normal ).
adj( 'gladsome', '-', '-', normal ).
adj( 'glamorous', '-', '-', normal ).
adj( 'glandular', '-', '-', normal ).
adj( 'glaring', '-', '-', normal ).
adj( 'glassy', 'glassier', 'glassiest', normal ).
adj( 'glaucous', '-', '-', normal ).
adj( 'gleeful', '-', '-', normal ).
adj( 'glib', 'glibber', 'glibbest', normal ).
adj( 'glissando', '-', '-', normal ).
adj( 'glittering', '-', '-', normal ).
adj( 'global', '-', '-', normal ).
adj( 'globular', '-', '-', normal ).
adj( 'gloomy', 'gloomier', 'gloomiest', normal ).
adj( 'glorious', '-', '-', normal ).
adj( 'glossy', 'glossier', 'glossiest', normal ).
adj( 'glottal', '-', '-', normal ).
adj( 'gloved', '-', '-', normal ).
adj( 'glowing', '-', '-', normal ).
adj( 'gluey', '-', '-', normal ).
adj( 'glum', 'glummer', 'glummest', normal ).
adj( 'glutinous', '-', '-', normal ).
adj( 'gluttonous', '-', '-', normal ).
adj( 'gnarled', '-', '-', normal ).
adj( 'go-as-you-please', '-', '-', attr ).
adj( 'go-to-meeting', '-', '-', attr ).
adj( 'goalless', '-', '-', normal ).
adj( 'god-damn', '-', '-', normal ).
adj( 'god-damned', '-', '-', normal ).
adj( 'goddam', '-', '-', normal ).
adj( 'godfearing', '-', '-', normal ).
adj( 'godforsaken', '-', '-', normal ).
adj( 'godless', '-', '-', normal ).
adj( 'godlike', '-', '-', normal ).
adj( 'godly', 'godlier', 'godliest', normal ).
adj( 'goggle-eyed', '-', '-', normal ).
adj( 'going', '-', '-', normal ).
adj( 'golden', '-', '-', normal ).
adj( 'good', 'better', 'best', normal ).
adj( 'good-for-naught', '-', '-', normal ).
adj( 'good-for-nothing', '-', '-', normal ).
adj( 'good-humoured', '-', '-', normal ).
adj( 'good-looking', '-', '-', normal ).
adj( 'good-natured', '-', '-', normal ).
adj( 'good-tempered', '-', '-', normal ).
adj( 'good-time', '-', '-', normal ).
adj( 'goodish', '-', '-', attr ).
adj( 'goodly', 'goodlier', 'goodliest', normal ).
adj( 'goody-goody', '-', '-', normal ).
adj( 'gooey', 'gooier', 'gooiest', normal ).
adj( 'goofy', 'goofier', 'goofiest', normal ).
adj( 'gorgeous', '-', '-', normal ).
adj( 'gormless', '-', '-', normal ).
adj( 'gory', 'gorier', 'goriest', normal ).
adj( 'gouty', 'goutier', 'goutiest', normal ).
adj( 'governing', '-', '-', normal ).
adj( 'governmental', '-', '-', normal ).
adj( 'graceful', '-', '-', normal ).
adj( 'graceless', '-', '-', normal ).
adj( 'gracious', '-', '-', normal ).
adj( 'gradual', '-', '-', normal ).
adj( 'grained', '-', '-', affix ).
adj( 'grammatical', '-', '-', normal ).
adj( 'grand', 'grander', 'grandest', normal ).
adj( 'grandiloquent', '-', '-', normal ).
adj( 'grandiose', '-', '-', normal ).
adj( 'granular', '-', '-', normal ).
adj( 'graphic', '-', '-', normal ).
adj( 'graphical', '-', '-', normal ).
adj( 'grasping', '-', '-', normal ).
adj( 'grassy', 'grassier', 'grassiest', normal ).
adj( 'grateful', '-', '-', normal ).
adj( 'gratifying', '-', '-', normal ).
adj( 'gratis', '-', '-', normal ).
adj( 'gratuitous', '-', '-', normal ).
adj( 'grave', 'graver', 'gravest', normal ).
adj( 'gravelly', '-', '-', normal ).
adj( 'graven', '-', '-', normal ).
adj( 'gray', 'grayer', 'grayest', normal ).
adj( 'greaseproof', '-', '-', normal ).
adj( 'greasy', 'greasier', 'greasiest', normal ).
adj( 'great', 'greater', 'greatest', normal ).
adj( 'greedy', 'greedier', 'greediest', normal ).
adj( 'green', 'greener', 'greenest', normal ).
adj( 'greeneyed', '-', '-', normal ).
adj( 'greenish', '-', '-', normal ).
adj( 'gregarious', '-', '-', normal ).
adj( 'grey', 'greyer', 'greyest', normal ).
adj( 'grey-headed', '-', '-', normal ).
adj( 'greyish', '-', '-', normal ).
adj( 'grievous', '-', '-', normal ).
adj( 'grim', 'grimmer', 'grimmest', normal ).
adj( 'grimy', 'grimier', 'grimiest', normal ).
adj( 'grisly', '-', '-', normal ).
adj( 'gritty', 'grittier', 'grittiest', normal ).
adj( 'grizzled', '-', '-', normal ).
adj( 'groggy', 'groggier', 'groggiest', normal ).
adj( 'groovy', 'groovier', 'grooviest', normal ).
adj( 'gross', '-', '-', normal ).
adj( 'grotesque', '-', '-', normal ).
adj( 'grotty', 'grottier', 'grottiest', normal ).
adj( 'grouchy', 'grouchier', 'grouchiest', normal ).
adj( 'groundless', '-', '-', normal ).
adj( 'grown-up', '-', '-', normal ).
adj( 'grubby', 'grubbier', 'grubbiest', normal ).
adj( 'gruelling', '-', '-', normal ).
adj( 'gruesome', '-', '-', normal ).
adj( 'gruff', 'gruffer', 'gruffest', normal ).
adj( 'grumpy', 'grumpier', 'grumpiest', normal ).
adj( 'guarded', '-', '-', normal ).
adj( 'guileful', '-', '-', normal ).
adj( 'guileless', '-', '-', normal ).
adj( 'guiltless', '-', '-', normal ).
adj( 'guilty', 'guiltier', 'guiltiest', normal ).
adj( 'gullible', '-', '-', normal ).
adj( 'gummy', 'gummier', 'gummiest', normal ).
adj( 'gushing', '-', '-', normal ).
adj( 'gusseted', '-', '-', normal ).
adj( 'gusty', 'gustier', 'gustiest', normal ).
adj( 'gutless', '-', '-', normal ).
adj( 'guttural', '-', '-', normal ).
adj( 'gymnastic', '-', '-', normal ).
adj( 'gynaecological', '-', '-', normal ).
adj( 'gynecological', '-', '-', normal ).
adj( 'gyroscopic', '-', '-', normal ).
adj( 'habitable', '-', '-', normal ).
adj( 'habitual', '-', '-', normal ).
adj( 'hackneyed', '-', '-', normal ).
adj( 'hag-ridden', '-', '-', normal ).
adj( 'haggard', '-', '-', normal ).
adj( 'hair-raising', '-', '-', normal ).
adj( 'hairless', '-', '-', normal ).
adj( 'hairlike', '-', '-', normal ).
adj( 'hairsplitting', '-', '-', normal ).
adj( 'hairy', 'hairier', 'hairiest', normal ).
adj( 'halcyon', '-', '-', normal ).
adj( 'hale', '-', '-', normal ).
adj( 'half-baked', '-', '-', normal ).
adj( 'half-crazed', '-', '-', normal ).
adj( 'half-hardy', '-', '-', normal ).
adj( 'half-hearted', '-', '-', normal ).
adj( 'half-hourly', '-', '-', normal ).
adj( 'half-length', '-', '-', normal ).
adj( 'half-seas-over', '-', '-', pred ).
adj( 'half-size', '-', '-', normal ).
adj( 'half-timbered', '-', '-', normal ).
adj( 'half-tracked', '-', '-', normal ).
adj( 'half-yearly', '-', '-', normal ).
adj( 'halfway', '-', '-', normal ).
adj( 'halfwitted', '-', '-', normal ).
adj( 'hallucinatory', '-', '-', normal ).
adj( 'hallucinogenic', '-', '-', normal ).
adj( 'halt', '-', '-', normal ).
adj( 'ham-fisted', '-', '-', normal ).
adj( 'ham-handed', '-', '-', normal ).
adj( 'hand-picked', '-', '-', normal ).
adj( 'handmade', '-', '-', normal ).
adj( 'handsome', '-', '-', normal ).
adj( 'handy', 'handier', 'handiest', normal ).
adj( 'hangdog', '-', '-', attr ).
adj( 'haphazard', '-', '-', normal ).
adj( 'hapless', '-', '-', normal ).
adj( 'happy', 'happier', 'happiest', normal ).
adj( 'happy-go-lucky', '-', '-', normal ).
adj( 'hard', 'harder', 'hardest', normal ).
adj( 'hard-baked', '-', '-', normal ).
adj( 'hard-bitten', '-', '-', normal ).
adj( 'hard-boiled', '-', '-', normal ).
adj( 'hard-headed', '-', '-', normal ).
adj( 'hardbacked', '-', '-', normal ).
adj( 'hardbound', '-', '-', normal ).
adj( 'hardcovered', '-', '-', normal ).
adj( 'hardhearted', '-', '-', normal ).
adj( 'hardhitting', '-', '-', normal ).
adj( 'hardworking', '-', '-', normal ).
adj( 'hardy', 'hardier', 'hardiest', normal ).
adj( 'harebrained', '-', '-', normal ).
adj( 'harmful', '-', '-', normal ).
adj( 'harmless', '-', '-', normal ).
adj( 'harmonious', '-', '-', normal ).
adj( 'harsh', 'harsher', 'harshest', normal ).
adj( 'harum-scarum', '-', '-', normal ).
adj( 'hasty', 'hastier', 'hastiest', normal ).
adj( 'hateful', '-', '-', normal ).
adj( 'hatless', '-', '-', normal ).
adj( 'haughty', 'haughtier', 'haughtiest', normal ).
adj( 'hawk-eyed', '-', '-', normal ).
adj( 'haywire', '-', '-', pred ).
adj( 'hazardous', '-', '-', normal ).
adj( 'hazy', 'hazier', 'haziest', normal ).
adj( 'head-on', '-', '-', normal ).
adj( 'headed', '-', '-', normal ).
adj( 'headless', '-', '-', normal ).
adj( 'headlong', '-', '-', normal ).
adj( 'headstrong', '-', '-', normal ).
adj( 'heady', 'headier', 'headiest', normal ).
adj( 'healing', '-', '-', normal ).
adj( 'healthful', '-', '-', normal ).
adj( 'healthy', 'healthier', 'healthiest', normal ).
adj( 'heart-rending', '-', '-', normal ).
adj( 'heartbreaking', '-', '-', normal ).
adj( 'heartbroken', '-', '-', normal ).
adj( 'hearted', '-', '-', affix ).
adj( 'heartfelt', '-', '-', normal ).
adj( 'heartless', '-', '-', normal ).
adj( 'heartsick', '-', '-', normal ).
adj( 'hearty', 'heartier', 'heartiest', normal ).
adj( 'heated', '-', '-', normal ).
adj( 'heathenish', '-', '-', normal ).
adj( 'heavenly', '-', '-', normal ).
adj( 'heavensent', '-', '-', normal ).
adj( 'heavenward', '-', '-', normal ).
adj( 'heavenwards', '-', '-', normal ).
adj( 'heavy', 'heavier', 'heaviest', normal ).
adj( 'heavy-handed', '-', '-', normal ).
adj( 'heavy-hearted', '-', '-', normal ).
adj( 'heavy-laden', '-', '-', normal ).
adj( 'hebdomadal', '-', '-', normal ).
adj( 'hectic', '-', '-', normal ).
adj( 'hedonistic', '-', '-', normal ).
adj( 'heedful', '-', '-', normal ).
adj( 'heedless', '-', '-', normal ).
adj( 'hefty', 'heftier', 'heftiest', normal ).
adj( 'heinous', '-', '-', normal ).
adj( 'hellish', '-', '-', normal ).
adj( 'helmeted', '-', '-', normal ).
adj( 'helpful', '-', '-', normal ).
adj( 'helpless', '-', '-', normal ).
adj( 'hempen', '-', '-', normal ).
adj( 'hennaed', '-', '-', normal ).
adj( 'henpecked', '-', '-', normal ).
adj( 'hep', '-', '-', normal ).
adj( 'her', '-', '-', attr ).
adj( 'heraldic', '-', '-', normal ).
adj( 'herbaceous', '-', '-', normal ).
adj( 'herbal', '-', '-', normal ).
adj( 'herbivorous', '-', '-', normal ).
adj( 'herculean', '-', '-', normal ).
adj( 'hereditary', '-', '-', normal ).
adj( 'heretical', '-', '-', normal ).
adj( 'heritable', '-', '-', normal ).
adj( 'hermetic', '-', '-', normal ).
adj( 'heroic', '-', '-', normal ).
adj( 'hesitant', '-', '-', normal ).
adj( 'het-up', '-', '-', normal ).
adj( 'heterodox', '-', '-', normal ).
adj( 'heterogeneous', '-', '-', normal ).
adj( 'heterosexual', '-', '-', normal ).
adj( 'heuristic', '-', '-', normal ).
adj( 'hexagonal', '-', '-', normal ).
adj( 'hi-fi', '-', '-', normal ).
adj( 'hick', '-', '-', normal ).
adj( 'hidebound', '-', '-', normal ).
adj( 'hideous', '-', '-', normal ).
adj( 'hierarchic', '-', '-', normal ).
adj( 'hierarchical', '-', '-', normal ).
adj( 'hieroglyphic', '-', '-', normal ).
adj( 'higgledy-piggledy', '-', '-', normal ).
adj( 'high', 'higher', 'highest', normal ).
adj( 'high-class', '-', '-', normal ).
adj( 'high-falutin', '-', '-', normal ).
adj( 'high-fidelity', '-', '-', normal ).
adj( 'high-grade', '-', '-', normal ).
adj( 'high-handed', '-', '-', normal ).
adj( 'high-keyed', '-', '-', normal ).
adj( 'high-level', '-', '-', normal ).
adj( 'high-minded', '-', '-', normal ).
adj( 'high-necked', '-', '-', normal ).
adj( 'high-octane', '-', '-', normal ).
adj( 'high-pitched', '-', '-', normal ).
adj( 'high-powered', '-', '-', normal ).
adj( 'high-priced', '-', '-', normal ).
adj( 'high-principled', '-', '-', normal ).
adj( 'high-ranking', '-', '-', normal ).
adj( 'high-rise', '-', '-', attr ).
adj( 'high-sounding', '-', '-', normal ).
adj( 'high-speed', '-', '-', normal ).
adj( 'high-spirited', '-', '-', normal ).
adj( 'high-tension', '-', '-', normal ).
adj( 'high-toned', '-', '-', normal ).
adj( 'highborn', '-', '-', normal ).
adj( 'highbrow', '-', '-', normal ).
adj( 'highflown', '-', '-', normal ).
adj( 'highflying', '-', '-', normal ).
adj( 'hilarious', '-', '-', normal ).
adj( 'hilly', 'hillier', 'hilliest', normal ).
adj( 'hind', '-', '-', normal ).
adj( 'hindmost', '-', '-', normal ).
adj( 'hip', '-', '-', normal ).
adj( 'hirsute', '-', '-', normal ).
adj( 'his', '-', '-', normal ).
adj( 'historic', '-', '-', normal ).
adj( 'historical', '-', '-', normal ).
adj( 'histrionic', '-', '-', normal ).
adj( 'hit-and-run', '-', '-', attr ).
adj( 'hoar', '-', '-', normal ).
adj( 'hoarse', 'hoarser', 'hoarsest', normal ).
adj( 'hoary', 'hoarier', 'hoariest', normal ).
adj( 'hobnailed', '-', '-', normal ).
adj( 'hoggish', '-', '-', normal ).
adj( 'hoity-toity', '-', '-', normal ).
adj( 'hole-and-corner', '-', '-', normal ).
adj( 'hollow', '-', '-', normal ).
adj( 'holy', 'holier', 'holiest', normal ).
adj( 'home-baked', '-', '-', normal ).
adj( 'home-brewed', '-', '-', normal ).
adj( 'home-cured', '-', '-', normal ).
adj( 'home-grown', '-', '-', normal ).
adj( 'home-made', '-', '-', normal ).
adj( 'homeless', '-', '-', normal ).
adj( 'homelike', '-', '-', normal ).
adj( 'homely', 'homelier', 'homeliest', normal ).
adj( 'homeopathic', '-', '-', normal ).
adj( 'homesick', '-', '-', normal ).
adj( 'homespun', '-', '-', normal ).
adj( 'homeward', '-', '-', normal ).
adj( 'homey', '-', '-', normal ).
adj( 'homicidal', '-', '-', normal ).
adj( 'homiletic', '-', '-', normal ).
adj( 'homing', '-', '-', normal ).
adj( 'homogeneous', '-', '-', normal ).
adj( 'homosexual', '-', '-', normal ).
adj( 'honest', '-', '-', normal ).
adj( 'honeyed', '-', '-', normal ).
adj( 'honorary', '-', '-', normal ).
adj( 'honorific', '-', '-', normal ).
adj( 'honourable', '-', '-', normal ).
adj( 'hook-nosed', '-', '-', normal ).
adj( 'hooked', '-', '-', normal ).
adj( 'hopeful', '-', '-', normal ).
adj( 'hopeless', '-', '-', normal ).
adj( 'hopped-up', '-', '-', normal ).
adj( 'horizontal', '-', '-', normal ).
adj( 'horn-rimmed', '-', '-', normal ).
adj( 'horned', '-', '-', normal ).
adj( 'hornless', '-', '-', normal ).
adj( 'hornlike', '-', '-', normal ).
adj( 'horny', 'hornier', 'horniest', normal ).
adj( 'horrible', '-', '-', normal ).
adj( 'horrid', '-', '-', normal ).
adj( 'horrific', '-', '-', normal ).
adj( 'horror-stricken', '-', '-', normal ).
adj( 'horror-struck', '-', '-', normal ).
adj( 'hors de combat', '-', '-', pred ).
adj( 'horsy', '-', '-', normal ).
adj( 'hortative', '-', '-', normal ).
adj( 'horticultural', '-', '-', normal ).
adj( 'hospitable', '-', '-', normal ).
adj( 'hostile', '-', '-', normal ).
adj( 'hot', 'hotter', 'hottest', normal ).
adj( 'hot-blooded', '-', '-', normal ).
adj( 'hot-headed', '-', '-', normal ).
adj( 'hot-tempered', '-', '-', normal ).
adj( 'hourly', '-', '-', normal ).
adj( 'housebound', '-', '-', normal ).
adj( 'houseproud', '-', '-', normal ).
adj( 'housetrained', '-', '-', normal ).
adj( 'housewifely', '-', '-', normal ).
adj( 'howling', '-', '-', normal ).
adj( 'hoydenish', '-', '-', normal ).
adj( 'hued', '-', '-', affix ).
adj( 'huffish', '-', '-', normal ).
adj( 'huffy', 'huffier', 'huffiest', normal ).
adj( 'huge', '-', '-', normal ).
adj( 'hugger-mugger', '-', '-', normal ).
adj( 'hulking', '-', '-', normal ).
adj( 'human', '-', '-', normal ).
adj( 'humane', '-', '-', normal ).
adj( 'humanitarian', '-', '-', normal ).
adj( 'humble', 'humbler', 'humblest', normal ).
adj( 'humdrum', '-', '-', normal ).
adj( 'humid', '-', '-', normal ).
adj( 'humorous', '-', '-', normal ).
adj( 'humourless', '-', '-', normal ).
adj( 'humpbacked', '-', '-', normal ).
adj( 'hunchback', '-', '-', normal ).
adj( 'hunchbacked', '-', '-', normal ).
adj( 'hundred', '-', '-', normal ).
adj( 'hundredth', '-', '-', normal ).
adj( 'hungry', 'hungrier', 'hungriest', normal ).
adj( 'hurried', '-', '-', normal ).
adj( 'hurtful', '-', '-', normal ).
adj( 'hush-hush', '-', '-', normal ).
adj( 'husky', 'huskier', 'huskiest', normal ).
adj( 'hutted', '-', '-', normal ).
adj( 'hybrid', '-', '-', normal ).
adj( 'hydraulic', '-', '-', normal ).
adj( 'hydrochloric', '-', '-', normal ).
adj( 'hydroelectric', '-', '-', normal ).
adj( 'hydropathic', '-', '-', normal ).
adj( 'hygienic', '-', '-', normal ).
adj( 'hyperbolic', '-', '-', normal ).
adj( 'hypercritical', '-', '-', normal ).
adj( 'hypersensitive', '-', '-', normal ).
adj( 'hypnotic', '-', '-', normal ).
adj( 'hypochondriac', '-', '-', normal ).
adj( 'hypochondriacal', '-', '-', normal ).
adj( 'hypocritical', '-', '-', normal ).
adj( 'hypodermic', '-', '-', normal ).
adj( 'hypothetical', '-', '-', normal ).
adj( 'hysterical', '-', '-', normal ).
adj( 'iambic', '-', '-', normal ).
adj( 'icebound', '-', '-', normal ).
adj( 'icefree', '-', '-', normal ).
adj( 'icy', 'icier', 'iciest', normal ).
adj( 'ideal', '-', '-', normal ).
adj( 'idealistic', '-', '-', normal ).
adj( 'identical', '-', '-', normal ).
adj( 'identifiable', '-', '-', normal ).
adj( 'ideographic', '-', '-', normal ).
adj( 'ideological', '-', '-', normal ).
adj( 'idiomatic', '-', '-', normal ).
adj( 'idiosyncratic', '-', '-', normal ).
adj( 'idiotic', '-', '-', normal ).
adj( 'idle', 'idler', 'idlest', normal ).
adj( 'idolatrous', '-', '-', normal ).
adj( 'idyllic', '-', '-', normal ).
adj( 'igneous', '-', '-', normal ).
adj( 'ignoble', '-', '-', normal ).
adj( 'ignominious', '-', '-', normal ).
adj( 'ignorant', '-', '-', normal ).
adj( 'ill', '-', '-', normal ).
adj( 'ill-advised', '-', '-', normal ).
adj( 'ill-affected', '-', '-', normal ).
adj( 'ill-bred', '-', '-', normal ).
adj( 'ill-disposed', '-', '-', normal ).
adj( 'ill-fated', '-', '-', normal ).
adj( 'ill-favoured', '-', '-', normal ).
adj( 'ill-gotten', '-', '-', normal ).
adj( 'ill-judged', '-', '-', normal ).
adj( 'ill-mannered', '-', '-', normal ).
adj( 'ill-natured', '-', '-', normal ).
adj( 'ill-omened', '-', '-', normal ).
adj( 'ill-starred', '-', '-', normal ).
adj( 'ill-timed', '-', '-', normal ).
adj( 'illegal', '-', '-', normal ).
adj( 'illegible', '-', '-', normal ).
adj( 'illegitimate', '-', '-', normal ).
adj( 'illiberal', '-', '-', normal ).
adj( 'illicit', '-', '-', normal ).
adj( 'illimitable', '-', '-', normal ).
adj( 'illiterate', '-', '-', normal ).
adj( 'illogical', '-', '-', normal ).
adj( 'illusive', '-', '-', normal ).
adj( 'illusory', '-', '-', normal ).
adj( 'illustrative', '-', '-', normal ).
adj( 'illustrious', '-', '-', normal ).
adj( 'imaginable', '-', '-', normal ).
adj( 'imaginary', '-', '-', normal ).
adj( 'imaginative', '-', '-', normal ).
adj( 'imbecile', '-', '-', normal ).
adj( 'imitative', '-', '-', normal ).
adj( 'immaculate', '-', '-', normal ).
adj( 'immanent', '-', '-', normal ).
adj( 'immaterial', '-', '-', normal ).
adj( 'immature', '-', '-', normal ).
adj( 'immeasurable', '-', '-', normal ).
adj( 'immediate', '-', '-', normal ).
adj( 'immemorial', '-', '-', normal ).
adj( 'immense', '-', '-', normal ).
adj( 'imminent', '-', '-', normal ).
adj( 'immobile', '-', '-', normal ).
adj( 'immoderate', '-', '-', normal ).
adj( 'immodest', '-', '-', normal ).
adj( 'immoral', '-', '-', normal ).
adj( 'immortal', '-', '-', normal ).
adj( 'immovable', '-', '-', normal ).
adj( 'immune', '-', '-', normal ).
adj( 'immutable', '-', '-', normal ).
adj( 'impalpable', '-', '-', normal ).
adj( 'impartial', '-', '-', normal ).
adj( 'impassable', '-', '-', normal ).
adj( 'impassioned', '-', '-', normal ).
adj( 'impassive', '-', '-', normal ).
adj( 'impatient', '-', '-', normal ).
adj( 'impeccable', '-', '-', normal ).
adj( 'impecunious', '-', '-', normal ).
adj( 'impenetrable', '-', '-', normal ).
adj( 'impenitent', '-', '-', normal ).
adj( 'imperative', '-', '-', normal ).
adj( 'imperceptible', '-', '-', normal ).
adj( 'imperfect', '-', '-', normal ).
adj( 'imperial', '-', '-', normal ).
adj( 'imperialistic', '-', '-', normal ).
adj( 'imperious', '-', '-', normal ).
adj( 'imperishable', '-', '-', normal ).
adj( 'impermanent', '-', '-', normal ).
adj( 'impermeable', '-', '-', normal ).
adj( 'impersonal', '-', '-', normal ).
adj( 'impertinent', '-', '-', normal ).
adj( 'imperturbable', '-', '-', normal ).
adj( 'impervious', '-', '-', normal ).
adj( 'impetuous', '-', '-', normal ).
adj( 'impious', '-', '-', normal ).
adj( 'impish', '-', '-', normal ).
adj( 'implacable', '-', '-', normal ).
adj( 'implausible', '-', '-', normal ).
adj( 'implicit', '-', '-', normal ).
adj( 'impolite', '-', '-', normal ).
adj( 'impolitic', '-', '-', normal ).
adj( 'imponderable', '-', '-', normal ).
adj( 'important', '-', '-', normal ).
adj( 'importunate', '-', '-', normal ).
adj( 'imposing', '-', '-', normal ).
adj( 'impossible', '-', '-', normal ).
adj( 'impotent', '-', '-', normal ).
adj( 'impracticable', '-', '-', normal ).
adj( 'impractical', '-', '-', normal ).
adj( 'imprecise', '-', '-', normal ).
adj( 'impregnable', '-', '-', normal ).
adj( 'impressionable', '-', '-', normal ).
adj( 'impressionistic', '-', '-', normal ).
adj( 'impressive', '-', '-', normal ).
adj( 'improbable', '-', '-', normal ).
adj( 'impromptu', '-', '-', normal ).
adj( 'improper', '-', '-', normal ).
adj( 'improvident', '-', '-', normal ).
adj( 'imprudent', '-', '-', normal ).
adj( 'impudent', '-', '-', normal ).
adj( 'impulsive', '-', '-', normal ).
adj( 'impure', '-', '-', normal ).
adj( 'in-chief', '-', '-', affix ).
adj( 'in-service', '-', '-', attr ).
adj( 'inaccessible', '-', '-', normal ).
adj( 'inaccurate', '-', '-', normal ).
adj( 'inactive', '-', '-', normal ).
adj( 'inadequate', '-', '-', normal ).
adj( 'inadmissible', '-', '-', normal ).
adj( 'inadvertent', '-', '-', normal ).
adj( 'inadvisable', '-', '-', normal ).
adj( 'inalienable', '-', '-', normal ).
adj( 'inane', '-', '-', normal ).
adj( 'inanimate', '-', '-', normal ).
adj( 'inapplicable', '-', '-', normal ).
adj( 'inappreciable', '-', '-', normal ).
adj( 'inappropriate', '-', '-', normal ).
adj( 'inapt', '-', '-', normal ).
adj( 'inarticulate', '-', '-', normal ).
adj( 'inattentive', '-', '-', normal ).
adj( 'inaudible', '-', '-', normal ).
adj( 'inaugural', '-', '-', normal ).
adj( 'inauspicious', '-', '-', normal ).
adj( 'inboard', '-', '-', normal ).
adj( 'inborn', '-', '-', normal ).
adj( 'inbound', '-', '-', normal ).
adj( 'inbred', '-', '-', normal ).
adj( 'inbuilt', '-', '-', normal ).
adj( 'incalculable', '-', '-', normal ).
adj( 'incandescent', '-', '-', normal ).
adj( 'incapable', '-', '-', normal ).
adj( 'incarnate', '-', '-', normal ).
adj( 'incautious', '-', '-', normal ).
adj( 'incessant', '-', '-', normal ).
adj( 'incestuous', '-', '-', normal ).
adj( 'inchoate', '-', '-', normal ).
adj( 'inchoative', '-', '-', normal ).
adj( 'incident', '-', '-', normal ).
adj( 'incidental', '-', '-', normal ).
adj( 'incipient', '-', '-', normal ).
adj( 'incisive', '-', '-', normal ).
adj( 'inclement', '-', '-', normal ).
adj( 'inclusive', '-', '-', normal ).
adj( 'incognito', '-', '-', normal ).
adj( 'incoherent', '-', '-', normal ).
adj( 'incombustible', '-', '-', normal ).
adj( 'incoming', '-', '-', normal ).
adj( 'incommensurable', '-', '-', normal ).
adj( 'incommensurate', '-', '-', normal ).
adj( 'incommunicado', '-', '-', normal ).
adj( 'incomparable', '-', '-', normal ).
adj( 'incompatible', '-', '-', normal ).
adj( 'incompetent', '-', '-', normal ).
adj( 'incomplete', '-', '-', normal ).
adj( 'incomprehensible', '-', '-', normal ).
adj( 'incompressible', '-', '-', normal ).
adj( 'inconceivable', '-', '-', normal ).
adj( 'inconclusive', '-', '-', normal ).
adj( 'incongruous', '-', '-', normal ).
adj( 'inconsequent', '-', '-', normal ).
adj( 'inconsequential', '-', '-', normal ).
adj( 'inconsiderable', '-', '-', normal ).
adj( 'inconsiderate', '-', '-', normal ).
adj( 'inconsistent', '-', '-', normal ).
adj( 'inconsolable', '-', '-', normal ).
adj( 'inconspicuous', '-', '-', normal ).
adj( 'inconstant', '-', '-', normal ).
adj( 'incontestable', '-', '-', normal ).
adj( 'incontinent', '-', '-', normal ).
adj( 'incontrovertible', '-', '-', normal ).
adj( 'inconvenient', '-', '-', normal ).
adj( 'inconvertible', '-', '-', normal ).
adj( 'incorporate', '-', '-', normal ).
adj( 'incorporeal', '-', '-', normal ).
adj( 'incorrect', '-', '-', normal ).
adj( 'incorrigible', '-', '-', normal ).
adj( 'incorruptible', '-', '-', normal ).
adj( 'incredible', '-', '-', normal ).
adj( 'incredulous', '-', '-', normal ).
adj( 'incremental', '-', '-', normal ).
adj( 'incumbent', '-', '-', normal ).
adj( 'incurable', '-', '-', normal ).
adj( 'incurious', '-', '-', normal ).
adj( 'incurved', '-', '-', normal ).
adj( 'indebted', '-', '-', normal ).
adj( 'indecent', '-', '-', normal ).
adj( 'indecipherable', '-', '-', normal ).
adj( 'indecisive', '-', '-', normal ).
adj( 'indecorous', '-', '-', normal ).
adj( 'indefatigable', '-', '-', normal ).
adj( 'indefeasible', '-', '-', normal ).
adj( 'indefensible', '-', '-', normal ).
adj( 'indefinable', '-', '-', normal ).
adj( 'indefinite', '-', '-', normal ).
adj( 'indelible', '-', '-', normal ).
adj( 'indelicate', '-', '-', normal ).
adj( 'independent', '-', '-', normal ).
adj( 'indescribable', '-', '-', normal ).
adj( 'indestructible', '-', '-', normal ).
adj( 'indeterminable', '-', '-', normal ).
adj( 'indeterminate', '-', '-', normal ).
adj( 'indicative', '-', '-', normal ).
adj( 'indictable', '-', '-', normal ).
adj( 'indifferent', '-', '-', normal ).
adj( 'indigenous', '-', '-', normal ).
adj( 'indigent', '-', '-', normal ).
adj( 'indigestible', '-', '-', normal ).
adj( 'indignant', '-', '-', normal ).
adj( 'indirect', '-', '-', normal ).
adj( 'indiscernible', '-', '-', normal ).
adj( 'indiscreet', '-', '-', normal ).
adj( 'indiscrete', '-', '-', normal ).
adj( 'indiscriminate', '-', '-', normal ).
adj( 'indispensable', '-', '-', normal ).
adj( 'indisposed', '-', '-', normal ).
adj( 'indisputable', '-', '-', normal ).
adj( 'indissoluble', '-', '-', normal ).
adj( 'indistinct', '-', '-', normal ).
adj( 'indistinguishable', '-', '-', normal ).
adj( 'individual', '-', '-', normal ).
adj( 'individualistic', '-', '-', normal ).
adj( 'indivisible', '-', '-', normal ).
adj( 'indolent', '-', '-', normal ).
adj( 'indomitable', '-', '-', normal ).
adj( 'indoor', '-', '-', normal ).
adj( 'indrawn', '-', '-', normal ).
adj( 'indubitable', '-', '-', normal ).
adj( 'inductive', '-', '-', normal ).
adj( 'indulgent', '-', '-', normal ).
adj( 'industrial', '-', '-', normal ).
adj( 'industrialized', '-', '-', normal ).
adj( 'industrious', '-', '-', normal ).
adj( 'indwelling', '-', '-', normal ).
adj( 'inebriate', '-', '-', normal ).
adj( 'inedible', '-', '-', normal ).
adj( 'ineffable', '-', '-', normal ).
adj( 'ineffective', '-', '-', normal ).
adj( 'ineffectual', '-', '-', normal ).
adj( 'inefficient', '-', '-', normal ).
adj( 'inelastic', '-', '-', normal ).
adj( 'inelegant', '-', '-', normal ).
adj( 'ineligible', '-', '-', normal ).
adj( 'ineluctable', '-', '-', normal ).
adj( 'inept', '-', '-', normal ).
adj( 'inequitable', '-', '-', normal ).
adj( 'ineradicable', '-', '-', normal ).
adj( 'inert', '-', '-', normal ).
adj( 'inescapable', '-', '-', normal ).
adj( 'inessential', '-', '-', normal ).
adj( 'inestimable', '-', '-', normal ).
adj( 'inevitable', '-', '-', normal ).
adj( 'inexact', '-', '-', normal ).
adj( 'inexcusable', '-', '-', normal ).
adj( 'inexhaustible', '-', '-', normal ).
adj( 'inexorable', '-', '-', normal ).
adj( 'inexpedient', '-', '-', normal ).
adj( 'inexpensive', '-', '-', normal ).
adj( 'inexperienced', '-', '-', normal ).
adj( 'inexpert', '-', '-', normal ).
adj( 'inexpiable', '-', '-', normal ).
adj( 'inexplicable', '-', '-', normal ).
adj( 'inexpressible', '-', '-', normal ).
adj( 'inextinguishable', '-', '-', normal ).
adj( 'inextricable', '-', '-', normal ).
adj( 'infallible', '-', '-', normal ).
adj( 'infamous', '-', '-', normal ).
adj( 'infantile', '-', '-', normal ).
adj( 'infectious', '-', '-', normal ).
adj( 'inferential', '-', '-', normal ).
adj( 'inferior', '-', '-', normal ).
adj( 'infernal', '-', '-', normal ).
adj( 'infertile', '-', '-', normal ).
adj( 'infinite', '-', '-', normal ).
adj( 'infinitesimal', '-', '-', normal ).
adj( 'infinitive', '-', '-', normal ).
adj( 'infirm', '-', '-', normal ).
adj( 'inflammable', '-', '-', normal ).
adj( 'inflammatory', '-', '-', normal ).
adj( 'inflatable', '-', '-', normal ).
adj( 'inflationary', '-', '-', normal ).
adj( 'inflectional', '-', '-', normal ).
adj( 'inflexible', '-', '-', normal ).
adj( 'influential', '-', '-', normal ).
adj( 'informal', '-', '-', normal ).
adj( 'informative', '-', '-', normal ).
adj( 'infra dig', '-', '-', pred ).
adj( 'infra-red', '-', '-', normal ).
adj( 'infrequent', '-', '-', normal ).
adj( 'ingenious', '-', '-', normal ).
adj( 'ingenuous', '-', '-', normal ).
adj( 'inglorious', '-', '-', normal ).
adj( 'ingoing', '-', '-', normal ).
adj( 'ingrained', '-', '-', normal ).
adj( 'ingrowing', '-', '-', normal ).
adj( 'inhabitable', '-', '-', normal ).
adj( 'inharmonious', '-', '-', normal ).
adj( 'inherent', '-', '-', normal ).
adj( 'inhibitory', '-', '-', normal ).
adj( 'inhospitable', '-', '-', normal ).
adj( 'inhuman', '-', '-', normal ).
adj( 'inhumane', '-', '-', normal ).
adj( 'inimical', '-', '-', normal ).
adj( 'inimitable', '-', '-', normal ).
adj( 'iniquitous', '-', '-', normal ).
adj( 'initial', '-', '-', normal ).
adj( 'initiate', '-', '-', normal ).
adj( 'injudicious', '-', '-', normal ).
adj( 'injured', '-', '-', normal ).
adj( 'injurious', '-', '-', normal ).
adj( 'inky', 'inkier', 'inkiest', normal ).
adj( 'inland', '-', '-', normal ).
adj( 'inmost', '-', '-', normal ).
adj( 'innate', '-', '-', normal ).
adj( 'inner', '-', '-', normal ).
adj( 'innermost', '-', '-', normal ).
adj( 'innocent', '-', '-', normal ).
adj( 'innocuous', '-', '-', normal ).
adj( 'innumerable', '-', '-', normal ).
adj( 'inoffensive', '-', '-', normal ).
adj( 'inoperable', '-', '-', normal ).
adj( 'inoperative', '-', '-', normal ).
adj( 'inopportune', '-', '-', normal ).
adj( 'inordinate', '-', '-', normal ).
adj( 'inorganic', '-', '-', normal ).
adj( 'inpouring', '-', '-', normal ).
adj( 'inquiring', '-', '-', normal ).
adj( 'inquisitive', '-', '-', normal ).
adj( 'inquisitorial', '-', '-', normal ).
adj( 'insane', '-', '-', normal ).
adj( 'insanitary', '-', '-', normal ).
adj( 'insatiable', '-', '-', normal ).
adj( 'insatiate', '-', '-', normal ).
adj( 'inscrutable', '-', '-', normal ).
adj( 'insectivorous', '-', '-', normal ).
adj( 'insecure', '-', '-', normal ).
adj( 'insensate', '-', '-', normal ).
adj( 'insensible', '-', '-', normal ).
adj( 'insensitive', '-', '-', normal ).
adj( 'insentient', '-', '-', normal ).
adj( 'inseparable', '-', '-', normal ).
adj( 'inshore', '-', '-', normal ).
adj( 'inside', '-', '-', normal ).
adj( 'insidious', '-', '-', normal ).
adj( 'insignificant', '-', '-', normal ).
adj( 'insincere', '-', '-', normal ).
adj( 'insipid', '-', '-', normal ).
adj( 'insistent', '-', '-', normal ).
adj( 'insolent', '-', '-', normal ).
adj( 'insoluble', '-', '-', normal ).
adj( 'insolvent', '-', '-', normal ).
adj( 'insouciant', '-', '-', normal ).
adj( 'inspirational', '-', '-', normal ).
adj( 'inspired', '-', '-', normal ).
adj( 'instant', '-', '-', normal ).
adj( 'instantaneous', '-', '-', normal ).
adj( 'instinct', '-', '-', pred ).
adj( 'instinctive', '-', '-', normal ).
adj( 'institutional', '-', '-', normal ).
adj( 'instructional', '-', '-', normal ).
adj( 'instructive', '-', '-', normal ).
adj( 'instrumental', '-', '-', normal ).
adj( 'insubordinate', '-', '-', normal ).
adj( 'insubstantial', '-', '-', normal ).
adj( 'insufferable', '-', '-', normal ).
adj( 'insufficient', '-', '-', normal ).
adj( 'insular', '-', '-', normal ).
adj( 'insulting', '-', '-', normal ).
adj( 'insuperable', '-', '-', normal ).
adj( 'insupportable', '-', '-', normal ).
adj( 'insurgent', '-', '-', normal ).
adj( 'insurmountable', '-', '-', normal ).
adj( 'intact', '-', '-', normal ).
adj( 'intangible', '-', '-', normal ).
adj( 'integral', '-', '-', normal ).
adj( 'intellectual', '-', '-', normal ).
adj( 'intelligent', '-', '-', normal ).
adj( 'intelligible', '-', '-', normal ).
adj( 'intemperate', '-', '-', normal ).
adj( 'intense', '-', '-', normal ).
adj( 'intensive', '-', '-', normal ).
adj( 'intent', '-', '-', normal ).
adj( 'intentional', '-', '-', normal ).
adj( 'interactive', '-', '-', normal ).
adj( 'intercalary', '-', '-', normal ).
adj( 'interchangeable', '-', '-', normal ).
adj( 'intercollegiate', '-', '-', normal ).
adj( 'intercontinental', '-', '-', normal ).
adj( 'interdenominational', '-', '-', normal ).
adj( 'interdepartmental', '-', '-', normal ).
adj( 'interdependent', '-', '-', normal ).
adj( 'interdisciplinary', '-', '-', normal ).
adj( 'interested', '-', '-', normal ).
adj( 'interesting', '-', '-', normal ).
adj( 'interior', '-', '-', normal ).
adj( 'intermediate', '-', '-', normal ).
adj( 'interminable', '-', '-', normal ).
adj( 'intermittent', '-', '-', normal ).
adj( 'internal', '-', '-', normal ).
adj( 'international', '-', '-', normal ).
adj( 'internecine', '-', '-', normal ).
adj( 'interplanetary', '-', '-', normal ).
adj( 'interpretative', '-', '-', normal ).
adj( 'interracial', '-', '-', normal ).
adj( 'interrogative', '-', '-', normal ).
adj( 'interrogatory', '-', '-', normal ).
adj( 'interstate', '-', '-', normal ).
adj( 'interstellar', '-', '-', normal ).
adj( 'intertribal', '-', '-', normal ).
adj( 'intestate', '-', '-', normal ).
adj( 'intestinal', '-', '-', normal ).
adj( 'intimate', '-', '-', normal ).
adj( 'intolerable', '-', '-', normal ).
adj( 'intolerant', '-', '-', normal ).
adj( 'intoxicant', '-', '-', normal ).
adj( 'intra-uterine', '-', '-', normal ).
adj( 'intractable', '-', '-', normal ).
adj( 'intramural', '-', '-', normal ).
adj( 'intransigent', '-', '-', normal ).
adj( 'intransitive', '-', '-', normal ).
adj( 'intravenous', '-', '-', normal ).
adj( 'intrepid', '-', '-', normal ).
adj( 'intricate', '-', '-', normal ).
adj( 'intrinsic', '-', '-', normal ).
adj( 'introductory', '-', '-', normal ).
adj( 'introspective', '-', '-', normal ).
adj( 'intrusive', '-', '-', normal ).
adj( 'intuitive', '-', '-', normal ).
adj( 'invalid', '-', '-', normal ).
adj( 'invalid', '-', '-', normal ).
adj( 'invaluable', '-', '-', normal ).
adj( 'invariable', '-', '-', normal ).
adj( 'invasive', '-', '-', normal ).
adj( 'inventive', '-', '-', normal ).
adj( 'inverse', '-', '-', normal ).
adj( 'invertebrate', '-', '-', normal ).
adj( 'inveterate', '-', '-', normal ).
adj( 'invidious', '-', '-', normal ).
adj( 'invincible', '-', '-', normal ).
adj( 'inviolable', '-', '-', normal ).
adj( 'inviolate', '-', '-', normal ).
adj( 'invisible', '-', '-', normal ).
adj( 'inviting', '-', '-', normal ).
adj( 'involuntary', '-', '-', normal ).
adj( 'involute', '-', '-', normal ).
adj( 'involved', '-', '-', normal ).
adj( 'invulnerable', '-', '-', normal ).
adj( 'inward', '-', '-', normal ).
adj( 'inwrought', '-', '-', normal ).
adj( 'irascible', '-', '-', normal ).
adj( 'irate', '-', '-', normal ).
adj( 'ireful', '-', '-', normal ).
adj( 'iridescent', '-', '-', normal ).
adj( 'irksome', '-', '-', normal ).
adj( 'iron-grey', '-', '-', normal ).
adj( 'ironclad', '-', '-', normal ).
adj( 'ironic', '-', '-', normal ).
adj( 'ironical', '-', '-', normal ).
adj( 'irrational', '-', '-', normal ).
adj( 'irreconcilable', '-', '-', normal ).
adj( 'irrecoverable', '-', '-', normal ).
adj( 'irredeemable', '-', '-', normal ).
adj( 'irreducible', '-', '-', normal ).
adj( 'irrefutable', '-', '-', normal ).
adj( 'irregular', '-', '-', normal ).
adj( 'irrelevant', '-', '-', normal ).
adj( 'irreligious', '-', '-', normal ).
adj( 'irremediable', '-', '-', normal ).
adj( 'irremovable', '-', '-', normal ).
adj( 'irreparable', '-', '-', normal ).
adj( 'irreplaceable', '-', '-', normal ).
adj( 'irrepressible', '-', '-', normal ).
adj( 'irreproachable', '-', '-', normal ).
adj( 'irresistible', '-', '-', normal ).
adj( 'irresolute', '-', '-', normal ).
adj( 'irrespective', '-', '-', normal ).
adj( 'irresponsible', '-', '-', normal ).
adj( 'irretrievable', '-', '-', normal ).
adj( 'irreverent', '-', '-', normal ).
adj( 'irreversible', '-', '-', normal ).
adj( 'irrevocable', '-', '-', normal ).
adj( 'irritable', '-', '-', normal ).
adj( 'irritant', '-', '-', normal ).
adj( 'isosceles', '-', '-', normal ).
adj( 'italic', '-', '-', normal ).
adj( 'itchy', 'itchier', 'itchiest', normal ).
adj( 'itinerant', '-', '-', normal ).
adj( 'its', '-', '-', normal ).
adj( 'ivied', '-', '-', normal ).
adj( 'jaded', '-', '-', normal ).
adj( 'jagged', '-', '-', normal ).
adj( 'jaggy', 'jaggier', 'jaggiest', normal ).
adj( 'jarring', '-', '-', normal ).
adj( 'jaunty', 'jauntier', 'jauntiest', normal ).
adj( 'jazzy', 'jazzier', 'jazziest', normal ).
adj( 'jealous', '-', '-', normal ).
adj( 'jejune', '-', '-', normal ).
adj( 'jellied', '-', '-', normal ).
adj( 'jerky', 'jerkier', 'jerkiest', normal ).
adj( 'jerry-built', '-', '-', normal ).
adj( 'jesting', '-', '-', normal ).
adj( 'jet-black', '-', '-', normal ).
adj( 'jet-propelled', '-', '-', normal ).
adj( 'jiggered', '-', '-', pred ).
adj( 'jingoistic', '-', '-', normal ).
adj( 'jittery', '-', '-', normal ).
adj( 'jocose', '-', '-', normal ).
adj( 'jocular', '-', '-', normal ).
adj( 'jocund', '-', '-', normal ).
adj( 'joint', '-', '-', normal ).
adj( 'jolly', 'jollier', 'jolliest', normal ).
adj( 'jolty', 'joltier', 'joltiest', normal ).
adj( 'journalistic', '-', '-', normal ).
adj( 'jovial', '-', '-', normal ).
adj( 'jowly', 'jowlier', 'jowliest', normal ).
adj( 'joyful', '-', '-', normal ).
adj( 'joyless', '-', '-', normal ).
adj( 'joyous', '-', '-', normal ).
adj( 'jubilant', '-', '-', normal ).
adj( 'judicial', '-', '-', normal ).
adj( 'judicious', '-', '-', normal ).
adj( 'jugular', '-', '-', normal ).
adj( 'juicy', 'juicier', 'juiciest', normal ).
adj( 'jumbo', '-', '-', normal ).
adj( 'jumped-up', '-', '-', normal ).
adj( 'jumpy', 'jumpier', 'jumpiest', normal ).
adj( 'jungly', 'junglier', 'jungliest', normal ).
adj( 'junior', '-', '-', normal ).
adj( 'juridical', '-', '-', normal ).
adj( 'just', '-', '-', normal ).
adj( 'justifiable', '-', '-', normal ).
adj( 'juvenile', '-', '-', normal ).
adj( 'kaleidoscopic', '-', '-', normal ).
adj( 'kaput', '-', '-', normal ).
adj( 'keen', 'keener', 'keenest', normal ).
adj( 'keyless', '-', '-', normal ).
adj( 'khaki', '-', '-', normal ).
adj( 'killing', '-', '-', normal ).
adj( 'kind', 'kinder', 'kindest', normal ).
adj( 'kind-hearted', '-', '-', normal ).
adj( 'kindly', 'kindlier', 'kindliest', normal ).
adj( 'kindred', '-', '-', normal ).
adj( 'kinetic', '-', '-', normal ).
adj( 'king-size', '-', '-', normal ).
adj( 'king-sized', '-', '-', normal ).
adj( 'kinglike', '-', '-', normal ).
adj( 'kingly', '-', '-', normal ).
adj( 'kinky', 'kinkier', 'kinkiest', normal ).
adj( 'kitsch', '-', '-', normal ).
adj( 'kittenish', '-', '-', normal ).
adj( 'knavish', '-', '-', normal ).
adj( 'knee-deep', '-', '-', normal ).
adj( 'knee-high', '-', '-', normal ).
adj( 'knightly', '-', '-', normal ).
adj( 'knobbly', 'knobblier', 'knobbliest', normal ).
adj( 'knock-kneed', '-', '-', normal ).
adj( 'knockabout', '-', '-', normal ).
adj( 'knockdown', '-', '-', normal ).
adj( 'knockout', '-', '-', normal ).
adj( 'knotty', 'knottier', 'knottiest', normal ).
adj( 'knowing', '-', '-', normal ).
adj( 'knowledgeable', '-', '-', normal ).
adj( 'kosher', '-', '-', normal ).
adj( 'la-di-da', '-', '-', normal ).
adj( 'labial', '-', '-', normal ).
adj( 'laborious', '-', '-', normal ).
adj( 'labour-saving', '-', '-', normal ).
adj( 'laboured', '-', '-', normal ).
adj( 'labyrinthine', '-', '-', normal ).
adj( 'lachrymal', '-', '-', normal ).
adj( 'lachrymose', '-', '-', normal ).
adj( 'lack-lustre', '-', '-', normal ).
adj( 'lackadaisical', '-', '-', normal ).
adj( 'laconic', '-', '-', normal ).
adj( 'lactic', '-', '-', normal ).
adj( 'lacy', 'lacier', 'laciest', normal ).
adj( 'ladder-proof', '-', '-', normal ).
adj( 'laden', '-', '-', normal ).
adj( 'ladylike', '-', '-', normal ).
adj( 'laic', '-', '-', normal ).
adj( 'lambent', '-', '-', normal ).
adj( 'lame', 'lamer', 'lamest', normal ).
adj( 'lamentable', '-', '-', normal ).
adj( 'lamplit', '-', '-', normal ).
adj( 'landed', '-', '-', normal ).
adj( 'landless', '-', '-', normal ).
adj( 'landlocked', '-', '-', normal ).
adj( 'languid', '-', '-', normal ).
adj( 'languorous', '-', '-', normal ).
adj( 'lank', '-', '-', normal ).
adj( 'lanky', 'lankier', 'lankiest', normal ).
adj( 'lantern-jawed', '-', '-', normal ).
adj( 'lapidary', '-', '-', normal ).
adj( 'larboard', '-', '-', normal ).
adj( 'large', 'larger', 'largest', normal ).
adj( 'large-scale', '-', '-', normal ).
adj( 'largish', '-', '-', normal ).
adj( 'larval', '-', '-', normal ).
adj( 'lascivious', '-', '-', normal ).
adj( 'last', '-', '-', normal ).
adj( 'lasting', '-', '-', normal ).
adj( 'late', 'later', 'latest', normal ).
adj( 'lateen', '-', '-', normal ).
adj( 'latent', '-', '-', normal ).
adj( 'lateral', '-', '-', normal ).
adj( 'latish', '-', '-', normal ).
adj( 'latitudinal', '-', '-', normal ).
adj( 'latitudinarian', '-', '-', normal ).
adj( 'latter', '-', '-', normal ).
adj( 'latter-day', '-', '-', normal ).
adj( 'latticed', '-', '-', normal ).
adj( 'laudable', '-', '-', normal ).
adj( 'laudatory', '-', '-', normal ).
adj( 'laughable', '-', '-', normal ).
adj( 'laughing', '-', '-', normal ).
adj( 'laureate', '-', '-', normal ).
adj( 'laurelled', '-', '-', normal ).
adj( 'lavish', '-', '-', normal ).
adj( 'law-abiding', '-', '-', normal ).
adj( 'lawful', '-', '-', normal ).
adj( 'lawless', '-', '-', normal ).
adj( 'lax', '-', '-', normal ).
adj( 'laxative', '-', '-', normal ).
adj( 'lay', '-', '-', normal ).
adj( 'lazy', 'lazier', 'laziest', normal ).
adj( 'leaded', '-', '-', normal ).
adj( 'leaden', '-', '-', normal ).
adj( 'leaderless', '-', '-', normal ).
adj( 'leading', '-', '-', normal ).
adj( 'leafless', '-', '-', normal ).
adj( 'leafy', 'leafier', 'leafiest', normal ).
adj( 'leaky', 'leakier', 'leakiest', normal ).
adj( 'leal', '-', '-', normal ).
adj( 'lean', 'leaner', 'leanest', normal ).
adj( 'learned', '-', '-', normal ).
adj( 'leasehold', '-', '-', normal ).
adj( 'least', '-', '-', normal ).
adj( 'leathery', '-', '-', normal ).
adj( 'lecherous', '-', '-', normal ).
adj( 'leeward', '-', '-', normal ).
adj( 'left', '-', '-', normal ).
adj( 'left-hand', '-', '-', normal ).
adj( 'left-handed', '-', '-', normal ).
adj( 'legal', '-', '-', normal ).
adj( 'legalistic', '-', '-', normal ).
adj( 'legato', '-', '-', normal ).
adj( 'legendary', '-', '-', normal ).
adj( 'legged', '-', '-', affix ).
adj( 'leggy', '-', '-', normal ).
adj( 'legible', '-', '-', normal ).
adj( 'legislative', '-', '-', normal ).
adj( 'legitimate', '-', '-', normal ).
adj( 'legless', '-', '-', normal ).
adj( 'leguminous', '-', '-', normal ).
adj( 'leisured', '-', '-', normal ).
adj( 'leisurely', '-', '-', normal ).
adj( 'lengthways', '-', '-', normal ).
adj( 'lengthy', 'lengthier', 'lengthiest', normal ).
adj( 'lenient', '-', '-', normal ).
adj( 'lento', '-', '-', normal ).
adj( 'leonine', '-', '-', normal ).
adj( 'leprous', '-', '-', normal ).
adj( 'less', '-', '-', normal ).
adj( 'lesser', '-', '-', normal ).
adj( 'lethal', '-', '-', normal ).
adj( 'lethargic', '-', '-', normal ).
adj( 'lettered', '-', '-', normal ).
adj( 'level', '-', '-', normal ).
adj( 'level-headed', '-', '-', normal ).
adj( 'lewd', 'lewder', 'lewdest', normal ).
adj( 'lexical', '-', '-', normal ).
adj( 'liable', '-', '-', normal ).
adj( 'libellous', '-', '-', normal ).
adj( 'liberal', '-', '-', normal ).
adj( 'libidinous', '-', '-', normal ).
adj( 'licentious', '-', '-', normal ).
adj( 'licit', '-', '-', normal ).
adj( 'lidless', '-', '-', normal ).
adj( 'liege', '-', '-', normal ).
adj( 'life-giving', '-', '-', normal ).
adj( 'life-size', '-', '-', normal ).
adj( 'life-sized', '-', '-', normal ).
adj( 'lifeless', '-', '-', normal ).
adj( 'lifelike', '-', '-', normal ).
adj( 'lifelong', '-', '-', normal ).
adj( 'light', 'lighter', 'lightest', normal ).
adj( 'light-armed', '-', '-', normal ).
adj( 'light-coloured', '-', '-', normal ).
adj( 'light-fingered', '-', '-', normal ).
adj( 'light-handed', '-', '-', normal ).
adj( 'light-headed', '-', '-', normal ).
adj( 'light-hearted', '-', '-', normal ).
adj( 'light-minded', '-', '-', normal ).
adj( 'lighting-up', '-', '-', attr ).
adj( 'lightsome', '-', '-', normal ).
adj( 'lightweight', '-', '-', normal ).
adj( 'ligneous', '-', '-', normal ).
adj( 'likable', '-', '-', normal ).
adj( 'like', '-', '-', normal ).
adj( 'like-minded', '-', '-', normal ).
adj( 'likeable', '-', '-', normal ).
adj( 'likely', 'likelier', 'likeliest', normal ).
adj( 'lily-livered', '-', '-', normal ).
adj( 'lily-white', '-', '-', normal ).
adj( 'limbed', '-', '-', affix ).
adj( 'limber', '-', '-', normal ).
adj( 'limbless', '-', '-', normal ).
adj( 'limitless', '-', '-', normal ).
adj( 'limp', '-', '-', normal ).
adj( 'limpid', '-', '-', normal ).
adj( 'lineal', '-', '-', normal ).
adj( 'linear', '-', '-', normal ).
adj( 'lingering', '-', '-', normal ).
adj( 'lingual', '-', '-', normal ).
adj( 'linguistic', '-', '-', normal ).
adj( 'lion-hearted', '-', '-', normal ).
adj( 'lipped', '-', '-', affix ).
adj( 'liquescent', '-', '-', normal ).
adj( 'liquid', '-', '-', normal ).
adj( 'lissom', '-', '-', normal ).
adj( 'lissome', '-', '-', normal ).
adj( 'listless', '-', '-', normal ).
adj( 'literal', '-', '-', normal ).
adj( 'literary', '-', '-', normal ).
adj( 'literate', '-', '-', normal ).
adj( 'lithe', '-', '-', normal ).
adj( 'lithographic', '-', '-', normal ).
adj( 'litigious', '-', '-', normal ).
adj( 'little', 'littler', 'littlest', normal ).
adj( 'littoral', '-', '-', normal ).
adj( 'liturgical', '-', '-', normal ).
adj( 'livable', '-', '-', normal ).
adj( 'live', '-', '-', normal ).
adj( 'liveable', '-', '-', normal ).
adj( 'livelong', '-', '-', normal ).
adj( 'lively', 'livelier', 'liveliest', normal ).
adj( 'liveried', '-', '-', normal ).
adj( 'liverish', '-', '-', normal ).
adj( 'livery', '-', '-', normal ).
adj( 'livid', '-', '-', normal ).
adj( 'living', '-', '-', normal ).
adj( 'loaded', '-', '-', normal ).
adj( 'loamy', 'loamier', 'loamiest', normal ).
adj( 'loath', '-', '-', normal ).
adj( 'loathly', 'loathlier', 'loathliest', normal ).
adj( 'loathsome', '-', '-', normal ).
adj( 'lobed', '-', '-', normal ).
adj( 'local', '-', '-', normal ).
adj( 'lockup', '-', '-', normal ).
adj( 'loco', '-', '-', normal ).
adj( 'locomotive', '-', '-', normal ).
adj( 'lofty', 'loftier', 'loftiest', normal ).
adj( 'logarithmic', '-', '-', normal ).
adj( 'logical', '-', '-', normal ).
adj( 'lone', '-', '-', normal ).
adj( 'lonely', 'lonelier', 'loneliest', normal ).
adj( 'lonesome', '-', '-', normal ).
adj( 'long', 'longer', 'longest', normal ).
adj( 'long-distance', '-', '-', attr ).
adj( 'long-drawn-out', '-', '-', normal ).
adj( 'long-haired', '-', '-', normal ).
adj( 'long-headed', '-', '-', normal ).
adj( 'long-lived', '-', '-', normal ).
adj( 'long-play', '-', '-', normal ).
adj( 'long-playing', '-', '-', normal ).
adj( 'long-range', '-', '-', attr ).
adj( 'long-sighted', '-', '-', normal ).
adj( 'long-standing', '-', '-', normal ).
adj( 'long-suffering', '-', '-', normal ).
adj( 'long-term', '-', '-', attr ).
adj( 'long-time', '-', '-', attr ).
adj( 'long-winded', '-', '-', normal ).
adj( 'longing', '-', '-', normal ).
adj( 'longish', '-', '-', normal ).
adj( 'longitudinal', '-', '-', normal ).
adj( 'loony', 'loonier', 'looniest', normal ).
adj( 'loopy', '-', '-', normal ).
adj( 'loose', 'looser', 'loosest', normal ).
adj( 'loose-leaf', '-', '-', attr ).
adj( 'lop-eared', '-', '-', normal ).
adj( 'lopsided', '-', '-', normal ).
adj( 'loquacious', '-', '-', normal ).
adj( 'lordless', '-', '-', normal ).
adj( 'lordly', 'lordlier', 'lordliest', normal ).
adj( 'lorn', '-', '-', normal ).
adj( 'loth', '-', '-', normal ).
adj( 'loud', 'louder', 'loudest', normal ).
adj( 'lousy', 'lousier', 'lousiest', normal ).
adj( 'loutish', '-', '-', normal ).
adj( 'louvered', '-', '-', normal ).
adj( 'lovable', '-', '-', normal ).
adj( 'loveless', '-', '-', normal ).
adj( 'lovelorn', '-', '-', normal ).
adj( 'lovely', 'lovelier', 'loveliest', normal ).
adj( 'loverlike', '-', '-', normal ).
adj( 'lovesick', '-', '-', normal ).
adj( 'loving', '-', '-', normal ).
adj( 'low', 'lower', 'lowest', normal ).
adj( 'low-down', '-', '-', normal ).
adj( 'low-keyed', '-', '-', normal ).
adj( 'low-pitched', '-', '-', normal ).
adj( 'low-spirited', '-', '-', normal ).
adj( 'lowborn', '-', '-', normal ).
adj( 'lowbred', '-', '-', normal ).
adj( 'lowbrow', '-', '-', normal ).
adj( 'lowermost', '-', '-', normal ).
adj( 'lowly', 'lowlier', 'lowliest', normal ).
adj( 'loyal', 'loyaller', 'loyallest', normal ).
adj( 'lubberly', '-', '-', normal ).
adj( 'lucent', '-', '-', normal ).
adj( 'lucid', '-', '-', normal ).
adj( 'luckless', '-', '-', normal ).
adj( 'lucky', 'luckier', 'luckiest', normal ).
adj( 'lucrative', '-', '-', normal ).
adj( 'ludicrous', '-', '-', normal ).
adj( 'lugubrious', '-', '-', normal ).
adj( 'lukewarm', '-', '-', normal ).
adj( 'lumbar', '-', '-', normal ).
adj( 'luminous', '-', '-', normal ).
adj( 'lumpish', '-', '-', normal ).
adj( 'lumpy', 'lumpier', 'lumpiest', normal ).
adj( 'lunar', '-', '-', normal ).
adj( 'lurid', '-', '-', normal ).
adj( 'luscious', '-', '-', normal ).
adj( 'lush', 'lusher', 'lushest', normal ).
adj( 'lustful', '-', '-', normal ).
adj( 'lustrous', '-', '-', normal ).
adj( 'lusty', 'lustier', 'lustiest', normal ).
adj( 'luxe', '-', '-', normal ).
adj( 'luxuriant', '-', '-', normal ).
adj( 'luxurious', '-', '-', normal ).
adj( 'lying-in', '-', '-', normal ).
adj( 'lymphatic', '-', '-', normal ).
adj( 'lynx-eyed', '-', '-', normal ).
adj( 'lyric', '-', '-', normal ).
adj( 'lyrical', '-', '-', normal ).
adj( 'macabre', '-', '-', normal ).
adj( 'machiavellian', '-', '-', normal ).
adj( 'machine-made', '-', '-', normal ).
adj( 'macrobiotic', '-', '-', normal ).
adj( 'mad', 'madder', 'maddest', normal ).
adj( 'magenta', '-', '-', normal ).
adj( 'maggoty', '-', '-', normal ).
adj( 'magic', '-', '-', normal ).
adj( 'magical', '-', '-', normal ).
adj( 'magisterial', '-', '-', normal ).
adj( 'magnanimous', '-', '-', normal ).
adj( 'magnetic', '-', '-', normal ).
adj( 'magnificent', '-', '-', normal ).
adj( 'magniloquent', '-', '-', normal ).
adj( 'maiden', '-', '-', normal ).
adj( 'maidenlike', '-', '-', normal ).
adj( 'maidenly', '-', '-', normal ).
adj( 'mailed', '-', '-', normal ).
adj( 'main', '-', '-', normal ).
adj( 'maintainable', '-', '-', normal ).
adj( 'majestic', '-', '-', normal ).
adj( 'major', '-', '-', normal ).
adj( 'maladjusted', '-', '-', normal ).
adj( 'maladroit', '-', '-', normal ).
adj( 'malapropos', '-', '-', normal ).
adj( 'malarial', '-', '-', normal ).
adj( 'malcontent', '-', '-', normal ).
adj( 'male', '-', '-', normal ).
adj( 'maleficent', '-', '-', normal ).
adj( 'malevolent', '-', '-', normal ).
adj( 'malformed', '-', '-', normal ).
adj( 'malicious', '-', '-', normal ).
adj( 'malign', '-', '-', normal ).
adj( 'malignant', '-', '-', normal ).
adj( 'malleable', '-', '-', normal ).
adj( 'malodorous', '-', '-', normal ).
adj( 'man-sized', '-', '-', normal ).
adj( 'manageable', '-', '-', normal ).
adj( 'managerial', '-', '-', normal ).
adj( 'mandatory', '-', '-', normal ).
adj( 'manful', '-', '-', normal ).
adj( 'mangy', 'mangier', 'mangiest', normal ).
adj( 'maniacal', '-', '-', normal ).
adj( 'manic-depressive', '-', '-', normal ).
adj( 'manifest', '-', '-', normal ).
adj( 'manifold', '-', '-', normal ).
adj( 'manipulative', '-', '-', normal ).
adj( 'manlike', '-', '-', normal ).
adj( 'manly', 'manlier', 'manliest', normal ).
adj( 'mannered', '-', '-', normal ).
adj( 'mannerly', '-', '-', normal ).
adj( 'mannish', '-', '-', normal ).
adj( 'manoeuvrable', '-', '-', normal ).
adj( 'manorial', '-', '-', normal ).
adj( 'manual', '-', '-', normal ).
adj( 'many', '-', '-', normal ).
adj( 'many-sided', '-', '-', normal ).
adj( 'marbled', '-', '-', normal ).
adj( 'marginal', '-', '-', normal ).
adj( 'marine', '-', '-', normal ).
adj( 'marital', '-', '-', normal ).
adj( 'maritime', '-', '-', normal ).
adj( 'marked', '-', '-', normal ).
adj( 'marketable', '-', '-', normal ).
adj( 'marmoreal', '-', '-', normal ).
adj( 'maroon', '-', '-', normal ).
adj( 'marriageable', '-', '-', normal ).
adj( 'married', '-', '-', normal ).
adj( 'marshy', 'marshier', 'marshiest', normal ).
adj( 'marsupial', '-', '-', normal ).
adj( 'martial', '-', '-', normal ).
adj( 'marvellous', '-', '-', normal ).
adj( 'marvelous', '-', '-', normal ).
adj( 'masculine', '-', '-', normal ).
adj( 'masochistic', '-', '-', normal ).
adj( 'masonic', '-', '-', normal ).
adj( 'massive', '-', '-', normal ).
adj( 'massy', 'massier', 'massiest', normal ).
adj( 'masterful', '-', '-', normal ).
adj( 'masterless', '-', '-', normal ).
adj( 'masterly', '-', '-', normal ).
adj( 'mat', '-', '-', normal ).
adj( 'matchless', '-', '-', normal ).
adj( 'material', '-', '-', normal ).
adj( 'materialistic', '-', '-', normal ).
adj( 'maternal', '-', '-', normal ).
adj( 'matey', '-', '-', normal ).
adj( 'mathematical', '-', '-', normal ).
adj( 'matriarchal', '-', '-', normal ).
adj( 'matrimonial', '-', '-', normal ).
adj( 'matronly', '-', '-', normal ).
adj( 'matt', '-', '-', normal ).
adj( 'matted', '-', '-', normal ).
adj( 'matter-of-course', '-', '-', normal ).
adj( 'matter-of-fact', '-', '-', normal ).
adj( 'mature', '-', '-', normal ).
adj( 'matutinal', '-', '-', normal ).
adj( 'maudlin', '-', '-', normal ).
adj( 'mauve', '-', '-', normal ).
adj( 'mawkish', '-', '-', normal ).
adj( 'maximal', '-', '-', normal ).
adj( 'maximum', '-', '-', normal ).
adj( 'mayoral', '-', '-', normal ).
adj( 'mazed', '-', '-', normal ).
adj( 'meagre', '-', '-', normal ).
adj( 'mealy', 'mealier', 'mealiest', normal ).
adj( 'mealy-mouthed', '-', '-', normal ).
adj( 'mean', 'meaner', 'meanest', normal ).
adj( 'meaning', '-', '-', normal ).
adj( 'meaningful', '-', '-', normal ).
adj( 'meaningless', '-', '-', normal ).
adj( 'measly', '-', '-', normal ).
adj( 'measurable', '-', '-', normal ).
adj( 'measured', '-', '-', normal ).
adj( 'measureless', '-', '-', normal ).
adj( 'meatless', '-', '-', normal ).
adj( 'meaty', 'meatier', 'meatiest', normal ).
adj( 'mechanical', '-', '-', normal ).
adj( 'mechanistic', '-', '-', normal ).
adj( 'meddlesome', '-', '-', normal ).
adj( 'mediaeval', '-', '-', normal ).
adj( 'medial', '-', '-', normal ).
adj( 'median', '-', '-', normal ).
adj( 'medical', '-', '-', normal ).
adj( 'medicinal', '-', '-', normal ).
adj( 'medieval', '-', '-', normal ).
adj( 'mediocre', '-', '-', normal ).
adj( 'meditative', '-', '-', normal ).
adj( 'medium', '-', '-', normal ).
adj( 'meek', 'meeker', 'meekest', normal ).
adj( 'meet', '-', '-', normal ).
adj( 'megalithic', '-', '-', normal ).
adj( 'melancholic', '-', '-', normal ).
adj( 'melancholy', '-', '-', normal ).
adj( 'mellifluous', '-', '-', normal ).
adj( 'mellow', 'mellower', 'mellowest', normal ).
adj( 'melodic', '-', '-', normal ).
adj( 'melodious', '-', '-', normal ).
adj( 'melodramatic', '-', '-', normal ).
adj( 'melting', '-', '-', normal ).
adj( 'membranous', '-', '-', normal ).
adj( 'memorable', '-', '-', normal ).
adj( 'mendacious', '-', '-', normal ).
adj( 'mendicant', '-', '-', normal ).
adj( 'menial', '-', '-', normal ).
adj( 'menstrual', '-', '-', normal ).
adj( 'mensurable', '-', '-', normal ).
adj( 'mental', '-', '-', normal ).
adj( 'mentholated', '-', '-', normal ).
adj( 'mentioned', '-', '-', affix ).
adj( 'mercantile', '-', '-', normal ).
adj( 'mercenary', '-', '-', normal ).
adj( 'merciful', '-', '-', normal ).
adj( 'merciless', '-', '-', normal ).
adj( 'mercurial', '-', '-', normal ).
adj( 'mere', '-', '-', normal ).
adj( 'meretricious', '-', '-', normal ).
adj( 'meridional', '-', '-', normal ).
adj( 'meritocratic', '-', '-', normal ).
adj( 'meritorious', '-', '-', normal ).
adj( 'merry', 'merrier', 'merriest', normal ).
adj( 'mesmeric', '-', '-', normal ).
adj( 'messianic', '-', '-', normal ).
adj( 'messy', 'messier', 'messiest', normal ).
adj( 'metabolic', '-', '-', normal ).
adj( 'metacarpal', '-', '-', normal ).
adj( 'metallic', '-', '-', normal ).
adj( 'metallurgical', '-', '-', normal ).
adj( 'metaphorical', '-', '-', normal ).
adj( 'metaphysical', '-', '-', normal ).
adj( 'metatarsal', '-', '-', normal ).
adj( 'meteoric', '-', '-', normal ).
adj( 'meteorological', '-', '-', normal ).
adj( 'methodical', '-', '-', normal ).
adj( 'methodological', '-', '-', normal ).
adj( 'methylated', '-', '-', normal ).
adj( 'meticulous', '-', '-', normal ).
adj( 'metric', '-', '-', normal ).
adj( 'metrical', '-', '-', normal ).
adj( 'metropolitan', '-', '-', normal ).
adj( 'mettlesome', '-', '-', normal ).
adj( 'mezzanine', '-', '-', normal ).
adj( 'microscopic', '-', '-', normal ).
adj( 'microscopical', '-', '-', normal ).
adj( 'mid', '-', '-', normal ).
adj( 'middle-aged', '-', '-', normal ).
adj( 'middle-class', '-', '-', normal ).
adj( 'middle-distance', '-', '-', normal ).
adj( 'middle-of-the-road', '-', '-', normal ).
adj( 'middleweight', '-', '-', normal ).
adj( 'middling', '-', '-', normal ).
adj( 'midmost', '-', '-', normal ).
adj( 'midway', '-', '-', normal ).
adj( 'midweek', '-', '-', attr ).
adj( 'mighty', 'mightier', 'mightiest', normal ).
adj( 'migratory', '-', '-', normal ).
adj( 'milch', '-', '-', normal ).
adj( 'mild', 'milder', 'mildest', normal ).
adj( 'militant', '-', '-', normal ).
adj( 'militaristic', '-', '-', normal ).
adj( 'military', '-', '-', normal ).
adj( 'milk-white', '-', '-', normal ).
adj( 'milky', 'milkier', 'milkiest', normal ).
adj( 'millennial', '-', '-', normal ).
adj( 'million', '-', '-', normal ).
adj( 'millionth', '-', '-', normal ).
adj( 'mimetic', '-', '-', normal ).
adj( 'mimic', '-', '-', attr ).
adj( 'minatory', '-', '-', normal ).
adj( 'mincing', '-', '-', normal ).
adj( 'mind-bending', '-', '-', normal ).
adj( 'mind-blowing', '-', '-', normal ).
adj( 'mind-boggling', '-', '-', normal ).
adj( 'minded', '-', '-', pred ).
adj( 'minded', '-', '-', affix ).
adj( 'mindful', '-', '-', normal ).
adj( 'mindless', '-', '-', normal ).
adj( 'mine', '-', '-', normal ).
adj( 'mineral', '-', '-', normal ).
adj( 'mingy', 'mingier', 'mingiest', normal ).
adj( 'minimal', '-', '-', normal ).
adj( 'minimum', '-', '-', normal ).
adj( 'ministerial', '-', '-', normal ).
adj( 'ministrant', '-', '-', attr ).
adj( 'minor', '-', '-', normal ).
adj( 'minus', '-', '-', normal ).
adj( 'minuscule', '-', '-', normal ).
adj( 'minute', 'minuter', 'minutest', normal ).
adj( 'miraculous', '-', '-', normal ).
adj( 'mirthful', '-', '-', normal ).
adj( 'mirthless', '-', '-', normal ).
adj( 'miry', 'mirier', 'miriest', normal ).
adj( 'misanthropic', '-', '-', normal ).
adj( 'misbegotten', '-', '-', normal ).
adj( 'miscellaneous', '-', '-', normal ).
adj( 'mischievous', '-', '-', normal ).
adj( 'miserable', '-', '-', normal ).
adj( 'miserly', '-', '-', normal ).
adj( 'misguided', '-', '-', normal ).
adj( 'misshapen', '-', '-', normal ).
adj( 'missing', '-', '-', normal ).
adj( 'mistaken', '-', '-', normal ).
adj( 'mistrustful', '-', '-', normal ).
adj( 'misty', 'mistier', 'mistiest', normal ).
adj( 'mixed', '-', '-', normal ).
adj( 'mixed-up', '-', '-', normal ).
adj( 'mnemonic', '-', '-', normal ).
adj( 'moated', '-', '-', normal ).
adj( 'mobile', '-', '-', normal ).
adj( 'mock', '-', '-', attr ).
adj( 'mod', '-', '-', normal ).
adj( 'modal', '-', '-', normal ).
adj( 'moderate', '-', '-', normal ).
adj( 'modern', '-', '-', normal ).
adj( 'modernistic', '-', '-', normal ).
adj( 'modest', '-', '-', normal ).
adj( 'modish', '-', '-', normal ).
adj( 'modular', '-', '-', normal ).
adj( 'moist', '-', '-', normal ).
adj( 'molar', '-', '-', normal ).
adj( 'moldy', 'moldier', 'moldiest', normal ).
adj( 'molecular', '-', '-', normal ).
adj( 'momentary', '-', '-', normal ).
adj( 'momentous', '-', '-', normal ).
adj( 'monarchic', '-', '-', normal ).
adj( 'monastic', '-', '-', normal ).
adj( 'monaural', '-', '-', normal ).
adj( 'monetary', '-', '-', normal ).
adj( 'moneyed', '-', '-', normal ).
adj( 'moneyless', '-', '-', normal ).
adj( 'mongol', '-', '-', attr ).
adj( 'mongrel', '-', '-', attr ).
adj( 'monkish', '-', '-', normal ).
adj( 'mono', '-', '-', normal ).
adj( 'monochrome', '-', '-', normal ).
adj( 'monocled', '-', '-', normal ).
adj( 'monogamous', '-', '-', normal ).
adj( 'monolithic', '-', '-', normal ).
adj( 'monopolistic', '-', '-', normal ).
adj( 'monosyllabic', '-', '-', normal ).
adj( 'monotheistic', '-', '-', normal ).
adj( 'monotonous', '-', '-', normal ).
adj( 'monstrous', '-', '-', normal ).
adj( 'monthly', '-', '-', normal ).
adj( 'monumental', '-', '-', normal ).
adj( 'moody', 'moodier', 'moodiest', normal ).
adj( 'moonless', '-', '-', normal ).
adj( 'moonlit', '-', '-', normal ).
adj( 'moonstruck', '-', '-', normal ).
adj( 'moony', 'moonier', 'mooniest', normal ).
adj( 'moot', '-', '-', normal ).
adj( 'moral', '-', '-', normal ).
adj( 'moralistic', '-', '-', normal ).
adj( 'morbid', '-', '-', normal ).
adj( 'mordant', '-', '-', normal ).
adj( 'morganatic', '-', '-', normal ).
adj( 'moribund', '-', '-', normal ).
adj( 'moronic', '-', '-', normal ).
adj( 'morose', '-', '-', normal ).
adj( 'morphemic', '-', '-', normal ).
adj( 'morphological', '-', '-', normal ).
adj( 'mortal', '-', '-', normal ).
adj( 'mosaic', '-', '-', normal ).
adj( 'moss-grown', '-', '-', normal ).
adj( 'mossy', 'mossier', 'mossiest', normal ).
adj( 'moth-eaten', '-', '-', normal ).
adj( 'motherless', '-', '-', normal ).
adj( 'motherlike', '-', '-', normal ).
adj( 'motherly', '-', '-', normal ).
adj( 'mothproof', '-', '-', normal ).
adj( 'motionless', '-', '-', normal ).
adj( 'motive', '-', '-', normal ).
adj( 'motiveless', '-', '-', normal ).
adj( 'motley', '-', '-', normal ).
adj( 'motor-assisted', '-', '-', normal ).
adj( 'mouldy', 'mouldier', 'mouldiest', normal ).
adj( 'mountainous', '-', '-', normal ).
adj( 'mournful', '-', '-', normal ).
adj( 'mousy', 'mousier', 'mousiest', normal ).
adj( 'mouth-watering', '-', '-', normal ).
adj( 'movable', '-', '-', normal ).
adj( 'much', '-', '-', normal ).
adj( 'mucky', 'muckier', 'muckiest', normal ).
adj( 'mucous', '-', '-', normal ).
adj( 'muddle-headed', '-', '-', normal ).
adj( 'muddy', 'muddier', 'muddiest', normal ).
adj( 'muggy', 'muggier', 'muggiest', normal ).
adj( 'mulish', '-', '-', normal ).
adj( 'mullioned', '-', '-', normal ).
adj( 'multi-lingual', '-', '-', normal ).
adj( 'multifarious', '-', '-', normal ).
adj( 'multiform', '-', '-', normal ).
adj( 'multilateral', '-', '-', normal ).
adj( 'multiple', '-', '-', normal ).
adj( 'multiplex', '-', '-', normal ).
adj( 'multitudinous', '-', '-', normal ).
adj( 'mum', '-', '-', normal ).
adj( 'mundane', '-', '-', normal ).
adj( 'municipal', '-', '-', normal ).
adj( 'munificent', '-', '-', normal ).
adj( 'mural', '-', '-', normal ).
adj( 'murderous', '-', '-', normal ).
adj( 'murky', 'murkier', 'murkiest', normal ).
adj( 'muscle-bound', '-', '-', normal ).
adj( 'muscular', '-', '-', normal ).
adj( 'mushy', 'mushier', 'mushiest', normal ).
adj( 'musical', '-', '-', normal ).
adj( 'musky', 'muskier', 'muskiest', normal ).
adj( 'musty', 'mustier', 'mustiest', normal ).
adj( 'mutable', '-', '-', normal ).
adj( 'mute', '-', '-', normal ).
adj( 'mutinous', '-', '-', normal ).
adj( 'mutual', '-', '-', normal ).
adj( 'muzzy', 'muzzier', 'muzziest', normal ).
adj( 'my', '-', '-', normal ).
adj( 'myopic', '-', '-', normal ).
adj( 'myriad', '-', '-', attr ).
adj( 'mysterious', '-', '-', normal ).
adj( 'mystic', '-', '-', normal ).
adj( 'mystical', '-', '-', normal ).
adj( 'mythical', '-', '-', normal ).
adj( 'mythological', '-', '-', normal ).
adj( 'n_ee', '-', '-', normal ).
adj( 'na\"ive', '-', '-', normal ).
adj( 'naive', '-', '-', normal ).
adj( 'naked', '-', '-', normal ).
adj( 'namby-pamby', '-', '-', normal ).
adj( 'nameless', '-', '-', normal ).
adj( 'narcotic', '-', '-', normal ).
adj( 'narrow', 'narrower', 'narrowest', normal ).
adj( 'narrow-minded', '-', '-', normal ).
adj( 'nasal', '-', '-', normal ).
adj( 'nascent', '-', '-', normal ).
adj( 'nasty', 'nastier', 'nastiest', normal ).
adj( 'natal', '-', '-', normal ).
adj( 'national', '-', '-', normal ).
adj( 'nationalist', '-', '-', normal ).
adj( 'nationalistic', '-', '-', normal ).
adj( 'nationwide', '-', '-', normal ).
adj( 'native', '-', '-', normal ).
adj( 'natty', 'nattier', 'nattiest', normal ).
adj( 'natural', '-', '-', normal ).
adj( 'naturalistic', '-', '-', normal ).
adj( 'naughty', 'naughtier', 'naughtiest', normal ).
adj( 'nauseous', '-', '-', normal ).
adj( 'nautical', '-', '-', normal ).
adj( 'naval', '-', '-', normal ).
adj( 'navigable', '-', '-', normal ).
adj( 'near', 'nearer', 'nearest', normal ).
adj( 'near-sighted', '-', '-', normal ).
adj( 'nearby', '-', '-', normal ).
adj( 'neat', 'neater', 'neatest', normal ).
adj( 'nebular', '-', '-', normal ).
adj( 'nebulous', '-', '-', normal ).
adj( 'necessary', '-', '-', normal ).
adj( 'necessitous', '-', '-', normal ).
adj( 'needful', '-', '-', normal ).
adj( 'needless', '-', '-', normal ).
adj( 'needy', 'needier', 'neediest', normal ).
adj( 'nefarious', '-', '-', normal ).
adj( 'negative', '-', '-', normal ).
adj( 'neglectful', '-', '-', normal ).
adj( 'negligent', '-', '-', normal ).
adj( 'negligible', '-', '-', normal ).
adj( 'negotiable', '-', '-', normal ).
adj( 'neighbourly', '-', '-', normal ).
adj( 'neither', '-', '-', normal ).
adj( 'neolithic', '-', '-', normal ).
adj( 'nerve-racking', '-', '-', normal ).
adj( 'nerveless', '-', '-', normal ).
adj( 'nervous', '-', '-', normal ).
adj( 'nervy', '-', '-', normal ).
adj( 'nescient', '-', '-', normal ).
adj( 'net', '-', '-', normal ).
adj( 'nether', '-', '-', normal ).
adj( 'nethermost', '-', '-', normal ).
adj( 'nett', '-', '-', normal ).
adj( 'neural', '-', '-', normal ).
adj( 'neuralgic', '-', '-', normal ).
adj( 'neurasthenic', '-', '-', normal ).
adj( 'neurotic', '-', '-', normal ).
adj( 'neuter', '-', '-', normal ).
adj( 'neutral', '-', '-', normal ).
adj( 'new', 'newer', 'newest', normal ).
adj( 'newfangled', '-', '-', normal ).
adj( 'newsless', '-', '-', normal ).
adj( 'newsworthy', '-', '-', normal ).
adj( 'newsy', 'newsier', 'newsiest', normal ).
adj( 'next', '-', '-', normal ).
adj( 'nice', 'nicer', 'nicest', normal ).
adj( 'niffy', 'niffier', 'niffiest', normal ).
adj( 'nifty', 'niftier', 'niftiest', normal ).
adj( 'niggardly', '-', '-', normal ).
adj( 'niggling', '-', '-', normal ).
adj( 'nightlong', '-', '-', normal ).
adj( 'nightly', '-', '-', normal ).
adj( 'nightmarish', '-', '-', normal ).
adj( 'nihilistic', '-', '-', normal ).
adj( 'nimble', 'nimbler', 'nimblest', normal ).
adj( 'niminy-piminy', '-', '-', normal ).
adj( 'nine', '-', '-', normal ).
adj( 'ninefold', '-', '-', normal ).
adj( 'ninepenny', '-', '-', normal ).
adj( 'nineteen', '-', '-', normal ).
adj( 'nineteenth', '-', '-', normal ).
adj( 'ninetieth', '-', '-', normal ).
adj( 'ninety', '-', '-', normal ).
adj( 'ninth', '-', '-', normal ).
adj( 'nipping', '-', '-', normal ).
adj( 'nippy', 'nippier', 'nippiest', normal ).
adj( 'nisi', '-', '-', normal ).
adj( 'nitric', '-', '-', normal ).
adj( 'nitrous', '-', '-', normal ).
adj( 'nitwitted', '-', '-', normal ).
adj( 'no', '-', '-', normal ).
adj( 'no-go', '-', '-', attr ).
adj( 'noble', 'nobler', 'noblest', normal ).
adj( 'nocturnal', '-', '-', normal ).
adj( 'nodular', '-', '-', normal ).
adj( 'nodulated', '-', '-', normal ).
adj( 'noiseless', '-', '-', normal ).
adj( 'noisome', '-', '-', normal ).
adj( 'noisy', 'noisier', 'noisiest', normal ).
adj( 'nomadic', '-', '-', normal ).
adj( 'nominal', '-', '-', normal ).
adj( 'nominative', '-', '-', normal ).
adj( 'non compos mentis', '-', '-', normal ).
adj( 'non-u', '-', '-', normal ).
adj( 'non-contentious', '-', '-', normal ).
adj( 'non-skid', '-', '-', normal ).
adj( 'nonagenarian', '-', '-', normal ).
adj( 'nonchalant', '-', '-', normal ).
adj( 'noncommissioned', '-', '-', normal ).
adj( 'noncommittal', '-', '-', normal ).
adj( 'nondescript', '-', '-', normal ).
adj( 'nonflammable', '-', '-', normal ).
adj( 'nonmoral', '-', '-', normal ).
adj( 'nonpareil', '-', '-', normal ).
adj( 'nonresident', '-', '-', normal ).
adj( 'nonsensical', '-', '-', normal ).
adj( 'nonstick', '-', '-', normal ).
adj( 'nonstop', '-', '-', normal ).
adj( 'nonunion', '-', '-', normal ).
adj( 'noonday', '-', '-', normal ).
adj( 'normal', '-', '-', normal ).
adj( 'normative', '-', '-', normal ).
adj( 'northeasterly', '-', '-', normal ).
adj( 'northeastern', '-', '-', normal ).
adj( 'northerly', '-', '-', normal ).
adj( 'northern', '-', '-', normal ).
adj( 'northernmost', '-', '-', normal ).
adj( 'northwesterly', '-', '-', normal ).
adj( 'northwestern', '-', '-', normal ).
adj( 'nosed', '-', '-', affix ).
adj( 'nosey', '-', '-', normal ).
adj( 'nostalgic', '-', '-', normal ).
adj( 'nosy', 'nosier', 'nosiest', normal ).
adj( 'notable', '-', '-', normal ).
adj( 'noted', '-', '-', normal ).
adj( 'noteworthy', '-', '-', normal ).
adj( 'noticeable', '-', '-', normal ).
adj( 'notifiable', '-', '-', normal ).
adj( 'notional', '-', '-', normal ).
adj( 'notorious', '-', '-', normal ).
adj( 'novel', '-', '-', normal ).
adj( 'noxious', '-', '-', normal ).
adj( 'nubile', '-', '-', normal ).
adj( 'nuclear', '-', '-', normal ).
adj( 'nucleic', '-', '-', normal ).
adj( 'nude', '-', '-', normal ).
adj( 'nugatory', '-', '-', normal ).
adj( 'null', '-', '-', normal ).
adj( 'numb', '-', '-', normal ).
adj( 'numberless', '-', '-', normal ).
adj( 'numerable', '-', '-', normal ).
adj( 'numeral', '-', '-', normal ).
adj( 'numerate', '-', '-', normal ).
adj( 'numeric', '-', '-', normal ).
adj( 'numerical', '-', '-', normal ).
adj( 'numerous', '-', '-', normal ).
adj( 'numinous', '-', '-', normal ).
adj( 'nuptial', '-', '-', normal ).
adj( 'nut-brown', '-', '-', normal ).
adj( 'nutrient', '-', '-', normal ).
adj( 'nutritional', '-', '-', normal ).
adj( 'nutritious', '-', '-', normal ).
adj( 'nutritive', '-', '-', normal ).
adj( 'nuts', '-', '-', normal ).
adj( 'nutty', 'nuttier', 'nuttiest', normal ).
adj( 'nymphomaniac', '-', '-', normal ).
adj( 'oafish', '-', '-', normal ).
adj( 'oaken', '-', '-', normal ).
adj( 'obdurate', '-', '-', normal ).
adj( 'obedient', '-', '-', normal ).
adj( 'obese', '-', '-', normal ).
adj( 'objectionable', '-', '-', normal ).
adj( 'objective', '-', '-', normal ).
adj( 'oblate', '-', '-', normal ).
adj( 'obligatory', '-', '-', normal ).
adj( 'obliging', '-', '-', normal ).
adj( 'oblique', '-', '-', normal ).
adj( 'oblivious', '-', '-', normal ).
adj( 'oblong', '-', '-', normal ).
adj( 'obnoxious', '-', '-', normal ).
adj( 'obscene', '-', '-', normal ).
adj( 'obscure', '-', '-', normal ).
adj( 'obsequious', '-', '-', normal ).
adj( 'observable', '-', '-', normal ).
adj( 'observant', '-', '-', normal ).
adj( 'observing', '-', '-', normal ).
adj( 'obsessional', '-', '-', normal ).
adj( 'obsessive', '-', '-', normal ).
adj( 'obsolescent', '-', '-', normal ).
adj( 'obsolete', '-', '-', normal ).
adj( 'obstetric', '-', '-', normal ).
adj( 'obstetrical', '-', '-', normal ).
adj( 'obstinate', '-', '-', normal ).
adj( 'obstreperous', '-', '-', normal ).
adj( 'obstructive', '-', '-', normal ).
adj( 'obtainable', '-', '-', normal ).
adj( 'obtrusive', '-', '-', normal ).
adj( 'obtuse', '-', '-', normal ).
adj( 'obvious', '-', '-', normal ).
adj( 'occasional', '-', '-', normal ).
adj( 'occult', '-', '-', normal ).
adj( 'occupational', '-', '-', normal ).
adj( 'oceanic', '-', '-', normal ).
adj( 'octagonal', '-', '-', normal ).
adj( 'octogenarian', '-', '-', normal ).
adj( 'ocular', '-', '-', normal ).
adj( 'odd', 'odder', 'oddest', normal ).
adj( 'odd-job', '-', '-', attr ).
adj( 'odds-on', '-', '-', normal ).
adj( 'odious', '-', '-', normal ).
adj( 'odoriferous', '-', '-', normal ).
adj( 'odorous', '-', '-', normal ).
adj( 'odourless', '-', '-', normal ).
adj( 'oecumenical', '-', '-', normal ).
adj( 'off', '-', '-', normal ).
adj( 'off-peak', '-', '-', attr ).
adj( 'off-putting', '-', '-', normal ).
adj( 'off-street', '-', '-', attr ).
adj( 'off-white', '-', '-', normal ).
adj( 'offbeat', '-', '-', normal ).
adj( 'offenceless', '-', '-', normal ).
adj( 'offensive', '-', '-', normal ).
adj( 'offhand', '-', '-', normal ).
adj( 'offhanded', '-', '-', normal ).
adj( 'offhandedly', '-', '-', normal ).
adj( 'official', '-', '-', normal ).
adj( 'officious', '-', '-', normal ).
adj( 'offish', '-', '-', normal ).
adj( 'offshore', '-', '-', normal ).
adj( 'offside', '-', '-', attr ).
adj( 'offstage', '-', '-', normal ).
adj( 'ogreish', '-', '-', normal ).
adj( 'oil-bearing', '-', '-', normal ).
adj( 'oiled', '-', '-', normal ).
adj( 'oilfired', '-', '-', normal ).
adj( 'oily', 'oilier', 'oiliest', normal ).
adj( 'okay', '-', '-', normal ).
adj( 'old', 'older', 'oldest', normal ).
adj( 'old-fashioned', '-', '-', normal ).
adj( 'old-maidish', '-', '-', normal ).
adj( 'old-time', '-', '-', normal ).
adj( 'old-womanish', '-', '-', normal ).
adj( 'old-world', '-', '-', normal ).
adj( 'olden', '-', '-', normal ).
adj( 'oldish', '-', '-', normal ).
adj( 'oleaginous', '-', '-', normal ).
adj( 'olfactory', '-', '-', normal ).
adj( 'olive', '-', '-', normal ).
adj( 'ominous', '-', '-', normal ).
adj( 'omnipotent', '-', '-', normal ).
adj( 'omniscient', '-', '-', normal ).
adj( 'omnivorous', '-', '-', normal ).
adj( 'oncoming', '-', '-', normal ).
adj( 'one', '-', '-', normal ).
adj( 'one-armed', '-', '-', normal ).
adj( 'one-eyed', '-', '-', normal ).
adj( 'one-horse', '-', '-', normal ).
adj( 'one-idea\'d', '-', '-', normal ).
adj( 'one-sided', '-', '-', normal ).
adj( 'one-time', '-', '-', normal ).
adj( 'onerous', '-', '-', normal ).
adj( 'ongoing', '-', '-', normal ).
adj( 'only', '-', '-', normal ).
adj( 'onshore', '-', '-', normal ).
adj( 'onward', '-', '-', normal ).
adj( 'oozy', 'oozier', 'ooziest', normal ).
adj( 'opalescent', '-', '-', normal ).
adj( 'opaque', '-', '-', normal ).
adj( 'open', '-', '-', normal ).
adj( 'open-air', '-', '-', attr ).
adj( 'open-ended', '-', '-', normal ).
adj( 'open-eyed', '-', '-', normal ).
adj( 'open-handed', '-', '-', normal ).
adj( 'open-hearted', '-', '-', normal ).
adj( 'open-minded', '-', '-', normal ).
adj( 'open-mouthed', '-', '-', normal ).
adj( 'opencast', '-', '-', normal ).
adj( 'opening', '-', '-', normal ).
adj( 'operable', '-', '-', normal ).
adj( 'operatic', '-', '-', normal ).
adj( 'operational', '-', '-', normal ).
adj( 'operative', '-', '-', normal ).
adj( 'ophthalmic', '-', '-', normal ).
adj( 'opinionated', '-', '-', normal ).
adj( 'opinionative', '-', '-', normal ).
adj( 'opportune', '-', '-', normal ).
adj( 'opposite', '-', '-', normal ).
adj( 'oppressive', '-', '-', normal ).
adj( 'opprobrious', '-', '-', normal ).
adj( 'optative', '-', '-', normal ).
adj( 'optic', '-', '-', normal ).
adj( 'optical', '-', '-', normal ).
adj( 'optimal', '-', '-', normal ).
adj( 'optimistic', '-', '-', normal ).
adj( 'optional', '-', '-', normal ).
adj( 'opulent', '-', '-', normal ).
adj( 'oracular', '-', '-', normal ).
adj( 'oral', '-', '-', normal ).
adj( 'orange', '-', '-', normal ).
adj( 'oratorical', '-', '-', normal ).
adj( 'orbital', '-', '-', normal ).
adj( 'orchestral', '-', '-', normal ).
adj( 'orderly', '-', '-', normal ).
adj( 'ordinal', '-', '-', normal ).
adj( 'ordinary', '-', '-', normal ).
adj( 'organic', '-', '-', normal ).
adj( 'organizational', '-', '-', normal ).
adj( 'organized', '-', '-', normal ).
adj( 'orgiastic', '-', '-', normal ).
adj( 'orient', '-', '-', normal ).
adj( 'oriental', '-', '-', normal ).
adj( 'original', '-', '-', normal ).
adj( 'ornamental', '-', '-', normal ).
adj( 'ornate', '-', '-', normal ).
adj( 'ornery', '-', '-', normal ).
adj( 'ornithological', '-', '-', normal ).
adj( 'orotund', '-', '-', normal ).
adj( 'orthodontic', '-', '-', normal ).
adj( 'orthodox', '-', '-', normal ).
adj( 'orthogonal', '-', '-', normal ).
adj( 'orthographic', '-', '-', normal ).
adj( 'orthopaedic', '-', '-', normal ).
adj( 'orthopedic', '-', '-', normal ).
adj( 'osseous', '-', '-', normal ).
adj( 'ostensible', '-', '-', normal ).
adj( 'ostentatious', '-', '-', normal ).
adj( 'other', '-', '-', normal ).
adj( 'otherworldly', '-', '-', normal ).
adj( 'otiose', '-', '-', normal ).
adj( 'our', '-', '-', normal ).
adj( 'ours', '-', '-', pred ).
adj( 'out-of-date', '-', '-', normal ).
adj( 'out-of-door', '-', '-', attr ).
adj( 'out-of-the-way', '-', '-', normal ).
adj( 'out-of-work', '-', '-', normal ).
adj( 'outback', '-', '-', normal ).
adj( 'outboard', '-', '-', attr ).
adj( 'outbound', '-', '-', normal ).
adj( 'outcast', '-', '-', normal ).
adj( 'outcaste', '-', '-', normal ).
adj( 'outdated', '-', '-', normal ).
adj( 'outdoor', '-', '-', attr ).
adj( 'outer', '-', '-', normal ).
adj( 'outermost', '-', '-', normal ).
adj( 'outgoing', '-', '-', normal ).
adj( 'outlandish', '-', '-', normal ).
adj( 'outlying', '-', '-', normal ).
adj( 'outmoded', '-', '-', normal ).
adj( 'outmost', '-', '-', normal ).
adj( 'outr_e', '-', '-', normal ).
adj( 'outrageous', '-', '-', normal ).
adj( 'outrigged', '-', '-', normal ).
adj( 'outright', '-', '-', normal ).
adj( 'outside', '-', '-', normal ).
adj( 'outsize', '-', '-', normal ).
adj( 'outspoken', '-', '-', normal ).
adj( 'outspread', '-', '-', normal ).
adj( 'outstanding', '-', '-', normal ).
adj( 'outstretched', '-', '-', normal ).
adj( 'outward', '-', '-', normal ).
adj( 'oval', '-', '-', normal ).
adj( 'over-ripe', '-', '-', normal ).
adj( 'overabundant', '-', '-', normal ).
adj( 'overactive', '-', '-', normal ).
adj( 'overall', '-', '-', normal ).
adj( 'overambitious', '-', '-', normal ).
adj( 'overanxious', '-', '-', normal ).
adj( 'overarm', '-', '-', normal ).
adj( 'overbearing', '-', '-', normal ).
adj( 'overblown', '-', '-', normal ).
adj( 'overbold', '-', '-', normal ).
adj( 'overbusy', '-', '-', normal ).
adj( 'overcareful', '-', '-', normal ).
adj( 'overcast', '-', '-', normal ).
adj( 'overcautious', '-', '-', normal ).
adj( 'overconfident', '-', '-', normal ).
adj( 'overcredulous', '-', '-', normal ).
adj( 'overcritical', '-', '-', normal ).
adj( 'overcurious', '-', '-', normal ).
adj( 'overdelicate', '-', '-', normal ).
adj( 'overdue', '-', '-', normal ).
adj( 'overeager', '-', '-', normal ).
adj( 'overemotional', '-', '-', normal ).
adj( 'overenthusiastic', '-', '-', normal ).
adj( 'overexcited', '-', '-', normal ).
adj( 'overfamiliar', '-', '-', normal ).
adj( 'overfond', '-', '-', normal ).
adj( 'overfull', '-', '-', normal ).
adj( 'overgenerous', '-', '-', normal ).
adj( 'overgreedy', '-', '-', normal ).
adj( 'overgrown', '-', '-', normal ).
adj( 'overhand', '-', '-', normal ).
adj( 'overhasty', '-', '-', normal ).
adj( 'overhead', '-', '-', normal ).
adj( 'overjealous', '-', '-', normal ).
adj( 'overjoyed', '-', '-', normal ).
adj( 'overladen', '-', '-', normal ).
adj( 'overland', '-', '-', normal ).
adj( 'overlarge', '-', '-', normal ).
adj( 'overmodest', '-', '-', normal ).
adj( 'overmuch', '-', '-', normal ).
adj( 'overnervous', '-', '-', normal ).
adj( 'overnight', '-', '-', normal ).
adj( 'overpowering', '-', '-', normal ).
adj( 'overproud', '-', '-', normal ).
adj( 'overripe', '-', '-', normal ).
adj( 'oversea', '-', '-', normal ).
adj( 'overseas', '-', '-', normal ).
adj( 'oversensitive', '-', '-', normal ).
adj( 'overserious', '-', '-', normal ).
adj( 'oversewn', '-', '-', normal ).
adj( 'oversexed', '-', '-', normal ).
adj( 'overshot', '-', '-', normal ).
adj( 'oversize', '-', '-', normal ).
adj( 'overstrung', '-', '-', normal ).
adj( 'overstuffed', '-', '-', normal ).
adj( 'oversubscribed', '-', '-', normal ).
adj( 'oversuspicious', '-', '-', normal ).
adj( 'overt', '-', '-', normal ).
adj( 'overweening', '-', '-', normal ).
adj( 'overweight', '-', '-', normal ).
adj( 'overweighted', '-', '-', normal ).
adj( 'overwrought', '-', '-', normal ).
adj( 'overzealous', '-', '-', normal ).
adj( 'oviparous', '-', '-', normal ).
adj( 'ovoid', '-', '-', normal ).
adj( 'owing', '-', '-', normal ).
adj( 'owlish', '-', '-', normal ).
adj( 'own', '-', '-', normal ).
adj( 'owner-driven', '-', '-', normal ).
adj( 'owner-occupied', '-', '-', normal ).
adj( 'ownerless', '-', '-', normal ).
adj( 'oxeyed', '-', '-', normal ).
adj( 'oxyacetylene', '-', '-', normal ).
adj( 'pacific', '-', '-', normal ).
adj( 'pagan', '-', '-', normal ).
adj( 'pained', '-', '-', normal ).
adj( 'painful', '-', '-', normal ).
adj( 'painless', '-', '-', normal ).
adj( 'painstaking', '-', '-', normal ).
adj( 'palaeolithic', '-', '-', normal ).
adj( 'palatable', '-', '-', normal ).
adj( 'palatal', '-', '-', normal ).
adj( 'palatial', '-', '-', normal ).
adj( 'pale', 'paler', 'palest', normal ).
adj( 'paleolithic', '-', '-', normal ).
adj( 'palish', '-', '-', normal ).
adj( 'palliative', '-', '-', normal ).
adj( 'pallid', '-', '-', normal ).
adj( 'pally', 'pallier', 'palliest', normal ).
adj( 'palmy', 'palmier', 'palmiest', normal ).
adj( 'palpable', '-', '-', normal ).
adj( 'paltry', 'paltrier', 'paltriest', normal ).
adj( 'panchromatic', '-', '-', normal ).
adj( 'pancreatic', '-', '-', normal ).
adj( 'pandemic', '-', '-', normal ).
adj( 'panic-stricken', '-', '-', normal ).
adj( 'panicky', '-', '-', normal ).
adj( 'panoplied', '-', '-', normal ).
adj( 'panoptic', '-', '-', normal ).
adj( 'panoramic', '-', '-', normal ).
adj( 'pantheistic', '-', '-', normal ).
adj( 'panzer', '-', '-', attr ).
adj( 'papal', '-', '-', normal ).
adj( 'paperbacked', '-', '-', normal ).
adj( 'paperless', '-', '-', normal ).
adj( 'papist', '-', '-', normal ).
adj( 'parabolic', '-', '-', normal ).
adj( 'parabolical', '-', '-', normal ).
adj( 'paradisiac', '-', '-', normal ).
adj( 'paradisiacal', '-', '-', normal ).
adj( 'paradoxical', '-', '-', normal ).
adj( 'parallel', '-', '-', normal ).
adj( 'paralytic', '-', '-', normal ).
adj( 'paramilitary', '-', '-', normal ).
adj( 'paramount', '-', '-', normal ).
adj( 'paranoid', '-', '-', normal ).
adj( 'paraplegic', '-', '-', normal ).
adj( 'parasitic', '-', '-', normal ).
adj( 'parasitical', '-', '-', normal ).
adj( 'pardonable', '-', '-', normal ).
adj( 'parental', '-', '-', normal ).
adj( 'parenthetic', '-', '-', normal ).
adj( 'parenthetical', '-', '-', normal ).
adj( 'parky', '-', '-', normal ).
adj( 'parliamentary', '-', '-', normal ).
adj( 'parlous', '-', '-', normal ).
adj( 'parochial', '-', '-', normal ).
adj( 'parsimonious', '-', '-', normal ).
adj( 'part-time', '-', '-', normal ).
adj( 'parti-coloured', '-', '-', normal ).
adj( 'partial', '-', '-', normal ).
adj( 'participial', '-', '-', normal ).
adj( 'particular', '-', '-', normal ).
adj( 'partisan', '-', '-', normal ).
adj( 'partitive', '-', '-', normal ).
adj( 'party-spirited', '-', '-', normal ).
adj( 'paschal', '-', '-', normal ).
adj( 'pass_e', '-', '-', normal ).
adj( 'pass_ee', '-', '-', normal ).
adj( 'passable', '-', '-', normal ).
adj( 'passing', '-', '-', normal ).
adj( 'passing-out', '-', '-', normal ).
adj( 'passionate', '-', '-', normal ).
adj( 'passionless', '-', '-', normal ).
adj( 'passive', '-', '-', normal ).
adj( 'past', '-', '-', normal ).
adj( 'pastoral', '-', '-', normal ).
adj( 'pasty', 'pastier', 'pastiest', normal ).
adj( 'patchy', 'patchier', 'patchiest', normal ).
adj( 'patent', '-', '-', normal ).
adj( 'paternal', '-', '-', normal ).
adj( 'pathetic', '-', '-', normal ).
adj( 'pathless', '-', '-', normal ).
adj( 'pathological', '-', '-', normal ).
adj( 'patient', '-', '-', normal ).
adj( 'patriarchal', '-', '-', normal ).
adj( 'patrician', '-', '-', normal ).
adj( 'patrimonial', '-', '-', normal ).
adj( 'patriotic', '-', '-', normal ).
adj( 'patronizing', '-', '-', normal ).
adj( 'patronymic', '-', '-', normal ).
adj( 'paunchy', 'paunchier', 'paunchiest', normal ).
adj( 'pawky', '-', '-', normal ).
adj( 'payable', '-', '-', normal ).
adj( 'pea-green', '-', '-', normal ).
adj( 'peaceable', '-', '-', normal ).
adj( 'peaceful', '-', '-', normal ).
adj( 'peacock-blue', '-', '-', normal ).
adj( 'peaked', '-', '-', normal ).
adj( 'peaky', 'peakier', 'peakiest', normal ).
adj( 'pearly', 'pearlier', 'pearliest', normal ).
adj( 'peaty', 'peatier', 'peatiest', normal ).
adj( 'pebbly', 'pebblier', 'pebbliest', normal ).
adj( 'peccable', '-', '-', normal ).
adj( 'peckish', '-', '-', normal ).
adj( 'pectic', '-', '-', normal ).
adj( 'pectoral', '-', '-', normal ).
adj( 'peculiar', '-', '-', normal ).
adj( 'pecuniary', '-', '-', normal ).
adj( 'pedagogic', '-', '-', normal ).
adj( 'pedagogical', '-', '-', normal ).
adj( 'pedal', '-', '-', normal ).
adj( 'pedantic', '-', '-', normal ).
adj( 'peddling', '-', '-', normal ).
adj( 'pedestrian', '-', '-', normal ).
adj( 'peerless', '-', '-', normal ).
adj( 'peeved', '-', '-', normal ).
adj( 'peevish', '-', '-', normal ).
adj( 'pejorative', '-', '-', normal ).
adj( 'pellucid', '-', '-', normal ).
adj( 'pelvic', '-', '-', normal ).
adj( 'penal', '-', '-', normal ).
adj( 'pendent', '-', '-', normal ).
adj( 'pending', '-', '-', normal ).
adj( 'pendulous', '-', '-', normal ).
adj( 'penetrable', '-', '-', normal ).
adj( 'penetrating', '-', '-', normal ).
adj( 'penetrative', '-', '-', normal ).
adj( 'peninsular', '-', '-', normal ).
adj( 'penitent', '-', '-', normal ).
adj( 'penitential', '-', '-', normal ).
adj( 'penitentiary', '-', '-', normal ).
adj( 'penniless', '-', '-', normal ).
adj( 'pensionable', '-', '-', normal ).
adj( 'pensive', '-', '-', normal ).
adj( 'pent-up', '-', '-', normal ).
adj( 'pentagonal', '-', '-', normal ).
adj( 'penultimate', '-', '-', normal ).
adj( 'penurious', '-', '-', normal ).
adj( 'peppery', '-', '-', normal ).
adj( 'peptic', '-', '-', normal ).
adj( 'perceivable', '-', '-', normal ).
adj( 'perceptible', '-', '-', normal ).
adj( 'perceptive', '-', '-', normal ).
adj( 'perceptual', '-', '-', normal ).
adj( 'percipient', '-', '-', normal ).
adj( 'peremptory', '-', '-', normal ).
adj( 'perennial', '-', '-', normal ).
adj( 'perfect', '-', '-', normal ).
adj( 'perfectible', '-', '-', normal ).
adj( 'perfervid', '-', '-', normal ).
adj( 'perfidious', '-', '-', normal ).
adj( 'perfunctory', '-', '-', normal ).
adj( 'perilous', '-', '-', normal ).
adj( 'periodic', '-', '-', normal ).
adj( 'periodical', '-', '-', normal ).
adj( 'peripatetic', '-', '-', normal ).
adj( 'peripheral', '-', '-', normal ).
adj( 'periphrastic', '-', '-', normal ).
adj( 'perishable', '-', '-', normal ).
adj( 'perky', 'perkier', 'perkiest', normal ).
adj( 'permanent', '-', '-', normal ).
adj( 'permeable', '-', '-', normal ).
adj( 'permissible', '-', '-', normal ).
adj( 'permissive', '-', '-', normal ).
adj( 'pernicious', '-', '-', normal ).
adj( 'pernickety', '-', '-', normal ).
adj( 'perpendicular', '-', '-', normal ).
adj( 'perpetual', '-', '-', normal ).
adj( 'perplexed', '-', '-', normal ).
adj( 'persevering', '-', '-', normal ).
adj( 'persistent', '-', '-', normal ).
adj( 'personable', '-', '-', normal ).
adj( 'personal', '-', '-', normal ).
adj( 'perspicacious', '-', '-', normal ).
adj( 'perspicuous', '-', '-', normal ).
adj( 'persuadable', '-', '-', normal ).
adj( 'persuasive', '-', '-', normal ).
adj( 'pert', '-', '-', normal ).
adj( 'pertinacious', '-', '-', normal ).
adj( 'pertinent', '-', '-', normal ).
adj( 'pervasive', '-', '-', normal ).
adj( 'perverse', '-', '-', normal ).
adj( 'pesky', 'peskier', 'peskiest', normal ).
adj( 'pessimistic', '-', '-', normal ).
adj( 'pestiferous', '-', '-', normal ).
adj( 'pestilent', '-', '-', normal ).
adj( 'pestilential', '-', '-', normal ).
adj( 'petaled', '-', '-', normal ).
adj( 'petalled', '-', '-', normal ).
adj( 'petite', '-', '-', normal ).
adj( 'pettifogging', '-', '-', normal ).
adj( 'pettish', '-', '-', normal ).
adj( 'petty', 'pettier', 'pettiest', normal ).
adj( 'petulant', '-', '-', normal ).
adj( 'phallic', '-', '-', normal ).
adj( 'phantasmal', '-', '-', normal ).
adj( 'pharisaic', '-', '-', normal ).
adj( 'pharisaical', '-', '-', normal ).
adj( 'pharmaceutical', '-', '-', normal ).
adj( 'pharmacological', '-', '-', normal ).
adj( 'phenomenal', '-', '-', normal ).
adj( 'philanthropic', '-', '-', normal ).
adj( 'philatelic', '-', '-', normal ).
adj( 'philharmonic', '-', '-', normal ).
adj( 'philhellene', '-', '-', normal ).
adj( 'philhellenic', '-', '-', normal ).
adj( 'philological', '-', '-', normal ).
adj( 'philosophic', '-', '-', normal ).
adj( 'philosophical', '-', '-', normal ).
adj( 'phlegmatic', '-', '-', normal ).
adj( 'phonemic', '-', '-', normal ).
adj( 'phonetic', '-', '-', normal ).
adj( 'phoney', '-', '-', normal ).
adj( 'phonic', '-', '-', normal ).
adj( 'phonological', '-', '-', normal ).
adj( 'phony', '-', '-', normal ).
adj( 'phosphorescent', '-', '-', normal ).
adj( 'phosphoric', '-', '-', normal ).
adj( 'phosphorous', '-', '-', normal ).
adj( 'photoelectric', '-', '-', normal ).
adj( 'photogenic', '-', '-', normal ).
adj( 'photographic', '-', '-', normal ).
adj( 'phrasal', '-', '-', normal ).
adj( 'phrenetic', '-', '-', normal ).
adj( 'physical', '-', '-', normal ).
adj( 'physiological', '-', '-', normal ).
adj( 'pianissimo', '-', '-', normal ).
adj( 'piano', '-', '-', normal ).
adj( 'picaresque', '-', '-', normal ).
adj( 'picric', '-', '-', normal ).
adj( 'pictorial', '-', '-', normal ).
adj( 'picturesque', '-', '-', normal ).
adj( 'piddling', '-', '-', normal ).
adj( 'piebald', '-', '-', normal ).
adj( 'piecemeal', '-', '-', normal ).
adj( 'pied', '-', '-', normal ).
adj( 'piercing', '-', '-', normal ).
adj( 'piffling', '-', '-', normal ).
adj( 'pig-headed', '-', '-', normal ).
adj( 'pigeon-breasted', '-', '-', normal ).
adj( 'pigeon-toed', '-', '-', normal ).
adj( 'piggish', '-', '-', normal ).
adj( 'piggy', 'piggier', 'piggiest', normal ).
adj( 'pillared', '-', '-', normal ).
adj( 'pimpled', '-', '-', normal ).
adj( 'pimply', 'pimplier', 'pimpliest', normal ).
adj( 'pinchbeck', '-', '-', normal ).
adj( 'pineal', '-', '-', normal ).
adj( 'pink', 'pinker', 'pinkest', normal ).
adj( 'pinkish', '-', '-', normal ).
adj( 'pinnate', '-', '-', normal ).
adj( 'pinstripe', '-', '-', normal ).
adj( 'pious', '-', '-', normal ).
adj( 'piping', '-', '-', normal ).
adj( 'piquant', '-', '-', normal ).
adj( 'piratical', '-', '-', normal ).
adj( 'piscatorial', '-', '-', normal ).
adj( 'pissed', '-', '-', normal ).
adj( 'pitch-black', '-', '-', normal ).
adj( 'pitch-dark', '-', '-', normal ).
adj( 'piteous', '-', '-', normal ).
adj( 'pithy', 'pithier', 'pithiest', normal ).
adj( 'pitiable', '-', '-', normal ).
adj( 'pitiful', '-', '-', normal ).
adj( 'pitiless', '-', '-', normal ).
adj( 'pituitary', '-', '-', normal ).
adj( 'pitying', '-', '-', normal ).
adj( 'pivotal', '-', '-', normal ).
adj( 'pixilated', '-', '-', normal ).
adj( 'pizzicato', '-', '-', normal ).
adj( 'placid', '-', '-', normal ).
adj( 'plaguy', '-', '-', normal ).
adj( 'plain', 'plainer', 'plainest', normal ).
adj( 'plain-spoken', '-', '-', normal ).
adj( 'plaintive', '-', '-', normal ).
adj( 'planetary', '-', '-', normal ).
adj( 'plangent', '-', '-', normal ).
adj( 'planless', '-', '-', normal ).
adj( 'plastered', '-', '-', normal ).
adj( 'plastic', '-', '-', normal ).
adj( 'platitudinous', '-', '-', normal ).
adj( 'plausible', '-', '-', normal ).
adj( 'playful', '-', '-', normal ).
adj( 'pleasant', 'pleasanter', 'pleasantest', normal ).
adj( 'pleased', '-', '-', normal ).
adj( 'pleasing', '-', '-', normal ).
adj( 'pleasurable', '-', '-', normal ).
adj( 'plebeian', '-', '-', normal ).
adj( 'plenary', '-', '-', normal ).
adj( 'plenteous', '-', '-', normal ).
adj( 'plentiful', '-', '-', normal ).
adj( 'pliable', '-', '-', normal ).
adj( 'pliant', '-', '-', normal ).
adj( 'plodding', '-', '-', normal ).
adj( 'plosive', '-', '-', normal ).
adj( 'plucky', 'pluckier', 'pluckiest', normal ).
adj( 'plummy', 'plummier', 'plummiest', normal ).
adj( 'plump', 'plumper', 'plumpest', normal ).
adj( 'pluperfect', '-', '-', normal ).
adj( 'plural', '-', '-', normal ).
adj( 'plus', '-', '-', normal ).
adj( 'plush', 'plusher', 'plushest', normal ).
adj( 'plushy', 'plushier', 'plushiest', normal ).
adj( 'plutocratic', '-', '-', normal ).
adj( 'pneumatic', '-', '-', normal ).
adj( 'pocked', '-', '-', normal ).
adj( 'pockmarked', '-', '-', normal ).
adj( 'podgy', 'podgier', 'podgiest', normal ).
adj( 'poetic', '-', '-', normal ).
adj( 'poetical', '-', '-', normal ).
adj( 'poignant', '-', '-', normal ).
adj( 'point-blank', '-', '-', normal ).
adj( 'pointed', '-', '-', normal ).
adj( 'pointless', '-', '-', normal ).
adj( 'poisonous', '-', '-', normal ).
adj( 'poky', 'pokier', 'pokiest', normal ).
adj( 'polar', '-', '-', normal ).
adj( 'polemic', '-', '-', normal ).
adj( 'polite', 'politer', 'politest', normal ).
adj( 'politic', '-', '-', normal ).
adj( 'political', '-', '-', normal ).
adj( 'polo-neck', '-', '-', normal ).
adj( 'polyandrous', '-', '-', normal ).
adj( 'polygamous', '-', '-', normal ).
adj( 'polyglot', '-', '-', normal ).
adj( 'polygonal', '-', '-', normal ).
adj( 'polymorphic', '-', '-', normal ).
adj( 'polymorphous', '-', '-', normal ).
adj( 'polynomial', '-', '-', normal ).
adj( 'polyphonic', '-', '-', normal ).
adj( 'polysyllabic', '-', '-', normal ).
adj( 'polytheistic', '-', '-', normal ).
adj( 'pompous', '-', '-', normal ).
adj( 'ponderable', '-', '-', normal ).
adj( 'ponderous', '-', '-', normal ).
adj( 'pontifical', '-', '-', normal ).
adj( 'poor', 'poorer', 'poorest', normal ).
adj( 'poor-spirited', '-', '-', normal ).
adj( 'poorly', '-', '-', pred ).
adj( 'pop', '-', '-', normal ).
adj( 'popeyed', '-', '-', normal ).
adj( 'popish', '-', '-', normal ).
adj( 'popular', '-', '-', normal ).
adj( 'populous', '-', '-', normal ).
adj( 'porcine', '-', '-', normal ).
adj( 'pornographic', '-', '-', normal ).
adj( 'porous', '-', '-', normal ).
adj( 'portable', '-', '-', normal ).
adj( 'portentous', '-', '-', normal ).
adj( 'portly', '-', '-', normal ).
adj( 'posh', 'posher', 'poshest', normal ).
adj( 'positive', '-', '-', normal ).
adj( 'possessive', '-', '-', normal ).
adj( 'possible', '-', '-', normal ).
adj( 'post-free', '-', '-', normal ).
adj( 'post-mortem', '-', '-', normal ).
adj( 'post-paid', '-', '-', normal ).
adj( 'postal', '-', '-', normal ).
adj( 'posterior', '-', '-', normal ).
adj( 'postgraduate', '-', '-', normal ).
adj( 'posthumous', '-', '-', normal ).
adj( 'postprandial', '-', '-', normal ).
adj( 'postural', '-', '-', normal ).
adj( 'pot-trained', '-', '-', normal ).
adj( 'potable', '-', '-', normal ).
adj( 'potbellied', '-', '-', normal ).
adj( 'potbound', '-', '-', normal ).
adj( 'potent', '-', '-', normal ).
adj( 'potential', '-', '-', normal ).
adj( 'potted', '-', '-', normal ).
adj( 'potty', 'pottier', 'pottiest', normal ).
adj( 'poverty-stricken', '-', '-', normal ).
adj( 'powdered', '-', '-', normal ).
adj( 'powdery', '-', '-', normal ).
adj( 'powered', '-', '-', normal ).
adj( 'powerful', '-', '-', normal ).
adj( 'powerless', '-', '-', normal ).
adj( 'practicable', '-', '-', normal ).
adj( 'practical', '-', '-', normal ).
adj( 'practised', '-', '-', normal ).
adj( 'praetorian', '-', '-', normal ).
adj( 'pragmatic', '-', '-', normal ).
adj( 'praiseworthy', '-', '-', normal ).
adj( 'pre-eminent', '-', '-', normal ).
adj( 'pre-emptive', '-', '-', normal ).
adj( 'pre-existent', '-', '-', normal ).
adj( 'pre-packaged', '-', '-', normal ).
adj( 'precarious', '-', '-', normal ).
adj( 'precast', '-', '-', normal ).
adj( 'precautionary', '-', '-', normal ).
adj( 'precedented', '-', '-', normal ).
adj( 'preceding', '-', '-', normal ).
adj( 'precious', '-', '-', normal ).
adj( 'precipitate', '-', '-', normal ).
adj( 'precipitous', '-', '-', normal ).
adj( 'precise', '-', '-', normal ).
adj( 'precocious', '-', '-', normal ).
adj( 'preconcerted', '-', '-', normal ).
adj( 'preconditioned', '-', '-', normal ).
adj( 'precursory', '-', '-', normal ).
adj( 'predatory', '-', '-', normal ).
adj( 'predestinate', '-', '-', normal ).
adj( 'predicative', '-', '-', normal ).
adj( 'predictable', '-', '-', normal ).
adj( 'predominant', '-', '-', normal ).
adj( 'prefatory', '-', '-', normal ).
adj( 'prefectural', '-', '-', normal ).
adj( 'preferable', '-', '-', normal ).
adj( 'preferential', '-', '-', normal ).
adj( 'pregnant', '-', '-', normal ).
adj( 'prehensile', '-', '-', normal ).
adj( 'prehistoric', '-', '-', normal ).
adj( 'prehistorical', '-', '-', normal ).
adj( 'prejudicial', '-', '-', normal ).
adj( 'preliminary', '-', '-', normal ).
adj( 'premarital', '-', '-', normal ).
adj( 'premature', '-', '-', normal ).
adj( 'premier', '-', '-', normal ).
adj( 'premonitory', '-', '-', normal ).
adj( 'prenatal', '-', '-', normal ).
adj( 'prepacked', '-', '-', normal ).
adj( 'preparatory', '-', '-', normal ).
adj( 'preponderant', '-', '-', normal ).
adj( 'prepositional', '-', '-', normal ).
adj( 'prepossessing', '-', '-', normal ).
adj( 'preposterous', '-', '-', normal ).
adj( 'prerequisite', '-', '-', normal ).
adj( 'prescient', '-', '-', normal ).
adj( 'prescriptive', '-', '-', normal ).
adj( 'present', '-', '-', normal ).
adj( 'present-day', '-', '-', attr ).
adj( 'presentable', '-', '-', normal ).
adj( 'preservable', '-', '-', normal ).
adj( 'preservative', '-', '-', normal ).
adj( 'presidential', '-', '-', normal ).
adj( 'pressing', '-', '-', normal ).
adj( 'pressurized', '-', '-', normal ).
adj( 'prestigious', '-', '-', normal ).
adj( 'prestissimo', '-', '-', normal ).
adj( 'presto', '-', '-', normal ).
adj( 'prestressed', '-', '-', normal ).
adj( 'presumable', '-', '-', normal ).
adj( 'presuming', '-', '-', normal ).
adj( 'presumptive', '-', '-', normal ).
adj( 'presumptuous', '-', '-', normal ).
adj( 'pretentious', '-', '-', normal ).
adj( 'preterit', '-', '-', normal ).
adj( 'preterite', '-', '-', normal ).
adj( 'preternatural', '-', '-', normal ).
adj( 'pretty', 'prettier', 'prettiest', normal ).
adj( 'pretty-pretty', '-', '-', normal ).
adj( 'prevailing', '-', '-', normal ).
adj( 'prevalent', '-', '-', normal ).
adj( 'preventable', '-', '-', normal ).
adj( 'preventive', '-', '-', normal ).
adj( 'previous', '-', '-', normal ).
adj( 'price-controlled', '-', '-', normal ).
adj( 'priceless', '-', '-', normal ).
adj( 'pricey', 'pricier', 'priciest', normal ).
adj( 'prickly', 'pricklier', 'prickliest', normal ).
adj( 'priest-ridden', '-', '-', normal ).
adj( 'priestlike', '-', '-', normal ).
adj( 'priestly', 'priestlier', 'priestliest', normal ).
adj( 'priggish', '-', '-', normal ).
adj( 'prim', 'primmer', 'primmest', normal ).
adj( 'prima', '-', '-', normal ).
adj( 'prima facie', '-', '-', normal ).
adj( 'primaeval', '-', '-', normal ).
adj( 'primal', '-', '-', normal ).
adj( 'primary', '-', '-', normal ).
adj( 'prime', '-', '-', normal ).
adj( 'primeval', '-', '-', normal ).
adj( 'primitive', '-', '-', normal ).
adj( 'primordial', '-', '-', normal ).
adj( 'princely', 'princelier', 'princeliest', normal ).
adj( 'principal', '-', '-', normal ).
adj( 'principled', '-', '-', normal ).
adj( 'printable', '-', '-', normal ).
adj( 'prior', '-', '-', normal ).
adj( 'prismatic', '-', '-', normal ).
adj( 'pristine', '-', '-', normal ).
adj( 'private', '-', '-', normal ).
adj( 'privileged', '-', '-', normal ).
adj( 'privy', '-', '-', normal ).
adj( 'pro forma', '-', '-', normal ).
adj( 'probabilistic', '-', '-', normal ).
adj( 'probable', '-', '-', normal ).
adj( 'probationary', '-', '-', normal ).
adj( 'problematic', '-', '-', normal ).
adj( 'procedural', '-', '-', normal ).
adj( 'processional', '-', '-', normal ).
adj( 'proconsular', '-', '-', normal ).
adj( 'procurable', '-', '-', normal ).
adj( 'prodigal', '-', '-', normal ).
adj( 'prodigious', '-', '-', normal ).
adj( 'productive', '-', '-', normal ).
adj( 'profane', '-', '-', normal ).
adj( 'professed', '-', '-', normal ).
adj( 'professional', '-', '-', normal ).
adj( 'professorial', '-', '-', normal ).
adj( 'proficient', '-', '-', normal ).
adj( 'profitable', '-', '-', normal ).
adj( 'profitless', '-', '-', normal ).
adj( 'profligate', '-', '-', normal ).
adj( 'profound', '-', '-', normal ).
adj( 'profuse', '-', '-', normal ).
adj( 'prognostic', '-', '-', normal ).
adj( 'progressive', '-', '-', normal ).
adj( 'prohibitive', '-', '-', normal ).
adj( 'prohibitory', '-', '-', normal ).
adj( 'projectile', '-', '-', normal ).
adj( 'proletarian', '-', '-', normal ).
adj( 'prolific', '-', '-', normal ).
adj( 'prolix', '-', '-', normal ).
adj( 'prolonged', '-', '-', normal ).
adj( 'prominent', '-', '-', normal ).
adj( 'promiscuous', '-', '-', normal ).
adj( 'promising', '-', '-', normal ).
adj( 'promissory', '-', '-', normal ).
adj( 'promotional', '-', '-', normal ).
adj( 'prompt', '-', '-', normal ).
adj( 'prone', '-', '-', normal ).
adj( 'pronged', '-', '-', normal ).
adj( 'pronominal', '-', '-', normal ).
adj( 'pronounceable', '-', '-', normal ).
adj( 'pronounced', '-', '-', normal ).
adj( 'proof', '-', '-', normal ).
adj( 'propellant', '-', '-', normal ).
adj( 'propellent', '-', '-', normal ).
adj( 'proper', '-', '-', normal ).
adj( 'propertied', '-', '-', normal ).
adj( 'prophetic', '-', '-', normal ).
adj( 'prophetical', '-', '-', normal ).
adj( 'prophylactic', '-', '-', normal ).
adj( 'propitiatory', '-', '-', normal ).
adj( 'propitious', '-', '-', normal ).
adj( 'proportionable', '-', '-', normal ).
adj( 'proportional', '-', '-', normal ).
adj( 'proportionate', '-', '-', normal ).
adj( 'proprietary', '-', '-', normal ).
adj( 'propulsive', '-', '-', normal ).
adj( 'prosaic', '-', '-', normal ).
adj( 'prospective', '-', '-', normal ).
adj( 'prosperous', '-', '-', normal ).
adj( 'prostrate', '-', '-', normal ).
adj( 'prosy', 'prosier', 'prosiest', normal ).
adj( 'protean', '-', '-', normal ).
adj( 'protective', '-', '-', normal ).
adj( 'protrusive', '-', '-', normal ).
adj( 'protuberant', '-', '-', normal ).
adj( 'proud', 'prouder', 'proudest', normal ).
adj( 'provable', '-', '-', normal ).
adj( 'proverbial', '-', '-', normal ).
adj( 'provident', '-', '-', normal ).
adj( 'providential', '-', '-', normal ).
adj( 'provincial', '-', '-', normal ).
adj( 'provisional', '-', '-', normal ).
adj( 'provisory', '-', '-', normal ).
adj( 'provocative', '-', '-', normal ).
adj( 'provoking', '-', '-', normal ).
adj( 'proximate', '-', '-', normal ).
adj( 'proximo', '-', '-', normal ).
adj( 'prudent', '-', '-', normal ).
adj( 'prudential', '-', '-', normal ).
adj( 'prudish', '-', '-', normal ).
adj( 'prurient', '-', '-', normal ).
adj( 'prussic', '-', '-', normal ).
adj( 'pseudo', '-', '-', normal ).
adj( 'pseudonymous', '-', '-', normal ).
adj( 'psychedelic', '-', '-', normal ).
adj( 'psychiatric', '-', '-', normal ).
adj( 'psychic', '-', '-', normal ).
adj( 'psychical', '-', '-', normal ).
adj( 'psychoanalytic', '-', '-', normal ).
adj( 'psychoanalytical', '-', '-', normal ).
adj( 'psychological', '-', '-', normal ).
adj( 'psychopathic', '-', '-', normal ).
adj( 'psychosomatic', '-', '-', normal ).
adj( 'pubic', '-', '-', normal ).
adj( 'public', '-', '-', normal ).
adj( 'public-spirited', '-', '-', normal ).
adj( 'puckish', '-', '-', normal ).
adj( 'pudgy', 'pudgier', 'pudgiest', normal ).
adj( 'puerile', '-', '-', normal ).
adj( 'puerperal', '-', '-', normal ).
adj( 'puffy', 'puffier', 'puffiest', normal ).
adj( 'pug-nose', '-', '-', normal ).
adj( 'pug-nosed', '-', '-', normal ).
adj( 'pugilistic', '-', '-', normal ).
adj( 'pugnacious', '-', '-', normal ).
adj( 'puissant', '-', '-', normal ).
adj( 'pukka', '-', '-', normal ).
adj( 'pulchritudinous', '-', '-', normal ).
adj( 'pulmonary', '-', '-', normal ).
adj( 'pulpy', 'pulpier', 'pulpiest', normal ).
adj( 'punch-drunk', '-', '-', normal ).
adj( 'punctilious', '-', '-', normal ).
adj( 'punctual', '-', '-', normal ).
adj( 'pungent', '-', '-', normal ).
adj( 'punishable', '-', '-', normal ).
adj( 'punitive', '-', '-', normal ).
adj( 'puny', 'punier', 'puniest', normal ).
adj( 'purblind', '-', '-', normal ).
adj( 'purchasable', '-', '-', normal ).
adj( 'pure', 'purer', 'purest', normal ).
adj( 'purgative', '-', '-', normal ).
adj( 'purgatorial', '-', '-', normal ).
adj( 'puritan', '-', '-', normal ).
adj( 'puritanical', '-', '-', normal ).
adj( 'purple', '-', '-', normal ).
adj( 'purplish', '-', '-', normal ).
adj( 'purpose-built', '-', '-', normal ).
adj( 'purposeful', '-', '-', normal ).
adj( 'purposeless', '-', '-', normal ).
adj( 'purposive', '-', '-', normal ).
adj( 'purse-proud', '-', '-', normal ).
adj( 'pursuant', '-', '-', normal ).
adj( 'pursy', '-', '-', normal ).
adj( 'purulent', '-', '-', normal ).
adj( 'pushful', '-', '-', normal ).
adj( 'pushing', '-', '-', normal ).
adj( 'pusillanimous', '-', '-', normal ).
adj( 'putative', '-', '-', normal ).
adj( 'putrescent', '-', '-', normal ).
adj( 'putrid', '-', '-', normal ).
adj( 'pyjama', '-', '-', attr ).
adj( 'pyrotechnic', '-', '-', normal ).
adj( 'quadrangular', '-', '-', normal ).
adj( 'quadratic', '-', '-', normal ).
adj( 'quadrilateral', '-', '-', normal ).
adj( 'quadrophonic', '-', '-', normal ).
adj( 'quadruple', '-', '-', normal ).
adj( 'quadruplicate', '-', '-', normal ).
adj( 'quaint', 'quainter', 'quaintest', normal ).
adj( 'qualified', '-', '-', normal ).
adj( 'qualitative', '-', '-', normal ).
adj( 'quantitative', '-', '-', normal ).
adj( 'quarrelsome', '-', '-', normal ).
adj( 'quarterly', '-', '-', normal ).
adj( 'queasy', 'queasier', 'queasiest', normal ).
adj( 'queenly', 'queenlier', 'queenliest', normal ).
adj( 'queer', 'queerer', 'queerest', normal ).
adj( 'quenchless', '-', '-', normal ).
adj( 'querulous', '-', '-', normal ).
adj( 'questionable', '-', '-', normal ).
adj( 'quibbling', '-', '-', normal ).
adj( 'quick', 'quicker', 'quickest', normal ).
adj( 'quick-change', '-', '-', attr ).
adj( 'quick-eared', '-', '-', normal ).
adj( 'quick-eyed', '-', '-', normal ).
adj( 'quick-sighted', '-', '-', normal ).
adj( 'quick-tempered', '-', '-', normal ).
adj( 'quick-witted', '-', '-', normal ).
adj( 'quickset', '-', '-', normal ).
adj( 'quiescent', '-', '-', normal ).
adj( 'quiet', 'quieter', 'quietest', normal ).
adj( 'quincentenary', '-', '-', normal ).
adj( 'quit', '-', '-', pred ).
adj( 'quits', '-', '-', pred ).
adj( 'quixotic', '-', '-', normal ).
adj( 'quizzical', '-', '-', normal ).
adj( 'quotable', '-', '-', normal ).
adj( 'quotidian', '-', '-', normal ).
adj( 'rabbinical', '-', '-', normal ).
adj( 'rabble-rousing', '-', '-', normal ).
adj( 'rabid', '-', '-', normal ).
adj( 'racial', '-', '-', normal ).
adj( 'racy', 'racier', 'raciest', normal ).
adj( 'raddled', '-', '-', normal ).
adj( 'radial', '-', '-', normal ).
adj( 'radiant', '-', '-', normal ).
adj( 'radical', '-', '-', normal ).
adj( 'radioactive', '-', '-', normal ).
adj( 'raffish', '-', '-', normal ).
adj( 'raftered', '-', '-', normal ).
adj( 'ragged', '-', '-', normal ).
adj( 'rainproof', '-', '-', normal ).
adj( 'rainy', 'rainier', 'rainiest', normal ).
adj( 'rakish', '-', '-', normal ).
adj( 'rallentando', '-', '-', normal ).
adj( 'rambling', '-', '-', normal ).
adj( 'rambunctious', '-', '-', normal ).
adj( 'rampageous', '-', '-', normal ).
adj( 'rampant', '-', '-', normal ).
adj( 'ramshackle', '-', '-', normal ).
adj( 'rancid', '-', '-', normal ).
adj( 'rancorous', '-', '-', normal ).
adj( 'randy', 'randier', 'randiest', normal ).
adj( 'rank', '-', '-', normal ).
adj( 'rapacious', '-', '-', normal ).
adj( 'rapid', '-', '-', normal ).
adj( 'rapt', '-', '-', normal ).
adj( 'rapturous', '-', '-', normal ).
adj( 'rare', 'rarer', 'rarest', normal ).
adj( 'raring', '-', '-', normal ).
adj( 'rascally', '-', '-', normal ).
adj( 'rash', 'rasher', 'rashest', normal ).
adj( 'ratable', '-', '-', normal ).
adj( 'rateable', '-', '-', normal ).
adj( 'rational', '-', '-', normal ).
adj( 'rationalistic', '-', '-', normal ).
adj( 'rattlebrained', '-', '-', normal ).
adj( 'rattlepated', '-', '-', normal ).
adj( 'rattling', '-', '-', normal ).
adj( 'ratty', 'rattier', 'rattiest', normal ).
adj( 'raucous', '-', '-', normal ).
adj( 'ravening', '-', '-', normal ).
adj( 'ravenous', '-', '-', normal ).
adj( 'raving', '-', '-', normal ).
adj( 'raw', '-', '-', normal ).
adj( 'rawboned', '-', '-', normal ).
adj( 'rawhide', '-', '-', normal ).
adj( 'razorbacked', '-', '-', normal ).
adj( 'readable', '-', '-', normal ).
adj( 'ready', 'readier', 'readiest', normal ).
adj( 'ready-made', '-', '-', normal ).
adj( 'real', '-', '-', normal ).
adj( 'realistic', '-', '-', normal ).
adj( 'realizable', '-', '-', normal ).
adj( 'rearmost', '-', '-', normal ).
adj( 'reasonable', '-', '-', normal ).
adj( 'reasonless', '-', '-', normal ).
adj( 'reassuring', '-', '-', normal ).
adj( 'rebarbative', '-', '-', normal ).
adj( 'rebellious', '-', '-', normal ).
adj( 'reborn', '-', '-', normal ).
adj( 'recalcitrant', '-', '-', normal ).
adj( 'receivable', '-', '-', normal ).
adj( 'received', '-', '-', normal ).
adj( 'recent', '-', '-', normal ).
adj( 'receptive', '-', '-', normal ).
adj( 'recessional', '-', '-', normal ).
adj( 'recessive', '-', '-', normal ).
adj( 'recherch_e', '-', '-', normal ).
adj( 'reciprocal', '-', '-', normal ).
adj( 'reckless', '-', '-', normal ).
adj( 'recognizable', '-', '-', normal ).
adj( 'reconcilable', '-', '-', normal ).
adj( 'recondite', '-', '-', normal ).
adj( 'record-breaking', '-', '-', normal ).
adj( 'recoverable', '-', '-', normal ).
adj( 'recreant', '-', '-', normal ).
adj( 'recreational', '-', '-', normal ).
adj( 'recriminatory', '-', '-', normal ).
adj( 'rectal', '-', '-', normal ).
adj( 'rectangular', '-', '-', normal ).
adj( 'rectilinear', '-', '-', normal ).
adj( 'recumbent', '-', '-', normal ).
adj( 'recuperative', '-', '-', normal ).
adj( 'recurrent', '-', '-', normal ).
adj( 'recusant', '-', '-', normal ).
adj( 'red', 'redder', 'reddest', normal ).
adj( 'red-hot', '-', '-', normal ).
adj( 'red-rimmed', '-', '-', normal ).
adj( 'reddish', '-', '-', normal ).
adj( 'redeemable', '-', '-', normal ).
adj( 'redemptive', '-', '-', normal ).
adj( 'redolent', '-', '-', normal ).
adj( 'redoubtable', '-', '-', normal ).
adj( 'reducible', '-', '-', normal ).
adj( 'redundant', '-', '-', normal ).
adj( 'reedy', 'reedier', 'reediest', normal ).
adj( 'referable', '-', '-', normal ).
adj( 'referential', '-', '-', normal ).
adj( 'reflective', '-', '-', normal ).
adj( 'reflex', '-', '-', normal ).
adj( 'reflexive', '-', '-', normal ).
adj( 'reformatory', '-', '-', normal ).
adj( 'refractory', '-', '-', normal ).
adj( 'refreshing', '-', '-', normal ).
adj( 'refrigerant', '-', '-', normal ).
adj( 'refulgent', '-', '-', normal ).
adj( 'refutable', '-', '-', normal ).
adj( 'regal', '-', '-', normal ).
adj( 'regardful', '-', '-', normal ).
adj( 'regardless', '-', '-', normal ).
adj( 'regenerate', '-', '-', normal ).
adj( 'regent', '-', '-', normal ).
adj( 'regimental', '-', '-', normal ).
adj( 'regional', '-', '-', normal ).
adj( 'regnant', '-', '-', normal ).
adj( 'regressive', '-', '-', normal ).
adj( 'regretful', '-', '-', normal ).
adj( 'regrettable', '-', '-', normal ).
adj( 'regular', '-', '-', normal ).
adj( 'reincarnate', '-', '-', normal ).
adj( 'relative', '-', '-', normal ).
adj( 'relentless', '-', '-', normal ).
adj( 'relevant', '-', '-', normal ).
adj( 'reliable', '-', '-', normal ).
adj( 'reliant', '-', '-', normal ).
adj( 'religious', '-', '-', normal ).
adj( 'reluctant', '-', '-', normal ).
adj( 'remarkable', '-', '-', normal ).
adj( 'remediable', '-', '-', normal ).
adj( 'remedial', '-', '-', normal ).
adj( 'reminiscent', '-', '-', normal ).
adj( 'remiss', '-', '-', normal ).
adj( 'remittent', '-', '-', normal ).
adj( 'remorseful', '-', '-', normal ).
adj( 'remorseless', '-', '-', normal ).
adj( 'remote', 'remoter', 'remotest', normal ).
adj( 'removable', '-', '-', normal ).
adj( 'removed', '-', '-', normal ).
adj( 'remunerative', '-', '-', normal ).
adj( 'renal', '-', '-', normal ).
adj( 'renascent', '-', '-', normal ).
adj( 'renewable', '-', '-', normal ).
adj( 'renowned', '-', '-', normal ).
adj( 'rent-free', '-', '-', normal ).
adj( 'rentable', '-', '-', normal ).
adj( 'repairable', '-', '-', normal ).
adj( 'reparable', '-', '-', normal ).
adj( 'repayable', '-', '-', normal ).
adj( 'repeatable', '-', '-', normal ).
adj( 'repeated', '-', '-', normal ).
adj( 'repellent', '-', '-', normal ).
adj( 'repentant', '-', '-', normal ).
adj( 'repetitious', '-', '-', normal ).
adj( 'repetitive', '-', '-', normal ).
adj( 'replaceable', '-', '-', normal ).
adj( 'replete', '-', '-', normal ).
adj( 'reply-paid', '-', '-', normal ).
adj( 'reportable', '-', '-', normal ).
adj( 'reposeful', '-', '-', normal ).
adj( 'reprehensible', '-', '-', normal ).
adj( 'representative', '-', '-', normal ).
adj( 'repressed', '-', '-', normal ).
adj( 'repressive', '-', '-', normal ).
adj( 'reproachful', '-', '-', normal ).
adj( 'reproducible', '-', '-', normal ).
adj( 'reproductive', '-', '-', normal ).
adj( 'reptilian', '-', '-', normal ).
adj( 'republican', '-', '-', normal ).
adj( 'repugnant', '-', '-', normal ).
adj( 'repulsive', '-', '-', normal ).
adj( 'reputable', '-', '-', normal ).
adj( 'reputed', '-', '-', attr ).
adj( 'requisite', '-', '-', normal ).
adj( 'resentful', '-', '-', normal ).
adj( 'reserved', '-', '-', normal ).
adj( 'resident', '-', '-', normal ).
adj( 'residential', '-', '-', normal ).
adj( 'residual', '-', '-', normal ).
adj( 'residuary', '-', '-', normal ).
adj( 'resigned', '-', '-', normal ).
adj( 'resilient', '-', '-', normal ).
adj( 'resinated', '-', '-', normal ).
adj( 'resinous', '-', '-', normal ).
adj( 'resistant', '-', '-', normal ).
adj( 'resistive', '-', '-', normal ).
adj( 'resistless', '-', '-', normal ).
adj( 'resolute', '-', '-', normal ).
adj( 'resolvable', '-', '-', normal ).
adj( 'resonant', '-', '-', normal ).
adj( 'resourceful', '-', '-', normal ).
adj( 'resourceless', '-', '-', normal ).
adj( 'respectable', '-', '-', normal ).
adj( 'respectful', '-', '-', normal ).
adj( 'respective', '-', '-', normal ).
adj( 'respiratory', '-', '-', normal ).
adj( 'resplendent', '-', '-', normal ).
adj( 'responsible', '-', '-', normal ).
adj( 'responsive', '-', '-', normal ).
adj( 'restful', '-', '-', normal ).
adj( 'restive', '-', '-', normal ).
adj( 'restless', '-', '-', normal ).
adj( 'restorative', '-', '-', normal ).
adj( 'restrained', '-', '-', normal ).
adj( 'restrictive', '-', '-', normal ).
adj( 'resultant', '-', '-', normal ).
adj( 'resurgent', '-', '-', normal ).
adj( 'retaliative', '-', '-', normal ).
adj( 'retaliatory', '-', '-', normal ).
adj( 'retentive', '-', '-', normal ).
adj( 'reticent', '-', '-', normal ).
adj( 'reticulate', '-', '-', normal ).
adj( 'retinal', '-', '-', normal ).
adj( 'retired', '-', '-', normal ).
adj( 'retiring', '-', '-', normal ).
adj( 'retractable', '-', '-', normal ).
adj( 'retractile', '-', '-', normal ).
adj( 'retributive', '-', '-', normal ).
adj( 'retrievable', '-', '-', normal ).
adj( 'retroactive', '-', '-', normal ).
adj( 'retrograde', '-', '-', normal ).
adj( 'retrogressive', '-', '-', normal ).
adj( 'retrospective', '-', '-', normal ).
adj( 'retrouss_e', '-', '-', normal ).
adj( 'returnable', '-', '-', normal ).
adj( 'revengeful', '-', '-', normal ).
adj( 'reverberant', '-', '-', normal ).
adj( 'reverend', '-', '-', normal ).
adj( 'reverent', '-', '-', normal ).
adj( 'reverential', '-', '-', normal ).
adj( 'reverse', '-', '-', normal ).
adj( 'reversible', '-', '-', normal ).
adj( 'reversionary', '-', '-', normal ).
adj( 'revertible', '-', '-', normal ).
adj( 'revocable', '-', '-', normal ).
adj( 'revolting', '-', '-', normal ).
adj( 'revolutionary', '-', '-', normal ).
adj( 'rhetorical', '-', '-', normal ).
adj( 'rheumatic', '-', '-', normal ).
adj( 'rheumatoid', '-', '-', normal ).
adj( 'rhinal', '-', '-', normal ).
adj( 'rhomboid', '-', '-', normal ).
adj( 'rhymed', '-', '-', normal ).
adj( 'rhythmic', '-', '-', normal ).
adj( 'rhythmical', '-', '-', normal ).
adj( 'ribald', '-', '-', normal ).
adj( 'rich', 'richer', 'richest', normal ).
adj( 'rickety', '-', '-', normal ).
adj( 'riderless', '-', '-', normal ).
adj( 'ridiculous', '-', '-', normal ).
adj( 'rife', '-', '-', pred ).
adj( 'right', '-', '-', normal ).
adj( 'right-angled', '-', '-', normal ).
adj( 'right-down', '-', '-', normal ).
adj( 'right-hand', '-', '-', normal ).
adj( 'right-handed', '-', '-', normal ).
adj( 'right-minded', '-', '-', normal ).
adj( 'righteous', '-', '-', normal ).
adj( 'rightful', '-', '-', normal ).
adj( 'rightist', '-', '-', normal ).
adj( 'rigid', '-', '-', normal ).
adj( 'rigorous', '-', '-', normal ).
adj( 'riotous', '-', '-', normal ).
adj( 'rip-roaring', '-', '-', normal ).
adj( 'riparian', '-', '-', normal ).
adj( 'ripe', 'riper', 'ripest', normal ).
adj( 'risible', '-', '-', normal ).
adj( 'risky', 'riskier', 'riskiest', normal ).
adj( 'risqu_e', '-', '-', normal ).
adj( 'ritual', '-', '-', normal ).
adj( 'ritualistic', '-', '-', normal ).
adj( 'ritzy', '-', '-', normal ).
adj( 'roadless', '-', '-', normal ).
adj( 'roadworthy', '-', '-', normal ).
adj( 'roan', '-', '-', normal ).
adj( 'roaring', '-', '-', normal ).
adj( 'roast', '-', '-', attr ).
adj( 'robust', '-', '-', normal ).
adj( 'rocky', 'rockier', 'rockiest', normal ).
adj( 'rococo', '-', '-', normal ).
adj( 'roguish', '-', '-', normal ).
adj( 'rollicking', '-', '-', normal ).
adj( 'romantic', '-', '-', normal ).
adj( 'roofless', '-', '-', normal ).
adj( 'roomed', '-', '-', affix ).
adj( 'roomy', 'roomier', 'roomiest', normal ).
adj( 'rootless', '-', '-', normal ).
adj( 'ropey', 'ropier', 'ropiest', normal ).
adj( 'rose-red', '-', '-', normal ).
adj( 'roseate', '-', '-', normal ).
adj( 'rosy', 'rosier', 'rosiest', normal ).
adj( 'rotary', '-', '-', normal ).
adj( 'rotational', '-', '-', normal ).
adj( 'rotatory', '-', '-', normal ).
adj( 'rotten', '-', '-', normal ).
adj( 'rotund', '-', '-', normal ).
adj( 'rough', 'rougher', 'roughest', normal ).
adj( 'rough-and-tumble', '-', '-', normal ).
adj( 'rough-hewn', '-', '-', normal ).
adj( 'rough-spoken', '-', '-', normal ).
adj( 'roughish', '-', '-', normal ).
adj( 'roughshod', '-', '-', normal ).
adj( 'round', 'rounder', 'roundest', normal ).
adj( 'round-arm', '-', '-', normal ).
adj( 'round-backed', '-', '-', normal ).
adj( 'round-eyed', '-', '-', normal ).
adj( 'round-shouldered', '-', '-', normal ).
adj( 'round-the-clock', '-', '-', attr ).
adj( 'roundabout', '-', '-', attr ).
adj( 'roundish', '-', '-', normal ).
adj( 'routine', '-', '-', normal ).
adj( 'rowdy', 'rowdier', 'rowdiest', normal ).
adj( 'royal', '-', '-', normal ).
adj( 'rubbery', '-', '-', normal ).
adj( 'rubbishy', '-', '-', normal ).
adj( 'rubicund', '-', '-', normal ).
adj( 'ruby', '-', '-', normal ).
adj( 'rudderless', '-', '-', normal ).
adj( 'ruddy', 'ruddier', 'ruddiest', normal ).
adj( 'rude', 'ruder', 'rudest', normal ).
adj( 'rudimentary', '-', '-', normal ).
adj( 'rueful', '-', '-', normal ).
adj( 'ruffianly', '-', '-', normal ).
adj( 'rugged', '-', '-', normal ).
adj( 'ruinous', '-', '-', normal ).
adj( 'ruling', '-', '-', normal ).
adj( 'rumbustious', '-', '-', normal ).
adj( 'ruminant', '-', '-', normal ).
adj( 'ruminative', '-', '-', normal ).
adj( 'rummy', 'rummer', 'rummest', normal ).
adj( 'run-of-the-mill', '-', '-', normal ).
adj( 'runaway', '-', '-', normal ).
adj( 'rundown', '-', '-', normal ).
adj( 'runic', '-', '-', normal ).
adj( 'running', '-', '-', normal ).
adj( 'runny', 'runnier', 'runniest', normal ).
adj( 'rural', '-', '-', normal ).
adj( 'rushy', 'rushier', 'rushiest', normal ).
adj( 'russet', '-', '-', normal ).
adj( 'rustic', '-', '-', normal ).
adj( 'rustless', '-', '-', normal ).
adj( 'rusty', 'rustier', 'rustiest', normal ).
adj( 'ruthless', '-', '-', normal ).
adj( 'sabbatarian', '-', '-', normal ).
adj( 'sabbatical', '-', '-', normal ).
adj( 'sable', '-', '-', normal ).
adj( 'sabre-toothed', '-', '-', normal ).
adj( 'saccharine', '-', '-', normal ).
adj( 'sacerdotal', '-', '-', normal ).
adj( 'sacramental', '-', '-', normal ).
adj( 'sacred', '-', '-', normal ).
adj( 'sacrificial', '-', '-', normal ).
adj( 'sacrilegious', '-', '-', normal ).
adj( 'sacrosanct', '-', '-', normal ).
adj( 'sad', 'sadder', 'saddest', normal ).
adj( 'saddle-sore', '-', '-', normal ).
adj( 'sadistic', '-', '-', normal ).
adj( 'safe', 'safer', 'safest', normal ).
adj( 'safe-deposit', '-', '-', normal ).
adj( 'sagacious', '-', '-', normal ).
adj( 'sage', '-', '-', normal ).
adj( 'sage-green', '-', '-', normal ).
adj( 'said', '-', '-', attr ).
adj( 'sainted', '-', '-', normal ).
adj( 'saintlike', '-', '-', normal ).
adj( 'saintly', 'saintlier', 'saintliest', normal ).
adj( 'salable', '-', '-', normal ).
adj( 'salacious', '-', '-', normal ).
adj( 'salaried', '-', '-', normal ).
adj( 'saleable', '-', '-', normal ).
adj( 'salient', '-', '-', normal ).
adj( 'saline', '-', '-', normal ).
adj( 'salivary', '-', '-', normal ).
adj( 'sallow', 'sallower', 'sallowest', normal ).
adj( 'salt', '-', '-', normal ).
adj( 'salty', 'saltier', 'saltiest', normal ).
adj( 'salubrious', '-', '-', normal ).
adj( 'salutary', '-', '-', normal ).
adj( 'same', '-', '-', normal ).
adj( 'sanctimonious', '-', '-', normal ).
adj( 'sandalled', '-', '-', normal ).
adj( 'sandy', 'sandier', 'sandiest', normal ).
adj( 'sane', 'saner', 'sanest', normal ).
adj( 'sanguinary', '-', '-', normal ).
adj( 'sanguine', '-', '-', normal ).
adj( 'sanitary', '-', '-', normal ).
adj( 'sapient', '-', '-', normal ).
adj( 'sapless', '-', '-', normal ).
adj( 'sappy', 'sappier', 'sappiest', normal ).
adj( 'sarcastic', '-', '-', normal ).
adj( 'sardonic', '-', '-', normal ).
adj( 'sartorial', '-', '-', normal ).
adj( 'satiable', '-', '-', normal ).
adj( 'satin', '-', '-', normal ).
adj( 'satirical', '-', '-', normal ).
adj( 'satisfactory', '-', '-', normal ).
adj( 'satisfying', '-', '-', normal ).
adj( 'saturnine', '-', '-', normal ).
adj( 'satyric', '-', '-', normal ).
adj( 'saucer-eyed', '-', '-', normal ).
adj( 'saucy', 'saucier', 'sauciest', normal ).
adj( 'saurian', '-', '-', normal ).
adj( 'saut_e', '-', '-', normal ).
adj( 'savage', '-', '-', normal ).
adj( 'saving', '-', '-', normal ).
adj( 'savoury', '-', '-', normal ).
adj( 'scabby', 'scabbier', 'scabbiest', normal ).
adj( 'scabrous', '-', '-', normal ).
adj( 'scaly', 'scalier', 'scaliest', normal ).
adj( 'scandalous', '-', '-', normal ).
adj( 'scant', '-', '-', normal ).
adj( 'scanty', 'scantier', 'scantiest', normal ).
adj( 'scarce', 'scarcer', 'scarcest', normal ).
adj( 'scarlet', '-', '-', normal ).
adj( 'scary', 'scarier', 'scariest', normal ).
adj( 'scathing', '-', '-', normal ).
adj( 'scatterbrained', '-', '-', normal ).
adj( 'scattered', '-', '-', normal ).
adj( 'scatty', 'scattier', 'scattiest', normal ).
adj( 'scenic', '-', '-', normal ).
adj( 'scentless', '-', '-', normal ).
adj( 'sceptered', '-', '-', normal ).
adj( 'sceptical', '-', '-', normal ).
adj( 'sceptred', '-', '-', normal ).
adj( 'schematic', '-', '-', normal ).
adj( 'schismatic', '-', '-', normal ).
adj( 'schizoid', '-', '-', normal ).
adj( 'schizophrenic', '-', '-', normal ).
adj( 'schmaltzy', 'schmaltzier', 'schmaltziest', normal ).
adj( 'schmalzy', 'schmalzier', 'schmalziest', normal ).
adj( 'scholarly', '-', '-', normal ).
adj( 'scholastic', '-', '-', normal ).
adj( 'sciatic', '-', '-', normal ).
adj( 'scientific', '-', '-', normal ).
adj( 'scorbutic', '-', '-', normal ).
adj( 'scorching', '-', '-', normal ).
adj( 'scornful', '-', '-', normal ).
adj( 'scoundrelly', '-', '-', normal ).
adj( 'scraggy', 'scraggier', 'scraggiest', normal ).
adj( 'scrappy', 'scrappier', 'scrappiest', normal ).
adj( 'scratchy', 'scratchier', 'scratchiest', normal ).
adj( 'scrawny', 'scrawnier', 'scrawniest', normal ).
adj( 'screw-topped', '-', '-', normal ).
adj( 'screwball', '-', '-', normal ).
adj( 'screwy', 'screwier', 'screwiest', normal ).
adj( 'scripted', '-', '-', normal ).
adj( 'scriptural', '-', '-', normal ).
adj( 'scrofulous', '-', '-', normal ).
adj( 'scrubby', 'scrubbier', 'scrubbiest', normal ).
adj( 'scruffy', 'scruffier', 'scruffiest', normal ).
adj( 'scrumptious', '-', '-', normal ).
adj( 'scrupulous', '-', '-', normal ).
adj( 'sculptural', '-', '-', normal ).
adj( 'scummy', 'scummier', 'scummiest', normal ).
adj( 'scurfy', 'scurfier', 'scurfiest', normal ).
adj( 'scurrilous', '-', '-', normal ).
adj( 'scurvy', '-', '-', normal ).
adj( 'sea-girt', '-', '-', normal ).
adj( 'sea-green', '-', '-', normal ).
adj( 'seaborne', '-', '-', normal ).
adj( 'seafaring', '-', '-', normal ).
adj( 'seagoing', '-', '-', normal ).
adj( 'seamanlike', '-', '-', normal ).
adj( 'seamless', '-', '-', normal ).
adj( 'seamy', 'seamier', 'seamiest', normal ).
adj( 'sear', '-', '-', normal ).
adj( 'searching', '-', '-', normal ).
adj( 'seasick', '-', '-', normal ).
adj( 'seasonable', '-', '-', normal ).
adj( 'seasonal', '-', '-', normal ).
adj( 'seaward', '-', '-', normal ).
adj( 'seaworthy', '-', '-', normal ).
adj( 'secluded', '-', '-', normal ).
adj( 'second', '-', '-', normal ).
adj( 'second-best', '-', '-', normal ).
adj( 'second-class', '-', '-', normal ).
adj( 'second-hand', '-', '-', normal ).
adj( 'second-rate', '-', '-', normal ).
adj( 'second-sighted', '-', '-', normal ).
adj( 'secondary', '-', '-', normal ).
adj( 'secret', '-', '-', normal ).
adj( 'secretarial', '-', '-', normal ).
adj( 'secretive', '-', '-', normal ).
adj( 'sectarian', '-', '-', normal ).
adj( 'sectional', '-', '-', normal ).
adj( 'secular', '-', '-', normal ).
adj( 'secure', '-', '-', normal ).
adj( 'sedate', '-', '-', normal ).
adj( 'sedative', '-', '-', normal ).
adj( 'sedentary', '-', '-', normal ).
adj( 'sedgy', 'sedgier', 'sedgiest', normal ).
adj( 'sedimentary', '-', '-', normal ).
adj( 'seditious', '-', '-', normal ).
adj( 'seductive', '-', '-', normal ).
adj( 'sedulous', '-', '-', normal ).
adj( 'see-through', '-', '-', normal ).
adj( 'seedless', '-', '-', normal ).
adj( 'seedy', 'seedier', 'seediest', normal ).
adj( 'seeming', '-', '-', normal ).
adj( 'seemly', 'seemlier', 'seemliest', normal ).
adj( 'seismic', '-', '-', normal ).
adj( 'select', '-', '-', normal ).
adj( 'selective', '-', '-', normal ).
adj( 'self-absorbed', '-', '-', normal ).
adj( 'self-acting', '-', '-', normal ).
adj( 'self-activating', '-', '-', normal ).
adj( 'self-addressed', '-', '-', normal ).
adj( 'self-appointed', '-', '-', normal ).
adj( 'self-assertive', '-', '-', normal ).
adj( 'self-assured', '-', '-', normal ).
adj( 'self-centred', '-', '-', normal ).
adj( 'self-collected', '-', '-', normal ).
adj( 'self-coloured', '-', '-', normal ).
adj( 'self-confessed', '-', '-', normal ).
adj( 'self-confident', '-', '-', normal ).
adj( 'self-conscious', '-', '-', normal ).
adj( 'self-contained', '-', '-', normal ).
adj( 'self-denying', '-', '-', normal ).
adj( 'self-educated', '-', '-', normal ).
adj( 'self-effacing', '-', '-', normal ).
adj( 'self-employed', '-', '-', normal ).
adj( 'self-evident', '-', '-', normal ).
adj( 'self-explanatory', '-', '-', normal ).
adj( 'self-important', '-', '-', normal ).
adj( 'self-imposed', '-', '-', normal ).
adj( 'self-indulgent', '-', '-', normal ).
adj( 'self-locking', '-', '-', normal ).
adj( 'self-made', '-', '-', normal ).
adj( 'self-opinionated', '-', '-', normal ).
adj( 'self-possessed', '-', '-', normal ).
adj( 'self-raising', '-', '-', normal ).
adj( 'self-reliant', '-', '-', normal ).
adj( 'self-respecting', '-', '-', normal ).
adj( 'self-righteous', '-', '-', normal ).
adj( 'self-sacrificing', '-', '-', normal ).
adj( 'self-same', '-', '-', normal ).
adj( 'self-sealing', '-', '-', normal ).
adj( 'self-seeking', '-', '-', normal ).
adj( 'self-sown', '-', '-', normal ).
adj( 'self-styled', '-', '-', normal ).
adj( 'self-sufficient', '-', '-', normal ).
adj( 'self-sufficing', '-', '-', normal ).
adj( 'self-supporting', '-', '-', normal ).
adj( 'self-willed', '-', '-', normal ).
adj( 'self-winding', '-', '-', normal ).
adj( 'selfish', '-', '-', normal ).
adj( 'semantic', '-', '-', normal ).
adj( 'semicircular', '-', '-', normal ).
adj( 'semiconducting', '-', '-', normal ).
adj( 'semiconscious', '-', '-', normal ).
adj( 'semidetached', '-', '-', normal ).
adj( 'seminal', '-', '-', normal ).
adj( 'semiofficial', '-', '-', normal ).
adj( 'semirigid', '-', '-', normal ).
adj( 'semitropical', '-', '-', normal ).
adj( 'senatorial', '-', '-', normal ).
adj( 'senescent', '-', '-', normal ).
adj( 'senile', '-', '-', normal ).
adj( 'senior', '-', '-', normal ).
adj( 'sensational', '-', '-', normal ).
adj( 'senseless', '-', '-', normal ).
adj( 'sensible', '-', '-', normal ).
adj( 'sensitive', '-', '-', normal ).
adj( 'sensory', '-', '-', normal ).
adj( 'sensual', '-', '-', normal ).
adj( 'sensuous', '-', '-', normal ).
adj( 'sententious', '-', '-', normal ).
adj( 'sentient', '-', '-', normal ).
adj( 'sentimental', '-', '-', normal ).
adj( 'separable', '-', '-', normal ).
adj( 'separate', '-', '-', normal ).
adj( 'septic', '-', '-', normal ).
adj( 'sepulchral', '-', '-', normal ).
adj( 'sequent', '-', '-', normal ).
adj( 'sequential', '-', '-', normal ).
adj( 'sequestered', '-', '-', normal ).
adj( 'seraphic', '-', '-', normal ).
adj( 'sere', '-', '-', normal ).
adj( 'serene', '-', '-', normal ).
adj( 'serial', '-', '-', normal ).
adj( 'sericultural', '-', '-', normal ).
adj( 'seriocomic', '-', '-', normal ).
adj( 'serious', '-', '-', normal ).
adj( 'serous', '-', '-', normal ).
adj( 'serpentine', '-', '-', normal ).
adj( 'serrated', '-', '-', normal ).
adj( 'serried', '-', '-', normal ).
adj( 'serviceable', '-', '-', normal ).
adj( 'servile', '-', '-', normal ).
adj( 'sesquipedalian', '-', '-', normal ).
adj( 'settled', '-', '-', normal ).
adj( 'seven', '-', '-', normal ).
adj( 'sevenfold', '-', '-', normal ).
adj( 'seventeen', '-', '-', normal ).
adj( 'seventeenth', '-', '-', normal ).
adj( 'seventh', '-', '-', normal ).
adj( 'seventieth', '-', '-', normal ).
adj( 'seventy', '-', '-', normal ).
adj( 'several', '-', '-', normal ).
adj( 'severe', 'severer', 'severest', normal ).
adj( 'sex-starved', '-', '-', normal ).
adj( 'sexagenarian', '-', '-', normal ).
adj( 'sexed', '-', '-', normal ).
adj( 'sexist', '-', '-', normal ).
adj( 'sexless', '-', '-', normal ).
adj( 'sexual', '-', '-', normal ).
adj( 'sexy', 'sexier', 'sexiest', normal ).
adj( 'shabby', 'shabbier', 'shabbiest', normal ).
adj( 'shabby-genteel', '-', '-', normal ).
adj( 'shadowy', 'shadowier', 'shadowiest', normal ).
adj( 'shady', 'shadier', 'shadiest', normal ).
adj( 'shaggy', 'shaggier', 'shaggiest', normal ).
adj( 'shaky', 'shakier', 'shakiest', normal ).
adj( 'shallow', 'shallower', 'shallowest', normal ).
adj( 'sham', '-', '-', normal ).
adj( 'shame-making', '-', '-', normal ).
adj( 'shamefaced', '-', '-', normal ).
adj( 'shameful', '-', '-', normal ).
adj( 'shameless', '-', '-', normal ).
adj( 'shapeless', '-', '-', normal ).
adj( 'shapely', 'shapelier', 'shapeliest', normal ).
adj( 'sharp', 'sharper', 'sharpest', normal ).
adj( 'sharp-eyed', '-', '-', normal ).
adj( 'sharp-set', '-', '-', normal ).
adj( 'sharp-sighted', '-', '-', normal ).
adj( 'sharp-witted', '-', '-', normal ).
adj( 'shatterproof', '-', '-', normal ).
adj( 'sheepish', '-', '-', normal ).
adj( 'sheer', 'sheerer', 'sheerest', normal ).
adj( 'shellproof', '-', '-', normal ).
adj( 'shiftless', '-', '-', normal ).
adj( 'shifty', 'shiftier', 'shiftiest', normal ).
adj( 'shingly', 'shinglier', 'shingliest', normal ).
adj( 'shiny', 'shinier', 'shiniest', normal ).
adj( 'shipboard', '-', '-', attr ).
adj( 'shipshape', '-', '-', normal ).
adj( 'shirty', 'shirtier', 'shirtiest', normal ).
adj( 'shivery', '-', '-', normal ).
adj( 'shock-headed', '-', '-', normal ).
adj( 'shockable', '-', '-', normal ).
adj( 'shocking', '-', '-', normal ).
adj( 'shoddy', 'shoddier', 'shoddiest', normal ).
adj( 'shopsoiled', '-', '-', normal ).
adj( 'shopworn', '-', '-', normal ).
adj( 'short', 'shorter', 'shortest', normal ).
adj( 'short-dated', '-', '-', normal ).
adj( 'short-handed', '-', '-', normal ).
adj( 'short-lived', '-', '-', normal ).
adj( 'short-range', '-', '-', normal ).
adj( 'short-sighted', '-', '-', normal ).
adj( 'short-tempered', '-', '-', normal ).
adj( 'short-term', '-', '-', attr ).
adj( 'short-winded', '-', '-', normal ).
adj( 'shortish', '-', '-', normal ).
adj( 'showery', 'showerier', 'showeriest', normal ).
adj( 'showy', 'showier', 'showiest', normal ).
adj( 'shrewd', 'shrewder', 'shrewdest', normal ).
adj( 'shrewish', '-', '-', normal ).
adj( 'shrill', 'shriller', 'shrillest', normal ).
adj( 'shrinkable', '-', '-', normal ).
adj( 'shy', 'shyer', 'shyest', normal ).
adj( 'sibilant', '-', '-', normal ).
adj( 'sibylline', '-', '-', normal ).
adj( 'sick', '-', '-', normal ).
adj( 'sickening', '-', '-', normal ).
adj( 'sickish', '-', '-', normal ).
adj( 'sickly', 'sicklier', 'sickliest', normal ).
adj( 'side-splitting', '-', '-', normal ).
adj( 'sided', '-', '-', affix ).
adj( 'sidelong', '-', '-', normal ).
adj( 'sidereal', '-', '-', normal ).
adj( 'sighted', '-', '-', affix ).
adj( 'sightless', '-', '-', normal ).
adj( 'signal', '-', '-', attr ).
adj( 'significant', '-', '-', normal ).
adj( 'significative', '-', '-', normal ).
adj( 'silent', '-', '-', normal ).
adj( 'silken', '-', '-', normal ).
adj( 'silky', 'silkier', 'silkiest', normal ).
adj( 'silly', 'sillier', 'silliest', normal ).
adj( 'silvan', '-', '-', normal ).
adj( 'silvern', '-', '-', normal ).
adj( 'silvery', 'silverier', 'silveriest', normal ).
adj( 'simian', '-', '-', normal ).
adj( 'similar', '-', '-', normal ).
adj( 'simple', 'simpler', 'simplest', normal ).
adj( 'simple-hearted', '-', '-', normal ).
adj( 'simple-minded', '-', '-', normal ).
adj( 'simultaneous', '-', '-', normal ).
adj( 'sincere', '-', '-', normal ).
adj( 'sinewy', '-', '-', normal ).
adj( 'sinful', '-', '-', normal ).
adj( 'singable', '-', '-', normal ).
adj( 'single', '-', '-', normal ).
adj( 'single-breasted', '-', '-', normal ).
adj( 'single-handed', '-', '-', normal ).
adj( 'single-minded', '-', '-', normal ).
adj( 'singular', '-', '-', normal ).
adj( 'sinister', '-', '-', normal ).
adj( 'sinkable', '-', '-', normal ).
adj( 'sinless', '-', '-', normal ).
adj( 'sinuous', '-', '-', normal ).
adj( 'sissified', '-', '-', normal ).
adj( 'sisterly', '-', '-', normal ).
adj( 'situated', '-', '-', pred ).
adj( 'six', '-', '-', normal ).
adj( 'sixfold', '-', '-', normal ).
adj( 'sixpenny', '-', '-', normal ).
adj( 'sixteen', '-', '-', normal ).
adj( 'sixteenth', '-', '-', normal ).
adj( 'sixth', '-', '-', normal ).
adj( 'sixtieth', '-', '-', normal ).
adj( 'sixty', '-', '-', normal ).
adj( 'sizable', '-', '-', normal ).
adj( 'sizeable', '-', '-', normal ).
adj( 'sized', '-', '-', affix ).
adj( 'skeptical', '-', '-', normal ).
adj( 'sketchy', 'sketchier', 'sketchiest', normal ).
adj( 'skew', '-', '-', normal ).
adj( 'skew-eyed', '-', '-', normal ).
adj( 'skilful', '-', '-', normal ).
adj( 'skilled', '-', '-', normal ).
adj( 'skimpy', 'skimpier', 'skimpiest', normal ).
adj( 'skin-deep', '-', '-', normal ).
adj( 'skin-tight', '-', '-', normal ).
adj( 'skinny', 'skinnier', 'skinniest', normal ).
adj( 'skint', '-', '-', normal ).
adj( 'skittish', '-', '-', normal ).
adj( 'skulled', '-', '-', affix ).
adj( 'sky-blue', '-', '-', normal ).
adj( 'skyward', '-', '-', normal ).
adj( 'skywards', '-', '-', normal ).
adj( 'slack', 'slacker', 'slackest', normal ).
adj( 'slanderous', '-', '-', normal ).
adj( 'slangy', 'slangier', 'slangiest', normal ).
adj( 'slap-happy', '-', '-', normal ).
adj( 'slap-up', '-', '-', normal ).
adj( 'slapdash', '-', '-', normal ).
adj( 'slatted', '-', '-', normal ).
adj( 'slatternly', '-', '-', normal ).
adj( 'slaty', 'slatier', 'slatiest', normal ).
adj( 'slavish', '-', '-', normal ).
adj( 'sleazy', 'sleazier', 'sleaziest', normal ).
adj( 'sleek', 'sleeker', 'sleekest', normal ).
adj( 'sleepless', '-', '-', normal ).
adj( 'sleepy', 'sleepier', 'sleepiest', normal ).
adj( 'sleety', 'sleetier', 'sleetiest', normal ).
adj( 'sleeved', '-', '-', affix ).
adj( 'sleeveless', '-', '-', normal ).
adj( 'slender', '-', '-', normal ).
adj( 'slick', 'slicker', 'slickest', normal ).
adj( 'slight', 'slighter', 'slightest', normal ).
adj( 'slim', 'slimmer', 'slimmest', normal ).
adj( 'slimy', 'slimier', 'slimiest', normal ).
adj( 'slippered', '-', '-', normal ).
adj( 'slippery', 'slipperier', 'slipperiest', normal ).
adj( 'slippy', '-', '-', normal ).
adj( 'slipshod', '-', '-', normal ).
adj( 'slithery', '-', '-', normal ).
adj( 'sloppy', 'sloppier', 'sloppiest', normal ).
adj( 'sloshed', '-', '-', normal ).
adj( 'slothful', '-', '-', normal ).
adj( 'slovenly', 'slovenlier', 'slovenliest', normal ).
adj( 'slow', 'slower', 'slowest', normal ).
adj( 'sluggish', '-', '-', normal ).
adj( 'slumberous', '-', '-', normal ).
adj( 'slummy', 'slummier', 'slummiest', normal ).
adj( 'slushy', 'slushier', 'slushiest', normal ).
adj( 'sluttish', '-', '-', normal ).
adj( 'sly', 'slyer', 'slyest', normal ).
adj( 'small', 'smaller', 'smallest', normal ).
adj( 'small-minded', '-', '-', normal ).
adj( 'smalltime', '-', '-', normal ).
adj( 'smarmy', '-', '-', normal ).
adj( 'smart', 'smarter', 'smartest', normal ).
adj( 'smashing', '-', '-', normal ).
adj( 'smelly', 'smellier', 'smelliest', normal ).
adj( 'smoke-cured', '-', '-', normal ).
adj( 'smoke-dried', '-', '-', normal ).
adj( 'smokeless', '-', '-', normal ).
adj( 'smoky', 'smokier', 'smokiest', normal ).
adj( 'smooth', 'smoother', 'smoothest', normal ).
adj( 'smooth-bore', '-', '-', normal ).
adj( 'smooth-faced', '-', '-', normal ).
adj( 'smooth-spoken', '-', '-', normal ).
adj( 'smooth-tongued', '-', '-', normal ).
adj( 'smudgy', '-', '-', normal ).
adj( 'smug', 'smugger', 'smuggest', normal ).
adj( 'smutty', 'smuttier', 'smuttiest', normal ).
adj( 'snaky', 'snakier', 'snakiest', normal ).
adj( 'snappish', '-', '-', normal ).
adj( 'snappy', 'snappier', 'snappiest', normal ).
adj( 'snazzy', '-', '-', normal ).
adj( 'sneaking', '-', '-', normal ).
adj( 'sneaky', 'sneakier', 'sneakiest', normal ).
adj( 'snide', '-', '-', normal ).
adj( 'sniffy', 'sniffier', 'sniffiest', normal ).
adj( 'snobbish', '-', '-', normal ).
adj( 'snooty', 'snootier', 'snootiest', normal ).
adj( 'snorty', 'snortier', 'snortiest', normal ).
adj( 'snot-nosed', '-', '-', normal ).
adj( 'snotty', 'snottier', 'snottiest', normal ).
adj( 'snow-clad', '-', '-', normal ).
adj( 'snow-covered', '-', '-', normal ).
adj( 'snow-white', '-', '-', normal ).
adj( 'snowblind', '-', '-', normal ).
adj( 'snowbound', '-', '-', normal ).
adj( 'snowcapped', '-', '-', normal ).
adj( 'snowy', 'snowier', 'snowiest', normal ).
adj( 'snub', '-', '-', normal ).
adj( 'snub-nosed', '-', '-', normal ).
adj( 'snuff-colour', '-', '-', normal ).
adj( 'snuff-coloured', '-', '-', normal ).
adj( 'snug', 'snugger', 'snuggest', normal ).
adj( 'so-called', '-', '-', normal ).
adj( 'so-so', '-', '-', pred ).
adj( 'soapy', 'soapier', 'soapiest', normal ).
adj( 'sober', '-', '-', normal ).
adj( 'sociable', '-', '-', normal ).
adj( 'social', '-', '-', normal ).
adj( 'socialist', '-', '-', normal ).
adj( 'sociological', '-', '-', normal ).
adj( 'sodden', '-', '-', normal ).
adj( 'sodding', '-', '-', attr ).
adj( 'soft', 'softer', 'softest', normal ).
adj( 'soft-boiled', '-', '-', normal ).
adj( 'soft-footed', '-', '-', normal ).
adj( 'soft-headed', '-', '-', normal ).
adj( 'soft-hearted', '-', '-', normal ).
adj( 'soft-spoken', '-', '-', normal ).
adj( 'soft-witted', '-', '-', normal ).
adj( 'softish', '-', '-', normal ).
adj( 'soggy', 'soggier', 'soggiest', normal ).
adj( 'soign_e', '-', '-', normal ).
adj( 'soignee', '-', '-', normal ).
adj( 'solar', '-', '-', normal ).
adj( 'sole', '-', '-', normal ).
adj( 'soled', '-', '-', affix ).
adj( 'solemn', '-', '-', normal ).
adj( 'solicitous', '-', '-', normal ).
adj( 'solid', '-', '-', normal ).
adj( 'solid-state', '-', '-', normal ).
adj( 'solitary', '-', '-', normal ).
adj( 'soluble', '-', '-', normal ).
adj( 'solvable', '-', '-', normal ).
adj( 'solvent', '-', '-', normal ).
adj( 'somatic', '-', '-', normal ).
adj( 'sombre', '-', '-', normal ).
adj( 'some', '-', '-', normal ).
adj( 'somnolent', '-', '-', normal ).
adj( 'sonic', '-', '-', normal ).
adj( 'sonorous', '-', '-', normal ).
adj( 'sonsy', '-', '-', normal ).
adj( 'sooty', 'sootier', 'sootiest', normal ).
adj( 'sophisticated', '-', '-', normal ).
adj( 'soporific', '-', '-', normal ).
adj( 'sopping', '-', '-', normal ).
adj( 'soppy', 'soppier', 'soppiest', normal ).
adj( 'sordid', '-', '-', normal ).
adj( 'sore', '-', '-', normal ).
adj( 'sorrel', '-', '-', normal ).
adj( 'sorrowful', '-', '-', normal ).
adj( 'sorry', 'sorrier', 'sorriest', normal ).
adj( 'sottish', '-', '-', normal ).
adj( 'soul-destroying', '-', '-', normal ).
adj( 'soul-stirring', '-', '-', normal ).
adj( 'soulful', '-', '-', normal ).
adj( 'soulless', '-', '-', normal ).
adj( 'sound', '-', '-', normal ).
adj( 'soundless', '-', '-', normal ).
adj( 'soundproof', '-', '-', normal ).
adj( 'sour', '-', '-', normal ).
adj( 'soused', '-', '-', normal ).
adj( 'southeasterly', '-', '-', normal ).
adj( 'southeastern', '-', '-', normal ).
adj( 'southerly', '-', '-', normal ).
adj( 'southern', '-', '-', normal ).
adj( 'southernmost', '-', '-', normal ).
adj( 'southwesterly', '-', '-', normal ).
adj( 'southwestern', '-', '-', normal ).
adj( 'sovereign', '-', '-', normal ).
adj( 'sozzled', '-', '-', normal ).
adj( 'spacious', '-', '-', normal ).
adj( 'spanking', '-', '-', normal ).
adj( 'spare', '-', '-', normal ).
adj( 'sparing', '-', '-', normal ).
adj( 'sparkling', '-', '-', normal ).
adj( 'sparse', 'sparser', 'sparsest', normal ).
adj( 'spasmodic', '-', '-', normal ).
adj( 'spastic', '-', '-', normal ).
adj( 'spatial', '-', '-', normal ).
adj( 'spavined', '-', '-', normal ).
adj( 'special', '-', '-', normal ).
adj( 'specifiable', '-', '-', normal ).
adj( 'specific', '-', '-', normal ).
adj( 'specious', '-', '-', normal ).
adj( 'specked', '-', '-', normal ).
adj( 'speckled', '-', '-', normal ).
adj( 'speckless', '-', '-', normal ).
adj( 'spectacled', '-', '-', normal ).
adj( 'spectacular', '-', '-', normal ).
adj( 'spectral', '-', '-', normal ).
adj( 'spectroscopic', '-', '-', normal ).
adj( 'speculative', '-', '-', normal ).
adj( 'speechless', '-', '-', normal ).
adj( 'speedy', 'speedier', 'speediest', normal ).
adj( 'spellbound', '-', '-', normal ).
adj( 'spent', '-', '-', normal ).
adj( 'spherical', '-', '-', normal ).
adj( 'spick', '-', '-', normal ).
adj( 'spicy', 'spicier', 'spiciest', normal ).
adj( 'spidery', '-', '-', normal ).
adj( 'spiky', 'spikier', 'spikiest', normal ).
adj( 'spinal', '-', '-', normal ).
adj( 'spindle-legged', '-', '-', normal ).
adj( 'spindle-shanked', '-', '-', normal ).
adj( 'spindly', 'spindlier', 'spindliest', normal ).
adj( 'spineless', '-', '-', normal ).
adj( 'spiny', 'spinier', 'spiniest', normal ).
adj( 'spiral', '-', '-', normal ).
adj( 'spirited', '-', '-', normal ).
adj( 'spiritless', '-', '-', normal ).
adj( 'spiritual', '-', '-', normal ).
adj( 'spiritualistic', '-', '-', normal ).
adj( 'spirituous', '-', '-', normal ).
adj( 'spiteful', '-', '-', normal ).
adj( 'splay', '-', '-', normal ).
adj( 'splayfooted', '-', '-', normal ).
adj( 'splendid', '-', '-', normal ).
adj( 'splendiferous', '-', '-', normal ).
adj( 'splenetic', '-', '-', normal ).
adj( 'splinter-proof', '-', '-', normal ).
adj( 'splintery', '-', '-', normal ).
adj( 'spondaic', '-', '-', normal ).
adj( 'spongy', 'spongier', 'spongiest', normal ).
adj( 'spontaneous', '-', '-', normal ).
adj( 'spooky', 'spookier', 'spookiest', normal ).
adj( 'sporadic', '-', '-', normal ).
adj( 'sporting', '-', '-', normal ).
adj( 'sportive', '-', '-', normal ).
adj( 'sportsmanlike', '-', '-', normal ).
adj( 'spotless', '-', '-', normal ).
adj( 'spotted', '-', '-', normal ).
adj( 'spotty', 'spottier', 'spottiest', normal ).
adj( 'sprigged', '-', '-', normal ).
adj( 'sprightly', 'sprightlier', 'sprightliest', normal ).
adj( 'springless', '-', '-', normal ).
adj( 'springlike', '-', '-', normal ).
adj( 'springy', 'springier', 'springiest', normal ).
adj( 'spruce', '-', '-', normal ).
adj( 'spry', 'spryer', 'spryest', normal ).
adj( 'spunky', 'spunkier', 'spunkiest', normal ).
adj( 'spurious', '-', '-', normal ).
adj( 'squalid', '-', '-', normal ).
adj( 'squally', '-', '-', normal ).
adj( 'square', '-', '-', normal ).
adj( 'square-built', '-', '-', normal ).
adj( 'square-rigged', '-', '-', normal ).
adj( 'square-shouldered', '-', '-', normal ).
adj( 'square-toed', '-', '-', normal ).
adj( 'squashy', 'squashier', 'squashiest', normal ).
adj( 'squat', '-', '-', normal ).
adj( 'squeaky', 'squeakier', 'squeakiest', normal ).
adj( 'squeamish', '-', '-', normal ).
adj( 'squiffy', '-', '-', normal ).
adj( 'squiggly', 'squigglier', 'squiggliest', normal ).
adj( 'squint-eyed', '-', '-', normal ).
adj( 'stable', '-', '-', normal ).
adj( 'staccato', '-', '-', normal ).
adj( 'stage-struck', '-', '-', normal ).
adj( 'stagnant', '-', '-', normal ).
adj( 'stagy', '-', '-', normal ).
adj( 'staid', '-', '-', normal ).
adj( 'stainless', '-', '-', normal ).
adj( 'stale', 'staler', 'stalest', normal ).
adj( 'stall-fed', '-', '-', normal ).
adj( 'stalwart', '-', '-', normal ).
adj( 'stand-up', '-', '-', normal ).
adj( 'standard', '-', '-', normal ).
adj( 'standing', '-', '-', normal ).
adj( 'standoffish', '-', '-', normal ).
adj( 'starchy', 'starchier', 'starchiest', normal ).
adj( 'staring', '-', '-', normal ).
adj( 'stark', '-', '-', normal ).
adj( 'starkers', '-', '-', pred ).
adj( 'starless', '-', '-', normal ).
adj( 'starlit', '-', '-', normal ).
adj( 'starry', 'starrier', 'starriest', normal ).
adj( 'starry-eyed', '-', '-', normal ).
adj( 'stated', '-', '-', normal ).
adj( 'stateless', '-', '-', normal ).
adj( 'stately', 'statelier', 'stateliest', normal ).
adj( 'statesmanlike', '-', '-', normal ).
adj( 'static', '-', '-', normal ).
adj( 'stationary', '-', '-', normal ).
adj( 'statistical', '-', '-', normal ).
adj( 'statuary', '-', '-', normal ).
adj( 'statuesque', '-', '-', normal ).
adj( 'statutory', '-', '-', normal ).
adj( 'staunch', '-', '-', normal ).
adj( 'steadfast', '-', '-', normal ).
adj( 'steady', 'steadier', 'steadiest', normal ).
adj( 'stealthy', 'stealthier', 'stealthiest', normal ).
adj( 'steamy', 'steamier', 'steamiest', normal ).
adj( 'steel-clad', '-', '-', normal ).
adj( 'steel-plated', '-', '-', normal ).
adj( 'steely', 'steelier', 'steeliest', normal ).
adj( 'steep', 'steeper', 'steepest', normal ).
adj( 'steepish', '-', '-', normal ).
adj( 'stellar', '-', '-', normal ).
adj( 'stemmed', '-', '-', affix ).
adj( 'stentorian', '-', '-', normal ).
adj( 'stereophonic', '-', '-', normal ).
adj( 'stereoscopic', '-', '-', normal ).
adj( 'sterile', '-', '-', normal ).
adj( 'sterling', '-', '-', normal ).
adj( 'stern', 'sterner', 'sternest', normal ).
adj( 'stertorous', '-', '-', normal ).
adj( 'stewed', '-', '-', normal ).
adj( 'stick-in-the-mud', '-', '-', attr ).
adj( 'stick-on', '-', '-', attr ).
adj( 'sticky', 'stickier', 'stickiest', normal ).
adj( 'stiff', 'stiffer', 'stiffest', normal ).
adj( 'stiff-necked', '-', '-', normal ).
adj( 'still', 'stiller', 'stillest', normal ).
adj( 'stillborn', '-', '-', normal ).
adj( 'stilly', '-', '-', normal ).
adj( 'stilted', '-', '-', normal ).
adj( 'stimulating', '-', '-', normal ).
adj( 'stingless', '-', '-', normal ).
adj( 'stingy', 'stingier', 'stingiest', normal ).
adj( 'stipendiary', '-', '-', normal ).
adj( 'stirring', '-', '-', normal ).
adj( 'stochastic', '-', '-', normal ).
adj( 'stockinged', '-', '-', normal ).
adj( 'stocky', 'stockier', 'stockiest', normal ).
adj( 'stodgy', 'stodgier', 'stodgiest', normal ).
adj( 'stoical', '-', '-', normal ).
adj( 'stolen', '-', '-', normal ).
adj( 'stolid', '-', '-', normal ).
adj( 'stone-blind', '-', '-', normal ).
adj( 'stone-cold', '-', '-', normal ).
adj( 'stone-dead', '-', '-', normal ).
adj( 'stone-deaf', '-', '-', normal ).
adj( 'stone-sober', '-', '-', normal ).
adj( 'stoned', '-', '-', normal ).
adj( 'stoneless', '-', '-', normal ).
adj( 'stony', 'stonier', 'stoniest', normal ).
adj( 'stony-broke', '-', '-', normal ).
adj( 'storeyed', '-', '-', affix ).
adj( 'storied', '-', '-', normal ).
adj( 'storm-beaten', '-', '-', normal ).
adj( 'storm-bound', '-', '-', normal ).
adj( 'storm-tossed', '-', '-', normal ).
adj( 'stormproof', '-', '-', normal ).
adj( 'stormy', 'stormier', 'stormiest', normal ).
adj( 'stout', 'stouter', 'stoutest', normal ).
adj( 'stouthearted', '-', '-', normal ).
adj( 'straggly', 'stragglier', 'straggliest', normal ).
adj( 'straight', '-', '-', normal ).
adj( 'straightforward', '-', '-', normal ).
adj( 'strained', '-', '-', normal ).
adj( 'strait', '-', '-', normal ).
adj( 'strait-laced', '-', '-', normal ).
adj( 'strange', 'stranger', 'strangest', normal ).
adj( 'strapping', '-', '-', normal ).
adj( 'strategic', '-', '-', normal ).
adj( 'strategical', '-', '-', normal ).
adj( 'straw-coloured', '-', '-', normal ).
adj( 'streaky', 'streakier', 'streakiest', normal ).
adj( 'streamlined', '-', '-', normal ).
adj( 'strenuous', '-', '-', normal ).
adj( 'striated', '-', '-', normal ).
adj( 'stricken', '-', '-', pred ).
adj( 'strict', 'stricter', 'strictest', normal ).
adj( 'strident', '-', '-', normal ).
adj( 'strikebound', '-', '-', normal ).
adj( 'striking', '-', '-', normal ).
adj( 'stringent', '-', '-', normal ).
adj( 'stringy', 'stringier', 'stringiest', normal ).
adj( 'striped', '-', '-', normal ).
adj( 'stripy', 'stripier', 'stripiest', normal ).
adj( 'strong', 'stronger', 'strongest', normal ).
adj( 'strong-arm', '-', '-', normal ).
adj( 'strong-boned', '-', '-', normal ).
adj( 'strong-minded', '-', '-', normal ).
adj( 'stroppy', '-', '-', normal ).
adj( 'struck', '-', '-', affix ).
adj( 'structural', '-', '-', normal ).
adj( 'structured', '-', '-', normal ).
adj( 'stubbly', 'stubblier', 'stubbliest', normal ).
adj( 'stubborn', '-', '-', normal ).
adj( 'stubby', 'stubbier', 'stubbiest', normal ).
adj( 'stuck-up', '-', '-', normal ).
adj( 'studied', '-', '-', normal ).
adj( 'studious', '-', '-', normal ).
adj( 'stuffy', 'stuffier', 'stuffiest', normal ).
adj( 'stumpy', 'stumpier', 'stumpiest', normal ).
adj( 'stunning', '-', '-', normal ).
adj( 'stupendous', '-', '-', normal ).
adj( 'stupid', '-', '-', normal ).
adj( 'stupifying', '-', '-', normal ).
adj( 'sturdy', 'sturdier', 'sturdiest', normal ).
adj( 'stylish', '-', '-', normal ).
adj( 'stylistic', '-', '-', normal ).
adj( 'styptic', '-', '-', normal ).
adj( 'suave', '-', '-', normal ).
adj( 'sub judice', '-', '-', normal ).
adj( 'subatomic', '-', '-', normal ).
adj( 'subconscious', '-', '-', normal ).
adj( 'subcutaneous', '-', '-', normal ).
adj( 'subfusc', '-', '-', normal ).
adj( 'subhuman', '-', '-', normal ).
adj( 'subject', '-', '-', normal ).
adj( 'subjective', '-', '-', normal ).
adj( 'subjunctive', '-', '-', normal ).
adj( 'sublimate', '-', '-', normal ).
adj( 'sublime', '-', '-', normal ).
adj( 'subliminal', '-', '-', normal ).
adj( 'submarine', '-', '-', normal ).
adj( 'submerged', '-', '-', normal ).
adj( 'submersible', '-', '-', normal ).
adj( 'submissive', '-', '-', normal ).
adj( 'subnormal', '-', '-', normal ).
adj( 'suborbital', '-', '-', normal ).
adj( 'subordinate', '-', '-', normal ).
adj( 'subordinative', '-', '-', normal ).
adj( 'subsequent', '-', '-', normal ).
adj( 'subservient', '-', '-', normal ).
adj( 'subsidiary', '-', '-', normal ).
adj( 'subsonic', '-', '-', normal ).
adj( 'substandard', '-', '-', normal ).
adj( 'substantial', '-', '-', normal ).
adj( 'substantival', '-', '-', normal ).
adj( 'substantive', '-', '-', normal ).
adj( 'subterranean', '-', '-', normal ).
adj( 'subtle', 'subtler', 'subtlest', normal ).
adj( 'subtropical', '-', '-', normal ).
adj( 'suburban', '-', '-', normal ).
adj( 'subversive', '-', '-', normal ).
adj( 'successful', '-', '-', normal ).
adj( 'successive', '-', '-', normal ).
adj( 'succinct', '-', '-', normal ).
adj( 'succulent', '-', '-', normal ).
adj( 'such', '-', '-', normal ).
adj( 'suchlike', '-', '-', normal ).
adj( 'sudden', '-', '-', normal ).
adj( 'suety', '-', '-', normal ).
adj( 'sufferable', '-', '-', normal ).
adj( 'sufficient', '-', '-', normal ).
adj( 'sugar-coated', '-', '-', normal ).
adj( 'sugary', 'sugarier', 'sugariest', normal ).
adj( 'suggestible', '-', '-', normal ).
adj( 'suggestive', '-', '-', normal ).
adj( 'suicidal', '-', '-', normal ).
adj( 'suitable', '-', '-', normal ).
adj( 'sulky', 'sulkier', 'sulkiest', normal ).
adj( 'sullen', '-', '-', normal ).
adj( 'sulphuretted', '-', '-', normal ).
adj( 'sulphuric', '-', '-', normal ).
adj( 'sulphurous', '-', '-', normal ).
adj( 'sultry', 'sultrier', 'sultriest', normal ).
adj( 'summary', '-', '-', normal ).
adj( 'summery', 'summerier', 'summeriest', normal ).
adj( 'sumptuary', '-', '-', normal ).
adj( 'sumptuous', '-', '-', normal ).
adj( 'sun-drenched', '-', '-', normal ).
adj( 'sun-dried', '-', '-', normal ).
adj( 'sunbaked', '-', '-', normal ).
adj( 'sunburned', '-', '-', normal ).
adj( 'sunburnt', '-', '-', normal ).
adj( 'sundry', '-', '-', normal ).
adj( 'sunless', '-', '-', normal ).
adj( 'sunlit', '-', '-', normal ).
adj( 'sunny', 'sunnier', 'sunniest', normal ).
adj( 'sunray', '-', '-', normal ).
adj( 'super', '-', '-', normal ).
adj( 'superabundant', '-', '-', normal ).
adj( 'superannuated', '-', '-', normal ).
adj( 'superb', '-', '-', normal ).
adj( 'supercharged', '-', '-', normal ).
adj( 'supercilious', '-', '-', normal ).
adj( 'superfatted', '-', '-', normal ).
adj( 'superficial', '-', '-', normal ).
adj( 'superfine', '-', '-', normal ).
adj( 'superfluous', '-', '-', normal ).
adj( 'superhuman', '-', '-', normal ).
adj( 'superior', '-', '-', normal ).
adj( 'superlative', '-', '-', normal ).
adj( 'supernal', '-', '-', normal ).
adj( 'supernatural', '-', '-', normal ).
adj( 'supernormal', '-', '-', normal ).
adj( 'supersonic', '-', '-', normal ).
adj( 'superstitious', '-', '-', normal ).
adj( 'supervisory', '-', '-', normal ).
adj( 'supine', '-', '-', normal ).
adj( 'supperless', '-', '-', normal ).
adj( 'supple', 'suppler', 'supplest', normal ).
adj( 'supplementary', '-', '-', normal ).
adj( 'suppliant', '-', '-', normal ).
adj( 'supportable', '-', '-', normal ).
adj( 'supposed', '-', '-', normal ).
adj( 'suppressive', '-', '-', normal ).
adj( 'supranational', '-', '-', normal ).
adj( 'supreme', '-', '-', normal ).
adj( 'sure', 'surer', 'surest', normal ).
adj( 'sure-footed', '-', '-', normal ).
adj( 'surface-to-air', '-', '-', normal ).
adj( 'surgical', '-', '-', normal ).
adj( 'surly', 'surlier', 'surliest', normal ).
adj( 'surmountable', '-', '-', normal ).
adj( 'surpassing', '-', '-', normal ).
adj( 'surpliced', '-', '-', normal ).
adj( 'surprised', '-', '-', normal ).
adj( 'surprising', '-', '-', normal ).
adj( 'surrealistic', '-', '-', normal ).
adj( 'surreptitious', '-', '-', normal ).
adj( 'surrogate', '-', '-', normal ).
adj( 'surrounding', '-', '-', normal ).
adj( 'susceptible', '-', '-', normal ).
adj( 'suspect', '-', '-', pred ).
adj( 'suspicious', '-', '-', normal ).
adj( 'svelte', '-', '-', normal ).
adj( 'swagger', '-', '-', normal ).
adj( 'swallow-tailed', '-', '-', normal ).
adj( 'swampy', 'swampier', 'swampiest', normal ).
adj( 'swanky', 'swankier', 'swankiest', normal ).
adj( 'swarthy', '-', '-', normal ).
adj( 'swashbuckling', '-', '-', normal ).
adj( 'sweaty', 'sweatier', 'sweatiest', normal ).
adj( 'sweeping', '-', '-', normal ).
adj( 'sweet', 'sweeter', 'sweetest', normal ).
adj( 'sweet-scented', '-', '-', normal ).
adj( 'sweetish', '-', '-', normal ).
adj( 'swell', '-', '-', normal ).
adj( 'sweptback', '-', '-', normal ).
adj( 'swift', 'swifter', 'swiftest', normal ).
adj( 'swingeing', '-', '-', normal ).
adj( 'swinging', '-', '-', normal ).
adj( 'swinish', '-', '-', normal ).
adj( 'swish', '-', '-', normal ).
adj( 'swollen-headed', '-', '-', normal ).
adj( 'sworn', '-', '-', normal ).
adj( 'sybaritic', '-', '-', normal ).
adj( 'sycophantic', '-', '-', normal ).
adj( 'syllabic', '-', '-', normal ).
adj( 'syllabled', '-', '-', affix ).
adj( 'syllogistic', '-', '-', normal ).
adj( 'sylph-like', '-', '-', normal ).
adj( 'sylvan', '-', '-', normal ).
adj( 'symbolic', '-', '-', normal ).
adj( 'symbolical', '-', '-', normal ).
adj( 'symmetric', '-', '-', normal ).
adj( 'symmetrical', '-', '-', normal ).
adj( 'sympathetic', '-', '-', normal ).
adj( 'symphonic', '-', '-', normal ).
adj( 'symptomatic', '-', '-', normal ).
adj( 'symptomless', '-', '-', normal ).
adj( 'synchronous', '-', '-', normal ).
adj( 'synonymous', '-', '-', normal ).
adj( 'synoptic', '-', '-', normal ).
adj( 'syntactic', '-', '-', normal ).
adj( 'synthetic', '-', '-', normal ).
adj( 'syphilitic', '-', '-', normal ).
adj( 'syrupy', '-', '-', normal ).
adj( 'systematic', '-', '-', normal ).
adj( 'tabby', '-', '-', attr ).
adj( 'table d\'h^ote', '-', '-', normal ).
adj( 'taboo', '-', '-', normal ).
adj( 'tabular', '-', '-', normal ).
adj( 'tacit', '-', '-', normal ).
adj( 'taciturn', '-', '-', normal ).
adj( 'tacky', 'tackier', 'tackiest', normal ).
adj( 'tactful', '-', '-', normal ).
adj( 'tactical', '-', '-', normal ).
adj( 'tactile', '-', '-', normal ).
adj( 'tactless', '-', '-', normal ).
adj( 'tactual', '-', '-', normal ).
adj( 'tailed', '-', '-', affix ).
adj( 'tailless', '-', '-', normal ).
adj( 'tailor-made', '-', '-', normal ).
adj( 'taintless', '-', '-', normal ).
adj( 'take-home', '-', '-', normal ).
adj( 'takeaway', '-', '-', attr ).
adj( 'taking', '-', '-', normal ).
adj( 'talented', '-', '-', normal ).
adj( 'talkative', '-', '-', normal ).
adj( 'tall', 'taller', 'tallest', normal ).
adj( 'tallish', '-', '-', normal ).
adj( 'tamable', '-', '-', normal ).
adj( 'tame', 'tamer', 'tamest', normal ).
adj( 'tan', '-', '-', normal ).
adj( 'tangerine', '-', '-', normal ).
adj( 'tangible', '-', '-', normal ).
adj( 'tangy', 'tangier', 'tangiest', normal ).
adj( 'tannic', '-', '-', normal ).
adj( 'tantamount', '-', '-', normal ).
adj( 'tapestried', '-', '-', normal ).
adj( 'tardy', 'tardier', 'tardiest', normal ).
adj( 'tarry', '-', '-', normal ).
adj( 'tarsal', '-', '-', normal ).
adj( 'tart', '-', '-', normal ).
adj( 'tartaric', '-', '-', normal ).
adj( 'tasseled', '-', '-', normal ).
adj( 'tasselled', '-', '-', normal ).
adj( 'tasteful', '-', '-', normal ).
adj( 'tasteless', '-', '-', normal ).
adj( 'tasty', 'tastier', 'tastiest', normal ).
adj( 'tattered', '-', '-', normal ).
adj( 'tatty', 'tattier', 'tattiest', normal ).
adj( 'taut', 'tauter', 'tautest', normal ).
adj( 'tautological', '-', '-', normal ).
adj( 'tawdry', 'tawdrier', 'tawdriest', normal ).
adj( 'tawny', '-', '-', normal ).
adj( 'tax-free', '-', '-', normal ).
adj( 'taxable', '-', '-', normal ).
adj( 'teachable', '-', '-', normal ).
adj( 'tearaway', '-', '-', normal ).
adj( 'tearful', '-', '-', normal ).
adj( 'tearing', '-', '-', attr ).
adj( 'tearless', '-', '-', normal ).
adj( 'technical', '-', '-', normal ).
adj( 'technological', '-', '-', normal ).
adj( 'techy', '-', '-', normal ).
adj( 'tedious', '-', '-', normal ).
adj( 'teenage', '-', '-', normal ).
adj( 'teeny', 'teenier', 'teeniest', normal ).
adj( 'teetotal', '-', '-', normal ).
adj( 'telegraphic', '-', '-', normal ).
adj( 'teleological', '-', '-', normal ).
adj( 'telepathic', '-', '-', normal ).
adj( 'telephonic', '-', '-', normal ).
adj( 'telescopic', '-', '-', normal ).
adj( 'telling', '-', '-', normal ).
adj( 'temperamental', '-', '-', normal ).
adj( 'temperate', '-', '-', normal ).
adj( 'tempered', '-', '-', affix ).
adj( 'tempest-swept', '-', '-', normal ).
adj( 'tempest-tossed', '-', '-', normal ).
adj( 'tempestuous', '-', '-', normal ).
adj( 'temporal', '-', '-', normal ).
adj( 'temporary', '-', '-', normal ).
adj( 'ten', '-', '-', normal ).
adj( 'tenable', '-', '-', normal ).
adj( 'tenacious', '-', '-', normal ).
adj( 'tendentious', '-', '-', normal ).
adj( 'tender', 'tenderer', 'tenderest', normal ).
adj( 'tenderhearted', '-', '-', normal ).
adj( 'tense', 'tenser', 'tensest', normal ).
adj( 'tensile', '-', '-', normal ).
adj( 'tentative', '-', '-', normal ).
adj( 'tenth', '-', '-', normal ).
adj( 'tenuous', '-', '-', normal ).
adj( 'tepid', '-', '-', normal ).
adj( 'terminable', '-', '-', normal ).
adj( 'terminal', '-', '-', normal ).
adj( 'terminological', '-', '-', normal ).
adj( 'terrestrial', '-', '-', normal ).
adj( 'terrible', '-', '-', normal ).
adj( 'terrific', '-', '-', normal ).
adj( 'territorial', '-', '-', normal ).
adj( 'terror-stricken', '-', '-', normal ).
adj( 'terror-struck', '-', '-', normal ).
adj( 'terse', 'terser', 'tersest', normal ).
adj( 'tertian', '-', '-', normal ).
adj( 'tertiary', '-', '-', normal ).
adj( 'tessellated', '-', '-', normal ).
adj( 'testamentary', '-', '-', normal ).
adj( 'testate', '-', '-', normal ).
adj( 'testy', 'testier', 'testiest', normal ).
adj( 'tetchy', 'tetchier', 'tetchiest', normal ).
adj( 'textile', '-', '-', attr ).
adj( 'textual', '-', '-', normal ).
adj( 'textured', '-', '-', normal ).
adj( 'thankful', '-', '-', normal ).
adj( 'thankless', '-', '-', normal ).
adj( 'that', '-', '-', normal ).
adj( 'theatrical', '-', '-', normal ).
adj( 'their', '-', '-', normal ).
adj( 'theistic', '-', '-', normal ).
adj( 'theistical', '-', '-', normal ).
adj( 'thematic', '-', '-', normal ).
adj( 'theocratic', '-', '-', normal ).
adj( 'theological', '-', '-', normal ).
adj( 'theoretic', '-', '-', normal ).
adj( 'theoretical', '-', '-', normal ).
adj( 'theosophical', '-', '-', normal ).
adj( 'therapeutic', '-', '-', normal ).
adj( 'therapeutical', '-', '-', normal ).
adj( 'thermal', '-', '-', normal ).
adj( 'thermionic', '-', '-', normal ).
adj( 'thermonuclear', '-', '-', normal ).
adj( 'thermoplastic', '-', '-', normal ).
adj( 'thermosetting', '-', '-', normal ).
adj( 'thermostatic', '-', '-', normal ).
adj( 'these', '-', '-', normal ).
adj( 'thick', 'thicker', 'thickest', normal ).
adj( 'thick-headed', '-', '-', normal ).
adj( 'thick-set', '-', '-', normal ).
adj( 'thick-skinned', '-', '-', normal ).
adj( 'thievish', '-', '-', normal ).
adj( 'thin', 'thinner', 'thinnest', normal ).
adj( 'thin-skinned', '-', '-', normal ).
adj( 'thine', '-', '-', normal ).
adj( 'thinkable', '-', '-', normal ).
adj( 'thinking', '-', '-', normal ).
adj( 'third', '-', '-', normal ).
adj( 'third-rate', '-', '-', normal ).
adj( 'thirsty', 'thirstier', 'thirstiest', normal ).
adj( 'thirteen', '-', '-', normal ).
adj( 'thirteenth', '-', '-', normal ).
adj( 'thirtieth', '-', '-', normal ).
adj( 'thirty', '-', '-', normal ).
adj( 'this', '-', '-', normal ).
adj( 'thorny', 'thornier', 'thorniest', normal ).
adj( 'thorough', '-', '-', normal ).
adj( 'thoroughbred', '-', '-', normal ).
adj( 'thoroughgoing', '-', '-', normal ).
adj( 'those', '-', '-', normal ).
adj( 'thoughtful', '-', '-', normal ).
adj( 'thoughtless', '-', '-', normal ).
adj( 'thousand', '-', '-', normal ).
adj( 'thousandfold', '-', '-', normal ).
adj( 'thousandth', '-', '-', normal ).
adj( 'threadbare', '-', '-', normal ).
adj( 'threadlike', '-', '-', normal ).
adj( 'three', '-', '-', normal ).
adj( 'three-d', '-', '-', normal ).
adj( 'three-cornered', '-', '-', normal ).
adj( 'three-dimensional', '-', '-', normal ).
adj( 'three-figure', '-', '-', normal ).
adj( 'three-funnelled', '-', '-', normal ).
adj( 'three-lane', '-', '-', normal ).
adj( 'three-legged', '-', '-', normal ).
adj( 'three-piece', '-', '-', normal ).
adj( 'three-ply', '-', '-', normal ).
adj( 'three-quarter', '-', '-', normal ).
adj( 'three-score', '-', '-', normal ).
adj( 'three-storey', '-', '-', normal ).
adj( 'three-storeyed', '-', '-', normal ).
adj( 'three-wheeled', '-', '-', normal ).
adj( 'threefold', '-', '-', normal ).
adj( 'threepenny', '-', '-', normal ).
adj( 'thriftless', '-', '-', normal ).
adj( 'thrifty', 'thriftier', 'thriftiest', normal ).
adj( 'throated', '-', '-', affix ).
adj( 'throaty', 'throatier', 'throatiest', normal ).
adj( 'throbbing', '-', '-', normal ).
adj( 'thumping', '-', '-', normal ).
adj( 'thunderous', '-', '-', normal ).
adj( 'thunderstruck', '-', '-', normal ).
adj( 'thundery', '-', '-', normal ).
adj( 'thy', '-', '-', normal ).
adj( 'ticklish', '-', '-', normal ).
adj( 'tidal', '-', '-', normal ).
adj( 'tiddley', '-', '-', normal ).
adj( 'tidy', 'tidier', 'tidiest', normal ).
adj( 'tie-on', '-', '-', attr ).
adj( 'tigerish', '-', '-', normal ).
adj( 'tight', 'tighter', 'tightest', normal ).
adj( 'tight-laced', '-', '-', normal ).
adj( 'tight-lipped', '-', '-', normal ).
adj( 'timbered', '-', '-', normal ).
adj( 'time-expired', '-', '-', normal ).
adj( 'time-honoured', '-', '-', normal ).
adj( 'timeless', '-', '-', normal ).
adj( 'timely', 'timelier', 'timeliest', normal ).
adj( 'timesaving', '-', '-', normal ).
adj( 'timeserving', '-', '-', normal ).
adj( 'timid', '-', '-', normal ).
adj( 'timorous', '-', '-', normal ).
adj( 'tined', '-', '-', affix ).
adj( 'tinny', 'tinnier', 'tinniest', normal ).
adj( 'tinselly', '-', '-', normal ).
adj( 'tiny', 'tinier', 'tiniest', normal ).
adj( 'tip-and-run', '-', '-', normal ).
adj( 'tip-top', '-', '-', normal ).
adj( 'tip-up', '-', '-', normal ).
adj( 'tipsy', '-', '-', normal ).
adj( 'tired', '-', '-', normal ).
adj( 'tireless', '-', '-', normal ).
adj( 'tiresome', '-', '-', normal ).
adj( 'titanic', '-', '-', normal ).
adj( 'titled', '-', '-', normal ).
adj( 'titular', '-', '-', normal ).
adj( 'toilsome', '-', '-', normal ).
adj( 'tolerable', '-', '-', normal ).
adj( 'tolerant', '-', '-', normal ).
adj( 'tonal', '-', '-', normal ).
adj( 'tone-deaf', '-', '-', normal ).
adj( 'toned', '-', '-', affix ).
adj( 'toneless', '-', '-', normal ).
adj( 'tongue-in-cheek', '-', '-', normal ).
adj( 'tongue-tied', '-', '-', normal ).
adj( 'tongued', '-', '-', affix ).
adj( 'tonic', '-', '-', normal ).
adj( 'tonsorial', '-', '-', normal ).
adj( 'toothed', '-', '-', affix ).
adj( 'toothless', '-', '-', normal ).
adj( 'toothsome', '-', '-', normal ).
adj( 'top-flight', '-', '-', attr ).
adj( 'top-heavy', '-', '-', normal ).
adj( 'top-hole', '-', '-', normal ).
adj( 'top-ranking', '-', '-', normal ).
adj( 'topgallant', '-', '-', normal ).
adj( 'topical', '-', '-', normal ).
adj( 'topless', '-', '-', normal ).
adj( 'topmost', '-', '-', normal ).
adj( 'topnotch', '-', '-', attr ).
adj( 'topographical', '-', '-', normal ).
adj( 'topping', '-', '-', normal ).
adj( 'topsy-turvy', '-', '-', normal ).
adj( 'torpid', '-', '-', normal ).
adj( 'torrential', '-', '-', normal ).
adj( 'torrid', '-', '-', normal ).
adj( 'tortuous', '-', '-', normal ).
adj( 'total', '-', '-', normal ).
adj( 'totalitarian', '-', '-', normal ).
adj( 'tottery', '-', '-', normal ).
adj( 'touch-and-go', '-', '-', normal ).
adj( 'touchable', '-', '-', normal ).
adj( 'touched', '-', '-', normal ).
adj( 'touching', '-', '-', normal ).
adj( 'touchy', 'touchier', 'touchiest', normal ).
adj( 'tough', 'tougher', 'toughest', normal ).
adj( 'touring', '-', '-', normal ).
adj( 'towering', '-', '-', normal ).
adj( 'toxic', '-', '-', normal ).
adj( 'traceable', '-', '-', normal ).
adj( 'tracked', '-', '-', normal ).
adj( 'trackless', '-', '-', normal ).
adj( 'tractable', '-', '-', normal ).
adj( 'traditional', '-', '-', normal ).
adj( 'tragic', '-', '-', normal ).
adj( 'tragicomic', '-', '-', normal ).
adj( 'traitorous', '-', '-', normal ).
adj( 'tranquil', '-', '-', normal ).
adj( 'transalpine', '-', '-', normal ).
adj( 'transatlantic', '-', '-', normal ).
adj( 'transcendent', '-', '-', normal ).
adj( 'transcendental', '-', '-', normal ).
adj( 'transcontinental', '-', '-', normal ).
adj( 'transferable', '-', '-', normal ).
adj( 'transformable', '-', '-', normal ).
adj( 'transient', '-', '-', normal ).
adj( 'transistorized', '-', '-', normal ).
adj( 'transitional', '-', '-', normal ).
adj( 'transitive', '-', '-', normal ).
adj( 'transitory', '-', '-', normal ).
adj( 'translatable', '-', '-', normal ).
adj( 'translucent', '-', '-', normal ).
adj( 'transmutable', '-', '-', normal ).
adj( 'transoceanic', '-', '-', normal ).
adj( 'transparent', '-', '-', normal ).
adj( 'transpolar', '-', '-', normal ).
adj( 'transportable', '-', '-', normal ).
adj( 'transverse', '-', '-', normal ).
adj( 'trashy', 'trashier', 'trashiest', normal ).
adj( 'traumatic', '-', '-', normal ).
adj( 'travel-soiled', '-', '-', normal ).
adj( 'travel-stained', '-', '-', normal ).
adj( 'travel-worn', '-', '-', normal ).
adj( 'traveled', '-', '-', normal ).
adj( 'travelled', '-', '-', normal ).
adj( 'treacherous', '-', '-', normal ).
adj( 'treacly', '-', '-', normal ).
adj( 'treasonable', '-', '-', normal ).
adj( 'treasonous', '-', '-', normal ).
adj( 'treble', '-', '-', normal ).
adj( 'treeless', '-', '-', normal ).
adj( 'tremendous', '-', '-', normal ).
adj( 'tremulous', '-', '-', normal ).
adj( 'trenchant', '-', '-', normal ).
adj( 'trendy', 'trendier', 'trendiest', normal ).
adj( 'triangular', '-', '-', normal ).
adj( 'tribal', '-', '-', normal ).
adj( 'tributary', '-', '-', normal ).
adj( 'tricksy', '-', '-', normal ).
adj( 'tricky', 'trickier', 'trickiest', normal ).
adj( 'tried', '-', '-', normal ).
adj( 'triennial', '-', '-', normal ).
adj( 'trifling', '-', '-', normal ).
adj( 'trigger-happy', '-', '-', normal ).
adj( 'trilateral', '-', '-', normal ).
adj( 'trillion', '-', '-', normal ).
adj( 'trillionth', '-', '-', normal ).
adj( 'trim', 'trimmer', 'trimmest', normal ).
adj( 'tripartite', '-', '-', normal ).
adj( 'triple', '-', '-', normal ).
adj( 'triplex', '-', '-', normal ).
adj( 'triplicate', '-', '-', normal ).
adj( 'tripping', '-', '-', normal ).
adj( 'trite', '-', '-', normal ).
adj( 'triumphal', '-', '-', normal ).
adj( 'triumphant', '-', '-', normal ).
adj( 'triune', '-', '-', normal ).
adj( 'trivial', '-', '-', normal ).
adj( 'trochaic', '-', '-', normal ).
adj( 'tropical', '-', '-', normal ).
adj( 'troublesome', '-', '-', normal ).
adj( 'troublous', '-', '-', normal ).
adj( 'truculent', '-', '-', normal ).
adj( 'true', 'truer', 'truest', normal ).
adj( 'true-blue', '-', '-', normal ).
adj( 'true-hearted', '-', '-', normal ).
adj( 'trumpery', '-', '-', normal ).
adj( 'trustful', '-', '-', normal ).
adj( 'trusting', '-', '-', normal ).
adj( 'trustworthy', '-', '-', normal ).
adj( 'trusty', 'trustier', 'trustiest', normal ).
adj( 'truthful', '-', '-', normal ).
adj( 'trying', '-', '-', normal ).
adj( 'tubby', 'tubbier', 'tubbiest', normal ).
adj( 'tubeless', '-', '-', normal ).
adj( 'tubercular', '-', '-', normal ).
adj( 'tuberculous', '-', '-', normal ).
adj( 'tubular', '-', '-', normal ).
adj( 'tufted', '-', '-', normal ).
adj( 'tumble-down', '-', '-', attr ).
adj( 'tumescent', '-', '-', normal ).
adj( 'tumid', '-', '-', normal ).
adj( 'tumultuous', '-', '-', normal ).
adj( 'tuneful', '-', '-', normal ).
adj( 'tuppenny', '-', '-', normal ).
adj( 'turbaned', '-', '-', normal ).
adj( 'turbid', '-', '-', normal ).
adj( 'turbulent', '-', '-', normal ).
adj( 'turgid', '-', '-', normal ).
adj( 'turtleneck', '-', '-', normal ).
adj( 'turtlenecked', '-', '-', normal ).
adj( 'tutelary', '-', '-', normal ).
adj( 'tutorial', '-', '-', normal ).
adj( 'twee', '-', '-', normal ).
adj( 'twelfth', '-', '-', normal ).
adj( 'twelve', '-', '-', normal ).
adj( 'twentieth', '-', '-', normal ).
adj( 'twenty', '-', '-', normal ).
adj( 'twiddly', '-', '-', normal ).
adj( 'twiggy', 'twiggier', 'twiggiest', normal ).
adj( 'twilit', '-', '-', normal ).
adj( 'twilled', '-', '-', normal ).
adj( 'twinned', '-', '-', attr ).
adj( 'twisty', 'twistier', 'twistiest', normal ).
adj( 'two', '-', '-', normal ).
adj( 'two-a-penny', '-', '-', normal ).
adj( 'two-edged', '-', '-', normal ).
adj( 'two-faced', '-', '-', normal ).
adj( 'two-funnelled', '-', '-', normal ).
adj( 'two-handed', '-', '-', normal ).
adj( 'two-ply', '-', '-', normal ).
adj( 'two-timing', '-', '-', normal ).
adj( 'two-way', '-', '-', attr ).
adj( 'twofold', '-', '-', normal ).
adj( 'twopenny', '-', '-', normal ).
adj( 'twopenny-halfpenny', '-', '-', normal ).
adj( 'typewritten', '-', '-', normal ).
adj( 'typical', '-', '-', normal ).
adj( 'typographic', '-', '-', normal ).
adj( 'tyrannical', '-', '-', normal ).
adj( 'tyrannous', '-', '-', normal ).
adj( 'ubiquitous', '-', '-', normal ).
adj( 'ugly', 'uglier', 'ugliest', normal ).
adj( 'ulcerous', '-', '-', normal ).
adj( 'ulterior', '-', '-', normal ).
adj( 'ultimate', '-', '-', normal ).
adj( 'ultimo', '-', '-', normal ).
adj( 'ultra vires', '-', '-', normal ).
adj( 'ultramarine', '-', '-', normal ).
adj( 'ultramontane', '-', '-', normal ).
adj( 'ultrasonic', '-', '-', normal ).
adj( 'ultraviolet', '-', '-', normal ).
adj( 'umber', '-', '-', normal ).
adj( 'umbilical', '-', '-', normal ).
adj( 'umpteen', '-', '-', normal ).
adj( 'umpteenth', '-', '-', normal ).
adj( 'un-come-at-able', '-', '-', normal ).
adj( 'un-get-at-able', '-', '-', normal ).
adj( 'unabashed', '-', '-', normal ).
adj( 'unabated', '-', '-', normal ).
adj( 'unable', '-', '-', pred ).
adj( 'unabridged', '-', '-', normal ).
adj( 'unacceptable', '-', '-', normal ).
adj( 'unaccommodating', '-', '-', normal ).
adj( 'unaccompanied', '-', '-', normal ).
adj( 'unaccountable', '-', '-', normal ).
adj( 'unaccounted', '-', '-', normal ).
adj( 'unaccustomed', '-', '-', normal ).
adj( 'unacknowledged', '-', '-', normal ).
adj( 'unacquainted', '-', '-', normal ).
adj( 'unadorned', '-', '-', normal ).
adj( 'unadulterated', '-', '-', normal ).
adj( 'unadventurous', '-', '-', normal ).
adj( 'unadvised', '-', '-', normal ).
adj( 'unaffected', '-', '-', normal ).
adj( 'unafraid', '-', '-', pred ).
adj( 'unaided', '-', '-', normal ).
adj( 'unalienable', '-', '-', normal ).
adj( 'unaligned', '-', '-', normal ).
adj( 'unalloyed', '-', '-', normal ).
adj( 'unalterable', '-', '-', normal ).
adj( 'unaltered', '-', '-', normal ).
adj( 'unambiguous', '-', '-', normal ).
adj( 'unamended', '-', '-', normal ).
adj( 'unanimous', '-', '-', normal ).
adj( 'unannounced', '-', '-', normal ).
adj( 'unanswerable', '-', '-', normal ).
adj( 'unanswered', '-', '-', normal ).
adj( 'unanticipated', '-', '-', normal ).
adj( 'unappealing', '-', '-', normal ).
adj( 'unappetizing', '-', '-', normal ).
adj( 'unappreciated', '-', '-', normal ).
adj( 'unappreciative', '-', '-', normal ).
adj( 'unapproachable', '-', '-', normal ).
adj( 'unarguable', '-', '-', normal ).
adj( 'unarmed', '-', '-', normal ).
adj( 'unarticulated', '-', '-', normal ).
adj( 'unashamed', '-', '-', pred ).
adj( 'unasked', '-', '-', normal ).
adj( 'unassailable', '-', '-', normal ).
adj( 'unassisted', '-', '-', normal ).
adj( 'unassuming', '-', '-', normal ).
adj( 'unattached', '-', '-', normal ).
adj( 'unattainable', '-', '-', normal ).
adj( 'unattended', '-', '-', normal ).
adj( 'unattractive', '-', '-', normal ).
adj( 'unauthorized', '-', '-', normal ).
adj( 'unavailable', '-', '-', normal ).
adj( 'unavailing', '-', '-', normal ).
adj( 'unavoidable', '-', '-', normal ).
adj( 'unawakened', '-', '-', normal ).
adj( 'unaware', '-', '-', pred ).
adj( 'unbacked', '-', '-', normal ).
adj( 'unbalanced', '-', '-', normal ).
adj( 'unbalanced', '-', '-', normal ).
adj( 'unbearable', '-', '-', normal ).
adj( 'unbeatable', '-', '-', normal ).
adj( 'unbeaten', '-', '-', normal ).
adj( 'unbecoming', '-', '-', normal ).
adj( 'unbeknown', '-', '-', normal ).
adj( 'unbeknownst', '-', '-', normal ).
adj( 'unbelievable', '-', '-', normal ).
adj( 'unbelieving', '-', '-', normal ).
adj( 'unbeloved', '-', '-', normal ).
adj( 'unbending', '-', '-', normal ).
adj( 'unbiased', '-', '-', normal ).
adj( 'unbiassed', '-', '-', normal ).
adj( 'unbidden', '-', '-', normal ).
adj( 'unblushing', '-', '-', normal ).
adj( 'unborn', '-', '-', normal ).
adj( 'unbounded', '-', '-', normal ).
adj( 'unbowed', '-', '-', normal ).
adj( 'unbreakable', '-', '-', normal ).
adj( 'unbridled', '-', '-', normal ).
adj( 'unbroken', '-', '-', normal ).
adj( 'unburied', '-', '-', normal ).
adj( 'unbuttoned', '-', '-', normal ).
adj( 'unbuttoned', '-', '-', normal ).
adj( 'uncalled-for', '-', '-', normal ).
adj( 'uncanny', '-', '-', normal ).
adj( 'uncared-for', '-', '-', normal ).
adj( 'uncarpeted', '-', '-', normal ).
adj( 'unceasing', '-', '-', normal ).
adj( 'uncensored', '-', '-', normal ).
adj( 'unceremonious', '-', '-', normal ).
adj( 'uncertain', '-', '-', normal ).
adj( 'unchallengeable', '-', '-', normal ).
adj( 'unchallenged', '-', '-', normal ).
adj( 'unchanged', '-', '-', normal ).
adj( 'unchanging', '-', '-', normal ).
adj( 'uncharacteristic', '-', '-', normal ).
adj( 'uncharitable', '-', '-', normal ).
adj( 'uncharted', '-', '-', normal ).
adj( 'unchecked', '-', '-', normal ).
adj( 'unchristian', '-', '-', normal ).
adj( 'uncivil', '-', '-', normal ).
adj( 'uncivilized', '-', '-', normal ).
adj( 'unclaimed', '-', '-', normal ).
adj( 'unclassified', '-', '-', normal ).
adj( 'unclean', '-', '-', normal ).
adj( 'unclear', '-', '-', normal ).
adj( 'unclouded', '-', '-', normal ).
adj( 'uncluttered', '-', '-', normal ).
adj( 'unco', '-', '-', normal ).
adj( 'unco-operative', '-', '-', normal ).
adj( 'unco-ordinated', '-', '-', normal ).
adj( 'uncoloured', '-', '-', normal ).
adj( 'uncomfortable', '-', '-', normal ).
adj( 'uncommercialized', '-', '-', normal ).
adj( 'uncommitted', '-', '-', normal ).
adj( 'uncommon', '-', '-', normal ).
adj( 'uncommunicative', '-', '-', normal ).
adj( 'uncompetitive', '-', '-', normal ).
adj( 'uncomplaining', '-', '-', normal ).
adj( 'uncompleted', '-', '-', normal ).
adj( 'uncomplicated', '-', '-', normal ).
adj( 'uncomplimentary', '-', '-', normal ).
adj( 'uncomprehending', '-', '-', normal ).
adj( 'uncompromising', '-', '-', normal ).
adj( 'unconcealed', '-', '-', normal ).
adj( 'unconcerned', '-', '-', normal ).
adj( 'unconditional', '-', '-', normal ).
adj( 'unconditioned', '-', '-', normal ).
adj( 'unconfined', '-', '-', normal ).
adj( 'unconfirmed', '-', '-', normal ).
adj( 'unconformable', '-', '-', normal ).
adj( 'uncongenial', '-', '-', normal ).
adj( 'unconnected', '-', '-', normal ).
adj( 'unconquered', '-', '-', normal ).
adj( 'unconscionable', '-', '-', normal ).
adj( 'unconscious', '-', '-', normal ).
adj( 'unconsidered', '-', '-', normal ).
adj( 'unconstitutional', '-', '-', normal ).
adj( 'unconstrained', '-', '-', normal ).
adj( 'unconstructive', '-', '-', normal ).
adj( 'unconsummated', '-', '-', normal ).
adj( 'uncontaminated', '-', '-', normal ).
adj( 'uncontrollable', '-', '-', normal ).
adj( 'uncontrolled', '-', '-', normal ).
adj( 'uncontroversial', '-', '-', normal ).
adj( 'unconventional', '-', '-', normal ).
adj( 'unconverted', '-', '-', normal ).
adj( 'unconvinced', '-', '-', normal ).
adj( 'unconvincing', '-', '-', normal ).
adj( 'uncooked', '-', '-', normal ).
adj( 'uncorrected', '-', '-', normal ).
adj( 'uncorrelated', '-', '-', normal ).
adj( 'uncorroborated', '-', '-', normal ).
adj( 'uncouth', '-', '-', normal ).
adj( 'uncritical', '-', '-', normal ).
adj( 'uncrossed', '-', '-', normal ).
adj( 'uncrossed', '-', '-', normal ).
adj( 'uncrowned', '-', '-', normal ).
adj( 'unctuous', '-', '-', normal ).
adj( 'uncultivated', '-', '-', normal ).
adj( 'uncultured', '-', '-', normal ).
adj( 'uncut', '-', '-', normal ).
adj( 'undamaged', '-', '-', normal ).
adj( 'undated', '-', '-', normal ).
adj( 'undaunted', '-', '-', normal ).
adj( 'undecided', '-', '-', normal ).
adj( 'undeclared', '-', '-', normal ).
adj( 'undefeated', '-', '-', normal ).
adj( 'undefended', '-', '-', normal ).
adj( 'undeferential', '-', '-', normal ).
adj( 'undefinable', '-', '-', normal ).
adj( 'undefined', '-', '-', normal ).
adj( 'undemanding', '-', '-', normal ).
adj( 'undemocratic', '-', '-', normal ).
adj( 'undemonstrative', '-', '-', normal ).
adj( 'undeniable', '-', '-', normal ).
adj( 'undenominational', '-', '-', normal ).
adj( 'under-the-counter', '-', '-', normal ).
adj( 'underarm', '-', '-', normal ).
adj( 'underbred', '-', '-', normal ).
adj( 'undercover', '-', '-', normal ).
adj( 'underdeveloped', '-', '-', normal ).
adj( 'underdone', '-', '-', normal ).
adj( 'underemployed', '-', '-', normal ).
adj( 'underfed', '-', '-', normal ).
adj( 'underfloor', '-', '-', normal ).
adj( 'underground', '-', '-', attr ).
adj( 'underhand', '-', '-', normal ).
adj( 'underhanded', '-', '-', normal ).
adj( 'underhung', '-', '-', normal ).
adj( 'undermanned', '-', '-', normal ).
adj( 'undermentioned', '-', '-', normal ).
adj( 'undernourished', '-', '-', normal ).
adj( 'underpopulated', '-', '-', normal ).
adj( 'underprivileged', '-', '-', normal ).
adj( 'undersea', '-', '-', normal ).
adj( 'undersealed', '-', '-', normal ).
adj( 'undersexed', '-', '-', normal ).
adj( 'undersized', '-', '-', normal ).
adj( 'underslung', '-', '-', normal ).
adj( 'underspent', '-', '-', normal ).
adj( 'understaffed', '-', '-', normal ).
adj( 'understandable', '-', '-', normal ).
adj( 'understanding', '-', '-', normal ).
adj( 'underwater', '-', '-', normal ).
adj( 'underweight', '-', '-', normal ).
adj( 'undeserved', '-', '-', normal ).
adj( 'undeserving', '-', '-', normal ).
adj( 'undesigned', '-', '-', normal ).
adj( 'undesirable', '-', '-', normal ).
adj( 'undetected', '-', '-', normal ).
adj( 'undetermined', '-', '-', normal ).
adj( 'undeterred', '-', '-', normal ).
adj( 'undeveloped', '-', '-', normal ).
adj( 'undiagnosed', '-', '-', normal ).
adj( 'undifferentiated', '-', '-', normal ).
adj( 'undigested', '-', '-', normal ).
adj( 'undignified', '-', '-', normal ).
adj( 'undiluted', '-', '-', normal ).
adj( 'undiminished', '-', '-', normal ).
adj( 'undimmed', '-', '-', normal ).
adj( 'undiplomatic', '-', '-', normal ).
adj( 'undischarged', '-', '-', normal ).
adj( 'undisciplined', '-', '-', normal ).
adj( 'undisclosed', '-', '-', normal ).
adj( 'undiscovered', '-', '-', normal ).
adj( 'undiscriminating', '-', '-', normal ).
adj( 'undisguised', '-', '-', normal ).
adj( 'undismayed', '-', '-', normal ).
adj( 'undisputed', '-', '-', normal ).
adj( 'undissolved', '-', '-', normal ).
adj( 'undistinguishable', '-', '-', normal ).
adj( 'undistinguished', '-', '-', normal ).
adj( 'undistributed', '-', '-', normal ).
adj( 'undisturbed', '-', '-', normal ).
adj( 'undivided', '-', '-', normal ).
adj( 'undocumented', '-', '-', normal ).
adj( 'undogmatic', '-', '-', normal ).
adj( 'undomesticated', '-', '-', normal ).
adj( 'undone', '-', '-', pred ).
adj( 'undoubted', '-', '-', normal ).
adj( 'undramatic', '-', '-', normal ).
adj( 'undreamed', '-', '-', normal ).
adj( 'undreamed-of', '-', '-', normal ).
adj( 'undreamt', '-', '-', normal ).
adj( 'undrinkable', '-', '-', normal ).
adj( 'undue', '-', '-', normal ).
adj( 'undying', '-', '-', normal ).
adj( 'unearned', '-', '-', normal ).
adj( 'unearthly', '-', '-', normal ).
adj( 'uneasy', '-', '-', normal ).
adj( 'uneatable', '-', '-', normal ).
adj( 'uneaten', '-', '-', normal ).
adj( 'uneconomic', '-', '-', normal ).
adj( 'uneconomical', '-', '-', normal ).
adj( 'unedifying', '-', '-', normal ).
adj( 'unedited', '-', '-', normal ).
adj( 'uneducated', '-', '-', normal ).
adj( 'uneffective', '-', '-', normal ).
adj( 'unelaborated', '-', '-', normal ).
adj( 'unemotional', '-', '-', normal ).
adj( 'unemployable', '-', '-', normal ).
adj( 'unemployed', '-', '-', normal ).
adj( 'unending', '-', '-', normal ).
adj( 'unendurable', '-', '-', normal ).
adj( 'unenlightened', '-', '-', normal ).
adj( 'unenterprising', '-', '-', normal ).
adj( 'unenthusiastic', '-', '-', normal ).
adj( 'unenviable', '-', '-', normal ).
adj( 'unequal', '-', '-', normal ).
adj( 'unequalled', '-', '-', normal ).
adj( 'unequipped', '-', '-', normal ).
adj( 'unequivocal', '-', '-', normal ).
adj( 'unerring', '-', '-', normal ).
adj( 'unethical', '-', '-', normal ).
adj( 'uneven', '-', '-', normal ).
adj( 'uneventful', '-', '-', normal ).
adj( 'unexampled', '-', '-', normal ).
adj( 'unexceeded', '-', '-', normal ).
adj( 'unexceptionable', '-', '-', normal ).
adj( 'unexceptional', '-', '-', normal ).
adj( 'unexciting', '-', '-', normal ).
adj( 'unexpected', '-', '-', normal ).
adj( 'unexpired', '-', '-', normal ).
adj( 'unexplained', '-', '-', normal ).
adj( 'unexplored', '-', '-', normal ).
adj( 'unexposed', '-', '-', normal ).
adj( 'unexpressed', '-', '-', normal ).
adj( 'unexpurgated', '-', '-', normal ).
adj( 'unfailing', '-', '-', normal ).
adj( 'unfair', '-', '-', normal ).
adj( 'unfaithful', '-', '-', normal ).
adj( 'unfaltering', '-', '-', normal ).
adj( 'unfamiliar', '-', '-', normal ).
adj( 'unfashionable', '-', '-', normal ).
adj( 'unfathomable', '-', '-', normal ).
adj( 'unfathomed', '-', '-', normal ).
adj( 'unfavourable', '-', '-', normal ).
adj( 'unfed', '-', '-', normal ).
adj( 'unfeeling', '-', '-', normal ).
adj( 'unfeigned', '-', '-', normal ).
adj( 'unfermented', '-', '-', normal ).
adj( 'unfertilized', '-', '-', normal ).
adj( 'unfettered', '-', '-', normal ).
adj( 'unfinished', '-', '-', normal ).
adj( 'unfit', '-', '-', normal ).
adj( 'unflagging', '-', '-', normal ).
adj( 'unflappable', '-', '-', normal ).
adj( 'unfledged', '-', '-', normal ).
adj( 'unflinching', '-', '-', normal ).
adj( 'unflurried', '-', '-', normal ).
adj( 'unforeseeable', '-', '-', normal ).
adj( 'unforeseen', '-', '-', normal ).
adj( 'unforgettable', '-', '-', normal ).
adj( 'unforgivable', '-', '-', normal ).
adj( 'unforgiving', '-', '-', normal ).
adj( 'unformed', '-', '-', normal ).
adj( 'unforthcoming', '-', '-', normal ).
adj( 'unfortunate', '-', '-', normal ).
adj( 'unfounded', '-', '-', normal ).
adj( 'unframed', '-', '-', normal ).
adj( 'unfrequented', '-', '-', normal ).
adj( 'unfriendly', '-', '-', normal ).
adj( 'unfruitful', '-', '-', normal ).
adj( 'unfulfilled', '-', '-', normal ).
adj( 'unfurnished', '-', '-', normal ).
adj( 'ungainly', '-', '-', normal ).
adj( 'ungenerous', '-', '-', normal ).
adj( 'ungentle', '-', '-', normal ).
adj( 'ungentlemanly', '-', '-', normal ).
adj( 'unglazed', '-', '-', normal ).
adj( 'ungodly', '-', '-', normal ).
adj( 'ungovernable', '-', '-', normal ).
adj( 'ungraceful', '-', '-', normal ).
adj( 'ungracious', '-', '-', normal ).
adj( 'ungrammatical', '-', '-', normal ).
adj( 'ungrateful', '-', '-', normal ).
adj( 'ungrudging', '-', '-', normal ).
adj( 'unguarded', '-', '-', normal ).
adj( 'unhallowed', '-', '-', normal ).
adj( 'unhampered', '-', '-', normal ).
adj( 'unhappy', 'unhappier', 'unhappiest', normal ).
adj( 'unharmed', '-', '-', normal ).
adj( 'unhealed', '-', '-', normal ).
adj( 'unhealthy', '-', '-', normal ).
adj( 'unheard', '-', '-', normal ).
adj( 'unheard-of', '-', '-', normal ).
adj( 'unhearing', '-', '-', normal ).
adj( 'unheated', '-', '-', normal ).
adj( 'unheeded', '-', '-', normal ).
adj( 'unhelpful', '-', '-', normal ).
adj( 'unheralded', '-', '-', normal ).
adj( 'unhesitating', '-', '-', normal ).
adj( 'unhindered', '-', '-', normal ).
adj( 'unholy', '-', '-', normal ).
adj( 'unhoped-for', '-', '-', normal ).
adj( 'unhurried', '-', '-', normal ).
adj( 'unhurt', '-', '-', normal ).
adj( 'unhygienic', '-', '-', normal ).
adj( 'unidentifiable', '-', '-', normal ).
adj( 'unidentified', '-', '-', normal ).
adj( 'uniform', '-', '-', normal ).
adj( 'uniformed', '-', '-', normal ).
adj( 'unilateral', '-', '-', normal ).
adj( 'unilateralist', '-', '-', normal ).
adj( 'unimaginable', '-', '-', normal ).
adj( 'unimaginative', '-', '-', normal ).
adj( 'unimpaired', '-', '-', normal ).
adj( 'unimpeachable', '-', '-', normal ).
adj( 'unimpeded', '-', '-', normal ).
adj( 'unimportant', '-', '-', normal ).
adj( 'unimpressed', '-', '-', normal ).
adj( 'unimpressive', '-', '-', normal ).
adj( 'uninfluenced', '-', '-', normal ).
adj( 'uninformative', '-', '-', normal ).
adj( 'uninformed', '-', '-', normal ).
adj( 'uninhabitable', '-', '-', normal ).
adj( 'uninhabited', '-', '-', normal ).
adj( 'uninhibited', '-', '-', normal ).
adj( 'uninitiated', '-', '-', normal ).
adj( 'uninjured', '-', '-', normal ).
adj( 'uninspired', '-', '-', normal ).
adj( 'uninspiring', '-', '-', normal ).
adj( 'uninsured', '-', '-', normal ).
adj( 'unintelligent', '-', '-', normal ).
adj( 'unintelligible', '-', '-', normal ).
adj( 'unintended', '-', '-', normal ).
adj( 'unintentional', '-', '-', normal ).
adj( 'uninterested', '-', '-', normal ).
adj( 'uninteresting', '-', '-', normal ).
adj( 'uninterrupted', '-', '-', normal ).
adj( 'uninvited', '-', '-', normal ).
adj( 'uninviting', '-', '-', normal ).
adj( 'unique', '-', '-', normal ).
adj( 'unisex', '-', '-', normal ).
adj( 'unitary', '-', '-', normal ).
adj( 'united', '-', '-', normal ).
adj( 'universal', '-', '-', normal ).
adj( 'unjust', '-', '-', normal ).
adj( 'unjustifiable', '-', '-', normal ).
adj( 'unjustified', '-', '-', normal ).
adj( 'unkempt', '-', '-', normal ).
adj( 'unkind', '-', '-', normal ).
adj( 'unkissed', '-', '-', normal ).
adj( 'unknowable', '-', '-', normal ).
adj( 'unknowing', '-', '-', normal ).
adj( 'unknown', '-', '-', normal ).
adj( 'unlabelled', '-', '-', normal ).
adj( 'unladylike', '-', '-', normal ).
adj( 'unlamented', '-', '-', normal ).
adj( 'unlatched', '-', '-', normal ).
adj( 'unlawful', '-', '-', normal ).
adj( 'unleavened', '-', '-', normal ).
adj( 'unlettered', '-', '-', normal ).
adj( 'unlicensed', '-', '-', normal ).
adj( 'unlighted', '-', '-', normal ).
adj( 'unlikable', '-', '-', normal ).
adj( 'unlike', '-', '-', pred ).
adj( 'unlikely', '-', '-', normal ).
adj( 'unlimited', '-', '-', normal ).
adj( 'unlined', '-', '-', normal ).
adj( 'unlisted', '-', '-', normal ).
adj( 'unlit', '-', '-', normal ).
adj( 'unliterary', '-', '-', normal ).
adj( 'unlocated', '-', '-', normal ).
adj( 'unlooked-for', '-', '-', normal ).
adj( 'unlovable', '-', '-', normal ).
adj( 'unloved', '-', '-', normal ).
adj( 'unlovely', '-', '-', normal ).
adj( 'unlucky', '-', '-', normal ).
adj( 'unmade', '-', '-', normal ).
adj( 'unmanageable', '-', '-', normal ).
adj( 'unmanly', '-', '-', normal ).
adj( 'unmanned', '-', '-', normal ).
adj( 'unmannered', '-', '-', normal ).
adj( 'unmannerly', '-', '-', normal ).
adj( 'unmarked', '-', '-', normal ).
adj( 'unmarried', '-', '-', normal ).
adj( 'unmatchable', '-', '-', normal ).
adj( 'unmatched', '-', '-', normal ).
adj( 'unmechanized', '-', '-', normal ).
adj( 'unmemorable', '-', '-', normal ).
adj( 'unmentionable', '-', '-', normal ).
adj( 'unmerciful', '-', '-', normal ).
adj( 'unmerited', '-', '-', normal ).
adj( 'unmindful', '-', '-', normal ).
adj( 'unmingled', '-', '-', normal ).
adj( 'unmistakable', '-', '-', normal ).
adj( 'unmitigated', '-', '-', normal ).
adj( 'unmixed', '-', '-', normal ).
adj( 'unmodernized', '-', '-', normal ).
adj( 'unmodified', '-', '-', normal ).
adj( 'unmolested', '-', '-', normal ).
adj( 'unmourned', '-', '-', normal ).
adj( 'unmovable', '-', '-', normal ).
adj( 'unmoved', '-', '-', normal ).
adj( 'unmoving', '-', '-', normal ).
adj( 'unmusical', '-', '-', normal ).
adj( 'unnamed', '-', '-', normal ).
adj( 'unnatural', '-', '-', normal ).
adj( 'unnecessary', '-', '-', normal ).
adj( 'unnoticeable', '-', '-', normal ).
adj( 'unnoticed', '-', '-', normal ).
adj( 'unnumbered', '-', '-', normal ).
adj( 'unobjectionable', '-', '-', normal ).
adj( 'unobservant', '-', '-', normal ).
adj( 'unobserved', '-', '-', normal ).
adj( 'unobtainable', '-', '-', normal ).
adj( 'unobtrusive', '-', '-', normal ).
adj( 'unobvious', '-', '-', normal ).
adj( 'unoccupied', '-', '-', normal ).
adj( 'unofficial', '-', '-', normal ).
adj( 'unopen', '-', '-', normal ).
adj( 'unopened', '-', '-', normal ).
adj( 'unopposed', '-', '-', normal ).
adj( 'unoriginal', '-', '-', normal ).
adj( 'unorthodox', '-', '-', normal ).
adj( 'unpaid', '-', '-', normal ).
adj( 'unpainted', '-', '-', normal ).
adj( 'unpalatable', '-', '-', normal ).
adj( 'unparalleled', '-', '-', normal ).
adj( 'unpardonable', '-', '-', normal ).
adj( 'unparliamentary', '-', '-', normal ).
adj( 'unpatriotic', '-', '-', normal ).
adj( 'unpaved', '-', '-', normal ).
adj( 'unpersuaded', '-', '-', normal ).
adj( 'unperturbed', '-', '-', normal ).
adj( 'unplaced', '-', '-', normal ).
adj( 'unplanned', '-', '-', normal ).
adj( 'unplayable', '-', '-', normal ).
adj( 'unpleasant', '-', '-', normal ).
adj( 'unplumbed', '-', '-', normal ).
adj( 'unpolluted', '-', '-', normal ).
adj( 'unpompous', '-', '-', normal ).
adj( 'unpopular', '-', '-', normal ).
adj( 'unpractised', '-', '-', normal ).
adj( 'unprecedented', '-', '-', normal ).
adj( 'unpredictable', '-', '-', normal ).
adj( 'unpredicted', '-', '-', normal ).
adj( 'unprejudiced', '-', '-', normal ).
adj( 'unpremeditated', '-', '-', normal ).
adj( 'unprepared', '-', '-', normal ).
adj( 'unprepossessing', '-', '-', normal ).
adj( 'unpretentious', '-', '-', normal ).
adj( 'unprincipled', '-', '-', normal ).
adj( 'unprintable', '-', '-', normal ).
adj( 'unproductive', '-', '-', normal ).
adj( 'unprofessional', '-', '-', normal ).
adj( 'unprofitable', '-', '-', normal ).
adj( 'unpromising', '-', '-', normal ).
adj( 'unprompted', '-', '-', normal ).
adj( 'unpronounceable', '-', '-', normal ).
adj( 'unpropitious', '-', '-', normal ).
adj( 'unprotected', '-', '-', normal ).
adj( 'unproved', '-', '-', normal ).
adj( 'unproven', '-', '-', pred ).
adj( 'unprovided', '-', '-', normal ).
adj( 'unprovoked', '-', '-', normal ).
adj( 'unpublished', '-', '-', normal ).
adj( 'unpunished', '-', '-', normal ).
adj( 'unputdownable', '-', '-', normal ).
adj( 'unqualified', '-', '-', normal ).
adj( 'unquestionable', '-', '-', normal ).
adj( 'unquestioned', '-', '-', normal ).
adj( 'unquestioning', '-', '-', normal ).
adj( 'unquiet', '-', '-', normal ).
adj( 'unquotable', '-', '-', normal ).
adj( 'unreached', '-', '-', normal ).
adj( 'unread', '-', '-', normal ).
adj( 'unreadable', '-', '-', normal ).
adj( 'unready', '-', '-', normal ).
adj( 'unreal', '-', '-', normal ).
adj( 'unrealistic', '-', '-', normal ).
adj( 'unrealized', '-', '-', normal ).
adj( 'unreasonable', '-', '-', normal ).
adj( 'unreasoning', '-', '-', normal ).
adj( 'unreceptive', '-', '-', normal ).
adj( 'unreciprocated', '-', '-', normal ).
adj( 'unrecognizable', '-', '-', normal ).
adj( 'unrecognized', '-', '-', normal ).
adj( 'unreconciled', '-', '-', normal ).
adj( 'unrecorded', '-', '-', normal ).
adj( 'unredeemable', '-', '-', normal ).
adj( 'unrefined', '-', '-', normal ).
adj( 'unreflective', '-', '-', normal ).
adj( 'unregenerate', '-', '-', normal ).
adj( 'unregistered', '-', '-', normal ).
adj( 'unrehearsed', '-', '-', normal ).
adj( 'unrelated', '-', '-', normal ).
adj( 'unrelaxed', '-', '-', normal ).
adj( 'unrelenting', '-', '-', normal ).
adj( 'unreliable', '-', '-', normal ).
adj( 'unrelieved', '-', '-', normal ).
adj( 'unremarkable', '-', '-', normal ).
adj( 'unremitting', '-', '-', normal ).
adj( 'unrepeatable', '-', '-', normal ).
adj( 'unrepentant', '-', '-', normal ).
adj( 'unrepresentative', '-', '-', normal ).
adj( 'unrequested', '-', '-', normal ).
adj( 'unrequited', '-', '-', normal ).
adj( 'unreserved', '-', '-', normal ).
adj( 'unresisting', '-', '-', normal ).
adj( 'unresolved', '-', '-', normal ).
adj( 'unrestrained', '-', '-', normal ).
adj( 'unrestricted', '-', '-', normal ).
adj( 'unrevised', '-', '-', normal ).
adj( 'unrewarded', '-', '-', normal ).
adj( 'unrewarding', '-', '-', normal ).
adj( 'unrhythmical', '-', '-', normal ).
adj( 'unrighteous', '-', '-', normal ).
adj( 'unripe', '-', '-', normal ).
adj( 'unripened', '-', '-', normal ).
adj( 'unrivalled', '-', '-', normal ).
adj( 'unromantic', '-', '-', normal ).
adj( 'unruffled', '-', '-', normal ).
adj( 'unruly', 'unrulier', 'unruliest', normal ).
adj( 'unsaddled', '-', '-', normal ).
adj( 'unsafe', '-', '-', normal ).
adj( 'unsaid', '-', '-', normal ).
adj( 'unsaleable', '-', '-', normal ).
adj( 'unsalted', '-', '-', normal ).
adj( 'unsatisfactory', '-', '-', normal ).
adj( 'unsatisfied', '-', '-', normal ).
adj( 'unsatisfying', '-', '-', normal ).
adj( 'unsaturated', '-', '-', normal ).
adj( 'unsavoury', '-', '-', normal ).
adj( 'unscathed', '-', '-', normal ).
adj( 'unscheduled', '-', '-', normal ).
adj( 'unscientific', '-', '-', normal ).
adj( 'unscripted', '-', '-', normal ).
adj( 'unscrupulous', '-', '-', normal ).
adj( 'unsealed', '-', '-', normal ).
adj( 'unseasonable', '-', '-', normal ).
adj( 'unseasoned', '-', '-', normal ).
adj( 'unseaworthy', '-', '-', normal ).
adj( 'unsecured', '-', '-', normal ).
adj( 'unseeded', '-', '-', normal ).
adj( 'unseeing', '-', '-', normal ).
adj( 'unseemly', '-', '-', normal ).
adj( 'unseen', '-', '-', normal ).
adj( 'unselected', '-', '-', normal ).
adj( 'unselective', '-', '-', normal ).
adj( 'unselfconscious', '-', '-', normal ).
adj( 'unselfish', '-', '-', normal ).
adj( 'unsexed', '-', '-', normal ).
adj( 'unshakable', '-', '-', normal ).
adj( 'unshaved', '-', '-', normal ).
adj( 'unshaven', '-', '-', normal ).
adj( 'unshielded', '-', '-', normal ).
adj( 'unshrinkable', '-', '-', normal ).
adj( 'unshrinking', '-', '-', normal ).
adj( 'unshuttered', '-', '-', normal ).
adj( 'unsighted', '-', '-', normal ).
adj( 'unsightly', '-', '-', normal ).
adj( 'unsigned', '-', '-', normal ).
adj( 'unsilenced', '-', '-', normal ).
adj( 'unsinkable', '-', '-', normal ).
adj( 'unskilled', '-', '-', normal ).
adj( 'unsmiling', '-', '-', normal ).
adj( 'unsociable', '-', '-', normal ).
adj( 'unsocial', '-', '-', normal ).
adj( 'unsold', '-', '-', normal ).
adj( 'unsolicited', '-', '-', normal ).
adj( 'unsolved', '-', '-', normal ).
adj( 'unsophisticated', '-', '-', normal ).
adj( 'unsound', '-', '-', normal ).
adj( 'unsparing', '-', '-', normal ).
adj( 'unspeakable', '-', '-', normal ).
adj( 'unspecialized', '-', '-', normal ).
adj( 'unspecific', '-', '-', normal ).
adj( 'unspecified', '-', '-', normal ).
adj( 'unspectacular', '-', '-', normal ).
adj( 'unspoiled', '-', '-', normal ).
adj( 'unspoilt', '-', '-', normal ).
adj( 'unspoken', '-', '-', normal ).
adj( 'unsporting', '-', '-', normal ).
adj( 'unsportsmanlike', '-', '-', normal ).
adj( 'unspotted', '-', '-', normal ).
adj( 'unstable', '-', '-', normal ).
adj( 'unstartling', '-', '-', normal ).
adj( 'unstated', '-', '-', normal ).
adj( 'unstatesmanlike', '-', '-', normal ).
adj( 'unsteady', '-', '-', pred ).
adj( 'unstoppable', '-', '-', normal ).
adj( 'unstrained', '-', '-', normal ).
adj( 'unstressed', '-', '-', normal ).
adj( 'unstructured', '-', '-', normal ).
adj( 'unstrung', '-', '-', normal ).
adj( 'unstuck', '-', '-', normal ).
adj( 'unstudied', '-', '-', normal ).
adj( 'unsubtle', '-', '-', normal ).
adj( 'unsuccessful', '-', '-', normal ).
adj( 'unsuitable', '-', '-', normal ).
adj( 'unsuited', '-', '-', normal ).
adj( 'unsullied', '-', '-', normal ).
adj( 'unsung', '-', '-', normal ).
adj( 'unsupervised', '-', '-', normal ).
adj( 'unsupported', '-', '-', normal ).
adj( 'unsure', '-', '-', normal ).
adj( 'unsurpassed', '-', '-', normal ).
adj( 'unsuspected', '-', '-', normal ).
adj( 'unsuspecting', '-', '-', normal ).
adj( 'unsuspicious', '-', '-', normal ).
adj( 'unswayed', '-', '-', normal ).
adj( 'unsweetened', '-', '-', normal ).
adj( 'unswept', '-', '-', normal ).
adj( 'unswerving', '-', '-', normal ).
adj( 'unsyllabic', '-', '-', normal ).
adj( 'unsympathetic', '-', '-', normal ).
adj( 'unsystematic', '-', '-', normal ).
adj( 'untainted', '-', '-', normal ).
adj( 'untamed', '-', '-', normal ).
adj( 'untapped', '-', '-', normal ).
adj( 'untarnished', '-', '-', normal ).
adj( 'untaxed', '-', '-', normal ).
adj( 'untenable', '-', '-', normal ).
adj( 'untenanted', '-', '-', normal ).
adj( 'untended', '-', '-', normal ).
adj( 'untested', '-', '-', normal ).
adj( 'untethered', '-', '-', normal ).
adj( 'unthinkable', '-', '-', normal ).
adj( 'unthinking', '-', '-', normal ).
adj( 'unthought-of', '-', '-', normal ).
adj( 'untidy', 'untidier', 'untidiest', normal ).
adj( 'untimely', '-', '-', normal ).
adj( 'untiring', '-', '-', normal ).
adj( 'untitled', '-', '-', normal ).
adj( 'untold', '-', '-', normal ).
adj( 'untouchable', '-', '-', normal ).
adj( 'untouched', '-', '-', normal ).
adj( 'untoward', '-', '-', normal ).
adj( 'untrained', '-', '-', normal ).
adj( 'untrammelled', '-', '-', normal ).
adj( 'untranslatable', '-', '-', normal ).
adj( 'untreated', '-', '-', normal ).
adj( 'untried', '-', '-', normal ).
adj( 'untroubled', '-', '-', normal ).
adj( 'untrue', '-', '-', normal ).
adj( 'untrustworthy', '-', '-', normal ).
adj( 'untruthful', '-', '-', normal ).
adj( 'untucked', '-', '-', normal ).
adj( 'unturned', '-', '-', normal ).
adj( 'untutored', '-', '-', normal ).
adj( 'untwisted', '-', '-', normal ).
adj( 'untypical', '-', '-', normal ).
adj( 'unused', '-', '-', normal ).
adj( 'unused', '-', '-', normal ).
adj( 'unusual', '-', '-', normal ).
adj( 'unutterable', '-', '-', normal ).
adj( 'unvaried', '-', '-', normal ).
adj( 'unvarnished', '-', '-', normal ).
adj( 'unvarying', '-', '-', normal ).
adj( 'unverified', '-', '-', normal ).
adj( 'unversed', '-', '-', normal ).
adj( 'unvigilant', '-', '-', normal ).
adj( 'unvoiced', '-', '-', normal ).
adj( 'unwanted', '-', '-', normal ).
adj( 'unwarranted', '-', '-', normal ).
adj( 'unwary', '-', '-', normal ).
adj( 'unwashed', '-', '-', normal ).
adj( 'unwavering', '-', '-', normal ).
adj( 'unweaned', '-', '-', normal ).
adj( 'unwearied', '-', '-', normal ).
adj( 'unwed', '-', '-', normal ).
adj( 'unwelcome', '-', '-', normal ).
adj( 'unwell', '-', '-', pred ).
adj( 'unwholesome', '-', '-', normal ).
adj( 'unwieldy', '-', '-', normal ).
adj( 'unwilling', '-', '-', normal ).
adj( 'unwise', '-', '-', normal ).
adj( 'unwitnessed', '-', '-', normal ).
adj( 'unwitting', '-', '-', normal ).
adj( 'unwonted', '-', '-', attr ).
adj( 'unworkable', '-', '-', normal ).
adj( 'unworkmanlike', '-', '-', normal ).
adj( 'unworldly', '-', '-', normal ).
adj( 'unworn', '-', '-', normal ).
adj( 'unworried', '-', '-', normal ).
adj( 'unworthy', '-', '-', normal ).
adj( 'unwritten', '-', '-', normal ).
adj( 'unyielding', '-', '-', normal ).
adj( 'up-and-coming', '-', '-', normal ).
adj( 'up-market', '-', '-', normal ).
adj( 'up-to-date', '-', '-', normal ).
adj( 'up-to-the-minute', '-', '-', attr ).
adj( 'upcountry', '-', '-', normal ).
adj( 'uphill', '-', '-', normal ).
adj( 'upmost', '-', '-', normal ).
adj( 'upper', '-', '-', normal ).
adj( 'uppermost', '-', '-', normal ).
adj( 'uppish', '-', '-', normal ).
adj( 'uppity', '-', '-', normal ).
adj( 'upright', '-', '-', normal ).
adj( 'uproarious', '-', '-', normal ).
adj( 'upstage', '-', '-', normal ).
adj( 'upstairs', '-', '-', attr ).
adj( 'upstanding', '-', '-', normal ).
adj( 'upstart', '-', '-', normal ).
adj( 'uptight', '-', '-', normal ).
adj( 'uptown', '-', '-', normal ).
adj( 'upturned', '-', '-', normal ).
adj( 'upward', '-', '-', normal ).
adj( 'urban', '-', '-', normal ).
adj( 'urbane', '-', '-', normal ).
adj( 'urgent', '-', '-', normal ).
adj( 'uric', '-', '-', normal ).
adj( 'urinary', '-', '-', normal ).
adj( 'usable', '-', '-', normal ).
adj( 'used', '-', '-', normal ).
adj( 'used', '-', '-', normal ).
adj( 'useful', '-', '-', normal ).
adj( 'useless', '-', '-', normal ).
adj( 'usual', '-', '-', normal ).
adj( 'usurious', '-', '-', normal ).
adj( 'uterine', '-', '-', normal ).
adj( 'utilitarian', '-', '-', normal ).
adj( 'utilizable', '-', '-', normal ).
adj( 'utmost', '-', '-', normal ).
adj( 'utter', '-', '-', normal ).
adj( 'uttermost', '-', '-', normal ).
adj( 'uvular', '-', '-', normal ).
adj( 'uxorious', '-', '-', normal ).
adj( 'vacant', '-', '-', normal ).
adj( 'vacuous', '-', '-', normal ).
adj( 'vagabond', '-', '-', normal ).
adj( 'vaginal', '-', '-', normal ).
adj( 'vagrant', '-', '-', normal ).
adj( 'vague', 'vaguer', 'vaguest', normal ).
adj( 'vain', 'vainer', 'vainest', normal ).
adj( 'vainglorious', '-', '-', normal ).
adj( 'valedictory', '-', '-', normal ).
adj( 'valetudinarian', '-', '-', normal ).
adj( 'valiant', '-', '-', normal ).
adj( 'valid', '-', '-', normal ).
adj( 'valorous', '-', '-', normal ).
adj( 'valuable', '-', '-', normal ).
adj( 'value-added', '-', '-', attr ).
adj( 'valueless', '-', '-', normal ).
adj( 'valvular', '-', '-', normal ).
adj( 'vapid', '-', '-', normal ).
adj( 'vaporous', '-', '-', normal ).
adj( 'variable', '-', '-', normal ).
adj( 'variant', '-', '-', normal ).
adj( 'varicoloured', '-', '-', normal ).
adj( 'varicose', '-', '-', normal ).
adj( 'varied', '-', '-', normal ).
adj( 'variegated', '-', '-', normal ).
adj( 'variform', '-', '-', normal ).
adj( 'variorum', '-', '-', normal ).
adj( 'various', '-', '-', normal ).
adj( 'vascular', '-', '-', normal ).
adj( 'vast', '-', '-', normal ).
adj( 'vaulted', '-', '-', normal ).
adj( 'vegetable', '-', '-', normal ).
adj( 'vehement', '-', '-', normal ).
adj( 'vehicular', '-', '-', normal ).
adj( 'veined', '-', '-', normal ).
adj( 'velvety', '-', '-', normal ).
adj( 'venal', '-', '-', normal ).
adj( 'venerable', '-', '-', normal ).
adj( 'venereal', '-', '-', normal ).
adj( 'vengeful', '-', '-', normal ).
adj( 'venial', '-', '-', normal ).
adj( 'venomed', '-', '-', normal ).
adj( 'venomous', '-', '-', normal ).
adj( 'venous', '-', '-', normal ).
adj( 'ventral', '-', '-', normal ).
adj( 'ventricular', '-', '-', normal ).
adj( 'venturesome', '-', '-', normal ).
adj( 'venturous', '-', '-', normal ).
adj( 'veracious', '-', '-', normal ).
adj( 'verbal', '-', '-', normal ).
adj( 'verbatim', '-', '-', normal ).
adj( 'verbose', '-', '-', normal ).
adj( 'verdant', '-', '-', normal ).
adj( 'verifiable', '-', '-', normal ).
adj( 'veritable', '-', '-', normal ).
adj( 'vermiform', '-', '-', normal ).
adj( 'vermilion', '-', '-', normal ).
adj( 'verminous', '-', '-', normal ).
adj( 'vernacular', '-', '-', normal ).
adj( 'vernal', '-', '-', normal ).
adj( 'versatile', '-', '-', normal ).
adj( 'versed', '-', '-', normal ).
adj( 'vertebrate', '-', '-', normal ).
adj( 'vertical', '-', '-', normal ).
adj( 'vertiginous', '-', '-', normal ).
adj( 'very', '-', '-', attr ).
adj( 'vesicular', '-', '-', normal ).
adj( 'vestal', '-', '-', normal ).
adj( 'vestigial', '-', '-', normal ).
adj( 'veterinary', '-', '-', normal ).
adj( 'vexatious', '-', '-', normal ).
adj( 'viable', '-', '-', normal ).
adj( 'vibrant', '-', '-', normal ).
adj( 'vicarious', '-', '-', normal ).
adj( 'viceregal', '-', '-', normal ).
adj( 'vicious', '-', '-', normal ).
adj( 'victorious', '-', '-', normal ).
adj( 'viewless', '-', '-', normal ).
adj( 'vigilant', '-', '-', normal ).
adj( 'vigorous', '-', '-', normal ).
adj( 'vile', 'viler', 'vilest', normal ).
adj( 'villainous', '-', '-', normal ).
adj( 'vindictive', '-', '-', normal ).
adj( 'vinegary', '-', '-', normal ).
adj( 'vinous', '-', '-', normal ).
adj( 'violent', '-', '-', normal ).
adj( 'virgin', '-', '-', normal ).
adj( 'virginal', '-', '-', normal ).
adj( 'virile', '-', '-', normal ).
adj( 'virtual', '-', '-', normal ).
adj( 'virtuous', '-', '-', normal ).
adj( 'virulent', '-', '-', normal ).
adj( 'visaged', '-', '-', affix ).
adj( 'visceral', '-', '-', normal ).
adj( 'viscid', '-', '-', normal ).
adj( 'viscous', '-', '-', normal ).
adj( 'visible', '-', '-', normal ).
adj( 'visionary', '-', '-', normal ).
adj( 'visual', '-', '-', normal ).
adj( 'vital', '-', '-', normal ).
adj( 'vitreous', '-', '-', normal ).
adj( 'vitriolic', '-', '-', normal ).
adj( 'vituperative', '-', '-', normal ).
adj( 'viva voce', '-', '-', normal ).
adj( 'vivacious', '-', '-', normal ).
adj( 'vivid', '-', '-', normal ).
adj( 'viviparous', '-', '-', normal ).
adj( 'vixenish', '-', '-', normal ).
adj( 'vocal', '-', '-', normal ).
adj( 'vocational', '-', '-', normal ).
adj( 'vocative', '-', '-', normal ).
adj( 'vociferous', '-', '-', normal ).
adj( 'voiced', '-', '-', affix ).
adj( 'voiceless', '-', '-', normal ).
adj( 'void', '-', '-', normal ).
adj( 'volatile', '-', '-', normal ).
adj( 'volcanic', '-', '-', normal ).
adj( 'volitional', '-', '-', normal ).
adj( 'voluble', '-', '-', normal ).
adj( 'voluminous', '-', '-', normal ).
adj( 'voluntary', '-', '-', normal ).
adj( 'voluptuous', '-', '-', normal ).
adj( 'voluted', '-', '-', normal ).
adj( 'voracious', '-', '-', normal ).
adj( 'voteless', '-', '-', normal ).
adj( 'votive', '-', '-', normal ).
adj( 'voyeuristic', '-', '-', normal ).
adj( 'vulgar', '-', '-', normal ).
adj( 'vulnerable', '-', '-', normal ).
adj( 'vulpine', '-', '-', normal ).
adj( 'waggish', '-', '-', normal ).
adj( 'wainscoted', '-', '-', normal ).
adj( 'waist-deep', '-', '-', normal ).
adj( 'waist-high', '-', '-', normal ).
adj( 'wakeful', '-', '-', normal ).
adj( 'waking', '-', '-', normal ).
adj( 'wall-eyed', '-', '-', normal ).
adj( 'walloping', '-', '-', normal ).
adj( 'wan', 'wanner', 'wannest', normal ).
adj( 'wanton', '-', '-', normal ).
adj( 'war-torn', '-', '-', normal ).
adj( 'warlike', '-', '-', normal ).
adj( 'warm', 'warmer', 'warmest', normal ).
adj( 'warm-blooded', '-', '-', normal ).
adj( 'warm-hearted', '-', '-', normal ).
adj( 'warning', '-', '-', normal ).
adj( 'wary', 'warier', 'wariest', normal ).
adj( 'washable', '-', '-', normal ).
adj( 'washy', '-', '-', normal ).
adj( 'wasp-waisted', '-', '-', normal ).
adj( 'waspish', '-', '-', normal ).
adj( 'waste', '-', '-', normal ).
adj( 'wasteful', '-', '-', normal ).
adj( 'watchful', '-', '-', normal ).
adj( 'water-worn', '-', '-', normal ).
adj( 'waterborne', '-', '-', normal ).
adj( 'waterless', '-', '-', normal ).
adj( 'waterlogged', '-', '-', normal ).
adj( 'waterproof', '-', '-', normal ).
adj( 'watertight', '-', '-', normal ).
adj( 'watery', 'waterier', 'wateriest', normal ).
adj( 'wavy', 'wavier', 'waviest', normal ).
adj( 'waxen', '-', '-', normal ).
adj( 'waxy', 'waxier', 'waxiest', normal ).
adj( 'way-out', '-', '-', normal ).
adj( 'wayfaring', '-', '-', normal ).
adj( 'wayward', '-', '-', normal ).
adj( 'weak', 'weaker', 'weakest', normal ).
adj( 'weak-kneed', '-', '-', normal ).
adj( 'weakly', 'weaklier', 'weakliest', normal ).
adj( 'wealthy', 'wealthier', 'wealthiest', normal ).
adj( 'weaponless', '-', '-', normal ).
adj( 'wearable', '-', '-', normal ).
adj( 'wearing', '-', '-', normal ).
adj( 'wearisome', '-', '-', normal ).
adj( 'weary', 'wearier', 'weariest', normal ).
adj( 'weather-beaten', '-', '-', normal ).
adj( 'weather-bound', '-', '-', normal ).
adj( 'weatherproof', '-', '-', normal ).
adj( 'web-footed', '-', '-', normal ).
adj( 'web-toed', '-', '-', normal ).
adj( 'webbed', '-', '-', normal ).
adj( 'wee', '-', '-', normal ).
adj( 'weedy', 'weedier', 'weediest', normal ).
adj( 'weekly', '-', '-', normal ).
adj( 'weeny', 'weenier', 'weeniest', normal ).
adj( 'weeping', '-', '-', normal ).
adj( 'weightless', '-', '-', normal ).
adj( 'weighty', 'weightier', 'weightiest', normal ).
adj( 'weird', 'weirder', 'weirdest', normal ).
adj( 'welcome', '-', '-', normal ).
adj( 'well', '-', '-', pred ).
adj( 'well-adjusted', '-', '-', normal ).
adj( 'well-advised', '-', '-', normal ).
adj( 'well-appointed', '-', '-', normal ).
adj( 'well-balanced', '-', '-', normal ).
adj( 'well-behaved', '-', '-', normal ).
adj( 'well-born', '-', '-', normal ).
adj( 'well-bred', '-', '-', normal ).
adj( 'well-conducted', '-', '-', normal ).
adj( 'well-connected', '-', '-', normal ).
adj( 'well-disposed', '-', '-', normal ).
adj( 'well-favoured', '-', '-', normal ).
adj( 'well-found', '-', '-', normal ).
adj( 'well-founded', '-', '-', normal ).
adj( 'well-groomed', '-', '-', normal ).
adj( 'well-grounded', '-', '-', normal ).
adj( 'well-heeled', '-', '-', normal ).
adj( 'well-informed', '-', '-', normal ).
adj( 'well-intentioned', '-', '-', normal ).
adj( 'well-knit', '-', '-', normal ).
adj( 'well-known', '-', '-', normal ).
adj( 'well-lined', '-', '-', normal ).
adj( 'well-marked', '-', '-', normal ).
adj( 'well-meaning', '-', '-', normal ).
adj( 'well-meant', '-', '-', normal ).
adj( 'well-read', '-', '-', normal ).
adj( 'well-rounded', '-', '-', normal ).
adj( 'well-set', '-', '-', normal ).
adj( 'well-shaven', '-', '-', normal ).
adj( 'well-spoken', '-', '-', normal ).
adj( 'well-timed', '-', '-', normal ).
adj( 'well-to-do', '-', '-', normal ).
adj( 'well-tried', '-', '-', normal ).
adj( 'well-turned', '-', '-', normal ).
adj( 'well-worn', '-', '-', normal ).
adj( 'welter', '-', '-', normal ).
adj( 'west-country', '-', '-', normal ).
adj( 'west-end', '-', '-', normal ).
adj( 'westerly', '-', '-', attr ).
adj( 'western', '-', '-', normal ).
adj( 'westernmost', '-', '-', normal ).
adj( 'westward', '-', '-', normal ).
adj( 'wet', 'wetter', 'wettest', normal ).
adj( 'whacked', '-', '-', normal ).
adj( 'whacking', '-', '-', normal ).
adj( 'what', '-', '-', normal ).
adj( 'whate\'er', '-', '-', normal ).
adj( 'whatever', '-', '-', normal ).
adj( 'whatsoe\'er', '-', '-', normal ).
adj( 'whatsoever', '-', '-', normal ).
adj( 'wheaten', '-', '-', normal ).
adj( 'wheezy', 'wheezier', 'wheeziest', normal ).
adj( 'whencesoever', '-', '-', normal ).
adj( 'which', '-', '-', normal ).
adj( 'whichever', '-', '-', normal ).
adj( 'whichsoever', '-', '-', normal ).
adj( 'whimsical', '-', '-', normal ).
adj( 'whippy', 'whippier', 'whippiest', normal ).
adj( 'whiskered', '-', '-', normal ).
adj( 'white', 'whiter', 'whitest', normal ).
adj( 'white-collar', '-', '-', normal ).
adj( 'white-hot', '-', '-', normal ).
adj( 'white-lipped', '-', '-', normal ).
adj( 'white-livered', '-', '-', normal ).
adj( 'whitish', '-', '-', normal ).
adj( 'whole', '-', '-', normal ).
adj( 'wholehearted', '-', '-', normal ).
adj( 'wholesale', '-', '-', normal ).
adj( 'wholesome', '-', '-', normal ).
adj( 'whopping', '-', '-', normal ).
adj( 'whorled', '-', '-', normal ).
adj( 'wicked', '-', '-', normal ).
adj( 'wide', 'wider', 'widest', normal ).
adj( 'wide-awake', '-', '-', normal ).
adj( 'widespread', '-', '-', normal ).
adj( 'widowed', '-', '-', normal ).
adj( 'wifelike', '-', '-', normal ).
adj( 'wifely', 'wifelier', 'wifeliest', normal ).
adj( 'wigged', '-', '-', normal ).
adj( 'wild', 'wilder', 'wildest', normal ).
adj( 'wildcat', '-', '-', attr ).
adj( 'wilful', '-', '-', normal ).
adj( 'willed', '-', '-', affix ).
adj( 'willful', '-', '-', normal ).
adj( 'willing', '-', '-', normal ).
adj( 'willowy', '-', '-', normal ).
adj( 'wily', 'wilier', 'wiliest', normal ).
adj( 'windblown', '-', '-', normal ).
adj( 'windless', '-', '-', normal ).
adj( 'windswept', '-', '-', normal ).
adj( 'windward', '-', '-', normal ).
adj( 'windy', 'windier', 'windiest', normal ).
adj( 'winged', '-', '-', normal ).
adj( 'wingless', '-', '-', normal ).
adj( 'winsome', '-', '-', normal ).
adj( 'wintery', 'winterier', 'winteriest', normal ).
adj( 'wintry', 'wintrier', 'wintriest', normal ).
adj( 'wire-haired', '-', '-', normal ).
adj( 'wireless', '-', '-', normal ).
adj( 'wiry', 'wirier', 'wiriest', normal ).
adj( 'wise', 'wiser', 'wisest', normal ).
adj( 'wishful', '-', '-', normal ).
adj( 'wishy-washy', '-', '-', normal ).
adj( 'wispy', 'wispier', 'wispiest', normal ).
adj( 'wistful', '-', '-', normal ).
adj( 'witching', '-', '-', normal ).
adj( 'witless', '-', '-', normal ).
adj( 'witty', 'wittier', 'wittiest', normal ).
adj( 'wizard', '-', '-', normal ).
adj( 'wizened', '-', '-', normal ).
adj( 'wobbly', 'wobblier', 'wobbliest', normal ).
adj( 'woebegone', '-', '-', normal ).
adj( 'woeful', '-', '-', normal ).
adj( 'wolfish', '-', '-', normal ).
adj( 'womanish', '-', '-', normal ).
adj( 'womanlike', '-', '-', normal ).
adj( 'womanly', 'womanlier', 'womanliest', normal ).
adj( 'wonder-struck', '-', '-', normal ).
adj( 'wonderful', '-', '-', normal ).
adj( 'wondrous', '-', '-', normal ).
adj( 'wonky', '-', '-', normal ).
adj( 'wont', '-', '-', pred ).
adj( 'wonted', '-', '-', attr ).
adj( 'wooded', '-', '-', normal ).
adj( 'wooden', '-', '-', normal ).
adj( 'woodenheaded', '-', '-', normal ).
adj( 'woody', 'woodier', 'woodiest', normal ).
adj( 'woolen', '-', '-', normal ).
adj( 'woolgathering', '-', '-', normal ).
adj( 'woollen', '-', '-', normal ).
adj( 'woolly', 'woollier', 'woolliest', normal ).
adj( 'wooly', 'woolier', 'wooliest', normal ).
adj( 'word-perfect', '-', '-', normal ).
adj( 'wordless', '-', '-', normal ).
adj( 'wordy', 'wordier', 'wordiest', normal ).
adj( 'workable', '-', '-', normal ).
adj( 'workaday', '-', '-', normal ).
adj( 'working', '-', '-', normal ).
adj( 'working-class', '-', '-', normal ).
adj( 'workmanlike', '-', '-', normal ).
adj( 'workshy', '-', '-', normal ).
adj( 'world-weary', '-', '-', normal ).
adj( 'worldly', 'worldlier', 'worldliest', normal ).
adj( 'worldwide', '-', '-', normal ).
adj( 'worm-eaten', '-', '-', normal ).
adj( 'wormy', 'wormier', 'wormiest', normal ).
adj( 'worried', '-', '-', normal ).
adj( 'worrisome', '-', '-', normal ).
adj( 'worrying', '-', '-', normal ).
adj( 'worshipful', '-', '-', normal ).
adj( 'worth', '-', '-', pred ).
adj( 'worthless', '-', '-', normal ).
adj( 'worthwhile', '-', '-', normal ).
adj( 'worthy', 'worthier', 'worthiest', normal ).
adj( 'would-be', '-', '-', attr ).
adj( 'wrathful', '-', '-', normal ).
adj( 'wretched', '-', '-', normal ).
adj( 'wrinkly', 'wrinklier', 'wrinkliest', normal ).
adj( 'writ large', '-', '-', normal ).
adj( 'wrong', '-', '-', normal ).
adj( 'wrong-headed', '-', '-', normal ).
adj( 'wrongful', '-', '-', normal ).
adj( 'wroth', '-', '-', normal ).
adj( 'wry', 'wrier', 'wriest', normal ).
adj( 'yearlong', '-', '-', normal ).
adj( 'yearly', '-', '-', normal ).
adj( 'yeasty', 'yeastier', 'yeastiest', normal ).
adj( 'yellow', 'yellower', 'yellowest', normal ).
adj( 'yellow-bellied', '-', '-', normal ).
adj( 'yellowish', '-', '-', normal ).
adj( 'yielding', '-', '-', normal ).
adj( 'yon', '-', '-', normal ).
adj( 'yonder', '-', '-', normal ).
adj( 'young', 'younger', 'youngest', normal ).
adj( 'youngish', '-', '-', normal ).
adj( 'your', '-', '-', normal ).
adj( 'yours', '-', '-', pred ).
adj( 'youthful', '-', '-', normal ).
adj( 'zany', 'zanier', 'zaniest', normal ).
adj( 'zealous', '-', '-', normal ).
adj( 'zenithal', '-', '-', normal ).
adj( 'zestful', '-', '-', normal ).
adj( 'zonal', '-', '-', normal ).
adj( 'zoological', '-', '-', normal ).

adv( '\'tween', normal ).
adv( '\'tween-decks', normal ).
adv( '`a la carte', normal ).
adv( '`a la mode', normal ).
adv( 'a fortiori', normal ).
adv( 'a posteriori', normal ).
adv( 'a priori', normal ).
adv( 'ab initio', normal ).
adv( 'aback', normal ).
adv( 'abaft', normal ).
adv( 'abeam', normal ).
adv( 'abed', normal ).
adv( 'abjectly', normal ).
adv( 'ablaze', normal ).
adv( 'ably', normal ).
adv( 'abnormally', normal ).
adv( 'aboard', normal ).
adv( 'abominably', normal ).
adv( 'abortively', normal ).
adv( 'about', normal ).
adv( 'above', normal ).
adv( 'above board', normal ).
adv( 'abreast', normal ).
adv( 'abroad', normal ).
adv( 'abruptly', normal ).
adv( 'absent-mindedly', normal ).
adv( 'absently', normal ).
adv( 'absolutely', normal ).
adv( 'abstemiously', normal ).
adv( 'abstractedly', normal ).
adv( 'abstrusely', normal ).
adv( 'absurdly', normal ).
adv( 'abundantly', normal ).
adv( 'abusively', normal ).
adv( 'abysmally', normal ).
adv( 'academically', normal ).
adv( 'accelerando', normal ).
adv( 'acceptably', normal ).
adv( 'accidentally', normal ).
adv( 'accordingly', normal ).
adv( 'accurately', normal ).
adv( 'accusingly', normal ).
adv( 'across', normal ).
adv( 'actively', normal ).
adv( 'actually', normal ).
adv( 'acutely', normal ).
adv( 'ad hoc', normal ).
adv( 'ad infinitum', normal ).
adv( 'ad interim', normal ).
adv( 'ad lib', normal ).
adv( 'ad libitum', normal ).
adv( 'ad nauseam', normal ).
adv( 'ad valorem', normal ).
adv( 'adagio', normal ).
adv( 'additionally', normal ).
adv( 'adequately', normal ).
adv( 'administratively', normal ).
adv( 'admirably', normal ).
adv( 'admiringly', normal ).
adv( 'admittedly', normal ).
adv( 'adorably', normal ).
adv( 'adoringly', normal ).
adv( 'adrift', normal ).
adv( 'adroitly', normal ).
adv( 'advantageously', normal ).
adv( 'adverbially', normal ).
adv( 'adversely', normal ).
adv( 'advisedly', normal ).
adv( 'aesthetically', normal ).
adv( 'afar', normal ).
adv( 'affably', normal ).
adv( 'affectingly', normal ).
adv( 'affectionately', normal ).
adv( 'afield', normal ).
adv( 'afore', normal ).
adv( 'aforethought', normal ).
adv( 'afoul', normal ).
adv( 'afresh', normal ).
adv( 'aft', normal ).
adv( 'after', normal ).
adv( 'afterwards', normal ).
adv( 'again', normal ).
adv( 'aggressively', normal ).
adv( 'agilely', normal ).
adv( 'ago', normal ).
adv( 'agonizingly', normal ).
adv( 'agreeably', normal ).
adv( 'aground', normal ).
adv( 'ahead', normal ).
adv( 'aimlessly', normal ).
adv( 'airily', normal ).
adv( 'akimbo', normal ).
adv( 'alarmingly', normal ).
adv( 'alee', normal ).
adv( 'alertly', normal ).
adv( 'alfresco', normal ).
adv( 'algebraically', normal ).
adv( 'alias', normal ).
adv( 'alike', normal ).
adv( 'all', normal ).
adv( 'allegedly', normal ).
adv( 'allegretto', normal ).
adv( 'allegro', normal ).
adv( 'alliteratively', normal ).
adv( 'almost', normal ).
adv( 'aloft', normal ).
adv( 'alone', normal ).
adv( 'along', normal ).
adv( 'alongside', normal ).
adv( 'aloof', normal ).
adv( 'aloud', normal ).
adv( 'alphabetically', normal ).
adv( 'already', normal ).
adv( 'alright', normal ).
adv( 'also', normal ).
adv( 'alternately', normal ).
adv( 'alternatively', normal ).
adv( 'altogether', normal ).
adv( 'altruistically', normal ).
adv( 'always', normal ).
adv( 'amain', normal ).
adv( 'amazingly', normal ).
adv( 'ambiguously', normal ).
adv( 'ambitiously', normal ).
adv( 'amiably', normal ).
adv( 'amicably', normal ).
adv( 'amidships', normal ).
adv( 'amiss', normal ).
adv( 'amok', normal ).
adv( 'amorously', normal ).
adv( 'amply', normal ).
adv( 'amuck', normal ).
adv( 'amusingly', normal ).
adv( 'analogously', normal ).
adv( 'analytically', normal ).
adv( 'anarchically', normal ).
adv( 'anatomically', normal ).
adv( 'andante', normal ).
adv( 'anew', normal ).
adv( 'angelically', normal ).
adv( 'angrily', normal ).
adv( 'annually', normal ).
adv( 'anomalously', normal ).
adv( 'anon', normal ).
adv( 'anonymously', normal ).
adv( 'antagonistically', normal ).
adv( 'ante meridiem', normal ).
adv( 'anticlockwise', normal ).
adv( 'antithetically', normal ).
adv( 'anxiously', normal ).
adv( 'any', normal ).
adv( 'anyhow', normal ).
adv( 'anyplace', normal ).
adv( 'anyway', normal ).
adv( 'anywhere', normal ).
adv( 'apace', normal ).
adv( 'apart', normal ).
adv( 'apathetically', normal ).
adv( 'apiece', normal ).
adv( 'apologetically', normal ).
adv( 'appallingly', normal ).
adv( 'apparently', normal ).
adv( 'appealingly', normal ).
adv( 'appositely', normal ).
adv( 'appreciably', normal ).
adv( 'appreciatively', normal ).
adv( 'appropriately', normal ).
adv( 'approvingly', normal ).
adv( 'approximately', normal ).
adv( 'apropos', normal ).
adv( 'aptly', normal ).
adv( 'arbitrarily', normal ).
adv( 'architecturally', normal ).
adv( 'archly', normal ).
adv( 'ardently', normal ).
adv( 'arduously', normal ).
adv( 'arguably', normal ).
adv( 'aright', normal ).
adv( 'aristocratically', normal ).
adv( 'arithmetically', normal ).
adv( 'around', normal ).
adv( 'arrogantly', normal ).
adv( 'artfully', normal ).
adv( 'articulately', normal ).
adv( 'artificially', normal ).
adv( 'artistically', normal ).
adv( 'artlessly', normal ).
adv( 'ascetically', normal ).
adv( 'ashamedly', normal ).
adv( 'ashore', normal ).
adv( 'aside', normal ).
adv( 'askance', normal ).
adv( 'askew', normal ).
adv( 'aslant', normal ).
adv( 'asleep', normal ).
adv( 'assertively', normal ).
adv( 'assiduously', normal ).
adv( 'assuredly', normal ).
adv( 'astern', normal ).
adv( 'astir', normal ).
adv( 'astonishingly', normal ).
adv( 'astray', normal ).
adv( 'astride', normal ).
adv( 'astronomically', normal ).
adv( 'astutely', normal ).
adv( 'asunder', normal ).
adv( 'asymmetrically', normal ).
adv( 'asymptotically', normal ).
adv( 'athwart', normal ).
adv( 'atop', normal ).
adv( 'atrociously', normal ).
adv( 'attentively', normal ).
adv( 'attractively', normal ).
adv( 'attributively', normal ).
adv( 'atypically', normal ).
adv( 'au fond', normal ).
adv( 'audaciously', normal ).
adv( 'audibly', normal ).
adv( 'auspiciously', normal ).
adv( 'austerely', normal ).
adv( 'authentically', normal ).
adv( 'authoritatively', normal ).
adv( 'autocratically', normal ).
adv( 'automatically', normal ).
adv( 'avariciously', normal ).
adv( 'avidly', normal ).
adv( 'avowedly', normal ).
adv( 'away', normal ).
adv( 'aweigh', normal ).
adv( 'awfully', normal ).
adv( 'awhile', normal ).
adv( 'awkwardly', normal ).
adv( 'awry', normal ).
adv( 'axiomatically', normal ).
adv( 'ay', normal ).
adv( 'aye', normal ).
adv( 'aye', normal ).
adv( 'back', normal ).
adv( 'backstage', normal ).
adv( 'backward', normal ).
adv( 'backwards', normal ).
adv( 'badly', normal ).
adv( 'baldly', normal ).
adv( 'balefully', normal ).
adv( 'bally', normal ).
adv( 'banefully', normal ).
adv( 'bang', normal ).
adv( 'banteringly', normal ).
adv( 'barbarously', normal ).
adv( 'bareback', normal ).
adv( 'barebacked', normal ).
adv( 'barefacedly', normal ).
adv( 'barefoot', normal ).
adv( 'barefooted', normal ).
adv( 'barely', normal ).
adv( 'bashfully', normal ).
adv( 'basically', normal ).
adv( 'bawdily', normal ).
adv( 'beastly', normal ).
adv( 'beautifully', normal ).
adv( 'becomingly', normal ).
adv( 'befittingly', normal ).
adv( 'before', normal ).
adv( 'beforehand', normal ).
adv( 'behind', normal ).
adv( 'belatedly', normal ).
adv( 'belike', normal ).
adv( 'belligerently', normal ).
adv( 'below', normal ).
adv( 'beneath', normal ).
adv( 'beneficially', normal ).
adv( 'benevolently', normal ).
adv( 'benignantly', normal ).
adv( 'benignly', normal ).
adv( 'beseechingly', normal ).
adv( 'besides', normal ).
adv( 'best', normal ).
adv( 'bestially', normal ).
adv( 'betimes', normal ).
adv( 'better', normal ).
adv( 'between', normal ).
adv( 'betwixt', normal ).
adv( 'bewitchingly', normal ).
adv( 'beyond', normal ).
adv( 'biennially', normal ).
adv( 'bilaterally', normal ).
adv( 'biologically', normal ).
adv( 'bitingly', normal ).
adv( 'bitterly', normal ).
adv( 'blamelessly', normal ).
adv( 'blandly', normal ).
adv( 'blankly', normal ).
adv( 'blasphemously', normal ).
adv( 'blatantly', normal ).
adv( 'bleakly', normal ).
adv( 'blindly', normal ).
adv( 'blissfully', normal ).
adv( 'blithely', normal ).
adv( 'bloodlessly', normal ).
adv( 'bloody', normal ).
adv( 'bluffly', normal ).
adv( 'bluntly', normal ).
adv( 'blushingly', normal ).
adv( 'boastfully', normal ).
adv( 'bodily', normal ).
adv( 'boisterously', normal ).
adv( 'boldly', normal ).
adv( 'bolt', normal ).
adv( 'bombastically', normal ).
adv( 'bona fide', normal ).
adv( 'bonnily', normal ).
adv( 'boorishly', normal ).
adv( 'both', normal ).
adv( 'boundlessly', normal ).
adv( 'bounteously', normal ).
adv( 'bountifully', normal ).
adv( 'boyishly', normal ).
adv( 'bravely', normal ).
adv( 'breadthways', normal ).
adv( 'breadthwise', normal ).
adv( 'breast-deep', normal ).
adv( 'breast-high', normal ).
adv( 'breathlessly', normal ).
adv( 'breezily', normal ).
adv( 'briefly', normal ).
adv( 'bright', normal ).
adv( 'brightly', normal ).
adv( 'brilliantly', normal ).
adv( 'briskly', normal ).
adv( 'broadcast', normal ).
adv( 'broadly', normal ).
adv( 'broadways', normal ).
adv( 'broadwise', normal ).
adv( 'brusquely', normal ).
adv( 'brutally', normal ).
adv( 'brutishly', normal ).
adv( 'bump', normal ).
adv( 'bumptiously', normal ).
adv( 'buoyantly', normal ).
adv( 'bureaucratically', normal ).
adv( 'busily', normal ).
adv( 'but', normal ).
adv( 'by', normal ).
adv( 'cagily', normal ).
adv( 'calmly', normal ).
adv( 'candidly', normal ).
adv( 'cannily', normal ).
adv( 'cantankerously', normal ).
adv( 'cap-`a-pie', normal ).
adv( 'capably', normal ).
adv( 'capriciously', normal ).
adv( 'captiously', normal ).
adv( 'carefully', normal ).
adv( 'carelessly', normal ).
adv( 'caressingly', normal ).
adv( 'carnally', normal ).
adv( 'casually', normal ).
adv( 'catastrophically', normal ).
adv( 'categorically', normal ).
adv( 'caustically', normal ).
adv( 'cautiously', normal ).
adv( 'cavalierly', normal ).
adv( 'ceaselessly', normal ).
adv( 'centennially', normal ).
adv( 'centrally', normal ).
adv( 'ceremonially', normal ).
adv( 'ceremoniously', normal ).
adv( 'certainly', normal ).
adv( 'champion', normal ).
adv( 'chaotically', normal ).
adv( 'characteristically', normal ).
adv( 'charily', normal ).
adv( 'charitably', normal ).
adv( 'charmingly', normal ).
adv( 'chastely', normal ).
adv( 'chattily', normal ).
adv( 'cheaply', normal ).
adv( 'cheekily', normal ).
adv( 'cheerfully', normal ).
adv( 'cheerily', normal ).
adv( 'cheerlessly', normal ).
adv( 'chemically', normal ).
adv( 'chiefly', normal ).
adv( 'childishly', normal ).
adv( 'chirpily', normal ).
adv( 'chivalrously', normal ).
adv( 'chock-a-block', normal ).
adv( 'chop-chop', normal ).
adv( 'chronically', normal ).
adv( 'chronologically', normal ).
adv( 'churlishly', normal ).
adv( 'circumspectly', normal ).
adv( 'circumstantially', normal ).
adv( 'civilly', normal ).
adv( 'clammily', normal ).
adv( 'clannishly', normal ).
adv( 'classically', normal ).
adv( 'clean', normal ).
adv( 'cleanly', normal ).
adv( 'clear', normal ).
adv( 'clear-cut', normal ).
adv( 'clearly', normal ).
adv( 'cleverly', normal ).
adv( 'climatically', normal ).
adv( 'clinically', normal ).
adv( 'clockwise', normal ).
adv( 'close', normal ).
adv( 'closely', normal ).
adv( 'clumsily', normal ).
adv( 'coarsely', normal ).
adv( 'coastwise', normal ).
adv( 'coaxingly', normal ).
adv( 'cock-a-hoop', normal ).
adv( 'coherently', normal ).
adv( 'coldly', normal ).
adv( 'collect', normal ).
adv( 'collectedly', normal ).
adv( 'colloquially', normal ).
adv( 'combatively', normal ).
adv( 'comfortably', normal ).
adv( 'comfortingly', normal ).
adv( 'comically', normal ).
adv( 'commercially', normal ).
adv( 'commonly', normal ).
adv( 'communally', normal ).
adv( 'compactly', normal ).
adv( 'comparatively', normal ).
adv( 'compassionately', normal ).
adv( 'compatibly', normal ).
adv( 'competently', normal ).
adv( 'complacently', normal ).
adv( 'complainingly', normal ).
adv( 'completely', normal ).
adv( 'composedly', normal ).
adv( 'comprehensively', normal ).
adv( 'compulsively', normal ).
adv( 'compulsorily', normal ).
adv( 'computationally', normal ).
adv( 'comradely', normal ).
adv( 'con', normal ).
adv( 'conceitedly', normal ).
adv( 'conceivably', normal ).
adv( 'conceptually', normal ).
adv( 'concernedly', normal ).
adv( 'concisely', normal ).
adv( 'conclusively', normal ).
adv( 'concretely', normal ).
adv( 'concurrently', normal ).
adv( 'condescendingly', normal ).
adv( 'conditionally', normal ).
adv( 'confessedly', normal ).
adv( 'confidentially', normal ).
adv( 'confidently', normal ).
adv( 'confidingly', normal ).
adv( 'confoundedly', normal ).
adv( 'confusedly', normal ).
adv( 'congenially', normal ).
adv( 'conjointly', normal ).
adv( 'conjugally', normal ).
adv( 'conscientiously', normal ).
adv( 'consciously', normal ).
adv( 'consecutively', normal ).
adv( 'consequentially', normal ).
adv( 'consequently', normal ).
adv( 'conservatively', normal ).
adv( 'considerably', normal ).
adv( 'considerately', normal ).
adv( 'consistently', normal ).
adv( 'conspicuously', normal ).
adv( 'constantly', normal ).
adv( 'constitutionally', normal ).
adv( 'constrainedly', normal ).
adv( 'constructively', normal ).
adv( 'contagiously', normal ).
adv( 'contemporaneously', normal ).
adv( 'contemptuously', normal ).
adv( 'contentedly', normal ).
adv( 'contiguously', normal ).
adv( 'continually', normal ).
adv( 'continuously', normal ).
adv( 'contrarily', normal ).
adv( 'contrariwise', normal ).
adv( 'contrastingly', normal ).
adv( 'contritely', normal ).
adv( 'controversially', normal ).
adv( 'contumaciously', normal ).
adv( 'contumeliously', normal ).
adv( 'conveniently', normal ).
adv( 'conventionally', normal ).
adv( 'conversationally', normal ).
adv( 'conversely', normal ).
adv( 'convexly', normal ).
adv( 'convincingly', normal ).
adv( 'convivially', normal ).
adv( 'convulsively', normal ).
adv( 'coolly', normal ).
adv( 'coordinately', normal ).
adv( 'copiously', normal ).
adv( 'coquettishly', normal ).
adv( 'cordially', normal ).
adv( 'correctly', normal ).
adv( 'correspondingly', normal ).
adv( 'corruptly', normal ).
adv( 'cosily', normal ).
adv( 'counter', normal ).
adv( 'counterclockwise', normal ).
adv( 'courageously', normal ).
adv( 'courteously', normal ).
adv( 'covertly', normal ).
adv( 'covetously', normal ).
adv( 'coyly', normal ).
adv( 'craftily', normal ).
adv( 'cram-full', normal ).
adv( 'crash', normal ).
adv( 'crazily', normal ).
adv( 'creakily', normal ).
adv( 'creatively', normal ).
adv( 'credibly', normal ).
adv( 'creditably', normal ).
adv( 'credulously', normal ).
adv( 'criminally', normal ).
adv( 'crisply', normal ).
adv( 'crisscross', normal ).
adv( 'critically', normal ).
adv( 'crookedly', normal ).
adv( 'cross-legged', normal ).
adv( 'crosscountry', normal ).
adv( 'crossly', normal ).
adv( 'crosswise', normal ).
adv( 'crucially', normal ).
adv( 'crudely', normal ).
adv( 'cruelly', normal ).
adv( 'crushingly', normal ).
adv( 'cryptically', normal ).
adv( 'culpably', normal ).
adv( 'cumulatively', normal ).
adv( 'cunningly', normal ).
adv( 'curiously', normal ).
adv( 'currently', normal ).
adv( 'currishly', normal ).
adv( 'cursedly', normal ).
adv( 'cursorily', normal ).
adv( 'curtly', normal ).
adv( 'cussedly', normal ).
adv( 'customarily', normal ).
adv( 'cutely', normal ).
adv( 'cynically', normal ).
adv( 'daftly', normal ).
adv( 'daily', normal ).
adv( 'daintily', normal ).
adv( 'damn', normal ).
adv( 'damnably', normal ).
adv( 'damned', normal ).
adv( 'damply', normal ).
adv( 'dangerously', normal ).
adv( 'daringly', normal ).
adv( 'darkly', normal ).
adv( 'dashingly', normal ).
adv( 'dauntlessly', normal ).
adv( 'daylong', normal ).
adv( 'dazedly', normal ).
adv( 'de facto', normal ).
adv( 'de jure', normal ).
adv( 'dead', normal ).
adv( 'deadly', normal ).
adv( 'deal', normal ).
adv( 'dear', normal ).
adv( 'dearly', normal ).
adv( 'deathly', normal ).
adv( 'deceitfully', normal ).
adv( 'deceivingly', normal ).
adv( 'decently', normal ).
adv( 'deceptively', normal ).
adv( 'decidedly', normal ).
adv( 'decisively', normal ).
adv( 'decorously', normal ).
adv( 'deep', normal ).
adv( 'deeply', normal ).
adv( 'defectively', normal ).
adv( 'defencelessly', normal ).
adv( 'defensively', normal ).
adv( 'deferentially', normal ).
adv( 'defiantly', normal ).
adv( 'definitely', normal ).
adv( 'deftly', normal ).
adv( 'dejectedly', normal ).
adv( 'deliberately', normal ).
adv( 'delicately', normal ).
adv( 'deliciously', normal ).
adv( 'delightedly', normal ).
adv( 'delightfully', normal ).
adv( 'deliriously', normal ).
adv( 'delusively', normal ).
adv( 'dementedly', normal ).
adv( 'democratically', normal ).
adv( 'demoniacally', normal ).
adv( 'demonstrably', normal ).
adv( 'demonstratively', normal ).
adv( 'demurely', normal ).
adv( 'densely', normal ).
adv( 'deplorably', normal ).
adv( 'derisively', normal ).
adv( 'descriptively', normal ).
adv( 'deservedly', normal ).
adv( 'designedly', normal ).
adv( 'desolately', normal ).
adv( 'despairingly', normal ).
adv( 'desperately', normal ).
adv( 'despicably', normal ).
adv( 'despitefully', normal ).
adv( 'despondently', normal ).
adv( 'destructively', normal ).
adv( 'determinedly', normal ).
adv( 'detestably', normal ).
adv( 'detrimentally', normal ).
adv( 'deucedly', normal ).
adv( 'devilish', normal ).
adv( 'devilishly', normal ).
adv( 'deviously', normal ).
adv( 'devotedly', normal ).
adv( 'devoutly', normal ).
adv( 'dexterously', normal ).
adv( 'diabolically', normal ).
adv( 'diagonally', normal ).
adv( 'diagrammatically', normal ).
adv( 'diametrically', normal ).
adv( 'dictatorially', normal ).
adv( 'didactically', normal ).
adv( 'differentially', normal ).
adv( 'differently', normal ).
adv( 'diffidently', normal ).
adv( 'diffusely', normal ).
adv( 'digitally', normal ).
adv( 'diligently', normal ).
adv( 'dimly', normal ).
adv( 'ding-dong', normal ).
adv( 'dingily', normal ).
adv( 'diplomatically', normal ).
adv( 'direct', normal ).
adv( 'directly', normal ).
adv( 'direfully', normal ).
adv( 'dirtily', normal ).
adv( 'disadvantageously', normal ).
adv( 'disagreeably', normal ).
adv( 'disappointedly', normal ).
adv( 'disappointingly', normal ).
adv( 'disapprovingly', normal ).
adv( 'disastrously', normal ).
adv( 'disbelievingly', normal ).
adv( 'disconcertingly', normal ).
adv( 'disconsolately', normal ).
adv( 'discontentedly', normal ).
adv( 'discordantly', normal ).
adv( 'discourteously', normal ).
adv( 'discreditably', normal ).
adv( 'discreetly', normal ).
adv( 'discursively', normal ).
adv( 'disdainfully', normal ).
adv( 'disgracefully', normal ).
adv( 'disgustedly', normal ).
adv( 'disgustingly', normal ).
adv( 'dishonestly', normal ).
adv( 'dishonourably', normal ).
adv( 'disingenuously', normal ).
adv( 'disinterestedly', normal ).
adv( 'disjointedly', normal ).
adv( 'disloyally', normal ).
adv( 'dismally', normal ).
adv( 'disobediently', normal ).
adv( 'disparagingly', normal ).
adv( 'dispassionately', normal ).
adv( 'dispiritedly', normal ).
adv( 'displeasingly', normal ).
adv( 'disproportionately', normal ).
adv( 'disputatiously', normal ).
adv( 'disquietingly', normal ).
adv( 'disreputably', normal ).
adv( 'disrespectfully', normal ).
adv( 'dissolutely', normal ).
adv( 'distantly', normal ).
adv( 'distastefully', normal ).
adv( 'distinctively', normal ).
adv( 'distinctly', normal ).
adv( 'distractedly', normal ).
adv( 'distressfully', normal ).
adv( 'distressingly', normal ).
adv( 'distributively', normal ).
adv( 'distrustfully', normal ).
adv( 'disturbingly', normal ).
adv( 'diversely', normal ).
adv( 'divertingly', normal ).
adv( 'divinely', normal ).
adv( 'dizzily', normal ).
adv( 'doggedly', normal ).
adv( 'doggo', normal ).
adv( 'dogmatically', normal ).
adv( 'dolefully', normal ).
adv( 'domestically', normal ).
adv( 'dominantly', normal ).
adv( 'domineeringly', normal ).
adv( 'double', normal ).
adv( 'double-quick', normal ).
adv( 'doubly', normal ).
adv( 'doubtfully', normal ).
adv( 'doubtless', normal ).
adv( 'dourly', normal ).
adv( 'dowdily', normal ).
adv( 'down', normal ).
adv( 'downhill', normal ).
adv( 'downright', normal ).
adv( 'downstairs', normal ).
adv( 'downstream', normal ).
adv( 'downtown', normal ).
adv( 'downward', normal ).
adv( 'downwards', normal ).
adv( 'drably', normal ).
adv( 'dramatically', normal ).
adv( 'drastically', normal ).
adv( 'dreadfully', normal ).
adv( 'dreamily', normal ).
adv( 'drearily', normal ).
adv( 'drily', normal ).
adv( 'droopingly', normal ).
adv( 'drowsily', normal ).
adv( 'drunkenly', normal ).
adv( 'dubiously', normal ).
adv( 'due', normal ).
adv( 'dully', normal ).
adv( 'duly', normal ).
adv( 'dumbly', normal ).
adv( 'dutifully', normal ).
adv( 'dynamically', normal ).
adv( 'e\'en', normal ).
adv( 'e\'er', normal ).
adv( 'eagerly', normal ).
adv( 'early', normal ).
adv( 'earnestly', normal ).
adv( 'easily', normal ).
adv( 'east', normal ).
adv( 'easterly', normal ).
adv( 'eastward', normal ).
adv( 'eastwards', normal ).
adv( 'easy', normal ).
adv( 'ebulliently', normal ).
adv( 'ecclesiastically', normal ).
adv( 'ecologically', normal ).
adv( 'economically', normal ).
adv( 'ecstatically', normal ).
adv( 'edgeways', normal ).
adv( 'edgewise', normal ).
adv( 'educationally', normal ).
adv( 'eerily', normal ).
adv( 'effectively', normal ).
adv( 'effectually', normal ).
adv( 'efficaciously', normal ).
adv( 'efficiently', normal ).
adv( 'effortlessly', normal ).
adv( 'effusively', normal ).
adv( 'egotistically', normal ).
adv( 'either', normal ).
adv( 'elaborately', normal ).
adv( 'electrically', normal ).
adv( 'electronically', normal ).
adv( 'elegantly', normal ).
adv( 'elementarily', normal ).
adv( 'eloquently', normal ).
adv( 'else', normal ).
adv( 'elsewhere', normal ).
adv( 'embarrassingly', normal ).
adv( 'eminently', normal ).
adv( 'emotionally', normal ).
adv( 'emphatically', normal ).
adv( 'empirically', normal ).
adv( 'emulously', normal ).
adv( 'en clair', normal ).
adv( 'en famille', normal ).
adv( 'en masse', normal ).
adv( 'en route', normal ).
adv( 'enchantingly', normal ).
adv( 'encouragingly', normal ).
adv( 'endearingly', normal ).
adv( 'endlessly', normal ).
adv( 'enduringly', normal ).
adv( 'endways', normal ).
adv( 'endwise', normal ).
adv( 'energetically', normal ).
adv( 'engagingly', normal ).
adv( 'enigmatically', normal ).
adv( 'enjoyably', normal ).
adv( 'enormously', normal ).
adv( 'enough', normal ).
adv( 'enquiringly', normal ).
adv( 'enterprisingly', normal ).
adv( 'entertainingly', normal ).
adv( 'enthusiastically', normal ).
adv( 'entirely', normal ).
adv( 'entreatingly', normal ).
adv( 'enviously', normal ).
adv( 'environmentally', normal ).
adv( 'equably', normal ).
adv( 'equally', normal ).
adv( 'equitably', normal ).
adv( 'ere', normal ).
adv( 'erectly', normal ).
adv( 'ergo', normal ).
adv( 'erratically', normal ).
adv( 'erroneously', normal ).
adv( 'eruditely', normal ).
adv( 'especially', normal ).
adv( 'essentially', normal ).
adv( 'eternally', normal ).
adv( 'ethically', normal ).
adv( 'ethnically', normal ).
adv( 'euphemistically', normal ).
adv( 'evasively', normal ).
adv( 'even', normal ).
adv( 'evenly', normal ).
adv( 'eventually', normal ).
adv( 'ever', normal ).
adv( 'evermore', normal ).
adv( 'everyplace', normal ).
adv( 'everywhere', normal ).
adv( 'evidently', normal ).
adv( 'evilly', normal ).
adv( 'ex officio', normal ).
adv( 'exactly', normal ).
adv( 'exceedingly', normal ).
adv( 'excellently', normal ).
adv( 'exceptionally', normal ).
adv( 'excessively', normal ).
adv( 'excitedly', normal ).
adv( 'excitingly', normal ).
adv( 'exclusively', normal ).
adv( 'excruciatingly', normal ).
adv( 'excusably', normal ).
adv( 'exhaustively', normal ).
adv( 'exorbitantly', normal ).
adv( 'expansively', normal ).
adv( 'expectantly', normal ).
adv( 'expediently', normal ).
adv( 'expeditiously', normal ).
adv( 'expensively', normal ).
adv( 'experimentally', normal ).
adv( 'expertly', normal ).
adv( 'explicitly', normal ).
adv( 'explosively', normal ).
adv( 'exponentially', normal ).
adv( 'express', normal ).
adv( 'expressively', normal ).
adv( 'expressly', normal ).
adv( 'exquisitely', normal ).
adv( 'extemporaneously', normal ).
adv( 'extemporarily', normal ).
adv( 'extempore', normal ).
adv( 'extensively', normal ).
adv( 'externally', normal ).
adv( 'extortionately', normal ).
adv( 'extra', normal ).
adv( 'extraordinarily', normal ).
adv( 'extravagantly', normal ).
adv( 'extremely', normal ).
adv( 'exuberantly', normal ).
adv( 'exultantly', normal ).
adv( 'fabulously', normal ).
adv( 'facetiously', normal ).
adv( 'factually', normal ).
adv( 'faddily', normal ).
adv( 'fain', normal ).
adv( 'faintly', normal ).
adv( 'fair', normal ).
adv( 'fairly', normal ).
adv( 'faithfully', normal ).
adv( 'faithlessly', normal ).
adv( 'false', normal ).
adv( 'falsely', normal ).
adv( 'falteringly', normal ).
adv( 'familiarly', normal ).
adv( 'famously', normal ).
adv( 'fanatically', normal ).
adv( 'fancifully', normal ).
adv( 'fantastically', normal ).
adv( 'far', normal ).
adv( 'farcically', normal ).
adv( 'farther', normal ).
adv( 'farthest', normal ).
adv( 'fascinatingly', normal ).
adv( 'fashionably', normal ).
adv( 'fast', normal ).
adv( 'fastidiously', normal ).
adv( 'fatefully', normal ).
adv( 'fatuously', normal ).
adv( 'faultily', normal ).
adv( 'faultlessly', normal ).
adv( 'favourably', normal ).
adv( 'fearfully', normal ).
adv( 'fearlessly', normal ).
adv( 'fearsomely', normal ).
adv( 'fecklessly', normal ).
adv( 'feebly', normal ).
adv( 'feelingly', normal ).
adv( 'felicitously', normal ).
adv( 'ferociously', normal ).
adv( 'fervently', normal ).
adv( 'fervidly', normal ).
adv( 'feverishly', normal ).
adv( 'fiendishly', normal ).
adv( 'fiercely', normal ).
adv( 'fierily', normal ).
adv( 'fifthly', normal ).
adv( 'figuratively', normal ).
adv( 'filthily', normal ).
adv( 'finally', normal ).
adv( 'financially', normal ).
adv( 'fine', normal ).
adv( 'finely', normal ).
adv( 'firm', normal ).
adv( 'firmly', normal ).
adv( 'first', normal ).
adv( 'first-class', normal ).
adv( 'first-hand', normal ).
adv( 'first-rate', normal ).
adv( 'firstly', normal ).
adv( 'fitfully', normal ).
adv( 'fitly', normal ).
adv( 'fixedly', normal ).
adv( 'flabbily', normal ).
adv( 'flagrantly', normal ).
adv( 'flamboyantly', normal ).
adv( 'flashily', normal ).
adv( 'flat', normal ).
adv( 'flatly', normal ).
adv( 'flawlessly', normal ).
adv( 'fleetly', normal ).
adv( 'flexibly', normal ).
adv( 'flimsily', normal ).
adv( 'flippantly', normal ).
adv( 'flop', normal ).
adv( 'floridly', normal ).
adv( 'fluently', normal ).
adv( 'fondly', normal ).
adv( 'foolishly', normal ).
adv( 'forbiddingly', normal ).
adv( 'forcefully', normal ).
adv( 'forcibly', normal ).
adv( 'fore', normal ).
adv( 'foremost', normal ).
adv( 'forever', normal ).
adv( 'forgetfully', normal ).
adv( 'forgivably', normal ).
adv( 'forgivingly', normal ).
adv( 'forlornly', normal ).
adv( 'formally', normal ).
adv( 'formerly', normal ).
adv( 'formidably', normal ).
adv( 'formlessly', normal ).
adv( 'forrader', normal ).
adv( 'forsooth', normal ).
adv( 'forte', normal ).
adv( 'forth', normal ).
adv( 'forthwith', normal ).
adv( 'fortissimo', normal ).
adv( 'fortnightly', normal ).
adv( 'fortuitously', normal ).
adv( 'fortunately', normal ).
adv( 'forward', normal ).
adv( 'forwards', normal ).
adv( 'foully', normal ).
adv( 'fourfold', normal ).
adv( 'fourthly', normal ).
adv( 'fractiously', normal ).
adv( 'frankly', normal ).
adv( 'frantically', normal ).
adv( 'fraternally', normal ).
adv( 'fraudulently', normal ).
adv( 'freakishly', normal ).
adv( 'freely', normal ).
adv( 'frenziedly', normal ).
adv( 'frequently', normal ).
adv( 'fresh', normal ).
adv( 'freshly', normal ).
adv( 'fretfully', normal ).
adv( 'frighteningly', normal ).
adv( 'frightfully', normal ).
adv( 'frigidly', normal ).
adv( 'friskily', normal ).
adv( 'frivolously', normal ).
adv( 'fro', normal ).
adv( 'frostily', normal ).
adv( 'frothily', normal ).
adv( 'frowningly', normal ).
adv( 'frugally', normal ).
adv( 'fruitfully', normal ).
adv( 'fruitlessly', normal ).
adv( 'full-time', normal ).
adv( 'fully', normal ).
adv( 'fulsomely', normal ).
adv( 'functionally', normal ).
adv( 'fundamentally', normal ).
adv( 'funnily', normal ).
adv( 'furiously', normal ).
adv( 'further', normal ).
adv( 'furthermore', normal ).
adv( 'furthest', normal ).
adv( 'furtively', normal ).
adv( 'fussily', normal ).
adv( 'gaily', normal ).
adv( 'gainfully', normal ).
adv( 'gallantly', normal ).
adv( 'galore', normal ).
adv( 'gamely', normal ).
adv( 'garishly', normal ).
adv( 'gaudily', normal ).
adv( 'genealogically', normal ).
adv( 'generally', normal ).
adv( 'generically', normal ).
adv( 'generously', normal ).
adv( 'genetically', normal ).
adv( 'genially', normal ).
adv( 'genteelly', normal ).
adv( 'gently', normal ).
adv( 'genuinely', normal ).
adv( 'geographically', normal ).
adv( 'geologically', normal ).
adv( 'geometrically', normal ).
adv( 'gibingly', normal ).
adv( 'giddily', normal ).
adv( 'gingerly', normal ).
adv( 'girlishly', normal ).
adv( 'gladly', normal ).
adv( 'gleefully', normal ).
adv( 'glibly', normal ).
adv( 'glissando', normal ).
adv( 'gloatingly', normal ).
adv( 'gloomily', normal ).
adv( 'gloriously', normal ).
adv( 'glossily', normal ).
adv( 'gloweringly', normal ).
adv( 'glowingly', normal ).
adv( 'glumly', normal ).
adv( 'gluttonously', normal ).
adv( 'goddam', normal ).
adv( 'gorgeously', normal ).
adv( 'gracefully', normal ).
adv( 'gracelessly', normal ).
adv( 'graciously', normal ).
adv( 'gradually', normal ).
adv( 'grammatically', normal ).
adv( 'grandly', normal ).
adv( 'graphically', normal ).
adv( 'gratefully', normal ).
adv( 'gratingly', normal ).
adv( 'gratis', normal ).
adv( 'gratuitously', normal ).
adv( 'gravely', normal ).
adv( 'greasily', normal ).
adv( 'greatly', normal ).
adv( 'greedily', normal ).
adv( 'gregariously', normal ).
adv( 'greyly', normal ).
adv( 'grievously', normal ).
adv( 'grimly', normal ).
adv( 'gropingly', normal ).
adv( 'grossly', normal ).
adv( 'grotesquely', normal ).
adv( 'growlingly', normal ).
adv( 'grudgingly', normal ).
adv( 'gruesomely', normal ).
adv( 'gruffly', normal ).
adv( 'grumpily', normal ).
adv( 'guardedly', normal ).
adv( 'guiltily', normal ).
adv( 'gushingly', normal ).
adv( 'gutturally', normal ).
adv( 'habitually', normal ).
adv( 'half-heartedly', normal ).
adv( 'half-hourly', normal ).
adv( 'half-price', normal ).
adv( 'half-yearly', normal ).
adv( 'halfway', normal ).
adv( 'haltingly', normal ).
adv( 'handily', normal ).
adv( 'handsomely', normal ).
adv( 'haphazard', normal ).
adv( 'haphazardly', normal ).
adv( 'haply', normal ).
adv( 'happily', normal ).
adv( 'hard', normal ).
adv( 'hardly', normal ).
adv( 'harmfully', normal ).
adv( 'harmlessly', normal ).
adv( 'harmoniously', normal ).
adv( 'harshly', normal ).
adv( 'hastily', normal ).
adv( 'hatefully', normal ).
adv( 'haughtily', normal ).
adv( 'hazily', normal ).
adv( 'head-on', normal ).
adv( 'headlong', normal ).
adv( 'healthily', normal ).
adv( 'heaps', normal ).
adv( 'heartily', normal ).
adv( 'heartlessly', normal ).
adv( 'heatedly', normal ).
adv( 'heavenward', normal ).
adv( 'heavenwards', normal ).
adv( 'heavily', normal ).
adv( 'heavy', normal ).
adv( 'heinously', normal ).
adv( 'helpfully', normal ).
adv( 'helplessly', normal ).
adv( 'helter-skelter', normal ).
adv( 'hence', normal ).
adv( 'henceforth', normal ).
adv( 'henceforward', normal ).
adv( 'here', normal ).
adv( 'hereabouts', normal ).
adv( 'hereafter', normal ).
adv( 'hereby', normal ).
adv( 'herein', normal ).
adv( 'hereinafter', normal ).
adv( 'hereinbefore', normal ).
adv( 'hereof', normal ).
adv( 'hereto', normal ).
adv( 'heretofore', normal ).
adv( 'hereupon', normal ).
adv( 'herewith', normal ).
adv( 'hermetically', normal ).
adv( 'heroically', normal ).
adv( 'hesitantly', normal ).
adv( 'hesitatingly', normal ).
adv( 'hideously', normal ).
adv( 'higgledy-piggledy', normal ).
adv( 'high', normal ).
adv( 'high-handedly', normal ).
adv( 'high-mindedly', normal ).
adv( 'highly', normal ).
adv( 'hilariously', normal ).
adv( 'historically', normal ).
adv( 'hither', normal ).
adv( 'hitherto', normal ).
adv( 'hoarsely', normal ).
adv( 'home', normal ).
adv( 'homeward', normal ).
adv( 'homewards', normal ).
adv( 'honestly', normal ).
adv( 'honourably', normal ).
adv( 'hopefully', normal ).
adv( 'hopelessly', normal ).
adv( 'horizontally', normal ).
adv( 'horribly', normal ).
adv( 'horridly', normal ).
adv( 'horrifyingly', normal ).
adv( 'hospitably', normal ).
adv( 'hostilely', normal ).
adv( 'hotfoot', normal ).
adv( 'hotly', normal ).
adv( 'hourly', normal ).
adv( 'how', normal ).
adv( 'however', normal ).
adv( 'huffily', normal ).
adv( 'hugely', normal ).
adv( 'hugger-mugger', normal ).
adv( 'humanely', normal ).
adv( 'humanly', normal ).
adv( 'humbly', normal ).
adv( 'humorously', normal ).
adv( 'hundredfold', normal ).
adv( 'hungrily', normal ).
adv( 'hurriedly', normal ).
adv( 'huskily', normal ).
adv( 'hydraulicly', normal ).
adv( 'hygienically', normal ).
adv( 'hypocritically', normal ).
adv( 'hysterically', normal ).
adv( 'ibidem', normal ).
adv( 'icily', normal ).
adv( 'ideally', normal ).
adv( 'identically', normal ).
adv( 'identifiably', normal ).
adv( 'ideologically', normal ).
adv( 'idiomatically', normal ).
adv( 'idiotically', normal ).
adv( 'idly', normal ).
adv( 'idolatrously', normal ).
adv( 'ignobly', normal ).
adv( 'ignominiously', normal ).
adv( 'ignorantly', normal ).
adv( 'ill', normal ).
adv( 'illegally', normal ).
adv( 'illegibly', normal ).
adv( 'illegitimately', normal ).
adv( 'illiberally', normal ).
adv( 'illicitly', normal ).
adv( 'illogically', normal ).
adv( 'illustriously', normal ).
adv( 'imaginatively', normal ).
adv( 'immaculately', normal ).
adv( 'immeasurably', normal ).
adv( 'immediately', normal ).
adv( 'immensely', normal ).
adv( 'imminently', normal ).
adv( 'immoderately', normal ).
adv( 'immodestly', normal ).
adv( 'immorally', normal ).
adv( 'immovably', normal ).
adv( 'immutably', normal ).
adv( 'impartially', normal ).
adv( 'impassively', normal ).
adv( 'impatiently', normal ).
adv( 'impeccably', normal ).
adv( 'impenitently', normal ).
adv( 'imperatively', normal ).
adv( 'imperceptibly', normal ).
adv( 'imperfectly', normal ).
adv( 'imperially', normal ).
adv( 'imperiously', normal ).
adv( 'impersonally', normal ).
adv( 'impertinently', normal ).
adv( 'impetuously', normal ).
adv( 'impiously', normal ).
adv( 'impishly', normal ).
adv( 'implausibly', normal ).
adv( 'implicitly', normal ).
adv( 'imploringly', normal ).
adv( 'impolitely', normal ).
adv( 'importantly', normal ).
adv( 'importunately', normal ).
adv( 'imposingly', normal ).
adv( 'impossibly', normal ).
adv( 'impotently', normal ).
adv( 'impracticably', normal ).
adv( 'imprecisely', normal ).
adv( 'impregnably', normal ).
adv( 'impressively', normal ).
adv( 'improbably', normal ).
adv( 'impromptu', normal ).
adv( 'improperly', normal ).
adv( 'improvidently', normal ).
adv( 'imprudently', normal ).
adv( 'impudently', normal ).
adv( 'impulsively', normal ).
adv( 'in', normal ).
adv( 'in loco parentis', normal ).
adv( 'in situ', normal ).
adv( 'in toto', normal ).
adv( 'inaccurately', normal ).
adv( 'inadequately', normal ).
adv( 'inadvertently', normal ).
adv( 'inadvisably', normal ).
adv( 'inanely', normal ).
adv( 'inappropriately', normal ).
adv( 'inasmuch as', normal ).
adv( 'inauspiciously', normal ).
adv( 'incautiously', normal ).
adv( 'incessantly', normal ).
adv( 'incidentally', normal ).
adv( 'incisively', normal ).
adv( 'inclusively', normal ).
adv( 'incognito', normal ).
adv( 'incoherently', normal ).
adv( 'incomparably', normal ).
adv( 'incompetently', normal ).
adv( 'incompletely', normal ).
adv( 'inconclusively', normal ).
adv( 'incongruously', normal ).
adv( 'inconsequentially', normal ).
adv( 'inconsequently', normal ).
adv( 'inconsiderately', normal ).
adv( 'inconsistently', normal ).
adv( 'inconspicuously', normal ).
adv( 'inconveniently', normal ).
adv( 'incorrectly', normal ).
adv( 'increasingly', normal ).
adv( 'incredibly', normal ).
adv( 'incredulously', normal ).
adv( 'incurably', normal ).
adv( 'indecently', normal ).
adv( 'indecisively', normal ).
adv( 'indecorously', normal ).
adv( 'indeed', normal ).
adv( 'indefinitely', normal ).
adv( 'indelibly', normal ).
adv( 'independently', normal ).
adv( 'indescribably', normal ).
adv( 'indeterminably', normal ).
adv( 'indifferently', normal ).
adv( 'indignantly', normal ).
adv( 'indirectly', normal ).
adv( 'indiscreetly', normal ).
adv( 'indiscriminately', normal ).
adv( 'indistinctly', normal ).
adv( 'individually', normal ).
adv( 'indolently', normal ).
adv( 'indoors', normal ).
adv( 'indubitably', normal ).
adv( 'indulgently', normal ).
adv( 'industriously', normal ).
adv( 'ineffably', normal ).
adv( 'ineffectively', normal ).
adv( 'ineffectually', normal ).
adv( 'inefficiently', normal ).
adv( 'inelegantly', normal ).
adv( 'ineptly', normal ).
adv( 'inescapably', normal ).
adv( 'inevitably', normal ).
adv( 'inexorably', normal ).
adv( 'inexpensively', normal ).
adv( 'inexpertly', normal ).
adv( 'inextricably', normal ).
adv( 'infernally', normal ).
adv( 'infinitely', normal ).
adv( 'inflexibly', normal ).
adv( 'influentially', normal ).
adv( 'informally', normal ).
adv( 'informatively', normal ).
adv( 'infra', normal ).
adv( 'infrequently', normal ).
adv( 'ingeniously', normal ).
adv( 'ingenuously', normal ).
adv( 'ingloriously', normal ).
adv( 'ingratiatingly', normal ).
adv( 'inherently', normal ).
adv( 'inhumanely', normal ).
adv( 'inimitably', normal ).
adv( 'iniquitously', normal ).
adv( 'initially', normal ).
adv( 'injudiciously', normal ).
adv( 'inland', normal ).
adv( 'innately', normal ).
adv( 'innocently', normal ).
adv( 'inoffensively', normal ).
adv( 'inopportunely', normal ).
adv( 'inordinately', normal ).
adv( 'inorganically', normal ).
adv( 'inquiringly', normal ).
adv( 'inquisitively', normal ).
adv( 'insanely', normal ).
adv( 'insatiably', normal ).
adv( 'insecurely', normal ).
adv( 'insensibly', normal ).
adv( 'insensitively', normal ).
adv( 'inshore', normal ).
adv( 'inside', normal ).
adv( 'insidiously', normal ).
adv( 'insignificantly', normal ).
adv( 'insincerely', normal ).
adv( 'insinuatingly', normal ).
adv( 'insipidly', normal ).
adv( 'insofar', normal ).
adv( 'insolently', normal ).
adv( 'insomuch', normal ).
adv( 'inspirationally', normal ).
adv( 'instantaneously', normal ).
adv( 'instantly', normal ).
adv( 'instead', normal ).
adv( 'instinctively', normal ).
adv( 'institutionally', normal ).
adv( 'instructively', normal ).
adv( 'insubstantially', normal ).
adv( 'insufficiently', normal ).
adv( 'insultingly', normal ).
adv( 'insuperably', normal ).
adv( 'integrally', normal ).
adv( 'intellectually', normal ).
adv( 'intelligently', normal ).
adv( 'intelligibly', normal ).
adv( 'intemperately', normal ).
adv( 'intensely', normal ).
adv( 'intensively', normal ).
adv( 'intentionally', normal ).
adv( 'intently', normal ).
adv( 'inter alia', normal ).
adv( 'interchangeably', normal ).
adv( 'interdepartmentally', normal ).
adv( 'interestingly', normal ).
adv( 'intermediately', normal ).
adv( 'interminably', normal ).
adv( 'intermittently', normal ).
adv( 'internally', normal ).
adv( 'internationally', normal ).
adv( 'interrogatively', normal ).
adv( 'intimately', normal ).
adv( 'intolerably', normal ).
adv( 'intolerantly', normal ).
adv( 'intransitively', normal ).
adv( 'intravenously', normal ).
adv( 'intrepidly', normal ).
adv( 'intricately', normal ).
adv( 'intrinsically', normal ).
adv( 'intuitively', normal ).
adv( 'invariably', normal ).
adv( 'inventively', normal ).
adv( 'inversely', normal ).
adv( 'invidiously', normal ).
adv( 'invincibly', normal ).
adv( 'invisibly', normal ).
adv( 'invitingly', normal ).
adv( 'involuntarily', normal ).
adv( 'inward', normal ).
adv( 'inwardly', normal ).
adv( 'inwards', normal ).
adv( 'ipso facto', normal ).
adv( 'irately', normal ).
adv( 'ironically', normal ).
adv( 'irrationally', normal ).
adv( 'irregularly', normal ).
adv( 'irrelevantly', normal ).
adv( 'irretrievably', normal ).
adv( 'irreverently', normal ).
adv( 'irreversibly', normal ).
adv( 'irritably', normal ).
adv( 'item', normal ).
adv( 'jarringly', normal ).
adv( 'jauntily', normal ).
adv( 'jealously', normal ).
adv( 'jeeringly', normal ).
adv( 'jejunely', normal ).
adv( 'jerkily', normal ).
adv( 'jestingly', normal ).
adv( 'jocosely', normal ).
adv( 'jocularly', normal ).
adv( 'jointly', normal ).
adv( 'jokingly', normal ).
adv( 'jolly', normal ).
adv( 'journalistically', normal ).
adv( 'jovially', normal ).
adv( 'joyfully', normal ).
adv( 'joylessly', normal ).
adv( 'joyously', normal ).
adv( 'jubilantly', normal ).
adv( 'judiciously', normal ).
adv( 'just', normal ).
adv( 'justifiably', normal ).
adv( 'justly', normal ).
adv( 'keenly', normal ).
adv( 'killingly', normal ).
adv( 'kinda', normal ).
adv( 'kindly', normal ).
adv( 'knavishly', normal ).
adv( 'knee-deep', normal ).
adv( 'knee-high', normal ).
adv( 'knowingly', normal ).
adv( 'laboriously', normal ).
adv( 'lackadaisically', normal ).
adv( 'laconically', normal ).
adv( 'lamely', normal ).
adv( 'lamentably', normal ).
adv( 'landward', normal ).
adv( 'langsyne', normal ).
adv( 'languidly', normal ).
adv( 'languorously', normal ).
adv( 'large', normal ).
adv( 'largely', normal ).
adv( 'lasciviously', normal ).
adv( 'last', normal ).
adv( 'lastly', normal ).
adv( 'late', normal ).
adv( 'lately', normal ).
adv( 'laterally', normal ).
adv( 'latterly', normal ).
adv( 'laudably', normal ).
adv( 'laughably', normal ).
adv( 'laughingly', normal ).
adv( 'lavishly', normal ).
adv( 'lawfully', normal ).
adv( 'lawlessly', normal ).
adv( 'laxly', normal ).
adv( 'lazily', normal ).
adv( 'learnedly', normal ).
adv( 'least', normal ).
adv( 'leastways', normal ).
adv( 'leastwise', normal ).
adv( 'leeward', normal ).
adv( 'left', normal ).
adv( 'legally', normal ).
adv( 'legato', normal ).
adv( 'legibly', normal ).
adv( 'legitimately', normal ).
adv( 'leisurely', normal ).
adv( 'lengthily', normal ).
adv( 'lengthways', normal ).
adv( 'lengthwise', normal ).
adv( 'leniently', normal ).
adv( 'lento', normal ).
adv( 'less', normal ).
adv( 'lethargically', normal ).
adv( 'lewdly', normal ).
adv( 'lexically', normal ).
adv( 'liberally', normal ).
adv( 'licentiously', normal ).
adv( 'lief', normal ).
adv( 'lifelessly', normal ).
adv( 'light', normal ).
adv( 'light-handedly', normal ).
adv( 'light-headedly', normal ).
adv( 'light-heartedly', normal ).
adv( 'lightly', normal ).
adv( 'lightsomely', normal ).
adv( 'like', normal ).
adv( 'likely', normal ).
adv( 'likewise', normal ).
adv( 'limpidly', normal ).
adv( 'limply', normal ).
adv( 'lineally', normal ).
adv( 'linearly', normal ).
adv( 'lingeringly', normal ).
adv( 'linguistically', normal ).
adv( 'lispingly', normal ).
adv( 'listlessly', normal ).
adv( 'literally', normal ).
adv( 'little', normal ).
adv( 'live', normal ).
adv( 'lividly', normal ).
adv( 'locally', normal ).
adv( 'loftily', normal ).
adv( 'logarithmically', normal ).
adv( 'logically', normal ).
adv( 'long', normal ).
adv( 'longer', normal ).
adv( 'longest', normal ).
adv( 'longingly', normal ).
adv( 'longitudinally', normal ).
adv( 'longways', normal ).
adv( 'longwise', normal ).
adv( 'loosely', normal ).
adv( 'loquaciously', normal ).
adv( 'loud', normal ).
adv( 'loudly', normal ).
adv( 'louringly', normal ).
adv( 'lovingly', normal ).
adv( 'low', normal ).
adv( 'lower', normal ).
adv( 'lowest', normal ).
adv( 'loyally', normal ).
adv( 'lucidly', normal ).
adv( 'luckily', normal ).
adv( 'ludicrously', normal ).
adv( 'lugubriously', normal ).
adv( 'lukewarmly', normal ).
adv( 'luridly', normal ).
adv( 'lusciously', normal ).
adv( 'lustfully', normal ).
adv( 'lustily', normal ).
adv( 'luxuriantly', normal ).
adv( 'luxuriously', normal ).
adv( 'lyrically', normal ).
adv( 'madly', normal ).
adv( 'magically', normal ).
adv( 'magisterially', normal ).
adv( 'magnanimously', normal ).
adv( 'magnetically', normal ).
adv( 'magnificently', normal ).
adv( 'magniloquently', normal ).
adv( 'mainly', normal ).
adv( 'majestically', normal ).
adv( 'maladroitly', normal ).
adv( 'malapropos', normal ).
adv( 'malevolently', normal ).
adv( 'maliciously', normal ).
adv( 'malignantly', normal ).
adv( 'manageably', normal ).
adv( 'manfully', normal ).
adv( 'mangily', normal ).
adv( 'maniacally', normal ).
adv( 'manifestly', normal ).
adv( 'manually', normal ).
adv( 'marginally', normal ).
adv( 'markedly', normal ).
adv( 'martially', normal ).
adv( 'marvellously', normal ).
adv( 'marvelously', normal ).
adv( 'masochistically', normal ).
adv( 'massively', normal ).
adv( 'masterfully', normal ).
adv( 'materialistically', normal ).
adv( 'materially', normal ).
adv( 'maternally', normal ).
adv( 'mathematically', normal ).
adv( 'maturely', normal ).
adv( 'mawkishly', normal ).
adv( 'maximally', normal ).
adv( 'maybe', normal ).
adv( 'meagrely', normal ).
adv( 'meanderingly', normal ).
adv( 'meaningfully', normal ).
adv( 'meaningly', normal ).
adv( 'meanly', normal ).
adv( 'meantime', normal ).
adv( 'meanwhile', normal ).
adv( 'measurably', normal ).
adv( 'mechanically', normal ).
adv( 'mechanistically', normal ).
adv( 'medially', normal ).
adv( 'medically', normal ).
adv( 'meditatively', normal ).
adv( 'meekly', normal ).
adv( 'mellowly', normal ).
adv( 'melodiously', normal ).
adv( 'melodramatically', normal ).
adv( 'memorably', normal ).
adv( 'menacingly', normal ).
adv( 'mendaciously', normal ).
adv( 'menially', normal ).
adv( 'mentally', normal ).
adv( 'mercifully', normal ).
adv( 'mercilessly', normal ).
adv( 'merely', normal ).
adv( 'meretriciously', normal ).
adv( 'meritoriously', normal ).
adv( 'merrily', normal ).
adv( 'meseems', normal ).
adv( 'messily', normal ).
adv( 'metaphorically', normal ).
adv( 'metaphysically', normal ).
adv( 'methinks', normal ).
adv( 'methodically', normal ).
adv( 'methodologically', normal ).
adv( 'methought', normal ).
adv( 'meticulously', normal ).
adv( 'metrically', normal ).
adv( 'mezzo', normal ).
adv( 'microscopically', normal ).
adv( 'middling', normal ).
adv( 'midmost', normal ).
adv( 'midships', normal ).
adv( 'midway', normal ).
adv( 'midweek', normal ).
adv( 'mightily', normal ).
adv( 'mighty', normal ).
adv( 'mildly', normal ).
adv( 'militarily', normal ).
adv( 'millionfold', normal ).
adv( 'mincingly', normal ).
adv( 'mindfully', normal ).
adv( 'mindlessly', normal ).
adv( 'minimally', normal ).
adv( 'ministerially', normal ).
adv( 'minutely', normal ).
adv( 'miraculously', normal ).
adv( 'mirthfully', normal ).
adv( 'mischievously', normal ).
adv( 'miserably', normal ).
adv( 'mistakenly', normal ).
adv( 'mistily', normal ).
adv( 'mistrustfully', normal ).
adv( 'mockingly', normal ).
adv( 'moderately', normal ).
adv( 'modestly', normal ).
adv( 'modishly', normal ).
adv( 'moistly', normal ).
adv( 'molto', normal ).
adv( 'momentarily', normal ).
adv( 'momentously', normal ).
adv( 'monotonously', normal ).
adv( 'monstrously', normal ).
adv( 'monthly', normal ).
adv( 'moodily', normal ).
adv( 'morally', normal ).
adv( 'morbidly', normal ).
adv( 'more', normal ).
adv( 'moreover', normal ).
adv( 'morosely', normal ).
adv( 'morphologically', normal ).
adv( 'mortally', normal ).
adv( 'most', normal ).
adv( 'mostly', normal ).
adv( 'motionlessly', normal ).
adv( 'mournfully', normal ).
adv( 'much', normal ).
adv( 'mulishly', normal ).
adv( 'multifariously', normal ).
adv( 'mundanely', normal ).
adv( 'municipally', normal ).
adv( 'munificently', normal ).
adv( 'murderously', normal ).
adv( 'murkily', normal ).
adv( 'musically', normal ).
adv( 'musingly', normal ).
adv( 'mutatis mutandis', normal ).
adv( 'mutely', normal ).
adv( 'mutually', normal ).
adv( 'mysteriously', normal ).
adv( 'naively', normal ).
adv( 'nakedly', normal ).
adv( 'namely', normal ).
adv( 'narrow-mindedly', normal ).
adv( 'narrowly', normal ).
adv( 'nastily', normal ).
adv( 'nationally', normal ).
adv( 'nationwide', normal ).
adv( 'nattily', normal ).
adv( 'naturally', normal ).
adv( 'naughtily', normal ).
adv( 'nay', normal ).
adv( 'ne\'er', normal ).
adv( 'near', normal ).
adv( 'nearer', normal ).
adv( 'nearest', normal ).
adv( 'nearly', normal ).
adv( 'neatly', normal ).
adv( 'necessarily', normal ).
adv( 'needfully', normal ).
adv( 'needlessly', normal ).
adv( 'needs', normal ).
adv( 'nefariously', normal ).
adv( 'negatively', normal ).
adv( 'neglectfully', normal ).
adv( 'negligently', normal ).
adv( 'neither', normal ).
adv( 'nem con', normal ).
adv( 'nervelessly', normal ).
adv( 'nervously', normal ).
adv( 'neurotically', normal ).
adv( 'never', normal ).
adv( 'nevermore', normal ).
adv( 'nevertheless', normal ).
adv( 'new', normal ).
adv( 'newly', normal ).
adv( 'next', normal ).
adv( 'nicely', normal ).
adv( 'nigh', normal ).
adv( 'nigher', normal ).
adv( 'nighest', normal ).
adv( 'nightly', normal ).
adv( 'nimbly', normal ).
adv( 'ninefold', normal ).
adv( 'no', normal ).
adv( 'nobly', normal ).
adv( 'nohow', normal ).
adv( 'noiselessly', normal ).
adv( 'noisily', normal ).
adv( 'nominally', normal ).
adv( 'nonchalantly', normal ).
adv( 'none', normal ).
adv( 'nonstop', normal ).
adv( 'nor\'-east', normal ).
adv( 'nor\'-nor\'-east', normal ).
adv( 'nor\'-nor\'-west', normal ).
adv( 'nor\'-west', normal ).
adv( 'normally', normal ).
adv( 'north', normal ).
adv( 'north-northeast', normal ).
adv( 'north-northwest', normal ).
adv( 'northeast', normal ).
adv( 'northerly', normal ).
adv( 'northwards', normal ).
adv( 'northwest', normal ).
adv( 'nostalgically', normal ).
adv( 'not', normal ).
adv( 'notably', normal ).
adv( 'nothing', normal ).
adv( 'noticeably', normal ).
adv( 'notoriously', normal ).
adv( 'notwithstanding', normal ).
adv( 'now', normal ).
adv( 'nowadays', normal ).
adv( 'nowhere', normal ).
adv( 'nowise', normal ).
adv( 'noxiously', normal ).
adv( 'numbly', normal ).
adv( 'numerically', normal ).
adv( 'nutritionally', normal ).
adv( 'o\'er', normal ).
adv( 'obdurately', normal ).
adv( 'obediently', normal ).
adv( 'objectionably', normal ).
adv( 'objectively', normal ).
adv( 'obligingly', normal ).
adv( 'obliquely', normal ).
adv( 'obnoxiously', normal ).
adv( 'obscenely', normal ).
adv( 'obscurely', normal ).
adv( 'obsequiously', normal ).
adv( 'observably', normal ).
adv( 'observantly', normal ).
adv( 'observingly', normal ).
adv( 'obsessionally', normal ).
adv( 'obsessively', normal ).
adv( 'obstinately', normal ).
adv( 'obstreperously', normal ).
adv( 'obstructively', normal ).
adv( 'obtrusively', normal ).
adv( 'obtusely', normal ).
adv( 'obviously', normal ).
adv( 'occasionally', normal ).
adv( 'oddly', normal ).
adv( 'odds-on', normal ).
adv( 'odiously', normal ).
adv( 'off', normal ).
adv( 'offensively', normal ).
adv( 'offhand', normal ).
adv( 'offhanded', normal ).
adv( 'offhandedly', normal ).
adv( 'officially', normal ).
adv( 'officiously', normal ).
adv( 'offside', normal ).
adv( 'offstage', normal ).
adv( 'oft', normal ).
adv( 'oft-times', normal ).
adv( 'often', normal ).
adv( 'oftener', normal ).
adv( 'okay', normal ).
adv( 'ominously', normal ).
adv( 'on', normal ).
adv( 'once', normal ).
adv( 'onerously', normal ).
adv( 'only', normal ).
adv( 'onshore', normal ).
adv( 'onward', normal ).
adv( 'onwards', normal ).
adv( 'opaquely', normal ).
adv( 'openly', normal ).
adv( 'opportunely', normal ).
adv( 'oppressively', normal ).
adv( 'opprobriously', normal ).
adv( 'optically', normal ).
adv( 'optimally', normal ).
adv( 'optimistically', normal ).
adv( 'optionally', normal ).
adv( 'opulently', normal ).
adv( 'orally', normal ).
adv( 'ordinarily', normal ).
adv( 'organically', normal ).
adv( 'organizationally', normal ).
adv( 'originally', normal ).
adv( 'ornately', normal ).
adv( 'orthogonally', normal ).
adv( 'ostensibly', normal ).
adv( 'ostentatiously', normal ).
adv( 'other', normal ).
adv( 'otherwise', normal ).
adv( 'out', normal ).
adv( 'out-of-doors', normal ).
adv( 'out-of-the-way', normal ).
adv( 'outdoors', normal ).
adv( 'outlandishly', normal ).
adv( 'outrageously', normal ).
adv( 'outright', normal ).
adv( 'outside', normal ).
adv( 'outspokenly', normal ).
adv( 'outstandingly', normal ).
adv( 'outward', normal ).
adv( 'outwardly', normal ).
adv( 'outwards', normal ).
adv( 'over', normal ).
adv( 'overarm', normal ).
adv( 'overbearingly', normal ).
adv( 'overboard', normal ).
adv( 'overhead', normal ).
adv( 'overleaf', normal ).
adv( 'overly', normal ).
adv( 'overmuch', normal ).
adv( 'overnight', normal ).
adv( 'oversea', normal ).
adv( 'overseas', normal ).
adv( 'overside', normal ).
adv( 'overtime', normal ).
adv( 'overtly', normal ).
adv( 'overwhelmingly', normal ).
adv( 'owlishly', normal ).
adv( 'pacifically', normal ).
adv( 'painfully', normal ).
adv( 'painlessly', normal ).
adv( 'painstakingly', normal ).
adv( 'palatably', normal ).
adv( 'palely', normal ).
adv( 'pallidly', normal ).
adv( 'palpably', normal ).
adv( 'pantingly', normal ).
adv( 'par excellence', normal ).
adv( 'paradoxically', normal ).
adv( 'pardonably', normal ).
adv( 'parentally', normal ).
adv( 'parenthetically', normal ).
adv( 'pari passu', normal ).
adv( 'parochially', normal ).
adv( 'part', normal ).
adv( 'part-time', normal ).
adv( 'partially', normal ).
adv( 'particularly', normal ).
adv( 'partly', normal ).
adv( 'passably', normal ).
adv( 'passim', normal ).
adv( 'passing', normal ).
adv( 'passionately', normal ).
adv( 'passively', normal ).
adv( 'past', normal ).
adv( 'pat', normal ).
adv( 'patchily', normal ).
adv( 'patently', normal ).
adv( 'paternally', normal ).
adv( 'pathetically', normal ).
adv( 'pathologically', normal ).
adv( 'patiently', normal ).
adv( 'patriotically', normal ).
adv( 'patronizingly', normal ).
adv( 'pawkily', normal ).
adv( 'peaceably', normal ).
adv( 'peacefully', normal ).
adv( 'peculiarly', normal ).
adv( 'pedantically', normal ).
adv( 'peevishly', normal ).
adv( 'pejoratively', normal ).
adv( 'pell-mell', normal ).
adv( 'pellucidly', normal ).
adv( 'penally', normal ).
adv( 'penetratingly', normal ).
adv( 'penitentially', normal ).
adv( 'penitently', normal ).
adv( 'pensively', normal ).
adv( 'penuriously', normal ).
adv( 'peradventure', normal ).
adv( 'perceptibly', normal ).
adv( 'perceptively', normal ).
adv( 'perceptually', normal ).
adv( 'perchance', normal ).
adv( 'peremptorily', normal ).
adv( 'perennially', normal ).
adv( 'perfectly', normal ).
adv( 'perfidiously', normal ).
adv( 'perforce', normal ).
adv( 'perfunctorily', normal ).
adv( 'perhaps', normal ).
adv( 'perilously', normal ).
adv( 'periodically', normal ).
adv( 'perkily', normal ).
adv( 'permanently', normal ).
adv( 'permissibly', normal ).
adv( 'perniciously', normal ).
adv( 'perpendicularly', normal ).
adv( 'perpetually', normal ).
adv( 'perplexedly', normal ).
adv( 'perseveringly', normal ).
adv( 'persistently', normal ).
adv( 'personally', normal ).
adv( 'perspicuously', normal ).
adv( 'persuasively', normal ).
adv( 'pertinaciously', normal ).
adv( 'pertinently', normal ).
adv( 'pertly', normal ).
adv( 'pervasively', normal ).
adv( 'perversely', normal ).
adv( 'pessimistically', normal ).
adv( 'pettily', normal ).
adv( 'pettishly', normal ).
adv( 'petulantly', normal ).
adv( 'pharmacologically', normal ).
adv( 'phenomenally', normal ).
adv( 'philanthropically', normal ).
adv( 'philosophically', normal ).
adv( 'phlegmatically', normal ).
adv( 'phonetically', normal ).
adv( 'photographically', normal ).
adv( 'phut', normal ).
adv( 'physically', normal ).
adv( 'pianissimo', normal ).
adv( 'piano', normal ).
adv( 'pickaback', normal ).
adv( 'picturesquely', normal ).
adv( 'piecemeal', normal ).
adv( 'piercingly', normal ).
adv( 'pig-headedly', normal ).
adv( 'piggishly', normal ).
adv( 'piously', normal ).
adv( 'piping', normal ).
adv( 'piquantly', normal ).
adv( 'piratically', normal ).
adv( 'pit-a-pat', normal ).
adv( 'piteously', normal ).
adv( 'pithily', normal ).
adv( 'pitiably', normal ).
adv( 'pitifully', normal ).
adv( 'pitilessly', normal ).
adv( 'pityingly', normal ).
adv( 'pizzicato', normal ).
adv( 'placatingly', normal ).
adv( 'placidly', normal ).
adv( 'plaguily', normal ).
adv( 'plain', normal ).
adv( 'plainly', normal ).
adv( 'plaintively', normal ).
adv( 'plausibly', normal ).
adv( 'playfully', normal ).
adv( 'pleadingly', normal ).
adv( 'pleasantly', normal ).
adv( 'pleasingly', normal ).
adv( 'pleasurably', normal ).
adv( 'plenarily', normal ).
adv( 'plenteously', normal ).
adv( 'plentifully', normal ).
adv( 'plenty', normal ).
adv( 'pliantly', normal ).
adv( 'ploddingly', normal ).
adv( 'plonk', normal ).
adv( 'plop', normal ).
adv( 'pluckily', normal ).
adv( 'plumb', normal ).
adv( 'plump', normal ).
adv( 'pneumatically', normal ).
adv( 'poetically', normal ).
adv( 'poignantly', normal ).
adv( 'point-blank', normal ).
adv( 'pointedly', normal ).
adv( 'pointlessly', normal ).
adv( 'poisonously', normal ).
adv( 'polemically', normal ).
adv( 'politely', normal ).
adv( 'politically', normal ).
adv( 'pompously', normal ).
adv( 'ponderously', normal ).
adv( 'poorly', normal ).
adv( 'pop', normal ).
adv( 'popishly', normal ).
adv( 'popularly', normal ).
adv( 'portentously', normal ).
adv( 'positively', normal ).
adv( 'possessively', normal ).
adv( 'possibly', normal ).
adv( 'post meridiem', normal ).
adv( 'post-free', normal ).
adv( 'post-haste', normal ).
adv( 'post-paid', normal ).
adv( 'posthumously', normal ).
adv( 'potentially', normal ).
adv( 'potently', normal ).
adv( 'poutingly', normal ).
adv( 'powerfully', normal ).
adv( 'powerlessly', normal ).
adv( 'practicably', normal ).
adv( 'practically', normal ).
adv( 'pragmatically', normal ).
adv( 'praiseworthily', normal ).
adv( 'pre-eminently', normal ).
adv( 'precariously', normal ).
adv( 'precious', normal ).
adv( 'preciously', normal ).
adv( 'precipitately', normal ).
adv( 'precipitously', normal ).
adv( 'precisely', normal ).
adv( 'precociously', normal ).
adv( 'predictably', normal ).
adv( 'predominantly', normal ).
adv( 'preferably', normal ).
adv( 'prematurely', normal ).
adv( 'preponderantly', normal ).
adv( 'preposterously', normal ).
adv( 'presciently', normal ).
adv( 'presentably', normal ).
adv( 'presently', normal ).
adv( 'pressingly', normal ).
adv( 'prestissimo', normal ).
adv( 'presto', normal ).
adv( 'presumably', normal ).
adv( 'presumptively', normal ).
adv( 'presumptuously', normal ).
adv( 'pretendedly', normal ).
adv( 'pretentiously', normal ).
adv( 'preternaturally', normal ).
adv( 'prettily', normal ).
adv( 'pretty', normal ).
adv( 'previously', normal ).
adv( 'priggishly', normal ).
adv( 'prima facie', normal ).
adv( 'primarily', normal ).
adv( 'primitively', normal ).
adv( 'primly', normal ).
adv( 'principally', normal ).
adv( 'privately', normal ).
adv( 'privily', normal ).
adv( 'pro', normal ).
adv( 'pro forma', normal ).
adv( 'pro rata', normal ).
adv( 'pro tem', normal ).
adv( 'pro tempore', normal ).
adv( 'probabilistically', normal ).
adv( 'probably', normal ).
adv( 'problematically', normal ).
adv( 'prodigally', normal ).
adv( 'prodigiously', normal ).
adv( 'productively', normal ).
adv( 'profanely', normal ).
adv( 'professedly', normal ).
adv( 'professionally', normal ).
adv( 'proficiently', normal ).
adv( 'profitably', normal ).
adv( 'profitlessly', normal ).
adv( 'profoundly', normal ).
adv( 'profusely', normal ).
adv( 'progressively', normal ).
adv( 'prohibitively', normal ).
adv( 'prominently', normal ).
adv( 'promiscuously', normal ).
adv( 'promisingly', normal ).
adv( 'promptly', normal ).
adv( 'pronto', normal ).
adv( 'properly', normal ).
adv( 'prophetically', normal ).
adv( 'propitiously', normal ).
adv( 'proportionally', normal ).
adv( 'proportionately', normal ).
adv( 'prosaically', normal ).
adv( 'prosily', normal ).
adv( 'prosperously', normal ).
adv( 'protectively', normal ).
adv( 'protestingly', normal ).
adv( 'proudly', normal ).
adv( 'proverbially', normal ).
adv( 'providentially', normal ).
adv( 'providently', normal ).
adv( 'provincially', normal ).
adv( 'provisionally', normal ).
adv( 'provocatively', normal ).
adv( 'provokingly', normal ).
adv( 'prox', normal ).
adv( 'prudently', normal ).
adv( 'prudishly', normal ).
adv( 'pruriently', normal ).
adv( 'pryingly', normal ).
adv( 'psychologically', normal ).
adv( 'publicly', normal ).
adv( 'puckishly', normal ).
adv( 'pugnaciously', normal ).
adv( 'punctiliously', normal ).
adv( 'punctually', normal ).
adv( 'pungently', normal ).
adv( 'punily', normal ).
adv( 'purely', normal ).
adv( 'puritanically', normal ).
adv( 'purposefully', normal ).
adv( 'purposelessly', normal ).
adv( 'purposely', normal ).
adv( 'quaintly', normal ).
adv( 'qualitatively', normal ).
adv( 'quantitatively', normal ).
adv( 'quarterly', normal ).
adv( 'queasily', normal ).
adv( 'queerly', normal ).
adv( 'querulously', normal ).
adv( 'questionably', normal ).
adv( 'questioningly', normal ).
adv( 'quick', normal ).
adv( 'quicker', normal ).
adv( 'quickest', normal ).
adv( 'quickly', normal ).
adv( 'quiescently', normal ).
adv( 'quietly', normal ).
adv( 'quite', normal ).
adv( 'quixotically', normal ).
adv( 'quizzically', normal ).
adv( 'racially', normal ).
adv( 'racily', normal ).
adv( 'radially', normal ).
adv( 'radiantly', normal ).
adv( 'radically', normal ).
adv( 'raffishly', normal ).
adv( 'raggedly', normal ).
adv( 'rakishly', normal ).
adv( 'rallentando', normal ).
adv( 'rampantly', normal ).
adv( 'randomly', normal ).
adv( 'rankly', normal ).
adv( 'rapaciously', normal ).
adv( 'rapidly', normal ).
adv( 'rapturously', normal ).
adv( 'rarely', normal ).
adv( 'rashly', normal ).
adv( 'raspingly', normal ).
adv( 'rather', normal ).
adv( 'rationally', normal ).
adv( 'rattling', normal ).
adv( 'raucously', normal ).
adv( 'ravenously', normal ).
adv( 'raving', normal ).
adv( 'ravishingly', normal ).
adv( 'readily', normal ).
adv( 'realistically', normal ).
adv( 'really', normal ).
adv( 'rearwards', normal ).
adv( 'reasonably', normal ).
adv( 'reassuringly', normal ).
adv( 'rebelliously', normal ).
adv( 'rebukingly', normal ).
adv( 'recently', normal ).
adv( 'receptively', normal ).
adv( 'reciprocally', normal ).
adv( 'recklessly', normal ).
adv( 'recognizably', normal ).
adv( 'reflectively', normal ).
adv( 'refreshingly', normal ).
adv( 'regally', normal ).
adv( 'regionally', normal ).
adv( 'regretfully', normal ).
adv( 'regrettably', normal ).
adv( 'regularly', normal ).
adv( 'relatively', normal ).
adv( 'relentlessly', normal ).
adv( 'relevantly', normal ).
adv( 'reliably', normal ).
adv( 'religiously', normal ).
adv( 'reluctantly', normal ).
adv( 'remarkably', normal ).
adv( 'reminiscently', normal ).
adv( 'remorsefully', normal ).
adv( 'remorselessly', normal ).
adv( 'remotely', normal ).
adv( 'rent-free', normal ).
adv( 'repeatedly', normal ).
adv( 'repentantly', normal ).
adv( 'repetitively', normal ).
adv( 'reportedly', normal ).
adv( 'reprehensibly', normal ).
adv( 'reproachfully', normal ).
adv( 'reprovingly', normal ).
adv( 'repulsively', normal ).
adv( 'reputably', normal ).
adv( 'reputedly', normal ).
adv( 'resentfully', normal ).
adv( 'reservedly', normal ).
adv( 'resignedly', normal ).
adv( 'resolutely', normal ).
adv( 'resoundingly', normal ).
adv( 'resourcefully', normal ).
adv( 'respectably', normal ).
adv( 'respectfully', normal ).
adv( 'respectively', normal ).
adv( 'resplendently', normal ).
adv( 'responsibly', normal ).
adv( 'restfully', normal ).
adv( 'restively', normal ).
adv( 'restlessly', normal ).
adv( 'restrictively', normal ).
adv( 'retail', normal ).
adv( 'retentively', normal ).
adv( 'reticently', normal ).
adv( 'retroactively', normal ).
adv( 'retrospectively', normal ).
adv( 'revengefully', normal ).
adv( 'reverentially', normal ).
adv( 'reverently', normal ).
adv( 'reversely', normal ).
adv( 'revoltingly', normal ).
adv( 'rhetorically', normal ).
adv( 'rhythmically', normal ).
adv( 'richly', normal ).
adv( 'ridiculously', normal ).
adv( 'right', normal ).
adv( 'right-down', normal ).
adv( 'righteously', normal ).
adv( 'rightfully', normal ).
adv( 'rightly', normal ).
adv( 'rigidly', normal ).
adv( 'rigorously', normal ).
adv( 'riotously', normal ).
adv( 'ripely', normal ).
adv( 'riskily', normal ).
adv( 'roaring', normal ).
adv( 'robustly', normal ).
adv( 'roguishly', normal ).
adv( 'romantically', normal ).
adv( 'roomily', normal ).
adv( 'rotationally', normal ).
adv( 'rottenly', normal ).
adv( 'rotundly', normal ).
adv( 'rough', normal ).
adv( 'roughly', normal ).
adv( 'round', normal ).
adv( 'round-arm', normal ).
adv( 'round-the-clock', normal ).
adv( 'roundly', normal ).
adv( 'rowdily', normal ).
adv( 'royally', normal ).
adv( 'rudely', normal ).
adv( 'ruefully', normal ).
adv( 'ruggedly', normal ).
adv( 'ruinously', normal ).
adv( 'ruthlessly', normal ).
adv( 'sacredly', normal ).
adv( 'sadly', normal ).
adv( 'safely', normal ).
adv( 'sagaciously', normal ).
adv( 'sagely', normal ).
adv( 'salaciously', normal ).
adv( 'same', normal ).
adv( 'sanctimoniously', normal ).
adv( 'sanely', normal ).
adv( 'sapiently', normal ).
adv( 'sarcastically', normal ).
adv( 'sardonically', normal ).
adv( 'satirically', normal ).
adv( 'satisfactorily', normal ).
adv( 'satisfyingly', normal ).
adv( 'saucily', normal ).
adv( 'savagely', normal ).
adv( 'scandalously', normal ).
adv( 'scantily', normal ).
adv( 'scarcely', normal ).
adv( 'scathingly', normal ).
adv( 'scenically', normal ).
adv( 'sceptically', normal ).
adv( 'schematically', normal ).
adv( 'scientifically', normal ).
adv( 'scoffingly', normal ).
adv( 'scorching', normal ).
adv( 'scornfully', normal ).
adv( 'scot-free', normal ).
adv( 'scrappily', normal ).
adv( 'screamingly', normal ).
adv( 'scrupulously', normal ).
adv( 'scurrilously', normal ).
adv( 'scurvily', normal ).
adv( 'searchingly', normal ).
adv( 'seasonally', normal ).
adv( 'seawards', normal ).
adv( 'second', normal ).
adv( 'second-best', normal ).
adv( 'second-class', normal ).
adv( 'secondarily', normal ).
adv( 'secondly', normal ).
adv( 'secretively', normal ).
adv( 'secretly', normal ).
adv( 'securely', normal ).
adv( 'sedately', normal ).
adv( 'seductively', normal ).
adv( 'sedulously', normal ).
adv( 'seedily', normal ).
adv( 'seemingly', normal ).
adv( 'seldom', normal ).
adv( 'selectively', normal ).
adv( 'selfconsciously', normal ).
adv( 'selfishly', normal ).
adv( 'sensationally', normal ).
adv( 'senselessly', normal ).
adv( 'sensibly', normal ).
adv( 'sensitively', normal ).
adv( 'sensuously', normal ).
adv( 'sententiously', normal ).
adv( 'sentimentally', normal ).
adv( 'separably', normal ).
adv( 'separately', normal ).
adv( 'sequentially', normal ).
adv( 'serenely', normal ).
adv( 'serially', normal ).
adv( 'seriatim', normal ).
adv( 'seriously', normal ).
adv( 'servilely', normal ).
adv( 'sevenfold', normal ).
adv( 'seventhly', normal ).
adv( 'severally', normal ).
adv( 'severely', normal ).
adv( 'sexually', normal ).
adv( 'shabbily', normal ).
adv( 'shaggily', normal ).
adv( 'shakily', normal ).
adv( 'shamefacedly', normal ).
adv( 'shamefully', normal ).
adv( 'shamelessly', normal ).
adv( 'shapelessly', normal ).
adv( 'sharp', normal ).
adv( 'sharply', normal ).
adv( 'sheepishly', normal ).
adv( 'sheer', normal ).
adv( 'shiftily', normal ).
adv( 'shipshape', normal ).
adv( 'shock', normal ).
adv( 'shockingly', normal ).
adv( 'short', normal ).
adv( 'shortly', normal ).
adv( 'showily', normal ).
adv( 'shrewdly', normal ).
adv( 'shrewishly', normal ).
adv( 'shrilly', normal ).
adv( 'shudderingly', normal ).
adv( 'shyly', normal ).
adv( 'sic', normal ).
adv( 'sickeningly', normal ).
adv( 'side-face', normal ).
adv( 'side-saddle', normal ).
adv( 'sidelong', normal ).
adv( 'sidewards', normal ).
adv( 'sideways', normal ).
adv( 'signally', normal ).
adv( 'significantly', normal ).
adv( 'silently', normal ).
adv( 'silkily', normal ).
adv( 'similarly', normal ).
adv( 'simperingly', normal ).
adv( 'simply', normal ).
adv( 'simultaneously', normal ).
adv( 'since', normal ).
adv( 'sincerely', normal ).
adv( 'sine die', normal ).
adv( 'single-handed', normal ).
adv( 'singly', normal ).
adv( 'singularly', normal ).
adv( 'sixfold', normal ).
adv( 'sixthly', normal ).
adv( 'skeptically', normal ).
adv( 'sketchily', normal ).
adv( 'skilfully', normal ).
adv( 'skimpily', normal ).
adv( 'skittishly', normal ).
adv( 'sky-high', normal ).
adv( 'skyward', normal ).
adv( 'skywards', normal ).
adv( 'slackly', normal ).
adv( 'slangily', normal ).
adv( 'slantingly', normal ).
adv( 'slantwise', normal ).
adv( 'slap', normal ).
adv( 'slap-bang', normal ).
adv( 'slapdash', normal ).
adv( 'slavishly', normal ).
adv( 'sleekly', normal ).
adv( 'sleepily', normal ).
adv( 'sleeplessly', normal ).
adv( 'slenderly', normal ).
adv( 'slick', normal ).
adv( 'slightingly', normal ).
adv( 'slightly', normal ).
adv( 'slimly', normal ).
adv( 'slopingly', normal ).
adv( 'sloppily', normal ).
adv( 'slouchingly', normal ).
adv( 'slow', normal ).
adv( 'slower', normal ).
adv( 'slowest', normal ).
adv( 'slowly', normal ).
adv( 'sluggishly', normal ).
adv( 'slyly', normal ).
adv( 'smack', normal ).
adv( 'small', normal ).
adv( 'smartly', normal ).
adv( 'smash', normal ).
adv( 'smilingly', normal ).
adv( 'smoothly', normal ).
adv( 'smugly', normal ).
adv( 'smuttily', normal ).
adv( 'snappishly', normal ).
adv( 'sneakingly', normal ).
adv( 'sneeringly', normal ).
adv( 'snobbishly', normal ).
adv( 'snootily', normal ).
adv( 'snugly', normal ).
adv( 'so', normal ).
adv( 'so-so', normal ).
adv( 'sobbingly', normal ).
adv( 'soberly', normal ).
adv( 'sociably', normal ).
adv( 'socially', normal ).
adv( 'sociologically', normal ).
adv( 'sock', normal ).
adv( 'softly', normal ).
adv( 'solely', normal ).
adv( 'solemnly', normal ).
adv( 'solicitously', normal ).
adv( 'solidly', normal ).
adv( 'solitarily', normal ).
adv( 'sombrely', normal ).
adv( 'some', normal ).
adv( 'someday', normal ).
adv( 'somehow', normal ).
adv( 'someplace', normal ).
adv( 'something', normal ).
adv( 'sometime', normal ).
adv( 'sometimes', normal ).
adv( 'someway', normal ).
adv( 'somewhat', normal ).
adv( 'somewhere', normal ).
adv( 'somnolently', normal ).
adv( 'sonorously', normal ).
adv( 'soon', normal ).
adv( 'sooner', normal ).
adv( 'soonest', normal ).
adv( 'soothingly', normal ).
adv( 'sopping', normal ).
adv( 'sordidly', normal ).
adv( 'sorely', normal ).
adv( 'sorrowfully', normal ).
adv( 'sottishly', normal ).
adv( 'sotto voce', normal ).
adv( 'sou\'-east', normal ).
adv( 'sou\'-sou\'-east', normal ).
adv( 'sou\'-sou\'-west', normal ).
adv( 'sou\'-west', normal ).
adv( 'soulfully', normal ).
adv( 'soullessly', normal ).
adv( 'sound', normal ).
adv( 'soundlessly', normal ).
adv( 'soundly', normal ).
adv( 'sourly', normal ).
adv( 'south', normal ).
adv( 'south-southeast', normal ).
adv( 'south-southwest', normal ).
adv( 'southeast', normal ).
adv( 'southerly', normal ).
adv( 'southward', normal ).
adv( 'southwards', normal ).
adv( 'southwest', normal ).
adv( 'spaciously', normal ).
adv( 'sparely', normal ).
adv( 'sparingly', normal ).
adv( 'sparsely', normal ).
adv( 'spasmodically', normal ).
adv( 'spatially', normal ).
adv( 'specially', normal ).
adv( 'specifically', normal ).
adv( 'speciously', normal ).
adv( 'spectacularly', normal ).
adv( 'speculatively', normal ).
adv( 'speechlessly', normal ).
adv( 'speedily', normal ).
adv( 'spicily', normal ).
adv( 'spirally', normal ).
adv( 'spiritually', normal ).
adv( 'spitefully', normal ).
adv( 'splendidly', normal ).
adv( 'spontaneously', normal ).
adv( 'sporadically', normal ).
adv( 'sportingly', normal ).
adv( 'sportively', normal ).
adv( 'spotlessly', normal ).
adv( 'sprucely', normal ).
adv( 'spuriously', normal ).
adv( 'squalidly', normal ).
adv( 'square', normal ).
adv( 'squarely', normal ).
adv( 'squeamishly', normal ).
adv( 'staccato', normal ).
adv( 'staggeringly', normal ).
adv( 'stagily', normal ).
adv( 'staidly', normal ).
adv( 'stammeringly', normal ).
adv( 'standoffishly', normal ).
adv( 'staring', normal ).
adv( 'stark', normal ).
adv( 'starkly', normal ).
adv( 'startlingly', normal ).
adv( 'statically', normal ).
adv( 'statistically', normal ).
adv( 'statutorily', normal ).
adv( 'staunchly', normal ).
adv( 'steadfastly', normal ).
adv( 'steadily', normal ).
adv( 'steady', normal ).
adv( 'stealthily', normal ).
adv( 'steeply', normal ).
adv( 'sternly', normal ).
adv( 'stertorously', normal ).
adv( 'stickily', normal ).
adv( 'stiff', normal ).
adv( 'stiffly', normal ).
adv( 'still', normal ).
adv( 'stiltedly', normal ).
adv( 'stingily', normal ).
adv( 'stirringly', normal ).
adv( 'stochastically', normal ).
adv( 'stock-still', normal ).
adv( 'stockily', normal ).
adv( 'stoically', normal ).
adv( 'stolidly', normal ).
adv( 'stonily', normal ).
adv( 'stormily', normal ).
adv( 'stoutly', normal ).
adv( 'straight', normal ).
adv( 'straightforwardly', normal ).
adv( 'straightway', normal ).
adv( 'strangely', normal ).
adv( 'strategically', normal ).
adv( 'strenuously', normal ).
adv( 'strictly', normal ).
adv( 'stridently', normal ).
adv( 'strikingly', normal ).
adv( 'stringently', normal ).
adv( 'strongly', normal ).
adv( 'structurally', normal ).
adv( 'stubbornly', normal ).
adv( 'studiously', normal ).
adv( 'stuffily', normal ).
adv( 'stunningly', normal ).
adv( 'stupendously', normal ).
adv( 'stupidly', normal ).
adv( 'sturdily', normal ).
adv( 'stutteringly', normal ).
adv( 'stylishly', normal ).
adv( 'stylistically', normal ).
adv( 'suavely', normal ).
adv( 'sub rosa', normal ).
adv( 'subconsciously', normal ).
adv( 'subjectively', normal ).
adv( 'sublimely', normal ).
adv( 'submissively', normal ).
adv( 'subsequently', normal ).
adv( 'subserviently', normal ).
adv( 'substantially', normal ).
adv( 'subtly', normal ).
adv( 'successfully', normal ).
adv( 'successively', normal ).
adv( 'succinctly', normal ).
adv( 'suddenly', normal ).
adv( 'sufficiently', normal ).
adv( 'suggestively', normal ).
adv( 'suitably', normal ).
adv( 'sulkily', normal ).
adv( 'sullenly', normal ).
adv( 'sultrily', normal ).
adv( 'summarily', normal ).
adv( 'sumptuously', normal ).
adv( 'sunnily', normal ).
adv( 'superbly', normal ).
adv( 'superciliously', normal ).
adv( 'superficially', normal ).
adv( 'superfluously', normal ).
adv( 'supernaturally', normal ).
adv( 'superstitiously', normal ).
adv( 'supinely', normal ).
adv( 'supposedly', normal ).
adv( 'supra', normal ).
adv( 'supremely', normal ).
adv( 'sure', normal ).
adv( 'surely', normal ).
adv( 'surgically', normal ).
adv( 'surlily', normal ).
adv( 'surpassingly', normal ).
adv( 'surprisedly', normal ).
adv( 'surprisingly', normal ).
adv( 'surreptitiously', normal ).
adv( 'suspiciously', normal ).
adv( 'sweepingly', normal ).
adv( 'sweetly', normal ).
adv( 'swiftly', normal ).
adv( 'swimmingly', normal ).
adv( 'symbolically', normal ).
adv( 'symmetrically', normal ).
adv( 'sympathetically', normal ).
adv( 'symptomatically', normal ).
adv( 'synchronously', normal ).
adv( 'synoptically', normal ).
adv( 'syntactically', normal ).
adv( 'synthetically', normal ).
adv( 'systematically', normal ).
adv( 't^ete-`a-t^ete', normal ).
adv( 'table d\'h^ote', normal ).
adv( 'tacitly', normal ).
adv( 'taciturnly', normal ).
adv( 'tactfully', normal ).
adv( 'tactically', normal ).
adv( 'tactlessly', normal ).
adv( 'tamely', normal ).
adv( 'tandem', normal ).
adv( 'tangibly', normal ).
adv( 'tardily', normal ).
adv( 'tartly', normal ).
adv( 'tastefully', normal ).
adv( 'tastelessly', normal ).
adv( 'tastily', normal ).
adv( 'tattily', normal ).
adv( 'tauntingly', normal ).
adv( 'tautly', normal ).
adv( 'tawdrily', normal ).
adv( 'tearfully', normal ).
adv( 'teasingly', normal ).
adv( 'technically', normal ).
adv( 'tediously', normal ).
adv( 'telegraphically', normal ).
adv( 'telescopically', normal ).
adv( 'tellingly', normal ).
adv( 'temperamentally', normal ).
adv( 'temperately', normal ).
adv( 'temporarily', normal ).
adv( 'temptingly', normal ).
adv( 'tenaciously', normal ).
adv( 'tendentiously', normal ).
adv( 'tenderly', normal ).
adv( 'tenfold', normal ).
adv( 'tensely', normal ).
adv( 'tentatively', normal ).
adv( 'tenthly', normal ).
adv( 'tenuously', normal ).
adv( 'tepidly', normal ).
adv( 'terminally', normal ).
adv( 'terribly', normal ).
adv( 'terrifically', normal ).
adv( 'tersely', normal ).
adv( 'testily', normal ).
adv( 'tetchily', normal ).
adv( 'thankfully', normal ).
adv( 'that', normal ).
adv( 'the', normal ).
adv( 'theatrically', normal ).
adv( 'then', normal ).
adv( 'thence', normal ).
adv( 'thenceforth', normal ).
adv( 'thenceforward', normal ).
adv( 'theologically', normal ).
adv( 'theoretically', normal ).
adv( 'there', normal ).
adv( 'thereabout', normal ).
adv( 'thereabouts', normal ).
adv( 'thereafter', normal ).
adv( 'thereby', normal ).
adv( 'therefore', normal ).
adv( 'therefrom', normal ).
adv( 'therein', normal ).
adv( 'thereinafter', normal ).
adv( 'thereof', normal ).
adv( 'thereon', normal ).
adv( 'thereto', normal ).
adv( 'thereunder', normal ).
adv( 'thereupon', normal ).
adv( 'therewith', normal ).
adv( 'therewithal', normal ).
adv( 'thermally', normal ).
adv( 'thermostatically', normal ).
adv( 'thick', normal ).
adv( 'thickly', normal ).
adv( 'thievishly', normal ).
adv( 'thin', normal ).
adv( 'thinly', normal ).
adv( 'thirdly', normal ).
adv( 'thirstily', normal ).
adv( 'this', normal ).
adv( 'thither', normal ).
adv( 'tho\'', normal ).
adv( 'thoroughly', normal ).
adv( 'though', normal ).
adv( 'thoughtfully', normal ).
adv( 'thoughtlessly', normal ).
adv( 'thousandfold', normal ).
adv( 'threateningly', normal ).
adv( 'threefold', normal ).
adv( 'thrice', normal ).
adv( 'thriftily', normal ).
adv( 'thriftlessly', normal ).
adv( 'through', normal ).
adv( 'throughout', normal ).
adv( 'thus', normal ).
adv( 'tidily', normal ).
adv( 'tight', normal ).
adv( 'tightly', normal ).
adv( 'timidly', normal ).
adv( 'timorously', normal ).
adv( 'tip-top', normal ).
adv( 'tiptoe', normal ).
adv( 'tirelessly', normal ).
adv( 'tiresomely', normal ).
adv( 'tiring', normal ).
adv( 'to', normal ).
adv( 'today', normal ).
adv( 'together', normal ).
adv( 'tolerably', normal ).
adv( 'tolerantly', normal ).
adv( 'tomorrow', normal ).
adv( 'tonelessly', normal ).
adv( 'tongue-in-cheek', normal ).
adv( 'tonight', normal ).
adv( 'too', normal ).
adv( 'topically', normal ).
adv( 'topographically', normal ).
adv( 'toppingly', normal ).
adv( 'topsy-turvy', normal ).
adv( 'torpidly', normal ).
adv( 'tortuously', normal ).
adv( 'totally', normal ).
adv( 'touchily', normal ).
adv( 'touchingly', normal ).
adv( 'toughly', normal ).
adv( 'tout ensemble', normal ).
adv( 'traditionally', normal ).
adv( 'tragically', normal ).
adv( 'traitorously', normal ).
adv( 'tranquilly', normal ).
adv( 'transcendentally', normal ).
adv( 'transiently', normal ).
adv( 'transitionally', normal ).
adv( 'transitively', normal ).
adv( 'transparently', normal ).
adv( 'transversely', normal ).
adv( 'treacherously', normal ).
adv( 'treasonably', normal ).
adv( 'tremendously', normal ).
adv( 'tremulously', normal ).
adv( 'trenchantly', normal ).
adv( 'trimly', normal ).
adv( 'trippingly', normal ).
adv( 'tritely', normal ).
adv( 'triumphantly', normal ).
adv( 'trivially', normal ).
adv( 'tropically', normal ).
adv( 'truculently', normal ).
adv( 'true', normal ).
adv( 'truly', normal ).
adv( 'trustfully', normal ).
adv( 'trustingly', normal ).
adv( 'truthfully', normal ).
adv( 'tumultuously', normal ).
adv( 'tunefully', normal ).
adv( 'turbulently', normal ).
adv( 'turgidly', normal ).
adv( 'tutorially', normal ).
adv( 'twice', normal ).
adv( 'twofold', normal ).
adv( 'typically', normal ).
adv( 'typographically', normal ).
adv( 'ultimately', normal ).
adv( 'ultra vires', normal ).
adv( 'unacceptably', normal ).
adv( 'unaccountably', normal ).
adv( 'unadvisedly', normal ).
adv( 'unalterably', normal ).
adv( 'unambiguously', normal ).
adv( 'unanimously', normal ).
adv( 'unarguably', normal ).
adv( 'unashamedly', normal ).
adv( 'unassailably', normal ).
adv( 'unassumingly', normal ).
adv( 'unattainably', normal ).
adv( 'unattractively', normal ).
adv( 'unavoidably', normal ).
adv( 'unawares', normal ).
adv( 'unbearably', normal ).
adv( 'unbecomingly', normal ).
adv( 'unbeknown', normal ).
adv( 'unbeknownst', normal ).
adv( 'unbelievably', normal ).
adv( 'unbelievingly', normal ).
adv( 'unblushingly', normal ).
adv( 'unblushingly', normal ).
adv( 'uncannily', normal ).
adv( 'unceasingly', normal ).
adv( 'unceremoniously', normal ).
adv( 'uncertainly', normal ).
adv( 'uncharacteristically', normal ).
adv( 'unchivalrously', normal ).
adv( 'unco', normal ).
adv( 'uncomfortably', normal ).
adv( 'uncommonly', normal ).
adv( 'uncomplainingly', normal ).
adv( 'unconcernedly', normal ).
adv( 'unconditionally', normal ).
adv( 'unconsciously', normal ).
adv( 'unconstitutionally', normal ).
adv( 'uncontrollably', normal ).
adv( 'uncouthly', normal ).
adv( 'uncritically', normal ).
adv( 'unctuously', normal ).
adv( 'undemocratically', normal ).
adv( 'undeniably', normal ).
adv( 'under', normal ).
adv( 'underarm', normal ).
adv( 'underfoot', normal ).
adv( 'underground', normal ).
adv( 'underhand', normal ).
adv( 'underneath', normal ).
adv( 'understandably', normal ).
adv( 'undiplomatically', normal ).
adv( 'undisputedly', normal ).
adv( 'undoubtedly', normal ).
adv( 'undramatically', normal ).
adv( 'unduly', normal ).
adv( 'uneasily', normal ).
adv( 'unemotionally', normal ).
adv( 'unendingly', normal ).
adv( 'unenthusiastically', normal ).
adv( 'unequally', normal ).
adv( 'unequivocally', normal ).
adv( 'unerringly', normal ).
adv( 'unethically', normal ).
adv( 'unevenly', normal ).
adv( 'uneventfully', normal ).
adv( 'unexpectedly', normal ).
adv( 'unfailingly', normal ).
adv( 'unfairly', normal ).
adv( 'unfaithfully', normal ).
adv( 'unfalteringly', normal ).
adv( 'unfashionably', normal ).
adv( 'unfavourably', normal ).
adv( 'unfeelingly', normal ).
adv( 'unfeignedly', normal ).
adv( 'unforgettably', normal ).
adv( 'unforgivably', normal ).
adv( 'unfortunately', normal ).
adv( 'ungraciously', normal ).
adv( 'ungrammatically', normal ).
adv( 'ungratefully', normal ).
adv( 'ungrudgingly', normal ).
adv( 'unhappily', normal ).
adv( 'unhelpfully', normal ).
adv( 'unhesitatingly', normal ).
adv( 'unhurriedly', normal ).
adv( 'unhygienically', normal ).
adv( 'uniformly', normal ).
adv( 'unilaterally', normal ).
adv( 'unimaginably', normal ).
adv( 'unimaginatively', normal ).
adv( 'unimpressively', normal ).
adv( 'unintelligently', normal ).
adv( 'unintelligibly', normal ).
adv( 'unintentionally', normal ).
adv( 'uninterestingly', normal ).
adv( 'uninterruptedly', normal ).
adv( 'uniquely', normal ).
adv( 'unitedly', normal ).
adv( 'universally', normal ).
adv( 'unjustifiably', normal ).
adv( 'unjustly', normal ).
adv( 'unkindly', normal ).
adv( 'unknowingly', normal ).
adv( 'unlawfully', normal ).
adv( 'unluckily', normal ).
adv( 'unmanageably', normal ).
adv( 'unmemorably', normal ).
adv( 'unmercifully', normal ).
adv( 'unmistakably', normal ).
adv( 'unmusically', normal ).
adv( 'unnaturally', normal ).
adv( 'unnecessarily', normal ).
adv( 'unobtrusively', normal ).
adv( 'unofficially', normal ).
adv( 'unpalatably', normal ).
adv( 'unpardonably', normal ).
adv( 'unpatriotically', normal ).
adv( 'unpleasantly', normal ).
adv( 'unprecedentedly', normal ).
adv( 'unpredictably', normal ).
adv( 'unproductively', normal ).
adv( 'unprofitably', normal ).
adv( 'unquestionably', normal ).
adv( 'unquote', normal ).
adv( 'unrealistically', normal ).
adv( 'unreasonably', normal ).
adv( 'unrecognizably', normal ).
adv( 'unreliably', normal ).
adv( 'unremarkably', normal ).
adv( 'unreservedly', normal ).
adv( 'unrighteously', normal ).
adv( 'unromantically', normal ).
adv( 'unsatisfactorily', normal ).
adv( 'unscientifically', normal ).
adv( 'unscrupulously', normal ).
adv( 'unseasonably', normal ).
adv( 'unselfconsciously', normal ).
adv( 'unselfishly', normal ).
adv( 'unshakably', normal ).
adv( 'unsociably', normal ).
adv( 'unsparingly', normal ).
adv( 'unspeakably', normal ).
adv( 'unspecifically', normal ).
adv( 'unsportingly', normal ).
adv( 'unsteadily', normal ).
adv( 'unsuccessfully', normal ).
adv( 'unsuitably', normal ).
adv( 'unsuspectingly', normal ).
adv( 'unswervingly', normal ).
adv( 'unsympathetically', normal ).
adv( 'unsystematically', normal ).
adv( 'unthinkingly', normal ).
adv( 'untidily', normal ).
adv( 'untruly', normal ).
adv( 'untruthfully', normal ).
adv( 'untypically', normal ).
adv( 'unusually', normal ).
adv( 'unutterably', normal ).
adv( 'unwarily', normal ).
adv( 'unwarrantably', normal ).
adv( 'unwaveringly', normal ).
adv( 'unwillingly', normal ).
adv( 'unwisely', normal ).
adv( 'unwittingly', normal ).
adv( 'unwontedly', normal ).
adv( 'unworthily', normal ).
adv( 'up', normal ).
adv( 'upcountry', normal ).
adv( 'uphill', normal ).
adv( 'uppermost', normal ).
adv( 'uppishly', normal ).
adv( 'uprightly', normal ).
adv( 'uproariously', normal ).
adv( 'upside-down', normal ).
adv( 'upstage', normal ).
adv( 'upstairs', normal ).
adv( 'upstream', normal ).
adv( 'uptown', normal ).
adv( 'upward', normal ).
adv( 'upwards', normal ).
adv( 'urbanely', normal ).
adv( 'urgently', normal ).
adv( 'usefully', normal ).
adv( 'uselessly', normal ).
adv( 'usually', normal ).
adv( 'utterly', normal ).
adv( 'uxoriously', normal ).
adv( 'vacantly', normal ).
adv( 'vacuously', normal ).
adv( 'vaguely', normal ).
adv( 'vainly', normal ).
adv( 'valiantly', normal ).
adv( 'validly', normal ).
adv( 'vapidly', normal ).
adv( 'variably', normal ).
adv( 'variously', normal ).
adv( 'vastly', normal ).
adv( 'vauntingly', normal ).
adv( 'vehemently', normal ).
adv( 'venally', normal ).
adv( 'venomously', normal ).
adv( 'veraciously', normal ).
adv( 'verbally', normal ).
adv( 'verbatim', normal ).
adv( 'verbosely', normal ).
adv( 'verily', normal ).
adv( 'vertically', normal ).
adv( 'very', normal ).
adv( 'vicariously', normal ).
adv( 'vice versa', normal ).
adv( 'viciously', normal ).
adv( 'victoriously', normal ).
adv( 'videlicet', normal ).
adv( 'vigilantly', normal ).
adv( 'vigorously', normal ).
adv( 'vilely', normal ).
adv( 'vindictively', normal ).
adv( 'violently', normal ).
adv( 'virtually', normal ).
adv( 'virtuously', normal ).
adv( 'virulently', normal ).
adv( 'vis-`a-vis', normal ).
adv( 'visibly', normal ).
adv( 'visually', normal ).
adv( 'vitally', normal ).
adv( 'viva voce', normal ).
adv( 'vivace', normal ).
adv( 'vivaciously', normal ).
adv( 'vividly', normal ).
adv( 'vocally', normal ).
adv( 'volubly', normal ).
adv( 'voluntarily', normal ).
adv( 'voluptuously', normal ).
adv( 'voraciously', normal ).
adv( 'voyeuristically', normal ).
adv( 'vulgarly', normal ).
adv( 'waggishly', normal ).
adv( 'waist-deep', normal ).
adv( 'waist-high', normal ).
adv( 'wanly', normal ).
adv( 'wantonly', normal ).
adv( 'warily', normal ).
adv( 'warmly', normal ).
adv( 'wastefully', normal ).
adv( 'watchfully', normal ).
adv( 'way', normal ).
adv( 'weakly', normal ).
adv( 'wealthily', normal ).
adv( 'wearily', normal ).
adv( 'weekly', normal ).
adv( 'weightily', normal ).
adv( 'weirdly', normal ).
adv( 'well', normal ).
adv( 'well-nigh', normal ).
adv( 'west', normal ).
adv( 'westerly', normal ).
adv( 'westward', normal ).
adv( 'westwards', normal ).
adv( 'whacking', normal ).
adv( 'whang', normal ).
adv( 'wheezily', normal ).
adv( 'when', whq ).
adv( 'when', whrel ).
adv( 'whence', normal ).
adv( 'whenever', normal ).
adv( 'where', whq ).
adv( 'where', whrel ).
adv( 'whereabouts', normal ).
adv( 'whereat', normal ).
adv( 'whereby', normal ).
adv( 'wherefore', normal ).
adv( 'wherein', normal ).
adv( 'whereof', normal ).
adv( 'whereon', normal ).
adv( 'wheresoever', normal ).
adv( 'whereto', normal ).
adv( 'whereunto', normal ).
adv( 'whereupon', normal ).
adv( 'wherever', normal ).
adv( 'wherewith', normal ).
adv( 'wherewithal', normal ).
adv( 'whimsically', normal ).
adv( 'whither', normal ).
adv( 'whithersoever', normal ).
adv( 'wholeheartedly', normal ).
adv( 'wholesale', normal ).
adv( 'wholesomely', normal ).
adv( 'wholly', normal ).
adv( 'whopping', normal ).
adv( 'why', whq ).
adv( 'why', whrel ).
adv( 'wickedly', normal ).
adv( 'wide', normal ).
adv( 'widely', normal ).
adv( 'wild', normal ).
adv( 'wildly', normal ).
adv( 'wilfully', normal ).
adv( 'willingly', normal ).
adv( 'willy-nilly', normal ).
adv( 'windily', normal ).
adv( 'winsomely', normal ).
adv( 'wisely', normal ).
adv( 'wishfully', normal ).
adv( 'wistfully', normal ).
adv( 'withal', normal ).
adv( 'witheringly', normal ).
adv( 'within', normal ).
adv( 'without', normal ).
adv( 'wittily', normal ).
adv( 'wittingly', normal ).
adv( 'woefully', normal ).
adv( 'wonderfully', normal ).
adv( 'wonderingly', normal ).
adv( 'wondrous', normal ).
adv( 'wordily', normal ).
adv( 'worryingly', normal ).
adv( 'worse', normal ).
adv( 'worst', normal ).
adv( 'worthily', normal ).
adv( 'worthlessly', normal ).
adv( 'wrathfully', normal ).
adv( 'wretchedly', normal ).
adv( 'wrong', normal ).
adv( 'wrong-headedly', normal ).
adv( 'wrongfully', normal ).
adv( 'wrongly', normal ).
adv( 'wryly', normal ).
adv( 'yea', normal ).
adv( 'yeah', normal ).
adv( 'yearly', normal ).
adv( 'yearningly', normal ).
adv( 'yesterday', normal ).
adv( 'yet', normal ).
adv( 'yieldingly', normal ).
adv( 'yon', normal ).
adv( 'yonder', normal ).
adv( 'youthfully', normal ).
adv( 'zealously', normal ).
adv( 'zestfully', normal ).
adv( 'zigzag', normal ).

pron( '\'em', acc, normal, _ ).
pron( '\'un', _, normal, _ ).
pron( 'i', nom, normal, per ).
pron( 'aforesaid', _, normal, _ ).
pron( 'all', _, normal, _ ).
pron( 'another', _, normal, _ ).
pron( 'any', _, normal, _ ).
pron( 'anybody', _, normal, per ).
pron( 'anyone', _, normal, per ).
pron( 'anything', _, normal, nper ).
pron( 'best', _, normal, _ ).
pron( 'both', _, normal, _ ).
pron( 'but', _, whrel, _ ).
pron( 'each', _, normal, _ ).
pron( 'either', _, normal, _ ).
pron( 'everybody', _, normal, per ).
pron( 'everyone', _, normal, per ).
pron( 'everything', _, normal, nper ).
pron( 'few', _, normal, _ ).
pron( 'he', nom, normal, per ).
pron( 'her', acc, normal, per ).
pron( 'hers', _, normal, _ ).
pron( 'herself', acc, normal, per ).
pron( 'him', acc, normal, per ).
pron( 'himself', acc, normal, per ).
pron( 'his', _, normal, _ ).
pron( 'it', _, normal, nper ).
pron( 'itself', acc, normal, nper ).
pron( 'me', acc, normal, per ).
pron( 'mine', _, normal, _ ).
pron( 'myself', acc, normal, per ).
pron( 'neither', _, normal, _ ).
pron( 'no one', _, normal, per ).
pron( 'no-one', _, normal, per ).
pron( 'nobody', _, normal, per ).
pron( 'none', _, normal, _ ).
pron( 'one', _, normal, _ ).
pron( 'oneself', acc, normal, per ).
pron( 'ours', _, normal, _ ).
pron( 'ourselves', acc, normal, per ).
pron( 'own', _, normal, _ ).
pron( 'same', _, normal, _ ).
pron( 'several', _, normal, _ ).
pron( 'she', nom, normal, per ).
pron( 'some', _, normal, _ ).
pron( 'somebody', _, normal, per ).
pron( 'someone', _, normal, per ).
pron( 'something', _, normal, nper ).
pron( 'such', _, normal, _ ).
pron( 'summat', _, normal, nper ).
pron( 'that', _, normal, nper ).
pron( 'that', _, whrel, _ ).
pron( 'thee', acc, normal, per ).
pron( 'theirs', _, normal, _ ).
pron( 'them', acc, normal, per ).
pron( 'themselves', acc, normal, per ).
pron( 'these', _, normal, nper ).
pron( 'they', nom, normal, per ).
pron( 'thine', _, normal, _ ).
pron( 'this', _, normal, nper ).
pron( 'those', _, normal, nper ).
pron( 'thou', nom, normal, per ).
pron( 'thyself', acc, normal, per ).
pron( 'us', acc, normal, per ).
pron( 'we', nom, normal, per ).
pron( 'what', _, whq, nper ).
pron( 'what', _, whrel, nper ).
pron( 'whatever', _, normal, nper ).
pron( 'which', _, whq, _ ).
pron( 'which', _, whrel, _ ).
pron( 'whichever', _, normal, _ ).
pron( 'whichsoever', _, normal, _ ).
pron( 'who', _, whq, per ).
pron( 'who', _, whrel, per ).
pron( 'whoever', _, normal, per ).
pron( 'whom', acc, whq, per ).
pron( 'whom', acc, whrel, per ).
pron( 'whose', _, normal, _ ).
pron( 'whoso', _, normal, per ).
pron( 'whosoever', _, normal, per ).
pron( 'ye', _, normal, per ).
pron( 'you', _, normal, per ).
pron( 'yours', _, normal, _ ).
pron( 'yourself', acc, normal, per ).
pron( 'yourselves', acc, normal, per ).

det( 'a', indef, exists ).
det( 'an', indef, exists ).
det( 'the', def, the ).
det( 'ye', def, the ).
% manual additions
det( 'every', indef, forall ).
det( 'some', indef, exists ).

prep( '\'neath', prep ).
prep( '\'tween', prep ).
prep( '\'twixt', prep ).
prep( 'abaft', prep ).
prep( 'aboard', prep ).
prep( 'about', prep ).
prep( 'above', prep ).
prep( 'according to', prep ).
prep( 'across', prep ).
prep( 'afore', prep ).
prep( 'after', prep ).
prep( 'against', prep ).
prep( 'agin', prep ).
prep( 'along', prep ).
prep( 'alongside', prep ).
prep( 'amid', prep ).
prep( 'amidst', prep ).
prep( 'among', prep ).
prep( 'amongst', prep ).
prep( 'anent', prep ).
prep( 'around', prep ).
prep( 'aslant', prep ).
prep( 'astride', prep ).
prep( 'at', prep ).
prep( 'athwart', prep ).
prep( 'bar', prep ).
prep( 'barring', prep ).
prep( 'before', prep ).
prep( 'behind', prep ).
prep( 'below', prep ).
prep( 'beneath', prep ).
prep( 'beside', prep ).
prep( 'besides', prep ).
prep( 'between', prep ).
prep( 'betwixt', prep ).
prep( 'beyond', prep ).
prep( 'but', prep ).
prep( 'by', prep ).
prep( 'circa', prep ).
prep( 'concerning', prep ).
prep( 'considering', prep ).
prep( 'despite', prep ).
prep( 'down', prep ).
prep( 'during', prep ).
prep( 'ere', prep ).
prep( 'except', prep ).
prep( 'excepting', prep ).
prep( 'failing', prep ).
prep( 'for', prep ).
prep( 'from', prep ).
prep( 'in', prep ).
prep( 'inside', prep ).
prep( 'into', prep ).
prep( 'less', prep ).
prep( 'like', prep ).
prep( 'mid', prep ).
prep( 'midst', prep ).
prep( 'minus', prep ).
prep( 'near', prep ).
prep( 'next', prep ).
prep( 'nigh', prep ).
prep( 'nigher', prep ).
prep( 'nighest', prep ).
prep( 'notwithstanding', prep ).
prep( 'o\'er', prep ).
prep( 'of', prep ).
prep( 'off', prep ).
prep( 'on', prep ).
prep( 'on to', prep ).
prep( 'onto', prep ).
prep( 'outside', prep ).
prep( 'over', prep ).
prep( 'past', prep ).
prep( 'pending', prep ).
prep( 'per', prep ).
prep( 'plus', prep ).
prep( 'qua', prep ).
prep( 're', prep ).
prep( 'respecting', prep ).
prep( 'round', prep ).
prep( 'sans', prep ).
prep( 'save', prep ).
prep( 'saving', prep ).
prep( 'since', prep ).
prep( 'thro\'', prep ).
prep( 'through', prep ).
prep( 'throughout', prep ).
prep( 'thru', prep ).
prep( 'till', prep ).
prep( 'to', prep ).
prep( 'touching', prep ).
prep( 'toward', prep ).
prep( 'towards', prep ).
prep( 'under', prep ).
prep( 'underneath', prep ).
prep( 'unlike', prep ).
prep( 'until', prep ).
prep( 'unto', prep ).
prep( 'up', prep ).
prep( 'upon', prep ).
prep( 'versus', prep ).
prep( 'via', prep ).
prep( 'vice', prep ).
prep( 'vis-`a-vis', prep ).
prep( 'wanting', prep ).
prep( 'with', prep ).
prep( 'within', prep ).
prep( 'without', prep ).

conj( 'according as', conj ).
conj( 'after', conj ).
conj( 'albeit', conj ).
conj( 'although', conj ).
conj( 'an', conj ).
conj( 'and', conj ).
conj( 'as', conj ).
conj( 'because', conj ).
conj( 'before', conj ).
conj( 'but', conj ).
conj( 'cos', conj ).
conj( 'directly', conj ).
conj( 'either', conj ).
conj( 'except', conj ).
conj( 'for', conj ).
conj( 'forasmuch as', conj ).
conj( 'howbeit', conj ).
conj( 'if', conj ).
conj( 'immediately', conj ).
conj( 'instantly', conj ).
conj( 'lest', conj ).
conj( 'like', conj ).
conj( 'likewise', conj ).
conj( 'neither', conj ).
conj( 'nevertheless', conj ).
conj( 'nisi', conj ).
conj( 'nor', conj ).
conj( 'notwithstanding', conj ).
conj( 'now', conj ).
conj( 'only', conj ).
conj( 'or', conj ).
conj( 'otherwise', conj ).
conj( 'provided', conj ).
conj( 'providing', conj ).
conj( 'qua', conj ).
conj( 'since', conj ).
conj( 'so', conj ).
conj( 'supposing', conj ).
conj( 'than', conj ).
conj( 'that', conj ).
conj( 'tho\'', conj ).
conj( 'though', conj ).
conj( 'till', conj ).
conj( 'unless', conj ).
conj( 'until', conj ).
conj( 'when', conj ).
conj( 'whencesoever', conj ).
conj( 'whenever', conj ).
conj( 'whereas', conj ).
conj( 'whether', conj ).
conj( 'while', conj ).
conj( 'whilst', conj ).
conj( 'yet', conj ).

misc( 'afro-', prefix, '-' ).
misc( 'anglo-', prefix, '-' ).
misc( 'franco-', prefix, '-' ).
misc( 'indo-', prefix, '-' ).
misc( 'after-', prefix, '-' ).
misc( 'all', prefix, '-' ).
misc( 'ante', prefix, '-' ).
misc( 'anti-', prefix, '-' ).
misc( 'arch-', prefix, '-' ).
misc( 'audio-', prefix, '-' ).
misc( 'auto-', prefix, '-' ).
misc( 'be-', prefix, '-' ).
misc( 'bi-', prefix, '-' ).
misc( 'bond-', prefix, '-' ).
misc( 'centi-', prefix, '-' ).
misc( 'cine-', prefix, '-' ).
misc( 'co-', prefix, '-' ).
misc( 'contra-', prefix, '-' ).
misc( 'counter-', prefix, '-' ).
misc( 'crypto-', prefix, '-' ).
misc( 'deca-', prefix, '-' ).
misc( 'deci-', prefix, '-' ).
misc( 'electro-', prefix, '-' ).
misc( 'ex-', prefix, '-' ).
misc( 'geo-', prefix, '-' ).
misc( 'grand-', prefix, '-' ).
misc( 'hecto-', prefix, '-' ).
misc( 'infra', prefix, '-' ).
misc( 'kilo-', prefix, '-' ).
misc( 'maxi-', prefix, '-' ).
misc( 'milli-', prefix, '-' ).
misc( 'mini-', prefix, '-' ).
misc( 'multi-', prefix, '-' ).
misc( 'neo-', prefix, '-' ).
misc( 'non-', prefix, '-' ).
misc( 'nor\'-', prefix, '-' ).
misc( 'off', prefix, '-' ).
misc( 'over-', prefix, '-' ).
misc( 'palaeo-', prefix, '-' ).
misc( 'pan-', prefix, '-' ).
misc( 'petro-', prefix, '-' ).
misc( 'photo-', prefix, '-' ).
misc( 'post-', prefix, '-' ).
misc( 'pre-', prefix, '-' ).
misc( 'pro-', prefix, '-' ).
misc( 'pseudo-', prefix, '-' ).
misc( 'quasi-', prefix, '-' ).
misc( 'radio-', prefix, '-' ).
misc( 're-', prefix, '-' ).
misc( 'self-', prefix, '-' ).
misc( 'semi-', prefix, '-' ).
misc( 'servo-', prefix, '-' ).
misc( 'socio-', prefix, '-' ).
misc( 'step-', prefix, '-' ).
misc( 'sub-', prefix, '-' ).
misc( 'thermo-', prefix, '-' ).
misc( 'trans-', prefix, '-' ).
misc( 'tri-', prefix, '-' ).
misc( 'ultra-', prefix, '-' ).
misc( 'un-', prefix, '-' ).
misc( 'under-', prefix, '-' ).
misc( 'up-', prefix, '-' ).
misc( 'vice-', prefix, '-' ).
misc( 'wash-', prefix, '-' ).
misc( 'well-', prefix, '-' ).
misc( 'wich-', prefix, '-' ).
misc( 'witch-', prefix, '-' ).
misc( 'wych-', prefix, '-' ).
misc( 'yester-', prefix, '-' ).

misc( '\'shun', interj, '-' ).
misc( 'all right', interj, '-' ).
misc( 'alright', interj, '-' ).
misc( 'cheers', interj, '-' ).
misc( 'hey presto', interj, '-' ).
misc( 'o', interj, '-' ).
misc( 'ok', interj, '-' ).
misc( 'shucks', interj, '-' ).
misc( 'abracadabra', interj, '-' ).
misc( 'adieu', interj, close ).
misc( 'ah', interj, '-' ).
misc( 'aha', interj, '-' ).
misc( 'ahem', interj, hail ).
misc( 'ahoy', interj, greet ).
misc( 'alack', interj, '-' ).
misc( 'alas', interj, '-' ).
misc( 'alleluia', interj, '-' ).
misc( 'amen', interj, close ).
misc( 'atishoo', interj, '-' ).
misc( 'attaboy', interj, '-' ).
misc( 'au revoir', interj, close ).
misc( 'avast', interj, greet ).
misc( 'avaunt', interj, greet ).
misc( 'ay', interj, '-' ).
misc( 'aye', interj, '-' ).
misc( 'bah', interj, '-' ).
misc( 'ball', interj, '-' ).
misc( 'balls', interj, '-' ).
misc( 'bang', interj, '-' ).
misc( 'begad', interj, '-' ).
misc( 'begorra', interj, '-' ).
misc( 'blimey', interj, '-' ).
misc( 'bo', interj, '-' ).
misc( 'boh', interj, '-' ).
misc( 'bosh', interj, '-' ).
misc( 'botheration', interj, '-' ).
misc( 'bow-wow', interj, '-' ).
misc( 'bravo', interj, '-' ).
misc( 'bye-bye', interj, close ).
misc( 'cheerio', interj, close ).
misc( 'crikey', interj, '-' ).
misc( 'cripes', interj, '-' ).
misc( 'dear', interj, '-' ).
misc( 'eh', interj, '-' ).
misc( 'encore', interj, '-' ).
misc( 'eureka', interj, '-' ).
misc( 'farewell', interj, close ).
misc( 'faugh', interj, '-' ).
misc( 'fiddlesticks', interj, '-' ).
misc( 'fie', interj, '-' ).
misc( 'fore', interj, '-' ).
misc( 'fudge', interj, '-' ).
misc( 'gad', interj, '-' ).
misc( 'gangway', interj, '-' ).
misc( 'gee', interj, '-' ).
misc( 'gee whiz', interj, '-' ).
misc( 'gee-up', interj, '-' ).
misc( 'golly', interj, '-' ).
misc( 'goodbye', interj, close ).
misc( 'gosh', interj, '-' ).
misc( 'h\'m', interj, '-' ).
misc( 'ha', interj, '-' ).
misc( 'hallelujah', interj, '-' ).
misc( 'hallo', interj, greet ).
misc( 'halloo', interj, greet ).
misc( 'haw-haw', interj, '-' ).
misc( 'heighho', interj, '-' ).
misc( 'hello', interj, greet ).
misc( 'hem', interj, '-' ).
misc( 'hey', interj, hail ).
misc( 'hi', interj, greet ).
misc( 'hip', interj, '-' ).
misc( 'ho', interj, '-' ).
misc( 'hoity-toity', interj, '-' ).
misc( 'holloa', interj, hail ).
misc( 'hooray', interj, '-' ).
misc( 'hosanna', interj, '-' ).
misc( 'hullo', interj, greet ).
misc( 'humbug', interj, '-' ).
misc( 'humph', interj, '-' ).
misc( 'hurrah', interj, '-' ).
misc( 'hurray', interj, '-' ).
misc( 'jiminy', interj, '-' ).
misc( 'lo', interj, '-' ).
misc( 'lor', interj, '-' ).
misc( 'lumme', interj, '-' ).
misc( 'lummy', interj, '-' ).
misc( 'mum', interj, '-' ).
misc( 'no', interj, '-' ).
misc( 'nope', interj, '-' ).
misc( 'oh', interj, '-' ).
misc( 'oho', interj, '-' ).
misc( 'ouch', interj, '-' ).
misc( 'oyes', interj, '-' ).
misc( 'oyez', interj, '-' ).
misc( 'pah', interj, '-' ).
misc( 'phew', interj, '-' ).
misc( 'phooey', interj, '-' ).
misc( 'pooh', interj, '-' ).
misc( 'prithee', interj, '-' ).
misc( 'pshaw', interj, '-' ).
misc( 'rah', interj, '-' ).
misc( 'roger', interj, '-' ).
misc( 'scat', interj, '-' ).
misc( 'shalom', interj, greet ).
misc( 'shit', interj, '-' ).
misc( 'strewth', interj, '-' ).
misc( 'struth', interj, '-' ).
misc( 'ta', interj, '-' ).
misc( 'ta ta', interj, close ).
misc( 'tally-ho', interj, '-' ).
misc( 'there', interj, '-' ).
misc( 'tu quoque', interj, '-' ).
misc( 'tut', interj, '-' ).
misc( 'tut-tut', interj, '-' ).
misc( 'ugh', interj, '-' ).
misc( 'well', interj, '-' ).
misc( 'whew', interj, '-' ).
misc( 'whoa', interj, '-' ).
misc( 'why', interj, '-' ).
misc( 'wo', interj, '-' ).
misc( 'wotcher', interj, greet ).
misc( 'wow', interj, '-' ).
misc( 'yah', interj, '-' ).
misc( 'yea', interj, '-' ).
misc( 'yes', interj, '-' ).
misc( 'yippee', interj, '-' ).
misc( 'yo-heave-ho', interj, '-' ).
% manual additions
misc( 'bye', interj, close ).
misc( 'exit', interj, close ).
misc( 'quit', interj, close ).

misc( 'to', partcl, '-' ).

misc( 'anno domini', unknown, '-' ).
misc( 'celsius', unknown, '-' ).
misc( 'fahrenheit', unknown, '-' ).
misc( 'anon', unknown, '-' ).
misc( 'begone', unknown, '-' ).
misc( 'da capo', unknown, '-' ).
misc( 'et cetera', unknown, '-' ).
misc( 'id est', unknown, '-' ).
misc( 'in memoriam', unknown, '-' ).
misc( 'nota bene', unknown, '-' ).
misc( 'o\'clock', unknown, '-' ).
misc( 'stet', unknown, '-' ).
misc( 'vide', unknown, '-' ).
